/* Generated by Cython 0.29.24 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "name": "Aspidites._vendor.pyparsing_extension",
        "sources": [
            "Aspidites/_vendor/pyparsing_extension.py"
        ]
    },
    "module_name": "Aspidites._vendor.pyparsing_extension"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_24"
#define CYTHON_HEX_VERSION 0x001D18F0
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX >= 0x030800A4 && PY_VERSION_HEX < 0x030800B2
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if defined(PyUnicode_IS_READY)
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #else
  #define __Pyx_PyUnicode_READY(op)       (0)
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if defined(PyUnicode_IS_READY) && defined(PyUnicode_GET_SIZE)
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
  #endif
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__Aspidites___vendor__pyparsing_extension
#define __PYX_HAVE_API__Aspidites___vendor__pyparsing_extension
/* Early includes */
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "Aspidites/_vendor/pyparsing_extension.py",
};

/*--- Type declarations ---*/
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct__genexpr;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_7___iadd__;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__;
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr;
struct __pyx_defaults;
typedef struct __pyx_defaults __pyx_defaults;
struct __pyx_defaults1;
typedef struct __pyx_defaults1 __pyx_defaults1;
struct __pyx_defaults2;
typedef struct __pyx_defaults2 __pyx_defaults2;
struct __pyx_defaults3;
typedef struct __pyx_defaults3 __pyx_defaults3;
struct __pyx_defaults4;
typedef struct __pyx_defaults4 __pyx_defaults4;
struct __pyx_defaults5;
typedef struct __pyx_defaults5 __pyx_defaults5;
struct __pyx_defaults6;
typedef struct __pyx_defaults6 __pyx_defaults6;
struct __pyx_defaults7;
typedef struct __pyx_defaults7 __pyx_defaults7;
struct __pyx_defaults {
  PyObject *__pyx_arg_isinstance;
};
struct __pyx_defaults1 {
  PyObject *__pyx_arg_isinstance;
};
struct __pyx_defaults2 {
  PyObject *__pyx_arg_maxMatches;
};
struct __pyx_defaults3 {
  PyObject *__pyx_arg_maxMatches;
};
struct __pyx_defaults4 {
  PyObject *__pyx_arg_maxsplit;
};
struct __pyx_defaults5 {
  PyObject *__pyx_arg_wordChars;
};
struct __pyx_defaults6 {
  PyObject *__pyx_arg_wordChars;
};
struct __pyx_defaults7 {
  PyObject *__pyx_arg_default;
};

/* "Aspidites/_vendor/pyparsing_extension.py":13
 * import sre_constants
 * 
 * _generatorType = type((y for y in range(1)))             # <<<<<<<<<<<<<<
 * 
 * from collections.abc import Iterable
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct__genexpr {
  PyObject_HEAD
  long __pyx_v_y;
  long __pyx_t_0;
};


/* "Aspidites/_vendor/pyparsing_extension.py":43
 * hexnums = nums + "ABCDEFabcdef"
 * alphanums = alphas + nums
 * printables = "".join(c for c in string.printable if c not in string.whitespace)             # <<<<<<<<<<<<<<
 * 
 * __diag__ = SimpleNamespace()
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr {
  PyObject_HEAD
  PyObject *__pyx_v_c;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "Aspidites/_vendor/pyparsing_extension.py":119
 *     # ampersand must be replaced first
 *     from_symbols = '&><"\''
 *     to_symbols = ('&' + s + ';' for s in "amp gt lt quot apos".split())             # <<<<<<<<<<<<<<
 *     for from_, to_ in zip(from_symbols, to_symbols):
 *         data = data.replace(from_, to_)
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr {
  PyObject_HEAD
  PyObject *__pyx_v_s;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "Aspidites/_vendor/pyparsing_extension.py":336
 *             return iter(self.__tokdict)
 * 
 *     def _itervalues(self):             # <<<<<<<<<<<<<<
 *         return (self[k] for k in self._iterkeys())
 * 
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues {
  PyObject_HEAD
  PyObject *__pyx_v_self;
};


/* "Aspidites/_vendor/pyparsing_extension.py":337
 * 
 *     def _itervalues(self):
 *         return (self[k] for k in self._iterkeys())             # <<<<<<<<<<<<<<
 * 
 *     def _iteritems(self):
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues *__pyx_outer_scope;
  PyObject *__pyx_v_k;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "Aspidites/_vendor/pyparsing_extension.py":339
 *         return (self[k] for k in self._iterkeys())
 * 
 *     def _iteritems(self):             # <<<<<<<<<<<<<<
 *         return ((k, self[k]) for k in self._iterkeys())
 * 
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems {
  PyObject_HEAD
  PyObject *__pyx_v_self;
};


/* "Aspidites/_vendor/pyparsing_extension.py":340
 * 
 *     def _iteritems(self):
 *         return ((k, self[k]) for k in self._iterkeys())             # <<<<<<<<<<<<<<
 * 
 *     if PY_3:
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems *__pyx_outer_scope;
  PyObject *__pyx_v_k;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "Aspidites/_vendor/pyparsing_extension.py":530
 *         return ret
 * 
 *     def __iadd__(self, other):             # <<<<<<<<<<<<<<
 *         if other.__tokdict:
 *             offset = len(self.__toklist)
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_7___iadd__ {
  PyObject_HEAD
  Py_ssize_t __pyx_v_offset;
};


/* "Aspidites/_vendor/pyparsing_extension.py":557
 *         return "(%s, %s)" % (repr(self.__toklist), repr(self.__tokdict))
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return '[' + ', '.join(str(i) if isinstance(i, ParseResults) else repr(i) for i in self.__toklist) + ']'
 * 
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__ {
  PyObject_HEAD
  PyObject *__pyx_v_self;
};


/* "Aspidites/_vendor/pyparsing_extension.py":558
 * 
 *     def __str__(self):
 *         return '[' + ', '.join(str(i) if isinstance(i, ParseResults) else repr(i) for i in self.__toklist) + ']'             # <<<<<<<<<<<<<<
 * 
 *     def _asStringList(self, sep=''):
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__ *__pyx_outer_scope;
  PyObject *__pyx_v_i;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "Aspidites/_vendor/pyparsing_extension.py":588
 *         return [res.asList() if isinstance(res, ParseResults) else res for res in self.__toklist]
 * 
 *     def asDict(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns the named parse results as a nested dictionary.
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict {
  PyObject_HEAD
  PyObject *__pyx_v_item_fn;
  PyObject *__pyx_v_toItem;
};


/* "Aspidites/_vendor/pyparsing_extension.py":622
 *                 return obj
 * 
 *         return dict((k, toItem(v)) for k, v in item_fn())             # <<<<<<<<<<<<<<
 * 
 *     def copy(self):
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict *__pyx_outer_scope;
  PyObject *__pyx_v_k;
  PyObject *__pyx_v_v;
};


/* "Aspidites/_vendor/pyparsing_extension.py":635
 *         return ret
 * 
 *     def asXML(self, doctag=None, namedItemsOnly=False, indent="", formatted=True):             # <<<<<<<<<<<<<<
 *         """
 *         (Deprecated) Returns the parse results as XML. Tags are created for tokens and lists that have defined results names.
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML {
  PyObject_HEAD
  PyObject *__pyx_v_self;
};


/* "Aspidites/_vendor/pyparsing_extension.py":641
 *         nl = "\n"
 *         out = []
 *         namedItems = dict((v[1], k) for (k, vlist) in self.__tokdict.items()             # <<<<<<<<<<<<<<
 *                           for v in vlist)
 *         nextLevelIndent = indent + "  "
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML *__pyx_outer_scope;
  PyObject *__pyx_v_k;
  PyObject *__pyx_v_v;
  PyObject *__pyx_v_vlist;
};


/* "Aspidites/_vendor/pyparsing_extension.py":743
 *             return None
 * 
 *     def dump(self, indent='', full=True, include_list=True, _depth=0):             # <<<<<<<<<<<<<<
 *         """
 *         Diagnostic method for listing out the contents of
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump {
  PyObject_HEAD
  PyObject *__pyx_v_self;
};


/* "Aspidites/_vendor/pyparsing_extension.py":773
 *         if full:
 *             if self.haskeys():
 *                 items = sorted((str(k), v) for k, v in self.items())             # <<<<<<<<<<<<<<
 *                 for k, v in items:
 *                     if out:
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump *__pyx_outer_scope;
  PyObject *__pyx_v_k;
  PyObject *__pyx_v_v;
};


/* "Aspidites/_vendor/pyparsing_extension.py":785
 *                     else:
 *                         out.append(repr(v))
 *             elif any(isinstance(vv, ParseResults) for vv in self):             # <<<<<<<<<<<<<<
 *                 v = self
 *                 for i, vv in enumerate(v):
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump *__pyx_outer_scope;
  PyObject *__pyx_v_vv;
};


/* "Aspidites/_vendor/pyparsing_extension.py":1238
 *         return newself
 * 
 *     def setBreak(self, breakFlag=True):             # <<<<<<<<<<<<<<
 *         """Method to invoke the Python pdb debugger when this element is
 *            about to be parsed. Set ``breakFlag`` to True to enable, False to
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak {
  PyObject_HEAD
  PyObject *__pyx_v__parseMethod;
};


/* "Aspidites/_vendor/pyparsing_extension.py":1259
 *         return self
 * 
 *     def setParseAction(self, *fns, **kwargs):             # <<<<<<<<<<<<<<
 *         from .pyparsing import _trim_arity
 *         """
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction {
  PyObject_HEAD
  PyObject *__pyx_v_fns;
};


/* "Aspidites/_vendor/pyparsing_extension.py":1303
 *             self.parseAction = []
 *         else:
 *             if not all(callable(fn) for fn in fns):             # <<<<<<<<<<<<<<
 *                 raise TypeError("parse actions must be callable")
 *             self.parseAction = list(map(_trim_arity, list(fns)))
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction *__pyx_outer_scope;
  PyObject *__pyx_v_fn;
};


/* "Aspidites/_vendor/pyparsing_extension.py":1494
 * 
 *     class _UnboundedCache(object):
 *         def __init__(self):             # <<<<<<<<<<<<<<
 *             cache = {}
 *             self.not_in_cache = not_in_cache = object()
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__ {
  PyObject_HEAD
  PyObject *__pyx_v_cache;
  PyObject *__pyx_v_not_in_cache;
};


/* "Aspidites/_vendor/pyparsing_extension.py":1517
 *     if _OrderedDict is not None:
 *         class _FifoCache(object):
 *             def __init__(self, size):             # <<<<<<<<<<<<<<
 *                 self.not_in_cache = not_in_cache = object()
 * 
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__ {
  PyObject_HEAD
  PyObject *__pyx_v_cache;
  PyObject *__pyx_v_not_in_cache;
  PyObject *__pyx_v_size;
};


/* "Aspidites/_vendor/pyparsing_extension.py":1546
 *     else:
 *         class _FifoCache(object):
 *             def __init__(self, size):             # <<<<<<<<<<<<<<
 *                 self.not_in_cache = not_in_cache = object()
 * 
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__ {
  PyObject_HEAD
  PyObject *__pyx_v_cache;
  PyObject *__pyx_v_key_fifo;
  PyObject *__pyx_v_not_in_cache;
  PyObject *__pyx_v_size;
};


/* "Aspidites/_vendor/pyparsing_extension.py":1706
 *             return tokens
 * 
 *     def scanString(self, instring, maxMatches=sys.maxsize, overlap=False):             # <<<<<<<<<<<<<<
 *         """
 *         Scan the input string for expression matches.  Each match will return the
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString {
  PyObject_HEAD
  PyObject *__pyx_v_e;
  PyObject *__pyx_v_exc;
  PyObject *__pyx_v_instring;
  Py_ssize_t __pyx_v_instrlen;
  PyObject *__pyx_v_loc;
  PyObject *__pyx_v_matches;
  PyObject *__pyx_v_maxMatches;
  PyObject *__pyx_v_nextLoc;
  PyObject *__pyx_v_nextloc;
  PyObject *__pyx_v_overlap;
  PyObject *__pyx_v_parseFn;
  PyObject *__pyx_v_preloc;
  PyObject *__pyx_v_preparseFn;
  PyObject *__pyx_v_self;
  PyObject *__pyx_v_tokens;
  PyObject *__pyx_t_0;
  PyObject *__pyx_t_1;
  PyObject *__pyx_t_2;
  PyObject *__pyx_t_3;
  PyObject *__pyx_t_4;
  PyObject *__pyx_t_5;
};


/* "Aspidites/_vendor/pyparsing_extension.py":1860
 *                 raise exc
 * 
 *     def split(self, instring, maxsplit=sys.maxsize, includeSeparators=False):             # <<<<<<<<<<<<<<
 *         """
 *         Generator method to split a string using the given expression as a separator.
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split {
  PyObject_HEAD
  PyObject *__pyx_v_e;
  PyObject *__pyx_v_includeSeparators;
  PyObject *__pyx_v_instring;
  PyObject *__pyx_v_last;
  PyObject *__pyx_v_maxsplit;
  PyObject *__pyx_v_s;
  PyObject *__pyx_v_self;
  long __pyx_v_splits;
  PyObject *__pyx_v_t;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "Aspidites/_vendor/pyparsing_extension.py":1965
 *         return other - self
 * 
 *     def __mul__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of * operator, allows use of ``expr * 3`` in place of
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__ {
  PyObject_HEAD
  PyObject *__pyx_v_makeOptionalList;
  PyObject *__pyx_v_other;
  PyObject *__pyx_v_self;
};


/* "Aspidites/_vendor/pyparsing_extension.py":1994
 *             minElements, optElements = other, 0
 *         elif isinstance(other, tuple):
 *             other = tuple(o if o is not Ellipsis else None for o in other)             # <<<<<<<<<<<<<<
 *             other = (other + (None, None))[:2]
 *             if other[0] is None:
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__ *__pyx_outer_scope;
  PyObject *__pyx_v_o;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "Aspidites/_vendor/pyparsing_extension.py":2557
 *     """
 * 
 *     def __init__(self, exprs, savelist=False):             # <<<<<<<<<<<<<<
 *         super(ParseExpression, self).__init__(savelist)
 *         if isinstance(exprs, _generatorType):
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__ {
  PyObject_HEAD
  PyObject *__pyx_v_exprs;
  PyObject *__pyx_v_self;
};


/* "Aspidites/_vendor/pyparsing_extension.py":2569
 *             exprs = list(exprs)
 *             # if sequence of strings provided, wrap with Literal
 *             if any(isinstance(expr, str) for expr in exprs):             # <<<<<<<<<<<<<<
 *                 exprs = (self._literalStringClass(e) if isinstance(e, str) else e for e in exprs)
 *             self.exprs = list(exprs)
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__ *__pyx_outer_scope;
  PyObject *__pyx_v_expr;
};


/* "Aspidites/_vendor/pyparsing_extension.py":2570
 *             # if sequence of strings provided, wrap with Literal
 *             if any(isinstance(expr, str) for expr in exprs):
 *                 exprs = (self._literalStringClass(e) if isinstance(e, str) else e for e in exprs)             # <<<<<<<<<<<<<<
 *             self.exprs = list(exprs)
 *         else:
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__ *__pyx_outer_scope;
  PyObject *__pyx_v_e;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "Aspidites/_vendor/pyparsing_extension.py":3006
 *     """
 * 
 *     def __init__(self, initChars, bodyChars=None, min=1, max=0, exact=0, asKeyword=False, excludeChars=None):             # <<<<<<<<<<<<<<
 *         super(Word, self).__init__()
 *         if excludeChars:
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__ {
  PyObject_HEAD
  PyObject *__pyx_v_bodyChars;
  PyObject *__pyx_v_excludeChars;
  PyObject *__pyx_v_initChars;
};


/* "Aspidites/_vendor/pyparsing_extension.py":3010
 *         if excludeChars:
 *             excludeChars = set(excludeChars)
 *             initChars = ''.join(c for c in initChars if c not in excludeChars)             # <<<<<<<<<<<<<<
 *             if bodyChars:
 *                 bodyChars = ''.join(c for c in bodyChars if c not in excludeChars)
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__ *__pyx_outer_scope;
  PyObject *__pyx_v_c;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "Aspidites/_vendor/pyparsing_extension.py":3012
 *             initChars = ''.join(c for c in initChars if c not in excludeChars)
 *             if bodyChars:
 *                 bodyChars = ''.join(c for c in bodyChars if c not in excludeChars)             # <<<<<<<<<<<<<<
 *         self.initCharsOrig = initChars
 *         self.initChars = set(initChars)
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__ *__pyx_outer_scope;
  PyObject *__pyx_v_c;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "Aspidites/_vendor/pyparsing_extension.py":3251
 *         return self.strRepr
 * 
 *     def sub(self, repl):             # <<<<<<<<<<<<<<
 *         r"""
 *         Return Regex with an attached parse action to transform the parsed
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub {
  PyObject_HEAD
  PyObject *__pyx_v_repl;
  PyObject *__pyx_v_self;
};


/* "Aspidites/_vendor/pyparsing_extension.py":3321
 *     """
 * 
 *     def __init__(self, quoteChar, escChar=None, escQuote=None, multiline=False,             # <<<<<<<<<<<<<<
 *                  unquoteResults=True, endQuoteChar=None, convertWhitespaceEscapes=True):
 *         super(QuotedString, self).__init__()
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__ {
  PyObject_HEAD
  PyObject *__pyx_v_self;
};


/* "Aspidites/_vendor/pyparsing_extension.py":3361
 *         if len(self.endQuoteChar) > 1:
 *             self.pattern += (
 *                     '|(?:' + ')|(?:'.join("%s[^%s]" % (re.escape(self.endQuoteChar[:i]),             # <<<<<<<<<<<<<<
 *                                                        _escapeRegexRangeChars(self.endQuoteChar[i]))
 *                                           for i in range(len(self.endQuoteChar) - 1, 0, -1)) + ')')
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__ *__pyx_outer_scope;
  Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_t_0;
  Py_ssize_t __pyx_t_1;
};


/* "Aspidites/_vendor/pyparsing_extension.py":3545
 *     }
 * 
 *     def __init__(self, ws=" \t\r\n", min=1, max=0, exact=0):             # <<<<<<<<<<<<<<
 *         super(White, self).__init__()
 *         self.matchWhite = ws
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__ {
  PyObject_HEAD
  PyObject *__pyx_v_self;
};


/* "Aspidites/_vendor/pyparsing_extension.py":3548
 *         super(White, self).__init__()
 *         self.matchWhite = ws
 *         self.setWhitespaceChars("".join(c for c in self.whiteChars if c not in self.matchWhite))             # <<<<<<<<<<<<<<
 *         # ~ self.leaveWhitespace()
 *         self.name = ("".join(White.whiteStrs[c] for c in self.matchWhite))
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__ *__pyx_outer_scope;
  PyObject *__pyx_v_c;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "Aspidites/_vendor/pyparsing_extension.py":3550
 *         self.setWhitespaceChars("".join(c for c in self.whiteChars if c not in self.matchWhite))
 *         # ~ self.leaveWhitespace()
 *         self.name = ("".join(White.whiteStrs[c] for c in self.matchWhite))             # <<<<<<<<<<<<<<
 *         self.mayReturnEmpty = True
 *         self.errmsg = "Expected " + self.name
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__ *__pyx_outer_scope;
  PyObject *__pyx_v_c;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "Aspidites/_vendor/pyparsing_extension.py":3778
 *             self.leaveWhitespace()
 * 
 *     def __init__(self, exprs, savelist=True):             # <<<<<<<<<<<<<<
 *         exprs = list(exprs)
 *         if exprs and Ellipsis in exprs:
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__ {
  PyObject_HEAD
  PyObject *__pyx_v_self;
};


/* "Aspidites/_vendor/pyparsing_extension.py":3793
 *             exprs[:] = tmp
 *         super(And, self).__init__(exprs, savelist)
 *         self.mayReturnEmpty = all(e.mayReturnEmpty for e in self.exprs)             # <<<<<<<<<<<<<<
 *         self.setWhitespaceChars(self.exprs[0].whiteChars)
 *         self.skipWhitespace = self.exprs[0].skipWhitespace
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__ *__pyx_outer_scope;
  PyObject *__pyx_v_e;
};


/* "Aspidites/_vendor/pyparsing_extension.py":3798
 *         self.callPreparse = True
 * 
 *     def streamline(self):             # <<<<<<<<<<<<<<
 *         # collapse any _PendingSkip's
 *         from .pyparsing import _PendingSkip
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline {
  PyObject_HEAD
  PyObject *__pyx_v__PendingSkip;
  PyObject *__pyx_v_self;
};


/* "Aspidites/_vendor/pyparsing_extension.py":3802
 *         from .pyparsing import _PendingSkip
 *         if self.exprs:
 *             if any(isinstance(e, ParseExpression) and e.exprs and isinstance(e.exprs[-1], _PendingSkip)             # <<<<<<<<<<<<<<
 *                    for e in self.exprs[:-1]):
 *                 for i, e in enumerate(self.exprs[:-1]):
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline *__pyx_outer_scope;
  PyObject *__pyx_v_e;
};


/* "Aspidites/_vendor/pyparsing_extension.py":3814
 * 
 *         super(And, self).streamline()
 *         self.mayReturnEmpty = all(e.mayReturnEmpty for e in self.exprs)             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline *__pyx_outer_scope;
  PyObject *__pyx_v_e;
};


/* "Aspidites/_vendor/pyparsing_extension.py":3854
 *                 break
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__ {
  PyObject_HEAD
  PyObject *__pyx_v_self;
};


/* "Aspidites/_vendor/pyparsing_extension.py":3859
 * 
 *         if self.strRepr is None:
 *             self.strRepr = "{" + " ".join(str(e) for e in self.exprs) + "}"             # <<<<<<<<<<<<<<
 * 
 *         return self.strRepr
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__ *__pyx_outer_scope;
  PyObject *__pyx_v_e;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "Aspidites/_vendor/pyparsing_extension.py":3882
 *     """
 * 
 *     def __init__(self, exprs, savelist=False):             # <<<<<<<<<<<<<<
 *         super(Or, self).__init__(exprs, savelist)
 *         if self.exprs:
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__ {
  PyObject_HEAD
  PyObject *__pyx_v_self;
};


/* "Aspidites/_vendor/pyparsing_extension.py":3885
 *         super(Or, self).__init__(exprs, savelist)
 *         if self.exprs:
 *             self.mayReturnEmpty = any(e.mayReturnEmpty for e in self.exprs)             # <<<<<<<<<<<<<<
 *         else:
 *             self.mayReturnEmpty = True
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__ *__pyx_outer_scope;
  PyObject *__pyx_v_e;
};


/* "Aspidites/_vendor/pyparsing_extension.py":3889
 *             self.mayReturnEmpty = True
 * 
 *     def streamline(self):             # <<<<<<<<<<<<<<
 *         super(Or, self).streamline()
 *         if __diag__.collect_all_And_tokens:
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline {
  PyObject_HEAD
  PyObject *__pyx_v_self;
};


/* "Aspidites/_vendor/pyparsing_extension.py":3892
 *         super(Or, self).streamline()
 *         if __diag__.collect_all_And_tokens:
 *             self.saveAsList = any(e.saveAsList for e in self.exprs)             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline *__pyx_outer_scope;
  PyObject *__pyx_v_e;
};


/* "Aspidites/_vendor/pyparsing_extension.py":3960
 *         return self.append(other)  # Or([self, other])
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__ {
  PyObject_HEAD
  PyObject *__pyx_v_self;
};


/* "Aspidites/_vendor/pyparsing_extension.py":3965
 * 
 *         if self.strRepr is None:
 *             self.strRepr = "{" + " ^ ".join(str(e) for e in self.exprs) + "}"             # <<<<<<<<<<<<<<
 * 
 *         return self.strRepr
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__ *__pyx_outer_scope;
  PyObject *__pyx_v_e;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "Aspidites/_vendor/pyparsing_extension.py":3974
 *             e.checkRecursion(subRecCheckList)
 * 
 *     def _setResultsName(self, name, listAllMatches=False):             # <<<<<<<<<<<<<<
 *         if (not __diag__.collect_all_And_tokens
 *                 and __diag__.warn_multiple_tokens_in_named_alternation):
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName {
  PyObject_HEAD
  PyObject *__pyx_v_self;
};


/* "Aspidites/_vendor/pyparsing_extension.py":3977
 *         if (not __diag__.collect_all_And_tokens
 *                 and __diag__.warn_multiple_tokens_in_named_alternation):
 *             if any(isinstance(e, And) for e in self.exprs):             # <<<<<<<<<<<<<<
 *                 warnings.warn("{0}: setting results name {1!r} on {2} expression "
 *                               "may only return a single token for an And alternative, "
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName *__pyx_outer_scope;
  PyObject *__pyx_v_e;
};


/* "Aspidites/_vendor/pyparsing_extension.py":4005
 *     """
 * 
 *     def __init__(self, exprs, savelist=False):             # <<<<<<<<<<<<<<
 *         super(MatchFirst, self).__init__(exprs, savelist)
 *         if self.exprs:
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__ {
  PyObject_HEAD
  PyObject *__pyx_v_self;
};


/* "Aspidites/_vendor/pyparsing_extension.py":4008
 *         super(MatchFirst, self).__init__(exprs, savelist)
 *         if self.exprs:
 *             self.mayReturnEmpty = any(e.mayReturnEmpty for e in self.exprs)             # <<<<<<<<<<<<<<
 *         else:
 *             self.mayReturnEmpty = True
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__ *__pyx_outer_scope;
  PyObject *__pyx_v_e;
};


/* "Aspidites/_vendor/pyparsing_extension.py":4012
 *             self.mayReturnEmpty = True
 * 
 *     def streamline(self):             # <<<<<<<<<<<<<<
 *         super(MatchFirst, self).streamline()
 *         if __diag__.collect_all_And_tokens:
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline {
  PyObject_HEAD
  PyObject *__pyx_v_self;
};


/* "Aspidites/_vendor/pyparsing_extension.py":4015
 *         super(MatchFirst, self).streamline()
 *         if __diag__.collect_all_And_tokens:
 *             self.saveAsList = any(e.saveAsList for e in self.exprs)             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline *__pyx_outer_scope;
  PyObject *__pyx_v_e;
};


/* "Aspidites/_vendor/pyparsing_extension.py":4047
 *         return self.append(other)  # MatchFirst([self, other])
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__ {
  PyObject_HEAD
  PyObject *__pyx_v_self;
};


/* "Aspidites/_vendor/pyparsing_extension.py":4052
 * 
 *         if self.strRepr is None:
 *             self.strRepr = "{" + " | ".join(str(e) for e in self.exprs) + "}"             # <<<<<<<<<<<<<<
 * 
 *         return self.strRepr
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__ *__pyx_outer_scope;
  PyObject *__pyx_v_e;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "Aspidites/_vendor/pyparsing_extension.py":4061
 *             e.checkRecursion(subRecCheckList)
 * 
 *     def _setResultsName(self, name, listAllMatches=False):             # <<<<<<<<<<<<<<
 *         if (not __diag__.collect_all_And_tokens
 *                 and __diag__.warn_multiple_tokens_in_named_alternation):
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName {
  PyObject_HEAD
  PyObject *__pyx_v_self;
};


/* "Aspidites/_vendor/pyparsing_extension.py":4064
 *         if (not __diag__.collect_all_And_tokens
 *                 and __diag__.warn_multiple_tokens_in_named_alternation):
 *             if any(isinstance(e, And) for e in self.exprs):             # <<<<<<<<<<<<<<
 *                 warnings.warn("{0}: setting results name {1!r} on {2} expression "
 *                               "may only return a single token for an And alternative, "
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName *__pyx_outer_scope;
  PyObject *__pyx_v_e;
};


/* "Aspidites/_vendor/pyparsing_extension.py":4132
 *     """
 * 
 *     def __init__(self, exprs, savelist=True):             # <<<<<<<<<<<<<<
 *         super(Each, self).__init__(exprs, savelist)
 *         self.mayReturnEmpty = all(e.mayReturnEmpty for e in self.exprs)
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__ {
  PyObject_HEAD
  PyObject *__pyx_v_self;
};


/* "Aspidites/_vendor/pyparsing_extension.py":4134
 *     def __init__(self, exprs, savelist=True):
 *         super(Each, self).__init__(exprs, savelist)
 *         self.mayReturnEmpty = all(e.mayReturnEmpty for e in self.exprs)             # <<<<<<<<<<<<<<
 *         self.skipWhitespace = True
 *         self.initExprGroups = True
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__ *__pyx_outer_scope;
  PyObject *__pyx_v_e;
};


/* "Aspidites/_vendor/pyparsing_extension.py":4139
 *         self.saveAsList = True
 * 
 *     def streamline(self):             # <<<<<<<<<<<<<<
 *         super(Each, self).streamline()
 *         self.mayReturnEmpty = all(e.mayReturnEmpty for e in self.exprs)
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline {
  PyObject_HEAD
  PyObject *__pyx_v_self;
};


/* "Aspidites/_vendor/pyparsing_extension.py":4141
 *     def streamline(self):
 *         super(Each, self).streamline()
 *         self.mayReturnEmpty = all(e.mayReturnEmpty for e in self.exprs)             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline *__pyx_outer_scope;
  PyObject *__pyx_v_e;
};


/* "Aspidites/_vendor/pyparsing_extension.py":4144
 *         return self
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if self.initExprGroups:
 *             self.opt1map = dict((id(e.expr), e) for e in self.exprs if isinstance(e, Optional))
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl {
  PyObject_HEAD
  PyObject *__pyx_v_self;
  PyObject *__pyx_v_tmpReqd;
};


/* "Aspidites/_vendor/pyparsing_extension.py":4146
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if self.initExprGroups:
 *             self.opt1map = dict((id(e.expr), e) for e in self.exprs if isinstance(e, Optional))             # <<<<<<<<<<<<<<
 *             opt1 = [e.expr for e in self.exprs if isinstance(e, Optional)]
 *             opt2 = [e for e in self.exprs if e.mayReturnEmpty and not isinstance(e, (Optional, Regex))]
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl *__pyx_outer_scope;
  PyObject *__pyx_v_e;
};


/* "Aspidites/_vendor/pyparsing_extension.py":4179
 * 
 *         if tmpReqd:
 *             missing = ", ".join(str(e) for e in tmpReqd)             # <<<<<<<<<<<<<<
 *             raise ParseException(instring, loc, "Missing one or more required elements (%s)" % missing)
 * 
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl *__pyx_outer_scope;
  PyObject *__pyx_v_e;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "Aspidites/_vendor/pyparsing_extension.py":4193
 *         return loc, finalResults
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__ {
  PyObject_HEAD
  PyObject *__pyx_v_self;
};


/* "Aspidites/_vendor/pyparsing_extension.py":4198
 * 
 *         if self.strRepr is None:
 *             self.strRepr = "{" + " & ".join(str(e) for e in self.exprs) + "}"             # <<<<<<<<<<<<<<
 * 
 *         return self.strRepr
 */
struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__ *__pyx_outer_scope;
  PyObject *__pyx_v_e;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* KeywordStringCheck.proto */
static int __Pyx_CheckKeywordStrings(PyObject *kwdict, const char* function_name, int kw_allowed);

/* ListExtend.proto */
static CYTHON_INLINE int __Pyx_PyList_Extend(PyObject* L, PyObject* v) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject* none = _PyList_Extend((PyListObject*)L, v);
    if (unlikely(!none))
        return -1;
    Py_DECREF(none);
    return 0;
#else
    return PyList_SetSlice(L, PY_SSIZE_T_MAX, PY_SSIZE_T_MAX, v);
#endif
}

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_SubtractObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceSubtract(op1, op2) : PyNumber_Subtract(op1, op2))
#endif

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* StrEquals.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyString_Equals __Pyx_PyUnicode_Equals
#else
#define __Pyx_PyString_Equals __Pyx_PyBytes_Equals
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname);

/* py_dict_items.proto */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d);

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod0.proto */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CallUnboundCMethod0(cfunc, self)\
    (likely((cfunc)->func) ?\
        (likely((cfunc)->flag == METH_NOARGS) ?  (*((cfunc)->func))(self, NULL) :\
         (PY_VERSION_HEX >= 0x030600B1 && likely((cfunc)->flag == METH_FASTCALL) ?\
            (PY_VERSION_HEX >= 0x030700A0 ?\
                (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0) :\
                (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL)) :\
          (PY_VERSION_HEX >= 0x030700A0 && (cfunc)->flag == (METH_FASTCALL | METH_KEYWORDS) ?\
            (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL) :\
            (likely((cfunc)->flag == (METH_VARARGS | METH_KEYWORDS)) ?  ((*(PyCFunctionWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, __pyx_empty_tuple, NULL)) :\
               ((cfunc)->flag == METH_VARARGS ?  (*((cfunc)->func))(self, __pyx_empty_tuple) :\
               __Pyx__CallUnboundCMethod0(cfunc, self)))))) :\
        __Pyx__CallUnboundCMethod0(cfunc, self))
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* append.proto */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x);

/* SliceObject.proto */
#define __Pyx_PyObject_DelSlice(obj, cstart, cstop, py_start, py_stop, py_slice, has_cstart, has_cstop, wraparound)\
    __Pyx_PyObject_SetSlice(obj, (PyObject*)NULL, cstart, cstop, py_start, py_stop, py_slice, has_cstart, has_cstop, wraparound)
static CYTHON_INLINE int __Pyx_PyObject_SetSlice(
        PyObject* obj, PyObject* value, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED 1
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    size_t defaults_size;  // used by FusedFunction for copying defaults
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_Check(obj)  (__Pyx_TypeCheck(obj, __pyx_CyFunctionType))
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* PyIntCompare.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* StringJoin.proto */
#if PY_MAJOR_VERSION < 3
#define __Pyx_PyString_Join __Pyx_PyBytes_Join
#define __Pyx_PyBaseString_Join(s, v) (PyUnicode_CheckExact(s) ? PyUnicode_Join(s, v) : __Pyx_PyBytes_Join(s, v))
#else
#define __Pyx_PyString_Join PyUnicode_Join
#define __Pyx_PyBaseString_Join PyUnicode_Join
#endif
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION < 3
    #define __Pyx_PyBytes_Join _PyString_Join
    #else
    #define __Pyx_PyBytes_Join _PyBytes_Join
    #endif
#else
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values);
#endif

/* PyDictContains.proto */
static CYTHON_INLINE int __Pyx_PyDict_ContainsTF(PyObject* item, PyObject* dict, int eq) {
    int result = PyDict_Contains(dict, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* IterNext.proto */
#define __Pyx_PyIter_Next(obj) __Pyx_PyIter_Next2(obj, NULL)
static CYTHON_INLINE PyObject *__Pyx_PyIter_Next2(PyObject *, PyObject *);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchInspect.proto */
static PyObject* __Pyx_patch_inspect(PyObject* module);

/* pyfrozenset_new.proto */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it);

/* PySetContains.proto */
static CYTHON_INLINE int __Pyx_PySet_ContainsTF(PyObject* key, PyObject* set, int eq);

/* CallableCheck.proto */
#if CYTHON_USE_TYPE_SLOTS && PY_MAJOR_VERSION >= 3
#define __Pyx_PyCallable_Check(obj)   (Py_TYPE(obj)->tp_call != NULL)
#else
#define __Pyx_PyCallable_Check(obj)   PyCallable_Check(obj)
#endif

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* dict_getitem_default.proto */
static PyObject* __Pyx_PyDict_GetItemDefault(PyObject* d, PyObject* key, PyObject* default_value);

/* CallUnboundCMethod1.proto */
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#else
#define __Pyx_CallUnboundCMethod1(cfunc, self, arg)  __Pyx__CallUnboundCMethod1(cfunc, self, arg)
#endif

/* CallUnboundCMethod2.proto */
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2);
#else
#define __Pyx_CallUnboundCMethod2(cfunc, self, arg1, arg2)  __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2)
#endif

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* py_dict_clear.proto */
#define __Pyx_PyDict_Clear(d) (PyDict_Clear(d), 0)

/* py_dict_pop.proto */
static CYTHON_INLINE PyObject *__Pyx_PyDict_Pop(PyObject *d, PyObject *key, PyObject *default_value);

/* PyObjectLookupSpecial.proto */
#if CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name) {
    PyObject *res;
    PyTypeObject *tp = Py_TYPE(obj);
#if PY_MAJOR_VERSION < 3
    if (unlikely(PyInstance_Check(obj)))
        return __Pyx_PyObject_GetAttrStr(obj, attr_name);
#endif
    res = _PyType_Lookup(tp, attr_name);
    if (likely(res)) {
        descrgetfunc f = Py_TYPE(res)->tp_descr_get;
        if (!f) {
            Py_INCREF(res);
        } else {
            res = f(res, obj, (PyObject *)tp);
        }
    } else {
        PyErr_SetObject(PyExc_AttributeError, attr_name);
    }
    return res;
}
#else
#define __Pyx_PyObject_LookupSpecial(o,n) __Pyx_PyObject_GetAttrStr(o,n)
#endif

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* PyIntCompare.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_NeObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* DelItemInt.proto */
#define __Pyx_DelItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_DelItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_DelItem_Generic(o, to_py_func(i))))
static int __Pyx_DelItem_Generic(PyObject *o, PyObject *j);
static CYTHON_INLINE int __Pyx_DelItemInt_Fast(PyObject *o, Py_ssize_t i,
                                               int is_list, int wraparound);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* SliceTupleAndList.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_GetSlice(PyObject* src, Py_ssize_t start, Py_ssize_t stop);
static CYTHON_INLINE PyObject* __Pyx_PyTuple_GetSlice(PyObject* src, Py_ssize_t start, Py_ssize_t stop);
#else
#define __Pyx_PyList_GetSlice(seq, start, stop)   PySequence_GetSlice(seq, start, stop)
#define __Pyx_PyTuple_GetSlice(seq, start, stop)  PySequence_GetSlice(seq, start, stop)
#endif

/* CalculateMetaclass.proto */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

/* Py3ClassCreate.proto */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

/* SetNameInClass.proto */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? _PyDict_SetItem_KnownHash(ns, name, value, ((PyASCIIObject *) name)->hash) : PyObject_SetItem(ns, name, value))
#elif CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? PyDict_SetItem(ns, name, value) : PyObject_SetItem(ns, name, value))
#else
#define __Pyx_SetNameInClass(ns, name, value)  PyObject_SetItem(ns, name, value)
#endif

/* ClassMethod.proto */
#include "descrobject.h"
static CYTHON_UNUSED PyObject* __Pyx_Method_ClassMethod(PyObject *method);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* Print.proto */
static int __Pyx_Print(PyObject*, PyObject *, int);
#if CYTHON_COMPILING_IN_PYPY || PY_MAJOR_VERSION >= 3
static PyObject* __pyx_print = 0;
static PyObject* __pyx_print_kwargs = 0;
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* PrintOne.proto */
static int __Pyx_PrintOne(PyObject* stream, PyObject *o);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CoroutineBase.proto */
typedef PyObject *(*__pyx_coroutine_body_t)(PyObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    PyObject *gi_frame;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
static PyTypeObject *__pyx_GeneratorType = 0;
#define __Pyx_Generator_CheckExact(obj) (Py_TYPE(obj) == __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'Aspidites._vendor.pyparsing_extension' */
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct__genexpr = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_7___iadd__ = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__ = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__ = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__ = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__ = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__ = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__ = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__ = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__ = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__ = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__ = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__ = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__ = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__ = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__ = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__ = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__ = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__ = 0;
static PyTypeObject *__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr = 0;
#define __Pyx_MODULE_NAME "Aspidites._vendor.pyparsing_extension"
extern int __pyx_module_is_main_Aspidites___vendor__pyparsing_extension;
int __pyx_module_is_main_Aspidites___vendor__pyparsing_extension = 0;

/* Implementation of 'Aspidites._vendor.pyparsing_extension' */
static PyObject *__pyx_builtin_ImportError;
static PyObject *__pyx_builtin_vars;
static PyObject *__pyx_builtin_object;
static PyObject *__pyx_builtin_staticmethod;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_chr;
static PyObject *__pyx_builtin_zip;
static PyObject *__pyx_builtin_KeyError;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_IndexError;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_AttributeError;
static PyObject *__pyx_builtin_map;
static PyObject *__pyx_builtin_Ellipsis;
static PyObject *__pyx_builtin_SyntaxWarning;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_open;
static PyObject *__pyx_builtin_id;
static PyObject *__pyx_builtin_super;
static PyObject *__pyx_builtin_SyntaxError;
static PyObject *__pyx_builtin_sum;
static const char __pyx_k_[] = " -> ";
static const char __pyx_k_0[] = "{0}";
static const char __pyx_k_L[] = "L";
static const char __pyx_k_b[] = "\\b";
static const char __pyx_k_c[] = "c";
static const char __pyx_k_d[] = "d";
static const char __pyx_k_e[] = "e";
static const char __pyx_k_f[] = "\\f";
static const char __pyx_k_i[] = "i";
static const char __pyx_k_j[] = "j";
static const char __pyx_k_k[] = "k";
static const char __pyx_k_l[] = "l";
static const char __pyx_k_n[] = "\\n";
static const char __pyx_k_o[] = "o";
static const char __pyx_k_r[] = "r";
static const char __pyx_k_s[] = "\"%s\"";
static const char __pyx_k_t[] = "\\t";
static const char __pyx_k_v[] = "v";
static const char __pyx_k_CR[] = "<CR>";
static const char __pyx_k_FF[] = "<FF>";
static const char __pyx_k_LF[] = "<LF>";
static const char __pyx_k_NL[] = "NL";
static const char __pyx_k_Or[] = "Or";
static const char __pyx_k_SP[] = "<SP>";
static const char __pyx_k__2[] = "\\^-[]";
static const char __pyx_k__4[] = "\n";
static const char __pyx_k__6[] = "\t";
static const char __pyx_k__8[] = "&";
static const char __pyx_k__9[] = ";";
static const char __pyx_k_eq[] = "__eq__";
static const char __pyx_k_ff[] = "ff";
static const char __pyx_k_fn[] = "fn";
static const char __pyx_k_id[] = "id";
static const char __pyx_k_ne[] = "__ne__";
static const char __pyx_k_nl[] = "nl";
static const char __pyx_k_nm[] = "nm";
static const char __pyx_k_or[] = "__or__";
static const char __pyx_k_p1[] = "p1";
static const char __pyx_k_p2[] = "p2";
static const char __pyx_k_pa[] = "pa";
static const char __pyx_k_pe[] = "pe";
static const char __pyx_k_re[] = "re";
static const char __pyx_k_se[] = "se";
static const char __pyx_k_tb[] = "tb";
static const char __pyx_k_to[] = "to_";
static const char __pyx_k_vv[] = "vv";
static const char __pyx_k_ws[] = "ws";
static const char __pyx_k_wt[] = "wt";
static const char __pyx_k_0_1[] = "{0}: {1}";
static const char __pyx_k_0_2[] = "... [{0}]";
static const char __pyx_k_And[] = "And";
static const char __pyx_k_BOM[] = "BOM";
static const char __pyx_k_HIT[] = "HIT";
static const char __pyx_k_L_2[] = "_L";
static const char __pyx_k_TAB[] = "<TAB>";
static const char __pyx_k_TRY[] = "TRY";
static const char __pyx_k_W_s[] = "W:(%s)";
static const char __pyx_k__10[] = "&><\"'";
static const char __pyx_k__12[] = "";
static const char __pyx_k__14[] = "[";
static const char __pyx_k__15[] = ", ";
static const char __pyx_k__16[] = "]";
static const char __pyx_k__19[] = "  ";
static const char __pyx_k__20[] = "<";
static const char __pyx_k__21[] = ">";
static const char __pyx_k__22[] = "</";
static const char __pyx_k__25[] = "\\\\";
static const char __pyx_k__26[] = "\\";
static const char __pyx_k__28[] = ">!<";
static const char __pyx_k__29[] = " ";
static const char __pyx_k__30[] = "^";
static const char __pyx_k__33[] = "*";
static const char __pyx_k__79[] = "#";
static const char __pyx_k__80[] = "\357\273\277";
static const char __pyx_k__85[] = "...";
static const char __pyx_k__98[] = "|(?:";
static const char __pyx_k__99[] = ")|(?:";
static const char __pyx_k_add[] = "__add__";
static const char __pyx_k_and[] = "__and__";
static const char __pyx_k_chr[] = "chr";
static const char __pyx_k_cls[] = "cls";
static const char __pyx_k_col[] = "col";
static const char __pyx_k_cpy[] = "cpy";
static const char __pyx_k_d_d[] = "(%d,%d)";
static const char __pyx_k_dir[] = "__dir__";
static const char __pyx_k_doc[] = "__doc__";
static const char __pyx_k_end[] = "end";
static const char __pyx_k_err[] = "err";
static const char __pyx_k_exc[] = "exc";
static const char __pyx_k_f_2[] = "f";
static const char __pyx_k_fns[] = "fns";
static const char __pyx_k_frm[] = "frm";
static const char __pyx_k_g_1[] = "\\g<1>";
static const char __pyx_k_get[] = "get";
static const char __pyx_k_ior[] = "__ior__";
static const char __pyx_k_key[] = "key";
static const char __pyx_k_len[] = "__len__";
static const char __pyx_k_loc[] = "loc";
static const char __pyx_k_map[] = "map";
static const char __pyx_k_mat[] = "mat";
static const char __pyx_k_max[] = "max";
static const char __pyx_k_min[] = "min";
static const char __pyx_k_msg[] = "msg";
static const char __pyx_k_mul[] = "__mul__";
static const char __pyx_k_n_2[] = "n";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_obj[] = "obj";
static const char __pyx_k_out[] = "out";
static const char __pyx_k_par[] = "par";
static const char __pyx_k_pbe[] = "pbe";
static const char __pyx_k_pdb[] = "pdb";
static const char __pyx_k_pop[] = "pop";
static const char __pyx_k_r_2[] = "\\r";
static const char __pyx_k_ref[] = "ref";
static const char __pyx_k_req[] = "__req__";
static const char __pyx_k_res[] = "res";
static const char __pyx_k_ret[] = "ret";
static const char __pyx_k_rne[] = "__rne__";
static const char __pyx_k_ror[] = "__ror__";
static const char __pyx_k_s_2[] = "'%s'";
static const char __pyx_k_s_3[] = "[%s]+";
static const char __pyx_k_s_4[] = "s";
static const char __pyx_k_s_5[] = "[%s]";
static const char __pyx_k_s_6[] = "|(?:%s)";
static const char __pyx_k_s_7[] = "|(?:%s.)";
static const char __pyx_k_s_8[] = ")*%s";
static const char __pyx_k_s_m[] = "s_m";
static const char __pyx_k_s_s[] = "(%s, %s)";
static const char __pyx_k_sep[] = "sep";
static const char __pyx_k_set[] = "set";
static const char __pyx_k_src[] = "src";
static const char __pyx_k_str[] = "__str__";
static const char __pyx_k_sub[] = "sub";
static const char __pyx_k_sum[] = "sum";
static const char __pyx_k_sys[] = "sys";
static const char __pyx_k_t_2[] = "t";
static const char __pyx_k_tmp[] = "tmp";
static const char __pyx_k_tok[] = "tok";
static const char __pyx_k_tup[] = "tup";
static const char __pyx_k_xor[] = "__xor__";
static const char __pyx_k_zip[] = "zip";
static const char __pyx_k_Char[] = "Char";
static const char __pyx_k_Dict[] = "Dict";
static const char __pyx_k_Each[] = "Each";
static const char __pyx_k_FAIL[] = "FAIL: ";
static const char __pyx_k_ITEM[] = "ITEM";
static const char __pyx_k_MISS[] = "MISS";
static const char __pyx_k_MMSP[] = "<MMSP>";
static const char __pyx_k_NBSP[] = "<NBSP>";
static const char __pyx_k_None[] = "None";
static const char __pyx_k_PY_3[] = "PY_3";
static const char __pyx_k_Re_s[] = "Re:(%s)";
static const char __pyx_k_Word[] = "Word";
static const char __pyx_k__100[] = ")";
static const char __pyx_k__101[] = "(.)";
static const char __pyx_k__102[] = "\014";
static const char __pyx_k__103[] = "\r";
static const char __pyx_k__105[] = " \t\r\n";
static const char __pyx_k__107[] = "-";
static const char __pyx_k__109[] = "{";
static const char __pyx_k__110[] = "}";
static const char __pyx_k__112[] = " ^ ";
static const char __pyx_k__113[] = " | ";
static const char __pyx_k__114[] = " & ";
static const char __pyx_k__115[] = "~{";
static const char __pyx_k__116[] = "}...";
static const char __pyx_k__117[] = "]...";
static const char __pyx_k__118[] = ": ...";
static const char __pyx_k__120[] = ": ";
static const char __pyx_k__268[] = " \n\t\r";
static const char __pyx_k__456[] = "_$";
static const char __pyx_k__527[] = "\302\240";
static const char __pyx_k__528[] = "\341\232\200";
static const char __pyx_k__529[] = "\341\240\216";
static const char __pyx_k__530[] = "\342\200\200";
static const char __pyx_k__531[] = "\342\200\201";
static const char __pyx_k__532[] = "\342\200\202";
static const char __pyx_k__533[] = "\342\200\203";
static const char __pyx_k__534[] = "\342\200\204";
static const char __pyx_k__535[] = "\342\200\205";
static const char __pyx_k__536[] = "\342\200\206";
static const char __pyx_k__537[] = "\342\200\207";
static const char __pyx_k__538[] = "\342\200\210";
static const char __pyx_k__539[] = "\342\200\211";
static const char __pyx_k__540[] = "\342\200\212";
static const char __pyx_k__541[] = "\342\200\213";
static const char __pyx_k__542[] = "\342\200\257";
static const char __pyx_k__543[] = "\342\201\237";
static const char __pyx_k__544[] = "\343\200\200";
static const char __pyx_k__673[] = "_";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_bool[] = "__bool__";
static const char __pyx_k_call[] = "__call__";
static const char __pyx_k_code[] = "code";
static const char __pyx_k_copy[] = "copy";
static const char __pyx_k_data[] = "data";
static const char __pyx_k_diag[] = "__diag__";
static const char __pyx_k_dump[] = "dump";
static const char __pyx_k_elem[] = "elem";
static const char __pyx_k_exit[] = "__exit__";
static const char __pyx_k_expr[] = "expr";
static const char __pyx_k_file[] = "file";
static const char __pyx_k_find[] = "find";
static const char __pyx_k_flag[] = "flag";
static const char __pyx_k_from[] = "from_";
static const char __pyx_k_full[] = "full";
static const char __pyx_k_hash[] = "__hash__";
static const char __pyx_k_iadd[] = "__iadd__";
static const char __pyx_k_ikey[] = "ikey";
static const char __pyx_k_init[] = "__init__";
static const char __pyx_k_item[] = "item";
static const char __pyx_k_iter[] = "__iter__";
static const char __pyx_k_ixor[] = "__ixor__";
static const char __pyx_k_join[] = "join";
static const char __pyx_k_keys[] = "keys";
static const char __pyx_k_last[] = "last";
static const char __pyx_k_line[] = "line";
static const char __pyx_k_loc1[] = "loc1";
static const char __pyx_k_loc2[] = "loc2";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "name";
static const char __pyx_k_nums[] = "nums";
static const char __pyx_k_open[] = "open";
static const char __pyx_k_opt1[] = "opt1";
static const char __pyx_k_opt2[] = "opt2";
static const char __pyx_k_pstr[] = "pstr";
static const char __pyx_k_radd[] = "__radd__";
static const char __pyx_k_rand[] = "__rand__";
static const char __pyx_k_read[] = "read";
static const char __pyx_k_repl[] = "repl";
static const char __pyx_k_repr[] = "__repr__";
static const char __pyx_k_rmul[] = "__rmul__";
static const char __pyx_k_rsub[] = "__rsub__";
static const char __pyx_k_rxor[] = "__rxor__";
static const char __pyx_k_seen[] = "seen";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_size[] = "size";
static const char __pyx_k_sort[] = "sort";
static const char __pyx_k_strg[] = "strg";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_toks[] = "toks";
static const char __pyx_k_vars[] = "vars";
static const char __pyx_k_warn[] = "warn";
static const char __pyx_k_0_1_2[] = "{0}.{1} - {2}";
static const char __pyx_k_0_1_3[] = "{0}.{1}";
static const char __pyx_k_Empty[] = "Empty";
static const char __pyx_k_FATAL[] = "(FATAL)";
static const char __pyx_k_Group[] = "Group";
static const char __pyx_k_MATCH[] = "MATCH";
static const char __pyx_k_Match[] = "Match ";
static const char __pyx_k_NNBSP[] = "<NNBSP>";
static const char __pyx_k_RLock[] = "RLock";
static const char __pyx_k_Regex[] = "Regex";
static const char __pyx_k_Token[] = "Token";
static const char __pyx_k_W_s_2[] = "!W:(%s...)";
static const char __pyx_k_W_s_3[] = "!W:(%s)";
static const char __pyx_k_W_s_s[] = "W:(%s, %s)";
static const char __pyx_k_White[] = "White";
static const char __pyx_k_aname[] = "aname";
static const char __pyx_k_asXML[] = "asXML";
static const char __pyx_k_b_s_b[] = "\\b%s\\b";
static const char __pyx_k_cache[] = "cache";
static const char __pyx_k_cause[] = "__cause__";
static const char __pyx_k_chars[] = "chars";
static const char __pyx_k_class[] = "__class__";
static const char __pyx_k_clear[] = "clear";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_colno[] = "colno";
static const char __pyx_k_count[] = "count";
static const char __pyx_k_debug[] = "debug";
static const char __pyx_k_depth[] = "_depth";
static const char __pyx_k_deque[] = "deque";
static const char __pyx_k_dummy[] = "dummy";
static const char __pyx_k_ender[] = "ender";
static const char __pyx_k_enter[] = "__enter__";
static const char __pyx_k_error[] = "error";
static const char __pyx_k_exact[] = "exact";
static const char __pyx_k_expr1[] = "expr1";
static const char __pyx_k_exprs[] = "exprs";
static const char __pyx_k_fatal[] = "fatal";
static const char __pyx_k_flags[] = "flags";
static const char __pyx_k_group[] = "group";
static const char __pyx_k_index[] = "index";
static const char __pyx_k_items[] = "items";
static const char __pyx_k_lastE[] = "lastE";
static const char __pyx_k_match[] = "match";
static const char __pyx_k_modal[] = "modal";
static const char __pyx_k_mylen[] = "mylen";
static const char __pyx_k_other[] = "other";
static const char __pyx_k_parse[] = "_parse";
static const char __pyx_k_print[] = "print_";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_reset[] = "reset";
static const char __pyx_k_rfind[] = "rfind";
static const char __pyx_k_s_s_2[] = "%s:(%s)";
static const char __pyx_k_s_s_3[] = "%s[%s]*";
static const char __pyx_k_s_s_4[] = "[%s][%s]*";
static const char __pyx_k_s_s_5[] = "%s[^%s]";
static const char __pyx_k_s_s_s[] = "%s%s- %s: ";
static const char __pyx_k_split[] = "split";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_state[] = "state";
static const char __pyx_k_strip[] = "strip";
static const char __pyx_k_sub_2[] = "__sub__";
static const char __pyx_k_super[] = "super";
static const char __pyx_k_tests[] = "tests";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_types[] = "types";
static const char __pyx_k_upper[] = "upper";
static const char __pyx_k_value[] = "value";
static const char __pyx_k_vlist[] = "vlist";
static const char __pyx_k_wkref[] = "wkref";
static const char __pyx_k_write[] = "write";
static const char __pyx_k_wslit[] = "wslit";
static const char __pyx_k_DOTALL[] = "DOTALL";
static const char __pyx_k_FAIL_2[] = "FAIL";
static const char __pyx_k_NotAny[] = "NotAny";
static const char __pyx_k_SkipTo[] = "SkipTo";
static const char __pyx_k_alphas[] = "alphas";
static const char __pyx_k_append[] = "append";
static const char __pyx_k_asDict[] = "asDict";
static const char __pyx_k_asList[] = "asList";
static const char __pyx_k_at_loc[] = " at loc ";
static const char __pyx_k_called[] = "called";
static const char __pyx_k_column[] = "column";
static const char __pyx_k_doctag[] = "doctag";
static const char __pyx_k_enable[] = "enable_";
static const char __pyx_k_endloc[] = "endloc";
static const char __pyx_k_errmsg[] = "errmsg";
static const char __pyx_k_escape[] = "escape";
static const char __pyx_k_expand[] = "expand";
static const char __pyx_k_extend[] = "extend";
static const char __pyx_k_f_code[] = "f_code";
static const char __pyx_k_f_self[] = "f_self";
static const char __pyx_k_failOn[] = "failOn";
static const char __pyx_k_failed[] = "failed";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_groups[] = "groups";
static const char __pyx_k_ignore[] = "ignore";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_indent[] = "indent";
static const char __pyx_k_insStr[] = "insStr";
static const char __pyx_k_insert[] = "insert";
static const char __pyx_k_invert[] = "__invert__";
static const char __pyx_k_kwargs[] = "kwargs";
static const char __pyx_k_lastCR[] = "lastCR";
static const char __pyx_k_lineno[] = "lineno";
static const char __pyx_k_lookup[] = "__lookup";
static const char __pyx_k_lshift[] = "__lshift__";
static const char __pyx_k_lstrip[] = "lstrip";
static const char __pyx_k_maxLen[] = "maxLen";
static const char __pyx_k_maxlen[] = "maxlen";
static const char __pyx_k_maxloc[] = "maxloc";
static const char __pyx_k_minLen[] = "minLen";
static const char __pyx_k_module[] = "__module__";
static const char __pyx_k_name_2[] = "__name__";
static const char __pyx_k_newloc[] = "newloc";
static const char __pyx_k_nextCR[] = "nextCR";
static const char __pyx_k_object[] = "object";
static const char __pyx_k_offset[] = "offset";
static const char __pyx_k_pprint[] = "pprint";
static const char __pyx_k_preloc[] = "preloc";
static const char __pyx_k_remove[] = "remove";
static const char __pyx_k_resTag[] = "resTag";
static const char __pyx_k_result[] = "result";
static const char __pyx_k_retobj[] = "retobj";
static const char __pyx_k_rstrip[] = "rstrip";
static const char __pyx_k_splits[] = "splits";
static const char __pyx_k_stdout[] = "stdout";
static const char __pyx_k_stopOn[] = "stopOn";
static const char __pyx_k_string[] = "string";
static const char __pyx_k_thread[] = "_thread";
static const char __pyx_k_tmpLoc[] = "tmpLoc";
static const char __pyx_k_tmpOpt[] = "tmpOpt";
static const char __pyx_k_tmploc[] = "tmploc";
static const char __pyx_k_toItem[] = "toItem";
static const char __pyx_k_tokens[] = "tokens";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_values[] = "values";
static const char __pyx_k_warn_2[] = "warn_";
static const char __pyx_k_ws_map[] = "ws_map";
static const char __pyx_k_wschar[] = "wschar";
static const char __pyx_k_Combine[] = "Combine";
static const char __pyx_k_EM_QUAD[] = "<EM_QUAD>";
static const char __pyx_k_EN_QUAD[] = "<EN_QUAD>";
static const char __pyx_k_Forward[] = "Forward";
static const char __pyx_k_Keyword[] = "Keyword";
static const char __pyx_k_LineEnd[] = "LineEnd";
static const char __pyx_k_Literal[] = "Literal";
static const char __pyx_k_Mapping[] = "Mapping";
static const char __pyx_k_Matched[] = "Matched ";
static const char __pyx_k_NoMatch[] = "NoMatch";
static const char __pyx_k_WordEnd[] = "WordEnd";
static const char __pyx_k_asMatch[] = "asMatch";
static const char __pyx_k_breaker[] = "breaker";
static const char __pyx_k_callers[] = "callers";
static const char __pyx_k_charset[] = "charset";
static const char __pyx_k_co_name[] = "co_name";
static const char __pyx_k_comment[] = "comment";
static const char __pyx_k_compile[] = "compile";
static const char __pyx_k_context[] = "context";
static const char __pyx_k_default[] = "default";
static const char __pyx_k_delitem[] = "__delitem__";
static const char __pyx_k_depth_2[] = "depth";
static const char __pyx_k_escChar[] = "escChar";
static const char __pyx_k_explain[] = "explain";
static const char __pyx_k_flatten[] = "_flatten";
static const char __pyx_k_found_r[] = ", found %r";
static const char __pyx_k_genexpr[] = "genexpr";
static const char __pyx_k_getName[] = "getName";
static const char __pyx_k_getattr[] = "__getattr__";
static const char __pyx_k_getitem[] = "__getitem__";
static const char __pyx_k_haskeys[] = "haskeys";
static const char __pyx_k_hexnums[] = "hexnums";
static const char __pyx_k_ilshift[] = "__ilshift__";
static const char __pyx_k_include[] = "include";
static const char __pyx_k_indices[] = "indices";
static const char __pyx_k_inspect[] = "inspect";
static const char __pyx_k_isspace[] = "isspace";
static const char __pyx_k_item_fn[] = "item_fn";
static const char __pyx_k_itemseq[] = "itemseq";
static const char __pyx_k_longest[] = "longest";
static const char __pyx_k_matches[] = "matches";
static const char __pyx_k_maxsize[] = "maxsize";
static const char __pyx_k_message[] = "message";
static const char __pyx_k_missing[] = "missing";
static const char __pyx_k_newself[] = "newself";
static const char __pyx_k_nextLoc[] = "nextLoc";
static const char __pyx_k_nextloc[] = "nextloc";
static const char __pyx_k_nonzero[] = "__nonzero__";
static const char __pyx_k_opt1map[] = "opt1map";
static const char __pyx_k_overlap[] = "overlap";
static const char __pyx_k_parseFn[] = "parseFn";
static const char __pyx_k_pattern[] = "pattern";
static const char __pyx_k_popitem[] = "popitem";
static const char __pyx_k_popleft[] = "popleft";
static const char __pyx_k_prepare[] = "__prepare__";
static const char __pyx_k_print_2[] = "print";
static const char __pyx_k_removed[] = "removed";
static const char __pyx_k_replace[] = "replace";
static const char __pyx_k_results[] = "results";
static const char __pyx_k_retToks[] = "retToks";
static const char __pyx_k_retreat[] = "retreat";
static const char __pyx_k_reverse[] = "reverse";
static const char __pyx_k_s_s_s_2[] = "%s(?:[^%s%s]";
static const char __pyx_k_selfTag[] = "selfTag";
static const char __pyx_k_setName[] = "setName";
static const char __pyx_k_setitem[] = "__setitem__";
static const char __pyx_k_skipped[] = "_skipped*";
static const char __pyx_k_strRepr[] = "strRepr";
static const char __pyx_k_success[] = "success";
static const char __pyx_k_tb_next[] = "tb_next";
static const char __pyx_k_thiscol[] = "thiscol";
static const char __pyx_k_tmpReqd[] = "tmpReqd";
static const char __pyx_k_toklist[] = "toklist";
static const char __pyx_k_weakref[] = "weakref";
static const char __pyx_k_wrapper[] = "wrapper";
static const char __pyx_k_EM_SPACE[] = "<EM_SPACE>";
static const char __pyx_k_EN_SPACE[] = "<EN_SPACE>";
static const char __pyx_k_Ellipsis[] = "Ellipsis";
static const char __pyx_k_Expected[] = "Expected ";
static const char __pyx_k_Iterable[] = "Iterable";
static const char __pyx_k_KeyError[] = "KeyError";
static const char __pyx_k_OnlyOnce[] = "OnlyOnce";
static const char __pyx_k_Optional[] = "Optional";
static const char __pyx_k_Or___str[] = "Or.__str__";
static const char __pyx_k_Suppress[] = "Suppress";
static const char __pyx_k_adjacent[] = "adjacent";
static const char __pyx_k_callable[] = "callable";
static const char __pyx_k_caseless[] = "caseless";
static const char __pyx_k_comments[] = "comments";
static const char __pyx_k_contains[] = "__contains__";
static const char __pyx_k_endswith[] = "endswith";
static const char __pyx_k_escQuote[] = "escQuote";
static const char __pyx_k_f_locals[] = "f_locals";
static const char __pyx_k_foundstr[] = "foundstr";
static const char __pyx_k_fullDump[] = "fullDump";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_instring[] = "instring";
static const char __pyx_k_instrlen[] = "instrlen";
static const char __pyx_k_iterkeys[] = "iterkeys";
static const char __pyx_k_keepTabs[] = "keepTabs";
static const char __pyx_k_key_fifo[] = "key_fifo";
static const char __pyx_k_line_str[] = "line_str";
static const char __pyx_k_lookup_2[] = "lookup";
static const char __pyx_k_matchLen[] = "matchLen";
static const char __pyx_k_maxsplit[] = "maxsplit";
static const char __pyx_k_module_2[] = "<module>";
static const char __pyx_k_notChars[] = "notChars";
static const char __pyx_k_notchars[] = "notchars";
static const char __pyx_k_operator[] = "operator";
static const char __pyx_k_original[] = "original";
static const char __pyx_k_parseAll[] = "parseAll";
static const char __pyx_k_position[] = "position";
static const char __pyx_k_pp_value[] = "pp_value";
static const char __pyx_k_preParse[] = "preParse";
static const char __pyx_k_qualname[] = "__qualname__";
static const char __pyx_k_reString[] = "reString";
static const char __pyx_k_re_match[] = "re_match";
static const char __pyx_k_required[] = "required";
static const char __pyx_k_reversed[] = "__reversed__";
static const char __pyx_k_runTests[] = "runTests";
static const char __pyx_k_savelist[] = "savelist";
static const char __pyx_k_setBreak[] = "setBreak";
static const char __pyx_k_setDebug[] = "setDebug";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_skiptext[] = "skiptext";
static const char __pyx_k_startloc[] = "startloc";
static const char __pyx_k_suppress[] = "suppress";
static const char __pyx_k_tmpExprs[] = "tmpExprs";
static const char __pyx_k_tryParse[] = "tryParse";
static const char __pyx_k_validate[] = "validate";
static const char __pyx_k_warnings[] = "warnings";
static const char __pyx_k_And___str[] = "And.__str__";
static const char __pyx_k_ErrorStop[] = "_ErrorStop";
static const char __pyx_k_FifoCache[] = "_FifoCache";
static const char __pyx_k_LineStart[] = "LineStart";
static const char __pyx_k_MULTILINE[] = "MULTILINE";
static const char __pyx_k_NullToken[] = "_NullToken";
static const char __pyx_k_OneOrMore[] = "OneOrMore";
static const char __pyx_k_Or___init[] = "Or.__init__";
static const char __pyx_k_Or___ixor[] = "Or.__ixor__";
static const char __pyx_k_Regex_sub[] = "Regex.sub";
static const char __pyx_k_StringEnd[] = "StringEnd";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_WordRegex[] = "_WordRegex";
static const char __pyx_k_WordStart[] = "WordStart";
static const char __pyx_k_addoffset[] = "addoffset";
static const char __pyx_k_all_names[] = "_all_names";
static const char __pyx_k_alphanums[] = "alphanums";
static const char __pyx_k_asKeyword[] = "asKeyword";
static const char __pyx_k_best_expr[] = "best_expr";
static const char __pyx_k_bodyChars[] = "bodyChars";
static const char __pyx_k_bodychars[] = "bodychars";
static const char __pyx_k_breakFlag[] = "breakFlag";
static const char __pyx_k_cache_len[] = "cache_len";
static const char __pyx_k_debugging[] = "debugging";
static const char __pyx_k_dictvalue[] = "dictvalue";
static const char __pyx_k_doActions[] = "doActions";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_errorStop[] = "errorStop";
static const char __pyx_k_formatted[] = "formatted";
static const char __pyx_k_from_dict[] = "from_dict";
static const char __pyx_k_groupdict[] = "groupdict";
static const char __pyx_k_initChars[] = "initChars";
static const char __pyx_k_iteritems[] = "iteritems";
static const char __pyx_k_last_expr[] = "last_expr";
static const char __pyx_k_maxExcLoc[] = "maxExcLoc";
static const char __pyx_k_metaclass[] = "__metaclass__";
static const char __pyx_k_multiline[] = "multiline";
static const char __pyx_k_must_skip[] = "must_skip";
static const char __pyx_k_not_ender[] = "not_ender";
static const char __pyx_k_optionals[] = "optionals";
static const char __pyx_k_parseFile[] = "parseFile";
static const char __pyx_k_parseImpl[] = "parseImpl";
static const char __pyx_k_postParse[] = "postParse";
static const char __pyx_k_printable[] = "printable";
static const char __pyx_k_pyparsing[] = "pyparsing";
static const char __pyx_k_quoteChar[] = "quoteChar";
static const char __pyx_k_retString[] = "retString";
static const char __pyx_k_retTokens[] = "retTokens";
static const char __pyx_k_s_s_n_r_s[] = "%s(?:[^%s\\n\\r%s]";
static const char __pyx_k_self_type[] = "self_type";
static const char __pyx_k_setOffset[] = "setOffset";
static const char __pyx_k_set_trace[] = "set_trace";
static const char __pyx_k_test_expr[] = "test_expr";
static const char __pyx_k_threading[] = "threading";
static const char __pyx_k_tmptokens[] = "tmptokens";
static const char __pyx_k_tokenlist[] = "tokenlist";
static const char __pyx_k_traceback[] = "__traceback__";
static const char __pyx_k_whiteStrs[] = "whiteStrs";
static const char __pyx_k_wordChars[] = "wordChars";
static const char __pyx_k_0123456789[] = "0123456789";
static const char __pyx_k_And___iadd[] = "And.__iadd__";
static const char __pyx_k_And___init[] = "And.__init__";
static const char __pyx_k_CharsNotIn[] = "CharsNotIn";
static const char __pyx_k_CloseMatch[] = "CloseMatch";
static const char __pyx_k_Each___str[] = "Each.__str__";
static const char __pyx_k_FollowedBy[] = "FollowedBy";
static const char __pyx_k_GoToColumn[] = "GoToColumn";
static const char __pyx_k_HAIR_SPACE[] = "<HAIR_SPACE>";
static const char __pyx_k_IndexError[] = "IndexError";
static const char __pyx_k_MatchFirst[] = "MatchFirst";
static const char __pyx_k_MethodType[] = "MethodType";
static const char __pyx_k_PrecededBy[] = "PrecededBy";
static const char __pyx_k_THIN_SPACE[] = "<THIN_SPACE>";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_Word___str[] = "Word.__str__";
static const char __pyx_k_ZeroOrMore[] = "ZeroOrMore";
static const char __pyx_k_allResults[] = "allResults";
static const char __pyx_k_charsAsStr[] = "charsAsStr";
static const char __pyx_k_expandtabs[] = "expandtabs";
static const char __pyx_k_expr_parse[] = "expr_parse";
static const char __pyx_k_exprsFound[] = "exprsFound";
static const char __pyx_k_exprtokens[] = "exprtokens";
static const char __pyx_k_failAction[] = "failAction";
static const char __pyx_k_getnewargs[] = "__getnewargs__";
static const char __pyx_k_identChars[] = "identChars";
static const char __pyx_k_ignoreExpr[] = "ignoreExpr";
static const char __pyx_k_isinstance[] = "isinstance";
static const char __pyx_k_itemgetter[] = "itemgetter";
static const char __pyx_k_iterkeys_2[] = "_iterkeys";
static const char __pyx_k_itervalues[] = "itervalues";
static const char __pyx_k_joinString[] = "joinString";
static const char __pyx_k_matchOrder[] = "matchOrder";
static const char __pyx_k_matchWhite[] = "matchWhite";
static const char __pyx_k_maxMatches[] = "maxMatches";
static const char __pyx_k_methodCall[] = "methodCall";
static const char __pyx_k_mismatches[] = "mismatches";
static const char __pyx_k_namedItems[] = "namedItems";
static const char __pyx_k_otheritems[] = "otheritems";
static const char __pyx_k_parseCache[] = "_parseCache";
static const char __pyx_k_preparseFn[] = "preparseFn";
static const char __pyx_k_printables[] = "printables";
static const char __pyx_k_resetCache[] = "resetCache";
static const char __pyx_k_resultlist[] = "resultlist";
static const char __pyx_k_saveAsList[] = "saveAsList";
static const char __pyx_k_scanString[] = "scanString";
static const char __pyx_k_skipresult[] = "skipresult";
static const char __pyx_k_skipto_arg[] = "skipto_arg";
static const char __pyx_k_splitlines[] = "splitlines";
static const char __pyx_k_stacklevel[] = "stacklevel";
static const char __pyx_k_startswith[] = "startswith";
static const char __pyx_k_streamline[] = "streamline";
static const char __pyx_k_testString[] = "testString";
static const char __pyx_k_to_symbols[] = "to_symbols";
static const char __pyx_k_trim_arity[] = "_trim_arity";
static const char __pyx_k_whiteChars[] = "whiteChars";
static const char __pyx_k_whitespace[] = "whitespace";
static const char __pyx_k_xml_escape[] = "_xml_escape";
static const char __pyx_k_Char___init[] = "Char.__init__";
static const char __pyx_k_Dict___init[] = "Dict.__init__";
static const char __pyx_k_Each___init[] = "Each.__init__";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_OrderedDict[] = "_OrderedDict";
static const char __pyx_k_PendingSkip[] = "_PendingSkip";
static const char __pyx_k_Regex___str[] = "Regex.__str__";
static const char __pyx_k_StringStart[] = "StringStart";
static const char __pyx_k_SyntaxError[] = "SyntaxError";
static const char __pyx_k_Word___init[] = "Word.__init__";
static const char __pyx_k_asGroupList[] = "asGroupList";
static const char __pyx_k_check_ender[] = "check_ender";
static const char __pyx_k_collections[] = "collections";
static const char __pyx_k_ignoreExprs[] = "ignoreExprs";
static const char __pyx_k_is_iterable[] = "is_iterable";
static const char __pyx_k_iteritems_2[] = "_iteritems";
static const char __pyx_k_line_column[] = "line_column";
static const char __pyx_k_matchString[] = "matchString";
static const char __pyx_k_minElements[] = "minElements";
static const char __pyx_k_occurrences[] = "occurrences";
static const char __pyx_k_optElements[] = "optElements";
static const char __pyx_k_ordereddict[] = "ordereddict";
static const char __pyx_k_parseAction[] = "parseAction";
static const char __pyx_k_parseMethod[] = "_parseMethod";
static const char __pyx_k_parseString[] = "parseString";
static const char __pyx_k_replaceWith[] = "replaceWith";
static const char __pyx_k_resultsName[] = "resultsName";
static const char __pyx_k_s_s_d_s_s_s[] = "\n%s%s[%d]:\n%s%s%s";
static const char __pyx_k_startAction[] = "startAction";
static const char __pyx_k_streamlined[] = "streamlined";
static const char __pyx_k_tokensStart[] = "tokensStart";
static const char __pyx_k_xmlBodyText[] = "xmlBodyText";
static const char __pyx_k_0_failed_1_2[] = "{0} failed: {1}: {2}";
static const char __pyx_k_ABCDEFabcdef[] = "ABCDEFabcdef";
static const char __pyx_k_Empty___init[] = "Empty.__init__";
static const char __pyx_k_FIGURE_SPACE[] = "<FIGURE_SPACE>";
static const char __pyx_k_Forward_copy[] = "Forward.copy";
static const char __pyx_k_Group___init[] = "Group.__init__";
static const char __pyx_k_Keyword_copy[] = "Keyword.copy";
static const char __pyx_k_NotAny___str[] = "NotAny.__str__";
static const char __pyx_k_Or_parseImpl[] = "Or.parseImpl";
static const char __pyx_k_ParseResults[] = "ParseResults";
static const char __pyx_k_QuotedString[] = "QuotedString";
static const char __pyx_k_Regex___init[] = "Regex.__init__";
static const char __pyx_k_Token___init[] = "Token.__init__";
static const char __pyx_k_White___init[] = "White.__init__";
static const char __pyx_k_addCondition[] = "addCondition";
static const char __pyx_k_asStringList[] = "_asStringList";
static const char __pyx_k_callPreParse[] = "callPreParse";
static const char __pyx_k_callPreparse[] = "callPreparse";
static const char __pyx_k_canParseNext[] = "canParseNext";
static const char __pyx_k_debugActions[] = "debugActions";
static const char __pyx_k_defaultValue[] = "defaultValue";
static const char __pyx_k_defaultvalue[] = "defaultvalue";
static const char __pyx_k_endQuoteChar[] = "endQuoteChar";
static const char __pyx_k_excludeChars[] = "excludeChars";
static const char __pyx_k_failureTests[] = "failureTests";
static const char __pyx_k_finalResults[] = "finalResults";
static const char __pyx_k_from_symbols[] = "from_symbols";
static const char __pyx_k_inAccumNames[] = "inAccumNames";
static const char __pyx_k_includeMatch[] = "includeMatch";
static const char __pyx_k_include_list[] = "include_list";
static const char __pyx_k_itervalues_2[] = "_itervalues";
static const char __pyx_k_keepMatching[] = "keepMatching";
static const char __pyx_k_markerString[] = "markerString";
static const char __pyx_k_match_string[] = "match_string";
static const char __pyx_k_maxException[] = "maxException";
static const char __pyx_k_maxSpecified[] = "maxSpecified";
static const char __pyx_k_modalResults[] = "modalResults";
static const char __pyx_k_not_in_cache[] = "not_in_cache";
static const char __pyx_k_parseNoCache[] = "_parseNoCache";
static const char __pyx_k_printResults[] = "printResults";
static const char __pyx_k_quoteCharLen[] = "quoteCharLen";
static const char __pyx_k_quotedString[] = "quotedString";
static const char __pyx_k_returnString[] = "returnString";
static const char __pyx_k_searchString[] = "searchString";
static const char __pyx_k_staticmethod[] = "staticmethod";
static const char __pyx_k_version_info[] = "version_info";
static const char __pyx_k_And_parseImpl[] = "And.parseImpl";
static const char __pyx_k_Forward___str[] = "Forward.__str__";
static const char __pyx_k_MultipleMatch[] = "_MultipleMatch";
static const char __pyx_k_NotAny___init[] = "NotAny.__init__";
static const char __pyx_k_Or_streamline[] = "Or.streamline";
static const char __pyx_k_OrderedDict_2[] = "OrderedDict";
static const char __pyx_k_ParserElement[] = "ParserElement";
static const char __pyx_k_PositionToken[] = "_PositionToken";
static const char __pyx_k_SkipTo___init[] = "SkipTo.__init__";
static const char __pyx_k_SyntaxWarning[] = "SyntaxWarning";
static const char __pyx_k_bodyCharsOrig[] = "bodyCharsOrig";
static const char __pyx_k_callDuringTry[] = "callDuringTry";
static const char __pyx_k_caselessmatch[] = "caselessmatch";
static const char __pyx_k_enablePackrat[] = "enablePackrat";
static const char __pyx_k_file_contents[] = "file_contents";
static const char __pyx_k_generatorType[] = "_generatorType";
static const char __pyx_k_initCharsOrig[] = "initCharsOrig";
static const char __pyx_k_markInputline[] = "markInputline";
static const char __pyx_k_maxMismatches[] = "maxMismatches";
static const char __pyx_k_mayIndexError[] = "mayIndexError";
static const char __pyx_k_multirequired[] = "multirequired";
static const char __pyx_k_packrat_cache[] = "packrat_cache";
static const char __pyx_k_parseWithTabs[] = "parseWithTabs";
static const char __pyx_k_parserElement[] = "parserElement";
static const char __pyx_k_setFailAction[] = "setFailAction";
static const char __pyx_k_sre_constants[] = "sre_constants";
static const char __pyx_k_successAction[] = "successAction";
static const char __pyx_k_try_not_ender[] = "try_not_ender";
static const char __pyx_k_validateTrace[] = "validateTrace";
static const char __pyx_k_And__ErrorStop[] = "And._ErrorStop";
static const char __pyx_k_And_streamline[] = "And.streamline";
static const char __pyx_k_AttributeError[] = "AttributeError";
static const char __pyx_k_Combine___init[] = "Combine.__init__";
static const char __pyx_k_Combine_ignore[] = "Combine.ignore";
static const char __pyx_k_Dict_postParse[] = "Dict.postParse";
static const char __pyx_k_Each_parseImpl[] = "Each.parseImpl";
static const char __pyx_k_FAIL_EXCEPTION[] = "FAIL-EXCEPTION: ";
static const char __pyx_k_Forward___init[] = "Forward.__init__";
static const char __pyx_k_Keyword___init[] = "Keyword.__init__";
static const char __pyx_k_LineEnd___init[] = "LineEnd.__init__";
static const char __pyx_k_Literal___init[] = "Literal.__init__";
static const char __pyx_k_MutableMapping[] = "MutableMapping";
static const char __pyx_k_NoMatch___init[] = "NoMatch.__init__";
static const char __pyx_k_OnlyOnce_reset[] = "OnlyOnce.reset";
static const char __pyx_k_Optional___str[] = "Optional.__str__";
static const char __pyx_k_ParseException[] = "ParseException";
static const char __pyx_k_TokenConverter[] = "TokenConverter";
static const char __pyx_k_UnboundedCache[] = "_UnboundedCache";
static const char __pyx_k_WordEnd___init[] = "WordEnd.__init__";
static const char __pyx_k_Word_parseImpl[] = "Word.parseImpl";
static const char __pyx_k_addParseAction[] = "addParseAction";
static const char __pyx_k_checkRecursion[] = "checkRecursion";
static const char __pyx_k_firstMatchChar[] = "firstMatchChar";
static const char __pyx_k_firstQuoteChar[] = "firstQuoteChar";
static const char __pyx_k_from_exception[] = "_from_exception";
static const char __pyx_k_getinnerframes[] = "getinnerframes";
static const char __pyx_k_hasIgnoreExprs[] = "hasIgnoreExprs";
static const char __pyx_k_initExprGroups[] = "initExprGroups";
static const char __pyx_k_instring_slice[] = "instring_slice";
static const char __pyx_k_listAllMatches[] = "listAllMatches";
static const char __pyx_k_mayReturnEmpty[] = "mayReturnEmpty";
static const char __pyx_k_multioptionals[] = "multioptionals";
static const char __pyx_k_namedItemsOnly[] = "namedItemsOnly";
static const char __pyx_k_otherdictitems[] = "otherdictitems";
static const char __pyx_k_packratEnabled[] = "_packratEnabled";
static const char __pyx_k_setParseAction[] = "setParseAction";
static const char __pyx_k_setResultsName[] = "_setResultsName";
static const char __pyx_k_skipIgnorables[] = "_skipIgnorables";
static const char __pyx_k_skipWhitespace[] = "skipWhitespace";
static const char __pyx_k_system_version[] = "system_version";
static const char __pyx_k_throwException[] = "throwException";
static const char __pyx_k_trim_traceback[] = "_trim_traceback";
static const char __pyx_k_unquoteResults[] = "unquoteResults";
static const char __pyx_k_CaselessKeyword[] = "CaselessKeyword";
static const char __pyx_k_CaselessLiteral[] = "CaselessLiteral";
static const char __pyx_k_Each_streamline[] = "Each.streamline";
static const char __pyx_k_Group_postParse[] = "Group.postParse";
static const char __pyx_k_NullToken___str[] = "_NullToken.__str__";
static const char __pyx_k_OneOrMore___str[] = "OneOrMore.__str__";
static const char __pyx_k_OnlyOnce___call[] = "OnlyOnce.__call__";
static const char __pyx_k_OnlyOnce___init[] = "OnlyOnce.__init__";
static const char __pyx_k_Optional___init[] = "Optional.__init__";
static const char __pyx_k_ParseExpression[] = "ParseExpression";
static const char __pyx_k_Regex_parseImpl[] = "Regex.parseImpl";
static const char __pyx_k_SimpleNamespace[] = "SimpleNamespace";
static const char __pyx_k_White_parseImpl[] = "White.parseImpl";
static const char __pyx_k_ascii_lowercase[] = "ascii_lowercase";
static const char __pyx_k_ascii_uppercase[] = "ascii_uppercase";
static const char __pyx_k_collections_abc[] = "collections.abc";
static const char __pyx_k_endQuoteCharLen[] = "endQuoteCharLen";
static const char __pyx_k_exceptionAction[] = "exceptionAction";
static const char __pyx_k_leaveWhitespace[] = "leaveWhitespace";
static const char __pyx_k_lineno_col_line[] = "lineno col line";
static const char __pyx_k_match_stringloc[] = "match_stringloc";
static const char __pyx_k_nextLevelIndent[] = "nextLevelIndent";
static const char __pyx_k_not_preceded_by[] = "not preceded by ";
static const char __pyx_k_nullDebugAction[] = "nullDebugAction";
static const char __pyx_k_self_expr_parse[] = "self_expr_parse";
static const char __pyx_k_setDebugActions[] = "setDebugActions";
static const char __pyx_k_subRecCheckList[] = "subRecCheckList";
static const char __pyx_k_transformString[] = "transformString";
static const char __pyx_k_CharsNotIn___str[] = "CharsNotIn.__str__";
static const char __pyx_k_Exception_raised[] = "Exception raised:";
static const char __pyx_k_Forward___lshift[] = "Forward.__lshift__";
static const char __pyx_k_Forward_validate[] = "Forward.validate";
static const char __pyx_k_LineStart___init[] = "LineStart.__init__";
static const char __pyx_k_MatchFirst___ior[] = "MatchFirst.__ior__";
static const char __pyx_k_MatchFirst___str[] = "MatchFirst.__str__";
static const char __pyx_k_NotAny_parseImpl[] = "NotAny.parseImpl";
static const char __pyx_k_NullToken___bool[] = "_NullToken.__bool__";
static const char __pyx_k_OGHAM_SPACE_MARK[] = "<OGHAM_SPACE_MARK>";
static const char __pyx_k_ParseResults_get[] = "ParseResults.get";
static const char __pyx_k_ParseResults_pop[] = "ParseResults.pop";
static const char __pyx_k_SIX_PER_EM_SPACE[] = "<SIX-PER-EM_SPACE>";
static const char __pyx_k_SkipTo_parseImpl[] = "SkipTo.parseImpl";
static const char __pyx_k_StringEnd___init[] = "StringEnd.__init__";
static const char __pyx_k_WordStart___init[] = "WordStart.__init__";
static const char __pyx_k_ZERO_WIDTH_SPACE[] = "<ZERO_WIDTH_SPACE>";
static const char __pyx_k_ZeroOrMore___str[] = "ZeroOrMore.__str__";
static const char __pyx_k_cache_size_limit[] = "cache_size_limit";
static const char __pyx_k_file_or_filename[] = "file_or_filename";
static const char __pyx_k_makeOptionalList[] = "makeOptionalList";
static const char __pyx_k_parseElementList[] = "parseElementList";
static const char __pyx_k_parseImplAsMatch[] = "parseImplAsMatch";
static const char __pyx_k_parse_action_exc[] = "parse_action_exc";
static const char __pyx_k_setResultsName_2[] = "setResultsName";
static const char __pyx_k_CharsNotIn___init[] = "CharsNotIn.__init__";
static const char __pyx_k_CloseMatch___init[] = "CloseMatch.__init__";
static const char __pyx_k_Combine_postParse[] = "Combine.postParse";
static const char __pyx_k_FOUR_PER_EM_SPACE[] = "<FOUR-PER-EM_SPACE>";
static const char __pyx_k_FollowedBy___init[] = "FollowedBy.__init__";
static const char __pyx_k_Forward___ilshift[] = "Forward.__ilshift__";
static const char __pyx_k_GoToColumn___init[] = "GoToColumn.__init__";
static const char __pyx_k_IDEOGRAPHIC_SPACE[] = "<IDEOGRAPHIC_SPACE>";
static const char __pyx_k_Keyword_parseImpl[] = "Keyword.parseImpl";
static const char __pyx_k_LineEnd_parseImpl[] = "LineEnd.parseImpl";
static const char __pyx_k_Literal_parseImpl[] = "Literal.parseImpl";
static const char __pyx_k_MatchFirst___init[] = "MatchFirst.__init__";
static const char __pyx_k_NoMatch_parseImpl[] = "NoMatch.parseImpl";
static const char __pyx_k_Or_checkRecursion[] = "Or.checkRecursion";
static const char __pyx_k_PUNCTUATION_SPACE[] = "<PUNCTUATION_SPACE>";
static const char __pyx_k_ParseResults_copy[] = "ParseResults.copy";
static const char __pyx_k_ParseResults_dump[] = "ParseResults.dump";
static const char __pyx_k_ParseResults_keys[] = "ParseResults.keys";
static const char __pyx_k_PrecededBy___init[] = "PrecededBy.__init__";
static const char __pyx_k_SingleCharLiteral[] = "_SingleCharLiteral";
static const char __pyx_k_Suppress_suppress[] = "Suppress.suppress";
static const char __pyx_k_Unmatchable_token[] = "Unmatchable token";
static const char __pyx_k_WordEnd_parseImpl[] = "WordEnd.parseImpl";
static const char __pyx_k_ZeroOrMore___init[] = "ZeroOrMore.__init__";
static const char __pyx_k_found_end_of_text[] = ", found end of text";
static const char __pyx_k_getrecursionlimit[] = "getrecursionlimit";
static const char __pyx_k_includeSeparators[] = "includeSeparators";
static const char __pyx_k_parseElementTrace[] = "parseElementTrace";
static const char __pyx_k_And_checkRecursion[] = "And.checkRecursion";
static const char __pyx_k_Forward_streamline[] = "Forward.streamline";
static const char __pyx_k_No_match_found_for[] = "No match found for ";
static const char __pyx_k_Optional_parseImpl[] = "Optional.parseImpl";
static const char __pyx_k_Or__setResultsName[] = "Or._setResultsName";
static const char __pyx_k_ParseBaseException[] = "ParseBaseException";
static const char __pyx_k_ParseResults___add[] = "ParseResults.__add__";
static const char __pyx_k_ParseResults___dir[] = "ParseResults.__dir__";
static const char __pyx_k_ParseResults___len[] = "ParseResults.__len__";
static const char __pyx_k_ParseResults___new[] = "ParseResults.__new__";
static const char __pyx_k_ParseResults___str[] = "ParseResults.__str__";
static const char __pyx_k_ParseResults__name[] = "_ParseResults__name";
static const char __pyx_k_ParseResults_asXML[] = "ParseResults.asXML";
static const char __pyx_k_ParseResults_clear[] = "ParseResults.clear";
static const char __pyx_k_ParseResults_items[] = "ParseResults.items";
static const char __pyx_k_ParserElement___eq[] = "ParserElement.__eq__";
static const char __pyx_k_ParserElement___ne[] = "ParserElement.__ne__";
static const char __pyx_k_ParserElement___or[] = "ParserElement.__or__";
static const char __pyx_k_ParserElement_copy[] = "ParserElement.copy";
static const char __pyx_k_QuotedString___str[] = "QuotedString.__str__";
static const char __pyx_k_StringStart___init[] = "StringStart.__init__";
static const char __pyx_k_Suppress_postParse[] = "Suppress.postParse";
static const char __pyx_k_THREE_PER_EM_SPACE[] = "<THREE-PER-EM_SPACE>";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_literalStringClass[] = "_literalStringClass";
static const char __pyx_k_packrat_cache_lock[] = "packrat_cache_lock";
static const char __pyx_k_setWhitespaceChars[] = "setWhitespaceChars";
static const char __pyx_k_verbose_stacktrace[] = "verbose_stacktrace";
static const char __pyx_k_DEFAULT_WHITE_CHARS[] = "DEFAULT_WHITE_CHARS";
static const char __pyx_k_Each_checkRecursion[] = "Each.checkRecursion";
static const char __pyx_k_GoToColumn_preParse[] = "GoToColumn.preParse";
static const char __pyx_k_LineStart_parseImpl[] = "LineStart.parseImpl";
static const char __pyx_k_ParseElementEnhance[] = "ParseElementEnhance";
static const char __pyx_k_ParseFatalException[] = "ParseFatalException";
static const char __pyx_k_ParseResults___bool[] = "ParseResults.__bool__";
static const char __pyx_k_ParseResults___iadd[] = "ParseResults.__iadd__";
static const char __pyx_k_ParseResults___init[] = "ParseResults.__init__";
static const char __pyx_k_ParseResults___iter[] = "ParseResults.__iter__";
static const char __pyx_k_ParseResults___radd[] = "ParseResults.__radd__";
static const char __pyx_k_ParseResults___repr[] = "ParseResults.__repr__";
static const char __pyx_k_ParseResults__modal[] = "_ParseResults__modal";
static const char __pyx_k_ParseResults_append[] = "ParseResults.append";
static const char __pyx_k_ParseResults_asDict[] = "ParseResults.asDict";
static const char __pyx_k_ParseResults_asList[] = "ParseResults.asList";
static const char __pyx_k_ParseResults_extend[] = "ParseResults.extend";
static const char __pyx_k_ParseResults_insert[] = "ParseResults.insert";
static const char __pyx_k_ParseResults_pprint[] = "ParseResults.pprint";
static const char __pyx_k_ParseResults_values[] = "ParseResults.values";
static const char __pyx_k_ParserElement___add[] = "ParserElement.__add__";
static const char __pyx_k_ParserElement___and[] = "ParserElement.__and__";
static const char __pyx_k_ParserElement___mul[] = "ParserElement.__mul__";
static const char __pyx_k_ParserElement___req[] = "ParserElement.__req__";
static const char __pyx_k_ParserElement___rne[] = "ParserElement.__rne__";
static const char __pyx_k_ParserElement___ror[] = "ParserElement.__ror__";
static const char __pyx_k_ParserElement___str[] = "ParserElement.__str__";
static const char __pyx_k_ParserElement___sub[] = "ParserElement.__sub__";
static const char __pyx_k_ParserElement___xor[] = "ParserElement.__xor__";
static const char __pyx_k_ParserElement_split[] = "ParserElement.split";
static const char __pyx_k_QuotedString___init[] = "QuotedString.__init__";
static const char __pyx_k_Regex_sub_locals_pa[] = "Regex.sub.<locals>.pa";
static const char __pyx_k_StringEnd_parseImpl[] = "StringEnd.parseImpl";
static const char __pyx_k_WordRegex_parseImpl[] = "_WordRegex.parseImpl";
static const char __pyx_k_WordStart_parseImpl[] = "WordStart.parseImpl";
static const char __pyx_k_amp_gt_lt_quot_apos[] = "amp gt lt quot apos";
static const char __pyx_k_enable_all_warnings[] = "_enable_all_warnings";
static const char __pyx_k_inlineLiteralsUsing[] = "inlineLiteralsUsing";
static const char __pyx_k_originalParseMethod[] = "_originalParseMethod";
static const char __pyx_k_packrat_cache_stats[] = "packrat_cache_stats";
static const char __pyx_k_CharsNotIn_parseImpl[] = "CharsNotIn.parseImpl";
static const char __pyx_k_CloseMatch_parseImpl[] = "CloseMatch.parseImpl";
static const char __pyx_k_Expected_end_of_line[] = "Expected end of line";
static const char __pyx_k_Expected_end_of_text[] = "Expected end of text";
static const char __pyx_k_FollowedBy_parseImpl[] = "FollowedBy.parseImpl";
static const char __pyx_k_Found_unwanted_token[] = "Found unwanted token, ";
static const char __pyx_k_GoToColumn_parseImpl[] = "GoToColumn.parseImpl";
static const char __pyx_k_MatchFirst_parseImpl[] = "MatchFirst.parseImpl";
static const char __pyx_k_MultipleMatch___init[] = "_MultipleMatch.__init__";
static const char __pyx_k_MultipleMatch_stopOn[] = "_MultipleMatch.stopOn";
static const char __pyx_k_ParseExpression_copy[] = "ParseExpression.copy";
static const char __pyx_k_ParseResults__asList[] = "_ParseResults__asList";
static const char __pyx_k_ParseResults__doinit[] = "_ParseResults__doinit";
static const char __pyx_k_ParseResults__lookup[] = "_ParseResults__lookup";
static const char __pyx_k_ParseResults__parent[] = "_ParseResults__parent";
static const char __pyx_k_ParseResults_getName[] = "ParseResults.getName";
static const char __pyx_k_ParseResults_haskeys[] = "ParseResults.haskeys";
static const char __pyx_k_ParseSyntaxException[] = "ParseSyntaxException";
static const char __pyx_k_ParserElement___call[] = "ParserElement.__call__";
static const char __pyx_k_ParserElement___hash[] = "ParserElement.__hash__";
static const char __pyx_k_ParserElement___init[] = "ParserElement.__init__";
static const char __pyx_k_ParserElement___iter[] = "ParserElement.__iter__";
static const char __pyx_k_ParserElement___radd[] = "ParserElement.__radd__";
static const char __pyx_k_ParserElement___rand[] = "ParserElement.__rand__";
static const char __pyx_k_ParserElement___repr[] = "ParserElement.__repr__";
static const char __pyx_k_ParserElement___rmul[] = "ParserElement.__rmul__";
static const char __pyx_k_ParserElement___rsub[] = "ParserElement.__rsub__";
static const char __pyx_k_ParserElement___rxor[] = "ParserElement.__rxor__";
static const char __pyx_k_ParserElement_ignore[] = "ParserElement.ignore";
static const char __pyx_k_PositionToken___init[] = "_PositionToken.__init__";
static const char __pyx_k_PrecededBy_parseImpl[] = "PrecededBy.parseImpl";
static const char __pyx_k_ZeroOrMore_parseImpl[] = "ZeroOrMore.parseImpl";
static const char __pyx_k_parseImplAsGroupList[] = "parseImplAsGroupList";
static const char __pyx_k_self_skip_ignorables[] = "self_skip_ignorables";
static const char __pyx_k_And__ErrorStop___init[] = "And._ErrorStop.__init__";
static const char __pyx_k_DEFAULT_KEYWORD_CHARS[] = "DEFAULT_KEYWORD_CHARS";
static const char __pyx_k_MatchFirst_streamline[] = "MatchFirst.streamline";
static const char __pyx_k_ParseExpression___str[] = "ParseExpression.__str__";
static const char __pyx_k_ParseResults___lookup[] = "ParseResults.__lookup";
static const char __pyx_k_ParseResults__tokdict[] = "_ParseResults__tokdict";
static const char __pyx_k_ParseResults__toklist[] = "_ParseResults__toklist";
static const char __pyx_k_ParserElement_matches[] = "ParserElement.matches";
static const char __pyx_k_ParserElement_setName[] = "ParserElement.setName";
static const char __pyx_k_StringStart_parseImpl[] = "StringStart.parseImpl";
static const char __pyx_k_TokenConverter___init[] = "TokenConverter.__init__";
static const char __pyx_k_copyDefaultWhiteChars[] = "copyDefaultWhiteChars";
static const char __pyx_k_enable_all_warnings_2[] = "enable_all_warnings";
static const char __pyx_k_escCharReplacePattern[] = "escCharReplacePattern";
static const char __pyx_k_escapeRegexRangeChars[] = "_escapeRegexRangeChars";
static const char __pyx_k_CaselessKeyword___init[] = "CaselessKeyword.__init__";
static const char __pyx_k_CaselessLiteral___init[] = "CaselessLiteral.__init__";
static const char __pyx_k_Expected_start_of_line[] = "Expected start of line";
static const char __pyx_k_Expected_start_of_text[] = "Expected start of text";
static const char __pyx_k_ParseException_explain[] = "ParseException.explain";
static const char __pyx_k_ParseExpression___init[] = "ParseExpression.__init__";
static const char __pyx_k_ParseExpression_append[] = "ParseExpression.append";
static const char __pyx_k_ParseExpression_ignore[] = "ParseExpression.ignore";
static const char __pyx_k_ParseResultsWithOffset[] = "_ParseResultsWithOffset";
static const char __pyx_k_ParseResults___delitem[] = "ParseResults.__delitem__";
static const char __pyx_k_ParseResults___getattr[] = "ParseResults.__getattr__";
static const char __pyx_k_ParseResults___getitem[] = "ParseResults.__getitem__";
static const char __pyx_k_ParseResults___setitem[] = "ParseResults.__setitem__";
static const char __pyx_k_ParseResults__iterkeys[] = "ParseResults._iterkeys";
static const char __pyx_k_ParseResults_from_dict[] = "ParseResults.from_dict";
static const char __pyx_k_ParserElement___invert[] = "ParserElement.__invert__";
static const char __pyx_k_ParserElement_preParse[] = "ParserElement.preParse";
static const char __pyx_k_ParserElement_runTests[] = "ParserElement.runTests";
static const char __pyx_k_ParserElement_setBreak[] = "ParserElement.setBreak";
static const char __pyx_k_ParserElement_setDebug[] = "ParserElement.setDebug";
static const char __pyx_k_ParserElement_suppress[] = "ParserElement.suppress";
static const char __pyx_k_ParserElement_tryParse[] = "ParserElement.tryParse";
static const char __pyx_k_ParserElement_validate[] = "ParserElement.validate";
static const char __pyx_k_QuotedString_parseImpl[] = "QuotedString.parseImpl";
static const char __pyx_k_Regex_parseImplAsMatch[] = "Regex.parseImplAsMatch";
static const char __pyx_k_collect_all_And_tokens[] = "collect_all_And_tokens";
static const char __pyx_k_conditionAsParseAction[] = "conditionAsParseAction";
static const char __pyx_k_setDefaultKeywordChars[] = "setDefaultKeywordChars";
static const char __pyx_k_Forward__setResultsName[] = "Forward._setResultsName";
static const char __pyx_k_Forward_leaveWhitespace[] = "Forward.leaveWhitespace";
static const char __pyx_k_MultipleMatch_parseImpl[] = "_MultipleMatch.parseImpl";
static const char __pyx_k_ParseResults___contains[] = "ParseResults.__contains__";
static const char __pyx_k_ParseResults___getstate[] = "ParseResults.__getstate__";
static const char __pyx_k_ParseResults___reversed[] = "ParseResults.__reversed__";
static const char __pyx_k_ParseResults___setstate[] = "ParseResults.__setstate__";
static const char __pyx_k_ParseResults__iteritems[] = "ParseResults._iteritems";
static const char __pyx_k_ParserElement___getitem[] = "ParserElement.__getitem__";
static const char __pyx_k_ParserElement_parseFile[] = "ParserElement.parseFile";
static const char __pyx_k_ParserElement_parseImpl[] = "ParserElement.parseImpl";
static const char __pyx_k_ParserElement_postParse[] = "ParserElement.postParse";
static const char __pyx_k_defaultStartDebugAction[] = "_defaultStartDebugAction";
static const char __pyx_k_Not_at_the_end_of_a_word[] = "Not at the end of a word";
static const char __pyx_k_ParseBaseException___dir[] = "ParseBaseException.__dir__";
static const char __pyx_k_ParseBaseException___str[] = "ParseBaseException.__str__";
static const char __pyx_k_ParseExpression_validate[] = "ParseExpression.validate";
static const char __pyx_k_ParseResults__accumNames[] = "_ParseResults__accumNames";
static const char __pyx_k_ParseResults__itervalues[] = "ParseResults._itervalues";
static const char __pyx_k_ParserElement__FifoCache[] = "ParserElement._FifoCache";
static const char __pyx_k_ParserElement_resetCache[] = "ParserElement.resetCache";
static const char __pyx_k_ParserElement_scanString[] = "ParserElement.scanString";
static const char __pyx_k_ParserElement_streamline[] = "ParserElement.streamline";
static const char __pyx_k_convertWhitespaceEscapes[] = "convertWhitespaceEscapes";
static const char __pyx_k_r_object_is_not_iterable[] = "%r object is not iterable";
static const char __pyx_k_self_failOn_canParseNext[] = "self_failOn_canParseNext";
static const char __pyx_k_self_ignoreExpr_tryParse[] = "self_ignoreExpr_tryParse";
static const char __pyx_k_CaselessLiteral_parseImpl[] = "CaselessLiteral.parseImpl";
static const char __pyx_k_MONGOLIAN_VOWEL_SEPARATOR[] = "<MONGOLIAN_VOWEL_SEPARATOR>";
static const char __pyx_k_MatchFirst_checkRecursion[] = "MatchFirst.checkRecursion";
static const char __pyx_k_Or___str___locals_genexpr[] = "Or.__str__.<locals>.genexpr";
static const char __pyx_k_ParseBaseException___init[] = "ParseBaseException.__init__";
static const char __pyx_k_ParseBaseException___repr[] = "ParseBaseException.__repr__";
static const char __pyx_k_ParseElementEnhance___str[] = "ParseElementEnhance.__str__";
static const char __pyx_k_ParseResults___getnewargs[] = "ParseResults.__getnewargs__";
static const char __pyx_k_ParserElement__parseCache[] = "ParserElement._parseCache";
static const char __pyx_k_ParserElement_parseString[] = "ParserElement.parseString";
static const char __pyx_k_RecursiveGrammarException[] = "RecursiveGrammarException";
static const char __pyx_k_defaultSuccessDebugAction[] = "_defaultSuccessDebugAction";
static const char __pyx_k_setDefaultWhitespaceChars[] = "setDefaultWhitespaceChars";
static const char __pyx_k_xml_escape_locals_genexpr[] = "_xml_escape.<locals>.genexpr";
static const char __pyx_k_And___str___locals_genexpr[] = "And.__str__.<locals>.genexpr";
static const char __pyx_k_MatchFirst__setResultsName[] = "MatchFirst._setResultsName";
static const char __pyx_k_Not_at_the_start_of_a_word[] = "Not at the start of a word";
static const char __pyx_k_Or___init___locals_genexpr[] = "Or.__init__.<locals>.genexpr";
static const char __pyx_k_ParseElementEnhance___init[] = "ParseElementEnhance.__init__";
static const char __pyx_k_ParseElementEnhance_ignore[] = "ParseElementEnhance.ignore";
static const char __pyx_k_ParseExpression_streamline[] = "ParseExpression.streamline";
static const char __pyx_k_ParseResults__asStringList[] = "ParseResults._asStringList";
static const char __pyx_k_ParserElement_addCondition[] = "ParserElement.addCondition";
static const char __pyx_k_ParserElement_canParseNext[] = "ParserElement.canParseNext";
static const char __pyx_k_ParserElement_searchString[] = "ParserElement.searchString";
static const char __pyx_k_Regex_parseImplAsGroupList[] = "Regex.parseImplAsGroupList";
static const char __pyx_k_s_s_at_char_d_line_d_col_d[] = "%s%s  (at char %d), (line:%d, col:%d)";
static const char __pyx_k_And___init___locals_genexpr[] = "And.__init__.<locals>.genexpr";
static const char __pyx_k_Each___str___locals_genexpr[] = "Each.__str__.<locals>.genexpr";
static const char __pyx_k_ParserElement__parseNoCache[] = "ParserElement._parseNoCache";
static const char __pyx_k_ParserElement_enablePackrat[] = "ParserElement.enablePackrat";
static const char __pyx_k_ParserElement_parseWithTabs[] = "ParserElement.parseWithTabs";
static const char __pyx_k_ParserElement_setFailAction[] = "ParserElement.setFailAction";
static const char __pyx_k_RecursiveGrammarException_s[] = "RecursiveGrammarException: %s";
static const char __pyx_k_SingleCharLiteral_parseImpl[] = "_SingleCharLiteral.parseImpl";
static const char __pyx_k_Text_not_in_expected_column[] = "Text not in expected column";
static const char __pyx_k_defaultExceptionDebugAction[] = "_defaultExceptionDebugAction";
static const char __pyx_k_Each___init___locals_genexpr[] = "Each.__init__.<locals>.genexpr";
static const char __pyx_k_Optional__optionalNotMatched[] = "_Optional__optionalNotMatched";
static const char __pyx_k_Or_streamline_locals_genexpr[] = "Or.streamline.<locals>.genexpr";
static const char __pyx_k_ParseBaseException___getattr[] = "ParseBaseException.__getattr__";
static const char __pyx_k_ParseElementEnhance_validate[] = "ParseElementEnhance.validate";
static const char __pyx_k_ParserElement_addParseAction[] = "ParserElement.addParseAction";
static const char __pyx_k_ParserElement_checkRecursion[] = "ParserElement.checkRecursion";
static const char __pyx_k_ParserElement_setParseAction[] = "ParserElement.setParseAction";
static const char __pyx_k_ParserElement_setResultsName[] = "ParserElement.setResultsName";
static const char __pyx_k_Word___init___locals_genexpr[] = "Word.__init__.<locals>.genexpr";
static const char __pyx_k_A_token_that_will_never_match[] = "A token that will never match.\n    ";
static const char __pyx_k_And_streamline_locals_genexpr[] = "And.streamline.<locals>.genexpr";
static const char __pyx_k_Each_parseImpl_locals_genexpr[] = "Each.parseImpl.<locals>.genexpr";
static const char __pyx_k_MultipleMatch__setResultsName[] = "_MultipleMatch._setResultsName";
static const char __pyx_k_ParseElementEnhance_parseImpl[] = "ParseElementEnhance.parseImpl";
static const char __pyx_k_ParseResultsWithOffset___init[] = "_ParseResultsWithOffset.__init__";
static const char __pyx_k_ParseResultsWithOffset___repr[] = "_ParseResultsWithOffset.__repr__";
static const char __pyx_k_ParserElement__UnboundedCache[] = "ParserElement._UnboundedCache.__init__.<locals>.get";
static const char __pyx_k_ParserElement__setResultsName[] = "ParserElement._setResultsName";
static const char __pyx_k_ParserElement__skipIgnorables[] = "ParserElement._skipIgnorables";
static const char __pyx_k_ParserElement__trim_traceback[] = "ParserElement._trim_traceback";
static const char __pyx_k_ParserElement_leaveWhitespace[] = "ParserElement.leaveWhitespace";
static const char __pyx_k_ParserElement_setDebugActions[] = "ParserElement.setDebugActions";
static const char __pyx_k_ParserElement_transformString[] = "ParserElement.transformString";
static const char __pyx_k_White___init___locals_genexpr[] = "White.__init__.<locals>.genexpr";
static const char __pyx_k_Each_streamline_locals_genexpr[] = "Each.streamline.<locals>.genexpr";
static const char __pyx_k_Keyword_setDefaultKeywordChars[] = "Keyword.setDefaultKeywordChars";
static const char __pyx_k_ParseElementEnhance_streamline[] = "ParseElementEnhance.streamline";
static const char __pyx_k_Word___str___locals_charsAsStr[] = "Word.__str__.<locals>.charsAsStr";
static const char __pyx_k_parse_actions_must_be_callable[] = "parse actions must be callable";
static const char __pyx_k_warn_name_set_on_empty_Forward[] = "warn_name_set_on_empty_Forward";
static const char __pyx_k_0_setting_results_name_0_r_on_1[] = "{0}: setting results name {0!r} on {1} expression that has no contained expression";
static const char __pyx_k_0_setting_results_name_1_r_on_2[] = "{0}: setting results name {1!r} on {2} expression collides with {3!r} on contained expression";
static const char __pyx_k_Abstract_subclass_of_class_Pars[] = "\n    Abstract subclass of :class:`ParseExpression`, for converting parsed results.\n    ";
static const char __pyx_k_Caseless_version_of_class_Keywo[] = "\n    Caseless version of :class:`Keyword`.\n\n    Example::\n\n        OneOrMore(CaselessKeyword(\"CMD\")).parseString(\"cmd CMD Cmd10\") # -> ['CMD', 'CMD']\n\n    (Contrast with example for :class:`CaselessLiteral`.)\n    ";
static const char __pyx_k_Converter_to_return_the_matched[] = "Converter to return the matched tokens as a list - useful for\n    returning tokens of :class:`ZeroOrMore` and :class:`OneOrMore` expressions.\n\n    Example::\n\n        ident = Word(alphas)\n        num = Word(nums)\n        term = ident | num\n        func = ident + Optional(delimitedList(term))\n        print(func.parseString(\"fn a, b, 100\"))  # -> ['fn', 'a', 'b', '100']\n\n        func = ident + Group(Optional(delimitedList(term)))\n        print(func.parseString(\"fn a, b, 100\"))  # -> ['fn', ['a', 'b', '100']]\n    ";
static const char __pyx_k_Diagnostic_configuration_all_de[] = "\nDiagnostic configuration (all default to False)\n     - warn_multiple_tokens_in_named_alternation - flag to enable warnings when a results\n       name is defined on a MatchFirst or Or expression with one or more And subexpressions\n       (only warns if __diag__.collect_all_And_tokens is False)\n     - warn_ungrouped_named_tokens_in_collection - flag to enable warnings when a results\n       name is defined on a containing expression with ungrouped subexpressions that also\n       have results names\n     - warn_name_set_on_empty_Forward - flag to enable warnings whan a Forward is defined\n       with a results name, but has no contents defined\n     - warn_on_multiple_string_args_to_oneof - flag to enable warnings whan oneOf is\n       incorrectly called with multiple str arguments\n     - enable_debug_on_named_expressions - flag to auto-enable debug on all subsequent\n       calls to ParserElement.setName()\n";
static const char __pyx_k_Exception_thrown_when_parse_exp[] = "\n    Exception thrown when parse expressions don't match class;\n    supported attributes by name are:\n    - lineno - returns the line number of the exception text\n    - col - returns the column number of the exception text\n    - line - returns the line containing the exception text\n\n    Example::\n\n        try:\n            Word(nums).setName(\"integer\").parseString(\"ABC\")\n        except ParseException as pe:\n            print(pe)\n            print(\"column: {}\".format(pe.col))\n\n    prints::\n\n       Expected integer (at char 0), (line:1, col:1)\n        column: 1\n\n    ";
static const char __pyx_k_Lookahead_matching_of_the_given[] = "Lookahead matching of the given parse expression.\n    ``FollowedBy`` does *not* advance the parsing position within\n    the input string, it only verifies that the specified parse\n    expression matches at the current position.  ``FollowedBy``\n    always returns a null token list. If any results names are defined\n    in the lookahead expression, those *will* be returned for access by\n    name.\n\n    Example::\n\n        # use FollowedBy to match a label only if it is followed by a ':'\n        data_word = Word(alphas)\n        label = data_word + FollowedBy(':')\n        attr_expr = Group(label + Suppress(':') + OneOrMore(data_word, stopOn=label).setParseAction(' '.join))\n\n        OneOrMore(attr_expr).parseString(\"shape: SQUARE color: BLACK posn: upper left\").pprint()\n\n    prints::\n\n        [['shape', 'SQUARE'], ['color', 'BLACK'], ['posn', 'upper left']]\n    ";
static const char __pyx_k_Matches_if_the_current_position[] = "Matches if the current position is at the beginning of a Word,\n    and is not preceded by any character in a given set of\n    ``wordChars`` (default= ``printables``). To emulate the\n    ``\010`` behavior of regular expressions, use\n    ``WordStart(alphanums)``. ``WordStart`` will also match at\n    the beginning of the string being parsed, or at the beginning of\n    a line.\n    ";
static const char __pyx_k_ParseExpression___init___locals[] = "ParseExpression.__init__.<locals>.genexpr";
static const char __pyx_k_ParseExpression__setResultsName[] = "ParseExpression._setResultsName";
static const char __pyx_k_ParseExpression_leaveWhitespace[] = "ParseExpression.leaveWhitespace";
static const char __pyx_k_ParseResultsWithOffset___getite[] = "_ParseResultsWithOffset.__getitem__";
static const char __pyx_k_ParseResultsWithOffset_setOffse[] = "_ParseResultsWithOffset.setOffset";
static const char __pyx_k_ParseResults__itervalues_locals[] = "ParseResults._itervalues.<locals>.genexpr";
static const char __pyx_k_ParserElement__FifoCache___init[] = "ParserElement._FifoCache.__init__.<locals>.get";
static const char __pyx_k_ParserElement__UnboundedCache_2[] = "ParserElement._UnboundedCache.__init__.<locals>.set";
static const char __pyx_k_ParserElement__UnboundedCache_3[] = "ParserElement._UnboundedCache.__init__.<locals>.clear";
static const char __pyx_k_ParserElement__UnboundedCache_4[] = "ParserElement._UnboundedCache.__init__.<locals>.cache_len";
static const char __pyx_k_ParserElement__UnboundedCache_5[] = "ParserElement._UnboundedCache";
static const char __pyx_k_ParserElement__UnboundedCache_6[] = "ParserElement._UnboundedCache.__init__";
static const char __pyx_k_RecursiveGrammarException___str[] = "RecursiveGrammarException.__str__";
static const char __pyx_k_Requires_all_given_class_ParseE[] = "\n    Requires all given :class:`ParseExpression` s to be found in the given order.\n    Expressions may be separated by whitespace.\n    May be constructed using the ``'+'`` operator.\n    May also be constructed using the ``'-'`` operator, which will\n    suppress backtracking.\n\n    Example::\n\n        integer = Word(nums)\n        name_expr = OneOrMore(Word(alphas))\n\n        expr = And([integer(\"id\"), name_expr(\"name\"), integer(\"age\")])\n        # more easily written as:\n        expr = integer(\"id\") + name_expr(\"name\") + integer(\"age\")\n    ";
static const char __pyx_k_Token_for_matching_strings_that[] = "Token for matching strings that match a given regular\n    expression. Defined with string specifying the regular expression in\n    a form recognized by the stdlib Python  `re module <https://docs.python.org/3/library/re.html>`_.\n    If the given regex contains named groups (defined using ``(?P<name>...)``),\n    these will be preserved as named parse results.\n\n    If instead of the Python stdlib re module you wish to use a different RE module\n    (such as the `regex` module), you can replace it by either building your\n    Regex object with a compiled RE that was compiled using regex:\n\n    Example::\n\n        realnum = Regex(r\"[+-]?\\d+\\.\\d*\")\n        date = Regex(r'(?P<year>\\d{4})-(?P<month>\\d\\d?)-(?P<day>\\d\\d?)')\n        # ref: https://stackoverflow.com/questions/267399/how-do-you-match-only-valid-roman-numerals-with-a-regular-expression\n        roman = Regex(r\"M{0,4}(CM|CD|D?{0,3})(XC|XL|L?X{0,3})(IX|IV|V?I{0,3})\")\n\n        # use regex module instead of stdlib re module to construct a Regex using\n        # a compiled regular expression\n        import regex\n        parser = pp.Regex(regex.compile(r'[0-9]'))\n\n    ";
static const char __pyx_k_cannot_specify_a_minimum_length[] = "cannot specify a minimum length < 1; use Optional(Word()) if zero-length word is permitted";
static const char __pyx_k_null_string_passed_to_Regex_use[] = "null string passed to Regex; use Empty() instead";
static const char __pyx_k_user_throwable_exception_thrown[] = "user-throwable exception thrown when inconsistent parse content\n       is found; stops all parsing immediately";
static const char __pyx_k_warn_on_multiple_string_args_to[] = "warn_on_multiple_string_args_to_oneof";
static const char __pyx_k_A_short_cut_class_for_defining_W[] = "A short-cut class for defining ``Word(characters, exact=1)``,\n    when defining a match of any single character in a string of\n    characters.\n    ";
static const char __pyx_k_A_variation_on_class_Literal_whi[] = "A variation on :class:`Literal` which matches \"close\" matches,\n    that is, strings with at most 'n' mismatching characters.\n    :class:`CloseMatch` takes parameters:\n\n     - ``match_string`` - string to be matched\n     - ``maxMismatches`` - (``default=1``) maximum number of\n       mismatches allowed to count as a match\n\n    The results from a successful parse will contain the matched text\n    from the input string and the following named results:\n\n     - ``mismatches`` - a list of the positions within the\n       match_string where mismatches were found\n     - ``original`` - the original match_string used to compare\n       against the input string\n\n    If ``mismatches`` is an empty list, then the match was an exact\n    match.\n\n    Example::\n\n        patt = CloseMatch(\"ATCATCGAATGGA\")\n        patt.parseString(\"ATCATCGAAXGGA\") # -> (['ATCATCGAAXGGA'], {'mismatches': [[9]], 'original': ['ATCATCGAATGGA']})\n        patt.parseString(\"ATCAXCGAAXGGA\") # -> Exception: Expected 'ATCATCGAATGGA' (with up to 1 mismatches) (at char 0), (line:1, col:1)\n\n        # exact match\n        patt.parseString(\"ATCATCGAATGGA\") # -> (['ATCATCGAATGGA'], {'mismatches': [[]], 'original': ['ATCATCGAATGGA']})\n\n        # close match allowing up to 2 mismatches\n        patt = CloseMatch(\"ATCATCGAATGGA\", maxMismatches=2)\n        patt.parseString(\"ATCAXCGAAXGGA\") # -> (['ATCAXCGAAXGGA'], {'mismatches': [[4, 9]], 'original': ['ATCATCGAATGGA']})\n    ";
static const char __pyx_k_Abstract_base_level_parser_eleme[] = "Abstract base level parser element class.";
static const char __pyx_k_Abstract_class_ParserElement_sub[] = "Abstract :class:`ParserElement` subclass, for defining atomic\n    matching patterns.\n    ";
static const char __pyx_k_Abstract_subclass_of_ParserEleme[] = "Abstract subclass of ParserElement, for combining and\n    post-processing parsed tokens.\n    ";
static const char __pyx_k_Abstract_subclass_of_class_Parse[] = "Abstract subclass of :class:`ParserElement`, for combining and\n    post-processing parsed tokens.\n    ";
static const char __pyx_k_An_empty_token_will_always_match[] = "An empty token, will always match.\n    ";
static const char __pyx_k_Aspidites__vendor_pyparsing_exte[] = "Aspidites._vendor.pyparsing_extension";
static const char __pyx_k_Cannot_combine_element_of_type_s[] = "Cannot combine element of type %s with ParserElement";
static const char __pyx_k_Converter_for_ignoring_the_resul[] = "Converter for ignoring the results of a parsed expression.\n\n    Example::\n\n        source = \"a, b, c,d\"\n        wd = Word(alphas)\n        wd_list1 = wd + ZeroOrMore(',' + wd)\n        print(wd_list1.parseString(source))\n\n        # often, delimiters that are useful during parsing are just in the\n        # way afterward - use Suppress to keep them out of the parsed output\n        wd_list2 = wd + ZeroOrMore(Suppress(',') + wd)\n        print(wd_list2.parseString(source))\n\n    prints::\n\n        ['a', ',', 'b', ',', 'c', ',', 'd']\n        ['a', 'b', 'c', 'd']\n\n    (See also :class:`delimitedList`.)\n    ";
static const char __pyx_k_Converter_to_concatenate_all_mat[] = "Converter to concatenate all matching tokens to a single string.\n    By default, the matching patterns must also be contiguous in the\n    input string; this can be disabled by specifying\n    ``'adjacent=False'`` in the constructor.\n\n    Example::\n\n        real = Word(nums) + '.' + Word(nums)\n        print(real.parseString('3.1416')) # -> ['3', '.', '1416']\n        # will also erroneously match the following\n        print(real.parseString('3. 1416')) # -> ['3', '.', '1416']\n\n        real = Combine(Word(nums) + '.' + Word(nums))\n        print(real.parseString('3.1416')) # -> ['3.1416']\n        # no match when there are internal spaces\n        print(real.parseString('3. 1416')) # -> Exception: Expected W:(0123...)\n    ";
static const char __pyx_k_Converter_to_return_a_repetitive[] = "Converter to return a repetitive expression as a list, but also\n    as a dictionary. Each element can also be referenced using the first\n    token in the expression as its key. Useful for tabular report\n    scraping when the first column can be used as a item key.\n\n    Example::\n\n        data_word = Word(alphas)\n        label = data_word + FollowedBy(':')\n        attr_expr = Group(label + Suppress(':') + OneOrMore(data_word).setParseAction(' '.join))\n\n        text = \"shape: SQUARE posn: upper left color: light blue texture: burlap\"\n        attr_expr = (label + Suppress(':') + OneOrMore(data_word, stopOn=label).setParseAction(' '.join))\n\n        # print attributes as plain groups\n        print(OneOrMore(attr_expr).parseString(text).dump())\n\n        # instead of OneOrMore(expr), parse using Dict(OneOrMore(Group(expr))) - Dict will auto-assign names\n        result = Dict(OneOrMore(Group(attr_expr))).parseString(text)\n        print(result.dump())\n\n        # access named fields as dict entries, or output as dict\n        print(result['shape'])\n        print(result.asDict())\n\n    prints::\n\n        ['shape', 'SQUARE', 'posn', 'upper left', 'color', 'light blue', 'texture', 'burlap']\n        [['shape', 'SQUARE'], ['posn', 'upper left'], ['color', 'light blue'], ['texture', 'burlap']]\n        - color: light blue\n        - posn: upper left\n        - shape: SQUARE\n        - texture: burlap\n        SQUARE\n        {'color': 'light blue', 'posn': 'upper left', 'texture': 'burlap', 'shape': 'SQUARE'}\n\n    See more examples at :class:`ParseResults` of accessing fields by results name.\n    ";
static const char __pyx_k_Expected_r_with_up_to_d_mismatch[] = "Expected %r (with up to %d mismatches)";
static const char __pyx_k_Forward_declaration_of_an_expres[] = "Forward declaration of an expression to be defined later -\n    used for recursive grammars, such as algebraic infix notation.\n    When the expression is known, it is assigned to the ``Forward``\n    variable using the '<<' operator.\n\n    Note: take care when assigning to ``Forward`` not to overlook\n    precedence of operators.\n\n    Specifically, '|' has a lower precedence than '<<', so that::\n\n        fwdExpr << a | b | c\n\n    will actually be evaluated as::\n\n        (fwdExpr << a) | b | c\n\n    thereby leaving b and c out as parseable alternatives.  It is recommended that you\n    explicitly group the values inserted into the ``Forward``::\n\n        fwdExpr << (a | b | c)\n\n    Converting to use the '<<=' operator instead will avoid this problem.\n\n    See :class:`ParseResults.pprint` for an example of a recursive\n    parser created using ``Forward``.\n    ";
static const char __pyx_k_Lookahead_to_disallow_matching_w[] = "Lookahead to disallow matching with the given parse expression.\n    ``NotAny`` does *not* advance the parsing position within the\n    input string, it only verifies that the specified parse expression\n    does *not* match at the current position.  Also, ``NotAny`` does\n    *not* skip over leading whitespace. ``NotAny`` always returns\n    a null token list.  May be constructed using the '~' operator.\n\n    Example::\n\n        AND, OR, NOT = map(CaselessKeyword, \"AND OR NOT\".split())\n\n        # take care not to mistake keywords for identifiers\n        ident = ~(AND | OR | NOT) + Word(alphas)\n        boolean_term = Optional(NOT) + ident\n\n        # very crude boolean expression - to support parenthesis groups and\n        # operation hierarchy, use infixNotation\n        boolean_expr = boolean_term + ZeroOrMore((AND | OR) + boolean_term)\n\n        # integers that are followed by \".\" are actually floats\n        integer = Word(nums) + ~Char(\".\")\n    ";
static const char __pyx_k_Lookbehind_matching_of_the_given[] = "Lookbehind matching of the given parse expression.\n    ``PrecededBy`` does not advance the parsing position within the\n    input string, it only verifies that the specified parse expression\n    matches prior to the current position.  ``PrecededBy`` always\n    returns a null token list, but if a results name is defined on the\n    given expression, it is returned.\n\n    Parameters:\n\n     - expr - expression that must match prior to the current parse\n       location\n     - retreat - (default= ``None``) - (int) maximum number of characters\n       to lookbehind prior to the current parse location\n\n    If the lookbehind expression is a string, Literal, Keyword, or\n    a Word or CharsNotIn with a specified exact or maximum length, then\n    the retreat parameter is not required. Otherwise, retreat must be\n    specified to give a maximum number of characters to look back from\n    the current parse position for a lookbehind match.\n\n    Example::\n\n        # VB-style variable names with type prefixes\n        int_var = PrecededBy(\"#\") + pyparsing_common.identifier\n        str_var = PrecededBy(\"$\") + pyparsing_common.identifier\n\n    ";
static const char __pyx_k_MatchFirst___init___locals_genex[] = "MatchFirst.__init__.<locals>.genexpr";
static const char __pyx_k_MatchFirst___str___locals_genexp[] = "MatchFirst.__str__.<locals>.genexpr";
static const char __pyx_k_MatchFirst__setResultsName_local[] = "MatchFirst._setResultsName.<locals>.genexpr";
static const char __pyx_k_MatchFirst_streamline_locals_gen[] = "MatchFirst.streamline.<locals>.genexpr";
static const char __pyx_k_Matches_if_current_position_is_a[] = "Matches if current position is at the beginning of a line within\n    the parse string\n\n    Example::\n\n        test = '''\\\n        AAA this line\n        AAA and this line\n          AAA but not this one\n        B AAA and definitely not this one\n        '''\n\n        for t in (LineStart() + 'AAA' + restOfLine).searchString(test):\n            print(t)\n\n    prints::\n\n        ['AAA', ' this line']\n        ['AAA', ' and this line']\n\n    ";
static const char __pyx_k_Missing_one_or_more_required_ele[] = "Missing one or more required elements (%s)";
static const char __pyx_k_Optional_matching_of_the_given_e[] = "Optional matching of the given expression.\n\n    Parameters:\n     - expr - expression that must match zero or more times\n     - default (optional) - value to be returned if the optional expression is not found.\n\n    Example::\n\n        # US postal code can be a 5-digit zip, plus optional 4-digit qualifier\n        zip = Combine(Word(nums, exact=5) + Optional('-' + Word(nums, exact=4)))\n        zip.runTests('''\n            # traditional ZIP code\n            12345\n\n            # ZIP+4 form\n            12101-0001\n\n            # invalid ZIP\n            98765-\n            ''')\n\n    prints::\n\n        # traditional ZIP code\n        12345\n        ['12345']\n\n        # ZIP+4 form\n        12101-0001\n        ['12101-0001']\n\n        # invalid ZIP\n        98765-\n             ^\n        FAIL: Expected end of text (at char 5), (line:1, col:6)\n    ";
static const char __pyx_k_Optional_repetition_of_zero_or_m[] = "Optional repetition of zero or more of the given expression.\n\n    Parameters:\n     - expr - expression that must match zero or more times\n     - stopOn - (default= ``None``) - expression for a terminating sentinel\n          (only required if the sentinel would ordinarily match the repetition\n          expression)\n\n    Example: similar to :class:`OneOrMore`\n    ";
static const char __pyx_k_Or__setResultsName_locals_genexp[] = "Or._setResultsName.<locals>.genexpr";
static const char __pyx_k_ParseBaseException__from_excepti[] = "ParseBaseException._from_exception";
static const char __pyx_k_ParseBaseException_markInputline[] = "ParseBaseException.markInputline";
static const char __pyx_k_ParseElementEnhance_checkRecursi[] = "ParseElementEnhance.checkRecursion";
static const char __pyx_k_ParseElementEnhance_leaveWhitesp[] = "ParseElementEnhance.leaveWhitespace";
static const char __pyx_k_ParseResults___iadd___locals_lam[] = "ParseResults.__iadd__.<locals>.<lambda>";
static const char __pyx_k_ParseResults___str___locals_gene[] = "ParseResults.__str__.<locals>.genexpr";
static const char __pyx_k_ParseResults__iteritems_locals_g[] = "ParseResults._iteritems.<locals>.genexpr";
static const char __pyx_k_ParseResults_asDict_locals_genex[] = "ParseResults.asDict.<locals>.genexpr";
static const char __pyx_k_ParseResults_asDict_locals_toIte[] = "ParseResults.asDict.<locals>.toItem";
static const char __pyx_k_ParseResults_asXML_locals_genexp[] = "ParseResults.asXML.<locals>.genexpr";
static const char __pyx_k_ParseResults_dump_locals_genexpr[] = "ParseResults.dump.<locals>.genexpr";
static const char __pyx_k_ParseResults_from_dict_locals_is[] = "ParseResults.from_dict.<locals>.is_iterable";
static const char __pyx_k_ParserElement___mul___locals_gen[] = "ParserElement.__mul__.<locals>.genexpr";
static const char __pyx_k_ParserElement___mul___locals_mak[] = "ParserElement.__mul__.<locals>.makeOptionalList";
static const char __pyx_k_ParserElement_inlineLiteralsUsin[] = "ParserElement.inlineLiteralsUsing";
static const char __pyx_k_ParserElement_setBreak_locals_br[] = "ParserElement.setBreak.<locals>.breaker";
static const char __pyx_k_ParserElement_setDefaultWhitespa[] = "ParserElement.setDefaultWhitespaceChars";
static const char __pyx_k_ParserElement_setParseAction_loc[] = "ParserElement.setParseAction.<locals>.genexpr";
static const char __pyx_k_ParserElement_setWhitespaceChars[] = "ParserElement.setWhitespaceChars";
static const char __pyx_k_PrecededBy___init___locals_lambd[] = "PrecededBy.__init__.<locals>.<lambda>";
static const char __pyx_k_QuotedString___init___locals_gen[] = "QuotedString.__init__.<locals>.genexpr";
static const char __pyx_k_RecursiveGrammarException___init[] = "RecursiveGrammarException.__init__";
static const char __pyx_k_Regex_may_only_be_constructed_wi[] = "Regex may only be constructed with a string or a compiled RE object";
static const char __pyx_k_Repetition_of_one_or_more_of_the[] = "Repetition of one or more of the given expression.\n\n    Parameters:\n     - expr - expression that must match one or more times\n     - stopOn - (default= ``None``) - expression for a terminating sentinel\n          (only required if the sentinel would ordinarily match the repetition\n          expression)\n\n    Example::\n\n        data_word = Word(alphas)\n        label = data_word + FollowedBy(':')\n        attr_expr = Group(label + Suppress(':') + OneOrMore(data_word).setParseAction(' '.join))\n\n        text = \"shape: SQUARE posn: upper left color: BLACK\"\n        OneOrMore(attr_expr).parseString(text).pprint()  # Fail! read 'color' as data instead of next label -> [['shape', 'SQUARE color']]\n\n        # use stopOn attribute for OneOrMore to avoid reading label string as part of the data\n        attr_expr = Group(label + Suppress(':') + OneOrMore(data_word, stopOn=label).setParseAction(' '.join))\n        OneOrMore(attr_expr).parseString(text).pprint() # Better -> [['shape', 'SQUARE'], ['posn', 'upper left'], ['color', 'BLACK']]\n\n        # could also be written as\n        (attr_expr * (1,)).parseString(text).pprint()\n    ";
static const char __pyx_k_Requires_all_given_class_ParseEx[] = "Requires all given :class:`ParseExpression` s to be found, but in\n    any order. Expressions may be separated by whitespace.\n\n    May be constructed using the ``'&'`` operator.\n\n    Example::\n\n        color = oneOf(\"RED ORANGE YELLOW GREEN BLUE PURPLE BLACK WHITE BROWN\")\n        shape_type = oneOf(\"SQUARE CIRCLE TRIANGLE STAR HEXAGON OCTAGON\")\n        integer = Word(nums)\n        shape_attr = \"shape:\" + shape_type(\"shape\")\n        posn_attr = \"posn:\" + Group(integer(\"x\") + ',' + integer(\"y\"))(\"posn\")\n        color_attr = \"color:\" + color(\"color\")\n        size_attr = \"size:\" + integer(\"size\")\n\n        # use Each (using operator '&') to accept attributes in any order\n        # (shape and posn are required, color and size are optional)\n        shape_spec = shape_attr & posn_attr & Optional(color_attr) & Optional(size_attr)\n\n        shape_spec.runTests('''\n            shape: SQUARE color: BLACK posn: 100, 120\n            shape: CIRCLE size: 50 color: BLUE posn: 50,80\n            color:GREEN size:20 shape:TRIANGLE posn:20,40\n            '''\n            )\n\n    prints::\n\n        shape: SQUARE color: BLACK posn: 100, 120\n        ['shape:', 'SQUARE', 'color:', 'BLACK', 'posn:', ['100', ',', '120']]\n        - color: BLACK\n        - posn: ['100', ',', '120']\n          - x: 100\n          - y: 120\n        - shape: SQUARE\n\n\n        shape: CIRCLE size: 50 color: BLUE posn: 50,80\n        ['shape:', 'CIRCLE', 'size:', '50', 'color:', 'BLUE', 'posn:', ['50', ',', '80']]\n        - color: BLUE\n        - posn: ['50', ',', '80']\n          - x: 50\n          - y: 80\n        - shape: CIRCLE\n        - size: 50\n\n\n        color: GREEN size: 20 shape: TRIANGLE posn: 20,40\n        ['color:', 'GREEN', 'size:', '20', 'shape:', 'TRIANGLE', 'posn:', ['20', ',', '40']]\n        - color: GREEN\n        - posn: ['20', ',', '40']\n          - x: 20\n          - y: 40\n        - shape: TRIANGLE\n        - size: 20\n    ";
static const char __pyx_k_Requires_that_at_least_one_class[] = "Requires that at least one :class:`ParseExpression` is found. If\n    two expressions match, the expression that matches the longest\n    string will be used. May be constructed using the ``'^'``\n    operator.\n\n    Example::\n\n        # construct Or using '^' operator\n\n        number = Word(nums) ^ Combine(Word(nums) + '.' + Word(nums))\n        print(number.searchString(\"123 3.1416 789\"))\n\n    prints::\n\n        [['123'], ['3.1416'], ['789']]\n    ";
static const char __pyx_k_Special_matching_class_for_match[] = "Special matching class for matching whitespace.  Normally,\n    whitespace is ignored by pyparsing grammars.  This class is included\n    when some whitespace structures are significant.  Define with\n    a string containing the whitespace characters to be matched; default\n    is ``\" \\t\\r\\n\"``.  Also takes optional ``min``,\n    ``max``, and ``exact`` arguments, as defined for the\n    :class:`Word` class.\n    ";
static const char __pyx_k_Structured_parse_results_to_prov[] = "Structured parse results, to provide multiple means of access to\n    the parsed data:\n\n       - as a list (``len(results)``)\n       - by list index (``results[0], results[1]``, etc.)\n       - by attribute (``results.<resultsName>`` - see :class:`ParserElement.setResultsName`)\n\n    Example::\n\n        integer = Word(nums)\n        date_str = (integer.setResultsName(\"year\") + '/'\n                        + integer.setResultsName(\"month\") + '/'\n                        + integer.setResultsName(\"day\"))\n        # equivalent form:\n        # date_str = integer(\"year\") + '/' + integer(\"month\") + '/' + integer(\"day\")\n\n        # parseString returns a ParseResults object\n        result = date_str.parseString(\"1999/12/31\")\n\n        def test(s, fn=repr):\n            print(\"%s -> %s\" % (s, fn(eval(s))))\n        test(\"list(result)\")\n        test(\"result[0]\")\n        test(\"result['month']\")\n        test(\"result.day\")\n        test(\"'month' in result\")\n        test(\"'minutes' in result\")\n        test(\"result.dump()\", str)\n\n    prints::\n\n        list(result) -> ['1999', '/', '12', '/', '31']\n        result[0] -> '1999'\n        result['month'] -> '12'\n        result.day -> '31'\n        'month' in result -> True\n        'minutes' in result -> False\n        result.dump() -> ['1999', '/', '12', '/', '31']\n        - day: 31\n        - month: 12\n        - year: 1999\n    ";
static const char __pyx_k_Token_for_matching_words_compose[] = "Token for matching words composed of allowed character sets.\n    Defined with string containing all allowed initial characters, an\n    optional string containing allowed body characters (if omitted,\n    defaults to the initial character set), and an optional minimum,\n    maximum, and/or exact length.  The default value for ``min`` is\n    1 (a minimum value < 1 is not valid); the default values for\n    ``max`` and ``exact`` are 0, meaning no maximum or exact\n    length restriction. An optional ``excludeChars`` parameter can\n    list characters that might be found in the input ``bodyChars``\n    string; useful to define a word of all printables except for one or\n    two characters, for instance.\n\n    :class:`srange` is useful for defining custom character set strings\n    for defining ``Word`` expressions, using range notation from\n    regular expression character sets.\n\n    A common mistake is to use :class:`Word` to match a specific literal\n    string, as in ``Word(\"Address\")``. Remember that :class:`Word`\n    uses the string argument to define *sets* of matchable characters.\n    This expression would match \"Add\", \"AAA\", \"dAred\", or any other word\n    made up of the characters 'A', 'd', 'r', 'e', and 's'. To match an\n    exact literal string, use :class:`Literal` or :class:`Keyword`.\n\n    pyparsing includes helper strings for building Words:\n\n     - :class:`alphas`\n     - :class:`nums`\n     - :class:`alphanums`\n     - :class:`hexnums`\n     - :class:`alphas8bit` (alphabetic characters in ASCII range 128-255\n       - accented, tilded, umlauted, etc.)\n     - :class:`punc8bit` (non-alphabetic characters in ASCII range\n       128-255 - currency, symbols, superscripts, diacriticals, etc.)\n     - :class:`printables` (any non-whitespace character)\n\n    Example::\n\n        # a word composed of digits\n        integer = Word(nums) # equivalent to Word(\"0123456789\") or Word(srange(\"0-9\"))\n\n        # a word with a leading capital,"" and zero or more lowercase\n        capital_word = Word(alphas.upper(), alphas.lower())\n\n        # hostnames are alphanumeric, with leading alpha, and '-'\n        hostname = Word(alphas, alphanums + '-')\n\n        # roman numeral (not a strict parser, accepts invalid mix of characters)\n        roman = Word(\"IVXLCDM\")\n\n        # any string of non-whitespace characters, except for ','\n        csv_value = Word(printables, excludeChars=\",\")\n    ";
static const char __pyx_k_Token_for_skipping_over_all_unde[] = "Token for skipping over all undefined text until the matched\n    expression is found.\n\n    Parameters:\n     - expr - target expression marking the end of the data to be skipped\n     - include - (default= ``False``) if True, the target expression is also parsed\n          (the skipped text and target expression are returned as a 2-element list).\n     - ignore - (default= ``None``) used to define grammars (typically quoted strings and\n          comments) that might contain false matches to the target expression\n     - failOn - (default= ``None``) define expressions that are not allowed to be\n          included in the skipped test; if found before the target expression is found,\n          the SkipTo is not a match\n\n    Example::\n\n        report = '''\n            Outstanding Issues Report - 1 Jan 2000\n\n               # | Severity | Description                               |  Days Open\n            -----+----------+-------------------------------------------+-----------\n             101 | Critical | Intermittent system crash                 |          6\n              94 | Cosmetic | Spelling error on Login ('log|n')         |         14\n              79 | Minor    | System slow when running too many reports |         47\n            '''\n        integer = Word(nums)\n        SEP = Suppress('|')\n        # use SkipTo to simply match everything up until the next SEP\n        # - ignore quoted strings, so that a '|' character inside a quoted string does not match\n        # - parse action will call token.strip() for each matched token, i.e., the description body\n        string_data = SkipTo(SEP, ignore=quotedString)\n        string_data.setParseAction(tokenMap(str.strip))\n        ticket_expr = (integer(\"issue_num\") + SEP\n                      + string_data(\"sev\") + SEP\n                      + string_data(\"desc\") + SEP\n                      + integer(\"days_open\"))\n\n        for tkt in ticket_expr.searchString(report):\n            print tkt"".dump()\n\n    prints::\n\n        ['101', 'Critical', 'Intermittent system crash', '6']\n        - days_open: 6\n        - desc: Intermittent system crash\n        - issue_num: 101\n        - sev: Critical\n        ['94', 'Cosmetic', \"Spelling error on Login ('log|n')\", '14']\n        - days_open: 14\n        - desc: Spelling error on Login ('log|n')\n        - issue_num: 94\n        - sev: Cosmetic\n        ['79', 'Minor', 'System slow when running too many reports', '47']\n        - days_open: 47\n        - desc: System slow when running too many reports\n        - issue_num: 79\n        - sev: Minor\n    ";
static const char __pyx_k_Token_to_advance_to_a_specific_c[] = "Token to advance to a specific column of input text; useful for\n    tabular report scraping.\n    ";
static const char __pyx_k_Token_to_exactly_match_a_specifi[] = "Token to exactly match a specified string.\n\n    Example::\n\n        Literal('blah').parseString('blah')  # -> ['blah']\n        Literal('blah').parseString('blahfooblah')  # -> ['blah']\n        Literal('blah').parseString('bla')  # -> Exception: Expected \"blah\"\n\n    For case-insensitive matching, use :class:`CaselessLiteral`.\n\n    For keyword matching (force word break before and after the matched string),\n    use :class:`Keyword` or :class:`CaselessKeyword`.\n    ";
static const char __pyx_k_Token_to_match_a_specified_strin[] = "Token to match a specified string, ignoring case of letters.\n    Note: the matched results will always be in the case of the given\n    match string, NOT the case of the input text.\n\n    Example::\n\n        OneOrMore(CaselessLiteral(\"CMD\")).parseString(\"cmd CMD Cmd10\") # -> ['CMD', 'CMD', 'CMD']\n\n    (Contrast with example for :class:`CaselessKeyword`.)\n    ";
static const char __pyx_k_Wrapper_for_parse_actions_to_ens[] = "Wrapper for parse actions, to ensure they are only called once.\n    ";
static const char __pyx_k_base_exception_class_for_all_par[] = "base exception class for all parsing runtime exceptions";
static const char __pyx_k_cannot_construct_And_with_sequen[] = "cannot construct And with sequence ending in ...";
static const char __pyx_k_cannot_multiply_ParserElement_an[] = "cannot multiply 'ParserElement' and ('%s', '%s') objects";
static const char __pyx_k_cannot_multiply_ParserElement_by[] = "cannot multiply ParserElement by negative value";
static const char __pyx_k_cannot_use_sub_with_Regex_asGrou[] = "cannot use sub() with Regex(asGroupList=True)";
static const char __pyx_k_cannot_use_sub_with_a_callable_w[] = "cannot use sub() with a callable with Regex(asMatch=True)";
static const char __pyx_k_enable_debug_on_named_expression[] = "enable_debug_on_named_expressions";
static const char __pyx_k_endQuoteChar_cannot_be_the_empty[] = "endQuoteChar cannot be the empty string";
static const char __pyx_k_exception_raised_in_parse_action[] = "exception raised in parse action";
static const char __pyx_k_exception_thrown_by_class_Parser[] = "exception thrown by :class:`ParserElement.validate` if the\n    grammar could be improperly recursive\n    ";
static const char __pyx_k_invalid_pattern_s_passed_to_Rege[] = "invalid pattern (%s) passed to Regex";
static const char __pyx_k_just_like_class_ParseFatalExcept[] = "just like :class:`ParseFatalException`, but thrown internally\n    when an :class:`ErrorStop<And._ErrorStop>` ('-' operator) indicates\n    that parsing is to stop immediately because an unbacktrackable\n    syntax error has been found.\n    ";
static const char __pyx_k_no_defined_alternatives_to_match[] = "no defined alternatives to match";
static const char __pyx_k_null_string_passed_to_Keyword_us[] = "null string passed to Keyword; use Empty() instead";
static const char __pyx_k_null_string_passed_to_Literal_us[] = "null string passed to Literal; use Empty() instead";
static const char __pyx_k_only_1_or_2_index_arguments_supp[] = "only 1 or 2 index arguments supported ({0}{1})";
static const char __pyx_k_pop_got_an_unexpected_keyword_ar[] = "pop() got an unexpected keyword argument '%s'";
static const char __pyx_k_quoteChar_cannot_be_the_empty_st[] = "quoteChar cannot be the empty string";
static const char __pyx_k_quoted_string_starting_with_s_en[] = "quoted string, starting with %s ending with %s";
static const char __pyx_k_second_tuple_value_must_be_great[] = "second tuple value must be greater or equal to first tuple value";
static const char __pyx_k_warn_multiple_tokens_in_named_al[] = "warn_multiple_tokens_in_named_alternation";
static const char __pyx_k_warn_ungrouped_named_tokens_in_c[] = "warn_ungrouped_named_tokens_in_collection";
static const char __pyx_k_0_setting_results_name_1_r_on_2_2[] = "{0}: setting results name {1!r} on {2} expression may only return a single token for an And alternative, in future will return the full list of tokens";
static const char __pyx_k_Matches_if_the_current_position_2[] = "Matches if the current position is at the end of a Word, and is\n    not followed by any character in a given set of ``wordChars``\n    (default= ``printables``). To emulate the ``\010`` behavior of\n    regular expressions, use ``WordEnd(alphanums)``. ``WordEnd``\n    will also match at the end of the string being parsed, or at the end\n    of a line.\n    ";
static const char __pyx_k_ParserElement__FifoCache___init_2[] = "ParserElement._FifoCache.__init__.<locals>.set";
static const char __pyx_k_ParserElement__FifoCache___init_3[] = "ParserElement._FifoCache.__init__.<locals>.clear";
static const char __pyx_k_ParserElement__FifoCache___init_4[] = "ParserElement._FifoCache.__init__.<locals>.cache_len";
static const char __pyx_k_ParserElement__FifoCache___init_5[] = "ParserElement._FifoCache.__init__";
static const char __pyx_k_Token_for_matching_strings_that_2[] = "\n    Token for matching strings that are delimited by quoting characters.\n\n    Defined with the following parameters:\n\n        - quoteChar - string of one or more characters defining the\n          quote delimiting string\n        - escChar - character to escape quotes, typically backslash\n          (default= ``None``)\n        - escQuote - special quote sequence to escape an embedded quote\n          string (such as SQL's ``\"\"`` to escape an embedded ``\"``)\n          (default= ``None``)\n        - multiline - boolean indicating whether quotes can span\n          multiple lines (default= ``False``)\n        - unquoteResults - boolean indicating whether the matched text\n          should be unquoted (default= ``True``)\n        - endQuoteChar - string of one or more characters defining the\n          end of the quote delimited string (default= ``None``  => same as\n          quoteChar)\n        - convertWhitespaceEscapes - convert escaped whitespace\n          (``'\\t'``, ``'\\n'``, etc.) to actual whitespace\n          (default= ``True``)\n\n    Example::\n\n        qs = QuotedString('\"')\n        print(qs.searchString('lsjdf \"This is the quote\" sldjf'))\n        complex_qs = QuotedString('{{', endQuoteChar='}}')\n        print(complex_qs.searchString('lsjdf {{This is the \"quote\"}} sldjf'))\n        sql_qs = QuotedString('\"', escQuote='\"\"')\n        print(sql_qs.searchString('lsjdf \"This is the quote with \"\"embedded\"\" quotes\" sldjf'))\n\n    prints::\n\n        [['This is the quote']]\n        [['This is the \"quote\"']]\n        [['This is the quote with \"embedded\" quotes']]\n    ";
static const char __pyx_k_cannot_specify_a_minimum_length_2[] = "cannot specify a minimum length < 1; use Optional(CharsNotIn()) if zero-length char group is permitted";
static const char __pyx_k_Aspidites__vendor_pyparsing_exte_2[] = "Aspidites/_vendor/pyparsing_extension.py";
static const char __pyx_k_Matches_if_current_position_is_a_2[] = "Matches if current position is at the end of a line within the\n    parse string\n    ";
static const char __pyx_k_Matches_if_current_position_is_a_3[] = "Matches if current position is at the beginning of the parse\n    string\n    ";
static const char __pyx_k_Matches_if_current_position_is_a_4[] = "Matches if current position is at the end of the parse string\n    ";
static const char __pyx_k_Requires_that_at_least_one_class_2[] = "Requires that at least one :class:`ParseExpression` is found. If\n    two expressions match, the first one listed is the one that will\n    match. May be constructed using the ``'|'`` operator.\n\n    Example::\n\n        # construct MatchFirst using '|' operator\n\n        # watch the order of expressions to match\n        number = Word(nums) | Combine(Word(nums) + '.' + Word(nums))\n        print(number.searchString(\"123 3.1416 789\")) #  Fail! -> [['123'], ['3'], ['1416'], ['789']]\n\n        # put more selective expression first\n        number = Combine(Word(nums) + '.' + Word(nums)) | Word(nums)\n        print(number.searchString(\"123 3.1416 789\")) #  Better -> [['123'], ['3.1416'], ['789']]\n    ";
static const char __pyx_k_Token_for_matching_words_compose_2[] = "Token for matching words composed of characters *not* in a given\n    set (will include whitespace in matched characters if not listed in\n    the provided exclusion set - see example). Defined with string\n    containing all disallowed characters, and an optional minimum,\n    maximum, and/or exact length.  The default value for ``min`` is\n    1 (a minimum value < 1 is not valid); the default values for\n    ``max`` and ``exact`` are 0, meaning no maximum or exact\n    length restriction.\n\n    Example::\n\n        # define a comma-separated-value as anything that is not a ','\n        csv_value = CharsNotIn(',')\n        print(delimitedList(csv_value).parseString(\"dkls,lsdkjf,s12 34,@!#,213\"))\n\n    prints::\n\n        ['dkls', 'lsdkjf', 's12 34', '@!#', '213']\n    ";
static const char __pyx_k_Token_to_exactly_match_a_specifi_2[] = "Token to exactly match a specified string as a keyword, that is,\n    it must be immediately followed by a non-keyword character.  Compare\n    with :class:`Literal`:\n\n     - ``Literal(\"if\")`` will match the leading ``'if'`` in\n       ``'ifAndOnlyIf'``.\n     - ``Keyword(\"if\")`` will not; it will only match the leading\n       ``'if'`` in ``'if x=1'``, or ``'if(y==2)'``\n\n    Accepts two optional constructor arguments in addition to the\n    keyword string:\n\n     - ``identChars`` is a string of characters that would be valid\n       identifier characters, defaulting to all alphanumerics + \"_\" and\n       \"$\"\n     - ``caseless`` allows case-insensitive matching, default is ``False``.\n\n    Example::\n\n        Keyword(\"start\").parseString(\"start\")  # -> ['start']\n        Keyword(\"start\").parseString(\"starting\")  # -> Exception\n\n    For case-insensitive matching, use :class:`CaselessKeyword`.\n    ";
static const char __pyx_k_cannot_multiply_ParserElement_an_2[] = "cannot multiply 'ParserElement' and '%s' objects";
static const char __pyx_k_cannot_multiply_ParserElement_by_2[] = "cannot multiply ParserElement by 0 or (0, 0)";
static PyObject *__pyx_kp_s_;
static PyObject *__pyx_kp_s_0;
static PyObject *__pyx_kp_s_0123456789;
static PyObject *__pyx_kp_s_0_1;
static PyObject *__pyx_kp_s_0_1_2;
static PyObject *__pyx_kp_s_0_1_3;
static PyObject *__pyx_kp_s_0_2;
static PyObject *__pyx_kp_s_0_failed_1_2;
static PyObject *__pyx_kp_s_0_setting_results_name_0_r_on_1;
static PyObject *__pyx_kp_s_0_setting_results_name_1_r_on_2;
static PyObject *__pyx_kp_s_0_setting_results_name_1_r_on_2_2;
static PyObject *__pyx_n_s_ABCDEFabcdef;
static PyObject *__pyx_kp_s_A_short_cut_class_for_defining_W;
static PyObject *__pyx_kp_s_A_token_that_will_never_match;
static PyObject *__pyx_kp_s_A_variation_on_class_Literal_whi;
static PyObject *__pyx_kp_s_Abstract_base_level_parser_eleme;
static PyObject *__pyx_kp_s_Abstract_class_ParserElement_sub;
static PyObject *__pyx_kp_s_Abstract_subclass_of_ParserEleme;
static PyObject *__pyx_kp_s_Abstract_subclass_of_class_Pars;
static PyObject *__pyx_kp_s_Abstract_subclass_of_class_Parse;
static PyObject *__pyx_kp_s_An_empty_token_will_always_match;
static PyObject *__pyx_n_s_And;
static PyObject *__pyx_n_s_And__ErrorStop;
static PyObject *__pyx_n_s_And__ErrorStop___init;
static PyObject *__pyx_n_s_And___iadd;
static PyObject *__pyx_n_s_And___init;
static PyObject *__pyx_n_s_And___init___locals_genexpr;
static PyObject *__pyx_n_s_And___str;
static PyObject *__pyx_n_s_And___str___locals_genexpr;
static PyObject *__pyx_n_s_And_checkRecursion;
static PyObject *__pyx_n_s_And_parseImpl;
static PyObject *__pyx_n_s_And_streamline;
static PyObject *__pyx_n_s_And_streamline_locals_genexpr;
static PyObject *__pyx_n_s_Aspidites__vendor_pyparsing_exte;
static PyObject *__pyx_kp_s_Aspidites__vendor_pyparsing_exte_2;
static PyObject *__pyx_n_s_AttributeError;
static PyObject *__pyx_n_s_BOM;
static PyObject *__pyx_kp_s_CR;
static PyObject *__pyx_kp_s_Cannot_combine_element_of_type_s;
static PyObject *__pyx_n_s_CaselessKeyword;
static PyObject *__pyx_n_s_CaselessKeyword___init;
static PyObject *__pyx_n_s_CaselessLiteral;
static PyObject *__pyx_n_s_CaselessLiteral___init;
static PyObject *__pyx_n_s_CaselessLiteral_parseImpl;
static PyObject *__pyx_kp_s_Caseless_version_of_class_Keywo;
static PyObject *__pyx_n_s_Char;
static PyObject *__pyx_n_s_Char___init;
static PyObject *__pyx_n_s_CharsNotIn;
static PyObject *__pyx_n_s_CharsNotIn___init;
static PyObject *__pyx_n_s_CharsNotIn___str;
static PyObject *__pyx_n_s_CharsNotIn_parseImpl;
static PyObject *__pyx_n_s_CloseMatch;
static PyObject *__pyx_n_s_CloseMatch___init;
static PyObject *__pyx_n_s_CloseMatch_parseImpl;
static PyObject *__pyx_n_s_Combine;
static PyObject *__pyx_n_s_Combine___init;
static PyObject *__pyx_n_s_Combine_ignore;
static PyObject *__pyx_n_s_Combine_postParse;
static PyObject *__pyx_kp_s_Converter_for_ignoring_the_resul;
static PyObject *__pyx_kp_s_Converter_to_concatenate_all_mat;
static PyObject *__pyx_kp_s_Converter_to_return_a_repetitive;
static PyObject *__pyx_kp_s_Converter_to_return_the_matched;
static PyObject *__pyx_n_s_DEFAULT_KEYWORD_CHARS;
static PyObject *__pyx_n_s_DEFAULT_WHITE_CHARS;
static PyObject *__pyx_n_s_DOTALL;
static PyObject *__pyx_kp_s_Diagnostic_configuration_all_de;
static PyObject *__pyx_n_s_Dict;
static PyObject *__pyx_n_s_Dict___init;
static PyObject *__pyx_n_s_Dict_postParse;
static PyObject *__pyx_kp_s_EM_QUAD;
static PyObject *__pyx_kp_s_EM_SPACE;
static PyObject *__pyx_kp_s_EN_QUAD;
static PyObject *__pyx_kp_s_EN_SPACE;
static PyObject *__pyx_n_s_Each;
static PyObject *__pyx_n_s_Each___init;
static PyObject *__pyx_n_s_Each___init___locals_genexpr;
static PyObject *__pyx_n_s_Each___str;
static PyObject *__pyx_n_s_Each___str___locals_genexpr;
static PyObject *__pyx_n_s_Each_checkRecursion;
static PyObject *__pyx_n_s_Each_parseImpl;
static PyObject *__pyx_n_s_Each_parseImpl_locals_genexpr;
static PyObject *__pyx_n_s_Each_streamline;
static PyObject *__pyx_n_s_Each_streamline_locals_genexpr;
static PyObject *__pyx_n_s_Ellipsis;
static PyObject *__pyx_n_s_Empty;
static PyObject *__pyx_n_s_Empty___init;
static PyObject *__pyx_n_s_ErrorStop;
static PyObject *__pyx_kp_s_Exception_raised;
static PyObject *__pyx_kp_s_Exception_thrown_when_parse_exp;
static PyObject *__pyx_kp_s_Expected;
static PyObject *__pyx_kp_s_Expected_end_of_line;
static PyObject *__pyx_kp_s_Expected_end_of_text;
static PyObject *__pyx_kp_s_Expected_r_with_up_to_d_mismatch;
static PyObject *__pyx_kp_s_Expected_start_of_line;
static PyObject *__pyx_kp_s_Expected_start_of_text;
static PyObject *__pyx_kp_s_FAIL;
static PyObject *__pyx_n_s_FAIL_2;
static PyObject *__pyx_kp_s_FAIL_EXCEPTION;
static PyObject *__pyx_kp_s_FATAL;
static PyObject *__pyx_kp_s_FF;
static PyObject *__pyx_kp_s_FIGURE_SPACE;
static PyObject *__pyx_kp_s_FOUR_PER_EM_SPACE;
static PyObject *__pyx_n_s_FifoCache;
static PyObject *__pyx_n_s_FollowedBy;
static PyObject *__pyx_n_s_FollowedBy___init;
static PyObject *__pyx_n_s_FollowedBy_parseImpl;
static PyObject *__pyx_n_s_Forward;
static PyObject *__pyx_n_s_Forward___ilshift;
static PyObject *__pyx_n_s_Forward___init;
static PyObject *__pyx_n_s_Forward___lshift;
static PyObject *__pyx_n_s_Forward___str;
static PyObject *__pyx_n_s_Forward__setResultsName;
static PyObject *__pyx_n_s_Forward_copy;
static PyObject *__pyx_kp_s_Forward_declaration_of_an_expres;
static PyObject *__pyx_n_s_Forward_leaveWhitespace;
static PyObject *__pyx_n_s_Forward_streamline;
static PyObject *__pyx_n_s_Forward_validate;
static PyObject *__pyx_kp_s_Found_unwanted_token;
static PyObject *__pyx_n_s_GoToColumn;
static PyObject *__pyx_n_s_GoToColumn___init;
static PyObject *__pyx_n_s_GoToColumn_parseImpl;
static PyObject *__pyx_n_s_GoToColumn_preParse;
static PyObject *__pyx_n_s_Group;
static PyObject *__pyx_n_s_Group___init;
static PyObject *__pyx_n_s_Group_postParse;
static PyObject *__pyx_kp_s_HAIR_SPACE;
static PyObject *__pyx_n_s_HIT;
static PyObject *__pyx_kp_s_IDEOGRAPHIC_SPACE;
static PyObject *__pyx_n_s_ITEM;
static PyObject *__pyx_n_s_ImportError;
static PyObject *__pyx_n_s_IndexError;
static PyObject *__pyx_n_s_Iterable;
static PyObject *__pyx_n_s_KeyError;
static PyObject *__pyx_n_s_Keyword;
static PyObject *__pyx_n_s_Keyword___init;
static PyObject *__pyx_n_s_Keyword_copy;
static PyObject *__pyx_n_s_Keyword_parseImpl;
static PyObject *__pyx_n_s_Keyword_setDefaultKeywordChars;
static PyObject *__pyx_n_s_L;
static PyObject *__pyx_kp_s_LF;
static PyObject *__pyx_n_s_L_2;
static PyObject *__pyx_n_s_LineEnd;
static PyObject *__pyx_n_s_LineEnd___init;
static PyObject *__pyx_n_s_LineEnd_parseImpl;
static PyObject *__pyx_n_s_LineStart;
static PyObject *__pyx_n_s_LineStart___init;
static PyObject *__pyx_n_s_LineStart_parseImpl;
static PyObject *__pyx_n_s_Literal;
static PyObject *__pyx_n_s_Literal___init;
static PyObject *__pyx_n_s_Literal_parseImpl;
static PyObject *__pyx_kp_s_Lookahead_matching_of_the_given;
static PyObject *__pyx_kp_s_Lookahead_to_disallow_matching_w;
static PyObject *__pyx_kp_s_Lookbehind_matching_of_the_given;
static PyObject *__pyx_n_s_MATCH;
static PyObject *__pyx_n_s_MISS;
static PyObject *__pyx_kp_s_MMSP;
static PyObject *__pyx_kp_s_MONGOLIAN_VOWEL_SEPARATOR;
static PyObject *__pyx_n_s_MULTILINE;
static PyObject *__pyx_n_s_Mapping;
static PyObject *__pyx_kp_s_Match;
static PyObject *__pyx_n_s_MatchFirst;
static PyObject *__pyx_n_s_MatchFirst___init;
static PyObject *__pyx_n_s_MatchFirst___init___locals_genex;
static PyObject *__pyx_n_s_MatchFirst___ior;
static PyObject *__pyx_n_s_MatchFirst___str;
static PyObject *__pyx_n_s_MatchFirst___str___locals_genexp;
static PyObject *__pyx_n_s_MatchFirst__setResultsName;
static PyObject *__pyx_n_s_MatchFirst__setResultsName_local;
static PyObject *__pyx_n_s_MatchFirst_checkRecursion;
static PyObject *__pyx_n_s_MatchFirst_parseImpl;
static PyObject *__pyx_n_s_MatchFirst_streamline;
static PyObject *__pyx_n_s_MatchFirst_streamline_locals_gen;
static PyObject *__pyx_kp_s_Matched;
static PyObject *__pyx_kp_s_Matches_if_current_position_is_a;
static PyObject *__pyx_kp_s_Matches_if_current_position_is_a_2;
static PyObject *__pyx_kp_s_Matches_if_current_position_is_a_3;
static PyObject *__pyx_kp_s_Matches_if_current_position_is_a_4;
static PyObject *__pyx_kp_s_Matches_if_the_current_position;
static PyObject *__pyx_kp_s_Matches_if_the_current_position_2;
static PyObject *__pyx_n_s_MethodType;
static PyObject *__pyx_kp_s_Missing_one_or_more_required_ele;
static PyObject *__pyx_n_s_MultipleMatch;
static PyObject *__pyx_n_s_MultipleMatch___init;
static PyObject *__pyx_n_s_MultipleMatch__setResultsName;
static PyObject *__pyx_n_s_MultipleMatch_parseImpl;
static PyObject *__pyx_n_s_MultipleMatch_stopOn;
static PyObject *__pyx_n_s_MutableMapping;
static PyObject *__pyx_kp_s_NBSP;
static PyObject *__pyx_n_s_NL;
static PyObject *__pyx_kp_s_NNBSP;
static PyObject *__pyx_n_s_NoMatch;
static PyObject *__pyx_n_s_NoMatch___init;
static PyObject *__pyx_n_s_NoMatch_parseImpl;
static PyObject *__pyx_kp_s_No_match_found_for;
static PyObject *__pyx_n_s_None;
static PyObject *__pyx_n_s_NotAny;
static PyObject *__pyx_n_s_NotAny___init;
static PyObject *__pyx_n_s_NotAny___str;
static PyObject *__pyx_n_s_NotAny_parseImpl;
static PyObject *__pyx_kp_s_Not_at_the_end_of_a_word;
static PyObject *__pyx_kp_s_Not_at_the_start_of_a_word;
static PyObject *__pyx_n_s_NullToken;
static PyObject *__pyx_n_s_NullToken___bool;
static PyObject *__pyx_n_s_NullToken___str;
static PyObject *__pyx_kp_s_OGHAM_SPACE_MARK;
static PyObject *__pyx_n_s_OneOrMore;
static PyObject *__pyx_n_s_OneOrMore___str;
static PyObject *__pyx_n_s_OnlyOnce;
static PyObject *__pyx_n_s_OnlyOnce___call;
static PyObject *__pyx_n_s_OnlyOnce___init;
static PyObject *__pyx_n_s_OnlyOnce_reset;
static PyObject *__pyx_n_s_Optional;
static PyObject *__pyx_n_s_Optional___init;
static PyObject *__pyx_n_s_Optional___str;
static PyObject *__pyx_n_s_Optional__optionalNotMatched;
static PyObject *__pyx_kp_s_Optional_matching_of_the_given_e;
static PyObject *__pyx_n_s_Optional_parseImpl;
static PyObject *__pyx_kp_s_Optional_repetition_of_zero_or_m;
static PyObject *__pyx_n_s_Or;
static PyObject *__pyx_n_s_Or___init;
static PyObject *__pyx_n_s_Or___init___locals_genexpr;
static PyObject *__pyx_n_s_Or___ixor;
static PyObject *__pyx_n_s_Or___str;
static PyObject *__pyx_n_s_Or___str___locals_genexpr;
static PyObject *__pyx_n_s_Or__setResultsName;
static PyObject *__pyx_n_s_Or__setResultsName_locals_genexp;
static PyObject *__pyx_n_s_Or_checkRecursion;
static PyObject *__pyx_n_s_Or_parseImpl;
static PyObject *__pyx_n_s_Or_streamline;
static PyObject *__pyx_n_s_Or_streamline_locals_genexpr;
static PyObject *__pyx_n_s_OrderedDict;
static PyObject *__pyx_n_s_OrderedDict_2;
static PyObject *__pyx_kp_s_PUNCTUATION_SPACE;
static PyObject *__pyx_n_s_PY_3;
static PyObject *__pyx_n_s_ParseBaseException;
static PyObject *__pyx_n_s_ParseBaseException___dir;
static PyObject *__pyx_n_s_ParseBaseException___getattr;
static PyObject *__pyx_n_s_ParseBaseException___init;
static PyObject *__pyx_n_s_ParseBaseException___repr;
static PyObject *__pyx_n_s_ParseBaseException___str;
static PyObject *__pyx_n_s_ParseBaseException__from_excepti;
static PyObject *__pyx_n_s_ParseBaseException_markInputline;
static PyObject *__pyx_n_s_ParseElementEnhance;
static PyObject *__pyx_n_s_ParseElementEnhance___init;
static PyObject *__pyx_n_s_ParseElementEnhance___str;
static PyObject *__pyx_n_s_ParseElementEnhance_checkRecursi;
static PyObject *__pyx_n_s_ParseElementEnhance_ignore;
static PyObject *__pyx_n_s_ParseElementEnhance_leaveWhitesp;
static PyObject *__pyx_n_s_ParseElementEnhance_parseImpl;
static PyObject *__pyx_n_s_ParseElementEnhance_streamline;
static PyObject *__pyx_n_s_ParseElementEnhance_validate;
static PyObject *__pyx_n_s_ParseException;
static PyObject *__pyx_n_s_ParseException_explain;
static PyObject *__pyx_n_s_ParseExpression;
static PyObject *__pyx_n_s_ParseExpression___init;
static PyObject *__pyx_n_s_ParseExpression___init___locals;
static PyObject *__pyx_n_s_ParseExpression___str;
static PyObject *__pyx_n_s_ParseExpression__setResultsName;
static PyObject *__pyx_n_s_ParseExpression_append;
static PyObject *__pyx_n_s_ParseExpression_copy;
static PyObject *__pyx_n_s_ParseExpression_ignore;
static PyObject *__pyx_n_s_ParseExpression_leaveWhitespace;
static PyObject *__pyx_n_s_ParseExpression_streamline;
static PyObject *__pyx_n_s_ParseExpression_validate;
static PyObject *__pyx_n_s_ParseFatalException;
static PyObject *__pyx_n_s_ParseResults;
static PyObject *__pyx_n_s_ParseResultsWithOffset;
static PyObject *__pyx_n_s_ParseResultsWithOffset___getite;
static PyObject *__pyx_n_s_ParseResultsWithOffset___init;
static PyObject *__pyx_n_s_ParseResultsWithOffset___repr;
static PyObject *__pyx_n_s_ParseResultsWithOffset_setOffse;
static PyObject *__pyx_n_s_ParseResults___add;
static PyObject *__pyx_n_s_ParseResults___bool;
static PyObject *__pyx_n_s_ParseResults___contains;
static PyObject *__pyx_n_s_ParseResults___delitem;
static PyObject *__pyx_n_s_ParseResults___dir;
static PyObject *__pyx_n_s_ParseResults___getattr;
static PyObject *__pyx_n_s_ParseResults___getitem;
static PyObject *__pyx_n_s_ParseResults___getnewargs;
static PyObject *__pyx_n_s_ParseResults___getstate;
static PyObject *__pyx_n_s_ParseResults___iadd;
static PyObject *__pyx_n_s_ParseResults___iadd___locals_lam;
static PyObject *__pyx_n_s_ParseResults___init;
static PyObject *__pyx_n_s_ParseResults___iter;
static PyObject *__pyx_n_s_ParseResults___len;
static PyObject *__pyx_n_s_ParseResults___lookup;
static PyObject *__pyx_n_s_ParseResults___new;
static PyObject *__pyx_n_s_ParseResults___radd;
static PyObject *__pyx_n_s_ParseResults___repr;
static PyObject *__pyx_n_s_ParseResults___reversed;
static PyObject *__pyx_n_s_ParseResults___setitem;
static PyObject *__pyx_n_s_ParseResults___setstate;
static PyObject *__pyx_n_s_ParseResults___str;
static PyObject *__pyx_n_s_ParseResults___str___locals_gene;
static PyObject *__pyx_n_s_ParseResults__accumNames;
static PyObject *__pyx_n_s_ParseResults__asList;
static PyObject *__pyx_n_s_ParseResults__asStringList;
static PyObject *__pyx_n_s_ParseResults__doinit;
static PyObject *__pyx_n_s_ParseResults__iteritems;
static PyObject *__pyx_n_s_ParseResults__iteritems_locals_g;
static PyObject *__pyx_n_s_ParseResults__iterkeys;
static PyObject *__pyx_n_s_ParseResults__itervalues;
static PyObject *__pyx_n_s_ParseResults__itervalues_locals;
static PyObject *__pyx_n_s_ParseResults__lookup;
static PyObject *__pyx_n_s_ParseResults__modal;
static PyObject *__pyx_n_s_ParseResults__name;
static PyObject *__pyx_n_s_ParseResults__parent;
static PyObject *__pyx_n_s_ParseResults__tokdict;
static PyObject *__pyx_n_s_ParseResults__toklist;
static PyObject *__pyx_n_s_ParseResults_append;
static PyObject *__pyx_n_s_ParseResults_asDict;
static PyObject *__pyx_n_s_ParseResults_asDict_locals_genex;
static PyObject *__pyx_n_s_ParseResults_asDict_locals_toIte;
static PyObject *__pyx_n_s_ParseResults_asList;
static PyObject *__pyx_n_s_ParseResults_asXML;
static PyObject *__pyx_n_s_ParseResults_asXML_locals_genexp;
static PyObject *__pyx_n_s_ParseResults_clear;
static PyObject *__pyx_n_s_ParseResults_copy;
static PyObject *__pyx_n_s_ParseResults_dump;
static PyObject *__pyx_n_s_ParseResults_dump_locals_genexpr;
static PyObject *__pyx_n_s_ParseResults_extend;
static PyObject *__pyx_n_s_ParseResults_from_dict;
static PyObject *__pyx_n_s_ParseResults_from_dict_locals_is;
static PyObject *__pyx_n_s_ParseResults_get;
static PyObject *__pyx_n_s_ParseResults_getName;
static PyObject *__pyx_n_s_ParseResults_haskeys;
static PyObject *__pyx_n_s_ParseResults_insert;
static PyObject *__pyx_n_s_ParseResults_items;
static PyObject *__pyx_n_s_ParseResults_keys;
static PyObject *__pyx_n_s_ParseResults_pop;
static PyObject *__pyx_n_s_ParseResults_pprint;
static PyObject *__pyx_n_s_ParseResults_values;
static PyObject *__pyx_n_s_ParseSyntaxException;
static PyObject *__pyx_n_s_ParserElement;
static PyObject *__pyx_n_s_ParserElement__FifoCache;
static PyObject *__pyx_n_s_ParserElement__FifoCache___init;
static PyObject *__pyx_n_s_ParserElement__FifoCache___init_2;
static PyObject *__pyx_n_s_ParserElement__FifoCache___init_3;
static PyObject *__pyx_n_s_ParserElement__FifoCache___init_4;
static PyObject *__pyx_n_s_ParserElement__FifoCache___init_5;
static PyObject *__pyx_n_s_ParserElement__UnboundedCache;
static PyObject *__pyx_n_s_ParserElement__UnboundedCache_2;
static PyObject *__pyx_n_s_ParserElement__UnboundedCache_3;
static PyObject *__pyx_n_s_ParserElement__UnboundedCache_4;
static PyObject *__pyx_n_s_ParserElement__UnboundedCache_5;
static PyObject *__pyx_n_s_ParserElement__UnboundedCache_6;
static PyObject *__pyx_n_s_ParserElement___add;
static PyObject *__pyx_n_s_ParserElement___and;
static PyObject *__pyx_n_s_ParserElement___call;
static PyObject *__pyx_n_s_ParserElement___eq;
static PyObject *__pyx_n_s_ParserElement___getitem;
static PyObject *__pyx_n_s_ParserElement___hash;
static PyObject *__pyx_n_s_ParserElement___init;
static PyObject *__pyx_n_s_ParserElement___invert;
static PyObject *__pyx_n_s_ParserElement___iter;
static PyObject *__pyx_n_s_ParserElement___mul;
static PyObject *__pyx_n_s_ParserElement___mul___locals_gen;
static PyObject *__pyx_n_s_ParserElement___mul___locals_mak;
static PyObject *__pyx_n_s_ParserElement___ne;
static PyObject *__pyx_n_s_ParserElement___or;
static PyObject *__pyx_n_s_ParserElement___radd;
static PyObject *__pyx_n_s_ParserElement___rand;
static PyObject *__pyx_n_s_ParserElement___repr;
static PyObject *__pyx_n_s_ParserElement___req;
static PyObject *__pyx_n_s_ParserElement___rmul;
static PyObject *__pyx_n_s_ParserElement___rne;
static PyObject *__pyx_n_s_ParserElement___ror;
static PyObject *__pyx_n_s_ParserElement___rsub;
static PyObject *__pyx_n_s_ParserElement___rxor;
static PyObject *__pyx_n_s_ParserElement___str;
static PyObject *__pyx_n_s_ParserElement___sub;
static PyObject *__pyx_n_s_ParserElement___xor;
static PyObject *__pyx_n_s_ParserElement__parseCache;
static PyObject *__pyx_n_s_ParserElement__parseNoCache;
static PyObject *__pyx_n_s_ParserElement__setResultsName;
static PyObject *__pyx_n_s_ParserElement__skipIgnorables;
static PyObject *__pyx_n_s_ParserElement__trim_traceback;
static PyObject *__pyx_n_s_ParserElement_addCondition;
static PyObject *__pyx_n_s_ParserElement_addParseAction;
static PyObject *__pyx_n_s_ParserElement_canParseNext;
static PyObject *__pyx_n_s_ParserElement_checkRecursion;
static PyObject *__pyx_n_s_ParserElement_copy;
static PyObject *__pyx_n_s_ParserElement_enablePackrat;
static PyObject *__pyx_n_s_ParserElement_ignore;
static PyObject *__pyx_n_s_ParserElement_inlineLiteralsUsin;
static PyObject *__pyx_n_s_ParserElement_leaveWhitespace;
static PyObject *__pyx_n_s_ParserElement_matches;
static PyObject *__pyx_n_s_ParserElement_parseFile;
static PyObject *__pyx_n_s_ParserElement_parseImpl;
static PyObject *__pyx_n_s_ParserElement_parseString;
static PyObject *__pyx_n_s_ParserElement_parseWithTabs;
static PyObject *__pyx_n_s_ParserElement_postParse;
static PyObject *__pyx_n_s_ParserElement_preParse;
static PyObject *__pyx_n_s_ParserElement_resetCache;
static PyObject *__pyx_n_s_ParserElement_runTests;
static PyObject *__pyx_n_s_ParserElement_scanString;
static PyObject *__pyx_n_s_ParserElement_searchString;
static PyObject *__pyx_n_s_ParserElement_setBreak;
static PyObject *__pyx_n_s_ParserElement_setBreak_locals_br;
static PyObject *__pyx_n_s_ParserElement_setDebug;
static PyObject *__pyx_n_s_ParserElement_setDebugActions;
static PyObject *__pyx_n_s_ParserElement_setDefaultWhitespa;
static PyObject *__pyx_n_s_ParserElement_setFailAction;
static PyObject *__pyx_n_s_ParserElement_setName;
static PyObject *__pyx_n_s_ParserElement_setParseAction;
static PyObject *__pyx_n_s_ParserElement_setParseAction_loc;
static PyObject *__pyx_n_s_ParserElement_setResultsName;
static PyObject *__pyx_n_s_ParserElement_setWhitespaceChars;
static PyObject *__pyx_n_s_ParserElement_split;
static PyObject *__pyx_n_s_ParserElement_streamline;
static PyObject *__pyx_n_s_ParserElement_suppress;
static PyObject *__pyx_n_s_ParserElement_transformString;
static PyObject *__pyx_n_s_ParserElement_tryParse;
static PyObject *__pyx_n_s_ParserElement_validate;
static PyObject *__pyx_n_s_PendingSkip;
static PyObject *__pyx_n_s_PositionToken;
static PyObject *__pyx_n_s_PositionToken___init;
static PyObject *__pyx_n_s_PrecededBy;
static PyObject *__pyx_n_s_PrecededBy___init;
static PyObject *__pyx_n_s_PrecededBy___init___locals_lambd;
static PyObject *__pyx_n_s_PrecededBy_parseImpl;
static PyObject *__pyx_n_s_QuotedString;
static PyObject *__pyx_n_s_QuotedString___init;
static PyObject *__pyx_n_s_QuotedString___init___locals_gen;
static PyObject *__pyx_n_s_QuotedString___str;
static PyObject *__pyx_n_s_QuotedString_parseImpl;
static PyObject *__pyx_n_s_RLock;
static PyObject *__pyx_kp_s_Re_s;
static PyObject *__pyx_n_s_RecursiveGrammarException;
static PyObject *__pyx_n_s_RecursiveGrammarException___init;
static PyObject *__pyx_n_s_RecursiveGrammarException___str;
static PyObject *__pyx_kp_s_RecursiveGrammarException_s;
static PyObject *__pyx_n_s_Regex;
static PyObject *__pyx_n_s_Regex___init;
static PyObject *__pyx_n_s_Regex___str;
static PyObject *__pyx_kp_s_Regex_may_only_be_constructed_wi;
static PyObject *__pyx_n_s_Regex_parseImpl;
static PyObject *__pyx_n_s_Regex_parseImplAsGroupList;
static PyObject *__pyx_n_s_Regex_parseImplAsMatch;
static PyObject *__pyx_n_s_Regex_sub;
static PyObject *__pyx_n_s_Regex_sub_locals_pa;
static PyObject *__pyx_kp_s_Repetition_of_one_or_more_of_the;
static PyObject *__pyx_kp_s_Requires_all_given_class_ParseE;
static PyObject *__pyx_kp_s_Requires_all_given_class_ParseEx;
static PyObject *__pyx_kp_s_Requires_that_at_least_one_class;
static PyObject *__pyx_kp_s_Requires_that_at_least_one_class_2;
static PyObject *__pyx_kp_s_SIX_PER_EM_SPACE;
static PyObject *__pyx_kp_s_SP;
static PyObject *__pyx_n_s_SimpleNamespace;
static PyObject *__pyx_n_s_SingleCharLiteral;
static PyObject *__pyx_n_s_SingleCharLiteral_parseImpl;
static PyObject *__pyx_n_s_SkipTo;
static PyObject *__pyx_n_s_SkipTo___init;
static PyObject *__pyx_n_s_SkipTo_parseImpl;
static PyObject *__pyx_kp_s_Special_matching_class_for_match;
static PyObject *__pyx_n_s_StringEnd;
static PyObject *__pyx_n_s_StringEnd___init;
static PyObject *__pyx_n_s_StringEnd_parseImpl;
static PyObject *__pyx_n_s_StringStart;
static PyObject *__pyx_n_s_StringStart___init;
static PyObject *__pyx_n_s_StringStart_parseImpl;
static PyObject *__pyx_kp_s_Structured_parse_results_to_prov;
static PyObject *__pyx_n_s_Suppress;
static PyObject *__pyx_n_s_Suppress_postParse;
static PyObject *__pyx_n_s_Suppress_suppress;
static PyObject *__pyx_n_s_SyntaxError;
static PyObject *__pyx_n_s_SyntaxWarning;
static PyObject *__pyx_kp_s_TAB;
static PyObject *__pyx_kp_s_THIN_SPACE;
static PyObject *__pyx_kp_s_THREE_PER_EM_SPACE;
static PyObject *__pyx_n_s_TRY;
static PyObject *__pyx_kp_s_Text_not_in_expected_column;
static PyObject *__pyx_n_s_Token;
static PyObject *__pyx_n_s_TokenConverter;
static PyObject *__pyx_n_s_TokenConverter___init;
static PyObject *__pyx_n_s_Token___init;
static PyObject *__pyx_kp_s_Token_for_matching_strings_that;
static PyObject *__pyx_kp_s_Token_for_matching_strings_that_2;
static PyObject *__pyx_kp_s_Token_for_matching_words_compose;
static PyObject *__pyx_kp_s_Token_for_matching_words_compose_2;
static PyObject *__pyx_kp_s_Token_for_skipping_over_all_unde;
static PyObject *__pyx_kp_s_Token_to_advance_to_a_specific_c;
static PyObject *__pyx_kp_s_Token_to_exactly_match_a_specifi;
static PyObject *__pyx_kp_s_Token_to_exactly_match_a_specifi_2;
static PyObject *__pyx_kp_s_Token_to_match_a_specified_strin;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_UnboundedCache;
static PyObject *__pyx_kp_s_Unmatchable_token;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_kp_s_W_s;
static PyObject *__pyx_kp_s_W_s_2;
static PyObject *__pyx_kp_s_W_s_3;
static PyObject *__pyx_kp_s_W_s_s;
static PyObject *__pyx_n_s_White;
static PyObject *__pyx_n_s_White___init;
static PyObject *__pyx_n_s_White___init___locals_genexpr;
static PyObject *__pyx_n_s_White_parseImpl;
static PyObject *__pyx_n_s_Word;
static PyObject *__pyx_n_s_WordEnd;
static PyObject *__pyx_n_s_WordEnd___init;
static PyObject *__pyx_n_s_WordEnd_parseImpl;
static PyObject *__pyx_n_s_WordRegex;
static PyObject *__pyx_n_s_WordRegex_parseImpl;
static PyObject *__pyx_n_s_WordStart;
static PyObject *__pyx_n_s_WordStart___init;
static PyObject *__pyx_n_s_WordStart_parseImpl;
static PyObject *__pyx_n_s_Word___init;
static PyObject *__pyx_n_s_Word___init___locals_genexpr;
static PyObject *__pyx_n_s_Word___str;
static PyObject *__pyx_n_s_Word___str___locals_charsAsStr;
static PyObject *__pyx_n_s_Word_parseImpl;
static PyObject *__pyx_kp_s_Wrapper_for_parse_actions_to_ens;
static PyObject *__pyx_kp_s_ZERO_WIDTH_SPACE;
static PyObject *__pyx_n_s_ZeroOrMore;
static PyObject *__pyx_n_s_ZeroOrMore___init;
static PyObject *__pyx_n_s_ZeroOrMore___str;
static PyObject *__pyx_n_s_ZeroOrMore_parseImpl;
static PyObject *__pyx_kp_s__10;
static PyObject *__pyx_kp_s__100;
static PyObject *__pyx_kp_s__101;
static PyObject *__pyx_kp_s__102;
static PyObject *__pyx_kp_s__103;
static PyObject *__pyx_kp_s__105;
static PyObject *__pyx_kp_s__107;
static PyObject *__pyx_kp_s__109;
static PyObject *__pyx_kp_s__110;
static PyObject *__pyx_kp_s__112;
static PyObject *__pyx_kp_s__113;
static PyObject *__pyx_kp_s__114;
static PyObject *__pyx_kp_s__115;
static PyObject *__pyx_kp_s__116;
static PyObject *__pyx_kp_s__117;
static PyObject *__pyx_kp_s__118;
static PyObject *__pyx_kp_s__12;
static PyObject *__pyx_kp_s__120;
static PyObject *__pyx_kp_s__14;
static PyObject *__pyx_kp_s__15;
static PyObject *__pyx_kp_s__16;
static PyObject *__pyx_kp_s__19;
static PyObject *__pyx_kp_s__2;
static PyObject *__pyx_kp_s__20;
static PyObject *__pyx_kp_s__21;
static PyObject *__pyx_kp_s__22;
static PyObject *__pyx_kp_s__25;
static PyObject *__pyx_kp_s__26;
static PyObject *__pyx_kp_s__268;
static PyObject *__pyx_kp_s__28;
static PyObject *__pyx_kp_s__29;
static PyObject *__pyx_kp_s__30;
static PyObject *__pyx_kp_s__33;
static PyObject *__pyx_kp_s__4;
static PyObject *__pyx_kp_s__456;
static PyObject *__pyx_kp_u__527;
static PyObject *__pyx_kp_u__528;
static PyObject *__pyx_kp_u__529;
static PyObject *__pyx_kp_u__530;
static PyObject *__pyx_kp_u__531;
static PyObject *__pyx_kp_u__532;
static PyObject *__pyx_kp_u__533;
static PyObject *__pyx_kp_u__534;
static PyObject *__pyx_kp_u__535;
static PyObject *__pyx_kp_u__536;
static PyObject *__pyx_kp_u__537;
static PyObject *__pyx_kp_u__538;
static PyObject *__pyx_kp_u__539;
static PyObject *__pyx_kp_u__540;
static PyObject *__pyx_kp_u__541;
static PyObject *__pyx_kp_u__542;
static PyObject *__pyx_kp_u__543;
static PyObject *__pyx_kp_u__544;
static PyObject *__pyx_kp_s__6;
static PyObject *__pyx_n_s__673;
static PyObject *__pyx_kp_s__79;
static PyObject *__pyx_kp_s__8;
static PyObject *__pyx_kp_u__80;
static PyObject *__pyx_kp_s__85;
static PyObject *__pyx_kp_s__9;
static PyObject *__pyx_kp_s__98;
static PyObject *__pyx_kp_s__99;
static PyObject *__pyx_n_s_add;
static PyObject *__pyx_n_s_addCondition;
static PyObject *__pyx_n_s_addParseAction;
static PyObject *__pyx_n_s_addoffset;
static PyObject *__pyx_n_s_adjacent;
static PyObject *__pyx_n_s_allResults;
static PyObject *__pyx_n_s_all_names;
static PyObject *__pyx_n_s_alphanums;
static PyObject *__pyx_n_s_alphas;
static PyObject *__pyx_kp_s_amp_gt_lt_quot_apos;
static PyObject *__pyx_n_s_aname;
static PyObject *__pyx_n_s_and;
static PyObject *__pyx_n_s_append;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_asDict;
static PyObject *__pyx_n_s_asGroupList;
static PyObject *__pyx_n_s_asKeyword;
static PyObject *__pyx_n_s_asList;
static PyObject *__pyx_n_s_asMatch;
static PyObject *__pyx_n_s_asStringList;
static PyObject *__pyx_n_s_asXML;
static PyObject *__pyx_n_s_ascii_lowercase;
static PyObject *__pyx_n_s_ascii_uppercase;
static PyObject *__pyx_kp_s_at_loc;
static PyObject *__pyx_kp_s_b;
static PyObject *__pyx_kp_s_b_s_b;
static PyObject *__pyx_kp_s_base_exception_class_for_all_par;
static PyObject *__pyx_n_s_best_expr;
static PyObject *__pyx_n_s_bodyChars;
static PyObject *__pyx_n_s_bodyCharsOrig;
static PyObject *__pyx_n_s_bodychars;
static PyObject *__pyx_n_s_bool;
static PyObject *__pyx_n_s_breakFlag;
static PyObject *__pyx_n_s_breaker;
static PyObject *__pyx_n_s_c;
static PyObject *__pyx_n_s_cache;
static PyObject *__pyx_n_s_cache_len;
static PyObject *__pyx_n_s_cache_size_limit;
static PyObject *__pyx_n_s_call;
static PyObject *__pyx_n_s_callDuringTry;
static PyObject *__pyx_n_s_callPreParse;
static PyObject *__pyx_n_s_callPreparse;
static PyObject *__pyx_n_s_callable;
static PyObject *__pyx_n_s_called;
static PyObject *__pyx_n_s_callers;
static PyObject *__pyx_n_s_canParseNext;
static PyObject *__pyx_kp_s_cannot_construct_And_with_sequen;
static PyObject *__pyx_kp_s_cannot_multiply_ParserElement_an;
static PyObject *__pyx_kp_s_cannot_multiply_ParserElement_an_2;
static PyObject *__pyx_kp_s_cannot_multiply_ParserElement_by;
static PyObject *__pyx_kp_s_cannot_multiply_ParserElement_by_2;
static PyObject *__pyx_kp_s_cannot_specify_a_minimum_length;
static PyObject *__pyx_kp_s_cannot_specify_a_minimum_length_2;
static PyObject *__pyx_kp_s_cannot_use_sub_with_Regex_asGrou;
static PyObject *__pyx_kp_s_cannot_use_sub_with_a_callable_w;
static PyObject *__pyx_n_s_caseless;
static PyObject *__pyx_n_s_caselessmatch;
static PyObject *__pyx_n_s_cause;
static PyObject *__pyx_n_s_chars;
static PyObject *__pyx_n_s_charsAsStr;
static PyObject *__pyx_n_s_charset;
static PyObject *__pyx_n_s_checkRecursion;
static PyObject *__pyx_n_s_check_ender;
static PyObject *__pyx_n_s_chr;
static PyObject *__pyx_n_s_class;
static PyObject *__pyx_n_s_clear;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_s_cls;
static PyObject *__pyx_n_s_co_name;
static PyObject *__pyx_n_s_code;
static PyObject *__pyx_n_s_col;
static PyObject *__pyx_n_s_collect_all_And_tokens;
static PyObject *__pyx_n_s_collections;
static PyObject *__pyx_n_s_collections_abc;
static PyObject *__pyx_n_s_colno;
static PyObject *__pyx_n_s_column;
static PyObject *__pyx_n_s_comment;
static PyObject *__pyx_n_s_comments;
static PyObject *__pyx_n_s_compile;
static PyObject *__pyx_n_s_conditionAsParseAction;
static PyObject *__pyx_n_s_contains;
static PyObject *__pyx_n_s_context;
static PyObject *__pyx_n_s_convertWhitespaceEscapes;
static PyObject *__pyx_n_s_copy;
static PyObject *__pyx_n_s_copyDefaultWhiteChars;
static PyObject *__pyx_n_s_count;
static PyObject *__pyx_n_s_cpy;
static PyObject *__pyx_n_s_d;
static PyObject *__pyx_kp_s_d_d;
static PyObject *__pyx_n_s_data;
static PyObject *__pyx_n_s_debug;
static PyObject *__pyx_n_s_debugActions;
static PyObject *__pyx_n_s_debugging;
static PyObject *__pyx_n_s_default;
static PyObject *__pyx_n_s_defaultExceptionDebugAction;
static PyObject *__pyx_n_s_defaultStartDebugAction;
static PyObject *__pyx_n_s_defaultSuccessDebugAction;
static PyObject *__pyx_n_s_defaultValue;
static PyObject *__pyx_n_s_defaultvalue;
static PyObject *__pyx_n_s_delitem;
static PyObject *__pyx_n_s_depth;
static PyObject *__pyx_n_s_depth_2;
static PyObject *__pyx_n_s_deque;
static PyObject *__pyx_n_s_diag;
static PyObject *__pyx_n_s_dictvalue;
static PyObject *__pyx_n_s_dir;
static PyObject *__pyx_n_s_doActions;
static PyObject *__pyx_n_s_doc;
static PyObject *__pyx_n_s_doctag;
static PyObject *__pyx_n_s_dummy;
static PyObject *__pyx_n_s_dump;
static PyObject *__pyx_n_s_e;
static PyObject *__pyx_n_s_elem;
static PyObject *__pyx_n_s_enable;
static PyObject *__pyx_n_s_enablePackrat;
static PyObject *__pyx_n_s_enable_all_warnings;
static PyObject *__pyx_n_s_enable_all_warnings_2;
static PyObject *__pyx_n_s_enable_debug_on_named_expression;
static PyObject *__pyx_n_s_end;
static PyObject *__pyx_n_s_endQuoteChar;
static PyObject *__pyx_n_s_endQuoteCharLen;
static PyObject *__pyx_kp_s_endQuoteChar_cannot_be_the_empty;
static PyObject *__pyx_n_s_ender;
static PyObject *__pyx_n_s_endloc;
static PyObject *__pyx_n_s_endswith;
static PyObject *__pyx_n_s_enter;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_n_s_eq;
static PyObject *__pyx_n_s_err;
static PyObject *__pyx_n_s_errmsg;
static PyObject *__pyx_n_s_error;
static PyObject *__pyx_n_s_errorStop;
static PyObject *__pyx_n_s_escChar;
static PyObject *__pyx_n_s_escCharReplacePattern;
static PyObject *__pyx_n_s_escQuote;
static PyObject *__pyx_n_s_escape;
static PyObject *__pyx_n_s_escapeRegexRangeChars;
static PyObject *__pyx_n_s_exact;
static PyObject *__pyx_n_s_exc;
static PyObject *__pyx_n_s_exceptionAction;
static PyObject *__pyx_kp_s_exception_raised_in_parse_action;
static PyObject *__pyx_kp_s_exception_thrown_by_class_Parser;
static PyObject *__pyx_n_s_excludeChars;
static PyObject *__pyx_n_s_exit;
static PyObject *__pyx_n_s_expand;
static PyObject *__pyx_n_s_expandtabs;
static PyObject *__pyx_n_s_explain;
static PyObject *__pyx_n_s_expr;
static PyObject *__pyx_n_s_expr1;
static PyObject *__pyx_n_s_expr_parse;
static PyObject *__pyx_n_s_exprs;
static PyObject *__pyx_n_s_exprsFound;
static PyObject *__pyx_n_s_exprtokens;
static PyObject *__pyx_n_s_extend;
static PyObject *__pyx_kp_s_f;
static PyObject *__pyx_n_s_f_2;
static PyObject *__pyx_n_s_f_code;
static PyObject *__pyx_n_s_f_locals;
static PyObject *__pyx_n_s_f_self;
static PyObject *__pyx_n_s_failAction;
static PyObject *__pyx_n_s_failOn;
static PyObject *__pyx_n_s_failed;
static PyObject *__pyx_n_s_failureTests;
static PyObject *__pyx_n_s_fatal;
static PyObject *__pyx_n_s_ff;
static PyObject *__pyx_n_s_file;
static PyObject *__pyx_n_s_file_contents;
static PyObject *__pyx_n_s_file_or_filename;
static PyObject *__pyx_n_s_finalResults;
static PyObject *__pyx_n_s_find;
static PyObject *__pyx_n_s_firstMatchChar;
static PyObject *__pyx_n_s_firstQuoteChar;
static PyObject *__pyx_n_s_flag;
static PyObject *__pyx_n_s_flags;
static PyObject *__pyx_n_s_flatten;
static PyObject *__pyx_n_s_fn;
static PyObject *__pyx_n_s_fns;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_s_formatted;
static PyObject *__pyx_kp_s_found_end_of_text;
static PyObject *__pyx_kp_s_found_r;
static PyObject *__pyx_n_s_foundstr;
static PyObject *__pyx_n_s_frm;
static PyObject *__pyx_n_s_from;
static PyObject *__pyx_n_s_from_dict;
static PyObject *__pyx_n_s_from_exception;
static PyObject *__pyx_n_s_from_symbols;
static PyObject *__pyx_n_s_full;
static PyObject *__pyx_n_s_fullDump;
static PyObject *__pyx_kp_s_g_1;
static PyObject *__pyx_n_s_generatorType;
static PyObject *__pyx_n_s_genexpr;
static PyObject *__pyx_n_s_get;
static PyObject *__pyx_n_s_getName;
static PyObject *__pyx_n_s_getattr;
static PyObject *__pyx_n_s_getinnerframes;
static PyObject *__pyx_n_s_getitem;
static PyObject *__pyx_n_s_getnewargs;
static PyObject *__pyx_n_s_getrecursionlimit;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_group;
static PyObject *__pyx_n_s_groupdict;
static PyObject *__pyx_n_s_groups;
static PyObject *__pyx_n_s_hasIgnoreExprs;
static PyObject *__pyx_n_s_hash;
static PyObject *__pyx_n_s_haskeys;
static PyObject *__pyx_n_s_hexnums;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_n_s_iadd;
static PyObject *__pyx_n_s_id;
static PyObject *__pyx_n_s_identChars;
static PyObject *__pyx_n_s_ignore;
static PyObject *__pyx_n_s_ignoreExpr;
static PyObject *__pyx_n_s_ignoreExprs;
static PyObject *__pyx_n_s_ikey;
static PyObject *__pyx_n_s_ilshift;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_inAccumNames;
static PyObject *__pyx_n_s_include;
static PyObject *__pyx_n_s_includeMatch;
static PyObject *__pyx_n_s_includeSeparators;
static PyObject *__pyx_n_s_include_list;
static PyObject *__pyx_n_s_indent;
static PyObject *__pyx_n_s_index;
static PyObject *__pyx_n_s_indices;
static PyObject *__pyx_n_s_init;
static PyObject *__pyx_n_s_initChars;
static PyObject *__pyx_n_s_initCharsOrig;
static PyObject *__pyx_n_s_initExprGroups;
static PyObject *__pyx_n_s_inlineLiteralsUsing;
static PyObject *__pyx_n_s_insStr;
static PyObject *__pyx_n_s_insert;
static PyObject *__pyx_n_s_inspect;
static PyObject *__pyx_n_s_instring;
static PyObject *__pyx_n_s_instring_slice;
static PyObject *__pyx_n_s_instrlen;
static PyObject *__pyx_kp_s_invalid_pattern_s_passed_to_Rege;
static PyObject *__pyx_n_s_invert;
static PyObject *__pyx_n_s_ior;
static PyObject *__pyx_n_s_is_iterable;
static PyObject *__pyx_n_s_isinstance;
static PyObject *__pyx_n_s_isspace;
static PyObject *__pyx_n_s_item;
static PyObject *__pyx_n_s_item_fn;
static PyObject *__pyx_n_s_itemgetter;
static PyObject *__pyx_n_s_items;
static PyObject *__pyx_n_s_itemseq;
static PyObject *__pyx_n_s_iter;
static PyObject *__pyx_n_s_iteritems;
static PyObject *__pyx_n_s_iteritems_2;
static PyObject *__pyx_n_s_iterkeys;
static PyObject *__pyx_n_s_iterkeys_2;
static PyObject *__pyx_n_s_itervalues;
static PyObject *__pyx_n_s_itervalues_2;
static PyObject *__pyx_n_s_ixor;
static PyObject *__pyx_n_s_j;
static PyObject *__pyx_n_s_join;
static PyObject *__pyx_n_s_joinString;
static PyObject *__pyx_kp_s_just_like_class_ParseFatalExcept;
static PyObject *__pyx_n_s_k;
static PyObject *__pyx_n_s_keepMatching;
static PyObject *__pyx_n_s_keepTabs;
static PyObject *__pyx_n_s_key;
static PyObject *__pyx_n_s_key_fifo;
static PyObject *__pyx_n_s_keys;
static PyObject *__pyx_n_s_kwargs;
static PyObject *__pyx_n_s_l;
static PyObject *__pyx_n_s_last;
static PyObject *__pyx_n_s_lastCR;
static PyObject *__pyx_n_s_lastE;
static PyObject *__pyx_n_s_last_expr;
static PyObject *__pyx_n_s_leaveWhitespace;
static PyObject *__pyx_n_s_len;
static PyObject *__pyx_n_s_line;
static PyObject *__pyx_n_s_line_column;
static PyObject *__pyx_n_s_line_str;
static PyObject *__pyx_n_s_lineno;
static PyObject *__pyx_kp_s_lineno_col_line;
static PyObject *__pyx_n_s_listAllMatches;
static PyObject *__pyx_n_s_literalStringClass;
static PyObject *__pyx_n_s_loc;
static PyObject *__pyx_n_s_loc1;
static PyObject *__pyx_n_s_loc2;
static PyObject *__pyx_n_s_longest;
static PyObject *__pyx_n_s_lookup;
static PyObject *__pyx_n_s_lookup_2;
static PyObject *__pyx_n_s_lshift;
static PyObject *__pyx_n_s_lstrip;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_makeOptionalList;
static PyObject *__pyx_n_s_map;
static PyObject *__pyx_n_s_markInputline;
static PyObject *__pyx_n_s_markerString;
static PyObject *__pyx_n_s_mat;
static PyObject *__pyx_n_s_match;
static PyObject *__pyx_n_s_matchLen;
static PyObject *__pyx_n_s_matchOrder;
static PyObject *__pyx_n_s_matchString;
static PyObject *__pyx_n_s_matchWhite;
static PyObject *__pyx_n_s_match_string;
static PyObject *__pyx_n_s_match_stringloc;
static PyObject *__pyx_n_s_matches;
static PyObject *__pyx_n_s_max;
static PyObject *__pyx_n_s_maxExcLoc;
static PyObject *__pyx_n_s_maxException;
static PyObject *__pyx_n_s_maxLen;
static PyObject *__pyx_n_s_maxMatches;
static PyObject *__pyx_n_s_maxMismatches;
static PyObject *__pyx_n_s_maxSpecified;
static PyObject *__pyx_n_s_maxlen;
static PyObject *__pyx_n_s_maxloc;
static PyObject *__pyx_n_s_maxsize;
static PyObject *__pyx_n_s_maxsplit;
static PyObject *__pyx_n_s_mayIndexError;
static PyObject *__pyx_n_s_mayReturnEmpty;
static PyObject *__pyx_n_s_message;
static PyObject *__pyx_n_s_metaclass;
static PyObject *__pyx_n_s_methodCall;
static PyObject *__pyx_n_s_min;
static PyObject *__pyx_n_s_minElements;
static PyObject *__pyx_n_s_minLen;
static PyObject *__pyx_n_s_mismatches;
static PyObject *__pyx_n_s_missing;
static PyObject *__pyx_n_s_modal;
static PyObject *__pyx_n_s_modalResults;
static PyObject *__pyx_n_s_module;
static PyObject *__pyx_kp_s_module_2;
static PyObject *__pyx_n_s_msg;
static PyObject *__pyx_n_s_mul;
static PyObject *__pyx_n_s_multiline;
static PyObject *__pyx_n_s_multioptionals;
static PyObject *__pyx_n_s_multirequired;
static PyObject *__pyx_n_s_must_skip;
static PyObject *__pyx_n_s_mylen;
static PyObject *__pyx_kp_s_n;
static PyObject *__pyx_n_s_n_2;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_name_2;
static PyObject *__pyx_n_s_namedItems;
static PyObject *__pyx_n_s_namedItemsOnly;
static PyObject *__pyx_n_s_ne;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_n_s_newloc;
static PyObject *__pyx_n_s_newself;
static PyObject *__pyx_n_s_nextCR;
static PyObject *__pyx_n_s_nextLevelIndent;
static PyObject *__pyx_n_s_nextLoc;
static PyObject *__pyx_n_s_nextloc;
static PyObject *__pyx_n_s_nl;
static PyObject *__pyx_n_s_nm;
static PyObject *__pyx_kp_s_no_defined_alternatives_to_match;
static PyObject *__pyx_n_s_nonzero;
static PyObject *__pyx_n_s_notChars;
static PyObject *__pyx_n_s_not_ender;
static PyObject *__pyx_n_s_not_in_cache;
static PyObject *__pyx_kp_s_not_preceded_by;
static PyObject *__pyx_n_s_notchars;
static PyObject *__pyx_n_s_nullDebugAction;
static PyObject *__pyx_kp_s_null_string_passed_to_Keyword_us;
static PyObject *__pyx_kp_s_null_string_passed_to_Literal_us;
static PyObject *__pyx_kp_s_null_string_passed_to_Regex_use;
static PyObject *__pyx_n_s_nums;
static PyObject *__pyx_n_s_o;
static PyObject *__pyx_n_s_obj;
static PyObject *__pyx_n_s_object;
static PyObject *__pyx_n_s_occurrences;
static PyObject *__pyx_n_s_offset;
static PyObject *__pyx_kp_s_only_1_or_2_index_arguments_supp;
static PyObject *__pyx_n_s_open;
static PyObject *__pyx_n_s_operator;
static PyObject *__pyx_n_s_opt1;
static PyObject *__pyx_n_s_opt1map;
static PyObject *__pyx_n_s_opt2;
static PyObject *__pyx_n_s_optElements;
static PyObject *__pyx_n_s_optionals;
static PyObject *__pyx_n_s_or;
static PyObject *__pyx_n_s_ordereddict;
static PyObject *__pyx_n_s_original;
static PyObject *__pyx_n_s_originalParseMethod;
static PyObject *__pyx_n_s_other;
static PyObject *__pyx_n_s_otherdictitems;
static PyObject *__pyx_n_s_otheritems;
static PyObject *__pyx_n_s_out;
static PyObject *__pyx_n_s_overlap;
static PyObject *__pyx_n_s_p1;
static PyObject *__pyx_n_s_p2;
static PyObject *__pyx_n_s_pa;
static PyObject *__pyx_n_s_packratEnabled;
static PyObject *__pyx_n_s_packrat_cache;
static PyObject *__pyx_n_s_packrat_cache_lock;
static PyObject *__pyx_n_s_packrat_cache_stats;
static PyObject *__pyx_n_s_par;
static PyObject *__pyx_n_s_parse;
static PyObject *__pyx_n_s_parseAction;
static PyObject *__pyx_n_s_parseAll;
static PyObject *__pyx_n_s_parseCache;
static PyObject *__pyx_n_s_parseElementList;
static PyObject *__pyx_n_s_parseElementTrace;
static PyObject *__pyx_n_s_parseFile;
static PyObject *__pyx_n_s_parseFn;
static PyObject *__pyx_n_s_parseImpl;
static PyObject *__pyx_n_s_parseImplAsGroupList;
static PyObject *__pyx_n_s_parseImplAsMatch;
static PyObject *__pyx_n_s_parseMethod;
static PyObject *__pyx_n_s_parseNoCache;
static PyObject *__pyx_n_s_parseString;
static PyObject *__pyx_n_s_parseWithTabs;
static PyObject *__pyx_n_s_parse_action_exc;
static PyObject *__pyx_kp_s_parse_actions_must_be_callable;
static PyObject *__pyx_n_s_parserElement;
static PyObject *__pyx_n_s_pattern;
static PyObject *__pyx_n_s_pbe;
static PyObject *__pyx_n_s_pdb;
static PyObject *__pyx_n_s_pe;
static PyObject *__pyx_n_s_pop;
static PyObject *__pyx_kp_s_pop_got_an_unexpected_keyword_ar;
static PyObject *__pyx_n_s_popitem;
static PyObject *__pyx_n_s_popleft;
static PyObject *__pyx_n_s_position;
static PyObject *__pyx_n_s_postParse;
static PyObject *__pyx_n_s_pp_value;
static PyObject *__pyx_n_s_pprint;
static PyObject *__pyx_n_s_preParse;
static PyObject *__pyx_n_s_preloc;
static PyObject *__pyx_n_s_prepare;
static PyObject *__pyx_n_s_preparseFn;
static PyObject *__pyx_n_s_print;
static PyObject *__pyx_n_s_printResults;
static PyObject *__pyx_n_s_print_2;
static PyObject *__pyx_n_s_printable;
static PyObject *__pyx_n_s_printables;
static PyObject *__pyx_n_s_pstr;
static PyObject *__pyx_n_s_pyparsing;
static PyObject *__pyx_n_s_qualname;
static PyObject *__pyx_n_s_quoteChar;
static PyObject *__pyx_n_s_quoteCharLen;
static PyObject *__pyx_kp_s_quoteChar_cannot_be_the_empty_st;
static PyObject *__pyx_n_s_quotedString;
static PyObject *__pyx_kp_s_quoted_string_starting_with_s_en;
static PyObject *__pyx_n_s_r;
static PyObject *__pyx_kp_s_r_2;
static PyObject *__pyx_kp_s_r_object_is_not_iterable;
static PyObject *__pyx_n_s_radd;
static PyObject *__pyx_n_s_rand;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_re;
static PyObject *__pyx_n_s_reString;
static PyObject *__pyx_n_s_re_match;
static PyObject *__pyx_n_s_read;
static PyObject *__pyx_n_s_ref;
static PyObject *__pyx_n_s_remove;
static PyObject *__pyx_n_s_removed;
static PyObject *__pyx_n_s_repl;
static PyObject *__pyx_n_s_replace;
static PyObject *__pyx_n_s_replaceWith;
static PyObject *__pyx_n_s_repr;
static PyObject *__pyx_n_s_req;
static PyObject *__pyx_n_s_required;
static PyObject *__pyx_n_s_res;
static PyObject *__pyx_n_s_resTag;
static PyObject *__pyx_n_s_reset;
static PyObject *__pyx_n_s_resetCache;
static PyObject *__pyx_n_s_result;
static PyObject *__pyx_n_s_resultlist;
static PyObject *__pyx_n_s_results;
static PyObject *__pyx_n_s_resultsName;
static PyObject *__pyx_n_s_ret;
static PyObject *__pyx_n_s_retString;
static PyObject *__pyx_n_s_retTokens;
static PyObject *__pyx_n_s_retToks;
static PyObject *__pyx_n_s_retobj;
static PyObject *__pyx_n_s_retreat;
static PyObject *__pyx_n_s_returnString;
static PyObject *__pyx_n_s_reverse;
static PyObject *__pyx_n_s_reversed;
static PyObject *__pyx_n_s_rfind;
static PyObject *__pyx_n_s_rmul;
static PyObject *__pyx_n_s_rne;
static PyObject *__pyx_n_s_ror;
static PyObject *__pyx_n_s_rstrip;
static PyObject *__pyx_n_s_rsub;
static PyObject *__pyx_n_s_runTests;
static PyObject *__pyx_n_s_rxor;
static PyObject *__pyx_kp_s_s;
static PyObject *__pyx_kp_s_s_2;
static PyObject *__pyx_kp_s_s_3;
static PyObject *__pyx_n_s_s_4;
static PyObject *__pyx_kp_s_s_5;
static PyObject *__pyx_kp_s_s_6;
static PyObject *__pyx_kp_s_s_7;
static PyObject *__pyx_kp_s_s_8;
static PyObject *__pyx_n_s_s_m;
static PyObject *__pyx_kp_s_s_s;
static PyObject *__pyx_kp_s_s_s_2;
static PyObject *__pyx_kp_s_s_s_3;
static PyObject *__pyx_kp_s_s_s_4;
static PyObject *__pyx_kp_s_s_s_5;
static PyObject *__pyx_kp_s_s_s_at_char_d_line_d_col_d;
static PyObject *__pyx_kp_s_s_s_d_s_s_s;
static PyObject *__pyx_kp_s_s_s_n_r_s;
static PyObject *__pyx_kp_s_s_s_s;
static PyObject *__pyx_kp_s_s_s_s_2;
static PyObject *__pyx_n_s_saveAsList;
static PyObject *__pyx_n_s_savelist;
static PyObject *__pyx_n_s_scanString;
static PyObject *__pyx_n_s_se;
static PyObject *__pyx_n_s_searchString;
static PyObject *__pyx_kp_s_second_tuple_value_must_be_great;
static PyObject *__pyx_n_s_seen;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_n_s_selfTag;
static PyObject *__pyx_n_s_self_expr_parse;
static PyObject *__pyx_n_s_self_failOn_canParseNext;
static PyObject *__pyx_n_s_self_ignoreExpr_tryParse;
static PyObject *__pyx_n_s_self_skip_ignorables;
static PyObject *__pyx_n_s_self_type;
static PyObject *__pyx_n_s_send;
static PyObject *__pyx_n_s_sep;
static PyObject *__pyx_n_s_set;
static PyObject *__pyx_n_s_setBreak;
static PyObject *__pyx_n_s_setDebug;
static PyObject *__pyx_n_s_setDebugActions;
static PyObject *__pyx_n_s_setDefaultKeywordChars;
static PyObject *__pyx_n_s_setDefaultWhitespaceChars;
static PyObject *__pyx_n_s_setFailAction;
static PyObject *__pyx_n_s_setName;
static PyObject *__pyx_n_s_setOffset;
static PyObject *__pyx_n_s_setParseAction;
static PyObject *__pyx_n_s_setResultsName;
static PyObject *__pyx_n_s_setResultsName_2;
static PyObject *__pyx_n_s_setWhitespaceChars;
static PyObject *__pyx_n_s_set_trace;
static PyObject *__pyx_n_s_setitem;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_size;
static PyObject *__pyx_n_s_skipIgnorables;
static PyObject *__pyx_n_s_skipWhitespace;
static PyObject *__pyx_kp_s_skipped;
static PyObject *__pyx_n_s_skipresult;
static PyObject *__pyx_n_s_skiptext;
static PyObject *__pyx_n_s_skipto_arg;
static PyObject *__pyx_n_s_sort;
static PyObject *__pyx_n_s_split;
static PyObject *__pyx_n_s_splitlines;
static PyObject *__pyx_n_s_splits;
static PyObject *__pyx_n_s_src;
static PyObject *__pyx_n_s_sre_constants;
static PyObject *__pyx_n_s_stacklevel;
static PyObject *__pyx_n_s_start;
static PyObject *__pyx_n_s_startAction;
static PyObject *__pyx_n_s_startloc;
static PyObject *__pyx_n_s_startswith;
static PyObject *__pyx_n_s_state;
static PyObject *__pyx_n_s_staticmethod;
static PyObject *__pyx_n_s_stdout;
static PyObject *__pyx_n_s_stopOn;
static PyObject *__pyx_n_s_str;
static PyObject *__pyx_n_s_strRepr;
static PyObject *__pyx_n_s_streamline;
static PyObject *__pyx_n_s_streamlined;
static PyObject *__pyx_n_s_strg;
static PyObject *__pyx_n_s_string;
static PyObject *__pyx_n_s_strip;
static PyObject *__pyx_n_s_sub;
static PyObject *__pyx_n_s_subRecCheckList;
static PyObject *__pyx_n_s_sub_2;
static PyObject *__pyx_n_s_success;
static PyObject *__pyx_n_s_successAction;
static PyObject *__pyx_n_s_sum;
static PyObject *__pyx_n_s_super;
static PyObject *__pyx_n_s_suppress;
static PyObject *__pyx_n_s_sys;
static PyObject *__pyx_n_s_system_version;
static PyObject *__pyx_kp_s_t;
static PyObject *__pyx_n_s_t_2;
static PyObject *__pyx_n_s_tb;
static PyObject *__pyx_n_s_tb_next;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_testString;
static PyObject *__pyx_n_s_test_expr;
static PyObject *__pyx_n_s_tests;
static PyObject *__pyx_n_s_thiscol;
static PyObject *__pyx_n_s_thread;
static PyObject *__pyx_n_s_threading;
static PyObject *__pyx_n_s_throw;
static PyObject *__pyx_n_s_throwException;
static PyObject *__pyx_n_s_tmp;
static PyObject *__pyx_n_s_tmpExprs;
static PyObject *__pyx_n_s_tmpLoc;
static PyObject *__pyx_n_s_tmpOpt;
static PyObject *__pyx_n_s_tmpReqd;
static PyObject *__pyx_n_s_tmploc;
static PyObject *__pyx_n_s_tmptokens;
static PyObject *__pyx_n_s_to;
static PyObject *__pyx_n_s_toItem;
static PyObject *__pyx_n_s_to_symbols;
static PyObject *__pyx_n_s_tok;
static PyObject *__pyx_n_s_tokenlist;
static PyObject *__pyx_n_s_tokens;
static PyObject *__pyx_n_s_tokensStart;
static PyObject *__pyx_n_s_toklist;
static PyObject *__pyx_n_s_toks;
static PyObject *__pyx_n_s_traceback;
static PyObject *__pyx_n_s_transformString;
static PyObject *__pyx_n_s_trim_arity;
static PyObject *__pyx_n_s_trim_traceback;
static PyObject *__pyx_n_s_tryParse;
static PyObject *__pyx_n_s_try_not_ender;
static PyObject *__pyx_n_s_tup;
static PyObject *__pyx_n_s_types;
static PyObject *__pyx_n_s_unquoteResults;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_n_s_upper;
static PyObject *__pyx_kp_s_user_throwable_exception_thrown;
static PyObject *__pyx_n_s_v;
static PyObject *__pyx_n_s_validate;
static PyObject *__pyx_n_s_validateTrace;
static PyObject *__pyx_n_s_value;
static PyObject *__pyx_n_s_values;
static PyObject *__pyx_n_s_vars;
static PyObject *__pyx_n_s_verbose_stacktrace;
static PyObject *__pyx_n_s_version_info;
static PyObject *__pyx_n_s_vlist;
static PyObject *__pyx_n_s_vv;
static PyObject *__pyx_n_s_warn;
static PyObject *__pyx_n_s_warn_2;
static PyObject *__pyx_n_s_warn_multiple_tokens_in_named_al;
static PyObject *__pyx_n_s_warn_name_set_on_empty_Forward;
static PyObject *__pyx_n_s_warn_on_multiple_string_args_to;
static PyObject *__pyx_n_s_warn_ungrouped_named_tokens_in_c;
static PyObject *__pyx_n_s_warnings;
static PyObject *__pyx_n_s_weakref;
static PyObject *__pyx_n_s_whiteChars;
static PyObject *__pyx_n_s_whiteStrs;
static PyObject *__pyx_n_s_whitespace;
static PyObject *__pyx_n_s_wkref;
static PyObject *__pyx_n_s_wordChars;
static PyObject *__pyx_n_s_wrapper;
static PyObject *__pyx_n_s_write;
static PyObject *__pyx_n_s_ws;
static PyObject *__pyx_n_s_ws_map;
static PyObject *__pyx_n_s_wschar;
static PyObject *__pyx_n_s_wslit;
static PyObject *__pyx_n_s_wt;
static PyObject *__pyx_n_s_xmlBodyText;
static PyObject *__pyx_n_s_xml_escape;
static PyObject *__pyx_n_s_xml_escape_locals_genexpr;
static PyObject *__pyx_n_s_xor;
static PyObject *__pyx_n_s_zip;
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_22genexpr(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_25genexpr(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension__enable_all_warnings(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_2_defaultStartDebugAction(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, PyObject *__pyx_v_expr); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4_defaultSuccessDebugAction(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_instring, CYTHON_UNUSED PyObject *__pyx_v_startloc, CYTHON_UNUSED PyObject *__pyx_v_endloc, PyObject *__pyx_v_expr, PyObject *__pyx_v_toks); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_6_defaultExceptionDebugAction(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_instring, CYTHON_UNUSED PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_expr, PyObject *__pyx_v_exc); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_8nullDebugAction(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_args); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_escapeRegexRangeChars(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_s); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12_flatten(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_L); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_11_xml_escape_genexpr(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_14_xml_escape(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_data); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_16col(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_loc, PyObject *__pyx_v_strg); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_18lineno(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_loc, PyObject *__pyx_v_strg); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_20line(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_loc, PyObject *__pyx_v_strg); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_23_ParseResultsWithOffset___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_p1, PyObject *__pyx_v_p2); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_23_ParseResultsWithOffset_2__getitem__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_23_ParseResultsWithOffset_4__repr__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_23_ParseResultsWithOffset_6setOffset(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults___new__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_cls, PyObject *__pyx_v_toklist, CYTHON_UNUSED PyObject *__pyx_v_name, CYTHON_UNUSED PyObject *__pyx_v_asList, CYTHON_UNUSED PyObject *__pyx_v_modal); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_28__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_2__init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_toklist, PyObject *__pyx_v_name, PyObject *__pyx_v_asList, PyObject *__pyx_v_modal, PyObject *__pyx_v_isinstance); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_4__getitem__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_30__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_6__setitem__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_k, PyObject *__pyx_v_v, PyObject *__pyx_v_isinstance); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_8__delitem__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_10__contains__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_k); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_12__len__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_14__bool__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_16__iter__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_18__reversed__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_20_iterkeys(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_11_itervalues_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_22_itervalues(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_10_iteritems_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_24_iteritems(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_26keys(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_28values(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_30items(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_32haskeys(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_34pop(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_36get(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_defaultValue); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_38insert(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_insStr); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_40append(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_item); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_42extend(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_itemseq); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_44clear(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_46__getattr__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_48__add__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda5(PyObject *__pyx_self, PyObject *__pyx_v_a); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_50__iadd__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_52__radd__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_54__repr__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_7__str___genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_56__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_58_asStringList(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_sep); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_60asList(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_6asDict_toItem(PyObject *__pyx_self, PyObject *__pyx_v_obj); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_6asDict_2genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_62asDict(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_64copy(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_5asXML_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_66asXML(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_doctag, PyObject *__pyx_v_namedItemsOnly, PyObject *__pyx_v_indent, PyObject *__pyx_v_formatted); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_68__lookup(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_sub); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_70getName(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_4dump_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_4dump_3genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_72dump(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_indent, PyObject *__pyx_v_full, PyObject *__pyx_v_include_list, PyObject *__pyx_v__depth); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_74pprint(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_76__getstate__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_78__setstate__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_state); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_80__getnewargs__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_82__dir__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_9from_dict_is_iterable(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_84from_dict(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_cls, PyObject *__pyx_v_other, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_pstr, PyObject *__pyx_v_loc, PyObject *__pyx_v_msg, PyObject *__pyx_v_elem); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_2_from_exception(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_cls, PyObject *__pyx_v_pe); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_4__getattr__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_aname); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_6__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_8__repr__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_10markInputline(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_markerString); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_12__dir__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_14ParseException_explain(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_exc, PyObject *__pyx_v_depth); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_25RecursiveGrammarException___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parseElementList); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_25RecursiveGrammarException_2__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_setDefaultWhitespaceChars(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_chars); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_2inlineLiteralsUsing(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_cls); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_4_trim_traceback(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_cls, PyObject *__pyx_v_tb); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_6__init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_savelist); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_8copy(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_10setName(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_12setResultsName(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_listAllMatches); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_14_setResultsName(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_listAllMatches); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_8setBreak_breaker(PyObject *__pyx_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, PyObject *__pyx_v_doActions, PyObject *__pyx_v_callPreParse); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_16setBreak(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_breakFlag); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_14setParseAction_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_18setParseAction(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_fns, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_20addParseAction(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_fns, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_22addCondition(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_fns, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_24setFailAction(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_fn); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_26_skipIgnorables(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_28preParse(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_30parseImpl(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_32postParse(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_instring, CYTHON_UNUSED PyObject *__pyx_v_loc, PyObject *__pyx_v_tokenlist); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_34_parseNoCache(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, PyObject *__pyx_v_doActions, PyObject *__pyx_v_callPreParse); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_36tryParse(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_38canParseNext(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15_UnboundedCache_8__init___get(PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15_UnboundedCache_8__init___2set(PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15_UnboundedCache_8__init___4clear(PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15_UnboundedCache_8__init___6cache_len(PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15_UnboundedCache___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___get(PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___2set(PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___4clear(PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___6cache_len(PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_size); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___8get(PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___10set(PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___12clear(PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___14cache_len(PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_2__init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_size); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_40_parseCache(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, PyObject *__pyx_v_doActions, PyObject *__pyx_v_callPreParse); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_42resetCache(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_44enablePackrat(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_cache_size_limit); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_46parseString(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_parseAll); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_32__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_48scanString(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_maxMatches, PyObject *__pyx_v_overlap); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_51transformString(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_34__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_53searchString(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_maxMatches); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_36__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_55split(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_maxsplit, PyObject *__pyx_v_includeSeparators); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_58__add__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_60__radd__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_62__sub__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_64__rsub__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_7__mul___makeOptionalList(PyObject *__pyx_self, PyObject *__pyx_v_n); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_7__mul___2genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_66__mul__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_68__rmul__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_70__or__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_72__ror__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_74__xor__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_76__rxor__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_78__and__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_80__rand__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_82__invert__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_84__iter__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_86__getitem__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_88__call__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_90suppress(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_92leaveWhitespace(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_94setWhitespaceChars(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_chars); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_96parseWithTabs(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_98ignore(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_100setDebugActions(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_startAction, PyObject *__pyx_v_successAction, PyObject *__pyx_v_exceptionAction); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_102setDebug(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_flag); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_104__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_106__repr__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_108streamline(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_110checkRecursion(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_parseElementList); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_112validate(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_validateTrace); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_114parseFile(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_file_or_filename, PyObject *__pyx_v_parseAll); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_116__eq__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_118__ne__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_120__hash__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_122__req__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_124__rne__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_126matches(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_testString, PyObject *__pyx_v_parseAll); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_128runTests(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_tests, PyObject *__pyx_v_parseAll, PyObject *__pyx_v_comment, PyObject *__pyx_v_fullDump, PyObject *__pyx_v_printResults, PyObject *__pyx_v_failureTests, PyObject *__pyx_v_postParse, PyObject *__pyx_v_file); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_8__init___genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_8__init___3genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_exprs, PyObject *__pyx_v_savelist); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_2append(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_4leaveWhitespace(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_6ignore(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_8__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_10streamline(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_12validate(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_validateTrace); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_14copy(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_16_setResultsName(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_listAllMatches); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5Token___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5Empty___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7NoMatch___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7NoMatch_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Literal___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_matchString); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Literal_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_18_SingleCharLiteral_parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Keyword___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_matchString, PyObject *__pyx_v_identChars, PyObject *__pyx_v_caseless); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Keyword_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Keyword_4copy(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Keyword_6setDefaultKeywordChars(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_chars); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15CaselessLiteral___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_matchString); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15CaselessLiteral_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15CaselessKeyword___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_matchString, PyObject *__pyx_v_identChars); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10CloseMatch___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_match_string, PyObject *__pyx_v_maxMismatches); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10CloseMatch_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Word_8__init___genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Word_8__init___3genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Word___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_initChars, PyObject *__pyx_v_bodyChars, PyObject *__pyx_v_min, PyObject *__pyx_v_max, PyObject *__pyx_v_exact, PyObject *__pyx_v_asKeyword, PyObject *__pyx_v_excludeChars); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Word_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Word_7__str___charsAsStr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_s); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Word_4__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_WordRegex_parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Char___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_charset, PyObject *__pyx_v_asKeyword, PyObject *__pyx_v_excludeChars); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5Regex___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_pattern, PyObject *__pyx_v_flags, PyObject *__pyx_v_asGroupList, PyObject *__pyx_v_asMatch); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5Regex_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5Regex_4parseImplAsGroupList(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5Regex_6parseImplAsMatch(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5Regex_8__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5Regex_3sub_pa(PyObject *__pyx_self, PyObject *__pyx_v_tokens); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5Regex_3sub_2pa(PyObject *__pyx_self, PyObject *__pyx_v_tokens); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5Regex_10sub(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_repl); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12QuotedString_8__init___genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12QuotedString___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_quoteChar, PyObject *__pyx_v_escChar, PyObject *__pyx_v_escQuote, PyObject *__pyx_v_multiline, PyObject *__pyx_v_unquoteResults, PyObject *__pyx_v_endQuoteChar, PyObject *__pyx_v_convertWhitespaceEscapes); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12QuotedString_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12QuotedString_4__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10CharsNotIn___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_notChars, PyObject *__pyx_v_min, PyObject *__pyx_v_max, PyObject *__pyx_v_exact); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10CharsNotIn_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10CharsNotIn_4__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5White_8__init___genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5White_8__init___3genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5White___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_ws, PyObject *__pyx_v_min, PyObject *__pyx_v_max, PyObject *__pyx_v_exact); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5White_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_14_PositionToken___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10GoToColumn___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_colno); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10GoToColumn_2preParse(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10GoToColumn_4parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_9LineStart___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_9LineStart_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7LineEnd___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7LineEnd_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_11StringStart___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_11StringStart_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_9StringEnd___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_9StringEnd_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_38__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_9WordStart___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_wordChars); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_9WordStart_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_40__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7WordEnd___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_wordChars); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7WordEnd_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_ErrorStop___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_3And_8__init___genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_3And___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_exprs, PyObject *__pyx_v_savelist); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_3And_10streamline_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_3And_10streamline_3genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_3And_2streamline(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_3And_4parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, PyObject *__pyx_v_doActions); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_3And_6__iadd__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_3And_8checkRecursion(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parseElementList); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_3And_7__str___genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_3And_10__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_2Or_8__init___genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_2Or___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_exprs, PyObject *__pyx_v_savelist); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_2Or_10streamline_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_2Or_2streamline(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_2Or_4parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, PyObject *__pyx_v_doActions); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_2Or_6__ixor__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_2Or_7__str___genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_2Or_8__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_2Or_10checkRecursion(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parseElementList); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_2Or_15_setResultsName_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_2Or_12_setResultsName(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_listAllMatches); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_8__init___genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_exprs, PyObject *__pyx_v_savelist); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_10streamline_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_2streamline(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_4parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, PyObject *__pyx_v_doActions); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_6__ior__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_7__str___genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_8__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_10checkRecursion(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parseElementList); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_15_setResultsName_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_12_setResultsName(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_listAllMatches); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Each_8__init___genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Each___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_exprs, PyObject *__pyx_v_savelist); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Each_10streamline_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Each_2streamline(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Each_9parseImpl_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Each_9parseImpl_3genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Each_4parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, PyObject *__pyx_v_doActions); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Each_7__str___genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Each_6__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Each_8checkRecursion(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parseElementList); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_expr, PyObject *__pyx_v_savelist); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, PyObject *__pyx_v_doActions); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_4leaveWhitespace(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_6ignore(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_8streamline(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_10checkRecursion(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parseElementList); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_12validate(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_validateTrace); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_14__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10FollowedBy___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_expr); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10FollowedBy_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, PyObject *__pyx_v_doActions); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda37(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_s, CYTHON_UNUSED PyObject *__pyx_v_l, PyObject *__pyx_v_t); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10PrecededBy___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_expr, PyObject *__pyx_v_retreat); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10PrecededBy_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_6NotAny___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_expr); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_6NotAny_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_6NotAny_4__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_14_MultipleMatch___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_expr, PyObject *__pyx_v_stopOn); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_14_MultipleMatch_2stopOn(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_ender); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_14_MultipleMatch_4parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, PyObject *__pyx_v_doActions); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_14_MultipleMatch_6_setResultsName(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_listAllMatches); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_9OneOrMore___str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10ZeroOrMore___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_expr, PyObject *__pyx_v_stopOn); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10ZeroOrMore_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, PyObject *__pyx_v_doActions); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10ZeroOrMore_4__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_NullToken___bool__(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_NullToken_2__str__(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_42__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_8Optional___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_expr, PyObject *__pyx_v_default); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_8Optional_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, PyObject *__pyx_v_doActions); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_8Optional_4__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_6SkipTo___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other, PyObject *__pyx_v_include, PyObject *__pyx_v_ignore, PyObject *__pyx_v_failOn); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_6SkipTo_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, PyObject *__pyx_v_doActions); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Forward___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Forward_2__lshift__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Forward_4__ilshift__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Forward_6leaveWhitespace(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Forward_8streamline(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Forward_10validate(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_validateTrace); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Forward_12__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Forward_14copy(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Forward_16_setResultsName(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_listAllMatches); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_14TokenConverter___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_expr, CYTHON_UNUSED PyObject *__pyx_v_savelist); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Combine___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_expr, PyObject *__pyx_v_joinString, PyObject *__pyx_v_adjacent); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Combine_2ignore(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Combine_4postParse(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_instring, CYTHON_UNUSED PyObject *__pyx_v_loc, PyObject *__pyx_v_tokenlist); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5Group___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_expr); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5Group_2postParse(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_instring, CYTHON_UNUSED PyObject *__pyx_v_loc, PyObject *__pyx_v_tokenlist); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Dict___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_expr); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Dict_2postParse(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_instring, CYTHON_UNUSED PyObject *__pyx_v_loc, PyObject *__pyx_v_tokenlist); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_8Suppress_postParse(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_instring, CYTHON_UNUSED PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_tokenlist); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_8Suppress_2suppress(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_8OnlyOnce___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_methodCall); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_8OnlyOnce_2__call__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_s, PyObject *__pyx_v_l, PyObject *__pyx_v_t); /* proto */
static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_8OnlyOnce_4reset(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct__genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_7___iadd__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_get = {0, &__pyx_n_s_get, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_items = {0, &__pyx_n_s_items, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_pop = {0, &__pyx_n_s_pop, 0, 0, 0};
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_3;
static PyObject *__pyx_int_4;
static PyObject *__pyx_int_5;
static PyObject *__pyx_int_16;
static PyObject *__pyx_int_92;
static PyObject *__pyx_int_128;
static PyObject *__pyx_int_1000;
static PyObject *__pyx_int_neg_1;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_slice__11;
static PyObject *__pyx_slice__13;
static PyObject *__pyx_slice__34;
static PyObject *__pyx_slice__66;
static PyObject *__pyx_slice__69;
static PyObject *__pyx_slice__71;
static PyObject *__pyx_slice__78;
static PyObject *__pyx_slice__84;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__27;
static PyObject *__pyx_tuple__31;
static PyObject *__pyx_tuple__32;
static PyObject *__pyx_tuple__35;
static PyObject *__pyx_tuple__37;
static PyObject *__pyx_tuple__38;
static PyObject *__pyx_tuple__39;
static PyObject *__pyx_tuple__41;
static PyObject *__pyx_tuple__43;
static PyObject *__pyx_tuple__45;
static PyObject *__pyx_tuple__47;
static PyObject *__pyx_tuple__49;
static PyObject *__pyx_tuple__51;
static PyObject *__pyx_tuple__53;
static PyObject *__pyx_tuple__55;
static PyObject *__pyx_tuple__57;
static PyObject *__pyx_tuple__59;
static PyObject *__pyx_tuple__61;
static PyObject *__pyx_tuple__65;
static PyObject *__pyx_tuple__67;
static PyObject *__pyx_tuple__68;
static PyObject *__pyx_tuple__70;
static PyObject *__pyx_tuple__72;
static PyObject *__pyx_tuple__73;
static PyObject *__pyx_tuple__74;
static PyObject *__pyx_tuple__75;
static PyObject *__pyx_tuple__76;
static PyObject *__pyx_tuple__81;
static PyObject *__pyx_tuple__82;
static PyObject *__pyx_tuple__83;
static PyObject *__pyx_tuple__86;
static PyObject *__pyx_tuple__88;
static PyObject *__pyx_tuple__89;
static PyObject *__pyx_tuple__90;
static PyObject *__pyx_tuple__91;
static PyObject *__pyx_tuple__92;
static PyObject *__pyx_tuple__94;
static PyObject *__pyx_tuple__96;
static PyObject *__pyx_tuple__97;
static PyObject *__pyx_slice__119;
static PyObject *__pyx_tuple__104;
static PyObject *__pyx_tuple__106;
static PyObject *__pyx_tuple__108;
static PyObject *__pyx_tuple__111;
static PyObject *__pyx_tuple__122;
static PyObject *__pyx_tuple__124;
static PyObject *__pyx_tuple__126;
static PyObject *__pyx_tuple__128;
static PyObject *__pyx_tuple__130;
static PyObject *__pyx_tuple__132;
static PyObject *__pyx_tuple__134;
static PyObject *__pyx_tuple__136;
static PyObject *__pyx_tuple__138;
static PyObject *__pyx_tuple__140;
static PyObject *__pyx_tuple__142;
static PyObject *__pyx_tuple__143;
static PyObject *__pyx_tuple__145;
static PyObject *__pyx_tuple__147;
static PyObject *__pyx_tuple__149;
static PyObject *__pyx_tuple__151;
static PyObject *__pyx_tuple__152;
static PyObject *__pyx_tuple__154;
static PyObject *__pyx_tuple__155;
static PyObject *__pyx_tuple__157;
static PyObject *__pyx_tuple__159;
static PyObject *__pyx_tuple__161;
static PyObject *__pyx_tuple__163;
static PyObject *__pyx_tuple__165;
static PyObject *__pyx_tuple__167;
static PyObject *__pyx_tuple__169;
static PyObject *__pyx_tuple__171;
static PyObject *__pyx_tuple__173;
static PyObject *__pyx_tuple__175;
static PyObject *__pyx_tuple__177;
static PyObject *__pyx_tuple__179;
static PyObject *__pyx_tuple__181;
static PyObject *__pyx_tuple__183;
static PyObject *__pyx_tuple__185;
static PyObject *__pyx_tuple__187;
static PyObject *__pyx_tuple__189;
static PyObject *__pyx_tuple__191;
static PyObject *__pyx_tuple__192;
static PyObject *__pyx_tuple__194;
static PyObject *__pyx_tuple__196;
static PyObject *__pyx_tuple__198;
static PyObject *__pyx_tuple__200;
static PyObject *__pyx_tuple__202;
static PyObject *__pyx_tuple__204;
static PyObject *__pyx_tuple__206;
static PyObject *__pyx_tuple__208;
static PyObject *__pyx_tuple__210;
static PyObject *__pyx_tuple__212;
static PyObject *__pyx_tuple__214;
static PyObject *__pyx_tuple__215;
static PyObject *__pyx_tuple__217;
static PyObject *__pyx_tuple__219;
static PyObject *__pyx_tuple__221;
static PyObject *__pyx_tuple__223;
static PyObject *__pyx_tuple__224;
static PyObject *__pyx_tuple__226;
static PyObject *__pyx_tuple__228;
static PyObject *__pyx_tuple__230;
static PyObject *__pyx_tuple__231;
static PyObject *__pyx_tuple__233;
static PyObject *__pyx_tuple__235;
static PyObject *__pyx_tuple__237;
static PyObject *__pyx_tuple__239;
static PyObject *__pyx_tuple__241;
static PyObject *__pyx_tuple__243;
static PyObject *__pyx_tuple__244;
static PyObject *__pyx_tuple__246;
static PyObject *__pyx_tuple__247;
static PyObject *__pyx_tuple__249;
static PyObject *__pyx_tuple__251;
static PyObject *__pyx_tuple__253;
static PyObject *__pyx_tuple__255;
static PyObject *__pyx_tuple__257;
static PyObject *__pyx_tuple__258;
static PyObject *__pyx_tuple__260;
static PyObject *__pyx_tuple__262;
static PyObject *__pyx_tuple__263;
static PyObject *__pyx_tuple__265;
static PyObject *__pyx_tuple__267;
static PyObject *__pyx_tuple__269;
static PyObject *__pyx_tuple__271;
static PyObject *__pyx_tuple__273;
static PyObject *__pyx_tuple__275;
static PyObject *__pyx_tuple__277;
static PyObject *__pyx_tuple__278;
static PyObject *__pyx_tuple__280;
static PyObject *__pyx_tuple__282;
static PyObject *__pyx_tuple__284;
static PyObject *__pyx_tuple__285;
static PyObject *__pyx_tuple__287;
static PyObject *__pyx_tuple__288;
static PyObject *__pyx_tuple__290;
static PyObject *__pyx_tuple__291;
static PyObject *__pyx_tuple__293;
static PyObject *__pyx_tuple__295;
static PyObject *__pyx_tuple__297;
static PyObject *__pyx_tuple__299;
static PyObject *__pyx_tuple__301;
static PyObject *__pyx_tuple__303;
static PyObject *__pyx_tuple__305;
static PyObject *__pyx_tuple__306;
static PyObject *__pyx_tuple__308;
static PyObject *__pyx_tuple__310;
static PyObject *__pyx_tuple__311;
static PyObject *__pyx_tuple__313;
static PyObject *__pyx_tuple__315;
static PyObject *__pyx_tuple__316;
static PyObject *__pyx_tuple__318;
static PyObject *__pyx_tuple__319;
static PyObject *__pyx_tuple__321;
static PyObject *__pyx_tuple__322;
static PyObject *__pyx_tuple__324;
static PyObject *__pyx_tuple__326;
static PyObject *__pyx_tuple__328;
static PyObject *__pyx_tuple__330;
static PyObject *__pyx_tuple__331;
static PyObject *__pyx_tuple__333;
static PyObject *__pyx_tuple__334;
static PyObject *__pyx_tuple__335;
static PyObject *__pyx_tuple__337;
static PyObject *__pyx_tuple__339;
static PyObject *__pyx_tuple__340;
static PyObject *__pyx_tuple__342;
static PyObject *__pyx_tuple__344;
static PyObject *__pyx_tuple__346;
static PyObject *__pyx_tuple__348;
static PyObject *__pyx_tuple__350;
static PyObject *__pyx_tuple__352;
static PyObject *__pyx_tuple__354;
static PyObject *__pyx_tuple__356;
static PyObject *__pyx_tuple__358;
static PyObject *__pyx_tuple__360;
static PyObject *__pyx_tuple__362;
static PyObject *__pyx_tuple__364;
static PyObject *__pyx_tuple__366;
static PyObject *__pyx_tuple__368;
static PyObject *__pyx_tuple__370;
static PyObject *__pyx_tuple__372;
static PyObject *__pyx_tuple__373;
static PyObject *__pyx_tuple__375;
static PyObject *__pyx_tuple__377;
static PyObject *__pyx_tuple__379;
static PyObject *__pyx_tuple__381;
static PyObject *__pyx_tuple__383;
static PyObject *__pyx_tuple__385;
static PyObject *__pyx_tuple__387;
static PyObject *__pyx_tuple__388;
static PyObject *__pyx_tuple__390;
static PyObject *__pyx_tuple__392;
static PyObject *__pyx_tuple__394;
static PyObject *__pyx_tuple__396;
static PyObject *__pyx_tuple__398;
static PyObject *__pyx_tuple__399;
static PyObject *__pyx_tuple__401;
static PyObject *__pyx_tuple__402;
static PyObject *__pyx_tuple__404;
static PyObject *__pyx_tuple__406;
static PyObject *__pyx_tuple__408;
static PyObject *__pyx_tuple__410;
static PyObject *__pyx_tuple__412;
static PyObject *__pyx_tuple__414;
static PyObject *__pyx_tuple__415;
static PyObject *__pyx_tuple__417;
static PyObject *__pyx_tuple__418;
static PyObject *__pyx_tuple__420;
static PyObject *__pyx_tuple__421;
static PyObject *__pyx_tuple__423;
static PyObject *__pyx_tuple__425;
static PyObject *__pyx_tuple__427;
static PyObject *__pyx_tuple__429;
static PyObject *__pyx_tuple__431;
static PyObject *__pyx_tuple__433;
static PyObject *__pyx_tuple__434;
static PyObject *__pyx_tuple__436;
static PyObject *__pyx_tuple__438;
static PyObject *__pyx_tuple__439;
static PyObject *__pyx_tuple__441;
static PyObject *__pyx_tuple__443;
static PyObject *__pyx_tuple__445;
static PyObject *__pyx_tuple__447;
static PyObject *__pyx_tuple__448;
static PyObject *__pyx_tuple__450;
static PyObject *__pyx_tuple__452;
static PyObject *__pyx_tuple__453;
static PyObject *__pyx_tuple__455;
static PyObject *__pyx_tuple__457;
static PyObject *__pyx_tuple__459;
static PyObject *__pyx_tuple__460;
static PyObject *__pyx_tuple__462;
static PyObject *__pyx_tuple__463;
static PyObject *__pyx_tuple__465;
static PyObject *__pyx_tuple__467;
static PyObject *__pyx_tuple__469;
static PyObject *__pyx_tuple__471;
static PyObject *__pyx_tuple__472;
static PyObject *__pyx_tuple__474;
static PyObject *__pyx_tuple__475;
static PyObject *__pyx_tuple__477;
static PyObject *__pyx_tuple__478;
static PyObject *__pyx_tuple__480;
static PyObject *__pyx_tuple__481;
static PyObject *__pyx_tuple__483;
static PyObject *__pyx_tuple__484;
static PyObject *__pyx_tuple__486;
static PyObject *__pyx_tuple__487;
static PyObject *__pyx_tuple__489;
static PyObject *__pyx_tuple__491;
static PyObject *__pyx_tuple__492;
static PyObject *__pyx_tuple__494;
static PyObject *__pyx_tuple__495;
static PyObject *__pyx_tuple__497;
static PyObject *__pyx_tuple__498;
static PyObject *__pyx_tuple__500;
static PyObject *__pyx_tuple__501;
static PyObject *__pyx_tuple__503;
static PyObject *__pyx_tuple__504;
static PyObject *__pyx_tuple__506;
static PyObject *__pyx_tuple__507;
static PyObject *__pyx_tuple__509;
static PyObject *__pyx_tuple__511;
static PyObject *__pyx_tuple__513;
static PyObject *__pyx_tuple__514;
static PyObject *__pyx_tuple__516;
static PyObject *__pyx_tuple__517;
static PyObject *__pyx_tuple__519;
static PyObject *__pyx_tuple__521;
static PyObject *__pyx_tuple__522;
static PyObject *__pyx_tuple__524;
static PyObject *__pyx_tuple__525;
static PyObject *__pyx_tuple__545;
static PyObject *__pyx_tuple__547;
static PyObject *__pyx_tuple__548;
static PyObject *__pyx_tuple__550;
static PyObject *__pyx_tuple__551;
static PyObject *__pyx_tuple__553;
static PyObject *__pyx_tuple__555;
static PyObject *__pyx_tuple__557;
static PyObject *__pyx_tuple__559;
static PyObject *__pyx_tuple__560;
static PyObject *__pyx_tuple__562;
static PyObject *__pyx_tuple__564;
static PyObject *__pyx_tuple__565;
static PyObject *__pyx_tuple__567;
static PyObject *__pyx_tuple__569;
static PyObject *__pyx_tuple__570;
static PyObject *__pyx_tuple__572;
static PyObject *__pyx_tuple__574;
static PyObject *__pyx_tuple__575;
static PyObject *__pyx_tuple__577;
static PyObject *__pyx_tuple__579;
static PyObject *__pyx_tuple__580;
static PyObject *__pyx_tuple__582;
static PyObject *__pyx_tuple__584;
static PyObject *__pyx_tuple__585;
static PyObject *__pyx_tuple__587;
static PyObject *__pyx_tuple__589;
static PyObject *__pyx_tuple__590;
static PyObject *__pyx_tuple__592;
static PyObject *__pyx_tuple__594;
static PyObject *__pyx_tuple__595;
static PyObject *__pyx_tuple__597;
static PyObject *__pyx_tuple__599;
static PyObject *__pyx_tuple__600;
static PyObject *__pyx_tuple__602;
static PyObject *__pyx_tuple__604;
static PyObject *__pyx_tuple__606;
static PyObject *__pyx_tuple__608;
static PyObject *__pyx_tuple__609;
static PyObject *__pyx_tuple__611;
static PyObject *__pyx_tuple__613;
static PyObject *__pyx_tuple__614;
static PyObject *__pyx_tuple__616;
static PyObject *__pyx_tuple__618;
static PyObject *__pyx_tuple__620;
static PyObject *__pyx_tuple__622;
static PyObject *__pyx_tuple__623;
static PyObject *__pyx_tuple__625;
static PyObject *__pyx_tuple__626;
static PyObject *__pyx_tuple__628;
static PyObject *__pyx_tuple__630;
static PyObject *__pyx_tuple__631;
static PyObject *__pyx_tuple__633;
static PyObject *__pyx_tuple__635;
static PyObject *__pyx_tuple__637;
static PyObject *__pyx_tuple__639;
static PyObject *__pyx_tuple__640;
static PyObject *__pyx_tuple__642;
static PyObject *__pyx_tuple__643;
static PyObject *__pyx_tuple__645;
static PyObject *__pyx_tuple__647;
static PyObject *__pyx_tuple__648;
static PyObject *__pyx_tuple__650;
static PyObject *__pyx_tuple__652;
static PyObject *__pyx_tuple__654;
static PyObject *__pyx_tuple__655;
static PyObject *__pyx_tuple__657;
static PyObject *__pyx_tuple__658;
static PyObject *__pyx_tuple__660;
static PyObject *__pyx_tuple__662;
static PyObject *__pyx_tuple__664;
static PyObject *__pyx_tuple__666;
static PyObject *__pyx_tuple__668;
static PyObject *__pyx_tuple__669;
static PyObject *__pyx_tuple__671;
static PyObject *__pyx_tuple__674;
static PyObject *__pyx_tuple__676;
static PyObject *__pyx_tuple__677;
static PyObject *__pyx_tuple__679;
static PyObject *__pyx_tuple__680;
static PyObject *__pyx_tuple__682;
static PyObject *__pyx_tuple__683;
static PyObject *__pyx_tuple__685;
static PyObject *__pyx_tuple__687;
static PyObject *__pyx_tuple__688;
static PyObject *__pyx_tuple__690;
static PyObject *__pyx_tuple__692;
static PyObject *__pyx_tuple__693;
static PyObject *__pyx_tuple__695;
static PyObject *__pyx_tuple__697;
static PyObject *__pyx_tuple__698;
static PyObject *__pyx_tuple__700;
static PyObject *__pyx_tuple__701;
static PyObject *__pyx_tuple__703;
static PyObject *__pyx_tuple__705;
static PyObject *__pyx_tuple__706;
static PyObject *__pyx_tuple__708;
static PyObject *__pyx_tuple__709;
static PyObject *__pyx_tuple__711;
static PyObject *__pyx_tuple__712;
static PyObject *__pyx_tuple__714;
static PyObject *__pyx_tuple__716;
static PyObject *__pyx_tuple__718;
static PyObject *__pyx_tuple__720;
static PyObject *__pyx_tuple__721;
static PyObject *__pyx_tuple__723;
static PyObject *__pyx_tuple__725;
static PyObject *__pyx_tuple__726;
static PyObject *__pyx_tuple__728;
static PyObject *__pyx_tuple__729;
static PyObject *__pyx_tuple__731;
static PyObject *__pyx_tuple__732;
static PyObject *__pyx_tuple__734;
static PyObject *__pyx_tuple__736;
static PyObject *__pyx_tuple__738;
static PyObject *__pyx_tuple__740;
static PyObject *__pyx_tuple__742;
static PyObject *__pyx_tuple__743;
static PyObject *__pyx_tuple__745;
static PyObject *__pyx_tuple__747;
static PyObject *__pyx_tuple__749;
static PyObject *__pyx_tuple__750;
static PyObject *__pyx_tuple__752;
static PyObject *__pyx_tuple__753;
static PyObject *__pyx_tuple__755;
static PyObject *__pyx_tuple__756;
static PyObject *__pyx_tuple__758;
static PyObject *__pyx_tuple__760;
static PyObject *__pyx_tuple__762;
static PyObject *__pyx_tuple__764;
static PyObject *__pyx_tuple__766;
static PyObject *__pyx_tuple__768;
static PyObject *__pyx_tuple__770;
static PyObject *__pyx_tuple__772;
static PyObject *__pyx_tuple__773;
static PyObject *__pyx_tuple__775;
static PyObject *__pyx_tuple__777;
static PyObject *__pyx_codeobj__18;
static PyObject *__pyx_codeobj__24;
static PyObject *__pyx_codeobj__36;
static PyObject *__pyx_codeobj__40;
static PyObject *__pyx_codeobj__42;
static PyObject *__pyx_codeobj__44;
static PyObject *__pyx_codeobj__46;
static PyObject *__pyx_codeobj__48;
static PyObject *__pyx_codeobj__50;
static PyObject *__pyx_codeobj__52;
static PyObject *__pyx_codeobj__54;
static PyObject *__pyx_codeobj__56;
static PyObject *__pyx_codeobj__58;
static PyObject *__pyx_codeobj__60;
static PyObject *__pyx_codeobj__62;
static PyObject *__pyx_codeobj__63;
static PyObject *__pyx_codeobj__64;
static PyObject *__pyx_codeobj__77;
static PyObject *__pyx_codeobj__87;
static PyObject *__pyx_codeobj__93;
static PyObject *__pyx_codeobj__95;
static PyObject *__pyx_codeobj__121;
static PyObject *__pyx_codeobj__123;
static PyObject *__pyx_codeobj__125;
static PyObject *__pyx_codeobj__127;
static PyObject *__pyx_codeobj__129;
static PyObject *__pyx_codeobj__131;
static PyObject *__pyx_codeobj__133;
static PyObject *__pyx_codeobj__135;
static PyObject *__pyx_codeobj__137;
static PyObject *__pyx_codeobj__139;
static PyObject *__pyx_codeobj__141;
static PyObject *__pyx_codeobj__144;
static PyObject *__pyx_codeobj__146;
static PyObject *__pyx_codeobj__148;
static PyObject *__pyx_codeobj__150;
static PyObject *__pyx_codeobj__153;
static PyObject *__pyx_codeobj__156;
static PyObject *__pyx_codeobj__158;
static PyObject *__pyx_codeobj__160;
static PyObject *__pyx_codeobj__162;
static PyObject *__pyx_codeobj__164;
static PyObject *__pyx_codeobj__166;
static PyObject *__pyx_codeobj__168;
static PyObject *__pyx_codeobj__170;
static PyObject *__pyx_codeobj__172;
static PyObject *__pyx_codeobj__174;
static PyObject *__pyx_codeobj__176;
static PyObject *__pyx_codeobj__178;
static PyObject *__pyx_codeobj__180;
static PyObject *__pyx_codeobj__182;
static PyObject *__pyx_codeobj__184;
static PyObject *__pyx_codeobj__186;
static PyObject *__pyx_codeobj__188;
static PyObject *__pyx_codeobj__190;
static PyObject *__pyx_codeobj__193;
static PyObject *__pyx_codeobj__195;
static PyObject *__pyx_codeobj__197;
static PyObject *__pyx_codeobj__199;
static PyObject *__pyx_codeobj__201;
static PyObject *__pyx_codeobj__203;
static PyObject *__pyx_codeobj__205;
static PyObject *__pyx_codeobj__207;
static PyObject *__pyx_codeobj__209;
static PyObject *__pyx_codeobj__211;
static PyObject *__pyx_codeobj__213;
static PyObject *__pyx_codeobj__216;
static PyObject *__pyx_codeobj__218;
static PyObject *__pyx_codeobj__220;
static PyObject *__pyx_codeobj__222;
static PyObject *__pyx_codeobj__225;
static PyObject *__pyx_codeobj__227;
static PyObject *__pyx_codeobj__229;
static PyObject *__pyx_codeobj__232;
static PyObject *__pyx_codeobj__234;
static PyObject *__pyx_codeobj__236;
static PyObject *__pyx_codeobj__238;
static PyObject *__pyx_codeobj__240;
static PyObject *__pyx_codeobj__242;
static PyObject *__pyx_codeobj__245;
static PyObject *__pyx_codeobj__248;
static PyObject *__pyx_codeobj__250;
static PyObject *__pyx_codeobj__252;
static PyObject *__pyx_codeobj__254;
static PyObject *__pyx_codeobj__256;
static PyObject *__pyx_codeobj__259;
static PyObject *__pyx_codeobj__261;
static PyObject *__pyx_codeobj__264;
static PyObject *__pyx_codeobj__266;
static PyObject *__pyx_codeobj__270;
static PyObject *__pyx_codeobj__272;
static PyObject *__pyx_codeobj__274;
static PyObject *__pyx_codeobj__276;
static PyObject *__pyx_codeobj__279;
static PyObject *__pyx_codeobj__281;
static PyObject *__pyx_codeobj__283;
static PyObject *__pyx_codeobj__286;
static PyObject *__pyx_codeobj__289;
static PyObject *__pyx_codeobj__292;
static PyObject *__pyx_codeobj__294;
static PyObject *__pyx_codeobj__296;
static PyObject *__pyx_codeobj__298;
static PyObject *__pyx_codeobj__300;
static PyObject *__pyx_codeobj__302;
static PyObject *__pyx_codeobj__304;
static PyObject *__pyx_codeobj__307;
static PyObject *__pyx_codeobj__309;
static PyObject *__pyx_codeobj__312;
static PyObject *__pyx_codeobj__314;
static PyObject *__pyx_codeobj__317;
static PyObject *__pyx_codeobj__320;
static PyObject *__pyx_codeobj__323;
static PyObject *__pyx_codeobj__325;
static PyObject *__pyx_codeobj__327;
static PyObject *__pyx_codeobj__329;
static PyObject *__pyx_codeobj__332;
static PyObject *__pyx_codeobj__336;
static PyObject *__pyx_codeobj__338;
static PyObject *__pyx_codeobj__341;
static PyObject *__pyx_codeobj__343;
static PyObject *__pyx_codeobj__345;
static PyObject *__pyx_codeobj__347;
static PyObject *__pyx_codeobj__349;
static PyObject *__pyx_codeobj__351;
static PyObject *__pyx_codeobj__353;
static PyObject *__pyx_codeobj__355;
static PyObject *__pyx_codeobj__357;
static PyObject *__pyx_codeobj__359;
static PyObject *__pyx_codeobj__361;
static PyObject *__pyx_codeobj__363;
static PyObject *__pyx_codeobj__365;
static PyObject *__pyx_codeobj__367;
static PyObject *__pyx_codeobj__369;
static PyObject *__pyx_codeobj__371;
static PyObject *__pyx_codeobj__374;
static PyObject *__pyx_codeobj__376;
static PyObject *__pyx_codeobj__378;
static PyObject *__pyx_codeobj__380;
static PyObject *__pyx_codeobj__382;
static PyObject *__pyx_codeobj__384;
static PyObject *__pyx_codeobj__386;
static PyObject *__pyx_codeobj__389;
static PyObject *__pyx_codeobj__391;
static PyObject *__pyx_codeobj__393;
static PyObject *__pyx_codeobj__395;
static PyObject *__pyx_codeobj__397;
static PyObject *__pyx_codeobj__400;
static PyObject *__pyx_codeobj__403;
static PyObject *__pyx_codeobj__405;
static PyObject *__pyx_codeobj__407;
static PyObject *__pyx_codeobj__409;
static PyObject *__pyx_codeobj__411;
static PyObject *__pyx_codeobj__413;
static PyObject *__pyx_codeobj__416;
static PyObject *__pyx_codeobj__419;
static PyObject *__pyx_codeobj__422;
static PyObject *__pyx_codeobj__424;
static PyObject *__pyx_codeobj__426;
static PyObject *__pyx_codeobj__428;
static PyObject *__pyx_codeobj__430;
static PyObject *__pyx_codeobj__432;
static PyObject *__pyx_codeobj__435;
static PyObject *__pyx_codeobj__437;
static PyObject *__pyx_codeobj__440;
static PyObject *__pyx_codeobj__442;
static PyObject *__pyx_codeobj__444;
static PyObject *__pyx_codeobj__446;
static PyObject *__pyx_codeobj__449;
static PyObject *__pyx_codeobj__451;
static PyObject *__pyx_codeobj__454;
static PyObject *__pyx_codeobj__458;
static PyObject *__pyx_codeobj__461;
static PyObject *__pyx_codeobj__464;
static PyObject *__pyx_codeobj__466;
static PyObject *__pyx_codeobj__468;
static PyObject *__pyx_codeobj__470;
static PyObject *__pyx_codeobj__473;
static PyObject *__pyx_codeobj__476;
static PyObject *__pyx_codeobj__479;
static PyObject *__pyx_codeobj__482;
static PyObject *__pyx_codeobj__485;
static PyObject *__pyx_codeobj__488;
static PyObject *__pyx_codeobj__490;
static PyObject *__pyx_codeobj__493;
static PyObject *__pyx_codeobj__496;
static PyObject *__pyx_codeobj__499;
static PyObject *__pyx_codeobj__502;
static PyObject *__pyx_codeobj__505;
static PyObject *__pyx_codeobj__508;
static PyObject *__pyx_codeobj__510;
static PyObject *__pyx_codeobj__512;
static PyObject *__pyx_codeobj__515;
static PyObject *__pyx_codeobj__518;
static PyObject *__pyx_codeobj__520;
static PyObject *__pyx_codeobj__523;
static PyObject *__pyx_codeobj__526;
static PyObject *__pyx_codeobj__546;
static PyObject *__pyx_codeobj__549;
static PyObject *__pyx_codeobj__552;
static PyObject *__pyx_codeobj__554;
static PyObject *__pyx_codeobj__556;
static PyObject *__pyx_codeobj__558;
static PyObject *__pyx_codeobj__561;
static PyObject *__pyx_codeobj__563;
static PyObject *__pyx_codeobj__566;
static PyObject *__pyx_codeobj__568;
static PyObject *__pyx_codeobj__571;
static PyObject *__pyx_codeobj__573;
static PyObject *__pyx_codeobj__576;
static PyObject *__pyx_codeobj__578;
static PyObject *__pyx_codeobj__581;
static PyObject *__pyx_codeobj__583;
static PyObject *__pyx_codeobj__586;
static PyObject *__pyx_codeobj__588;
static PyObject *__pyx_codeobj__591;
static PyObject *__pyx_codeobj__593;
static PyObject *__pyx_codeobj__596;
static PyObject *__pyx_codeobj__598;
static PyObject *__pyx_codeobj__601;
static PyObject *__pyx_codeobj__603;
static PyObject *__pyx_codeobj__605;
static PyObject *__pyx_codeobj__607;
static PyObject *__pyx_codeobj__610;
static PyObject *__pyx_codeobj__612;
static PyObject *__pyx_codeobj__615;
static PyObject *__pyx_codeobj__617;
static PyObject *__pyx_codeobj__619;
static PyObject *__pyx_codeobj__621;
static PyObject *__pyx_codeobj__624;
static PyObject *__pyx_codeobj__627;
static PyObject *__pyx_codeobj__629;
static PyObject *__pyx_codeobj__632;
static PyObject *__pyx_codeobj__634;
static PyObject *__pyx_codeobj__636;
static PyObject *__pyx_codeobj__638;
static PyObject *__pyx_codeobj__641;
static PyObject *__pyx_codeobj__644;
static PyObject *__pyx_codeobj__646;
static PyObject *__pyx_codeobj__649;
static PyObject *__pyx_codeobj__651;
static PyObject *__pyx_codeobj__653;
static PyObject *__pyx_codeobj__656;
static PyObject *__pyx_codeobj__659;
static PyObject *__pyx_codeobj__661;
static PyObject *__pyx_codeobj__663;
static PyObject *__pyx_codeobj__665;
static PyObject *__pyx_codeobj__667;
static PyObject *__pyx_codeobj__670;
static PyObject *__pyx_codeobj__672;
static PyObject *__pyx_codeobj__675;
static PyObject *__pyx_codeobj__678;
static PyObject *__pyx_codeobj__681;
static PyObject *__pyx_codeobj__684;
static PyObject *__pyx_codeobj__686;
static PyObject *__pyx_codeobj__689;
static PyObject *__pyx_codeobj__691;
static PyObject *__pyx_codeobj__694;
static PyObject *__pyx_codeobj__696;
static PyObject *__pyx_codeobj__699;
static PyObject *__pyx_codeobj__702;
static PyObject *__pyx_codeobj__704;
static PyObject *__pyx_codeobj__707;
static PyObject *__pyx_codeobj__710;
static PyObject *__pyx_codeobj__713;
static PyObject *__pyx_codeobj__715;
static PyObject *__pyx_codeobj__717;
static PyObject *__pyx_codeobj__719;
static PyObject *__pyx_codeobj__722;
static PyObject *__pyx_codeobj__724;
static PyObject *__pyx_codeobj__727;
static PyObject *__pyx_codeobj__730;
static PyObject *__pyx_codeobj__733;
static PyObject *__pyx_codeobj__735;
static PyObject *__pyx_codeobj__737;
static PyObject *__pyx_codeobj__739;
static PyObject *__pyx_codeobj__741;
static PyObject *__pyx_codeobj__744;
static PyObject *__pyx_codeobj__746;
static PyObject *__pyx_codeobj__748;
static PyObject *__pyx_codeobj__751;
static PyObject *__pyx_codeobj__754;
static PyObject *__pyx_codeobj__757;
static PyObject *__pyx_codeobj__759;
static PyObject *__pyx_codeobj__761;
static PyObject *__pyx_codeobj__763;
static PyObject *__pyx_codeobj__765;
static PyObject *__pyx_codeobj__767;
static PyObject *__pyx_codeobj__769;
static PyObject *__pyx_codeobj__771;
static PyObject *__pyx_codeobj__774;
static PyObject *__pyx_codeobj__776;
static PyObject *__pyx_codeobj__778;
/* Late includes */
static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_24generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/pyparsing_extension.py":13
 * import sre_constants
 * 
 * _generatorType = type((y for y in range(1)))             # <<<<<<<<<<<<<<
 * 
 * from collections.abc import Iterable
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_22genexpr(CYTHON_UNUSED PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct__genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct__genexpr *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct__genexpr(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct__genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct__genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 13, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_24generator2, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_genexpr, __pyx_n_s_Aspidites__vendor_pyparsing_exte); if (unlikely(!gen)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_24generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct__genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct__genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  long __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 13, __pyx_L1_error)
  for (__pyx_t_1 = 0; __pyx_t_1 < 1; __pyx_t_1+=1) {
    __pyx_cur_scope->__pyx_v_y = __pyx_t_1;
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_cur_scope->__pyx_v_y); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 13, __pyx_L1_error)
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_27generator3(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/pyparsing_extension.py":43
 * hexnums = nums + "ABCDEFabcdef"
 * alphanums = alphas + nums
 * printables = "".join(c for c in string.printable if c not in string.whitespace)             # <<<<<<<<<<<<<<
 * 
 * __diag__ = SimpleNamespace()
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_25genexpr(CYTHON_UNUSED PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 43, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_27generator3, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_genexpr, __pyx_n_s_Aspidites__vendor_pyparsing_exte); if (unlikely(!gen)) __PYX_ERR(0, 43, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_27generator3(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L7_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_string); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_printable); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 43, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 43, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 43, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 43, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 43, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 43, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_4(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 43, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_c);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_c, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_string); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 43, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_whitespace); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 43, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = (__Pyx_PySequence_ContainsTF(__pyx_cur_scope->__pyx_v_c, __pyx_t_5, Py_NE)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 43, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_7 = (__pyx_t_6 != 0);
    if (__pyx_t_7) {
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_c);
      __pyx_r = __pyx_cur_scope->__pyx_v_c;
      __Pyx_XGIVEREF(__pyx_t_1);
      __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
      __pyx_cur_scope->__pyx_t_1 = __pyx_t_3;
      __pyx_cur_scope->__pyx_t_2 = __pyx_t_4;
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
      /* return from generator, yielding value */
      __pyx_generator->resume_label = 1;
      return __pyx_r;
      __pyx_L7_resume_from_yield:;
      __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
      __pyx_cur_scope->__pyx_t_0 = 0;
      __Pyx_XGOTREF(__pyx_t_1);
      __pyx_t_3 = __pyx_cur_scope->__pyx_t_1;
      __pyx_t_4 = __pyx_cur_scope->__pyx_t_2;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 43, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":70
 * 
 * 
 * def _enable_all_warnings():             # <<<<<<<<<<<<<<
 *     __diag__.warn_multiple_tokens_in_named_alternation = True
 *     __diag__.warn_ungrouped_named_tokens_in_collection = True
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_1_enable_all_warnings(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_1_enable_all_warnings = {"_enable_all_warnings", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_1_enable_all_warnings, METH_NOARGS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_1_enable_all_warnings(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_enable_all_warnings (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension__enable_all_warnings(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension__enable_all_warnings(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_enable_all_warnings", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":71
 * 
 * def _enable_all_warnings():
 *     __diag__.warn_multiple_tokens_in_named_alternation = True             # <<<<<<<<<<<<<<
 *     __diag__.warn_ungrouped_named_tokens_in_collection = True
 *     __diag__.warn_name_set_on_empty_Forward = True
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_diag); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_1, __pyx_n_s_warn_multiple_tokens_in_named_al, Py_True) < 0) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":72
 * def _enable_all_warnings():
 *     __diag__.warn_multiple_tokens_in_named_alternation = True
 *     __diag__.warn_ungrouped_named_tokens_in_collection = True             # <<<<<<<<<<<<<<
 *     __diag__.warn_name_set_on_empty_Forward = True
 *     __diag__.warn_on_multiple_string_args_to_oneof = True
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_diag); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_1, __pyx_n_s_warn_ungrouped_named_tokens_in_c, Py_True) < 0) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":73
 *     __diag__.warn_multiple_tokens_in_named_alternation = True
 *     __diag__.warn_ungrouped_named_tokens_in_collection = True
 *     __diag__.warn_name_set_on_empty_Forward = True             # <<<<<<<<<<<<<<
 *     __diag__.warn_on_multiple_string_args_to_oneof = True
 * __diag__.enable_all_warnings = _enable_all_warnings
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_diag); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_1, __pyx_n_s_warn_name_set_on_empty_Forward, Py_True) < 0) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":74
 *     __diag__.warn_ungrouped_named_tokens_in_collection = True
 *     __diag__.warn_name_set_on_empty_Forward = True
 *     __diag__.warn_on_multiple_string_args_to_oneof = True             # <<<<<<<<<<<<<<
 * __diag__.enable_all_warnings = _enable_all_warnings
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_diag); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_1, __pyx_n_s_warn_on_multiple_string_args_to, Py_True) < 0) __PYX_ERR(0, 74, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":70
 * 
 * 
 * def _enable_all_warnings():             # <<<<<<<<<<<<<<
 *     __diag__.warn_multiple_tokens_in_named_alternation = True
 *     __diag__.warn_ungrouped_named_tokens_in_collection = True
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._enable_all_warnings", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":78
 * 
 * 
 * def _defaultStartDebugAction(instring, loc, expr):             # <<<<<<<<<<<<<<
 *     print(("Match " + str(expr) + " at loc " + str(loc) + "(%d,%d)" % (lineno(loc, instring), col(loc, instring))))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_3_defaultStartDebugAction(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_3_defaultStartDebugAction = {"_defaultStartDebugAction", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_3_defaultStartDebugAction, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_3_defaultStartDebugAction(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  PyObject *__pyx_v_expr = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_defaultStartDebugAction (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_expr,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_defaultStartDebugAction", 1, 3, 3, 1); __PYX_ERR(0, 78, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_expr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_defaultStartDebugAction", 1, 3, 3, 2); __PYX_ERR(0, 78, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_defaultStartDebugAction") < 0)) __PYX_ERR(0, 78, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_instring = values[0];
    __pyx_v_loc = values[1];
    __pyx_v_expr = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_defaultStartDebugAction", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 78, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._defaultStartDebugAction", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_2_defaultStartDebugAction(__pyx_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_expr);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_2_defaultStartDebugAction(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, PyObject *__pyx_v_expr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_defaultStartDebugAction", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":79
 * 
 * def _defaultStartDebugAction(instring, loc, expr):
 *     print(("Match " + str(expr) + " at loc " + str(loc) + "(%d,%d)" % (lineno(loc, instring), col(loc, instring))))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_v_expr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_Add(__pyx_kp_s_Match, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Add(__pyx_t_2, __pyx_kp_s_at_loc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_v_loc); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Add(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_lineno); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_loc, __pyx_v_instring};
    __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_2);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_loc, __pyx_v_instring};
    __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_2);
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_loc);
    __Pyx_GIVEREF(__pyx_v_loc);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_loc);
    __Pyx_INCREF(__pyx_v_instring);
    __Pyx_GIVEREF(__pyx_v_instring);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_instring);
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_col); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_loc, __pyx_v_instring};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_loc, __pyx_v_instring};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_loc);
    __Pyx_GIVEREF(__pyx_v_loc);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_5, __pyx_v_loc);
    __Pyx_INCREF(__pyx_v_instring);
    __Pyx_GIVEREF(__pyx_v_instring);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_5, __pyx_v_instring);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_d_d, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = PyNumber_Add(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_PrintOne(0, __pyx_t_6) < 0) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":78
 * 
 * 
 * def _defaultStartDebugAction(instring, loc, expr):             # <<<<<<<<<<<<<<
 *     print(("Match " + str(expr) + " at loc " + str(loc) + "(%d,%d)" % (lineno(loc, instring), col(loc, instring))))
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._defaultStartDebugAction", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":82
 * 
 * 
 * def _defaultSuccessDebugAction(instring, startloc, endloc, expr, toks):             # <<<<<<<<<<<<<<
 *     print("Matched " + str(expr) + " -> " + str(toks.asList()))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5_defaultSuccessDebugAction(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_5_defaultSuccessDebugAction = {"_defaultSuccessDebugAction", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5_defaultSuccessDebugAction, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5_defaultSuccessDebugAction(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_instring = 0;
  CYTHON_UNUSED PyObject *__pyx_v_startloc = 0;
  CYTHON_UNUSED PyObject *__pyx_v_endloc = 0;
  PyObject *__pyx_v_expr = 0;
  PyObject *__pyx_v_toks = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_defaultSuccessDebugAction (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_instring,&__pyx_n_s_startloc,&__pyx_n_s_endloc,&__pyx_n_s_expr,&__pyx_n_s_toks,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_startloc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_defaultSuccessDebugAction", 1, 5, 5, 1); __PYX_ERR(0, 82, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_endloc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_defaultSuccessDebugAction", 1, 5, 5, 2); __PYX_ERR(0, 82, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_expr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_defaultSuccessDebugAction", 1, 5, 5, 3); __PYX_ERR(0, 82, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_toks)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_defaultSuccessDebugAction", 1, 5, 5, 4); __PYX_ERR(0, 82, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_defaultSuccessDebugAction") < 0)) __PYX_ERR(0, 82, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 5) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
    }
    __pyx_v_instring = values[0];
    __pyx_v_startloc = values[1];
    __pyx_v_endloc = values[2];
    __pyx_v_expr = values[3];
    __pyx_v_toks = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_defaultSuccessDebugAction", 1, 5, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 82, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._defaultSuccessDebugAction", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4_defaultSuccessDebugAction(__pyx_self, __pyx_v_instring, __pyx_v_startloc, __pyx_v_endloc, __pyx_v_expr, __pyx_v_toks);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4_defaultSuccessDebugAction(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_instring, CYTHON_UNUSED PyObject *__pyx_v_startloc, CYTHON_UNUSED PyObject *__pyx_v_endloc, PyObject *__pyx_v_expr, PyObject *__pyx_v_toks) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_defaultSuccessDebugAction", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":83
 * 
 * def _defaultSuccessDebugAction(instring, startloc, endloc, expr, toks):
 *     print("Matched " + str(expr) + " -> " + str(toks.asList()))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_v_expr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 83, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_Add(__pyx_kp_s_Matched, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 83, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Add(__pyx_t_2, __pyx_kp_s_); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 83, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_toks, __pyx_n_s_asList); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 83, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 83, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 83, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_Add(__pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 83, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_PrintOne(0, __pyx_t_2) < 0) __PYX_ERR(0, 83, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":82
 * 
 * 
 * def _defaultSuccessDebugAction(instring, startloc, endloc, expr, toks):             # <<<<<<<<<<<<<<
 *     print("Matched " + str(expr) + " -> " + str(toks.asList()))
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._defaultSuccessDebugAction", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":86
 * 
 * 
 * def _defaultExceptionDebugAction(instring, loc, expr, exc):             # <<<<<<<<<<<<<<
 *     print("Exception raised:" + str(exc))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7_defaultExceptionDebugAction(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7_defaultExceptionDebugAction = {"_defaultExceptionDebugAction", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7_defaultExceptionDebugAction, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7_defaultExceptionDebugAction(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_instring = 0;
  CYTHON_UNUSED PyObject *__pyx_v_loc = 0;
  CYTHON_UNUSED PyObject *__pyx_v_expr = 0;
  PyObject *__pyx_v_exc = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_defaultExceptionDebugAction (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_expr,&__pyx_n_s_exc,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_defaultExceptionDebugAction", 1, 4, 4, 1); __PYX_ERR(0, 86, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_expr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_defaultExceptionDebugAction", 1, 4, 4, 2); __PYX_ERR(0, 86, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_exc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_defaultExceptionDebugAction", 1, 4, 4, 3); __PYX_ERR(0, 86, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_defaultExceptionDebugAction") < 0)) __PYX_ERR(0, 86, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_instring = values[0];
    __pyx_v_loc = values[1];
    __pyx_v_expr = values[2];
    __pyx_v_exc = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_defaultExceptionDebugAction", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 86, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._defaultExceptionDebugAction", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_6_defaultExceptionDebugAction(__pyx_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_expr, __pyx_v_exc);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_6_defaultExceptionDebugAction(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_instring, CYTHON_UNUSED PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_expr, PyObject *__pyx_v_exc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_defaultExceptionDebugAction", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":87
 * 
 * def _defaultExceptionDebugAction(instring, loc, expr, exc):
 *     print("Exception raised:" + str(exc))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_v_exc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_Add(__pyx_kp_s_Exception_raised, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_PrintOne(0, __pyx_t_2) < 0) __PYX_ERR(0, 87, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":86
 * 
 * 
 * def _defaultExceptionDebugAction(instring, loc, expr, exc):             # <<<<<<<<<<<<<<
 *     print("Exception raised:" + str(exc))
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._defaultExceptionDebugAction", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":90
 * 
 * 
 * def nullDebugAction(*args):             # <<<<<<<<<<<<<<
 *     """'Do-nothing' debug action, to suppress debugging output during parsing."""
 *     pass
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_9nullDebugAction(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_8nullDebugAction[] = "'Do-nothing' debug action, to suppress debugging output during parsing.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_9nullDebugAction = {"nullDebugAction", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_9nullDebugAction, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_8nullDebugAction};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_9nullDebugAction(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_args = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("nullDebugAction (wrapper)", 0);
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "nullDebugAction", 0))) return NULL;
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_8nullDebugAction(__pyx_self, __pyx_v_args);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_8nullDebugAction(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_args) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("nullDebugAction", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":95
 * 
 * 
 * def _escapeRegexRangeChars(s):             # <<<<<<<<<<<<<<
 *     # ~  escape these chars: ^-[]
 *     for c in r"\^-[]":
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_11_escapeRegexRangeChars(PyObject *__pyx_self, PyObject *__pyx_v_s); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_11_escapeRegexRangeChars = {"_escapeRegexRangeChars", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_11_escapeRegexRangeChars, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_11_escapeRegexRangeChars(PyObject *__pyx_self, PyObject *__pyx_v_s) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_escapeRegexRangeChars (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_escapeRegexRangeChars(__pyx_self, ((PyObject *)__pyx_v_s));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_escapeRegexRangeChars(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_s) {
  PyObject *__pyx_v_c = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *(*__pyx_t_2)(PyObject *);
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_escapeRegexRangeChars", 0);
  __Pyx_INCREF(__pyx_v_s);

  /* "Aspidites/_vendor/pyparsing_extension.py":97
 * def _escapeRegexRangeChars(s):
 *     # ~  escape these chars: ^-[]
 *     for c in r"\^-[]":             # <<<<<<<<<<<<<<
 *         s = s.replace(c, chr(92) + c)
 *     s = s.replace("\n", r"\n")
 */
  __pyx_t_1 = PyObject_GetIter(__pyx_kp_s__2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 97, __pyx_L1_error)
  for (;;) {
    {
      __pyx_t_3 = __pyx_t_2(__pyx_t_1);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 97, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XDECREF_SET(__pyx_v_c, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":98
 *     # ~  escape these chars: ^-[]
 *     for c in r"\^-[]":
 *         s = s.replace(c, chr(92) + c)             # <<<<<<<<<<<<<<
 *     s = s.replace("\n", r"\n")
 *     s = s.replace("\t", r"\t")
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_s, __pyx_n_s_replace); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 98, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_chr, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 98, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = PyNumber_Add(__pyx_t_5, __pyx_v_c); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 98, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_c, __pyx_t_6};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 98, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_c, __pyx_t_6};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 98, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 98, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_INCREF(__pyx_v_c);
      __Pyx_GIVEREF(__pyx_v_c);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_v_c);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_8, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 98, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_s, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":97
 * def _escapeRegexRangeChars(s):
 *     # ~  escape these chars: ^-[]
 *     for c in r"\^-[]":             # <<<<<<<<<<<<<<
 *         s = s.replace(c, chr(92) + c)
 *     s = s.replace("\n", r"\n")
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":99
 *     for c in r"\^-[]":
 *         s = s.replace(c, chr(92) + c)
 *     s = s.replace("\n", r"\n")             # <<<<<<<<<<<<<<
 *     s = s.replace("\t", r"\t")
 *     return str(s)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_s, __pyx_n_s_replace); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF_SET(__pyx_v_s, __pyx_t_3);
  __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":100
 *         s = s.replace(c, chr(92) + c)
 *     s = s.replace("\n", r"\n")
 *     s = s.replace("\t", r"\t")             # <<<<<<<<<<<<<<
 *     return str(s)
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_s, __pyx_n_s_replace); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF_SET(__pyx_v_s, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":101
 *     s = s.replace("\n", r"\n")
 *     s = s.replace("\t", r"\t")
 *     return str(s)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_v_s); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":95
 * 
 * 
 * def _escapeRegexRangeChars(s):             # <<<<<<<<<<<<<<
 *     # ~  escape these chars: ^-[]
 *     for c in r"\^-[]":
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._escapeRegexRangeChars", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_c);
  __Pyx_XDECREF(__pyx_v_s);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":104
 * 
 * 
 * def _flatten(L):             # <<<<<<<<<<<<<<
 *     ret = []
 *     for i in L:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13_flatten(PyObject *__pyx_self, PyObject *__pyx_v_L); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13_flatten = {"_flatten", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13_flatten, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13_flatten(PyObject *__pyx_self, PyObject *__pyx_v_L) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_flatten (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12_flatten(__pyx_self, ((PyObject *)__pyx_v_L));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12_flatten(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_L) {
  PyObject *__pyx_v_ret = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_flatten", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":105
 * 
 * def _flatten(L):
 *     ret = []             # <<<<<<<<<<<<<<
 *     for i in L:
 *         if isinstance(i, list):
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ret = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":106
 * def _flatten(L):
 *     ret = []
 *     for i in L:             # <<<<<<<<<<<<<<
 *         if isinstance(i, list):
 *             ret.extend(_flatten(i))
 */
  if (likely(PyList_CheckExact(__pyx_v_L)) || PyTuple_CheckExact(__pyx_v_L)) {
    __pyx_t_1 = __pyx_v_L; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_L); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 106, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 106, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 106, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 106, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 106, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 106, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":107
 *     ret = []
 *     for i in L:
 *         if isinstance(i, list):             # <<<<<<<<<<<<<<
 *             ret.extend(_flatten(i))
 *         else:
 */
    __pyx_t_5 = PyList_Check(__pyx_v_i); 
    __pyx_t_6 = (__pyx_t_5 != 0);
    if (__pyx_t_6) {

      /* "Aspidites/_vendor/pyparsing_extension.py":108
 *     for i in L:
 *         if isinstance(i, list):
 *             ret.extend(_flatten(i))             # <<<<<<<<<<<<<<
 *         else:
 *             ret.append(i)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_flatten); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 108, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      __pyx_t_4 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_v_i) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_i);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 108, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_9 = __Pyx_PyList_Extend(__pyx_v_ret, __pyx_t_4); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 108, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":107
 *     ret = []
 *     for i in L:
 *         if isinstance(i, list):             # <<<<<<<<<<<<<<
 *             ret.extend(_flatten(i))
 *         else:
 */
      goto __pyx_L5;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":110
 *             ret.extend(_flatten(i))
 *         else:
 *             ret.append(i)             # <<<<<<<<<<<<<<
 *     return ret
 * 
 */
    /*else*/ {
      __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_ret, __pyx_v_i); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 110, __pyx_L1_error)
    }
    __pyx_L5:;

    /* "Aspidites/_vendor/pyparsing_extension.py":106
 * def _flatten(L):
 *     ret = []
 *     for i in L:             # <<<<<<<<<<<<<<
 *         if isinstance(i, list):
 *             ret.extend(_flatten(i))
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":111
 *         else:
 *             ret.append(i)
 *     return ret             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_ret);
  __pyx_r = __pyx_v_ret;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":104
 * 
 * 
 * def _flatten(L):             # <<<<<<<<<<<<<<
 *     ret = []
 *     for i in L:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._flatten", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":114
 * 
 * 
 * def _xml_escape(data):             # <<<<<<<<<<<<<<
 *     """Escape &, <, >, ", ', etc. in a string of data."""
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15_xml_escape(PyObject *__pyx_self, PyObject *__pyx_v_data); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_14_xml_escape[] = "Escape &, <, >, \", ', etc. in a string of data.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_15_xml_escape = {"_xml_escape", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15_xml_escape, METH_O, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_14_xml_escape};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15_xml_escape(PyObject *__pyx_self, PyObject *__pyx_v_data) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_xml_escape (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_14_xml_escape(__pyx_self, ((PyObject *)__pyx_v_data));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_11_xml_escape_2generator4(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/pyparsing_extension.py":119
 *     # ampersand must be replaced first
 *     from_symbols = '&><"\''
 *     to_symbols = ('&' + s + ';' for s in "amp gt lt quot apos".split())             # <<<<<<<<<<<<<<
 *     for from_, to_ in zip(from_symbols, to_symbols):
 *         data = data.replace(from_, to_)
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_11_xml_escape_genexpr(CYTHON_UNUSED PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 119, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_11_xml_escape_2generator4, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_xml_escape_locals_genexpr, __pyx_n_s_Aspidites__vendor_pyparsing_exte); if (unlikely(!gen)) __PYX_ERR(0, 119, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._xml_escape.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_11_xml_escape_2generator4(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 119, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_amp_gt_lt_quot_apos, __pyx_n_s_split); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 119, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 119, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 119, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 119, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 119, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_5(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 119, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_s);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_s, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_Add(__pyx_kp_s__8, __pyx_cur_scope->__pyx_v_s); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyNumber_Add(__pyx_t_1, __pyx_kp_s__9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    __Pyx_XGIVEREF(__pyx_t_2);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_4;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_5;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_2);
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_5 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 119, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":114
 * 
 * 
 * def _xml_escape(data):             # <<<<<<<<<<<<<<
 *     """Escape &, <, >, ", ', etc. in a string of data."""
 * 
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_14_xml_escape(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_data) {
  PyObject *__pyx_v_from_symbols = NULL;
  PyObject *__pyx_v_to_symbols = NULL;
  PyObject *__pyx_v_from_ = NULL;
  PyObject *__pyx_v_to_ = NULL;
  PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_11_xml_escape_2generator4 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_xml_escape", 0);
  __Pyx_INCREF(__pyx_v_data);

  /* "Aspidites/_vendor/pyparsing_extension.py":118
 * 
 *     # ampersand must be replaced first
 *     from_symbols = '&><"\''             # <<<<<<<<<<<<<<
 *     to_symbols = ('&' + s + ';' for s in "amp gt lt quot apos".split())
 *     for from_, to_ in zip(from_symbols, to_symbols):
 */
  __Pyx_INCREF(__pyx_kp_s__10);
  __pyx_v_from_symbols = __pyx_kp_s__10;

  /* "Aspidites/_vendor/pyparsing_extension.py":119
 *     # ampersand must be replaced first
 *     from_symbols = '&><"\''
 *     to_symbols = ('&' + s + ';' for s in "amp gt lt quot apos".split())             # <<<<<<<<<<<<<<
 *     for from_, to_ in zip(from_symbols, to_symbols):
 *         data = data.replace(from_, to_)
 */
  __pyx_t_1 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_11_xml_escape_genexpr(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_to_symbols = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":120
 *     from_symbols = '&><"\''
 *     to_symbols = ('&' + s + ';' for s in "amp gt lt quot apos".split())
 *     for from_, to_ in zip(from_symbols, to_symbols):             # <<<<<<<<<<<<<<
 *         data = data.replace(from_, to_)
 *     return data
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_from_symbols);
  __Pyx_GIVEREF(__pyx_v_from_symbols);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_from_symbols);
  __Pyx_INCREF(__pyx_v_to_symbols);
  __Pyx_GIVEREF(__pyx_v_to_symbols);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_to_symbols);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 120, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 120, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 120, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 120, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 120, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 120, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_4(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 120, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
      PyObject* sequence = __pyx_t_2;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 120, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 120, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 120, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_7 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 120, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_8 = Py_TYPE(__pyx_t_7)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_5)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_6 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 2) < 0) __PYX_ERR(0, 120, __pyx_L1_error)
      __pyx_t_8 = NULL;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 120, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_from_, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_XDECREF_SET(__pyx_v_to_, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":121
 *     to_symbols = ('&' + s + ';' for s in "amp gt lt quot apos".split())
 *     for from_, to_ in zip(from_symbols, to_symbols):
 *         data = data.replace(from_, to_)             # <<<<<<<<<<<<<<
 *     return data
 * 
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_replace); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_from_, __pyx_v_to_};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 121, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_from_, __pyx_v_to_};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 121, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else
    #endif
    {
      __pyx_t_7 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 121, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_INCREF(__pyx_v_from_);
      __Pyx_GIVEREF(__pyx_v_from_);
      PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_9, __pyx_v_from_);
      __Pyx_INCREF(__pyx_v_to_);
      __Pyx_GIVEREF(__pyx_v_to_);
      PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_9, __pyx_v_to_);
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 121, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF_SET(__pyx_v_data, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":120
 *     from_symbols = '&><"\''
 *     to_symbols = ('&' + s + ';' for s in "amp gt lt quot apos".split())
 *     for from_, to_ in zip(from_symbols, to_symbols):             # <<<<<<<<<<<<<<
 *         data = data.replace(from_, to_)
 *     return data
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":122
 *     for from_, to_ in zip(from_symbols, to_symbols):
 *         data = data.replace(from_, to_)
 *     return data             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_data);
  __pyx_r = __pyx_v_data;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":114
 * 
 * 
 * def _xml_escape(data):             # <<<<<<<<<<<<<<
 *     """Escape &, <, >, ", ', etc. in a string of data."""
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._xml_escape", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_from_symbols);
  __Pyx_XDECREF(__pyx_v_to_symbols);
  __Pyx_XDECREF(__pyx_v_from_);
  __Pyx_XDECREF(__pyx_v_to_);
  __Pyx_XDECREF(__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_11_xml_escape_2generator4);
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":125
 * 
 * 
 * def col(loc, strg):             # <<<<<<<<<<<<<<
 *     """Returns current column within a string, counting newlines as line separators.
 *    The first column is number 1.
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_17col(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_16col[] = "Returns current column within a string, counting newlines as line separators.\n   The first column is number 1.\n\n   Note: the default parsing behavior is to expand tabs in the input string\n   before starting the parsing process.  See\n   :class:`ParserElement.parseString` for more\n   information on parsing strings containing ``<TAB>`` s, and suggested\n   methods to maintain a consistent view of the parsed string, the parse\n   location, and line and column positions within the parsed string.\n   ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_17col = {"col", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_17col, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_16col};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_17col(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_loc = 0;
  PyObject *__pyx_v_strg = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("col (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_loc,&__pyx_n_s_strg,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strg)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("col", 1, 2, 2, 1); __PYX_ERR(0, 125, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "col") < 0)) __PYX_ERR(0, 125, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_loc = values[0];
    __pyx_v_strg = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("col", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 125, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.col", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_16col(__pyx_self, __pyx_v_loc, __pyx_v_strg);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_16col(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_loc, PyObject *__pyx_v_strg) {
  PyObject *__pyx_v_s = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("col", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":136
 *    location, and line and column positions within the parsed string.
 *    """
 *     s = strg             # <<<<<<<<<<<<<<
 *     return 1 if 0 < loc < len(s) and s[loc - 1] == '\n' else loc - s.rfind("\n", 0, loc)
 * 
 */
  __Pyx_INCREF(__pyx_v_strg);
  __pyx_v_s = __pyx_v_strg;

  /* "Aspidites/_vendor/pyparsing_extension.py":137
 *    """
 *     s = strg
 *     return 1 if 0 < loc < len(s) and s[loc - 1] == '\n' else loc - s.rfind("\n", 0, loc)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = PyObject_RichCompare(__pyx_int_0, __pyx_v_loc, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 137, __pyx_L1_error)
  if (__Pyx_PyObject_IsTrue(__pyx_t_3)) {
    __Pyx_DECREF(__pyx_t_3);
    __pyx_t_4 = PyObject_Length(__pyx_v_s); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 137, __pyx_L1_error)
    __pyx_t_5 = PyInt_FromSsize_t(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 137, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = PyObject_RichCompare(__pyx_v_loc, __pyx_t_5, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 137, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_6) {
  } else {
    __pyx_t_2 = __pyx_t_6;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyInt_SubtractObjC(__pyx_v_loc, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_s, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = (__Pyx_PyString_Equals(__pyx_t_5, __pyx_kp_s__4, Py_EQ)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_2 = __pyx_t_6;
  __pyx_L3_bool_binop_done:;
  if (__pyx_t_2) {
    __Pyx_INCREF(__pyx_int_1);
    __pyx_t_1 = __pyx_int_1;
  } else {
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_s, __pyx_n_s_rfind); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 137, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[4] = {__pyx_t_7, __pyx_kp_s__4, __pyx_int_0, __pyx_v_loc};
      __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_8, 3+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 137, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_5);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[4] = {__pyx_t_7, __pyx_kp_s__4, __pyx_int_0, __pyx_v_loc};
      __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_8, 3+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 137, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_5);
    } else
    #endif
    {
      __pyx_t_9 = PyTuple_New(3+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 137, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_INCREF(__pyx_kp_s__4);
      __Pyx_GIVEREF(__pyx_kp_s__4);
      PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_kp_s__4);
      __Pyx_INCREF(__pyx_int_0);
      __Pyx_GIVEREF(__pyx_int_0);
      PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_int_0);
      __Pyx_INCREF(__pyx_v_loc);
      __Pyx_GIVEREF(__pyx_v_loc);
      PyTuple_SET_ITEM(__pyx_t_9, 2+__pyx_t_8, __pyx_v_loc);
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_9, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 137, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Subtract(__pyx_v_loc, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 137, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":125
 * 
 * 
 * def col(loc, strg):             # <<<<<<<<<<<<<<
 *     """Returns current column within a string, counting newlines as line separators.
 *    The first column is number 1.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.col", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_s);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":140
 * 
 * 
 * def lineno(loc, strg):             # <<<<<<<<<<<<<<
 *     """Returns current line number within a string, counting newlines as line separators.
 *     The first line is number 1.
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_19lineno(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_18lineno[] = "Returns current line number within a string, counting newlines as line separators.\n    The first line is number 1.\n\n    Note - the default parsing behavior is to expand tabs in the input string\n    before starting the parsing process.  See :class:`ParserElement.parseString`\n    for more information on parsing strings containing ``<TAB>`` s, and\n    suggested methods to maintain a consistent view of the parsed string, the\n    parse location, and line and column positions within the parsed string.\n    ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_19lineno = {"lineno", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_19lineno, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_18lineno};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_19lineno(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_loc = 0;
  PyObject *__pyx_v_strg = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lineno (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_loc,&__pyx_n_s_strg,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strg)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("lineno", 1, 2, 2, 1); __PYX_ERR(0, 140, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "lineno") < 0)) __PYX_ERR(0, 140, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_loc = values[0];
    __pyx_v_strg = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("lineno", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 140, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.lineno", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_18lineno(__pyx_self, __pyx_v_loc, __pyx_v_strg);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_18lineno(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_loc, PyObject *__pyx_v_strg) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lineno", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":150
 *     parse location, and line and column positions within the parsed string.
 *     """
 *     return strg.count("\n", 0, loc) + 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_strg, __pyx_n_s_count); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[4] = {__pyx_t_3, __pyx_kp_s__4, __pyx_int_0, __pyx_v_loc};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 3+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 150, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[4] = {__pyx_t_3, __pyx_kp_s__4, __pyx_int_0, __pyx_v_loc};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 3+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 150, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(3+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_kp_s__4);
    __Pyx_GIVEREF(__pyx_kp_s__4);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_kp_s__4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_int_0);
    __Pyx_INCREF(__pyx_v_loc);
    __Pyx_GIVEREF(__pyx_v_loc);
    PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_4, __pyx_v_loc);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":140
 * 
 * 
 * def lineno(loc, strg):             # <<<<<<<<<<<<<<
 *     """Returns current line number within a string, counting newlines as line separators.
 *     The first line is number 1.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.lineno", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":153
 * 
 * 
 * def line(loc, strg):             # <<<<<<<<<<<<<<
 *     """Returns the line of text containing loc within a string, counting newlines as line separators.
 *        """
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_21line(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_20line[] = "Returns the line of text containing loc within a string, counting newlines as line separators.\n       ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_21line = {"line", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_21line, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_20line};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_21line(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_loc = 0;
  PyObject *__pyx_v_strg = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("line (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_loc,&__pyx_n_s_strg,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strg)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("line", 1, 2, 2, 1); __PYX_ERR(0, 153, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "line") < 0)) __PYX_ERR(0, 153, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_loc = values[0];
    __pyx_v_strg = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("line", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 153, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.line", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_20line(__pyx_self, __pyx_v_loc, __pyx_v_strg);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_20line(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_loc, PyObject *__pyx_v_strg) {
  PyObject *__pyx_v_lastCR = NULL;
  PyObject *__pyx_v_nextCR = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("line", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":156
 *     """Returns the line of text containing loc within a string, counting newlines as line separators.
 *        """
 *     lastCR = strg.rfind("\n", 0, loc)             # <<<<<<<<<<<<<<
 *     nextCR = strg.find("\n", loc)
 *     if nextCR >= 0:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_strg, __pyx_n_s_rfind); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[4] = {__pyx_t_3, __pyx_kp_s__4, __pyx_int_0, __pyx_v_loc};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 3+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 156, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[4] = {__pyx_t_3, __pyx_kp_s__4, __pyx_int_0, __pyx_v_loc};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 3+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 156, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(3+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 156, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_kp_s__4);
    __Pyx_GIVEREF(__pyx_kp_s__4);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_kp_s__4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_int_0);
    __Pyx_INCREF(__pyx_v_loc);
    __Pyx_GIVEREF(__pyx_v_loc);
    PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_4, __pyx_v_loc);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 156, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_lastCR = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":157
 *        """
 *     lastCR = strg.rfind("\n", 0, loc)
 *     nextCR = strg.find("\n", loc)             # <<<<<<<<<<<<<<
 *     if nextCR >= 0:
 *         return strg[lastCR + 1:nextCR]
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_strg, __pyx_n_s_find); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_kp_s__4, __pyx_v_loc};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 157, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_kp_s__4, __pyx_v_loc};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 157, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_3 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 157, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_kp_s__4);
    __Pyx_GIVEREF(__pyx_kp_s__4);
    PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_4, __pyx_kp_s__4);
    __Pyx_INCREF(__pyx_v_loc);
    __Pyx_GIVEREF(__pyx_v_loc);
    PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_4, __pyx_v_loc);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 157, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_nextCR = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":158
 *     lastCR = strg.rfind("\n", 0, loc)
 *     nextCR = strg.find("\n", loc)
 *     if nextCR >= 0:             # <<<<<<<<<<<<<<
 *         return strg[lastCR + 1:nextCR]
 *     else:
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_nextCR, __pyx_int_0, Py_GE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 158, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 158, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_6) {

    /* "Aspidites/_vendor/pyparsing_extension.py":159
 *     nextCR = strg.find("\n", loc)
 *     if nextCR >= 0:
 *         return strg[lastCR + 1:nextCR]             # <<<<<<<<<<<<<<
 *     else:
 *         return strg[lastCR + 1:]
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyInt_AddObjC(__pyx_v_lastCR, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 159, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetSlice(__pyx_v_strg, 0, 0, &__pyx_t_1, &__pyx_v_nextCR, NULL, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 159, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":158
 *     lastCR = strg.rfind("\n", 0, loc)
 *     nextCR = strg.find("\n", loc)
 *     if nextCR >= 0:             # <<<<<<<<<<<<<<
 *         return strg[lastCR + 1:nextCR]
 *     else:
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":161
 *         return strg[lastCR + 1:nextCR]
 *     else:
 *         return strg[lastCR + 1:]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_v_lastCR, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 161, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_strg, 0, 0, &__pyx_t_2, NULL, NULL, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 161, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":153
 * 
 * 
 * def line(loc, strg):             # <<<<<<<<<<<<<<
 *     """Returns the line of text containing loc within a string, counting newlines as line separators.
 *        """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.line", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_lastCR);
  __Pyx_XDECREF(__pyx_v_nextCR);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":165
 * 
 * class _ParseResultsWithOffset(object):
 *     def __init__(self, p1, p2):             # <<<<<<<<<<<<<<
 *         self.tup = (p1, p2)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_23_ParseResultsWithOffset_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_23_ParseResultsWithOffset_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_23_ParseResultsWithOffset_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_23_ParseResultsWithOffset_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_p1 = 0;
  PyObject *__pyx_v_p2 = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_p1,&__pyx_n_s_p2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 165, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 165, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 165, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_p1 = values[1];
    __pyx_v_p2 = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 165, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._ParseResultsWithOffset.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_23_ParseResultsWithOffset___init__(__pyx_self, __pyx_v_self, __pyx_v_p1, __pyx_v_p2);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_23_ParseResultsWithOffset___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_p1, PyObject *__pyx_v_p2) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":166
 * class _ParseResultsWithOffset(object):
 *     def __init__(self, p1, p2):
 *         self.tup = (p1, p2)             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, i):
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_p1);
  __Pyx_GIVEREF(__pyx_v_p1);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_p1);
  __Pyx_INCREF(__pyx_v_p2);
  __Pyx_GIVEREF(__pyx_v_p2);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_p2);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_tup, __pyx_t_1) < 0) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":165
 * 
 * class _ParseResultsWithOffset(object):
 *     def __init__(self, p1, p2):             # <<<<<<<<<<<<<<
 *         self.tup = (p1, p2)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._ParseResultsWithOffset.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":168
 *         self.tup = (p1, p2)
 * 
 *     def __getitem__(self, i):             # <<<<<<<<<<<<<<
 *         return self.tup[i]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_23_ParseResultsWithOffset_3__getitem__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_23_ParseResultsWithOffset_3__getitem__ = {"__getitem__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_23_ParseResultsWithOffset_3__getitem__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_23_ParseResultsWithOffset_3__getitem__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_i)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__getitem__", 1, 2, 2, 1); __PYX_ERR(0, 168, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__getitem__") < 0)) __PYX_ERR(0, 168, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__getitem__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 168, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._ParseResultsWithOffset.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_23_ParseResultsWithOffset_2__getitem__(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_23_ParseResultsWithOffset_2__getitem__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":169
 * 
 *     def __getitem__(self, i):
 *         return self.tup[i]             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_tup); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_i); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":168
 *         self.tup = (p1, p2)
 * 
 *     def __getitem__(self, i):             # <<<<<<<<<<<<<<
 *         return self.tup[i]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._ParseResultsWithOffset.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":171
 *         return self.tup[i]
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.tup[0])
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_23_ParseResultsWithOffset_5__repr__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_23_ParseResultsWithOffset_5__repr__ = {"__repr__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_23_ParseResultsWithOffset_5__repr__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_23_ParseResultsWithOffset_5__repr__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_23_ParseResultsWithOffset_4__repr__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_23_ParseResultsWithOffset_4__repr__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":172
 * 
 *     def __repr__(self):
 *         return repr(self.tup[0])             # <<<<<<<<<<<<<<
 * 
 *     def setOffset(self, i):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_tup); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_Repr(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":171
 *         return self.tup[i]
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.tup[0])
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._ParseResultsWithOffset.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":174
 *         return repr(self.tup[0])
 * 
 *     def setOffset(self, i):             # <<<<<<<<<<<<<<
 *         self.tup = (self.tup[0], i)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_23_ParseResultsWithOffset_7setOffset(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_23_ParseResultsWithOffset_7setOffset = {"setOffset", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_23_ParseResultsWithOffset_7setOffset, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_23_ParseResultsWithOffset_7setOffset(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setOffset (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_i)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setOffset", 1, 2, 2, 1); __PYX_ERR(0, 174, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setOffset") < 0)) __PYX_ERR(0, 174, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setOffset", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 174, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._ParseResultsWithOffset.setOffset", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_23_ParseResultsWithOffset_6setOffset(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_23_ParseResultsWithOffset_6setOffset(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setOffset", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":175
 * 
 *     def setOffset(self, i):
 *         self.tup = (self.tup[0], i)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_tup); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_i);
  __Pyx_GIVEREF(__pyx_v_i);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_i);
  __pyx_t_2 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_tup, __pyx_t_1) < 0) __PYX_ERR(0, 175, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":174
 *         return repr(self.tup[0])
 * 
 *     def setOffset(self, i):             # <<<<<<<<<<<<<<
 *         self.tup = (self.tup[0], i)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._ParseResultsWithOffset.setOffset", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":222
 *     """
 * 
 *     def __new__(cls, toklist=None, name=None, asList=True, modal=True):             # <<<<<<<<<<<<<<
 *         if isinstance(toklist, cls):
 *             return toklist
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_1__new__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_1__new__ = {"__new__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_1__new__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_1__new__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_cls = 0;
  PyObject *__pyx_v_toklist = 0;
  CYTHON_UNUSED PyObject *__pyx_v_name = 0;
  CYTHON_UNUSED PyObject *__pyx_v_asList = 0;
  CYTHON_UNUSED PyObject *__pyx_v_modal = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__new__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_cls,&__pyx_n_s_toklist,&__pyx_n_s_name,&__pyx_n_s_asList,&__pyx_n_s_modal,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)Py_True));
    values[4] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cls)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_toklist);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_asList);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_modal);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__new__") < 0)) __PYX_ERR(0, 222, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_cls = values[0];
    __pyx_v_toklist = values[1];
    __pyx_v_name = values[2];
    __pyx_v_asList = values[3];
    __pyx_v_modal = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__new__", 0, 1, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 222, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.__new__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults___new__(__pyx_self, __pyx_v_cls, __pyx_v_toklist, __pyx_v_name, __pyx_v_asList, __pyx_v_modal);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults___new__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_cls, PyObject *__pyx_v_toklist, CYTHON_UNUSED PyObject *__pyx_v_name, CYTHON_UNUSED PyObject *__pyx_v_asList, CYTHON_UNUSED PyObject *__pyx_v_modal) {
  PyObject *__pyx_v_retobj = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__new__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":223
 * 
 *     def __new__(cls, toklist=None, name=None, asList=True, modal=True):
 *         if isinstance(toklist, cls):             # <<<<<<<<<<<<<<
 *             return toklist
 *         retobj = object.__new__(cls)
 */
  __pyx_t_1 = PyObject_IsInstance(__pyx_v_toklist, __pyx_v_cls); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 223, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":224
 *     def __new__(cls, toklist=None, name=None, asList=True, modal=True):
 *         if isinstance(toklist, cls):
 *             return toklist             # <<<<<<<<<<<<<<
 *         retobj = object.__new__(cls)
 *         retobj.__doinit = True
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_toklist);
    __pyx_r = __pyx_v_toklist;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":223
 * 
 *     def __new__(cls, toklist=None, name=None, asList=True, modal=True):
 *         if isinstance(toklist, cls):             # <<<<<<<<<<<<<<
 *             return toklist
 *         retobj = object.__new__(cls)
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":225
 *         if isinstance(toklist, cls):
 *             return toklist
 *         retobj = object.__new__(cls)             # <<<<<<<<<<<<<<
 *         retobj.__doinit = True
 *         return retobj
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_builtin_object, __pyx_n_s_new); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_cls) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_cls);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_retobj = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":226
 *             return toklist
 *         retobj = object.__new__(cls)
 *         retobj.__doinit = True             # <<<<<<<<<<<<<<
 *         return retobj
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_retobj, __pyx_n_s_ParseResults__doinit, Py_True) < 0) __PYX_ERR(0, 226, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":227
 *         retobj = object.__new__(cls)
 *         retobj.__doinit = True
 *         return retobj             # <<<<<<<<<<<<<<
 * 
 *     # Performance tuning: we construct a *lot* of these, so keep this
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_retobj);
  __pyx_r = __pyx_v_retobj;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":222
 *     """
 * 
 *     def __new__(cls, toklist=None, name=None, asList=True, modal=True):             # <<<<<<<<<<<<<<
 *         if isinstance(toklist, cls):
 *             return toklist
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.__new__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_retobj);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":231
 *     # Performance tuning: we construct a *lot* of these, so keep this
 *     # constructor as small and fast as possible
 *     def __init__(self, toklist=None, name=None, asList=True, modal=True, isinstance=isinstance):             # <<<<<<<<<<<<<<
 *         if self.__doinit:
 *             self.__doinit = False
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_28__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__defaults__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)Py_None));
  __Pyx_GIVEREF(((PyObject *)Py_None));
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_None));
  __Pyx_INCREF(((PyObject *)Py_None));
  __Pyx_GIVEREF(((PyObject *)Py_None));
  PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)Py_None));
  __Pyx_INCREF(((PyObject *)Py_True));
  __Pyx_GIVEREF(((PyObject *)Py_True));
  PyTuple_SET_ITEM(__pyx_t_1, 2, ((PyObject *)Py_True));
  __Pyx_INCREF(((PyObject *)Py_True));
  __Pyx_GIVEREF(((PyObject *)Py_True));
  PyTuple_SET_ITEM(__pyx_t_1, 3, ((PyObject *)Py_True));
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_isinstance);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_isinstance);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_isinstance);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_2, 1, Py_None);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_3__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_3__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_3__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_3__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_toklist = 0;
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_asList = 0;
  PyObject *__pyx_v_modal = 0;
  PyObject *__pyx_v_isinstance = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_toklist,&__pyx_n_s_name,&__pyx_n_s_asList,&__pyx_n_s_modal,&__pyx_n_s_isinstance,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    __pyx_defaults *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self);
    values[1] = ((PyObject *)((PyObject *)Py_None));
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)Py_True));
    values[4] = ((PyObject *)((PyObject *)Py_True));
    values[5] = __pyx_dynamic_args->__pyx_arg_isinstance;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_toklist);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_asList);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_modal);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_isinstance);
          if (value) { values[5] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 231, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_toklist = values[1];
    __pyx_v_name = values[2];
    __pyx_v_asList = values[3];
    __pyx_v_modal = values[4];
    __pyx_v_isinstance = values[5];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 1, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 231, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_2__init__(__pyx_self, __pyx_v_self, __pyx_v_toklist, __pyx_v_name, __pyx_v_asList, __pyx_v_modal, __pyx_v_isinstance);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_2__init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_toklist, PyObject *__pyx_v_name, PyObject *__pyx_v_asList, PyObject *__pyx_v_modal, PyObject *__pyx_v_isinstance) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);
  __Pyx_INCREF(__pyx_v_toklist);
  __Pyx_INCREF(__pyx_v_name);

  /* "Aspidites/_vendor/pyparsing_extension.py":232
 *     # constructor as small and fast as possible
 *     def __init__(self, toklist=None, name=None, asList=True, modal=True, isinstance=isinstance):
 *         if self.__doinit:             # <<<<<<<<<<<<<<
 *             self.__doinit = False
 *             self.__name = None
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__doinit); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 232, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":233
 *     def __init__(self, toklist=None, name=None, asList=True, modal=True, isinstance=isinstance):
 *         if self.__doinit:
 *             self.__doinit = False             # <<<<<<<<<<<<<<
 *             self.__name = None
 *             self.__parent = None
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__doinit, Py_False) < 0) __PYX_ERR(0, 233, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":234
 *         if self.__doinit:
 *             self.__doinit = False
 *             self.__name = None             # <<<<<<<<<<<<<<
 *             self.__parent = None
 *             self.__accumNames = {}
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__name, Py_None) < 0) __PYX_ERR(0, 234, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":235
 *             self.__doinit = False
 *             self.__name = None
 *             self.__parent = None             # <<<<<<<<<<<<<<
 *             self.__accumNames = {}
 *             self.__asList = asList
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__parent, Py_None) < 0) __PYX_ERR(0, 235, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":236
 *             self.__name = None
 *             self.__parent = None
 *             self.__accumNames = {}             # <<<<<<<<<<<<<<
 *             self.__asList = asList
 *             self.__modal = modal
 */
    __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 236, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__accumNames, __pyx_t_1) < 0) __PYX_ERR(0, 236, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":237
 *             self.__parent = None
 *             self.__accumNames = {}
 *             self.__asList = asList             # <<<<<<<<<<<<<<
 *             self.__modal = modal
 *             if toklist is None:
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__asList, __pyx_v_asList) < 0) __PYX_ERR(0, 237, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":238
 *             self.__accumNames = {}
 *             self.__asList = asList
 *             self.__modal = modal             # <<<<<<<<<<<<<<
 *             if toklist is None:
 *                 toklist = []
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__modal, __pyx_v_modal) < 0) __PYX_ERR(0, 238, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":239
 *             self.__asList = asList
 *             self.__modal = modal
 *             if toklist is None:             # <<<<<<<<<<<<<<
 *                 toklist = []
 *             if isinstance(toklist, list):
 */
    __pyx_t_2 = (__pyx_v_toklist == Py_None);
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (__pyx_t_3) {

      /* "Aspidites/_vendor/pyparsing_extension.py":240
 *             self.__modal = modal
 *             if toklist is None:
 *                 toklist = []             # <<<<<<<<<<<<<<
 *             if isinstance(toklist, list):
 *                 self.__toklist = toklist[:]
 */
      __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 240, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_toklist, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":239
 *             self.__asList = asList
 *             self.__modal = modal
 *             if toklist is None:             # <<<<<<<<<<<<<<
 *                 toklist = []
 *             if isinstance(toklist, list):
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":241
 *             if toklist is None:
 *                 toklist = []
 *             if isinstance(toklist, list):             # <<<<<<<<<<<<<<
 *                 self.__toklist = toklist[:]
 *             elif isinstance(toklist, _generatorType):
 */
    __Pyx_INCREF(__pyx_v_isinstance);
    __pyx_t_4 = __pyx_v_isinstance; __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_toklist, ((PyObject *)(&PyList_Type))};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 241, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_toklist, ((PyObject *)(&PyList_Type))};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 241, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 241, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_INCREF(__pyx_v_toklist);
      __Pyx_GIVEREF(__pyx_v_toklist);
      PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_toklist);
      __Pyx_INCREF(((PyObject *)(&PyList_Type)));
      __Pyx_GIVEREF(((PyObject *)(&PyList_Type)));
      PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, ((PyObject *)(&PyList_Type)));
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 241, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 241, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_3) {

      /* "Aspidites/_vendor/pyparsing_extension.py":242
 *                 toklist = []
 *             if isinstance(toklist, list):
 *                 self.__toklist = toklist[:]             # <<<<<<<<<<<<<<
 *             elif isinstance(toklist, _generatorType):
 *                 self.__toklist = list(toklist)
 */
      __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_toklist, 0, 0, NULL, NULL, &__pyx_slice__11, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 242, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__toklist, __pyx_t_1) < 0) __PYX_ERR(0, 242, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":241
 *             if toklist is None:
 *                 toklist = []
 *             if isinstance(toklist, list):             # <<<<<<<<<<<<<<
 *                 self.__toklist = toklist[:]
 *             elif isinstance(toklist, _generatorType):
 */
      goto __pyx_L5;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":243
 *             if isinstance(toklist, list):
 *                 self.__toklist = toklist[:]
 *             elif isinstance(toklist, _generatorType):             # <<<<<<<<<<<<<<
 *                 self.__toklist = list(toklist)
 *             else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_generatorType); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 243, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_isinstance);
    __pyx_t_7 = __pyx_v_isinstance; __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_7)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_toklist, __pyx_t_4};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 243, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_toklist, __pyx_t_4};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 243, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 243, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_INCREF(__pyx_v_toklist);
      __Pyx_GIVEREF(__pyx_v_toklist);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_6, __pyx_v_toklist);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_6, __pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 243, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 243, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_3) {

      /* "Aspidites/_vendor/pyparsing_extension.py":244
 *                 self.__toklist = toklist[:]
 *             elif isinstance(toklist, _generatorType):
 *                 self.__toklist = list(toklist)             # <<<<<<<<<<<<<<
 *             else:
 *                 self.__toklist = [toklist]
 */
      __pyx_t_1 = PySequence_List(__pyx_v_toklist); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__toklist, __pyx_t_1) < 0) __PYX_ERR(0, 244, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":243
 *             if isinstance(toklist, list):
 *                 self.__toklist = toklist[:]
 *             elif isinstance(toklist, _generatorType):             # <<<<<<<<<<<<<<
 *                 self.__toklist = list(toklist)
 *             else:
 */
      goto __pyx_L5;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":246
 *                 self.__toklist = list(toklist)
 *             else:
 *                 self.__toklist = [toklist]             # <<<<<<<<<<<<<<
 *             self.__tokdict = dict()
 * 
 */
    /*else*/ {
      __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 246, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_v_toklist);
      __Pyx_GIVEREF(__pyx_v_toklist);
      PyList_SET_ITEM(__pyx_t_1, 0, __pyx_v_toklist);
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__toklist, __pyx_t_1) < 0) __PYX_ERR(0, 246, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L5:;

    /* "Aspidites/_vendor/pyparsing_extension.py":247
 *             else:
 *                 self.__toklist = [toklist]
 *             self.__tokdict = dict()             # <<<<<<<<<<<<<<
 * 
 *         if name is not None and name:
 */
    __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__tokdict, __pyx_t_1) < 0) __PYX_ERR(0, 247, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":232
 *     # constructor as small and fast as possible
 *     def __init__(self, toklist=None, name=None, asList=True, modal=True, isinstance=isinstance):
 *         if self.__doinit:             # <<<<<<<<<<<<<<
 *             self.__doinit = False
 *             self.__name = None
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":249
 *             self.__tokdict = dict()
 * 
 *         if name is not None and name:             # <<<<<<<<<<<<<<
 *             if not modal:
 *                 self.__accumNames[name] = 0
 */
  __pyx_t_2 = (__pyx_v_name != Py_None);
  __pyx_t_9 = (__pyx_t_2 != 0);
  if (__pyx_t_9) {
  } else {
    __pyx_t_3 = __pyx_t_9;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_v_name); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 249, __pyx_L1_error)
  __pyx_t_3 = __pyx_t_9;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/pyparsing_extension.py":250
 * 
 *         if name is not None and name:
 *             if not modal:             # <<<<<<<<<<<<<<
 *                 self.__accumNames[name] = 0
 *             if isinstance(name, int):
 */
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_modal); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 250, __pyx_L1_error)
    __pyx_t_9 = ((!__pyx_t_3) != 0);
    if (__pyx_t_9) {

      /* "Aspidites/_vendor/pyparsing_extension.py":251
 *         if name is not None and name:
 *             if not modal:
 *                 self.__accumNames[name] = 0             # <<<<<<<<<<<<<<
 *             if isinstance(name, int):
 *                 name = str(name)  # will always return a str, but use str for consistency
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__accumNames); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 251, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (unlikely(PyObject_SetItem(__pyx_t_1, __pyx_v_name, __pyx_int_0) < 0)) __PYX_ERR(0, 251, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":250
 * 
 *         if name is not None and name:
 *             if not modal:             # <<<<<<<<<<<<<<
 *                 self.__accumNames[name] = 0
 *             if isinstance(name, int):
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":252
 *             if not modal:
 *                 self.__accumNames[name] = 0
 *             if isinstance(name, int):             # <<<<<<<<<<<<<<
 *                 name = str(name)  # will always return a str, but use str for consistency
 *             self.__name = name
 */
    __Pyx_INCREF(__pyx_v_isinstance);
    __pyx_t_7 = __pyx_v_isinstance; __pyx_t_8 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_7)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_name, ((PyObject *)(&PyInt_Type))};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 252, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_name, ((PyObject *)(&PyInt_Type))};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 252, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 252, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_INCREF(__pyx_v_name);
      __Pyx_GIVEREF(__pyx_v_name);
      PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_v_name);
      __Pyx_INCREF(((PyObject *)(&PyInt_Type)));
      __Pyx_GIVEREF(((PyObject *)(&PyInt_Type)));
      PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, ((PyObject *)(&PyInt_Type)));
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 252, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 252, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_9) {

      /* "Aspidites/_vendor/pyparsing_extension.py":253
 *                 self.__accumNames[name] = 0
 *             if isinstance(name, int):
 *                 name = str(name)  # will always return a str, but use str for consistency             # <<<<<<<<<<<<<<
 *             self.__name = name
 *             if not (isinstance(toklist, (type(None), str, list)) and toklist in (None, '', [])):
 */
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_v_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 253, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_name, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":252
 *             if not modal:
 *                 self.__accumNames[name] = 0
 *             if isinstance(name, int):             # <<<<<<<<<<<<<<
 *                 name = str(name)  # will always return a str, but use str for consistency
 *             self.__name = name
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":254
 *             if isinstance(name, int):
 *                 name = str(name)  # will always return a str, but use str for consistency
 *             self.__name = name             # <<<<<<<<<<<<<<
 *             if not (isinstance(toklist, (type(None), str, list)) and toklist in (None, '', [])):
 *                 if isinstance(toklist, str):
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__name, __pyx_v_name) < 0) __PYX_ERR(0, 254, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":255
 *                 name = str(name)  # will always return a str, but use str for consistency
 *             self.__name = name
 *             if not (isinstance(toklist, (type(None), str, list)) and toklist in (None, '', [])):             # <<<<<<<<<<<<<<
 *                 if isinstance(toklist, str):
 *                     toklist = [toklist]
 */
    __pyx_t_7 = PyTuple_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 255, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(((PyObject *)Py_TYPE(Py_None)));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(Py_None)));
    PyTuple_SET_ITEM(__pyx_t_7, 0, ((PyObject *)Py_TYPE(Py_None)));
    __Pyx_INCREF(((PyObject *)(&PyString_Type)));
    __Pyx_GIVEREF(((PyObject *)(&PyString_Type)));
    PyTuple_SET_ITEM(__pyx_t_7, 1, ((PyObject *)(&PyString_Type)));
    __Pyx_INCREF(((PyObject *)(&PyList_Type)));
    __Pyx_GIVEREF(((PyObject *)(&PyList_Type)));
    PyTuple_SET_ITEM(__pyx_t_7, 2, ((PyObject *)(&PyList_Type)));
    __Pyx_INCREF(__pyx_v_isinstance);
    __pyx_t_4 = __pyx_v_isinstance; __pyx_t_8 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_toklist, __pyx_t_7};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 255, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_toklist, __pyx_t_7};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 255, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 255, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_INCREF(__pyx_v_toklist);
      __Pyx_GIVEREF(__pyx_v_toklist);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_v_toklist);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_7);
      __pyx_t_7 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 255, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 255, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_3) {
    } else {
      __pyx_t_9 = __pyx_t_3;
      goto __pyx_L12_bool_binop_done;
    }
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 255, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_toklist);
    __pyx_t_4 = __pyx_v_toklist;
    __pyx_t_5 = PyObject_RichCompare(__pyx_t_4, Py_None, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 255, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 255, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_t_4, __pyx_kp_s__12, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 255, __pyx_L1_error)
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_t_5 = PyObject_RichCompare(__pyx_t_4, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 255, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 255, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_3 = __pyx_t_2;
    __pyx_L14_bool_binop_done:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_2 = (__pyx_t_3 != 0);
    __pyx_t_9 = __pyx_t_2;
    __pyx_L12_bool_binop_done:;
    __pyx_t_2 = ((!__pyx_t_9) != 0);
    if (__pyx_t_2) {

      /* "Aspidites/_vendor/pyparsing_extension.py":256
 *             self.__name = name
 *             if not (isinstance(toklist, (type(None), str, list)) and toklist in (None, '', [])):
 *                 if isinstance(toklist, str):             # <<<<<<<<<<<<<<
 *                     toklist = [toklist]
 *                 if asList:
 */
      __Pyx_INCREF(__pyx_v_isinstance);
      __pyx_t_4 = __pyx_v_isinstance; __pyx_t_5 = NULL;
      __pyx_t_6 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_6 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_toklist, ((PyObject *)(&PyString_Type))};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 256, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_toklist, ((PyObject *)(&PyString_Type))};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 256, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 256, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        if (__pyx_t_5) {
          __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        }
        __Pyx_INCREF(__pyx_v_toklist);
        __Pyx_GIVEREF(__pyx_v_toklist);
        PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_toklist);
        __Pyx_INCREF(((PyObject *)(&PyString_Type)));
        __Pyx_GIVEREF(((PyObject *)(&PyString_Type)));
        PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, ((PyObject *)(&PyString_Type)));
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 256, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 256, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_2) {

        /* "Aspidites/_vendor/pyparsing_extension.py":257
 *             if not (isinstance(toklist, (type(None), str, list)) and toklist in (None, '', [])):
 *                 if isinstance(toklist, str):
 *                     toklist = [toklist]             # <<<<<<<<<<<<<<
 *                 if asList:
 *                     if isinstance(toklist, ParseResults):
 */
        __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 257, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_v_toklist);
        __Pyx_GIVEREF(__pyx_v_toklist);
        PyList_SET_ITEM(__pyx_t_1, 0, __pyx_v_toklist);
        __Pyx_DECREF_SET(__pyx_v_toklist, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":256
 *             self.__name = name
 *             if not (isinstance(toklist, (type(None), str, list)) and toklist in (None, '', [])):
 *                 if isinstance(toklist, str):             # <<<<<<<<<<<<<<
 *                     toklist = [toklist]
 *                 if asList:
 */
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":258
 *                 if isinstance(toklist, str):
 *                     toklist = [toklist]
 *                 if asList:             # <<<<<<<<<<<<<<
 *                     if isinstance(toklist, ParseResults):
 *                         self[name] = _ParseResultsWithOffset(ParseResults(toklist.__toklist), 0)
 */
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_asList); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 258, __pyx_L1_error)
      if (__pyx_t_2) {

        /* "Aspidites/_vendor/pyparsing_extension.py":259
 *                     toklist = [toklist]
 *                 if asList:
 *                     if isinstance(toklist, ParseResults):             # <<<<<<<<<<<<<<
 *                         self[name] = _ParseResultsWithOffset(ParseResults(toklist.__toklist), 0)
 *                     else:
 */
        __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ParseResults); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 259, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_v_isinstance);
        __pyx_t_7 = __pyx_v_isinstance; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_toklist, __pyx_t_4};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 259, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_toklist, __pyx_t_4};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 259, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_8 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 259, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(__pyx_v_toklist);
          __Pyx_GIVEREF(__pyx_v_toklist);
          PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_6, __pyx_v_toklist);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 259, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 259, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_2) {

          /* "Aspidites/_vendor/pyparsing_extension.py":260
 *                 if asList:
 *                     if isinstance(toklist, ParseResults):
 *                         self[name] = _ParseResultsWithOffset(ParseResults(toklist.__toklist), 0)             # <<<<<<<<<<<<<<
 *                     else:
 *                         self[name] = _ParseResultsWithOffset(ParseResults(toklist[0]), 0)
 */
          __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_ParseResultsWithOffset); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 260, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ParseResults); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 260, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_toklist, __pyx_n_s_ParseResults__toklist); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 260, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_10 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_10)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_10);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          __pyx_t_8 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_10, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5);
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 260, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_8, __pyx_int_0};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 260, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_8, __pyx_int_0};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 260, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          } else
          #endif
          {
            __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 260, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            if (__pyx_t_4) {
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
            }
            __Pyx_GIVEREF(__pyx_t_8);
            PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_t_8);
            __Pyx_INCREF(__pyx_int_0);
            __Pyx_GIVEREF(__pyx_int_0);
            PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_int_0);
            __pyx_t_8 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 260, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (unlikely(PyObject_SetItem(__pyx_v_self, __pyx_v_name, __pyx_t_1) < 0)) __PYX_ERR(0, 260, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":259
 *                     toklist = [toklist]
 *                 if asList:
 *                     if isinstance(toklist, ParseResults):             # <<<<<<<<<<<<<<
 *                         self[name] = _ParseResultsWithOffset(ParseResults(toklist.__toklist), 0)
 *                     else:
 */
          goto __pyx_L19;
        }

        /* "Aspidites/_vendor/pyparsing_extension.py":262
 *                         self[name] = _ParseResultsWithOffset(ParseResults(toklist.__toklist), 0)
 *                     else:
 *                         self[name] = _ParseResultsWithOffset(ParseResults(toklist[0]), 0)             # <<<<<<<<<<<<<<
 *                     self[name].__name = name
 *                 else:
 */
        /*else*/ {
          __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_ParseResultsWithOffset); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 262, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_ParseResults); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 262, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_toklist, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 262, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_10 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
            __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_8);
            if (likely(__pyx_t_10)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
              __Pyx_INCREF(__pyx_t_10);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_8, function);
            }
          }
          __pyx_t_5 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_10, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_4);
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 262, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __pyx_t_8 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_8)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_8);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_5, __pyx_int_0};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 262, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
            PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_5, __pyx_int_0};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 262, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 262, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (__pyx_t_8) {
              __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_8); __pyx_t_8 = NULL;
            }
            __Pyx_GIVEREF(__pyx_t_5);
            PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_t_5);
            __Pyx_INCREF(__pyx_int_0);
            __Pyx_GIVEREF(__pyx_int_0);
            PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_int_0);
            __pyx_t_5 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 262, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (unlikely(PyObject_SetItem(__pyx_v_self, __pyx_v_name, __pyx_t_1) < 0)) __PYX_ERR(0, 262, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        }
        __pyx_L19:;

        /* "Aspidites/_vendor/pyparsing_extension.py":263
 *                     else:
 *                         self[name] = _ParseResultsWithOffset(ParseResults(toklist[0]), 0)
 *                     self[name].__name = name             # <<<<<<<<<<<<<<
 *                 else:
 *                     try:
 */
        __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_self, __pyx_v_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 263, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (__Pyx_PyObject_SetAttrStr(__pyx_t_1, __pyx_n_s_ParseResults__name, __pyx_v_name) < 0) __PYX_ERR(0, 263, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":258
 *                 if isinstance(toklist, str):
 *                     toklist = [toklist]
 *                 if asList:             # <<<<<<<<<<<<<<
 *                     if isinstance(toklist, ParseResults):
 *                         self[name] = _ParseResultsWithOffset(ParseResults(toklist.__toklist), 0)
 */
        goto __pyx_L18;
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":265
 *                     self[name].__name = name
 *                 else:
 *                     try:             # <<<<<<<<<<<<<<
 *                         self[name] = toklist[0]
 *                     except (KeyError, TypeError, IndexError):
 */
      /*else*/ {
        {
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __Pyx_ExceptionSave(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_11);
          __Pyx_XGOTREF(__pyx_t_12);
          __Pyx_XGOTREF(__pyx_t_13);
          /*try:*/ {

            /* "Aspidites/_vendor/pyparsing_extension.py":266
 *                 else:
 *                     try:
 *                         self[name] = toklist[0]             # <<<<<<<<<<<<<<
 *                     except (KeyError, TypeError, IndexError):
 *                         self[name] = toklist
 */
            __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_toklist, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 266, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_1);
            if (unlikely(PyObject_SetItem(__pyx_v_self, __pyx_v_name, __pyx_t_1) < 0)) __PYX_ERR(0, 266, __pyx_L20_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "Aspidites/_vendor/pyparsing_extension.py":265
 *                     self[name].__name = name
 *                 else:
 *                     try:             # <<<<<<<<<<<<<<
 *                         self[name] = toklist[0]
 *                     except (KeyError, TypeError, IndexError):
 */
          }
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          goto __pyx_L25_try_end;
          __pyx_L20_error:;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":267
 *                     try:
 *                         self[name] = toklist[0]
 *                     except (KeyError, TypeError, IndexError):             # <<<<<<<<<<<<<<
 *                         self[name] = toklist
 * 
 */
          __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_IndexError);
          if (__pyx_t_6) {
            __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
            if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_7, &__pyx_t_4) < 0) __PYX_ERR(0, 267, __pyx_L22_except_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_GOTREF(__pyx_t_4);

            /* "Aspidites/_vendor/pyparsing_extension.py":268
 *                         self[name] = toklist[0]
 *                     except (KeyError, TypeError, IndexError):
 *                         self[name] = toklist             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, i):
 */
            if (unlikely(PyObject_SetItem(__pyx_v_self, __pyx_v_name, __pyx_v_toklist) < 0)) __PYX_ERR(0, 268, __pyx_L22_except_error)
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            goto __pyx_L21_exception_handled;
          }
          goto __pyx_L22_except_error;
          __pyx_L22_except_error:;

          /* "Aspidites/_vendor/pyparsing_extension.py":265
 *                     self[name].__name = name
 *                 else:
 *                     try:             # <<<<<<<<<<<<<<
 *                         self[name] = toklist[0]
 *                     except (KeyError, TypeError, IndexError):
 */
          __Pyx_XGIVEREF(__pyx_t_11);
          __Pyx_XGIVEREF(__pyx_t_12);
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
          goto __pyx_L1_error;
          __pyx_L21_exception_handled:;
          __Pyx_XGIVEREF(__pyx_t_11);
          __Pyx_XGIVEREF(__pyx_t_12);
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
          __pyx_L25_try_end:;
        }
      }
      __pyx_L18:;

      /* "Aspidites/_vendor/pyparsing_extension.py":255
 *                 name = str(name)  # will always return a str, but use str for consistency
 *             self.__name = name
 *             if not (isinstance(toklist, (type(None), str, list)) and toklist in (None, '', [])):             # <<<<<<<<<<<<<<
 *                 if isinstance(toklist, str):
 *                     toklist = [toklist]
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":249
 *             self.__tokdict = dict()
 * 
 *         if name is not None and name:             # <<<<<<<<<<<<<<
 *             if not modal:
 *                 self.__accumNames[name] = 0
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":231
 *     # Performance tuning: we construct a *lot* of these, so keep this
 *     # constructor as small and fast as possible
 *     def __init__(self, toklist=None, name=None, asList=True, modal=True, isinstance=isinstance):             # <<<<<<<<<<<<<<
 *         if self.__doinit:
 *             self.__doinit = False
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_toklist);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":270
 *                         self[name] = toklist
 * 
 *     def __getitem__(self, i):             # <<<<<<<<<<<<<<
 *         if isinstance(i, (int, slice)):
 *             return self.__toklist[i]
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_5__getitem__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_5__getitem__ = {"__getitem__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_5__getitem__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_5__getitem__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_i)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__getitem__", 1, 2, 2, 1); __PYX_ERR(0, 270, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__getitem__") < 0)) __PYX_ERR(0, 270, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__getitem__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 270, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_4__getitem__(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_4__getitem__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  PyObject *(*__pyx_t_10)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":271
 * 
 *     def __getitem__(self, i):
 *         if isinstance(i, (int, slice)):             # <<<<<<<<<<<<<<
 *             return self.__toklist[i]
 *         else:
 */
  __pyx_t_2 = PyInt_Check(__pyx_v_i); 
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = PySlice_Check(__pyx_v_i); 
  __pyx_t_2 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":272
 *     def __getitem__(self, i):
 *         if isinstance(i, (int, slice)):
 *             return self.__toklist[i]             # <<<<<<<<<<<<<<
 *         else:
 *             if i not in self.__accumNames:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__toklist); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_t_4, __pyx_v_i); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":271
 * 
 *     def __getitem__(self, i):
 *         if isinstance(i, (int, slice)):             # <<<<<<<<<<<<<<
 *             return self.__toklist[i]
 *         else:
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":274
 *             return self.__toklist[i]
 *         else:
 *             if i not in self.__accumNames:             # <<<<<<<<<<<<<<
 *                 return self.__tokdict[i][-1][0]
 *             else:
 */
  /*else*/ {
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__accumNames); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_v_i, __pyx_t_5, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 274, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {

      /* "Aspidites/_vendor/pyparsing_extension.py":275
 *         else:
 *             if i not in self.__accumNames:
 *                 return self.__tokdict[i][-1][0]             # <<<<<<<<<<<<<<
 *             else:
 *                 return ParseResults([v[0] for v in self.__tokdict[i]])
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__tokdict); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 275, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_t_5, __pyx_v_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 275, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_4, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 275, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_5, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 275, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_r = __pyx_t_4;
      __pyx_t_4 = 0;
      goto __pyx_L0;

      /* "Aspidites/_vendor/pyparsing_extension.py":274
 *             return self.__toklist[i]
 *         else:
 *             if i not in self.__accumNames:             # <<<<<<<<<<<<<<
 *                 return self.__tokdict[i][-1][0]
 *             else:
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":277
 *                 return self.__tokdict[i][-1][0]
 *             else:
 *                 return ParseResults([v[0] for v in self.__tokdict[i]])             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, k, v, isinstance=isinstance):
 */
    /*else*/ {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_ParseResults); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 277, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 277, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__tokdict); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 277, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = __Pyx_PyObject_GetItem(__pyx_t_7, __pyx_v_i); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 277, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (likely(PyList_CheckExact(__pyx_t_8)) || PyTuple_CheckExact(__pyx_t_8)) {
        __pyx_t_7 = __pyx_t_8; __Pyx_INCREF(__pyx_t_7); __pyx_t_9 = 0;
        __pyx_t_10 = NULL;
      } else {
        __pyx_t_9 = -1; __pyx_t_7 = PyObject_GetIter(__pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 277, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_10 = Py_TYPE(__pyx_t_7)->tp_iternext; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 277, __pyx_L1_error)
      }
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      for (;;) {
        if (likely(!__pyx_t_10)) {
          if (likely(PyList_CheckExact(__pyx_t_7))) {
            if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_7)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_8 = PyList_GET_ITEM(__pyx_t_7, __pyx_t_9); __Pyx_INCREF(__pyx_t_8); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 277, __pyx_L1_error)
            #else
            __pyx_t_8 = PySequence_ITEM(__pyx_t_7, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 277, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_8);
            #endif
          } else {
            if (__pyx_t_9 >= PyTuple_GET_SIZE(__pyx_t_7)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_9); __Pyx_INCREF(__pyx_t_8); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 277, __pyx_L1_error)
            #else
            __pyx_t_8 = PySequence_ITEM(__pyx_t_7, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 277, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_8);
            #endif
          }
        } else {
          __pyx_t_8 = __pyx_t_10(__pyx_t_7);
          if (unlikely(!__pyx_t_8)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 277, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_8);
        }
        __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_8);
        __pyx_t_8 = 0;
        __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_v, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 277, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_6, (PyObject*)__pyx_t_8))) __PYX_ERR(0, 277, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_4 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_7, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 277, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_r = __pyx_t_4;
      __pyx_t_4 = 0;
      goto __pyx_L0;
    }
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":270
 *                         self[name] = toklist
 * 
 *     def __getitem__(self, i):             # <<<<<<<<<<<<<<
 *         if isinstance(i, (int, slice)):
 *             return self.__toklist[i]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":279
 *                 return ParseResults([v[0] for v in self.__tokdict[i]])
 * 
 *     def __setitem__(self, k, v, isinstance=isinstance):             # <<<<<<<<<<<<<<
 *         if isinstance(v, _ParseResultsWithOffset):
 *             self.__tokdict[k] = self.__tokdict.get(k, list()) + [v]
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_30__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__defaults__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_self)->__pyx_arg_isinstance);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_self)->__pyx_arg_isinstance);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_self)->__pyx_arg_isinstance);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_2, 1, Py_None);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_7__setitem__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_7__setitem__ = {"__setitem__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_7__setitem__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_7__setitem__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_k = 0;
  PyObject *__pyx_v_v = 0;
  PyObject *__pyx_v_isinstance = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_k,&__pyx_n_s_v,&__pyx_n_s_isinstance,0};
    PyObject* values[4] = {0,0,0,0};
    __pyx_defaults1 *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_self);
    values[3] = __pyx_dynamic_args->__pyx_arg_isinstance;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__setitem__", 0, 3, 4, 1); __PYX_ERR(0, 279, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_v)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__setitem__", 0, 3, 4, 2); __PYX_ERR(0, 279, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_isinstance);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__setitem__") < 0)) __PYX_ERR(0, 279, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_k = values[1];
    __pyx_v_v = values[2];
    __pyx_v_isinstance = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setitem__", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 279, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_6__setitem__(__pyx_self, __pyx_v_self, __pyx_v_k, __pyx_v_v, __pyx_v_isinstance);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_6__setitem__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_k, PyObject *__pyx_v_v, PyObject *__pyx_v_isinstance) {
  PyObject *__pyx_v_sub = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setitem__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":280
 * 
 *     def __setitem__(self, k, v, isinstance=isinstance):
 *         if isinstance(v, _ParseResultsWithOffset):             # <<<<<<<<<<<<<<
 *             self.__tokdict[k] = self.__tokdict.get(k, list()) + [v]
 *             sub = v[0]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParseResultsWithOffset); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_isinstance);
  __pyx_t_3 = __pyx_v_isinstance; __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_v, __pyx_t_2};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 280, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_v, __pyx_t_2};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 280, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 280, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_v);
    __Pyx_GIVEREF(__pyx_v_v);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_v);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 280, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 280, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_7) {

    /* "Aspidites/_vendor/pyparsing_extension.py":281
 *     def __setitem__(self, k, v, isinstance=isinstance):
 *         if isinstance(v, _ParseResultsWithOffset):
 *             self.__tokdict[k] = self.__tokdict.get(k, list()) + [v]             # <<<<<<<<<<<<<<
 *             sub = v[0]
 *         elif isinstance(k, (int, slice)):
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__tokdict); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_get); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_5 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_k, __pyx_t_3};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 281, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_k, __pyx_t_3};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 281, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 281, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (__pyx_t_2) {
        __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
      }
      __Pyx_INCREF(__pyx_v_k);
      __Pyx_GIVEREF(__pyx_v_k);
      PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_5, __pyx_v_k);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_5, __pyx_t_3);
      __pyx_t_3 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 281, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = PyList_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_v_v);
    __Pyx_GIVEREF(__pyx_v_v);
    PyList_SET_ITEM(__pyx_t_6, 0, __pyx_v_v);
    __pyx_t_4 = PyNumber_Add(__pyx_t_1, __pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__tokdict); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (unlikely(PyObject_SetItem(__pyx_t_6, __pyx_v_k, __pyx_t_4) < 0)) __PYX_ERR(0, 281, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":282
 *         if isinstance(v, _ParseResultsWithOffset):
 *             self.__tokdict[k] = self.__tokdict.get(k, list()) + [v]
 *             sub = v[0]             # <<<<<<<<<<<<<<
 *         elif isinstance(k, (int, slice)):
 *             self.__toklist[k] = v
 */
    __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_v, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 282, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_v_sub = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":280
 * 
 *     def __setitem__(self, k, v, isinstance=isinstance):
 *         if isinstance(v, _ParseResultsWithOffset):             # <<<<<<<<<<<<<<
 *             self.__tokdict[k] = self.__tokdict.get(k, list()) + [v]
 *             sub = v[0]
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":283
 *             self.__tokdict[k] = self.__tokdict.get(k, list()) + [v]
 *             sub = v[0]
 *         elif isinstance(k, (int, slice)):             # <<<<<<<<<<<<<<
 *             self.__toklist[k] = v
 *             sub = v
 */
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(((PyObject *)(&PyInt_Type)));
  __Pyx_GIVEREF(((PyObject *)(&PyInt_Type)));
  PyTuple_SET_ITEM(__pyx_t_6, 0, ((PyObject *)(&PyInt_Type)));
  __Pyx_INCREF(((PyObject *)(&PySlice_Type)));
  __Pyx_GIVEREF(((PyObject *)(&PySlice_Type)));
  PyTuple_SET_ITEM(__pyx_t_6, 1, ((PyObject *)(&PySlice_Type)));
  __Pyx_INCREF(__pyx_v_isinstance);
  __pyx_t_1 = __pyx_v_isinstance; __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_k, __pyx_t_6};
    __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 283, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_k, __pyx_t_6};
    __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 283, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 283, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_k);
    __Pyx_GIVEREF(__pyx_v_k);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_5, __pyx_v_k);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_5, __pyx_t_6);
    __pyx_t_6 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 283, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 283, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_7) {

    /* "Aspidites/_vendor/pyparsing_extension.py":284
 *             sub = v[0]
 *         elif isinstance(k, (int, slice)):
 *             self.__toklist[k] = v             # <<<<<<<<<<<<<<
 *             sub = v
 *         else:
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__toklist); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 284, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (unlikely(PyObject_SetItem(__pyx_t_4, __pyx_v_k, __pyx_v_v) < 0)) __PYX_ERR(0, 284, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":285
 *         elif isinstance(k, (int, slice)):
 *             self.__toklist[k] = v
 *             sub = v             # <<<<<<<<<<<<<<
 *         else:
 *             self.__tokdict[k] = self.__tokdict.get(k, list()) + [_ParseResultsWithOffset(v, 0)]
 */
    __Pyx_INCREF(__pyx_v_v);
    __pyx_v_sub = __pyx_v_v;

    /* "Aspidites/_vendor/pyparsing_extension.py":283
 *             self.__tokdict[k] = self.__tokdict.get(k, list()) + [v]
 *             sub = v[0]
 *         elif isinstance(k, (int, slice)):             # <<<<<<<<<<<<<<
 *             self.__toklist[k] = v
 *             sub = v
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":287
 *             sub = v
 *         else:
 *             self.__tokdict[k] = self.__tokdict.get(k, list()) + [_ParseResultsWithOffset(v, 0)]             # <<<<<<<<<<<<<<
 *             sub = v
 *         if isinstance(sub, ParseResults):
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__tokdict); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 287, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_get); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 287, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 287, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_5 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_k, __pyx_t_1};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 287, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_k, __pyx_t_1};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 287, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else
    #endif
    {
      __pyx_t_3 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 287, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_INCREF(__pyx_v_k);
      __Pyx_GIVEREF(__pyx_v_k);
      PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_5, __pyx_v_k);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_5, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 287, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ParseResultsWithOffset); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 287, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_v, __pyx_int_0};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 287, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_v, __pyx_int_0};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 287, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 287, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_1) {
        __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1); __pyx_t_1 = NULL;
      }
      __Pyx_INCREF(__pyx_v_v);
      __Pyx_GIVEREF(__pyx_v_v);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_v);
      __Pyx_INCREF(__pyx_int_0);
      __Pyx_GIVEREF(__pyx_int_0);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 287, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 287, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_2);
    PyList_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_2 = PyNumber_Add(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 287, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__tokdict); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 287, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (unlikely(PyObject_SetItem(__pyx_t_3, __pyx_v_k, __pyx_t_2) < 0)) __PYX_ERR(0, 287, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":288
 *         else:
 *             self.__tokdict[k] = self.__tokdict.get(k, list()) + [_ParseResultsWithOffset(v, 0)]
 *             sub = v             # <<<<<<<<<<<<<<
 *         if isinstance(sub, ParseResults):
 *             sub.__parent = wkref(self)
 */
    __Pyx_INCREF(__pyx_v_v);
    __pyx_v_sub = __pyx_v_v;
  }
  __pyx_L3:;

  /* "Aspidites/_vendor/pyparsing_extension.py":289
 *             self.__tokdict[k] = self.__tokdict.get(k, list()) + [_ParseResultsWithOffset(v, 0)]
 *             sub = v
 *         if isinstance(sub, ParseResults):             # <<<<<<<<<<<<<<
 *             sub.__parent = wkref(self)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ParseResults); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_isinstance);
  __pyx_t_4 = __pyx_v_isinstance; __pyx_t_6 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_4)) {
    PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_sub, __pyx_t_3};
    __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 289, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
    PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_sub, __pyx_t_3};
    __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 289, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  {
    __pyx_t_1 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 289, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (__pyx_t_6) {
      __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_6); __pyx_t_6 = NULL;
    }
    __Pyx_INCREF(__pyx_v_sub);
    __Pyx_GIVEREF(__pyx_v_sub);
    PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_5, __pyx_v_sub);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_5, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 289, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 289, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_7) {

    /* "Aspidites/_vendor/pyparsing_extension.py":290
 *             sub = v
 *         if isinstance(sub, ParseResults):
 *             sub.__parent = wkref(self)             # <<<<<<<<<<<<<<
 * 
 *     def __delitem__(self, i):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_wkref); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 290, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_1, __pyx_v_self) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 290, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_sub, __pyx_n_s_ParseResults__parent, __pyx_t_2) < 0) __PYX_ERR(0, 290, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":289
 *             self.__tokdict[k] = self.__tokdict.get(k, list()) + [_ParseResultsWithOffset(v, 0)]
 *             sub = v
 *         if isinstance(sub, ParseResults):             # <<<<<<<<<<<<<<
 *             sub.__parent = wkref(self)
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":279
 *                 return ParseResults([v[0] for v in self.__tokdict[i]])
 * 
 *     def __setitem__(self, k, v, isinstance=isinstance):             # <<<<<<<<<<<<<<
 *         if isinstance(v, _ParseResultsWithOffset):
 *             self.__tokdict[k] = self.__tokdict.get(k, list()) + [v]
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_sub);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":292
 *             sub.__parent = wkref(self)
 * 
 *     def __delitem__(self, i):             # <<<<<<<<<<<<<<
 *         if isinstance(i, (int, slice)):
 *             mylen = len(self.__toklist)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_9__delitem__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_9__delitem__ = {"__delitem__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_9__delitem__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_9__delitem__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_i = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__delitem__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_i)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__delitem__", 1, 2, 2, 1); __PYX_ERR(0, 292, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__delitem__") < 0)) __PYX_ERR(0, 292, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_i = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__delitem__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 292, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.__delitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_8__delitem__(__pyx_self, __pyx_v_self, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_8__delitem__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_i) {
  PyObject *__pyx_v_mylen = NULL;
  PyObject *__pyx_v_removed = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_v_occurrences = NULL;
  PyObject *__pyx_v_j = NULL;
  PyObject *__pyx_v_k = NULL;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_v_position = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *(*__pyx_t_12)(PyObject *);
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  PyObject *(*__pyx_t_15)(PyObject *);
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  int __pyx_t_19;
  PyObject *__pyx_t_20 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__delitem__", 0);
  __Pyx_INCREF(__pyx_v_i);

  /* "Aspidites/_vendor/pyparsing_extension.py":293
 * 
 *     def __delitem__(self, i):
 *         if isinstance(i, (int, slice)):             # <<<<<<<<<<<<<<
 *             mylen = len(self.__toklist)
 *             del self.__toklist[i]
 */
  __pyx_t_2 = PyInt_Check(__pyx_v_i); 
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = PySlice_Check(__pyx_v_i); 
  __pyx_t_2 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":294
 *     def __delitem__(self, i):
 *         if isinstance(i, (int, slice)):
 *             mylen = len(self.__toklist)             # <<<<<<<<<<<<<<
 *             del self.__toklist[i]
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__toklist); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 294, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyObject_Length(__pyx_t_4); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 294, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 294, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_v_mylen = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":295
 *         if isinstance(i, (int, slice)):
 *             mylen = len(self.__toklist)
 *             del self.__toklist[i]             # <<<<<<<<<<<<<<
 * 
 *             # convert int to slice
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__toklist); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 295, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (unlikely(PyObject_DelItem(__pyx_t_4, __pyx_v_i) < 0)) __PYX_ERR(0, 295, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":298
 * 
 *             # convert int to slice
 *             if isinstance(i, int):             # <<<<<<<<<<<<<<
 *                 if i < 0:
 *                     i += mylen
 */
    __pyx_t_2 = PyInt_Check(__pyx_v_i); 
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {

      /* "Aspidites/_vendor/pyparsing_extension.py":299
 *             # convert int to slice
 *             if isinstance(i, int):
 *                 if i < 0:             # <<<<<<<<<<<<<<
 *                     i += mylen
 *                 i = slice(i, i + 1)
 */
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_i, __pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 299, __pyx_L1_error)
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 299, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_1) {

        /* "Aspidites/_vendor/pyparsing_extension.py":300
 *             if isinstance(i, int):
 *                 if i < 0:
 *                     i += mylen             # <<<<<<<<<<<<<<
 *                 i = slice(i, i + 1)
 *             # get removed indices
 */
        __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_i, __pyx_v_mylen); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 300, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF_SET(__pyx_v_i, __pyx_t_4);
        __pyx_t_4 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":299
 *             # convert int to slice
 *             if isinstance(i, int):
 *                 if i < 0:             # <<<<<<<<<<<<<<
 *                     i += mylen
 *                 i = slice(i, i + 1)
 */
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":301
 *                 if i < 0:
 *                     i += mylen
 *                 i = slice(i, i + 1)             # <<<<<<<<<<<<<<
 *             # get removed indices
 *             removed = list(range(*i.indices(mylen)))
 */
      __pyx_t_4 = __Pyx_PyInt_AddObjC(__pyx_v_i, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 301, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = PySlice_New(__pyx_v_i, __pyx_t_4, Py_None); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 301, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF_SET(__pyx_v_i, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":298
 * 
 *             # convert int to slice
 *             if isinstance(i, int):             # <<<<<<<<<<<<<<
 *                 if i < 0:
 *                     i += mylen
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":303
 *                 i = slice(i, i + 1)
 *             # get removed indices
 *             removed = list(range(*i.indices(mylen)))             # <<<<<<<<<<<<<<
 *             removed.reverse()
 *             # fixup indices in token dictionary
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_i, __pyx_n_s_indices); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 303, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_6 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_7, __pyx_v_mylen) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_mylen);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 303, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PySequence_Tuple(__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 303, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_4, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 303, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PySequence_List(__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 303, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_removed = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":304
 *             # get removed indices
 *             removed = list(range(*i.indices(mylen)))
 *             removed.reverse()             # <<<<<<<<<<<<<<
 *             # fixup indices in token dictionary
 *             for name, occurrences in self.__tokdict.items():
 */
    __pyx_t_8 = PyList_Reverse(__pyx_v_removed); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 304, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":306
 *             removed.reverse()
 *             # fixup indices in token dictionary
 *             for name, occurrences in self.__tokdict.items():             # <<<<<<<<<<<<<<
 *                 for j in removed:
 *                     for k, (value, position) in enumerate(occurrences):
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__tokdict); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 306, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_items); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 306, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 306, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
      __pyx_t_7 = __pyx_t_4; __Pyx_INCREF(__pyx_t_7); __pyx_t_5 = 0;
      __pyx_t_9 = NULL;
    } else {
      __pyx_t_5 = -1; __pyx_t_7 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 306, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_9 = Py_TYPE(__pyx_t_7)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 306, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    for (;;) {
      if (likely(!__pyx_t_9)) {
        if (likely(PyList_CheckExact(__pyx_t_7))) {
          if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_7)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyList_GET_ITEM(__pyx_t_7, __pyx_t_5); __Pyx_INCREF(__pyx_t_4); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 306, __pyx_L1_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_7, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 306, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        } else {
          if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_7)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_5); __Pyx_INCREF(__pyx_t_4); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 306, __pyx_L1_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_7, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 306, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        }
      } else {
        __pyx_t_4 = __pyx_t_9(__pyx_t_7);
        if (unlikely(!__pyx_t_4)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 306, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_4);
      }
      if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
        PyObject* sequence = __pyx_t_4;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 306, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_6 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_6 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_10);
        #else
        __pyx_t_6 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 306, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_10 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 306, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        #endif
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_11 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 306, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_12 = Py_TYPE(__pyx_t_11)->tp_iternext;
        index = 0; __pyx_t_6 = __pyx_t_12(__pyx_t_11); if (unlikely(!__pyx_t_6)) goto __pyx_L10_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_6);
        index = 1; __pyx_t_10 = __pyx_t_12(__pyx_t_11); if (unlikely(!__pyx_t_10)) goto __pyx_L10_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_10);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_11), 2) < 0) __PYX_ERR(0, 306, __pyx_L1_error)
        __pyx_t_12 = NULL;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        goto __pyx_L11_unpacking_done;
        __pyx_L10_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_12 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 306, __pyx_L1_error)
        __pyx_L11_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_6);
      __pyx_t_6 = 0;
      __Pyx_XDECREF_SET(__pyx_v_occurrences, __pyx_t_10);
      __pyx_t_10 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":307
 *             # fixup indices in token dictionary
 *             for name, occurrences in self.__tokdict.items():
 *                 for j in removed:             # <<<<<<<<<<<<<<
 *                     for k, (value, position) in enumerate(occurrences):
 *                         occurrences[k] = _ParseResultsWithOffset(value, position - (position > j))
 */
      __pyx_t_4 = __pyx_v_removed; __Pyx_INCREF(__pyx_t_4); __pyx_t_13 = 0;
      for (;;) {
        if (__pyx_t_13 >= PyList_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_10 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_13); __Pyx_INCREF(__pyx_t_10); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 307, __pyx_L1_error)
        #else
        __pyx_t_10 = PySequence_ITEM(__pyx_t_4, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 307, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        #endif
        __Pyx_XDECREF_SET(__pyx_v_j, __pyx_t_10);
        __pyx_t_10 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":308
 *             for name, occurrences in self.__tokdict.items():
 *                 for j in removed:
 *                     for k, (value, position) in enumerate(occurrences):             # <<<<<<<<<<<<<<
 *                         occurrences[k] = _ParseResultsWithOffset(value, position - (position > j))
 *         else:
 */
        __Pyx_INCREF(__pyx_int_0);
        __pyx_t_10 = __pyx_int_0;
        if (likely(PyList_CheckExact(__pyx_v_occurrences)) || PyTuple_CheckExact(__pyx_v_occurrences)) {
          __pyx_t_6 = __pyx_v_occurrences; __Pyx_INCREF(__pyx_t_6); __pyx_t_14 = 0;
          __pyx_t_15 = NULL;
        } else {
          __pyx_t_14 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_v_occurrences); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 308, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_15 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 308, __pyx_L1_error)
        }
        for (;;) {
          if (likely(!__pyx_t_15)) {
            if (likely(PyList_CheckExact(__pyx_t_6))) {
              if (__pyx_t_14 >= PyList_GET_SIZE(__pyx_t_6)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_11 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_14); __Pyx_INCREF(__pyx_t_11); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 308, __pyx_L1_error)
              #else
              __pyx_t_11 = PySequence_ITEM(__pyx_t_6, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 308, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_11);
              #endif
            } else {
              if (__pyx_t_14 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_11 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_14); __Pyx_INCREF(__pyx_t_11); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 308, __pyx_L1_error)
              #else
              __pyx_t_11 = PySequence_ITEM(__pyx_t_6, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 308, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_11);
              #endif
            }
          } else {
            __pyx_t_11 = __pyx_t_15(__pyx_t_6);
            if (unlikely(!__pyx_t_11)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 308, __pyx_L1_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_11);
          }
          if ((likely(PyTuple_CheckExact(__pyx_t_11))) || (PyList_CheckExact(__pyx_t_11))) {
            PyObject* sequence = __pyx_t_11;
            Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
            if (unlikely(size != 2)) {
              if (size > 2) __Pyx_RaiseTooManyValuesError(2);
              else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
              __PYX_ERR(0, 308, __pyx_L1_error)
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            if (likely(PyTuple_CheckExact(sequence))) {
              __pyx_t_16 = PyTuple_GET_ITEM(sequence, 0); 
              __pyx_t_17 = PyTuple_GET_ITEM(sequence, 1); 
            } else {
              __pyx_t_16 = PyList_GET_ITEM(sequence, 0); 
              __pyx_t_17 = PyList_GET_ITEM(sequence, 1); 
            }
            __Pyx_INCREF(__pyx_t_16);
            __Pyx_INCREF(__pyx_t_17);
            #else
            __pyx_t_16 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 308, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_16);
            __pyx_t_17 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 308, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_17);
            #endif
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          } else {
            Py_ssize_t index = -1;
            __pyx_t_18 = PyObject_GetIter(__pyx_t_11); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 308, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_18);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            __pyx_t_12 = Py_TYPE(__pyx_t_18)->tp_iternext;
            index = 0; __pyx_t_16 = __pyx_t_12(__pyx_t_18); if (unlikely(!__pyx_t_16)) goto __pyx_L16_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_16);
            index = 1; __pyx_t_17 = __pyx_t_12(__pyx_t_18); if (unlikely(!__pyx_t_17)) goto __pyx_L16_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_17);
            if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_18), 2) < 0) __PYX_ERR(0, 308, __pyx_L1_error)
            __pyx_t_12 = NULL;
            __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
            goto __pyx_L17_unpacking_done;
            __pyx_L16_unpacking_failed:;
            __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
            __pyx_t_12 = NULL;
            if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
            __PYX_ERR(0, 308, __pyx_L1_error)
            __pyx_L17_unpacking_done:;
          }
          __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_16);
          __pyx_t_16 = 0;
          __Pyx_XDECREF_SET(__pyx_v_position, __pyx_t_17);
          __pyx_t_17 = 0;
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_XDECREF_SET(__pyx_v_k, __pyx_t_10);
          __pyx_t_11 = __Pyx_PyInt_AddObjC(__pyx_t_10, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 308, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_10);
          __pyx_t_10 = __pyx_t_11;
          __pyx_t_11 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":309
 *                 for j in removed:
 *                     for k, (value, position) in enumerate(occurrences):
 *                         occurrences[k] = _ParseResultsWithOffset(value, position - (position > j))             # <<<<<<<<<<<<<<
 *         else:
 *             del self.__tokdict[i]
 */
          __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_n_s_ParseResultsWithOffset); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 309, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_17);
          __pyx_t_16 = PyObject_RichCompare(__pyx_v_position, __pyx_v_j, Py_GT); __Pyx_XGOTREF(__pyx_t_16); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 309, __pyx_L1_error)
          __pyx_t_18 = PyNumber_Subtract(__pyx_v_position, __pyx_t_16); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 309, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_18);
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          __pyx_t_16 = NULL;
          __pyx_t_19 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_17))) {
            __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_17);
            if (likely(__pyx_t_16)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_17);
              __Pyx_INCREF(__pyx_t_16);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_17, function);
              __pyx_t_19 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_17)) {
            PyObject *__pyx_temp[3] = {__pyx_t_16, __pyx_v_value, __pyx_t_18};
            __pyx_t_11 = __Pyx_PyFunction_FastCall(__pyx_t_17, __pyx_temp+1-__pyx_t_19, 2+__pyx_t_19); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 309, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_17)) {
            PyObject *__pyx_temp[3] = {__pyx_t_16, __pyx_v_value, __pyx_t_18};
            __pyx_t_11 = __Pyx_PyCFunction_FastCall(__pyx_t_17, __pyx_temp+1-__pyx_t_19, 2+__pyx_t_19); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 309, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
          } else
          #endif
          {
            __pyx_t_20 = PyTuple_New(2+__pyx_t_19); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 309, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_20);
            if (__pyx_t_16) {
              __Pyx_GIVEREF(__pyx_t_16); PyTuple_SET_ITEM(__pyx_t_20, 0, __pyx_t_16); __pyx_t_16 = NULL;
            }
            __Pyx_INCREF(__pyx_v_value);
            __Pyx_GIVEREF(__pyx_v_value);
            PyTuple_SET_ITEM(__pyx_t_20, 0+__pyx_t_19, __pyx_v_value);
            __Pyx_GIVEREF(__pyx_t_18);
            PyTuple_SET_ITEM(__pyx_t_20, 1+__pyx_t_19, __pyx_t_18);
            __pyx_t_18 = 0;
            __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_17, __pyx_t_20, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 309, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
          }
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
          if (unlikely(PyObject_SetItem(__pyx_v_occurrences, __pyx_v_k, __pyx_t_11) < 0)) __PYX_ERR(0, 309, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":308
 *             for name, occurrences in self.__tokdict.items():
 *                 for j in removed:
 *                     for k, (value, position) in enumerate(occurrences):             # <<<<<<<<<<<<<<
 *                         occurrences[k] = _ParseResultsWithOffset(value, position - (position > j))
 *         else:
 */
        }
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":307
 *             # fixup indices in token dictionary
 *             for name, occurrences in self.__tokdict.items():
 *                 for j in removed:             # <<<<<<<<<<<<<<
 *                     for k, (value, position) in enumerate(occurrences):
 *                         occurrences[k] = _ParseResultsWithOffset(value, position - (position > j))
 */
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":306
 *             removed.reverse()
 *             # fixup indices in token dictionary
 *             for name, occurrences in self.__tokdict.items():             # <<<<<<<<<<<<<<
 *                 for j in removed:
 *                     for k, (value, position) in enumerate(occurrences):
 */
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":293
 * 
 *     def __delitem__(self, i):
 *         if isinstance(i, (int, slice)):             # <<<<<<<<<<<<<<
 *             mylen = len(self.__toklist)
 *             del self.__toklist[i]
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":311
 *                         occurrences[k] = _ParseResultsWithOffset(value, position - (position > j))
 *         else:
 *             del self.__tokdict[i]             # <<<<<<<<<<<<<<
 * 
 *     def __contains__(self, k):
 */
  /*else*/ {
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__tokdict); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 311, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (unlikely(PyObject_DelItem(__pyx_t_7, __pyx_v_i) < 0)) __PYX_ERR(0, 311, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __pyx_L3:;

  /* "Aspidites/_vendor/pyparsing_extension.py":292
 *             sub.__parent = wkref(self)
 * 
 *     def __delitem__(self, i):             # <<<<<<<<<<<<<<
 *         if isinstance(i, (int, slice)):
 *             mylen = len(self.__toklist)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.__delitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_mylen);
  __Pyx_XDECREF(__pyx_v_removed);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_occurrences);
  __Pyx_XDECREF(__pyx_v_j);
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XDECREF(__pyx_v_position);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":313
 *             del self.__tokdict[i]
 * 
 *     def __contains__(self, k):             # <<<<<<<<<<<<<<
 *         return k in self.__tokdict
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_11__contains__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_11__contains__ = {"__contains__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_11__contains__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_11__contains__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_k = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__contains__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_k,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__contains__", 1, 2, 2, 1); __PYX_ERR(0, 313, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__contains__") < 0)) __PYX_ERR(0, 313, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_k = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__contains__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 313, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.__contains__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_10__contains__(__pyx_self, __pyx_v_self, __pyx_v_k);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_10__contains__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_k) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__contains__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":314
 * 
 *     def __contains__(self, k):
 *         return k in self.__tokdict             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__tokdict); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_v_k, __pyx_t_1, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 314, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":313
 *             del self.__tokdict[i]
 * 
 *     def __contains__(self, k):             # <<<<<<<<<<<<<<
 *         return k in self.__tokdict
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.__contains__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":316
 *         return k in self.__tokdict
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return len(self.__toklist)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_13__len__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_13__len__ = {"__len__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_13__len__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_13__len__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_12__len__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_12__len__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":317
 * 
 *     def __len__(self):
 *         return len(self.__toklist)             # <<<<<<<<<<<<<<
 * 
 *     def __bool__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__toklist); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 317, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 317, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 317, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":316
 *         return k in self.__tokdict
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return len(self.__toklist)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.__len__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":319
 *         return len(self.__toklist)
 * 
 *     def __bool__(self):             # <<<<<<<<<<<<<<
 *         return (not not self.__toklist)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_15__bool__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_15__bool__ = {"__bool__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_15__bool__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_15__bool__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__bool__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_14__bool__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_14__bool__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__bool__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":320
 * 
 *     def __bool__(self):
 *         return (not not self.__toklist)             # <<<<<<<<<<<<<<
 * 
 *     __nonzero__ = __bool__
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__toklist); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyBool_FromLong((!((!__pyx_t_2) != 0))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":319
 *         return len(self.__toklist)
 * 
 *     def __bool__(self):             # <<<<<<<<<<<<<<
 *         return (not not self.__toklist)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.__bool__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":324
 *     __nonzero__ = __bool__
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return iter(self.__toklist)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_17__iter__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_17__iter__ = {"__iter__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_17__iter__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_17__iter__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_16__iter__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_16__iter__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__iter__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":325
 * 
 *     def __iter__(self):
 *         return iter(self.__toklist)             # <<<<<<<<<<<<<<
 * 
 *     def __reversed__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__toklist); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":324
 *     __nonzero__ = __bool__
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return iter(self.__toklist)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":327
 *         return iter(self.__toklist)
 * 
 *     def __reversed__(self):             # <<<<<<<<<<<<<<
 *         return iter(self.__toklist[::-1])
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_19__reversed__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_19__reversed__ = {"__reversed__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_19__reversed__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_19__reversed__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reversed__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_18__reversed__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_18__reversed__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reversed__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":328
 * 
 *     def __reversed__(self):
 *         return iter(self.__toklist[::-1])             # <<<<<<<<<<<<<<
 * 
 *     def _iterkeys(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__toklist); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_slice__13); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":327
 *         return iter(self.__toklist)
 * 
 *     def __reversed__(self):             # <<<<<<<<<<<<<<
 *         return iter(self.__toklist[::-1])
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.__reversed__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":330
 *         return iter(self.__toklist[::-1])
 * 
 *     def _iterkeys(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self.__tokdict, "iterkeys"):
 *             return self.__tokdict.iterkeys()
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_21_iterkeys(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_21_iterkeys = {"_iterkeys", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_21_iterkeys, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_21_iterkeys(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_iterkeys (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_20_iterkeys(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_20_iterkeys(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_iterkeys", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":331
 * 
 *     def _iterkeys(self):
 *         if hasattr(self.__tokdict, "iterkeys"):             # <<<<<<<<<<<<<<
 *             return self.__tokdict.iterkeys()
 *         else:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__tokdict); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 331, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_HasAttr(__pyx_t_1, __pyx_n_s_iterkeys); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 331, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/pyparsing_extension.py":332
 *     def _iterkeys(self):
 *         if hasattr(self.__tokdict, "iterkeys"):
 *             return self.__tokdict.iterkeys()             # <<<<<<<<<<<<<<
 *         else:
 *             return iter(self.__tokdict)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__tokdict); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 332, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_iterkeys); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 332, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 332, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":331
 * 
 *     def _iterkeys(self):
 *         if hasattr(self.__tokdict, "iterkeys"):             # <<<<<<<<<<<<<<
 *             return self.__tokdict.iterkeys()
 *         else:
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":334
 *             return self.__tokdict.iterkeys()
 *         else:
 *             return iter(self.__tokdict)             # <<<<<<<<<<<<<<
 * 
 *     def _itervalues(self):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__tokdict); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 334, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 334, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":330
 *         return iter(self.__toklist[::-1])
 * 
 *     def _iterkeys(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self.__tokdict, "iterkeys"):
 *             return self.__tokdict.iterkeys()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults._iterkeys", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":336
 *             return iter(self.__tokdict)
 * 
 *     def _itervalues(self):             # <<<<<<<<<<<<<<
 *         return (self[k] for k in self._iterkeys())
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_23_itervalues(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_23_itervalues = {"_itervalues", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_23_itervalues, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_23_itervalues(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_itervalues (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_22_itervalues(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_11_itervalues_2generator5(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/pyparsing_extension.py":337
 * 
 *     def _itervalues(self):
 *         return (self[k] for k in self._iterkeys())             # <<<<<<<<<<<<<<
 * 
 *     def _iteritems(self):
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_11_itervalues_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 337, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_11_itervalues_2generator5, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_ParseResults__itervalues_locals, __pyx_n_s_Aspidites__vendor_pyparsing_exte); if (unlikely(!gen)) __PYX_ERR(0, 337, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults._itervalues.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_11_itervalues_2generator5(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 337, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 337, __pyx_L1_error) }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_iterkeys_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 337, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 337, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 337, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 337, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 337, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 337, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_5(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 337, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_k);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_k, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 337, __pyx_L1_error) }
    __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_cur_scope->__pyx_v_k); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 337, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    __Pyx_XGIVEREF(__pyx_t_2);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_4;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_5;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_2);
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_5 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 337, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":336
 *             return iter(self.__tokdict)
 * 
 *     def _itervalues(self):             # <<<<<<<<<<<<<<
 *         return (self[k] for k in self._iterkeys())
 * 
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_22_itervalues(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues *__pyx_cur_scope;
  PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_11_itervalues_2generator5 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_itervalues", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 336, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);

  /* "Aspidites/_vendor/pyparsing_extension.py":337
 * 
 *     def _itervalues(self):
 *         return (self[k] for k in self._iterkeys())             # <<<<<<<<<<<<<<
 * 
 *     def _iteritems(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_11_itervalues_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":336
 *             return iter(self.__tokdict)
 * 
 *     def _itervalues(self):             # <<<<<<<<<<<<<<
 *         return (self[k] for k in self._iterkeys())
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults._itervalues", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_11_itervalues_2generator5);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":339
 *         return (self[k] for k in self._iterkeys())
 * 
 *     def _iteritems(self):             # <<<<<<<<<<<<<<
 *         return ((k, self[k]) for k in self._iterkeys())
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_25_iteritems(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_25_iteritems = {"_iteritems", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_25_iteritems, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_25_iteritems(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_iteritems (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_24_iteritems(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_10_iteritems_2generator6(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/pyparsing_extension.py":340
 * 
 *     def _iteritems(self):
 *         return ((k, self[k]) for k in self._iterkeys())             # <<<<<<<<<<<<<<
 * 
 *     if PY_3:
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_10_iteritems_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 340, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_10_iteritems_2generator6, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_ParseResults__iteritems_locals_g, __pyx_n_s_Aspidites__vendor_pyparsing_exte); if (unlikely(!gen)) __PYX_ERR(0, 340, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults._iteritems.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_10_iteritems_2generator6(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 340, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 340, __pyx_L1_error) }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_iterkeys_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 340, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 340, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 340, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 340, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 340, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 340, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 340, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 340, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_5(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 340, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_k);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_k, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 340, __pyx_L1_error) }
    __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_cur_scope->__pyx_v_k); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 340, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 340, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_k);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_k);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_cur_scope->__pyx_v_k);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    __Pyx_XGIVEREF(__pyx_t_2);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_4;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_5;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_2);
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_5 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 340, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":339
 *         return (self[k] for k in self._iterkeys())
 * 
 *     def _iteritems(self):             # <<<<<<<<<<<<<<
 *         return ((k, self[k]) for k in self._iterkeys())
 * 
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_24_iteritems(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems *__pyx_cur_scope;
  PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_10_iteritems_2generator6 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_iteritems", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 339, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);

  /* "Aspidites/_vendor/pyparsing_extension.py":340
 * 
 *     def _iteritems(self):
 *         return ((k, self[k]) for k in self._iterkeys())             # <<<<<<<<<<<<<<
 * 
 *     if PY_3:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_10_iteritems_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 340, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":339
 *         return (self[k] for k in self._iterkeys())
 * 
 *     def _iteritems(self):             # <<<<<<<<<<<<<<
 *         return ((k, self[k]) for k in self._iterkeys())
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults._iteritems", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_10_iteritems_2generator6);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":362
 *         """Returns an iterator of all named result key-value tuples (Python 2.x only)."""
 * 
 *         def keys(self):             # <<<<<<<<<<<<<<
 *             """Returns all named result keys (as a list in Python 2.x, as an iterator in Python 3.x)."""
 *             return list(self.iterkeys())
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_27keys(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_26keys[] = "Returns all named result keys (as a list in Python 2.x, as an iterator in Python 3.x).";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_27keys = {"keys", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_27keys, METH_O, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_26keys};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_27keys(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("keys (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_26keys(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_26keys(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("keys", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":364
 *         def keys(self):
 *             """Returns all named result keys (as a list in Python 2.x, as an iterator in Python 3.x)."""
 *             return list(self.iterkeys())             # <<<<<<<<<<<<<<
 * 
 *         def values(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_iterkeys); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 364, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 364, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PySequence_List(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 364, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":362
 *         """Returns an iterator of all named result key-value tuples (Python 2.x only)."""
 * 
 *         def keys(self):             # <<<<<<<<<<<<<<
 *             """Returns all named result keys (as a list in Python 2.x, as an iterator in Python 3.x)."""
 *             return list(self.iterkeys())
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.keys", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":366
 *             return list(self.iterkeys())
 * 
 *         def values(self):             # <<<<<<<<<<<<<<
 *             """Returns all named result values (as a list in Python 2.x, as an iterator in Python 3.x)."""
 *             return list(self.itervalues())
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_29values(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_28values[] = "Returns all named result values (as a list in Python 2.x, as an iterator in Python 3.x).";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_29values = {"values", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_29values, METH_O, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_28values};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_29values(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("values (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_28values(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_28values(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("values", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":368
 *         def values(self):
 *             """Returns all named result values (as a list in Python 2.x, as an iterator in Python 3.x)."""
 *             return list(self.itervalues())             # <<<<<<<<<<<<<<
 * 
 *         def items(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_itervalues); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PySequence_List(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":366
 *             return list(self.iterkeys())
 * 
 *         def values(self):             # <<<<<<<<<<<<<<
 *             """Returns all named result values (as a list in Python 2.x, as an iterator in Python 3.x)."""
 *             return list(self.itervalues())
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.values", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":370
 *             return list(self.itervalues())
 * 
 *         def items(self):             # <<<<<<<<<<<<<<
 *             """Returns all named result key-values (as a list of tuples in Python 2.x, as an iterator in Python 3.x)."""
 *             return list(self.iteritems())
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_31items(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_30items[] = "Returns all named result key-values (as a list of tuples in Python 2.x, as an iterator in Python 3.x).";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_31items = {"items", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_31items, METH_O, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_30items};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_31items(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("items (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_30items(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_30items(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("items", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":372
 *         def items(self):
 *             """Returns all named result key-values (as a list of tuples in Python 2.x, as an iterator in Python 3.x)."""
 *             return list(self.iteritems())             # <<<<<<<<<<<<<<
 * 
 *     def haskeys(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_iteritems); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PySequence_List(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":370
 *             return list(self.itervalues())
 * 
 *         def items(self):             # <<<<<<<<<<<<<<
 *             """Returns all named result key-values (as a list of tuples in Python 2.x, as an iterator in Python 3.x)."""
 *             return list(self.iteritems())
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.items", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":374
 *             return list(self.iteritems())
 * 
 *     def haskeys(self):             # <<<<<<<<<<<<<<
 *         """Since keys() returns an iterator, this method is helpful in bypassing
 *            code that looks for the existence of any defined results names."""
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_33haskeys(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_32haskeys[] = "Since keys() returns an iterator, this method is helpful in bypassing\n           code that looks for the existence of any defined results names.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_33haskeys = {"haskeys", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_33haskeys, METH_O, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_32haskeys};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_33haskeys(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("haskeys (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_32haskeys(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_32haskeys(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("haskeys", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":377
 *         """Since keys() returns an iterator, this method is helpful in bypassing
 *            code that looks for the existence of any defined results names."""
 *         return bool(self.__tokdict)             # <<<<<<<<<<<<<<
 * 
 *     def pop(self, *args, **kwargs):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__tokdict); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 377, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyBool_FromLong((!(!__pyx_t_2))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":374
 *             return list(self.iteritems())
 * 
 *     def haskeys(self):             # <<<<<<<<<<<<<<
 *         """Since keys() returns an iterator, this method is helpful in bypassing
 *            code that looks for the existence of any defined results names."""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.haskeys", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":379
 *         return bool(self.__tokdict)
 * 
 *     def pop(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         """
 *         Removes and returns item at specified index (default= ``last``).
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_35pop(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_34pop[] = "\n        Removes and returns item at specified index (default= ``last``).\n        Supports both ``list`` and ``dict`` semantics for ``pop()``. If\n        passed no argument or an integer argument, it will use ``list``\n        semantics and pop tokens from the list of parsed tokens. If passed\n        a non-integer argument (most likely a string), it will use ``dict``\n        semantics and pop the corresponding value from any defined results\n        names. A second default return value argument is supported, just as in\n        ``dict.pop()``.\n\n        Example::\n\n            def remove_first(tokens):\n                tokens.pop(0)\n            print(OneOrMore(Word(nums)).parseString(\"0 123 321\")) # -> ['0', '123', '321']\n            print(OneOrMore(Word(nums)).addParseAction(remove_first).parseString(\"0 123 321\")) # -> ['123', '321']\n\n            label = Word(alphas)\n            patt = label(\"LABEL\") + OneOrMore(Word(nums))\n            print(patt.parseString(\"AAB 123 321\").dump())\n\n            # Use pop() in a parse action to remove named result (note that corresponding value is not\n            # removed from list form of results)\n            def remove_LABEL(tokens):\n                tokens.pop(\"LABEL\")\n                return tokens\n            patt.addParseAction(remove_LABEL)\n            print(patt.parseString(\"AAB 123 321\").dump())\n\n        prints::\n\n            ['AAB', '123', '321']\n            - LABEL: AAB\n\n            ['AAB', '123', '321']\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_35pop = {"pop", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_35pop, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_34pop};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_35pop(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pop (wrapper)", 0);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  if (PyTuple_GET_SIZE(__pyx_args) > 1) {
    __pyx_v_args = PyTuple_GetSlice(__pyx_args, 1, PyTuple_GET_SIZE(__pyx_args));
    if (unlikely(!__pyx_v_args)) {
      __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
      __Pyx_RefNannyFinishContext();
      return NULL;
    }
    __Pyx_GOTREF(__pyx_v_args);
  } else {
    __pyx_v_args = __pyx_empty_tuple; __Pyx_INCREF(__pyx_empty_tuple);
  }
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        default:
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t used_pos_args = (pos_args < 1) ? pos_args : 1;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwargs, values, used_pos_args, "pop") < 0)) __PYX_ERR(0, 379, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) < 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_self = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("pop", 0, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 379, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_args); __pyx_v_args = 0;
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.pop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_34pop(__pyx_self, __pyx_v_self, __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_34pop(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_v_k = NULL;
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_v_index = NULL;
  PyObject *__pyx_v_ret = NULL;
  PyObject *__pyx_v_defaultvalue = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *(*__pyx_t_10)(PyObject *);
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pop", 0);
  __Pyx_INCREF(__pyx_v_args);

  /* "Aspidites/_vendor/pyparsing_extension.py":416
 *             ['AAB', '123', '321']
 *         """
 *         if not args:             # <<<<<<<<<<<<<<
 *             args = [-1]
 *         for k, v in kwargs.items():
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_args); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 416, __pyx_L1_error)
  __pyx_t_2 = ((!__pyx_t_1) != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":417
 *         """
 *         if not args:
 *             args = [-1]             # <<<<<<<<<<<<<<
 *         for k, v in kwargs.items():
 *             if k == 'default':
 */
    __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 417, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_int_neg_1);
    __Pyx_GIVEREF(__pyx_int_neg_1);
    PyList_SET_ITEM(__pyx_t_3, 0, __pyx_int_neg_1);
    __Pyx_DECREF_SET(__pyx_v_args, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":416
 *             ['AAB', '123', '321']
 *         """
 *         if not args:             # <<<<<<<<<<<<<<
 *             args = [-1]
 *         for k, v in kwargs.items():
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":418
 *         if not args:
 *             args = [-1]
 *         for k, v in kwargs.items():             # <<<<<<<<<<<<<<
 *             if k == 'default':
 *                 args = (args[0], v)
 */
  __pyx_t_3 = __Pyx_PyDict_Items(__pyx_v_kwargs); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 418, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_4 = __pyx_t_3; __Pyx_INCREF(__pyx_t_4); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 418, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_4))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_3); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 418, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 418, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_3); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 418, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 418, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_6(__pyx_t_4);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 418, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 418, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_7 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_8 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_7 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_8 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_8);
      #else
      __pyx_t_7 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 418, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 418, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_9 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 418, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_10 = Py_TYPE(__pyx_t_9)->tp_iternext;
      index = 0; __pyx_t_7 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_7)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_7);
      index = 1; __pyx_t_8 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_8)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_8);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_9), 2) < 0) __PYX_ERR(0, 418, __pyx_L1_error)
      __pyx_t_10 = NULL;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L7_unpacking_done;
      __pyx_L6_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_10 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 418, __pyx_L1_error)
      __pyx_L7_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_k, __pyx_t_7);
    __pyx_t_7 = 0;
    __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_8);
    __pyx_t_8 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":419
 *             args = [-1]
 *         for k, v in kwargs.items():
 *             if k == 'default':             # <<<<<<<<<<<<<<
 *                 args = (args[0], v)
 *             else:
 */
    __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_v_k, __pyx_n_s_default, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 419, __pyx_L1_error)
    if (likely(__pyx_t_2)) {

      /* "Aspidites/_vendor/pyparsing_extension.py":420
 *         for k, v in kwargs.items():
 *             if k == 'default':
 *                 args = (args[0], v)             # <<<<<<<<<<<<<<
 *             else:
 *                 raise TypeError("pop() got an unexpected keyword argument '%s'" % k)
 */
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_args, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 420, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 420, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_3);
      __Pyx_INCREF(__pyx_v_v);
      __Pyx_GIVEREF(__pyx_v_v);
      PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_v_v);
      __pyx_t_3 = 0;
      __Pyx_DECREF_SET(__pyx_v_args, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":419
 *             args = [-1]
 *         for k, v in kwargs.items():
 *             if k == 'default':             # <<<<<<<<<<<<<<
 *                 args = (args[0], v)
 *             else:
 */
      goto __pyx_L8;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":422
 *                 args = (args[0], v)
 *             else:
 *                 raise TypeError("pop() got an unexpected keyword argument '%s'" % k)             # <<<<<<<<<<<<<<
 *         if (isinstance(args[0], int)
 *                 or len(args) == 1
 */
    /*else*/ {
      __pyx_t_8 = __Pyx_PyString_FormatSafe(__pyx_kp_s_pop_got_an_unexpected_keyword_ar, __pyx_v_k); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 422, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 422, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 422, __pyx_L1_error)
    }
    __pyx_L8:;

    /* "Aspidites/_vendor/pyparsing_extension.py":418
 *         if not args:
 *             args = [-1]
 *         for k, v in kwargs.items():             # <<<<<<<<<<<<<<
 *             if k == 'default':
 *                 args = (args[0], v)
 */
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":423
 *             else:
 *                 raise TypeError("pop() got an unexpected keyword argument '%s'" % k)
 *         if (isinstance(args[0], int)             # <<<<<<<<<<<<<<
 *                 or len(args) == 1
 *                 or args[0] in self):
 */
  __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_args, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = PyInt_Check(__pyx_t_4); 
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_11 = (__pyx_t_1 != 0);
  if (!__pyx_t_11) {
  } else {
    __pyx_t_2 = __pyx_t_11;
    goto __pyx_L10_bool_binop_done;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":424
 *                 raise TypeError("pop() got an unexpected keyword argument '%s'" % k)
 *         if (isinstance(args[0], int)
 *                 or len(args) == 1             # <<<<<<<<<<<<<<
 *                 or args[0] in self):
 *             index = args[0]
 */
  __pyx_t_5 = PyObject_Length(__pyx_v_args); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 424, __pyx_L1_error)
  __pyx_t_11 = ((__pyx_t_5 == 1) != 0);
  if (!__pyx_t_11) {
  } else {
    __pyx_t_2 = __pyx_t_11;
    goto __pyx_L10_bool_binop_done;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":425
 *         if (isinstance(args[0], int)
 *                 or len(args) == 1
 *                 or args[0] in self):             # <<<<<<<<<<<<<<
 *             index = args[0]
 *             ret = self[index]
 */
  __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_args, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_11 = (__Pyx_PySequence_ContainsTF(__pyx_t_4, __pyx_v_self, Py_EQ)); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 425, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = (__pyx_t_11 != 0);
  __pyx_t_2 = __pyx_t_1;
  __pyx_L10_bool_binop_done:;

  /* "Aspidites/_vendor/pyparsing_extension.py":423
 *             else:
 *                 raise TypeError("pop() got an unexpected keyword argument '%s'" % k)
 *         if (isinstance(args[0], int)             # <<<<<<<<<<<<<<
 *                 or len(args) == 1
 *                 or args[0] in self):
 */
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":426
 *                 or len(args) == 1
 *                 or args[0] in self):
 *             index = args[0]             # <<<<<<<<<<<<<<
 *             ret = self[index]
 *             del self[index]
 */
    __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_args, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 426, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_v_index = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":427
 *                 or args[0] in self):
 *             index = args[0]
 *             ret = self[index]             # <<<<<<<<<<<<<<
 *             del self[index]
 *             return ret
 */
    __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_self, __pyx_v_index); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 427, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_v_ret = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":428
 *             index = args[0]
 *             ret = self[index]
 *             del self[index]             # <<<<<<<<<<<<<<
 *             return ret
 *         else:
 */
    if (unlikely(PyObject_DelItem(__pyx_v_self, __pyx_v_index) < 0)) __PYX_ERR(0, 428, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":429
 *             ret = self[index]
 *             del self[index]
 *             return ret             # <<<<<<<<<<<<<<
 *         else:
 *             defaultvalue = args[1]
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_ret);
    __pyx_r = __pyx_v_ret;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":423
 *             else:
 *                 raise TypeError("pop() got an unexpected keyword argument '%s'" % k)
 *         if (isinstance(args[0], int)             # <<<<<<<<<<<<<<
 *                 or len(args) == 1
 *                 or args[0] in self):
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":431
 *             return ret
 *         else:
 *             defaultvalue = args[1]             # <<<<<<<<<<<<<<
 *             return defaultvalue
 * 
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_args, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 431, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_v_defaultvalue = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":432
 *         else:
 *             defaultvalue = args[1]
 *             return defaultvalue             # <<<<<<<<<<<<<<
 * 
 *     def get(self, key, defaultValue=None):
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_defaultvalue);
    __pyx_r = __pyx_v_defaultvalue;
    goto __pyx_L0;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":379
 *         return bool(self.__tokdict)
 * 
 *     def pop(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         """
 *         Removes and returns item at specified index (default= ``last``).
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.pop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_index);
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_XDECREF(__pyx_v_defaultvalue);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":434
 *             return defaultvalue
 * 
 *     def get(self, key, defaultValue=None):             # <<<<<<<<<<<<<<
 *         """
 *         Returns named result matching the given key, or if there is no
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_37get(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_36get[] = "\n        Returns named result matching the given key, or if there is no\n        such name, then returns the given ``defaultValue`` or ``None`` if no\n        ``defaultValue`` is specified.\n\n        Similar to ``dict.get()``.\n\n        Example::\n\n            integer = Word(nums)\n            date_str = integer(\"year\") + '/' + integer(\"month\") + '/' + integer(\"day\")\n\n            result = date_str.parseString(\"1999/12/31\")\n            print(result.get(\"year\")) # -> '1999'\n            print(result.get(\"hour\", \"not specified\")) # -> 'not specified'\n            print(result.get(\"hour\")) # -> None\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_37get = {"get", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_37get, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_36get};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_37get(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_defaultValue = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_key,&__pyx_n_s_defaultValue,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_key)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get", 0, 2, 3, 1); __PYX_ERR(0, 434, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defaultValue);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "get") < 0)) __PYX_ERR(0, 434, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_key = values[1];
    __pyx_v_defaultValue = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 434, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_36get(__pyx_self, __pyx_v_self, __pyx_v_key, __pyx_v_defaultValue);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_36get(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_defaultValue) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":452
 *             print(result.get("hour")) # -> None
 *         """
 *         if key in self:             # <<<<<<<<<<<<<<
 *             return self[key]
 *         else:
 */
  __pyx_t_1 = (__Pyx_PySequence_ContainsTF(__pyx_v_key, __pyx_v_self, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 452, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":453
 *         """
 *         if key in self:
 *             return self[key]             # <<<<<<<<<<<<<<
 *         else:
 *             return defaultValue
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_self, __pyx_v_key); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 453, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":452
 *             print(result.get("hour")) # -> None
 *         """
 *         if key in self:             # <<<<<<<<<<<<<<
 *             return self[key]
 *         else:
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":455
 *             return self[key]
 *         else:
 *             return defaultValue             # <<<<<<<<<<<<<<
 * 
 *     def insert(self, index, insStr):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_defaultValue);
    __pyx_r = __pyx_v_defaultValue;
    goto __pyx_L0;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":434
 *             return defaultvalue
 * 
 *     def get(self, key, defaultValue=None):             # <<<<<<<<<<<<<<
 *         """
 *         Returns named result matching the given key, or if there is no
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":457
 *             return defaultValue
 * 
 *     def insert(self, index, insStr):             # <<<<<<<<<<<<<<
 *         """
 *         Inserts new element at location index in the list of parsed tokens.
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_39insert(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_38insert[] = "\n        Inserts new element at location index in the list of parsed tokens.\n\n        Similar to ``list.insert()``.\n\n        Example::\n\n            print(OneOrMore(Word(nums)).parseString(\"0 123 321\")) # -> ['0', '123', '321']\n\n            # use a parse action to insert the parse location in the front of the parsed results\n            def insert_locn(locn, tokens):\n                tokens.insert(0, locn)\n            print(OneOrMore(Word(nums)).addParseAction(insert_locn).parseString(\"0 123 321\")) # -> [0, '0', '123', '321']\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_39insert = {"insert", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_39insert, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_38insert};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_39insert(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_index = 0;
  PyObject *__pyx_v_insStr = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("insert (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_index,&__pyx_n_s_insStr,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_index)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("insert", 1, 3, 3, 1); __PYX_ERR(0, 457, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_insStr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("insert", 1, 3, 3, 2); __PYX_ERR(0, 457, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "insert") < 0)) __PYX_ERR(0, 457, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_index = values[1];
    __pyx_v_insStr = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("insert", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 457, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_38insert(__pyx_self, __pyx_v_self, __pyx_v_index, __pyx_v_insStr);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_38insert(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_insStr) {
  CYTHON_UNUSED PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_v_occurrences = NULL;
  PyObject *__pyx_v_k = NULL;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_v_position = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  PyObject *(*__pyx_t_9)(PyObject *);
  Py_ssize_t __pyx_t_10;
  PyObject *(*__pyx_t_11)(PyObject *);
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("insert", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":472
 *             print(OneOrMore(Word(nums)).addParseAction(insert_locn).parseString("0 123 321")) # -> [0, '0', '123', '321']
 *         """
 *         self.__toklist.insert(index, insStr)             # <<<<<<<<<<<<<<
 *         # fixup indices in token dictionary
 *         for name, occurrences in self.__tokdict.items():
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__toklist); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 472, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_insert); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 472, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_index, __pyx_v_insStr};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 472, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_index, __pyx_v_insStr};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 472, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 472, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_index);
    __Pyx_GIVEREF(__pyx_v_index);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_index);
    __Pyx_INCREF(__pyx_v_insStr);
    __Pyx_GIVEREF(__pyx_v_insStr);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_insStr);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 472, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":474
 *         self.__toklist.insert(index, insStr)
 *         # fixup indices in token dictionary
 *         for name, occurrences in self.__tokdict.items():             # <<<<<<<<<<<<<<
 *             for k, (value, position) in enumerate(occurrences):
 *                 occurrences[k] = _ParseResultsWithOffset(value, position + (position > index))
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__tokdict); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_items); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_5 = __pyx_t_1; __Pyx_INCREF(__pyx_t_5); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 474, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 474, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_5))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_6); __Pyx_INCREF(__pyx_t_1); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 474, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_5, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 474, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_6); __Pyx_INCREF(__pyx_t_1); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 474, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_5, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 474, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_7(__pyx_t_5);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 474, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
      PyObject* sequence = __pyx_t_1;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 474, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      #else
      __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 474, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 474, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_8 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 474, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_9 = Py_TYPE(__pyx_t_8)->tp_iternext;
      index = 0; __pyx_t_3 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_3)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_3);
      index = 1; __pyx_t_2 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_2)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_2);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_9(__pyx_t_8), 2) < 0) __PYX_ERR(0, 474, __pyx_L1_error)
      __pyx_t_9 = NULL;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_9 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 474, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_occurrences, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":475
 *         # fixup indices in token dictionary
 *         for name, occurrences in self.__tokdict.items():
 *             for k, (value, position) in enumerate(occurrences):             # <<<<<<<<<<<<<<
 *                 occurrences[k] = _ParseResultsWithOffset(value, position + (position > index))
 * 
 */
    __Pyx_INCREF(__pyx_int_0);
    __pyx_t_1 = __pyx_int_0;
    if (likely(PyList_CheckExact(__pyx_v_occurrences)) || PyTuple_CheckExact(__pyx_v_occurrences)) {
      __pyx_t_2 = __pyx_v_occurrences; __Pyx_INCREF(__pyx_t_2); __pyx_t_10 = 0;
      __pyx_t_11 = NULL;
    } else {
      __pyx_t_10 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_occurrences); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 475, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_11 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 475, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_11)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_10); __Pyx_INCREF(__pyx_t_3); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 475, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 475, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        } else {
          if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_10); __Pyx_INCREF(__pyx_t_3); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 475, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 475, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        }
      } else {
        __pyx_t_3 = __pyx_t_11(__pyx_t_2);
        if (unlikely(!__pyx_t_3)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 475, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_3);
      }
      if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
        PyObject* sequence = __pyx_t_3;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 475, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_8 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_12 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_8 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_12 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_12);
        #else
        __pyx_t_8 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 475, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_12 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 475, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        #endif
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_13 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 475, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_9 = Py_TYPE(__pyx_t_13)->tp_iternext;
        index = 0; __pyx_t_8 = __pyx_t_9(__pyx_t_13); if (unlikely(!__pyx_t_8)) goto __pyx_L9_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_8);
        index = 1; __pyx_t_12 = __pyx_t_9(__pyx_t_13); if (unlikely(!__pyx_t_12)) goto __pyx_L9_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_12);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_9(__pyx_t_13), 2) < 0) __PYX_ERR(0, 475, __pyx_L1_error)
        __pyx_t_9 = NULL;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        goto __pyx_L10_unpacking_done;
        __pyx_L9_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __pyx_t_9 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 475, __pyx_L1_error)
        __pyx_L10_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_8);
      __pyx_t_8 = 0;
      __Pyx_XDECREF_SET(__pyx_v_position, __pyx_t_12);
      __pyx_t_12 = 0;
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_k, __pyx_t_1);
      __pyx_t_3 = __Pyx_PyInt_AddObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 475, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1);
      __pyx_t_1 = __pyx_t_3;
      __pyx_t_3 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":476
 *         for name, occurrences in self.__tokdict.items():
 *             for k, (value, position) in enumerate(occurrences):
 *                 occurrences[k] = _ParseResultsWithOffset(value, position + (position > index))             # <<<<<<<<<<<<<<
 * 
 *     def append(self, item):
 */
      __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_ParseResultsWithOffset); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 476, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_t_8 = PyObject_RichCompare(__pyx_v_position, __pyx_v_index, Py_GT); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 476, __pyx_L1_error)
      __pyx_t_13 = PyNumber_Add(__pyx_v_position, __pyx_t_8); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 476, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = NULL;
      __pyx_t_4 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_12))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_12);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_12, function);
          __pyx_t_4 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_12)) {
        PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_value, __pyx_t_13};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 476, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_12)) {
        PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_value, __pyx_t_13};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 476, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      } else
      #endif
      {
        __pyx_t_14 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 476, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
        if (__pyx_t_8) {
          __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_8); __pyx_t_8 = NULL;
        }
        __Pyx_INCREF(__pyx_v_value);
        __Pyx_GIVEREF(__pyx_v_value);
        PyTuple_SET_ITEM(__pyx_t_14, 0+__pyx_t_4, __pyx_v_value);
        __Pyx_GIVEREF(__pyx_t_13);
        PyTuple_SET_ITEM(__pyx_t_14, 1+__pyx_t_4, __pyx_t_13);
        __pyx_t_13 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_14, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 476, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      }
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      if (unlikely(PyObject_SetItem(__pyx_v_occurrences, __pyx_v_k, __pyx_t_3) < 0)) __PYX_ERR(0, 476, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":475
 *         # fixup indices in token dictionary
 *         for name, occurrences in self.__tokdict.items():
 *             for k, (value, position) in enumerate(occurrences):             # <<<<<<<<<<<<<<
 *                 occurrences[k] = _ParseResultsWithOffset(value, position + (position > index))
 * 
 */
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":474
 *         self.__toklist.insert(index, insStr)
 *         # fixup indices in token dictionary
 *         for name, occurrences in self.__tokdict.items():             # <<<<<<<<<<<<<<
 *             for k, (value, position) in enumerate(occurrences):
 *                 occurrences[k] = _ParseResultsWithOffset(value, position + (position > index))
 */
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":457
 *             return defaultValue
 * 
 *     def insert(self, index, insStr):             # <<<<<<<<<<<<<<
 *         """
 *         Inserts new element at location index in the list of parsed tokens.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_occurrences);
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XDECREF(__pyx_v_position);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":478
 *                 occurrences[k] = _ParseResultsWithOffset(value, position + (position > index))
 * 
 *     def append(self, item):             # <<<<<<<<<<<<<<
 *         """
 *         Add single element to end of ParseResults list of elements.
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_41append(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_40append[] = "\n        Add single element to end of ParseResults list of elements.\n\n        Example::\n\n            print(OneOrMore(Word(nums)).parseString(\"0 123 321\")) # -> ['0', '123', '321']\n\n            # use a parse action to compute the sum of the parsed integers, and add it to the end\n            def append_sum(tokens):\n                tokens.append(sum(map(int, tokens)))\n            print(OneOrMore(Word(nums)).addParseAction(append_sum).parseString(\"0 123 321\")) # -> ['0', '123', '321', 444]\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_41append = {"append", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_41append, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_40append};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_41append(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_item = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("append (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_item,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_item)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("append", 1, 2, 2, 1); __PYX_ERR(0, 478, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "append") < 0)) __PYX_ERR(0, 478, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_item = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("append", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 478, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_40append(__pyx_self, __pyx_v_self, __pyx_v_item);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_40append(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_item) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("append", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":491
 *             print(OneOrMore(Word(nums)).addParseAction(append_sum).parseString("0 123 321")) # -> ['0', '123', '321', 444]
 *         """
 *         self.__toklist.append(item)             # <<<<<<<<<<<<<<
 * 
 *     def extend(self, itemseq):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__toklist); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Append(__pyx_t_1, __pyx_v_item); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 491, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":478
 *                 occurrences[k] = _ParseResultsWithOffset(value, position + (position > index))
 * 
 *     def append(self, item):             # <<<<<<<<<<<<<<
 *         """
 *         Add single element to end of ParseResults list of elements.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":493
 *         self.__toklist.append(item)
 * 
 *     def extend(self, itemseq):             # <<<<<<<<<<<<<<
 *         """
 *         Add sequence of elements to end of ParseResults list of elements.
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_43extend(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_42extend[] = "\n        Add sequence of elements to end of ParseResults list of elements.\n\n        Example::\n\n            patt = OneOrMore(Word(alphas))\n\n            # use a parse action to append the reverse of the matched strings, to make a palindrome\n            def make_palindrome(tokens):\n                tokens.extend(reversed([t[::-1] for t in tokens]))\n                return ''.join(tokens)\n            print(patt.addParseAction(make_palindrome).parseString(\"lskdj sdlkjf lksd\")) # -> 'lskdjsdlkjflksddsklfjkldsjdksl'\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_43extend = {"extend", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_43extend, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_42extend};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_43extend(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_itemseq = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("extend (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_itemseq,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_itemseq)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("extend", 1, 2, 2, 1); __PYX_ERR(0, 493, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "extend") < 0)) __PYX_ERR(0, 493, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_itemseq = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("extend", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 493, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.extend", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_42extend(__pyx_self, __pyx_v_self, __pyx_v_itemseq);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_42extend(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_itemseq) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("extend", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":507
 *             print(patt.addParseAction(make_palindrome).parseString("lskdj sdlkjf lksd")) # -> 'lskdjsdlkjflksddsklfjkldsjdksl'
 *         """
 *         if isinstance(itemseq, ParseResults):             # <<<<<<<<<<<<<<
 *             self.__iadd__(itemseq)
 *         else:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParseResults); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 507, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_IsInstance(__pyx_v_itemseq, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 507, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/pyparsing_extension.py":508
 *         """
 *         if isinstance(itemseq, ParseResults):
 *             self.__iadd__(itemseq)             # <<<<<<<<<<<<<<
 *         else:
 *             self.__toklist.extend(itemseq)
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_iadd); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 508, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_itemseq) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_itemseq);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 508, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":507
 *             print(patt.addParseAction(make_palindrome).parseString("lskdj sdlkjf lksd")) # -> 'lskdjsdlkjflksddsklfjkldsjdksl'
 *         """
 *         if isinstance(itemseq, ParseResults):             # <<<<<<<<<<<<<<
 *             self.__iadd__(itemseq)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":510
 *             self.__iadd__(itemseq)
 *         else:
 *             self.__toklist.extend(itemseq)             # <<<<<<<<<<<<<<
 * 
 *     def clear(self):
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__toklist); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_extend); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_v_itemseq) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_itemseq);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_L3:;

  /* "Aspidites/_vendor/pyparsing_extension.py":493
 *         self.__toklist.append(item)
 * 
 *     def extend(self, itemseq):             # <<<<<<<<<<<<<<
 *         """
 *         Add sequence of elements to end of ParseResults list of elements.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.extend", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":512
 *             self.__toklist.extend(itemseq)
 * 
 *     def clear(self):             # <<<<<<<<<<<<<<
 *         """
 *         Clear all elements and results names.
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_45clear(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_44clear[] = "\n        Clear all elements and results names.\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_45clear = {"clear", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_45clear, METH_O, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_44clear};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_45clear(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clear (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_44clear(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_44clear(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clear", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":516
 *         Clear all elements and results names.
 *         """
 *         del self.__toklist[:]             # <<<<<<<<<<<<<<
 *         self.__tokdict.clear()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__toklist); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_DelSlice(__pyx_t_1, 0, 0, NULL, NULL, &__pyx_slice__11, 0, 0, 1) < 0) __PYX_ERR(0, 516, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":517
 *         """
 *         del self.__toklist[:]
 *         self.__tokdict.clear()             # <<<<<<<<<<<<<<
 * 
 *     def __getattr__(self, name):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__tokdict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_clear); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":512
 *             self.__toklist.extend(itemseq)
 * 
 *     def clear(self):             # <<<<<<<<<<<<<<
 *         """
 *         Clear all elements and results names.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.clear", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":519
 *         self.__tokdict.clear()
 * 
 *     def __getattr__(self, name):             # <<<<<<<<<<<<<<
 *         try:
 *             return self[name]
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_47__getattr__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_47__getattr__ = {"__getattr__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_47__getattr__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_47__getattr__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_name = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getattr__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_name,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__getattr__", 1, 2, 2, 1); __PYX_ERR(0, 519, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__getattr__") < 0)) __PYX_ERR(0, 519, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_name = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__getattr__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 519, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.__getattr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_46__getattr__(__pyx_self, __pyx_v_self, __pyx_v_name);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_46__getattr__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_name) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getattr__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":520
 * 
 *     def __getattr__(self, name):
 *         try:             # <<<<<<<<<<<<<<
 *             return self[name]
 *         except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "Aspidites/_vendor/pyparsing_extension.py":521
 *     def __getattr__(self, name):
 *         try:
 *             return self[name]             # <<<<<<<<<<<<<<
 *         except KeyError:
 *             return ""
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_self, __pyx_v_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 521, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_r = __pyx_t_4;
      __pyx_t_4 = 0;
      goto __pyx_L7_try_return;

      /* "Aspidites/_vendor/pyparsing_extension.py":520
 * 
 *     def __getattr__(self, name):
 *         try:             # <<<<<<<<<<<<<<
 *             return self[name]
 *         except KeyError:
 */
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":522
 *         try:
 *             return self[name]
 *         except KeyError:             # <<<<<<<<<<<<<<
 *             return ""
 * 
 */
    __pyx_t_5 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_5) {
      __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.__getattr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(0, 522, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "Aspidites/_vendor/pyparsing_extension.py":523
 *             return self[name]
 *         except KeyError:
 *             return ""             # <<<<<<<<<<<<<<
 * 
 *     def __add__(self, other):
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_kp_s__12);
      __pyx_r = __pyx_kp_s__12;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L6_except_return;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "Aspidites/_vendor/pyparsing_extension.py":520
 * 
 *     def __getattr__(self, name):
 *         try:             # <<<<<<<<<<<<<<
 *             return self[name]
 *         except KeyError:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":519
 *         self.__tokdict.clear()
 * 
 *     def __getattr__(self, name):             # <<<<<<<<<<<<<<
 *         try:
 *             return self[name]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.__getattr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":525
 *             return ""
 * 
 *     def __add__(self, other):             # <<<<<<<<<<<<<<
 *         ret = self.copy()
 *         ret += other
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_49__add__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_49__add__ = {"__add__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_49__add__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_49__add__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_other = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__add__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_other,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_other)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__add__", 1, 2, 2, 1); __PYX_ERR(0, 525, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__add__") < 0)) __PYX_ERR(0, 525, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_other = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__add__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 525, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.__add__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_48__add__(__pyx_self, __pyx_v_self, __pyx_v_other);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_48__add__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_v_ret = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__add__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":526
 * 
 *     def __add__(self, other):
 *         ret = self.copy()             # <<<<<<<<<<<<<<
 *         ret += other
 *         return ret
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_ret = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":527
 *     def __add__(self, other):
 *         ret = self.copy()
 *         ret += other             # <<<<<<<<<<<<<<
 *         return ret
 * 
 */
  __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_ret, __pyx_v_other); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 527, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_ret, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":528
 *         ret = self.copy()
 *         ret += other
 *         return ret             # <<<<<<<<<<<<<<
 * 
 *     def __iadd__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_ret);
  __pyx_r = __pyx_v_ret;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":525
 *             return ""
 * 
 *     def __add__(self, other):             # <<<<<<<<<<<<<<
 *         ret = self.copy()
 *         ret += other
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.__add__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":530
 *         return ret
 * 
 *     def __iadd__(self, other):             # <<<<<<<<<<<<<<
 *         if other.__tokdict:
 *             offset = len(self.__toklist)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_51__iadd__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_51__iadd__ = {"__iadd__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_51__iadd__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_51__iadd__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_other = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iadd__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_other,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_other)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__iadd__", 1, 2, 2, 1); __PYX_ERR(0, 530, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__iadd__") < 0)) __PYX_ERR(0, 530, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_other = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__iadd__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 530, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.__iadd__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_50__iadd__(__pyx_self, __pyx_v_self, __pyx_v_other);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":533
 *         if other.__tokdict:
 *             offset = len(self.__toklist)
 *             addoffset = lambda a: offset if a < 0 else a + offset             # <<<<<<<<<<<<<<
 *             otheritems = other.__tokdict.items()
 *             otherdictitems = [(k, _ParseResultsWithOffset(v[0], addoffset(v[1])))
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_8__iadd___lambda5(PyObject *__pyx_self, PyObject *__pyx_v_a); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_8__iadd___lambda5 = {"lambda5", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_8__iadd___lambda5, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_8__iadd___lambda5(PyObject *__pyx_self, PyObject *__pyx_v_a) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda5 (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_lambda5(__pyx_self, ((PyObject *)__pyx_v_a));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda5(PyObject *__pyx_self, PyObject *__pyx_v_a) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_7___iadd__ *__pyx_cur_scope;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_7___iadd__ *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda5", 0);
  __pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_7___iadd__ *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_a, __pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 533, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 533, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_cur_scope->__pyx_v_offset); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 533, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __pyx_t_2;
    __pyx_t_2 = 0;
  } else {
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_cur_scope->__pyx_v_offset); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 533, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyNumber_Add(__pyx_v_a, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 533, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_1 = __pyx_t_4;
    __pyx_t_4 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.__iadd__.lambda5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":530
 *         return ret
 * 
 *     def __iadd__(self, other):             # <<<<<<<<<<<<<<
 *         if other.__tokdict:
 *             offset = len(self.__toklist)
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_50__iadd__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_7___iadd__ *__pyx_cur_scope;
  PyObject *__pyx_v_addoffset = NULL;
  PyObject *__pyx_v_otheritems = NULL;
  PyObject *__pyx_v_otherdictitems = NULL;
  PyObject *__pyx_v_k = NULL;
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_v_vlist = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *(*__pyx_t_6)(PyObject *);
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *(*__pyx_t_10)(PyObject *);
  Py_ssize_t __pyx_t_11;
  PyObject *(*__pyx_t_12)(PyObject *);
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  int __pyx_t_17;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__iadd__", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_7___iadd__ *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_7___iadd__(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_7___iadd__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_7___iadd__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 530, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":531
 * 
 *     def __iadd__(self, other):
 *         if other.__tokdict:             # <<<<<<<<<<<<<<
 *             offset = len(self.__toklist)
 *             addoffset = lambda a: offset if a < 0 else a + offset
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_ParseResults__tokdict); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 531, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":532
 *     def __iadd__(self, other):
 *         if other.__tokdict:
 *             offset = len(self.__toklist)             # <<<<<<<<<<<<<<
 *             addoffset = lambda a: offset if a < 0 else a + offset
 *             otheritems = other.__tokdict.items()
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__toklist); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 532, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 532, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_cur_scope->__pyx_v_offset = __pyx_t_3;

    /* "Aspidites/_vendor/pyparsing_extension.py":533
 *         if other.__tokdict:
 *             offset = len(self.__toklist)
 *             addoffset = lambda a: offset if a < 0 else a + offset             # <<<<<<<<<<<<<<
 *             otheritems = other.__tokdict.items()
 *             otherdictitems = [(k, _ParseResultsWithOffset(v[0], addoffset(v[1])))
 */
    __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_8__iadd___lambda5, 0, __pyx_n_s_ParseResults___iadd___locals_lam, ((PyObject*)__pyx_cur_scope), __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 533, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_addoffset = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":534
 *             offset = len(self.__toklist)
 *             addoffset = lambda a: offset if a < 0 else a + offset
 *             otheritems = other.__tokdict.items()             # <<<<<<<<<<<<<<
 *             otherdictitems = [(k, _ParseResultsWithOffset(v[0], addoffset(v[1])))
 *                               for k, vlist in otheritems for v in vlist]
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_ParseResults__tokdict); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 534, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_items); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 534, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 534, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_otheritems = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":535
 *             addoffset = lambda a: offset if a < 0 else a + offset
 *             otheritems = other.__tokdict.items()
 *             otherdictitems = [(k, _ParseResultsWithOffset(v[0], addoffset(v[1])))             # <<<<<<<<<<<<<<
 *                               for k, vlist in otheritems for v in vlist]
 *             for k, v in otherdictitems:
 */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 535, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "Aspidites/_vendor/pyparsing_extension.py":536
 *             otheritems = other.__tokdict.items()
 *             otherdictitems = [(k, _ParseResultsWithOffset(v[0], addoffset(v[1])))
 *                               for k, vlist in otheritems for v in vlist]             # <<<<<<<<<<<<<<
 *             for k, v in otherdictitems:
 *                 self[k] = v
 */
    if (likely(PyList_CheckExact(__pyx_v_otheritems)) || PyTuple_CheckExact(__pyx_v_otheritems)) {
      __pyx_t_5 = __pyx_v_otheritems; __Pyx_INCREF(__pyx_t_5); __pyx_t_3 = 0;
      __pyx_t_6 = NULL;
    } else {
      __pyx_t_3 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_v_otheritems); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 536, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 536, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_6)) {
        if (likely(PyList_CheckExact(__pyx_t_5))) {
          if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_5)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_3); __Pyx_INCREF(__pyx_t_4); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 536, __pyx_L1_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_5, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 536, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        } else {
          if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_3); __Pyx_INCREF(__pyx_t_4); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 536, __pyx_L1_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_5, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 536, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        }
      } else {
        __pyx_t_4 = __pyx_t_6(__pyx_t_5);
        if (unlikely(!__pyx_t_4)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 536, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_4);
      }
      if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
        PyObject* sequence = __pyx_t_4;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 536, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_7 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_8 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_7 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_8 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        #else
        __pyx_t_7 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 536, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 536, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        #endif
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_9 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 536, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_10 = Py_TYPE(__pyx_t_9)->tp_iternext;
        index = 0; __pyx_t_7 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_7)) goto __pyx_L6_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_7);
        index = 1; __pyx_t_8 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_8)) goto __pyx_L6_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_8);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_9), 2) < 0) __PYX_ERR(0, 536, __pyx_L1_error)
        __pyx_t_10 = NULL;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L7_unpacking_done;
        __pyx_L6_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_10 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 536, __pyx_L1_error)
        __pyx_L7_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_k, __pyx_t_7);
      __pyx_t_7 = 0;
      __Pyx_XDECREF_SET(__pyx_v_vlist, __pyx_t_8);
      __pyx_t_8 = 0;
      if (likely(PyList_CheckExact(__pyx_v_vlist)) || PyTuple_CheckExact(__pyx_v_vlist)) {
        __pyx_t_4 = __pyx_v_vlist; __Pyx_INCREF(__pyx_t_4); __pyx_t_11 = 0;
        __pyx_t_12 = NULL;
      } else {
        __pyx_t_11 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_vlist); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 536, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_12 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 536, __pyx_L1_error)
      }
      for (;;) {
        if (likely(!__pyx_t_12)) {
          if (likely(PyList_CheckExact(__pyx_t_4))) {
            if (__pyx_t_11 >= PyList_GET_SIZE(__pyx_t_4)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_8 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_11); __Pyx_INCREF(__pyx_t_8); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 536, __pyx_L1_error)
            #else
            __pyx_t_8 = PySequence_ITEM(__pyx_t_4, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 536, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_8);
            #endif
          } else {
            if (__pyx_t_11 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_11); __Pyx_INCREF(__pyx_t_8); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 536, __pyx_L1_error)
            #else
            __pyx_t_8 = PySequence_ITEM(__pyx_t_4, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 536, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_8);
            #endif
          }
        } else {
          __pyx_t_8 = __pyx_t_12(__pyx_t_4);
          if (unlikely(!__pyx_t_8)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 536, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_8);
        }
        __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_8);
        __pyx_t_8 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":535
 *             addoffset = lambda a: offset if a < 0 else a + offset
 *             otheritems = other.__tokdict.items()
 *             otherdictitems = [(k, _ParseResultsWithOffset(v[0], addoffset(v[1])))             # <<<<<<<<<<<<<<
 *                               for k, vlist in otheritems for v in vlist]
 *             for k, v in otherdictitems:
 */
        __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_ParseResultsWithOffset); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 535, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_9 = __Pyx_GetItemInt(__pyx_v_v, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 535, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_13 = __Pyx_GetItemInt(__pyx_v_v, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 535, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        __pyx_t_14 = __pyx_lambda_funcdef_lambda5(__pyx_v_addoffset, __pyx_t_13); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 535, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __pyx_t_13 = NULL;
        __pyx_t_15 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_13)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_13);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_15 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_13, __pyx_t_9, __pyx_t_14};
          __pyx_t_8 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_15, 2+__pyx_t_15); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 535, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_13, __pyx_t_9, __pyx_t_14};
          __pyx_t_8 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_15, 2+__pyx_t_15); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 535, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        } else
        #endif
        {
          __pyx_t_16 = PyTuple_New(2+__pyx_t_15); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 535, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_16);
          if (__pyx_t_13) {
            __Pyx_GIVEREF(__pyx_t_13); PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_t_13); __pyx_t_13 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_9);
          PyTuple_SET_ITEM(__pyx_t_16, 0+__pyx_t_15, __pyx_t_9);
          __Pyx_GIVEREF(__pyx_t_14);
          PyTuple_SET_ITEM(__pyx_t_16, 1+__pyx_t_15, __pyx_t_14);
          __pyx_t_9 = 0;
          __pyx_t_14 = 0;
          __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_16, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 535, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 535, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_v_k);
        __Pyx_GIVEREF(__pyx_v_k);
        PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_k);
        __Pyx_GIVEREF(__pyx_t_8);
        PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_8);
        __pyx_t_8 = 0;
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_7))) __PYX_ERR(0, 535, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":536
 *             otheritems = other.__tokdict.items()
 *             otherdictitems = [(k, _ParseResultsWithOffset(v[0], addoffset(v[1])))
 *                               for k, vlist in otheritems for v in vlist]             # <<<<<<<<<<<<<<
 *             for k, v in otherdictitems:
 *                 self[k] = v
 */
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_otherdictitems = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":537
 *             otherdictitems = [(k, _ParseResultsWithOffset(v[0], addoffset(v[1])))
 *                               for k, vlist in otheritems for v in vlist]
 *             for k, v in otherdictitems:             # <<<<<<<<<<<<<<
 *                 self[k] = v
 *                 if isinstance(v[0], ParseResults):
 */
    __pyx_t_1 = __pyx_v_otherdictitems; __Pyx_INCREF(__pyx_t_1); __pyx_t_3 = 0;
    for (;;) {
      if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_1)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_5 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 537, __pyx_L1_error)
      #else
      __pyx_t_5 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 537, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
      if ((likely(PyTuple_CheckExact(__pyx_t_5))) || (PyList_CheckExact(__pyx_t_5))) {
        PyObject* sequence = __pyx_t_5;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 537, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_7 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_7 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        #else
        __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 537, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_7 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 537, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_8 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 537, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_10 = Py_TYPE(__pyx_t_8)->tp_iternext;
        index = 0; __pyx_t_4 = __pyx_t_10(__pyx_t_8); if (unlikely(!__pyx_t_4)) goto __pyx_L12_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_4);
        index = 1; __pyx_t_7 = __pyx_t_10(__pyx_t_8); if (unlikely(!__pyx_t_7)) goto __pyx_L12_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_7);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_8), 2) < 0) __PYX_ERR(0, 537, __pyx_L1_error)
        __pyx_t_10 = NULL;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L13_unpacking_done;
        __pyx_L12_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_10 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 537, __pyx_L1_error)
        __pyx_L13_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_k, __pyx_t_4);
      __pyx_t_4 = 0;
      __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":538
 *                               for k, vlist in otheritems for v in vlist]
 *             for k, v in otherdictitems:
 *                 self[k] = v             # <<<<<<<<<<<<<<
 *                 if isinstance(v[0], ParseResults):
 *                     v[0].__parent = wkref(self)
 */
      if (unlikely(PyObject_SetItem(__pyx_v_self, __pyx_v_k, __pyx_v_v) < 0)) __PYX_ERR(0, 538, __pyx_L1_error)

      /* "Aspidites/_vendor/pyparsing_extension.py":539
 *             for k, v in otherdictitems:
 *                 self[k] = v
 *                 if isinstance(v[0], ParseResults):             # <<<<<<<<<<<<<<
 *                     v[0].__parent = wkref(self)
 * 
 */
      __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_v, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 539, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_ParseResults); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 539, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_2 = PyObject_IsInstance(__pyx_t_5, __pyx_t_7); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 539, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_17 = (__pyx_t_2 != 0);
      if (__pyx_t_17) {

        /* "Aspidites/_vendor/pyparsing_extension.py":540
 *                 self[k] = v
 *                 if isinstance(v[0], ParseResults):
 *                     v[0].__parent = wkref(self)             # <<<<<<<<<<<<<<
 * 
 *         self.__toklist += other.__toklist
 */
        __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_wkref); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 540, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
          }
        }
        __pyx_t_7 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_v_self) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_self);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 540, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_v, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 540, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_ParseResults__parent, __pyx_t_7) < 0) __PYX_ERR(0, 540, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":539
 *             for k, v in otherdictitems:
 *                 self[k] = v
 *                 if isinstance(v[0], ParseResults):             # <<<<<<<<<<<<<<
 *                     v[0].__parent = wkref(self)
 * 
 */
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":537
 *             otherdictitems = [(k, _ParseResultsWithOffset(v[0], addoffset(v[1])))
 *                               for k, vlist in otheritems for v in vlist]
 *             for k, v in otherdictitems:             # <<<<<<<<<<<<<<
 *                 self[k] = v
 *                 if isinstance(v[0], ParseResults):
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":531
 * 
 *     def __iadd__(self, other):
 *         if other.__tokdict:             # <<<<<<<<<<<<<<
 *             offset = len(self.__toklist)
 *             addoffset = lambda a: offset if a < 0 else a + offset
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":542
 *                     v[0].__parent = wkref(self)
 * 
 *         self.__toklist += other.__toklist             # <<<<<<<<<<<<<<
 *         self.__accumNames.update(other.__accumNames)
 *         return self
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__toklist); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_ParseResults__toklist); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyNumber_InPlaceAdd(__pyx_t_1, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__toklist, __pyx_t_7) < 0) __PYX_ERR(0, 542, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":543
 * 
 *         self.__toklist += other.__toklist
 *         self.__accumNames.update(other.__accumNames)             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__accumNames); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_ParseResults__accumNames); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_7 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":544
 *         self.__toklist += other.__toklist
 *         self.__accumNames.update(other.__accumNames)
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __radd__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":530
 *         return ret
 * 
 *     def __iadd__(self, other):             # <<<<<<<<<<<<<<
 *         if other.__tokdict:
 *             offset = len(self.__toklist)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.__iadd__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_addoffset);
  __Pyx_XDECREF(__pyx_v_otheritems);
  __Pyx_XDECREF(__pyx_v_otherdictitems);
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_vlist);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":546
 *         return self
 * 
 *     def __radd__(self, other):             # <<<<<<<<<<<<<<
 *         if isinstance(other, int) and other == 0:
 *             # useful for merging many ParseResults using sum() builtin
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_53__radd__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_53__radd__ = {"__radd__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_53__radd__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_53__radd__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_other = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__radd__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_other,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_other)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__radd__", 1, 2, 2, 1); __PYX_ERR(0, 546, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__radd__") < 0)) __PYX_ERR(0, 546, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_other = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__radd__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 546, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.__radd__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_52__radd__(__pyx_self, __pyx_v_self, __pyx_v_other);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_52__radd__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__radd__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":547
 * 
 *     def __radd__(self, other):
 *         if isinstance(other, int) and other == 0:             # <<<<<<<<<<<<<<
 *             # useful for merging many ParseResults using sum() builtin
 *             return self.copy()
 */
  __pyx_t_2 = PyInt_Check(__pyx_v_other); 
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyInt_EqObjC(__pyx_v_other, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 547, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 547, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "Aspidites/_vendor/pyparsing_extension.py":549
 *         if isinstance(other, int) and other == 0:
 *             # useful for merging many ParseResults using sum() builtin
 *             return self.copy()             # <<<<<<<<<<<<<<
 *         else:
 *             # this may raise a TypeError - so be it
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copy); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 549, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 549, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":547
 * 
 *     def __radd__(self, other):
 *         if isinstance(other, int) and other == 0:             # <<<<<<<<<<<<<<
 *             # useful for merging many ParseResults using sum() builtin
 *             return self.copy()
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":552
 *         else:
 *             # this may raise a TypeError - so be it
 *             return other + self             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = PyNumber_Add(__pyx_v_other, __pyx_v_self); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 552, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":546
 *         return self
 * 
 *     def __radd__(self, other):             # <<<<<<<<<<<<<<
 *         if isinstance(other, int) and other == 0:
 *             # useful for merging many ParseResults using sum() builtin
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.__radd__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":554
 *             return other + self
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "(%s, %s)" % (repr(self.__toklist), repr(self.__tokdict))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_55__repr__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_55__repr__ = {"__repr__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_55__repr__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_55__repr__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_54__repr__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_54__repr__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":555
 * 
 *     def __repr__(self):
 *         return "(%s, %s)" % (repr(self.__toklist), repr(self.__tokdict))             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__toklist); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 555, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 555, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__tokdict); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 555, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 555, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 555, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_s_s, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 555, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":554
 *             return other + self
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "(%s, %s)" % (repr(self.__toklist), repr(self.__tokdict))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":557
 *         return "(%s, %s)" % (repr(self.__toklist), repr(self.__tokdict))
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return '[' + ', '.join(str(i) if isinstance(i, ParseResults) else repr(i) for i in self.__toklist) + ']'
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_57__str__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_57__str__ = {"__str__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_57__str__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_57__str__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_56__str__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_7__str___2generator7(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/pyparsing_extension.py":558
 * 
 *     def __str__(self):
 *         return '[' + ', '.join(str(i) if isinstance(i, ParseResults) else repr(i) for i in self.__toklist) + ']'             # <<<<<<<<<<<<<<
 * 
 *     def _asStringList(self, sep=''):
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_7__str___genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 558, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__ *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_7__str___2generator7, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_ParseResults___str___locals_gene, __pyx_n_s_Aspidites__vendor_pyparsing_exte); if (unlikely(!gen)) __PYX_ERR(0, 558, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.__str__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_7__str___2generator7(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 558, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 558, __pyx_L1_error) }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_ParseResults__toklist); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 558, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 558, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 558, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 558, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 558, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 558, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 558, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 558, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_i);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_i, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_ParseResults); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 558, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = PyObject_IsInstance(__pyx_cur_scope->__pyx_v_i, __pyx_t_5); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 558, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if ((__pyx_t_6 != 0)) {
      __pyx_t_5 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_cur_scope->__pyx_v_i); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 558, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_1 = __pyx_t_5;
      __pyx_t_5 = 0;
    } else {
      __pyx_t_5 = PyObject_Repr(__pyx_cur_scope->__pyx_v_i); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 558, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_1 = __pyx_t_5;
      __pyx_t_5 = 0;
    }
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    __Pyx_XGIVEREF(__pyx_t_2);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_3;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_4;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_2);
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 558, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":557
 *         return "(%s, %s)" % (repr(self.__toklist), repr(self.__tokdict))
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return '[' + ', '.join(str(i) if isinstance(i, ParseResults) else repr(i) for i in self.__toklist) + ']'
 * 
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_56__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__ *__pyx_cur_scope;
  PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_7__str___2generator7 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__ *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 557, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);

  /* "Aspidites/_vendor/pyparsing_extension.py":558
 * 
 *     def __str__(self):
 *         return '[' + ', '.join(str(i) if isinstance(i, ParseResults) else repr(i) for i in self.__toklist) + ']'             # <<<<<<<<<<<<<<
 * 
 *     def _asStringList(self, sep=''):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_7__str___genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 558, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyString_Join(__pyx_kp_s__15, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 558, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Add(__pyx_kp_s__14, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 558, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_Add(__pyx_t_1, __pyx_kp_s__16); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 558, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":557
 *         return "(%s, %s)" % (repr(self.__toklist), repr(self.__tokdict))
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return '[' + ', '.join(str(i) if isinstance(i, ParseResults) else repr(i) for i in self.__toklist) + ']'
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_7__str___2generator7);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":560
 *         return '[' + ', '.join(str(i) if isinstance(i, ParseResults) else repr(i) for i in self.__toklist) + ']'
 * 
 *     def _asStringList(self, sep=''):             # <<<<<<<<<<<<<<
 *         out = []
 *         for item in self.__toklist:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_59_asStringList(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_59_asStringList = {"_asStringList", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_59_asStringList, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_59_asStringList(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_sep = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_asStringList (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_sep,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject*)__pyx_kp_s__12));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sep);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_asStringList") < 0)) __PYX_ERR(0, 560, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_sep = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_asStringList", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 560, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults._asStringList", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_58_asStringList(__pyx_self, __pyx_v_self, __pyx_v_sep);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_58_asStringList(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_sep) {
  PyObject *__pyx_v_out = NULL;
  PyObject *__pyx_v_item = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_asStringList", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":561
 * 
 *     def _asStringList(self, sep=''):
 *         out = []             # <<<<<<<<<<<<<<
 *         for item in self.__toklist:
 *             if out and sep:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_out = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":562
 *     def _asStringList(self, sep=''):
 *         out = []
 *         for item in self.__toklist:             # <<<<<<<<<<<<<<
 *             if out and sep:
 *                 out.append(sep)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__toklist); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 562, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 562, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 562, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 562, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 562, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 562, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 562, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 562, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":563
 *         out = []
 *         for item in self.__toklist:
 *             if out and sep:             # <<<<<<<<<<<<<<
 *                 out.append(sep)
 *             if isinstance(item, ParseResults):
 */
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_out); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 563, __pyx_L1_error)
    if (__pyx_t_6) {
    } else {
      __pyx_t_5 = __pyx_t_6;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_sep); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 563, __pyx_L1_error)
    __pyx_t_5 = __pyx_t_6;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_5) {

      /* "Aspidites/_vendor/pyparsing_extension.py":564
 *         for item in self.__toklist:
 *             if out and sep:
 *                 out.append(sep)             # <<<<<<<<<<<<<<
 *             if isinstance(item, ParseResults):
 *                 out += item._asStringList()
 */
      __pyx_t_7 = __Pyx_PyObject_Append(__pyx_v_out, __pyx_v_sep); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 564, __pyx_L1_error)

      /* "Aspidites/_vendor/pyparsing_extension.py":563
 *         out = []
 *         for item in self.__toklist:
 *             if out and sep:             # <<<<<<<<<<<<<<
 *                 out.append(sep)
 *             if isinstance(item, ParseResults):
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":565
 *             if out and sep:
 *                 out.append(sep)
 *             if isinstance(item, ParseResults):             # <<<<<<<<<<<<<<
 *                 out += item._asStringList()
 *             else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParseResults); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 565, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = PyObject_IsInstance(__pyx_v_item, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 565, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = (__pyx_t_5 != 0);
    if (__pyx_t_6) {

      /* "Aspidites/_vendor/pyparsing_extension.py":566
 *                 out.append(sep)
 *             if isinstance(item, ParseResults):
 *                 out += item._asStringList()             # <<<<<<<<<<<<<<
 *             else:
 *                 out.append(str(item))
 */
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_asStringList); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 566, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
        }
      }
      __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_9) : __Pyx_PyObject_CallNoArg(__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 566, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = PyNumber_InPlaceAdd(__pyx_v_out, __pyx_t_1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 566, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF_SET(__pyx_v_out, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":565
 *             if out and sep:
 *                 out.append(sep)
 *             if isinstance(item, ParseResults):             # <<<<<<<<<<<<<<
 *                 out += item._asStringList()
 *             else:
 */
      goto __pyx_L8;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":568
 *                 out += item._asStringList()
 *             else:
 *                 out.append(str(item))             # <<<<<<<<<<<<<<
 *         return out
 * 
 */
    /*else*/ {
      __pyx_t_8 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_v_item); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 568, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_7 = __Pyx_PyObject_Append(__pyx_v_out, __pyx_t_8); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 568, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __pyx_L8:;

    /* "Aspidites/_vendor/pyparsing_extension.py":562
 *     def _asStringList(self, sep=''):
 *         out = []
 *         for item in self.__toklist:             # <<<<<<<<<<<<<<
 *             if out and sep:
 *                 out.append(sep)
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":569
 *             else:
 *                 out.append(str(item))
 *         return out             # <<<<<<<<<<<<<<
 * 
 *     def asList(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_out);
  __pyx_r = __pyx_v_out;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":560
 *         return '[' + ', '.join(str(i) if isinstance(i, ParseResults) else repr(i) for i in self.__toklist) + ']'
 * 
 *     def _asStringList(self, sep=''):             # <<<<<<<<<<<<<<
 *         out = []
 *         for item in self.__toklist:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults._asStringList", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_out);
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":571
 *         return out
 * 
 *     def asList(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns the parse results as a nested list of matching tokens, all converted to strings.
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_61asList(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_60asList[] = "\n        Returns the parse results as a nested list of matching tokens, all converted to strings.\n\n        Example::\n\n            patt = OneOrMore(Word(alphas))\n            result = patt.parseString(\"sldkj lsdkj sldkj\")\n            # even though the result prints in string-like form, it is actually a pyparsing ParseResults\n            print(type(result), result) # -> <class 'pyparsing.ParseResults'> ['sldkj', 'lsdkj', 'sldkj']\n\n            # Use asList() to create an actual list\n            result_list = result.asList()\n            print(type(result_list), result_list) # -> <class 'list'> ['sldkj', 'lsdkj', 'sldkj']\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_61asList = {"asList", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_61asList, METH_O, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_60asList};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_61asList(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("asList (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_60asList(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_60asList(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_res = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("asList", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":586
 *             print(type(result_list), result_list) # -> <class 'list'> ['sldkj', 'lsdkj', 'sldkj']
 *         """
 *         return [res.asList() if isinstance(res, ParseResults) else res for res in self.__toklist]             # <<<<<<<<<<<<<<
 * 
 *     def asDict(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__toklist); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 586, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 586, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 586, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 586, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 586, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 586, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_5(__pyx_t_3);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 586, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_res, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_ParseResults); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 586, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = PyObject_IsInstance(__pyx_v_res, __pyx_t_6); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 586, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if ((__pyx_t_7 != 0)) {
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_res, __pyx_n_s_asList); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 586, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
        }
      }
      __pyx_t_6 = (__pyx_t_9) ? __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_9) : __Pyx_PyObject_CallNoArg(__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 586, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_2 = __pyx_t_6;
      __pyx_t_6 = 0;
    } else {
      __Pyx_INCREF(__pyx_v_res);
      __pyx_t_2 = __pyx_v_res;
    }
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 586, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":571
 *         return out
 * 
 *     def asList(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns the parse results as a nested list of matching tokens, all converted to strings.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.asList", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_res);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":588
 *         return [res.asList() if isinstance(res, ParseResults) else res for res in self.__toklist]
 * 
 *     def asDict(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns the named parse results as a nested dictionary.
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_63asDict(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_62asDict[] = "\n        Returns the named parse results as a nested dictionary.\n\n        Example::\n\n            integer = Word(nums)\n            date_str = integer(\"year\") + '/' + integer(\"month\") + '/' + integer(\"day\")\n\n            result = date_str.parseString('12/31/1999')\n            print(type(result), repr(result)) # -> <class 'pyparsing.ParseResults'> (['12', '/', '31', '/', '1999'], {'day': [('1999', 4)], 'year': [('12', 0)], 'month': [('31', 2)]})\n\n            result_dict = result.asDict()\n            print(type(result_dict), repr(result_dict)) # -> <class 'dict'> {'day': '1999', 'year': '12', 'month': '31'}\n\n            # even though a ParseResults supports dict-like access, sometime you just need to have a dict\n            import json\n            print(json.dumps(result)) # -> Exception: TypeError: ... is not JSON serializable\n            print(json.dumps(result.asDict())) # -> {\"month\": \"31\", \"day\": \"1999\", \"year\": \"12\"}\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_63asDict = {"asDict", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_63asDict, METH_O, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_62asDict};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_63asDict(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("asDict (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_62asDict(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":613
 *             item_fn = self.iteritems
 * 
 *         def toItem(obj):             # <<<<<<<<<<<<<<
 *             if isinstance(obj, ParseResults):
 *                 if obj.haskeys():
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_6asDict_1toItem(PyObject *__pyx_self, PyObject *__pyx_v_obj); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_6asDict_1toItem = {"toItem", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_6asDict_1toItem, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_6asDict_1toItem(PyObject *__pyx_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("toItem (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_6asDict_toItem(__pyx_self, ((PyObject *)__pyx_v_obj));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_6asDict_toItem(PyObject *__pyx_self, PyObject *__pyx_v_obj) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict *__pyx_cur_scope;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict *__pyx_outer_scope;
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("toItem", 0);
  __pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "Aspidites/_vendor/pyparsing_extension.py":614
 * 
 *         def toItem(obj):
 *             if isinstance(obj, ParseResults):             # <<<<<<<<<<<<<<
 *                 if obj.haskeys():
 *                     return obj.asDict()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParseResults); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 614, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_IsInstance(__pyx_v_obj, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 614, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/pyparsing_extension.py":615
 *         def toItem(obj):
 *             if isinstance(obj, ParseResults):
 *                 if obj.haskeys():             # <<<<<<<<<<<<<<
 *                     return obj.asDict()
 *                 else:
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_haskeys); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 615, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 615, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 615, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_3) {

      /* "Aspidites/_vendor/pyparsing_extension.py":616
 *             if isinstance(obj, ParseResults):
 *                 if obj.haskeys():
 *                     return obj.asDict()             # <<<<<<<<<<<<<<
 *                 else:
 *                     return [toItem(v) for v in obj]
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_asDict); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 616, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 616, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_1;
      __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "Aspidites/_vendor/pyparsing_extension.py":615
 *         def toItem(obj):
 *             if isinstance(obj, ParseResults):
 *                 if obj.haskeys():             # <<<<<<<<<<<<<<
 *                     return obj.asDict()
 *                 else:
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":618
 *                     return obj.asDict()
 *                 else:
 *                     return [toItem(v) for v in obj]             # <<<<<<<<<<<<<<
 *             else:
 *                 return obj
 */
    /*else*/ {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 618, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (likely(PyList_CheckExact(__pyx_v_obj)) || PyTuple_CheckExact(__pyx_v_obj)) {
        __pyx_t_4 = __pyx_v_obj; __Pyx_INCREF(__pyx_t_4); __pyx_t_6 = 0;
        __pyx_t_7 = NULL;
      } else {
        __pyx_t_6 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_obj); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 618, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_7 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 618, __pyx_L1_error)
      }
      for (;;) {
        if (likely(!__pyx_t_7)) {
          if (likely(PyList_CheckExact(__pyx_t_4))) {
            if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_4)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_5 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_5); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 618, __pyx_L1_error)
            #else
            __pyx_t_5 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 618, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            #endif
          } else {
            if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_5); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 618, __pyx_L1_error)
            #else
            __pyx_t_5 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 618, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            #endif
          }
        } else {
          __pyx_t_5 = __pyx_t_7(__pyx_t_4);
          if (unlikely(!__pyx_t_5)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 618, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_5);
        }
        __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_5);
        __pyx_t_5 = 0;
        if (unlikely(!__pyx_cur_scope->__pyx_v_toItem)) { __Pyx_RaiseClosureNameError("toItem"); __PYX_ERR(0, 618, __pyx_L1_error) }
        __pyx_t_5 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_6asDict_toItem(__pyx_cur_scope->__pyx_v_toItem, __pyx_v_v); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 618, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 618, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_1;
      __pyx_t_1 = 0;
      goto __pyx_L0;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":614
 * 
 *         def toItem(obj):
 *             if isinstance(obj, ParseResults):             # <<<<<<<<<<<<<<
 *                 if obj.haskeys():
 *                     return obj.asDict()
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":620
 *                     return [toItem(v) for v in obj]
 *             else:
 *                 return obj             # <<<<<<<<<<<<<<
 * 
 *         return dict((k, toItem(v)) for k, v in item_fn())
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_obj);
    __pyx_r = __pyx_v_obj;
    goto __pyx_L0;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":613
 *             item_fn = self.iteritems
 * 
 *         def toItem(obj):             # <<<<<<<<<<<<<<
 *             if isinstance(obj, ParseResults):
 *                 if obj.haskeys():
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.asDict.toItem", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_6asDict_4generator8(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/pyparsing_extension.py":622
 *                 return obj
 * 
 *         return dict((k, toItem(v)) for k, v in item_fn())             # <<<<<<<<<<<<<<
 * 
 *     def copy(self):
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_6asDict_2genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 622, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_6asDict_4generator8, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_ParseResults_asDict_locals_genex, __pyx_n_s_Aspidites__vendor_pyparsing_exte); if (unlikely(!gen)) __PYX_ERR(0, 622, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.asDict.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_6asDict_4generator8(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 622, __pyx_L1_error)
  __pyx_r = PyDict_New(); if (unlikely(!__pyx_r)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_item_fn)) { __Pyx_RaiseClosureNameError("item_fn"); __PYX_ERR(0, 622, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_item_fn);
  __pyx_t_2 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_item_fn; __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 622, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 622, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 622, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 622, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 622, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 622, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_5(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 622, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
      PyObject* sequence = __pyx_t_1;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 622, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_6);
      #else
      __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 622, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 622, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_7 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 622, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_8 = Py_TYPE(__pyx_t_7)->tp_iternext;
      index = 0; __pyx_t_3 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_3)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_3);
      index = 1; __pyx_t_6 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 2) < 0) __PYX_ERR(0, 622, __pyx_L1_error)
      __pyx_t_8 = NULL;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L7_unpacking_done;
      __pyx_L6_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 622, __pyx_L1_error)
      __pyx_L7_unpacking_done:;
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_k);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_k, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_v);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_v, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    __pyx_t_6 = 0;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_toItem)) { __Pyx_RaiseClosureNameError("toItem"); __PYX_ERR(0, 622, __pyx_L1_error) }
    __pyx_t_1 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_6asDict_toItem(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_toItem, __pyx_cur_scope->__pyx_v_v); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 622, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(PyDict_SetItem(__pyx_r, (PyObject*)__pyx_cur_scope->__pyx_v_k, (PyObject*)__pyx_t_1))) __PYX_ERR(0, 622, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":588
 *         return [res.asList() if isinstance(res, ParseResults) else res for res in self.__toklist]
 * 
 *     def asDict(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns the named parse results as a nested dictionary.
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_62asDict(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict *__pyx_cur_scope;
  PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_6asDict_4generator8 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("asDict", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 588, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":608
 *             print(json.dumps(result.asDict())) # -> {"month": "31", "day": "1999", "year": "12"}
 *         """
 *         if PY_3:             # <<<<<<<<<<<<<<
 *             item_fn = self.items
 *         else:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_PY_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 608, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 608, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":609
 *         """
 *         if PY_3:
 *             item_fn = self.items             # <<<<<<<<<<<<<<
 *         else:
 *             item_fn = self.iteritems
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_items); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 609, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_v_item_fn = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":608
 *             print(json.dumps(result.asDict())) # -> {"month": "31", "day": "1999", "year": "12"}
 *         """
 *         if PY_3:             # <<<<<<<<<<<<<<
 *             item_fn = self.items
 *         else:
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":611
 *             item_fn = self.items
 *         else:
 *             item_fn = self.iteritems             # <<<<<<<<<<<<<<
 * 
 *         def toItem(obj):
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_iteritems); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 611, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_v_item_fn = __pyx_t_1;
    __pyx_t_1 = 0;
  }
  __pyx_L3:;

  /* "Aspidites/_vendor/pyparsing_extension.py":613
 *             item_fn = self.iteritems
 * 
 *         def toItem(obj):             # <<<<<<<<<<<<<<
 *             if isinstance(obj, ParseResults):
 *                 if obj.haskeys():
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_6asDict_1toItem, 0, __pyx_n_s_ParseResults_asDict_locals_toIte, ((PyObject*)__pyx_cur_scope), __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__18)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_toItem = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":622
 *                 return obj
 * 
 *         return dict((k, toItem(v)) for k, v in item_fn())             # <<<<<<<<<<<<<<
 * 
 *     def copy(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_6asDict_2genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_Generator_Next(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":588
 *         return [res.asList() if isinstance(res, ParseResults) else res for res in self.__toklist]
 * 
 *     def asDict(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns the named parse results as a nested dictionary.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.asDict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_6asDict_4generator8);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":624
 *         return dict((k, toItem(v)) for k, v in item_fn())
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns a new copy of a :class:`ParseResults` object.
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_65copy(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_64copy[] = "\n        Returns a new copy of a :class:`ParseResults` object.\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_65copy = {"copy", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_65copy, METH_O, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_64copy};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_65copy(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_64copy(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_64copy(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_ret = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":628
 *         Returns a new copy of a :class:`ParseResults` object.
 *         """
 *         ret = ParseResults(self.__toklist)             # <<<<<<<<<<<<<<
 *         ret.__tokdict = dict(self.__tokdict.items())
 *         ret.__parent = self.__parent
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParseResults); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__toklist); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_ret = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":629
 *         """
 *         ret = ParseResults(self.__toklist)
 *         ret.__tokdict = dict(self.__tokdict.items())             # <<<<<<<<<<<<<<
 *         ret.__parent = self.__parent
 *         ret.__accumNames.update(self.__accumNames)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__tokdict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_items); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_ret, __pyx_n_s_ParseResults__tokdict, __pyx_t_3) < 0) __PYX_ERR(0, 629, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":630
 *         ret = ParseResults(self.__toklist)
 *         ret.__tokdict = dict(self.__tokdict.items())
 *         ret.__parent = self.__parent             # <<<<<<<<<<<<<<
 *         ret.__accumNames.update(self.__accumNames)
 *         ret.__name = self.__name
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__parent); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_ret, __pyx_n_s_ParseResults__parent, __pyx_t_3) < 0) __PYX_ERR(0, 630, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":631
 *         ret.__tokdict = dict(self.__tokdict.items())
 *         ret.__parent = self.__parent
 *         ret.__accumNames.update(self.__accumNames)             # <<<<<<<<<<<<<<
 *         ret.__name = self.__name
 *         return ret
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_ret, __pyx_n_s_ParseResults__accumNames); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_update); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__accumNames); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":632
 *         ret.__parent = self.__parent
 *         ret.__accumNames.update(self.__accumNames)
 *         ret.__name = self.__name             # <<<<<<<<<<<<<<
 *         return ret
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_ret, __pyx_n_s_ParseResults__name, __pyx_t_3) < 0) __PYX_ERR(0, 632, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":633
 *         ret.__accumNames.update(self.__accumNames)
 *         ret.__name = self.__name
 *         return ret             # <<<<<<<<<<<<<<
 * 
 *     def asXML(self, doctag=None, namedItemsOnly=False, indent="", formatted=True):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_ret);
  __pyx_r = __pyx_v_ret;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":624
 *         return dict((k, toItem(v)) for k, v in item_fn())
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns a new copy of a :class:`ParseResults` object.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":635
 *         return ret
 * 
 *     def asXML(self, doctag=None, namedItemsOnly=False, indent="", formatted=True):             # <<<<<<<<<<<<<<
 *         """
 *         (Deprecated) Returns the parse results as XML. Tags are created for tokens and lists that have defined results names.
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_67asXML(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_66asXML[] = "\n        (Deprecated) Returns the parse results as XML. Tags are created for tokens and lists that have defined results names.\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_67asXML = {"asXML", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_67asXML, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_66asXML};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_67asXML(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_doctag = 0;
  PyObject *__pyx_v_namedItemsOnly = 0;
  PyObject *__pyx_v_indent = 0;
  PyObject *__pyx_v_formatted = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("asXML (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_doctag,&__pyx_n_s_namedItemsOnly,&__pyx_n_s_indent,&__pyx_n_s_formatted,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    values[2] = ((PyObject *)((PyObject *)Py_False));
    values[3] = ((PyObject *)((PyObject*)__pyx_kp_s__12));
    values[4] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_doctag);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_namedItemsOnly);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_indent);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_formatted);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "asXML") < 0)) __PYX_ERR(0, 635, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_doctag = values[1];
    __pyx_v_namedItemsOnly = values[2];
    __pyx_v_indent = values[3];
    __pyx_v_formatted = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("asXML", 0, 1, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 635, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.asXML", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_66asXML(__pyx_self, __pyx_v_self, __pyx_v_doctag, __pyx_v_namedItemsOnly, __pyx_v_indent, __pyx_v_formatted);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_5asXML_2generator9(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/pyparsing_extension.py":641
 *         nl = "\n"
 *         out = []
 *         namedItems = dict((v[1], k) for (k, vlist) in self.__tokdict.items()             # <<<<<<<<<<<<<<
 *                           for v in vlist)
 *         nextLevelIndent = indent + "  "
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_5asXML_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 641, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_5asXML_2generator9, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_ParseResults_asXML_locals_genexp, __pyx_n_s_Aspidites__vendor_pyparsing_exte); if (unlikely(!gen)) __PYX_ERR(0, 641, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.asXML.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_5asXML_2generator9(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  Py_ssize_t __pyx_t_9;
  PyObject *(*__pyx_t_10)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 641, __pyx_L1_error)
  __pyx_r = PyDict_New(); if (unlikely(!__pyx_r)) __PYX_ERR(0, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 641, __pyx_L1_error) }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_ParseResults__tokdict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_items); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_3 = __pyx_t_1; __Pyx_INCREF(__pyx_t_3); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 641, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 641, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 641, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 641, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 641, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 641, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_5(__pyx_t_3);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 641, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
      PyObject* sequence = __pyx_t_1;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 641, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_6);
      #else
      __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 641, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 641, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_7 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 641, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_8 = Py_TYPE(__pyx_t_7)->tp_iternext;
      index = 0; __pyx_t_2 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_2)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_2);
      index = 1; __pyx_t_6 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 2) < 0) __PYX_ERR(0, 641, __pyx_L1_error)
      __pyx_t_8 = NULL;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L7_unpacking_done;
      __pyx_L6_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 641, __pyx_L1_error)
      __pyx_L7_unpacking_done:;
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_k);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_k, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_vlist);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_vlist, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    __pyx_t_6 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":642
 *         out = []
 *         namedItems = dict((v[1], k) for (k, vlist) in self.__tokdict.items()
 *                           for v in vlist)             # <<<<<<<<<<<<<<
 *         nextLevelIndent = indent + "  "
 * 
 */
    if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_v_vlist)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_v_vlist)) {
      __pyx_t_1 = __pyx_cur_scope->__pyx_v_vlist; __Pyx_INCREF(__pyx_t_1); __pyx_t_9 = 0;
      __pyx_t_10 = NULL;
    } else {
      __pyx_t_9 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_v_vlist); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 642, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_10 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 642, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_10)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_9); __Pyx_INCREF(__pyx_t_6); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 642, __pyx_L1_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_1, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 642, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        } else {
          if (__pyx_t_9 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_9); __Pyx_INCREF(__pyx_t_6); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 642, __pyx_L1_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_1, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 642, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        }
      } else {
        __pyx_t_6 = __pyx_t_10(__pyx_t_1);
        if (unlikely(!__pyx_t_6)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 642, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_6);
      }
      __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_v);
      __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_v, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_6);
      __pyx_t_6 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":641
 *         nl = "\n"
 *         out = []
 *         namedItems = dict((v[1], k) for (k, vlist) in self.__tokdict.items()             # <<<<<<<<<<<<<<
 *                           for v in vlist)
 *         nextLevelIndent = indent + "  "
 */
      __pyx_t_6 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_v, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 641, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (unlikely(PyDict_SetItem(__pyx_r, (PyObject*)__pyx_t_6, (PyObject*)__pyx_cur_scope->__pyx_v_k))) __PYX_ERR(0, 641, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":642
 *         out = []
 *         namedItems = dict((v[1], k) for (k, vlist) in self.__tokdict.items()
 *                           for v in vlist)             # <<<<<<<<<<<<<<
 *         nextLevelIndent = indent + "  "
 * 
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":641
 *         nl = "\n"
 *         out = []
 *         namedItems = dict((v[1], k) for (k, vlist) in self.__tokdict.items()             # <<<<<<<<<<<<<<
 *                           for v in vlist)
 *         nextLevelIndent = indent + "  "
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":635
 *         return ret
 * 
 *     def asXML(self, doctag=None, namedItemsOnly=False, indent="", formatted=True):             # <<<<<<<<<<<<<<
 *         """
 *         (Deprecated) Returns the parse results as XML. Tags are created for tokens and lists that have defined results names.
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_66asXML(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_doctag, PyObject *__pyx_v_namedItemsOnly, PyObject *__pyx_v_indent, PyObject *__pyx_v_formatted) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML *__pyx_cur_scope;
  PyObject *__pyx_v_nl = NULL;
  PyObject *__pyx_v_out = NULL;
  PyObject *__pyx_v_namedItems = NULL;
  PyObject *__pyx_v_nextLevelIndent = NULL;
  PyObject *__pyx_v_selfTag = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_res = NULL;
  PyObject *__pyx_v_resTag = NULL;
  PyObject *__pyx_v_xmlBodyText = NULL;
  PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_5asXML_2generator9 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("asXML", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 635, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_INCREF(__pyx_v_indent);

  /* "Aspidites/_vendor/pyparsing_extension.py":639
 *         (Deprecated) Returns the parse results as XML. Tags are created for tokens and lists that have defined results names.
 *         """
 *         nl = "\n"             # <<<<<<<<<<<<<<
 *         out = []
 *         namedItems = dict((v[1], k) for (k, vlist) in self.__tokdict.items()
 */
  __Pyx_INCREF(__pyx_kp_s__4);
  __pyx_v_nl = __pyx_kp_s__4;

  /* "Aspidites/_vendor/pyparsing_extension.py":640
 *         """
 *         nl = "\n"
 *         out = []             # <<<<<<<<<<<<<<
 *         namedItems = dict((v[1], k) for (k, vlist) in self.__tokdict.items()
 *                           for v in vlist)
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 640, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_out = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":641
 *         nl = "\n"
 *         out = []
 *         namedItems = dict((v[1], k) for (k, vlist) in self.__tokdict.items()             # <<<<<<<<<<<<<<
 *                           for v in vlist)
 *         nextLevelIndent = indent + "  "
 */
  __pyx_t_1 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_5asXML_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_Generator_Next(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_namedItems = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":643
 *         namedItems = dict((v[1], k) for (k, vlist) in self.__tokdict.items()
 *                           for v in vlist)
 *         nextLevelIndent = indent + "  "             # <<<<<<<<<<<<<<
 * 
 *         # collapse out indents if formatting is not desired
 */
  __pyx_t_2 = PyNumber_Add(__pyx_v_indent, __pyx_kp_s__19); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_nextLevelIndent = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":646
 * 
 *         # collapse out indents if formatting is not desired
 *         if not formatted:             # <<<<<<<<<<<<<<
 *             indent = ""
 *             nextLevelIndent = ""
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_formatted); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 646, __pyx_L1_error)
  __pyx_t_4 = ((!__pyx_t_3) != 0);
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":647
 *         # collapse out indents if formatting is not desired
 *         if not formatted:
 *             indent = ""             # <<<<<<<<<<<<<<
 *             nextLevelIndent = ""
 *             nl = ""
 */
    __Pyx_INCREF(__pyx_kp_s__12);
    __Pyx_DECREF_SET(__pyx_v_indent, __pyx_kp_s__12);

    /* "Aspidites/_vendor/pyparsing_extension.py":648
 *         if not formatted:
 *             indent = ""
 *             nextLevelIndent = ""             # <<<<<<<<<<<<<<
 *             nl = ""
 * 
 */
    __Pyx_INCREF(__pyx_kp_s__12);
    __Pyx_DECREF_SET(__pyx_v_nextLevelIndent, __pyx_kp_s__12);

    /* "Aspidites/_vendor/pyparsing_extension.py":649
 *             indent = ""
 *             nextLevelIndent = ""
 *             nl = ""             # <<<<<<<<<<<<<<
 * 
 *         selfTag = None
 */
    __Pyx_INCREF(__pyx_kp_s__12);
    __Pyx_DECREF_SET(__pyx_v_nl, __pyx_kp_s__12);

    /* "Aspidites/_vendor/pyparsing_extension.py":646
 * 
 *         # collapse out indents if formatting is not desired
 *         if not formatted:             # <<<<<<<<<<<<<<
 *             indent = ""
 *             nextLevelIndent = ""
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":651
 *             nl = ""
 * 
 *         selfTag = None             # <<<<<<<<<<<<<<
 *         if doctag is not None:
 *             selfTag = doctag
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_selfTag = Py_None;

  /* "Aspidites/_vendor/pyparsing_extension.py":652
 * 
 *         selfTag = None
 *         if doctag is not None:             # <<<<<<<<<<<<<<
 *             selfTag = doctag
 *         else:
 */
  __pyx_t_4 = (__pyx_v_doctag != Py_None);
  __pyx_t_3 = (__pyx_t_4 != 0);
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/pyparsing_extension.py":653
 *         selfTag = None
 *         if doctag is not None:
 *             selfTag = doctag             # <<<<<<<<<<<<<<
 *         else:
 *             if self.__name:
 */
    __Pyx_INCREF(__pyx_v_doctag);
    __Pyx_DECREF_SET(__pyx_v_selfTag, __pyx_v_doctag);

    /* "Aspidites/_vendor/pyparsing_extension.py":652
 * 
 *         selfTag = None
 *         if doctag is not None:             # <<<<<<<<<<<<<<
 *             selfTag = doctag
 *         else:
 */
    goto __pyx_L4;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":655
 *             selfTag = doctag
 *         else:
 *             if self.__name:             # <<<<<<<<<<<<<<
 *                 selfTag = self.__name
 * 
 */
  /*else*/ {
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_ParseResults__name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 655, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 655, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_3) {

      /* "Aspidites/_vendor/pyparsing_extension.py":656
 *         else:
 *             if self.__name:
 *                 selfTag = self.__name             # <<<<<<<<<<<<<<
 * 
 *         if not selfTag:
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_ParseResults__name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 656, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF_SET(__pyx_v_selfTag, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":655
 *             selfTag = doctag
 *         else:
 *             if self.__name:             # <<<<<<<<<<<<<<
 *                 selfTag = self.__name
 * 
 */
    }
  }
  __pyx_L4:;

  /* "Aspidites/_vendor/pyparsing_extension.py":658
 *                 selfTag = self.__name
 * 
 *         if not selfTag:             # <<<<<<<<<<<<<<
 *             if namedItemsOnly:
 *                 return ""
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_selfTag); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 658, __pyx_L1_error)
  __pyx_t_4 = ((!__pyx_t_3) != 0);
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":659
 * 
 *         if not selfTag:
 *             if namedItemsOnly:             # <<<<<<<<<<<<<<
 *                 return ""
 *             else:
 */
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_namedItemsOnly); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 659, __pyx_L1_error)
    if (__pyx_t_4) {

      /* "Aspidites/_vendor/pyparsing_extension.py":660
 *         if not selfTag:
 *             if namedItemsOnly:
 *                 return ""             # <<<<<<<<<<<<<<
 *             else:
 *                 selfTag = "ITEM"
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_kp_s__12);
      __pyx_r = __pyx_kp_s__12;
      goto __pyx_L0;

      /* "Aspidites/_vendor/pyparsing_extension.py":659
 * 
 *         if not selfTag:
 *             if namedItemsOnly:             # <<<<<<<<<<<<<<
 *                 return ""
 *             else:
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":662
 *                 return ""
 *             else:
 *                 selfTag = "ITEM"             # <<<<<<<<<<<<<<
 * 
 *         out += [nl, indent, "<", selfTag, ">"]
 */
    /*else*/ {
      __Pyx_INCREF(__pyx_n_s_ITEM);
      __Pyx_DECREF_SET(__pyx_v_selfTag, __pyx_n_s_ITEM);
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":658
 *                 selfTag = self.__name
 * 
 *         if not selfTag:             # <<<<<<<<<<<<<<
 *             if namedItemsOnly:
 *                 return ""
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":664
 *                 selfTag = "ITEM"
 * 
 *         out += [nl, indent, "<", selfTag, ">"]             # <<<<<<<<<<<<<<
 * 
 *         for i, res in enumerate(self.__toklist):
 */
  __pyx_t_2 = PyList_New(5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 664, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_nl);
  __Pyx_GIVEREF(__pyx_v_nl);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_v_nl);
  __Pyx_INCREF(__pyx_v_indent);
  __Pyx_GIVEREF(__pyx_v_indent);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_v_indent);
  __Pyx_INCREF(__pyx_kp_s__20);
  __Pyx_GIVEREF(__pyx_kp_s__20);
  PyList_SET_ITEM(__pyx_t_2, 2, __pyx_kp_s__20);
  __Pyx_INCREF(__pyx_v_selfTag);
  __Pyx_GIVEREF(__pyx_v_selfTag);
  PyList_SET_ITEM(__pyx_t_2, 3, __pyx_v_selfTag);
  __Pyx_INCREF(__pyx_kp_s__21);
  __Pyx_GIVEREF(__pyx_kp_s__21);
  PyList_SET_ITEM(__pyx_t_2, 4, __pyx_kp_s__21);
  __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_out, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 664, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_out, ((PyObject*)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":666
 *         out += [nl, indent, "<", selfTag, ">"]
 * 
 *         for i, res in enumerate(self.__toklist):             # <<<<<<<<<<<<<<
 *             if isinstance(res, ParseResults):
 *                 if i in namedItems:
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_t_1 = __pyx_int_0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_ParseResults__toklist); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_5 = __pyx_t_2; __Pyx_INCREF(__pyx_t_5); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 666, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 666, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_5))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 666, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_5, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 666, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 666, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_5, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 666, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_7(__pyx_t_5);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 666, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_res, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_1);
    __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 666, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1);
    __pyx_t_1 = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":667
 * 
 *         for i, res in enumerate(self.__toklist):
 *             if isinstance(res, ParseResults):             # <<<<<<<<<<<<<<
 *                 if i in namedItems:
 *                     out += [res.asXML(namedItems[i],
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParseResults); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 667, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyObject_IsInstance(__pyx_v_res, __pyx_t_2); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 667, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_3 = (__pyx_t_4 != 0);
    if (__pyx_t_3) {

      /* "Aspidites/_vendor/pyparsing_extension.py":668
 *         for i, res in enumerate(self.__toklist):
 *             if isinstance(res, ParseResults):
 *                 if i in namedItems:             # <<<<<<<<<<<<<<
 *                     out += [res.asXML(namedItems[i],
 *                                       namedItemsOnly and doctag is None,
 */
      if (unlikely(__pyx_v_namedItems == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
        __PYX_ERR(0, 668, __pyx_L1_error)
      }
      __pyx_t_3 = (__Pyx_PyDict_ContainsTF(__pyx_v_i, __pyx_v_namedItems, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 668, __pyx_L1_error)
      __pyx_t_4 = (__pyx_t_3 != 0);
      if (__pyx_t_4) {

        /* "Aspidites/_vendor/pyparsing_extension.py":669
 *             if isinstance(res, ParseResults):
 *                 if i in namedItems:
 *                     out += [res.asXML(namedItems[i],             # <<<<<<<<<<<<<<
 *                                       namedItemsOnly and doctag is None,
 *                                       nextLevelIndent,
 */
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_res, __pyx_n_s_asXML); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 669, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        if (unlikely(__pyx_v_namedItems == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 669, __pyx_L1_error)
        }
        __pyx_t_9 = __Pyx_PyDict_GetItem(__pyx_v_namedItems, __pyx_v_i); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 669, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);

        /* "Aspidites/_vendor/pyparsing_extension.py":670
 *                 if i in namedItems:
 *                     out += [res.asXML(namedItems[i],
 *                                       namedItemsOnly and doctag is None,             # <<<<<<<<<<<<<<
 *                                       nextLevelIndent,
 *                                       formatted)]
 */
        __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_namedItemsOnly); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 670, __pyx_L1_error)
        if (__pyx_t_4) {
        } else {
          __Pyx_INCREF(__pyx_v_namedItemsOnly);
          __pyx_t_10 = __pyx_v_namedItemsOnly;
          goto __pyx_L12_bool_binop_done;
        }
        __pyx_t_4 = (__pyx_v_doctag == Py_None);
        __pyx_t_11 = __Pyx_PyBool_FromLong(__pyx_t_4); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 670, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_10 = __pyx_t_11;
        __pyx_t_11 = 0;
        __pyx_L12_bool_binop_done:;

        /* "Aspidites/_vendor/pyparsing_extension.py":672
 *                                       namedItemsOnly and doctag is None,
 *                                       nextLevelIndent,
 *                                       formatted)]             # <<<<<<<<<<<<<<
 *                 else:
 *                     out += [res.asXML(None,
 */
        __pyx_t_11 = NULL;
        __pyx_t_12 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
          __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_8);
          if (likely(__pyx_t_11)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
            __Pyx_INCREF(__pyx_t_11);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_8, function);
            __pyx_t_12 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_8)) {
          PyObject *__pyx_temp[5] = {__pyx_t_11, __pyx_t_9, __pyx_t_10, __pyx_v_nextLevelIndent, __pyx_v_formatted};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_12, 4+__pyx_t_12); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 669, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
          PyObject *__pyx_temp[5] = {__pyx_t_11, __pyx_t_9, __pyx_t_10, __pyx_v_nextLevelIndent, __pyx_v_formatted};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_12, 4+__pyx_t_12); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 669, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        } else
        #endif
        {
          __pyx_t_13 = PyTuple_New(4+__pyx_t_12); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 669, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_13);
          if (__pyx_t_11) {
            __Pyx_GIVEREF(__pyx_t_11); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_11); __pyx_t_11 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_9);
          PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_12, __pyx_t_9);
          __Pyx_GIVEREF(__pyx_t_10);
          PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_12, __pyx_t_10);
          __Pyx_INCREF(__pyx_v_nextLevelIndent);
          __Pyx_GIVEREF(__pyx_v_nextLevelIndent);
          PyTuple_SET_ITEM(__pyx_t_13, 2+__pyx_t_12, __pyx_v_nextLevelIndent);
          __Pyx_INCREF(__pyx_v_formatted);
          __Pyx_GIVEREF(__pyx_v_formatted);
          PyTuple_SET_ITEM(__pyx_t_13, 3+__pyx_t_12, __pyx_v_formatted);
          __pyx_t_9 = 0;
          __pyx_t_10 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_13, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 669, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        }
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":669
 *             if isinstance(res, ParseResults):
 *                 if i in namedItems:
 *                     out += [res.asXML(namedItems[i],             # <<<<<<<<<<<<<<
 *                                       namedItemsOnly and doctag is None,
 *                                       nextLevelIndent,
 */
        __pyx_t_8 = PyList_New(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 669, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_GIVEREF(__pyx_t_2);
        PyList_SET_ITEM(__pyx_t_8, 0, __pyx_t_2);
        __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_out, __pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 669, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF_SET(__pyx_v_out, ((PyObject*)__pyx_t_2));
        __pyx_t_2 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":668
 *         for i, res in enumerate(self.__toklist):
 *             if isinstance(res, ParseResults):
 *                 if i in namedItems:             # <<<<<<<<<<<<<<
 *                     out += [res.asXML(namedItems[i],
 *                                       namedItemsOnly and doctag is None,
 */
        goto __pyx_L11;
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":674
 *                                       formatted)]
 *                 else:
 *                     out += [res.asXML(None,             # <<<<<<<<<<<<<<
 *                                       namedItemsOnly and doctag is None,
 *                                       nextLevelIndent,
 */
      /*else*/ {
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_res, __pyx_n_s_asXML); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 674, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);

        /* "Aspidites/_vendor/pyparsing_extension.py":675
 *                 else:
 *                     out += [res.asXML(None,
 *                                       namedItemsOnly and doctag is None,             # <<<<<<<<<<<<<<
 *                                       nextLevelIndent,
 *                                       formatted)]
 */
        __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_namedItemsOnly); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 675, __pyx_L1_error)
        if (__pyx_t_4) {
        } else {
          __Pyx_INCREF(__pyx_v_namedItemsOnly);
          __pyx_t_13 = __pyx_v_namedItemsOnly;
          goto __pyx_L14_bool_binop_done;
        }
        __pyx_t_4 = (__pyx_v_doctag == Py_None);
        __pyx_t_10 = __Pyx_PyBool_FromLong(__pyx_t_4); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 675, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_13 = __pyx_t_10;
        __pyx_t_10 = 0;
        __pyx_L14_bool_binop_done:;

        /* "Aspidites/_vendor/pyparsing_extension.py":677
 *                                       namedItemsOnly and doctag is None,
 *                                       nextLevelIndent,
 *                                       formatted)]             # <<<<<<<<<<<<<<
 *             else:
 *                 # individual token, see if there is a name for it
 */
        __pyx_t_10 = NULL;
        __pyx_t_12 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
          __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_8);
          if (likely(__pyx_t_10)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
            __Pyx_INCREF(__pyx_t_10);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_8, function);
            __pyx_t_12 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_8)) {
          PyObject *__pyx_temp[5] = {__pyx_t_10, Py_None, __pyx_t_13, __pyx_v_nextLevelIndent, __pyx_v_formatted};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_12, 4+__pyx_t_12); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 674, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
          PyObject *__pyx_temp[5] = {__pyx_t_10, Py_None, __pyx_t_13, __pyx_v_nextLevelIndent, __pyx_v_formatted};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_12, 4+__pyx_t_12); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 674, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        } else
        #endif
        {
          __pyx_t_9 = PyTuple_New(4+__pyx_t_12); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 674, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          if (__pyx_t_10) {
            __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_10); __pyx_t_10 = NULL;
          }
          __Pyx_INCREF(Py_None);
          __Pyx_GIVEREF(Py_None);
          PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_12, Py_None);
          __Pyx_GIVEREF(__pyx_t_13);
          PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_12, __pyx_t_13);
          __Pyx_INCREF(__pyx_v_nextLevelIndent);
          __Pyx_GIVEREF(__pyx_v_nextLevelIndent);
          PyTuple_SET_ITEM(__pyx_t_9, 2+__pyx_t_12, __pyx_v_nextLevelIndent);
          __Pyx_INCREF(__pyx_v_formatted);
          __Pyx_GIVEREF(__pyx_v_formatted);
          PyTuple_SET_ITEM(__pyx_t_9, 3+__pyx_t_12, __pyx_v_formatted);
          __pyx_t_13 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_9, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 674, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        }
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":674
 *                                       formatted)]
 *                 else:
 *                     out += [res.asXML(None,             # <<<<<<<<<<<<<<
 *                                       namedItemsOnly and doctag is None,
 *                                       nextLevelIndent,
 */
        __pyx_t_8 = PyList_New(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 674, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_GIVEREF(__pyx_t_2);
        PyList_SET_ITEM(__pyx_t_8, 0, __pyx_t_2);
        __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_out, __pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 674, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF_SET(__pyx_v_out, ((PyObject*)__pyx_t_2));
        __pyx_t_2 = 0;
      }
      __pyx_L11:;

      /* "Aspidites/_vendor/pyparsing_extension.py":667
 * 
 *         for i, res in enumerate(self.__toklist):
 *             if isinstance(res, ParseResults):             # <<<<<<<<<<<<<<
 *                 if i in namedItems:
 *                     out += [res.asXML(namedItems[i],
 */
      goto __pyx_L10;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":680
 *             else:
 *                 # individual token, see if there is a name for it
 *                 resTag = None             # <<<<<<<<<<<<<<
 *                 if i in namedItems:
 *                     resTag = namedItems[i]
 */
    /*else*/ {
      __Pyx_INCREF(Py_None);
      __Pyx_XDECREF_SET(__pyx_v_resTag, Py_None);

      /* "Aspidites/_vendor/pyparsing_extension.py":681
 *                 # individual token, see if there is a name for it
 *                 resTag = None
 *                 if i in namedItems:             # <<<<<<<<<<<<<<
 *                     resTag = namedItems[i]
 *                 if not resTag:
 */
      if (unlikely(__pyx_v_namedItems == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
        __PYX_ERR(0, 681, __pyx_L1_error)
      }
      __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_v_i, __pyx_v_namedItems, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 681, __pyx_L1_error)
      __pyx_t_3 = (__pyx_t_4 != 0);
      if (__pyx_t_3) {

        /* "Aspidites/_vendor/pyparsing_extension.py":682
 *                 resTag = None
 *                 if i in namedItems:
 *                     resTag = namedItems[i]             # <<<<<<<<<<<<<<
 *                 if not resTag:
 *                     if namedItemsOnly:
 */
        if (unlikely(__pyx_v_namedItems == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 682, __pyx_L1_error)
        }
        __pyx_t_2 = __Pyx_PyDict_GetItem(__pyx_v_namedItems, __pyx_v_i); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 682, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF_SET(__pyx_v_resTag, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":681
 *                 # individual token, see if there is a name for it
 *                 resTag = None
 *                 if i in namedItems:             # <<<<<<<<<<<<<<
 *                     resTag = namedItems[i]
 *                 if not resTag:
 */
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":683
 *                 if i in namedItems:
 *                     resTag = namedItems[i]
 *                 if not resTag:             # <<<<<<<<<<<<<<
 *                     if namedItemsOnly:
 *                         continue
 */
      __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_resTag); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 683, __pyx_L1_error)
      __pyx_t_4 = ((!__pyx_t_3) != 0);
      if (__pyx_t_4) {

        /* "Aspidites/_vendor/pyparsing_extension.py":684
 *                     resTag = namedItems[i]
 *                 if not resTag:
 *                     if namedItemsOnly:             # <<<<<<<<<<<<<<
 *                         continue
 *                     else:
 */
        __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_namedItemsOnly); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 684, __pyx_L1_error)
        if (__pyx_t_4) {

          /* "Aspidites/_vendor/pyparsing_extension.py":685
 *                 if not resTag:
 *                     if namedItemsOnly:
 *                         continue             # <<<<<<<<<<<<<<
 *                     else:
 *                         resTag = "ITEM"
 */
          goto __pyx_L8_continue;

          /* "Aspidites/_vendor/pyparsing_extension.py":684
 *                     resTag = namedItems[i]
 *                 if not resTag:
 *                     if namedItemsOnly:             # <<<<<<<<<<<<<<
 *                         continue
 *                     else:
 */
        }

        /* "Aspidites/_vendor/pyparsing_extension.py":687
 *                         continue
 *                     else:
 *                         resTag = "ITEM"             # <<<<<<<<<<<<<<
 *                 xmlBodyText = _xml_escape(str(res))
 *                 out += [nl, nextLevelIndent, "<", resTag, ">",
 */
        /*else*/ {
          __Pyx_INCREF(__pyx_n_s_ITEM);
          __Pyx_DECREF_SET(__pyx_v_resTag, __pyx_n_s_ITEM);
        }

        /* "Aspidites/_vendor/pyparsing_extension.py":683
 *                 if i in namedItems:
 *                     resTag = namedItems[i]
 *                 if not resTag:             # <<<<<<<<<<<<<<
 *                     if namedItemsOnly:
 *                         continue
 */
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":688
 *                     else:
 *                         resTag = "ITEM"
 *                 xmlBodyText = _xml_escape(str(res))             # <<<<<<<<<<<<<<
 *                 out += [nl, nextLevelIndent, "<", resTag, ">",
 *                         xmlBodyText,
 */
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_xml_escape); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 688, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_v_res); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 688, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_13 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_13)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_13);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
        }
      }
      __pyx_t_2 = (__pyx_t_13) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_13, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_9);
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 688, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF_SET(__pyx_v_xmlBodyText, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":689
 *                         resTag = "ITEM"
 *                 xmlBodyText = _xml_escape(str(res))
 *                 out += [nl, nextLevelIndent, "<", resTag, ">",             # <<<<<<<<<<<<<<
 *                         xmlBodyText,
 *                         "</", resTag, ">"]
 */
      __pyx_t_2 = PyList_New(9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 689, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_v_nl);
      __Pyx_GIVEREF(__pyx_v_nl);
      PyList_SET_ITEM(__pyx_t_2, 0, __pyx_v_nl);
      __Pyx_INCREF(__pyx_v_nextLevelIndent);
      __Pyx_GIVEREF(__pyx_v_nextLevelIndent);
      PyList_SET_ITEM(__pyx_t_2, 1, __pyx_v_nextLevelIndent);
      __Pyx_INCREF(__pyx_kp_s__20);
      __Pyx_GIVEREF(__pyx_kp_s__20);
      PyList_SET_ITEM(__pyx_t_2, 2, __pyx_kp_s__20);
      __Pyx_INCREF(__pyx_v_resTag);
      __Pyx_GIVEREF(__pyx_v_resTag);
      PyList_SET_ITEM(__pyx_t_2, 3, __pyx_v_resTag);
      __Pyx_INCREF(__pyx_kp_s__21);
      __Pyx_GIVEREF(__pyx_kp_s__21);
      PyList_SET_ITEM(__pyx_t_2, 4, __pyx_kp_s__21);
      __Pyx_INCREF(__pyx_v_xmlBodyText);
      __Pyx_GIVEREF(__pyx_v_xmlBodyText);
      PyList_SET_ITEM(__pyx_t_2, 5, __pyx_v_xmlBodyText);
      __Pyx_INCREF(__pyx_kp_s__22);
      __Pyx_GIVEREF(__pyx_kp_s__22);
      PyList_SET_ITEM(__pyx_t_2, 6, __pyx_kp_s__22);
      __Pyx_INCREF(__pyx_v_resTag);
      __Pyx_GIVEREF(__pyx_v_resTag);
      PyList_SET_ITEM(__pyx_t_2, 7, __pyx_v_resTag);
      __Pyx_INCREF(__pyx_kp_s__21);
      __Pyx_GIVEREF(__pyx_kp_s__21);
      PyList_SET_ITEM(__pyx_t_2, 8, __pyx_kp_s__21);
      __pyx_t_8 = PyNumber_InPlaceAdd(__pyx_v_out, __pyx_t_2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 689, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF_SET(__pyx_v_out, ((PyObject*)__pyx_t_8));
      __pyx_t_8 = 0;
    }
    __pyx_L10:;

    /* "Aspidites/_vendor/pyparsing_extension.py":666
 *         out += [nl, indent, "<", selfTag, ">"]
 * 
 *         for i, res in enumerate(self.__toklist):             # <<<<<<<<<<<<<<
 *             if isinstance(res, ParseResults):
 *                 if i in namedItems:
 */
    __pyx_L8_continue:;
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":693
 *                         "</", resTag, ">"]
 * 
 *         out += [nl, indent, "</", selfTag, ">"]             # <<<<<<<<<<<<<<
 *         return "".join(out)
 * 
 */
  __pyx_t_1 = PyList_New(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 693, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_nl);
  __Pyx_GIVEREF(__pyx_v_nl);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_v_nl);
  __Pyx_INCREF(__pyx_v_indent);
  __Pyx_GIVEREF(__pyx_v_indent);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_v_indent);
  __Pyx_INCREF(__pyx_kp_s__22);
  __Pyx_GIVEREF(__pyx_kp_s__22);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_kp_s__22);
  __Pyx_INCREF(__pyx_v_selfTag);
  __Pyx_GIVEREF(__pyx_v_selfTag);
  PyList_SET_ITEM(__pyx_t_1, 3, __pyx_v_selfTag);
  __Pyx_INCREF(__pyx_kp_s__21);
  __Pyx_GIVEREF(__pyx_kp_s__21);
  PyList_SET_ITEM(__pyx_t_1, 4, __pyx_kp_s__21);
  __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_v_out, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 693, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF_SET(__pyx_v_out, ((PyObject*)__pyx_t_5));
  __pyx_t_5 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":694
 * 
 *         out += [nl, indent, "</", selfTag, ">"]
 *         return "".join(out)             # <<<<<<<<<<<<<<
 * 
 *     def __lookup(self, sub):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = __Pyx_PyString_Join(__pyx_kp_s__12, __pyx_v_out); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":635
 *         return ret
 * 
 *     def asXML(self, doctag=None, namedItemsOnly=False, indent="", formatted=True):             # <<<<<<<<<<<<<<
 *         """
 *         (Deprecated) Returns the parse results as XML. Tags are created for tokens and lists that have defined results names.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.asXML", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_nl);
  __Pyx_XDECREF(__pyx_v_out);
  __Pyx_XDECREF(__pyx_v_namedItems);
  __Pyx_XDECREF(__pyx_v_nextLevelIndent);
  __Pyx_XDECREF(__pyx_v_selfTag);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_res);
  __Pyx_XDECREF(__pyx_v_resTag);
  __Pyx_XDECREF(__pyx_v_xmlBodyText);
  __Pyx_XDECREF(__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_5asXML_2generator9);
  __Pyx_XDECREF(__pyx_v_indent);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":696
 *         return "".join(out)
 * 
 *     def __lookup(self, sub):             # <<<<<<<<<<<<<<
 *         for k, vlist in self.__tokdict.items():
 *             for v, loc in vlist:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_69__lookup(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_69__lookup = {"__lookup", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_69__lookup, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_69__lookup(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_sub = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__lookup (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_sub,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sub)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__lookup", 1, 2, 2, 1); __PYX_ERR(0, 696, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__lookup") < 0)) __PYX_ERR(0, 696, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_sub = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__lookup", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 696, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.__lookup", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_68__lookup(__pyx_self, __pyx_v_self, __pyx_v_sub);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_68__lookup(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_sub) {
  PyObject *__pyx_v_k = NULL;
  PyObject *__pyx_v_vlist = NULL;
  PyObject *__pyx_v_v = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_loc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  Py_ssize_t __pyx_t_9;
  PyObject *(*__pyx_t_10)(PyObject *);
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  int __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__lookup", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":697
 * 
 *     def __lookup(self, sub):
 *         for k, vlist in self.__tokdict.items():             # <<<<<<<<<<<<<<
 *             for v, loc in vlist:
 *                 if sub is v:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__tokdict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_items); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_3 = __pyx_t_1; __Pyx_INCREF(__pyx_t_3); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 697, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 697, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 697, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 697, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 697, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 697, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_5(__pyx_t_3);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 697, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
      PyObject* sequence = __pyx_t_1;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 697, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_6);
      #else
      __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 697, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 697, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_7 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 697, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_8 = Py_TYPE(__pyx_t_7)->tp_iternext;
      index = 0; __pyx_t_2 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_2)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_2);
      index = 1; __pyx_t_6 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 2) < 0) __PYX_ERR(0, 697, __pyx_L1_error)
      __pyx_t_8 = NULL;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 697, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_k, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_XDECREF_SET(__pyx_v_vlist, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":698
 *     def __lookup(self, sub):
 *         for k, vlist in self.__tokdict.items():
 *             for v, loc in vlist:             # <<<<<<<<<<<<<<
 *                 if sub is v:
 *                     return k
 */
    if (likely(PyList_CheckExact(__pyx_v_vlist)) || PyTuple_CheckExact(__pyx_v_vlist)) {
      __pyx_t_1 = __pyx_v_vlist; __Pyx_INCREF(__pyx_t_1); __pyx_t_9 = 0;
      __pyx_t_10 = NULL;
    } else {
      __pyx_t_9 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_vlist); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 698, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_10 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 698, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_10)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_9); __Pyx_INCREF(__pyx_t_6); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 698, __pyx_L1_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_1, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 698, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        } else {
          if (__pyx_t_9 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_9); __Pyx_INCREF(__pyx_t_6); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 698, __pyx_L1_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_1, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 698, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        }
      } else {
        __pyx_t_6 = __pyx_t_10(__pyx_t_1);
        if (unlikely(!__pyx_t_6)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 698, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_6);
      }
      if ((likely(PyTuple_CheckExact(__pyx_t_6))) || (PyList_CheckExact(__pyx_t_6))) {
        PyObject* sequence = __pyx_t_6;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 698, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_7 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_7 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_7);
        #else
        __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 698, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 698, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_11 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 698, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_8 = Py_TYPE(__pyx_t_11)->tp_iternext;
        index = 0; __pyx_t_2 = __pyx_t_8(__pyx_t_11); if (unlikely(!__pyx_t_2)) goto __pyx_L9_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_2);
        index = 1; __pyx_t_7 = __pyx_t_8(__pyx_t_11); if (unlikely(!__pyx_t_7)) goto __pyx_L9_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_7);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_11), 2) < 0) __PYX_ERR(0, 698, __pyx_L1_error)
        __pyx_t_8 = NULL;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        goto __pyx_L10_unpacking_done;
        __pyx_L9_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_8 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 698, __pyx_L1_error)
        __pyx_L10_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_XDECREF_SET(__pyx_v_loc, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":699
 *         for k, vlist in self.__tokdict.items():
 *             for v, loc in vlist:
 *                 if sub is v:             # <<<<<<<<<<<<<<
 *                     return k
 *         return None
 */
      __pyx_t_12 = (__pyx_v_sub == __pyx_v_v);
      __pyx_t_13 = (__pyx_t_12 != 0);
      if (__pyx_t_13) {

        /* "Aspidites/_vendor/pyparsing_extension.py":700
 *             for v, loc in vlist:
 *                 if sub is v:
 *                     return k             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_v_k);
        __pyx_r = __pyx_v_k;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        goto __pyx_L0;

        /* "Aspidites/_vendor/pyparsing_extension.py":699
 *         for k, vlist in self.__tokdict.items():
 *             for v, loc in vlist:
 *                 if sub is v:             # <<<<<<<<<<<<<<
 *                     return k
 *         return None
 */
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":698
 *     def __lookup(self, sub):
 *         for k, vlist in self.__tokdict.items():
 *             for v, loc in vlist:             # <<<<<<<<<<<<<<
 *                 if sub is v:
 *                     return k
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":697
 * 
 *     def __lookup(self, sub):
 *         for k, vlist in self.__tokdict.items():             # <<<<<<<<<<<<<<
 *             for v, loc in vlist:
 *                 if sub is v:
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":701
 *                 if sub is v:
 *                     return k
 *         return None             # <<<<<<<<<<<<<<
 * 
 *     def getName(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":696
 *         return "".join(out)
 * 
 *     def __lookup(self, sub):             # <<<<<<<<<<<<<<
 *         for k, vlist in self.__tokdict.items():
 *             for v, loc in vlist:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.__lookup", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XDECREF(__pyx_v_vlist);
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_loc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":703
 *         return None
 * 
 *     def getName(self):             # <<<<<<<<<<<<<<
 *         r"""
 *         Returns the results name for this token expression. Useful when several
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_71getName(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_70getName[] = "\n        Returns the results name for this token expression. Useful when several\n        different expressions might match at a particular location.\n\n        Example::\n\n            integer = Word(nums)\n            ssn_expr = Regex(r\"\\d\\d\\d-\\d\\d-\\d\\d\\d\\d\")\n            house_number_expr = Suppress('#') + Word(nums, alphanums)\n            user_data = (Group(house_number_expr)(\"house_number\")\n                        | Group(ssn_expr)(\"ssn\")\n                        | Group(integer)(\"age\"))\n            user_info = OneOrMore(user_data)\n\n            result = user_info.parseString(\"22 111-22-3333 #221B\")\n            for item in result:\n                print(item.getName(), ':', item[0])\n\n        prints::\n\n            age : 22\n            ssn : 111-22-3333\n            house_number : 221B\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_71getName = {"getName", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_71getName, METH_O, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_70getName};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_71getName(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getName (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_70getName(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_70getName(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_par = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getName", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":728
 *             house_number : 221B
 *         """
 *         if self.__name:             # <<<<<<<<<<<<<<
 *             return self.__name
 *         elif self.__parent:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 728, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 728, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":729
 *         """
 *         if self.__name:
 *             return self.__name             # <<<<<<<<<<<<<<
 *         elif self.__parent:
 *             par = self.__parent()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 729, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":728
 *             house_number : 221B
 *         """
 *         if self.__name:             # <<<<<<<<<<<<<<
 *             return self.__name
 *         elif self.__parent:
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":730
 *         if self.__name:
 *             return self.__name
 *         elif self.__parent:             # <<<<<<<<<<<<<<
 *             par = self.__parent()
 *             if par:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":731
 *             return self.__name
 *         elif self.__parent:
 *             par = self.__parent()             # <<<<<<<<<<<<<<
 *             if par:
 *                 return par.__lookup(self)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__parent); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 731, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 731, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_par = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":732
 *         elif self.__parent:
 *             par = self.__parent()
 *             if par:             # <<<<<<<<<<<<<<
 *                 return par.__lookup(self)
 *             else:
 */
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_par); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 732, __pyx_L1_error)
    if (__pyx_t_2) {

      /* "Aspidites/_vendor/pyparsing_extension.py":733
 *             par = self.__parent()
 *             if par:
 *                 return par.__lookup(self)             # <<<<<<<<<<<<<<
 *             else:
 *                 return None
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_par, __pyx_n_s_ParseResults__lookup); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 733, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_self) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_self);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 733, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_1;
      __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "Aspidites/_vendor/pyparsing_extension.py":732
 *         elif self.__parent:
 *             par = self.__parent()
 *             if par:             # <<<<<<<<<<<<<<
 *                 return par.__lookup(self)
 *             else:
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":735
 *                 return par.__lookup(self)
 *             else:
 *                 return None             # <<<<<<<<<<<<<<
 *         elif (len(self) == 1
 *               and len(self.__tokdict) == 1
 */
    /*else*/ {
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":730
 *         if self.__name:
 *             return self.__name
 *         elif self.__parent:             # <<<<<<<<<<<<<<
 *             par = self.__parent()
 *             if par:
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":736
 *             else:
 *                 return None
 *         elif (len(self) == 1             # <<<<<<<<<<<<<<
 *               and len(self.__tokdict) == 1
 *               and next(iter(self.__tokdict.values()))[0][1] in (0, -1)):
 */
  __pyx_t_5 = PyObject_Length(__pyx_v_self); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 736, __pyx_L1_error)
  __pyx_t_6 = ((__pyx_t_5 == 1) != 0);
  if (__pyx_t_6) {
  } else {
    __pyx_t_2 = __pyx_t_6;
    goto __pyx_L5_bool_binop_done;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":737
 *                 return None
 *         elif (len(self) == 1
 *               and len(self.__tokdict) == 1             # <<<<<<<<<<<<<<
 *               and next(iter(self.__tokdict.values()))[0][1] in (0, -1)):
 *             return next(iter(self.__tokdict.keys()))
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__tokdict); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 737, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 737, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = ((__pyx_t_5 == 1) != 0);
  if (__pyx_t_6) {
  } else {
    __pyx_t_2 = __pyx_t_6;
    goto __pyx_L5_bool_binop_done;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":738
 *         elif (len(self) == 1
 *               and len(self.__tokdict) == 1
 *               and next(iter(self.__tokdict.values()))[0][1] in (0, -1)):             # <<<<<<<<<<<<<<
 *             return next(iter(self.__tokdict.keys()))
 *         else:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__tokdict); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 738, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_values); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 738, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 738, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 738, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyIter_Next(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 738, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 738, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_4, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 738, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_EqObjC(__pyx_t_1, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 738, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 738, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (!__pyx_t_7) {
  } else {
    __pyx_t_6 = __pyx_t_7;
    goto __pyx_L8_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyInt_EqObjC(__pyx_t_1, __pyx_int_neg_1, -1L, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 738, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 738, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_6 = __pyx_t_7;
  __pyx_L8_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = (__pyx_t_6 != 0);
  __pyx_t_2 = __pyx_t_7;
  __pyx_L5_bool_binop_done:;

  /* "Aspidites/_vendor/pyparsing_extension.py":736
 *             else:
 *                 return None
 *         elif (len(self) == 1             # <<<<<<<<<<<<<<
 *               and len(self.__tokdict) == 1
 *               and next(iter(self.__tokdict.values()))[0][1] in (0, -1)):
 */
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":739
 *               and len(self.__tokdict) == 1
 *               and next(iter(self.__tokdict.values()))[0][1] in (0, -1)):
 *             return next(iter(self.__tokdict.keys()))             # <<<<<<<<<<<<<<
 *         else:
 *             return None
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__tokdict); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 739, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_keys); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 739, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 739, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 739, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyIter_Next(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 739, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":736
 *             else:
 *                 return None
 *         elif (len(self) == 1             # <<<<<<<<<<<<<<
 *               and len(self.__tokdict) == 1
 *               and next(iter(self.__tokdict.values()))[0][1] in (0, -1)):
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":741
 *             return next(iter(self.__tokdict.keys()))
 *         else:
 *             return None             # <<<<<<<<<<<<<<
 * 
 *     def dump(self, indent='', full=True, include_list=True, _depth=0):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":703
 *         return None
 * 
 *     def getName(self):             # <<<<<<<<<<<<<<
 *         r"""
 *         Returns the results name for this token expression. Useful when several
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.getName", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_par);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":743
 *             return None
 * 
 *     def dump(self, indent='', full=True, include_list=True, _depth=0):             # <<<<<<<<<<<<<<
 *         """
 *         Diagnostic method for listing out the contents of
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_73dump(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_72dump[] = "\n        Diagnostic method for listing out the contents of\n        a :class:`ParseResults`. Accepts an optional ``indent`` argument so\n        that this string can be embedded in a nested display of other data.\n\n        Example::\n\n            integer = Word(nums)\n            date_str = integer(\"year\") + '/' + integer(\"month\") + '/' + integer(\"day\")\n\n            result = date_str.parseString('12/31/1999')\n            print(result.dump())\n\n        prints::\n\n            ['12', '/', '31', '/', '1999']\n            - day: 1999\n            - month: 31\n            - year: 12\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_73dump = {"dump", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_73dump, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_72dump};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_73dump(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_indent = 0;
  PyObject *__pyx_v_full = 0;
  PyObject *__pyx_v_include_list = 0;
  PyObject *__pyx_v__depth = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dump (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_indent,&__pyx_n_s_full,&__pyx_n_s_include_list,&__pyx_n_s_depth,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[1] = ((PyObject *)((PyObject*)__pyx_kp_s__12));
    values[2] = ((PyObject *)((PyObject *)Py_True));
    values[3] = ((PyObject *)((PyObject *)Py_True));
    values[4] = ((PyObject *)((PyObject *)__pyx_int_0));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_indent);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_full);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_include_list);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_depth);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dump") < 0)) __PYX_ERR(0, 743, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_indent = values[1];
    __pyx_v_full = values[2];
    __pyx_v_include_list = values[3];
    __pyx_v__depth = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dump", 0, 1, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 743, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.dump", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_72dump(__pyx_self, __pyx_v_self, __pyx_v_indent, __pyx_v_full, __pyx_v_include_list, __pyx_v__depth);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_4dump_2generator10(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/pyparsing_extension.py":773
 *         if full:
 *             if self.haskeys():
 *                 items = sorted((str(k), v) for k, v in self.items())             # <<<<<<<<<<<<<<
 *                 for k, v in items:
 *                     if out:
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_4dump_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 773, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_4dump_2generator10, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_ParseResults_dump_locals_genexpr, __pyx_n_s_Aspidites__vendor_pyparsing_exte); if (unlikely(!gen)) __PYX_ERR(0, 773, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.dump.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_4dump_2generator10(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 773, __pyx_L1_error)
  __pyx_r = PyList_New(0); if (unlikely(!__pyx_r)) __PYX_ERR(0, 773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 773, __pyx_L1_error) }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_items); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 773, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 773, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 773, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 773, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 773, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 773, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_5(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 773, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
      PyObject* sequence = __pyx_t_1;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 773, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_6);
      #else
      __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 773, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 773, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_7 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 773, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_8 = Py_TYPE(__pyx_t_7)->tp_iternext;
      index = 0; __pyx_t_3 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_3)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_3);
      index = 1; __pyx_t_6 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 2) < 0) __PYX_ERR(0, 773, __pyx_L1_error)
      __pyx_t_8 = NULL;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L7_unpacking_done;
      __pyx_L6_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 773, __pyx_L1_error)
      __pyx_L7_unpacking_done:;
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_k);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_k, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_v);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_v, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    __pyx_t_6 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_cur_scope->__pyx_v_k); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 773, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 773, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_v);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_v);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_cur_scope->__pyx_v_v);
    __pyx_t_1 = 0;
    if (unlikely(__Pyx_ListComp_Append(__pyx_r, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 773, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_4dump_5generator11(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/pyparsing_extension.py":785
 *                     else:
 *                         out.append(repr(v))
 *             elif any(isinstance(vv, ParseResults) for vv in self):             # <<<<<<<<<<<<<<
 *                 v = self
 *                 for i, vv in enumerate(v):
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_4dump_3genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 785, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_4dump_5generator11, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_ParseResults_dump_locals_genexpr, __pyx_n_s_Aspidites__vendor_pyparsing_exte); if (unlikely(!gen)) __PYX_ERR(0, 785, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.dump.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_4dump_5generator11(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 785, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 785, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_self; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 785, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 785, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 785, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 785, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 785, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 785, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 785, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_vv);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_vv, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ParseResults); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 785, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyObject_IsInstance(__pyx_cur_scope->__pyx_v_vv, __pyx_t_4); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 785, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_6 = (__pyx_t_5 != 0);
    if (__pyx_t_6) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_True);
      __pyx_r = Py_True;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":743
 *             return None
 * 
 *     def dump(self, indent='', full=True, include_list=True, _depth=0):             # <<<<<<<<<<<<<<
 *         """
 *         Diagnostic method for listing out the contents of
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_72dump(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_indent, PyObject *__pyx_v_full, PyObject *__pyx_v_include_list, PyObject *__pyx_v__depth) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump *__pyx_cur_scope;
  PyObject *__pyx_v_out = NULL;
  PyObject *__pyx_v_NL = NULL;
  PyObject *__pyx_v_items = NULL;
  PyObject *__pyx_v_k = NULL;
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_vv = NULL;
  PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_4dump_2generator10 = 0;
  PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_4dump_5generator11 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *(*__pyx_t_9)(PyObject *);
  int __pyx_t_10;
  PyObject *(*__pyx_t_11)(PyObject *);
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("dump", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 743, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);

  /* "Aspidites/_vendor/pyparsing_extension.py":764
 *             - year: 12
 *         """
 *         out = []             # <<<<<<<<<<<<<<
 *         NL = '\n'
 *         if include_list:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 764, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_out = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":765
 *         """
 *         out = []
 *         NL = '\n'             # <<<<<<<<<<<<<<
 *         if include_list:
 *             out.append(indent + str(self.asList()))
 */
  __Pyx_INCREF(__pyx_kp_s__4);
  __pyx_v_NL = __pyx_kp_s__4;

  /* "Aspidites/_vendor/pyparsing_extension.py":766
 *         out = []
 *         NL = '\n'
 *         if include_list:             # <<<<<<<<<<<<<<
 *             out.append(indent + str(self.asList()))
 *         else:
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_include_list); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 766, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":767
 *         NL = '\n'
 *         if include_list:
 *             out.append(indent + str(self.asList()))             # <<<<<<<<<<<<<<
 *         else:
 *             out.append('')
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_asList); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 767, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 767, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 767, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_Add(__pyx_v_indent, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 767, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_out, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 767, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":766
 *         out = []
 *         NL = '\n'
 *         if include_list:             # <<<<<<<<<<<<<<
 *             out.append(indent + str(self.asList()))
 *         else:
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":769
 *             out.append(indent + str(self.asList()))
 *         else:
 *             out.append('')             # <<<<<<<<<<<<<<
 * 
 *         if full:
 */
  /*else*/ {
    __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_out, __pyx_kp_s__12); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 769, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "Aspidites/_vendor/pyparsing_extension.py":771
 *             out.append('')
 * 
 *         if full:             # <<<<<<<<<<<<<<
 *             if self.haskeys():
 *                 items = sorted((str(k), v) for k, v in self.items())
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_full); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 771, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":772
 * 
 *         if full:
 *             if self.haskeys():             # <<<<<<<<<<<<<<
 *                 items = sorted((str(k), v) for k, v in self.items())
 *                 for k, v in items:
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_haskeys); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 772, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 772, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 772, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_2) {

      /* "Aspidites/_vendor/pyparsing_extension.py":773
 *         if full:
 *             if self.haskeys():
 *                 items = sorted((str(k), v) for k, v in self.items())             # <<<<<<<<<<<<<<
 *                 for k, v in items:
 *                     if out:
 */
      __pyx_t_3 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_4dump_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 773, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_Generator_Next(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 773, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_1 = ((PyObject*)__pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_t_5 = PyList_Sort(__pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 773, __pyx_L1_error)
      __pyx_v_items = ((PyObject*)__pyx_t_1);
      __pyx_t_1 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":774
 *             if self.haskeys():
 *                 items = sorted((str(k), v) for k, v in self.items())
 *                 for k, v in items:             # <<<<<<<<<<<<<<
 *                     if out:
 *                         out.append(NL)
 */
      if (unlikely(__pyx_v_items == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
        __PYX_ERR(0, 774, __pyx_L1_error)
      }
      __pyx_t_1 = __pyx_v_items; __Pyx_INCREF(__pyx_t_1); __pyx_t_6 = 0;
      for (;;) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_4); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 774, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 774, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
        if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
          PyObject* sequence = __pyx_t_4;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 774, __pyx_L1_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_7 = PyTuple_GET_ITEM(sequence, 1); 
          } else {
            __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_7 = PyList_GET_ITEM(sequence, 1); 
          }
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_7);
          #else
          __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 774, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_7 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 774, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          #endif
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else {
          Py_ssize_t index = -1;
          __pyx_t_8 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 774, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_9 = Py_TYPE(__pyx_t_8)->tp_iternext;
          index = 0; __pyx_t_3 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_3)) goto __pyx_L8_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_3);
          index = 1; __pyx_t_7 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_7)) goto __pyx_L8_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_7);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_9(__pyx_t_8), 2) < 0) __PYX_ERR(0, 774, __pyx_L1_error)
          __pyx_t_9 = NULL;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          goto __pyx_L9_unpacking_done;
          __pyx_L8_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __pyx_t_9 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 774, __pyx_L1_error)
          __pyx_L9_unpacking_done:;
        }
        __Pyx_XDECREF_SET(__pyx_v_k, __pyx_t_3);
        __pyx_t_3 = 0;
        __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":775
 *                 items = sorted((str(k), v) for k, v in self.items())
 *                 for k, v in items:
 *                     if out:             # <<<<<<<<<<<<<<
 *                         out.append(NL)
 *                     out.append("%s%s- %s: " % (indent, ('  ' * _depth), k))
 */
        __pyx_t_2 = (PyList_GET_SIZE(__pyx_v_out) != 0);
        if (__pyx_t_2) {

          /* "Aspidites/_vendor/pyparsing_extension.py":776
 *                 for k, v in items:
 *                     if out:
 *                         out.append(NL)             # <<<<<<<<<<<<<<
 *                     out.append("%s%s- %s: " % (indent, ('  ' * _depth), k))
 *                     if isinstance(v, ParseResults):
 */
          __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_out, __pyx_v_NL); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 776, __pyx_L1_error)

          /* "Aspidites/_vendor/pyparsing_extension.py":775
 *                 items = sorted((str(k), v) for k, v in self.items())
 *                 for k, v in items:
 *                     if out:             # <<<<<<<<<<<<<<
 *                         out.append(NL)
 *                     out.append("%s%s- %s: " % (indent, ('  ' * _depth), k))
 */
        }

        /* "Aspidites/_vendor/pyparsing_extension.py":777
 *                     if out:
 *                         out.append(NL)
 *                     out.append("%s%s- %s: " % (indent, ('  ' * _depth), k))             # <<<<<<<<<<<<<<
 *                     if isinstance(v, ParseResults):
 *                         if v:
 */
        __pyx_t_4 = PyNumber_Multiply(__pyx_kp_s__19, __pyx_v__depth); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 777, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_7 = PyTuple_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 777, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_v_indent);
        __Pyx_GIVEREF(__pyx_v_indent);
        PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_indent);
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_4);
        __Pyx_INCREF(__pyx_v_k);
        __Pyx_GIVEREF(__pyx_v_k);
        PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_v_k);
        __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_s_s_s, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 777, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_out, __pyx_t_4); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 777, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":778
 *                         out.append(NL)
 *                     out.append("%s%s- %s: " % (indent, ('  ' * _depth), k))
 *                     if isinstance(v, ParseResults):             # <<<<<<<<<<<<<<
 *                         if v:
 *                             out.append(v.dump(indent=indent, full=full, include_list=include_list, _depth=_depth + 1))
 */
        __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ParseResults); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 778, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = PyObject_IsInstance(__pyx_v_v, __pyx_t_4); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 778, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_10 = (__pyx_t_2 != 0);
        if (__pyx_t_10) {

          /* "Aspidites/_vendor/pyparsing_extension.py":779
 *                     out.append("%s%s- %s: " % (indent, ('  ' * _depth), k))
 *                     if isinstance(v, ParseResults):
 *                         if v:             # <<<<<<<<<<<<<<
 *                             out.append(v.dump(indent=indent, full=full, include_list=include_list, _depth=_depth + 1))
 *                         else:
 */
          __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_v_v); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 779, __pyx_L1_error)
          if (__pyx_t_10) {

            /* "Aspidites/_vendor/pyparsing_extension.py":780
 *                     if isinstance(v, ParseResults):
 *                         if v:
 *                             out.append(v.dump(indent=indent, full=full, include_list=include_list, _depth=_depth + 1))             # <<<<<<<<<<<<<<
 *                         else:
 *                             out.append(str(v))
 */
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_v, __pyx_n_s_dump); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 780, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_7 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 780, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_indent, __pyx_v_indent) < 0) __PYX_ERR(0, 780, __pyx_L1_error)
            if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_full, __pyx_v_full) < 0) __PYX_ERR(0, 780, __pyx_L1_error)
            if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_include_list, __pyx_v_include_list) < 0) __PYX_ERR(0, 780, __pyx_L1_error)
            __pyx_t_3 = __Pyx_PyInt_AddObjC(__pyx_v__depth, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 780, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_depth, __pyx_t_3) < 0) __PYX_ERR(0, 780, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_empty_tuple, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 780, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_out, __pyx_t_3); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 780, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

            /* "Aspidites/_vendor/pyparsing_extension.py":779
 *                     out.append("%s%s- %s: " % (indent, ('  ' * _depth), k))
 *                     if isinstance(v, ParseResults):
 *                         if v:             # <<<<<<<<<<<<<<
 *                             out.append(v.dump(indent=indent, full=full, include_list=include_list, _depth=_depth + 1))
 *                         else:
 */
            goto __pyx_L12;
          }

          /* "Aspidites/_vendor/pyparsing_extension.py":782
 *                             out.append(v.dump(indent=indent, full=full, include_list=include_list, _depth=_depth + 1))
 *                         else:
 *                             out.append(str(v))             # <<<<<<<<<<<<<<
 *                     else:
 *                         out.append(repr(v))
 */
          /*else*/ {
            __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_v_v); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 782, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_out, __pyx_t_3); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 782, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __pyx_L12:;

          /* "Aspidites/_vendor/pyparsing_extension.py":778
 *                         out.append(NL)
 *                     out.append("%s%s- %s: " % (indent, ('  ' * _depth), k))
 *                     if isinstance(v, ParseResults):             # <<<<<<<<<<<<<<
 *                         if v:
 *                             out.append(v.dump(indent=indent, full=full, include_list=include_list, _depth=_depth + 1))
 */
          goto __pyx_L11;
        }

        /* "Aspidites/_vendor/pyparsing_extension.py":784
 *                             out.append(str(v))
 *                     else:
 *                         out.append(repr(v))             # <<<<<<<<<<<<<<
 *             elif any(isinstance(vv, ParseResults) for vv in self):
 *                 v = self
 */
        /*else*/ {
          __pyx_t_3 = PyObject_Repr(__pyx_v_v); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 784, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_out, __pyx_t_3); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 784, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __pyx_L11:;

        /* "Aspidites/_vendor/pyparsing_extension.py":774
 *             if self.haskeys():
 *                 items = sorted((str(k), v) for k, v in self.items())
 *                 for k, v in items:             # <<<<<<<<<<<<<<
 *                     if out:
 *                         out.append(NL)
 */
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":772
 * 
 *         if full:
 *             if self.haskeys():             # <<<<<<<<<<<<<<
 *                 items = sorted((str(k), v) for k, v in self.items())
 *                 for k, v in items:
 */
      goto __pyx_L5;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":785
 *                     else:
 *                         out.append(repr(v))
 *             elif any(isinstance(vv, ParseResults) for vv in self):             # <<<<<<<<<<<<<<
 *                 v = self
 *                 for i, vv in enumerate(v):
 */
    __pyx_t_1 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_4dump_3genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 785, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_Generator_Next(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 785, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 785, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_10) {

      /* "Aspidites/_vendor/pyparsing_extension.py":786
 *                         out.append(repr(v))
 *             elif any(isinstance(vv, ParseResults) for vv in self):
 *                 v = self             # <<<<<<<<<<<<<<
 *                 for i, vv in enumerate(v):
 *                     if isinstance(vv, ParseResults):
 */
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
      __pyx_v_v = __pyx_cur_scope->__pyx_v_self;

      /* "Aspidites/_vendor/pyparsing_extension.py":787
 *             elif any(isinstance(vv, ParseResults) for vv in self):
 *                 v = self
 *                 for i, vv in enumerate(v):             # <<<<<<<<<<<<<<
 *                     if isinstance(vv, ParseResults):
 *                         out.append("\n%s%s[%d]:\n%s%s%s" % (indent,
 */
      __Pyx_INCREF(__pyx_int_0);
      __pyx_t_3 = __pyx_int_0;
      if (likely(PyList_CheckExact(__pyx_v_v)) || PyTuple_CheckExact(__pyx_v_v)) {
        __pyx_t_1 = __pyx_v_v; __Pyx_INCREF(__pyx_t_1); __pyx_t_6 = 0;
        __pyx_t_11 = NULL;
      } else {
        __pyx_t_6 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_v); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 787, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 787, __pyx_L1_error)
      }
      for (;;) {
        if (likely(!__pyx_t_11)) {
          if (likely(PyList_CheckExact(__pyx_t_1))) {
            if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_1)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_7 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_7); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 787, __pyx_L1_error)
            #else
            __pyx_t_7 = PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 787, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_7);
            #endif
          } else {
            if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_7); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 787, __pyx_L1_error)
            #else
            __pyx_t_7 = PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 787, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_7);
            #endif
          }
        } else {
          __pyx_t_7 = __pyx_t_11(__pyx_t_1);
          if (unlikely(!__pyx_t_7)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 787, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_7);
        }
        __Pyx_XDECREF_SET(__pyx_v_vv, __pyx_t_7);
        __pyx_t_7 = 0;
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_3);
        __pyx_t_7 = __Pyx_PyInt_AddObjC(__pyx_t_3, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 787, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3);
        __pyx_t_3 = __pyx_t_7;
        __pyx_t_7 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":788
 *                 v = self
 *                 for i, vv in enumerate(v):
 *                     if isinstance(vv, ParseResults):             # <<<<<<<<<<<<<<
 *                         out.append("\n%s%s[%d]:\n%s%s%s" % (indent,
 *                                                             ('  ' * (_depth)),
 */
        __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_ParseResults); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 788, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_10 = PyObject_IsInstance(__pyx_v_vv, __pyx_t_7); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 788, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_2 = (__pyx_t_10 != 0);
        if (__pyx_t_2) {

          /* "Aspidites/_vendor/pyparsing_extension.py":790
 *                     if isinstance(vv, ParseResults):
 *                         out.append("\n%s%s[%d]:\n%s%s%s" % (indent,
 *                                                             ('  ' * (_depth)),             # <<<<<<<<<<<<<<
 *                                                             i,
 *                                                             indent,
 */
          __pyx_t_7 = PyNumber_Multiply(__pyx_kp_s__19, __pyx_v__depth); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 790, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);

          /* "Aspidites/_vendor/pyparsing_extension.py":793
 *                                                             i,
 *                                                             indent,
 *                                                             ('  ' * (_depth + 1)),             # <<<<<<<<<<<<<<
 *                                                             vv.dump(indent=indent,
 *                                                                     full=full,
 */
          __pyx_t_4 = __Pyx_PyInt_AddObjC(__pyx_v__depth, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 793, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_8 = PyNumber_Multiply(__pyx_kp_s__19, __pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 793, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":794
 *                                                             indent,
 *                                                             ('  ' * (_depth + 1)),
 *                                                             vv.dump(indent=indent,             # <<<<<<<<<<<<<<
 *                                                                     full=full,
 *                                                                     include_list=include_list,
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_vv, __pyx_n_s_dump); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 794, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_12 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 794, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_12);
          if (PyDict_SetItem(__pyx_t_12, __pyx_n_s_indent, __pyx_v_indent) < 0) __PYX_ERR(0, 794, __pyx_L1_error)

          /* "Aspidites/_vendor/pyparsing_extension.py":795
 *                                                             ('  ' * (_depth + 1)),
 *                                                             vv.dump(indent=indent,
 *                                                                     full=full,             # <<<<<<<<<<<<<<
 *                                                                     include_list=include_list,
 *                                                                     _depth=_depth + 1)))
 */
          if (PyDict_SetItem(__pyx_t_12, __pyx_n_s_full, __pyx_v_full) < 0) __PYX_ERR(0, 794, __pyx_L1_error)

          /* "Aspidites/_vendor/pyparsing_extension.py":796
 *                                                             vv.dump(indent=indent,
 *                                                                     full=full,
 *                                                                     include_list=include_list,             # <<<<<<<<<<<<<<
 *                                                                     _depth=_depth + 1)))
 *                     else:
 */
          if (PyDict_SetItem(__pyx_t_12, __pyx_n_s_include_list, __pyx_v_include_list) < 0) __PYX_ERR(0, 794, __pyx_L1_error)

          /* "Aspidites/_vendor/pyparsing_extension.py":797
 *                                                                     full=full,
 *                                                                     include_list=include_list,
 *                                                                     _depth=_depth + 1)))             # <<<<<<<<<<<<<<
 *                     else:
 *                         out.append("\n%s%s[%d]:\n%s%s%s" % (indent,
 */
          __pyx_t_13 = __Pyx_PyInt_AddObjC(__pyx_v__depth, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 797, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_13);
          if (PyDict_SetItem(__pyx_t_12, __pyx_n_s_depth, __pyx_t_13) < 0) __PYX_ERR(0, 794, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":794
 *                                                             indent,
 *                                                             ('  ' * (_depth + 1)),
 *                                                             vv.dump(indent=indent,             # <<<<<<<<<<<<<<
 *                                                                     full=full,
 *                                                                     include_list=include_list,
 */
          __pyx_t_13 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_empty_tuple, __pyx_t_12); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 794, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":789
 *                 for i, vv in enumerate(v):
 *                     if isinstance(vv, ParseResults):
 *                         out.append("\n%s%s[%d]:\n%s%s%s" % (indent,             # <<<<<<<<<<<<<<
 *                                                             ('  ' * (_depth)),
 *                                                             i,
 */
          __pyx_t_12 = PyTuple_New(6); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 789, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_INCREF(__pyx_v_indent);
          __Pyx_GIVEREF(__pyx_v_indent);
          PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_v_indent);
          __Pyx_GIVEREF(__pyx_t_7);
          PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_t_7);
          __Pyx_INCREF(__pyx_v_i);
          __Pyx_GIVEREF(__pyx_v_i);
          PyTuple_SET_ITEM(__pyx_t_12, 2, __pyx_v_i);
          __Pyx_INCREF(__pyx_v_indent);
          __Pyx_GIVEREF(__pyx_v_indent);
          PyTuple_SET_ITEM(__pyx_t_12, 3, __pyx_v_indent);
          __Pyx_GIVEREF(__pyx_t_8);
          PyTuple_SET_ITEM(__pyx_t_12, 4, __pyx_t_8);
          __Pyx_GIVEREF(__pyx_t_13);
          PyTuple_SET_ITEM(__pyx_t_12, 5, __pyx_t_13);
          __pyx_t_7 = 0;
          __pyx_t_8 = 0;
          __pyx_t_13 = 0;
          __pyx_t_13 = __Pyx_PyString_Format(__pyx_kp_s_s_s_d_s_s_s, __pyx_t_12); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 789, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_out, __pyx_t_13); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 789, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":788
 *                 v = self
 *                 for i, vv in enumerate(v):
 *                     if isinstance(vv, ParseResults):             # <<<<<<<<<<<<<<
 *                         out.append("\n%s%s[%d]:\n%s%s%s" % (indent,
 *                                                             ('  ' * (_depth)),
 */
          goto __pyx_L15;
        }

        /* "Aspidites/_vendor/pyparsing_extension.py":799
 *                                                                     _depth=_depth + 1)))
 *                     else:
 *                         out.append("\n%s%s[%d]:\n%s%s%s" % (indent,             # <<<<<<<<<<<<<<
 *                                                             ('  ' * (_depth)),
 *                                                             i,
 */
        /*else*/ {

          /* "Aspidites/_vendor/pyparsing_extension.py":800
 *                     else:
 *                         out.append("\n%s%s[%d]:\n%s%s%s" % (indent,
 *                                                             ('  ' * (_depth)),             # <<<<<<<<<<<<<<
 *                                                             i,
 *                                                             indent,
 */
          __pyx_t_13 = PyNumber_Multiply(__pyx_kp_s__19, __pyx_v__depth); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 800, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_13);

          /* "Aspidites/_vendor/pyparsing_extension.py":803
 *                                                             i,
 *                                                             indent,
 *                                                             ('  ' * (_depth + 1)),             # <<<<<<<<<<<<<<
 *                                                             str(vv)))
 * 
 */
          __pyx_t_12 = __Pyx_PyInt_AddObjC(__pyx_v__depth, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 803, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_8 = PyNumber_Multiply(__pyx_kp_s__19, __pyx_t_12); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 803, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":804
 *                                                             indent,
 *                                                             ('  ' * (_depth + 1)),
 *                                                             str(vv)))             # <<<<<<<<<<<<<<
 * 
 *         return "".join(out)
 */
          __pyx_t_12 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_v_vv); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 804, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_12);

          /* "Aspidites/_vendor/pyparsing_extension.py":799
 *                                                                     _depth=_depth + 1)))
 *                     else:
 *                         out.append("\n%s%s[%d]:\n%s%s%s" % (indent,             # <<<<<<<<<<<<<<
 *                                                             ('  ' * (_depth)),
 *                                                             i,
 */
          __pyx_t_7 = PyTuple_New(6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 799, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_INCREF(__pyx_v_indent);
          __Pyx_GIVEREF(__pyx_v_indent);
          PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_indent);
          __Pyx_GIVEREF(__pyx_t_13);
          PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_13);
          __Pyx_INCREF(__pyx_v_i);
          __Pyx_GIVEREF(__pyx_v_i);
          PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_v_i);
          __Pyx_INCREF(__pyx_v_indent);
          __Pyx_GIVEREF(__pyx_v_indent);
          PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_v_indent);
          __Pyx_GIVEREF(__pyx_t_8);
          PyTuple_SET_ITEM(__pyx_t_7, 4, __pyx_t_8);
          __Pyx_GIVEREF(__pyx_t_12);
          PyTuple_SET_ITEM(__pyx_t_7, 5, __pyx_t_12);
          __pyx_t_13 = 0;
          __pyx_t_8 = 0;
          __pyx_t_12 = 0;
          __pyx_t_12 = __Pyx_PyString_Format(__pyx_kp_s_s_s_d_s_s_s, __pyx_t_7); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 799, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_out, __pyx_t_12); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 799, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        }
        __pyx_L15:;

        /* "Aspidites/_vendor/pyparsing_extension.py":787
 *             elif any(isinstance(vv, ParseResults) for vv in self):
 *                 v = self
 *                 for i, vv in enumerate(v):             # <<<<<<<<<<<<<<
 *                     if isinstance(vv, ParseResults):
 *                         out.append("\n%s%s[%d]:\n%s%s%s" % (indent,
 */
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":785
 *                     else:
 *                         out.append(repr(v))
 *             elif any(isinstance(vv, ParseResults) for vv in self):             # <<<<<<<<<<<<<<
 *                 v = self
 *                 for i, vv in enumerate(v):
 */
    }
    __pyx_L5:;

    /* "Aspidites/_vendor/pyparsing_extension.py":771
 *             out.append('')
 * 
 *         if full:             # <<<<<<<<<<<<<<
 *             if self.haskeys():
 *                 items = sorted((str(k), v) for k, v in self.items())
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":806
 *                                                             str(vv)))
 * 
 *         return "".join(out)             # <<<<<<<<<<<<<<
 * 
 *     def pprint(self, *args, **kwargs):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyString_Join(__pyx_kp_s__12, __pyx_v_out); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 806, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":743
 *             return None
 * 
 *     def dump(self, indent='', full=True, include_list=True, _depth=0):             # <<<<<<<<<<<<<<
 *         """
 *         Diagnostic method for listing out the contents of
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.dump", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_out);
  __Pyx_XDECREF(__pyx_v_NL);
  __Pyx_XDECREF(__pyx_v_items);
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_vv);
  __Pyx_XDECREF(__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_4dump_2generator10);
  __Pyx_XDECREF(__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_4dump_5generator11);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":808
 *         return "".join(out)
 * 
 *     def pprint(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         """
 *         Pretty-printer for parsed results as a list, using the
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_75pprint(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_74pprint[] = "\n        Pretty-printer for parsed results as a list, using the\n        `pprint <https://docs.python.org/3/library/pprint.html>`_ module.\n        Accepts additional positional or keyword args as defined for\n        `pprint.pprint <https://docs.python.org/3/library/pprint.html#pprint.pprint>`_ .\n\n        Example::\n\n            ident = Word(alphas, alphanums)\n            num = Word(nums)\n            func = Forward()\n            term = ident | num | Group('(' + func + ')')\n            func <<= ident + Group(Optional(delimitedList(term)))\n            result = func.parseString(\"fna a,b,(fnb c,d,200),100\")\n            result.pprint(width=40)\n\n        prints::\n\n            ['fna',\n             ['a',\n              'b',\n              ['(', 'fnb', ['c', 'd', '200'], ')'],\n              '100']]\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_75pprint = {"pprint", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_75pprint, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_74pprint};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_75pprint(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pprint (wrapper)", 0);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  if (PyTuple_GET_SIZE(__pyx_args) > 1) {
    __pyx_v_args = PyTuple_GetSlice(__pyx_args, 1, PyTuple_GET_SIZE(__pyx_args));
    if (unlikely(!__pyx_v_args)) {
      __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
      __Pyx_RefNannyFinishContext();
      return NULL;
    }
    __Pyx_GOTREF(__pyx_v_args);
  } else {
    __pyx_v_args = __pyx_empty_tuple; __Pyx_INCREF(__pyx_empty_tuple);
  }
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        default:
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t used_pos_args = (pos_args < 1) ? pos_args : 1;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwargs, values, used_pos_args, "pprint") < 0)) __PYX_ERR(0, 808, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) < 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_self = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("pprint", 0, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 808, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_args); __pyx_v_args = 0;
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.pprint", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_74pprint(__pyx_self, __pyx_v_self, __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_74pprint(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pprint", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":833
 *               '100']]
 *         """
 *         pprint.pprint(self.asList(), *args, **kwargs)             # <<<<<<<<<<<<<<
 * 
 *     # add support for pickle protocol
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pprint); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 833, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_pprint); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 833, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_asList); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 833, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 833, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 833, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Add(__pyx_t_3, __pyx_v_args); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 833, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 833, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 833, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":808
 *         return "".join(out)
 * 
 *     def pprint(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         """
 *         Pretty-printer for parsed results as a list, using the
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.pprint", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":836
 * 
 *     # add support for pickle protocol
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return (self.__toklist,
 *                 (self.__tokdict.copy(),
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_77__getstate__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_77__getstate__ = {"__getstate__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_77__getstate__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_77__getstate__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getstate__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_76__getstate__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_76__getstate__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getstate__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":837
 *     # add support for pickle protocol
 *     def __getstate__(self):
 *         return (self.__toklist,             # <<<<<<<<<<<<<<
 *                 (self.__tokdict.copy(),
 *                  self.__parent is not None and self.__parent() or None,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__toklist); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 837, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "Aspidites/_vendor/pyparsing_extension.py":838
 *     def __getstate__(self):
 *         return (self.__toklist,
 *                 (self.__tokdict.copy(),             # <<<<<<<<<<<<<<
 *                  self.__parent is not None and self.__parent() or None,
 *                  self.__accumNames,
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__tokdict); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 838, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_copy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 838, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 838, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":839
 *         return (self.__toklist,
 *                 (self.__tokdict.copy(),
 *                  self.__parent is not None and self.__parent() or None,             # <<<<<<<<<<<<<<
 *                  self.__accumNames,
 *                  self.__name))
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__parent); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 839, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = (__pyx_t_3 != Py_None);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!__pyx_t_5) {
    goto __pyx_L4_next_or;
  } else {
  }
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__parent); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 839, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_3 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 839, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 839, __pyx_L1_error)
  if (!__pyx_t_5) {
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_4 = __pyx_t_3;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_L4_next_or:;
  __Pyx_INCREF(Py_None);
  __pyx_t_4 = Py_None;
  __pyx_L3_bool_binop_done:;

  /* "Aspidites/_vendor/pyparsing_extension.py":840
 *                 (self.__tokdict.copy(),
 *                  self.__parent is not None and self.__parent() or None,
 *                  self.__accumNames,             # <<<<<<<<<<<<<<
 *                  self.__name))
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__accumNames); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "Aspidites/_vendor/pyparsing_extension.py":841
 *                  self.__parent is not None and self.__parent() or None,
 *                  self.__accumNames,
 *                  self.__name))             # <<<<<<<<<<<<<<
 * 
 *     def __setstate__(self, state):
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__name); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 841, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);

  /* "Aspidites/_vendor/pyparsing_extension.py":838
 *     def __getstate__(self):
 *         return (self.__toklist,
 *                 (self.__tokdict.copy(),             # <<<<<<<<<<<<<<
 *                  self.__parent is not None and self.__parent() or None,
 *                  self.__accumNames,
 */
  __pyx_t_7 = PyTuple_New(4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 838, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_t_6);
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_t_3 = 0;
  __pyx_t_6 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":837
 *     # add support for pickle protocol
 *     def __getstate__(self):
 *         return (self.__toklist,             # <<<<<<<<<<<<<<
 *                 (self.__tokdict.copy(),
 *                  self.__parent is not None and self.__parent() or None,
 */
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 837, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_7);
  __pyx_t_1 = 0;
  __pyx_t_7 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":836
 * 
 *     # add support for pickle protocol
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return (self.__toklist,
 *                 (self.__tokdict.copy(),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.__getstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":843
 *                  self.__name))
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self.__toklist = state[0]
 *         self.__tokdict, par, inAccumNames, self.__name = state[1]
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_79__setstate__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_79__setstate__ = {"__setstate__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_79__setstate__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_79__setstate__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_state = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_state,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__setstate__", 1, 2, 2, 1); __PYX_ERR(0, 843, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__setstate__") < 0)) __PYX_ERR(0, 843, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_state = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 843, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.__setstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_78__setstate__(__pyx_self, __pyx_v_self, __pyx_v_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_78__setstate__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_state) {
  PyObject *__pyx_v_par = NULL;
  PyObject *__pyx_v_inAccumNames = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *(*__pyx_t_7)(PyObject *);
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":844
 * 
 *     def __setstate__(self, state):
 *         self.__toklist = state[0]             # <<<<<<<<<<<<<<
 *         self.__tokdict, par, inAccumNames, self.__name = state[1]
 *         self.__accumNames = {}
 */
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__toklist, __pyx_t_1) < 0) __PYX_ERR(0, 844, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":845
 *     def __setstate__(self, state):
 *         self.__toklist = state[0]
 *         self.__tokdict, par, inAccumNames, self.__name = state[1]             # <<<<<<<<<<<<<<
 *         self.__accumNames = {}
 *         self.__accumNames.update(inAccumNames)
 */
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 845, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 4)) {
      if (size > 4) __Pyx_RaiseTooManyValuesError(4);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 845, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 2); 
      __pyx_t_5 = PyTuple_GET_ITEM(sequence, 3); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_4 = PyList_GET_ITEM(sequence, 2); 
      __pyx_t_5 = PyList_GET_ITEM(sequence, 3); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    #else
    {
      Py_ssize_t i;
      PyObject** temps[4] = {&__pyx_t_2,&__pyx_t_3,&__pyx_t_4,&__pyx_t_5};
      for (i=0; i < 4; i++) {
        PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 845, __pyx_L1_error)
        __Pyx_GOTREF(item);
        *(temps[i]) = item;
      }
    }
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    PyObject** temps[4] = {&__pyx_t_2,&__pyx_t_3,&__pyx_t_4,&__pyx_t_5};
    __pyx_t_6 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 845, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_7 = Py_TYPE(__pyx_t_6)->tp_iternext;
    for (index=0; index < 4; index++) {
      PyObject* item = __pyx_t_7(__pyx_t_6); if (unlikely(!item)) goto __pyx_L3_unpacking_failed;
      __Pyx_GOTREF(item);
      *(temps[index]) = item;
    }
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_6), 4) < 0) __PYX_ERR(0, 845, __pyx_L1_error)
    __pyx_t_7 = NULL;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 845, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__tokdict, __pyx_t_2) < 0) __PYX_ERR(0, 845, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_par = __pyx_t_3;
  __pyx_t_3 = 0;
  __pyx_v_inAccumNames = __pyx_t_4;
  __pyx_t_4 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__name, __pyx_t_5) < 0) __PYX_ERR(0, 845, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":846
 *         self.__toklist = state[0]
 *         self.__tokdict, par, inAccumNames, self.__name = state[1]
 *         self.__accumNames = {}             # <<<<<<<<<<<<<<
 *         self.__accumNames.update(inAccumNames)
 *         if par is not None:
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 846, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__accumNames, __pyx_t_1) < 0) __PYX_ERR(0, 846, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":847
 *         self.__tokdict, par, inAccumNames, self.__name = state[1]
 *         self.__accumNames = {}
 *         self.__accumNames.update(inAccumNames)             # <<<<<<<<<<<<<<
 *         if par is not None:
 *             self.__parent = wkref(par)
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__accumNames); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 847, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 847, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_inAccumNames) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_inAccumNames);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 847, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":848
 *         self.__accumNames = {}
 *         self.__accumNames.update(inAccumNames)
 *         if par is not None:             # <<<<<<<<<<<<<<
 *             self.__parent = wkref(par)
 *         else:
 */
  __pyx_t_8 = (__pyx_v_par != Py_None);
  __pyx_t_9 = (__pyx_t_8 != 0);
  if (__pyx_t_9) {

    /* "Aspidites/_vendor/pyparsing_extension.py":849
 *         self.__accumNames.update(inAccumNames)
 *         if par is not None:
 *             self.__parent = wkref(par)             # <<<<<<<<<<<<<<
 *         else:
 *             self.__parent = None
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_wkref); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 849, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_par) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_par);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 849, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__parent, __pyx_t_1) < 0) __PYX_ERR(0, 849, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":848
 *         self.__accumNames = {}
 *         self.__accumNames.update(inAccumNames)
 *         if par is not None:             # <<<<<<<<<<<<<<
 *             self.__parent = wkref(par)
 *         else:
 */
    goto __pyx_L5;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":851
 *             self.__parent = wkref(par)
 *         else:
 *             self.__parent = None             # <<<<<<<<<<<<<<
 * 
 *     def __getnewargs__(self):
 */
  /*else*/ {
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__parent, Py_None) < 0) __PYX_ERR(0, 851, __pyx_L1_error)
  }
  __pyx_L5:;

  /* "Aspidites/_vendor/pyparsing_extension.py":843
 *                  self.__name))
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self.__toklist = state[0]
 *         self.__tokdict, par, inAccumNames, self.__name = state[1]
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.__setstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_par);
  __Pyx_XDECREF(__pyx_v_inAccumNames);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":853
 *             self.__parent = None
 * 
 *     def __getnewargs__(self):             # <<<<<<<<<<<<<<
 *         return self.__toklist, self.__name, self.__asList, self.__modal
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_81__getnewargs__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_81__getnewargs__ = {"__getnewargs__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_81__getnewargs__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_81__getnewargs__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getnewargs__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_80__getnewargs__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_80__getnewargs__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getnewargs__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":854
 * 
 *     def __getnewargs__(self):
 *         return self.__toklist, self.__name, self.__asList, self.__modal             # <<<<<<<<<<<<<<
 * 
 *     def __dir__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__toklist); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 854, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 854, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__asList); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 854, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ParseResults__modal); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 854, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 854, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_4);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":853
 *             self.__parent = None
 * 
 *     def __getnewargs__(self):             # <<<<<<<<<<<<<<
 *         return self.__toklist, self.__name, self.__asList, self.__modal
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.__getnewargs__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":856
 *         return self.__toklist, self.__name, self.__asList, self.__modal
 * 
 *     def __dir__(self):             # <<<<<<<<<<<<<<
 *         return dir(type(self)) + list(self.keys())
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_83__dir__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_83__dir__ = {"__dir__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_83__dir__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_83__dir__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dir__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_82__dir__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_82__dir__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__dir__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":857
 * 
 *     def __dir__(self):
 *         return dir(type(self)) + list(self.keys())             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyObject_Dir(((PyObject *)Py_TYPE(__pyx_v_self))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 857, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_keys); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 857, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 857, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PySequence_List(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 857, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_Add(__pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 857, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":856
 *         return self.__toklist, self.__name, self.__asList, self.__modal
 * 
 *     def __dir__(self):             # <<<<<<<<<<<<<<
 *         return dir(type(self)) + list(self.keys())
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.__dir__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":860
 * 
 *     @classmethod
 *     def from_dict(cls, other, name=None):             # <<<<<<<<<<<<<<
 *         """
 *         Helper classmethod to construct a ParseResults from a dict, preserving the
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_85from_dict(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_84from_dict[] = "\n        Helper classmethod to construct a ParseResults from a dict, preserving the\n        name-value relations as results names. If an optional 'name' argument is\n        given, a nested ParseResults will be returned\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_85from_dict = {"from_dict", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_85from_dict, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_84from_dict};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_85from_dict(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_cls = 0;
  PyObject *__pyx_v_other = 0;
  PyObject *__pyx_v_name = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("from_dict (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_cls,&__pyx_n_s_other,&__pyx_n_s_name,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cls)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_other)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("from_dict", 0, 2, 3, 1); __PYX_ERR(0, 860, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "from_dict") < 0)) __PYX_ERR(0, 860, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_cls = values[0];
    __pyx_v_other = values[1];
    __pyx_v_name = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("from_dict", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 860, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.from_dict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_84from_dict(__pyx_self, __pyx_v_cls, __pyx_v_other, __pyx_v_name);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":867
 *         """
 * 
 *         def is_iterable(obj):             # <<<<<<<<<<<<<<
 *             try:
 *                 iter(obj)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_9from_dict_1is_iterable(PyObject *__pyx_self, PyObject *__pyx_v_obj); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_9from_dict_1is_iterable = {"is_iterable", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_9from_dict_1is_iterable, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_9from_dict_1is_iterable(PyObject *__pyx_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_iterable (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_9from_dict_is_iterable(__pyx_self, ((PyObject *)__pyx_v_obj));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_9from_dict_is_iterable(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_iterable", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":868
 * 
 *         def is_iterable(obj):
 *             try:             # <<<<<<<<<<<<<<
 *                 iter(obj)
 *             except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "Aspidites/_vendor/pyparsing_extension.py":869
 *         def is_iterable(obj):
 *             try:
 *                 iter(obj)             # <<<<<<<<<<<<<<
 *             except Exception:
 *                 return False
 */
      __pyx_t_4 = PyObject_GetIter(__pyx_v_obj); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 869, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":868
 * 
 *         def is_iterable(obj):
 *             try:             # <<<<<<<<<<<<<<
 *                 iter(obj)
 *             except Exception:
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":873
 *                 return False
 *             else:
 *                 if PY_3:             # <<<<<<<<<<<<<<
 *                     return not isinstance(obj, (str, bytes))
 *                 else:
 */
    /*else:*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_PY_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 873, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 873, __pyx_L5_except_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_5) {

        /* "Aspidites/_vendor/pyparsing_extension.py":874
 *             else:
 *                 if PY_3:
 *                     return not isinstance(obj, (str, bytes))             # <<<<<<<<<<<<<<
 *                 else:
 *                     return not isinstance(obj, str)
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_6 = PyString_Check(__pyx_v_obj); 
        __pyx_t_7 = (__pyx_t_6 != 0);
        if (!__pyx_t_7) {
        } else {
          __pyx_t_5 = __pyx_t_7;
          goto __pyx_L10_bool_binop_done;
        }
        __pyx_t_7 = PyBytes_Check(__pyx_v_obj); 
        __pyx_t_6 = (__pyx_t_7 != 0);
        __pyx_t_5 = __pyx_t_6;
        __pyx_L10_bool_binop_done:;
        __pyx_t_4 = __Pyx_PyBool_FromLong((!(__pyx_t_5 != 0))); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 874, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_r = __pyx_t_4;
        __pyx_t_4 = 0;
        goto __pyx_L6_except_return;

        /* "Aspidites/_vendor/pyparsing_extension.py":873
 *                 return False
 *             else:
 *                 if PY_3:             # <<<<<<<<<<<<<<
 *                     return not isinstance(obj, (str, bytes))
 *                 else:
 */
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":876
 *                     return not isinstance(obj, (str, bytes))
 *                 else:
 *                     return not isinstance(obj, str)             # <<<<<<<<<<<<<<
 * 
 *         ret = cls([])
 */
      /*else*/ {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_5 = PyString_Check(__pyx_v_obj); 
        __pyx_t_4 = __Pyx_PyBool_FromLong((!(__pyx_t_5 != 0))); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 876, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_r = __pyx_t_4;
        __pyx_t_4 = 0;
        goto __pyx_L6_except_return;
      }
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":870
 *             try:
 *                 iter(obj)
 *             except Exception:             # <<<<<<<<<<<<<<
 *                 return False
 *             else:
 */
    __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_8) {
      __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.from_dict.is_iterable", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_9, &__pyx_t_10) < 0) __PYX_ERR(0, 870, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_GOTREF(__pyx_t_10);

      /* "Aspidites/_vendor/pyparsing_extension.py":871
 *                 iter(obj)
 *             except Exception:
 *                 return False             # <<<<<<<<<<<<<<
 *             else:
 *                 if PY_3:
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L6_except_return;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "Aspidites/_vendor/pyparsing_extension.py":868
 * 
 *         def is_iterable(obj):
 *             try:             # <<<<<<<<<<<<<<
 *                 iter(obj)
 *             except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":867
 *         """
 * 
 *         def is_iterable(obj):             # <<<<<<<<<<<<<<
 *             try:
 *                 iter(obj)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.from_dict.is_iterable", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":860
 * 
 *     @classmethod
 *     def from_dict(cls, other, name=None):             # <<<<<<<<<<<<<<
 *         """
 *         Helper classmethod to construct a ParseResults from a dict, preserving the
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_84from_dict(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_cls, PyObject *__pyx_v_other, PyObject *__pyx_v_name) {
  PyObject *__pyx_v_is_iterable = 0;
  PyObject *__pyx_v_ret = NULL;
  PyObject *__pyx_v_k = NULL;
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("from_dict", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":867
 *         """
 * 
 *         def is_iterable(obj):             # <<<<<<<<<<<<<<
 *             try:
 *                 iter(obj)
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_9from_dict_1is_iterable, 0, __pyx_n_s_ParseResults_from_dict_locals_is, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__24)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 867, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_is_iterable = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":878
 *                     return not isinstance(obj, str)
 * 
 *         ret = cls([])             # <<<<<<<<<<<<<<
 *         for k, v in other.items():
 *             if isinstance(v, Mapping):
 */
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 878, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_cls);
  __pyx_t_3 = __pyx_v_cls; __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 878, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_ret = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":879
 * 
 *         ret = cls([])
 *         for k, v in other.items():             # <<<<<<<<<<<<<<
 *             if isinstance(v, Mapping):
 *                 ret += cls.from_dict(v, name=k)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_items); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_3 = __pyx_t_1; __Pyx_INCREF(__pyx_t_3); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 879, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 879, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 879, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 879, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 879, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 879, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_6(__pyx_t_3);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 879, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
      PyObject* sequence = __pyx_t_1;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 879, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 879, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 879, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_7 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 879, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_8 = Py_TYPE(__pyx_t_7)->tp_iternext;
      index = 0; __pyx_t_2 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_2)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_2);
      index = 1; __pyx_t_4 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_4)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 2) < 0) __PYX_ERR(0, 879, __pyx_L1_error)
      __pyx_t_8 = NULL;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 879, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_k, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":880
 *         ret = cls([])
 *         for k, v in other.items():
 *             if isinstance(v, Mapping):             # <<<<<<<<<<<<<<
 *                 ret += cls.from_dict(v, name=k)
 *             else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Mapping); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 880, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = PyObject_IsInstance(__pyx_v_v, __pyx_t_1); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 880, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_10 = (__pyx_t_9 != 0);
    if (__pyx_t_10) {

      /* "Aspidites/_vendor/pyparsing_extension.py":881
 *         for k, v in other.items():
 *             if isinstance(v, Mapping):
 *                 ret += cls.from_dict(v, name=k)             # <<<<<<<<<<<<<<
 *             else:
 *                 ret += cls([v], name=k, asList=is_iterable(v))
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_cls, __pyx_n_s_from_dict); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 881, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 881, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_v_v);
      __Pyx_GIVEREF(__pyx_v_v);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_v);
      __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 881, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_name, __pyx_v_k) < 0) __PYX_ERR(0, 881, __pyx_L1_error)
      __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 881, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_ret, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 881, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF_SET(__pyx_v_ret, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":880
 *         ret = cls([])
 *         for k, v in other.items():
 *             if isinstance(v, Mapping):             # <<<<<<<<<<<<<<
 *                 ret += cls.from_dict(v, name=k)
 *             else:
 */
      goto __pyx_L7;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":883
 *                 ret += cls.from_dict(v, name=k)
 *             else:
 *                 ret += cls([v], name=k, asList=is_iterable(v))             # <<<<<<<<<<<<<<
 *         if name is not None:
 *             ret = cls([ret], name=name)
 */
    /*else*/ {
      __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 883, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_v_v);
      __Pyx_GIVEREF(__pyx_v_v);
      PyList_SET_ITEM(__pyx_t_2, 0, __pyx_v_v);
      __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 883, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 883, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_name, __pyx_v_k) < 0) __PYX_ERR(0, 883, __pyx_L1_error)
      __pyx_t_4 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_9from_dict_is_iterable(__pyx_v_is_iterable, __pyx_v_v); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 883, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_asList, __pyx_t_4) < 0) __PYX_ERR(0, 883, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_v_cls, __pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 883, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_ret, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 883, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF_SET(__pyx_v_ret, __pyx_t_2);
      __pyx_t_2 = 0;
    }
    __pyx_L7:;

    /* "Aspidites/_vendor/pyparsing_extension.py":879
 * 
 *         ret = cls([])
 *         for k, v in other.items():             # <<<<<<<<<<<<<<
 *             if isinstance(v, Mapping):
 *                 ret += cls.from_dict(v, name=k)
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":884
 *             else:
 *                 ret += cls([v], name=k, asList=is_iterable(v))
 *         if name is not None:             # <<<<<<<<<<<<<<
 *             ret = cls([ret], name=name)
 *         return ret
 */
  __pyx_t_10 = (__pyx_v_name != Py_None);
  __pyx_t_9 = (__pyx_t_10 != 0);
  if (__pyx_t_9) {

    /* "Aspidites/_vendor/pyparsing_extension.py":885
 *                 ret += cls([v], name=k, asList=is_iterable(v))
 *         if name is not None:
 *             ret = cls([ret], name=name)             # <<<<<<<<<<<<<<
 *         return ret
 * 
 */
    __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 885, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_ret);
    __Pyx_GIVEREF(__pyx_v_ret);
    PyList_SET_ITEM(__pyx_t_3, 0, __pyx_v_ret);
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 885, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 885, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_name, __pyx_v_name) < 0) __PYX_ERR(0, 885, __pyx_L1_error)
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_v_cls, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 885, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_ret, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":884
 *             else:
 *                 ret += cls([v], name=k, asList=is_iterable(v))
 *         if name is not None:             # <<<<<<<<<<<<<<
 *             ret = cls([ret], name=name)
 *         return ret
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":886
 *         if name is not None:
 *             ret = cls([ret], name=name)
 *         return ret             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_ret);
  __pyx_r = __pyx_v_ret;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":860
 * 
 *     @classmethod
 *     def from_dict(cls, other, name=None):             # <<<<<<<<<<<<<<
 *         """
 *         Helper classmethod to construct a ParseResults from a dict, preserving the
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseResults.from_dict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_is_iterable);
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":894
 *     # Performance tuning: we construct a *lot* of these, so keep this
 *     # constructor as small and fast as possible
 *     def __init__(self, pstr, loc=0, msg=None, elem=None):             # <<<<<<<<<<<<<<
 *         self.loc = loc
 *         if msg is None:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_pstr = 0;
  PyObject *__pyx_v_loc = 0;
  PyObject *__pyx_v_msg = 0;
  PyObject *__pyx_v_elem = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_pstr,&__pyx_n_s_loc,&__pyx_n_s_msg,&__pyx_n_s_elem,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)__pyx_int_0));
    values[3] = ((PyObject *)((PyObject *)Py_None));
    values[4] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pstr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 5, 1); __PYX_ERR(0, 894, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_msg);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_elem);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 894, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_pstr = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_msg = values[3];
    __pyx_v_elem = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 894, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseBaseException.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException___init__(__pyx_self, __pyx_v_self, __pyx_v_pstr, __pyx_v_loc, __pyx_v_msg, __pyx_v_elem);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_pstr, PyObject *__pyx_v_loc, PyObject *__pyx_v_msg, PyObject *__pyx_v_elem) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":895
 *     # constructor as small and fast as possible
 *     def __init__(self, pstr, loc=0, msg=None, elem=None):
 *         self.loc = loc             # <<<<<<<<<<<<<<
 *         if msg is None:
 *             self.msg = pstr
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_loc, __pyx_v_loc) < 0) __PYX_ERR(0, 895, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":896
 *     def __init__(self, pstr, loc=0, msg=None, elem=None):
 *         self.loc = loc
 *         if msg is None:             # <<<<<<<<<<<<<<
 *             self.msg = pstr
 *             self.pstr = ""
 */
  __pyx_t_1 = (__pyx_v_msg == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":897
 *         self.loc = loc
 *         if msg is None:
 *             self.msg = pstr             # <<<<<<<<<<<<<<
 *             self.pstr = ""
 *         else:
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_msg, __pyx_v_pstr) < 0) __PYX_ERR(0, 897, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":898
 *         if msg is None:
 *             self.msg = pstr
 *             self.pstr = ""             # <<<<<<<<<<<<<<
 *         else:
 *             self.msg = msg
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_pstr, __pyx_kp_s__12) < 0) __PYX_ERR(0, 898, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":896
 *     def __init__(self, pstr, loc=0, msg=None, elem=None):
 *         self.loc = loc
 *         if msg is None:             # <<<<<<<<<<<<<<
 *             self.msg = pstr
 *             self.pstr = ""
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":900
 *             self.pstr = ""
 *         else:
 *             self.msg = msg             # <<<<<<<<<<<<<<
 *             self.pstr = pstr
 *         self.parserElement = elem
 */
  /*else*/ {
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_msg, __pyx_v_msg) < 0) __PYX_ERR(0, 900, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":901
 *         else:
 *             self.msg = msg
 *             self.pstr = pstr             # <<<<<<<<<<<<<<
 *         self.parserElement = elem
 *         self.args = (pstr, loc, msg)
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_pstr, __pyx_v_pstr) < 0) __PYX_ERR(0, 901, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "Aspidites/_vendor/pyparsing_extension.py":902
 *             self.msg = msg
 *             self.pstr = pstr
 *         self.parserElement = elem             # <<<<<<<<<<<<<<
 *         self.args = (pstr, loc, msg)
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parserElement, __pyx_v_elem) < 0) __PYX_ERR(0, 902, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":903
 *             self.pstr = pstr
 *         self.parserElement = elem
 *         self.args = (pstr, loc, msg)             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 903, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_pstr);
  __Pyx_GIVEREF(__pyx_v_pstr);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_pstr);
  __Pyx_INCREF(__pyx_v_loc);
  __Pyx_GIVEREF(__pyx_v_loc);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_loc);
  __Pyx_INCREF(__pyx_v_msg);
  __Pyx_GIVEREF(__pyx_v_msg);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_msg);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_args, __pyx_t_3) < 0) __PYX_ERR(0, 903, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":894
 *     # Performance tuning: we construct a *lot* of these, so keep this
 *     # constructor as small and fast as possible
 *     def __init__(self, pstr, loc=0, msg=None, elem=None):             # <<<<<<<<<<<<<<
 *         self.loc = loc
 *         if msg is None:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseBaseException.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":906
 * 
 *     @classmethod
 *     def _from_exception(cls, pe):             # <<<<<<<<<<<<<<
 *         """
 *         internal factory method to simplify creating one type of ParseException
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_3_from_exception(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_2_from_exception[] = "\n        internal factory method to simplify creating one type of ParseException\n        from another - avoids having __init__ signature conflicts among subclasses\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_3_from_exception = {"_from_exception", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_3_from_exception, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_2_from_exception};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_3_from_exception(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_cls = 0;
  PyObject *__pyx_v_pe = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_from_exception (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_cls,&__pyx_n_s_pe,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cls)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pe)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_from_exception", 1, 2, 2, 1); __PYX_ERR(0, 906, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_from_exception") < 0)) __PYX_ERR(0, 906, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_cls = values[0];
    __pyx_v_pe = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_from_exception", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 906, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseBaseException._from_exception", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_2_from_exception(__pyx_self, __pyx_v_cls, __pyx_v_pe);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_2_from_exception(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_cls, PyObject *__pyx_v_pe) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_from_exception", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":911
 *         from another - avoids having __init__ signature conflicts among subclasses
 *         """
 *         return cls(pe.pstr, pe.loc, pe.msg, pe.parserElement)             # <<<<<<<<<<<<<<
 * 
 *     def __getattr__(self, aname):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_pe, __pyx_n_s_pstr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 911, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_pe, __pyx_n_s_loc); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 911, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_pe, __pyx_n_s_msg); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 911, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_pe, __pyx_n_s_parserElement); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 911, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_cls);
  __pyx_t_6 = __pyx_v_cls; __pyx_t_7 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
      __pyx_t_8 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[5] = {__pyx_t_7, __pyx_t_2, __pyx_t_3, __pyx_t_4, __pyx_t_5};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_8, 4+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 911, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[5] = {__pyx_t_7, __pyx_t_2, __pyx_t_3, __pyx_t_4, __pyx_t_5};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_8, 4+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 911, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else
  #endif
  {
    __pyx_t_9 = PyTuple_New(4+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 911, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (__pyx_t_7) {
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_9, 2+__pyx_t_8, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_9, 3+__pyx_t_8, __pyx_t_5);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 911, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":906
 * 
 *     @classmethod
 *     def _from_exception(cls, pe):             # <<<<<<<<<<<<<<
 *         """
 *         internal factory method to simplify creating one type of ParseException
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseBaseException._from_exception", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":913
 *         return cls(pe.pstr, pe.loc, pe.msg, pe.parserElement)
 * 
 *     def __getattr__(self, aname):             # <<<<<<<<<<<<<<
 *         """supported attributes by name are:
 *            - lineno - returns the line number of the exception text
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_5__getattr__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_4__getattr__[] = "supported attributes by name are:\n           - lineno - returns the line number of the exception text\n           - col - returns the column number of the exception text\n           - line - returns the line containing the exception text\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_5__getattr__ = {"__getattr__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_5__getattr__, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_4__getattr__};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_5__getattr__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_aname = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getattr__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_aname,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_aname)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__getattr__", 1, 2, 2, 1); __PYX_ERR(0, 913, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__getattr__") < 0)) __PYX_ERR(0, 913, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_aname = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__getattr__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 913, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseBaseException.__getattr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_4__getattr__(__pyx_self, __pyx_v_self, __pyx_v_aname);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_4__getattr__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_aname) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getattr__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":919
 *            - line - returns the line containing the exception text
 *         """
 *         if aname == "lineno":             # <<<<<<<<<<<<<<
 *             return lineno(self.loc, self.pstr)
 *         elif aname in ("col", "column"):
 */
  __pyx_t_1 = (__Pyx_PyString_Equals(__pyx_v_aname, __pyx_n_s_lineno, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 919, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "Aspidites/_vendor/pyparsing_extension.py":920
 *         """
 *         if aname == "lineno":
 *             return lineno(self.loc, self.pstr)             # <<<<<<<<<<<<<<
 *         elif aname in ("col", "column"):
 *             return col(self.loc, self.pstr)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_lineno); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 920, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_loc); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 920, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_pstr); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 920, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_4, __pyx_t_5};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 920, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_4, __pyx_t_5};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 920, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 920, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_5);
      __pyx_t_4 = 0;
      __pyx_t_5 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 920, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":919
 *            - line - returns the line containing the exception text
 *         """
 *         if aname == "lineno":             # <<<<<<<<<<<<<<
 *             return lineno(self.loc, self.pstr)
 *         elif aname in ("col", "column"):
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":921
 *         if aname == "lineno":
 *             return lineno(self.loc, self.pstr)
 *         elif aname in ("col", "column"):             # <<<<<<<<<<<<<<
 *             return col(self.loc, self.pstr)
 *         elif aname == "line":
 */
  __Pyx_INCREF(__pyx_v_aname);
  __pyx_t_2 = __pyx_v_aname;
  __pyx_t_9 = (__Pyx_PyString_Equals(__pyx_t_2, __pyx_n_s_col, Py_EQ)); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 921, __pyx_L1_error)
  if (!__pyx_t_9) {
  } else {
    __pyx_t_1 = __pyx_t_9;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_9 = (__Pyx_PyString_Equals(__pyx_t_2, __pyx_n_s_column, Py_EQ)); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 921, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_9;
  __pyx_L4_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_9 = (__pyx_t_1 != 0);
  if (__pyx_t_9) {

    /* "Aspidites/_vendor/pyparsing_extension.py":922
 *             return lineno(self.loc, self.pstr)
 *         elif aname in ("col", "column"):
 *             return col(self.loc, self.pstr)             # <<<<<<<<<<<<<<
 *         elif aname == "line":
 *             return line(self.loc, self.pstr)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_col); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 922, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_loc); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 922, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_pstr); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 922, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_8, __pyx_t_5};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 922, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_8, __pyx_t_5};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 922, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 922, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_4) {
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_7, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_7, __pyx_t_5);
      __pyx_t_8 = 0;
      __pyx_t_5 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 922, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":921
 *         if aname == "lineno":
 *             return lineno(self.loc, self.pstr)
 *         elif aname in ("col", "column"):             # <<<<<<<<<<<<<<
 *             return col(self.loc, self.pstr)
 *         elif aname == "line":
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":923
 *         elif aname in ("col", "column"):
 *             return col(self.loc, self.pstr)
 *         elif aname == "line":             # <<<<<<<<<<<<<<
 *             return line(self.loc, self.pstr)
 *         else:
 */
  __pyx_t_9 = (__Pyx_PyString_Equals(__pyx_v_aname, __pyx_n_s_line, Py_EQ)); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 923, __pyx_L1_error)
  if (likely(__pyx_t_9)) {

    /* "Aspidites/_vendor/pyparsing_extension.py":924
 *             return col(self.loc, self.pstr)
 *         elif aname == "line":
 *             return line(self.loc, self.pstr)             # <<<<<<<<<<<<<<
 *         else:
 *             raise AttributeError(aname)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_line); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 924, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_loc); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 924, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_pstr); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 924, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_8 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_5};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 924, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_5};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 924, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 924, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_7, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_7, __pyx_t_5);
      __pyx_t_6 = 0;
      __pyx_t_5 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 924, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":923
 *         elif aname in ("col", "column"):
 *             return col(self.loc, self.pstr)
 *         elif aname == "line":             # <<<<<<<<<<<<<<
 *             return line(self.loc, self.pstr)
 *         else:
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":926
 *             return line(self.loc, self.pstr)
 *         else:
 *             raise AttributeError(aname)             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  /*else*/ {
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_AttributeError, __pyx_v_aname); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 926, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 926, __pyx_L1_error)
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":913
 *         return cls(pe.pstr, pe.loc, pe.msg, pe.parserElement)
 * 
 *     def __getattr__(self, aname):             # <<<<<<<<<<<<<<
 *         """supported attributes by name are:
 *            - lineno - returns the line number of the exception text
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseBaseException.__getattr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":928
 *             raise AttributeError(aname)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if self.pstr:
 *             if self.loc >= len(self.pstr):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_7__str__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_7__str__ = {"__str__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_7__str__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_7__str__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_6__str__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_6__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_foundstr = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":929
 * 
 *     def __str__(self):
 *         if self.pstr:             # <<<<<<<<<<<<<<
 *             if self.loc >= len(self.pstr):
 *                 foundstr = ', found end of text'
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_pstr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 929, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 929, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":930
 *     def __str__(self):
 *         if self.pstr:
 *             if self.loc >= len(self.pstr):             # <<<<<<<<<<<<<<
 *                 foundstr = ', found end of text'
 *             else:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_loc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 930, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_pstr); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 930, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyObject_Length(__pyx_t_3); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 930, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 930, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = PyObject_RichCompare(__pyx_t_1, __pyx_t_3, Py_GE); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 930, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 930, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_2) {

      /* "Aspidites/_vendor/pyparsing_extension.py":931
 *         if self.pstr:
 *             if self.loc >= len(self.pstr):
 *                 foundstr = ', found end of text'             # <<<<<<<<<<<<<<
 *             else:
 *                 foundstr = (', found %r' % self.pstr[self.loc:self.loc + 1]).replace(r'\\', '\\')
 */
      __Pyx_INCREF(__pyx_kp_s_found_end_of_text);
      __pyx_v_foundstr = __pyx_kp_s_found_end_of_text;

      /* "Aspidites/_vendor/pyparsing_extension.py":930
 *     def __str__(self):
 *         if self.pstr:
 *             if self.loc >= len(self.pstr):             # <<<<<<<<<<<<<<
 *                 foundstr = ', found end of text'
 *             else:
 */
      goto __pyx_L4;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":933
 *                 foundstr = ', found end of text'
 *             else:
 *                 foundstr = (', found %r' % self.pstr[self.loc:self.loc + 1]).replace(r'\\', '\\')             # <<<<<<<<<<<<<<
 *         else:
 *             foundstr = ''
 */
    /*else*/ {
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_pstr); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 933, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_loc); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 933, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_loc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 933, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyInt_AddObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 933, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_t_5, 0, 0, &__pyx_t_3, &__pyx_t_6, NULL, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 933, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyString_FormatSafe(__pyx_kp_s_found_r, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 933, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_replace); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 933, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__27, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 933, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_foundstr = __pyx_t_6;
      __pyx_t_6 = 0;
    }
    __pyx_L4:;

    /* "Aspidites/_vendor/pyparsing_extension.py":929
 * 
 *     def __str__(self):
 *         if self.pstr:             # <<<<<<<<<<<<<<
 *             if self.loc >= len(self.pstr):
 *                 foundstr = ', found end of text'
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":935
 *                 foundstr = (', found %r' % self.pstr[self.loc:self.loc + 1]).replace(r'\\', '\\')
 *         else:
 *             foundstr = ''             # <<<<<<<<<<<<<<
 *         return ("%s%s  (at char %d), (line:%d, col:%d)" %
 *                 (self.msg, foundstr, self.loc, self.lineno, self.column))
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_kp_s__12);
    __pyx_v_foundstr = __pyx_kp_s__12;
  }
  __pyx_L3:;

  /* "Aspidites/_vendor/pyparsing_extension.py":936
 *         else:
 *             foundstr = ''
 *         return ("%s%s  (at char %d), (line:%d, col:%d)" %             # <<<<<<<<<<<<<<
 *                 (self.msg, foundstr, self.loc, self.lineno, self.column))
 * 
 */
  __Pyx_XDECREF(__pyx_r);

  /* "Aspidites/_vendor/pyparsing_extension.py":937
 *             foundstr = ''
 *         return ("%s%s  (at char %d), (line:%d, col:%d)" %
 *                 (self.msg, foundstr, self.loc, self.lineno, self.column))             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_msg); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 937, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_loc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 937, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_lineno); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 937, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_column); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 937, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_New(5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 937, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_6);
  __Pyx_INCREF(__pyx_v_foundstr);
  __Pyx_GIVEREF(__pyx_v_foundstr);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_v_foundstr);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 4, __pyx_t_5);
  __pyx_t_6 = 0;
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;
  __pyx_t_5 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":936
 *         else:
 *             foundstr = ''
 *         return ("%s%s  (at char %d), (line:%d, col:%d)" %             # <<<<<<<<<<<<<<
 *                 (self.msg, foundstr, self.loc, self.lineno, self.column))
 * 
 */
  __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_s_s_at_char_d_line_d_col_d, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 936, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":928
 *             raise AttributeError(aname)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if self.pstr:
 *             if self.loc >= len(self.pstr):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseBaseException.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_foundstr);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":939
 *                 (self.msg, foundstr, self.loc, self.lineno, self.column))
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return str(self)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_9__repr__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_9__repr__ = {"__repr__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_9__repr__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_9__repr__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_8__repr__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_8__repr__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":940
 * 
 *     def __repr__(self):
 *         return str(self)             # <<<<<<<<<<<<<<
 * 
 *     def markInputline(self, markerString=">!<"):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 940, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":939
 *                 (self.msg, foundstr, self.loc, self.lineno, self.column))
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return str(self)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseBaseException.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":942
 *         return str(self)
 * 
 *     def markInputline(self, markerString=">!<"):             # <<<<<<<<<<<<<<
 *         """Extracts the exception line from the input string, and marks
 *            the location of the exception with a special symbol.
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_11markInputline(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_10markInputline[] = "Extracts the exception line from the input string, and marks\n           the location of the exception with a special symbol.\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_11markInputline = {"markInputline", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_11markInputline, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_10markInputline};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_11markInputline(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_markerString = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("markInputline (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_markerString,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject*)__pyx_kp_s__28));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_markerString);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "markInputline") < 0)) __PYX_ERR(0, 942, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_markerString = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("markInputline", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 942, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseBaseException.markInputline", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_10markInputline(__pyx_self, __pyx_v_self, __pyx_v_markerString);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_10markInputline(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_markerString) {
  PyObject *__pyx_v_line_str = NULL;
  PyObject *__pyx_v_line_column = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("markInputline", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":946
 *            the location of the exception with a special symbol.
 *         """
 *         line_str = self.line             # <<<<<<<<<<<<<<
 *         line_column = self.column - 1
 *         if markerString:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_line); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_line_str = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":947
 *         """
 *         line_str = self.line
 *         line_column = self.column - 1             # <<<<<<<<<<<<<<
 *         if markerString:
 *             line_str = "".join((line_str[:line_column],
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_column); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 947, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 947, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_line_column = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":948
 *         line_str = self.line
 *         line_column = self.column - 1
 *         if markerString:             # <<<<<<<<<<<<<<
 *             line_str = "".join((line_str[:line_column],
 *                                 markerString, line_str[line_column:]))
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_markerString); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 948, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/pyparsing_extension.py":949
 *         line_column = self.column - 1
 *         if markerString:
 *             line_str = "".join((line_str[:line_column],             # <<<<<<<<<<<<<<
 *                                 markerString, line_str[line_column:]))
 *         return line_str.strip()
 */
    __pyx_t_2 = __Pyx_PyObject_GetSlice(__pyx_v_line_str, 0, 0, NULL, &__pyx_v_line_column, NULL, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 949, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "Aspidites/_vendor/pyparsing_extension.py":950
 *         if markerString:
 *             line_str = "".join((line_str[:line_column],
 *                                 markerString, line_str[line_column:]))             # <<<<<<<<<<<<<<
 *         return line_str.strip()
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_line_str, 0, 0, &__pyx_v_line_column, NULL, NULL, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 950, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "Aspidites/_vendor/pyparsing_extension.py":949
 *         line_column = self.column - 1
 *         if markerString:
 *             line_str = "".join((line_str[:line_column],             # <<<<<<<<<<<<<<
 *                                 markerString, line_str[line_column:]))
 *         return line_str.strip()
 */
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 949, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
    __Pyx_INCREF(__pyx_v_markerString);
    __Pyx_GIVEREF(__pyx_v_markerString);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_markerString);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_1);
    __pyx_t_2 = 0;
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyString_Join(__pyx_kp_s__12, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 949, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_line_str, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":948
 *         line_str = self.line
 *         line_column = self.column - 1
 *         if markerString:             # <<<<<<<<<<<<<<
 *             line_str = "".join((line_str[:line_column],
 *                                 markerString, line_str[line_column:]))
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":951
 *             line_str = "".join((line_str[:line_column],
 *                                 markerString, line_str[line_column:]))
 *         return line_str.strip()             # <<<<<<<<<<<<<<
 * 
 *     def __dir__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_line_str, __pyx_n_s_strip); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 951, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 951, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":942
 *         return str(self)
 * 
 *     def markInputline(self, markerString=">!<"):             # <<<<<<<<<<<<<<
 *         """Extracts the exception line from the input string, and marks
 *            the location of the exception with a special symbol.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseBaseException.markInputline", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_line_str);
  __Pyx_XDECREF(__pyx_v_line_column);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":953
 *         return line_str.strip()
 * 
 *     def __dir__(self):             # <<<<<<<<<<<<<<
 *         return "lineno col line".split() + dir(type(self))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_13__dir__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_13__dir__ = {"__dir__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_13__dir__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_13__dir__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dir__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_12__dir__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_12__dir__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__dir__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":954
 * 
 *     def __dir__(self):
 *         return "lineno col line".split() + dir(type(self))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_lineno_col_line, __pyx_n_s_split); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 954, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 954, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyObject_Dir(((PyObject *)Py_TYPE(__pyx_v_self))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 954, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Add(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 954, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":953
 *         return line_str.strip()
 * 
 *     def __dir__(self):             # <<<<<<<<<<<<<<
 *         return "lineno col line".split() + dir(type(self))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseBaseException.__dir__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":981
 * 
 *     @staticmethod
 *     def explain(exc, depth=16):             # <<<<<<<<<<<<<<
 *         """
 *         Method to take an exception and translate the Python internal traceback into a list
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_14ParseException_1explain(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_14ParseException_explain[] = "\n        Method to take an exception and translate the Python internal traceback into a list\n        of the pyparsing expressions that caused the exception to be raised.\n\n        Parameters:\n\n         - exc - exception raised during parsing (need not be a ParseException, in support\n           of Python exceptions that might be raised in a parse action)\n         - depth (default=16) - number of levels back in the stack trace to list expression\n           and function names; if None, the full stack trace names will be listed; if 0, only\n           the failing input line, marker, and exception string will be shown\n\n        Returns a multi-line string listing the ParserElements and/or function names in the\n        exception's stack trace.\n\n        Note: the diagnostic output will include string representations of the expressions\n        that failed to parse. These representations will be more helpful if you use `setName` to\n        give identifiable names to your expressions. Otherwise they will use the default string\n        forms, which may be cryptic to read.\n\n        explain() is only supported under Python 3.\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_14ParseException_1explain = {"explain", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_14ParseException_1explain, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_14ParseException_explain};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_14ParseException_1explain(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_exc = 0;
  PyObject *__pyx_v_depth = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("explain (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_exc,&__pyx_n_s_depth_2,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)__pyx_int_16));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_exc)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_depth_2);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "explain") < 0)) __PYX_ERR(0, 981, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_exc = values[0];
    __pyx_v_depth = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("explain", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 981, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseException.explain", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_14ParseException_explain(__pyx_self, __pyx_v_exc, __pyx_v_depth);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_14ParseException_explain(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_exc, PyObject *__pyx_v_depth) {
  PyObject *__pyx_v_inspect = NULL;
  PyObject *__pyx_v_ret = NULL;
  PyObject *__pyx_v_callers = NULL;
  PyObject *__pyx_v_seen = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_ff = NULL;
  PyObject *__pyx_v_frm = NULL;
  PyObject *__pyx_v_f_self = NULL;
  PyTypeObject *__pyx_v_self_type = NULL;
  PyObject *__pyx_v_code = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  Py_ssize_t __pyx_t_10;
  PyObject *(*__pyx_t_11)(PyObject *);
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("explain", 0);
  __Pyx_INCREF(__pyx_v_depth);

  /* "Aspidites/_vendor/pyparsing_extension.py":1004
 *         explain() is only supported under Python 3.
 *         """
 *         import inspect             # <<<<<<<<<<<<<<
 * 
 *         if depth is None:
 */
  __pyx_t_1 = __Pyx_patch_inspect(__Pyx_Import(__pyx_n_s_inspect, 0, -1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1004, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_inspect = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1006
 *         import inspect
 * 
 *         if depth is None:             # <<<<<<<<<<<<<<
 *             depth = sys.getrecursionlimit()
 *         ret = []
 */
  __pyx_t_2 = (__pyx_v_depth == Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/pyparsing_extension.py":1007
 * 
 *         if depth is None:
 *             depth = sys.getrecursionlimit()             # <<<<<<<<<<<<<<
 *         ret = []
 *         if isinstance(exc, ParseBaseException):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_sys); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1007, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_getrecursionlimit); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1007, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1007, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_depth, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1006
 *         import inspect
 * 
 *         if depth is None:             # <<<<<<<<<<<<<<
 *             depth = sys.getrecursionlimit()
 *         ret = []
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1008
 *         if depth is None:
 *             depth = sys.getrecursionlimit()
 *         ret = []             # <<<<<<<<<<<<<<
 *         if isinstance(exc, ParseBaseException):
 *             ret.append(exc.line)
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1008, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ret = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1009
 *             depth = sys.getrecursionlimit()
 *         ret = []
 *         if isinstance(exc, ParseBaseException):             # <<<<<<<<<<<<<<
 *             ret.append(exc.line)
 *             ret.append(' ' * (exc.col - 1) + '^')
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParseBaseException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1009, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_exc, __pyx_t_1); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 1009, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":1010
 *         ret = []
 *         if isinstance(exc, ParseBaseException):
 *             ret.append(exc.line)             # <<<<<<<<<<<<<<
 *             ret.append(' ' * (exc.col - 1) + '^')
 *         ret.append("{0}: {1}".format(type(exc).__name__, exc))
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_exc, __pyx_n_s_line); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1010, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_ret, __pyx_t_1); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1010, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1011
 *         if isinstance(exc, ParseBaseException):
 *             ret.append(exc.line)
 *             ret.append(' ' * (exc.col - 1) + '^')             # <<<<<<<<<<<<<<
 *         ret.append("{0}: {1}".format(type(exc).__name__, exc))
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_exc, __pyx_n_s_col); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1011, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyInt_SubtractObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1011, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_Multiply(__pyx_kp_s__29, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1011, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_Add(__pyx_t_1, __pyx_kp_s__30); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1011, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_ret, __pyx_t_5); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1011, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1009
 *             depth = sys.getrecursionlimit()
 *         ret = []
 *         if isinstance(exc, ParseBaseException):             # <<<<<<<<<<<<<<
 *             ret.append(exc.line)
 *             ret.append(' ' * (exc.col - 1) + '^')
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1012
 *             ret.append(exc.line)
 *             ret.append(' ' * (exc.col - 1) + '^')
 *         ret.append("{0}: {1}".format(type(exc).__name__, exc))             # <<<<<<<<<<<<<<
 * 
 *         if depth > 0:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_0_1, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1012, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(__pyx_v_exc)), __pyx_n_s_name_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1012, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_8 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_4, __pyx_v_exc};
    __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1012, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_4, __pyx_v_exc};
    __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1012, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_9 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1012, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (__pyx_t_7) {
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_t_4);
    __Pyx_INCREF(__pyx_v_exc);
    __Pyx_GIVEREF(__pyx_v_exc);
    PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_v_exc);
    __pyx_t_4 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_9, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1012, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_ret, __pyx_t_5); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1012, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1014
 *         ret.append("{0}: {1}".format(type(exc).__name__, exc))
 * 
 *         if depth > 0:             # <<<<<<<<<<<<<<
 *             callers = inspect.getinnerframes(exc.__traceback__, context=depth)
 *             seen = set()
 */
  __pyx_t_5 = PyObject_RichCompare(__pyx_v_depth, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1014, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1014, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":1015
 * 
 *         if depth > 0:
 *             callers = inspect.getinnerframes(exc.__traceback__, context=depth)             # <<<<<<<<<<<<<<
 *             seen = set()
 *             for i, ff in enumerate(callers[-depth:]):
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_inspect, __pyx_n_s_getinnerframes); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1015, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_exc, __pyx_n_s_traceback); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1015, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = PyTuple_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1015, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1015, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_context, __pyx_v_depth) < 0) __PYX_ERR(0, 1015, __pyx_L1_error)
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_9, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1015, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_callers = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1016
 *         if depth > 0:
 *             callers = inspect.getinnerframes(exc.__traceback__, context=depth)
 *             seen = set()             # <<<<<<<<<<<<<<
 *             for i, ff in enumerate(callers[-depth:]):
 *                 frm = ff[0]
 */
    __pyx_t_4 = PySet_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1016, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_v_seen = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1017
 *             callers = inspect.getinnerframes(exc.__traceback__, context=depth)
 *             seen = set()
 *             for i, ff in enumerate(callers[-depth:]):             # <<<<<<<<<<<<<<
 *                 frm = ff[0]
 * 
 */
    __Pyx_INCREF(__pyx_int_0);
    __pyx_t_4 = __pyx_int_0;
    __pyx_t_1 = PyNumber_Negative(__pyx_v_depth); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1017, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = __Pyx_PyObject_GetSlice(__pyx_v_callers, 0, 0, &__pyx_t_1, NULL, NULL, 0, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1017, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (likely(PyList_CheckExact(__pyx_t_9)) || PyTuple_CheckExact(__pyx_t_9)) {
      __pyx_t_1 = __pyx_t_9; __Pyx_INCREF(__pyx_t_1); __pyx_t_10 = 0;
      __pyx_t_11 = NULL;
    } else {
      __pyx_t_10 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1017, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_11 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1017, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    for (;;) {
      if (likely(!__pyx_t_11)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_9 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_10); __Pyx_INCREF(__pyx_t_9); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 1017, __pyx_L1_error)
          #else
          __pyx_t_9 = PySequence_ITEM(__pyx_t_1, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1017, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          #endif
        } else {
          if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_9 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_10); __Pyx_INCREF(__pyx_t_9); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 1017, __pyx_L1_error)
          #else
          __pyx_t_9 = PySequence_ITEM(__pyx_t_1, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1017, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          #endif
        }
      } else {
        __pyx_t_9 = __pyx_t_11(__pyx_t_1);
        if (unlikely(!__pyx_t_9)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1017, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_9);
      }
      __Pyx_XDECREF_SET(__pyx_v_ff, __pyx_t_9);
      __pyx_t_9 = 0;
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_4);
      __pyx_t_9 = __Pyx_PyInt_AddObjC(__pyx_t_4, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1017, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_4);
      __pyx_t_4 = __pyx_t_9;
      __pyx_t_9 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":1018
 *             seen = set()
 *             for i, ff in enumerate(callers[-depth:]):
 *                 frm = ff[0]             # <<<<<<<<<<<<<<
 * 
 *                 f_self = frm.f_locals.get('self', None)
 */
      __pyx_t_9 = __Pyx_GetItemInt(__pyx_v_ff, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1018, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_XDECREF_SET(__pyx_v_frm, __pyx_t_9);
      __pyx_t_9 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":1020
 *                 frm = ff[0]
 * 
 *                 f_self = frm.f_locals.get('self', None)             # <<<<<<<<<<<<<<
 *                 if isinstance(f_self, ParserElement):
 *                     if frm.f_code.co_name not in ('parseImpl', '_parseNoCache'):
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_frm, __pyx_n_s_f_locals); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1020, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_get); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1020, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_tuple__31, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1020, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF_SET(__pyx_v_f_self, __pyx_t_9);
      __pyx_t_9 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":1021
 * 
 *                 f_self = frm.f_locals.get('self', None)
 *                 if isinstance(f_self, ParserElement):             # <<<<<<<<<<<<<<
 *                     if frm.f_code.co_name not in ('parseImpl', '_parseNoCache'):
 *                         continue
 */
      __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1021, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_2 = PyObject_IsInstance(__pyx_v_f_self, __pyx_t_9); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 1021, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {

        /* "Aspidites/_vendor/pyparsing_extension.py":1022
 *                 f_self = frm.f_locals.get('self', None)
 *                 if isinstance(f_self, ParserElement):
 *                     if frm.f_code.co_name not in ('parseImpl', '_parseNoCache'):             # <<<<<<<<<<<<<<
 *                         continue
 *                     if f_self in seen:
 */
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_frm, __pyx_n_s_f_code); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1022, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_co_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1022, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_t_5, __pyx_n_s_parseImpl, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1022, __pyx_L1_error)
        if (__pyx_t_2) {
        } else {
          __pyx_t_3 = __pyx_t_2;
          goto __pyx_L10_bool_binop_done;
        }
        __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_t_5, __pyx_n_s_parseNoCache, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1022, __pyx_L1_error)
        __pyx_t_3 = __pyx_t_2;
        __pyx_L10_bool_binop_done:;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_2 = (__pyx_t_3 != 0);
        if (__pyx_t_2) {

          /* "Aspidites/_vendor/pyparsing_extension.py":1023
 *                 if isinstance(f_self, ParserElement):
 *                     if frm.f_code.co_name not in ('parseImpl', '_parseNoCache'):
 *                         continue             # <<<<<<<<<<<<<<
 *                     if f_self in seen:
 *                         continue
 */
          goto __pyx_L6_continue;

          /* "Aspidites/_vendor/pyparsing_extension.py":1022
 *                 f_self = frm.f_locals.get('self', None)
 *                 if isinstance(f_self, ParserElement):
 *                     if frm.f_code.co_name not in ('parseImpl', '_parseNoCache'):             # <<<<<<<<<<<<<<
 *                         continue
 *                     if f_self in seen:
 */
        }

        /* "Aspidites/_vendor/pyparsing_extension.py":1024
 *                     if frm.f_code.co_name not in ('parseImpl', '_parseNoCache'):
 *                         continue
 *                     if f_self in seen:             # <<<<<<<<<<<<<<
 *                         continue
 *                     seen.add(f_self)
 */
        __pyx_t_2 = (__Pyx_PySet_ContainsTF(__pyx_v_f_self, __pyx_v_seen, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1024, __pyx_L1_error)
        __pyx_t_3 = (__pyx_t_2 != 0);
        if (__pyx_t_3) {

          /* "Aspidites/_vendor/pyparsing_extension.py":1025
 *                         continue
 *                     if f_self in seen:
 *                         continue             # <<<<<<<<<<<<<<
 *                     seen.add(f_self)
 * 
 */
          goto __pyx_L6_continue;

          /* "Aspidites/_vendor/pyparsing_extension.py":1024
 *                     if frm.f_code.co_name not in ('parseImpl', '_parseNoCache'):
 *                         continue
 *                     if f_self in seen:             # <<<<<<<<<<<<<<
 *                         continue
 *                     seen.add(f_self)
 */
        }

        /* "Aspidites/_vendor/pyparsing_extension.py":1026
 *                     if f_self in seen:
 *                         continue
 *                     seen.add(f_self)             # <<<<<<<<<<<<<<
 * 
 *                     self_type = type(f_self)
 */
        __pyx_t_6 = PySet_Add(__pyx_v_seen, __pyx_v_f_self); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1026, __pyx_L1_error)

        /* "Aspidites/_vendor/pyparsing_extension.py":1028
 *                     seen.add(f_self)
 * 
 *                     self_type = type(f_self)             # <<<<<<<<<<<<<<
 *                     ret.append("{0}.{1} - {2}".format(self_type.__module__,
 *                                                       self_type.__name__,
 */
        __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_f_self)));
        __Pyx_XDECREF_SET(__pyx_v_self_type, ((PyTypeObject*)((PyObject *)Py_TYPE(__pyx_v_f_self))));

        /* "Aspidites/_vendor/pyparsing_extension.py":1029
 * 
 *                     self_type = type(f_self)
 *                     ret.append("{0}.{1} - {2}".format(self_type.__module__,             # <<<<<<<<<<<<<<
 *                                                       self_type.__name__,
 *                                                       f_self))
 */
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_0_1_2, __pyx_n_s_format); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1029, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self_type), __pyx_n_s_module); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1029, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);

        /* "Aspidites/_vendor/pyparsing_extension.py":1030
 *                     self_type = type(f_self)
 *                     ret.append("{0}.{1} - {2}".format(self_type.__module__,
 *                                                       self_type.__name__,             # <<<<<<<<<<<<<<
 *                                                       f_self))
 *                 elif f_self is not None:
 */
        __pyx_t_12 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self_type), __pyx_n_s_name_2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1030, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);

        /* "Aspidites/_vendor/pyparsing_extension.py":1031
 *                     ret.append("{0}.{1} - {2}".format(self_type.__module__,
 *                                                       self_type.__name__,
 *                                                       f_self))             # <<<<<<<<<<<<<<
 *                 elif f_self is not None:
 *                     self_type = type(f_self)
 */
        __pyx_t_13 = NULL;
        __pyx_t_8 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
          __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_9);
          if (likely(__pyx_t_13)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
            __Pyx_INCREF(__pyx_t_13);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_9, function);
            __pyx_t_8 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_9)) {
          PyObject *__pyx_temp[4] = {__pyx_t_13, __pyx_t_7, __pyx_t_12, __pyx_v_f_self};
          __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_8, 3+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1029, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_9)) {
          PyObject *__pyx_temp[4] = {__pyx_t_13, __pyx_t_7, __pyx_t_12, __pyx_v_f_self};
          __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_8, 3+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1029, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        } else
        #endif
        {
          __pyx_t_14 = PyTuple_New(3+__pyx_t_8); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1029, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_14);
          if (__pyx_t_13) {
            __Pyx_GIVEREF(__pyx_t_13); PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_13); __pyx_t_13 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_7);
          PyTuple_SET_ITEM(__pyx_t_14, 0+__pyx_t_8, __pyx_t_7);
          __Pyx_GIVEREF(__pyx_t_12);
          PyTuple_SET_ITEM(__pyx_t_14, 1+__pyx_t_8, __pyx_t_12);
          __Pyx_INCREF(__pyx_v_f_self);
          __Pyx_GIVEREF(__pyx_v_f_self);
          PyTuple_SET_ITEM(__pyx_t_14, 2+__pyx_t_8, __pyx_v_f_self);
          __pyx_t_7 = 0;
          __pyx_t_12 = 0;
          __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_14, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1029, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        }
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":1029
 * 
 *                     self_type = type(f_self)
 *                     ret.append("{0}.{1} - {2}".format(self_type.__module__,             # <<<<<<<<<<<<<<
 *                                                       self_type.__name__,
 *                                                       f_self))
 */
        __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_ret, __pyx_t_5); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1029, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":1021
 * 
 *                 f_self = frm.f_locals.get('self', None)
 *                 if isinstance(f_self, ParserElement):             # <<<<<<<<<<<<<<
 *                     if frm.f_code.co_name not in ('parseImpl', '_parseNoCache'):
 *                         continue
 */
        goto __pyx_L8;
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":1032
 *                                                       self_type.__name__,
 *                                                       f_self))
 *                 elif f_self is not None:             # <<<<<<<<<<<<<<
 *                     self_type = type(f_self)
 *                     ret.append("{0}.{1}".format(self_type.__module__,
 */
      __pyx_t_3 = (__pyx_v_f_self != Py_None);
      __pyx_t_2 = (__pyx_t_3 != 0);
      if (__pyx_t_2) {

        /* "Aspidites/_vendor/pyparsing_extension.py":1033
 *                                                       f_self))
 *                 elif f_self is not None:
 *                     self_type = type(f_self)             # <<<<<<<<<<<<<<
 *                     ret.append("{0}.{1}".format(self_type.__module__,
 *                                                 self_type.__name__))
 */
        __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_f_self)));
        __Pyx_XDECREF_SET(__pyx_v_self_type, ((PyTypeObject*)((PyObject *)Py_TYPE(__pyx_v_f_self))));

        /* "Aspidites/_vendor/pyparsing_extension.py":1034
 *                 elif f_self is not None:
 *                     self_type = type(f_self)
 *                     ret.append("{0}.{1}".format(self_type.__module__,             # <<<<<<<<<<<<<<
 *                                                 self_type.__name__))
 *                 else:
 */
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_0_1_3, __pyx_n_s_format); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1034, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_14 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self_type), __pyx_n_s_module); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1034, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);

        /* "Aspidites/_vendor/pyparsing_extension.py":1035
 *                     self_type = type(f_self)
 *                     ret.append("{0}.{1}".format(self_type.__module__,
 *                                                 self_type.__name__))             # <<<<<<<<<<<<<<
 *                 else:
 *                     code = frm.f_code
 */
        __pyx_t_12 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self_type), __pyx_n_s_name_2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1035, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_7 = NULL;
        __pyx_t_8 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_9);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_9, function);
            __pyx_t_8 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_9)) {
          PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_14, __pyx_t_12};
          __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1034, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_9)) {
          PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_14, __pyx_t_12};
          __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1034, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        } else
        #endif
        {
          __pyx_t_13 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1034, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_13);
          if (__pyx_t_7) {
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_7); __pyx_t_7 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_14);
          PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_8, __pyx_t_14);
          __Pyx_GIVEREF(__pyx_t_12);
          PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_8, __pyx_t_12);
          __pyx_t_14 = 0;
          __pyx_t_12 = 0;
          __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_13, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1034, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        }
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":1034
 *                 elif f_self is not None:
 *                     self_type = type(f_self)
 *                     ret.append("{0}.{1}".format(self_type.__module__,             # <<<<<<<<<<<<<<
 *                                                 self_type.__name__))
 *                 else:
 */
        __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_ret, __pyx_t_5); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1034, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":1032
 *                                                       self_type.__name__,
 *                                                       f_self))
 *                 elif f_self is not None:             # <<<<<<<<<<<<<<
 *                     self_type = type(f_self)
 *                     ret.append("{0}.{1}".format(self_type.__module__,
 */
        goto __pyx_L8;
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":1037
 *                                                 self_type.__name__))
 *                 else:
 *                     code = frm.f_code             # <<<<<<<<<<<<<<
 *                     if code.co_name in ('wrapper', '<module>'):
 *                         continue
 */
      /*else*/ {
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_frm, __pyx_n_s_f_code); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1037, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_XDECREF_SET(__pyx_v_code, __pyx_t_5);
        __pyx_t_5 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":1038
 *                 else:
 *                     code = frm.f_code
 *                     if code.co_name in ('wrapper', '<module>'):             # <<<<<<<<<<<<<<
 *                         continue
 * 
 */
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_code, __pyx_n_s_co_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1038, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_t_5, __pyx_n_s_wrapper, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1038, __pyx_L1_error)
        if (!__pyx_t_3) {
        } else {
          __pyx_t_2 = __pyx_t_3;
          goto __pyx_L14_bool_binop_done;
        }
        __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_t_5, __pyx_kp_s_module_2, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1038, __pyx_L1_error)
        __pyx_t_2 = __pyx_t_3;
        __pyx_L14_bool_binop_done:;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_3 = (__pyx_t_2 != 0);
        if (__pyx_t_3) {

          /* "Aspidites/_vendor/pyparsing_extension.py":1039
 *                     code = frm.f_code
 *                     if code.co_name in ('wrapper', '<module>'):
 *                         continue             # <<<<<<<<<<<<<<
 * 
 *                     ret.append("{0}".format(code.co_name))
 */
          goto __pyx_L6_continue;

          /* "Aspidites/_vendor/pyparsing_extension.py":1038
 *                 else:
 *                     code = frm.f_code
 *                     if code.co_name in ('wrapper', '<module>'):             # <<<<<<<<<<<<<<
 *                         continue
 * 
 */
        }

        /* "Aspidites/_vendor/pyparsing_extension.py":1041
 *                         continue
 * 
 *                     ret.append("{0}".format(code.co_name))             # <<<<<<<<<<<<<<
 * 
 *                 depth -= 1
 */
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_0, __pyx_n_s_format); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1041, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_code, __pyx_n_s_co_name); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1041, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        __pyx_t_12 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
          __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_9);
          if (likely(__pyx_t_12)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
            __Pyx_INCREF(__pyx_t_12);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_9, function);
          }
        }
        __pyx_t_5 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_12, __pyx_t_13) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_13);
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1041, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_ret, __pyx_t_5); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1041, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __pyx_L8:;

      /* "Aspidites/_vendor/pyparsing_extension.py":1043
 *                     ret.append("{0}".format(code.co_name))
 * 
 *                 depth -= 1             # <<<<<<<<<<<<<<
 *                 if not depth:
 *                     break
 */
      __pyx_t_5 = __Pyx_PyInt_SubtractObjC(__pyx_v_depth, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1043, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF_SET(__pyx_v_depth, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":1044
 * 
 *                 depth -= 1
 *                 if not depth:             # <<<<<<<<<<<<<<
 *                     break
 * 
 */
      __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_depth); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1044, __pyx_L1_error)
      __pyx_t_2 = ((!__pyx_t_3) != 0);
      if (__pyx_t_2) {

        /* "Aspidites/_vendor/pyparsing_extension.py":1045
 *                 depth -= 1
 *                 if not depth:
 *                     break             # <<<<<<<<<<<<<<
 * 
 *         return '\n'.join(ret)
 */
        goto __pyx_L7_break;

        /* "Aspidites/_vendor/pyparsing_extension.py":1044
 * 
 *                 depth -= 1
 *                 if not depth:             # <<<<<<<<<<<<<<
 *                     break
 * 
 */
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":1017
 *             callers = inspect.getinnerframes(exc.__traceback__, context=depth)
 *             seen = set()
 *             for i, ff in enumerate(callers[-depth:]):             # <<<<<<<<<<<<<<
 *                 frm = ff[0]
 * 
 */
      __pyx_L6_continue:;
    }
    __pyx_L7_break:;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1014
 *         ret.append("{0}: {1}".format(type(exc).__name__, exc))
 * 
 *         if depth > 0:             # <<<<<<<<<<<<<<
 *             callers = inspect.getinnerframes(exc.__traceback__, context=depth)
 *             seen = set()
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1047
 *                     break
 * 
 *         return '\n'.join(ret)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyString_Join(__pyx_kp_s__4, __pyx_v_ret); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1047, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":981
 * 
 *     @staticmethod
 *     def explain(exc, depth=16):             # <<<<<<<<<<<<<<
 *         """
 *         Method to take an exception and translate the Python internal traceback into a list
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseException.explain", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_inspect);
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_XDECREF(__pyx_v_callers);
  __Pyx_XDECREF(__pyx_v_seen);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_ff);
  __Pyx_XDECREF(__pyx_v_frm);
  __Pyx_XDECREF(__pyx_v_f_self);
  __Pyx_XDECREF(__pyx_v_self_type);
  __Pyx_XDECREF(__pyx_v_code);
  __Pyx_XDECREF(__pyx_v_depth);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1083
 *     """
 * 
 *     def __init__(self, parseElementList):             # <<<<<<<<<<<<<<
 *         self.parseElementTrace = parseElementList
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_25RecursiveGrammarException_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_25RecursiveGrammarException_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_25RecursiveGrammarException_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_25RecursiveGrammarException_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parseElementList = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parseElementList,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_parseElementList)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(0, 1083, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 1083, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_parseElementList = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1083, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.RecursiveGrammarException.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_25RecursiveGrammarException___init__(__pyx_self, __pyx_v_self, __pyx_v_parseElementList);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_25RecursiveGrammarException___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parseElementList) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":1084
 * 
 *     def __init__(self, parseElementList):
 *         self.parseElementTrace = parseElementList             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parseElementTrace, __pyx_v_parseElementList) < 0) __PYX_ERR(0, 1084, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1083
 *     """
 * 
 *     def __init__(self, parseElementList):             # <<<<<<<<<<<<<<
 *         self.parseElementTrace = parseElementList
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.RecursiveGrammarException.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1086
 *         self.parseElementTrace = parseElementList
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "RecursiveGrammarException: %s" % self.parseElementTrace
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_25RecursiveGrammarException_3__str__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_25RecursiveGrammarException_3__str__ = {"__str__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_25RecursiveGrammarException_3__str__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_25RecursiveGrammarException_3__str__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_25RecursiveGrammarException_2__str__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_25RecursiveGrammarException_2__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":1087
 * 
 *     def __str__(self):
 *         return "RecursiveGrammarException: %s" % self.parseElementTrace             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_parseElementTrace); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1087, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyString_FormatSafe(__pyx_kp_s_RecursiveGrammarException_s, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1087, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1086
 *         self.parseElementTrace = parseElementList
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "RecursiveGrammarException: %s" % self.parseElementTrace
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.RecursiveGrammarException.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1096
 * 
 *     @staticmethod
 *     def setDefaultWhitespaceChars(chars):             # <<<<<<<<<<<<<<
 *         r"""
 *         Overrides the default whitespace chars
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_1setDefaultWhitespaceChars(PyObject *__pyx_self, PyObject *__pyx_v_chars); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_setDefaultWhitespaceChars[] = "\n        Overrides the default whitespace chars\n\n        Example::\n\n            # default whitespace chars are space, <TAB> and newline\n            OneOrMore(Word(alphas)).parseString(\"abc def\\nghi jkl\")  # -> ['abc', 'def', 'ghi', 'jkl']\n\n            # change to just treat newline as significant\n            ParserElement.setDefaultWhitespaceChars(\" \\t\")\n            OneOrMore(Word(alphas)).parseString(\"abc def\\nghi jkl\")  # -> ['abc', 'def']\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_1setDefaultWhitespaceChars = {"setDefaultWhitespaceChars", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_1setDefaultWhitespaceChars, METH_O, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_setDefaultWhitespaceChars};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_1setDefaultWhitespaceChars(PyObject *__pyx_self, PyObject *__pyx_v_chars) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setDefaultWhitespaceChars (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_setDefaultWhitespaceChars(__pyx_self, ((PyObject *)__pyx_v_chars));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_setDefaultWhitespaceChars(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_chars) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setDefaultWhitespaceChars", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":1109
 *             OneOrMore(Word(alphas)).parseString("abc def\nghi jkl")  # -> ['abc', 'def']
 *         """
 *         ParserElement.DEFAULT_WHITE_CHARS = chars             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_1, __pyx_n_s_DEFAULT_WHITE_CHARS, __pyx_v_chars) < 0) __PYX_ERR(0, 1109, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1096
 * 
 *     @staticmethod
 *     def setDefaultWhitespaceChars(chars):             # <<<<<<<<<<<<<<
 *         r"""
 *         Overrides the default whitespace chars
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.setDefaultWhitespaceChars", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1112
 * 
 *     @staticmethod
 *     def inlineLiteralsUsing(cls):             # <<<<<<<<<<<<<<
 *         """
 *         Set class to be used for inclusion of string literals into a parser.
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_3inlineLiteralsUsing(PyObject *__pyx_self, PyObject *__pyx_v_cls); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_2inlineLiteralsUsing[] = "\n        Set class to be used for inclusion of string literals into a parser.\n\n        Example::\n\n            # default literal class used is Literal\n            integer = Word(nums)\n            date_str = integer(\"year\") + '/' + integer(\"month\") + '/' + integer(\"day\")\n\n            date_str.parseString(\"1999/12/31\")  # -> ['1999', '/', '12', '/', '31']\n\n\n            # change to Suppress\n            ParserElement.inlineLiteralsUsing(Suppress)\n            date_str = integer(\"year\") + '/' + integer(\"month\") + '/' + integer(\"day\")\n\n            date_str.parseString(\"1999/12/31\")  # -> ['1999', '12', '31']\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_3inlineLiteralsUsing = {"inlineLiteralsUsing", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_3inlineLiteralsUsing, METH_O, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_2inlineLiteralsUsing};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_3inlineLiteralsUsing(PyObject *__pyx_self, PyObject *__pyx_v_cls) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("inlineLiteralsUsing (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_2inlineLiteralsUsing(__pyx_self, ((PyObject *)__pyx_v_cls));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_2inlineLiteralsUsing(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_cls) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("inlineLiteralsUsing", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":1131
 *             date_str.parseString("1999/12/31")  # -> ['1999', '12', '31']
 *         """
 *         ParserElement._literalStringClass = cls             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_1, __pyx_n_s_literalStringClass, __pyx_v_cls) < 0) __PYX_ERR(0, 1131, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1112
 * 
 *     @staticmethod
 *     def inlineLiteralsUsing(cls):             # <<<<<<<<<<<<<<
 *         """
 *         Set class to be used for inclusion of string literals into a parser.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.inlineLiteralsUsing", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1134
 * 
 *     @classmethod
 *     def _trim_traceback(cls, tb):             # <<<<<<<<<<<<<<
 *         while tb.tb_next:
 *             tb = tb.tb_next
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_5_trim_traceback(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_5_trim_traceback = {"_trim_traceback", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_5_trim_traceback, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_5_trim_traceback(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_cls = 0;
  PyObject *__pyx_v_tb = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_trim_traceback (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_cls,&__pyx_n_s_tb,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cls)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tb)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_trim_traceback", 1, 2, 2, 1); __PYX_ERR(0, 1134, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_trim_traceback") < 0)) __PYX_ERR(0, 1134, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_cls = values[0];
    __pyx_v_tb = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_trim_traceback", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1134, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement._trim_traceback", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_4_trim_traceback(__pyx_self, __pyx_v_cls, __pyx_v_tb);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_4_trim_traceback(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_cls, PyObject *__pyx_v_tb) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_trim_traceback", 0);
  __Pyx_INCREF(__pyx_v_tb);

  /* "Aspidites/_vendor/pyparsing_extension.py":1135
 *     @classmethod
 *     def _trim_traceback(cls, tb):
 *         while tb.tb_next:             # <<<<<<<<<<<<<<
 *             tb = tb.tb_next
 *         return tb
 */
  while (1) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_tb, __pyx_n_s_tb_next); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1135, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!__pyx_t_2) break;

    /* "Aspidites/_vendor/pyparsing_extension.py":1136
 *     def _trim_traceback(cls, tb):
 *         while tb.tb_next:
 *             tb = tb.tb_next             # <<<<<<<<<<<<<<
 *         return tb
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_tb, __pyx_n_s_tb_next); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1136, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_tb, __pyx_t_1);
    __pyx_t_1 = 0;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1137
 *         while tb.tb_next:
 *             tb = tb.tb_next
 *         return tb             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, savelist=False):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_tb);
  __pyx_r = __pyx_v_tb;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1134
 * 
 *     @classmethod
 *     def _trim_traceback(cls, tb):             # <<<<<<<<<<<<<<
 *         while tb.tb_next:
 *             tb = tb.tb_next
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement._trim_traceback", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tb);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1139
 *         return tb
 * 
 *     def __init__(self, savelist=False):             # <<<<<<<<<<<<<<
 *         self.parseAction = list()
 *         self.failAction = None
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_7__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_7__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_7__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_7__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_savelist = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_savelist,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_False));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_savelist);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 1139, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_savelist = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1139, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_6__init__(__pyx_self, __pyx_v_self, __pyx_v_savelist);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_6__init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_savelist) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":1140
 * 
 *     def __init__(self, savelist=False):
 *         self.parseAction = list()             # <<<<<<<<<<<<<<
 *         self.failAction = None
 *         # ~ self.name = "<unknown>"  # don't define self.name, let subclasses try/except upcall
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parseAction, __pyx_t_1) < 0) __PYX_ERR(0, 1140, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1141
 *     def __init__(self, savelist=False):
 *         self.parseAction = list()
 *         self.failAction = None             # <<<<<<<<<<<<<<
 *         # ~ self.name = "<unknown>"  # don't define self.name, let subclasses try/except upcall
 *         self.strRepr = None
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_failAction, Py_None) < 0) __PYX_ERR(0, 1141, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1143
 *         self.failAction = None
 *         # ~ self.name = "<unknown>"  # don't define self.name, let subclasses try/except upcall
 *         self.strRepr = None             # <<<<<<<<<<<<<<
 *         self.resultsName = None
 *         self.saveAsList = savelist
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_strRepr, Py_None) < 0) __PYX_ERR(0, 1143, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1144
 *         # ~ self.name = "<unknown>"  # don't define self.name, let subclasses try/except upcall
 *         self.strRepr = None
 *         self.resultsName = None             # <<<<<<<<<<<<<<
 *         self.saveAsList = savelist
 *         self.skipWhitespace = True
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_resultsName, Py_None) < 0) __PYX_ERR(0, 1144, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1145
 *         self.strRepr = None
 *         self.resultsName = None
 *         self.saveAsList = savelist             # <<<<<<<<<<<<<<
 *         self.skipWhitespace = True
 *         self.whiteChars = set(ParserElement.DEFAULT_WHITE_CHARS)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_saveAsList, __pyx_v_savelist) < 0) __PYX_ERR(0, 1145, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1146
 *         self.resultsName = None
 *         self.saveAsList = savelist
 *         self.skipWhitespace = True             # <<<<<<<<<<<<<<
 *         self.whiteChars = set(ParserElement.DEFAULT_WHITE_CHARS)
 *         self.copyDefaultWhiteChars = True
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_skipWhitespace, Py_True) < 0) __PYX_ERR(0, 1146, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1147
 *         self.saveAsList = savelist
 *         self.skipWhitespace = True
 *         self.whiteChars = set(ParserElement.DEFAULT_WHITE_CHARS)             # <<<<<<<<<<<<<<
 *         self.copyDefaultWhiteChars = True
 *         self.mayReturnEmpty = False  # used when checking for left-recursion
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_DEFAULT_WHITE_CHARS); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PySet_New(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_whiteChars, __pyx_t_1) < 0) __PYX_ERR(0, 1147, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1148
 *         self.skipWhitespace = True
 *         self.whiteChars = set(ParserElement.DEFAULT_WHITE_CHARS)
 *         self.copyDefaultWhiteChars = True             # <<<<<<<<<<<<<<
 *         self.mayReturnEmpty = False  # used when checking for left-recursion
 *         self.keepTabs = False
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_copyDefaultWhiteChars, Py_True) < 0) __PYX_ERR(0, 1148, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1149
 *         self.whiteChars = set(ParserElement.DEFAULT_WHITE_CHARS)
 *         self.copyDefaultWhiteChars = True
 *         self.mayReturnEmpty = False  # used when checking for left-recursion             # <<<<<<<<<<<<<<
 *         self.keepTabs = False
 *         self.ignoreExprs = list()
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_mayReturnEmpty, Py_False) < 0) __PYX_ERR(0, 1149, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1150
 *         self.copyDefaultWhiteChars = True
 *         self.mayReturnEmpty = False  # used when checking for left-recursion
 *         self.keepTabs = False             # <<<<<<<<<<<<<<
 *         self.ignoreExprs = list()
 *         self.debug = False
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_keepTabs, Py_False) < 0) __PYX_ERR(0, 1150, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1151
 *         self.mayReturnEmpty = False  # used when checking for left-recursion
 *         self.keepTabs = False
 *         self.ignoreExprs = list()             # <<<<<<<<<<<<<<
 *         self.debug = False
 *         self.streamlined = False
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_ignoreExprs, __pyx_t_1) < 0) __PYX_ERR(0, 1151, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1152
 *         self.keepTabs = False
 *         self.ignoreExprs = list()
 *         self.debug = False             # <<<<<<<<<<<<<<
 *         self.streamlined = False
 *         self.mayIndexError = True  # used to optimize exception handling for subclasses that don't advance parse index
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_debug, Py_False) < 0) __PYX_ERR(0, 1152, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1153
 *         self.ignoreExprs = list()
 *         self.debug = False
 *         self.streamlined = False             # <<<<<<<<<<<<<<
 *         self.mayIndexError = True  # used to optimize exception handling for subclasses that don't advance parse index
 *         self.errmsg = ""
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_streamlined, Py_False) < 0) __PYX_ERR(0, 1153, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1154
 *         self.debug = False
 *         self.streamlined = False
 *         self.mayIndexError = True  # used to optimize exception handling for subclasses that don't advance parse index             # <<<<<<<<<<<<<<
 *         self.errmsg = ""
 *         self.modalResults = True  # used to mark results names as modal (report only last) or cumulative (list all)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_mayIndexError, Py_True) < 0) __PYX_ERR(0, 1154, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1155
 *         self.streamlined = False
 *         self.mayIndexError = True  # used to optimize exception handling for subclasses that don't advance parse index
 *         self.errmsg = ""             # <<<<<<<<<<<<<<
 *         self.modalResults = True  # used to mark results names as modal (report only last) or cumulative (list all)
 *         self.debugActions = (None, None, None)  # custom debug actions
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_errmsg, __pyx_kp_s__12) < 0) __PYX_ERR(0, 1155, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1156
 *         self.mayIndexError = True  # used to optimize exception handling for subclasses that don't advance parse index
 *         self.errmsg = ""
 *         self.modalResults = True  # used to mark results names as modal (report only last) or cumulative (list all)             # <<<<<<<<<<<<<<
 *         self.debugActions = (None, None, None)  # custom debug actions
 *         self.re = None
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_modalResults, Py_True) < 0) __PYX_ERR(0, 1156, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1157
 *         self.errmsg = ""
 *         self.modalResults = True  # used to mark results names as modal (report only last) or cumulative (list all)
 *         self.debugActions = (None, None, None)  # custom debug actions             # <<<<<<<<<<<<<<
 *         self.re = None
 *         self.callPreparse = True  # used to avoid redundant calls to preParse
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_debugActions, __pyx_tuple__32) < 0) __PYX_ERR(0, 1157, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1158
 *         self.modalResults = True  # used to mark results names as modal (report only last) or cumulative (list all)
 *         self.debugActions = (None, None, None)  # custom debug actions
 *         self.re = None             # <<<<<<<<<<<<<<
 *         self.callPreparse = True  # used to avoid redundant calls to preParse
 *         self.callDuringTry = False
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_re, Py_None) < 0) __PYX_ERR(0, 1158, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1159
 *         self.debugActions = (None, None, None)  # custom debug actions
 *         self.re = None
 *         self.callPreparse = True  # used to avoid redundant calls to preParse             # <<<<<<<<<<<<<<
 *         self.callDuringTry = False
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_callPreparse, Py_True) < 0) __PYX_ERR(0, 1159, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1160
 *         self.re = None
 *         self.callPreparse = True  # used to avoid redundant calls to preParse
 *         self.callDuringTry = False             # <<<<<<<<<<<<<<
 * 
 *     def copy(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_callDuringTry, Py_False) < 0) __PYX_ERR(0, 1160, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1139
 *         return tb
 * 
 *     def __init__(self, savelist=False):             # <<<<<<<<<<<<<<
 *         self.parseAction = list()
 *         self.failAction = None
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1162
 *         self.callDuringTry = False
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         """
 *         Make a copy of this :class:`ParserElement`.  Useful for defining
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_9copy(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_8copy[] = "\n        Make a copy of this :class:`ParserElement`.  Useful for defining\n        different parse actions for the same parsing pattern, using copies of\n        the original parse element.\n\n        Example::\n\n            integer = Word(nums).setParseAction(lambda toks: int(toks[0]))\n            integerK = integer.copy().addParseAction(lambda toks: toks[0] * 1024) + Suppress(\"K\")\n            integerM = integer.copy().addParseAction(lambda toks: toks[0] * 1024 * 1024) + Suppress(\"M\")\n\n            print(OneOrMore(integerK | integerM | integer).parseString(\"5K 100 640K 256M\"))\n\n        prints::\n\n            [5120, 100, 655360, 268435456]\n\n        Equivalent form of ``expr.copy()`` is just ``expr()``::\n\n            integerM = integer().addParseAction(lambda toks: toks[0] * 1024 * 1024) + Suppress(\"M\")\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_9copy = {"copy", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_9copy, METH_O, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_8copy};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_9copy(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_8copy(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_8copy(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_cpy = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":1184
 *             integerM = integer().addParseAction(lambda toks: toks[0] * 1024 * 1024) + Suppress("M")
 *         """
 *         cpy = copy.copy(self)             # <<<<<<<<<<<<<<
 *         cpy.parseAction = self.parseAction[:]
 *         cpy.ignoreExprs = self.ignoreExprs[:]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_copy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_copy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_self) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_self);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_cpy = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1185
 *         """
 *         cpy = copy.copy(self)
 *         cpy.parseAction = self.parseAction[:]             # <<<<<<<<<<<<<<
 *         cpy.ignoreExprs = self.ignoreExprs[:]
 *         if self.copyDefaultWhiteChars:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_parseAction); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_t_1, 0, 0, NULL, NULL, &__pyx_slice__11, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_cpy, __pyx_n_s_parseAction, __pyx_t_3) < 0) __PYX_ERR(0, 1185, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1186
 *         cpy = copy.copy(self)
 *         cpy.parseAction = self.parseAction[:]
 *         cpy.ignoreExprs = self.ignoreExprs[:]             # <<<<<<<<<<<<<<
 *         if self.copyDefaultWhiteChars:
 *             cpy.whiteChars = ParserElement.DEFAULT_WHITE_CHARS
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ignoreExprs); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_t_3, 0, 0, NULL, NULL, &__pyx_slice__11, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_cpy, __pyx_n_s_ignoreExprs, __pyx_t_1) < 0) __PYX_ERR(0, 1186, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1187
 *         cpy.parseAction = self.parseAction[:]
 *         cpy.ignoreExprs = self.ignoreExprs[:]
 *         if self.copyDefaultWhiteChars:             # <<<<<<<<<<<<<<
 *             cpy.whiteChars = ParserElement.DEFAULT_WHITE_CHARS
 *         return cpy
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copyDefaultWhiteChars); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1187, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":1188
 *         cpy.ignoreExprs = self.ignoreExprs[:]
 *         if self.copyDefaultWhiteChars:
 *             cpy.whiteChars = ParserElement.DEFAULT_WHITE_CHARS             # <<<<<<<<<<<<<<
 *         return cpy
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1188, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_DEFAULT_WHITE_CHARS); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1188, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_cpy, __pyx_n_s_whiteChars, __pyx_t_3) < 0) __PYX_ERR(0, 1188, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1187
 *         cpy.parseAction = self.parseAction[:]
 *         cpy.ignoreExprs = self.ignoreExprs[:]
 *         if self.copyDefaultWhiteChars:             # <<<<<<<<<<<<<<
 *             cpy.whiteChars = ParserElement.DEFAULT_WHITE_CHARS
 *         return cpy
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1189
 *         if self.copyDefaultWhiteChars:
 *             cpy.whiteChars = ParserElement.DEFAULT_WHITE_CHARS
 *         return cpy             # <<<<<<<<<<<<<<
 * 
 *     def setName(self, name):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_cpy);
  __pyx_r = __pyx_v_cpy;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1162
 *         self.callDuringTry = False
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         """
 *         Make a copy of this :class:`ParserElement`.  Useful for defining
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_cpy);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1191
 *         return cpy
 * 
 *     def setName(self, name):             # <<<<<<<<<<<<<<
 *         """
 *         Define name for this expression, makes debugging and exception messages clearer.
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_11setName(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_10setName[] = "\n        Define name for this expression, makes debugging and exception messages clearer.\n\n        Example::\n\n            Word(nums).parseString(\"ABC\")  # -> Exception: Expected W:(0123...) (at char 0), (line:1, col:1)\n            Word(nums).setName(\"integer\").parseString(\"ABC\")  # -> Exception: Expected integer (at char 0), (line:1, col:1)\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_11setName = {"setName", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_11setName, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_10setName};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_11setName(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_name = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setName (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_name,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setName", 1, 2, 2, 1); __PYX_ERR(0, 1191, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setName") < 0)) __PYX_ERR(0, 1191, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_name = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setName", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1191, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.setName", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_10setName(__pyx_self, __pyx_v_self, __pyx_v_name);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_10setName(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_name) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setName", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":1200
 *             Word(nums).setName("integer").parseString("ABC")  # -> Exception: Expected integer (at char 0), (line:1, col:1)
 *         """
 *         self.name = name             # <<<<<<<<<<<<<<
 *         self.errmsg = "Expected " + self.name
 *         if __diag__.enable_debug_on_named_expressions:
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_name, __pyx_v_name) < 0) __PYX_ERR(0, 1200, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1201
 *         """
 *         self.name = name
 *         self.errmsg = "Expected " + self.name             # <<<<<<<<<<<<<<
 *         if __diag__.enable_debug_on_named_expressions:
 *             self.setDebug()
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_Add(__pyx_kp_s_Expected, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_errmsg, __pyx_t_2) < 0) __PYX_ERR(0, 1201, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1202
 *         self.name = name
 *         self.errmsg = "Expected " + self.name
 *         if __diag__.enable_debug_on_named_expressions:             # <<<<<<<<<<<<<<
 *             self.setDebug()
 *         return self
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_diag); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_enable_debug_on_named_expression); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1202, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/pyparsing_extension.py":1203
 *         self.errmsg = "Expected " + self.name
 *         if __diag__.enable_debug_on_named_expressions:
 *             self.setDebug()             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_setDebug); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1203, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1203, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1202
 *         self.name = name
 *         self.errmsg = "Expected " + self.name
 *         if __diag__.enable_debug_on_named_expressions:             # <<<<<<<<<<<<<<
 *             self.setDebug()
 *         return self
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1204
 *         if __diag__.enable_debug_on_named_expressions:
 *             self.setDebug()
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def setResultsName(self, name, listAllMatches=False):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1191
 *         return cpy
 * 
 *     def setName(self, name):             # <<<<<<<<<<<<<<
 *         """
 *         Define name for this expression, makes debugging and exception messages clearer.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.setName", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1206
 *         return self
 * 
 *     def setResultsName(self, name, listAllMatches=False):             # <<<<<<<<<<<<<<
 *         """
 *         Define name for referencing matching tokens as a nested attribute
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_13setResultsName(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_12setResultsName[] = "\n        Define name for referencing matching tokens as a nested attribute\n        of the returned parse results.\n        NOTE: this returns a *copy* of the original :class:`ParserElement` object;\n        this is so that the client can define a basic element, such as an\n        integer, and reference it in multiple places with different names.\n\n        You can also set results names using the abbreviated syntax,\n        ``expr(\"name\")`` in place of ``expr.setResultsName(\"name\")``\n        - see :class:`__call__`.\n\n        Example::\n\n            date_str = (integer.setResultsName(\"year\") + '/'\n                        + integer.setResultsName(\"month\") + '/'\n                        + integer.setResultsName(\"day\"))\n\n            # equivalent form:\n            date_str = integer(\"year\") + '/' + integer(\"month\") + '/' + integer(\"day\")\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_13setResultsName = {"setResultsName", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_13setResultsName, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_12setResultsName};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_13setResultsName(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_listAllMatches = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setResultsName (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_name,&__pyx_n_s_listAllMatches,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_False));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setResultsName", 0, 2, 3, 1); __PYX_ERR(0, 1206, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_listAllMatches);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setResultsName") < 0)) __PYX_ERR(0, 1206, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_name = values[1];
    __pyx_v_listAllMatches = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setResultsName", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1206, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.setResultsName", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_12setResultsName(__pyx_self, __pyx_v_self, __pyx_v_name, __pyx_v_listAllMatches);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_12setResultsName(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_listAllMatches) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setResultsName", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":1227
 *             date_str = integer("year") + '/' + integer("month") + '/' + integer("day")
 *         """
 *         return self._setResultsName(name, listAllMatches)             # <<<<<<<<<<<<<<
 * 
 *     def _setResultsName(self, name, listAllMatches=False):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_setResultsName); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_name, __pyx_v_listAllMatches};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1227, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_name, __pyx_v_listAllMatches};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1227, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1227, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_name);
    __Pyx_GIVEREF(__pyx_v_name);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_name);
    __Pyx_INCREF(__pyx_v_listAllMatches);
    __Pyx_GIVEREF(__pyx_v_listAllMatches);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_listAllMatches);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1227, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1206
 *         return self
 * 
 *     def setResultsName(self, name, listAllMatches=False):             # <<<<<<<<<<<<<<
 *         """
 *         Define name for referencing matching tokens as a nested attribute
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.setResultsName", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1229
 *         return self._setResultsName(name, listAllMatches)
 * 
 *     def _setResultsName(self, name, listAllMatches=False):             # <<<<<<<<<<<<<<
 *         newself = self.copy()
 *         if name.endswith("*"):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_15_setResultsName(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_15_setResultsName = {"_setResultsName", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_15_setResultsName, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_15_setResultsName(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_listAllMatches = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_setResultsName (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_name,&__pyx_n_s_listAllMatches,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_False));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_setResultsName", 0, 2, 3, 1); __PYX_ERR(0, 1229, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_listAllMatches);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_setResultsName") < 0)) __PYX_ERR(0, 1229, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_name = values[1];
    __pyx_v_listAllMatches = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_setResultsName", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1229, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement._setResultsName", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_14_setResultsName(__pyx_self, __pyx_v_self, __pyx_v_name, __pyx_v_listAllMatches);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_14_setResultsName(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_listAllMatches) {
  PyObject *__pyx_v_newself = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_setResultsName", 0);
  __Pyx_INCREF(__pyx_v_name);
  __Pyx_INCREF(__pyx_v_listAllMatches);

  /* "Aspidites/_vendor/pyparsing_extension.py":1230
 * 
 *     def _setResultsName(self, name, listAllMatches=False):
 *         newself = self.copy()             # <<<<<<<<<<<<<<
 *         if name.endswith("*"):
 *             name = name[:-1]
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_newself = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1231
 *     def _setResultsName(self, name, listAllMatches=False):
 *         newself = self.copy()
 *         if name.endswith("*"):             # <<<<<<<<<<<<<<
 *             name = name[:-1]
 *             listAllMatches = True
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_name, __pyx_n_s_endswith); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_s__33) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s__33);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1231, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":1232
 *         newself = self.copy()
 *         if name.endswith("*"):
 *             name = name[:-1]             # <<<<<<<<<<<<<<
 *             listAllMatches = True
 *         newself.resultsName = name
 */
    __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_name, 0, -1L, NULL, NULL, &__pyx_slice__34, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_name, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1233
 *         if name.endswith("*"):
 *             name = name[:-1]
 *             listAllMatches = True             # <<<<<<<<<<<<<<
 *         newself.resultsName = name
 *         newself.modalResults = not listAllMatches
 */
    __Pyx_INCREF(Py_True);
    __Pyx_DECREF_SET(__pyx_v_listAllMatches, Py_True);

    /* "Aspidites/_vendor/pyparsing_extension.py":1231
 *     def _setResultsName(self, name, listAllMatches=False):
 *         newself = self.copy()
 *         if name.endswith("*"):             # <<<<<<<<<<<<<<
 *             name = name[:-1]
 *             listAllMatches = True
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1234
 *             name = name[:-1]
 *             listAllMatches = True
 *         newself.resultsName = name             # <<<<<<<<<<<<<<
 *         newself.modalResults = not listAllMatches
 *         return newself
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_newself, __pyx_n_s_resultsName, __pyx_v_name) < 0) __PYX_ERR(0, 1234, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1235
 *             listAllMatches = True
 *         newself.resultsName = name
 *         newself.modalResults = not listAllMatches             # <<<<<<<<<<<<<<
 *         return newself
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_listAllMatches); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1235, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyBool_FromLong((!__pyx_t_4)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_newself, __pyx_n_s_modalResults, __pyx_t_1) < 0) __PYX_ERR(0, 1235, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1236
 *         newself.resultsName = name
 *         newself.modalResults = not listAllMatches
 *         return newself             # <<<<<<<<<<<<<<
 * 
 *     def setBreak(self, breakFlag=True):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_newself);
  __pyx_r = __pyx_v_newself;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1229
 *         return self._setResultsName(name, listAllMatches)
 * 
 *     def _setResultsName(self, name, listAllMatches=False):             # <<<<<<<<<<<<<<
 *         newself = self.copy()
 *         if name.endswith("*"):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement._setResultsName", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_newself);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_listAllMatches);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1238
 *         return newself
 * 
 *     def setBreak(self, breakFlag=True):             # <<<<<<<<<<<<<<
 *         """Method to invoke the Python pdb debugger when this element is
 *            about to be parsed. Set ``breakFlag`` to True to enable, False to
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_17setBreak(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_16setBreak[] = "Method to invoke the Python pdb debugger when this element is\n           about to be parsed. Set ``breakFlag`` to True to enable, False to\n           disable.\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_17setBreak = {"setBreak", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_17setBreak, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_16setBreak};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_17setBreak(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_breakFlag = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setBreak (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_breakFlag,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_breakFlag);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setBreak") < 0)) __PYX_ERR(0, 1238, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_breakFlag = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setBreak", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1238, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.setBreak", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_16setBreak(__pyx_self, __pyx_v_self, __pyx_v_breakFlag);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1246
 *             _parseMethod = self._parse
 * 
 *             def breaker(instring, loc, doActions=True, callPreParse=True):             # <<<<<<<<<<<<<<
 *                 import pdb
 *                 # this call to pdb.set_trace() is intentional, not a checkin error
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_8setBreak_1breaker(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_8setBreak_1breaker = {"breaker", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_8setBreak_1breaker, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_8setBreak_1breaker(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  PyObject *__pyx_v_doActions = 0;
  PyObject *__pyx_v_callPreParse = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("breaker (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_doActions,&__pyx_n_s_callPreParse,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_True));
    values[3] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("breaker", 0, 2, 4, 1); __PYX_ERR(0, 1246, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_doActions);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_callPreParse);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "breaker") < 0)) __PYX_ERR(0, 1246, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_instring = values[0];
    __pyx_v_loc = values[1];
    __pyx_v_doActions = values[2];
    __pyx_v_callPreParse = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("breaker", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1246, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.setBreak.breaker", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_8setBreak_breaker(__pyx_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions, __pyx_v_callPreParse);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_8setBreak_breaker(PyObject *__pyx_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, PyObject *__pyx_v_doActions, PyObject *__pyx_v_callPreParse) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak *__pyx_cur_scope;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak *__pyx_outer_scope;
  PyObject *__pyx_v_pdb = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("breaker", 0);
  __pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "Aspidites/_vendor/pyparsing_extension.py":1247
 * 
 *             def breaker(instring, loc, doActions=True, callPreParse=True):
 *                 import pdb             # <<<<<<<<<<<<<<
 *                 # this call to pdb.set_trace() is intentional, not a checkin error
 *                 pdb.set_trace()
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_pdb, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_pdb = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1249
 *                 import pdb
 *                 # this call to pdb.set_trace() is intentional, not a checkin error
 *                 pdb.set_trace()             # <<<<<<<<<<<<<<
 *                 return _parseMethod(instring, loc, doActions, callPreParse)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_pdb, __pyx_n_s_set_trace); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1250
 *                 # this call to pdb.set_trace() is intentional, not a checkin error
 *                 pdb.set_trace()
 *                 return _parseMethod(instring, loc, doActions, callPreParse)             # <<<<<<<<<<<<<<
 * 
 *             breaker._originalParseMethod = _parseMethod
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_v__parseMethod)) { __Pyx_RaiseClosureNameError("_parseMethod"); __PYX_ERR(0, 1250, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v__parseMethod);
  __pyx_t_2 = __pyx_cur_scope->__pyx_v__parseMethod; __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_3, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions, __pyx_v_callPreParse};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 4+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1250, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_3, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions, __pyx_v_callPreParse};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 4+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1250, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(4+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_instring);
    __Pyx_GIVEREF(__pyx_v_instring);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_instring);
    __Pyx_INCREF(__pyx_v_loc);
    __Pyx_GIVEREF(__pyx_v_loc);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_loc);
    __Pyx_INCREF(__pyx_v_doActions);
    __Pyx_GIVEREF(__pyx_v_doActions);
    PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_4, __pyx_v_doActions);
    __Pyx_INCREF(__pyx_v_callPreParse);
    __Pyx_GIVEREF(__pyx_v_callPreParse);
    PyTuple_SET_ITEM(__pyx_t_5, 3+__pyx_t_4, __pyx_v_callPreParse);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1246
 *             _parseMethod = self._parse
 * 
 *             def breaker(instring, loc, doActions=True, callPreParse=True):             # <<<<<<<<<<<<<<
 *                 import pdb
 *                 # this call to pdb.set_trace() is intentional, not a checkin error
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.setBreak.breaker", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pdb);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1238
 *         return newself
 * 
 *     def setBreak(self, breakFlag=True):             # <<<<<<<<<<<<<<
 *         """Method to invoke the Python pdb debugger when this element is
 *            about to be parsed. Set ``breakFlag`` to True to enable, False to
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_16setBreak(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_breakFlag) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak *__pyx_cur_scope;
  PyObject *__pyx_v_breaker = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setBreak", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1238, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1243
 *            disable.
 *         """
 *         if breakFlag:             # <<<<<<<<<<<<<<
 *             _parseMethod = self._parse
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_breakFlag); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1243, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "Aspidites/_vendor/pyparsing_extension.py":1244
 *         """
 *         if breakFlag:
 *             _parseMethod = self._parse             # <<<<<<<<<<<<<<
 * 
 *             def breaker(instring, loc, doActions=True, callPreParse=True):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_parse); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1244, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __pyx_cur_scope->__pyx_v__parseMethod = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1246
 *             _parseMethod = self._parse
 * 
 *             def breaker(instring, loc, doActions=True, callPreParse=True):             # <<<<<<<<<<<<<<
 *                 import pdb
 *                 # this call to pdb.set_trace() is intentional, not a checkin error
 */
    __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_8setBreak_1breaker, 0, __pyx_n_s_ParserElement_setBreak_locals_br, ((PyObject*)__pyx_cur_scope), __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__36)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1246, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_tuple__37);
    __pyx_v_breaker = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1252
 *                 return _parseMethod(instring, loc, doActions, callPreParse)
 * 
 *             breaker._originalParseMethod = _parseMethod             # <<<<<<<<<<<<<<
 *             self._parse = breaker
 *         else:
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_breaker, __pyx_n_s_originalParseMethod, __pyx_cur_scope->__pyx_v__parseMethod) < 0) __PYX_ERR(0, 1252, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":1253
 * 
 *             breaker._originalParseMethod = _parseMethod
 *             self._parse = breaker             # <<<<<<<<<<<<<<
 *         else:
 *             if hasattr(self._parse, "_originalParseMethod"):
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parse, __pyx_v_breaker) < 0) __PYX_ERR(0, 1253, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":1243
 *            disable.
 *         """
 *         if breakFlag:             # <<<<<<<<<<<<<<
 *             _parseMethod = self._parse
 * 
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1255
 *             self._parse = breaker
 *         else:
 *             if hasattr(self._parse, "_originalParseMethod"):             # <<<<<<<<<<<<<<
 *                 self._parse = self._parse._originalParseMethod
 *         return self
 */
  /*else*/ {
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_parse); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1255, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_HasAttr(__pyx_t_2, __pyx_n_s_originalParseMethod); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 1255, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_3 = (__pyx_t_1 != 0);
    if (__pyx_t_3) {

      /* "Aspidites/_vendor/pyparsing_extension.py":1256
 *         else:
 *             if hasattr(self._parse, "_originalParseMethod"):
 *                 self._parse = self._parse._originalParseMethod             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_parse); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1256, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_originalParseMethod); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1256, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parse, __pyx_t_4) < 0) __PYX_ERR(0, 1256, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":1255
 *             self._parse = breaker
 *         else:
 *             if hasattr(self._parse, "_originalParseMethod"):             # <<<<<<<<<<<<<<
 *                 self._parse = self._parse._originalParseMethod
 *         return self
 */
    }
  }
  __pyx_L3:;

  /* "Aspidites/_vendor/pyparsing_extension.py":1257
 *             if hasattr(self._parse, "_originalParseMethod"):
 *                 self._parse = self._parse._originalParseMethod
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def setParseAction(self, *fns, **kwargs):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1238
 *         return newself
 * 
 *     def setBreak(self, breakFlag=True):             # <<<<<<<<<<<<<<
 *         """Method to invoke the Python pdb debugger when this element is
 *            about to be parsed. Set ``breakFlag`` to True to enable, False to
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.setBreak", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_breaker);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1259
 *         return self
 * 
 *     def setParseAction(self, *fns, **kwargs):             # <<<<<<<<<<<<<<
 *         from .pyparsing import _trim_arity
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_19setParseAction(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_19setParseAction = {"setParseAction", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_19setParseAction, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_19setParseAction(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_fns = 0;
  PyObject *__pyx_v_kwargs = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setParseAction (wrapper)", 0);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  if (PyTuple_GET_SIZE(__pyx_args) > 1) {
    __pyx_v_fns = PyTuple_GetSlice(__pyx_args, 1, PyTuple_GET_SIZE(__pyx_args));
    if (unlikely(!__pyx_v_fns)) {
      __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
      __Pyx_RefNannyFinishContext();
      return NULL;
    }
    __Pyx_GOTREF(__pyx_v_fns);
  } else {
    __pyx_v_fns = __pyx_empty_tuple; __Pyx_INCREF(__pyx_empty_tuple);
  }
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        default:
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t used_pos_args = (pos_args < 1) ? pos_args : 1;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwargs, values, used_pos_args, "setParseAction") < 0)) __PYX_ERR(0, 1259, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) < 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_self = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setParseAction", 0, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1259, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_CLEAR(__pyx_v_fns);
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.setParseAction", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_18setParseAction(__pyx_self, __pyx_v_self, __pyx_v_fns, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_fns);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_14setParseAction_2generator12(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/pyparsing_extension.py":1303
 *             self.parseAction = []
 *         else:
 *             if not all(callable(fn) for fn in fns):             # <<<<<<<<<<<<<<
 *                 raise TypeError("parse actions must be callable")
 *             self.parseAction = list(map(_trim_arity, list(fns)))
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_14setParseAction_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1303, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_14setParseAction_2generator12, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_ParserElement_setParseAction_loc, __pyx_n_s_Aspidites__vendor_pyparsing_exte); if (unlikely(!gen)) __PYX_ERR(0, 1303, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.setParseAction.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_14setParseAction_2generator12(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1303, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_fns)) { __Pyx_RaiseClosureNameError("fns"); __PYX_ERR(0, 1303, __pyx_L1_error) }
  if (unlikely(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_fns == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 1303, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_fns; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 1303, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1303, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_fn);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_fn, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_4 = __Pyx_PyCallable_Check(__pyx_cur_scope->__pyx_v_fn); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 1303, __pyx_L1_error)
    __pyx_t_5 = ((!(__pyx_t_4 != 0)) != 0);
    if (__pyx_t_5) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_True);
    __pyx_r = Py_True;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1259
 *         return self
 * 
 *     def setParseAction(self, *fns, **kwargs):             # <<<<<<<<<<<<<<
 *         from .pyparsing import _trim_arity
 *         """
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_18setParseAction(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_fns, PyObject *__pyx_v_kwargs) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction *__pyx_cur_scope;
  PyObject *__pyx_v__trim_arity = NULL;
  PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_14setParseAction_2generator12 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setParseAction", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1259, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_fns = __pyx_v_fns;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_fns);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_fns);

  /* "Aspidites/_vendor/pyparsing_extension.py":1260
 * 
 *     def setParseAction(self, *fns, **kwargs):
 *         from .pyparsing import _trim_arity             # <<<<<<<<<<<<<<
 *         """
 *         Define one or more actions to perform when successfully matching parse element definition.
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_trim_arity);
  __Pyx_GIVEREF(__pyx_n_s_trim_arity);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_trim_arity);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_pyparsing, __pyx_t_1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_trim_arity); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v__trim_arity = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1300
 *             date_str.parseString("1999/12/31")  # -> [1999, '/', 12, '/', 31]
 *         """
 *         if list(fns) == [None, ]:             # <<<<<<<<<<<<<<
 *             self.parseAction = []
 *         else:
 */
  __pyx_t_2 = PySequence_List(__pyx_cur_scope->__pyx_v_fns); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyList_SET_ITEM(__pyx_t_1, 0, Py_None);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1300, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1300, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":1301
 *         """
 *         if list(fns) == [None, ]:
 *             self.parseAction = []             # <<<<<<<<<<<<<<
 *         else:
 *             if not all(callable(fn) for fn in fns):
 */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1301, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parseAction, __pyx_t_3) < 0) __PYX_ERR(0, 1301, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1300
 *             date_str.parseString("1999/12/31")  # -> [1999, '/', 12, '/', 31]
 *         """
 *         if list(fns) == [None, ]:             # <<<<<<<<<<<<<<
 *             self.parseAction = []
 *         else:
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1303
 *             self.parseAction = []
 *         else:
 *             if not all(callable(fn) for fn in fns):             # <<<<<<<<<<<<<<
 *                 raise TypeError("parse actions must be callable")
 *             self.parseAction = list(map(_trim_arity, list(fns)))
 */
  /*else*/ {
    __pyx_t_3 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_14setParseAction_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1303, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_Generator_Next(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1303, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1303, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = ((!__pyx_t_4) != 0);
    if (unlikely(__pyx_t_5)) {

      /* "Aspidites/_vendor/pyparsing_extension.py":1304
 *         else:
 *             if not all(callable(fn) for fn in fns):
 *                 raise TypeError("parse actions must be callable")             # <<<<<<<<<<<<<<
 *             self.parseAction = list(map(_trim_arity, list(fns)))
 *             self.callDuringTry = kwargs.get("callDuringTry", False)
 */
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__38, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1304, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 1304, __pyx_L1_error)

      /* "Aspidites/_vendor/pyparsing_extension.py":1303
 *             self.parseAction = []
 *         else:
 *             if not all(callable(fn) for fn in fns):             # <<<<<<<<<<<<<<
 *                 raise TypeError("parse actions must be callable")
 *             self.parseAction = list(map(_trim_arity, list(fns)))
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":1305
 *             if not all(callable(fn) for fn in fns):
 *                 raise TypeError("parse actions must be callable")
 *             self.parseAction = list(map(_trim_arity, list(fns)))             # <<<<<<<<<<<<<<
 *             self.callDuringTry = kwargs.get("callDuringTry", False)
 *         return self
 */
    __pyx_t_1 = PySequence_List(__pyx_cur_scope->__pyx_v_fns); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1305, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1305, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v__trim_arity);
    __Pyx_GIVEREF(__pyx_v__trim_arity);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v__trim_arity);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_map, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1305, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PySequence_List(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1305, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parseAction, __pyx_t_3) < 0) __PYX_ERR(0, 1305, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1306
 *                 raise TypeError("parse actions must be callable")
 *             self.parseAction = list(map(_trim_arity, list(fns)))
 *             self.callDuringTry = kwargs.get("callDuringTry", False)             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
    __pyx_t_3 = __Pyx_PyDict_GetItemDefault(__pyx_v_kwargs, __pyx_n_s_callDuringTry, Py_False); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1306, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_callDuringTry, __pyx_t_3) < 0) __PYX_ERR(0, 1306, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_L3:;

  /* "Aspidites/_vendor/pyparsing_extension.py":1307
 *             self.parseAction = list(map(_trim_arity, list(fns)))
 *             self.callDuringTry = kwargs.get("callDuringTry", False)
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def addParseAction(self, *fns, **kwargs):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1259
 *         return self
 * 
 *     def setParseAction(self, *fns, **kwargs):             # <<<<<<<<<<<<<<
 *         from .pyparsing import _trim_arity
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.setParseAction", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v__trim_arity);
  __Pyx_XDECREF(__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_14setParseAction_2generator12);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1309
 *         return self
 * 
 *     def addParseAction(self, *fns, **kwargs):             # <<<<<<<<<<<<<<
 *         from .pyparsing import _trim_arity
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_21addParseAction(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_21addParseAction = {"addParseAction", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_21addParseAction, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_21addParseAction(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_fns = 0;
  PyObject *__pyx_v_kwargs = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("addParseAction (wrapper)", 0);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  if (PyTuple_GET_SIZE(__pyx_args) > 1) {
    __pyx_v_fns = PyTuple_GetSlice(__pyx_args, 1, PyTuple_GET_SIZE(__pyx_args));
    if (unlikely(!__pyx_v_fns)) {
      __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
      __Pyx_RefNannyFinishContext();
      return NULL;
    }
    __Pyx_GOTREF(__pyx_v_fns);
  } else {
    __pyx_v_fns = __pyx_empty_tuple; __Pyx_INCREF(__pyx_empty_tuple);
  }
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        default:
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t used_pos_args = (pos_args < 1) ? pos_args : 1;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwargs, values, used_pos_args, "addParseAction") < 0)) __PYX_ERR(0, 1309, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) < 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_self = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("addParseAction", 0, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1309, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_fns); __pyx_v_fns = 0;
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.addParseAction", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_20addParseAction(__pyx_self, __pyx_v_self, __pyx_v_fns, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_fns);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_20addParseAction(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_fns, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_v__trim_arity = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("addParseAction", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":1310
 * 
 *     def addParseAction(self, *fns, **kwargs):
 *         from .pyparsing import _trim_arity             # <<<<<<<<<<<<<<
 *         """
 *         Add one or more parse actions to expression's list of parse actions. See :class:`setParseAction`.
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_trim_arity);
  __Pyx_GIVEREF(__pyx_n_s_trim_arity);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_trim_arity);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_pyparsing, __pyx_t_1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_trim_arity); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v__trim_arity = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1316
 *         See examples in :class:`copy`.
 *         """
 *         self.parseAction += list(map(_trim_arity, list(fns)))             # <<<<<<<<<<<<<<
 *         self.callDuringTry = self.callDuringTry or kwargs.get("callDuringTry", False)
 *         return self
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_parseAction); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PySequence_List(__pyx_v_fns); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v__trim_arity);
  __Pyx_GIVEREF(__pyx_v__trim_arity);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v__trim_arity);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_map, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PySequence_List(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parseAction, __pyx_t_1) < 0) __PYX_ERR(0, 1316, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1317
 *         """
 *         self.parseAction += list(map(_trim_arity, list(fns)))
 *         self.callDuringTry = self.callDuringTry or kwargs.get("callDuringTry", False)             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_callDuringTry); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1317, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1317, __pyx_L1_error)
  if (!__pyx_t_4) {
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_1 = __pyx_t_3;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyDict_GetItemDefault(__pyx_v_kwargs, __pyx_n_s_callDuringTry, Py_False); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1317, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_1 = __pyx_t_3;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_L3_bool_binop_done:;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_callDuringTry, __pyx_t_1) < 0) __PYX_ERR(0, 1317, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1318
 *         self.parseAction += list(map(_trim_arity, list(fns)))
 *         self.callDuringTry = self.callDuringTry or kwargs.get("callDuringTry", False)
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def addCondition(self, *fns, **kwargs):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1309
 *         return self
 * 
 *     def addParseAction(self, *fns, **kwargs):             # <<<<<<<<<<<<<<
 *         from .pyparsing import _trim_arity
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.addParseAction", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v__trim_arity);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1320
 *         return self
 * 
 *     def addCondition(self, *fns, **kwargs):             # <<<<<<<<<<<<<<
 *         from .pyparsing import conditionAsParseAction
 *         """Add a boolean predicate function to expression's list of parse actions. See
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_23addCondition(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_23addCondition = {"addCondition", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_23addCondition, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_23addCondition(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_fns = 0;
  PyObject *__pyx_v_kwargs = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("addCondition (wrapper)", 0);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  if (PyTuple_GET_SIZE(__pyx_args) > 1) {
    __pyx_v_fns = PyTuple_GetSlice(__pyx_args, 1, PyTuple_GET_SIZE(__pyx_args));
    if (unlikely(!__pyx_v_fns)) {
      __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
      __Pyx_RefNannyFinishContext();
      return NULL;
    }
    __Pyx_GOTREF(__pyx_v_fns);
  } else {
    __pyx_v_fns = __pyx_empty_tuple; __Pyx_INCREF(__pyx_empty_tuple);
  }
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        default:
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t used_pos_args = (pos_args < 1) ? pos_args : 1;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwargs, values, used_pos_args, "addCondition") < 0)) __PYX_ERR(0, 1320, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) < 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_self = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("addCondition", 0, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1320, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_fns); __pyx_v_fns = 0;
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.addCondition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_22addCondition(__pyx_self, __pyx_v_self, __pyx_v_fns, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_fns);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_22addCondition(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_fns, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_v_conditionAsParseAction = NULL;
  PyObject *__pyx_v_fn = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("addCondition", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":1321
 * 
 *     def addCondition(self, *fns, **kwargs):
 *         from .pyparsing import conditionAsParseAction             # <<<<<<<<<<<<<<
 *         """Add a boolean predicate function to expression's list of parse actions. See
 *         :class:`setParseAction` for function call signatures. Unlike ``setParseAction``,
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_conditionAsParseAction);
  __Pyx_GIVEREF(__pyx_n_s_conditionAsParseAction);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_conditionAsParseAction);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_pyparsing, __pyx_t_1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_conditionAsParseAction); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_conditionAsParseAction = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1339
 *             result = date_str.parseString("1999/12/31")  # -> Exception: Only support years 2000 and later (at char 0), (line:1, col:1)
 *         """
 *         for fn in fns:             # <<<<<<<<<<<<<<
 *             self.parseAction.append(conditionAsParseAction(fn, message=kwargs.get('message'),
 *                                                            fatal=kwargs.get('fatal', False)))
 */
  __pyx_t_2 = __pyx_v_fns; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
  for (;;) {
    if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 1339, __pyx_L1_error)
    #else
    __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1339, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_fn, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1340
 *         """
 *         for fn in fns:
 *             self.parseAction.append(conditionAsParseAction(fn, message=kwargs.get('message'),             # <<<<<<<<<<<<<<
 *                                                            fatal=kwargs.get('fatal', False)))
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_parseAction); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1340, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1340, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_fn);
    __Pyx_GIVEREF(__pyx_v_fn);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_fn);
    __pyx_t_5 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1340, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyDict_GetItemDefault(__pyx_v_kwargs, __pyx_n_s_message, Py_None); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1340, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_message, __pyx_t_6) < 0) __PYX_ERR(0, 1340, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1341
 *         for fn in fns:
 *             self.parseAction.append(conditionAsParseAction(fn, message=kwargs.get('message'),
 *                                                            fatal=kwargs.get('fatal', False)))             # <<<<<<<<<<<<<<
 * 
 *         self.callDuringTry = self.callDuringTry or kwargs.get("callDuringTry", False)
 */
    __pyx_t_6 = __Pyx_PyDict_GetItemDefault(__pyx_v_kwargs, __pyx_n_s_fatal, Py_False); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1341, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_fatal, __pyx_t_6) < 0) __PYX_ERR(0, 1340, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1340
 *         """
 *         for fn in fns:
 *             self.parseAction.append(conditionAsParseAction(fn, message=kwargs.get('message'),             # <<<<<<<<<<<<<<
 *                                                            fatal=kwargs.get('fatal', False)))
 * 
 */
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_v_conditionAsParseAction, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1340, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_7 = __Pyx_PyObject_Append(__pyx_t_1, __pyx_t_6); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 1340, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1339
 *             result = date_str.parseString("1999/12/31")  # -> Exception: Only support years 2000 and later (at char 0), (line:1, col:1)
 *         """
 *         for fn in fns:             # <<<<<<<<<<<<<<
 *             self.parseAction.append(conditionAsParseAction(fn, message=kwargs.get('message'),
 *                                                            fatal=kwargs.get('fatal', False)))
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1343
 *                                                            fatal=kwargs.get('fatal', False)))
 * 
 *         self.callDuringTry = self.callDuringTry or kwargs.get("callDuringTry", False)             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_callDuringTry); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 1343, __pyx_L1_error)
  if (!__pyx_t_8) {
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_6);
    __pyx_t_2 = __pyx_t_6;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_6 = __Pyx_PyDict_GetItemDefault(__pyx_v_kwargs, __pyx_n_s_callDuringTry, Py_False); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_t_6);
  __pyx_t_2 = __pyx_t_6;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_L5_bool_binop_done:;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_callDuringTry, __pyx_t_2) < 0) __PYX_ERR(0, 1343, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1344
 * 
 *         self.callDuringTry = self.callDuringTry or kwargs.get("callDuringTry", False)
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def setFailAction(self, fn):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1320
 *         return self
 * 
 *     def addCondition(self, *fns, **kwargs):             # <<<<<<<<<<<<<<
 *         from .pyparsing import conditionAsParseAction
 *         """Add a boolean predicate function to expression's list of parse actions. See
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.addCondition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_conditionAsParseAction);
  __Pyx_XDECREF(__pyx_v_fn);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1346
 *         return self
 * 
 *     def setFailAction(self, fn):             # <<<<<<<<<<<<<<
 *         """Define action to perform if parsing fails at this expression.
 *            Fail acton fn is a callable function that takes the arguments
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_25setFailAction(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_24setFailAction[] = "Define action to perform if parsing fails at this expression.\n           Fail acton fn is a callable function that takes the arguments\n           ``fn(s, loc, expr, err)`` where:\n           - s = string being parsed\n           - loc = location where expression match was attempted and failed\n           - expr = the parse expression that failed\n           - err = the exception thrown\n           The function returns no value.  It may throw :class:`ParseFatalException`\n           if it is desired to stop parsing immediately.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_25setFailAction = {"setFailAction", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_25setFailAction, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_24setFailAction};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_25setFailAction(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_fn = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setFailAction (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_fn,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fn)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setFailAction", 1, 2, 2, 1); __PYX_ERR(0, 1346, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setFailAction") < 0)) __PYX_ERR(0, 1346, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_fn = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setFailAction", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1346, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.setFailAction", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_24setFailAction(__pyx_self, __pyx_v_self, __pyx_v_fn);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_24setFailAction(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_fn) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setFailAction", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":1356
 *            The function returns no value.  It may throw :class:`ParseFatalException`
 *            if it is desired to stop parsing immediately."""
 *         self.failAction = fn             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_failAction, __pyx_v_fn) < 0) __PYX_ERR(0, 1356, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1357
 *            if it is desired to stop parsing immediately."""
 *         self.failAction = fn
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def _skipIgnorables(self, instring, loc):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1346
 *         return self
 * 
 *     def setFailAction(self, fn):             # <<<<<<<<<<<<<<
 *         """Define action to perform if parsing fails at this expression.
 *            Fail acton fn is a callable function that takes the arguments
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.setFailAction", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1359
 *         return self
 * 
 *     def _skipIgnorables(self, instring, loc):             # <<<<<<<<<<<<<<
 *         exprsFound = True
 *         while exprsFound:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_27_skipIgnorables(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_27_skipIgnorables = {"_skipIgnorables", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_27_skipIgnorables, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_27_skipIgnorables(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_skipIgnorables (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_skipIgnorables", 1, 3, 3, 1); __PYX_ERR(0, 1359, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_skipIgnorables", 1, 3, 3, 2); __PYX_ERR(0, 1359, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_skipIgnorables") < 0)) __PYX_ERR(0, 1359, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_skipIgnorables", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1359, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement._skipIgnorables", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_26_skipIgnorables(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_26_skipIgnorables(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc) {
  int __pyx_v_exprsFound;
  PyObject *__pyx_v_e = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_dummy = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *(*__pyx_t_13)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_skipIgnorables", 0);
  __Pyx_INCREF(__pyx_v_loc);

  /* "Aspidites/_vendor/pyparsing_extension.py":1360
 * 
 *     def _skipIgnorables(self, instring, loc):
 *         exprsFound = True             # <<<<<<<<<<<<<<
 *         while exprsFound:
 *             exprsFound = False
 */
  __pyx_v_exprsFound = 1;

  /* "Aspidites/_vendor/pyparsing_extension.py":1361
 *     def _skipIgnorables(self, instring, loc):
 *         exprsFound = True
 *         while exprsFound:             # <<<<<<<<<<<<<<
 *             exprsFound = False
 *             for e in self.ignoreExprs:
 */
  while (1) {
    __pyx_t_1 = (__pyx_v_exprsFound != 0);
    if (!__pyx_t_1) break;

    /* "Aspidites/_vendor/pyparsing_extension.py":1362
 *         exprsFound = True
 *         while exprsFound:
 *             exprsFound = False             # <<<<<<<<<<<<<<
 *             for e in self.ignoreExprs:
 *                 try:
 */
    __pyx_v_exprsFound = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1363
 *         while exprsFound:
 *             exprsFound = False
 *             for e in self.ignoreExprs:             # <<<<<<<<<<<<<<
 *                 try:
 *                     while 1:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ignoreExprs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1363, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3); __pyx_t_4 = 0;
      __pyx_t_5 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1363, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1363, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_5)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 1363, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1363, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 1363, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1363, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_5(__pyx_t_3);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1363, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":1364
 *             exprsFound = False
 *             for e in self.ignoreExprs:
 *                 try:             # <<<<<<<<<<<<<<
 *                     while 1:
 *                         loc, dummy = e._parse(instring, loc)
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        /*try:*/ {

          /* "Aspidites/_vendor/pyparsing_extension.py":1365
 *             for e in self.ignoreExprs:
 *                 try:
 *                     while 1:             # <<<<<<<<<<<<<<
 *                         loc, dummy = e._parse(instring, loc)
 *                         exprsFound = True
 */
          while (1) {

            /* "Aspidites/_vendor/pyparsing_extension.py":1366
 *                 try:
 *                     while 1:
 *                         loc, dummy = e._parse(instring, loc)             # <<<<<<<<<<<<<<
 *                         exprsFound = True
 *                 except ParseException:
 */
            __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_parse); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1366, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_9);
            __pyx_t_10 = NULL;
            __pyx_t_11 = 0;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
              __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
              if (likely(__pyx_t_10)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
                __Pyx_INCREF(__pyx_t_10);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_9, function);
                __pyx_t_11 = 1;
              }
            }
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_9)) {
              PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_v_instring, __pyx_v_loc};
              __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1366, __pyx_L7_error)
              __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
              __Pyx_GOTREF(__pyx_t_2);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_9)) {
              PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_v_instring, __pyx_v_loc};
              __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1366, __pyx_L7_error)
              __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
              __Pyx_GOTREF(__pyx_t_2);
            } else
            #endif
            {
              __pyx_t_12 = PyTuple_New(2+__pyx_t_11); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1366, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_12);
              if (__pyx_t_10) {
                __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_10); __pyx_t_10 = NULL;
              }
              __Pyx_INCREF(__pyx_v_instring);
              __Pyx_GIVEREF(__pyx_v_instring);
              PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_11, __pyx_v_instring);
              __Pyx_INCREF(__pyx_v_loc);
              __Pyx_GIVEREF(__pyx_v_loc);
              PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_11, __pyx_v_loc);
              __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_12, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1366, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            }
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
            if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
              PyObject* sequence = __pyx_t_2;
              Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
              if (unlikely(size != 2)) {
                if (size > 2) __Pyx_RaiseTooManyValuesError(2);
                else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                __PYX_ERR(0, 1366, __pyx_L7_error)
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              if (likely(PyTuple_CheckExact(sequence))) {
                __pyx_t_9 = PyTuple_GET_ITEM(sequence, 0); 
                __pyx_t_12 = PyTuple_GET_ITEM(sequence, 1); 
              } else {
                __pyx_t_9 = PyList_GET_ITEM(sequence, 0); 
                __pyx_t_12 = PyList_GET_ITEM(sequence, 1); 
              }
              __Pyx_INCREF(__pyx_t_9);
              __Pyx_INCREF(__pyx_t_12);
              #else
              __pyx_t_9 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1366, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_9);
              __pyx_t_12 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1366, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_12);
              #endif
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            } else {
              Py_ssize_t index = -1;
              __pyx_t_10 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1366, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_10);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __pyx_t_13 = Py_TYPE(__pyx_t_10)->tp_iternext;
              index = 0; __pyx_t_9 = __pyx_t_13(__pyx_t_10); if (unlikely(!__pyx_t_9)) goto __pyx_L17_unpacking_failed;
              __Pyx_GOTREF(__pyx_t_9);
              index = 1; __pyx_t_12 = __pyx_t_13(__pyx_t_10); if (unlikely(!__pyx_t_12)) goto __pyx_L17_unpacking_failed;
              __Pyx_GOTREF(__pyx_t_12);
              if (__Pyx_IternextUnpackEndCheck(__pyx_t_13(__pyx_t_10), 2) < 0) __PYX_ERR(0, 1366, __pyx_L7_error)
              __pyx_t_13 = NULL;
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              goto __pyx_L18_unpacking_done;
              __pyx_L17_unpacking_failed:;
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __pyx_t_13 = NULL;
              if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
              __PYX_ERR(0, 1366, __pyx_L7_error)
              __pyx_L18_unpacking_done:;
            }
            __Pyx_DECREF_SET(__pyx_v_loc, __pyx_t_9);
            __pyx_t_9 = 0;
            __Pyx_XDECREF_SET(__pyx_v_dummy, __pyx_t_12);
            __pyx_t_12 = 0;

            /* "Aspidites/_vendor/pyparsing_extension.py":1367
 *                     while 1:
 *                         loc, dummy = e._parse(instring, loc)
 *                         exprsFound = True             # <<<<<<<<<<<<<<
 *                 except ParseException:
 *                     pass
 */
            __pyx_v_exprsFound = 1;
          }

          /* "Aspidites/_vendor/pyparsing_extension.py":1364
 *             exprsFound = False
 *             for e in self.ignoreExprs:
 *                 try:             # <<<<<<<<<<<<<<
 *                     while 1:
 *                         loc, dummy = e._parse(instring, loc)
 */
        }
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L14_try_end;
        __pyx_L7_error:;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":1368
 *                         loc, dummy = e._parse(instring, loc)
 *                         exprsFound = True
 *                 except ParseException:             # <<<<<<<<<<<<<<
 *                     pass
 *         return loc
 */
        __Pyx_ErrFetch(&__pyx_t_2, &__pyx_t_12, &__pyx_t_9);
        __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1368, __pyx_L9_except_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_11 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_2, __pyx_t_10);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_ErrRestore(__pyx_t_2, __pyx_t_12, __pyx_t_9);
        __pyx_t_2 = 0; __pyx_t_12 = 0; __pyx_t_9 = 0;
        if (__pyx_t_11) {
          __Pyx_ErrRestore(0,0,0);
          goto __pyx_L8_exception_handled;
        }
        goto __pyx_L9_except_error;
        __pyx_L9_except_error:;

        /* "Aspidites/_vendor/pyparsing_extension.py":1364
 *             exprsFound = False
 *             for e in self.ignoreExprs:
 *                 try:             # <<<<<<<<<<<<<<
 *                     while 1:
 *                         loc, dummy = e._parse(instring, loc)
 */
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
        goto __pyx_L1_error;
        __pyx_L8_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
        __pyx_L14_try_end:;
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":1363
 *         while exprsFound:
 *             exprsFound = False
 *             for e in self.ignoreExprs:             # <<<<<<<<<<<<<<
 *                 try:
 *                     while 1:
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1370
 *                 except ParseException:
 *                     pass
 *         return loc             # <<<<<<<<<<<<<<
 * 
 *     def preParse(self, instring, loc):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_loc);
  __pyx_r = __pyx_v_loc;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1359
 *         return self
 * 
 *     def _skipIgnorables(self, instring, loc):             # <<<<<<<<<<<<<<
 *         exprsFound = True
 *         while exprsFound:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement._skipIgnorables", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XDECREF(__pyx_v_dummy);
  __Pyx_XDECREF(__pyx_v_loc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1372
 *         return loc
 * 
 *     def preParse(self, instring, loc):             # <<<<<<<<<<<<<<
 *         if self.ignoreExprs:
 *             loc = self._skipIgnorables(instring, loc)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_29preParse(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_29preParse = {"preParse", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_29preParse, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_29preParse(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("preParse (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("preParse", 1, 3, 3, 1); __PYX_ERR(0, 1372, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("preParse", 1, 3, 3, 2); __PYX_ERR(0, 1372, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "preParse") < 0)) __PYX_ERR(0, 1372, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("preParse", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1372, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.preParse", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_28preParse(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_28preParse(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc) {
  PyObject *__pyx_v_wt = NULL;
  Py_ssize_t __pyx_v_instrlen;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("preParse", 0);
  __Pyx_INCREF(__pyx_v_loc);

  /* "Aspidites/_vendor/pyparsing_extension.py":1373
 * 
 *     def preParse(self, instring, loc):
 *         if self.ignoreExprs:             # <<<<<<<<<<<<<<
 *             loc = self._skipIgnorables(instring, loc)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ignoreExprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1373, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":1374
 *     def preParse(self, instring, loc):
 *         if self.ignoreExprs:
 *             loc = self._skipIgnorables(instring, loc)             # <<<<<<<<<<<<<<
 * 
 *         if self.skipWhitespace:
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_skipIgnorables); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1374, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_instring, __pyx_v_loc};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1374, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_instring, __pyx_v_loc};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1374, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1374, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_4) {
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
      }
      __Pyx_INCREF(__pyx_v_instring);
      __Pyx_GIVEREF(__pyx_v_instring);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_instring);
      __Pyx_INCREF(__pyx_v_loc);
      __Pyx_GIVEREF(__pyx_v_loc);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_loc);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1374, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_loc, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1373
 * 
 *     def preParse(self, instring, loc):
 *         if self.ignoreExprs:             # <<<<<<<<<<<<<<
 *             loc = self._skipIgnorables(instring, loc)
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1376
 *             loc = self._skipIgnorables(instring, loc)
 * 
 *         if self.skipWhitespace:             # <<<<<<<<<<<<<<
 *             wt = self.whiteChars
 *             instrlen = len(instring)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_skipWhitespace); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1376, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":1377
 * 
 *         if self.skipWhitespace:
 *             wt = self.whiteChars             # <<<<<<<<<<<<<<
 *             instrlen = len(instring)
 *             while loc < instrlen and instring[loc] in wt:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_whiteChars); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1377, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_wt = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1378
 *         if self.skipWhitespace:
 *             wt = self.whiteChars
 *             instrlen = len(instring)             # <<<<<<<<<<<<<<
 *             while loc < instrlen and instring[loc] in wt:
 *                 loc += 1
 */
    __pyx_t_7 = PyObject_Length(__pyx_v_instring); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1378, __pyx_L1_error)
    __pyx_v_instrlen = __pyx_t_7;

    /* "Aspidites/_vendor/pyparsing_extension.py":1379
 *             wt = self.whiteChars
 *             instrlen = len(instring)
 *             while loc < instrlen and instring[loc] in wt:             # <<<<<<<<<<<<<<
 *                 loc += 1
 * 
 */
    while (1) {
      __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_instrlen); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1379, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_loc, __pyx_t_1, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1379, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 1379, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_8) {
      } else {
        __pyx_t_2 = __pyx_t_8;
        goto __pyx_L7_bool_binop_done;
      }
      __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_instring, __pyx_v_loc); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1379, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = (__Pyx_PySequence_ContainsTF(__pyx_t_3, __pyx_v_wt, Py_EQ)); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 1379, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_9 = (__pyx_t_8 != 0);
      __pyx_t_2 = __pyx_t_9;
      __pyx_L7_bool_binop_done:;
      if (!__pyx_t_2) break;

      /* "Aspidites/_vendor/pyparsing_extension.py":1380
 *             instrlen = len(instring)
 *             while loc < instrlen and instring[loc] in wt:
 *                 loc += 1             # <<<<<<<<<<<<<<
 * 
 *         return loc
 */
      __pyx_t_3 = __Pyx_PyInt_AddObjC(__pyx_v_loc, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1380, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF_SET(__pyx_v_loc, __pyx_t_3);
      __pyx_t_3 = 0;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":1376
 *             loc = self._skipIgnorables(instring, loc)
 * 
 *         if self.skipWhitespace:             # <<<<<<<<<<<<<<
 *             wt = self.whiteChars
 *             instrlen = len(instring)
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1382
 *                 loc += 1
 * 
 *         return loc             # <<<<<<<<<<<<<<
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_loc);
  __pyx_r = __pyx_v_loc;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1372
 *         return loc
 * 
 *     def preParse(self, instring, loc):             # <<<<<<<<<<<<<<
 *         if self.ignoreExprs:
 *             loc = self._skipIgnorables(instring, loc)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.preParse", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_wt);
  __Pyx_XDECREF(__pyx_v_loc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1384
 *         return loc
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         return loc, []
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_31parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_31parseImpl = {"parseImpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_31parseImpl, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_31parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  CYTHON_UNUSED PyObject *__pyx_v_doActions = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parseImpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_doActions,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 1); __PYX_ERR(0, 1384, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 2); __PYX_ERR(0, 1384, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_doActions);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "parseImpl") < 0)) __PYX_ERR(0, 1384, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_doActions = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1384, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_30parseImpl(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_30parseImpl(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parseImpl", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":1385
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         return loc, []             # <<<<<<<<<<<<<<
 * 
 *     def postParse(self, instring, loc, tokenlist):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_loc);
  __Pyx_GIVEREF(__pyx_v_loc);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_loc);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1384
 *         return loc
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         return loc, []
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1387
 *         return loc, []
 * 
 *     def postParse(self, instring, loc, tokenlist):             # <<<<<<<<<<<<<<
 *         return tokenlist
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_33postParse(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_33postParse = {"postParse", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_33postParse, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_33postParse(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_instring = 0;
  CYTHON_UNUSED PyObject *__pyx_v_loc = 0;
  PyObject *__pyx_v_tokenlist = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("postParse (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_tokenlist,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("postParse", 1, 4, 4, 1); __PYX_ERR(0, 1387, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("postParse", 1, 4, 4, 2); __PYX_ERR(0, 1387, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tokenlist)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("postParse", 1, 4, 4, 3); __PYX_ERR(0, 1387, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "postParse") < 0)) __PYX_ERR(0, 1387, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_tokenlist = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("postParse", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1387, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.postParse", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_32postParse(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_tokenlist);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_32postParse(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_instring, CYTHON_UNUSED PyObject *__pyx_v_loc, PyObject *__pyx_v_tokenlist) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("postParse", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":1388
 * 
 *     def postParse(self, instring, loc, tokenlist):
 *         return tokenlist             # <<<<<<<<<<<<<<
 * 
 *     # ~ @profile
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_tokenlist);
  __pyx_r = __pyx_v_tokenlist;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1387
 *         return loc, []
 * 
 *     def postParse(self, instring, loc, tokenlist):             # <<<<<<<<<<<<<<
 *         return tokenlist
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1391
 * 
 *     # ~ @profile
 *     def _parseNoCache(self, instring, loc, doActions=True, callPreParse=True):             # <<<<<<<<<<<<<<
 *         TRY, MATCH, FAIL = 0, 1, 2
 *         debugging = (self.debug)  # and doActions)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_35_parseNoCache(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_35_parseNoCache = {"_parseNoCache", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_35_parseNoCache, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_35_parseNoCache(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  PyObject *__pyx_v_doActions = 0;
  PyObject *__pyx_v_callPreParse = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_parseNoCache (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_doActions,&__pyx_n_s_callPreParse,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[3] = ((PyObject *)((PyObject *)Py_True));
    values[4] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_parseNoCache", 0, 3, 5, 1); __PYX_ERR(0, 1391, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_parseNoCache", 0, 3, 5, 2); __PYX_ERR(0, 1391, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_doActions);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_callPreParse);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_parseNoCache") < 0)) __PYX_ERR(0, 1391, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_doActions = values[3];
    __pyx_v_callPreParse = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_parseNoCache", 0, 3, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1391, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement._parseNoCache", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_34_parseNoCache(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions, __pyx_v_callPreParse);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_34_parseNoCache(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, PyObject *__pyx_v_doActions, PyObject *__pyx_v_callPreParse) {
  long __pyx_v_TRY;
  long __pyx_v_MATCH;
  long __pyx_v_FAIL;
  PyObject *__pyx_v_debugging = NULL;
  PyObject *__pyx_v_preloc = NULL;
  PyObject *__pyx_v_tokensStart = NULL;
  PyObject *__pyx_v_tokens = NULL;
  PyObject *__pyx_v_err = NULL;
  PyObject *__pyx_v_retTokens = NULL;
  PyObject *__pyx_v_fn = NULL;
  PyObject *__pyx_v_parse_action_exc = NULL;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  long __pyx_t_2;
  long __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  Py_ssize_t __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *(*__pyx_t_18)(PyObject *);
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  PyObject *(*__pyx_t_24)(PyObject *);
  int __pyx_t_25;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_parseNoCache", 0);
  __Pyx_INCREF(__pyx_v_loc);

  /* "Aspidites/_vendor/pyparsing_extension.py":1392
 *     # ~ @profile
 *     def _parseNoCache(self, instring, loc, doActions=True, callPreParse=True):
 *         TRY, MATCH, FAIL = 0, 1, 2             # <<<<<<<<<<<<<<
 *         debugging = (self.debug)  # and doActions)
 * 
 */
  __pyx_t_1 = 0;
  __pyx_t_2 = 1;
  __pyx_t_3 = 2;
  __pyx_v_TRY = __pyx_t_1;
  __pyx_v_MATCH = __pyx_t_2;
  __pyx_v_FAIL = __pyx_t_3;

  /* "Aspidites/_vendor/pyparsing_extension.py":1393
 *     def _parseNoCache(self, instring, loc, doActions=True, callPreParse=True):
 *         TRY, MATCH, FAIL = 0, 1, 2
 *         debugging = (self.debug)  # and doActions)             # <<<<<<<<<<<<<<
 * 
 *         if debugging or self.failAction:
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_debug); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1393, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_debugging = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1395
 *         debugging = (self.debug)  # and doActions)
 * 
 *         if debugging or self.failAction:             # <<<<<<<<<<<<<<
 *             # ~ print ("Match", self, "at loc", loc, "(%d, %d)" % (lineno(loc, instring), col(loc, instring)))
 *             if self.debugActions[TRY]:
 */
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_debugging); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1395, __pyx_L1_error)
  if (!__pyx_t_6) {
  } else {
    __pyx_t_5 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_failAction); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1395, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_5 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/pyparsing_extension.py":1397
 *         if debugging or self.failAction:
 *             # ~ print ("Match", self, "at loc", loc, "(%d, %d)" % (lineno(loc, instring), col(loc, instring)))
 *             if self.debugActions[TRY]:             # <<<<<<<<<<<<<<
 *                 self.debugActions[TRY](instring, loc, self)
 *             try:
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_debugActions); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_4, __pyx_v_TRY, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1397, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (__pyx_t_5) {

      /* "Aspidites/_vendor/pyparsing_extension.py":1398
 *             # ~ print ("Match", self, "at loc", loc, "(%d, %d)" % (lineno(loc, instring), col(loc, instring)))
 *             if self.debugActions[TRY]:
 *                 self.debugActions[TRY](instring, loc, self)             # <<<<<<<<<<<<<<
 *             try:
 *                 if callPreParse and self.callPreparse:
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_debugActions); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1398, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = __Pyx_GetItemInt(__pyx_t_4, __pyx_v_TRY, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1398, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = NULL;
      __pyx_t_9 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
          __pyx_t_9 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_8)) {
        PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_instring, __pyx_v_loc, __pyx_v_self};
        __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1398, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_7);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
        PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_instring, __pyx_v_loc, __pyx_v_self};
        __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1398, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_7);
      } else
      #endif
      {
        __pyx_t_10 = PyTuple_New(3+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1398, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        if (__pyx_t_4) {
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_4); __pyx_t_4 = NULL;
        }
        __Pyx_INCREF(__pyx_v_instring);
        __Pyx_GIVEREF(__pyx_v_instring);
        PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_v_instring);
        __Pyx_INCREF(__pyx_v_loc);
        __Pyx_GIVEREF(__pyx_v_loc);
        PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_v_loc);
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_10, 2+__pyx_t_9, __pyx_v_self);
        __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_10, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1398, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":1397
 *         if debugging or self.failAction:
 *             # ~ print ("Match", self, "at loc", loc, "(%d, %d)" % (lineno(loc, instring), col(loc, instring)))
 *             if self.debugActions[TRY]:             # <<<<<<<<<<<<<<
 *                 self.debugActions[TRY](instring, loc, self)
 *             try:
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":1399
 *             if self.debugActions[TRY]:
 *                 self.debugActions[TRY](instring, loc, self)
 *             try:             # <<<<<<<<<<<<<<
 *                 if callPreParse and self.callPreparse:
 *                     preloc = self.preParse(instring, loc)
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      /*try:*/ {

        /* "Aspidites/_vendor/pyparsing_extension.py":1400
 *                 self.debugActions[TRY](instring, loc, self)
 *             try:
 *                 if callPreParse and self.callPreparse:             # <<<<<<<<<<<<<<
 *                     preloc = self.preParse(instring, loc)
 *                 else:
 */
        __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_callPreParse); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1400, __pyx_L7_error)
        if (__pyx_t_6) {
        } else {
          __pyx_t_5 = __pyx_t_6;
          goto __pyx_L14_bool_binop_done;
        }
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_callPreparse); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1400, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1400, __pyx_L7_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_5 = __pyx_t_6;
        __pyx_L14_bool_binop_done:;
        if (__pyx_t_5) {

          /* "Aspidites/_vendor/pyparsing_extension.py":1401
 *             try:
 *                 if callPreParse and self.callPreparse:
 *                     preloc = self.preParse(instring, loc)             # <<<<<<<<<<<<<<
 *                 else:
 *                     preloc = loc
 */
          __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_preParse); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1401, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_10 = NULL;
          __pyx_t_9 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
            __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_8);
            if (likely(__pyx_t_10)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
              __Pyx_INCREF(__pyx_t_10);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_8, function);
              __pyx_t_9 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_8)) {
            PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_v_instring, __pyx_v_loc};
            __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1401, __pyx_L7_error)
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
            PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_v_instring, __pyx_v_loc};
            __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1401, __pyx_L7_error)
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1401, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (__pyx_t_10) {
              __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_10); __pyx_t_10 = NULL;
            }
            __Pyx_INCREF(__pyx_v_instring);
            __Pyx_GIVEREF(__pyx_v_instring);
            PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_9, __pyx_v_instring);
            __Pyx_INCREF(__pyx_v_loc);
            __Pyx_GIVEREF(__pyx_v_loc);
            PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_9, __pyx_v_loc);
            __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_4, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1401, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __pyx_v_preloc = __pyx_t_7;
          __pyx_t_7 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":1400
 *                 self.debugActions[TRY](instring, loc, self)
 *             try:
 *                 if callPreParse and self.callPreparse:             # <<<<<<<<<<<<<<
 *                     preloc = self.preParse(instring, loc)
 *                 else:
 */
          goto __pyx_L13;
        }

        /* "Aspidites/_vendor/pyparsing_extension.py":1403
 *                     preloc = self.preParse(instring, loc)
 *                 else:
 *                     preloc = loc             # <<<<<<<<<<<<<<
 *                 tokensStart = preloc
 *                 if self.mayIndexError or preloc >= len(instring):
 */
        /*else*/ {
          __Pyx_INCREF(__pyx_v_loc);
          __pyx_v_preloc = __pyx_v_loc;
        }
        __pyx_L13:;

        /* "Aspidites/_vendor/pyparsing_extension.py":1404
 *                 else:
 *                     preloc = loc
 *                 tokensStart = preloc             # <<<<<<<<<<<<<<
 *                 if self.mayIndexError or preloc >= len(instring):
 *                     try:
 */
        __Pyx_INCREF(__pyx_v_preloc);
        __pyx_v_tokensStart = __pyx_v_preloc;

        /* "Aspidites/_vendor/pyparsing_extension.py":1405
 *                     preloc = loc
 *                 tokensStart = preloc
 *                 if self.mayIndexError or preloc >= len(instring):             # <<<<<<<<<<<<<<
 *                     try:
 *                         loc, tokens = self.parseImpl(instring, preloc, doActions)
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_mayIndexError); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1405, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1405, __pyx_L7_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (!__pyx_t_6) {
        } else {
          __pyx_t_5 = __pyx_t_6;
          goto __pyx_L17_bool_binop_done;
        }
        __pyx_t_14 = PyObject_Length(__pyx_v_instring); if (unlikely(__pyx_t_14 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1405, __pyx_L7_error)
        __pyx_t_7 = PyInt_FromSsize_t(__pyx_t_14); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1405, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = PyObject_RichCompare(__pyx_v_preloc, __pyx_t_7, Py_GE); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1405, __pyx_L7_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1405, __pyx_L7_error)
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_5 = __pyx_t_6;
        __pyx_L17_bool_binop_done:;
        if (__pyx_t_5) {

          /* "Aspidites/_vendor/pyparsing_extension.py":1406
 *                 tokensStart = preloc
 *                 if self.mayIndexError or preloc >= len(instring):
 *                     try:             # <<<<<<<<<<<<<<
 *                         loc, tokens = self.parseImpl(instring, preloc, doActions)
 *                     except IndexError:
 */
          {
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __Pyx_ExceptionSave(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            /*try:*/ {

              /* "Aspidites/_vendor/pyparsing_extension.py":1407
 *                 if self.mayIndexError or preloc >= len(instring):
 *                     try:
 *                         loc, tokens = self.parseImpl(instring, preloc, doActions)             # <<<<<<<<<<<<<<
 *                     except IndexError:
 *                         raise ParseException(instring, len(instring), self.errmsg, self)
 */
              __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_parseImpl); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1407, __pyx_L19_error)
              __Pyx_GOTREF(__pyx_t_7);
              __pyx_t_4 = NULL;
              __pyx_t_9 = 0;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
                __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
                if (likely(__pyx_t_4)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
                  __Pyx_INCREF(__pyx_t_4);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_7, function);
                  __pyx_t_9 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_7)) {
                PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_instring, __pyx_v_preloc, __pyx_v_doActions};
                __pyx_t_8 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1407, __pyx_L19_error)
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_GOTREF(__pyx_t_8);
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
                PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_instring, __pyx_v_preloc, __pyx_v_doActions};
                __pyx_t_8 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1407, __pyx_L19_error)
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_GOTREF(__pyx_t_8);
              } else
              #endif
              {
                __pyx_t_10 = PyTuple_New(3+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1407, __pyx_L19_error)
                __Pyx_GOTREF(__pyx_t_10);
                if (__pyx_t_4) {
                  __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_4); __pyx_t_4 = NULL;
                }
                __Pyx_INCREF(__pyx_v_instring);
                __Pyx_GIVEREF(__pyx_v_instring);
                PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_v_instring);
                __Pyx_INCREF(__pyx_v_preloc);
                __Pyx_GIVEREF(__pyx_v_preloc);
                PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_v_preloc);
                __Pyx_INCREF(__pyx_v_doActions);
                __Pyx_GIVEREF(__pyx_v_doActions);
                PyTuple_SET_ITEM(__pyx_t_10, 2+__pyx_t_9, __pyx_v_doActions);
                __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_10, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1407, __pyx_L19_error)
                __Pyx_GOTREF(__pyx_t_8);
                __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              }
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              if ((likely(PyTuple_CheckExact(__pyx_t_8))) || (PyList_CheckExact(__pyx_t_8))) {
                PyObject* sequence = __pyx_t_8;
                Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
                if (unlikely(size != 2)) {
                  if (size > 2) __Pyx_RaiseTooManyValuesError(2);
                  else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                  __PYX_ERR(0, 1407, __pyx_L19_error)
                }
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                if (likely(PyTuple_CheckExact(sequence))) {
                  __pyx_t_7 = PyTuple_GET_ITEM(sequence, 0); 
                  __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
                } else {
                  __pyx_t_7 = PyList_GET_ITEM(sequence, 0); 
                  __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
                }
                __Pyx_INCREF(__pyx_t_7);
                __Pyx_INCREF(__pyx_t_10);
                #else
                __pyx_t_7 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1407, __pyx_L19_error)
                __Pyx_GOTREF(__pyx_t_7);
                __pyx_t_10 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1407, __pyx_L19_error)
                __Pyx_GOTREF(__pyx_t_10);
                #endif
                __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              } else {
                Py_ssize_t index = -1;
                __pyx_t_4 = PyObject_GetIter(__pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1407, __pyx_L19_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                __pyx_t_18 = Py_TYPE(__pyx_t_4)->tp_iternext;
                index = 0; __pyx_t_7 = __pyx_t_18(__pyx_t_4); if (unlikely(!__pyx_t_7)) goto __pyx_L25_unpacking_failed;
                __Pyx_GOTREF(__pyx_t_7);
                index = 1; __pyx_t_10 = __pyx_t_18(__pyx_t_4); if (unlikely(!__pyx_t_10)) goto __pyx_L25_unpacking_failed;
                __Pyx_GOTREF(__pyx_t_10);
                if (__Pyx_IternextUnpackEndCheck(__pyx_t_18(__pyx_t_4), 2) < 0) __PYX_ERR(0, 1407, __pyx_L19_error)
                __pyx_t_18 = NULL;
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                goto __pyx_L26_unpacking_done;
                __pyx_L25_unpacking_failed:;
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                __pyx_t_18 = NULL;
                if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
                __PYX_ERR(0, 1407, __pyx_L19_error)
                __pyx_L26_unpacking_done:;
              }
              __Pyx_DECREF_SET(__pyx_v_loc, __pyx_t_7);
              __pyx_t_7 = 0;
              __pyx_v_tokens = __pyx_t_10;
              __pyx_t_10 = 0;

              /* "Aspidites/_vendor/pyparsing_extension.py":1406
 *                 tokensStart = preloc
 *                 if self.mayIndexError or preloc >= len(instring):
 *                     try:             # <<<<<<<<<<<<<<
 *                         loc, tokens = self.parseImpl(instring, preloc, doActions)
 *                     except IndexError:
 */
            }
            __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
            __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
            goto __pyx_L24_try_end;
            __pyx_L19_error:;
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

            /* "Aspidites/_vendor/pyparsing_extension.py":1408
 *                     try:
 *                         loc, tokens = self.parseImpl(instring, preloc, doActions)
 *                     except IndexError:             # <<<<<<<<<<<<<<
 *                         raise ParseException(instring, len(instring), self.errmsg, self)
 *                 else:
 */
            __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_IndexError);
            if (__pyx_t_9) {
              __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement._parseNoCache", __pyx_clineno, __pyx_lineno, __pyx_filename);
              if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_10, &__pyx_t_7) < 0) __PYX_ERR(0, 1408, __pyx_L21_except_error)
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_GOTREF(__pyx_t_10);
              __Pyx_GOTREF(__pyx_t_7);

              /* "Aspidites/_vendor/pyparsing_extension.py":1409
 *                         loc, tokens = self.parseImpl(instring, preloc, doActions)
 *                     except IndexError:
 *                         raise ParseException(instring, len(instring), self.errmsg, self)             # <<<<<<<<<<<<<<
 *                 else:
 *                     loc, tokens = self.parseImpl(instring, preloc, doActions)
 */
              __Pyx_GetModuleGlobalName(__pyx_t_19, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1409, __pyx_L21_except_error)
              __Pyx_GOTREF(__pyx_t_19);
              __pyx_t_14 = PyObject_Length(__pyx_v_instring); if (unlikely(__pyx_t_14 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1409, __pyx_L21_except_error)
              __pyx_t_20 = PyInt_FromSsize_t(__pyx_t_14); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1409, __pyx_L21_except_error)
              __Pyx_GOTREF(__pyx_t_20);
              __pyx_t_21 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errmsg); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1409, __pyx_L21_except_error)
              __Pyx_GOTREF(__pyx_t_21);
              __pyx_t_22 = NULL;
              __pyx_t_9 = 0;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_19))) {
                __pyx_t_22 = PyMethod_GET_SELF(__pyx_t_19);
                if (likely(__pyx_t_22)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_19);
                  __Pyx_INCREF(__pyx_t_22);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_19, function);
                  __pyx_t_9 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_19)) {
                PyObject *__pyx_temp[5] = {__pyx_t_22, __pyx_v_instring, __pyx_t_20, __pyx_t_21, __pyx_v_self};
                __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_19, __pyx_temp+1-__pyx_t_9, 4+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1409, __pyx_L21_except_error)
                __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
                __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_19)) {
                PyObject *__pyx_temp[5] = {__pyx_t_22, __pyx_v_instring, __pyx_t_20, __pyx_t_21, __pyx_v_self};
                __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_19, __pyx_temp+1-__pyx_t_9, 4+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1409, __pyx_L21_except_error)
                __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
                __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
              } else
              #endif
              {
                __pyx_t_23 = PyTuple_New(4+__pyx_t_9); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1409, __pyx_L21_except_error)
                __Pyx_GOTREF(__pyx_t_23);
                if (__pyx_t_22) {
                  __Pyx_GIVEREF(__pyx_t_22); PyTuple_SET_ITEM(__pyx_t_23, 0, __pyx_t_22); __pyx_t_22 = NULL;
                }
                __Pyx_INCREF(__pyx_v_instring);
                __Pyx_GIVEREF(__pyx_v_instring);
                PyTuple_SET_ITEM(__pyx_t_23, 0+__pyx_t_9, __pyx_v_instring);
                __Pyx_GIVEREF(__pyx_t_20);
                PyTuple_SET_ITEM(__pyx_t_23, 1+__pyx_t_9, __pyx_t_20);
                __Pyx_GIVEREF(__pyx_t_21);
                PyTuple_SET_ITEM(__pyx_t_23, 2+__pyx_t_9, __pyx_t_21);
                __Pyx_INCREF(__pyx_v_self);
                __Pyx_GIVEREF(__pyx_v_self);
                PyTuple_SET_ITEM(__pyx_t_23, 3+__pyx_t_9, __pyx_v_self);
                __pyx_t_20 = 0;
                __pyx_t_21 = 0;
                __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_19, __pyx_t_23, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1409, __pyx_L21_except_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
              }
              __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
              __Pyx_Raise(__pyx_t_4, 0, 0, 0);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __PYX_ERR(0, 1409, __pyx_L21_except_error)
            }
            goto __pyx_L21_except_error;
            __pyx_L21_except_error:;

            /* "Aspidites/_vendor/pyparsing_extension.py":1406
 *                 tokensStart = preloc
 *                 if self.mayIndexError or preloc >= len(instring):
 *                     try:             # <<<<<<<<<<<<<<
 *                         loc, tokens = self.parseImpl(instring, preloc, doActions)
 *                     except IndexError:
 */
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_ExceptionReset(__pyx_t_15, __pyx_t_16, __pyx_t_17);
            goto __pyx_L7_error;
            __pyx_L24_try_end:;
          }

          /* "Aspidites/_vendor/pyparsing_extension.py":1405
 *                     preloc = loc
 *                 tokensStart = preloc
 *                 if self.mayIndexError or preloc >= len(instring):             # <<<<<<<<<<<<<<
 *                     try:
 *                         loc, tokens = self.parseImpl(instring, preloc, doActions)
 */
          goto __pyx_L16;
        }

        /* "Aspidites/_vendor/pyparsing_extension.py":1411
 *                         raise ParseException(instring, len(instring), self.errmsg, self)
 *                 else:
 *                     loc, tokens = self.parseImpl(instring, preloc, doActions)             # <<<<<<<<<<<<<<
 *             except Exception as err:
 *                 # ~ print ("Exception raised:", err)
 */
        /*else*/ {
          __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_parseImpl); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1411, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_8 = NULL;
          __pyx_t_9 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
            __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_10);
            if (likely(__pyx_t_8)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
              __Pyx_INCREF(__pyx_t_8);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_10, function);
              __pyx_t_9 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_10)) {
            PyObject *__pyx_temp[4] = {__pyx_t_8, __pyx_v_instring, __pyx_v_preloc, __pyx_v_doActions};
            __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1411, __pyx_L7_error)
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
            PyObject *__pyx_temp[4] = {__pyx_t_8, __pyx_v_instring, __pyx_v_preloc, __pyx_v_doActions};
            __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1411, __pyx_L7_error)
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(3+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1411, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (__pyx_t_8) {
              __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_8); __pyx_t_8 = NULL;
            }
            __Pyx_INCREF(__pyx_v_instring);
            __Pyx_GIVEREF(__pyx_v_instring);
            PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_9, __pyx_v_instring);
            __Pyx_INCREF(__pyx_v_preloc);
            __Pyx_GIVEREF(__pyx_v_preloc);
            PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_9, __pyx_v_preloc);
            __Pyx_INCREF(__pyx_v_doActions);
            __Pyx_GIVEREF(__pyx_v_doActions);
            PyTuple_SET_ITEM(__pyx_t_4, 2+__pyx_t_9, __pyx_v_doActions);
            __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_4, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1411, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          if ((likely(PyTuple_CheckExact(__pyx_t_7))) || (PyList_CheckExact(__pyx_t_7))) {
            PyObject* sequence = __pyx_t_7;
            Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
            if (unlikely(size != 2)) {
              if (size > 2) __Pyx_RaiseTooManyValuesError(2);
              else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
              __PYX_ERR(0, 1411, __pyx_L7_error)
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            if (likely(PyTuple_CheckExact(sequence))) {
              __pyx_t_10 = PyTuple_GET_ITEM(sequence, 0); 
              __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
            } else {
              __pyx_t_10 = PyList_GET_ITEM(sequence, 0); 
              __pyx_t_4 = PyList_GET_ITEM(sequence, 1); 
            }
            __Pyx_INCREF(__pyx_t_10);
            __Pyx_INCREF(__pyx_t_4);
            #else
            __pyx_t_10 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1411, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_10);
            __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1411, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_4);
            #endif
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          } else {
            Py_ssize_t index = -1;
            __pyx_t_8 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1411, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_8);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __pyx_t_18 = Py_TYPE(__pyx_t_8)->tp_iternext;
            index = 0; __pyx_t_10 = __pyx_t_18(__pyx_t_8); if (unlikely(!__pyx_t_10)) goto __pyx_L29_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_10);
            index = 1; __pyx_t_4 = __pyx_t_18(__pyx_t_8); if (unlikely(!__pyx_t_4)) goto __pyx_L29_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_4);
            if (__Pyx_IternextUnpackEndCheck(__pyx_t_18(__pyx_t_8), 2) < 0) __PYX_ERR(0, 1411, __pyx_L7_error)
            __pyx_t_18 = NULL;
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            goto __pyx_L30_unpacking_done;
            __pyx_L29_unpacking_failed:;
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            __pyx_t_18 = NULL;
            if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
            __PYX_ERR(0, 1411, __pyx_L7_error)
            __pyx_L30_unpacking_done:;
          }
          __Pyx_DECREF_SET(__pyx_v_loc, __pyx_t_10);
          __pyx_t_10 = 0;
          __pyx_v_tokens = __pyx_t_4;
          __pyx_t_4 = 0;
        }
        __pyx_L16:;

        /* "Aspidites/_vendor/pyparsing_extension.py":1399
 *             if self.debugActions[TRY]:
 *                 self.debugActions[TRY](instring, loc, self)
 *             try:             # <<<<<<<<<<<<<<
 *                 if callPreParse and self.callPreparse:
 *                     preloc = self.preParse(instring, loc)
 */
      }
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      goto __pyx_L12_try_end;
      __pyx_L7_error:;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
      __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
      __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
      __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
      __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":1412
 *                 else:
 *                     loc, tokens = self.parseImpl(instring, preloc, doActions)
 *             except Exception as err:             # <<<<<<<<<<<<<<
 *                 # ~ print ("Exception raised:", err)
 *                 if self.debugActions[FAIL]:
 */
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
      if (__pyx_t_9) {
        __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement._parseNoCache", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_4, &__pyx_t_10) < 0) __PYX_ERR(0, 1412, __pyx_L9_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_4);
        __pyx_v_err = __pyx_t_4;

        /* "Aspidites/_vendor/pyparsing_extension.py":1414
 *             except Exception as err:
 *                 # ~ print ("Exception raised:", err)
 *                 if self.debugActions[FAIL]:             # <<<<<<<<<<<<<<
 *                     self.debugActions[FAIL](instring, tokensStart, self, err)
 *                 if self.failAction:
 */
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_debugActions); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1414, __pyx_L9_except_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_19 = __Pyx_GetItemInt(__pyx_t_8, __pyx_v_FAIL, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1414, __pyx_L9_except_error)
        __Pyx_GOTREF(__pyx_t_19);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_19); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1414, __pyx_L9_except_error)
        __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
        if (__pyx_t_5) {

          /* "Aspidites/_vendor/pyparsing_extension.py":1415
 *                 # ~ print ("Exception raised:", err)
 *                 if self.debugActions[FAIL]:
 *                     self.debugActions[FAIL](instring, tokensStart, self, err)             # <<<<<<<<<<<<<<
 *                 if self.failAction:
 *                     self.failAction(instring, tokensStart, self, err)
 */
          __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_debugActions); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1415, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_23 = __Pyx_GetItemInt(__pyx_t_8, __pyx_v_FAIL, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1415, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_23);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          if (unlikely(!__pyx_v_tokensStart)) { __Pyx_RaiseUnboundLocalError("tokensStart"); __PYX_ERR(0, 1415, __pyx_L9_except_error) }
          __pyx_t_8 = NULL;
          __pyx_t_9 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_23))) {
            __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_23);
            if (likely(__pyx_t_8)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_23);
              __Pyx_INCREF(__pyx_t_8);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_23, function);
              __pyx_t_9 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_23)) {
            PyObject *__pyx_temp[5] = {__pyx_t_8, __pyx_v_instring, __pyx_v_tokensStart, __pyx_v_self, __pyx_v_err};
            __pyx_t_19 = __Pyx_PyFunction_FastCall(__pyx_t_23, __pyx_temp+1-__pyx_t_9, 4+__pyx_t_9); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1415, __pyx_L9_except_error)
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_GOTREF(__pyx_t_19);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_23)) {
            PyObject *__pyx_temp[5] = {__pyx_t_8, __pyx_v_instring, __pyx_v_tokensStart, __pyx_v_self, __pyx_v_err};
            __pyx_t_19 = __Pyx_PyCFunction_FastCall(__pyx_t_23, __pyx_temp+1-__pyx_t_9, 4+__pyx_t_9); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1415, __pyx_L9_except_error)
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_GOTREF(__pyx_t_19);
          } else
          #endif
          {
            __pyx_t_21 = PyTuple_New(4+__pyx_t_9); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1415, __pyx_L9_except_error)
            __Pyx_GOTREF(__pyx_t_21);
            if (__pyx_t_8) {
              __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_21, 0, __pyx_t_8); __pyx_t_8 = NULL;
            }
            __Pyx_INCREF(__pyx_v_instring);
            __Pyx_GIVEREF(__pyx_v_instring);
            PyTuple_SET_ITEM(__pyx_t_21, 0+__pyx_t_9, __pyx_v_instring);
            __Pyx_INCREF(__pyx_v_tokensStart);
            __Pyx_GIVEREF(__pyx_v_tokensStart);
            PyTuple_SET_ITEM(__pyx_t_21, 1+__pyx_t_9, __pyx_v_tokensStart);
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_21, 2+__pyx_t_9, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_err);
            __Pyx_GIVEREF(__pyx_v_err);
            PyTuple_SET_ITEM(__pyx_t_21, 3+__pyx_t_9, __pyx_v_err);
            __pyx_t_19 = __Pyx_PyObject_Call(__pyx_t_23, __pyx_t_21, NULL); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1415, __pyx_L9_except_error)
            __Pyx_GOTREF(__pyx_t_19);
            __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
          }
          __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
          __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":1414
 *             except Exception as err:
 *                 # ~ print ("Exception raised:", err)
 *                 if self.debugActions[FAIL]:             # <<<<<<<<<<<<<<
 *                     self.debugActions[FAIL](instring, tokensStart, self, err)
 *                 if self.failAction:
 */
        }

        /* "Aspidites/_vendor/pyparsing_extension.py":1416
 *                 if self.debugActions[FAIL]:
 *                     self.debugActions[FAIL](instring, tokensStart, self, err)
 *                 if self.failAction:             # <<<<<<<<<<<<<<
 *                     self.failAction(instring, tokensStart, self, err)
 *                 raise
 */
        __pyx_t_19 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_failAction); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1416, __pyx_L9_except_error)
        __Pyx_GOTREF(__pyx_t_19);
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_19); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1416, __pyx_L9_except_error)
        __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
        if (__pyx_t_5) {

          /* "Aspidites/_vendor/pyparsing_extension.py":1417
 *                     self.debugActions[FAIL](instring, tokensStart, self, err)
 *                 if self.failAction:
 *                     self.failAction(instring, tokensStart, self, err)             # <<<<<<<<<<<<<<
 *                 raise
 *         else:
 */
          __pyx_t_23 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_failAction); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1417, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_23);
          if (unlikely(!__pyx_v_tokensStart)) { __Pyx_RaiseUnboundLocalError("tokensStart"); __PYX_ERR(0, 1417, __pyx_L9_except_error) }
          __pyx_t_21 = NULL;
          __pyx_t_9 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_23))) {
            __pyx_t_21 = PyMethod_GET_SELF(__pyx_t_23);
            if (likely(__pyx_t_21)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_23);
              __Pyx_INCREF(__pyx_t_21);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_23, function);
              __pyx_t_9 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_23)) {
            PyObject *__pyx_temp[5] = {__pyx_t_21, __pyx_v_instring, __pyx_v_tokensStart, __pyx_v_self, __pyx_v_err};
            __pyx_t_19 = __Pyx_PyFunction_FastCall(__pyx_t_23, __pyx_temp+1-__pyx_t_9, 4+__pyx_t_9); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1417, __pyx_L9_except_error)
            __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
            __Pyx_GOTREF(__pyx_t_19);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_23)) {
            PyObject *__pyx_temp[5] = {__pyx_t_21, __pyx_v_instring, __pyx_v_tokensStart, __pyx_v_self, __pyx_v_err};
            __pyx_t_19 = __Pyx_PyCFunction_FastCall(__pyx_t_23, __pyx_temp+1-__pyx_t_9, 4+__pyx_t_9); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1417, __pyx_L9_except_error)
            __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
            __Pyx_GOTREF(__pyx_t_19);
          } else
          #endif
          {
            __pyx_t_8 = PyTuple_New(4+__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1417, __pyx_L9_except_error)
            __Pyx_GOTREF(__pyx_t_8);
            if (__pyx_t_21) {
              __Pyx_GIVEREF(__pyx_t_21); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_21); __pyx_t_21 = NULL;
            }
            __Pyx_INCREF(__pyx_v_instring);
            __Pyx_GIVEREF(__pyx_v_instring);
            PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_9, __pyx_v_instring);
            __Pyx_INCREF(__pyx_v_tokensStart);
            __Pyx_GIVEREF(__pyx_v_tokensStart);
            PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_9, __pyx_v_tokensStart);
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_9, __pyx_v_self);
            __Pyx_INCREF(__pyx_v_err);
            __Pyx_GIVEREF(__pyx_v_err);
            PyTuple_SET_ITEM(__pyx_t_8, 3+__pyx_t_9, __pyx_v_err);
            __pyx_t_19 = __Pyx_PyObject_Call(__pyx_t_23, __pyx_t_8, NULL); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1417, __pyx_L9_except_error)
            __Pyx_GOTREF(__pyx_t_19);
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          }
          __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
          __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":1416
 *                 if self.debugActions[FAIL]:
 *                     self.debugActions[FAIL](instring, tokensStart, self, err)
 *                 if self.failAction:             # <<<<<<<<<<<<<<
 *                     self.failAction(instring, tokensStart, self, err)
 *                 raise
 */
        }

        /* "Aspidites/_vendor/pyparsing_extension.py":1418
 *                 if self.failAction:
 *                     self.failAction(instring, tokensStart, self, err)
 *                 raise             # <<<<<<<<<<<<<<
 *         else:
 *             if callPreParse and self.callPreparse:
 */
        __Pyx_GIVEREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_4);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_ErrRestoreWithState(__pyx_t_7, __pyx_t_4, __pyx_t_10);
        __pyx_t_7 = 0; __pyx_t_4 = 0; __pyx_t_10 = 0; 
        __PYX_ERR(0, 1418, __pyx_L9_except_error)
      }
      goto __pyx_L9_except_error;
      __pyx_L9_except_error:;

      /* "Aspidites/_vendor/pyparsing_extension.py":1399
 *             if self.debugActions[TRY]:
 *                 self.debugActions[TRY](instring, loc, self)
 *             try:             # <<<<<<<<<<<<<<
 *                 if callPreParse and self.callPreparse:
 *                     preloc = self.preParse(instring, loc)
 */
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
      goto __pyx_L1_error;
      __pyx_L12_try_end:;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":1395
 *         debugging = (self.debug)  # and doActions)
 * 
 *         if debugging or self.failAction:             # <<<<<<<<<<<<<<
 *             # ~ print ("Match", self, "at loc", loc, "(%d, %d)" % (lineno(loc, instring), col(loc, instring)))
 *             if self.debugActions[TRY]:
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1420
 *                 raise
 *         else:
 *             if callPreParse and self.callPreparse:             # <<<<<<<<<<<<<<
 *                 preloc = self.preParse(instring, loc)
 *             else:
 */
  /*else*/ {
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_callPreParse); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1420, __pyx_L1_error)
    if (__pyx_t_6) {
    } else {
      __pyx_t_5 = __pyx_t_6;
      goto __pyx_L36_bool_binop_done;
    }
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_callPreparse); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1420, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1420, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_5 = __pyx_t_6;
    __pyx_L36_bool_binop_done:;
    if (__pyx_t_5) {

      /* "Aspidites/_vendor/pyparsing_extension.py":1421
 *         else:
 *             if callPreParse and self.callPreparse:
 *                 preloc = self.preParse(instring, loc)             # <<<<<<<<<<<<<<
 *             else:
 *                 preloc = loc
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_preParse); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1421, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = NULL;
      __pyx_t_9 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_9 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_instring, __pyx_v_loc};
        __pyx_t_10 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1421, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_10);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_instring, __pyx_v_loc};
        __pyx_t_10 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1421, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_10);
      } else
      #endif
      {
        __pyx_t_19 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1421, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_19);
        if (__pyx_t_7) {
          __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_19, 0, __pyx_t_7); __pyx_t_7 = NULL;
        }
        __Pyx_INCREF(__pyx_v_instring);
        __Pyx_GIVEREF(__pyx_v_instring);
        PyTuple_SET_ITEM(__pyx_t_19, 0+__pyx_t_9, __pyx_v_instring);
        __Pyx_INCREF(__pyx_v_loc);
        __Pyx_GIVEREF(__pyx_v_loc);
        PyTuple_SET_ITEM(__pyx_t_19, 1+__pyx_t_9, __pyx_v_loc);
        __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_19, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1421, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_preloc = __pyx_t_10;
      __pyx_t_10 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":1420
 *                 raise
 *         else:
 *             if callPreParse and self.callPreparse:             # <<<<<<<<<<<<<<
 *                 preloc = self.preParse(instring, loc)
 *             else:
 */
      goto __pyx_L35;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":1423
 *                 preloc = self.preParse(instring, loc)
 *             else:
 *                 preloc = loc             # <<<<<<<<<<<<<<
 *             tokensStart = preloc
 *             if self.mayIndexError or preloc >= len(instring):
 */
    /*else*/ {
      __Pyx_INCREF(__pyx_v_loc);
      __pyx_v_preloc = __pyx_v_loc;
    }
    __pyx_L35:;

    /* "Aspidites/_vendor/pyparsing_extension.py":1424
 *             else:
 *                 preloc = loc
 *             tokensStart = preloc             # <<<<<<<<<<<<<<
 *             if self.mayIndexError or preloc >= len(instring):
 *                 try:
 */
    __Pyx_INCREF(__pyx_v_preloc);
    __pyx_v_tokensStart = __pyx_v_preloc;

    /* "Aspidites/_vendor/pyparsing_extension.py":1425
 *                 preloc = loc
 *             tokensStart = preloc
 *             if self.mayIndexError or preloc >= len(instring):             # <<<<<<<<<<<<<<
 *                 try:
 *                     loc, tokens = self.parseImpl(instring, preloc, doActions)
 */
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_mayIndexError); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1425, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1425, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (!__pyx_t_6) {
    } else {
      __pyx_t_5 = __pyx_t_6;
      goto __pyx_L39_bool_binop_done;
    }
    __pyx_t_14 = PyObject_Length(__pyx_v_instring); if (unlikely(__pyx_t_14 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1425, __pyx_L1_error)
    __pyx_t_10 = PyInt_FromSsize_t(__pyx_t_14); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1425, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_4 = PyObject_RichCompare(__pyx_v_preloc, __pyx_t_10, Py_GE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1425, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1425, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = __pyx_t_6;
    __pyx_L39_bool_binop_done:;
    if (__pyx_t_5) {

      /* "Aspidites/_vendor/pyparsing_extension.py":1426
 *             tokensStart = preloc
 *             if self.mayIndexError or preloc >= len(instring):
 *                 try:             # <<<<<<<<<<<<<<
 *                     loc, tokens = self.parseImpl(instring, preloc, doActions)
 *                 except IndexError:
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_13, &__pyx_t_12, &__pyx_t_11);
        __Pyx_XGOTREF(__pyx_t_13);
        __Pyx_XGOTREF(__pyx_t_12);
        __Pyx_XGOTREF(__pyx_t_11);
        /*try:*/ {

          /* "Aspidites/_vendor/pyparsing_extension.py":1427
 *             if self.mayIndexError or preloc >= len(instring):
 *                 try:
 *                     loc, tokens = self.parseImpl(instring, preloc, doActions)             # <<<<<<<<<<<<<<
 *                 except IndexError:
 *                     raise ParseException(instring, len(instring), self.errmsg, self)
 */
          __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_parseImpl); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1427, __pyx_L41_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_19 = NULL;
          __pyx_t_9 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
            __pyx_t_19 = PyMethod_GET_SELF(__pyx_t_10);
            if (likely(__pyx_t_19)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
              __Pyx_INCREF(__pyx_t_19);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_10, function);
              __pyx_t_9 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_10)) {
            PyObject *__pyx_temp[4] = {__pyx_t_19, __pyx_v_instring, __pyx_v_preloc, __pyx_v_doActions};
            __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1427, __pyx_L41_error)
            __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
            PyObject *__pyx_temp[4] = {__pyx_t_19, __pyx_v_instring, __pyx_v_preloc, __pyx_v_doActions};
            __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1427, __pyx_L41_error)
            __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(3+__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1427, __pyx_L41_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (__pyx_t_19) {
              __Pyx_GIVEREF(__pyx_t_19); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_19); __pyx_t_19 = NULL;
            }
            __Pyx_INCREF(__pyx_v_instring);
            __Pyx_GIVEREF(__pyx_v_instring);
            PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_9, __pyx_v_instring);
            __Pyx_INCREF(__pyx_v_preloc);
            __Pyx_GIVEREF(__pyx_v_preloc);
            PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_9, __pyx_v_preloc);
            __Pyx_INCREF(__pyx_v_doActions);
            __Pyx_GIVEREF(__pyx_v_doActions);
            PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_9, __pyx_v_doActions);
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_7, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1427, __pyx_L41_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
            PyObject* sequence = __pyx_t_4;
            Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
            if (unlikely(size != 2)) {
              if (size > 2) __Pyx_RaiseTooManyValuesError(2);
              else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
              __PYX_ERR(0, 1427, __pyx_L41_error)
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            if (likely(PyTuple_CheckExact(sequence))) {
              __pyx_t_10 = PyTuple_GET_ITEM(sequence, 0); 
              __pyx_t_7 = PyTuple_GET_ITEM(sequence, 1); 
            } else {
              __pyx_t_10 = PyList_GET_ITEM(sequence, 0); 
              __pyx_t_7 = PyList_GET_ITEM(sequence, 1); 
            }
            __Pyx_INCREF(__pyx_t_10);
            __Pyx_INCREF(__pyx_t_7);
            #else
            __pyx_t_10 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1427, __pyx_L41_error)
            __Pyx_GOTREF(__pyx_t_10);
            __pyx_t_7 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1427, __pyx_L41_error)
            __Pyx_GOTREF(__pyx_t_7);
            #endif
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else {
            Py_ssize_t index = -1;
            __pyx_t_19 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1427, __pyx_L41_error)
            __Pyx_GOTREF(__pyx_t_19);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_18 = Py_TYPE(__pyx_t_19)->tp_iternext;
            index = 0; __pyx_t_10 = __pyx_t_18(__pyx_t_19); if (unlikely(!__pyx_t_10)) goto __pyx_L47_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_10);
            index = 1; __pyx_t_7 = __pyx_t_18(__pyx_t_19); if (unlikely(!__pyx_t_7)) goto __pyx_L47_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_7);
            if (__Pyx_IternextUnpackEndCheck(__pyx_t_18(__pyx_t_19), 2) < 0) __PYX_ERR(0, 1427, __pyx_L41_error)
            __pyx_t_18 = NULL;
            __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
            goto __pyx_L48_unpacking_done;
            __pyx_L47_unpacking_failed:;
            __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
            __pyx_t_18 = NULL;
            if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
            __PYX_ERR(0, 1427, __pyx_L41_error)
            __pyx_L48_unpacking_done:;
          }
          __Pyx_DECREF_SET(__pyx_v_loc, __pyx_t_10);
          __pyx_t_10 = 0;
          __pyx_v_tokens = __pyx_t_7;
          __pyx_t_7 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":1426
 *             tokensStart = preloc
 *             if self.mayIndexError or preloc >= len(instring):
 *                 try:             # <<<<<<<<<<<<<<
 *                     loc, tokens = self.parseImpl(instring, preloc, doActions)
 *                 except IndexError:
 */
        }
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        goto __pyx_L46_try_end;
        __pyx_L41_error:;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
        __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
        __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
        __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
        __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":1428
 *                 try:
 *                     loc, tokens = self.parseImpl(instring, preloc, doActions)
 *                 except IndexError:             # <<<<<<<<<<<<<<
 *                     raise ParseException(instring, len(instring), self.errmsg, self)
 *             else:
 */
        __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_IndexError);
        if (__pyx_t_9) {
          __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement._parseNoCache", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_7, &__pyx_t_10) < 0) __PYX_ERR(0, 1428, __pyx_L43_except_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_GOTREF(__pyx_t_10);

          /* "Aspidites/_vendor/pyparsing_extension.py":1429
 *                     loc, tokens = self.parseImpl(instring, preloc, doActions)
 *                 except IndexError:
 *                     raise ParseException(instring, len(instring), self.errmsg, self)             # <<<<<<<<<<<<<<
 *             else:
 *                 loc, tokens = self.parseImpl(instring, preloc, doActions)
 */
          __Pyx_GetModuleGlobalName(__pyx_t_23, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1429, __pyx_L43_except_error)
          __Pyx_GOTREF(__pyx_t_23);
          __pyx_t_14 = PyObject_Length(__pyx_v_instring); if (unlikely(__pyx_t_14 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1429, __pyx_L43_except_error)
          __pyx_t_8 = PyInt_FromSsize_t(__pyx_t_14); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1429, __pyx_L43_except_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_21 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errmsg); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1429, __pyx_L43_except_error)
          __Pyx_GOTREF(__pyx_t_21);
          __pyx_t_20 = NULL;
          __pyx_t_9 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_23))) {
            __pyx_t_20 = PyMethod_GET_SELF(__pyx_t_23);
            if (likely(__pyx_t_20)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_23);
              __Pyx_INCREF(__pyx_t_20);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_23, function);
              __pyx_t_9 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_23)) {
            PyObject *__pyx_temp[5] = {__pyx_t_20, __pyx_v_instring, __pyx_t_8, __pyx_t_21, __pyx_v_self};
            __pyx_t_19 = __Pyx_PyFunction_FastCall(__pyx_t_23, __pyx_temp+1-__pyx_t_9, 4+__pyx_t_9); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1429, __pyx_L43_except_error)
            __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
            __Pyx_GOTREF(__pyx_t_19);
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_23)) {
            PyObject *__pyx_temp[5] = {__pyx_t_20, __pyx_v_instring, __pyx_t_8, __pyx_t_21, __pyx_v_self};
            __pyx_t_19 = __Pyx_PyCFunction_FastCall(__pyx_t_23, __pyx_temp+1-__pyx_t_9, 4+__pyx_t_9); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1429, __pyx_L43_except_error)
            __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
            __Pyx_GOTREF(__pyx_t_19);
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
          } else
          #endif
          {
            __pyx_t_22 = PyTuple_New(4+__pyx_t_9); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1429, __pyx_L43_except_error)
            __Pyx_GOTREF(__pyx_t_22);
            if (__pyx_t_20) {
              __Pyx_GIVEREF(__pyx_t_20); PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_t_20); __pyx_t_20 = NULL;
            }
            __Pyx_INCREF(__pyx_v_instring);
            __Pyx_GIVEREF(__pyx_v_instring);
            PyTuple_SET_ITEM(__pyx_t_22, 0+__pyx_t_9, __pyx_v_instring);
            __Pyx_GIVEREF(__pyx_t_8);
            PyTuple_SET_ITEM(__pyx_t_22, 1+__pyx_t_9, __pyx_t_8);
            __Pyx_GIVEREF(__pyx_t_21);
            PyTuple_SET_ITEM(__pyx_t_22, 2+__pyx_t_9, __pyx_t_21);
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_22, 3+__pyx_t_9, __pyx_v_self);
            __pyx_t_8 = 0;
            __pyx_t_21 = 0;
            __pyx_t_19 = __Pyx_PyObject_Call(__pyx_t_23, __pyx_t_22, NULL); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1429, __pyx_L43_except_error)
            __Pyx_GOTREF(__pyx_t_19);
            __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
          }
          __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
          __Pyx_Raise(__pyx_t_19, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
          __PYX_ERR(0, 1429, __pyx_L43_except_error)
        }
        goto __pyx_L43_except_error;
        __pyx_L43_except_error:;

        /* "Aspidites/_vendor/pyparsing_extension.py":1426
 *             tokensStart = preloc
 *             if self.mayIndexError or preloc >= len(instring):
 *                 try:             # <<<<<<<<<<<<<<
 *                     loc, tokens = self.parseImpl(instring, preloc, doActions)
 *                 except IndexError:
 */
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_12, __pyx_t_11);
        goto __pyx_L1_error;
        __pyx_L46_try_end:;
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":1425
 *                 preloc = loc
 *             tokensStart = preloc
 *             if self.mayIndexError or preloc >= len(instring):             # <<<<<<<<<<<<<<
 *                 try:
 *                     loc, tokens = self.parseImpl(instring, preloc, doActions)
 */
      goto __pyx_L38;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":1431
 *                     raise ParseException(instring, len(instring), self.errmsg, self)
 *             else:
 *                 loc, tokens = self.parseImpl(instring, preloc, doActions)             # <<<<<<<<<<<<<<
 * 
 *         tokens = self.postParse(instring, loc, tokens)
 */
    /*else*/ {
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_parseImpl); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1431, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_4 = NULL;
      __pyx_t_9 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
          __pyx_t_9 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_7)) {
        PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_instring, __pyx_v_preloc, __pyx_v_doActions};
        __pyx_t_10 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1431, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_10);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
        PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_instring, __pyx_v_preloc, __pyx_v_doActions};
        __pyx_t_10 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1431, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_10);
      } else
      #endif
      {
        __pyx_t_19 = PyTuple_New(3+__pyx_t_9); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1431, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_19);
        if (__pyx_t_4) {
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_19, 0, __pyx_t_4); __pyx_t_4 = NULL;
        }
        __Pyx_INCREF(__pyx_v_instring);
        __Pyx_GIVEREF(__pyx_v_instring);
        PyTuple_SET_ITEM(__pyx_t_19, 0+__pyx_t_9, __pyx_v_instring);
        __Pyx_INCREF(__pyx_v_preloc);
        __Pyx_GIVEREF(__pyx_v_preloc);
        PyTuple_SET_ITEM(__pyx_t_19, 1+__pyx_t_9, __pyx_v_preloc);
        __Pyx_INCREF(__pyx_v_doActions);
        __Pyx_GIVEREF(__pyx_v_doActions);
        PyTuple_SET_ITEM(__pyx_t_19, 2+__pyx_t_9, __pyx_v_doActions);
        __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_19, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1431, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if ((likely(PyTuple_CheckExact(__pyx_t_10))) || (PyList_CheckExact(__pyx_t_10))) {
        PyObject* sequence = __pyx_t_10;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 1431, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_7 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_19 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_7 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_19 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_19);
        #else
        __pyx_t_7 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1431, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_19 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1431, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_19);
        #endif
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_4 = PyObject_GetIter(__pyx_t_10); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1431, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_18 = Py_TYPE(__pyx_t_4)->tp_iternext;
        index = 0; __pyx_t_7 = __pyx_t_18(__pyx_t_4); if (unlikely(!__pyx_t_7)) goto __pyx_L51_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_7);
        index = 1; __pyx_t_19 = __pyx_t_18(__pyx_t_4); if (unlikely(!__pyx_t_19)) goto __pyx_L51_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_19);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_18(__pyx_t_4), 2) < 0) __PYX_ERR(0, 1431, __pyx_L1_error)
        __pyx_t_18 = NULL;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        goto __pyx_L52_unpacking_done;
        __pyx_L51_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_18 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 1431, __pyx_L1_error)
        __pyx_L52_unpacking_done:;
      }
      __Pyx_DECREF_SET(__pyx_v_loc, __pyx_t_7);
      __pyx_t_7 = 0;
      __pyx_v_tokens = __pyx_t_19;
      __pyx_t_19 = 0;
    }
    __pyx_L38:;
  }
  __pyx_L3:;

  /* "Aspidites/_vendor/pyparsing_extension.py":1433
 *                 loc, tokens = self.parseImpl(instring, preloc, doActions)
 * 
 *         tokens = self.postParse(instring, loc, tokens)             # <<<<<<<<<<<<<<
 * 
 *         retTokens = ParseResults(tokens, self.resultsName, asList=self.saveAsList, modal=self.modalResults)
 */
  __pyx_t_19 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_postParse); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __pyx_t_7 = NULL;
  __pyx_t_9 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_19))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_19);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_19);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_19, function);
      __pyx_t_9 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_19)) {
    PyObject *__pyx_temp[4] = {__pyx_t_7, __pyx_v_instring, __pyx_v_loc, __pyx_v_tokens};
    __pyx_t_10 = __Pyx_PyFunction_FastCall(__pyx_t_19, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1433, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_10);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_19)) {
    PyObject *__pyx_temp[4] = {__pyx_t_7, __pyx_v_instring, __pyx_v_loc, __pyx_v_tokens};
    __pyx_t_10 = __Pyx_PyCFunction_FastCall(__pyx_t_19, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1433, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_10);
  } else
  #endif
  {
    __pyx_t_4 = PyTuple_New(3+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1433, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (__pyx_t_7) {
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __pyx_t_7 = NULL;
    }
    __Pyx_INCREF(__pyx_v_instring);
    __Pyx_GIVEREF(__pyx_v_instring);
    PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_9, __pyx_v_instring);
    __Pyx_INCREF(__pyx_v_loc);
    __Pyx_GIVEREF(__pyx_v_loc);
    PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_9, __pyx_v_loc);
    __Pyx_INCREF(__pyx_v_tokens);
    __Pyx_GIVEREF(__pyx_v_tokens);
    PyTuple_SET_ITEM(__pyx_t_4, 2+__pyx_t_9, __pyx_v_tokens);
    __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_19, __pyx_t_4, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1433, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __Pyx_DECREF_SET(__pyx_v_tokens, __pyx_t_10);
  __pyx_t_10 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1435
 *         tokens = self.postParse(instring, loc, tokens)
 * 
 *         retTokens = ParseResults(tokens, self.resultsName, asList=self.saveAsList, modal=self.modalResults)             # <<<<<<<<<<<<<<
 *         if self.parseAction and (doActions or self.callDuringTry):
 *             if debugging:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_ParseResults); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_19 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_resultsName); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_tokens);
  __Pyx_GIVEREF(__pyx_v_tokens);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_tokens);
  __Pyx_GIVEREF(__pyx_t_19);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_19);
  __pyx_t_19 = 0;
  __pyx_t_19 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_saveAsList); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_19, __pyx_n_s_asList, __pyx_t_7) < 0) __PYX_ERR(0, 1435, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_modalResults); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_19, __pyx_n_s_modal, __pyx_t_7) < 0) __PYX_ERR(0, 1435, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_4, __pyx_t_19); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __pyx_v_retTokens = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1436
 * 
 *         retTokens = ParseResults(tokens, self.resultsName, asList=self.saveAsList, modal=self.modalResults)
 *         if self.parseAction and (doActions or self.callDuringTry):             # <<<<<<<<<<<<<<
 *             if debugging:
 *                 try:
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_parseAction); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1436, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__pyx_t_6) {
  } else {
    __pyx_t_5 = __pyx_t_6;
    goto __pyx_L54_bool_binop_done;
  }
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_doActions); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1436, __pyx_L1_error)
  if (!__pyx_t_6) {
  } else {
    __pyx_t_5 = __pyx_t_6;
    goto __pyx_L54_bool_binop_done;
  }
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_callDuringTry); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1436, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_5 = __pyx_t_6;
  __pyx_L54_bool_binop_done:;
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/pyparsing_extension.py":1437
 *         retTokens = ParseResults(tokens, self.resultsName, asList=self.saveAsList, modal=self.modalResults)
 *         if self.parseAction and (doActions or self.callDuringTry):
 *             if debugging:             # <<<<<<<<<<<<<<
 *                 try:
 *                     for fn in self.parseAction:
 */
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_debugging); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1437, __pyx_L1_error)
    if (__pyx_t_5) {

      /* "Aspidites/_vendor/pyparsing_extension.py":1438
 *         if self.parseAction and (doActions or self.callDuringTry):
 *             if debugging:
 *                 try:             # <<<<<<<<<<<<<<
 *                     for fn in self.parseAction:
 *                         try:
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);
        __Pyx_XGOTREF(__pyx_t_11);
        __Pyx_XGOTREF(__pyx_t_12);
        __Pyx_XGOTREF(__pyx_t_13);
        /*try:*/ {

          /* "Aspidites/_vendor/pyparsing_extension.py":1439
 *             if debugging:
 *                 try:
 *                     for fn in self.parseAction:             # <<<<<<<<<<<<<<
 *                         try:
 *                             tokens = fn(instring, tokensStart, retTokens)
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_parseAction); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1439, __pyx_L58_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (likely(PyList_CheckExact(__pyx_t_7)) || PyTuple_CheckExact(__pyx_t_7)) {
            __pyx_t_19 = __pyx_t_7; __Pyx_INCREF(__pyx_t_19); __pyx_t_14 = 0;
            __pyx_t_24 = NULL;
          } else {
            __pyx_t_14 = -1; __pyx_t_19 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1439, __pyx_L58_error)
            __Pyx_GOTREF(__pyx_t_19);
            __pyx_t_24 = Py_TYPE(__pyx_t_19)->tp_iternext; if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1439, __pyx_L58_error)
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          for (;;) {
            if (likely(!__pyx_t_24)) {
              if (likely(PyList_CheckExact(__pyx_t_19))) {
                if (__pyx_t_14 >= PyList_GET_SIZE(__pyx_t_19)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_7 = PyList_GET_ITEM(__pyx_t_19, __pyx_t_14); __Pyx_INCREF(__pyx_t_7); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 1439, __pyx_L58_error)
                #else
                __pyx_t_7 = PySequence_ITEM(__pyx_t_19, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1439, __pyx_L58_error)
                __Pyx_GOTREF(__pyx_t_7);
                #endif
              } else {
                if (__pyx_t_14 >= PyTuple_GET_SIZE(__pyx_t_19)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_19, __pyx_t_14); __Pyx_INCREF(__pyx_t_7); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 1439, __pyx_L58_error)
                #else
                __pyx_t_7 = PySequence_ITEM(__pyx_t_19, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1439, __pyx_L58_error)
                __Pyx_GOTREF(__pyx_t_7);
                #endif
              }
            } else {
              __pyx_t_7 = __pyx_t_24(__pyx_t_19);
              if (unlikely(!__pyx_t_7)) {
                PyObject* exc_type = PyErr_Occurred();
                if (exc_type) {
                  if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                  else __PYX_ERR(0, 1439, __pyx_L58_error)
                }
                break;
              }
              __Pyx_GOTREF(__pyx_t_7);
            }
            __Pyx_XDECREF_SET(__pyx_v_fn, __pyx_t_7);
            __pyx_t_7 = 0;

            /* "Aspidites/_vendor/pyparsing_extension.py":1440
 *                 try:
 *                     for fn in self.parseAction:
 *                         try:             # <<<<<<<<<<<<<<
 *                             tokens = fn(instring, tokensStart, retTokens)
 *                         except IndexError as parse_action_exc:
 */
            {
              __Pyx_PyThreadState_declare
              __Pyx_PyThreadState_assign
              __Pyx_ExceptionSave(&__pyx_t_17, &__pyx_t_16, &__pyx_t_15);
              __Pyx_XGOTREF(__pyx_t_17);
              __Pyx_XGOTREF(__pyx_t_16);
              __Pyx_XGOTREF(__pyx_t_15);
              /*try:*/ {

                /* "Aspidites/_vendor/pyparsing_extension.py":1441
 *                     for fn in self.parseAction:
 *                         try:
 *                             tokens = fn(instring, tokensStart, retTokens)             # <<<<<<<<<<<<<<
 *                         except IndexError as parse_action_exc:
 *                             exc = ParseException("exception raised in parse action")
 */
                __Pyx_INCREF(__pyx_v_fn);
                __pyx_t_4 = __pyx_v_fn; __pyx_t_10 = NULL;
                __pyx_t_9 = 0;
                if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
                  __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_4);
                  if (likely(__pyx_t_10)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                    __Pyx_INCREF(__pyx_t_10);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_4, function);
                    __pyx_t_9 = 1;
                  }
                }
                #if CYTHON_FAST_PYCALL
                if (PyFunction_Check(__pyx_t_4)) {
                  PyObject *__pyx_temp[4] = {__pyx_t_10, __pyx_v_instring, __pyx_v_tokensStart, __pyx_v_retTokens};
                  __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1441, __pyx_L66_error)
                  __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
                  __Pyx_GOTREF(__pyx_t_7);
                } else
                #endif
                #if CYTHON_FAST_PYCCALL
                if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
                  PyObject *__pyx_temp[4] = {__pyx_t_10, __pyx_v_instring, __pyx_v_tokensStart, __pyx_v_retTokens};
                  __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1441, __pyx_L66_error)
                  __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
                  __Pyx_GOTREF(__pyx_t_7);
                } else
                #endif
                {
                  __pyx_t_23 = PyTuple_New(3+__pyx_t_9); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1441, __pyx_L66_error)
                  __Pyx_GOTREF(__pyx_t_23);
                  if (__pyx_t_10) {
                    __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_23, 0, __pyx_t_10); __pyx_t_10 = NULL;
                  }
                  __Pyx_INCREF(__pyx_v_instring);
                  __Pyx_GIVEREF(__pyx_v_instring);
                  PyTuple_SET_ITEM(__pyx_t_23, 0+__pyx_t_9, __pyx_v_instring);
                  __Pyx_INCREF(__pyx_v_tokensStart);
                  __Pyx_GIVEREF(__pyx_v_tokensStart);
                  PyTuple_SET_ITEM(__pyx_t_23, 1+__pyx_t_9, __pyx_v_tokensStart);
                  __Pyx_INCREF(__pyx_v_retTokens);
                  __Pyx_GIVEREF(__pyx_v_retTokens);
                  PyTuple_SET_ITEM(__pyx_t_23, 2+__pyx_t_9, __pyx_v_retTokens);
                  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_23, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1441, __pyx_L66_error)
                  __Pyx_GOTREF(__pyx_t_7);
                  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
                }
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_DECREF_SET(__pyx_v_tokens, __pyx_t_7);
                __pyx_t_7 = 0;

                /* "Aspidites/_vendor/pyparsing_extension.py":1440
 *                 try:
 *                     for fn in self.parseAction:
 *                         try:             # <<<<<<<<<<<<<<
 *                             tokens = fn(instring, tokensStart, retTokens)
 *                         except IndexError as parse_action_exc:
 */
              }
              __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
              __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
              __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
              goto __pyx_L73_try_end;
              __pyx_L66_error:;
              __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
              __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
              __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
              __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
              __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

              /* "Aspidites/_vendor/pyparsing_extension.py":1442
 *                         try:
 *                             tokens = fn(instring, tokensStart, retTokens)
 *                         except IndexError as parse_action_exc:             # <<<<<<<<<<<<<<
 *                             exc = ParseException("exception raised in parse action")
 *                             exc.__cause__ = parse_action_exc
 */
              __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_IndexError);
              if (__pyx_t_9) {
                __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement._parseNoCache", __pyx_clineno, __pyx_lineno, __pyx_filename);
                if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_4, &__pyx_t_23) < 0) __PYX_ERR(0, 1442, __pyx_L68_except_error)
                __Pyx_GOTREF(__pyx_t_7);
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_GOTREF(__pyx_t_23);
                __Pyx_INCREF(__pyx_t_4);
                __pyx_v_parse_action_exc = __pyx_t_4;

                /* "Aspidites/_vendor/pyparsing_extension.py":1443
 *                             tokens = fn(instring, tokensStart, retTokens)
 *                         except IndexError as parse_action_exc:
 *                             exc = ParseException("exception raised in parse action")             # <<<<<<<<<<<<<<
 *                             exc.__cause__ = parse_action_exc
 *                             raise exc
 */
                __Pyx_GetModuleGlobalName(__pyx_t_22, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1443, __pyx_L68_except_error)
                __Pyx_GOTREF(__pyx_t_22);
                __pyx_t_21 = NULL;
                if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_22))) {
                  __pyx_t_21 = PyMethod_GET_SELF(__pyx_t_22);
                  if (likely(__pyx_t_21)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_22);
                    __Pyx_INCREF(__pyx_t_21);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_22, function);
                  }
                }
                __pyx_t_10 = (__pyx_t_21) ? __Pyx_PyObject_Call2Args(__pyx_t_22, __pyx_t_21, __pyx_kp_s_exception_raised_in_parse_action) : __Pyx_PyObject_CallOneArg(__pyx_t_22, __pyx_kp_s_exception_raised_in_parse_action);
                __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
                if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1443, __pyx_L68_except_error)
                __Pyx_GOTREF(__pyx_t_10);
                __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
                __pyx_v_exc = __pyx_t_10;
                __pyx_t_10 = 0;

                /* "Aspidites/_vendor/pyparsing_extension.py":1444
 *                         except IndexError as parse_action_exc:
 *                             exc = ParseException("exception raised in parse action")
 *                             exc.__cause__ = parse_action_exc             # <<<<<<<<<<<<<<
 *                             raise exc
 * 
 */
                if (__Pyx_PyObject_SetAttrStr(__pyx_v_exc, __pyx_n_s_cause, __pyx_v_parse_action_exc) < 0) __PYX_ERR(0, 1444, __pyx_L68_except_error)

                /* "Aspidites/_vendor/pyparsing_extension.py":1445
 *                             exc = ParseException("exception raised in parse action")
 *                             exc.__cause__ = parse_action_exc
 *                             raise exc             # <<<<<<<<<<<<<<
 * 
 *                         if tokens is not None and tokens is not retTokens:
 */
                __Pyx_Raise(__pyx_v_exc, 0, 0, 0);
                __PYX_ERR(0, 1445, __pyx_L68_except_error)
              }
              goto __pyx_L68_except_error;
              __pyx_L68_except_error:;

              /* "Aspidites/_vendor/pyparsing_extension.py":1440
 *                 try:
 *                     for fn in self.parseAction:
 *                         try:             # <<<<<<<<<<<<<<
 *                             tokens = fn(instring, tokensStart, retTokens)
 *                         except IndexError as parse_action_exc:
 */
              __Pyx_XGIVEREF(__pyx_t_17);
              __Pyx_XGIVEREF(__pyx_t_16);
              __Pyx_XGIVEREF(__pyx_t_15);
              __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_16, __pyx_t_15);
              goto __pyx_L58_error;
              __pyx_L73_try_end:;
            }

            /* "Aspidites/_vendor/pyparsing_extension.py":1447
 *                             raise exc
 * 
 *                         if tokens is not None and tokens is not retTokens:             # <<<<<<<<<<<<<<
 *                             retTokens = ParseResults(tokens,
 *                                                      self.resultsName,
 */
            __pyx_t_6 = (__pyx_v_tokens != Py_None);
            __pyx_t_25 = (__pyx_t_6 != 0);
            if (__pyx_t_25) {
            } else {
              __pyx_t_5 = __pyx_t_25;
              goto __pyx_L77_bool_binop_done;
            }
            __pyx_t_25 = (__pyx_v_tokens != __pyx_v_retTokens);
            __pyx_t_6 = (__pyx_t_25 != 0);
            __pyx_t_5 = __pyx_t_6;
            __pyx_L77_bool_binop_done:;
            if (__pyx_t_5) {

              /* "Aspidites/_vendor/pyparsing_extension.py":1448
 * 
 *                         if tokens is not None and tokens is not retTokens:
 *                             retTokens = ParseResults(tokens,             # <<<<<<<<<<<<<<
 *                                                      self.resultsName,
 *                                                      asList=self.saveAsList and isinstance(tokens,
 */
              __Pyx_GetModuleGlobalName(__pyx_t_23, __pyx_n_s_ParseResults); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1448, __pyx_L58_error)
              __Pyx_GOTREF(__pyx_t_23);

              /* "Aspidites/_vendor/pyparsing_extension.py":1449
 *                         if tokens is not None and tokens is not retTokens:
 *                             retTokens = ParseResults(tokens,
 *                                                      self.resultsName,             # <<<<<<<<<<<<<<
 *                                                      asList=self.saveAsList and isinstance(tokens,
 *                                                                                            (ParseResults, list)),
 */
              __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_resultsName); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1449, __pyx_L58_error)
              __Pyx_GOTREF(__pyx_t_4);

              /* "Aspidites/_vendor/pyparsing_extension.py":1448
 * 
 *                         if tokens is not None and tokens is not retTokens:
 *                             retTokens = ParseResults(tokens,             # <<<<<<<<<<<<<<
 *                                                      self.resultsName,
 *                                                      asList=self.saveAsList and isinstance(tokens,
 */
              __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1448, __pyx_L58_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_INCREF(__pyx_v_tokens);
              __Pyx_GIVEREF(__pyx_v_tokens);
              PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_tokens);
              __Pyx_GIVEREF(__pyx_t_4);
              PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_4);
              __pyx_t_4 = 0;

              /* "Aspidites/_vendor/pyparsing_extension.py":1450
 *                             retTokens = ParseResults(tokens,
 *                                                      self.resultsName,
 *                                                      asList=self.saveAsList and isinstance(tokens,             # <<<<<<<<<<<<<<
 *                                                                                            (ParseResults, list)),
 *                                                      modal=self.modalResults)
 */
              __pyx_t_4 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1450, __pyx_L58_error)
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_22 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_saveAsList); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1450, __pyx_L58_error)
              __Pyx_GOTREF(__pyx_t_22);
              __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_22); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1450, __pyx_L58_error)
              if (__pyx_t_5) {
                __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
              } else {
                __Pyx_INCREF(__pyx_t_22);
                __pyx_t_10 = __pyx_t_22;
                __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
                goto __pyx_L79_bool_binop_done;
              }

              /* "Aspidites/_vendor/pyparsing_extension.py":1451
 *                                                      self.resultsName,
 *                                                      asList=self.saveAsList and isinstance(tokens,
 *                                                                                            (ParseResults, list)),             # <<<<<<<<<<<<<<
 *                                                      modal=self.modalResults)
 *                 except Exception as err:
 */
              __Pyx_GetModuleGlobalName(__pyx_t_22, __pyx_n_s_ParseResults); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1451, __pyx_L58_error)
              __Pyx_GOTREF(__pyx_t_22);

              /* "Aspidites/_vendor/pyparsing_extension.py":1450
 *                             retTokens = ParseResults(tokens,
 *                                                      self.resultsName,
 *                                                      asList=self.saveAsList and isinstance(tokens,             # <<<<<<<<<<<<<<
 *                                                                                            (ParseResults, list)),
 *                                                      modal=self.modalResults)
 */
              __pyx_t_6 = PyObject_IsInstance(__pyx_v_tokens, __pyx_t_22); 

              /* "Aspidites/_vendor/pyparsing_extension.py":1451
 *                                                      self.resultsName,
 *                                                      asList=self.saveAsList and isinstance(tokens,
 *                                                                                            (ParseResults, list)),             # <<<<<<<<<<<<<<
 *                                                      modal=self.modalResults)
 *                 except Exception as err:
 */
              __pyx_t_25 = (__pyx_t_6 != 0);
              if (!__pyx_t_25) {
              } else {
                __pyx_t_5 = __pyx_t_25;
                goto __pyx_L81_bool_binop_done;
              }

              /* "Aspidites/_vendor/pyparsing_extension.py":1450
 *                             retTokens = ParseResults(tokens,
 *                                                      self.resultsName,
 *                                                      asList=self.saveAsList and isinstance(tokens,             # <<<<<<<<<<<<<<
 *                                                                                            (ParseResults, list)),
 *                                                      modal=self.modalResults)
 */
              __pyx_t_25 = PyList_Check(__pyx_v_tokens); 

              /* "Aspidites/_vendor/pyparsing_extension.py":1451
 *                                                      self.resultsName,
 *                                                      asList=self.saveAsList and isinstance(tokens,
 *                                                                                            (ParseResults, list)),             # <<<<<<<<<<<<<<
 *                                                      modal=self.modalResults)
 *                 except Exception as err:
 */
              __pyx_t_6 = (__pyx_t_25 != 0);
              __pyx_t_5 = __pyx_t_6;
              __pyx_L81_bool_binop_done:;
              __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

              /* "Aspidites/_vendor/pyparsing_extension.py":1450
 *                             retTokens = ParseResults(tokens,
 *                                                      self.resultsName,
 *                                                      asList=self.saveAsList and isinstance(tokens,             # <<<<<<<<<<<<<<
 *                                                                                            (ParseResults, list)),
 *                                                      modal=self.modalResults)
 */
              __pyx_t_22 = __Pyx_PyBool_FromLong(__pyx_t_5); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1450, __pyx_L58_error)
              __Pyx_GOTREF(__pyx_t_22);
              __pyx_t_10 = __pyx_t_22;
              __pyx_t_22 = 0;
              __pyx_L79_bool_binop_done:;
              if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_asList, __pyx_t_10) < 0) __PYX_ERR(0, 1450, __pyx_L58_error)
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

              /* "Aspidites/_vendor/pyparsing_extension.py":1452
 *                                                      asList=self.saveAsList and isinstance(tokens,
 *                                                                                            (ParseResults, list)),
 *                                                      modal=self.modalResults)             # <<<<<<<<<<<<<<
 *                 except Exception as err:
 *                     # ~ print "Exception raised in user parse action:", err
 */
              __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_modalResults); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1452, __pyx_L58_error)
              __Pyx_GOTREF(__pyx_t_10);
              if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_modal, __pyx_t_10) < 0) __PYX_ERR(0, 1450, __pyx_L58_error)
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

              /* "Aspidites/_vendor/pyparsing_extension.py":1448
 * 
 *                         if tokens is not None and tokens is not retTokens:
 *                             retTokens = ParseResults(tokens,             # <<<<<<<<<<<<<<
 *                                                      self.resultsName,
 *                                                      asList=self.saveAsList and isinstance(tokens,
 */
              __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_23, __pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1448, __pyx_L58_error)
              __Pyx_GOTREF(__pyx_t_10);
              __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_DECREF_SET(__pyx_v_retTokens, __pyx_t_10);
              __pyx_t_10 = 0;

              /* "Aspidites/_vendor/pyparsing_extension.py":1447
 *                             raise exc
 * 
 *                         if tokens is not None and tokens is not retTokens:             # <<<<<<<<<<<<<<
 *                             retTokens = ParseResults(tokens,
 *                                                      self.resultsName,
 */
            }

            /* "Aspidites/_vendor/pyparsing_extension.py":1439
 *             if debugging:
 *                 try:
 *                     for fn in self.parseAction:             # <<<<<<<<<<<<<<
 *                         try:
 *                             tokens = fn(instring, tokensStart, retTokens)
 */
          }
          __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":1438
 *         if self.parseAction and (doActions or self.callDuringTry):
 *             if debugging:
 *                 try:             # <<<<<<<<<<<<<<
 *                     for fn in self.parseAction:
 *                         try:
 */
        }
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        goto __pyx_L63_try_end;
        __pyx_L58_error:;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
        __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
        __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
        __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
        __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":1453
 *                                                                                            (ParseResults, list)),
 *                                                      modal=self.modalResults)
 *                 except Exception as err:             # <<<<<<<<<<<<<<
 *                     # ~ print "Exception raised in user parse action:", err
 *                     if self.debugActions[FAIL]:
 */
        __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
        if (__pyx_t_9) {
          __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement._parseNoCache", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_19, &__pyx_t_10, &__pyx_t_4) < 0) __PYX_ERR(0, 1453, __pyx_L60_except_error)
          __Pyx_GOTREF(__pyx_t_19);
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_10);
          __pyx_v_err = __pyx_t_10;

          /* "Aspidites/_vendor/pyparsing_extension.py":1455
 *                 except Exception as err:
 *                     # ~ print "Exception raised in user parse action:", err
 *                     if self.debugActions[FAIL]:             # <<<<<<<<<<<<<<
 *                         self.debugActions[FAIL](instring, tokensStart, self, err)
 *                     raise
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_debugActions); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1455, __pyx_L60_except_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_23 = __Pyx_GetItemInt(__pyx_t_7, __pyx_v_FAIL, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1455, __pyx_L60_except_error)
          __Pyx_GOTREF(__pyx_t_23);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_23); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1455, __pyx_L60_except_error)
          __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
          if (__pyx_t_5) {

            /* "Aspidites/_vendor/pyparsing_extension.py":1456
 *                     # ~ print "Exception raised in user parse action:", err
 *                     if self.debugActions[FAIL]:
 *                         self.debugActions[FAIL](instring, tokensStart, self, err)             # <<<<<<<<<<<<<<
 *                     raise
 *             else:
 */
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_debugActions); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1456, __pyx_L60_except_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_22 = __Pyx_GetItemInt(__pyx_t_7, __pyx_v_FAIL, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1456, __pyx_L60_except_error)
            __Pyx_GOTREF(__pyx_t_22);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __pyx_t_7 = NULL;
            __pyx_t_9 = 0;
            if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_22))) {
              __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_22);
              if (likely(__pyx_t_7)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_22);
                __Pyx_INCREF(__pyx_t_7);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_22, function);
                __pyx_t_9 = 1;
              }
            }
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_22)) {
              PyObject *__pyx_temp[5] = {__pyx_t_7, __pyx_v_instring, __pyx_v_tokensStart, __pyx_v_self, __pyx_v_err};
              __pyx_t_23 = __Pyx_PyFunction_FastCall(__pyx_t_22, __pyx_temp+1-__pyx_t_9, 4+__pyx_t_9); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1456, __pyx_L60_except_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_23);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_22)) {
              PyObject *__pyx_temp[5] = {__pyx_t_7, __pyx_v_instring, __pyx_v_tokensStart, __pyx_v_self, __pyx_v_err};
              __pyx_t_23 = __Pyx_PyCFunction_FastCall(__pyx_t_22, __pyx_temp+1-__pyx_t_9, 4+__pyx_t_9); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1456, __pyx_L60_except_error)
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_GOTREF(__pyx_t_23);
            } else
            #endif
            {
              __pyx_t_21 = PyTuple_New(4+__pyx_t_9); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1456, __pyx_L60_except_error)
              __Pyx_GOTREF(__pyx_t_21);
              if (__pyx_t_7) {
                __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_21, 0, __pyx_t_7); __pyx_t_7 = NULL;
              }
              __Pyx_INCREF(__pyx_v_instring);
              __Pyx_GIVEREF(__pyx_v_instring);
              PyTuple_SET_ITEM(__pyx_t_21, 0+__pyx_t_9, __pyx_v_instring);
              __Pyx_INCREF(__pyx_v_tokensStart);
              __Pyx_GIVEREF(__pyx_v_tokensStart);
              PyTuple_SET_ITEM(__pyx_t_21, 1+__pyx_t_9, __pyx_v_tokensStart);
              __Pyx_INCREF(__pyx_v_self);
              __Pyx_GIVEREF(__pyx_v_self);
              PyTuple_SET_ITEM(__pyx_t_21, 2+__pyx_t_9, __pyx_v_self);
              __Pyx_INCREF(__pyx_v_err);
              __Pyx_GIVEREF(__pyx_v_err);
              PyTuple_SET_ITEM(__pyx_t_21, 3+__pyx_t_9, __pyx_v_err);
              __pyx_t_23 = __Pyx_PyObject_Call(__pyx_t_22, __pyx_t_21, NULL); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1456, __pyx_L60_except_error)
              __Pyx_GOTREF(__pyx_t_23);
              __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
            }
            __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
            __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

            /* "Aspidites/_vendor/pyparsing_extension.py":1455
 *                 except Exception as err:
 *                     # ~ print "Exception raised in user parse action:", err
 *                     if self.debugActions[FAIL]:             # <<<<<<<<<<<<<<
 *                         self.debugActions[FAIL](instring, tokensStart, self, err)
 *                     raise
 */
          }

          /* "Aspidites/_vendor/pyparsing_extension.py":1457
 *                     if self.debugActions[FAIL]:
 *                         self.debugActions[FAIL](instring, tokensStart, self, err)
 *                     raise             # <<<<<<<<<<<<<<
 *             else:
 *                 for fn in self.parseAction:
 */
          __Pyx_GIVEREF(__pyx_t_19);
          __Pyx_GIVEREF(__pyx_t_10);
          __Pyx_XGIVEREF(__pyx_t_4);
          __Pyx_ErrRestoreWithState(__pyx_t_19, __pyx_t_10, __pyx_t_4);
          __pyx_t_19 = 0; __pyx_t_10 = 0; __pyx_t_4 = 0; 
          __PYX_ERR(0, 1457, __pyx_L60_except_error)
        }
        goto __pyx_L60_except_error;
        __pyx_L60_except_error:;

        /* "Aspidites/_vendor/pyparsing_extension.py":1438
 *         if self.parseAction and (doActions or self.callDuringTry):
 *             if debugging:
 *                 try:             # <<<<<<<<<<<<<<
 *                     for fn in self.parseAction:
 *                         try:
 */
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
        goto __pyx_L1_error;
        __pyx_L63_try_end:;
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":1437
 *         retTokens = ParseResults(tokens, self.resultsName, asList=self.saveAsList, modal=self.modalResults)
 *         if self.parseAction and (doActions or self.callDuringTry):
 *             if debugging:             # <<<<<<<<<<<<<<
 *                 try:
 *                     for fn in self.parseAction:
 */
      goto __pyx_L57;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":1459
 *                     raise
 *             else:
 *                 for fn in self.parseAction:             # <<<<<<<<<<<<<<
 *                     try:
 *                         tokens = fn(instring, tokensStart, retTokens)
 */
    /*else*/ {
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_parseAction); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1459, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
        __pyx_t_10 = __pyx_t_4; __Pyx_INCREF(__pyx_t_10); __pyx_t_14 = 0;
        __pyx_t_24 = NULL;
      } else {
        __pyx_t_14 = -1; __pyx_t_10 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1459, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_24 = Py_TYPE(__pyx_t_10)->tp_iternext; if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 1459, __pyx_L1_error)
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      for (;;) {
        if (likely(!__pyx_t_24)) {
          if (likely(PyList_CheckExact(__pyx_t_10))) {
            if (__pyx_t_14 >= PyList_GET_SIZE(__pyx_t_10)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_4 = PyList_GET_ITEM(__pyx_t_10, __pyx_t_14); __Pyx_INCREF(__pyx_t_4); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 1459, __pyx_L1_error)
            #else
            __pyx_t_4 = PySequence_ITEM(__pyx_t_10, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1459, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            #endif
          } else {
            if (__pyx_t_14 >= PyTuple_GET_SIZE(__pyx_t_10)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_10, __pyx_t_14); __Pyx_INCREF(__pyx_t_4); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 1459, __pyx_L1_error)
            #else
            __pyx_t_4 = PySequence_ITEM(__pyx_t_10, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1459, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            #endif
          }
        } else {
          __pyx_t_4 = __pyx_t_24(__pyx_t_10);
          if (unlikely(!__pyx_t_4)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 1459, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_4);
        }
        __Pyx_XDECREF_SET(__pyx_v_fn, __pyx_t_4);
        __pyx_t_4 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":1460
 *             else:
 *                 for fn in self.parseAction:
 *                     try:             # <<<<<<<<<<<<<<
 *                         tokens = fn(instring, tokensStart, retTokens)
 *                     except IndexError as parse_action_exc:
 */
        {
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __Pyx_ExceptionSave(&__pyx_t_13, &__pyx_t_12, &__pyx_t_11);
          __Pyx_XGOTREF(__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_12);
          __Pyx_XGOTREF(__pyx_t_11);
          /*try:*/ {

            /* "Aspidites/_vendor/pyparsing_extension.py":1461
 *                 for fn in self.parseAction:
 *                     try:
 *                         tokens = fn(instring, tokensStart, retTokens)             # <<<<<<<<<<<<<<
 *                     except IndexError as parse_action_exc:
 *                         exc = ParseException("exception raised in parse action")
 */
            __Pyx_INCREF(__pyx_v_fn);
            __pyx_t_19 = __pyx_v_fn; __pyx_t_23 = NULL;
            __pyx_t_9 = 0;
            if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_19))) {
              __pyx_t_23 = PyMethod_GET_SELF(__pyx_t_19);
              if (likely(__pyx_t_23)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_19);
                __Pyx_INCREF(__pyx_t_23);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_19, function);
                __pyx_t_9 = 1;
              }
            }
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_19)) {
              PyObject *__pyx_temp[4] = {__pyx_t_23, __pyx_v_instring, __pyx_v_tokensStart, __pyx_v_retTokens};
              __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_19, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1461, __pyx_L88_error)
              __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
              __Pyx_GOTREF(__pyx_t_4);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_19)) {
              PyObject *__pyx_temp[4] = {__pyx_t_23, __pyx_v_instring, __pyx_v_tokensStart, __pyx_v_retTokens};
              __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_19, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1461, __pyx_L88_error)
              __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
              __Pyx_GOTREF(__pyx_t_4);
            } else
            #endif
            {
              __pyx_t_22 = PyTuple_New(3+__pyx_t_9); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1461, __pyx_L88_error)
              __Pyx_GOTREF(__pyx_t_22);
              if (__pyx_t_23) {
                __Pyx_GIVEREF(__pyx_t_23); PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_t_23); __pyx_t_23 = NULL;
              }
              __Pyx_INCREF(__pyx_v_instring);
              __Pyx_GIVEREF(__pyx_v_instring);
              PyTuple_SET_ITEM(__pyx_t_22, 0+__pyx_t_9, __pyx_v_instring);
              __Pyx_INCREF(__pyx_v_tokensStart);
              __Pyx_GIVEREF(__pyx_v_tokensStart);
              PyTuple_SET_ITEM(__pyx_t_22, 1+__pyx_t_9, __pyx_v_tokensStart);
              __Pyx_INCREF(__pyx_v_retTokens);
              __Pyx_GIVEREF(__pyx_v_retTokens);
              PyTuple_SET_ITEM(__pyx_t_22, 2+__pyx_t_9, __pyx_v_retTokens);
              __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_19, __pyx_t_22, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1461, __pyx_L88_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
            }
            __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
            __Pyx_DECREF_SET(__pyx_v_tokens, __pyx_t_4);
            __pyx_t_4 = 0;

            /* "Aspidites/_vendor/pyparsing_extension.py":1460
 *             else:
 *                 for fn in self.parseAction:
 *                     try:             # <<<<<<<<<<<<<<
 *                         tokens = fn(instring, tokensStart, retTokens)
 *                     except IndexError as parse_action_exc:
 */
          }
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          goto __pyx_L95_try_end;
          __pyx_L88_error:;
          __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
          __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
          __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
          __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
          __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":1462
 *                     try:
 *                         tokens = fn(instring, tokensStart, retTokens)
 *                     except IndexError as parse_action_exc:             # <<<<<<<<<<<<<<
 *                         exc = ParseException("exception raised in parse action")
 *                         exc.__cause__ = parse_action_exc
 */
          __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_IndexError);
          if (__pyx_t_9) {
            __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement._parseNoCache", __pyx_clineno, __pyx_lineno, __pyx_filename);
            if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_19, &__pyx_t_22) < 0) __PYX_ERR(0, 1462, __pyx_L90_except_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GOTREF(__pyx_t_19);
            __Pyx_GOTREF(__pyx_t_22);
            __Pyx_INCREF(__pyx_t_19);
            __pyx_v_parse_action_exc = __pyx_t_19;

            /* "Aspidites/_vendor/pyparsing_extension.py":1463
 *                         tokens = fn(instring, tokensStart, retTokens)
 *                     except IndexError as parse_action_exc:
 *                         exc = ParseException("exception raised in parse action")             # <<<<<<<<<<<<<<
 *                         exc.__cause__ = parse_action_exc
 *                         raise exc
 */
            __Pyx_GetModuleGlobalName(__pyx_t_21, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1463, __pyx_L90_except_error)
            __Pyx_GOTREF(__pyx_t_21);
            __pyx_t_7 = NULL;
            if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_21))) {
              __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_21);
              if (likely(__pyx_t_7)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_21);
                __Pyx_INCREF(__pyx_t_7);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_21, function);
              }
            }
            __pyx_t_23 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_21, __pyx_t_7, __pyx_kp_s_exception_raised_in_parse_action) : __Pyx_PyObject_CallOneArg(__pyx_t_21, __pyx_kp_s_exception_raised_in_parse_action);
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1463, __pyx_L90_except_error)
            __Pyx_GOTREF(__pyx_t_23);
            __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
            __pyx_v_exc = __pyx_t_23;
            __pyx_t_23 = 0;

            /* "Aspidites/_vendor/pyparsing_extension.py":1464
 *                     except IndexError as parse_action_exc:
 *                         exc = ParseException("exception raised in parse action")
 *                         exc.__cause__ = parse_action_exc             # <<<<<<<<<<<<<<
 *                         raise exc
 * 
 */
            if (__Pyx_PyObject_SetAttrStr(__pyx_v_exc, __pyx_n_s_cause, __pyx_v_parse_action_exc) < 0) __PYX_ERR(0, 1464, __pyx_L90_except_error)

            /* "Aspidites/_vendor/pyparsing_extension.py":1465
 *                         exc = ParseException("exception raised in parse action")
 *                         exc.__cause__ = parse_action_exc
 *                         raise exc             # <<<<<<<<<<<<<<
 * 
 *                     if tokens is not None and tokens is not retTokens:
 */
            __Pyx_Raise(__pyx_v_exc, 0, 0, 0);
            __PYX_ERR(0, 1465, __pyx_L90_except_error)
          }
          goto __pyx_L90_except_error;
          __pyx_L90_except_error:;

          /* "Aspidites/_vendor/pyparsing_extension.py":1460
 *             else:
 *                 for fn in self.parseAction:
 *                     try:             # <<<<<<<<<<<<<<
 *                         tokens = fn(instring, tokensStart, retTokens)
 *                     except IndexError as parse_action_exc:
 */
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_XGIVEREF(__pyx_t_12);
          __Pyx_XGIVEREF(__pyx_t_11);
          __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_12, __pyx_t_11);
          goto __pyx_L1_error;
          __pyx_L95_try_end:;
        }

        /* "Aspidites/_vendor/pyparsing_extension.py":1467
 *                         raise exc
 * 
 *                     if tokens is not None and tokens is not retTokens:             # <<<<<<<<<<<<<<
 *                         retTokens = ParseResults(tokens,
 *                                                  self.resultsName,
 */
        __pyx_t_6 = (__pyx_v_tokens != Py_None);
        __pyx_t_25 = (__pyx_t_6 != 0);
        if (__pyx_t_25) {
        } else {
          __pyx_t_5 = __pyx_t_25;
          goto __pyx_L99_bool_binop_done;
        }
        __pyx_t_25 = (__pyx_v_tokens != __pyx_v_retTokens);
        __pyx_t_6 = (__pyx_t_25 != 0);
        __pyx_t_5 = __pyx_t_6;
        __pyx_L99_bool_binop_done:;
        if (__pyx_t_5) {

          /* "Aspidites/_vendor/pyparsing_extension.py":1468
 * 
 *                     if tokens is not None and tokens is not retTokens:
 *                         retTokens = ParseResults(tokens,             # <<<<<<<<<<<<<<
 *                                                  self.resultsName,
 *                                                  asList=self.saveAsList and isinstance(tokens, (ParseResults, list)),
 */
          __Pyx_GetModuleGlobalName(__pyx_t_22, __pyx_n_s_ParseResults); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 1468, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_22);

          /* "Aspidites/_vendor/pyparsing_extension.py":1469
 *                     if tokens is not None and tokens is not retTokens:
 *                         retTokens = ParseResults(tokens,
 *                                                  self.resultsName,             # <<<<<<<<<<<<<<
 *                                                  asList=self.saveAsList and isinstance(tokens, (ParseResults, list)),
 *                                                  modal=self.modalResults)
 */
          __pyx_t_19 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_resultsName); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1469, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_19);

          /* "Aspidites/_vendor/pyparsing_extension.py":1468
 * 
 *                     if tokens is not None and tokens is not retTokens:
 *                         retTokens = ParseResults(tokens,             # <<<<<<<<<<<<<<
 *                                                  self.resultsName,
 *                                                  asList=self.saveAsList and isinstance(tokens, (ParseResults, list)),
 */
          __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1468, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_INCREF(__pyx_v_tokens);
          __Pyx_GIVEREF(__pyx_v_tokens);
          PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_tokens);
          __Pyx_GIVEREF(__pyx_t_19);
          PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_19);
          __pyx_t_19 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":1470
 *                         retTokens = ParseResults(tokens,
 *                                                  self.resultsName,
 *                                                  asList=self.saveAsList and isinstance(tokens, (ParseResults, list)),             # <<<<<<<<<<<<<<
 *                                                  modal=self.modalResults)
 *         if debugging:
 */
          __pyx_t_19 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1470, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_19);
          __pyx_t_21 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_saveAsList); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1470, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_21);
          __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_21); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1470, __pyx_L1_error)
          if (__pyx_t_5) {
            __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
          } else {
            __Pyx_INCREF(__pyx_t_21);
            __pyx_t_23 = __pyx_t_21;
            __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
            goto __pyx_L101_bool_binop_done;
          }
          __Pyx_GetModuleGlobalName(__pyx_t_21, __pyx_n_s_ParseResults); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1470, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_21);
          __pyx_t_6 = PyObject_IsInstance(__pyx_v_tokens, __pyx_t_21); 
          __pyx_t_25 = (__pyx_t_6 != 0);
          if (!__pyx_t_25) {
          } else {
            __pyx_t_5 = __pyx_t_25;
            goto __pyx_L103_bool_binop_done;
          }
          __pyx_t_25 = PyList_Check(__pyx_v_tokens); 
          __pyx_t_6 = (__pyx_t_25 != 0);
          __pyx_t_5 = __pyx_t_6;
          __pyx_L103_bool_binop_done:;
          __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
          __pyx_t_21 = __Pyx_PyBool_FromLong(__pyx_t_5); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 1470, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_21);
          __pyx_t_23 = __pyx_t_21;
          __pyx_t_21 = 0;
          __pyx_L101_bool_binop_done:;
          if (PyDict_SetItem(__pyx_t_19, __pyx_n_s_asList, __pyx_t_23) < 0) __PYX_ERR(0, 1470, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":1471
 *                                                  self.resultsName,
 *                                                  asList=self.saveAsList and isinstance(tokens, (ParseResults, list)),
 *                                                  modal=self.modalResults)             # <<<<<<<<<<<<<<
 *         if debugging:
 *             # ~ print ("Matched", self, "->", retTokens.asList())
 */
          __pyx_t_23 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_modalResults); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1471, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_23);
          if (PyDict_SetItem(__pyx_t_19, __pyx_n_s_modal, __pyx_t_23) < 0) __PYX_ERR(0, 1470, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":1468
 * 
 *                     if tokens is not None and tokens is not retTokens:
 *                         retTokens = ParseResults(tokens,             # <<<<<<<<<<<<<<
 *                                                  self.resultsName,
 *                                                  asList=self.saveAsList and isinstance(tokens, (ParseResults, list)),
 */
          __pyx_t_23 = __Pyx_PyObject_Call(__pyx_t_22, __pyx_t_4, __pyx_t_19); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1468, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_23);
          __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
          __Pyx_DECREF_SET(__pyx_v_retTokens, __pyx_t_23);
          __pyx_t_23 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":1467
 *                         raise exc
 * 
 *                     if tokens is not None and tokens is not retTokens:             # <<<<<<<<<<<<<<
 *                         retTokens = ParseResults(tokens,
 *                                                  self.resultsName,
 */
        }

        /* "Aspidites/_vendor/pyparsing_extension.py":1459
 *                     raise
 *             else:
 *                 for fn in self.parseAction:             # <<<<<<<<<<<<<<
 *                     try:
 *                         tokens = fn(instring, tokensStart, retTokens)
 */
      }
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __pyx_L57:;

    /* "Aspidites/_vendor/pyparsing_extension.py":1436
 * 
 *         retTokens = ParseResults(tokens, self.resultsName, asList=self.saveAsList, modal=self.modalResults)
 *         if self.parseAction and (doActions or self.callDuringTry):             # <<<<<<<<<<<<<<
 *             if debugging:
 *                 try:
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1472
 *                                                  asList=self.saveAsList and isinstance(tokens, (ParseResults, list)),
 *                                                  modal=self.modalResults)
 *         if debugging:             # <<<<<<<<<<<<<<
 *             # ~ print ("Matched", self, "->", retTokens.asList())
 *             if self.debugActions[MATCH]:
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_debugging); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1472, __pyx_L1_error)
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/pyparsing_extension.py":1474
 *         if debugging:
 *             # ~ print ("Matched", self, "->", retTokens.asList())
 *             if self.debugActions[MATCH]:             # <<<<<<<<<<<<<<
 *                 self.debugActions[MATCH](instring, tokensStart, loc, self, retTokens)
 * 
 */
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_debugActions); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1474, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_23 = __Pyx_GetItemInt(__pyx_t_10, __pyx_v_MATCH, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1474, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_23); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1474, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    if (__pyx_t_5) {

      /* "Aspidites/_vendor/pyparsing_extension.py":1475
 *             # ~ print ("Matched", self, "->", retTokens.asList())
 *             if self.debugActions[MATCH]:
 *                 self.debugActions[MATCH](instring, tokensStart, loc, self, retTokens)             # <<<<<<<<<<<<<<
 * 
 *         return loc, retTokens
 */
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_debugActions); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1475, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_19 = __Pyx_GetItemInt(__pyx_t_10, __pyx_v_MATCH, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1475, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_19);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = NULL;
      __pyx_t_9 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_19))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_19);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_19);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_19, function);
          __pyx_t_9 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_19)) {
        PyObject *__pyx_temp[6] = {__pyx_t_10, __pyx_v_instring, __pyx_v_tokensStart, __pyx_v_loc, __pyx_v_self, __pyx_v_retTokens};
        __pyx_t_23 = __Pyx_PyFunction_FastCall(__pyx_t_19, __pyx_temp+1-__pyx_t_9, 5+__pyx_t_9); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1475, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_GOTREF(__pyx_t_23);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_19)) {
        PyObject *__pyx_temp[6] = {__pyx_t_10, __pyx_v_instring, __pyx_v_tokensStart, __pyx_v_loc, __pyx_v_self, __pyx_v_retTokens};
        __pyx_t_23 = __Pyx_PyCFunction_FastCall(__pyx_t_19, __pyx_temp+1-__pyx_t_9, 5+__pyx_t_9); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1475, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_GOTREF(__pyx_t_23);
      } else
      #endif
      {
        __pyx_t_4 = PyTuple_New(5+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1475, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        if (__pyx_t_10) {
          __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_10); __pyx_t_10 = NULL;
        }
        __Pyx_INCREF(__pyx_v_instring);
        __Pyx_GIVEREF(__pyx_v_instring);
        PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_9, __pyx_v_instring);
        __Pyx_INCREF(__pyx_v_tokensStart);
        __Pyx_GIVEREF(__pyx_v_tokensStart);
        PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_9, __pyx_v_tokensStart);
        __Pyx_INCREF(__pyx_v_loc);
        __Pyx_GIVEREF(__pyx_v_loc);
        PyTuple_SET_ITEM(__pyx_t_4, 2+__pyx_t_9, __pyx_v_loc);
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_4, 3+__pyx_t_9, __pyx_v_self);
        __Pyx_INCREF(__pyx_v_retTokens);
        __Pyx_GIVEREF(__pyx_v_retTokens);
        PyTuple_SET_ITEM(__pyx_t_4, 4+__pyx_t_9, __pyx_v_retTokens);
        __pyx_t_23 = __Pyx_PyObject_Call(__pyx_t_19, __pyx_t_4, NULL); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1475, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_23);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
      __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":1474
 *         if debugging:
 *             # ~ print ("Matched", self, "->", retTokens.asList())
 *             if self.debugActions[MATCH]:             # <<<<<<<<<<<<<<
 *                 self.debugActions[MATCH](instring, tokensStart, loc, self, retTokens)
 * 
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":1472
 *                                                  asList=self.saveAsList and isinstance(tokens, (ParseResults, list)),
 *                                                  modal=self.modalResults)
 *         if debugging:             # <<<<<<<<<<<<<<
 *             # ~ print ("Matched", self, "->", retTokens.asList())
 *             if self.debugActions[MATCH]:
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1477
 *                 self.debugActions[MATCH](instring, tokensStart, loc, self, retTokens)
 * 
 *         return loc, retTokens             # <<<<<<<<<<<<<<
 * 
 *     def tryParse(self, instring, loc):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_23 = PyTuple_New(2); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 1477, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_INCREF(__pyx_v_loc);
  __Pyx_GIVEREF(__pyx_v_loc);
  PyTuple_SET_ITEM(__pyx_t_23, 0, __pyx_v_loc);
  __Pyx_INCREF(__pyx_v_retTokens);
  __Pyx_GIVEREF(__pyx_v_retTokens);
  PyTuple_SET_ITEM(__pyx_t_23, 1, __pyx_v_retTokens);
  __pyx_r = __pyx_t_23;
  __pyx_t_23 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1391
 * 
 *     # ~ @profile
 *     def _parseNoCache(self, instring, loc, doActions=True, callPreParse=True):             # <<<<<<<<<<<<<<
 *         TRY, MATCH, FAIL = 0, 1, 2
 *         debugging = (self.debug)  # and doActions)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_XDECREF(__pyx_t_22);
  __Pyx_XDECREF(__pyx_t_23);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement._parseNoCache", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_debugging);
  __Pyx_XDECREF(__pyx_v_preloc);
  __Pyx_XDECREF(__pyx_v_tokensStart);
  __Pyx_XDECREF(__pyx_v_tokens);
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XDECREF(__pyx_v_retTokens);
  __Pyx_XDECREF(__pyx_v_fn);
  __Pyx_XDECREF(__pyx_v_parse_action_exc);
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XDECREF(__pyx_v_loc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1479
 *         return loc, retTokens
 * 
 *     def tryParse(self, instring, loc):             # <<<<<<<<<<<<<<
 *         try:
 *             return self._parse(instring, loc, doActions=False)[0]
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_37tryParse(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_37tryParse = {"tryParse", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_37tryParse, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_37tryParse(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("tryParse (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("tryParse", 1, 3, 3, 1); __PYX_ERR(0, 1479, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("tryParse", 1, 3, 3, 2); __PYX_ERR(0, 1479, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "tryParse") < 0)) __PYX_ERR(0, 1479, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("tryParse", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1479, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.tryParse", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_36tryParse(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_36tryParse(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("tryParse", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":1480
 * 
 *     def tryParse(self, instring, loc):
 *         try:             # <<<<<<<<<<<<<<
 *             return self._parse(instring, loc, doActions=False)[0]
 *         except ParseFatalException:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "Aspidites/_vendor/pyparsing_extension.py":1481
 *     def tryParse(self, instring, loc):
 *         try:
 *             return self._parse(instring, loc, doActions=False)[0]             # <<<<<<<<<<<<<<
 *         except ParseFatalException:
 *             raise ParseException(instring, loc, self.errmsg, self)
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_parse); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1481, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1481, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_v_instring);
      __Pyx_GIVEREF(__pyx_v_instring);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_instring);
      __Pyx_INCREF(__pyx_v_loc);
      __Pyx_GIVEREF(__pyx_v_loc);
      PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_loc);
      __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1481, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_doActions, Py_False) < 0) __PYX_ERR(0, 1481, __pyx_L3_error)
      __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1481, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_7, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1481, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_r = __pyx_t_6;
      __pyx_t_6 = 0;
      goto __pyx_L7_try_return;

      /* "Aspidites/_vendor/pyparsing_extension.py":1480
 * 
 *     def tryParse(self, instring, loc):
 *         try:             # <<<<<<<<<<<<<<
 *             return self._parse(instring, loc, doActions=False)[0]
 *         except ParseFatalException:
 */
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1482
 *         try:
 *             return self._parse(instring, loc, doActions=False)[0]
 *         except ParseFatalException:             # <<<<<<<<<<<<<<
 *             raise ParseException(instring, loc, self.errmsg, self)
 * 
 */
    __Pyx_ErrFetch(&__pyx_t_6, &__pyx_t_7, &__pyx_t_5);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ParseFatalException); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1482, __pyx_L5_except_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_6, __pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_ErrRestore(__pyx_t_6, __pyx_t_7, __pyx_t_5);
    __pyx_t_6 = 0; __pyx_t_7 = 0; __pyx_t_5 = 0;
    if (__pyx_t_8) {
      __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.tryParse", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_7, &__pyx_t_6) < 0) __PYX_ERR(0, 1482, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_t_6);

      /* "Aspidites/_vendor/pyparsing_extension.py":1483
 *             return self._parse(instring, loc, doActions=False)[0]
 *         except ParseFatalException:
 *             raise ParseException(instring, loc, self.errmsg, self)             # <<<<<<<<<<<<<<
 * 
 *     def canParseNext(self, instring, loc):
 */
      __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1483, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errmsg); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1483, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_11 = NULL;
      __pyx_t_8 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_9);
        if (likely(__pyx_t_11)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_11);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_9, function);
          __pyx_t_8 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_9)) {
        PyObject *__pyx_temp[5] = {__pyx_t_11, __pyx_v_instring, __pyx_v_loc, __pyx_t_10, __pyx_v_self};
        __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_8, 4+__pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1483, __pyx_L5_except_error)
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_9)) {
        PyObject *__pyx_temp[5] = {__pyx_t_11, __pyx_v_instring, __pyx_v_loc, __pyx_t_10, __pyx_v_self};
        __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_8, 4+__pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1483, __pyx_L5_except_error)
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      } else
      #endif
      {
        __pyx_t_12 = PyTuple_New(4+__pyx_t_8); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1483, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_12);
        if (__pyx_t_11) {
          __Pyx_GIVEREF(__pyx_t_11); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_11); __pyx_t_11 = NULL;
        }
        __Pyx_INCREF(__pyx_v_instring);
        __Pyx_GIVEREF(__pyx_v_instring);
        PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_8, __pyx_v_instring);
        __Pyx_INCREF(__pyx_v_loc);
        __Pyx_GIVEREF(__pyx_v_loc);
        PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_8, __pyx_v_loc);
        __Pyx_GIVEREF(__pyx_t_10);
        PyTuple_SET_ITEM(__pyx_t_12, 2+__pyx_t_8, __pyx_t_10);
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_12, 3+__pyx_t_8, __pyx_v_self);
        __pyx_t_10 = 0;
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_12, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1483, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      }
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __PYX_ERR(0, 1483, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "Aspidites/_vendor/pyparsing_extension.py":1480
 * 
 *     def tryParse(self, instring, loc):
 *         try:             # <<<<<<<<<<<<<<
 *             return self._parse(instring, loc, doActions=False)[0]
 *         except ParseFatalException:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1479
 *         return loc, retTokens
 * 
 *     def tryParse(self, instring, loc):             # <<<<<<<<<<<<<<
 *         try:
 *             return self._parse(instring, loc, doActions=False)[0]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.tryParse", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1485
 *             raise ParseException(instring, loc, self.errmsg, self)
 * 
 *     def canParseNext(self, instring, loc):             # <<<<<<<<<<<<<<
 *         try:
 *             self.tryParse(instring, loc)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_39canParseNext(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_39canParseNext = {"canParseNext", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_39canParseNext, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_39canParseNext(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("canParseNext (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("canParseNext", 1, 3, 3, 1); __PYX_ERR(0, 1485, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("canParseNext", 1, 3, 3, 2); __PYX_ERR(0, 1485, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "canParseNext") < 0)) __PYX_ERR(0, 1485, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("canParseNext", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1485, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.canParseNext", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_38canParseNext(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_38canParseNext(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("canParseNext", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":1486
 * 
 *     def canParseNext(self, instring, loc):
 *         try:             # <<<<<<<<<<<<<<
 *             self.tryParse(instring, loc)
 *         except (ParseException, IndexError):
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "Aspidites/_vendor/pyparsing_extension.py":1487
 *     def canParseNext(self, instring, loc):
 *         try:
 *             self.tryParse(instring, loc)             # <<<<<<<<<<<<<<
 *         except (ParseException, IndexError):
 *             return False
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_tryParse); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1487, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_7 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_instring, __pyx_v_loc};
        __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1487, __pyx_L3_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_4);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_instring, __pyx_v_loc};
        __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1487, __pyx_L3_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_4);
      } else
      #endif
      {
        __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1487, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_8);
        if (__pyx_t_6) {
          __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
        }
        __Pyx_INCREF(__pyx_v_instring);
        __Pyx_GIVEREF(__pyx_v_instring);
        PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_v_instring);
        __Pyx_INCREF(__pyx_v_loc);
        __Pyx_GIVEREF(__pyx_v_loc);
        PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_v_loc);
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1487, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":1486
 * 
 *     def canParseNext(self, instring, loc):
 *         try:             # <<<<<<<<<<<<<<
 *             self.tryParse(instring, loc)
 *         except (ParseException, IndexError):
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":1491
 *             return False
 *         else:
 *             return True             # <<<<<<<<<<<<<<
 * 
 *     class _UnboundedCache(object):
 */
    /*else:*/ {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_True);
      __pyx_r = Py_True;
      goto __pyx_L6_except_return;
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1488
 *         try:
 *             self.tryParse(instring, loc)
 *         except (ParseException, IndexError):             # <<<<<<<<<<<<<<
 *             return False
 *         else:
 */
    __Pyx_ErrFetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_8);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1488, __pyx_L5_except_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_4, __pyx_t_6) || __Pyx_PyErr_GivenExceptionMatches(__pyx_t_4, __pyx_builtin_IndexError);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_ErrRestore(__pyx_t_4, __pyx_t_5, __pyx_t_8);
    __pyx_t_4 = 0; __pyx_t_5 = 0; __pyx_t_8 = 0;
    if (__pyx_t_7) {
      __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.canParseNext", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_5, &__pyx_t_4) < 0) __PYX_ERR(0, 1488, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_4);

      /* "Aspidites/_vendor/pyparsing_extension.py":1489
 *             self.tryParse(instring, loc)
 *         except (ParseException, IndexError):
 *             return False             # <<<<<<<<<<<<<<
 *         else:
 *             return True
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L6_except_return;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "Aspidites/_vendor/pyparsing_extension.py":1486
 * 
 *     def canParseNext(self, instring, loc):
 *         try:             # <<<<<<<<<<<<<<
 *             self.tryParse(instring, loc)
 *         except (ParseException, IndexError):
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1485
 *             raise ParseException(instring, loc, self.errmsg, self)
 * 
 *     def canParseNext(self, instring, loc):             # <<<<<<<<<<<<<<
 *         try:
 *             self.tryParse(instring, loc)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.canParseNext", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1494
 * 
 *     class _UnboundedCache(object):
 *         def __init__(self):             # <<<<<<<<<<<<<<
 *             cache = {}
 *             self.not_in_cache = not_in_cache = object()
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15_UnboundedCache_1__init__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_15_UnboundedCache_1__init__ = {"__init__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15_UnboundedCache_1__init__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15_UnboundedCache_1__init__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15_UnboundedCache___init__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1498
 *             self.not_in_cache = not_in_cache = object()
 * 
 *             def get(self, key):             # <<<<<<<<<<<<<<
 *                 return cache.get(key, not_in_cache)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15_UnboundedCache_8__init___1get(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_15_UnboundedCache_8__init___1get = {"get", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15_UnboundedCache_8__init___1get, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15_UnboundedCache_8__init___1get(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_key = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_key,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_key)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get", 1, 2, 2, 1); __PYX_ERR(0, 1498, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "get") < 0)) __PYX_ERR(0, 1498, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_key = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1498, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._UnboundedCache.__init__.get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15_UnboundedCache_8__init___get(__pyx_self, __pyx_v_self, __pyx_v_key);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15_UnboundedCache_8__init___get(PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__ *__pyx_cur_scope;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__ *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get", 0);
  __pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__ *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "Aspidites/_vendor/pyparsing_extension.py":1499
 * 
 *             def get(self, key):
 *                 return cache.get(key, not_in_cache)             # <<<<<<<<<<<<<<
 * 
 *             def set(self, key, value):
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_v_cache)) { __Pyx_RaiseClosureNameError("cache"); __PYX_ERR(0, 1499, __pyx_L1_error) }
  if (unlikely(__pyx_cur_scope->__pyx_v_cache == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "get");
    __PYX_ERR(0, 1499, __pyx_L1_error)
  }
  if (unlikely(!__pyx_cur_scope->__pyx_v_not_in_cache)) { __Pyx_RaiseClosureNameError("not_in_cache"); __PYX_ERR(0, 1499, __pyx_L1_error) }
  __pyx_t_1 = __Pyx_PyDict_GetItemDefault(__pyx_cur_scope->__pyx_v_cache, __pyx_v_key, __pyx_cur_scope->__pyx_v_not_in_cache); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1498
 *             self.not_in_cache = not_in_cache = object()
 * 
 *             def get(self, key):             # <<<<<<<<<<<<<<
 *                 return cache.get(key, not_in_cache)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._UnboundedCache.__init__.get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1501
 *                 return cache.get(key, not_in_cache)
 * 
 *             def set(self, key, value):             # <<<<<<<<<<<<<<
 *                 cache[key] = value
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15_UnboundedCache_8__init___3set(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_15_UnboundedCache_8__init___3set = {"set", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15_UnboundedCache_8__init___3set, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15_UnboundedCache_8__init___3set(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_value = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_key,&__pyx_n_s_value,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_key)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("set", 1, 3, 3, 1); __PYX_ERR(0, 1501, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_value)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("set", 1, 3, 3, 2); __PYX_ERR(0, 1501, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "set") < 0)) __PYX_ERR(0, 1501, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_key = values[1];
    __pyx_v_value = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1501, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._UnboundedCache.__init__.set", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15_UnboundedCache_8__init___2set(__pyx_self, __pyx_v_self, __pyx_v_key, __pyx_v_value);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15_UnboundedCache_8__init___2set(PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_value) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__ *__pyx_cur_scope;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__ *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set", 0);
  __pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__ *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "Aspidites/_vendor/pyparsing_extension.py":1502
 * 
 *             def set(self, key, value):
 *                 cache[key] = value             # <<<<<<<<<<<<<<
 * 
 *             def clear(self):
 */
  if (unlikely(!__pyx_cur_scope->__pyx_v_cache)) { __Pyx_RaiseClosureNameError("cache"); __PYX_ERR(0, 1502, __pyx_L1_error) }
  if (unlikely(__pyx_cur_scope->__pyx_v_cache == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 1502, __pyx_L1_error)
  }
  if (unlikely(PyDict_SetItem(__pyx_cur_scope->__pyx_v_cache, __pyx_v_key, __pyx_v_value) < 0)) __PYX_ERR(0, 1502, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1501
 *                 return cache.get(key, not_in_cache)
 * 
 *             def set(self, key, value):             # <<<<<<<<<<<<<<
 *                 cache[key] = value
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._UnboundedCache.__init__.set", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1504
 *                 cache[key] = value
 * 
 *             def clear(self):             # <<<<<<<<<<<<<<
 *                 cache.clear()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15_UnboundedCache_8__init___5clear(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_15_UnboundedCache_8__init___5clear = {"clear", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15_UnboundedCache_8__init___5clear, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15_UnboundedCache_8__init___5clear(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clear (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15_UnboundedCache_8__init___4clear(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15_UnboundedCache_8__init___4clear(PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__ *__pyx_cur_scope;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__ *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clear", 0);
  __pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__ *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "Aspidites/_vendor/pyparsing_extension.py":1505
 * 
 *             def clear(self):
 *                 cache.clear()             # <<<<<<<<<<<<<<
 * 
 *             def cache_len(self):
 */
  if (unlikely(!__pyx_cur_scope->__pyx_v_cache)) { __Pyx_RaiseClosureNameError("cache"); __PYX_ERR(0, 1505, __pyx_L1_error) }
  if (unlikely(__pyx_cur_scope->__pyx_v_cache == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "clear");
    __PYX_ERR(0, 1505, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_Clear(__pyx_cur_scope->__pyx_v_cache); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 1505, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1504
 *                 cache[key] = value
 * 
 *             def clear(self):             # <<<<<<<<<<<<<<
 *                 cache.clear()
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._UnboundedCache.__init__.clear", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1507
 *                 cache.clear()
 * 
 *             def cache_len(self):             # <<<<<<<<<<<<<<
 *                 return len(cache)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15_UnboundedCache_8__init___7cache_len(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_15_UnboundedCache_8__init___7cache_len = {"cache_len", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15_UnboundedCache_8__init___7cache_len, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15_UnboundedCache_8__init___7cache_len(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cache_len (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15_UnboundedCache_8__init___6cache_len(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15_UnboundedCache_8__init___6cache_len(PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__ *__pyx_cur_scope;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__ *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cache_len", 0);
  __pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__ *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "Aspidites/_vendor/pyparsing_extension.py":1508
 * 
 *             def cache_len(self):
 *                 return len(cache)             # <<<<<<<<<<<<<<
 * 
 *             self.get = types.MethodType(get, self)
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_v_cache)) { __Pyx_RaiseClosureNameError("cache"); __PYX_ERR(0, 1508, __pyx_L1_error) }
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_cache;
  __Pyx_INCREF(__pyx_t_1);
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 1508, __pyx_L1_error)
  }
  __pyx_t_2 = PyDict_Size(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1508, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1508, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1507
 *                 cache.clear()
 * 
 *             def cache_len(self):             # <<<<<<<<<<<<<<
 *                 return len(cache)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._UnboundedCache.__init__.cache_len", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1494
 * 
 *     class _UnboundedCache(object):
 *         def __init__(self):             # <<<<<<<<<<<<<<
 *             cache = {}
 *             self.not_in_cache = not_in_cache = object()
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15_UnboundedCache___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__ *__pyx_cur_scope;
  PyObject *__pyx_v_get = 0;
  PyObject *__pyx_v_set = 0;
  PyObject *__pyx_v_clear = 0;
  PyObject *__pyx_v_cache_len = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__ *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1494, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1495
 *     class _UnboundedCache(object):
 *         def __init__(self):
 *             cache = {}             # <<<<<<<<<<<<<<
 *             self.not_in_cache = not_in_cache = object()
 * 
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_cache = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1496
 *         def __init__(self):
 *             cache = {}
 *             self.not_in_cache = not_in_cache = object()             # <<<<<<<<<<<<<<
 * 
 *             def get(self, key):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_builtin_object); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_not_in_cache, __pyx_t_1) < 0) __PYX_ERR(0, 1496, __pyx_L1_error)
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_not_in_cache = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1498
 *             self.not_in_cache = not_in_cache = object()
 * 
 *             def get(self, key):             # <<<<<<<<<<<<<<
 *                 return cache.get(key, not_in_cache)
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_15_UnboundedCache_8__init___1get, 0, __pyx_n_s_ParserElement__UnboundedCache, ((PyObject*)__pyx_cur_scope), __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__40)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_get = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1501
 *                 return cache.get(key, not_in_cache)
 * 
 *             def set(self, key, value):             # <<<<<<<<<<<<<<
 *                 cache[key] = value
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_15_UnboundedCache_8__init___3set, 0, __pyx_n_s_ParserElement__UnboundedCache_2, ((PyObject*)__pyx_cur_scope), __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__42)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_set = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1504
 *                 cache[key] = value
 * 
 *             def clear(self):             # <<<<<<<<<<<<<<
 *                 cache.clear()
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_15_UnboundedCache_8__init___5clear, 0, __pyx_n_s_ParserElement__UnboundedCache_3, ((PyObject*)__pyx_cur_scope), __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__44)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_clear = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1507
 *                 cache.clear()
 * 
 *             def cache_len(self):             # <<<<<<<<<<<<<<
 *                 return len(cache)
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_15_UnboundedCache_8__init___7cache_len, 0, __pyx_n_s_ParserElement__UnboundedCache_4, ((PyObject*)__pyx_cur_scope), __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__46)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1507, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_cache_len = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1510
 *                 return len(cache)
 * 
 *             self.get = types.MethodType(get, self)             # <<<<<<<<<<<<<<
 *             self.set = types.MethodType(set, self)
 *             self.clear = types.MethodType(clear, self)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_types); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_MethodType); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_get, __pyx_v_self};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1510, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_get, __pyx_v_self};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1510, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_get);
    __Pyx_GIVEREF(__pyx_v_get);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_get);
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_self);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_get, __pyx_t_1) < 0) __PYX_ERR(0, 1510, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1511
 * 
 *             self.get = types.MethodType(get, self)
 *             self.set = types.MethodType(set, self)             # <<<<<<<<<<<<<<
 *             self.clear = types.MethodType(clear, self)
 *             self.__len__ = types.MethodType(cache_len, self)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_types); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_MethodType); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_set, __pyx_v_self};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1511, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_set, __pyx_v_self};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1511, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1511, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_set);
    __Pyx_GIVEREF(__pyx_v_set);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_4, __pyx_v_set);
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_4, __pyx_v_self);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1511, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_set, __pyx_t_1) < 0) __PYX_ERR(0, 1511, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1512
 *             self.get = types.MethodType(get, self)
 *             self.set = types.MethodType(set, self)
 *             self.clear = types.MethodType(clear, self)             # <<<<<<<<<<<<<<
 *             self.__len__ = types.MethodType(cache_len, self)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_types); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_MethodType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_clear, __pyx_v_self};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1512, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_clear, __pyx_v_self};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1512, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_3 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_clear);
    __Pyx_GIVEREF(__pyx_v_clear);
    PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_4, __pyx_v_clear);
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_4, __pyx_v_self);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_clear, __pyx_t_1) < 0) __PYX_ERR(0, 1512, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1513
 *             self.set = types.MethodType(set, self)
 *             self.clear = types.MethodType(clear, self)
 *             self.__len__ = types.MethodType(cache_len, self)             # <<<<<<<<<<<<<<
 * 
 *     if _OrderedDict is not None:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_types); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_MethodType); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_cache_len, __pyx_v_self};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1513, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_cache_len, __pyx_v_self};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1513, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1513, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_cache_len);
    __Pyx_GIVEREF(__pyx_v_cache_len);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_cache_len);
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_self);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1513, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_len, __pyx_t_1) < 0) __PYX_ERR(0, 1513, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1494
 * 
 *     class _UnboundedCache(object):
 *         def __init__(self):             # <<<<<<<<<<<<<<
 *             cache = {}
 *             self.not_in_cache = not_in_cache = object()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._UnboundedCache.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_get);
  __Pyx_XDECREF(__pyx_v_set);
  __Pyx_XDECREF(__pyx_v_clear);
  __Pyx_XDECREF(__pyx_v_cache_len);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1517
 *     if _OrderedDict is not None:
 *         class _FifoCache(object):
 *             def __init__(self, size):             # <<<<<<<<<<<<<<
 *                 self.not_in_cache = not_in_cache = object()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_size = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_size,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(0, 1517, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 1517, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_size = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1517, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._FifoCache.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache___init__(__pyx_self, __pyx_v_self, __pyx_v_size);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1522
 *                 cache = _OrderedDict()
 * 
 *                 def get(self, key):             # <<<<<<<<<<<<<<
 *                     return cache.get(key, not_in_cache)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___1get(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___1get = {"get", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___1get, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___1get(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_key = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_key,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_key)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get", 1, 2, 2, 1); __PYX_ERR(0, 1522, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "get") < 0)) __PYX_ERR(0, 1522, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_key = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1522, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._FifoCache.__init__.get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___get(__pyx_self, __pyx_v_self, __pyx_v_key);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___get(PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__ *__pyx_cur_scope;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__ *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get", 0);
  __pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__ *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "Aspidites/_vendor/pyparsing_extension.py":1523
 * 
 *                 def get(self, key):
 *                     return cache.get(key, not_in_cache)             # <<<<<<<<<<<<<<
 * 
 *                 def set(self, key, value):
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_v_cache)) { __Pyx_RaiseClosureNameError("cache"); __PYX_ERR(0, 1523, __pyx_L1_error) }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_cache, __pyx_n_s_get); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1523, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(!__pyx_cur_scope->__pyx_v_not_in_cache)) { __Pyx_RaiseClosureNameError("not_in_cache"); __PYX_ERR(0, 1523, __pyx_L1_error) }
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_key, __pyx_cur_scope->__pyx_v_not_in_cache};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1523, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_key, __pyx_cur_scope->__pyx_v_not_in_cache};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1523, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1523, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_key);
    __Pyx_GIVEREF(__pyx_v_key);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_key);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_not_in_cache);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_not_in_cache);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_cur_scope->__pyx_v_not_in_cache);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1523, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1522
 *                 cache = _OrderedDict()
 * 
 *                 def get(self, key):             # <<<<<<<<<<<<<<
 *                     return cache.get(key, not_in_cache)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._FifoCache.__init__.get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1525
 *                     return cache.get(key, not_in_cache)
 * 
 *                 def set(self, key, value):             # <<<<<<<<<<<<<<
 *                     cache[key] = value
 *                     while len(cache) > size:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___3set(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___3set = {"set", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___3set, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___3set(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_value = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_key,&__pyx_n_s_value,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_key)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("set", 1, 3, 3, 1); __PYX_ERR(0, 1525, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_value)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("set", 1, 3, 3, 2); __PYX_ERR(0, 1525, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "set") < 0)) __PYX_ERR(0, 1525, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_key = values[1];
    __pyx_v_value = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1525, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._FifoCache.__init__.set", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___2set(__pyx_self, __pyx_v_self, __pyx_v_key, __pyx_v_value);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___2set(PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_value) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__ *__pyx_cur_scope;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__ *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set", 0);
  __pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__ *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "Aspidites/_vendor/pyparsing_extension.py":1526
 * 
 *                 def set(self, key, value):
 *                     cache[key] = value             # <<<<<<<<<<<<<<
 *                     while len(cache) > size:
 *                         try:
 */
  if (unlikely(!__pyx_cur_scope->__pyx_v_cache)) { __Pyx_RaiseClosureNameError("cache"); __PYX_ERR(0, 1526, __pyx_L1_error) }
  if (unlikely(PyObject_SetItem(__pyx_cur_scope->__pyx_v_cache, __pyx_v_key, __pyx_v_value) < 0)) __PYX_ERR(0, 1526, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1527
 *                 def set(self, key, value):
 *                     cache[key] = value
 *                     while len(cache) > size:             # <<<<<<<<<<<<<<
 *                         try:
 *                             cache.popitem(False)
 */
  while (1) {
    if (unlikely(!__pyx_cur_scope->__pyx_v_cache)) { __Pyx_RaiseClosureNameError("cache"); __PYX_ERR(0, 1527, __pyx_L1_error) }
    __pyx_t_1 = __pyx_cur_scope->__pyx_v_cache;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1527, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1527, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(!__pyx_cur_scope->__pyx_v_size)) { __Pyx_RaiseClosureNameError("size"); __PYX_ERR(0, 1527, __pyx_L1_error) }
    __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_cur_scope->__pyx_v_size, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1527, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1527, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (!__pyx_t_4) break;

    /* "Aspidites/_vendor/pyparsing_extension.py":1528
 *                     cache[key] = value
 *                     while len(cache) > size:
 *                         try:             # <<<<<<<<<<<<<<
 *                             cache.popitem(False)
 *                         except KeyError:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      /*try:*/ {

        /* "Aspidites/_vendor/pyparsing_extension.py":1529
 *                     while len(cache) > size:
 *                         try:
 *                             cache.popitem(False)             # <<<<<<<<<<<<<<
 *                         except KeyError:
 *                             pass
 */
        if (unlikely(!__pyx_cur_scope->__pyx_v_cache)) { __Pyx_RaiseClosureNameError("cache"); __PYX_ERR(0, 1529, __pyx_L5_error) }
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_cache, __pyx_n_s_popitem); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1529, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_8 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
          if (likely(__pyx_t_8)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_1, function);
          }
        }
        __pyx_t_3 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_8, Py_False) : __Pyx_PyObject_CallOneArg(__pyx_t_1, Py_False);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1529, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":1528
 *                     cache[key] = value
 *                     while len(cache) > size:
 *                         try:             # <<<<<<<<<<<<<<
 *                             cache.popitem(False)
 *                         except KeyError:
 */
      }
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L12_try_end;
      __pyx_L5_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":1530
 *                         try:
 *                             cache.popitem(False)
 *                         except KeyError:             # <<<<<<<<<<<<<<
 *                             pass
 * 
 */
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
      if (__pyx_t_9) {
        __Pyx_ErrRestore(0,0,0);
        goto __pyx_L6_exception_handled;
      }
      goto __pyx_L7_except_error;
      __pyx_L7_except_error:;

      /* "Aspidites/_vendor/pyparsing_extension.py":1528
 *                     cache[key] = value
 *                     while len(cache) > size:
 *                         try:             # <<<<<<<<<<<<<<
 *                             cache.popitem(False)
 *                         except KeyError:
 */
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
      goto __pyx_L1_error;
      __pyx_L6_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
      __pyx_L12_try_end:;
    }
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1525
 *                     return cache.get(key, not_in_cache)
 * 
 *                 def set(self, key, value):             # <<<<<<<<<<<<<<
 *                     cache[key] = value
 *                     while len(cache) > size:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._FifoCache.__init__.set", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1533
 *                             pass
 * 
 *                 def clear(self):             # <<<<<<<<<<<<<<
 *                     cache.clear()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___5clear(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___5clear = {"clear", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___5clear, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___5clear(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clear (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___4clear(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___4clear(PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__ *__pyx_cur_scope;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__ *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clear", 0);
  __pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__ *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "Aspidites/_vendor/pyparsing_extension.py":1534
 * 
 *                 def clear(self):
 *                     cache.clear()             # <<<<<<<<<<<<<<
 * 
 *                 def cache_len(self):
 */
  if (unlikely(!__pyx_cur_scope->__pyx_v_cache)) { __Pyx_RaiseClosureNameError("cache"); __PYX_ERR(0, 1534, __pyx_L1_error) }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_cache, __pyx_n_s_clear); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1533
 *                             pass
 * 
 *                 def clear(self):             # <<<<<<<<<<<<<<
 *                     cache.clear()
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._FifoCache.__init__.clear", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1536
 *                     cache.clear()
 * 
 *                 def cache_len(self):             # <<<<<<<<<<<<<<
 *                     return len(cache)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___7cache_len(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___7cache_len = {"cache_len", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___7cache_len, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___7cache_len(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cache_len (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___6cache_len(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___6cache_len(PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__ *__pyx_cur_scope;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__ *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cache_len", 0);
  __pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__ *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "Aspidites/_vendor/pyparsing_extension.py":1537
 * 
 *                 def cache_len(self):
 *                     return len(cache)             # <<<<<<<<<<<<<<
 * 
 *                 self.get = types.MethodType(get, self)
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_v_cache)) { __Pyx_RaiseClosureNameError("cache"); __PYX_ERR(0, 1537, __pyx_L1_error) }
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_cache;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1537, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1536
 *                     cache.clear()
 * 
 *                 def cache_len(self):             # <<<<<<<<<<<<<<
 *                     return len(cache)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._FifoCache.__init__.cache_len", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1517
 *     if _OrderedDict is not None:
 *         class _FifoCache(object):
 *             def __init__(self, size):             # <<<<<<<<<<<<<<
 *                 self.not_in_cache = not_in_cache = object()
 * 
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_size) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__ *__pyx_cur_scope;
  PyObject *__pyx_v_get = 0;
  PyObject *__pyx_v_set = 0;
  PyObject *__pyx_v_clear = 0;
  PyObject *__pyx_v_cache_len = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__ *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1517, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_size = __pyx_v_size;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_size);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_size);

  /* "Aspidites/_vendor/pyparsing_extension.py":1518
 *         class _FifoCache(object):
 *             def __init__(self, size):
 *                 self.not_in_cache = not_in_cache = object()             # <<<<<<<<<<<<<<
 * 
 *                 cache = _OrderedDict()
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_builtin_object); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_not_in_cache, __pyx_t_1) < 0) __PYX_ERR(0, 1518, __pyx_L1_error)
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_not_in_cache = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1520
 *                 self.not_in_cache = not_in_cache = object()
 * 
 *                 cache = _OrderedDict()             # <<<<<<<<<<<<<<
 * 
 *                 def get(self, key):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_OrderedDict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_cache = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1522
 *                 cache = _OrderedDict()
 * 
 *                 def get(self, key):             # <<<<<<<<<<<<<<
 *                     return cache.get(key, not_in_cache)
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___1get, 0, __pyx_n_s_ParserElement__FifoCache___init, ((PyObject*)__pyx_cur_scope), __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__48)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1522, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_get = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1525
 *                     return cache.get(key, not_in_cache)
 * 
 *                 def set(self, key, value):             # <<<<<<<<<<<<<<
 *                     cache[key] = value
 *                     while len(cache) > size:
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___3set, 0, __pyx_n_s_ParserElement__FifoCache___init_2, ((PyObject*)__pyx_cur_scope), __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__50)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_set = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1533
 *                             pass
 * 
 *                 def clear(self):             # <<<<<<<<<<<<<<
 *                     cache.clear()
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___5clear, 0, __pyx_n_s_ParserElement__FifoCache___init_3, ((PyObject*)__pyx_cur_scope), __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__52)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_clear = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1536
 *                     cache.clear()
 * 
 *                 def cache_len(self):             # <<<<<<<<<<<<<<
 *                     return len(cache)
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___7cache_len, 0, __pyx_n_s_ParserElement__FifoCache___init_4, ((PyObject*)__pyx_cur_scope), __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__54)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_cache_len = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1539
 *                     return len(cache)
 * 
 *                 self.get = types.MethodType(get, self)             # <<<<<<<<<<<<<<
 *                 self.set = types.MethodType(set, self)
 *                 self.clear = types.MethodType(clear, self)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_types); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1539, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_MethodType); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1539, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_get, __pyx_v_self};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1539, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_get, __pyx_v_self};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1539, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1539, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_get);
    __Pyx_GIVEREF(__pyx_v_get);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_get);
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_self);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1539, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_get, __pyx_t_1) < 0) __PYX_ERR(0, 1539, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1540
 * 
 *                 self.get = types.MethodType(get, self)
 *                 self.set = types.MethodType(set, self)             # <<<<<<<<<<<<<<
 *                 self.clear = types.MethodType(clear, self)
 *                 self.__len__ = types.MethodType(cache_len, self)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_types); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_MethodType); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_set, __pyx_v_self};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1540, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_set, __pyx_v_self};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1540, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_set);
    __Pyx_GIVEREF(__pyx_v_set);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_4, __pyx_v_set);
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_4, __pyx_v_self);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_set, __pyx_t_1) < 0) __PYX_ERR(0, 1540, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1541
 *                 self.get = types.MethodType(get, self)
 *                 self.set = types.MethodType(set, self)
 *                 self.clear = types.MethodType(clear, self)             # <<<<<<<<<<<<<<
 *                 self.__len__ = types.MethodType(cache_len, self)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_types); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_MethodType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_clear, __pyx_v_self};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1541, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_clear, __pyx_v_self};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1541, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_3 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_clear);
    __Pyx_GIVEREF(__pyx_v_clear);
    PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_4, __pyx_v_clear);
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_4, __pyx_v_self);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_clear, __pyx_t_1) < 0) __PYX_ERR(0, 1541, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1542
 *                 self.set = types.MethodType(set, self)
 *                 self.clear = types.MethodType(clear, self)
 *                 self.__len__ = types.MethodType(cache_len, self)             # <<<<<<<<<<<<<<
 * 
 *     else:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_types); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_MethodType); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_cache_len, __pyx_v_self};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1542, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_cache_len, __pyx_v_self};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1542, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1542, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_cache_len);
    __Pyx_GIVEREF(__pyx_v_cache_len);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_cache_len);
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_self);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1542, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_len, __pyx_t_1) < 0) __PYX_ERR(0, 1542, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1517
 *     if _OrderedDict is not None:
 *         class _FifoCache(object):
 *             def __init__(self, size):             # <<<<<<<<<<<<<<
 *                 self.not_in_cache = not_in_cache = object()
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._FifoCache.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_get);
  __Pyx_XDECREF(__pyx_v_set);
  __Pyx_XDECREF(__pyx_v_clear);
  __Pyx_XDECREF(__pyx_v_cache_len);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1546
 *     else:
 *         class _FifoCache(object):
 *             def __init__(self, size):             # <<<<<<<<<<<<<<
 *                 self.not_in_cache = not_in_cache = object()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_3__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_3__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_3__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_3__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_size = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_size,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(0, 1546, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 1546, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_size = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1546, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._FifoCache.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_2__init__(__pyx_self, __pyx_v_self, __pyx_v_size);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1552
 *                 key_fifo = collections.deque([], size)
 * 
 *                 def get(self, key):             # <<<<<<<<<<<<<<
 *                     return cache.get(key, not_in_cache)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___9get(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___9get = {"get", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___9get, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___9get(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_key = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_key,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_key)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get", 1, 2, 2, 1); __PYX_ERR(0, 1552, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "get") < 0)) __PYX_ERR(0, 1552, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_key = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1552, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._FifoCache.__init__.get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___8get(__pyx_self, __pyx_v_self, __pyx_v_key);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___8get(PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__ *__pyx_cur_scope;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__ *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get", 0);
  __pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__ *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "Aspidites/_vendor/pyparsing_extension.py":1553
 * 
 *                 def get(self, key):
 *                     return cache.get(key, not_in_cache)             # <<<<<<<<<<<<<<
 * 
 *                 def set(self, key, value):
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_v_cache)) { __Pyx_RaiseClosureNameError("cache"); __PYX_ERR(0, 1553, __pyx_L1_error) }
  if (unlikely(__pyx_cur_scope->__pyx_v_cache == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "get");
    __PYX_ERR(0, 1553, __pyx_L1_error)
  }
  if (unlikely(!__pyx_cur_scope->__pyx_v_not_in_cache)) { __Pyx_RaiseClosureNameError("not_in_cache"); __PYX_ERR(0, 1553, __pyx_L1_error) }
  __pyx_t_1 = __Pyx_PyDict_GetItemDefault(__pyx_cur_scope->__pyx_v_cache, __pyx_v_key, __pyx_cur_scope->__pyx_v_not_in_cache); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1553, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1552
 *                 key_fifo = collections.deque([], size)
 * 
 *                 def get(self, key):             # <<<<<<<<<<<<<<
 *                     return cache.get(key, not_in_cache)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._FifoCache.__init__.get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1555
 *                     return cache.get(key, not_in_cache)
 * 
 *                 def set(self, key, value):             # <<<<<<<<<<<<<<
 *                     cache[key] = value
 *                     while len(key_fifo) > size:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___11set(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___11set = {"set", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___11set, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___11set(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_value = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_key,&__pyx_n_s_value,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_key)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("set", 1, 3, 3, 1); __PYX_ERR(0, 1555, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_value)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("set", 1, 3, 3, 2); __PYX_ERR(0, 1555, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "set") < 0)) __PYX_ERR(0, 1555, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_key = values[1];
    __pyx_v_value = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1555, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._FifoCache.__init__.set", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___10set(__pyx_self, __pyx_v_self, __pyx_v_key, __pyx_v_value);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___10set(PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_value) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__ *__pyx_cur_scope;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__ *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set", 0);
  __pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__ *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "Aspidites/_vendor/pyparsing_extension.py":1556
 * 
 *                 def set(self, key, value):
 *                     cache[key] = value             # <<<<<<<<<<<<<<
 *                     while len(key_fifo) > size:
 *                         cache.pop(key_fifo.popleft(), None)
 */
  if (unlikely(!__pyx_cur_scope->__pyx_v_cache)) { __Pyx_RaiseClosureNameError("cache"); __PYX_ERR(0, 1556, __pyx_L1_error) }
  if (unlikely(__pyx_cur_scope->__pyx_v_cache == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 1556, __pyx_L1_error)
  }
  if (unlikely(PyDict_SetItem(__pyx_cur_scope->__pyx_v_cache, __pyx_v_key, __pyx_v_value) < 0)) __PYX_ERR(0, 1556, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1557
 *                 def set(self, key, value):
 *                     cache[key] = value
 *                     while len(key_fifo) > size:             # <<<<<<<<<<<<<<
 *                         cache.pop(key_fifo.popleft(), None)
 *                     key_fifo.append(key)
 */
  while (1) {
    if (unlikely(!__pyx_cur_scope->__pyx_v_key_fifo)) { __Pyx_RaiseClosureNameError("key_fifo"); __PYX_ERR(0, 1557, __pyx_L1_error) }
    __pyx_t_1 = __pyx_cur_scope->__pyx_v_key_fifo;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1557, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1557, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(!__pyx_cur_scope->__pyx_v_size)) { __Pyx_RaiseClosureNameError("size"); __PYX_ERR(0, 1557, __pyx_L1_error) }
    __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_cur_scope->__pyx_v_size, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1557, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1557, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (!__pyx_t_4) break;

    /* "Aspidites/_vendor/pyparsing_extension.py":1558
 *                     cache[key] = value
 *                     while len(key_fifo) > size:
 *                         cache.pop(key_fifo.popleft(), None)             # <<<<<<<<<<<<<<
 *                     key_fifo.append(key)
 * 
 */
    if (unlikely(!__pyx_cur_scope->__pyx_v_cache)) { __Pyx_RaiseClosureNameError("cache"); __PYX_ERR(0, 1558, __pyx_L1_error) }
    if (unlikely(__pyx_cur_scope->__pyx_v_cache == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "pop");
      __PYX_ERR(0, 1558, __pyx_L1_error)
    }
    if (unlikely(!__pyx_cur_scope->__pyx_v_key_fifo)) { __Pyx_RaiseClosureNameError("key_fifo"); __PYX_ERR(0, 1558, __pyx_L1_error) }
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_key_fifo, __pyx_n_s_popleft); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1558, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1558, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyDict_Pop(__pyx_cur_scope->__pyx_v_cache, __pyx_t_3, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1558, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1559
 *                     while len(key_fifo) > size:
 *                         cache.pop(key_fifo.popleft(), None)
 *                     key_fifo.append(key)             # <<<<<<<<<<<<<<
 * 
 *                 def clear(self):
 */
  if (unlikely(!__pyx_cur_scope->__pyx_v_key_fifo)) { __Pyx_RaiseClosureNameError("key_fifo"); __PYX_ERR(0, 1559, __pyx_L1_error) }
  __pyx_t_6 = __Pyx_PyObject_Append(__pyx_cur_scope->__pyx_v_key_fifo, __pyx_v_key); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1559, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1555
 *                     return cache.get(key, not_in_cache)
 * 
 *                 def set(self, key, value):             # <<<<<<<<<<<<<<
 *                     cache[key] = value
 *                     while len(key_fifo) > size:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._FifoCache.__init__.set", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1561
 *                     key_fifo.append(key)
 * 
 *                 def clear(self):             # <<<<<<<<<<<<<<
 *                     cache.clear()
 *                     key_fifo.clear()
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___13clear(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___13clear = {"clear", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___13clear, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___13clear(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clear (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___12clear(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___12clear(PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__ *__pyx_cur_scope;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__ *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clear", 0);
  __pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__ *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "Aspidites/_vendor/pyparsing_extension.py":1562
 * 
 *                 def clear(self):
 *                     cache.clear()             # <<<<<<<<<<<<<<
 *                     key_fifo.clear()
 * 
 */
  if (unlikely(!__pyx_cur_scope->__pyx_v_cache)) { __Pyx_RaiseClosureNameError("cache"); __PYX_ERR(0, 1562, __pyx_L1_error) }
  if (unlikely(__pyx_cur_scope->__pyx_v_cache == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "clear");
    __PYX_ERR(0, 1562, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_Clear(__pyx_cur_scope->__pyx_v_cache); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 1562, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1563
 *                 def clear(self):
 *                     cache.clear()
 *                     key_fifo.clear()             # <<<<<<<<<<<<<<
 * 
 *                 def cache_len(self):
 */
  if (unlikely(!__pyx_cur_scope->__pyx_v_key_fifo)) { __Pyx_RaiseClosureNameError("key_fifo"); __PYX_ERR(0, 1563, __pyx_L1_error) }
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_key_fifo, __pyx_n_s_clear); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1561
 *                     key_fifo.append(key)
 * 
 *                 def clear(self):             # <<<<<<<<<<<<<<
 *                     cache.clear()
 *                     key_fifo.clear()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._FifoCache.__init__.clear", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1565
 *                     key_fifo.clear()
 * 
 *                 def cache_len(self):             # <<<<<<<<<<<<<<
 *                     return len(cache)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___15cache_len(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___15cache_len = {"cache_len", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___15cache_len, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___15cache_len(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cache_len (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___14cache_len(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___14cache_len(PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__ *__pyx_cur_scope;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__ *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cache_len", 0);
  __pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__ *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "Aspidites/_vendor/pyparsing_extension.py":1566
 * 
 *                 def cache_len(self):
 *                     return len(cache)             # <<<<<<<<<<<<<<
 * 
 *                 self.get = types.MethodType(get, self)
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_v_cache)) { __Pyx_RaiseClosureNameError("cache"); __PYX_ERR(0, 1566, __pyx_L1_error) }
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_cache;
  __Pyx_INCREF(__pyx_t_1);
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 1566, __pyx_L1_error)
  }
  __pyx_t_2 = PyDict_Size(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1566, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1565
 *                     key_fifo.clear()
 * 
 *                 def cache_len(self):             # <<<<<<<<<<<<<<
 *                     return len(cache)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._FifoCache.__init__.cache_len", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1546
 *     else:
 *         class _FifoCache(object):
 *             def __init__(self, size):             # <<<<<<<<<<<<<<
 *                 self.not_in_cache = not_in_cache = object()
 * 
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_2__init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_size) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__ *__pyx_cur_scope;
  PyObject *__pyx_v_get = 0;
  PyObject *__pyx_v_set = 0;
  PyObject *__pyx_v_clear = 0;
  PyObject *__pyx_v_cache_len = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__ *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1546, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_size = __pyx_v_size;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_size);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_size);

  /* "Aspidites/_vendor/pyparsing_extension.py":1547
 *         class _FifoCache(object):
 *             def __init__(self, size):
 *                 self.not_in_cache = not_in_cache = object()             # <<<<<<<<<<<<<<
 * 
 *                 cache = {}
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_builtin_object); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1547, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_not_in_cache, __pyx_t_1) < 0) __PYX_ERR(0, 1547, __pyx_L1_error)
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_not_in_cache = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1549
 *                 self.not_in_cache = not_in_cache = object()
 * 
 *                 cache = {}             # <<<<<<<<<<<<<<
 *                 key_fifo = collections.deque([], size)
 * 
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_cache = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1550
 * 
 *                 cache = {}
 *                 key_fifo = collections.deque([], size)             # <<<<<<<<<<<<<<
 * 
 *                 def get(self, key):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_collections); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_deque); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_2, __pyx_cur_scope->__pyx_v_size};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1550, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_2, __pyx_cur_scope->__pyx_v_size};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1550, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1550, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_2);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_size);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_size);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_cur_scope->__pyx_v_size);
    __pyx_t_2 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1550, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_key_fifo = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1552
 *                 key_fifo = collections.deque([], size)
 * 
 *                 def get(self, key):             # <<<<<<<<<<<<<<
 *                     return cache.get(key, not_in_cache)
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___9get, 0, __pyx_n_s_ParserElement__FifoCache___init, ((PyObject*)__pyx_cur_scope), __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__56)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_get = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1555
 *                     return cache.get(key, not_in_cache)
 * 
 *                 def set(self, key, value):             # <<<<<<<<<<<<<<
 *                     cache[key] = value
 *                     while len(key_fifo) > size:
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___11set, 0, __pyx_n_s_ParserElement__FifoCache___init_2, ((PyObject*)__pyx_cur_scope), __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__58)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1555, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_set = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1561
 *                     key_fifo.append(key)
 * 
 *                 def clear(self):             # <<<<<<<<<<<<<<
 *                     cache.clear()
 *                     key_fifo.clear()
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___13clear, 0, __pyx_n_s_ParserElement__FifoCache___init_3, ((PyObject*)__pyx_cur_scope), __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__60)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_clear = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1565
 *                     key_fifo.clear()
 * 
 *                 def cache_len(self):             # <<<<<<<<<<<<<<
 *                     return len(cache)
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_8__init___15cache_len, 0, __pyx_n_s_ParserElement__FifoCache___init_4, ((PyObject*)__pyx_cur_scope), __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__62)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_cache_len = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1568
 *                     return len(cache)
 * 
 *                 self.get = types.MethodType(get, self)             # <<<<<<<<<<<<<<
 *                 self.set = types.MethodType(set, self)
 *                 self.clear = types.MethodType(clear, self)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_types); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1568, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_MethodType); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1568, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_get, __pyx_v_self};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1568, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_get, __pyx_v_self};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1568, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1568, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_get);
    __Pyx_GIVEREF(__pyx_v_get);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_5, __pyx_v_get);
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_5, __pyx_v_self);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1568, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_get, __pyx_t_1) < 0) __PYX_ERR(0, 1568, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1569
 * 
 *                 self.get = types.MethodType(get, self)
 *                 self.set = types.MethodType(set, self)             # <<<<<<<<<<<<<<
 *                 self.clear = types.MethodType(clear, self)
 *                 self.__len__ = types.MethodType(cache_len, self)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_types); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_MethodType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_set, __pyx_v_self};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1569, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_set, __pyx_v_self};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1569, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_3 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1569, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (__pyx_t_6) {
      __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_6); __pyx_t_6 = NULL;
    }
    __Pyx_INCREF(__pyx_v_set);
    __Pyx_GIVEREF(__pyx_v_set);
    PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_5, __pyx_v_set);
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_5, __pyx_v_self);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1569, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_set, __pyx_t_1) < 0) __PYX_ERR(0, 1569, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1570
 *                 self.get = types.MethodType(get, self)
 *                 self.set = types.MethodType(set, self)
 *                 self.clear = types.MethodType(clear, self)             # <<<<<<<<<<<<<<
 *                 self.__len__ = types.MethodType(cache_len, self)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_types); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_MethodType); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_clear, __pyx_v_self};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1570, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_clear, __pyx_v_self};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1570, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1570, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_clear);
    __Pyx_GIVEREF(__pyx_v_clear);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_clear);
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_self);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1570, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_clear, __pyx_t_1) < 0) __PYX_ERR(0, 1570, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1571
 *                 self.set = types.MethodType(set, self)
 *                 self.clear = types.MethodType(clear, self)
 *                 self.__len__ = types.MethodType(cache_len, self)             # <<<<<<<<<<<<<<
 * 
 *     # argument cache for optimizing repeated calls when backtracking through recursive expressions
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_types); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1571, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_MethodType); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1571, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_cache_len, __pyx_v_self};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1571, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_cache_len, __pyx_v_self};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1571, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1571, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_cache_len);
    __Pyx_GIVEREF(__pyx_v_cache_len);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_5, __pyx_v_cache_len);
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_5, __pyx_v_self);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1571, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_len, __pyx_t_1) < 0) __PYX_ERR(0, 1571, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1546
 *     else:
 *         class _FifoCache(object):
 *             def __init__(self, size):             # <<<<<<<<<<<<<<
 *                 self.not_in_cache = not_in_cache = object()
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._FifoCache.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_get);
  __Pyx_XDECREF(__pyx_v_set);
  __Pyx_XDECREF(__pyx_v_clear);
  __Pyx_XDECREF(__pyx_v_cache_len);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1580
 *     # this method gets repeatedly called during backtracking with the same arguments -
 *     # we can cache these arguments and save ourselves the trouble of re-parsing the contained expression
 *     def _parseCache(self, instring, loc, doActions=True, callPreParse=True):             # <<<<<<<<<<<<<<
 *         HIT, MISS = 0, 1
 *         lookup = (self, instring, loc, callPreParse, doActions)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_41_parseCache(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_41_parseCache = {"_parseCache", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_41_parseCache, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_41_parseCache(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  PyObject *__pyx_v_doActions = 0;
  PyObject *__pyx_v_callPreParse = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_parseCache (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_doActions,&__pyx_n_s_callPreParse,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[3] = ((PyObject *)((PyObject *)Py_True));
    values[4] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_parseCache", 0, 3, 5, 1); __PYX_ERR(0, 1580, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_parseCache", 0, 3, 5, 2); __PYX_ERR(0, 1580, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_doActions);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_callPreParse);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_parseCache") < 0)) __PYX_ERR(0, 1580, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_doActions = values[3];
    __pyx_v_callPreParse = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_parseCache", 0, 3, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1580, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement._parseCache", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_40_parseCache(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions, __pyx_v_callPreParse);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_40_parseCache(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, PyObject *__pyx_v_doActions, PyObject *__pyx_v_callPreParse) {
  long __pyx_v_HIT;
  long __pyx_v_MISS;
  PyObject *__pyx_v_lookup = NULL;
  PyObject *__pyx_v_cache = NULL;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_v_pe = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  long __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_parseCache", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":1581
 *     # we can cache these arguments and save ourselves the trouble of re-parsing the contained expression
 *     def _parseCache(self, instring, loc, doActions=True, callPreParse=True):
 *         HIT, MISS = 0, 1             # <<<<<<<<<<<<<<
 *         lookup = (self, instring, loc, callPreParse, doActions)
 *         with ParserElement.packrat_cache_lock:
 */
  __pyx_t_1 = 0;
  __pyx_t_2 = 1;
  __pyx_v_HIT = __pyx_t_1;
  __pyx_v_MISS = __pyx_t_2;

  /* "Aspidites/_vendor/pyparsing_extension.py":1582
 *     def _parseCache(self, instring, loc, doActions=True, callPreParse=True):
 *         HIT, MISS = 0, 1
 *         lookup = (self, instring, loc, callPreParse, doActions)             # <<<<<<<<<<<<<<
 *         with ParserElement.packrat_cache_lock:
 *             cache = ParserElement.packrat_cache
 */
  __pyx_t_3 = PyTuple_New(5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_self);
  __Pyx_INCREF(__pyx_v_instring);
  __Pyx_GIVEREF(__pyx_v_instring);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_instring);
  __Pyx_INCREF(__pyx_v_loc);
  __Pyx_GIVEREF(__pyx_v_loc);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_loc);
  __Pyx_INCREF(__pyx_v_callPreParse);
  __Pyx_GIVEREF(__pyx_v_callPreParse);
  PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_v_callPreParse);
  __Pyx_INCREF(__pyx_v_doActions);
  __Pyx_GIVEREF(__pyx_v_doActions);
  PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_v_doActions);
  __pyx_v_lookup = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1583
 *         HIT, MISS = 0, 1
 *         lookup = (self, instring, loc, callPreParse, doActions)
 *         with ParserElement.packrat_cache_lock:             # <<<<<<<<<<<<<<
 *             cache = ParserElement.packrat_cache
 *             value = cache.get(lookup)
 */
  /*with:*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1583, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_packrat_cache_lock); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1583, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_5 = __Pyx_PyObject_LookupSpecial(__pyx_t_4, __pyx_n_s_exit); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1583, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_LookupSpecial(__pyx_t_4, __pyx_n_s_enter); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1583, __pyx_L3_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_3 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1583, __pyx_L3_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    /*try:*/ {
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_10);
        /*try:*/ {

          /* "Aspidites/_vendor/pyparsing_extension.py":1584
 *         lookup = (self, instring, loc, callPreParse, doActions)
 *         with ParserElement.packrat_cache_lock:
 *             cache = ParserElement.packrat_cache             # <<<<<<<<<<<<<<
 *             value = cache.get(lookup)
 *             if value is cache.not_in_cache:
 */
          __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1584, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_packrat_cache); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1584, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_v_cache = __pyx_t_3;
          __pyx_t_3 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":1585
 *         with ParserElement.packrat_cache_lock:
 *             cache = ParserElement.packrat_cache
 *             value = cache.get(lookup)             # <<<<<<<<<<<<<<
 *             if value is cache.not_in_cache:
 *                 ParserElement.packrat_cache_stats[MISS] += 1
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_cache, __pyx_n_s_get); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1585, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_6 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_6)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_6);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_v_lookup) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_lookup);
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1585, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_v_value = __pyx_t_3;
          __pyx_t_3 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":1586
 *             cache = ParserElement.packrat_cache
 *             value = cache.get(lookup)
 *             if value is cache.not_in_cache:             # <<<<<<<<<<<<<<
 *                 ParserElement.packrat_cache_stats[MISS] += 1
 *                 try:
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_cache, __pyx_n_s_not_in_cache); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1586, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_11 = (__pyx_v_value == __pyx_t_3);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_12 = (__pyx_t_11 != 0);
          if (__pyx_t_12) {

            /* "Aspidites/_vendor/pyparsing_extension.py":1587
 *             value = cache.get(lookup)
 *             if value is cache.not_in_cache:
 *                 ParserElement.packrat_cache_stats[MISS] += 1             # <<<<<<<<<<<<<<
 *                 try:
 *                     value = self._parseNoCache(instring, loc, doActions, callPreParse)
 */
            __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1587, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_packrat_cache_stats); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1587, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_2 = __pyx_v_MISS;
            __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_4, __pyx_t_2, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1587, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_6 = __Pyx_PyInt_AddObjC(__pyx_t_3, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1587, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (unlikely(__Pyx_SetItemInt(__pyx_t_4, __pyx_t_2, __pyx_t_6, long, 1, __Pyx_PyInt_From_long, 0, 1, 1) < 0)) __PYX_ERR(0, 1587, __pyx_L7_error)
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

            /* "Aspidites/_vendor/pyparsing_extension.py":1588
 *             if value is cache.not_in_cache:
 *                 ParserElement.packrat_cache_stats[MISS] += 1
 *                 try:             # <<<<<<<<<<<<<<
 *                     value = self._parseNoCache(instring, loc, doActions, callPreParse)
 *                 except ParseBaseException as pe:
 */
            {
              __Pyx_PyThreadState_declare
              __Pyx_PyThreadState_assign
              __Pyx_ExceptionSave(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
              __Pyx_XGOTREF(__pyx_t_13);
              __Pyx_XGOTREF(__pyx_t_14);
              __Pyx_XGOTREF(__pyx_t_15);
              /*try:*/ {

                /* "Aspidites/_vendor/pyparsing_extension.py":1589
 *                 ParserElement.packrat_cache_stats[MISS] += 1
 *                 try:
 *                     value = self._parseNoCache(instring, loc, doActions, callPreParse)             # <<<<<<<<<<<<<<
 *                 except ParseBaseException as pe:
 *                     # cache a copy of the exception, without the traceback
 */
                __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_parseNoCache); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1589, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_6);
                __pyx_t_3 = NULL;
                __pyx_t_16 = 0;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
                  __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
                  if (likely(__pyx_t_3)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                    __Pyx_INCREF(__pyx_t_3);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_6, function);
                    __pyx_t_16 = 1;
                  }
                }
                #if CYTHON_FAST_PYCALL
                if (PyFunction_Check(__pyx_t_6)) {
                  PyObject *__pyx_temp[5] = {__pyx_t_3, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions, __pyx_v_callPreParse};
                  __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_16, 4+__pyx_t_16); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1589, __pyx_L14_error)
                  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                  __Pyx_GOTREF(__pyx_t_4);
                } else
                #endif
                #if CYTHON_FAST_PYCCALL
                if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
                  PyObject *__pyx_temp[5] = {__pyx_t_3, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions, __pyx_v_callPreParse};
                  __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_16, 4+__pyx_t_16); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1589, __pyx_L14_error)
                  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                  __Pyx_GOTREF(__pyx_t_4);
                } else
                #endif
                {
                  __pyx_t_7 = PyTuple_New(4+__pyx_t_16); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1589, __pyx_L14_error)
                  __Pyx_GOTREF(__pyx_t_7);
                  if (__pyx_t_3) {
                    __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
                  }
                  __Pyx_INCREF(__pyx_v_instring);
                  __Pyx_GIVEREF(__pyx_v_instring);
                  PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_16, __pyx_v_instring);
                  __Pyx_INCREF(__pyx_v_loc);
                  __Pyx_GIVEREF(__pyx_v_loc);
                  PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_16, __pyx_v_loc);
                  __Pyx_INCREF(__pyx_v_doActions);
                  __Pyx_GIVEREF(__pyx_v_doActions);
                  PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_16, __pyx_v_doActions);
                  __Pyx_INCREF(__pyx_v_callPreParse);
                  __Pyx_GIVEREF(__pyx_v_callPreParse);
                  PyTuple_SET_ITEM(__pyx_t_7, 3+__pyx_t_16, __pyx_v_callPreParse);
                  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_7, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1589, __pyx_L14_error)
                  __Pyx_GOTREF(__pyx_t_4);
                  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
                }
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_4);
                __pyx_t_4 = 0;

                /* "Aspidites/_vendor/pyparsing_extension.py":1588
 *             if value is cache.not_in_cache:
 *                 ParserElement.packrat_cache_stats[MISS] += 1
 *                 try:             # <<<<<<<<<<<<<<
 *                     value = self._parseNoCache(instring, loc, doActions, callPreParse)
 *                 except ParseBaseException as pe:
 */
              }

              /* "Aspidites/_vendor/pyparsing_extension.py":1595
 *                     raise
 *                 else:
 *                     cache.set(lookup, (value[0], value[1].copy()))             # <<<<<<<<<<<<<<
 *                     return value
 *             else:
 */
              /*else:*/ {
                __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_cache, __pyx_n_s_set); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1595, __pyx_L16_except_error)
                __Pyx_GOTREF(__pyx_t_6);
                __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_value, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1595, __pyx_L16_except_error)
                __Pyx_GOTREF(__pyx_t_7);
                __pyx_t_17 = __Pyx_GetItemInt(__pyx_v_value, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1595, __pyx_L16_except_error)
                __Pyx_GOTREF(__pyx_t_17);
                __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_t_17, __pyx_n_s_copy); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1595, __pyx_L16_except_error)
                __Pyx_GOTREF(__pyx_t_18);
                __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                __pyx_t_17 = NULL;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_18))) {
                  __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_18);
                  if (likely(__pyx_t_17)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_18);
                    __Pyx_INCREF(__pyx_t_17);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_18, function);
                  }
                }
                __pyx_t_3 = (__pyx_t_17) ? __Pyx_PyObject_CallOneArg(__pyx_t_18, __pyx_t_17) : __Pyx_PyObject_CallNoArg(__pyx_t_18);
                __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
                if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1595, __pyx_L16_except_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
                __pyx_t_18 = PyTuple_New(2); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1595, __pyx_L16_except_error)
                __Pyx_GOTREF(__pyx_t_18);
                __Pyx_GIVEREF(__pyx_t_7);
                PyTuple_SET_ITEM(__pyx_t_18, 0, __pyx_t_7);
                __Pyx_GIVEREF(__pyx_t_3);
                PyTuple_SET_ITEM(__pyx_t_18, 1, __pyx_t_3);
                __pyx_t_7 = 0;
                __pyx_t_3 = 0;
                __pyx_t_3 = NULL;
                __pyx_t_16 = 0;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
                  __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
                  if (likely(__pyx_t_3)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                    __Pyx_INCREF(__pyx_t_3);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_6, function);
                    __pyx_t_16 = 1;
                  }
                }
                #if CYTHON_FAST_PYCALL
                if (PyFunction_Check(__pyx_t_6)) {
                  PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_lookup, __pyx_t_18};
                  __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_16, 2+__pyx_t_16); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1595, __pyx_L16_except_error)
                  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                  __Pyx_GOTREF(__pyx_t_4);
                  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
                } else
                #endif
                #if CYTHON_FAST_PYCCALL
                if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
                  PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_lookup, __pyx_t_18};
                  __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_16, 2+__pyx_t_16); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1595, __pyx_L16_except_error)
                  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                  __Pyx_GOTREF(__pyx_t_4);
                  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
                } else
                #endif
                {
                  __pyx_t_7 = PyTuple_New(2+__pyx_t_16); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1595, __pyx_L16_except_error)
                  __Pyx_GOTREF(__pyx_t_7);
                  if (__pyx_t_3) {
                    __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
                  }
                  __Pyx_INCREF(__pyx_v_lookup);
                  __Pyx_GIVEREF(__pyx_v_lookup);
                  PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_16, __pyx_v_lookup);
                  __Pyx_GIVEREF(__pyx_t_18);
                  PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_16, __pyx_t_18);
                  __pyx_t_18 = 0;
                  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_7, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1595, __pyx_L16_except_error)
                  __Pyx_GOTREF(__pyx_t_4);
                  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
                }
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

                /* "Aspidites/_vendor/pyparsing_extension.py":1596
 *                 else:
 *                     cache.set(lookup, (value[0], value[1].copy()))
 *                     return value             # <<<<<<<<<<<<<<
 *             else:
 *                 ParserElement.packrat_cache_stats[HIT] += 1
 */
                __Pyx_XDECREF(__pyx_r);
                __Pyx_INCREF(__pyx_v_value);
                __pyx_r = __pyx_v_value;
                goto __pyx_L17_except_return;
              }
              __pyx_L14_error:;
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

              /* "Aspidites/_vendor/pyparsing_extension.py":1590
 *                 try:
 *                     value = self._parseNoCache(instring, loc, doActions, callPreParse)
 *                 except ParseBaseException as pe:             # <<<<<<<<<<<<<<
 *                     # cache a copy of the exception, without the traceback
 *                     cache.set(lookup, pe.__class__(*pe.args))
 */
              __Pyx_ErrFetch(&__pyx_t_4, &__pyx_t_6, &__pyx_t_7);
              __Pyx_GetModuleGlobalName(__pyx_t_18, __pyx_n_s_ParseBaseException); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1590, __pyx_L16_except_error)
              __Pyx_GOTREF(__pyx_t_18);
              __pyx_t_16 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_4, __pyx_t_18);
              __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
              __Pyx_ErrRestore(__pyx_t_4, __pyx_t_6, __pyx_t_7);
              __pyx_t_4 = 0; __pyx_t_6 = 0; __pyx_t_7 = 0;
              if (__pyx_t_16) {
                __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement._parseCache", __pyx_clineno, __pyx_lineno, __pyx_filename);
                if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_6, &__pyx_t_4) < 0) __PYX_ERR(0, 1590, __pyx_L16_except_error)
                __Pyx_GOTREF(__pyx_t_7);
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_INCREF(__pyx_t_6);
                __pyx_v_pe = __pyx_t_6;

                /* "Aspidites/_vendor/pyparsing_extension.py":1592
 *                 except ParseBaseException as pe:
 *                     # cache a copy of the exception, without the traceback
 *                     cache.set(lookup, pe.__class__(*pe.args))             # <<<<<<<<<<<<<<
 *                     raise
 *                 else:
 */
                __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_cache, __pyx_n_s_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1592, __pyx_L16_except_error)
                __Pyx_GOTREF(__pyx_t_3);
                __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_v_pe, __pyx_n_s_class); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1592, __pyx_L16_except_error)
                __Pyx_GOTREF(__pyx_t_17);
                __pyx_t_19 = __Pyx_PyObject_GetAttrStr(__pyx_v_pe, __pyx_n_s_args); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1592, __pyx_L16_except_error)
                __Pyx_GOTREF(__pyx_t_19);
                __pyx_t_20 = __Pyx_PySequence_Tuple(__pyx_t_19); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1592, __pyx_L16_except_error)
                __Pyx_GOTREF(__pyx_t_20);
                __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
                __pyx_t_19 = __Pyx_PyObject_Call(__pyx_t_17, __pyx_t_20, NULL); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1592, __pyx_L16_except_error)
                __Pyx_GOTREF(__pyx_t_19);
                __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
                __pyx_t_20 = NULL;
                __pyx_t_16 = 0;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
                  __pyx_t_20 = PyMethod_GET_SELF(__pyx_t_3);
                  if (likely(__pyx_t_20)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                    __Pyx_INCREF(__pyx_t_20);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_3, function);
                    __pyx_t_16 = 1;
                  }
                }
                #if CYTHON_FAST_PYCALL
                if (PyFunction_Check(__pyx_t_3)) {
                  PyObject *__pyx_temp[3] = {__pyx_t_20, __pyx_v_lookup, __pyx_t_19};
                  __pyx_t_18 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_16, 2+__pyx_t_16); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1592, __pyx_L16_except_error)
                  __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
                  __Pyx_GOTREF(__pyx_t_18);
                  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
                } else
                #endif
                #if CYTHON_FAST_PYCCALL
                if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
                  PyObject *__pyx_temp[3] = {__pyx_t_20, __pyx_v_lookup, __pyx_t_19};
                  __pyx_t_18 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_16, 2+__pyx_t_16); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1592, __pyx_L16_except_error)
                  __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
                  __Pyx_GOTREF(__pyx_t_18);
                  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
                } else
                #endif
                {
                  __pyx_t_17 = PyTuple_New(2+__pyx_t_16); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1592, __pyx_L16_except_error)
                  __Pyx_GOTREF(__pyx_t_17);
                  if (__pyx_t_20) {
                    __Pyx_GIVEREF(__pyx_t_20); PyTuple_SET_ITEM(__pyx_t_17, 0, __pyx_t_20); __pyx_t_20 = NULL;
                  }
                  __Pyx_INCREF(__pyx_v_lookup);
                  __Pyx_GIVEREF(__pyx_v_lookup);
                  PyTuple_SET_ITEM(__pyx_t_17, 0+__pyx_t_16, __pyx_v_lookup);
                  __Pyx_GIVEREF(__pyx_t_19);
                  PyTuple_SET_ITEM(__pyx_t_17, 1+__pyx_t_16, __pyx_t_19);
                  __pyx_t_19 = 0;
                  __pyx_t_18 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_17, NULL); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1592, __pyx_L16_except_error)
                  __Pyx_GOTREF(__pyx_t_18);
                  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                }
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

                /* "Aspidites/_vendor/pyparsing_extension.py":1593
 *                     # cache a copy of the exception, without the traceback
 *                     cache.set(lookup, pe.__class__(*pe.args))
 *                     raise             # <<<<<<<<<<<<<<
 *                 else:
 *                     cache.set(lookup, (value[0], value[1].copy()))
 */
                __Pyx_GIVEREF(__pyx_t_7);
                __Pyx_GIVEREF(__pyx_t_6);
                __Pyx_XGIVEREF(__pyx_t_4);
                __Pyx_ErrRestoreWithState(__pyx_t_7, __pyx_t_6, __pyx_t_4);
                __pyx_t_7 = 0; __pyx_t_6 = 0; __pyx_t_4 = 0; 
                __PYX_ERR(0, 1593, __pyx_L16_except_error)
              }
              goto __pyx_L16_except_error;
              __pyx_L16_except_error:;

              /* "Aspidites/_vendor/pyparsing_extension.py":1588
 *             if value is cache.not_in_cache:
 *                 ParserElement.packrat_cache_stats[MISS] += 1
 *                 try:             # <<<<<<<<<<<<<<
 *                     value = self._parseNoCache(instring, loc, doActions, callPreParse)
 *                 except ParseBaseException as pe:
 */
              __Pyx_XGIVEREF(__pyx_t_13);
              __Pyx_XGIVEREF(__pyx_t_14);
              __Pyx_XGIVEREF(__pyx_t_15);
              __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_14, __pyx_t_15);
              goto __pyx_L7_error;
              __pyx_L17_except_return:;
              __Pyx_XGIVEREF(__pyx_t_13);
              __Pyx_XGIVEREF(__pyx_t_14);
              __Pyx_XGIVEREF(__pyx_t_15);
              __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_14, __pyx_t_15);
              goto __pyx_L11_try_return;
            }

            /* "Aspidites/_vendor/pyparsing_extension.py":1586
 *             cache = ParserElement.packrat_cache
 *             value = cache.get(lookup)
 *             if value is cache.not_in_cache:             # <<<<<<<<<<<<<<
 *                 ParserElement.packrat_cache_stats[MISS] += 1
 *                 try:
 */
          }

          /* "Aspidites/_vendor/pyparsing_extension.py":1598
 *                     return value
 *             else:
 *                 ParserElement.packrat_cache_stats[HIT] += 1             # <<<<<<<<<<<<<<
 *                 if isinstance(value, Exception):
 *                     raise value
 */
          /*else*/ {
            __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1598, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_packrat_cache_stats); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1598, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_2 = __pyx_v_HIT;
            __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_6, __pyx_t_2, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1598, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_7 = __Pyx_PyInt_AddObjC(__pyx_t_4, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1598, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            if (unlikely(__Pyx_SetItemInt(__pyx_t_6, __pyx_t_2, __pyx_t_7, long, 1, __Pyx_PyInt_From_long, 0, 1, 1) < 0)) __PYX_ERR(0, 1598, __pyx_L7_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

            /* "Aspidites/_vendor/pyparsing_extension.py":1599
 *             else:
 *                 ParserElement.packrat_cache_stats[HIT] += 1
 *                 if isinstance(value, Exception):             # <<<<<<<<<<<<<<
 *                     raise value
 *                 return value[0], value[1].copy()
 */
            __pyx_t_12 = __Pyx_PyException_Check(__pyx_v_value); 
            __pyx_t_11 = (__pyx_t_12 != 0);
            if (unlikely(__pyx_t_11)) {

              /* "Aspidites/_vendor/pyparsing_extension.py":1600
 *                 ParserElement.packrat_cache_stats[HIT] += 1
 *                 if isinstance(value, Exception):
 *                     raise value             # <<<<<<<<<<<<<<
 *                 return value[0], value[1].copy()
 * 
 */
              __Pyx_Raise(__pyx_v_value, 0, 0, 0);
              __PYX_ERR(0, 1600, __pyx_L7_error)

              /* "Aspidites/_vendor/pyparsing_extension.py":1599
 *             else:
 *                 ParserElement.packrat_cache_stats[HIT] += 1
 *                 if isinstance(value, Exception):             # <<<<<<<<<<<<<<
 *                     raise value
 *                 return value[0], value[1].copy()
 */
            }

            /* "Aspidites/_vendor/pyparsing_extension.py":1601
 *                 if isinstance(value, Exception):
 *                     raise value
 *                 return value[0], value[1].copy()             # <<<<<<<<<<<<<<
 * 
 *     _parse = _parseNoCache
 */
            __Pyx_XDECREF(__pyx_r);
            __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_value, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1601, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_6);
            __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_value, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1601, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_copy); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1601, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_18);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_18))) {
              __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_18);
              if (likely(__pyx_t_4)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_18);
                __Pyx_INCREF(__pyx_t_4);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_18, function);
              }
            }
            __pyx_t_7 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_18, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_18);
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1601, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
            __pyx_t_18 = PyTuple_New(2); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1601, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_18);
            __Pyx_GIVEREF(__pyx_t_6);
            PyTuple_SET_ITEM(__pyx_t_18, 0, __pyx_t_6);
            __Pyx_GIVEREF(__pyx_t_7);
            PyTuple_SET_ITEM(__pyx_t_18, 1, __pyx_t_7);
            __pyx_t_6 = 0;
            __pyx_t_7 = 0;
            __pyx_r = __pyx_t_18;
            __pyx_t_18 = 0;
            goto __pyx_L11_try_return;
          }

          /* "Aspidites/_vendor/pyparsing_extension.py":1583
 *         HIT, MISS = 0, 1
 *         lookup = (self, instring, loc, callPreParse, doActions)
 *         with ParserElement.packrat_cache_lock:             # <<<<<<<<<<<<<<
 *             cache = ParserElement.packrat_cache
 *             value = cache.get(lookup)
 */
        }
        __pyx_L7_error:;
        __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
        __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
        __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
        __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        /*except:*/ {
          __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement._parseCache", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_18, &__pyx_t_7, &__pyx_t_6) < 0) __PYX_ERR(0, 1583, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_18);
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_4 = PyTuple_Pack(3, __pyx_t_18, __pyx_t_7, __pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1583, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_4, NULL);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1583, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_15);
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_15);
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          if (__pyx_t_11 < 0) __PYX_ERR(0, 1583, __pyx_L9_except_error)
          __pyx_t_12 = ((!(__pyx_t_11 != 0)) != 0);
          if (__pyx_t_12) {
            __Pyx_GIVEREF(__pyx_t_18);
            __Pyx_GIVEREF(__pyx_t_7);
            __Pyx_XGIVEREF(__pyx_t_6);
            __Pyx_ErrRestoreWithState(__pyx_t_18, __pyx_t_7, __pyx_t_6);
            __pyx_t_18 = 0; __pyx_t_7 = 0; __pyx_t_6 = 0; 
            __PYX_ERR(0, 1583, __pyx_L9_except_error)
          }
          __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          goto __pyx_L8_exception_handled;
        }
        __pyx_L9_except_error:;
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
        goto __pyx_L1_error;
        __pyx_L11_try_return:;
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
        goto __pyx_L4_return;
        __pyx_L8_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      }
    }
    /*finally:*/ {
      /*normal exit:*/{
        if (__pyx_t_5) {
          __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_tuple__32, NULL);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1583, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        }
        goto __pyx_L6;
      }
      __pyx_L4_return: {
        __pyx_t_10 = __pyx_r;
        __pyx_r = 0;
        if (__pyx_t_5) {
          __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_tuple__32, NULL);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1583, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        }
        __pyx_r = __pyx_t_10;
        __pyx_t_10 = 0;
        goto __pyx_L0;
      }
      __pyx_L6:;
    }
    goto __pyx_L26;
    __pyx_L3_error:;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L1_error;
    __pyx_L26:;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1580
 *     # this method gets repeatedly called during backtracking with the same arguments -
 *     # we can cache these arguments and save ourselves the trouble of re-parsing the contained expression
 *     def _parseCache(self, instring, loc, doActions=True, callPreParse=True):             # <<<<<<<<<<<<<<
 *         HIT, MISS = 0, 1
 *         lookup = (self, instring, loc, callPreParse, doActions)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement._parseCache", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_lookup);
  __Pyx_XDECREF(__pyx_v_cache);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XDECREF(__pyx_v_pe);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1606
 * 
 *     @staticmethod
 *     def resetCache():             # <<<<<<<<<<<<<<
 *         ParserElement.packrat_cache.clear()
 *         ParserElement.packrat_cache_stats[:] = [0] * len(ParserElement.packrat_cache_stats)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_43resetCache(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_43resetCache = {"resetCache", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_43resetCache, METH_NOARGS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_43resetCache(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("resetCache (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_42resetCache(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_42resetCache(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("resetCache", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":1607
 *     @staticmethod
 *     def resetCache():
 *         ParserElement.packrat_cache.clear()             # <<<<<<<<<<<<<<
 *         ParserElement.packrat_cache_stats[:] = [0] * len(ParserElement.packrat_cache_stats)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1607, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_packrat_cache); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1607, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_clear); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1607, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1607, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1608
 *     def resetCache():
 *         ParserElement.packrat_cache.clear()
 *         ParserElement.packrat_cache_stats[:] = [0] * len(ParserElement.packrat_cache_stats)             # <<<<<<<<<<<<<<
 * 
 *     _packratEnabled = False
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1608, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_packrat_cache_stats); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1608, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = PyObject_Length(__pyx_t_2); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1608, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyList_New(1 * ((__pyx_t_4<0) ? 0:__pyx_t_4)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1608, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  { Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < __pyx_t_4; __pyx_temp++) {
      __Pyx_INCREF(__pyx_int_0);
      __Pyx_GIVEREF(__pyx_int_0);
      PyList_SET_ITEM(__pyx_t_2, __pyx_temp, __pyx_int_0);
    }
  }
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1608, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_packrat_cache_stats); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1608, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_PyObject_SetSlice(__pyx_t_3, __pyx_t_2, 0, 0, NULL, NULL, &__pyx_slice__11, 0, 0, 1) < 0) __PYX_ERR(0, 1608, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1606
 * 
 *     @staticmethod
 *     def resetCache():             # <<<<<<<<<<<<<<
 *         ParserElement.packrat_cache.clear()
 *         ParserElement.packrat_cache_stats[:] = [0] * len(ParserElement.packrat_cache_stats)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.resetCache", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1613
 * 
 *     @staticmethod
 *     def enablePackrat(cache_size_limit=128):             # <<<<<<<<<<<<<<
 *         """Enables "packrat" parsing, which adds memoizing to the parsing logic.
 *            Repeated parse attempts at the same string location (which happens
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_45enablePackrat(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_44enablePackrat[] = "Enables \"packrat\" parsing, which adds memoizing to the parsing logic.\n           Repeated parse attempts at the same string location (which happens\n           often in many complex grammars) can immediately return a cached value,\n           instead of re-executing parsing/validating code.  Memoizing is done of\n           both valid results and parsing exceptions.\n\n           Parameters:\n\n           - cache_size_limit - (default= ``128``) - if an integer value is provided\n             will limit the size of the packrat cache; if None is passed, then\n             the cache size will be unbounded; if 0 is passed, the cache will\n             be effectively disabled.\n\n           This speedup may break existing programs that use parse actions that\n           have side-effects.  For this reason, packrat parsing is disabled when\n           you first import pyparsing.  To activate the packrat feature, your\n           program must call the class method :class:`ParserElement.enablePackrat`.\n           For best results, call ``enablePackrat()`` immediately after\n           importing pyparsing.\n\n           Example::\n\n               import pyparsing\n               pyparsing.ParserElement.enablePackrat()\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_45enablePackrat = {"enablePackrat", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_45enablePackrat, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_44enablePackrat};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_45enablePackrat(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_cache_size_limit = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enablePackrat (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_cache_size_limit,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject *)((PyObject *)__pyx_int_128));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cache_size_limit);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enablePackrat") < 0)) __PYX_ERR(0, 1613, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_cache_size_limit = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enablePackrat", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1613, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.enablePackrat", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_44enablePackrat(__pyx_self, __pyx_v_cache_size_limit);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_44enablePackrat(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_cache_size_limit) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("enablePackrat", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":1639
 *                pyparsing.ParserElement.enablePackrat()
 *         """
 *         if not ParserElement._packratEnabled:             # <<<<<<<<<<<<<<
 *             ParserElement._packratEnabled = True
 *             if cache_size_limit is None:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_packratEnabled); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1639, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = ((!__pyx_t_3) != 0);
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":1640
 *         """
 *         if not ParserElement._packratEnabled:
 *             ParserElement._packratEnabled = True             # <<<<<<<<<<<<<<
 *             if cache_size_limit is None:
 *                 ParserElement.packrat_cache = ParserElement._UnboundedCache()
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1640, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_2, __pyx_n_s_packratEnabled, Py_True) < 0) __PYX_ERR(0, 1640, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1641
 *         if not ParserElement._packratEnabled:
 *             ParserElement._packratEnabled = True
 *             if cache_size_limit is None:             # <<<<<<<<<<<<<<
 *                 ParserElement.packrat_cache = ParserElement._UnboundedCache()
 *             else:
 */
    __pyx_t_4 = (__pyx_v_cache_size_limit == Py_None);
    __pyx_t_3 = (__pyx_t_4 != 0);
    if (__pyx_t_3) {

      /* "Aspidites/_vendor/pyparsing_extension.py":1642
 *             ParserElement._packratEnabled = True
 *             if cache_size_limit is None:
 *                 ParserElement.packrat_cache = ParserElement._UnboundedCache()             # <<<<<<<<<<<<<<
 *             else:
 *                 ParserElement.packrat_cache = ParserElement._FifoCache(cache_size_limit)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1642, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_UnboundedCache); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1642, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1642, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1642, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_packrat_cache, __pyx_t_2) < 0) __PYX_ERR(0, 1642, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":1641
 *         if not ParserElement._packratEnabled:
 *             ParserElement._packratEnabled = True
 *             if cache_size_limit is None:             # <<<<<<<<<<<<<<
 *                 ParserElement.packrat_cache = ParserElement._UnboundedCache()
 *             else:
 */
      goto __pyx_L4;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":1644
 *                 ParserElement.packrat_cache = ParserElement._UnboundedCache()
 *             else:
 *                 ParserElement.packrat_cache = ParserElement._FifoCache(cache_size_limit)             # <<<<<<<<<<<<<<
 *             ParserElement._parse = ParserElement._parseCache
 * 
 */
    /*else*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1644, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_FifoCache); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1644, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_5 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_v_cache_size_limit) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_cache_size_limit);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1644, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1644, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (__Pyx_PyObject_SetAttrStr(__pyx_t_1, __pyx_n_s_packrat_cache, __pyx_t_5) < 0) __PYX_ERR(0, 1644, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L4:;

    /* "Aspidites/_vendor/pyparsing_extension.py":1645
 *             else:
 *                 ParserElement.packrat_cache = ParserElement._FifoCache(cache_size_limit)
 *             ParserElement._parse = ParserElement._parseCache             # <<<<<<<<<<<<<<
 * 
 *     def parseString(self, instring, parseAll=False):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1645, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_parseCache); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1645, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1645, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_1, __pyx_n_s_parse, __pyx_t_5) < 0) __PYX_ERR(0, 1645, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1639
 *                pyparsing.ParserElement.enablePackrat()
 *         """
 *         if not ParserElement._packratEnabled:             # <<<<<<<<<<<<<<
 *             ParserElement._packratEnabled = True
 *             if cache_size_limit is None:
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1613
 * 
 *     @staticmethod
 *     def enablePackrat(cache_size_limit=128):             # <<<<<<<<<<<<<<
 *         """Enables "packrat" parsing, which adds memoizing to the parsing logic.
 *            Repeated parse attempts at the same string location (which happens
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.enablePackrat", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1647
 *             ParserElement._parse = ParserElement._parseCache
 * 
 *     def parseString(self, instring, parseAll=False):             # <<<<<<<<<<<<<<
 *         """
 *         Execute the parse expression with the given string.
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_47parseString(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_46parseString[] = "\n        Execute the parse expression with the given string.\n        This is the main interface to the client code, once the complete\n        expression has been built.\n\n        Returns the parsed data as a :class:`ParseResults` object, which may be\n        accessed as a list, or as a dict or object with attributes if the given parser\n        includes results names.\n\n        If you want the grammar to require that the entire input string be\n        successfully parsed, then set ``parseAll`` to True (equivalent to ending\n        the grammar with ``StringEnd()``).\n\n        Note: ``parseString`` implicitly calls ``expandtabs()`` on the input string,\n        in order to report proper column numbers in parse actions.\n        If the input string contains tabs and\n        the grammar uses parse actions that use the ``loc`` argument to index into the\n        string being parsed, you can ensure you have a consistent view of the input\n        string by:\n\n        - calling ``parseWithTabs`` on your grammar before calling ``parseString``\n          (see :class:`parseWithTabs`)\n        - define your parse action using the full ``(s, loc, toks)`` signature, and\n          reference the input string using the parse action's ``s`` argument\n        - explictly expand the tabs in your input string before calling\n          ``parseString``\n\n        Example::\n\n            Word('a').parseString('aaaaabaaa')  # -> ['aaaaa']\n            Word('a').parseString('aaaaabaaa', parseAll=True)  # -> Exception: Expected end of text\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_47parseString = {"parseString", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_47parseString, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_46parseString};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_47parseString(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_parseAll = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parseString (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_parseAll,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_False));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseString", 0, 2, 3, 1); __PYX_ERR(0, 1647, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_parseAll);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "parseString") < 0)) __PYX_ERR(0, 1647, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_parseAll = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("parseString", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1647, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.parseString", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_46parseString(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_parseAll);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_46parseString(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_parseAll) {
  PyObject *__pyx_v_Empty = NULL;
  PyObject *__pyx_v_StringEnd = NULL;
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_v_loc = NULL;
  PyObject *__pyx_v_tokens = NULL;
  PyObject *__pyx_v_se = NULL;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  PyObject *(*__pyx_t_13)(PyObject *);
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parseString", 0);
  __Pyx_INCREF(__pyx_v_instring);

  /* "Aspidites/_vendor/pyparsing_extension.py":1680
 *             Word('a').parseString('aaaaabaaa', parseAll=True)  # -> Exception: Expected end of text
 *         """
 *         from .pyparsing import Empty, StringEnd             # <<<<<<<<<<<<<<
 *         ParserElement.resetCache()
 *         if not self.streamlined:
 */
  __pyx_t_1 = PyList_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1680, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_Empty);
  __Pyx_GIVEREF(__pyx_n_s_Empty);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_Empty);
  __Pyx_INCREF(__pyx_n_s_StringEnd);
  __Pyx_GIVEREF(__pyx_n_s_StringEnd);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_StringEnd);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_pyparsing, __pyx_t_1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1680, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_Empty); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1680, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_Empty = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_StringEnd); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1680, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_StringEnd = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1681
 *         """
 *         from .pyparsing import Empty, StringEnd
 *         ParserElement.resetCache()             # <<<<<<<<<<<<<<
 *         if not self.streamlined:
 *             self.streamline()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_resetCache); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1682
 *         from .pyparsing import Empty, StringEnd
 *         ParserElement.resetCache()
 *         if not self.streamlined:             # <<<<<<<<<<<<<<
 *             self.streamline()
 *             # ~ self.saveAsList = True
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_streamlined); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1682, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = ((!__pyx_t_4) != 0);
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/pyparsing_extension.py":1683
 *         ParserElement.resetCache()
 *         if not self.streamlined:
 *             self.streamline()             # <<<<<<<<<<<<<<
 *             # ~ self.saveAsList = True
 *         for e in self.ignoreExprs:
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_streamline); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1683, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1683, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1682
 *         from .pyparsing import Empty, StringEnd
 *         ParserElement.resetCache()
 *         if not self.streamlined:             # <<<<<<<<<<<<<<
 *             self.streamline()
 *             # ~ self.saveAsList = True
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1685
 *             self.streamline()
 *             # ~ self.saveAsList = True
 *         for e in self.ignoreExprs:             # <<<<<<<<<<<<<<
 *             e.streamline()
 *         if not self.keepTabs:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ignoreExprs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1685, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1685, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1685, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 1685, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1685, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 1685, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1685, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_7(__pyx_t_3);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1685, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1686
 *             # ~ self.saveAsList = True
 *         for e in self.ignoreExprs:
 *             e.streamline()             # <<<<<<<<<<<<<<
 *         if not self.keepTabs:
 *             instring = instring.expandtabs()
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_streamline); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1686, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1686, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1685
 *             self.streamline()
 *             # ~ self.saveAsList = True
 *         for e in self.ignoreExprs:             # <<<<<<<<<<<<<<
 *             e.streamline()
 *         if not self.keepTabs:
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1687
 *         for e in self.ignoreExprs:
 *             e.streamline()
 *         if not self.keepTabs:             # <<<<<<<<<<<<<<
 *             instring = instring.expandtabs()
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_keepTabs); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1687, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1687, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = ((!__pyx_t_5) != 0);
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":1688
 *             e.streamline()
 *         if not self.keepTabs:
 *             instring = instring.expandtabs()             # <<<<<<<<<<<<<<
 *         try:
 *             loc, tokens = self._parse(instring, 0)
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_instring, __pyx_n_s_expandtabs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1688, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1688, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_instring, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1687
 *         for e in self.ignoreExprs:
 *             e.streamline()
 *         if not self.keepTabs:             # <<<<<<<<<<<<<<
 *             instring = instring.expandtabs()
 *         try:
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1689
 *         if not self.keepTabs:
 *             instring = instring.expandtabs()
 *         try:             # <<<<<<<<<<<<<<
 *             loc, tokens = self._parse(instring, 0)
 *             if parseAll:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
    __Pyx_XGOTREF(__pyx_t_9);
    __Pyx_XGOTREF(__pyx_t_10);
    __Pyx_XGOTREF(__pyx_t_11);
    /*try:*/ {

      /* "Aspidites/_vendor/pyparsing_extension.py":1690
 *             instring = instring.expandtabs()
 *         try:
 *             loc, tokens = self._parse(instring, 0)             # <<<<<<<<<<<<<<
 *             if parseAll:
 *                 loc = self.preParse(instring, loc)
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_parse); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1690, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = NULL;
      __pyx_t_12 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_12 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_2)) {
        PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_instring, __pyx_int_0};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1690, __pyx_L7_error)
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
        PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_instring, __pyx_int_0};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1690, __pyx_L7_error)
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_8 = PyTuple_New(2+__pyx_t_12); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1690, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_8);
        if (__pyx_t_1) {
          __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_1); __pyx_t_1 = NULL;
        }
        __Pyx_INCREF(__pyx_v_instring);
        __Pyx_GIVEREF(__pyx_v_instring);
        PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_12, __pyx_v_instring);
        __Pyx_INCREF(__pyx_int_0);
        __Pyx_GIVEREF(__pyx_int_0);
        PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_12, __pyx_int_0);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_8, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1690, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
        PyObject* sequence = __pyx_t_3;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 1690, __pyx_L7_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_8 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_8 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_8);
        #else
        __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1690, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_8 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1690, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_8);
        #endif
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_1 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1690, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_13 = Py_TYPE(__pyx_t_1)->tp_iternext;
        index = 0; __pyx_t_2 = __pyx_t_13(__pyx_t_1); if (unlikely(!__pyx_t_2)) goto __pyx_L13_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_2);
        index = 1; __pyx_t_8 = __pyx_t_13(__pyx_t_1); if (unlikely(!__pyx_t_8)) goto __pyx_L13_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_8);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_13(__pyx_t_1), 2) < 0) __PYX_ERR(0, 1690, __pyx_L7_error)
        __pyx_t_13 = NULL;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L14_unpacking_done;
        __pyx_L13_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_13 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 1690, __pyx_L7_error)
        __pyx_L14_unpacking_done:;
      }
      __pyx_v_loc = __pyx_t_2;
      __pyx_t_2 = 0;
      __pyx_v_tokens = __pyx_t_8;
      __pyx_t_8 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":1691
 *         try:
 *             loc, tokens = self._parse(instring, 0)
 *             if parseAll:             # <<<<<<<<<<<<<<
 *                 loc = self.preParse(instring, loc)
 *                 se = Empty() + StringEnd()
 */
      __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_parseAll); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1691, __pyx_L7_error)
      if (__pyx_t_4) {

        /* "Aspidites/_vendor/pyparsing_extension.py":1692
 *             loc, tokens = self._parse(instring, 0)
 *             if parseAll:
 *                 loc = self.preParse(instring, loc)             # <<<<<<<<<<<<<<
 *                 se = Empty() + StringEnd()
 *                 se._parse(instring, loc)
 */
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_preParse); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1692, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_2 = NULL;
        __pyx_t_12 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_8);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_8, function);
            __pyx_t_12 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_8)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_instring, __pyx_v_loc};
          __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1692, __pyx_L7_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_3);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_instring, __pyx_v_loc};
          __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1692, __pyx_L7_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_3);
        } else
        #endif
        {
          __pyx_t_1 = PyTuple_New(2+__pyx_t_12); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1692, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_instring);
          __Pyx_GIVEREF(__pyx_v_instring);
          PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_12, __pyx_v_instring);
          __Pyx_INCREF(__pyx_v_loc);
          __Pyx_GIVEREF(__pyx_v_loc);
          PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_12, __pyx_v_loc);
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_1, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1692, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        }
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF_SET(__pyx_v_loc, __pyx_t_3);
        __pyx_t_3 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":1693
 *             if parseAll:
 *                 loc = self.preParse(instring, loc)
 *                 se = Empty() + StringEnd()             # <<<<<<<<<<<<<<
 *                 se._parse(instring, loc)
 *         except ParseBaseException as exc:
 */
        __Pyx_INCREF(__pyx_v_Empty);
        __pyx_t_8 = __pyx_v_Empty; __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_8);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_8, function);
          }
        }
        __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1693, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_INCREF(__pyx_v_StringEnd);
        __pyx_t_1 = __pyx_v_StringEnd; __pyx_t_2 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_1, function);
          }
        }
        __pyx_t_8 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1693, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyNumber_Add(__pyx_t_3, __pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1693, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_v_se = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":1694
 *                 loc = self.preParse(instring, loc)
 *                 se = Empty() + StringEnd()
 *                 se._parse(instring, loc)             # <<<<<<<<<<<<<<
 *         except ParseBaseException as exc:
 *             if ParserElement.verbose_stacktrace:
 */
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_se, __pyx_n_s_parse); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1694, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_3 = NULL;
        __pyx_t_12 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_8);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_8, function);
            __pyx_t_12 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_8)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_instring, __pyx_v_loc};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1694, __pyx_L7_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_instring, __pyx_v_loc};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1694, __pyx_L7_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_2 = PyTuple_New(2+__pyx_t_12); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1694, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_2);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(__pyx_v_instring);
          __Pyx_GIVEREF(__pyx_v_instring);
          PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_12, __pyx_v_instring);
          __Pyx_INCREF(__pyx_v_loc);
          __Pyx_GIVEREF(__pyx_v_loc);
          PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_12, __pyx_v_loc);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1694, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":1691
 *         try:
 *             loc, tokens = self._parse(instring, 0)
 *             if parseAll:             # <<<<<<<<<<<<<<
 *                 loc = self.preParse(instring, loc)
 *                 se = Empty() + StringEnd()
 */
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":1689
 *         if not self.keepTabs:
 *             instring = instring.expandtabs()
 *         try:             # <<<<<<<<<<<<<<
 *             loc, tokens = self._parse(instring, 0)
 *             if parseAll:
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":1704
 *                 raise exc
 *         else:
 *             return tokens             # <<<<<<<<<<<<<<
 * 
 *     def scanString(self, instring, maxMatches=sys.maxsize, overlap=False):
 */
    /*else:*/ {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_tokens);
      __pyx_r = __pyx_v_tokens;
      goto __pyx_L10_except_return;
    }
    __pyx_L7_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1695
 *                 se = Empty() + StringEnd()
 *                 se._parse(instring, loc)
 *         except ParseBaseException as exc:             # <<<<<<<<<<<<<<
 *             if ParserElement.verbose_stacktrace:
 *                 raise
 */
    __Pyx_ErrFetch(&__pyx_t_1, &__pyx_t_8, &__pyx_t_2);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ParseBaseException); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1695, __pyx_L9_except_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_12 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_1, __pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_ErrRestore(__pyx_t_1, __pyx_t_8, __pyx_t_2);
    __pyx_t_1 = 0; __pyx_t_8 = 0; __pyx_t_2 = 0;
    if (__pyx_t_12) {
      __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.parseString", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_8, &__pyx_t_1) < 0) __PYX_ERR(0, 1695, __pyx_L9_except_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_8);
      __pyx_v_exc = __pyx_t_8;

      /* "Aspidites/_vendor/pyparsing_extension.py":1696
 *                 se._parse(instring, loc)
 *         except ParseBaseException as exc:
 *             if ParserElement.verbose_stacktrace:             # <<<<<<<<<<<<<<
 *                 raise
 *             else:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1696, __pyx_L9_except_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_verbose_stacktrace); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1696, __pyx_L9_except_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_14); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1696, __pyx_L9_except_error)
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      if (unlikely(__pyx_t_4)) {

        /* "Aspidites/_vendor/pyparsing_extension.py":1697
 *         except ParseBaseException as exc:
 *             if ParserElement.verbose_stacktrace:
 *                 raise             # <<<<<<<<<<<<<<
 *             else:
 *                 # catch and re-raise exception from here, clearing out pyparsing internal stack trace
 */
        __Pyx_GIVEREF(__pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_1);
        __Pyx_ErrRestoreWithState(__pyx_t_2, __pyx_t_8, __pyx_t_1);
        __pyx_t_2 = 0; __pyx_t_8 = 0; __pyx_t_1 = 0; 
        __PYX_ERR(0, 1697, __pyx_L9_except_error)

        /* "Aspidites/_vendor/pyparsing_extension.py":1696
 *                 se._parse(instring, loc)
 *         except ParseBaseException as exc:
 *             if ParserElement.verbose_stacktrace:             # <<<<<<<<<<<<<<
 *                 raise
 *             else:
 */
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":1700
 *             else:
 *                 # catch and re-raise exception from here, clearing out pyparsing internal stack trace
 *                 if getattr(exc, '__traceback__', None) is not None:             # <<<<<<<<<<<<<<
 *                     exc.__traceback__ = self._trim_traceback(exc.__traceback__)
 *                 raise exc
 */
      /*else*/ {
        __pyx_t_14 = __Pyx_GetAttr3(__pyx_v_exc, __pyx_n_s_traceback, Py_None); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1700, __pyx_L9_except_error)
        __Pyx_GOTREF(__pyx_t_14);
        __pyx_t_4 = (__pyx_t_14 != Py_None);
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        __pyx_t_5 = (__pyx_t_4 != 0);
        if (__pyx_t_5) {

          /* "Aspidites/_vendor/pyparsing_extension.py":1701
 *                 # catch and re-raise exception from here, clearing out pyparsing internal stack trace
 *                 if getattr(exc, '__traceback__', None) is not None:
 *                     exc.__traceback__ = self._trim_traceback(exc.__traceback__)             # <<<<<<<<<<<<<<
 *                 raise exc
 *         else:
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_trim_traceback); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1701, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_v_exc, __pyx_n_s_traceback); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1701, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_15);
          __pyx_t_16 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_16)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_16);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          __pyx_t_14 = (__pyx_t_16) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_16, __pyx_t_15) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_15);
          __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1701, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_exc, __pyx_n_s_traceback, __pyx_t_14) < 0) __PYX_ERR(0, 1701, __pyx_L9_except_error)
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":1700
 *             else:
 *                 # catch and re-raise exception from here, clearing out pyparsing internal stack trace
 *                 if getattr(exc, '__traceback__', None) is not None:             # <<<<<<<<<<<<<<
 *                     exc.__traceback__ = self._trim_traceback(exc.__traceback__)
 *                 raise exc
 */
        }

        /* "Aspidites/_vendor/pyparsing_extension.py":1702
 *                 if getattr(exc, '__traceback__', None) is not None:
 *                     exc.__traceback__ = self._trim_traceback(exc.__traceback__)
 *                 raise exc             # <<<<<<<<<<<<<<
 *         else:
 *             return tokens
 */
        __Pyx_Raise(__pyx_v_exc, 0, 0, 0);
        __PYX_ERR(0, 1702, __pyx_L9_except_error)
      }
    }
    goto __pyx_L9_except_error;
    __pyx_L9_except_error:;

    /* "Aspidites/_vendor/pyparsing_extension.py":1689
 *         if not self.keepTabs:
 *             instring = instring.expandtabs()
 *         try:             # <<<<<<<<<<<<<<
 *             loc, tokens = self._parse(instring, 0)
 *             if parseAll:
 */
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_XGIVEREF(__pyx_t_10);
    __Pyx_XGIVEREF(__pyx_t_11);
    __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
    goto __pyx_L1_error;
    __pyx_L10_except_return:;
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_XGIVEREF(__pyx_t_10);
    __Pyx_XGIVEREF(__pyx_t_11);
    __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
    goto __pyx_L0;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1647
 *             ParserElement._parse = ParserElement._parseCache
 * 
 *     def parseString(self, instring, parseAll=False):             # <<<<<<<<<<<<<<
 *         """
 *         Execute the parse expression with the given string.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.parseString", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_Empty);
  __Pyx_XDECREF(__pyx_v_StringEnd);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XDECREF(__pyx_v_loc);
  __Pyx_XDECREF(__pyx_v_tokens);
  __Pyx_XDECREF(__pyx_v_se);
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XDECREF(__pyx_v_instring);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_50generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/pyparsing_extension.py":1706
 *             return tokens
 * 
 *     def scanString(self, instring, maxMatches=sys.maxsize, overlap=False):             # <<<<<<<<<<<<<<
 *         """
 *         Scan the input string for expression matches.  Each match will return the
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_32__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__defaults__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults2, __pyx_self)->__pyx_arg_maxMatches);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults2, __pyx_self)->__pyx_arg_maxMatches);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __Pyx_CyFunction_Defaults(__pyx_defaults2, __pyx_self)->__pyx_arg_maxMatches);
  __Pyx_INCREF(((PyObject *)Py_False));
  __Pyx_GIVEREF(((PyObject *)Py_False));
  PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)Py_False));
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_2, 1, Py_None);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_49scanString(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_48scanString[] = "\n        Scan the input string for expression matches.  Each match will return the\n        matching tokens, start location, and end location.  May be called with optional\n        ``maxMatches`` argument, to clip scanning after 'n' matches are found.  If\n        ``overlap`` is specified, then overlapping matches will be reported.\n\n        Note that the start and end locations are reported relative to the string\n        being parsed.  See :class:`parseString` for more information on parsing\n        strings with embedded tabs.\n\n        Example::\n\n            source = \"sldjf123lsdjjkf345sldkjf879lkjsfd987\"\n            print(source)\n            for tokens, start, end in Word(alphas).scanString(source):\n                print(' '*start + '^'*(end-start))\n                print(' '*start + tokens[0])\n\n        prints::\n\n            sldjf123lsdjjkf345sldkjf879lkjsfd987\n            ^^^^^\n            sldjf\n                    ^^^^^^^\n                    lsdjjkf\n                              ^^^^^^\n                              sldkjf\n                                       ^^^^^^\n                                       lkjsfd\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_49scanString = {"scanString", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_49scanString, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_48scanString};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_49scanString(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_maxMatches = 0;
  PyObject *__pyx_v_overlap = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("scanString (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_maxMatches,&__pyx_n_s_overlap,0};
    PyObject* values[4] = {0,0,0,0};
    __pyx_defaults2 *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults2, __pyx_self);
    values[2] = __pyx_dynamic_args->__pyx_arg_maxMatches;
    values[3] = ((PyObject *)((PyObject *)Py_False));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scanString", 0, 2, 4, 1); __PYX_ERR(0, 1706, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_maxMatches);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_overlap);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "scanString") < 0)) __PYX_ERR(0, 1706, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_maxMatches = values[2];
    __pyx_v_overlap = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("scanString", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1706, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.scanString", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_48scanString(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_maxMatches, __pyx_v_overlap);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_48scanString(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_maxMatches, PyObject *__pyx_v_overlap) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("scanString", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1706, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_instring = __pyx_v_instring;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_instring);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_instring);
  __pyx_cur_scope->__pyx_v_maxMatches = __pyx_v_maxMatches;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_maxMatches);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_maxMatches);
  __pyx_cur_scope->__pyx_v_overlap = __pyx_v_overlap;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_overlap);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_overlap);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_50generator, __pyx_codeobj__63, (PyObject *) __pyx_cur_scope, __pyx_n_s_scanString, __pyx_n_s_ParserElement_scanString, __pyx_n_s_Aspidites__vendor_pyparsing_exte); if (unlikely(!gen)) __PYX_ERR(0, 1706, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.scanString", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_50generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  PyObject *(*__pyx_t_16)(PyObject *);
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("scanString", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L29_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1706, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1737
 *                                        lkjsfd
 *         """
 *         if not self.streamlined:             # <<<<<<<<<<<<<<
 *             self.streamline()
 *         for e in self.ignoreExprs:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_streamlined); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1737, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1737, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = ((!__pyx_t_2) != 0);
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/pyparsing_extension.py":1738
 *         """
 *         if not self.streamlined:
 *             self.streamline()             # <<<<<<<<<<<<<<
 *         for e in self.ignoreExprs:
 *             e.streamline()
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_streamline); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1738, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1738, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1737
 *                                        lkjsfd
 *         """
 *         if not self.streamlined:             # <<<<<<<<<<<<<<
 *             self.streamline()
 *         for e in self.ignoreExprs:
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1739
 *         if not self.streamlined:
 *             self.streamline()
 *         for e in self.ignoreExprs:             # <<<<<<<<<<<<<<
 *             e.streamline()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_ignoreExprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1739, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_4 = __pyx_t_1; __Pyx_INCREF(__pyx_t_4); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1739, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1739, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_4))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_1); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 1739, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1739, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_1); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 1739, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1739, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_7(__pyx_t_4);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1739, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_e);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_e, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1740
 *             self.streamline()
 *         for e in self.ignoreExprs:
 *             e.streamline()             # <<<<<<<<<<<<<<
 * 
 *         if not self.keepTabs:
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_e, __pyx_n_s_streamline); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1740, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1740, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1739
 *         if not self.streamlined:
 *             self.streamline()
 *         for e in self.ignoreExprs:             # <<<<<<<<<<<<<<
 *             e.streamline()
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1742
 *             e.streamline()
 * 
 *         if not self.keepTabs:             # <<<<<<<<<<<<<<
 *             instring = str(instring).expandtabs()
 *         instrlen = len(instring)
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_keepTabs); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1742, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_2 = ((!__pyx_t_3) != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":1743
 * 
 *         if not self.keepTabs:
 *             instring = str(instring).expandtabs()             # <<<<<<<<<<<<<<
 *         instrlen = len(instring)
 *         loc = 0
 */
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_cur_scope->__pyx_v_instring); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1743, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_expandtabs); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1743, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_4 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1743, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_instring);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_instring, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1742
 *             e.streamline()
 * 
 *         if not self.keepTabs:             # <<<<<<<<<<<<<<
 *             instring = str(instring).expandtabs()
 *         instrlen = len(instring)
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1744
 *         if not self.keepTabs:
 *             instring = str(instring).expandtabs()
 *         instrlen = len(instring)             # <<<<<<<<<<<<<<
 *         loc = 0
 *         preparseFn = self.preParse
 */
  __pyx_t_6 = PyObject_Length(__pyx_cur_scope->__pyx_v_instring); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1744, __pyx_L1_error)
  __pyx_cur_scope->__pyx_v_instrlen = __pyx_t_6;

  /* "Aspidites/_vendor/pyparsing_extension.py":1745
 *             instring = str(instring).expandtabs()
 *         instrlen = len(instring)
 *         loc = 0             # <<<<<<<<<<<<<<
 *         preparseFn = self.preParse
 *         parseFn = self._parse
 */
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __pyx_cur_scope->__pyx_v_loc = __pyx_int_0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1746
 *         instrlen = len(instring)
 *         loc = 0
 *         preparseFn = self.preParse             # <<<<<<<<<<<<<<
 *         parseFn = self._parse
 *         ParserElement.resetCache()
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_preParse); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1746, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_cur_scope->__pyx_v_preparseFn = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1747
 *         loc = 0
 *         preparseFn = self.preParse
 *         parseFn = self._parse             # <<<<<<<<<<<<<<
 *         ParserElement.resetCache()
 *         matches = 0
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_parse); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1747, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_cur_scope->__pyx_v_parseFn = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1748
 *         preparseFn = self.preParse
 *         parseFn = self._parse
 *         ParserElement.resetCache()             # <<<<<<<<<<<<<<
 *         matches = 0
 *         try:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1748, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_resetCache); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1748, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1748, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1749
 *         parseFn = self._parse
 *         ParserElement.resetCache()
 *         matches = 0             # <<<<<<<<<<<<<<
 *         try:
 *             while loc <= instrlen and matches < maxMatches:
 */
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __pyx_cur_scope->__pyx_v_matches = __pyx_int_0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1750
 *         ParserElement.resetCache()
 *         matches = 0
 *         try:             # <<<<<<<<<<<<<<
 *             while loc <= instrlen and matches < maxMatches:
 *                 try:
 */
  {
    __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
    __Pyx_XGOTREF(__pyx_t_9);
    __Pyx_XGOTREF(__pyx_t_10);
    __Pyx_XGOTREF(__pyx_t_11);
    /*try:*/ {

      /* "Aspidites/_vendor/pyparsing_extension.py":1751
 *         matches = 0
 *         try:
 *             while loc <= instrlen and matches < maxMatches:             # <<<<<<<<<<<<<<
 *                 try:
 *                     preloc = preparseFn(instring, loc)
 */
      while (1) {
        __pyx_t_4 = PyInt_FromSsize_t(__pyx_cur_scope->__pyx_v_instrlen); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1751, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_1 = PyObject_RichCompare(__pyx_cur_scope->__pyx_v_loc, __pyx_t_4, Py_LE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1751, __pyx_L8_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1751, __pyx_L8_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_3) {
        } else {
          __pyx_t_2 = __pyx_t_3;
          goto __pyx_L16_bool_binop_done;
        }
        __pyx_t_1 = PyObject_RichCompare(__pyx_cur_scope->__pyx_v_matches, __pyx_cur_scope->__pyx_v_maxMatches, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1751, __pyx_L8_error)
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1751, __pyx_L8_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_2 = __pyx_t_3;
        __pyx_L16_bool_binop_done:;
        if (!__pyx_t_2) break;

        /* "Aspidites/_vendor/pyparsing_extension.py":1752
 *         try:
 *             while loc <= instrlen and matches < maxMatches:
 *                 try:             # <<<<<<<<<<<<<<
 *                     preloc = preparseFn(instring, loc)
 *                     nextLoc, tokens = parseFn(instring, preloc, callPreParse=False)
 */
        {
          __Pyx_ExceptionSave(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_12);
          __Pyx_XGOTREF(__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_14);
          /*try:*/ {

            /* "Aspidites/_vendor/pyparsing_extension.py":1753
 *             while loc <= instrlen and matches < maxMatches:
 *                 try:
 *                     preloc = preparseFn(instring, loc)             # <<<<<<<<<<<<<<
 *                     nextLoc, tokens = parseFn(instring, preloc, callPreParse=False)
 *                 except ParseException:
 */
            __Pyx_INCREF(__pyx_cur_scope->__pyx_v_preparseFn);
            __pyx_t_4 = __pyx_cur_scope->__pyx_v_preparseFn; __pyx_t_5 = NULL;
            __pyx_t_15 = 0;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_4, function);
                __pyx_t_15 = 1;
              }
            }
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_cur_scope->__pyx_v_instring, __pyx_cur_scope->__pyx_v_loc};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_15, 2+__pyx_t_15); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1753, __pyx_L18_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
              PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_cur_scope->__pyx_v_instring, __pyx_cur_scope->__pyx_v_loc};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_15, 2+__pyx_t_15); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1753, __pyx_L18_error)
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_8 = PyTuple_New(2+__pyx_t_15); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1753, __pyx_L18_error)
              __Pyx_GOTREF(__pyx_t_8);
              if (__pyx_t_5) {
                __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_5); __pyx_t_5 = NULL;
              }
              __Pyx_INCREF(__pyx_cur_scope->__pyx_v_instring);
              __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_instring);
              PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_15, __pyx_cur_scope->__pyx_v_instring);
              __Pyx_INCREF(__pyx_cur_scope->__pyx_v_loc);
              __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_loc);
              PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_15, __pyx_cur_scope->__pyx_v_loc);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1753, __pyx_L18_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            }
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_preloc);
            __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_preloc, __pyx_t_1);
            __Pyx_GIVEREF(__pyx_t_1);
            __pyx_t_1 = 0;

            /* "Aspidites/_vendor/pyparsing_extension.py":1754
 *                 try:
 *                     preloc = preparseFn(instring, loc)
 *                     nextLoc, tokens = parseFn(instring, preloc, callPreParse=False)             # <<<<<<<<<<<<<<
 *                 except ParseException:
 *                     loc = preloc + 1
 */
            __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1754, __pyx_L18_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_INCREF(__pyx_cur_scope->__pyx_v_instring);
            __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_instring);
            PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_cur_scope->__pyx_v_instring);
            __Pyx_INCREF(__pyx_cur_scope->__pyx_v_preloc);
            __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_preloc);
            PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_cur_scope->__pyx_v_preloc);
            __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1754, __pyx_L18_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_callPreParse, Py_False) < 0) __PYX_ERR(0, 1754, __pyx_L18_error)
            __pyx_t_8 = __Pyx_PyObject_Call(__pyx_cur_scope->__pyx_v_parseFn, __pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1754, __pyx_L18_error)
            __Pyx_GOTREF(__pyx_t_8);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            if ((likely(PyTuple_CheckExact(__pyx_t_8))) || (PyList_CheckExact(__pyx_t_8))) {
              PyObject* sequence = __pyx_t_8;
              Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
              if (unlikely(size != 2)) {
                if (size > 2) __Pyx_RaiseTooManyValuesError(2);
                else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                __PYX_ERR(0, 1754, __pyx_L18_error)
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              if (likely(PyTuple_CheckExact(sequence))) {
                __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
                __pyx_t_1 = PyTuple_GET_ITEM(sequence, 1); 
              } else {
                __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
                __pyx_t_1 = PyList_GET_ITEM(sequence, 1); 
              }
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_1);
              #else
              __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1754, __pyx_L18_error)
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_1 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1754, __pyx_L18_error)
              __Pyx_GOTREF(__pyx_t_1);
              #endif
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            } else {
              Py_ssize_t index = -1;
              __pyx_t_5 = PyObject_GetIter(__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1754, __pyx_L18_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              __pyx_t_16 = Py_TYPE(__pyx_t_5)->tp_iternext;
              index = 0; __pyx_t_4 = __pyx_t_16(__pyx_t_5); if (unlikely(!__pyx_t_4)) goto __pyx_L26_unpacking_failed;
              __Pyx_GOTREF(__pyx_t_4);
              index = 1; __pyx_t_1 = __pyx_t_16(__pyx_t_5); if (unlikely(!__pyx_t_1)) goto __pyx_L26_unpacking_failed;
              __Pyx_GOTREF(__pyx_t_1);
              if (__Pyx_IternextUnpackEndCheck(__pyx_t_16(__pyx_t_5), 2) < 0) __PYX_ERR(0, 1754, __pyx_L18_error)
              __pyx_t_16 = NULL;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              goto __pyx_L27_unpacking_done;
              __pyx_L26_unpacking_failed:;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __pyx_t_16 = NULL;
              if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
              __PYX_ERR(0, 1754, __pyx_L18_error)
              __pyx_L27_unpacking_done:;
            }
            __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_nextLoc);
            __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_nextLoc, __pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_4);
            __pyx_t_4 = 0;
            __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_tokens);
            __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_tokens, __pyx_t_1);
            __Pyx_GIVEREF(__pyx_t_1);
            __pyx_t_1 = 0;

            /* "Aspidites/_vendor/pyparsing_extension.py":1752
 *         try:
 *             while loc <= instrlen and matches < maxMatches:
 *                 try:             # <<<<<<<<<<<<<<
 *                     preloc = preparseFn(instring, loc)
 *                     nextLoc, tokens = parseFn(instring, preloc, callPreParse=False)
 */
          }

          /* "Aspidites/_vendor/pyparsing_extension.py":1758
 *                     loc = preloc + 1
 *                 else:
 *                     if nextLoc > loc:             # <<<<<<<<<<<<<<
 *                         matches += 1
 *                         yield tokens, preloc, nextLoc
 */
          /*else:*/ {
            __pyx_t_8 = PyObject_RichCompare(__pyx_cur_scope->__pyx_v_nextLoc, __pyx_cur_scope->__pyx_v_loc, Py_GT); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1758, __pyx_L20_except_error)
            __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1758, __pyx_L20_except_error)
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            if (__pyx_t_2) {

              /* "Aspidites/_vendor/pyparsing_extension.py":1759
 *                 else:
 *                     if nextLoc > loc:
 *                         matches += 1             # <<<<<<<<<<<<<<
 *                         yield tokens, preloc, nextLoc
 *                         if overlap:
 */
              __pyx_t_8 = __Pyx_PyInt_AddObjC(__pyx_cur_scope->__pyx_v_matches, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1759, __pyx_L20_except_error)
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_matches);
              __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_matches, __pyx_t_8);
              __Pyx_GIVEREF(__pyx_t_8);
              __pyx_t_8 = 0;

              /* "Aspidites/_vendor/pyparsing_extension.py":1760
 *                     if nextLoc > loc:
 *                         matches += 1
 *                         yield tokens, preloc, nextLoc             # <<<<<<<<<<<<<<
 *                         if overlap:
 *                             nextloc = preparseFn(instring, loc)
 */
              __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1760, __pyx_L20_except_error)
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_INCREF(__pyx_cur_scope->__pyx_v_tokens);
              __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_tokens);
              PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_cur_scope->__pyx_v_tokens);
              __Pyx_INCREF(__pyx_cur_scope->__pyx_v_preloc);
              __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_preloc);
              PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_cur_scope->__pyx_v_preloc);
              __Pyx_INCREF(__pyx_cur_scope->__pyx_v_nextLoc);
              __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_nextLoc);
              PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_cur_scope->__pyx_v_nextLoc);
              __pyx_r = __pyx_t_8;
              __pyx_t_8 = 0;
              __Pyx_XGIVEREF(__pyx_t_9);
              __pyx_cur_scope->__pyx_t_0 = __pyx_t_9;
              __Pyx_XGIVEREF(__pyx_t_10);
              __pyx_cur_scope->__pyx_t_1 = __pyx_t_10;
              __Pyx_XGIVEREF(__pyx_t_11);
              __pyx_cur_scope->__pyx_t_2 = __pyx_t_11;
              __Pyx_XGIVEREF(__pyx_t_12);
              __pyx_cur_scope->__pyx_t_3 = __pyx_t_12;
              __Pyx_XGIVEREF(__pyx_t_13);
              __pyx_cur_scope->__pyx_t_4 = __pyx_t_13;
              __Pyx_XGIVEREF(__pyx_t_14);
              __pyx_cur_scope->__pyx_t_5 = __pyx_t_14;
              __Pyx_XGIVEREF(__pyx_r);
              __Pyx_RefNannyFinishContext();
              __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
              /* return from generator, yielding value */
              __pyx_generator->resume_label = 1;
              return __pyx_r;
              __pyx_L29_resume_from_yield:;
              __pyx_t_9 = __pyx_cur_scope->__pyx_t_0;
              __pyx_cur_scope->__pyx_t_0 = 0;
              __Pyx_XGOTREF(__pyx_t_9);
              __pyx_t_10 = __pyx_cur_scope->__pyx_t_1;
              __pyx_cur_scope->__pyx_t_1 = 0;
              __Pyx_XGOTREF(__pyx_t_10);
              __pyx_t_11 = __pyx_cur_scope->__pyx_t_2;
              __pyx_cur_scope->__pyx_t_2 = 0;
              __Pyx_XGOTREF(__pyx_t_11);
              __pyx_t_12 = __pyx_cur_scope->__pyx_t_3;
              __pyx_cur_scope->__pyx_t_3 = 0;
              __Pyx_XGOTREF(__pyx_t_12);
              __pyx_t_13 = __pyx_cur_scope->__pyx_t_4;
              __pyx_cur_scope->__pyx_t_4 = 0;
              __Pyx_XGOTREF(__pyx_t_13);
              __pyx_t_14 = __pyx_cur_scope->__pyx_t_5;
              __pyx_cur_scope->__pyx_t_5 = 0;
              __Pyx_XGOTREF(__pyx_t_14);
              if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1760, __pyx_L20_except_error)

              /* "Aspidites/_vendor/pyparsing_extension.py":1761
 *                         matches += 1
 *                         yield tokens, preloc, nextLoc
 *                         if overlap:             # <<<<<<<<<<<<<<
 *                             nextloc = preparseFn(instring, loc)
 *                             if nextloc > loc:
 */
              __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_overlap); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1761, __pyx_L20_except_error)
              if (__pyx_t_2) {

                /* "Aspidites/_vendor/pyparsing_extension.py":1762
 *                         yield tokens, preloc, nextLoc
 *                         if overlap:
 *                             nextloc = preparseFn(instring, loc)             # <<<<<<<<<<<<<<
 *                             if nextloc > loc:
 *                                 loc = nextLoc
 */
                __Pyx_INCREF(__pyx_cur_scope->__pyx_v_preparseFn);
                __pyx_t_1 = __pyx_cur_scope->__pyx_v_preparseFn; __pyx_t_4 = NULL;
                __pyx_t_15 = 0;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
                  __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
                  if (likely(__pyx_t_4)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
                    __Pyx_INCREF(__pyx_t_4);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_1, function);
                    __pyx_t_15 = 1;
                  }
                }
                #if CYTHON_FAST_PYCALL
                if (PyFunction_Check(__pyx_t_1)) {
                  PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_cur_scope->__pyx_v_instring, __pyx_cur_scope->__pyx_v_loc};
                  __pyx_t_8 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_15, 2+__pyx_t_15); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1762, __pyx_L20_except_error)
                  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                  __Pyx_GOTREF(__pyx_t_8);
                } else
                #endif
                #if CYTHON_FAST_PYCCALL
                if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
                  PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_cur_scope->__pyx_v_instring, __pyx_cur_scope->__pyx_v_loc};
                  __pyx_t_8 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_15, 2+__pyx_t_15); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1762, __pyx_L20_except_error)
                  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                  __Pyx_GOTREF(__pyx_t_8);
                } else
                #endif
                {
                  __pyx_t_5 = PyTuple_New(2+__pyx_t_15); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1762, __pyx_L20_except_error)
                  __Pyx_GOTREF(__pyx_t_5);
                  if (__pyx_t_4) {
                    __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
                  }
                  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_instring);
                  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_instring);
                  PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_15, __pyx_cur_scope->__pyx_v_instring);
                  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_loc);
                  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_loc);
                  PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_15, __pyx_cur_scope->__pyx_v_loc);
                  __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_5, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1762, __pyx_L20_except_error)
                  __Pyx_GOTREF(__pyx_t_8);
                  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                }
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_nextloc);
                __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_nextloc, __pyx_t_8);
                __Pyx_GIVEREF(__pyx_t_8);
                __pyx_t_8 = 0;

                /* "Aspidites/_vendor/pyparsing_extension.py":1763
 *                         if overlap:
 *                             nextloc = preparseFn(instring, loc)
 *                             if nextloc > loc:             # <<<<<<<<<<<<<<
 *                                 loc = nextLoc
 *                             else:
 */
                __pyx_t_8 = PyObject_RichCompare(__pyx_cur_scope->__pyx_v_nextloc, __pyx_cur_scope->__pyx_v_loc, Py_GT); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1763, __pyx_L20_except_error)
                __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1763, __pyx_L20_except_error)
                __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                if (__pyx_t_2) {

                  /* "Aspidites/_vendor/pyparsing_extension.py":1764
 *                             nextloc = preparseFn(instring, loc)
 *                             if nextloc > loc:
 *                                 loc = nextLoc             # <<<<<<<<<<<<<<
 *                             else:
 *                                 loc += 1
 */
                  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_nextLoc);
                  __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_loc);
                  __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_loc, __pyx_cur_scope->__pyx_v_nextLoc);
                  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_nextLoc);

                  /* "Aspidites/_vendor/pyparsing_extension.py":1763
 *                         if overlap:
 *                             nextloc = preparseFn(instring, loc)
 *                             if nextloc > loc:             # <<<<<<<<<<<<<<
 *                                 loc = nextLoc
 *                             else:
 */
                  goto __pyx_L31;
                }

                /* "Aspidites/_vendor/pyparsing_extension.py":1766
 *                                 loc = nextLoc
 *                             else:
 *                                 loc += 1             # <<<<<<<<<<<<<<
 *                         else:
 *                             loc = nextLoc
 */
                /*else*/ {
                  __pyx_t_8 = __Pyx_PyInt_AddObjC(__pyx_cur_scope->__pyx_v_loc, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1766, __pyx_L20_except_error)
                  __Pyx_GOTREF(__pyx_t_8);
                  __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_loc);
                  __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_loc, __pyx_t_8);
                  __Pyx_GIVEREF(__pyx_t_8);
                  __pyx_t_8 = 0;
                }
                __pyx_L31:;

                /* "Aspidites/_vendor/pyparsing_extension.py":1761
 *                         matches += 1
 *                         yield tokens, preloc, nextLoc
 *                         if overlap:             # <<<<<<<<<<<<<<
 *                             nextloc = preparseFn(instring, loc)
 *                             if nextloc > loc:
 */
                goto __pyx_L30;
              }

              /* "Aspidites/_vendor/pyparsing_extension.py":1768
 *                                 loc += 1
 *                         else:
 *                             loc = nextLoc             # <<<<<<<<<<<<<<
 *                     else:
 *                         loc = preloc + 1
 */
              /*else*/ {
                __Pyx_INCREF(__pyx_cur_scope->__pyx_v_nextLoc);
                __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_loc);
                __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_loc, __pyx_cur_scope->__pyx_v_nextLoc);
                __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_nextLoc);
              }
              __pyx_L30:;

              /* "Aspidites/_vendor/pyparsing_extension.py":1758
 *                     loc = preloc + 1
 *                 else:
 *                     if nextLoc > loc:             # <<<<<<<<<<<<<<
 *                         matches += 1
 *                         yield tokens, preloc, nextLoc
 */
              goto __pyx_L28;
            }

            /* "Aspidites/_vendor/pyparsing_extension.py":1770
 *                             loc = nextLoc
 *                     else:
 *                         loc = preloc + 1             # <<<<<<<<<<<<<<
 *         except ParseBaseException as exc:
 *             if ParserElement.verbose_stacktrace:
 */
            /*else*/ {
              __pyx_t_8 = __Pyx_PyInt_AddObjC(__pyx_cur_scope->__pyx_v_preloc, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1770, __pyx_L20_except_error)
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_loc);
              __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_loc, __pyx_t_8);
              __Pyx_GIVEREF(__pyx_t_8);
              __pyx_t_8 = 0;
            }
            __pyx_L28:;
          }
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          goto __pyx_L25_try_end;
          __pyx_L18_error:;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":1755
 *                     preloc = preparseFn(instring, loc)
 *                     nextLoc, tokens = parseFn(instring, preloc, callPreParse=False)
 *                 except ParseException:             # <<<<<<<<<<<<<<
 *                     loc = preloc + 1
 *                 else:
 */
          __Pyx_ErrFetch(&__pyx_t_8, &__pyx_t_1, &__pyx_t_5);
          __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1755, __pyx_L20_except_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_15 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_8, __pyx_t_4);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_ErrRestore(__pyx_t_8, __pyx_t_1, __pyx_t_5);
          __pyx_t_8 = 0; __pyx_t_1 = 0; __pyx_t_5 = 0;
          if (__pyx_t_15) {
            __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.scanString", __pyx_clineno, __pyx_lineno, __pyx_filename);
            if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_1, &__pyx_t_8) < 0) __PYX_ERR(0, 1755, __pyx_L20_except_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_GOTREF(__pyx_t_8);

            /* "Aspidites/_vendor/pyparsing_extension.py":1756
 *                     nextLoc, tokens = parseFn(instring, preloc, callPreParse=False)
 *                 except ParseException:
 *                     loc = preloc + 1             # <<<<<<<<<<<<<<
 *                 else:
 *                     if nextLoc > loc:
 */
            if (unlikely(!__pyx_cur_scope->__pyx_v_preloc)) { __Pyx_RaiseUnboundLocalError("preloc"); __PYX_ERR(0, 1756, __pyx_L20_except_error) }
            __pyx_t_4 = __Pyx_PyInt_AddObjC(__pyx_cur_scope->__pyx_v_preloc, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1756, __pyx_L20_except_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_loc);
            __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_loc, __pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_4);
            __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            goto __pyx_L19_exception_handled;
          }
          goto __pyx_L20_except_error;
          __pyx_L20_except_error:;

          /* "Aspidites/_vendor/pyparsing_extension.py":1752
 *         try:
 *             while loc <= instrlen and matches < maxMatches:
 *                 try:             # <<<<<<<<<<<<<<
 *                     preloc = preparseFn(instring, loc)
 *                     nextLoc, tokens = parseFn(instring, preloc, callPreParse=False)
 */
          __Pyx_XGIVEREF(__pyx_t_12);
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
          goto __pyx_L8_error;
          __pyx_L19_exception_handled:;
          __Pyx_XGIVEREF(__pyx_t_12);
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
          __pyx_L25_try_end:;
        }
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":1750
 *         ParserElement.resetCache()
 *         matches = 0
 *         try:             # <<<<<<<<<<<<<<
 *             while loc <= instrlen and matches < maxMatches:
 *                 try:
 */
    }
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    goto __pyx_L13_try_end;
    __pyx_L8_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1771
 *                     else:
 *                         loc = preloc + 1
 *         except ParseBaseException as exc:             # <<<<<<<<<<<<<<
 *             if ParserElement.verbose_stacktrace:
 *                 raise
 */
    __Pyx_ErrFetch(&__pyx_t_8, &__pyx_t_1, &__pyx_t_5);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ParseBaseException); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1771, __pyx_L10_except_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_15 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_8, __pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_ErrRestore(__pyx_t_8, __pyx_t_1, __pyx_t_5);
    __pyx_t_8 = 0; __pyx_t_1 = 0; __pyx_t_5 = 0;
    if (__pyx_t_15) {
      __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.scanString", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_1, &__pyx_t_8) < 0) __PYX_ERR(0, 1771, __pyx_L10_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_1);
      __pyx_cur_scope->__pyx_v_exc = __pyx_t_1;

      /* "Aspidites/_vendor/pyparsing_extension.py":1772
 *                         loc = preloc + 1
 *         except ParseBaseException as exc:
 *             if ParserElement.verbose_stacktrace:             # <<<<<<<<<<<<<<
 *                 raise
 *             else:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1772, __pyx_L10_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_verbose_stacktrace); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1772, __pyx_L10_except_error)
      __Pyx_GOTREF(__pyx_t_17);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_17); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1772, __pyx_L10_except_error)
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
      if (unlikely(__pyx_t_2)) {

        /* "Aspidites/_vendor/pyparsing_extension.py":1773
 *         except ParseBaseException as exc:
 *             if ParserElement.verbose_stacktrace:
 *                 raise             # <<<<<<<<<<<<<<
 *             else:
 *                 # catch and re-raise exception from here, clearing out pyparsing internal stack trace
 */
        __Pyx_GIVEREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_1);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_ErrRestoreWithState(__pyx_t_5, __pyx_t_1, __pyx_t_8);
        __pyx_t_5 = 0; __pyx_t_1 = 0; __pyx_t_8 = 0; 
        __PYX_ERR(0, 1773, __pyx_L10_except_error)

        /* "Aspidites/_vendor/pyparsing_extension.py":1772
 *                         loc = preloc + 1
 *         except ParseBaseException as exc:
 *             if ParserElement.verbose_stacktrace:             # <<<<<<<<<<<<<<
 *                 raise
 *             else:
 */
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":1776
 *             else:
 *                 # catch and re-raise exception from here, clearing out pyparsing internal stack trace
 *                 if getattr(exc, '__traceback__', None) is not None:             # <<<<<<<<<<<<<<
 *                     exc.__traceback__ = self._trim_traceback(exc.__traceback__)
 *                 raise exc
 */
      /*else*/ {
        __pyx_t_17 = __Pyx_GetAttr3(__pyx_cur_scope->__pyx_v_exc, __pyx_n_s_traceback, Py_None); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1776, __pyx_L10_except_error)
        __Pyx_GOTREF(__pyx_t_17);
        __pyx_t_2 = (__pyx_t_17 != Py_None);
        __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
        __pyx_t_3 = (__pyx_t_2 != 0);
        if (__pyx_t_3) {

          /* "Aspidites/_vendor/pyparsing_extension.py":1777
 *                 # catch and re-raise exception from here, clearing out pyparsing internal stack trace
 *                 if getattr(exc, '__traceback__', None) is not None:
 *                     exc.__traceback__ = self._trim_traceback(exc.__traceback__)             # <<<<<<<<<<<<<<
 *                 raise exc
 * 
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_trim_traceback); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1777, __pyx_L10_except_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_exc, __pyx_n_s_traceback); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1777, __pyx_L10_except_error)
          __Pyx_GOTREF(__pyx_t_18);
          __pyx_t_19 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_19 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_19)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_19);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          __pyx_t_17 = (__pyx_t_19) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_19, __pyx_t_18) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_18);
          __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
          __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
          if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1777, __pyx_L10_except_error)
          __Pyx_GOTREF(__pyx_t_17);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_exc, __pyx_n_s_traceback, __pyx_t_17) < 0) __PYX_ERR(0, 1777, __pyx_L10_except_error)
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":1776
 *             else:
 *                 # catch and re-raise exception from here, clearing out pyparsing internal stack trace
 *                 if getattr(exc, '__traceback__', None) is not None:             # <<<<<<<<<<<<<<
 *                     exc.__traceback__ = self._trim_traceback(exc.__traceback__)
 *                 raise exc
 */
        }

        /* "Aspidites/_vendor/pyparsing_extension.py":1778
 *                 if getattr(exc, '__traceback__', None) is not None:
 *                     exc.__traceback__ = self._trim_traceback(exc.__traceback__)
 *                 raise exc             # <<<<<<<<<<<<<<
 * 
 *     def transformString(self, instring):
 */
        __Pyx_Raise(__pyx_cur_scope->__pyx_v_exc, 0, 0, 0);
        __PYX_ERR(0, 1778, __pyx_L10_except_error)
      }
    }
    goto __pyx_L10_except_error;
    __pyx_L10_except_error:;

    /* "Aspidites/_vendor/pyparsing_extension.py":1750
 *         ParserElement.resetCache()
 *         matches = 0
 *         try:             # <<<<<<<<<<<<<<
 *             while loc <= instrlen and matches < maxMatches:
 *                 try:
 */
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_XGIVEREF(__pyx_t_10);
    __Pyx_XGIVEREF(__pyx_t_11);
    __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
    goto __pyx_L1_error;
    __pyx_L13_try_end:;
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "Aspidites/_vendor/pyparsing_extension.py":1706
 *             return tokens
 * 
 *     def scanString(self, instring, maxMatches=sys.maxsize, overlap=False):             # <<<<<<<<<<<<<<
 *         """
 *         Scan the input string for expression matches.  Each match will return the
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_AddTraceback("scanString", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1780
 *                 raise exc
 * 
 *     def transformString(self, instring):             # <<<<<<<<<<<<<<
 *         """
 *         Extension to :class:`scanString`, to modify matching text with modified tokens that may
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_52transformString(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_51transformString[] = "\n        Extension to :class:`scanString`, to modify matching text with modified tokens that may\n        be returned from a parse action.  To use ``transformString``, define a grammar and\n        attach a parse action to it that modifies the returned token list.\n        Invoking ``transformString()`` on a target string will then scan for matches,\n        and replace the matched text patterns according to the logic in the parse\n        action.  ``transformString()`` returns the resulting transformed string.\n\n        Example::\n\n            wd = Word(alphas)\n            wd.setParseAction(lambda toks: toks[0].title())\n\n            print(wd.transformString(\"now is the winter of our discontent made glorious summer by this sun of york.\"))\n\n        prints::\n\n            Now Is The Winter Of Our Discontent Made Glorious Summer By This Sun Of York.\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_52transformString = {"transformString", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_52transformString, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_51transformString};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_52transformString(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("transformString (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("transformString", 1, 2, 2, 1); __PYX_ERR(0, 1780, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "transformString") < 0)) __PYX_ERR(0, 1780, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("transformString", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1780, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.transformString", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_51transformString(__pyx_self, __pyx_v_self, __pyx_v_instring);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_51transformString(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring) {
  PyObject *__pyx_v_out = NULL;
  PyObject *__pyx_v_lastE = NULL;
  PyObject *__pyx_v_t = NULL;
  PyObject *__pyx_v_s = NULL;
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_v_o = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *(*__pyx_t_12)(PyObject *);
  int __pyx_t_13;
  int __pyx_t_14;
  int __pyx_t_15;
  int __pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("transformString", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":1800
 *             Now Is The Winter Of Our Discontent Made Glorious Summer By This Sun Of York.
 *         """
 *         out = []             # <<<<<<<<<<<<<<
 *         lastE = 0
 *         # force preservation of <TAB>s, to minimize unwanted transformation of string, and to
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1800, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_out = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1801
 *         """
 *         out = []
 *         lastE = 0             # <<<<<<<<<<<<<<
 *         # force preservation of <TAB>s, to minimize unwanted transformation of string, and to
 *         # keep string locs straight between transformString and scanString
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_v_lastE = __pyx_int_0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1804
 *         # force preservation of <TAB>s, to minimize unwanted transformation of string, and to
 *         # keep string locs straight between transformString and scanString
 *         self.keepTabs = True             # <<<<<<<<<<<<<<
 *         try:
 *             for t, s, e in self.scanString(instring):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_keepTabs, Py_True) < 0) __PYX_ERR(0, 1804, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1805
 *         # keep string locs straight between transformString and scanString
 *         self.keepTabs = True
 *         try:             # <<<<<<<<<<<<<<
 *             for t, s, e in self.scanString(instring):
 *                 out.append(instring[lastE:s])
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    /*try:*/ {

      /* "Aspidites/_vendor/pyparsing_extension.py":1806
 *         self.keepTabs = True
 *         try:
 *             for t, s, e in self.scanString(instring):             # <<<<<<<<<<<<<<
 *                 out.append(instring[lastE:s])
 *                 if t:
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_scanString); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1806, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_v_instring) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_instring);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1806, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
        __pyx_t_5 = __pyx_t_1; __Pyx_INCREF(__pyx_t_5); __pyx_t_7 = 0;
        __pyx_t_8 = NULL;
      } else {
        __pyx_t_7 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1806, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_8 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1806, __pyx_L3_error)
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      for (;;) {
        if (likely(!__pyx_t_8)) {
          if (likely(PyList_CheckExact(__pyx_t_5))) {
            if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_5)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_1 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 1806, __pyx_L3_error)
            #else
            __pyx_t_1 = PySequence_ITEM(__pyx_t_5, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1806, __pyx_L3_error)
            __Pyx_GOTREF(__pyx_t_1);
            #endif
          } else {
            if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 1806, __pyx_L3_error)
            #else
            __pyx_t_1 = PySequence_ITEM(__pyx_t_5, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1806, __pyx_L3_error)
            __Pyx_GOTREF(__pyx_t_1);
            #endif
          }
        } else {
          __pyx_t_1 = __pyx_t_8(__pyx_t_5);
          if (unlikely(!__pyx_t_1)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 1806, __pyx_L3_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_1);
        }
        if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
          PyObject* sequence = __pyx_t_1;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 3)) {
            if (size > 3) __Pyx_RaiseTooManyValuesError(3);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 1806, __pyx_L3_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_6 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
            __pyx_t_10 = PyTuple_GET_ITEM(sequence, 2); 
          } else {
            __pyx_t_6 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
            __pyx_t_10 = PyList_GET_ITEM(sequence, 2); 
          }
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_10);
          #else
          __pyx_t_6 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1806, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1806, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_9);
          __pyx_t_10 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1806, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_10);
          #endif
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          Py_ssize_t index = -1;
          __pyx_t_11 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1806, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_12 = Py_TYPE(__pyx_t_11)->tp_iternext;
          index = 0; __pyx_t_6 = __pyx_t_12(__pyx_t_11); if (unlikely(!__pyx_t_6)) goto __pyx_L11_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_6);
          index = 1; __pyx_t_9 = __pyx_t_12(__pyx_t_11); if (unlikely(!__pyx_t_9)) goto __pyx_L11_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_9);
          index = 2; __pyx_t_10 = __pyx_t_12(__pyx_t_11); if (unlikely(!__pyx_t_10)) goto __pyx_L11_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_10);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_11), 3) < 0) __PYX_ERR(0, 1806, __pyx_L3_error)
          __pyx_t_12 = NULL;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          goto __pyx_L12_unpacking_done;
          __pyx_L11_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_12 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 1806, __pyx_L3_error)
          __pyx_L12_unpacking_done:;
        }
        __Pyx_XDECREF_SET(__pyx_v_t, __pyx_t_6);
        __pyx_t_6 = 0;
        __Pyx_XDECREF_SET(__pyx_v_s, __pyx_t_9);
        __pyx_t_9 = 0;
        __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_10);
        __pyx_t_10 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":1807
 *         try:
 *             for t, s, e in self.scanString(instring):
 *                 out.append(instring[lastE:s])             # <<<<<<<<<<<<<<
 *                 if t:
 *                     if isinstance(t, ParseResults):
 */
        __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_instring, 0, 0, &__pyx_v_lastE, &__pyx_v_s, NULL, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1807, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_13 = __Pyx_PyObject_Append(__pyx_v_out, __pyx_t_1); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 1807, __pyx_L3_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":1808
 *             for t, s, e in self.scanString(instring):
 *                 out.append(instring[lastE:s])
 *                 if t:             # <<<<<<<<<<<<<<
 *                     if isinstance(t, ParseResults):
 *                         out += t.asList()
 */
        __pyx_t_14 = __Pyx_PyObject_IsTrue(__pyx_v_t); if (unlikely(__pyx_t_14 < 0)) __PYX_ERR(0, 1808, __pyx_L3_error)
        if (__pyx_t_14) {

          /* "Aspidites/_vendor/pyparsing_extension.py":1809
 *                 out.append(instring[lastE:s])
 *                 if t:
 *                     if isinstance(t, ParseResults):             # <<<<<<<<<<<<<<
 *                         out += t.asList()
 *                     elif isinstance(t, list):
 */
          __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParseResults); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1809, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_14 = PyObject_IsInstance(__pyx_v_t, __pyx_t_1); if (unlikely(__pyx_t_14 == ((int)-1))) __PYX_ERR(0, 1809, __pyx_L3_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_15 = (__pyx_t_14 != 0);
          if (__pyx_t_15) {

            /* "Aspidites/_vendor/pyparsing_extension.py":1810
 *                 if t:
 *                     if isinstance(t, ParseResults):
 *                         out += t.asList()             # <<<<<<<<<<<<<<
 *                     elif isinstance(t, list):
 *                         out += t
 */
            __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_t, __pyx_n_s_asList); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1810, __pyx_L3_error)
            __Pyx_GOTREF(__pyx_t_10);
            __pyx_t_9 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
              __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_10);
              if (likely(__pyx_t_9)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
                __Pyx_INCREF(__pyx_t_9);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_10, function);
              }
            }
            __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_9) : __Pyx_PyObject_CallNoArg(__pyx_t_10);
            __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
            if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1810, __pyx_L3_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            __pyx_t_10 = PyNumber_InPlaceAdd(__pyx_v_out, __pyx_t_1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1810, __pyx_L3_error)
            __Pyx_GOTREF(__pyx_t_10);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF_SET(__pyx_v_out, __pyx_t_10);
            __pyx_t_10 = 0;

            /* "Aspidites/_vendor/pyparsing_extension.py":1809
 *                 out.append(instring[lastE:s])
 *                 if t:
 *                     if isinstance(t, ParseResults):             # <<<<<<<<<<<<<<
 *                         out += t.asList()
 *                     elif isinstance(t, list):
 */
            goto __pyx_L14;
          }

          /* "Aspidites/_vendor/pyparsing_extension.py":1811
 *                     if isinstance(t, ParseResults):
 *                         out += t.asList()
 *                     elif isinstance(t, list):             # <<<<<<<<<<<<<<
 *                         out += t
 *                     else:
 */
          __pyx_t_15 = PyList_Check(__pyx_v_t); 
          __pyx_t_14 = (__pyx_t_15 != 0);
          if (__pyx_t_14) {

            /* "Aspidites/_vendor/pyparsing_extension.py":1812
 *                         out += t.asList()
 *                     elif isinstance(t, list):
 *                         out += t             # <<<<<<<<<<<<<<
 *                     else:
 *                         out.append(t)
 */
            __pyx_t_10 = PyNumber_InPlaceAdd(__pyx_v_out, __pyx_v_t); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1812, __pyx_L3_error)
            __Pyx_GOTREF(__pyx_t_10);
            __Pyx_DECREF_SET(__pyx_v_out, __pyx_t_10);
            __pyx_t_10 = 0;

            /* "Aspidites/_vendor/pyparsing_extension.py":1811
 *                     if isinstance(t, ParseResults):
 *                         out += t.asList()
 *                     elif isinstance(t, list):             # <<<<<<<<<<<<<<
 *                         out += t
 *                     else:
 */
            goto __pyx_L14;
          }

          /* "Aspidites/_vendor/pyparsing_extension.py":1814
 *                         out += t
 *                     else:
 *                         out.append(t)             # <<<<<<<<<<<<<<
 *                 lastE = e
 *             out.append(instring[lastE:])
 */
          /*else*/ {
            __pyx_t_13 = __Pyx_PyObject_Append(__pyx_v_out, __pyx_v_t); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 1814, __pyx_L3_error)
          }
          __pyx_L14:;

          /* "Aspidites/_vendor/pyparsing_extension.py":1808
 *             for t, s, e in self.scanString(instring):
 *                 out.append(instring[lastE:s])
 *                 if t:             # <<<<<<<<<<<<<<
 *                     if isinstance(t, ParseResults):
 *                         out += t.asList()
 */
        }

        /* "Aspidites/_vendor/pyparsing_extension.py":1815
 *                     else:
 *                         out.append(t)
 *                 lastE = e             # <<<<<<<<<<<<<<
 *             out.append(instring[lastE:])
 *             out = [o for o in out if o]
 */
        __Pyx_INCREF(__pyx_v_e);
        __Pyx_DECREF_SET(__pyx_v_lastE, __pyx_v_e);

        /* "Aspidites/_vendor/pyparsing_extension.py":1806
 *         self.keepTabs = True
 *         try:
 *             for t, s, e in self.scanString(instring):             # <<<<<<<<<<<<<<
 *                 out.append(instring[lastE:s])
 *                 if t:
 */
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":1816
 *                         out.append(t)
 *                 lastE = e
 *             out.append(instring[lastE:])             # <<<<<<<<<<<<<<
 *             out = [o for o in out if o]
 *             return "".join(map(str, _flatten(out)))
 */
      __pyx_t_5 = __Pyx_PyObject_GetSlice(__pyx_v_instring, 0, 0, &__pyx_v_lastE, NULL, NULL, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1816, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_13 = __Pyx_PyObject_Append(__pyx_v_out, __pyx_t_5); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 1816, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":1817
 *                 lastE = e
 *             out.append(instring[lastE:])
 *             out = [o for o in out if o]             # <<<<<<<<<<<<<<
 *             return "".join(map(str, _flatten(out)))
 *         except ParseBaseException as exc:
 */
      __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1817, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (likely(PyList_CheckExact(__pyx_v_out)) || PyTuple_CheckExact(__pyx_v_out)) {
        __pyx_t_10 = __pyx_v_out; __Pyx_INCREF(__pyx_t_10); __pyx_t_7 = 0;
        __pyx_t_8 = NULL;
      } else {
        __pyx_t_7 = -1; __pyx_t_10 = PyObject_GetIter(__pyx_v_out); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1817, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_8 = Py_TYPE(__pyx_t_10)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1817, __pyx_L3_error)
      }
      for (;;) {
        if (likely(!__pyx_t_8)) {
          if (likely(PyList_CheckExact(__pyx_t_10))) {
            if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_10)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_1 = PyList_GET_ITEM(__pyx_t_10, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 1817, __pyx_L3_error)
            #else
            __pyx_t_1 = PySequence_ITEM(__pyx_t_10, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1817, __pyx_L3_error)
            __Pyx_GOTREF(__pyx_t_1);
            #endif
          } else {
            if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_10)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_10, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 1817, __pyx_L3_error)
            #else
            __pyx_t_1 = PySequence_ITEM(__pyx_t_10, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1817, __pyx_L3_error)
            __Pyx_GOTREF(__pyx_t_1);
            #endif
          }
        } else {
          __pyx_t_1 = __pyx_t_8(__pyx_t_10);
          if (unlikely(!__pyx_t_1)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 1817, __pyx_L3_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_1);
        }
        __Pyx_XDECREF_SET(__pyx_v_o, __pyx_t_1);
        __pyx_t_1 = 0;
        __pyx_t_14 = __Pyx_PyObject_IsTrue(__pyx_v_o); if (unlikely(__pyx_t_14 < 0)) __PYX_ERR(0, 1817, __pyx_L3_error)
        if (__pyx_t_14) {
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_5, (PyObject*)__pyx_v_o))) __PYX_ERR(0, 1817, __pyx_L3_error)
        }
      }
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF_SET(__pyx_v_out, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":1818
 *             out.append(instring[lastE:])
 *             out = [o for o in out if o]
 *             return "".join(map(str, _flatten(out)))             # <<<<<<<<<<<<<<
 *         except ParseBaseException as exc:
 *             if ParserElement.verbose_stacktrace:
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_flatten); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1818, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_10);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_10, function);
        }
      }
      __pyx_t_5 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_1, __pyx_v_out) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_v_out);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1818, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1818, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_INCREF(((PyObject *)(&PyString_Type)));
      __Pyx_GIVEREF(((PyObject *)(&PyString_Type)));
      PyTuple_SET_ITEM(__pyx_t_10, 0, ((PyObject *)(&PyString_Type)));
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_map, __pyx_t_10, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1818, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = __Pyx_PyString_Join(__pyx_kp_s__12, __pyx_t_5); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1818, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_r = __pyx_t_10;
      __pyx_t_10 = 0;
      goto __pyx_L7_try_return;

      /* "Aspidites/_vendor/pyparsing_extension.py":1805
 *         # keep string locs straight between transformString and scanString
 *         self.keepTabs = True
 *         try:             # <<<<<<<<<<<<<<
 *             for t, s, e in self.scanString(instring):
 *                 out.append(instring[lastE:s])
 */
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1819
 *             out = [o for o in out if o]
 *             return "".join(map(str, _flatten(out)))
 *         except ParseBaseException as exc:             # <<<<<<<<<<<<<<
 *             if ParserElement.verbose_stacktrace:
 *                 raise
 */
    __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_5, &__pyx_t_1);
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_ParseBaseException); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1819, __pyx_L5_except_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_16 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_10, __pyx_t_9);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_ErrRestore(__pyx_t_10, __pyx_t_5, __pyx_t_1);
    __pyx_t_10 = 0; __pyx_t_5 = 0; __pyx_t_1 = 0;
    if (__pyx_t_16) {
      __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.transformString", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_5, &__pyx_t_10) < 0) __PYX_ERR(0, 1819, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_5);
      __pyx_v_exc = __pyx_t_5;

      /* "Aspidites/_vendor/pyparsing_extension.py":1820
 *             return "".join(map(str, _flatten(out)))
 *         except ParseBaseException as exc:
 *             if ParserElement.verbose_stacktrace:             # <<<<<<<<<<<<<<
 *                 raise
 *             else:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1820, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_verbose_stacktrace); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1820, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_14 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_14 < 0)) __PYX_ERR(0, 1820, __pyx_L5_except_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(__pyx_t_14)) {

        /* "Aspidites/_vendor/pyparsing_extension.py":1821
 *         except ParseBaseException as exc:
 *             if ParserElement.verbose_stacktrace:
 *                 raise             # <<<<<<<<<<<<<<
 *             else:
 *                 # catch and re-raise exception from here, clearing out pyparsing internal stack trace
 */
        __Pyx_GIVEREF(__pyx_t_1);
        __Pyx_GIVEREF(__pyx_t_5);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_ErrRestoreWithState(__pyx_t_1, __pyx_t_5, __pyx_t_10);
        __pyx_t_1 = 0; __pyx_t_5 = 0; __pyx_t_10 = 0; 
        __PYX_ERR(0, 1821, __pyx_L5_except_error)

        /* "Aspidites/_vendor/pyparsing_extension.py":1820
 *             return "".join(map(str, _flatten(out)))
 *         except ParseBaseException as exc:
 *             if ParserElement.verbose_stacktrace:             # <<<<<<<<<<<<<<
 *                 raise
 *             else:
 */
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":1824
 *             else:
 *                 # catch and re-raise exception from here, clearing out pyparsing internal stack trace
 *                 if getattr(exc, '__traceback__', None) is not None:             # <<<<<<<<<<<<<<
 *                     exc.__traceback__ = self._trim_traceback(exc.__traceback__)
 *                 raise exc
 */
      /*else*/ {
        __pyx_t_6 = __Pyx_GetAttr3(__pyx_v_exc, __pyx_n_s_traceback, Py_None); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1824, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_14 = (__pyx_t_6 != Py_None);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_15 = (__pyx_t_14 != 0);
        if (__pyx_t_15) {

          /* "Aspidites/_vendor/pyparsing_extension.py":1825
 *                 # catch and re-raise exception from here, clearing out pyparsing internal stack trace
 *                 if getattr(exc, '__traceback__', None) is not None:
 *                     exc.__traceback__ = self._trim_traceback(exc.__traceback__)             # <<<<<<<<<<<<<<
 *                 raise exc
 * 
 */
          __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_trim_traceback); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1825, __pyx_L5_except_error)
          __Pyx_GOTREF(__pyx_t_9);
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_exc, __pyx_n_s_traceback); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1825, __pyx_L5_except_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_17 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
            __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_9);
            if (likely(__pyx_t_17)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
              __Pyx_INCREF(__pyx_t_17);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_9, function);
            }
          }
          __pyx_t_6 = (__pyx_t_17) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_17, __pyx_t_11) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_11);
          __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1825, __pyx_L5_except_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_exc, __pyx_n_s_traceback, __pyx_t_6) < 0) __PYX_ERR(0, 1825, __pyx_L5_except_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":1824
 *             else:
 *                 # catch and re-raise exception from here, clearing out pyparsing internal stack trace
 *                 if getattr(exc, '__traceback__', None) is not None:             # <<<<<<<<<<<<<<
 *                     exc.__traceback__ = self._trim_traceback(exc.__traceback__)
 *                 raise exc
 */
        }

        /* "Aspidites/_vendor/pyparsing_extension.py":1826
 *                 if getattr(exc, '__traceback__', None) is not None:
 *                     exc.__traceback__ = self._trim_traceback(exc.__traceback__)
 *                 raise exc             # <<<<<<<<<<<<<<
 * 
 *     def searchString(self, instring, maxMatches=sys.maxsize):
 */
        __Pyx_Raise(__pyx_v_exc, 0, 0, 0);
        __PYX_ERR(0, 1826, __pyx_L5_except_error)
      }
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "Aspidites/_vendor/pyparsing_extension.py":1805
 *         # keep string locs straight between transformString and scanString
 *         self.keepTabs = True
 *         try:             # <<<<<<<<<<<<<<
 *             for t, s, e in self.scanString(instring):
 *                 out.append(instring[lastE:s])
 */
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L0;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1780
 *                 raise exc
 * 
 *     def transformString(self, instring):             # <<<<<<<<<<<<<<
 *         """
 *         Extension to :class:`scanString`, to modify matching text with modified tokens that may
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.transformString", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_out);
  __Pyx_XDECREF(__pyx_v_lastE);
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_XDECREF(__pyx_v_s);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XDECREF(__pyx_v_o);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1828
 *                 raise exc
 * 
 *     def searchString(self, instring, maxMatches=sys.maxsize):             # <<<<<<<<<<<<<<
 *         """
 *         Another extension to :class:`scanString`, simplifying the access to the tokens found
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_34__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__defaults__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults3, __pyx_self)->__pyx_arg_maxMatches);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults3, __pyx_self)->__pyx_arg_maxMatches);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __Pyx_CyFunction_Defaults(__pyx_defaults3, __pyx_self)->__pyx_arg_maxMatches);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_2, 1, Py_None);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_54searchString(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_53searchString[] = "\n        Another extension to :class:`scanString`, simplifying the access to the tokens found\n        to match the given parse expression.  May be called with optional\n        ``maxMatches`` argument, to clip searching after 'n' matches are found.\n\n        Example::\n\n            # a capitalized word starts with an uppercase letter, followed by zero or more lowercase letters\n            cap_word = Word(alphas.upper(), alphas.lower())\n\n            print(cap_word.searchString(\"More than Iron, more than Lead, more than Gold I need Electricity\"))\n\n            # the sum() builtin can be used to merge results into a single ParseResults object\n            print(sum(cap_word.searchString(\"More than Iron, more than Lead, more than Gold I need Electricity\")))\n\n        prints::\n\n            [['More'], ['Iron'], ['Lead'], ['Gold'], ['I'], ['Electricity']]\n            ['More', 'Iron', 'Lead', 'Gold', 'I', 'Electricity']\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_54searchString = {"searchString", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_54searchString, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_53searchString};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_54searchString(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_maxMatches = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("searchString (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_maxMatches,0};
    PyObject* values[3] = {0,0,0};
    __pyx_defaults3 *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults3, __pyx_self);
    values[2] = __pyx_dynamic_args->__pyx_arg_maxMatches;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("searchString", 0, 2, 3, 1); __PYX_ERR(0, 1828, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_maxMatches);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "searchString") < 0)) __PYX_ERR(0, 1828, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_maxMatches = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("searchString", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1828, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.searchString", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_53searchString(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_maxMatches);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_53searchString(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_maxMatches) {
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_v_t = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_s = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  Py_ssize_t __pyx_t_12;
  PyObject *(*__pyx_t_13)(PyObject *);
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *(*__pyx_t_16)(PyObject *);
  int __pyx_t_17;
  int __pyx_t_18;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("searchString", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":1849
 *             ['More', 'Iron', 'Lead', 'Gold', 'I', 'Electricity']
 *         """
 *         try:             # <<<<<<<<<<<<<<
 *             return ParseResults([t for t, s, e in self.scanString(instring, maxMatches)])
 *         except ParseBaseException as exc:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "Aspidites/_vendor/pyparsing_extension.py":1850
 *         """
 *         try:
 *             return ParseResults([t for t, s, e in self.scanString(instring, maxMatches)])             # <<<<<<<<<<<<<<
 *         except ParseBaseException as exc:
 *             if ParserElement.verbose_stacktrace:
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_ParseResults); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1850, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1850, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_scanString); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1850, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = NULL;
      __pyx_t_10 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
          __pyx_t_10 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_8)) {
        PyObject *__pyx_temp[3] = {__pyx_t_9, __pyx_v_instring, __pyx_v_maxMatches};
        __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1850, __pyx_L3_error)
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_GOTREF(__pyx_t_7);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
        PyObject *__pyx_temp[3] = {__pyx_t_9, __pyx_v_instring, __pyx_v_maxMatches};
        __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1850, __pyx_L3_error)
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_GOTREF(__pyx_t_7);
      } else
      #endif
      {
        __pyx_t_11 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1850, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_11);
        if (__pyx_t_9) {
          __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_9); __pyx_t_9 = NULL;
        }
        __Pyx_INCREF(__pyx_v_instring);
        __Pyx_GIVEREF(__pyx_v_instring);
        PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_10, __pyx_v_instring);
        __Pyx_INCREF(__pyx_v_maxMatches);
        __Pyx_GIVEREF(__pyx_v_maxMatches);
        PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_10, __pyx_v_maxMatches);
        __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_11, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1850, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      }
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (likely(PyList_CheckExact(__pyx_t_7)) || PyTuple_CheckExact(__pyx_t_7)) {
        __pyx_t_8 = __pyx_t_7; __Pyx_INCREF(__pyx_t_8); __pyx_t_12 = 0;
        __pyx_t_13 = NULL;
      } else {
        __pyx_t_12 = -1; __pyx_t_8 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1850, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_13 = Py_TYPE(__pyx_t_8)->tp_iternext; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1850, __pyx_L3_error)
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      for (;;) {
        if (likely(!__pyx_t_13)) {
          if (likely(PyList_CheckExact(__pyx_t_8))) {
            if (__pyx_t_12 >= PyList_GET_SIZE(__pyx_t_8)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_7 = PyList_GET_ITEM(__pyx_t_8, __pyx_t_12); __Pyx_INCREF(__pyx_t_7); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 1850, __pyx_L3_error)
            #else
            __pyx_t_7 = PySequence_ITEM(__pyx_t_8, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1850, __pyx_L3_error)
            __Pyx_GOTREF(__pyx_t_7);
            #endif
          } else {
            if (__pyx_t_12 >= PyTuple_GET_SIZE(__pyx_t_8)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_8, __pyx_t_12); __Pyx_INCREF(__pyx_t_7); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 1850, __pyx_L3_error)
            #else
            __pyx_t_7 = PySequence_ITEM(__pyx_t_8, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1850, __pyx_L3_error)
            __Pyx_GOTREF(__pyx_t_7);
            #endif
          }
        } else {
          __pyx_t_7 = __pyx_t_13(__pyx_t_8);
          if (unlikely(!__pyx_t_7)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 1850, __pyx_L3_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_7);
        }
        if ((likely(PyTuple_CheckExact(__pyx_t_7))) || (PyList_CheckExact(__pyx_t_7))) {
          PyObject* sequence = __pyx_t_7;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 3)) {
            if (size > 3) __Pyx_RaiseTooManyValuesError(3);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 1850, __pyx_L3_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_11 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
            __pyx_t_14 = PyTuple_GET_ITEM(sequence, 2); 
          } else {
            __pyx_t_11 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
            __pyx_t_14 = PyList_GET_ITEM(sequence, 2); 
          }
          __Pyx_INCREF(__pyx_t_11);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_14);
          #else
          __pyx_t_11 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1850, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1850, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_9);
          __pyx_t_14 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1850, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_14);
          #endif
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else {
          Py_ssize_t index = -1;
          __pyx_t_15 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1850, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_16 = Py_TYPE(__pyx_t_15)->tp_iternext;
          index = 0; __pyx_t_11 = __pyx_t_16(__pyx_t_15); if (unlikely(!__pyx_t_11)) goto __pyx_L11_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_11);
          index = 1; __pyx_t_9 = __pyx_t_16(__pyx_t_15); if (unlikely(!__pyx_t_9)) goto __pyx_L11_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_9);
          index = 2; __pyx_t_14 = __pyx_t_16(__pyx_t_15); if (unlikely(!__pyx_t_14)) goto __pyx_L11_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_14);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_16(__pyx_t_15), 3) < 0) __PYX_ERR(0, 1850, __pyx_L3_error)
          __pyx_t_16 = NULL;
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          goto __pyx_L12_unpacking_done;
          __pyx_L11_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          __pyx_t_16 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 1850, __pyx_L3_error)
          __pyx_L12_unpacking_done:;
        }
        __Pyx_XDECREF_SET(__pyx_v_t, __pyx_t_11);
        __pyx_t_11 = 0;
        __Pyx_XDECREF_SET(__pyx_v_s, __pyx_t_9);
        __pyx_t_9 = 0;
        __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_14);
        __pyx_t_14 = 0;
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_6, (PyObject*)__pyx_v_t))) __PYX_ERR(0, 1850, __pyx_L3_error)
      }
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_4 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1850, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_r = __pyx_t_4;
      __pyx_t_4 = 0;
      goto __pyx_L7_try_return;

      /* "Aspidites/_vendor/pyparsing_extension.py":1849
 *             ['More', 'Iron', 'Lead', 'Gold', 'I', 'Electricity']
 *         """
 *         try:             # <<<<<<<<<<<<<<
 *             return ParseResults([t for t, s, e in self.scanString(instring, maxMatches)])
 *         except ParseBaseException as exc:
 */
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1851
 *         try:
 *             return ParseResults([t for t, s, e in self.scanString(instring, maxMatches)])
 *         except ParseBaseException as exc:             # <<<<<<<<<<<<<<
 *             if ParserElement.verbose_stacktrace:
 *                 raise
 */
    __Pyx_ErrFetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_ParseBaseException); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1851, __pyx_L5_except_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_10 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_4, __pyx_t_8);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_ErrRestore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
    __pyx_t_4 = 0; __pyx_t_5 = 0; __pyx_t_6 = 0;
    if (__pyx_t_10) {
      __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.searchString", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4) < 0) __PYX_ERR(0, 1851, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __pyx_v_exc = __pyx_t_5;

      /* "Aspidites/_vendor/pyparsing_extension.py":1852
 *             return ParseResults([t for t, s, e in self.scanString(instring, maxMatches)])
 *         except ParseBaseException as exc:
 *             if ParserElement.verbose_stacktrace:             # <<<<<<<<<<<<<<
 *                 raise
 *             else:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1852, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_verbose_stacktrace); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1852, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_17 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_17 < 0)) __PYX_ERR(0, 1852, __pyx_L5_except_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(__pyx_t_17)) {

        /* "Aspidites/_vendor/pyparsing_extension.py":1853
 *         except ParseBaseException as exc:
 *             if ParserElement.verbose_stacktrace:
 *                 raise             # <<<<<<<<<<<<<<
 *             else:
 *                 # catch and re-raise exception from here, clearing out pyparsing internal stack trace
 */
        __Pyx_GIVEREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5);
        __Pyx_XGIVEREF(__pyx_t_4);
        __Pyx_ErrRestoreWithState(__pyx_t_6, __pyx_t_5, __pyx_t_4);
        __pyx_t_6 = 0; __pyx_t_5 = 0; __pyx_t_4 = 0; 
        __PYX_ERR(0, 1853, __pyx_L5_except_error)

        /* "Aspidites/_vendor/pyparsing_extension.py":1852
 *             return ParseResults([t for t, s, e in self.scanString(instring, maxMatches)])
 *         except ParseBaseException as exc:
 *             if ParserElement.verbose_stacktrace:             # <<<<<<<<<<<<<<
 *                 raise
 *             else:
 */
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":1856
 *             else:
 *                 # catch and re-raise exception from here, clearing out pyparsing internal stack trace
 *                 if getattr(exc, '__traceback__', None) is not None:             # <<<<<<<<<<<<<<
 *                     exc.__traceback__ = self._trim_traceback(exc.__traceback__)
 *                 raise exc
 */
      /*else*/ {
        __pyx_t_7 = __Pyx_GetAttr3(__pyx_v_exc, __pyx_n_s_traceback, Py_None); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1856, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_17 = (__pyx_t_7 != Py_None);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_18 = (__pyx_t_17 != 0);
        if (__pyx_t_18) {

          /* "Aspidites/_vendor/pyparsing_extension.py":1857
 *                 # catch and re-raise exception from here, clearing out pyparsing internal stack trace
 *                 if getattr(exc, '__traceback__', None) is not None:
 *                     exc.__traceback__ = self._trim_traceback(exc.__traceback__)             # <<<<<<<<<<<<<<
 *                 raise exc
 * 
 */
          __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_trim_traceback); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1857, __pyx_L5_except_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_exc, __pyx_n_s_traceback); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1857, __pyx_L5_except_error)
          __Pyx_GOTREF(__pyx_t_14);
          __pyx_t_9 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
            __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
            if (likely(__pyx_t_9)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
              __Pyx_INCREF(__pyx_t_9);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_8, function);
            }
          }
          __pyx_t_7 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_t_14) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_14);
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1857, __pyx_L5_except_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_exc, __pyx_n_s_traceback, __pyx_t_7) < 0) __PYX_ERR(0, 1857, __pyx_L5_except_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":1856
 *             else:
 *                 # catch and re-raise exception from here, clearing out pyparsing internal stack trace
 *                 if getattr(exc, '__traceback__', None) is not None:             # <<<<<<<<<<<<<<
 *                     exc.__traceback__ = self._trim_traceback(exc.__traceback__)
 *                 raise exc
 */
        }

        /* "Aspidites/_vendor/pyparsing_extension.py":1858
 *                 if getattr(exc, '__traceback__', None) is not None:
 *                     exc.__traceback__ = self._trim_traceback(exc.__traceback__)
 *                 raise exc             # <<<<<<<<<<<<<<
 * 
 *     def split(self, instring, maxsplit=sys.maxsize, includeSeparators=False):
 */
        __Pyx_Raise(__pyx_v_exc, 0, 0, 0);
        __PYX_ERR(0, 1858, __pyx_L5_except_error)
      }
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "Aspidites/_vendor/pyparsing_extension.py":1849
 *             ['More', 'Iron', 'Lead', 'Gold', 'I', 'Electricity']
 *         """
 *         try:             # <<<<<<<<<<<<<<
 *             return ParseResults([t for t, s, e in self.scanString(instring, maxMatches)])
 *         except ParseBaseException as exc:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1828
 *                 raise exc
 * 
 *     def searchString(self, instring, maxMatches=sys.maxsize):             # <<<<<<<<<<<<<<
 *         """
 *         Another extension to :class:`scanString`, simplifying the access to the tokens found
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.searchString", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_XDECREF(__pyx_v_s);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_57generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/pyparsing_extension.py":1860
 *                 raise exc
 * 
 *     def split(self, instring, maxsplit=sys.maxsize, includeSeparators=False):             # <<<<<<<<<<<<<<
 *         """
 *         Generator method to split a string using the given expression as a separator.
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_36__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__defaults__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1860, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults4, __pyx_self)->__pyx_arg_maxsplit);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults4, __pyx_self)->__pyx_arg_maxsplit);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __Pyx_CyFunction_Defaults(__pyx_defaults4, __pyx_self)->__pyx_arg_maxsplit);
  __Pyx_INCREF(((PyObject *)Py_False));
  __Pyx_GIVEREF(((PyObject *)Py_False));
  PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)Py_False));
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1860, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_2, 1, Py_None);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_56split(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_55split[] = "\n        Generator method to split a string using the given expression as a separator.\n        May be called with optional ``maxsplit`` argument, to limit the number of splits;\n        and the optional ``includeSeparators`` argument (default= ``False``), if the separating\n        matching text should be included in the split results.\n\n        Example::\n\n            punc = oneOf(list(\".,;:/-!?\"))\n            print(list(punc.split(\"This, this?, this sentence, is badly punctuated!\")))\n\n        prints::\n\n            ['This', ' this', '', ' this sentence', ' is badly punctuated', '']\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_56split = {"split", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_56split, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_55split};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_56split(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_maxsplit = 0;
  PyObject *__pyx_v_includeSeparators = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("split (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_maxsplit,&__pyx_n_s_includeSeparators,0};
    PyObject* values[4] = {0,0,0,0};
    __pyx_defaults4 *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults4, __pyx_self);
    values[2] = __pyx_dynamic_args->__pyx_arg_maxsplit;
    values[3] = ((PyObject *)((PyObject *)Py_False));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("split", 0, 2, 4, 1); __PYX_ERR(0, 1860, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_maxsplit);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_includeSeparators);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "split") < 0)) __PYX_ERR(0, 1860, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_maxsplit = values[2];
    __pyx_v_includeSeparators = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("split", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1860, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.split", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_55split(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_maxsplit, __pyx_v_includeSeparators);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_55split(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_maxsplit, PyObject *__pyx_v_includeSeparators) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("split", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1860, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_instring = __pyx_v_instring;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_instring);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_instring);
  __pyx_cur_scope->__pyx_v_maxsplit = __pyx_v_maxsplit;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_maxsplit);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_maxsplit);
  __pyx_cur_scope->__pyx_v_includeSeparators = __pyx_v_includeSeparators;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_includeSeparators);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_includeSeparators);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_57generator1, __pyx_codeobj__64, (PyObject *) __pyx_cur_scope, __pyx_n_s_split, __pyx_n_s_ParserElement_split, __pyx_n_s_Aspidites__vendor_pyparsing_exte); if (unlikely(!gen)) __PYX_ERR(0, 1860, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.split", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_57generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *(*__pyx_t_9)(PyObject *);
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("split", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L8_resume_from_yield;
    case 2: goto __pyx_L10_resume_from_yield;
    case 3: goto __pyx_L11_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1860, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1876
 *             ['This', ' this', '', ' this sentence', ' is badly punctuated', '']
 *         """
 *         splits = 0             # <<<<<<<<<<<<<<
 *         last = 0
 *         for t, s, e in self.scanString(instring, maxMatches=maxsplit):
 */
  __pyx_cur_scope->__pyx_v_splits = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1877
 *         """
 *         splits = 0
 *         last = 0             # <<<<<<<<<<<<<<
 *         for t, s, e in self.scanString(instring, maxMatches=maxsplit):
 *             yield instring[last:s]
 */
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __pyx_cur_scope->__pyx_v_last = __pyx_int_0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1878
 *         splits = 0
 *         last = 0
 *         for t, s, e in self.scanString(instring, maxMatches=maxsplit):             # <<<<<<<<<<<<<<
 *             yield instring[last:s]
 *             if includeSeparators:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_scanString); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1878, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1878, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_instring);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_instring);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_cur_scope->__pyx_v_instring);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1878, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_maxMatches, __pyx_cur_scope->__pyx_v_maxsplit) < 0) __PYX_ERR(0, 1878, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1878, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
    __pyx_t_3 = __pyx_t_4; __Pyx_INCREF(__pyx_t_3); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1878, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1878, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_5); __Pyx_INCREF(__pyx_t_4); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 1878, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_3, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1878, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_5); __Pyx_INCREF(__pyx_t_4); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 1878, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_3, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1878, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_6(__pyx_t_3);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1878, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
      PyObject* sequence = __pyx_t_4;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 3)) {
        if (size > 3) __Pyx_RaiseTooManyValuesError(3);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 1878, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_1 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_7 = PyTuple_GET_ITEM(sequence, 2); 
      } else {
        __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_1 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_7 = PyList_GET_ITEM(sequence, 2); 
      }
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_7);
      #else
      __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1878, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1878, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1878, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      #endif
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_8 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1878, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_9 = Py_TYPE(__pyx_t_8)->tp_iternext;
      index = 0; __pyx_t_2 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_2)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_2);
      index = 1; __pyx_t_1 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_1)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_1);
      index = 2; __pyx_t_7 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_7)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_7);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_9(__pyx_t_8), 3) < 0) __PYX_ERR(0, 1878, __pyx_L1_error)
      __pyx_t_9 = NULL;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L7_unpacking_done;
      __pyx_L6_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_9 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 1878, __pyx_L1_error)
      __pyx_L7_unpacking_done:;
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_t);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_t, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_s);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_s, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_e);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_e, __pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_7);
    __pyx_t_7 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1879
 *         last = 0
 *         for t, s, e in self.scanString(instring, maxMatches=maxsplit):
 *             yield instring[last:s]             # <<<<<<<<<<<<<<
 *             if includeSeparators:
 *                 yield t[0]
 */
    __pyx_t_4 = __Pyx_PyObject_GetSlice(__pyx_cur_scope->__pyx_v_instring, 0, 0, &__pyx_cur_scope->__pyx_v_last, &__pyx_cur_scope->__pyx_v_s, NULL, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1879, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    __Pyx_XGIVEREF(__pyx_t_3);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_3;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_5;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_6;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L8_resume_from_yield:;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_3);
    __pyx_t_5 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_6 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1879, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":1880
 *         for t, s, e in self.scanString(instring, maxMatches=maxsplit):
 *             yield instring[last:s]
 *             if includeSeparators:             # <<<<<<<<<<<<<<
 *                 yield t[0]
 *             last = e
 */
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_includeSeparators); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 1880, __pyx_L1_error)
    if (__pyx_t_10) {

      /* "Aspidites/_vendor/pyparsing_extension.py":1881
 *             yield instring[last:s]
 *             if includeSeparators:
 *                 yield t[0]             # <<<<<<<<<<<<<<
 *             last = e
 *         yield instring[last:]
 */
      __pyx_t_4 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_t, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1881, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_r = __pyx_t_4;
      __pyx_t_4 = 0;
      __Pyx_XGIVEREF(__pyx_t_3);
      __pyx_cur_scope->__pyx_t_0 = __pyx_t_3;
      __pyx_cur_scope->__pyx_t_1 = __pyx_t_5;
      __pyx_cur_scope->__pyx_t_2 = __pyx_t_6;
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
      /* return from generator, yielding value */
      __pyx_generator->resume_label = 2;
      return __pyx_r;
      __pyx_L10_resume_from_yield:;
      __pyx_t_3 = __pyx_cur_scope->__pyx_t_0;
      __pyx_cur_scope->__pyx_t_0 = 0;
      __Pyx_XGOTREF(__pyx_t_3);
      __pyx_t_5 = __pyx_cur_scope->__pyx_t_1;
      __pyx_t_6 = __pyx_cur_scope->__pyx_t_2;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1881, __pyx_L1_error)

      /* "Aspidites/_vendor/pyparsing_extension.py":1880
 *         for t, s, e in self.scanString(instring, maxMatches=maxsplit):
 *             yield instring[last:s]
 *             if includeSeparators:             # <<<<<<<<<<<<<<
 *                 yield t[0]
 *             last = e
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":1882
 *             if includeSeparators:
 *                 yield t[0]
 *             last = e             # <<<<<<<<<<<<<<
 *         yield instring[last:]
 * 
 */
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_e);
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_last);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_last, __pyx_cur_scope->__pyx_v_e);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_e);

    /* "Aspidites/_vendor/pyparsing_extension.py":1878
 *         splits = 0
 *         last = 0
 *         for t, s, e in self.scanString(instring, maxMatches=maxsplit):             # <<<<<<<<<<<<<<
 *             yield instring[last:s]
 *             if includeSeparators:
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1883
 *                 yield t[0]
 *             last = e
 *         yield instring[last:]             # <<<<<<<<<<<<<<
 * 
 *     def __add__(self, other):
 */
  __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_cur_scope->__pyx_v_instring, 0, 0, &__pyx_cur_scope->__pyx_v_last, NULL, NULL, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1883, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  /* return from generator, yielding value */
  __pyx_generator->resume_label = 3;
  return __pyx_r;
  __pyx_L11_resume_from_yield:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1883, __pyx_L1_error)
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "Aspidites/_vendor/pyparsing_extension.py":1860
 *                 raise exc
 * 
 *     def split(self, instring, maxsplit=sys.maxsize, includeSeparators=False):             # <<<<<<<<<<<<<<
 *         """
 *         Generator method to split a string using the given expression as a separator.
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("split", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1885
 *         yield instring[last:]
 * 
 *     def __add__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of + operator - returns :class:`And`. Adding strings to a ParserElement
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_59__add__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_58__add__[] = "\n        Implementation of + operator - returns :class:`And`. Adding strings to a ParserElement\n        converts them to :class:`Literal`s by default.\n\n        Example::\n\n            greet = Word(alphas) + \",\" + Word(alphas) + \"!\"\n            hello = \"Hello, World!\"\n            print (hello, \"->\", greet.parseString(hello))\n\n        prints::\n\n            Hello, World! -> ['Hello', ',', 'World', '!']\n\n        ``...`` may be used as a parse expression as a short form of :class:`SkipTo`.\n\n            Literal('start') + ... + Literal('end')\n\n        is equivalent to:\n\n            Literal('start') + SkipTo('end')(\"_skipped*\") + Literal('end')\n\n        Note that the skipped text is returned with '_skipped' as a results name,\n        and to support having multiple skips in the same parser, the value returned is\n        a list of all skipped text.\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_59__add__ = {"__add__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_59__add__, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_58__add__};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_59__add__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_other = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__add__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_other,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_other)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__add__", 1, 2, 2, 1); __PYX_ERR(0, 1885, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__add__") < 0)) __PYX_ERR(0, 1885, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_other = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__add__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1885, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__add__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_58__add__(__pyx_self, __pyx_v_self, __pyx_v_other);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_58__add__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_v__PendingSkip = NULL;
  PyObject *__pyx_v_And = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__add__", 0);
  __Pyx_INCREF(__pyx_v_other);

  /* "Aspidites/_vendor/pyparsing_extension.py":1912
 *         a list of all skipped text.
 *         """
 *         from .pyparsing import _PendingSkip, And             # <<<<<<<<<<<<<<
 *         if other is Ellipsis:
 *             return _PendingSkip(self)
 */
  __pyx_t_1 = PyList_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1912, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_PendingSkip);
  __Pyx_GIVEREF(__pyx_n_s_PendingSkip);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PendingSkip);
  __Pyx_INCREF(__pyx_n_s_And);
  __Pyx_GIVEREF(__pyx_n_s_And);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_And);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_pyparsing, __pyx_t_1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1912, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_PendingSkip); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1912, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v__PendingSkip = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_And); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1912, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_And = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1913
 *         """
 *         from .pyparsing import _PendingSkip, And
 *         if other is Ellipsis:             # <<<<<<<<<<<<<<
 *             return _PendingSkip(self)
 * 
 */
  __pyx_t_3 = (__pyx_v_other == __pyx_builtin_Ellipsis);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":1914
 *         from .pyparsing import _PendingSkip, And
 *         if other is Ellipsis:
 *             return _PendingSkip(self)             # <<<<<<<<<<<<<<
 * 
 *         if isinstance(other, str):
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v__PendingSkip);
    __pyx_t_1 = __pyx_v__PendingSkip; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v_self) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_self);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1914, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1913
 *         """
 *         from .pyparsing import _PendingSkip, And
 *         if other is Ellipsis:             # <<<<<<<<<<<<<<
 *             return _PendingSkip(self)
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1916
 *             return _PendingSkip(self)
 * 
 *         if isinstance(other, str):             # <<<<<<<<<<<<<<
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):
 */
  __pyx_t_4 = PyString_Check(__pyx_v_other); 
  __pyx_t_3 = (__pyx_t_4 != 0);
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/pyparsing_extension.py":1917
 * 
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)             # <<<<<<<<<<<<<<
 *         if not isinstance(other, ParserElement):
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_literalStringClass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1917, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v_other) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_other);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1917, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_other, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1916
 *             return _PendingSkip(self)
 * 
 *         if isinstance(other, str):             # <<<<<<<<<<<<<<
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1918
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):             # <<<<<<<<<<<<<<
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 *                           SyntaxWarning, stacklevel=2)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1918, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_other, __pyx_t_2); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 1918, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = ((!(__pyx_t_3 != 0)) != 0);
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":1919
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),             # <<<<<<<<<<<<<<
 *                           SyntaxWarning, stacklevel=2)
 *             return None
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_warnings); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1919, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_warn); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1919, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Cannot_combine_element_of_type_s, ((PyObject *)Py_TYPE(__pyx_v_other))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1919, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "Aspidites/_vendor/pyparsing_extension.py":1920
 *         if not isinstance(other, ParserElement):
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 *                           SyntaxWarning, stacklevel=2)             # <<<<<<<<<<<<<<
 *             return None
 *         return And([self, other])
 */
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1919, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
    __Pyx_INCREF(__pyx_builtin_SyntaxWarning);
    __Pyx_GIVEREF(__pyx_builtin_SyntaxWarning);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_builtin_SyntaxWarning);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1920, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_stacklevel, __pyx_int_2) < 0) __PYX_ERR(0, 1920, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":1919
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),             # <<<<<<<<<<<<<<
 *                           SyntaxWarning, stacklevel=2)
 *             return None
 */
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_5, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1919, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1921
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 *                           SyntaxWarning, stacklevel=2)
 *             return None             # <<<<<<<<<<<<<<
 *         return And([self, other])
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1918
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):             # <<<<<<<<<<<<<<
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 *                           SyntaxWarning, stacklevel=2)
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1922
 *                           SyntaxWarning, stacklevel=2)
 *             return None
 *         return And([self, other])             # <<<<<<<<<<<<<<
 * 
 *     def __radd__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyList_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1922, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_v_self);
  __Pyx_INCREF(__pyx_v_other);
  __Pyx_GIVEREF(__pyx_v_other);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_v_other);
  __Pyx_INCREF(__pyx_v_And);
  __pyx_t_5 = __pyx_v_And; __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_6 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_1, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1922, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1885
 *         yield instring[last:]
 * 
 *     def __add__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of + operator - returns :class:`And`. Adding strings to a ParserElement
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__add__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v__PendingSkip);
  __Pyx_XDECREF(__pyx_v_And);
  __Pyx_XDECREF(__pyx_v_other);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1924
 *         return And([self, other])
 * 
 *     def __radd__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of + operator when left operand is not a :class:`ParserElement`
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_61__radd__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_60__radd__[] = "\n        Implementation of + operator when left operand is not a :class:`ParserElement`\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_61__radd__ = {"__radd__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_61__radd__, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_60__radd__};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_61__radd__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_other = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__radd__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_other,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_other)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__radd__", 1, 2, 2, 1); __PYX_ERR(0, 1924, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__radd__") < 0)) __PYX_ERR(0, 1924, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_other = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__radd__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1924, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__radd__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_60__radd__(__pyx_self, __pyx_v_self, __pyx_v_other);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_60__radd__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_v_SkipTo = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__radd__", 0);
  __Pyx_INCREF(__pyx_v_other);

  /* "Aspidites/_vendor/pyparsing_extension.py":1928
 *         Implementation of + operator when left operand is not a :class:`ParserElement`
 *         """
 *         from .pyparsing import SkipTo             # <<<<<<<<<<<<<<
 *         if other is Ellipsis:
 *             return SkipTo(self)("_skipped*") + self
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1928, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_SkipTo);
  __Pyx_GIVEREF(__pyx_n_s_SkipTo);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_SkipTo);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_pyparsing, __pyx_t_1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1928, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_SkipTo); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1928, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_SkipTo = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1929
 *         """
 *         from .pyparsing import SkipTo
 *         if other is Ellipsis:             # <<<<<<<<<<<<<<
 *             return SkipTo(self)("_skipped*") + self
 * 
 */
  __pyx_t_3 = (__pyx_v_other == __pyx_builtin_Ellipsis);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":1930
 *         from .pyparsing import SkipTo
 *         if other is Ellipsis:
 *             return SkipTo(self)("_skipped*") + self             # <<<<<<<<<<<<<<
 * 
 *         if isinstance(other, str):
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_SkipTo);
    __pyx_t_5 = __pyx_v_SkipTo; __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_v_self) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_self);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1930, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_kp_s_skipped) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_kp_s_skipped);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1930, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_Add(__pyx_t_2, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1930, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1929
 *         """
 *         from .pyparsing import SkipTo
 *         if other is Ellipsis:             # <<<<<<<<<<<<<<
 *             return SkipTo(self)("_skipped*") + self
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1932
 *             return SkipTo(self)("_skipped*") + self
 * 
 *         if isinstance(other, str):             # <<<<<<<<<<<<<<
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):
 */
  __pyx_t_4 = PyString_Check(__pyx_v_other); 
  __pyx_t_3 = (__pyx_t_4 != 0);
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/pyparsing_extension.py":1933
 * 
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)             # <<<<<<<<<<<<<<
 *         if not isinstance(other, ParserElement):
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_literalStringClass); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1933, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_v_other) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_other);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1933, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_other, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1932
 *             return SkipTo(self)("_skipped*") + self
 * 
 *         if isinstance(other, str):             # <<<<<<<<<<<<<<
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1934
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):             # <<<<<<<<<<<<<<
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 *                           SyntaxWarning, stacklevel=2)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1934, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_other, __pyx_t_1); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 1934, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = ((!(__pyx_t_3 != 0)) != 0);
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":1935
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),             # <<<<<<<<<<<<<<
 *                           SyntaxWarning, stacklevel=2)
 *             return None
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_warnings); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1935, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_warn); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1935, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Cannot_combine_element_of_type_s, ((PyObject *)Py_TYPE(__pyx_v_other))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1935, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "Aspidites/_vendor/pyparsing_extension.py":1936
 *         if not isinstance(other, ParserElement):
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 *                           SyntaxWarning, stacklevel=2)             # <<<<<<<<<<<<<<
 *             return None
 *         return other + self
 */
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1935, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
    __Pyx_INCREF(__pyx_builtin_SyntaxWarning);
    __Pyx_GIVEREF(__pyx_builtin_SyntaxWarning);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_builtin_SyntaxWarning);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1936, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_stacklevel, __pyx_int_2) < 0) __PYX_ERR(0, 1936, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":1935
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),             # <<<<<<<<<<<<<<
 *                           SyntaxWarning, stacklevel=2)
 *             return None
 */
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1935, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1937
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 *                           SyntaxWarning, stacklevel=2)
 *             return None             # <<<<<<<<<<<<<<
 *         return other + self
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1934
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):             # <<<<<<<<<<<<<<
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 *                           SyntaxWarning, stacklevel=2)
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1938
 *                           SyntaxWarning, stacklevel=2)
 *             return None
 *         return other + self             # <<<<<<<<<<<<<<
 * 
 *     def __sub__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_6 = PyNumber_Add(__pyx_v_other, __pyx_v_self); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1938, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1924
 *         return And([self, other])
 * 
 *     def __radd__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of + operator when left operand is not a :class:`ParserElement`
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__radd__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_SkipTo);
  __Pyx_XDECREF(__pyx_v_other);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1940
 *         return other + self
 * 
 *     def __sub__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of - operator, returns :class:`And` with error stop
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_63__sub__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_62__sub__[] = "\n        Implementation of - operator, returns :class:`And` with error stop\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_63__sub__ = {"__sub__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_63__sub__, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_62__sub__};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_63__sub__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_other = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__sub__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_other,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_other)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__sub__", 1, 2, 2, 1); __PYX_ERR(0, 1940, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__sub__") < 0)) __PYX_ERR(0, 1940, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_other = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__sub__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1940, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__sub__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_62__sub__(__pyx_self, __pyx_v_self, __pyx_v_other);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_62__sub__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_v_And = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__sub__", 0);
  __Pyx_INCREF(__pyx_v_other);

  /* "Aspidites/_vendor/pyparsing_extension.py":1944
 *         Implementation of - operator, returns :class:`And` with error stop
 *         """
 *         from .pyparsing import And             # <<<<<<<<<<<<<<
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1944, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_And);
  __Pyx_GIVEREF(__pyx_n_s_And);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_And);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_pyparsing, __pyx_t_1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1944, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_And); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1944, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_And = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1945
 *         """
 *         from .pyparsing import And
 *         if isinstance(other, str):             # <<<<<<<<<<<<<<
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):
 */
  __pyx_t_3 = PyString_Check(__pyx_v_other); 
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":1946
 *         from .pyparsing import And
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)             # <<<<<<<<<<<<<<
 *         if not isinstance(other, ParserElement):
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_literalStringClass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1946, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v_other) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_other);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1946, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_other, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1945
 *         """
 *         from .pyparsing import And
 *         if isinstance(other, str):             # <<<<<<<<<<<<<<
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1947
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):             # <<<<<<<<<<<<<<
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 *                           SyntaxWarning, stacklevel=2)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1947, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyObject_IsInstance(__pyx_v_other, __pyx_t_2); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 1947, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = ((!(__pyx_t_4 != 0)) != 0);
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/pyparsing_extension.py":1948
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),             # <<<<<<<<<<<<<<
 *                           SyntaxWarning, stacklevel=2)
 *             return None
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_warnings); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1948, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_warn); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1948, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Cannot_combine_element_of_type_s, ((PyObject *)Py_TYPE(__pyx_v_other))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1948, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "Aspidites/_vendor/pyparsing_extension.py":1949
 *         if not isinstance(other, ParserElement):
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 *                           SyntaxWarning, stacklevel=2)             # <<<<<<<<<<<<<<
 *             return None
 *         return self + And._ErrorStop() + other
 */
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1948, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
    __Pyx_INCREF(__pyx_builtin_SyntaxWarning);
    __Pyx_GIVEREF(__pyx_builtin_SyntaxWarning);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_builtin_SyntaxWarning);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1949, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_stacklevel, __pyx_int_2) < 0) __PYX_ERR(0, 1949, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":1948
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),             # <<<<<<<<<<<<<<
 *                           SyntaxWarning, stacklevel=2)
 *             return None
 */
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_5, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1948, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1950
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 *                           SyntaxWarning, stacklevel=2)
 *             return None             # <<<<<<<<<<<<<<
 *         return self + And._ErrorStop() + other
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1947
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):             # <<<<<<<<<<<<<<
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 *                           SyntaxWarning, stacklevel=2)
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1951
 *                           SyntaxWarning, stacklevel=2)
 *             return None
 *         return self + And._ErrorStop() + other             # <<<<<<<<<<<<<<
 * 
 *     def __rsub__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_And, __pyx_n_s_ErrorStop); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1951, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_6 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1951, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_Add(__pyx_v_self, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1951, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = PyNumber_Add(__pyx_t_2, __pyx_v_other); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1951, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1940
 *         return other + self
 * 
 *     def __sub__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of - operator, returns :class:`And` with error stop
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__sub__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_And);
  __Pyx_XDECREF(__pyx_v_other);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1953
 *         return self + And._ErrorStop() + other
 * 
 *     def __rsub__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of - operator when left operand is not a :class:`ParserElement`
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_65__rsub__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_64__rsub__[] = "\n        Implementation of - operator when left operand is not a :class:`ParserElement`\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_65__rsub__ = {"__rsub__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_65__rsub__, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_64__rsub__};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_65__rsub__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_other = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__rsub__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_other,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_other)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__rsub__", 1, 2, 2, 1); __PYX_ERR(0, 1953, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__rsub__") < 0)) __PYX_ERR(0, 1953, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_other = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__rsub__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1953, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__rsub__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_64__rsub__(__pyx_self, __pyx_v_self, __pyx_v_other);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_64__rsub__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__rsub__", 0);
  __Pyx_INCREF(__pyx_v_other);

  /* "Aspidites/_vendor/pyparsing_extension.py":1957
 *         Implementation of - operator when left operand is not a :class:`ParserElement`
 *         """
 *         if isinstance(other, str):             # <<<<<<<<<<<<<<
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):
 */
  __pyx_t_1 = PyString_Check(__pyx_v_other); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":1958
 *         """
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)             # <<<<<<<<<<<<<<
 *         if not isinstance(other, ParserElement):
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_literalStringClass); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1958, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_other) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_other);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1958, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_other, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1957
 *         Implementation of - operator when left operand is not a :class:`ParserElement`
 *         """
 *         if isinstance(other, str):             # <<<<<<<<<<<<<<
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1959
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):             # <<<<<<<<<<<<<<
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 *                           SyntaxWarning, stacklevel=2)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1959, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyObject_IsInstance(__pyx_v_other, __pyx_t_3); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 1959, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = ((!(__pyx_t_2 != 0)) != 0);
  if (__pyx_t_1) {

    /* "Aspidites/_vendor/pyparsing_extension.py":1960
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),             # <<<<<<<<<<<<<<
 *                           SyntaxWarning, stacklevel=2)
 *             return None
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_warnings); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1960, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_warn); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1960, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Cannot_combine_element_of_type_s, ((PyObject *)Py_TYPE(__pyx_v_other))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1960, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "Aspidites/_vendor/pyparsing_extension.py":1961
 *         if not isinstance(other, ParserElement):
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 *                           SyntaxWarning, stacklevel=2)             # <<<<<<<<<<<<<<
 *             return None
 *         return other - self
 */
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1960, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3);
    __Pyx_INCREF(__pyx_builtin_SyntaxWarning);
    __Pyx_GIVEREF(__pyx_builtin_SyntaxWarning);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_builtin_SyntaxWarning);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1961, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_stacklevel, __pyx_int_2) < 0) __PYX_ERR(0, 1961, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":1960
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),             # <<<<<<<<<<<<<<
 *                           SyntaxWarning, stacklevel=2)
 *             return None
 */
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1960, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1962
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 *                           SyntaxWarning, stacklevel=2)
 *             return None             # <<<<<<<<<<<<<<
 *         return other - self
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1959
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):             # <<<<<<<<<<<<<<
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 *                           SyntaxWarning, stacklevel=2)
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1963
 *                           SyntaxWarning, stacklevel=2)
 *             return None
 *         return other - self             # <<<<<<<<<<<<<<
 * 
 *     def __mul__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_6 = PyNumber_Subtract(__pyx_v_other, __pyx_v_self); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1963, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1953
 *         return self + And._ErrorStop() + other
 * 
 *     def __rsub__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of - operator when left operand is not a :class:`ParserElement`
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__rsub__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_other);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1965
 *         return other - self
 * 
 *     def __mul__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of * operator, allows use of ``expr * 3`` in place of
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_67__mul__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_66__mul__[] = "\n        Implementation of * operator, allows use of ``expr * 3`` in place of\n        ``expr + expr + expr``.  Expressions may also me multiplied by a 2-integer\n        tuple, similar to ``{min, max}`` multipliers in regular expressions.  Tuples\n        may also include ``None`` as in:\n         - ``expr*(n, None)`` or ``expr*(n, )`` is equivalent\n              to ``expr*n + ZeroOrMore(expr)``\n              (read as \"at least n instances of ``expr``\")\n         - ``expr*(None, n)`` is equivalent to ``expr*(0, n)``\n              (read as \"0 to n instances of ``expr``\")\n         - ``expr*(None, None)`` is equivalent to ``ZeroOrMore(expr)``\n         - ``expr*(1, None)`` is equivalent to ``OneOrMore(expr)``\n\n        Note that ``expr*(None, n)`` does not raise an exception if\n        more than n exprs exist in the input stream; that is,\n        ``expr*(None, n)`` does not enforce a maximum number of expr\n        occurrences.  If this behavior is desired, then write\n        ``expr*(None, n) + ~expr``\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_67__mul__ = {"__mul__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_67__mul__, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_66__mul__};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_67__mul__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_other = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__mul__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_other,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_other)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__mul__", 1, 2, 2, 1); __PYX_ERR(0, 1965, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__mul__") < 0)) __PYX_ERR(0, 1965, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_other = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__mul__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1965, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__mul__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_66__mul__(__pyx_self, __pyx_v_self, __pyx_v_other);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2022
 * 
 *         if optElements:
 *             def makeOptionalList(n):             # <<<<<<<<<<<<<<
 *                 from .pyparsing import Optional
 *                 if n > 1:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_7__mul___1makeOptionalList(PyObject *__pyx_self, PyObject *__pyx_v_n); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_7__mul___1makeOptionalList = {"makeOptionalList", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_7__mul___1makeOptionalList, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_7__mul___1makeOptionalList(PyObject *__pyx_self, PyObject *__pyx_v_n) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("makeOptionalList (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_7__mul___makeOptionalList(__pyx_self, ((PyObject *)__pyx_v_n));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_7__mul___makeOptionalList(PyObject *__pyx_self, PyObject *__pyx_v_n) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__ *__pyx_cur_scope;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__ *__pyx_outer_scope;
  PyObject *__pyx_v_Optional = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("makeOptionalList", 0);
  __pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__ *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "Aspidites/_vendor/pyparsing_extension.py":2023
 *         if optElements:
 *             def makeOptionalList(n):
 *                 from .pyparsing import Optional             # <<<<<<<<<<<<<<
 *                 if n > 1:
 *                     return Optional(self + makeOptionalList(n - 1))
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2023, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_Optional);
  __Pyx_GIVEREF(__pyx_n_s_Optional);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_Optional);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_pyparsing, __pyx_t_1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2023, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_Optional); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2023, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_Optional = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2024
 *             def makeOptionalList(n):
 *                 from .pyparsing import Optional
 *                 if n > 1:             # <<<<<<<<<<<<<<
 *                     return Optional(self + makeOptionalList(n - 1))
 *                 else:
 */
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_n, __pyx_int_1, Py_GT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2024, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 2024, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2025
 *                 from .pyparsing import Optional
 *                 if n > 1:
 *                     return Optional(self + makeOptionalList(n - 1))             # <<<<<<<<<<<<<<
 *                 else:
 *                     return Optional(self)
 */
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 2025, __pyx_L1_error) }
    __pyx_t_1 = __Pyx_PyInt_SubtractObjC(__pyx_v_n, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2025, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(!__pyx_cur_scope->__pyx_v_makeOptionalList)) { __Pyx_RaiseClosureNameError("makeOptionalList"); __PYX_ERR(0, 2025, __pyx_L1_error) }
    __pyx_t_4 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_7__mul___makeOptionalList(__pyx_cur_scope->__pyx_v_makeOptionalList, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2025, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_Add(__pyx_cur_scope->__pyx_v_self, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2025, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_INCREF(__pyx_v_Optional);
    __pyx_t_4 = __pyx_v_Optional; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2025, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2024
 *             def makeOptionalList(n):
 *                 from .pyparsing import Optional
 *                 if n > 1:             # <<<<<<<<<<<<<<
 *                     return Optional(self + makeOptionalList(n - 1))
 *                 else:
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2027
 *                     return Optional(self + makeOptionalList(n - 1))
 *                 else:
 *                     return Optional(self)             # <<<<<<<<<<<<<<
 * 
 *             if minElements:
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 2027, __pyx_L1_error) }
    __Pyx_INCREF(__pyx_v_Optional);
    __pyx_t_4 = __pyx_v_Optional; __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_1, __pyx_cur_scope->__pyx_v_self) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_cur_scope->__pyx_v_self);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2027, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2022
 * 
 *         if optElements:
 *             def makeOptionalList(n):             # <<<<<<<<<<<<<<
 *                 from .pyparsing import Optional
 *                 if n > 1:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__mul__.makeOptionalList", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_Optional);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_7__mul___4generator13(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/pyparsing_extension.py":1994
 *             minElements, optElements = other, 0
 *         elif isinstance(other, tuple):
 *             other = tuple(o if o is not Ellipsis else None for o in other)             # <<<<<<<<<<<<<<
 *             other = (other + (None, None))[:2]
 *             if other[0] is None:
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_7__mul___2genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1994, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__ *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_7__mul___4generator13, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_ParserElement___mul___locals_gen, __pyx_n_s_Aspidites__vendor_pyparsing_exte); if (unlikely(!gen)) __PYX_ERR(0, 1994, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__mul__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_7__mul___4generator13(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1994, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_other)) { __Pyx_RaiseClosureNameError("other"); __PYX_ERR(0, 1994, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_other)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_other)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_other; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_other); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1994, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1994, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 1994, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1994, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 1994, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1994, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1994, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_o);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_o, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_5 = (__pyx_cur_scope->__pyx_v_o != __pyx_builtin_Ellipsis);
    if ((__pyx_t_5 != 0)) {
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_o);
      __pyx_t_4 = __pyx_cur_scope->__pyx_v_o;
    } else {
      __Pyx_INCREF(Py_None);
      __pyx_t_4 = Py_None;
    }
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1994, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":1965
 *         return other - self
 * 
 *     def __mul__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of * operator, allows use of ``expr * 3`` in place of
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_66__mul__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__ *__pyx_cur_scope;
  PyObject *__pyx_v_ZeroOrMore = NULL;
  PyObject *__pyx_v_OneOrMore = NULL;
  PyObject *__pyx_v_And = NULL;
  PyObject *__pyx_v_minElements = NULL;
  PyObject *__pyx_v_optElements = NULL;
  PyObject *__pyx_v_ret = NULL;
  PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_7__mul___4generator13 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__mul__", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__ *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1965, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_other = __pyx_v_other;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_other);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_other);

  /* "Aspidites/_vendor/pyparsing_extension.py":1985
 *         ``expr*(None, n) + ~expr``
 *         """
 *         from .pyparsing import ZeroOrMore, OneOrMore, And             # <<<<<<<<<<<<<<
 *         if other is Ellipsis:
 *             other = (0, None)
 */
  __pyx_t_1 = PyList_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1985, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_ZeroOrMore);
  __Pyx_GIVEREF(__pyx_n_s_ZeroOrMore);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_ZeroOrMore);
  __Pyx_INCREF(__pyx_n_s_OneOrMore);
  __Pyx_GIVEREF(__pyx_n_s_OneOrMore);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_OneOrMore);
  __Pyx_INCREF(__pyx_n_s_And);
  __Pyx_GIVEREF(__pyx_n_s_And);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_s_And);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_pyparsing, __pyx_t_1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1985, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_ZeroOrMore); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1985, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_ZeroOrMore = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_OneOrMore); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1985, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_OneOrMore = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_And); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1985, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_And = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1986
 *         """
 *         from .pyparsing import ZeroOrMore, OneOrMore, And
 *         if other is Ellipsis:             # <<<<<<<<<<<<<<
 *             other = (0, None)
 *         elif isinstance(other, tuple) and other[:1] == (Ellipsis,):
 */
  __pyx_t_3 = (__pyx_cur_scope->__pyx_v_other == __pyx_builtin_Ellipsis);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":1987
 *         from .pyparsing import ZeroOrMore, OneOrMore, And
 *         if other is Ellipsis:
 *             other = (0, None)             # <<<<<<<<<<<<<<
 *         elif isinstance(other, tuple) and other[:1] == (Ellipsis,):
 *             other = ((0,) + other[1:] + (None,))[:2]
 */
    __Pyx_INCREF(__pyx_tuple__65);
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_other);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_other, __pyx_tuple__65);
    __Pyx_GIVEREF(__pyx_tuple__65);

    /* "Aspidites/_vendor/pyparsing_extension.py":1986
 *         """
 *         from .pyparsing import ZeroOrMore, OneOrMore, And
 *         if other is Ellipsis:             # <<<<<<<<<<<<<<
 *             other = (0, None)
 *         elif isinstance(other, tuple) and other[:1] == (Ellipsis,):
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1988
 *         if other is Ellipsis:
 *             other = (0, None)
 *         elif isinstance(other, tuple) and other[:1] == (Ellipsis,):             # <<<<<<<<<<<<<<
 *             other = ((0,) + other[1:] + (None,))[:2]
 * 
 */
  __pyx_t_2 = __pyx_cur_scope->__pyx_v_other;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_Check(__pyx_t_2); 
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = (__pyx_t_3 != 0);
  if (__pyx_t_5) {
  } else {
    __pyx_t_4 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = __Pyx_PyObject_GetSlice(__pyx_cur_scope->__pyx_v_other, 0, 1, NULL, NULL, &__pyx_slice__66, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1988, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyObject_RichCompare(__pyx_t_2, __pyx_tuple__67, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1988, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1988, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":1989
 *             other = (0, None)
 *         elif isinstance(other, tuple) and other[:1] == (Ellipsis,):
 *             other = ((0,) + other[1:] + (None,))[:2]             # <<<<<<<<<<<<<<
 * 
 *         if isinstance(other, int):
 */
    __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_cur_scope->__pyx_v_other, 1, 0, NULL, NULL, &__pyx_slice__69, 1, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1989, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyNumber_Add(__pyx_tuple__68, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1989, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_Add(__pyx_t_2, __pyx_tuple__70); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1989, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetSlice(__pyx_t_1, 0, 2, NULL, NULL, &__pyx_slice__71, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1989, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_other);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_other, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1988
 *         if other is Ellipsis:
 *             other = (0, None)
 *         elif isinstance(other, tuple) and other[:1] == (Ellipsis,):             # <<<<<<<<<<<<<<
 *             other = ((0,) + other[1:] + (None,))[:2]
 * 
 */
  }
  __pyx_L3:;

  /* "Aspidites/_vendor/pyparsing_extension.py":1991
 *             other = ((0,) + other[1:] + (None,))[:2]
 * 
 *         if isinstance(other, int):             # <<<<<<<<<<<<<<
 *             minElements, optElements = other, 0
 *         elif isinstance(other, tuple):
 */
  __pyx_t_2 = __pyx_cur_scope->__pyx_v_other;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_4 = PyInt_Check(__pyx_t_2); 
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/pyparsing_extension.py":1992
 * 
 *         if isinstance(other, int):
 *             minElements, optElements = other, 0             # <<<<<<<<<<<<<<
 *         elif isinstance(other, tuple):
 *             other = tuple(o if o is not Ellipsis else None for o in other)
 */
    __pyx_t_2 = __pyx_cur_scope->__pyx_v_other;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_1 = __pyx_int_0;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_minElements = __pyx_t_2;
    __pyx_t_2 = 0;
    __pyx_v_optElements = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1991
 *             other = ((0,) + other[1:] + (None,))[:2]
 * 
 *         if isinstance(other, int):             # <<<<<<<<<<<<<<
 *             minElements, optElements = other, 0
 *         elif isinstance(other, tuple):
 */
    goto __pyx_L6;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1993
 *         if isinstance(other, int):
 *             minElements, optElements = other, 0
 *         elif isinstance(other, tuple):             # <<<<<<<<<<<<<<
 *             other = tuple(o if o is not Ellipsis else None for o in other)
 *             other = (other + (None, None))[:2]
 */
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_other;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_5 = PyTuple_Check(__pyx_t_1); 
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = (__pyx_t_5 != 0);
  if (likely(__pyx_t_4)) {

    /* "Aspidites/_vendor/pyparsing_extension.py":1994
 *             minElements, optElements = other, 0
 *         elif isinstance(other, tuple):
 *             other = tuple(o if o is not Ellipsis else None for o in other)             # <<<<<<<<<<<<<<
 *             other = (other + (None, None))[:2]
 *             if other[0] is None:
 */
    __pyx_t_1 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_7__mul___2genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1994, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PySequence_Tuple(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1994, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_other);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_other, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1995
 *         elif isinstance(other, tuple):
 *             other = tuple(o if o is not Ellipsis else None for o in other)
 *             other = (other + (None, None))[:2]             # <<<<<<<<<<<<<<
 *             if other[0] is None:
 *                 other = (0, other[1])
 */
    __pyx_t_2 = PyNumber_Add(__pyx_cur_scope->__pyx_v_other, __pyx_tuple__72); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1995, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_t_2, 0, 2, NULL, NULL, &__pyx_slice__71, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1995, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_other);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_other, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1996
 *             other = tuple(o if o is not Ellipsis else None for o in other)
 *             other = (other + (None, None))[:2]
 *             if other[0] is None:             # <<<<<<<<<<<<<<
 *                 other = (0, other[1])
 *             if isinstance(other[0], int) and other[1] is None:
 */
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_other, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1996, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = (__pyx_t_1 == Py_None);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = (__pyx_t_4 != 0);
    if (__pyx_t_5) {

      /* "Aspidites/_vendor/pyparsing_extension.py":1997
 *             other = (other + (None, None))[:2]
 *             if other[0] is None:
 *                 other = (0, other[1])             # <<<<<<<<<<<<<<
 *             if isinstance(other[0], int) and other[1] is None:
 *                 if other[0] == 0:
 */
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_other, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1997, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1997, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_int_0);
      __Pyx_GIVEREF(__pyx_int_0);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_int_0);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
      __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_other);
      __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_other, __pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_2);
      __pyx_t_2 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":1996
 *             other = tuple(o if o is not Ellipsis else None for o in other)
 *             other = (other + (None, None))[:2]
 *             if other[0] is None:             # <<<<<<<<<<<<<<
 *                 other = (0, other[1])
 *             if isinstance(other[0], int) and other[1] is None:
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":1998
 *             if other[0] is None:
 *                 other = (0, other[1])
 *             if isinstance(other[0], int) and other[1] is None:             # <<<<<<<<<<<<<<
 *                 if other[0] == 0:
 *                     return ZeroOrMore(self)
 */
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_other, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1998, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyInt_Check(__pyx_t_2); 
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_3 = (__pyx_t_4 != 0);
    if (__pyx_t_3) {
    } else {
      __pyx_t_5 = __pyx_t_3;
      goto __pyx_L9_bool_binop_done;
    }
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_other, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1998, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = (__pyx_t_2 == Py_None);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = (__pyx_t_3 != 0);
    __pyx_t_5 = __pyx_t_4;
    __pyx_L9_bool_binop_done:;
    if (__pyx_t_5) {

      /* "Aspidites/_vendor/pyparsing_extension.py":1999
 *                 other = (0, other[1])
 *             if isinstance(other[0], int) and other[1] is None:
 *                 if other[0] == 0:             # <<<<<<<<<<<<<<
 *                     return ZeroOrMore(self)
 *                 if other[0] == 1:
 */
      __pyx_t_2 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_other, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1999, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_t_2, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1999, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1999, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_5) {

        /* "Aspidites/_vendor/pyparsing_extension.py":2000
 *             if isinstance(other[0], int) and other[1] is None:
 *                 if other[0] == 0:
 *                     return ZeroOrMore(self)             # <<<<<<<<<<<<<<
 *                 if other[0] == 1:
 *                     return OneOrMore(self)
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_v_ZeroOrMore);
        __pyx_t_2 = __pyx_v_ZeroOrMore; __pyx_t_6 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_6, __pyx_cur_scope->__pyx_v_self) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_cur_scope->__pyx_v_self);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2000, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_1;
        __pyx_t_1 = 0;
        goto __pyx_L0;

        /* "Aspidites/_vendor/pyparsing_extension.py":1999
 *                 other = (0, other[1])
 *             if isinstance(other[0], int) and other[1] is None:
 *                 if other[0] == 0:             # <<<<<<<<<<<<<<
 *                     return ZeroOrMore(self)
 *                 if other[0] == 1:
 */
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":2001
 *                 if other[0] == 0:
 *                     return ZeroOrMore(self)
 *                 if other[0] == 1:             # <<<<<<<<<<<<<<
 *                     return OneOrMore(self)
 *                 else:
 */
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_other, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2001, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_t_1, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2001, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 2001, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (__pyx_t_5) {

        /* "Aspidites/_vendor/pyparsing_extension.py":2002
 *                     return ZeroOrMore(self)
 *                 if other[0] == 1:
 *                     return OneOrMore(self)             # <<<<<<<<<<<<<<
 *                 else:
 *                     return self * other[0] + ZeroOrMore(self)
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_v_OneOrMore);
        __pyx_t_1 = __pyx_v_OneOrMore; __pyx_t_6 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_1, function);
          }
        }
        __pyx_t_2 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_cur_scope->__pyx_v_self) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_cur_scope->__pyx_v_self);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2002, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        goto __pyx_L0;

        /* "Aspidites/_vendor/pyparsing_extension.py":2001
 *                 if other[0] == 0:
 *                     return ZeroOrMore(self)
 *                 if other[0] == 1:             # <<<<<<<<<<<<<<
 *                     return OneOrMore(self)
 *                 else:
 */
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":2004
 *                     return OneOrMore(self)
 *                 else:
 *                     return self * other[0] + ZeroOrMore(self)             # <<<<<<<<<<<<<<
 *             elif isinstance(other[0], int) and isinstance(other[1], int):
 *                 minElements, optElements = other
 */
      /*else*/ {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_other, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2004, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = PyNumber_Multiply(__pyx_cur_scope->__pyx_v_self, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2004, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_INCREF(__pyx_v_ZeroOrMore);
        __pyx_t_6 = __pyx_v_ZeroOrMore; __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
          }
        }
        __pyx_t_2 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_7, __pyx_cur_scope->__pyx_v_self) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_cur_scope->__pyx_v_self);
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2004, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = PyNumber_Add(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2004, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_6;
        __pyx_t_6 = 0;
        goto __pyx_L0;
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":1998
 *             if other[0] is None:
 *                 other = (0, other[1])
 *             if isinstance(other[0], int) and other[1] is None:             # <<<<<<<<<<<<<<
 *                 if other[0] == 0:
 *                     return ZeroOrMore(self)
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":2005
 *                 else:
 *                     return self * other[0] + ZeroOrMore(self)
 *             elif isinstance(other[0], int) and isinstance(other[1], int):             # <<<<<<<<<<<<<<
 *                 minElements, optElements = other
 *                 optElements -= minElements
 */
    __pyx_t_6 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_other, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2005, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = PyInt_Check(__pyx_t_6); 
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_3 = (__pyx_t_4 != 0);
    if (__pyx_t_3) {
    } else {
      __pyx_t_5 = __pyx_t_3;
      goto __pyx_L13_bool_binop_done;
    }
    __pyx_t_6 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_other, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2005, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_3 = PyInt_Check(__pyx_t_6); 
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_4 = (__pyx_t_3 != 0);
    __pyx_t_5 = __pyx_t_4;
    __pyx_L13_bool_binop_done:;
    if (likely(__pyx_t_5)) {

      /* "Aspidites/_vendor/pyparsing_extension.py":2006
 *                     return self * other[0] + ZeroOrMore(self)
 *             elif isinstance(other[0], int) and isinstance(other[1], int):
 *                 minElements, optElements = other             # <<<<<<<<<<<<<<
 *                 optElements -= minElements
 *             else:
 */
      if ((likely(PyTuple_CheckExact(__pyx_cur_scope->__pyx_v_other))) || (PyList_CheckExact(__pyx_cur_scope->__pyx_v_other))) {
        PyObject* sequence = __pyx_cur_scope->__pyx_v_other;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 2006, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_6 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_6 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_2);
        #else
        __pyx_t_6 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2006, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2006, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        Py_ssize_t index = -1;
        __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_v_other); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2006, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_8 = Py_TYPE(__pyx_t_1)->tp_iternext;
        index = 0; __pyx_t_6 = __pyx_t_8(__pyx_t_1); if (unlikely(!__pyx_t_6)) goto __pyx_L15_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_6);
        index = 1; __pyx_t_2 = __pyx_t_8(__pyx_t_1); if (unlikely(!__pyx_t_2)) goto __pyx_L15_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_2);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_1), 2) < 0) __PYX_ERR(0, 2006, __pyx_L1_error)
        __pyx_t_8 = NULL;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L16_unpacking_done;
        __pyx_L15_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_8 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 2006, __pyx_L1_error)
        __pyx_L16_unpacking_done:;
      }
      __pyx_v_minElements = __pyx_t_6;
      __pyx_t_6 = 0;
      __pyx_v_optElements = __pyx_t_2;
      __pyx_t_2 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":2007
 *             elif isinstance(other[0], int) and isinstance(other[1], int):
 *                 minElements, optElements = other
 *                 optElements -= minElements             # <<<<<<<<<<<<<<
 *             else:
 *                 raise TypeError("cannot multiply 'ParserElement' and ('%s', '%s') objects", type(other[0]),
 */
      __pyx_t_2 = PyNumber_InPlaceSubtract(__pyx_v_optElements, __pyx_v_minElements); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2007, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF_SET(__pyx_v_optElements, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":2005
 *                 else:
 *                     return self * other[0] + ZeroOrMore(self)
 *             elif isinstance(other[0], int) and isinstance(other[1], int):             # <<<<<<<<<<<<<<
 *                 minElements, optElements = other
 *                 optElements -= minElements
 */
      goto __pyx_L8;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":2009
 *                 optElements -= minElements
 *             else:
 *                 raise TypeError("cannot multiply 'ParserElement' and ('%s', '%s') objects", type(other[0]),             # <<<<<<<<<<<<<<
 *                                 type(other[1]))
 *         else:
 */
    /*else*/ {
      __pyx_t_2 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_other, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2009, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);

      /* "Aspidites/_vendor/pyparsing_extension.py":2010
 *             else:
 *                 raise TypeError("cannot multiply 'ParserElement' and ('%s', '%s') objects", type(other[0]),
 *                                 type(other[1]))             # <<<<<<<<<<<<<<
 *         else:
 *             raise TypeError("cannot multiply 'ParserElement' and '%s' objects", type(other))
 */
      __pyx_t_6 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_other, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2010, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);

      /* "Aspidites/_vendor/pyparsing_extension.py":2009
 *                 optElements -= minElements
 *             else:
 *                 raise TypeError("cannot multiply 'ParserElement' and ('%s', '%s') objects", type(other[0]),             # <<<<<<<<<<<<<<
 *                                 type(other[1]))
 *         else:
 */
      __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2009, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_kp_s_cannot_multiply_ParserElement_an);
      __Pyx_GIVEREF(__pyx_kp_s_cannot_multiply_ParserElement_an);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_s_cannot_multiply_ParserElement_an);
      __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_t_2)));
      __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_t_2)));
      PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)Py_TYPE(__pyx_t_2)));
      __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_t_6)));
      __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_t_6)));
      PyTuple_SET_ITEM(__pyx_t_1, 2, ((PyObject *)Py_TYPE(__pyx_t_6)));
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_t_1, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2009, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(0, 2009, __pyx_L1_error)
    }
    __pyx_L8:;

    /* "Aspidites/_vendor/pyparsing_extension.py":1993
 *         if isinstance(other, int):
 *             minElements, optElements = other, 0
 *         elif isinstance(other, tuple):             # <<<<<<<<<<<<<<
 *             other = tuple(o if o is not Ellipsis else None for o in other)
 *             other = (other + (None, None))[:2]
 */
    goto __pyx_L6;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2012
 *                                 type(other[1]))
 *         else:
 *             raise TypeError("cannot multiply 'ParserElement' and '%s' objects", type(other))             # <<<<<<<<<<<<<<
 * 
 *         if minElements < 0:
 */
  /*else*/ {
    __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2012, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_kp_s_cannot_multiply_ParserElement_an_2);
    __Pyx_GIVEREF(__pyx_kp_s_cannot_multiply_ParserElement_an_2);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_kp_s_cannot_multiply_ParserElement_an_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_cur_scope->__pyx_v_other)));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_cur_scope->__pyx_v_other)));
    PyTuple_SET_ITEM(__pyx_t_6, 1, ((PyObject *)Py_TYPE(__pyx_cur_scope->__pyx_v_other)));
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2012, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 2012, __pyx_L1_error)
  }
  __pyx_L6:;

  /* "Aspidites/_vendor/pyparsing_extension.py":2014
 *             raise TypeError("cannot multiply 'ParserElement' and '%s' objects", type(other))
 * 
 *         if minElements < 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("cannot multiply ParserElement by negative value")
 *         if optElements < 0:
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_minElements, __pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2014, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 2014, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_5)) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2015
 * 
 *         if minElements < 0:
 *             raise ValueError("cannot multiply ParserElement by negative value")             # <<<<<<<<<<<<<<
 *         if optElements < 0:
 *             raise ValueError("second tuple value must be greater or equal to first tuple value")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__73, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2015, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 2015, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":2014
 *             raise TypeError("cannot multiply 'ParserElement' and '%s' objects", type(other))
 * 
 *         if minElements < 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("cannot multiply ParserElement by negative value")
 *         if optElements < 0:
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2016
 *         if minElements < 0:
 *             raise ValueError("cannot multiply ParserElement by negative value")
 *         if optElements < 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("second tuple value must be greater or equal to first tuple value")
 *         if minElements == optElements == 0:
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_optElements, __pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2016, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 2016, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_5)) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2017
 *             raise ValueError("cannot multiply ParserElement by negative value")
 *         if optElements < 0:
 *             raise ValueError("second tuple value must be greater or equal to first tuple value")             # <<<<<<<<<<<<<<
 *         if minElements == optElements == 0:
 *             raise ValueError("cannot multiply ParserElement by 0 or (0, 0)")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__74, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2017, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 2017, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":2016
 *         if minElements < 0:
 *             raise ValueError("cannot multiply ParserElement by negative value")
 *         if optElements < 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("second tuple value must be greater or equal to first tuple value")
 *         if minElements == optElements == 0:
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2018
 *         if optElements < 0:
 *             raise ValueError("second tuple value must be greater or equal to first tuple value")
 *         if minElements == optElements == 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("cannot multiply ParserElement by 0 or (0, 0)")
 * 
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_minElements, __pyx_v_optElements, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2018, __pyx_L1_error)
  if (__Pyx_PyObject_IsTrue(__pyx_t_1)) {
    __Pyx_DECREF(__pyx_t_1);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_optElements, __pyx_int_0, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2018, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 2018, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_5)) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2019
 *             raise ValueError("second tuple value must be greater or equal to first tuple value")
 *         if minElements == optElements == 0:
 *             raise ValueError("cannot multiply ParserElement by 0 or (0, 0)")             # <<<<<<<<<<<<<<
 * 
 *         if optElements:
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__75, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2019, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 2019, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":2018
 *         if optElements < 0:
 *             raise ValueError("second tuple value must be greater or equal to first tuple value")
 *         if minElements == optElements == 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("cannot multiply ParserElement by 0 or (0, 0)")
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2021
 *             raise ValueError("cannot multiply ParserElement by 0 or (0, 0)")
 * 
 *         if optElements:             # <<<<<<<<<<<<<<
 *             def makeOptionalList(n):
 *                 from .pyparsing import Optional
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_optElements); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 2021, __pyx_L1_error)
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2022
 * 
 *         if optElements:
 *             def makeOptionalList(n):             # <<<<<<<<<<<<<<
 *                 from .pyparsing import Optional
 *                 if n > 1:
 */
    __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_7__mul___1makeOptionalList, 0, __pyx_n_s_ParserElement___mul___locals_mak, ((PyObject*)__pyx_cur_scope), __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__77)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2022, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_v_makeOptionalList = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2029
 *                     return Optional(self)
 * 
 *             if minElements:             # <<<<<<<<<<<<<<
 *                 if minElements == 1:
 *                     ret = self + makeOptionalList(optElements)
 */
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_minElements); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 2029, __pyx_L1_error)
    if (__pyx_t_5) {

      /* "Aspidites/_vendor/pyparsing_extension.py":2030
 * 
 *             if minElements:
 *                 if minElements == 1:             # <<<<<<<<<<<<<<
 *                     ret = self + makeOptionalList(optElements)
 *                 else:
 */
      __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_minElements, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2030, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 2030, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_5) {

        /* "Aspidites/_vendor/pyparsing_extension.py":2031
 *             if minElements:
 *                 if minElements == 1:
 *                     ret = self + makeOptionalList(optElements)             # <<<<<<<<<<<<<<
 *                 else:
 *                     ret = And([self] * minElements) + makeOptionalList(optElements)
 */
        __pyx_t_1 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_7__mul___makeOptionalList(__pyx_cur_scope->__pyx_v_makeOptionalList, __pyx_v_optElements); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2031, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_6 = PyNumber_Add(__pyx_cur_scope->__pyx_v_self, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2031, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_v_ret = __pyx_t_6;
        __pyx_t_6 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":2030
 * 
 *             if minElements:
 *                 if minElements == 1:             # <<<<<<<<<<<<<<
 *                     ret = self + makeOptionalList(optElements)
 *                 else:
 */
        goto __pyx_L22;
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":2033
 *                     ret = self + makeOptionalList(optElements)
 *                 else:
 *                     ret = And([self] * minElements) + makeOptionalList(optElements)             # <<<<<<<<<<<<<<
 *             else:
 *                 ret = makeOptionalList(optElements)
 */
      /*else*/ {
        __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2033, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
        __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);
        PyList_SET_ITEM(__pyx_t_1, 0, __pyx_cur_scope->__pyx_v_self);
        { PyObject* __pyx_temp = PyNumber_InPlaceMultiply(__pyx_t_1, __pyx_v_minElements); if (unlikely(!__pyx_temp)) __PYX_ERR(0, 2033, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_temp);
          __Pyx_DECREF(__pyx_t_1);
          __pyx_t_1 = __pyx_temp;
        }
        __Pyx_INCREF(__pyx_v_And);
        __pyx_t_2 = __pyx_v_And; __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        __pyx_t_6 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_7, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1);
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2033, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_7__mul___makeOptionalList(__pyx_cur_scope->__pyx_v_makeOptionalList, __pyx_v_optElements); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2033, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = PyNumber_Add(__pyx_t_6, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2033, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_v_ret = __pyx_t_1;
        __pyx_t_1 = 0;
      }
      __pyx_L22:;

      /* "Aspidites/_vendor/pyparsing_extension.py":2029
 *                     return Optional(self)
 * 
 *             if minElements:             # <<<<<<<<<<<<<<
 *                 if minElements == 1:
 *                     ret = self + makeOptionalList(optElements)
 */
      goto __pyx_L21;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":2035
 *                     ret = And([self] * minElements) + makeOptionalList(optElements)
 *             else:
 *                 ret = makeOptionalList(optElements)             # <<<<<<<<<<<<<<
 *         else:
 *             if minElements == 1:
 */
    /*else*/ {
      __pyx_t_1 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_7__mul___makeOptionalList(__pyx_cur_scope->__pyx_v_makeOptionalList, __pyx_v_optElements); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2035, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_v_ret = __pyx_t_1;
      __pyx_t_1 = 0;
    }
    __pyx_L21:;

    /* "Aspidites/_vendor/pyparsing_extension.py":2021
 *             raise ValueError("cannot multiply ParserElement by 0 or (0, 0)")
 * 
 *         if optElements:             # <<<<<<<<<<<<<<
 *             def makeOptionalList(n):
 *                 from .pyparsing import Optional
 */
    goto __pyx_L20;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2037
 *                 ret = makeOptionalList(optElements)
 *         else:
 *             if minElements == 1:             # <<<<<<<<<<<<<<
 *                 ret = self
 *             else:
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_minElements, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2037, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 2037, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_5) {

      /* "Aspidites/_vendor/pyparsing_extension.py":2038
 *         else:
 *             if minElements == 1:
 *                 ret = self             # <<<<<<<<<<<<<<
 *             else:
 *                 ret = And([self] * minElements)
 */
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
      __pyx_v_ret = __pyx_cur_scope->__pyx_v_self;

      /* "Aspidites/_vendor/pyparsing_extension.py":2037
 *                 ret = makeOptionalList(optElements)
 *         else:
 *             if minElements == 1:             # <<<<<<<<<<<<<<
 *                 ret = self
 *             else:
 */
      goto __pyx_L23;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":2040
 *                 ret = self
 *             else:
 *                 ret = And([self] * minElements)             # <<<<<<<<<<<<<<
 *         return ret
 * 
 */
    /*else*/ {
      __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2040, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);
      PyList_SET_ITEM(__pyx_t_2, 0, __pyx_cur_scope->__pyx_v_self);
      { PyObject* __pyx_temp = PyNumber_InPlaceMultiply(__pyx_t_2, __pyx_v_minElements); if (unlikely(!__pyx_temp)) __PYX_ERR(0, 2040, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_temp);
        __Pyx_DECREF(__pyx_t_2);
        __pyx_t_2 = __pyx_temp;
      }
      __Pyx_INCREF(__pyx_v_And);
      __pyx_t_6 = __pyx_v_And; __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_7, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_2);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2040, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_v_ret = __pyx_t_1;
      __pyx_t_1 = 0;
    }
    __pyx_L23:;
  }
  __pyx_L20:;

  /* "Aspidites/_vendor/pyparsing_extension.py":2041
 *             else:
 *                 ret = And([self] * minElements)
 *         return ret             # <<<<<<<<<<<<<<
 * 
 *     def __rmul__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_ret);
  __pyx_r = __pyx_v_ret;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1965
 *         return other - self
 * 
 *     def __mul__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of * operator, allows use of ``expr * 3`` in place of
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__mul__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ZeroOrMore);
  __Pyx_XDECREF(__pyx_v_OneOrMore);
  __Pyx_XDECREF(__pyx_v_And);
  __Pyx_XDECREF(__pyx_v_minElements);
  __Pyx_XDECREF(__pyx_v_optElements);
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_XDECREF(__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_7__mul___4generator13);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2043
 *         return ret
 * 
 *     def __rmul__(self, other):             # <<<<<<<<<<<<<<
 *         return self.__mul__(other)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_69__rmul__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_69__rmul__ = {"__rmul__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_69__rmul__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_69__rmul__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_other = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__rmul__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_other,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_other)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__rmul__", 1, 2, 2, 1); __PYX_ERR(0, 2043, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__rmul__") < 0)) __PYX_ERR(0, 2043, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_other = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__rmul__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2043, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__rmul__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_68__rmul__(__pyx_self, __pyx_v_self, __pyx_v_other);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_68__rmul__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__rmul__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2044
 * 
 *     def __rmul__(self, other):
 *         return self.__mul__(other)             # <<<<<<<<<<<<<<
 * 
 *     def __or__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_mul); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_other) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_other);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2043
 *         return ret
 * 
 *     def __rmul__(self, other):             # <<<<<<<<<<<<<<
 *         return self.__mul__(other)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__rmul__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2046
 *         return self.__mul__(other)
 * 
 *     def __or__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of | operator - returns :class:`MatchFirst`
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_71__or__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_70__or__[] = "\n        Implementation of | operator - returns :class:`MatchFirst`\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_71__or__ = {"__or__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_71__or__, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_70__or__};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_71__or__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_other = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__or__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_other,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_other)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__or__", 1, 2, 2, 1); __PYX_ERR(0, 2046, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__or__") < 0)) __PYX_ERR(0, 2046, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_other = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__or__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2046, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__or__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_70__or__(__pyx_self, __pyx_v_self, __pyx_v_other);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_70__or__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_v__PendingSkip = NULL;
  PyObject *__pyx_v_MatchFirst = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__or__", 0);
  __Pyx_INCREF(__pyx_v_other);

  /* "Aspidites/_vendor/pyparsing_extension.py":2050
 *         Implementation of | operator - returns :class:`MatchFirst`
 *         """
 *         from .pyparsing import _PendingSkip, MatchFirst             # <<<<<<<<<<<<<<
 *         if other is Ellipsis:
 *             return _PendingSkip(self, must_skip=True)
 */
  __pyx_t_1 = PyList_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2050, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_PendingSkip);
  __Pyx_GIVEREF(__pyx_n_s_PendingSkip);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PendingSkip);
  __Pyx_INCREF(__pyx_n_s_MatchFirst);
  __Pyx_GIVEREF(__pyx_n_s_MatchFirst);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_MatchFirst);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_pyparsing, __pyx_t_1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2050, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_PendingSkip); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2050, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v__PendingSkip = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_MatchFirst); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2050, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_MatchFirst = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2051
 *         """
 *         from .pyparsing import _PendingSkip, MatchFirst
 *         if other is Ellipsis:             # <<<<<<<<<<<<<<
 *             return _PendingSkip(self, must_skip=True)
 * 
 */
  __pyx_t_3 = (__pyx_v_other == __pyx_builtin_Ellipsis);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2052
 *         from .pyparsing import _PendingSkip, MatchFirst
 *         if other is Ellipsis:
 *             return _PendingSkip(self, must_skip=True)             # <<<<<<<<<<<<<<
 * 
 *         if isinstance(other, str):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2052, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_self);
    __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2052, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_must_skip, Py_True) < 0) __PYX_ERR(0, 2052, __pyx_L1_error)
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_v__PendingSkip, __pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2052, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2051
 *         """
 *         from .pyparsing import _PendingSkip, MatchFirst
 *         if other is Ellipsis:             # <<<<<<<<<<<<<<
 *             return _PendingSkip(self, must_skip=True)
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2054
 *             return _PendingSkip(self, must_skip=True)
 * 
 *         if isinstance(other, str):             # <<<<<<<<<<<<<<
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):
 */
  __pyx_t_4 = PyString_Check(__pyx_v_other); 
  __pyx_t_3 = (__pyx_t_4 != 0);
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2055
 * 
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)             # <<<<<<<<<<<<<<
 *         if not isinstance(other, ParserElement):
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_literalStringClass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2055, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_5 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_v_other) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_other);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2055, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_other, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2054
 *             return _PendingSkip(self, must_skip=True)
 * 
 *         if isinstance(other, str):             # <<<<<<<<<<<<<<
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2056
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):             # <<<<<<<<<<<<<<
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 *                           SyntaxWarning, stacklevel=2)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2056, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_other, __pyx_t_5); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 2056, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_4 = ((!(__pyx_t_3 != 0)) != 0);
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2057
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),             # <<<<<<<<<<<<<<
 *                           SyntaxWarning, stacklevel=2)
 *             return None
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_warnings); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2057, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_warn); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2057, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Cannot_combine_element_of_type_s, ((PyObject *)Py_TYPE(__pyx_v_other))); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2057, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "Aspidites/_vendor/pyparsing_extension.py":2058
 *         if not isinstance(other, ParserElement):
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 *                           SyntaxWarning, stacklevel=2)             # <<<<<<<<<<<<<<
 *             return None
 *         return MatchFirst([self, other])
 */
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2057, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
    __Pyx_INCREF(__pyx_builtin_SyntaxWarning);
    __Pyx_GIVEREF(__pyx_builtin_SyntaxWarning);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_builtin_SyntaxWarning);
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2058, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_stacklevel, __pyx_int_2) < 0) __PYX_ERR(0, 2058, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":2057
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),             # <<<<<<<<<<<<<<
 *                           SyntaxWarning, stacklevel=2)
 *             return None
 */
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2057, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2059
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 *                           SyntaxWarning, stacklevel=2)
 *             return None             # <<<<<<<<<<<<<<
 *         return MatchFirst([self, other])
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2056
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):             # <<<<<<<<<<<<<<
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 *                           SyntaxWarning, stacklevel=2)
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2060
 *                           SyntaxWarning, stacklevel=2)
 *             return None
 *         return MatchFirst([self, other])             # <<<<<<<<<<<<<<
 * 
 *     def __ror__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = PyList_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2060, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyList_SET_ITEM(__pyx_t_5, 0, __pyx_v_self);
  __Pyx_INCREF(__pyx_v_other);
  __Pyx_GIVEREF(__pyx_v_other);
  PyList_SET_ITEM(__pyx_t_5, 1, __pyx_v_other);
  __Pyx_INCREF(__pyx_v_MatchFirst);
  __pyx_t_2 = __pyx_v_MatchFirst; __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_6 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_1, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2060, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2046
 *         return self.__mul__(other)
 * 
 *     def __or__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of | operator - returns :class:`MatchFirst`
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__or__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v__PendingSkip);
  __Pyx_XDECREF(__pyx_v_MatchFirst);
  __Pyx_XDECREF(__pyx_v_other);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2062
 *         return MatchFirst([self, other])
 * 
 *     def __ror__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of | operator when left operand is not a :class:`ParserElement`
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_73__ror__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_72__ror__[] = "\n        Implementation of | operator when left operand is not a :class:`ParserElement`\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_73__ror__ = {"__ror__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_73__ror__, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_72__ror__};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_73__ror__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_other = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__ror__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_other,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_other)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__ror__", 1, 2, 2, 1); __PYX_ERR(0, 2062, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__ror__") < 0)) __PYX_ERR(0, 2062, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_other = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__ror__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2062, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__ror__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_72__ror__(__pyx_self, __pyx_v_self, __pyx_v_other);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_72__ror__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__ror__", 0);
  __Pyx_INCREF(__pyx_v_other);

  /* "Aspidites/_vendor/pyparsing_extension.py":2066
 *         Implementation of | operator when left operand is not a :class:`ParserElement`
 *         """
 *         if isinstance(other, str):             # <<<<<<<<<<<<<<
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):
 */
  __pyx_t_1 = PyString_Check(__pyx_v_other); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2067
 *         """
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)             # <<<<<<<<<<<<<<
 *         if not isinstance(other, ParserElement):
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_literalStringClass); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2067, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_other) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_other);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2067, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_other, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2066
 *         Implementation of | operator when left operand is not a :class:`ParserElement`
 *         """
 *         if isinstance(other, str):             # <<<<<<<<<<<<<<
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2068
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):             # <<<<<<<<<<<<<<
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 *                           SyntaxWarning, stacklevel=2)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyObject_IsInstance(__pyx_v_other, __pyx_t_3); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 2068, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = ((!(__pyx_t_2 != 0)) != 0);
  if (__pyx_t_1) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2069
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),             # <<<<<<<<<<<<<<
 *                           SyntaxWarning, stacklevel=2)
 *             return None
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_warnings); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2069, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_warn); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2069, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Cannot_combine_element_of_type_s, ((PyObject *)Py_TYPE(__pyx_v_other))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2069, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "Aspidites/_vendor/pyparsing_extension.py":2070
 *         if not isinstance(other, ParserElement):
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 *                           SyntaxWarning, stacklevel=2)             # <<<<<<<<<<<<<<
 *             return None
 *         return other | self
 */
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2069, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3);
    __Pyx_INCREF(__pyx_builtin_SyntaxWarning);
    __Pyx_GIVEREF(__pyx_builtin_SyntaxWarning);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_builtin_SyntaxWarning);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2070, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_stacklevel, __pyx_int_2) < 0) __PYX_ERR(0, 2070, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":2069
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),             # <<<<<<<<<<<<<<
 *                           SyntaxWarning, stacklevel=2)
 *             return None
 */
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2069, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2071
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 *                           SyntaxWarning, stacklevel=2)
 *             return None             # <<<<<<<<<<<<<<
 *         return other | self
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2068
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):             # <<<<<<<<<<<<<<
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 *                           SyntaxWarning, stacklevel=2)
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2072
 *                           SyntaxWarning, stacklevel=2)
 *             return None
 *         return other | self             # <<<<<<<<<<<<<<
 * 
 *     def __xor__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_6 = PyNumber_Or(__pyx_v_other, __pyx_v_self); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2072, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2062
 *         return MatchFirst([self, other])
 * 
 *     def __ror__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of | operator when left operand is not a :class:`ParserElement`
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__ror__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_other);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2074
 *         return other | self
 * 
 *     def __xor__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of ^ operator - returns :class:`Or`
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_75__xor__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_74__xor__[] = "\n        Implementation of ^ operator - returns :class:`Or`\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_75__xor__ = {"__xor__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_75__xor__, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_74__xor__};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_75__xor__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_other = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__xor__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_other,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_other)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__xor__", 1, 2, 2, 1); __PYX_ERR(0, 2074, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__xor__") < 0)) __PYX_ERR(0, 2074, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_other = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__xor__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2074, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__xor__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_74__xor__(__pyx_self, __pyx_v_self, __pyx_v_other);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_74__xor__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_v_Or = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__xor__", 0);
  __Pyx_INCREF(__pyx_v_other);

  /* "Aspidites/_vendor/pyparsing_extension.py":2078
 *         Implementation of ^ operator - returns :class:`Or`
 *         """
 *         from .pyparsing import Or             # <<<<<<<<<<<<<<
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_Or);
  __Pyx_GIVEREF(__pyx_n_s_Or);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_Or);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_pyparsing, __pyx_t_1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_Or); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_Or = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2079
 *         """
 *         from .pyparsing import Or
 *         if isinstance(other, str):             # <<<<<<<<<<<<<<
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):
 */
  __pyx_t_3 = PyString_Check(__pyx_v_other); 
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2080
 *         from .pyparsing import Or
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)             # <<<<<<<<<<<<<<
 *         if not isinstance(other, ParserElement):
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_literalStringClass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2080, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v_other) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_other);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2080, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_other, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2079
 *         """
 *         from .pyparsing import Or
 *         if isinstance(other, str):             # <<<<<<<<<<<<<<
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2081
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):             # <<<<<<<<<<<<<<
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 *                           SyntaxWarning, stacklevel=2)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2081, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyObject_IsInstance(__pyx_v_other, __pyx_t_2); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 2081, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = ((!(__pyx_t_4 != 0)) != 0);
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2082
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),             # <<<<<<<<<<<<<<
 *                           SyntaxWarning, stacklevel=2)
 *             return None
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_warnings); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2082, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_warn); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2082, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Cannot_combine_element_of_type_s, ((PyObject *)Py_TYPE(__pyx_v_other))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2082, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "Aspidites/_vendor/pyparsing_extension.py":2083
 *         if not isinstance(other, ParserElement):
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 *                           SyntaxWarning, stacklevel=2)             # <<<<<<<<<<<<<<
 *             return None
 *         return Or([self, other])
 */
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2082, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
    __Pyx_INCREF(__pyx_builtin_SyntaxWarning);
    __Pyx_GIVEREF(__pyx_builtin_SyntaxWarning);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_builtin_SyntaxWarning);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2083, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_stacklevel, __pyx_int_2) < 0) __PYX_ERR(0, 2083, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":2082
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),             # <<<<<<<<<<<<<<
 *                           SyntaxWarning, stacklevel=2)
 *             return None
 */
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_5, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2082, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2084
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 *                           SyntaxWarning, stacklevel=2)
 *             return None             # <<<<<<<<<<<<<<
 *         return Or([self, other])
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2081
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):             # <<<<<<<<<<<<<<
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 *                           SyntaxWarning, stacklevel=2)
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2085
 *                           SyntaxWarning, stacklevel=2)
 *             return None
 *         return Or([self, other])             # <<<<<<<<<<<<<<
 * 
 *     def __rxor__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyList_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2085, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_v_self);
  __Pyx_INCREF(__pyx_v_other);
  __Pyx_GIVEREF(__pyx_v_other);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_v_other);
  __Pyx_INCREF(__pyx_v_Or);
  __pyx_t_5 = __pyx_v_Or; __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_6 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_1, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2085, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2074
 *         return other | self
 * 
 *     def __xor__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of ^ operator - returns :class:`Or`
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__xor__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_Or);
  __Pyx_XDECREF(__pyx_v_other);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2087
 *         return Or([self, other])
 * 
 *     def __rxor__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of ^ operator when left operand is not a :class:`ParserElement`
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_77__rxor__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_76__rxor__[] = "\n        Implementation of ^ operator when left operand is not a :class:`ParserElement`\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_77__rxor__ = {"__rxor__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_77__rxor__, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_76__rxor__};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_77__rxor__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_other = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__rxor__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_other,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_other)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__rxor__", 1, 2, 2, 1); __PYX_ERR(0, 2087, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__rxor__") < 0)) __PYX_ERR(0, 2087, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_other = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__rxor__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2087, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__rxor__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_76__rxor__(__pyx_self, __pyx_v_self, __pyx_v_other);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_76__rxor__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__rxor__", 0);
  __Pyx_INCREF(__pyx_v_other);

  /* "Aspidites/_vendor/pyparsing_extension.py":2091
 *         Implementation of ^ operator when left operand is not a :class:`ParserElement`
 *         """
 *         if isinstance(other, str):             # <<<<<<<<<<<<<<
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):
 */
  __pyx_t_1 = PyString_Check(__pyx_v_other); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2092
 *         """
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)             # <<<<<<<<<<<<<<
 *         if not isinstance(other, ParserElement):
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_literalStringClass); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2092, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_other) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_other);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2092, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_other, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2091
 *         Implementation of ^ operator when left operand is not a :class:`ParserElement`
 *         """
 *         if isinstance(other, str):             # <<<<<<<<<<<<<<
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2093
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):             # <<<<<<<<<<<<<<
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 *                           SyntaxWarning, stacklevel=2)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2093, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyObject_IsInstance(__pyx_v_other, __pyx_t_3); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 2093, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = ((!(__pyx_t_2 != 0)) != 0);
  if (__pyx_t_1) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2094
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),             # <<<<<<<<<<<<<<
 *                           SyntaxWarning, stacklevel=2)
 *             return None
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_warnings); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2094, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_warn); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2094, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Cannot_combine_element_of_type_s, ((PyObject *)Py_TYPE(__pyx_v_other))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2094, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "Aspidites/_vendor/pyparsing_extension.py":2095
 *         if not isinstance(other, ParserElement):
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 *                           SyntaxWarning, stacklevel=2)             # <<<<<<<<<<<<<<
 *             return None
 *         return other ^ self
 */
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2094, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3);
    __Pyx_INCREF(__pyx_builtin_SyntaxWarning);
    __Pyx_GIVEREF(__pyx_builtin_SyntaxWarning);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_builtin_SyntaxWarning);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2095, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_stacklevel, __pyx_int_2) < 0) __PYX_ERR(0, 2095, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":2094
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),             # <<<<<<<<<<<<<<
 *                           SyntaxWarning, stacklevel=2)
 *             return None
 */
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2094, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2096
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 *                           SyntaxWarning, stacklevel=2)
 *             return None             # <<<<<<<<<<<<<<
 *         return other ^ self
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2093
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):             # <<<<<<<<<<<<<<
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 *                           SyntaxWarning, stacklevel=2)
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2097
 *                           SyntaxWarning, stacklevel=2)
 *             return None
 *         return other ^ self             # <<<<<<<<<<<<<<
 * 
 *     def __and__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_6 = PyNumber_Xor(__pyx_v_other, __pyx_v_self); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2097, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2087
 *         return Or([self, other])
 * 
 *     def __rxor__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of ^ operator when left operand is not a :class:`ParserElement`
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__rxor__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_other);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2099
 *         return other ^ self
 * 
 *     def __and__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of & operator - returns :class:`Each`
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_79__and__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_78__and__[] = "\n        Implementation of & operator - returns :class:`Each`\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_79__and__ = {"__and__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_79__and__, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_78__and__};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_79__and__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_other = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__and__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_other,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_other)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__and__", 1, 2, 2, 1); __PYX_ERR(0, 2099, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__and__") < 0)) __PYX_ERR(0, 2099, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_other = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__and__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2099, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__and__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_78__and__(__pyx_self, __pyx_v_self, __pyx_v_other);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_78__and__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_v_Each = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__and__", 0);
  __Pyx_INCREF(__pyx_v_other);

  /* "Aspidites/_vendor/pyparsing_extension.py":2103
 *         Implementation of & operator - returns :class:`Each`
 *         """
 *         from .pyparsing import Each             # <<<<<<<<<<<<<<
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_Each);
  __Pyx_GIVEREF(__pyx_n_s_Each);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_Each);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_pyparsing, __pyx_t_1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_Each); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_Each = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2104
 *         """
 *         from .pyparsing import Each
 *         if isinstance(other, str):             # <<<<<<<<<<<<<<
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):
 */
  __pyx_t_3 = PyString_Check(__pyx_v_other); 
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2105
 *         from .pyparsing import Each
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)             # <<<<<<<<<<<<<<
 *         if not isinstance(other, ParserElement):
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_literalStringClass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v_other) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_other);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_other, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2104
 *         """
 *         from .pyparsing import Each
 *         if isinstance(other, str):             # <<<<<<<<<<<<<<
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2106
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):             # <<<<<<<<<<<<<<
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 *                           SyntaxWarning, stacklevel=2)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyObject_IsInstance(__pyx_v_other, __pyx_t_2); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 2106, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = ((!(__pyx_t_4 != 0)) != 0);
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2107
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),             # <<<<<<<<<<<<<<
 *                           SyntaxWarning, stacklevel=2)
 *             return None
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_warnings); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2107, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_warn); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2107, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Cannot_combine_element_of_type_s, ((PyObject *)Py_TYPE(__pyx_v_other))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2107, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "Aspidites/_vendor/pyparsing_extension.py":2108
 *         if not isinstance(other, ParserElement):
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 *                           SyntaxWarning, stacklevel=2)             # <<<<<<<<<<<<<<
 *             return None
 *         return Each([self, other])
 */
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2107, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
    __Pyx_INCREF(__pyx_builtin_SyntaxWarning);
    __Pyx_GIVEREF(__pyx_builtin_SyntaxWarning);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_builtin_SyntaxWarning);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2108, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_stacklevel, __pyx_int_2) < 0) __PYX_ERR(0, 2108, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":2107
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),             # <<<<<<<<<<<<<<
 *                           SyntaxWarning, stacklevel=2)
 *             return None
 */
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_5, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2107, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2109
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 *                           SyntaxWarning, stacklevel=2)
 *             return None             # <<<<<<<<<<<<<<
 *         return Each([self, other])
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2106
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):             # <<<<<<<<<<<<<<
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 *                           SyntaxWarning, stacklevel=2)
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2110
 *                           SyntaxWarning, stacklevel=2)
 *             return None
 *         return Each([self, other])             # <<<<<<<<<<<<<<
 * 
 *     def __rand__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyList_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_v_self);
  __Pyx_INCREF(__pyx_v_other);
  __Pyx_GIVEREF(__pyx_v_other);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_v_other);
  __Pyx_INCREF(__pyx_v_Each);
  __pyx_t_5 = __pyx_v_Each; __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_6 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_1, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2099
 *         return other ^ self
 * 
 *     def __and__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of & operator - returns :class:`Each`
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__and__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_Each);
  __Pyx_XDECREF(__pyx_v_other);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2112
 *         return Each([self, other])
 * 
 *     def __rand__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of & operator when left operand is not a :class:`ParserElement`
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_81__rand__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_80__rand__[] = "\n        Implementation of & operator when left operand is not a :class:`ParserElement`\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_81__rand__ = {"__rand__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_81__rand__, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_80__rand__};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_81__rand__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_other = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__rand__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_other,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_other)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__rand__", 1, 2, 2, 1); __PYX_ERR(0, 2112, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__rand__") < 0)) __PYX_ERR(0, 2112, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_other = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__rand__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2112, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__rand__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_80__rand__(__pyx_self, __pyx_v_self, __pyx_v_other);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_80__rand__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__rand__", 0);
  __Pyx_INCREF(__pyx_v_other);

  /* "Aspidites/_vendor/pyparsing_extension.py":2116
 *         Implementation of & operator when left operand is not a :class:`ParserElement`
 *         """
 *         if isinstance(other, str):             # <<<<<<<<<<<<<<
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):
 */
  __pyx_t_1 = PyString_Check(__pyx_v_other); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2117
 *         """
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)             # <<<<<<<<<<<<<<
 *         if not isinstance(other, ParserElement):
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_literalStringClass); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_other) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_other);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_other, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2116
 *         Implementation of & operator when left operand is not a :class:`ParserElement`
 *         """
 *         if isinstance(other, str):             # <<<<<<<<<<<<<<
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2118
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):             # <<<<<<<<<<<<<<
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 *                           SyntaxWarning, stacklevel=2)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyObject_IsInstance(__pyx_v_other, __pyx_t_3); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 2118, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = ((!(__pyx_t_2 != 0)) != 0);
  if (__pyx_t_1) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2119
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),             # <<<<<<<<<<<<<<
 *                           SyntaxWarning, stacklevel=2)
 *             return None
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_warnings); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_warn); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Cannot_combine_element_of_type_s, ((PyObject *)Py_TYPE(__pyx_v_other))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "Aspidites/_vendor/pyparsing_extension.py":2120
 *         if not isinstance(other, ParserElement):
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 *                           SyntaxWarning, stacklevel=2)             # <<<<<<<<<<<<<<
 *             return None
 *         return other & self
 */
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3);
    __Pyx_INCREF(__pyx_builtin_SyntaxWarning);
    __Pyx_GIVEREF(__pyx_builtin_SyntaxWarning);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_builtin_SyntaxWarning);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2120, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_stacklevel, __pyx_int_2) < 0) __PYX_ERR(0, 2120, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":2119
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),             # <<<<<<<<<<<<<<
 *                           SyntaxWarning, stacklevel=2)
 *             return None
 */
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2121
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 *                           SyntaxWarning, stacklevel=2)
 *             return None             # <<<<<<<<<<<<<<
 *         return other & self
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2118
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 *         if not isinstance(other, ParserElement):             # <<<<<<<<<<<<<<
 *             warnings.warn("Cannot combine element of type %s with ParserElement" % type(other),
 *                           SyntaxWarning, stacklevel=2)
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2122
 *                           SyntaxWarning, stacklevel=2)
 *             return None
 *         return other & self             # <<<<<<<<<<<<<<
 * 
 *     def __invert__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_6 = PyNumber_And(__pyx_v_other, __pyx_v_self); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2112
 *         return Each([self, other])
 * 
 *     def __rand__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of & operator when left operand is not a :class:`ParserElement`
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__rand__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_other);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2124
 *         return other & self
 * 
 *     def __invert__(self):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of ~ operator - returns :class:`NotAny`
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_83__invert__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_82__invert__[] = "\n        Implementation of ~ operator - returns :class:`NotAny`\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_83__invert__ = {"__invert__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_83__invert__, METH_O, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_82__invert__};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_83__invert__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__invert__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_82__invert__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_82__invert__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_NotAny = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__invert__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2128
 *         Implementation of ~ operator - returns :class:`NotAny`
 *         """
 *         from .pyparsing import NotAny             # <<<<<<<<<<<<<<
 *         return NotAny(self)
 * 
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_NotAny);
  __Pyx_GIVEREF(__pyx_n_s_NotAny);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_NotAny);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_pyparsing, __pyx_t_1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_NotAny); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_NotAny = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2129
 *         """
 *         from .pyparsing import NotAny
 *         return NotAny(self)             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_NotAny);
  __pyx_t_1 = __pyx_v_NotAny; __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_v_self) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_self);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2124
 *         return other & self
 * 
 *     def __invert__(self):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of ~ operator - returns :class:`NotAny`
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__invert__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_NotAny);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2131
 *         return NotAny(self)
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         # must implement __iter__ to override legacy use of sequential access to __getitem__ to
 *         # iterate over a sequence
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_85__iter__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_85__iter__ = {"__iter__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_85__iter__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_85__iter__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_84__iter__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_84__iter__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__iter__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2134
 *         # must implement __iter__ to override legacy use of sequential access to __getitem__ to
 *         # iterate over a sequence
 *         raise TypeError('%r object is not iterable' % self.__class__.__name__)             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, key):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_name_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyString_FormatSafe(__pyx_kp_s_r_object_is_not_iterable, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_Raise(__pyx_t_2, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __PYX_ERR(0, 2134, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2131
 *         return NotAny(self)
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         # must implement __iter__ to override legacy use of sequential access to __getitem__ to
 *         # iterate over a sequence
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2136
 *         raise TypeError('%r object is not iterable' % self.__class__.__name__)
 * 
 *     def __getitem__(self, key):             # <<<<<<<<<<<<<<
 *         """
 *         use ``[]`` indexing notation as a short form for expression repetition:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_87__getitem__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_86__getitem__[] = "\n        use ``[]`` indexing notation as a short form for expression repetition:\n         - ``expr[n]`` is equivalent to ``expr*n``\n         - ``expr[m, n]`` is equivalent to ``expr*(m, n)``\n         - ``expr[n, ...]`` or ``expr[n,]`` is equivalent\n              to ``expr*n + ZeroOrMore(expr)``\n              (read as \"at least n instances of ``expr``\")\n         - ``expr[..., n]`` is equivalent to ``expr*(0, n)``\n              (read as \"0 to n instances of ``expr``\")\n         - ``expr[...]`` and ``expr[0, ...]`` are equivalent to ``ZeroOrMore(expr)``\n         - ``expr[1, ...]`` is equivalent to ``OneOrMore(expr)``\n         ``None`` may be used in place of ``...``.\n\n        Note that ``expr[..., n]`` and ``expr[m, n]``do not raise an exception\n        if more than ``n`` ``expr``s exist in the input stream.  If this behavior is\n        desired, then write ``expr[..., n] + ~expr``.\n       ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_87__getitem__ = {"__getitem__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_87__getitem__, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_86__getitem__};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_87__getitem__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_key = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_key,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_key)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__getitem__", 1, 2, 2, 1); __PYX_ERR(0, 2136, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__getitem__") < 0)) __PYX_ERR(0, 2136, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_key = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__getitem__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2136, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_86__getitem__(__pyx_self, __pyx_v_self, __pyx_v_key);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_86__getitem__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_v_ret = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  Py_ssize_t __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  Py_ssize_t __pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);
  __Pyx_INCREF(__pyx_v_key);

  /* "Aspidites/_vendor/pyparsing_extension.py":2156
 * 
 *         # convert single arg keys to tuples
 *         try:             # <<<<<<<<<<<<<<
 *             if isinstance(key, str):
 *                 key = (key,)
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "Aspidites/_vendor/pyparsing_extension.py":2157
 *         # convert single arg keys to tuples
 *         try:
 *             if isinstance(key, str):             # <<<<<<<<<<<<<<
 *                 key = (key,)
 *             iter(key)
 */
      __pyx_t_4 = PyString_Check(__pyx_v_key); 
      __pyx_t_5 = (__pyx_t_4 != 0);
      if (__pyx_t_5) {

        /* "Aspidites/_vendor/pyparsing_extension.py":2158
 *         try:
 *             if isinstance(key, str):
 *                 key = (key,)             # <<<<<<<<<<<<<<
 *             iter(key)
 *         except TypeError:
 */
        __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2158, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_INCREF(__pyx_v_key);
        __Pyx_GIVEREF(__pyx_v_key);
        PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_key);
        __Pyx_DECREF_SET(__pyx_v_key, __pyx_t_6);
        __pyx_t_6 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":2157
 *         # convert single arg keys to tuples
 *         try:
 *             if isinstance(key, str):             # <<<<<<<<<<<<<<
 *                 key = (key,)
 *             iter(key)
 */
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":2159
 *             if isinstance(key, str):
 *                 key = (key,)
 *             iter(key)             # <<<<<<<<<<<<<<
 *         except TypeError:
 *             key = (key, key)
 */
      __pyx_t_6 = PyObject_GetIter(__pyx_v_key); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2159, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":2156
 * 
 *         # convert single arg keys to tuples
 *         try:             # <<<<<<<<<<<<<<
 *             if isinstance(key, str):
 *                 key = (key,)
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2160
 *                 key = (key,)
 *             iter(key)
 *         except TypeError:             # <<<<<<<<<<<<<<
 *             key = (key, key)
 * 
 */
    __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
    if (__pyx_t_7) {
      __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_8, &__pyx_t_9) < 0) __PYX_ERR(0, 2160, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GOTREF(__pyx_t_9);

      /* "Aspidites/_vendor/pyparsing_extension.py":2161
 *             iter(key)
 *         except TypeError:
 *             key = (key, key)             # <<<<<<<<<<<<<<
 * 
 *         if len(key) > 2:
 */
      __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2161, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_v_key);
      __Pyx_GIVEREF(__pyx_v_key);
      PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_v_key);
      __Pyx_INCREF(__pyx_v_key);
      __Pyx_GIVEREF(__pyx_v_key);
      PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_v_key);
      __Pyx_DECREF_SET(__pyx_v_key, __pyx_t_10);
      __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "Aspidites/_vendor/pyparsing_extension.py":2156
 * 
 *         # convert single arg keys to tuples
 *         try:             # <<<<<<<<<<<<<<
 *             if isinstance(key, str):
 *                 key = (key,)
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L8_try_end:;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2163
 *             key = (key, key)
 * 
 *         if len(key) > 2:             # <<<<<<<<<<<<<<
 *             warnings.warn("only 1 or 2 index arguments supported ({0}{1})".format(key[:5],
 *                                                                                   '... [{0}]'.format(len(key))
 */
  __pyx_t_11 = PyObject_Length(__pyx_v_key); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2163, __pyx_L1_error)
  __pyx_t_5 = ((__pyx_t_11 > 2) != 0);
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2164
 * 
 *         if len(key) > 2:
 *             warnings.warn("only 1 or 2 index arguments supported ({0}{1})".format(key[:5],             # <<<<<<<<<<<<<<
 *                                                                                   '... [{0}]'.format(len(key))
 *                                                                                   if len(key) > 5 else ''))
 */
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_warnings); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_warn); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_only_1_or_2_index_arguments_supp, __pyx_n_s_format); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_12 = __Pyx_PyObject_GetSlice(__pyx_v_key, 0, 5, NULL, NULL, &__pyx_slice__78, 0, 1, 1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 2164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);

    /* "Aspidites/_vendor/pyparsing_extension.py":2166
 *             warnings.warn("only 1 or 2 index arguments supported ({0}{1})".format(key[:5],
 *                                                                                   '... [{0}]'.format(len(key))
 *                                                                                   if len(key) > 5 else ''))             # <<<<<<<<<<<<<<
 * 
 *         # clip to 2 elements
 */
    __pyx_t_11 = PyObject_Length(__pyx_v_key); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2166, __pyx_L1_error)
    if (((__pyx_t_11 > 5) != 0)) {

      /* "Aspidites/_vendor/pyparsing_extension.py":2165
 *         if len(key) > 2:
 *             warnings.warn("only 1 or 2 index arguments supported ({0}{1})".format(key[:5],
 *                                                                                   '... [{0}]'.format(len(key))             # <<<<<<<<<<<<<<
 *                                                                                   if len(key) > 5 else ''))
 * 
 */
      __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_0_2, __pyx_n_s_format); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 2165, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __pyx_t_16 = PyObject_Length(__pyx_v_key); if (unlikely(__pyx_t_16 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2165, __pyx_L1_error)
      __pyx_t_17 = PyInt_FromSsize_t(__pyx_t_16); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 2165, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      __pyx_t_18 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_15))) {
        __pyx_t_18 = PyMethod_GET_SELF(__pyx_t_15);
        if (likely(__pyx_t_18)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
          __Pyx_INCREF(__pyx_t_18);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_15, function);
        }
      }
      __pyx_t_14 = (__pyx_t_18) ? __Pyx_PyObject_Call2Args(__pyx_t_15, __pyx_t_18, __pyx_t_17) : __Pyx_PyObject_CallOneArg(__pyx_t_15, __pyx_t_17);
      __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
      if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2165, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      __pyx_t_13 = __pyx_t_14;
      __pyx_t_14 = 0;
    } else {
      __Pyx_INCREF(__pyx_kp_s__12);
      __pyx_t_13 = __pyx_kp_s__12;
    }
    __pyx_t_14 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
      __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_10);
      if (likely(__pyx_t_14)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_14);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_10, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_10)) {
      PyObject *__pyx_temp[3] = {__pyx_t_14, __pyx_t_12, __pyx_t_13};
      __pyx_t_8 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2164, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
      PyObject *__pyx_temp[3] = {__pyx_t_14, __pyx_t_12, __pyx_t_13};
      __pyx_t_8 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2164, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    } else
    #endif
    {
      __pyx_t_15 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 2164, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      if (__pyx_t_14) {
        __Pyx_GIVEREF(__pyx_t_14); PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_14); __pyx_t_14 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_12);
      PyTuple_SET_ITEM(__pyx_t_15, 0+__pyx_t_7, __pyx_t_12);
      __Pyx_GIVEREF(__pyx_t_13);
      PyTuple_SET_ITEM(__pyx_t_15, 1+__pyx_t_7, __pyx_t_13);
      __pyx_t_12 = 0;
      __pyx_t_13 = 0;
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_15, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2164, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    }
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_9 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_10, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_8);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2163
 *             key = (key, key)
 * 
 *         if len(key) > 2:             # <<<<<<<<<<<<<<
 *             warnings.warn("only 1 or 2 index arguments supported ({0}{1})".format(key[:5],
 *                                                                                   '... [{0}]'.format(len(key))
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2169
 * 
 *         # clip to 2 elements
 *         ret = self * tuple(key[:2])             # <<<<<<<<<<<<<<
 *         return ret
 * 
 */
  __pyx_t_9 = __Pyx_PyObject_GetSlice(__pyx_v_key, 0, 2, NULL, NULL, &__pyx_slice__71, 0, 1, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_6 = __Pyx_PySequence_Tuple(__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = PyNumber_Multiply(__pyx_v_self, __pyx_t_6); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_ret = __pyx_t_9;
  __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2170
 *         # clip to 2 elements
 *         ret = self * tuple(key[:2])
 *         return ret             # <<<<<<<<<<<<<<
 * 
 *     def __call__(self, name=None):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_ret);
  __pyx_r = __pyx_v_ret;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2136
 *         raise TypeError('%r object is not iterable' % self.__class__.__name__)
 * 
 *     def __getitem__(self, key):             # <<<<<<<<<<<<<<
 *         """
 *         use ``[]`` indexing notation as a short form for expression repetition:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2172
 *         return ret
 * 
 *     def __call__(self, name=None):             # <<<<<<<<<<<<<<
 *         """
 *         Shortcut for :class:`setResultsName`, with ``listAllMatches=False``.
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_89__call__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_88__call__[] = "\n        Shortcut for :class:`setResultsName`, with ``listAllMatches=False``.\n\n        If ``name`` is given with a trailing ``'*'`` character, then ``listAllMatches`` will be\n        passed as ``True``.\n\n        If ``name` is omitted, same as calling :class:`copy`.\n\n        Example::\n\n            # these are equivalent\n            userdata = Word(alphas).setResultsName(\"name\") + Word(nums + \"-\").setResultsName(\"socsecno\")\n            userdata = Word(alphas)(\"name\") + Word(nums + \"-\")(\"socsecno\")\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_89__call__ = {"__call__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_89__call__, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_88__call__};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_89__call__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_name = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__call__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_name,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__call__") < 0)) __PYX_ERR(0, 2172, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_name = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__call__", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2172, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__call__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_88__call__(__pyx_self, __pyx_v_self, __pyx_v_name);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_88__call__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_name) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__call__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2187
 *             userdata = Word(alphas)("name") + Word(nums + "-")("socsecno")
 *         """
 *         if name is not None:             # <<<<<<<<<<<<<<
 *             return self._setResultsName(name)
 *         else:
 */
  __pyx_t_1 = (__pyx_v_name != Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2188
 *         """
 *         if name is not None:
 *             return self._setResultsName(name)             # <<<<<<<<<<<<<<
 *         else:
 *             return self.copy()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_setResultsName); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2188, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_name) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_name);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2188, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2187
 *             userdata = Word(alphas)("name") + Word(nums + "-")("socsecno")
 *         """
 *         if name is not None:             # <<<<<<<<<<<<<<
 *             return self._setResultsName(name)
 *         else:
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2190
 *             return self._setResultsName(name)
 *         else:
 *             return self.copy()             # <<<<<<<<<<<<<<
 * 
 *     def suppress(self):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_copy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2172
 *         return ret
 * 
 *     def __call__(self, name=None):             # <<<<<<<<<<<<<<
 *         """
 *         Shortcut for :class:`setResultsName`, with ``listAllMatches=False``.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__call__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2192
 *             return self.copy()
 * 
 *     def suppress(self):             # <<<<<<<<<<<<<<
 *         """
 *         Suppresses the output of this :class:`ParserElement`; useful to keep punctuation from
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_91suppress(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_90suppress[] = "\n        Suppresses the output of this :class:`ParserElement`; useful to keep punctuation from\n        cluttering up returned output.\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_91suppress = {"suppress", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_91suppress, METH_O, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_90suppress};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_91suppress(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("suppress (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_90suppress(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_90suppress(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_Suppress = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("suppress", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2197
 *         cluttering up returned output.
 *         """
 *         from .pyparsing import Suppress             # <<<<<<<<<<<<<<
 *         return Suppress(self)
 * 
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_Suppress);
  __Pyx_GIVEREF(__pyx_n_s_Suppress);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_Suppress);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_pyparsing, __pyx_t_1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_Suppress); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_Suppress = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2198
 *         """
 *         from .pyparsing import Suppress
 *         return Suppress(self)             # <<<<<<<<<<<<<<
 * 
 *     def leaveWhitespace(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_Suppress);
  __pyx_t_1 = __pyx_v_Suppress; __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_v_self) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_self);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2192
 *             return self.copy()
 * 
 *     def suppress(self):             # <<<<<<<<<<<<<<
 *         """
 *         Suppresses the output of this :class:`ParserElement`; useful to keep punctuation from
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.suppress", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_Suppress);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2200
 *         return Suppress(self)
 * 
 *     def leaveWhitespace(self):             # <<<<<<<<<<<<<<
 *         """
 *         Disables the skipping of whitespace before matching the characters in the
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_93leaveWhitespace(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_92leaveWhitespace[] = "\n        Disables the skipping of whitespace before matching the characters in the\n        :class:`ParserElement`'s defined pattern.  This is normally only used internally by\n        the pyparsing module, but may be needed in some whitespace-sensitive grammars.\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_93leaveWhitespace = {"leaveWhitespace", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_93leaveWhitespace, METH_O, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_92leaveWhitespace};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_93leaveWhitespace(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("leaveWhitespace (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_92leaveWhitespace(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_92leaveWhitespace(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("leaveWhitespace", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2206
 *         the pyparsing module, but may be needed in some whitespace-sensitive grammars.
 *         """
 *         self.skipWhitespace = False             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_skipWhitespace, Py_False) < 0) __PYX_ERR(0, 2206, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2207
 *         """
 *         self.skipWhitespace = False
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def setWhitespaceChars(self, chars):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2200
 *         return Suppress(self)
 * 
 *     def leaveWhitespace(self):             # <<<<<<<<<<<<<<
 *         """
 *         Disables the skipping of whitespace before matching the characters in the
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.leaveWhitespace", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2209
 *         return self
 * 
 *     def setWhitespaceChars(self, chars):             # <<<<<<<<<<<<<<
 *         """
 *         Overrides the default whitespace chars
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_95setWhitespaceChars(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_94setWhitespaceChars[] = "\n        Overrides the default whitespace chars\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_95setWhitespaceChars = {"setWhitespaceChars", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_95setWhitespaceChars, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_94setWhitespaceChars};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_95setWhitespaceChars(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_chars = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setWhitespaceChars (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_chars,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chars)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setWhitespaceChars", 1, 2, 2, 1); __PYX_ERR(0, 2209, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setWhitespaceChars") < 0)) __PYX_ERR(0, 2209, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_chars = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setWhitespaceChars", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2209, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.setWhitespaceChars", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_94setWhitespaceChars(__pyx_self, __pyx_v_self, __pyx_v_chars);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_94setWhitespaceChars(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_chars) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setWhitespaceChars", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2213
 *         Overrides the default whitespace chars
 *         """
 *         self.skipWhitespace = True             # <<<<<<<<<<<<<<
 *         self.whiteChars = chars
 *         self.copyDefaultWhiteChars = False
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_skipWhitespace, Py_True) < 0) __PYX_ERR(0, 2213, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2214
 *         """
 *         self.skipWhitespace = True
 *         self.whiteChars = chars             # <<<<<<<<<<<<<<
 *         self.copyDefaultWhiteChars = False
 *         return self
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_whiteChars, __pyx_v_chars) < 0) __PYX_ERR(0, 2214, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2215
 *         self.skipWhitespace = True
 *         self.whiteChars = chars
 *         self.copyDefaultWhiteChars = False             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_copyDefaultWhiteChars, Py_False) < 0) __PYX_ERR(0, 2215, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2216
 *         self.whiteChars = chars
 *         self.copyDefaultWhiteChars = False
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def parseWithTabs(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2209
 *         return self
 * 
 *     def setWhitespaceChars(self, chars):             # <<<<<<<<<<<<<<
 *         """
 *         Overrides the default whitespace chars
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.setWhitespaceChars", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2218
 *         return self
 * 
 *     def parseWithTabs(self):             # <<<<<<<<<<<<<<
 *         """
 *         Overrides default behavior to expand ``<TAB>``s to spaces before parsing the input string.
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_97parseWithTabs(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_96parseWithTabs[] = "\n        Overrides default behavior to expand ``<TAB>``s to spaces before parsing the input string.\n        Must be called before ``parseString`` when the input grammar contains elements that\n        match ``<TAB>`` characters.\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_97parseWithTabs = {"parseWithTabs", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_97parseWithTabs, METH_O, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_96parseWithTabs};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_97parseWithTabs(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parseWithTabs (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_96parseWithTabs(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_96parseWithTabs(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parseWithTabs", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2224
 *         match ``<TAB>`` characters.
 *         """
 *         self.keepTabs = True             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_keepTabs, Py_True) < 0) __PYX_ERR(0, 2224, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2225
 *         """
 *         self.keepTabs = True
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def ignore(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2218
 *         return self
 * 
 *     def parseWithTabs(self):             # <<<<<<<<<<<<<<
 *         """
 *         Overrides default behavior to expand ``<TAB>``s to spaces before parsing the input string.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.parseWithTabs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2227
 *         return self
 * 
 *     def ignore(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Define expression to be ignored (e.g., comments) while doing pattern
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_99ignore(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_98ignore[] = "\n        Define expression to be ignored (e.g., comments) while doing pattern\n        matching; may be called repeatedly, to define multiple comment or other\n        ignorable patterns.\n\n        Example::\n\n            patt = OneOrMore(Word(alphas))\n            patt.parseString('ablaj /* comment */ lskjd') # -> ['ablaj']\n\n            patt.ignore(cStyleComment)\n            patt.parseString('ablaj /* comment */ lskjd') # -> ['ablaj', 'lskjd']\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_99ignore = {"ignore", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_99ignore, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_98ignore};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_99ignore(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_other = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ignore (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_other,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_other)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ignore", 1, 2, 2, 1); __PYX_ERR(0, 2227, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "ignore") < 0)) __PYX_ERR(0, 2227, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_other = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ignore", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2227, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.ignore", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_98ignore(__pyx_self, __pyx_v_self, __pyx_v_other);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_98ignore(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_v_Suppress = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ignore", 0);
  __Pyx_INCREF(__pyx_v_other);

  /* "Aspidites/_vendor/pyparsing_extension.py":2241
 *             patt.parseString('ablaj /[inserted by cython to avoid comment start]* comment *[inserted by cython to avoid comment closer]/ lskjd') # -> ['ablaj', 'lskjd']
 *         """
 *         from .pyparsing import Suppress             # <<<<<<<<<<<<<<
 *         if isinstance(other, str):
 *             other = Suppress(other)
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_Suppress);
  __Pyx_GIVEREF(__pyx_n_s_Suppress);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_Suppress);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_pyparsing, __pyx_t_1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_Suppress); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_Suppress = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2242
 *         """
 *         from .pyparsing import Suppress
 *         if isinstance(other, str):             # <<<<<<<<<<<<<<
 *             other = Suppress(other)
 * 
 */
  __pyx_t_3 = PyString_Check(__pyx_v_other); 
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2243
 *         from .pyparsing import Suppress
 *         if isinstance(other, str):
 *             other = Suppress(other)             # <<<<<<<<<<<<<<
 * 
 *         if isinstance(other, Suppress):
 */
    __Pyx_INCREF(__pyx_v_Suppress);
    __pyx_t_1 = __pyx_v_Suppress; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v_other) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_other);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2243, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_other, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2242
 *         """
 *         from .pyparsing import Suppress
 *         if isinstance(other, str):             # <<<<<<<<<<<<<<
 *             other = Suppress(other)
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2245
 *             other = Suppress(other)
 * 
 *         if isinstance(other, Suppress):             # <<<<<<<<<<<<<<
 *             if other not in self.ignoreExprs:
 *                 self.ignoreExprs.append(other)
 */
  __pyx_t_4 = PyObject_IsInstance(__pyx_v_other, __pyx_v_Suppress); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 2245, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_4 != 0);
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2246
 * 
 *         if isinstance(other, Suppress):
 *             if other not in self.ignoreExprs:             # <<<<<<<<<<<<<<
 *                 self.ignoreExprs.append(other)
 *         else:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ignoreExprs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2246, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = (__Pyx_PySequence_ContainsTF(__pyx_v_other, __pyx_t_2, Py_NE)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 2246, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = (__pyx_t_3 != 0);
    if (__pyx_t_4) {

      /* "Aspidites/_vendor/pyparsing_extension.py":2247
 *         if isinstance(other, Suppress):
 *             if other not in self.ignoreExprs:
 *                 self.ignoreExprs.append(other)             # <<<<<<<<<<<<<<
 *         else:
 *             self.ignoreExprs.append(Suppress(other.copy()))
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ignoreExprs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2247, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = __Pyx_PyObject_Append(__pyx_t_2, __pyx_v_other); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 2247, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":2246
 * 
 *         if isinstance(other, Suppress):
 *             if other not in self.ignoreExprs:             # <<<<<<<<<<<<<<
 *                 self.ignoreExprs.append(other)
 *         else:
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":2245
 *             other = Suppress(other)
 * 
 *         if isinstance(other, Suppress):             # <<<<<<<<<<<<<<
 *             if other not in self.ignoreExprs:
 *                 self.ignoreExprs.append(other)
 */
    goto __pyx_L4;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2249
 *                 self.ignoreExprs.append(other)
 *         else:
 *             self.ignoreExprs.append(Suppress(other.copy()))             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  /*else*/ {
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ignoreExprs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_copy); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_5 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_INCREF(__pyx_v_Suppress);
    __pyx_t_7 = __pyx_v_Suppress; __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_6 = __Pyx_PyObject_Append(__pyx_t_2, __pyx_t_1); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 2249, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_L4:;

  /* "Aspidites/_vendor/pyparsing_extension.py":2250
 *         else:
 *             self.ignoreExprs.append(Suppress(other.copy()))
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def setDebugActions(self, startAction, successAction, exceptionAction):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2227
 *         return self
 * 
 *     def ignore(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Define expression to be ignored (e.g., comments) while doing pattern
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.ignore", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_Suppress);
  __Pyx_XDECREF(__pyx_v_other);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2252
 *         return self
 * 
 *     def setDebugActions(self, startAction, successAction, exceptionAction):             # <<<<<<<<<<<<<<
 *         """
 *         Enable display of debugging messages while doing pattern matching.
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_101setDebugActions(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_100setDebugActions[] = "\n        Enable display of debugging messages while doing pattern matching.\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_101setDebugActions = {"setDebugActions", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_101setDebugActions, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_100setDebugActions};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_101setDebugActions(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_startAction = 0;
  PyObject *__pyx_v_successAction = 0;
  PyObject *__pyx_v_exceptionAction = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setDebugActions (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_startAction,&__pyx_n_s_successAction,&__pyx_n_s_exceptionAction,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_startAction)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setDebugActions", 1, 4, 4, 1); __PYX_ERR(0, 2252, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_successAction)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setDebugActions", 1, 4, 4, 2); __PYX_ERR(0, 2252, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_exceptionAction)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setDebugActions", 1, 4, 4, 3); __PYX_ERR(0, 2252, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setDebugActions") < 0)) __PYX_ERR(0, 2252, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_self = values[0];
    __pyx_v_startAction = values[1];
    __pyx_v_successAction = values[2];
    __pyx_v_exceptionAction = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setDebugActions", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2252, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.setDebugActions", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_100setDebugActions(__pyx_self, __pyx_v_self, __pyx_v_startAction, __pyx_v_successAction, __pyx_v_exceptionAction);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_100setDebugActions(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_startAction, PyObject *__pyx_v_successAction, PyObject *__pyx_v_exceptionAction) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setDebugActions", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2256
 *         Enable display of debugging messages while doing pattern matching.
 *         """
 *         self.debugActions = (startAction or _defaultStartDebugAction,             # <<<<<<<<<<<<<<
 *                              successAction or _defaultSuccessDebugAction,
 *                              exceptionAction or _defaultExceptionDebugAction)
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_startAction); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 2256, __pyx_L1_error)
  if (!__pyx_t_2) {
  } else {
    __Pyx_INCREF(__pyx_v_startAction);
    __pyx_t_1 = __pyx_v_startAction;
    goto __pyx_L3_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_defaultStartDebugAction); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_1 = __pyx_t_3;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_L3_bool_binop_done:;

  /* "Aspidites/_vendor/pyparsing_extension.py":2257
 *         """
 *         self.debugActions = (startAction or _defaultStartDebugAction,
 *                              successAction or _defaultSuccessDebugAction,             # <<<<<<<<<<<<<<
 *                              exceptionAction or _defaultExceptionDebugAction)
 *         self.debug = True
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_successAction); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 2257, __pyx_L1_error)
  if (!__pyx_t_2) {
  } else {
    __Pyx_INCREF(__pyx_v_successAction);
    __pyx_t_3 = __pyx_v_successAction;
    goto __pyx_L5_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_defaultSuccessDebugAction); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_3 = __pyx_t_4;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_L5_bool_binop_done:;

  /* "Aspidites/_vendor/pyparsing_extension.py":2258
 *         self.debugActions = (startAction or _defaultStartDebugAction,
 *                              successAction or _defaultSuccessDebugAction,
 *                              exceptionAction or _defaultExceptionDebugAction)             # <<<<<<<<<<<<<<
 *         self.debug = True
 *         return self
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_exceptionAction); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 2258, __pyx_L1_error)
  if (!__pyx_t_2) {
  } else {
    __Pyx_INCREF(__pyx_v_exceptionAction);
    __pyx_t_4 = __pyx_v_exceptionAction;
    goto __pyx_L7_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_defaultExceptionDebugAction); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_t_5);
  __pyx_t_4 = __pyx_t_5;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_L7_bool_binop_done:;

  /* "Aspidites/_vendor/pyparsing_extension.py":2256
 *         Enable display of debugging messages while doing pattern matching.
 *         """
 *         self.debugActions = (startAction or _defaultStartDebugAction,             # <<<<<<<<<<<<<<
 *                              successAction or _defaultSuccessDebugAction,
 *                              exceptionAction or _defaultExceptionDebugAction)
 */
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_4);
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_debugActions, __pyx_t_5) < 0) __PYX_ERR(0, 2256, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2259
 *                              successAction or _defaultSuccessDebugAction,
 *                              exceptionAction or _defaultExceptionDebugAction)
 *         self.debug = True             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_debug, Py_True) < 0) __PYX_ERR(0, 2259, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2260
 *                              exceptionAction or _defaultExceptionDebugAction)
 *         self.debug = True
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def setDebug(self, flag=True):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2252
 *         return self
 * 
 *     def setDebugActions(self, startAction, successAction, exceptionAction):             # <<<<<<<<<<<<<<
 *         """
 *         Enable display of debugging messages while doing pattern matching.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.setDebugActions", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2262
 *         return self
 * 
 *     def setDebug(self, flag=True):             # <<<<<<<<<<<<<<
 *         """
 *         Enable display of debugging messages while doing pattern matching.
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_103setDebug(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_102setDebug[] = "\n        Enable display of debugging messages while doing pattern matching.\n        Set ``flag`` to True to enable, False to disable.\n\n        Example::\n\n            wd = Word(alphas).setName(\"alphaword\")\n            integer = Word(nums).setName(\"numword\")\n            term = wd | integer\n\n            # turn on debugging for wd\n            wd.setDebug()\n\n            OneOrMore(term).parseString(\"abc 123 xyz 890\")\n\n        prints::\n\n            Match alphaword at loc 0(1,1)\n            Matched alphaword -> ['abc']\n            Match alphaword at loc 3(1,4)\n            Exception raised:Expected alphaword (at char 4), (line:1, col:5)\n            Match alphaword at loc 7(1,8)\n            Matched alphaword -> ['xyz']\n            Match alphaword at loc 11(1,12)\n            Exception raised:Expected alphaword (at char 12), (line:1, col:13)\n            Match alphaword at loc 15(1,16)\n            Exception raised:Expected alphaword (at char 15), (line:1, col:16)\n\n        The output shown is that produced by the default debug actions - custom debug actions can be\n        specified using :class:`setDebugActions`. Prior to attempting\n        to match the ``wd`` expression, the debugging message ``\"Match <exprname> at loc <n>(<line>,<col>)\"``\n        is shown. Then if the parse succeeds, a ``\"Matched\"`` message is shown, or an ``\"Exception raised\"``\n        message is shown. Also note the use of :class:`setName` to assign a human-readable name to the expression,\n        which makes debugging and exception messages easier to understand - for instance, the default\n        name created for the :class:`Word` expression without calling ``setName`` is ``\"W:(ABCD...)\"``.\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_103setDebug = {"setDebug", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_103setDebug, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_102setDebug};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_103setDebug(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_flag = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setDebug (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_flag,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_flag);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setDebug") < 0)) __PYX_ERR(0, 2262, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_flag = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setDebug", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2262, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.setDebug", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_102setDebug(__pyx_self, __pyx_v_self, __pyx_v_flag);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_102setDebug(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_flag) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setDebug", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2299
 *         name created for the :class:`Word` expression without calling ``setName`` is ``"W:(ABCD...)"``.
 *         """
 *         if flag:             # <<<<<<<<<<<<<<
 *             self.setDebugActions(_defaultStartDebugAction, _defaultSuccessDebugAction, _defaultExceptionDebugAction)
 *         else:
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_flag); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 2299, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2300
 *         """
 *         if flag:
 *             self.setDebugActions(_defaultStartDebugAction, _defaultSuccessDebugAction, _defaultExceptionDebugAction)             # <<<<<<<<<<<<<<
 *         else:
 *             self.debug = False
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_setDebugActions); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2300, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_defaultStartDebugAction); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2300, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_defaultSuccessDebugAction); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2300, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_defaultExceptionDebugAction); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2300, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[4] = {__pyx_t_7, __pyx_t_4, __pyx_t_5, __pyx_t_6};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_8, 3+__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2300, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[4] = {__pyx_t_7, __pyx_t_4, __pyx_t_5, __pyx_t_6};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_8, 3+__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2300, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_9 = PyTuple_New(3+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2300, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_9, 2+__pyx_t_8, __pyx_t_6);
      __pyx_t_4 = 0;
      __pyx_t_5 = 0;
      __pyx_t_6 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_9, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2300, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2299
 *         name created for the :class:`Word` expression without calling ``setName`` is ``"W:(ABCD...)"``.
 *         """
 *         if flag:             # <<<<<<<<<<<<<<
 *             self.setDebugActions(_defaultStartDebugAction, _defaultSuccessDebugAction, _defaultExceptionDebugAction)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2302
 *             self.setDebugActions(_defaultStartDebugAction, _defaultSuccessDebugAction, _defaultExceptionDebugAction)
 *         else:
 *             self.debug = False             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  /*else*/ {
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_debug, Py_False) < 0) __PYX_ERR(0, 2302, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "Aspidites/_vendor/pyparsing_extension.py":2303
 *         else:
 *             self.debug = False
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2262
 *         return self
 * 
 *     def setDebug(self, flag=True):             # <<<<<<<<<<<<<<
 *         """
 *         Enable display of debugging messages while doing pattern matching.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.setDebug", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2305
 *         return self
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return self.name
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_105__str__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_105__str__ = {"__str__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_105__str__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_105__str__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_104__str__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_104__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2306
 * 
 *     def __str__(self):
 *         return self.name             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2305
 *         return self
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return self.name
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2308
 *         return self.name
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return str(self)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_107__repr__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_107__repr__ = {"__repr__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_107__repr__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_107__repr__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_106__repr__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_106__repr__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2309
 * 
 *     def __repr__(self):
 *         return str(self)             # <<<<<<<<<<<<<<
 * 
 *     def streamline(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2308
 *         return self.name
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return str(self)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2311
 *         return str(self)
 * 
 *     def streamline(self):             # <<<<<<<<<<<<<<
 *         self.streamlined = True
 *         self.strRepr = None
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_109streamline(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_109streamline = {"streamline", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_109streamline, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_109streamline(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("streamline (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_108streamline(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_108streamline(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("streamline", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2312
 * 
 *     def streamline(self):
 *         self.streamlined = True             # <<<<<<<<<<<<<<
 *         self.strRepr = None
 *         return self
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_streamlined, Py_True) < 0) __PYX_ERR(0, 2312, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2313
 *     def streamline(self):
 *         self.streamlined = True
 *         self.strRepr = None             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_strRepr, Py_None) < 0) __PYX_ERR(0, 2313, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2314
 *         self.streamlined = True
 *         self.strRepr = None
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def checkRecursion(self, parseElementList):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2311
 *         return str(self)
 * 
 *     def streamline(self):             # <<<<<<<<<<<<<<
 *         self.streamlined = True
 *         self.strRepr = None
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.streamline", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2316
 *         return self
 * 
 *     def checkRecursion(self, parseElementList):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_111checkRecursion(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_111checkRecursion = {"checkRecursion", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_111checkRecursion, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_111checkRecursion(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_parseElementList = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("checkRecursion (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parseElementList,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_parseElementList)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("checkRecursion", 1, 2, 2, 1); __PYX_ERR(0, 2316, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "checkRecursion") < 0)) __PYX_ERR(0, 2316, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_parseElementList = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("checkRecursion", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2316, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.checkRecursion", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_110checkRecursion(__pyx_self, __pyx_v_self, __pyx_v_parseElementList);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_110checkRecursion(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_parseElementList) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("checkRecursion", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2319
 *         pass
 * 
 *     def validate(self, validateTrace=None):             # <<<<<<<<<<<<<<
 *         """
 *         Check defined expressions for valid structure, check for infinite recursive definitions.
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_113validate(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_112validate[] = "\n        Check defined expressions for valid structure, check for infinite recursive definitions.\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_113validate = {"validate", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_113validate, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_112validate};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_113validate(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_validateTrace = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("validate (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_validateTrace,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_validateTrace);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "validate") < 0)) __PYX_ERR(0, 2319, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_validateTrace = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("validate", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2319, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_112validate(__pyx_self, __pyx_v_self, __pyx_v_validateTrace);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_112validate(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_validateTrace) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("validate", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2323
 *         Check defined expressions for valid structure, check for infinite recursive definitions.
 *         """
 *         self.checkRecursion([])             # <<<<<<<<<<<<<<
 * 
 *     def parseFile(self, file_or_filename, parseAll=False):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_checkRecursion); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2319
 *         pass
 * 
 *     def validate(self, validateTrace=None):             # <<<<<<<<<<<<<<
 *         """
 *         Check defined expressions for valid structure, check for infinite recursive definitions.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2325
 *         self.checkRecursion([])
 * 
 *     def parseFile(self, file_or_filename, parseAll=False):             # <<<<<<<<<<<<<<
 *         """
 *         Execute the parse expression on the given file or filename.
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_115parseFile(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_114parseFile[] = "\n        Execute the parse expression on the given file or filename.\n        If a filename is specified (instead of a file object),\n        the entire file is opened, read, and closed before parsing.\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_115parseFile = {"parseFile", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_115parseFile, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_114parseFile};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_115parseFile(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_file_or_filename = 0;
  PyObject *__pyx_v_parseAll = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parseFile (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_file_or_filename,&__pyx_n_s_parseAll,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_False));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_file_or_filename)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseFile", 0, 2, 3, 1); __PYX_ERR(0, 2325, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_parseAll);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "parseFile") < 0)) __PYX_ERR(0, 2325, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_file_or_filename = values[1];
    __pyx_v_parseAll = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("parseFile", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2325, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.parseFile", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_114parseFile(__pyx_self, __pyx_v_self, __pyx_v_file_or_filename, __pyx_v_parseAll);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_114parseFile(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_file_or_filename, PyObject *__pyx_v_parseAll) {
  PyObject *__pyx_v_file_contents = NULL;
  PyObject *__pyx_v_f = NULL;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  int __pyx_t_17;
  int __pyx_t_18;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parseFile", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2331
 *         the entire file is opened, read, and closed before parsing.
 *         """
 *         try:             # <<<<<<<<<<<<<<
 *             file_contents = file_or_filename.read()
 *         except AttributeError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "Aspidites/_vendor/pyparsing_extension.py":2332
 *         """
 *         try:
 *             file_contents = file_or_filename.read()             # <<<<<<<<<<<<<<
 *         except AttributeError:
 *             with open(file_or_filename, "r") as f:
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_file_or_filename, __pyx_n_s_read); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2332, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2332, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_v_file_contents = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":2331
 *         the entire file is opened, read, and closed before parsing.
 *         """
 *         try:             # <<<<<<<<<<<<<<
 *             file_contents = file_or_filename.read()
 *         except AttributeError:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2333
 *         try:
 *             file_contents = file_or_filename.read()
 *         except AttributeError:             # <<<<<<<<<<<<<<
 *             with open(file_or_filename, "r") as f:
 *                 file_contents = f.read()
 */
    __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError);
    if (__pyx_t_7) {
      __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.parseFile", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6) < 0) __PYX_ERR(0, 2333, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);

      /* "Aspidites/_vendor/pyparsing_extension.py":2334
 *             file_contents = file_or_filename.read()
 *         except AttributeError:
 *             with open(file_or_filename, "r") as f:             # <<<<<<<<<<<<<<
 *                 file_contents = f.read()
 *         try:
 */
      /*with:*/ {
        __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2334, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_INCREF(__pyx_v_file_or_filename);
        __Pyx_GIVEREF(__pyx_v_file_or_filename);
        PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_file_or_filename);
        __Pyx_INCREF(__pyx_n_s_r);
        __Pyx_GIVEREF(__pyx_n_s_r);
        PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_n_s_r);
        __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_open, __pyx_t_8, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2334, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_10 = __Pyx_PyObject_LookupSpecial(__pyx_t_9, __pyx_n_s_exit); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2334, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_11 = __Pyx_PyObject_LookupSpecial(__pyx_t_9, __pyx_n_s_enter); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2334, __pyx_L11_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_12 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
          __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
          if (likely(__pyx_t_12)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
            __Pyx_INCREF(__pyx_t_12);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_11, function);
          }
        }
        __pyx_t_8 = (__pyx_t_12) ? __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_12) : __Pyx_PyObject_CallNoArg(__pyx_t_11);
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2334, __pyx_L11_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_11 = __pyx_t_8;
        __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        /*try:*/ {
          {
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __Pyx_ExceptionSave(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_13);
            __Pyx_XGOTREF(__pyx_t_14);
            __Pyx_XGOTREF(__pyx_t_15);
            /*try:*/ {
              __pyx_v_f = __pyx_t_11;
              __pyx_t_11 = 0;

              /* "Aspidites/_vendor/pyparsing_extension.py":2335
 *         except AttributeError:
 *             with open(file_or_filename, "r") as f:
 *                 file_contents = f.read()             # <<<<<<<<<<<<<<
 *         try:
 *             return self.parseString(file_contents, parseAll)
 */
              __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_read); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2335, __pyx_L17_error)
              __Pyx_GOTREF(__pyx_t_9);
              __pyx_t_8 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
                __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_9);
                if (likely(__pyx_t_8)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
                  __Pyx_INCREF(__pyx_t_8);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_9, function);
                }
              }
              __pyx_t_11 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_9);
              __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
              if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2335, __pyx_L17_error)
              __Pyx_GOTREF(__pyx_t_11);
              __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
              __Pyx_XDECREF_SET(__pyx_v_file_contents, __pyx_t_11);
              __pyx_t_11 = 0;

              /* "Aspidites/_vendor/pyparsing_extension.py":2334
 *             file_contents = file_or_filename.read()
 *         except AttributeError:
 *             with open(file_or_filename, "r") as f:             # <<<<<<<<<<<<<<
 *                 file_contents = f.read()
 *         try:
 */
            }
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
            __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
            goto __pyx_L24_try_end;
            __pyx_L17_error:;
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
            /*except:*/ {
              __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.parseFile", __pyx_clineno, __pyx_lineno, __pyx_filename);
              if (__Pyx_GetException(&__pyx_t_11, &__pyx_t_9, &__pyx_t_8) < 0) __PYX_ERR(0, 2334, __pyx_L19_except_error)
              __Pyx_GOTREF(__pyx_t_11);
              __Pyx_GOTREF(__pyx_t_9);
              __Pyx_GOTREF(__pyx_t_8);
              __pyx_t_12 = PyTuple_Pack(3, __pyx_t_11, __pyx_t_9, __pyx_t_8); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 2334, __pyx_L19_except_error)
              __Pyx_GOTREF(__pyx_t_12);
              __pyx_t_16 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_12, NULL);
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
              if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 2334, __pyx_L19_except_error)
              __Pyx_GOTREF(__pyx_t_16);
              __pyx_t_17 = __Pyx_PyObject_IsTrue(__pyx_t_16);
              __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
              if (__pyx_t_17 < 0) __PYX_ERR(0, 2334, __pyx_L19_except_error)
              __pyx_t_18 = ((!(__pyx_t_17 != 0)) != 0);
              if (__pyx_t_18) {
                __Pyx_GIVEREF(__pyx_t_11);
                __Pyx_GIVEREF(__pyx_t_9);
                __Pyx_XGIVEREF(__pyx_t_8);
                __Pyx_ErrRestoreWithState(__pyx_t_11, __pyx_t_9, __pyx_t_8);
                __pyx_t_11 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; 
                __PYX_ERR(0, 2334, __pyx_L19_except_error)
              }
              __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
              __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
              __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
              goto __pyx_L18_exception_handled;
            }
            __pyx_L19_except_error:;
            __Pyx_XGIVEREF(__pyx_t_13);
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_14, __pyx_t_15);
            goto __pyx_L5_except_error;
            __pyx_L18_exception_handled:;
            __Pyx_XGIVEREF(__pyx_t_13);
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_14, __pyx_t_15);
            __pyx_L24_try_end:;
          }
        }
        /*finally:*/ {
          /*normal exit:*/{
            if (__pyx_t_10) {
              __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_tuple__32, NULL);
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 2334, __pyx_L5_except_error)
              __Pyx_GOTREF(__pyx_t_15);
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            }
            goto __pyx_L16;
          }
          __pyx_L16:;
        }
        goto __pyx_L28;
        __pyx_L11_error:;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        goto __pyx_L5_except_error;
        __pyx_L28:;
      }
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "Aspidites/_vendor/pyparsing_extension.py":2331
 *         the entire file is opened, read, and closed before parsing.
 *         """
 *         try:             # <<<<<<<<<<<<<<
 *             file_contents = file_or_filename.read()
 *         except AttributeError:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L8_try_end:;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2336
 *             with open(file_or_filename, "r") as f:
 *                 file_contents = f.read()
 *         try:             # <<<<<<<<<<<<<<
 *             return self.parseString(file_contents, parseAll)
 *         except ParseBaseException as exc:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_2, &__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_1);
    /*try:*/ {

      /* "Aspidites/_vendor/pyparsing_extension.py":2337
 *                 file_contents = f.read()
 *         try:
 *             return self.parseString(file_contents, parseAll)             # <<<<<<<<<<<<<<
 *         except ParseBaseException as exc:
 *             if ParserElement.verbose_stacktrace:
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_parseString); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2337, __pyx_L29_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (unlikely(!__pyx_v_file_contents)) { __Pyx_RaiseUnboundLocalError("file_contents"); __PYX_ERR(0, 2337, __pyx_L29_error) }
      __pyx_t_4 = NULL;
      __pyx_t_7 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_7 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_file_contents, __pyx_v_parseAll};
        __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2337, __pyx_L29_error)
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_6);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_file_contents, __pyx_v_parseAll};
        __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2337, __pyx_L29_error)
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_6);
      } else
      #endif
      {
        __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2337, __pyx_L29_error)
        __Pyx_GOTREF(__pyx_t_8);
        if (__pyx_t_4) {
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_4); __pyx_t_4 = NULL;
        }
        __Pyx_INCREF(__pyx_v_file_contents);
        __Pyx_GIVEREF(__pyx_v_file_contents);
        PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_v_file_contents);
        __Pyx_INCREF(__pyx_v_parseAll);
        __Pyx_GIVEREF(__pyx_v_parseAll);
        PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_v_parseAll);
        __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2337, __pyx_L29_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_r = __pyx_t_6;
      __pyx_t_6 = 0;
      goto __pyx_L33_try_return;

      /* "Aspidites/_vendor/pyparsing_extension.py":2336
 *             with open(file_or_filename, "r") as f:
 *                 file_contents = f.read()
 *         try:             # <<<<<<<<<<<<<<
 *             return self.parseString(file_contents, parseAll)
 *         except ParseBaseException as exc:
 */
    }
    __pyx_L29_error:;
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2338
 *         try:
 *             return self.parseString(file_contents, parseAll)
 *         except ParseBaseException as exc:             # <<<<<<<<<<<<<<
 *             if ParserElement.verbose_stacktrace:
 *                 raise
 */
    __Pyx_ErrFetch(&__pyx_t_6, &__pyx_t_5, &__pyx_t_8);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ParseBaseException); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2338, __pyx_L31_except_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_6, __pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_ErrRestore(__pyx_t_6, __pyx_t_5, __pyx_t_8);
    __pyx_t_6 = 0; __pyx_t_5 = 0; __pyx_t_8 = 0;
    if (__pyx_t_7) {
      __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.parseFile", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_5, &__pyx_t_6) < 0) __PYX_ERR(0, 2338, __pyx_L31_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_5);
      __pyx_v_exc = __pyx_t_5;

      /* "Aspidites/_vendor/pyparsing_extension.py":2339
 *             return self.parseString(file_contents, parseAll)
 *         except ParseBaseException as exc:
 *             if ParserElement.verbose_stacktrace:             # <<<<<<<<<<<<<<
 *                 raise
 *             else:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2339, __pyx_L31_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_verbose_stacktrace); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2339, __pyx_L31_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_18 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_18 < 0)) __PYX_ERR(0, 2339, __pyx_L31_except_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(__pyx_t_18)) {

        /* "Aspidites/_vendor/pyparsing_extension.py":2340
 *         except ParseBaseException as exc:
 *             if ParserElement.verbose_stacktrace:
 *                 raise             # <<<<<<<<<<<<<<
 *             else:
 *                 # catch and re-raise exception from here, clearing out pyparsing internal stack trace
 */
        __Pyx_GIVEREF(__pyx_t_8);
        __Pyx_GIVEREF(__pyx_t_5);
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_ErrRestoreWithState(__pyx_t_8, __pyx_t_5, __pyx_t_6);
        __pyx_t_8 = 0; __pyx_t_5 = 0; __pyx_t_6 = 0; 
        __PYX_ERR(0, 2340, __pyx_L31_except_error)

        /* "Aspidites/_vendor/pyparsing_extension.py":2339
 *             return self.parseString(file_contents, parseAll)
 *         except ParseBaseException as exc:
 *             if ParserElement.verbose_stacktrace:             # <<<<<<<<<<<<<<
 *                 raise
 *             else:
 */
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":2343
 *             else:
 *                 # catch and re-raise exception from here, clearing out pyparsing internal stack trace
 *                 if getattr(exc, '__traceback__', None) is not None:             # <<<<<<<<<<<<<<
 *                     exc.__traceback__ = self._trim_traceback(exc.__traceback__)
 *                 raise exc
 */
      /*else*/ {
        __pyx_t_9 = __Pyx_GetAttr3(__pyx_v_exc, __pyx_n_s_traceback, Py_None); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2343, __pyx_L31_except_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_18 = (__pyx_t_9 != Py_None);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_17 = (__pyx_t_18 != 0);
        if (__pyx_t_17) {

          /* "Aspidites/_vendor/pyparsing_extension.py":2344
 *                 # catch and re-raise exception from here, clearing out pyparsing internal stack trace
 *                 if getattr(exc, '__traceback__', None) is not None:
 *                     exc.__traceback__ = self._trim_traceback(exc.__traceback__)             # <<<<<<<<<<<<<<
 *                 raise exc
 * 
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_trim_traceback); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2344, __pyx_L31_except_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_exc, __pyx_n_s_traceback); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2344, __pyx_L31_except_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_12 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_12)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_12);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          __pyx_t_9 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_12, __pyx_t_11) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_11);
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2344, __pyx_L31_except_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_exc, __pyx_n_s_traceback, __pyx_t_9) < 0) __PYX_ERR(0, 2344, __pyx_L31_except_error)
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":2343
 *             else:
 *                 # catch and re-raise exception from here, clearing out pyparsing internal stack trace
 *                 if getattr(exc, '__traceback__', None) is not None:             # <<<<<<<<<<<<<<
 *                     exc.__traceback__ = self._trim_traceback(exc.__traceback__)
 *                 raise exc
 */
        }

        /* "Aspidites/_vendor/pyparsing_extension.py":2345
 *                 if getattr(exc, '__traceback__', None) is not None:
 *                     exc.__traceback__ = self._trim_traceback(exc.__traceback__)
 *                 raise exc             # <<<<<<<<<<<<<<
 * 
 *     def __eq__(self, other):
 */
        __Pyx_Raise(__pyx_v_exc, 0, 0, 0);
        __PYX_ERR(0, 2345, __pyx_L31_except_error)
      }
    }
    goto __pyx_L31_except_error;
    __pyx_L31_except_error:;

    /* "Aspidites/_vendor/pyparsing_extension.py":2336
 *             with open(file_or_filename, "r") as f:
 *                 file_contents = f.read()
 *         try:             # <<<<<<<<<<<<<<
 *             return self.parseString(file_contents, parseAll)
 *         except ParseBaseException as exc:
 */
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_2, __pyx_t_1);
    goto __pyx_L1_error;
    __pyx_L33_try_return:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_2, __pyx_t_1);
    goto __pyx_L0;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2325
 *         self.checkRecursion([])
 * 
 *     def parseFile(self, file_or_filename, parseAll=False):             # <<<<<<<<<<<<<<
 *         """
 *         Execute the parse expression on the given file or filename.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.parseFile", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_file_contents);
  __Pyx_XDECREF(__pyx_v_f);
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2347
 *                 raise exc
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         if self is other:
 *             return True
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_117__eq__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_117__eq__ = {"__eq__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_117__eq__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_117__eq__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_other = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__eq__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_other,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_other)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__eq__", 1, 2, 2, 1); __PYX_ERR(0, 2347, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__eq__") < 0)) __PYX_ERR(0, 2347, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_other = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__eq__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2347, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__eq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_116__eq__(__pyx_self, __pyx_v_self, __pyx_v_other);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_116__eq__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__eq__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2348
 * 
 *     def __eq__(self, other):
 *         if self is other:             # <<<<<<<<<<<<<<
 *             return True
 *         elif isinstance(other, str):
 */
  __pyx_t_1 = (__pyx_v_self == __pyx_v_other);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2349
 *     def __eq__(self, other):
 *         if self is other:
 *             return True             # <<<<<<<<<<<<<<
 *         elif isinstance(other, str):
 *             return self.matches(other)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_True);
    __pyx_r = Py_True;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2348
 * 
 *     def __eq__(self, other):
 *         if self is other:             # <<<<<<<<<<<<<<
 *             return True
 *         elif isinstance(other, str):
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2350
 *         if self is other:
 *             return True
 *         elif isinstance(other, str):             # <<<<<<<<<<<<<<
 *             return self.matches(other)
 *         elif isinstance(other, ParserElement):
 */
  __pyx_t_2 = PyString_Check(__pyx_v_other); 
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2351
 *             return True
 *         elif isinstance(other, str):
 *             return self.matches(other)             # <<<<<<<<<<<<<<
 *         elif isinstance(other, ParserElement):
 *             return vars(self) == vars(other)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_matches); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2351, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_other) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_other);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2351, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2350
 *         if self is other:
 *             return True
 *         elif isinstance(other, str):             # <<<<<<<<<<<<<<
 *             return self.matches(other)
 *         elif isinstance(other, ParserElement):
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2352
 *         elif isinstance(other, str):
 *             return self.matches(other)
 *         elif isinstance(other, ParserElement):             # <<<<<<<<<<<<<<
 *             return vars(self) == vars(other)
 *         return False
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyObject_IsInstance(__pyx_v_other, __pyx_t_3); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2352, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2353
 *             return self.matches(other)
 *         elif isinstance(other, ParserElement):
 *             return vars(self) == vars(other)             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_vars, __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2353, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_vars, __pyx_v_other); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2353, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyObject_RichCompare(__pyx_t_3, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2353, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2352
 *         elif isinstance(other, str):
 *             return self.matches(other)
 *         elif isinstance(other, ParserElement):             # <<<<<<<<<<<<<<
 *             return vars(self) == vars(other)
 *         return False
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2354
 *         elif isinstance(other, ParserElement):
 *             return vars(self) == vars(other)
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     def __ne__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_False);
  __pyx_r = Py_False;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2347
 *                 raise exc
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         if self is other:
 *             return True
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__eq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2356
 *         return False
 * 
 *     def __ne__(self, other):             # <<<<<<<<<<<<<<
 *         return not (self == other)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_119__ne__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_119__ne__ = {"__ne__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_119__ne__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_119__ne__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_other = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__ne__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_other,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_other)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__ne__", 1, 2, 2, 1); __PYX_ERR(0, 2356, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__ne__") < 0)) __PYX_ERR(0, 2356, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_other = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__ne__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2356, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__ne__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_118__ne__(__pyx_self, __pyx_v_self, __pyx_v_other);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_118__ne__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__ne__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2357
 * 
 *     def __ne__(self, other):
 *         return not (self == other)             # <<<<<<<<<<<<<<
 * 
 *     def __hash__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_self, __pyx_v_other, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2357, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 2357, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyBool_FromLong((!__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2356
 *         return False
 * 
 *     def __ne__(self, other):             # <<<<<<<<<<<<<<
 *         return not (self == other)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__ne__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2359
 *         return not (self == other)
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return id(self)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_121__hash__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_121__hash__ = {"__hash__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_121__hash__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_121__hash__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__hash__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_120__hash__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_120__hash__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__hash__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2360
 * 
 *     def __hash__(self):
 *         return id(self)             # <<<<<<<<<<<<<<
 * 
 *     def __req__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_id, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2359
 *         return not (self == other)
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return id(self)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__hash__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2362
 *         return id(self)
 * 
 *     def __req__(self, other):             # <<<<<<<<<<<<<<
 *         return self == other
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_123__req__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_123__req__ = {"__req__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_123__req__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_123__req__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_other = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__req__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_other,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_other)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__req__", 1, 2, 2, 1); __PYX_ERR(0, 2362, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__req__") < 0)) __PYX_ERR(0, 2362, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_other = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__req__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2362, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__req__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_122__req__(__pyx_self, __pyx_v_self, __pyx_v_other);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_122__req__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__req__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2363
 * 
 *     def __req__(self, other):
 *         return self == other             # <<<<<<<<<<<<<<
 * 
 *     def __rne__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_self, __pyx_v_other, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2363, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2362
 *         return id(self)
 * 
 *     def __req__(self, other):             # <<<<<<<<<<<<<<
 *         return self == other
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__req__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2365
 *         return self == other
 * 
 *     def __rne__(self, other):             # <<<<<<<<<<<<<<
 *         return not (self == other)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_125__rne__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_125__rne__ = {"__rne__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_125__rne__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_125__rne__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_other = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__rne__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_other,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_other)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__rne__", 1, 2, 2, 1); __PYX_ERR(0, 2365, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__rne__") < 0)) __PYX_ERR(0, 2365, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_other = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__rne__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2365, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__rne__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_124__rne__(__pyx_self, __pyx_v_self, __pyx_v_other);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_124__rne__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__rne__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2366
 * 
 *     def __rne__(self, other):
 *         return not (self == other)             # <<<<<<<<<<<<<<
 * 
 *     def matches(self, testString, parseAll=True):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_self, __pyx_v_other, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2366, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 2366, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyBool_FromLong((!__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2365
 *         return self == other
 * 
 *     def __rne__(self, other):             # <<<<<<<<<<<<<<
 *         return not (self == other)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.__rne__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2368
 *         return not (self == other)
 * 
 *     def matches(self, testString, parseAll=True):             # <<<<<<<<<<<<<<
 *         """
 *         Method for quick testing of a parser against a test string. Good for simple
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_127matches(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_126matches[] = "\n        Method for quick testing of a parser against a test string. Good for simple\n        inline microtests of sub expressions while building up larger parser.\n\n        Parameters:\n         - testString - to test against this expression for a match\n         - parseAll - (default= ``True``) - flag to pass to :class:`parseString` when running tests\n\n        Example::\n\n            expr = Word(nums)\n            assert expr.matches(\"100\")\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_127matches = {"matches", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_127matches, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_126matches};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_127matches(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_testString = 0;
  PyObject *__pyx_v_parseAll = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("matches (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_testString,&__pyx_n_s_parseAll,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_testString)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("matches", 0, 2, 3, 1); __PYX_ERR(0, 2368, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_parseAll);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "matches") < 0)) __PYX_ERR(0, 2368, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_testString = values[1];
    __pyx_v_parseAll = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("matches", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2368, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.matches", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_126matches(__pyx_self, __pyx_v_self, __pyx_v_testString, __pyx_v_parseAll);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_126matches(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_testString, PyObject *__pyx_v_parseAll) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("matches", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2382
 *             assert expr.matches("100")
 *         """
 *         try:             # <<<<<<<<<<<<<<
 *             self.parseString(str(testString), parseAll=parseAll)
 *             return True
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "Aspidites/_vendor/pyparsing_extension.py":2383
 *         """
 *         try:
 *             self.parseString(str(testString), parseAll=parseAll)             # <<<<<<<<<<<<<<
 *             return True
 *         except ParseBaseException:
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_parseString); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2383, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_v_testString); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2383, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2383, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2383, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_parseAll, __pyx_v_parseAll) < 0) __PYX_ERR(0, 2383, __pyx_L3_error)
      __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2383, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":2384
 *         try:
 *             self.parseString(str(testString), parseAll=parseAll)
 *             return True             # <<<<<<<<<<<<<<
 *         except ParseBaseException:
 *             return False
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_True);
      __pyx_r = Py_True;
      goto __pyx_L7_try_return;

      /* "Aspidites/_vendor/pyparsing_extension.py":2382
 *             assert expr.matches("100")
 *         """
 *         try:             # <<<<<<<<<<<<<<
 *             self.parseString(str(testString), parseAll=parseAll)
 *             return True
 */
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2385
 *             self.parseString(str(testString), parseAll=parseAll)
 *             return True
 *         except ParseBaseException:             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
    __Pyx_ErrFetch(&__pyx_t_7, &__pyx_t_5, &__pyx_t_6);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ParseBaseException); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2385, __pyx_L5_except_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_7, __pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_ErrRestore(__pyx_t_7, __pyx_t_5, __pyx_t_6);
    __pyx_t_7 = 0; __pyx_t_5 = 0; __pyx_t_6 = 0;
    if (__pyx_t_8) {
      __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.matches", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_5, &__pyx_t_7) < 0) __PYX_ERR(0, 2385, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_7);

      /* "Aspidites/_vendor/pyparsing_extension.py":2386
 *             return True
 *         except ParseBaseException:
 *             return False             # <<<<<<<<<<<<<<
 * 
 *     def runTests(self, tests, parseAll=True, comment='#',
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L6_except_return;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "Aspidites/_vendor/pyparsing_extension.py":2382
 *             assert expr.matches("100")
 *         """
 *         try:             # <<<<<<<<<<<<<<
 *             self.parseString(str(testString), parseAll=parseAll)
 *             return True
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2368
 *         return not (self == other)
 * 
 *     def matches(self, testString, parseAll=True):             # <<<<<<<<<<<<<<
 *         """
 *         Method for quick testing of a parser against a test string. Good for simple
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.matches", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2388
 *             return False
 * 
 *     def runTests(self, tests, parseAll=True, comment='#',             # <<<<<<<<<<<<<<
 *                  fullDump=True, printResults=True, failureTests=False, postParse=None,
 *                  file=None):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_129runTests(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_128runTests[] = "\n        Execute the parse expression on a series of test strings, showing each\n        test, the parsed results or where the parse failed. Quick and easy way to\n        run a parse expression against a list of sample strings.\n\n        Parameters:\n         - tests - a list of separate test strings, or a multiline string of test strings\n         - parseAll - (default= ``True``) - flag to pass to :class:`parseString` when running tests\n         - comment - (default= ``'#'``) - expression for indicating embedded comments in the test\n              string; pass None to disable comment filtering\n         - fullDump - (default= ``True``) - dump results as list followed by results names in nested outline;\n              if False, only dump nested list\n         - printResults - (default= ``True``) prints test output to stdout\n         - failureTests - (default= ``False``) indicates if these tests are expected to fail parsing\n         - postParse - (default= ``None``) optional callback for successful parse results; called as\n              `fn(test_string, parse_results)` and returns a string to be added to the test output\n         - file - (default=``None``) optional file-like object to which test output will be written;\n              if None, will default to ``sys.stdout``\n\n        Returns: a (success, results) tuple, where success indicates that all tests succeeded\n        (or failed if ``failureTests`` is True), and the results contain a list of lines of each\n        test's output\n\n        Example::\n\n            number_expr = pyparsing_common.number.copy()\n\n            result = number_expr.runTests('''\n                # unsigned integer\n                100\n                # negative integer\n                -100\n                # float with scientific notation\n                6.02e23\n                # integer with scientific notation\n                1e-12\n                ''')\n            print(\"Success\" if result[0] else \"Failed!\")""\n\n            result = number_expr.runTests('''\n                # stray character\n                100Z\n                # missing leading digit before '.'\n                -.100\n                # too many '.'\n                3.14.159\n                ''', failureTests=True)\n            print(\"Success\" if result[0] else \"Failed!\")\n\n        prints::\n\n            # unsigned integer\n            100\n            [100]\n\n            # negative integer\n            -100\n            [-100]\n\n            # float with scientific notation\n            6.02e23\n            [6.02e+23]\n\n            # integer with scientific notation\n            1e-12\n            [1e-12]\n\n            Success\n\n            # stray character\n            100Z\n               ^\n            FAIL: Expected end of text (at char 3), (line:1, col:4)\n\n            # missing leading digit before '.'\n            -.100\n            ^\n            FAIL: Expected {real number with scientific notation | real number | signed integer} (at char 0), (line:1, col:1)\n\n            # too many '.'\n            3.14.159\n                ^\n            FAIL: Expected end of text (at char 4), (line:1, col:5)\n\n            Success\n\n        Each test string must be on a single line. If you want to test a string that spans multiple\n        lines, create a test like this::\n\n            expr.runTest(r\"this is a test\\n of strings that spans \\n 3 lines\")\n\n        (Note that this is a raw string literal, you must include the leading 'r'.)\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_129runTests = {"runTests", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_129runTests, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_128runTests};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_129runTests(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_tests = 0;
  PyObject *__pyx_v_parseAll = 0;
  PyObject *__pyx_v_comment = 0;
  PyObject *__pyx_v_fullDump = 0;
  PyObject *__pyx_v_printResults = 0;
  PyObject *__pyx_v_failureTests = 0;
  PyObject *__pyx_v_postParse = 0;
  PyObject *__pyx_v_file = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("runTests (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_tests,&__pyx_n_s_parseAll,&__pyx_n_s_comment,&__pyx_n_s_fullDump,&__pyx_n_s_printResults,&__pyx_n_s_failureTests,&__pyx_n_s_postParse,&__pyx_n_s_file,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_True));
    values[3] = ((PyObject *)((PyObject*)__pyx_kp_s__79));

    /* "Aspidites/_vendor/pyparsing_extension.py":2389
 * 
 *     def runTests(self, tests, parseAll=True, comment='#',
 *                  fullDump=True, printResults=True, failureTests=False, postParse=None,             # <<<<<<<<<<<<<<
 *                  file=None):
 *         """
 */
    values[4] = ((PyObject *)((PyObject *)Py_True));
    values[5] = ((PyObject *)((PyObject *)Py_True));
    values[6] = ((PyObject *)((PyObject *)Py_False));
    values[7] = ((PyObject *)((PyObject *)Py_None));

    /* "Aspidites/_vendor/pyparsing_extension.py":2390
 *     def runTests(self, tests, parseAll=True, comment='#',
 *                  fullDump=True, printResults=True, failureTests=False, postParse=None,
 *                  file=None):             # <<<<<<<<<<<<<<
 *         """
 *         Execute the parse expression on a series of test strings, showing each
 */
    values[8] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tests)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("runTests", 0, 2, 9, 1); __PYX_ERR(0, 2388, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_parseAll);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_comment);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fullDump);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_printResults);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_failureTests);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_postParse);
          if (value) { values[7] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_file);
          if (value) { values[8] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "runTests") < 0)) __PYX_ERR(0, 2388, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_tests = values[1];
    __pyx_v_parseAll = values[2];
    __pyx_v_comment = values[3];
    __pyx_v_fullDump = values[4];
    __pyx_v_printResults = values[5];
    __pyx_v_failureTests = values[6];
    __pyx_v_postParse = values[7];
    __pyx_v_file = values[8];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("runTests", 0, 2, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2388, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.runTests", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_128runTests(__pyx_self, __pyx_v_self, __pyx_v_tests, __pyx_v_parseAll, __pyx_v_comment, __pyx_v_fullDump, __pyx_v_printResults, __pyx_v_failureTests, __pyx_v_postParse, __pyx_v_file);

  /* "Aspidites/_vendor/pyparsing_extension.py":2388
 *             return False
 * 
 *     def runTests(self, tests, parseAll=True, comment='#',             # <<<<<<<<<<<<<<
 *                  fullDump=True, printResults=True, failureTests=False, postParse=None,
 *                  file=None):
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_128runTests(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_tests, PyObject *__pyx_v_parseAll, PyObject *__pyx_v_comment, PyObject *__pyx_v_fullDump, PyObject *__pyx_v_printResults, PyObject *__pyx_v_failureTests, PyObject *__pyx_v_postParse, PyObject *__pyx_v_file) {
  PyObject *__pyx_v_Literal = NULL;
  PyObject *__pyx_v_replaceWith = NULL;
  PyObject *__pyx_v_quotedString = NULL;
  PyObject *__pyx_v_print_ = NULL;
  PyObject *__pyx_v_allResults = NULL;
  PyObject *__pyx_v_comments = NULL;
  PyObject *__pyx_v_success = NULL;
  PyObject *__pyx_v_NL = NULL;
  PyObject *__pyx_v_BOM = NULL;
  PyObject *__pyx_v_t = NULL;
  PyObject *__pyx_v_out = NULL;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_v_pe = NULL;
  PyObject *__pyx_v_fatal = NULL;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_v_pp_value = NULL;
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  PyObject *(*__pyx_t_10)(PyObject *);
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("runTests", 0);
  __Pyx_INCREF(__pyx_v_tests);
  __Pyx_INCREF(__pyx_v_comment);
  __Pyx_INCREF(__pyx_v_file);

  /* "Aspidites/_vendor/pyparsing_extension.py":2484
 *         (Note that this is a raw string literal, you must include the leading 'r'.)
 *         """
 *         from .pyparsing import Literal, replaceWith, quotedString             # <<<<<<<<<<<<<<
 *         if isinstance(tests, str):
 *             tests = list(map(str.strip, tests.rstrip().splitlines()))
 */
  __pyx_t_1 = PyList_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_Literal);
  __Pyx_GIVEREF(__pyx_n_s_Literal);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_Literal);
  __Pyx_INCREF(__pyx_n_s_replaceWith);
  __Pyx_GIVEREF(__pyx_n_s_replaceWith);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_replaceWith);
  __Pyx_INCREF(__pyx_n_s_quotedString);
  __Pyx_GIVEREF(__pyx_n_s_quotedString);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_s_quotedString);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_pyparsing, __pyx_t_1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_Literal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_Literal = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_replaceWith); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_replaceWith = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_quotedString); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_quotedString = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2485
 *         """
 *         from .pyparsing import Literal, replaceWith, quotedString
 *         if isinstance(tests, str):             # <<<<<<<<<<<<<<
 *             tests = list(map(str.strip, tests.rstrip().splitlines()))
 *         if isinstance(comment, str):
 */
  __pyx_t_3 = PyString_Check(__pyx_v_tests); 
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2486
 *         from .pyparsing import Literal, replaceWith, quotedString
 *         if isinstance(tests, str):
 *             tests = list(map(str.strip, tests.rstrip().splitlines()))             # <<<<<<<<<<<<<<
 *         if isinstance(comment, str):
 *             comment = Literal(comment)
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)(&PyString_Type)), __pyx_n_s_strip); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2486, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_tests, __pyx_n_s_rstrip); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2486, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_5 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2486, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_splitlines); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2486, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2486, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2486, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_1);
    __pyx_t_2 = 0;
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_map, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2486, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = PySequence_List(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2486, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_tests, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2485
 *         """
 *         from .pyparsing import Literal, replaceWith, quotedString
 *         if isinstance(tests, str):             # <<<<<<<<<<<<<<
 *             tests = list(map(str.strip, tests.rstrip().splitlines()))
 *         if isinstance(comment, str):
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2487
 *         if isinstance(tests, str):
 *             tests = list(map(str.strip, tests.rstrip().splitlines()))
 *         if isinstance(comment, str):             # <<<<<<<<<<<<<<
 *             comment = Literal(comment)
 *         if file is None:
 */
  __pyx_t_4 = PyString_Check(__pyx_v_comment); 
  __pyx_t_3 = (__pyx_t_4 != 0);
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2488
 *             tests = list(map(str.strip, tests.rstrip().splitlines()))
 *         if isinstance(comment, str):
 *             comment = Literal(comment)             # <<<<<<<<<<<<<<
 *         if file is None:
 *             file = sys.stdout
 */
    __Pyx_INCREF(__pyx_v_Literal);
    __pyx_t_1 = __pyx_v_Literal; __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_6 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_v_comment) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_comment);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2488, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_comment, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2487
 *         if isinstance(tests, str):
 *             tests = list(map(str.strip, tests.rstrip().splitlines()))
 *         if isinstance(comment, str):             # <<<<<<<<<<<<<<
 *             comment = Literal(comment)
 *         if file is None:
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2489
 *         if isinstance(comment, str):
 *             comment = Literal(comment)
 *         if file is None:             # <<<<<<<<<<<<<<
 *             file = sys.stdout
 *         print_ = file.write
 */
  __pyx_t_3 = (__pyx_v_file == Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2490
 *             comment = Literal(comment)
 *         if file is None:
 *             file = sys.stdout             # <<<<<<<<<<<<<<
 *         print_ = file.write
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_sys); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2490, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_stdout); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2490, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF_SET(__pyx_v_file, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2489
 *         if isinstance(comment, str):
 *             comment = Literal(comment)
 *         if file is None:             # <<<<<<<<<<<<<<
 *             file = sys.stdout
 *         print_ = file.write
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2491
 *         if file is None:
 *             file = sys.stdout
 *         print_ = file.write             # <<<<<<<<<<<<<<
 * 
 *         allResults = []
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_file, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_print_ = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2493
 *         print_ = file.write
 * 
 *         allResults = []             # <<<<<<<<<<<<<<
 *         comments = []
 *         success = True
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_allResults = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2494
 * 
 *         allResults = []
 *         comments = []             # <<<<<<<<<<<<<<
 *         success = True
 *         NL = Literal(r'\n').addParseAction(replaceWith('\n')).ignore(quotedString)
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_comments = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2495
 *         allResults = []
 *         comments = []
 *         success = True             # <<<<<<<<<<<<<<
 *         NL = Literal(r'\n').addParseAction(replaceWith('\n')).ignore(quotedString)
 *         BOM = u'\ufeff'
 */
  __Pyx_INCREF(Py_True);
  __pyx_v_success = Py_True;

  /* "Aspidites/_vendor/pyparsing_extension.py":2496
 *         comments = []
 *         success = True
 *         NL = Literal(r'\n').addParseAction(replaceWith('\n')).ignore(quotedString)             # <<<<<<<<<<<<<<
 *         BOM = u'\ufeff'
 *         for t in tests:
 */
  __Pyx_INCREF(__pyx_v_Literal);
  __pyx_t_5 = __pyx_v_Literal; __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_2 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_7, __pyx_kp_s_n) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_kp_s_n);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_addParseAction); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_INCREF(__pyx_v_replaceWith);
  __pyx_t_7 = __pyx_v_replaceWith; __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_2 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_kp_s__4) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_kp_s__4);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_6 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_7, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_ignore); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_v_quotedString) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_quotedString);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_NL = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2497
 *         success = True
 *         NL = Literal(r'\n').addParseAction(replaceWith('\n')).ignore(quotedString)
 *         BOM = u'\ufeff'             # <<<<<<<<<<<<<<
 *         for t in tests:
 *             if comment is not None and comment.matches(t, False) or comments and not t:
 */
  __Pyx_INCREF(__pyx_kp_u__80);
  __pyx_v_BOM = __pyx_kp_u__80;

  /* "Aspidites/_vendor/pyparsing_extension.py":2498
 *         NL = Literal(r'\n').addParseAction(replaceWith('\n')).ignore(quotedString)
 *         BOM = u'\ufeff'
 *         for t in tests:             # <<<<<<<<<<<<<<
 *             if comment is not None and comment.matches(t, False) or comments and not t:
 *                 comments.append(t)
 */
  if (likely(PyList_CheckExact(__pyx_v_tests)) || PyTuple_CheckExact(__pyx_v_tests)) {
    __pyx_t_1 = __pyx_v_tests; __Pyx_INCREF(__pyx_t_1); __pyx_t_9 = 0;
    __pyx_t_10 = NULL;
  } else {
    __pyx_t_9 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_tests); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2498, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_10 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2498, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_10)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_9); __Pyx_INCREF(__pyx_t_5); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 2498, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_1, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2498, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      } else {
        if (__pyx_t_9 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_9); __Pyx_INCREF(__pyx_t_5); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 2498, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_1, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2498, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      }
    } else {
      __pyx_t_5 = __pyx_t_10(__pyx_t_1);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 2498, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_XDECREF_SET(__pyx_v_t, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2499
 *         BOM = u'\ufeff'
 *         for t in tests:
 *             if comment is not None and comment.matches(t, False) or comments and not t:             # <<<<<<<<<<<<<<
 *                 comments.append(t)
 *                 continue
 */
    __pyx_t_3 = (__pyx_v_comment != Py_None);
    __pyx_t_11 = (__pyx_t_3 != 0);
    if (!__pyx_t_11) {
      goto __pyx_L10_next_or;
    } else {
    }
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_comment, __pyx_n_s_matches); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2499, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = NULL;
    __pyx_t_12 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_12 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_t, Py_False};
      __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2499, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_5);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_t, Py_False};
      __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2499, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_5);
    } else
    #endif
    {
      __pyx_t_7 = PyTuple_New(2+__pyx_t_12); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2499, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (__pyx_t_2) {
        __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
      }
      __Pyx_INCREF(__pyx_v_t);
      __Pyx_GIVEREF(__pyx_v_t);
      PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_12, __pyx_v_t);
      __Pyx_INCREF(Py_False);
      __Pyx_GIVEREF(Py_False);
      PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_12, Py_False);
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_7, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2499, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2499, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (!__pyx_t_11) {
    } else {
      __pyx_t_4 = __pyx_t_11;
      goto __pyx_L9_bool_binop_done;
    }
    __pyx_L10_next_or:;
    __pyx_t_11 = (PyList_GET_SIZE(__pyx_v_comments) != 0);
    if (__pyx_t_11) {
    } else {
      __pyx_t_4 = __pyx_t_11;
      goto __pyx_L9_bool_binop_done;
    }
    __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_v_t); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2499, __pyx_L1_error)
    __pyx_t_3 = ((!__pyx_t_11) != 0);
    __pyx_t_4 = __pyx_t_3;
    __pyx_L9_bool_binop_done:;
    if (__pyx_t_4) {

      /* "Aspidites/_vendor/pyparsing_extension.py":2500
 *         for t in tests:
 *             if comment is not None and comment.matches(t, False) or comments and not t:
 *                 comments.append(t)             # <<<<<<<<<<<<<<
 *                 continue
 *             if not t:
 */
      __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_comments, __pyx_v_t); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 2500, __pyx_L1_error)

      /* "Aspidites/_vendor/pyparsing_extension.py":2501
 *             if comment is not None and comment.matches(t, False) or comments and not t:
 *                 comments.append(t)
 *                 continue             # <<<<<<<<<<<<<<
 *             if not t:
 *                 continue
 */
      goto __pyx_L6_continue;

      /* "Aspidites/_vendor/pyparsing_extension.py":2499
 *         BOM = u'\ufeff'
 *         for t in tests:
 *             if comment is not None and comment.matches(t, False) or comments and not t:             # <<<<<<<<<<<<<<
 *                 comments.append(t)
 *                 continue
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":2502
 *                 comments.append(t)
 *                 continue
 *             if not t:             # <<<<<<<<<<<<<<
 *                 continue
 *             out = ['\n' + '\n'.join(comments) if comments else '', t]
 */
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_t); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 2502, __pyx_L1_error)
    __pyx_t_3 = ((!__pyx_t_4) != 0);
    if (__pyx_t_3) {

      /* "Aspidites/_vendor/pyparsing_extension.py":2503
 *                 continue
 *             if not t:
 *                 continue             # <<<<<<<<<<<<<<
 *             out = ['\n' + '\n'.join(comments) if comments else '', t]
 *             comments = []
 */
      goto __pyx_L6_continue;

      /* "Aspidites/_vendor/pyparsing_extension.py":2502
 *                 comments.append(t)
 *                 continue
 *             if not t:             # <<<<<<<<<<<<<<
 *                 continue
 *             out = ['\n' + '\n'.join(comments) if comments else '', t]
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":2504
 *             if not t:
 *                 continue
 *             out = ['\n' + '\n'.join(comments) if comments else '', t]             # <<<<<<<<<<<<<<
 *             comments = []
 *             try:
 */
    __pyx_t_3 = (PyList_GET_SIZE(__pyx_v_comments) != 0);
    if (__pyx_t_3) {
      __pyx_t_6 = __Pyx_PyString_Join(__pyx_kp_s__4, __pyx_v_comments); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2504, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = PyNumber_Add(__pyx_kp_s__4, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2504, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_5 = __pyx_t_7;
      __pyx_t_7 = 0;
    } else {
      __Pyx_INCREF(__pyx_kp_s__12);
      __pyx_t_5 = __pyx_kp_s__12;
    }
    __pyx_t_7 = PyList_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2504, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_5);
    PyList_SET_ITEM(__pyx_t_7, 0, __pyx_t_5);
    __Pyx_INCREF(__pyx_v_t);
    __Pyx_GIVEREF(__pyx_v_t);
    PyList_SET_ITEM(__pyx_t_7, 1, __pyx_v_t);
    __pyx_t_5 = 0;
    __Pyx_XDECREF_SET(__pyx_v_out, ((PyObject*)__pyx_t_7));
    __pyx_t_7 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2505
 *                 continue
 *             out = ['\n' + '\n'.join(comments) if comments else '', t]
 *             comments = []             # <<<<<<<<<<<<<<
 *             try:
 *                 # convert newline marks to actual newlines, and strip leading BOM if present
 */
    __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2505, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF_SET(__pyx_v_comments, ((PyObject*)__pyx_t_7));
    __pyx_t_7 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2506
 *             out = ['\n' + '\n'.join(comments) if comments else '', t]
 *             comments = []
 *             try:             # <<<<<<<<<<<<<<
 *                 # convert newline marks to actual newlines, and strip leading BOM if present
 *                 t = NL.transformString(t.lstrip(BOM))
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
      __Pyx_XGOTREF(__pyx_t_14);
      __Pyx_XGOTREF(__pyx_t_15);
      __Pyx_XGOTREF(__pyx_t_16);
      /*try:*/ {

        /* "Aspidites/_vendor/pyparsing_extension.py":2508
 *             try:
 *                 # convert newline marks to actual newlines, and strip leading BOM if present
 *                 t = NL.transformString(t.lstrip(BOM))             # <<<<<<<<<<<<<<
 *                 result = self.parseString(t, parseAll=parseAll)
 *             except ParseBaseException as pe:
 */
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_NL, __pyx_n_s_transformString); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2508, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_t, __pyx_n_s_lstrip); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2508, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_8 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_8)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        __pyx_t_6 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_8, __pyx_v_BOM) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_BOM);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2508, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
          }
        }
        __pyx_t_7 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_2, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6);
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2508, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF_SET(__pyx_v_t, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":2509
 *                 # convert newline marks to actual newlines, and strip leading BOM if present
 *                 t = NL.transformString(t.lstrip(BOM))
 *                 result = self.parseString(t, parseAll=parseAll)             # <<<<<<<<<<<<<<
 *             except ParseBaseException as pe:
 *                 fatal = "(FATAL)" if isinstance(pe, ParseFatalException) else ""
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_parseString); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2509, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2509, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_INCREF(__pyx_v_t);
        __Pyx_GIVEREF(__pyx_v_t);
        PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_t);
        __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2509, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_6);
        if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_parseAll, __pyx_v_parseAll) < 0) __PYX_ERR(0, 2509, __pyx_L14_error)
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2509, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF_SET(__pyx_v_result, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":2506
 *             out = ['\n' + '\n'.join(comments) if comments else '', t]
 *             comments = []
 *             try:             # <<<<<<<<<<<<<<
 *                 # convert newline marks to actual newlines, and strip leading BOM if present
 *                 t = NL.transformString(t.lstrip(BOM))
 */
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":2525
 *                 result = exc
 *             else:
 *                 success = success and not failureTests             # <<<<<<<<<<<<<<
 *                 if postParse is not None:
 *                     try:
 */
      /*else:*/ {
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_success); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 2525, __pyx_L16_except_error)
        if (__pyx_t_3) {
        } else {
          __Pyx_INCREF(__pyx_v_success);
          __pyx_t_2 = __pyx_v_success;
          goto __pyx_L22_bool_binop_done;
        }
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_failureTests); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 2525, __pyx_L16_except_error)
        __pyx_t_4 = (!__pyx_t_3);
        __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2525, __pyx_L16_except_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_2 = __pyx_t_6;
        __pyx_t_6 = 0;
        __pyx_L22_bool_binop_done:;
        __Pyx_DECREF_SET(__pyx_v_success, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":2526
 *             else:
 *                 success = success and not failureTests
 *                 if postParse is not None:             # <<<<<<<<<<<<<<
 *                     try:
 *                         pp_value = postParse(t, result)
 */
        __pyx_t_4 = (__pyx_v_postParse != Py_None);
        __pyx_t_3 = (__pyx_t_4 != 0);
        if (__pyx_t_3) {

          /* "Aspidites/_vendor/pyparsing_extension.py":2527
 *                 success = success and not failureTests
 *                 if postParse is not None:
 *                     try:             # <<<<<<<<<<<<<<
 *                         pp_value = postParse(t, result)
 *                         if pp_value is not None:
 */
          {
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __Pyx_ExceptionSave(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            /*try:*/ {

              /* "Aspidites/_vendor/pyparsing_extension.py":2528
 *                 if postParse is not None:
 *                     try:
 *                         pp_value = postParse(t, result)             # <<<<<<<<<<<<<<
 *                         if pp_value is not None:
 *                             if isinstance(pp_value, ParseResults):
 */
              __Pyx_INCREF(__pyx_v_postParse);
              __pyx_t_6 = __pyx_v_postParse; __pyx_t_5 = NULL;
              __pyx_t_12 = 0;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
                __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
                if (likely(__pyx_t_5)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                  __Pyx_INCREF(__pyx_t_5);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_6, function);
                  __pyx_t_12 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_6)) {
                PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_t, __pyx_v_result};
                __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2528, __pyx_L25_error)
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_GOTREF(__pyx_t_2);
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
                PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_t, __pyx_v_result};
                __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2528, __pyx_L25_error)
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_GOTREF(__pyx_t_2);
              } else
              #endif
              {
                __pyx_t_7 = PyTuple_New(2+__pyx_t_12); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2528, __pyx_L25_error)
                __Pyx_GOTREF(__pyx_t_7);
                if (__pyx_t_5) {
                  __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
                }
                __Pyx_INCREF(__pyx_v_t);
                __Pyx_GIVEREF(__pyx_v_t);
                PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_12, __pyx_v_t);
                __Pyx_INCREF(__pyx_v_result);
                __Pyx_GIVEREF(__pyx_v_result);
                PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_12, __pyx_v_result);
                __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2528, __pyx_L25_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              }
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_XDECREF_SET(__pyx_v_pp_value, __pyx_t_2);
              __pyx_t_2 = 0;

              /* "Aspidites/_vendor/pyparsing_extension.py":2529
 *                     try:
 *                         pp_value = postParse(t, result)
 *                         if pp_value is not None:             # <<<<<<<<<<<<<<
 *                             if isinstance(pp_value, ParseResults):
 *                                 out.append(pp_value.dump())
 */
              __pyx_t_3 = (__pyx_v_pp_value != Py_None);
              __pyx_t_4 = (__pyx_t_3 != 0);
              if (__pyx_t_4) {

                /* "Aspidites/_vendor/pyparsing_extension.py":2530
 *                         pp_value = postParse(t, result)
 *                         if pp_value is not None:
 *                             if isinstance(pp_value, ParseResults):             # <<<<<<<<<<<<<<
 *                                 out.append(pp_value.dump())
 *                             else:
 */
                __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParseResults); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2530, __pyx_L25_error)
                __Pyx_GOTREF(__pyx_t_2);
                __pyx_t_4 = PyObject_IsInstance(__pyx_v_pp_value, __pyx_t_2); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 2530, __pyx_L25_error)
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                __pyx_t_3 = (__pyx_t_4 != 0);
                if (__pyx_t_3) {

                  /* "Aspidites/_vendor/pyparsing_extension.py":2531
 *                         if pp_value is not None:
 *                             if isinstance(pp_value, ParseResults):
 *                                 out.append(pp_value.dump())             # <<<<<<<<<<<<<<
 *                             else:
 *                                 out.append(str(pp_value))
 */
                  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_pp_value, __pyx_n_s_dump); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2531, __pyx_L25_error)
                  __Pyx_GOTREF(__pyx_t_6);
                  __pyx_t_7 = NULL;
                  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
                    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
                    if (likely(__pyx_t_7)) {
                      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                      __Pyx_INCREF(__pyx_t_7);
                      __Pyx_INCREF(function);
                      __Pyx_DECREF_SET(__pyx_t_6, function);
                    }
                  }
                  __pyx_t_2 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
                  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2531, __pyx_L25_error)
                  __Pyx_GOTREF(__pyx_t_2);
                  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                  __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_out, __pyx_t_2); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 2531, __pyx_L25_error)
                  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

                  /* "Aspidites/_vendor/pyparsing_extension.py":2530
 *                         pp_value = postParse(t, result)
 *                         if pp_value is not None:
 *                             if isinstance(pp_value, ParseResults):             # <<<<<<<<<<<<<<
 *                                 out.append(pp_value.dump())
 *                             else:
 */
                  goto __pyx_L34;
                }

                /* "Aspidites/_vendor/pyparsing_extension.py":2533
 *                                 out.append(pp_value.dump())
 *                             else:
 *                                 out.append(str(pp_value))             # <<<<<<<<<<<<<<
 *                         else:
 *                             out.append(result.dump())
 */
                /*else*/ {
                  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_v_pp_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2533, __pyx_L25_error)
                  __Pyx_GOTREF(__pyx_t_2);
                  __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_out, __pyx_t_2); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 2533, __pyx_L25_error)
                  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                }
                __pyx_L34:;

                /* "Aspidites/_vendor/pyparsing_extension.py":2529
 *                     try:
 *                         pp_value = postParse(t, result)
 *                         if pp_value is not None:             # <<<<<<<<<<<<<<
 *                             if isinstance(pp_value, ParseResults):
 *                                 out.append(pp_value.dump())
 */
                goto __pyx_L33;
              }

              /* "Aspidites/_vendor/pyparsing_extension.py":2535
 *                                 out.append(str(pp_value))
 *                         else:
 *                             out.append(result.dump())             # <<<<<<<<<<<<<<
 *                     except Exception as e:
 *                         out.append(result.dump(full=fullDump))
 */
              /*else*/ {
                __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_result, __pyx_n_s_dump); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2535, __pyx_L25_error)
                __Pyx_GOTREF(__pyx_t_6);
                __pyx_t_7 = NULL;
                if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
                  __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
                  if (likely(__pyx_t_7)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                    __Pyx_INCREF(__pyx_t_7);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_6, function);
                  }
                }
                __pyx_t_2 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
                __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2535, __pyx_L25_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_out, __pyx_t_2); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 2535, __pyx_L25_error)
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              }
              __pyx_L33:;

              /* "Aspidites/_vendor/pyparsing_extension.py":2527
 *                 success = success and not failureTests
 *                 if postParse is not None:
 *                     try:             # <<<<<<<<<<<<<<
 *                         pp_value = postParse(t, result)
 *                         if pp_value is not None:
 */
            }
            __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
            __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
            __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
            goto __pyx_L32_try_end;
            __pyx_L25_error:;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

            /* "Aspidites/_vendor/pyparsing_extension.py":2536
 *                         else:
 *                             out.append(result.dump())
 *                     except Exception as e:             # <<<<<<<<<<<<<<
 *                         out.append(result.dump(full=fullDump))
 *                         out.append("{0} failed: {1}: {2}".format(postParse.__name__, type(e).__name__, e))
 */
            __pyx_t_12 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
            if (__pyx_t_12) {
              __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.runTests", __pyx_clineno, __pyx_lineno, __pyx_filename);
              if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(0, 2536, __pyx_L27_except_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_6);
              __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_6);

              /* "Aspidites/_vendor/pyparsing_extension.py":2537
 *                             out.append(result.dump())
 *                     except Exception as e:
 *                         out.append(result.dump(full=fullDump))             # <<<<<<<<<<<<<<
 *                         out.append("{0} failed: {1}: {2}".format(postParse.__name__, type(e).__name__, e))
 *                 else:
 */
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_result, __pyx_n_s_dump); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2537, __pyx_L27_except_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2537, __pyx_L27_except_error)
              __Pyx_GOTREF(__pyx_t_8);
              if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_full, __pyx_v_fullDump) < 0) __PYX_ERR(0, 2537, __pyx_L27_except_error)
              __pyx_t_20 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_empty_tuple, __pyx_t_8); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 2537, __pyx_L27_except_error)
              __Pyx_GOTREF(__pyx_t_20);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_out, __pyx_t_20); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 2537, __pyx_L27_except_error)
              __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

              /* "Aspidites/_vendor/pyparsing_extension.py":2538
 *                     except Exception as e:
 *                         out.append(result.dump(full=fullDump))
 *                         out.append("{0} failed: {1}: {2}".format(postParse.__name__, type(e).__name__, e))             # <<<<<<<<<<<<<<
 *                 else:
 *                     out.append(result.dump(full=fullDump))
 */
              __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_0_failed_1_2, __pyx_n_s_format); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2538, __pyx_L27_except_error)
              __Pyx_GOTREF(__pyx_t_8);
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_postParse, __pyx_n_s_name_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2538, __pyx_L27_except_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_21 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(__pyx_v_e)), __pyx_n_s_name_2); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 2538, __pyx_L27_except_error)
              __Pyx_GOTREF(__pyx_t_21);
              __pyx_t_22 = NULL;
              __pyx_t_12 = 0;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
                __pyx_t_22 = PyMethod_GET_SELF(__pyx_t_8);
                if (likely(__pyx_t_22)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
                  __Pyx_INCREF(__pyx_t_22);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_8, function);
                  __pyx_t_12 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_8)) {
                PyObject *__pyx_temp[4] = {__pyx_t_22, __pyx_t_5, __pyx_t_21, __pyx_v_e};
                __pyx_t_20 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_12, 3+__pyx_t_12); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 2538, __pyx_L27_except_error)
                __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
                __Pyx_GOTREF(__pyx_t_20);
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
                PyObject *__pyx_temp[4] = {__pyx_t_22, __pyx_t_5, __pyx_t_21, __pyx_v_e};
                __pyx_t_20 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_12, 3+__pyx_t_12); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 2538, __pyx_L27_except_error)
                __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
                __Pyx_GOTREF(__pyx_t_20);
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
              } else
              #endif
              {
                __pyx_t_23 = PyTuple_New(3+__pyx_t_12); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 2538, __pyx_L27_except_error)
                __Pyx_GOTREF(__pyx_t_23);
                if (__pyx_t_22) {
                  __Pyx_GIVEREF(__pyx_t_22); PyTuple_SET_ITEM(__pyx_t_23, 0, __pyx_t_22); __pyx_t_22 = NULL;
                }
                __Pyx_GIVEREF(__pyx_t_5);
                PyTuple_SET_ITEM(__pyx_t_23, 0+__pyx_t_12, __pyx_t_5);
                __Pyx_GIVEREF(__pyx_t_21);
                PyTuple_SET_ITEM(__pyx_t_23, 1+__pyx_t_12, __pyx_t_21);
                __Pyx_INCREF(__pyx_v_e);
                __Pyx_GIVEREF(__pyx_v_e);
                PyTuple_SET_ITEM(__pyx_t_23, 2+__pyx_t_12, __pyx_v_e);
                __pyx_t_5 = 0;
                __pyx_t_21 = 0;
                __pyx_t_20 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_23, NULL); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 2538, __pyx_L27_except_error)
                __Pyx_GOTREF(__pyx_t_20);
                __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
              }
              __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_out, __pyx_t_20); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 2538, __pyx_L27_except_error)
              __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              goto __pyx_L26_exception_handled;
            }
            goto __pyx_L27_except_error;
            __pyx_L27_except_error:;

            /* "Aspidites/_vendor/pyparsing_extension.py":2527
 *                 success = success and not failureTests
 *                 if postParse is not None:
 *                     try:             # <<<<<<<<<<<<<<
 *                         pp_value = postParse(t, result)
 *                         if pp_value is not None:
 */
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
            goto __pyx_L16_except_error;
            __pyx_L26_exception_handled:;
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
            __pyx_L32_try_end:;
          }

          /* "Aspidites/_vendor/pyparsing_extension.py":2526
 *             else:
 *                 success = success and not failureTests
 *                 if postParse is not None:             # <<<<<<<<<<<<<<
 *                     try:
 *                         pp_value = postParse(t, result)
 */
          goto __pyx_L24;
        }

        /* "Aspidites/_vendor/pyparsing_extension.py":2540
 *                         out.append("{0} failed: {1}: {2}".format(postParse.__name__, type(e).__name__, e))
 *                 else:
 *                     out.append(result.dump(full=fullDump))             # <<<<<<<<<<<<<<
 * 
 *             if printResults:
 */
        /*else*/ {
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_result, __pyx_n_s_dump); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2540, __pyx_L16_except_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2540, __pyx_L16_except_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_full, __pyx_v_fullDump) < 0) __PYX_ERR(0, 2540, __pyx_L16_except_error)
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_empty_tuple, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2540, __pyx_L16_except_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_out, __pyx_t_2); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 2540, __pyx_L16_except_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }
        __pyx_L24:;
      }
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
      goto __pyx_L21_try_end;
      __pyx_L14_error:;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":2510
 *                 t = NL.transformString(t.lstrip(BOM))
 *                 result = self.parseString(t, parseAll=parseAll)
 *             except ParseBaseException as pe:             # <<<<<<<<<<<<<<
 *                 fatal = "(FATAL)" if isinstance(pe, ParseFatalException) else ""
 *                 if '\n' in t:
 */
      __Pyx_ErrFetch(&__pyx_t_2, &__pyx_t_6, &__pyx_t_7);
      __Pyx_GetModuleGlobalName(__pyx_t_20, __pyx_n_s_ParseBaseException); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 2510, __pyx_L16_except_error)
      __Pyx_GOTREF(__pyx_t_20);
      __pyx_t_12 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_2, __pyx_t_20);
      __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
      __Pyx_ErrRestore(__pyx_t_2, __pyx_t_6, __pyx_t_7);
      __pyx_t_2 = 0; __pyx_t_6 = 0; __pyx_t_7 = 0;
      if (__pyx_t_12) {
        __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.runTests", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_6, &__pyx_t_2) < 0) __PYX_ERR(0, 2510, __pyx_L16_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_XDECREF_SET(__pyx_v_pe, __pyx_t_6);

        /* "Aspidites/_vendor/pyparsing_extension.py":2511
 *                 result = self.parseString(t, parseAll=parseAll)
 *             except ParseBaseException as pe:
 *                 fatal = "(FATAL)" if isinstance(pe, ParseFatalException) else ""             # <<<<<<<<<<<<<<
 *                 if '\n' in t:
 *                     out.append(line(pe.loc, t))
 */
        __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_ParseFatalException); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2511, __pyx_L16_except_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_3 = PyObject_IsInstance(__pyx_v_pe, __pyx_t_8); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 2511, __pyx_L16_except_error)
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if ((__pyx_t_3 != 0)) {
          __Pyx_INCREF(__pyx_kp_s_FATAL);
          __pyx_t_20 = __pyx_kp_s_FATAL;
        } else {
          __Pyx_INCREF(__pyx_kp_s__12);
          __pyx_t_20 = __pyx_kp_s__12;
        }
        __Pyx_XDECREF_SET(__pyx_v_fatal, __pyx_t_20);
        __pyx_t_20 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":2512
 *             except ParseBaseException as pe:
 *                 fatal = "(FATAL)" if isinstance(pe, ParseFatalException) else ""
 *                 if '\n' in t:             # <<<<<<<<<<<<<<
 *                     out.append(line(pe.loc, t))
 *                     out.append(' ' * (col(pe.loc, t) - 1) + '^' + fatal)
 */
        __pyx_t_3 = (__Pyx_PySequence_ContainsTF(__pyx_kp_s__4, __pyx_v_t, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 2512, __pyx_L16_except_error)
        __pyx_t_4 = (__pyx_t_3 != 0);
        if (__pyx_t_4) {

          /* "Aspidites/_vendor/pyparsing_extension.py":2513
 *                 fatal = "(FATAL)" if isinstance(pe, ParseFatalException) else ""
 *                 if '\n' in t:
 *                     out.append(line(pe.loc, t))             # <<<<<<<<<<<<<<
 *                     out.append(' ' * (col(pe.loc, t) - 1) + '^' + fatal)
 *                 else:
 */
          __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_line); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2513, __pyx_L16_except_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_23 = __Pyx_PyObject_GetAttrStr(__pyx_v_pe, __pyx_n_s_loc); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 2513, __pyx_L16_except_error)
          __Pyx_GOTREF(__pyx_t_23);
          __pyx_t_21 = NULL;
          __pyx_t_12 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
            __pyx_t_21 = PyMethod_GET_SELF(__pyx_t_8);
            if (likely(__pyx_t_21)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
              __Pyx_INCREF(__pyx_t_21);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_8, function);
              __pyx_t_12 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_8)) {
            PyObject *__pyx_temp[3] = {__pyx_t_21, __pyx_t_23, __pyx_v_t};
            __pyx_t_20 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 2513, __pyx_L16_except_error)
            __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
            __Pyx_GOTREF(__pyx_t_20);
            __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
            PyObject *__pyx_temp[3] = {__pyx_t_21, __pyx_t_23, __pyx_v_t};
            __pyx_t_20 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 2513, __pyx_L16_except_error)
            __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
            __Pyx_GOTREF(__pyx_t_20);
            __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
          } else
          #endif
          {
            __pyx_t_5 = PyTuple_New(2+__pyx_t_12); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2513, __pyx_L16_except_error)
            __Pyx_GOTREF(__pyx_t_5);
            if (__pyx_t_21) {
              __Pyx_GIVEREF(__pyx_t_21); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_21); __pyx_t_21 = NULL;
            }
            __Pyx_GIVEREF(__pyx_t_23);
            PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_12, __pyx_t_23);
            __Pyx_INCREF(__pyx_v_t);
            __Pyx_GIVEREF(__pyx_v_t);
            PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_12, __pyx_v_t);
            __pyx_t_23 = 0;
            __pyx_t_20 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_5, NULL); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 2513, __pyx_L16_except_error)
            __Pyx_GOTREF(__pyx_t_20);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          }
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_out, __pyx_t_20); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 2513, __pyx_L16_except_error)
          __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":2514
 *                 if '\n' in t:
 *                     out.append(line(pe.loc, t))
 *                     out.append(' ' * (col(pe.loc, t) - 1) + '^' + fatal)             # <<<<<<<<<<<<<<
 *                 else:
 *                     out.append(' ' * pe.loc + '^' + fatal)
 */
          __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_col); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2514, __pyx_L16_except_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_pe, __pyx_n_s_loc); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2514, __pyx_L16_except_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_23 = NULL;
          __pyx_t_12 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
            __pyx_t_23 = PyMethod_GET_SELF(__pyx_t_8);
            if (likely(__pyx_t_23)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
              __Pyx_INCREF(__pyx_t_23);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_8, function);
              __pyx_t_12 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_8)) {
            PyObject *__pyx_temp[3] = {__pyx_t_23, __pyx_t_5, __pyx_v_t};
            __pyx_t_20 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 2514, __pyx_L16_except_error)
            __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
            __Pyx_GOTREF(__pyx_t_20);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
            PyObject *__pyx_temp[3] = {__pyx_t_23, __pyx_t_5, __pyx_v_t};
            __pyx_t_20 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 2514, __pyx_L16_except_error)
            __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
            __Pyx_GOTREF(__pyx_t_20);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          } else
          #endif
          {
            __pyx_t_21 = PyTuple_New(2+__pyx_t_12); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 2514, __pyx_L16_except_error)
            __Pyx_GOTREF(__pyx_t_21);
            if (__pyx_t_23) {
              __Pyx_GIVEREF(__pyx_t_23); PyTuple_SET_ITEM(__pyx_t_21, 0, __pyx_t_23); __pyx_t_23 = NULL;
            }
            __Pyx_GIVEREF(__pyx_t_5);
            PyTuple_SET_ITEM(__pyx_t_21, 0+__pyx_t_12, __pyx_t_5);
            __Pyx_INCREF(__pyx_v_t);
            __Pyx_GIVEREF(__pyx_v_t);
            PyTuple_SET_ITEM(__pyx_t_21, 1+__pyx_t_12, __pyx_v_t);
            __pyx_t_5 = 0;
            __pyx_t_20 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_21, NULL); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 2514, __pyx_L16_except_error)
            __Pyx_GOTREF(__pyx_t_20);
            __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
          }
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __pyx_t_8 = __Pyx_PyInt_SubtractObjC(__pyx_t_20, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2514, __pyx_L16_except_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
          __pyx_t_20 = PyNumber_Multiply(__pyx_kp_s__29, __pyx_t_8); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 2514, __pyx_L16_except_error)
          __Pyx_GOTREF(__pyx_t_20);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __pyx_t_8 = PyNumber_Add(__pyx_t_20, __pyx_kp_s__30); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2514, __pyx_L16_except_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
          __pyx_t_20 = PyNumber_Add(__pyx_t_8, __pyx_v_fatal); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 2514, __pyx_L16_except_error)
          __Pyx_GOTREF(__pyx_t_20);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_out, __pyx_t_20); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 2514, __pyx_L16_except_error)
          __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":2512
 *             except ParseBaseException as pe:
 *                 fatal = "(FATAL)" if isinstance(pe, ParseFatalException) else ""
 *                 if '\n' in t:             # <<<<<<<<<<<<<<
 *                     out.append(line(pe.loc, t))
 *                     out.append(' ' * (col(pe.loc, t) - 1) + '^' + fatal)
 */
          goto __pyx_L39;
        }

        /* "Aspidites/_vendor/pyparsing_extension.py":2516
 *                     out.append(' ' * (col(pe.loc, t) - 1) + '^' + fatal)
 *                 else:
 *                     out.append(' ' * pe.loc + '^' + fatal)             # <<<<<<<<<<<<<<
 *                 out.append("FAIL: " + str(pe))
 *                 success = success and failureTests
 */
        /*else*/ {
          __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_v_pe, __pyx_n_s_loc); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 2516, __pyx_L16_except_error)
          __Pyx_GOTREF(__pyx_t_20);
          __pyx_t_8 = PyNumber_Multiply(__pyx_kp_s__29, __pyx_t_20); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2516, __pyx_L16_except_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
          __pyx_t_20 = PyNumber_Add(__pyx_t_8, __pyx_kp_s__30); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 2516, __pyx_L16_except_error)
          __Pyx_GOTREF(__pyx_t_20);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __pyx_t_8 = PyNumber_Add(__pyx_t_20, __pyx_v_fatal); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2516, __pyx_L16_except_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
          __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_out, __pyx_t_8); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 2516, __pyx_L16_except_error)
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        __pyx_L39:;

        /* "Aspidites/_vendor/pyparsing_extension.py":2517
 *                 else:
 *                     out.append(' ' * pe.loc + '^' + fatal)
 *                 out.append("FAIL: " + str(pe))             # <<<<<<<<<<<<<<
 *                 success = success and failureTests
 *                 result = pe
 */
        __pyx_t_8 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_v_pe); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2517, __pyx_L16_except_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_20 = PyNumber_Add(__pyx_kp_s_FAIL, __pyx_t_8); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 2517, __pyx_L16_except_error)
        __Pyx_GOTREF(__pyx_t_20);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_out, __pyx_t_20); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 2517, __pyx_L16_except_error)
        __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":2518
 *                     out.append(' ' * pe.loc + '^' + fatal)
 *                 out.append("FAIL: " + str(pe))
 *                 success = success and failureTests             # <<<<<<<<<<<<<<
 *                 result = pe
 *             except Exception as exc:
 */
        __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_success); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 2518, __pyx_L16_except_error)
        if (__pyx_t_4) {
        } else {
          __Pyx_INCREF(__pyx_v_success);
          __pyx_t_20 = __pyx_v_success;
          goto __pyx_L40_bool_binop_done;
        }
        __Pyx_INCREF(__pyx_v_failureTests);
        __pyx_t_20 = __pyx_v_failureTests;
        __pyx_L40_bool_binop_done:;
        __Pyx_DECREF_SET(__pyx_v_success, __pyx_t_20);
        __pyx_t_20 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":2519
 *                 out.append("FAIL: " + str(pe))
 *                 success = success and failureTests
 *                 result = pe             # <<<<<<<<<<<<<<
 *             except Exception as exc:
 *                 out.append("FAIL-EXCEPTION: " + str(exc))
 */
        __Pyx_INCREF(__pyx_v_pe);
        __Pyx_XDECREF_SET(__pyx_v_result, __pyx_v_pe);
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        goto __pyx_L15_exception_handled;
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":2520
 *                 success = success and failureTests
 *                 result = pe
 *             except Exception as exc:             # <<<<<<<<<<<<<<
 *                 out.append("FAIL-EXCEPTION: " + str(exc))
 *                 success = success and failureTests
 */
      __pyx_t_12 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
      if (__pyx_t_12) {
        __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.runTests", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(0, 2520, __pyx_L16_except_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_XDECREF_SET(__pyx_v_exc, __pyx_t_6);

        /* "Aspidites/_vendor/pyparsing_extension.py":2521
 *                 result = pe
 *             except Exception as exc:
 *                 out.append("FAIL-EXCEPTION: " + str(exc))             # <<<<<<<<<<<<<<
 *                 success = success and failureTests
 *                 result = exc
 */
        __pyx_t_20 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_v_exc); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 2521, __pyx_L16_except_error)
        __Pyx_GOTREF(__pyx_t_20);
        __pyx_t_8 = PyNumber_Add(__pyx_kp_s_FAIL_EXCEPTION, __pyx_t_20); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2521, __pyx_L16_except_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
        __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_out, __pyx_t_8); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 2521, __pyx_L16_except_error)
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":2522
 *             except Exception as exc:
 *                 out.append("FAIL-EXCEPTION: " + str(exc))
 *                 success = success and failureTests             # <<<<<<<<<<<<<<
 *                 result = exc
 *             else:
 */
        __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_success); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 2522, __pyx_L16_except_error)
        if (__pyx_t_4) {
        } else {
          __Pyx_INCREF(__pyx_v_success);
          __pyx_t_8 = __pyx_v_success;
          goto __pyx_L44_bool_binop_done;
        }
        __Pyx_INCREF(__pyx_v_failureTests);
        __pyx_t_8 = __pyx_v_failureTests;
        __pyx_L44_bool_binop_done:;
        __Pyx_DECREF_SET(__pyx_v_success, __pyx_t_8);
        __pyx_t_8 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":2523
 *                 out.append("FAIL-EXCEPTION: " + str(exc))
 *                 success = success and failureTests
 *                 result = exc             # <<<<<<<<<<<<<<
 *             else:
 *                 success = success and not failureTests
 */
        __Pyx_INCREF(__pyx_v_exc);
        __Pyx_XDECREF_SET(__pyx_v_result, __pyx_v_exc);
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L15_exception_handled;
      }
      goto __pyx_L16_except_error;
      __pyx_L16_except_error:;

      /* "Aspidites/_vendor/pyparsing_extension.py":2506
 *             out = ['\n' + '\n'.join(comments) if comments else '', t]
 *             comments = []
 *             try:             # <<<<<<<<<<<<<<
 *                 # convert newline marks to actual newlines, and strip leading BOM if present
 *                 t = NL.transformString(t.lstrip(BOM))
 */
      __Pyx_XGIVEREF(__pyx_t_14);
      __Pyx_XGIVEREF(__pyx_t_15);
      __Pyx_XGIVEREF(__pyx_t_16);
      __Pyx_ExceptionReset(__pyx_t_14, __pyx_t_15, __pyx_t_16);
      goto __pyx_L1_error;
      __pyx_L15_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_14);
      __Pyx_XGIVEREF(__pyx_t_15);
      __Pyx_XGIVEREF(__pyx_t_16);
      __Pyx_ExceptionReset(__pyx_t_14, __pyx_t_15, __pyx_t_16);
      __pyx_L21_try_end:;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":2542
 *                     out.append(result.dump(full=fullDump))
 * 
 *             if printResults:             # <<<<<<<<<<<<<<
 *                 if fullDump:
 *                     out.append('')
 */
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_printResults); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 2542, __pyx_L1_error)
    if (__pyx_t_4) {

      /* "Aspidites/_vendor/pyparsing_extension.py":2543
 * 
 *             if printResults:
 *                 if fullDump:             # <<<<<<<<<<<<<<
 *                     out.append('')
 *                 print_('\n'.join(out))
 */
      __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_fullDump); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 2543, __pyx_L1_error)
      if (__pyx_t_4) {

        /* "Aspidites/_vendor/pyparsing_extension.py":2544
 *             if printResults:
 *                 if fullDump:
 *                     out.append('')             # <<<<<<<<<<<<<<
 *                 print_('\n'.join(out))
 * 
 */
        __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_out, __pyx_kp_s__12); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 2544, __pyx_L1_error)

        /* "Aspidites/_vendor/pyparsing_extension.py":2543
 * 
 *             if printResults:
 *                 if fullDump:             # <<<<<<<<<<<<<<
 *                     out.append('')
 *                 print_('\n'.join(out))
 */
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":2545
 *                 if fullDump:
 *                     out.append('')
 *                 print_('\n'.join(out))             # <<<<<<<<<<<<<<
 * 
 *             allResults.append((t, result))
 */
      __pyx_t_6 = __Pyx_PyString_Join(__pyx_kp_s__4, __pyx_v_out); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2545, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_v_print_);
      __pyx_t_2 = __pyx_v_print_; __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_7 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2545, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":2542
 *                     out.append(result.dump(full=fullDump))
 * 
 *             if printResults:             # <<<<<<<<<<<<<<
 *                 if fullDump:
 *                     out.append('')
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":2547
 *                 print_('\n'.join(out))
 * 
 *             allResults.append((t, result))             # <<<<<<<<<<<<<<
 * 
 *         return success, allResults
 */
    __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2547, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_v_t);
    __Pyx_GIVEREF(__pyx_v_t);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_t);
    __Pyx_INCREF(__pyx_v_result);
    __Pyx_GIVEREF(__pyx_v_result);
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_v_result);
    __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_allResults, __pyx_t_7); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 2547, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2498
 *         NL = Literal(r'\n').addParseAction(replaceWith('\n')).ignore(quotedString)
 *         BOM = u'\ufeff'
 *         for t in tests:             # <<<<<<<<<<<<<<
 *             if comment is not None and comment.matches(t, False) or comments and not t:
 *                 comments.append(t)
 */
    __pyx_L6_continue:;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2549
 *             allResults.append((t, result))
 * 
 *         return success, allResults             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_success);
  __Pyx_GIVEREF(__pyx_v_success);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_success);
  __Pyx_INCREF(__pyx_v_allResults);
  __Pyx_GIVEREF(__pyx_v_allResults);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_allResults);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2388
 *             return False
 * 
 *     def runTests(self, tests, parseAll=True, comment='#',             # <<<<<<<<<<<<<<
 *                  fullDump=True, printResults=True, failureTests=False, postParse=None,
 *                  file=None):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_XDECREF(__pyx_t_22);
  __Pyx_XDECREF(__pyx_t_23);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParserElement.runTests", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_Literal);
  __Pyx_XDECREF(__pyx_v_replaceWith);
  __Pyx_XDECREF(__pyx_v_quotedString);
  __Pyx_XDECREF(__pyx_v_print_);
  __Pyx_XDECREF(__pyx_v_allResults);
  __Pyx_XDECREF(__pyx_v_comments);
  __Pyx_XDECREF(__pyx_v_success);
  __Pyx_XDECREF(__pyx_v_NL);
  __Pyx_XDECREF(__pyx_v_BOM);
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_XDECREF(__pyx_v_out);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_pe);
  __Pyx_XDECREF(__pyx_v_fatal);
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XDECREF(__pyx_v_pp_value);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XDECREF(__pyx_v_tests);
  __Pyx_XDECREF(__pyx_v_comment);
  __Pyx_XDECREF(__pyx_v_file);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2557
 *     """
 * 
 *     def __init__(self, exprs, savelist=False):             # <<<<<<<<<<<<<<
 *         super(ParseExpression, self).__init__(savelist)
 *         if isinstance(exprs, _generatorType):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_exprs = 0;
  PyObject *__pyx_v_savelist = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_exprs,&__pyx_n_s_savelist,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_False));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_exprs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, 1); __PYX_ERR(0, 2557, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_savelist);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 2557, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_exprs = values[1];
    __pyx_v_savelist = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2557, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseExpression.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression___init__(__pyx_self, __pyx_v_self, __pyx_v_exprs, __pyx_v_savelist);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_8__init___2generator14(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/pyparsing_extension.py":2569
 *             exprs = list(exprs)
 *             # if sequence of strings provided, wrap with Literal
 *             if any(isinstance(expr, str) for expr in exprs):             # <<<<<<<<<<<<<<
 *                 exprs = (self._literalStringClass(e) if isinstance(e, str) else e for e in exprs)
 *             self.exprs = list(exprs)
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_8__init___genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 2569, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__ *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_8__init___2generator14, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_ParseExpression___init___locals, __pyx_n_s_Aspidites__vendor_pyparsing_exte); if (unlikely(!gen)) __PYX_ERR(0, 2569, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseExpression.__init__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_8__init___2generator14(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2569, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_exprs)) { __Pyx_RaiseClosureNameError("exprs"); __PYX_ERR(0, 2569, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_exprs)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_exprs)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_exprs; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_exprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2569, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2569, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 2569, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2569, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 2569, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2569, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 2569, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_expr);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_expr, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_5 = PyString_Check(__pyx_cur_scope->__pyx_v_expr); 
    __pyx_t_6 = (__pyx_t_5 != 0);
    if (__pyx_t_6) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_True);
      __pyx_r = Py_True;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_8__init___5generator15(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/pyparsing_extension.py":2570
 *             # if sequence of strings provided, wrap with Literal
 *             if any(isinstance(expr, str) for expr in exprs):
 *                 exprs = (self._literalStringClass(e) if isinstance(e, str) else e for e in exprs)             # <<<<<<<<<<<<<<
 *             self.exprs = list(exprs)
 *         else:
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_8__init___3genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 2570, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__ *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_8__init___5generator15, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_ParseExpression___init___locals, __pyx_n_s_Aspidites__vendor_pyparsing_exte); if (unlikely(!gen)) __PYX_ERR(0, 2570, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseExpression.__init__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_8__init___5generator15(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2570, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_exprs)) { __Pyx_RaiseClosureNameError("exprs"); __PYX_ERR(0, 2570, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_exprs)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_exprs)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_exprs; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_exprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2570, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2570, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 2570, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2570, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 2570, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2570, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 2570, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_e);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_e, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_5 = PyString_Check(__pyx_cur_scope->__pyx_v_e); 
    if ((__pyx_t_5 != 0)) {
      if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 2570, __pyx_L1_error) }
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_literalStringClass); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2570, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      __pyx_t_6 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_cur_scope->__pyx_v_e) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_cur_scope->__pyx_v_e);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2570, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_4 = __pyx_t_6;
      __pyx_t_6 = 0;
    } else {
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_e);
      __pyx_t_4 = __pyx_cur_scope->__pyx_v_e;
    }
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2570, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2557
 *     """
 * 
 *     def __init__(self, exprs, savelist=False):             # <<<<<<<<<<<<<<
 *         super(ParseExpression, self).__init__(savelist)
 *         if isinstance(exprs, _generatorType):
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_exprs, PyObject *__pyx_v_savelist) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__ *__pyx_cur_scope;
  PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_8__init___2generator14 = 0;
  PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_8__init___5generator15 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__ *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 2557, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_exprs = __pyx_v_exprs;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_exprs);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_exprs);

  /* "Aspidites/_vendor/pyparsing_extension.py":2558
 * 
 *     def __init__(self, exprs, savelist=False):
 *         super(ParseExpression, self).__init__(savelist)             # <<<<<<<<<<<<<<
 *         if isinstance(exprs, _generatorType):
 *             exprs = list(exprs)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParseExpression); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2558, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2558, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_cur_scope->__pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2558, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2558, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_savelist) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_savelist);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2558, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2559
 *     def __init__(self, exprs, savelist=False):
 *         super(ParseExpression, self).__init__(savelist)
 *         if isinstance(exprs, _generatorType):             # <<<<<<<<<<<<<<
 *             exprs = list(exprs)
 * 
 */
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_exprs;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_generatorType); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2559, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyObject_IsInstance(__pyx_t_1, __pyx_t_3); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 2559, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2560
 *         super(ParseExpression, self).__init__(savelist)
 *         if isinstance(exprs, _generatorType):
 *             exprs = list(exprs)             # <<<<<<<<<<<<<<
 * 
 *         if isinstance(exprs, str):
 */
    __pyx_t_3 = PySequence_List(__pyx_cur_scope->__pyx_v_exprs); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2560, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_exprs);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_exprs, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2559
 *     def __init__(self, exprs, savelist=False):
 *         super(ParseExpression, self).__init__(savelist)
 *         if isinstance(exprs, _generatorType):             # <<<<<<<<<<<<<<
 *             exprs = list(exprs)
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2562
 *             exprs = list(exprs)
 * 
 *         if isinstance(exprs, str):             # <<<<<<<<<<<<<<
 *             self.exprs = [self._literalStringClass(exprs)]
 *         elif isinstance(exprs, ParserElement):
 */
  __pyx_t_3 = __pyx_cur_scope->__pyx_v_exprs;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_5 = PyString_Check(__pyx_t_3); 
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = (__pyx_t_5 != 0);
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2563
 * 
 *         if isinstance(exprs, str):
 *             self.exprs = [self._literalStringClass(exprs)]             # <<<<<<<<<<<<<<
 *         elif isinstance(exprs, ParserElement):
 *             self.exprs = [exprs]
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_literalStringClass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2563, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_cur_scope->__pyx_v_exprs) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_cur_scope->__pyx_v_exprs);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2563, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2563, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_3);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
    __pyx_t_3 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_exprs, __pyx_t_1) < 0) __PYX_ERR(0, 2563, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2562
 *             exprs = list(exprs)
 * 
 *         if isinstance(exprs, str):             # <<<<<<<<<<<<<<
 *             self.exprs = [self._literalStringClass(exprs)]
 *         elif isinstance(exprs, ParserElement):
 */
    goto __pyx_L4;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2564
 *         if isinstance(exprs, str):
 *             self.exprs = [self._literalStringClass(exprs)]
 *         elif isinstance(exprs, ParserElement):             # <<<<<<<<<<<<<<
 *             self.exprs = [exprs]
 *         elif isinstance(exprs, Iterable):
 */
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_exprs;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyObject_IsInstance(__pyx_t_1, __pyx_t_3); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 2564, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2565
 *             self.exprs = [self._literalStringClass(exprs)]
 *         elif isinstance(exprs, ParserElement):
 *             self.exprs = [exprs]             # <<<<<<<<<<<<<<
 *         elif isinstance(exprs, Iterable):
 *             exprs = list(exprs)
 */
    __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2565, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_exprs);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_exprs);
    PyList_SET_ITEM(__pyx_t_3, 0, __pyx_cur_scope->__pyx_v_exprs);
    if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_exprs, __pyx_t_3) < 0) __PYX_ERR(0, 2565, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2564
 *         if isinstance(exprs, str):
 *             self.exprs = [self._literalStringClass(exprs)]
 *         elif isinstance(exprs, ParserElement):             # <<<<<<<<<<<<<<
 *             self.exprs = [exprs]
 *         elif isinstance(exprs, Iterable):
 */
    goto __pyx_L4;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2566
 *         elif isinstance(exprs, ParserElement):
 *             self.exprs = [exprs]
 *         elif isinstance(exprs, Iterable):             # <<<<<<<<<<<<<<
 *             exprs = list(exprs)
 *             # if sequence of strings provided, wrap with Literal
 */
  __pyx_t_3 = __pyx_cur_scope->__pyx_v_exprs;
  __Pyx_INCREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Iterable); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyObject_IsInstance(__pyx_t_3, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 2566, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = (__pyx_t_5 != 0);
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2567
 *             self.exprs = [exprs]
 *         elif isinstance(exprs, Iterable):
 *             exprs = list(exprs)             # <<<<<<<<<<<<<<
 *             # if sequence of strings provided, wrap with Literal
 *             if any(isinstance(expr, str) for expr in exprs):
 */
    __pyx_t_1 = PySequence_List(__pyx_cur_scope->__pyx_v_exprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2567, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_exprs);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_exprs, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2569
 *             exprs = list(exprs)
 *             # if sequence of strings provided, wrap with Literal
 *             if any(isinstance(expr, str) for expr in exprs):             # <<<<<<<<<<<<<<
 *                 exprs = (self._literalStringClass(e) if isinstance(e, str) else e for e in exprs)
 *             self.exprs = list(exprs)
 */
    __pyx_t_1 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_8__init___genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2569, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_Generator_Next(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2569, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 2569, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_4) {

      /* "Aspidites/_vendor/pyparsing_extension.py":2570
 *             # if sequence of strings provided, wrap with Literal
 *             if any(isinstance(expr, str) for expr in exprs):
 *                 exprs = (self._literalStringClass(e) if isinstance(e, str) else e for e in exprs)             # <<<<<<<<<<<<<<
 *             self.exprs = list(exprs)
 *         else:
 */
      __pyx_t_3 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_8__init___3genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2570, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_exprs);
      __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_exprs, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_3);
      __pyx_t_3 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":2569
 *             exprs = list(exprs)
 *             # if sequence of strings provided, wrap with Literal
 *             if any(isinstance(expr, str) for expr in exprs):             # <<<<<<<<<<<<<<
 *                 exprs = (self._literalStringClass(e) if isinstance(e, str) else e for e in exprs)
 *             self.exprs = list(exprs)
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":2571
 *             if any(isinstance(expr, str) for expr in exprs):
 *                 exprs = (self._literalStringClass(e) if isinstance(e, str) else e for e in exprs)
 *             self.exprs = list(exprs)             # <<<<<<<<<<<<<<
 *         else:
 *             try:
 */
    __pyx_t_3 = PySequence_List(__pyx_cur_scope->__pyx_v_exprs); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2571, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_exprs, __pyx_t_3) < 0) __PYX_ERR(0, 2571, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2566
 *         elif isinstance(exprs, ParserElement):
 *             self.exprs = [exprs]
 *         elif isinstance(exprs, Iterable):             # <<<<<<<<<<<<<<
 *             exprs = list(exprs)
 *             # if sequence of strings provided, wrap with Literal
 */
    goto __pyx_L4;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2573
 *             self.exprs = list(exprs)
 *         else:
 *             try:             # <<<<<<<<<<<<<<
 *                 self.exprs = list(exprs)
 *             except TypeError:
 */
  /*else*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      /*try:*/ {

        /* "Aspidites/_vendor/pyparsing_extension.py":2574
 *         else:
 *             try:
 *                 self.exprs = list(exprs)             # <<<<<<<<<<<<<<
 *             except TypeError:
 *                 self.exprs = [exprs]
 */
        __pyx_t_3 = PySequence_List(__pyx_cur_scope->__pyx_v_exprs); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2574, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_exprs, __pyx_t_3) < 0) __PYX_ERR(0, 2574, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":2573
 *             self.exprs = list(exprs)
 *         else:
 *             try:             # <<<<<<<<<<<<<<
 *                 self.exprs = list(exprs)
 *             except TypeError:
 */
      }
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L11_try_end;
      __pyx_L6_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":2575
 *             try:
 *                 self.exprs = list(exprs)
 *             except TypeError:             # <<<<<<<<<<<<<<
 *                 self.exprs = [exprs]
 *         self.callPreparse = False
 */
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
      if (__pyx_t_9) {
        __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseExpression.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_1, &__pyx_t_2) < 0) __PYX_ERR(0, 2575, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_2);

        /* "Aspidites/_vendor/pyparsing_extension.py":2576
 *                 self.exprs = list(exprs)
 *             except TypeError:
 *                 self.exprs = [exprs]             # <<<<<<<<<<<<<<
 *         self.callPreparse = False
 * 
 */
        __pyx_t_10 = PyList_New(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2576, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_INCREF(__pyx_cur_scope->__pyx_v_exprs);
        __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_exprs);
        PyList_SET_ITEM(__pyx_t_10, 0, __pyx_cur_scope->__pyx_v_exprs);
        if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_exprs, __pyx_t_10) < 0) __PYX_ERR(0, 2576, __pyx_L8_except_error)
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;
      __pyx_L8_except_error:;

      /* "Aspidites/_vendor/pyparsing_extension.py":2573
 *             self.exprs = list(exprs)
 *         else:
 *             try:             # <<<<<<<<<<<<<<
 *                 self.exprs = list(exprs)
 *             except TypeError:
 */
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
      goto __pyx_L1_error;
      __pyx_L7_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
      __pyx_L11_try_end:;
    }
  }
  __pyx_L4:;

  /* "Aspidites/_vendor/pyparsing_extension.py":2577
 *             except TypeError:
 *                 self.exprs = [exprs]
 *         self.callPreparse = False             # <<<<<<<<<<<<<<
 * 
 *     def append(self, other):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_callPreparse, Py_False) < 0) __PYX_ERR(0, 2577, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2557
 *     """
 * 
 *     def __init__(self, exprs, savelist=False):             # <<<<<<<<<<<<<<
 *         super(ParseExpression, self).__init__(savelist)
 *         if isinstance(exprs, _generatorType):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseExpression.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_8__init___2generator14);
  __Pyx_XDECREF(__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_8__init___5generator15);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2579
 *         self.callPreparse = False
 * 
 *     def append(self, other):             # <<<<<<<<<<<<<<
 *         self.exprs.append(other)
 *         self.strRepr = None
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_3append(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_3append = {"append", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_3append, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_3append(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_other = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("append (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_other,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_other)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("append", 1, 2, 2, 1); __PYX_ERR(0, 2579, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "append") < 0)) __PYX_ERR(0, 2579, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_other = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("append", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2579, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseExpression.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_2append(__pyx_self, __pyx_v_self, __pyx_v_other);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_2append(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("append", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2580
 * 
 *     def append(self, other):
 *         self.exprs.append(other)             # <<<<<<<<<<<<<<
 *         self.strRepr = None
 *         return self
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2580, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Append(__pyx_t_1, __pyx_v_other); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 2580, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2581
 *     def append(self, other):
 *         self.exprs.append(other)
 *         self.strRepr = None             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_strRepr, Py_None) < 0) __PYX_ERR(0, 2581, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2582
 *         self.exprs.append(other)
 *         self.strRepr = None
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def leaveWhitespace(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2579
 *         self.callPreparse = False
 * 
 *     def append(self, other):             # <<<<<<<<<<<<<<
 *         self.exprs.append(other)
 *         self.strRepr = None
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseExpression.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2584
 *         return self
 * 
 *     def leaveWhitespace(self):             # <<<<<<<<<<<<<<
 *         """Extends ``leaveWhitespace`` defined in base class, and also invokes ``leaveWhitespace`` on
 *            all contained expressions."""
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_5leaveWhitespace(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_4leaveWhitespace[] = "Extends ``leaveWhitespace`` defined in base class, and also invokes ``leaveWhitespace`` on\n           all contained expressions.";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_5leaveWhitespace = {"leaveWhitespace", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_5leaveWhitespace, METH_O, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_4leaveWhitespace};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_5leaveWhitespace(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("leaveWhitespace (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_4leaveWhitespace(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_4leaveWhitespace(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("leaveWhitespace", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2587
 *         """Extends ``leaveWhitespace`` defined in base class, and also invokes ``leaveWhitespace`` on
 *            all contained expressions."""
 *         self.skipWhitespace = False             # <<<<<<<<<<<<<<
 *         self.exprs = [e.copy() for e in self.exprs]
 *         for e in self.exprs:
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_skipWhitespace, Py_False) < 0) __PYX_ERR(0, 2587, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2588
 *            all contained expressions."""
 *         self.skipWhitespace = False
 *         self.exprs = [e.copy() for e in self.exprs]             # <<<<<<<<<<<<<<
 *         for e in self.exprs:
 *             e.leaveWhitespace()
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2588, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2588, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 2588, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2588, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 2588, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2588, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_5(__pyx_t_3);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 2588, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_copy); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2588, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_2 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2588, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 2588, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_exprs, __pyx_t_1) < 0) __PYX_ERR(0, 2588, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2589
 *         self.skipWhitespace = False
 *         self.exprs = [e.copy() for e in self.exprs]
 *         for e in self.exprs:             # <<<<<<<<<<<<<<
 *             e.leaveWhitespace()
 *         return self
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2589, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_3 = __pyx_t_1; __Pyx_INCREF(__pyx_t_3); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2589, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2589, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 2589, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2589, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 2589, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2589, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_5(__pyx_t_3);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 2589, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2590
 *         self.exprs = [e.copy() for e in self.exprs]
 *         for e in self.exprs:
 *             e.leaveWhitespace()             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_leaveWhitespace); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2590, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2590, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2589
 *         self.skipWhitespace = False
 *         self.exprs = [e.copy() for e in self.exprs]
 *         for e in self.exprs:             # <<<<<<<<<<<<<<
 *             e.leaveWhitespace()
 *         return self
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2591
 *         for e in self.exprs:
 *             e.leaveWhitespace()
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def ignore(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2584
 *         return self
 * 
 *     def leaveWhitespace(self):             # <<<<<<<<<<<<<<
 *         """Extends ``leaveWhitespace`` defined in base class, and also invokes ``leaveWhitespace`` on
 *            all contained expressions."""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseExpression.leaveWhitespace", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2593
 *         return self
 * 
 *     def ignore(self, other):             # <<<<<<<<<<<<<<
 *         from .pyparsing import Suppress
 *         if isinstance(other, Suppress):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_7ignore(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_7ignore = {"ignore", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_7ignore, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_7ignore(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_other = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ignore (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_other,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_other)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ignore", 1, 2, 2, 1); __PYX_ERR(0, 2593, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "ignore") < 0)) __PYX_ERR(0, 2593, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_other = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ignore", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2593, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseExpression.ignore", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_6ignore(__pyx_self, __pyx_v_self, __pyx_v_other);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_6ignore(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_v_Suppress = NULL;
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ignore", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2594
 * 
 *     def ignore(self, other):
 *         from .pyparsing import Suppress             # <<<<<<<<<<<<<<
 *         if isinstance(other, Suppress):
 *             if other not in self.ignoreExprs:
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2594, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_Suppress);
  __Pyx_GIVEREF(__pyx_n_s_Suppress);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_Suppress);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_pyparsing, __pyx_t_1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2594, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_Suppress); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2594, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_Suppress = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2595
 *     def ignore(self, other):
 *         from .pyparsing import Suppress
 *         if isinstance(other, Suppress):             # <<<<<<<<<<<<<<
 *             if other not in self.ignoreExprs:
 *                 super(ParseExpression, self).ignore(other)
 */
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_other, __pyx_v_Suppress); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 2595, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2596
 *         from .pyparsing import Suppress
 *         if isinstance(other, Suppress):
 *             if other not in self.ignoreExprs:             # <<<<<<<<<<<<<<
 *                 super(ParseExpression, self).ignore(other)
 *                 for e in self.exprs:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ignoreExprs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2596, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = (__Pyx_PySequence_ContainsTF(__pyx_v_other, __pyx_t_2, Py_NE)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 2596, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_3 = (__pyx_t_4 != 0);
    if (__pyx_t_3) {

      /* "Aspidites/_vendor/pyparsing_extension.py":2597
 *         if isinstance(other, Suppress):
 *             if other not in self.ignoreExprs:
 *                 super(ParseExpression, self).ignore(other)             # <<<<<<<<<<<<<<
 *                 for e in self.exprs:
 *                     e.ignore(self.ignoreExprs[-1])
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParseExpression); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2597, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2597, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
      __Pyx_INCREF(__pyx_v_self);
      __Pyx_GIVEREF(__pyx_v_self);
      PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_self);
      __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2597, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ignore); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2597, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_1, __pyx_v_other) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_other);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2597, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":2598
 *             if other not in self.ignoreExprs:
 *                 super(ParseExpression, self).ignore(other)
 *                 for e in self.exprs:             # <<<<<<<<<<<<<<
 *                     e.ignore(self.ignoreExprs[-1])
 *         else:
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2598, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
        __pyx_t_5 = __pyx_t_2; __Pyx_INCREF(__pyx_t_5); __pyx_t_6 = 0;
        __pyx_t_7 = NULL;
      } else {
        __pyx_t_6 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2598, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_7 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2598, __pyx_L1_error)
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      for (;;) {
        if (likely(!__pyx_t_7)) {
          if (likely(PyList_CheckExact(__pyx_t_5))) {
            if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_5)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 2598, __pyx_L1_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_5, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2598, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
          } else {
            if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 2598, __pyx_L1_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_5, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2598, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
          }
        } else {
          __pyx_t_2 = __pyx_t_7(__pyx_t_5);
          if (unlikely(!__pyx_t_2)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 2598, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_2);
        }
        __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":2599
 *                 super(ParseExpression, self).ignore(other)
 *                 for e in self.exprs:
 *                     e.ignore(self.ignoreExprs[-1])             # <<<<<<<<<<<<<<
 *         else:
 *             super(ParseExpression, self).ignore(other)
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_ignore); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2599, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ignoreExprs); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2599, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_9 = __Pyx_GetItemInt(__pyx_t_8, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2599, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_8 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
          if (likely(__pyx_t_8)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_1, function);
          }
        }
        __pyx_t_2 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_8, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_9);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2599, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":2598
 *             if other not in self.ignoreExprs:
 *                 super(ParseExpression, self).ignore(other)
 *                 for e in self.exprs:             # <<<<<<<<<<<<<<
 *                     e.ignore(self.ignoreExprs[-1])
 *         else:
 */
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":2596
 *         from .pyparsing import Suppress
 *         if isinstance(other, Suppress):
 *             if other not in self.ignoreExprs:             # <<<<<<<<<<<<<<
 *                 super(ParseExpression, self).ignore(other)
 *                 for e in self.exprs:
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":2595
 *     def ignore(self, other):
 *         from .pyparsing import Suppress
 *         if isinstance(other, Suppress):             # <<<<<<<<<<<<<<
 *             if other not in self.ignoreExprs:
 *                 super(ParseExpression, self).ignore(other)
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2601
 *                     e.ignore(self.ignoreExprs[-1])
 *         else:
 *             super(ParseExpression, self).ignore(other)             # <<<<<<<<<<<<<<
 *             for e in self.exprs:
 *                 e.ignore(self.ignoreExprs[-1])
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParseExpression); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2601, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2601, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_self);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2601, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ignore); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2601, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_5 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_v_other) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_other);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2601, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2602
 *         else:
 *             super(ParseExpression, self).ignore(other)
 *             for e in self.exprs:             # <<<<<<<<<<<<<<
 *                 e.ignore(self.ignoreExprs[-1])
 *         return self
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2602, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (likely(PyList_CheckExact(__pyx_t_5)) || PyTuple_CheckExact(__pyx_t_5)) {
      __pyx_t_1 = __pyx_t_5; __Pyx_INCREF(__pyx_t_1); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2602, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2602, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_5); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 2602, __pyx_L1_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2602, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        } else {
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_5); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 2602, __pyx_L1_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2602, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        }
      } else {
        __pyx_t_5 = __pyx_t_7(__pyx_t_1);
        if (unlikely(!__pyx_t_5)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 2602, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_5);
      }
      __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":2603
 *             super(ParseExpression, self).ignore(other)
 *             for e in self.exprs:
 *                 e.ignore(self.ignoreExprs[-1])             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_ignore); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2603, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ignoreExprs); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2603, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_8 = __Pyx_GetItemInt(__pyx_t_9, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2603, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_5 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_9, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_8);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2603, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":2602
 *         else:
 *             super(ParseExpression, self).ignore(other)
 *             for e in self.exprs:             # <<<<<<<<<<<<<<
 *                 e.ignore(self.ignoreExprs[-1])
 *         return self
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_L3:;

  /* "Aspidites/_vendor/pyparsing_extension.py":2604
 *             for e in self.exprs:
 *                 e.ignore(self.ignoreExprs[-1])
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2593
 *         return self
 * 
 *     def ignore(self, other):             # <<<<<<<<<<<<<<
 *         from .pyparsing import Suppress
 *         if isinstance(other, Suppress):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseExpression.ignore", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_Suppress);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2606
 *         return self
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         try:
 *             return super(ParseExpression, self).__str__()
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_9__str__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_9__str__ = {"__str__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_9__str__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_9__str__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_8__str__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_8__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2607
 * 
 *     def __str__(self):
 *         try:             # <<<<<<<<<<<<<<
 *             return super(ParseExpression, self).__str__()
 *         except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "Aspidites/_vendor/pyparsing_extension.py":2608
 *     def __str__(self):
 *         try:
 *             return super(ParseExpression, self).__str__()             # <<<<<<<<<<<<<<
 *         except Exception:
 *             pass
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_ParseExpression); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2608, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2608, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5);
      __Pyx_INCREF(__pyx_v_self);
      __Pyx_GIVEREF(__pyx_v_self);
      PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_v_self);
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_6, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2608, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_str); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2608, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2608, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_r = __pyx_t_4;
      __pyx_t_4 = 0;
      goto __pyx_L7_try_return;

      /* "Aspidites/_vendor/pyparsing_extension.py":2607
 * 
 *     def __str__(self):
 *         try:             # <<<<<<<<<<<<<<
 *             return super(ParseExpression, self).__str__()
 *         except Exception:
 */
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2609
 *         try:
 *             return super(ParseExpression, self).__str__()
 *         except Exception:             # <<<<<<<<<<<<<<
 *             pass
 * 
 */
    __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_7) {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "Aspidites/_vendor/pyparsing_extension.py":2607
 * 
 *     def __str__(self):
 *         try:             # <<<<<<<<<<<<<<
 *             return super(ParseExpression, self).__str__()
 *         except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2612
 *             pass
 * 
 *         if self.strRepr is None:             # <<<<<<<<<<<<<<
 *             self.strRepr = "%s:(%s)" % (self.__class__.__name__, str(self.exprs))
 *         return self.strRepr
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_strRepr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_8 = (__pyx_t_4 == Py_None);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_9 = (__pyx_t_8 != 0);
  if (__pyx_t_9) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2613
 * 
 *         if self.strRepr is None:
 *             self.strRepr = "%s:(%s)" % (self.__class__.__name__, str(self.exprs))             # <<<<<<<<<<<<<<
 *         return self.strRepr
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_class); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2613, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_name_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2613, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2613, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2613, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2613, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_5);
    __pyx_t_6 = 0;
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_s_s_2, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2613, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_strRepr, __pyx_t_5) < 0) __PYX_ERR(0, 2613, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2612
 *             pass
 * 
 *         if self.strRepr is None:             # <<<<<<<<<<<<<<
 *             self.strRepr = "%s:(%s)" % (self.__class__.__name__, str(self.exprs))
 *         return self.strRepr
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2614
 *         if self.strRepr is None:
 *             self.strRepr = "%s:(%s)" % (self.__class__.__name__, str(self.exprs))
 *         return self.strRepr             # <<<<<<<<<<<<<<
 * 
 *     def streamline(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_strRepr); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2614, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2606
 *         return self
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         try:
 *             return super(ParseExpression, self).__str__()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseExpression.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2616
 *         return self.strRepr
 * 
 *     def streamline(self):             # <<<<<<<<<<<<<<
 *         super(ParseExpression, self).streamline()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_11streamline(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_11streamline = {"streamline", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_11streamline, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_11streamline(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("streamline (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_10streamline(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_10streamline(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_v_other = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("streamline", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2617
 * 
 *     def streamline(self):
 *         super(ParseExpression, self).streamline()             # <<<<<<<<<<<<<<
 * 
 *         for e in self.exprs:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParseExpression); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_streamline); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2619
 *         super(ParseExpression, self).streamline()
 * 
 *         for e in self.exprs:             # <<<<<<<<<<<<<<
 *             e.streamline()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_3 = __pyx_t_1; __Pyx_INCREF(__pyx_t_3); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2619, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2619, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 2619, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2619, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 2619, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2619, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_5(__pyx_t_3);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 2619, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2620
 * 
 *         for e in self.exprs:
 *             e.streamline()             # <<<<<<<<<<<<<<
 * 
 *         # collapse nested And's of the form And(And(And(a, b), c), d) to And(a, b, c, d)
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_streamline); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2620, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2620, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2619
 *         super(ParseExpression, self).streamline()
 * 
 *         for e in self.exprs:             # <<<<<<<<<<<<<<
 *             e.streamline()
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2625
 *         # but only if there are no parse actions or resultsNames on the nested And's
 *         # (likewise for Or's and MatchFirst's)
 *         if len(self.exprs) == 2:             # <<<<<<<<<<<<<<
 *             other = self.exprs[0]
 *             if (isinstance(other, self.__class__)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2625, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyObject_Length(__pyx_t_3); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2625, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_7 = ((__pyx_t_4 == 2) != 0);
  if (__pyx_t_7) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2626
 *         # (likewise for Or's and MatchFirst's)
 *         if len(self.exprs) == 2:
 *             other = self.exprs[0]             # <<<<<<<<<<<<<<
 *             if (isinstance(other, self.__class__)
 *                     and not other.parseAction
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2626, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2626, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_other = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2627
 *         if len(self.exprs) == 2:
 *             other = self.exprs[0]
 *             if (isinstance(other, self.__class__)             # <<<<<<<<<<<<<<
 *                     and not other.parseAction
 *                     and other.resultsName is None
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2627, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = PyObject_IsInstance(__pyx_v_other, __pyx_t_1); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 2627, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_9 = (__pyx_t_8 != 0);
    if (__pyx_t_9) {
    } else {
      __pyx_t_7 = __pyx_t_9;
      goto __pyx_L7_bool_binop_done;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":2628
 *             other = self.exprs[0]
 *             if (isinstance(other, self.__class__)
 *                     and not other.parseAction             # <<<<<<<<<<<<<<
 *                     and other.resultsName is None
 *                     and not other.debug):
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_parseAction); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2628, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 2628, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_8 = ((!__pyx_t_9) != 0);
    if (__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L7_bool_binop_done;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":2629
 *             if (isinstance(other, self.__class__)
 *                     and not other.parseAction
 *                     and other.resultsName is None             # <<<<<<<<<<<<<<
 *                     and not other.debug):
 *                 self.exprs = other.exprs[:] + [self.exprs[1]]
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_resultsName); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2629, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = (__pyx_t_1 == Py_None);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_9 = (__pyx_t_8 != 0);
    if (__pyx_t_9) {
    } else {
      __pyx_t_7 = __pyx_t_9;
      goto __pyx_L7_bool_binop_done;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":2630
 *                     and not other.parseAction
 *                     and other.resultsName is None
 *                     and not other.debug):             # <<<<<<<<<<<<<<
 *                 self.exprs = other.exprs[:] + [self.exprs[1]]
 *                 self.strRepr = None
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_debug); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2630, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 2630, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_8 = ((!__pyx_t_9) != 0);
    __pyx_t_7 = __pyx_t_8;
    __pyx_L7_bool_binop_done:;

    /* "Aspidites/_vendor/pyparsing_extension.py":2627
 *         if len(self.exprs) == 2:
 *             other = self.exprs[0]
 *             if (isinstance(other, self.__class__)             # <<<<<<<<<<<<<<
 *                     and not other.parseAction
 *                     and other.resultsName is None
 */
    if (__pyx_t_7) {

      /* "Aspidites/_vendor/pyparsing_extension.py":2631
 *                     and other.resultsName is None
 *                     and not other.debug):
 *                 self.exprs = other.exprs[:] + [self.exprs[1]]             # <<<<<<<<<<<<<<
 *                 self.strRepr = None
 *                 self.mayReturnEmpty |= other.mayReturnEmpty
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_exprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2631, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_t_1, 0, 0, NULL, NULL, &__pyx_slice__11, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2631, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2631, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_1, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2631, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2631, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_2);
      PyList_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_2 = PyNumber_Add(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2631, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_exprs, __pyx_t_2) < 0) __PYX_ERR(0, 2631, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":2632
 *                     and not other.debug):
 *                 self.exprs = other.exprs[:] + [self.exprs[1]]
 *                 self.strRepr = None             # <<<<<<<<<<<<<<
 *                 self.mayReturnEmpty |= other.mayReturnEmpty
 *                 self.mayIndexError |= other.mayIndexError
 */
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_strRepr, Py_None) < 0) __PYX_ERR(0, 2632, __pyx_L1_error)

      /* "Aspidites/_vendor/pyparsing_extension.py":2633
 *                 self.exprs = other.exprs[:] + [self.exprs[1]]
 *                 self.strRepr = None
 *                 self.mayReturnEmpty |= other.mayReturnEmpty             # <<<<<<<<<<<<<<
 *                 self.mayIndexError |= other.mayIndexError
 * 
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_mayReturnEmpty); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2633, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_mayReturnEmpty); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2633, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = PyNumber_InPlaceOr(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2633, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_mayReturnEmpty, __pyx_t_3) < 0) __PYX_ERR(0, 2633, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":2634
 *                 self.strRepr = None
 *                 self.mayReturnEmpty |= other.mayReturnEmpty
 *                 self.mayIndexError |= other.mayIndexError             # <<<<<<<<<<<<<<
 * 
 *             other = self.exprs[-1]
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_mayIndexError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2634, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_mayIndexError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2634, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = PyNumber_InPlaceOr(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2634, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_mayIndexError, __pyx_t_2) < 0) __PYX_ERR(0, 2634, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":2627
 *         if len(self.exprs) == 2:
 *             other = self.exprs[0]
 *             if (isinstance(other, self.__class__)             # <<<<<<<<<<<<<<
 *                     and not other.parseAction
 *                     and other.resultsName is None
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":2636
 *                 self.mayIndexError |= other.mayIndexError
 * 
 *             other = self.exprs[-1]             # <<<<<<<<<<<<<<
 *             if (isinstance(other, self.__class__)
 *                     and not other.parseAction
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2636, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_2, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2636, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_other, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2637
 * 
 *             other = self.exprs[-1]
 *             if (isinstance(other, self.__class__)             # <<<<<<<<<<<<<<
 *                     and not other.parseAction
 *                     and other.resultsName is None
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2637, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = PyObject_IsInstance(__pyx_v_other, __pyx_t_1); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 2637, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_9 = (__pyx_t_8 != 0);
    if (__pyx_t_9) {
    } else {
      __pyx_t_7 = __pyx_t_9;
      goto __pyx_L12_bool_binop_done;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":2638
 *             other = self.exprs[-1]
 *             if (isinstance(other, self.__class__)
 *                     and not other.parseAction             # <<<<<<<<<<<<<<
 *                     and other.resultsName is None
 *                     and not other.debug):
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_parseAction); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2638, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 2638, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_8 = ((!__pyx_t_9) != 0);
    if (__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L12_bool_binop_done;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":2639
 *             if (isinstance(other, self.__class__)
 *                     and not other.parseAction
 *                     and other.resultsName is None             # <<<<<<<<<<<<<<
 *                     and not other.debug):
 *                 self.exprs = self.exprs[:-1] + other.exprs[:]
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_resultsName); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2639, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = (__pyx_t_1 == Py_None);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_9 = (__pyx_t_8 != 0);
    if (__pyx_t_9) {
    } else {
      __pyx_t_7 = __pyx_t_9;
      goto __pyx_L12_bool_binop_done;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":2640
 *                     and not other.parseAction
 *                     and other.resultsName is None
 *                     and not other.debug):             # <<<<<<<<<<<<<<
 *                 self.exprs = self.exprs[:-1] + other.exprs[:]
 *                 self.strRepr = None
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_debug); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2640, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 2640, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_8 = ((!__pyx_t_9) != 0);
    __pyx_t_7 = __pyx_t_8;
    __pyx_L12_bool_binop_done:;

    /* "Aspidites/_vendor/pyparsing_extension.py":2637
 * 
 *             other = self.exprs[-1]
 *             if (isinstance(other, self.__class__)             # <<<<<<<<<<<<<<
 *                     and not other.parseAction
 *                     and other.resultsName is None
 */
    if (__pyx_t_7) {

      /* "Aspidites/_vendor/pyparsing_extension.py":2641
 *                     and other.resultsName is None
 *                     and not other.debug):
 *                 self.exprs = self.exprs[:-1] + other.exprs[:]             # <<<<<<<<<<<<<<
 *                 self.strRepr = None
 *                 self.mayReturnEmpty |= other.mayReturnEmpty
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2641, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = __Pyx_PyObject_GetSlice(__pyx_t_1, 0, -1L, NULL, NULL, &__pyx_slice__34, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2641, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_exprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2641, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_t_1, 0, 0, NULL, NULL, &__pyx_slice__11, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2641, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = PyNumber_Add(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2641, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_exprs, __pyx_t_1) < 0) __PYX_ERR(0, 2641, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":2642
 *                     and not other.debug):
 *                 self.exprs = self.exprs[:-1] + other.exprs[:]
 *                 self.strRepr = None             # <<<<<<<<<<<<<<
 *                 self.mayReturnEmpty |= other.mayReturnEmpty
 *                 self.mayIndexError |= other.mayIndexError
 */
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_strRepr, Py_None) < 0) __PYX_ERR(0, 2642, __pyx_L1_error)

      /* "Aspidites/_vendor/pyparsing_extension.py":2643
 *                 self.exprs = self.exprs[:-1] + other.exprs[:]
 *                 self.strRepr = None
 *                 self.mayReturnEmpty |= other.mayReturnEmpty             # <<<<<<<<<<<<<<
 *                 self.mayIndexError |= other.mayIndexError
 * 
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_mayReturnEmpty); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2643, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_mayReturnEmpty); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2643, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = PyNumber_InPlaceOr(__pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2643, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_mayReturnEmpty, __pyx_t_2) < 0) __PYX_ERR(0, 2643, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":2644
 *                 self.strRepr = None
 *                 self.mayReturnEmpty |= other.mayReturnEmpty
 *                 self.mayIndexError |= other.mayIndexError             # <<<<<<<<<<<<<<
 * 
 *         self.errmsg = "Expected " + str(self)
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_mayIndexError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2644, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_mayIndexError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2644, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = PyNumber_InPlaceOr(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2644, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_mayIndexError, __pyx_t_1) < 0) __PYX_ERR(0, 2644, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":2637
 * 
 *             other = self.exprs[-1]
 *             if (isinstance(other, self.__class__)             # <<<<<<<<<<<<<<
 *                     and not other.parseAction
 *                     and other.resultsName is None
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":2625
 *         # but only if there are no parse actions or resultsNames on the nested And's
 *         # (likewise for Or's and MatchFirst's)
 *         if len(self.exprs) == 2:             # <<<<<<<<<<<<<<
 *             other = self.exprs[0]
 *             if (isinstance(other, self.__class__)
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2646
 *                 self.mayIndexError |= other.mayIndexError
 * 
 *         self.errmsg = "Expected " + str(self)             # <<<<<<<<<<<<<<
 * 
 *         return self
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyNumber_Add(__pyx_kp_s_Expected, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_errmsg, __pyx_t_3) < 0) __PYX_ERR(0, 2646, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2648
 *         self.errmsg = "Expected " + str(self)
 * 
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def validate(self, validateTrace=None):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2616
 *         return self.strRepr
 * 
 *     def streamline(self):             # <<<<<<<<<<<<<<
 *         super(ParseExpression, self).streamline()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseExpression.streamline", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XDECREF(__pyx_v_other);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2650
 *         return self
 * 
 *     def validate(self, validateTrace=None):             # <<<<<<<<<<<<<<
 *         tmp = (validateTrace if validateTrace is not None else [])[:] + [self]
 *         for e in self.exprs:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_13validate(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_13validate = {"validate", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_13validate, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_13validate(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_validateTrace = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("validate (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_validateTrace,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_validateTrace);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "validate") < 0)) __PYX_ERR(0, 2650, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_validateTrace = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("validate", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2650, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseExpression.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_12validate(__pyx_self, __pyx_v_self, __pyx_v_validateTrace);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_12validate(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_validateTrace) {
  PyObject *__pyx_v_tmp = NULL;
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("validate", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2651
 * 
 *     def validate(self, validateTrace=None):
 *         tmp = (validateTrace if validateTrace is not None else [])[:] + [self]             # <<<<<<<<<<<<<<
 *         for e in self.exprs:
 *             e.validate(tmp)
 */
  __pyx_t_2 = (__pyx_v_validateTrace != Py_None);
  if ((__pyx_t_2 != 0)) {
    __Pyx_INCREF(__pyx_v_validateTrace);
    __pyx_t_1 = __pyx_v_validateTrace;
  } else {
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2651, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
  }
  __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_t_1, 0, 0, NULL, NULL, &__pyx_slice__11, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_v_self);
  __pyx_t_4 = PyNumber_Add(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_tmp = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2652
 *     def validate(self, validateTrace=None):
 *         tmp = (validateTrace if validateTrace is not None else [])[:] + [self]
 *         for e in self.exprs:             # <<<<<<<<<<<<<<
 *             e.validate(tmp)
 *         self.checkRecursion([])
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
    __pyx_t_1 = __pyx_t_4; __Pyx_INCREF(__pyx_t_1); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2652, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2652, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_5); __Pyx_INCREF(__pyx_t_4); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 2652, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2652, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_5); __Pyx_INCREF(__pyx_t_4); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 2652, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2652, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_6(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 2652, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2653
 *         tmp = (validateTrace if validateTrace is not None else [])[:] + [self]
 *         for e in self.exprs:
 *             e.validate(tmp)             # <<<<<<<<<<<<<<
 *         self.checkRecursion([])
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_validate); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2653, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_4 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_7, __pyx_v_tmp) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_tmp);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2653, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2652
 *     def validate(self, validateTrace=None):
 *         tmp = (validateTrace if validateTrace is not None else [])[:] + [self]
 *         for e in self.exprs:             # <<<<<<<<<<<<<<
 *             e.validate(tmp)
 *         self.checkRecursion([])
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2654
 *         for e in self.exprs:
 *             e.validate(tmp)
 *         self.checkRecursion([])             # <<<<<<<<<<<<<<
 * 
 *     def copy(self):
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_checkRecursion); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_7, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2650
 *         return self
 * 
 *     def validate(self, validateTrace=None):             # <<<<<<<<<<<<<<
 *         tmp = (validateTrace if validateTrace is not None else [])[:] + [self]
 *         for e in self.exprs:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseExpression.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tmp);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2656
 *         self.checkRecursion([])
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         ret = super(ParseExpression, self).copy()
 *         ret.exprs = [e.copy() for e in self.exprs]
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_15copy(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_15copy = {"copy", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_15copy, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_15copy(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_14copy(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_14copy(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_ret = NULL;
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2657
 * 
 *     def copy(self):
 *         ret = super(ParseExpression, self).copy()             # <<<<<<<<<<<<<<
 *         ret.exprs = [e.copy() for e in self.exprs]
 *         return ret
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParseExpression); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_copy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_ret = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2658
 *     def copy(self):
 *         ret = super(ParseExpression, self).copy()
 *         ret.exprs = [e.copy() for e in self.exprs]             # <<<<<<<<<<<<<<
 *         return ret
 * 
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_2 = __pyx_t_3; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2658, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2658, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 2658, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2658, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 2658, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2658, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_5(__pyx_t_2);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 2658, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_copy); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2658, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_3 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2658, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 2658, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_ret, __pyx_n_s_exprs, __pyx_t_1) < 0) __PYX_ERR(0, 2658, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2659
 *         ret = super(ParseExpression, self).copy()
 *         ret.exprs = [e.copy() for e in self.exprs]
 *         return ret             # <<<<<<<<<<<<<<
 * 
 *     def _setResultsName(self, name, listAllMatches=False):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_ret);
  __pyx_r = __pyx_v_ret;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2656
 *         self.checkRecursion([])
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         ret = super(ParseExpression, self).copy()
 *         ret.exprs = [e.copy() for e in self.exprs]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseExpression.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2661
 *         return ret
 * 
 *     def _setResultsName(self, name, listAllMatches=False):             # <<<<<<<<<<<<<<
 *         if __diag__.warn_ungrouped_named_tokens_in_collection:
 *             for e in self.exprs:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_17_setResultsName(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_17_setResultsName = {"_setResultsName", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_17_setResultsName, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_17_setResultsName(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_listAllMatches = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_setResultsName (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_name,&__pyx_n_s_listAllMatches,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_False));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_setResultsName", 0, 2, 3, 1); __PYX_ERR(0, 2661, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_listAllMatches);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_setResultsName") < 0)) __PYX_ERR(0, 2661, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_name = values[1];
    __pyx_v_listAllMatches = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_setResultsName", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2661, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseExpression._setResultsName", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_16_setResultsName(__pyx_self, __pyx_v_self, __pyx_v_name, __pyx_v_listAllMatches);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_16_setResultsName(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_listAllMatches) {
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_setResultsName", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2662
 * 
 *     def _setResultsName(self, name, listAllMatches=False):
 *         if __diag__.warn_ungrouped_named_tokens_in_collection:             # <<<<<<<<<<<<<<
 *             for e in self.exprs:
 *                 if isinstance(e, ParserElement) and e.resultsName:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_diag); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_warn_ungrouped_named_tokens_in_c); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 2662, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2663
 *     def _setResultsName(self, name, listAllMatches=False):
 *         if __diag__.warn_ungrouped_named_tokens_in_collection:
 *             for e in self.exprs:             # <<<<<<<<<<<<<<
 *                 if isinstance(e, ParserElement) and e.resultsName:
 *                     warnings.warn("{0}: setting results name {1!r} on {2} expression "
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2663, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1); __pyx_t_4 = 0;
      __pyx_t_5 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2663, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2663, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_5)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 2663, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2663, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 2663, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2663, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_5(__pyx_t_1);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 2663, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":2664
 *         if __diag__.warn_ungrouped_named_tokens_in_collection:
 *             for e in self.exprs:
 *                 if isinstance(e, ParserElement) and e.resultsName:             # <<<<<<<<<<<<<<
 *                     warnings.warn("{0}: setting results name {1!r} on {2} expression "
 *                                   "collides with {3!r} on contained expression".format(
 */
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2664, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = PyObject_IsInstance(__pyx_v_e, __pyx_t_2); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 2664, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_7 = (__pyx_t_6 != 0);
      if (__pyx_t_7) {
      } else {
        __pyx_t_3 = __pyx_t_7;
        goto __pyx_L7_bool_binop_done;
      }
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_resultsName); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2664, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 2664, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_3 = __pyx_t_7;
      __pyx_L7_bool_binop_done:;
      if (__pyx_t_3) {

        /* "Aspidites/_vendor/pyparsing_extension.py":2665
 *             for e in self.exprs:
 *                 if isinstance(e, ParserElement) and e.resultsName:
 *                     warnings.warn("{0}: setting results name {1!r} on {2} expression "             # <<<<<<<<<<<<<<
 *                                   "collides with {3!r} on contained expression".format(
 *                         "warn_ungrouped_named_tokens_in_collection",
 */
        __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_warnings); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2665, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_warn); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2665, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":2666
 *                 if isinstance(e, ParserElement) and e.resultsName:
 *                     warnings.warn("{0}: setting results name {1!r} on {2} expression "
 *                                   "collides with {3!r} on contained expression".format(             # <<<<<<<<<<<<<<
 *                         "warn_ungrouped_named_tokens_in_collection",
 *                         name,
 */
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_0_setting_results_name_1_r_on_2, __pyx_n_s_format); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2666, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);

        /* "Aspidites/_vendor/pyparsing_extension.py":2669
 *                         "warn_ungrouped_named_tokens_in_collection",
 *                         name,
 *                         type(self).__name__,             # <<<<<<<<<<<<<<
 *                         e.resultsName),
 *                                   stacklevel=3)
 */
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(__pyx_v_self)), __pyx_n_s_name_2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2669, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);

        /* "Aspidites/_vendor/pyparsing_extension.py":2670
 *                         name,
 *                         type(self).__name__,
 *                         e.resultsName),             # <<<<<<<<<<<<<<
 *                                   stacklevel=3)
 * 
 */
        __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_resultsName); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2670, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_12 = NULL;
        __pyx_t_13 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
          __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_9);
          if (likely(__pyx_t_12)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
            __Pyx_INCREF(__pyx_t_12);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_9, function);
            __pyx_t_13 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_9)) {
          PyObject *__pyx_temp[5] = {__pyx_t_12, __pyx_n_s_warn_ungrouped_named_tokens_in_c, __pyx_v_name, __pyx_t_10, __pyx_t_11};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_13, 4+__pyx_t_13); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2666, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_9)) {
          PyObject *__pyx_temp[5] = {__pyx_t_12, __pyx_n_s_warn_ungrouped_named_tokens_in_c, __pyx_v_name, __pyx_t_10, __pyx_t_11};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_13, 4+__pyx_t_13); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2666, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        } else
        #endif
        {
          __pyx_t_14 = PyTuple_New(4+__pyx_t_13); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2666, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_14);
          if (__pyx_t_12) {
            __Pyx_GIVEREF(__pyx_t_12); PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_12); __pyx_t_12 = NULL;
          }
          __Pyx_INCREF(__pyx_n_s_warn_ungrouped_named_tokens_in_c);
          __Pyx_GIVEREF(__pyx_n_s_warn_ungrouped_named_tokens_in_c);
          PyTuple_SET_ITEM(__pyx_t_14, 0+__pyx_t_13, __pyx_n_s_warn_ungrouped_named_tokens_in_c);
          __Pyx_INCREF(__pyx_v_name);
          __Pyx_GIVEREF(__pyx_v_name);
          PyTuple_SET_ITEM(__pyx_t_14, 1+__pyx_t_13, __pyx_v_name);
          __Pyx_GIVEREF(__pyx_t_10);
          PyTuple_SET_ITEM(__pyx_t_14, 2+__pyx_t_13, __pyx_t_10);
          __Pyx_GIVEREF(__pyx_t_11);
          PyTuple_SET_ITEM(__pyx_t_14, 3+__pyx_t_13, __pyx_t_11);
          __pyx_t_10 = 0;
          __pyx_t_11 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_14, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2666, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        }
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":2665
 *             for e in self.exprs:
 *                 if isinstance(e, ParserElement) and e.resultsName:
 *                     warnings.warn("{0}: setting results name {1!r} on {2} expression "             # <<<<<<<<<<<<<<
 *                                   "collides with {3!r} on contained expression".format(
 *                         "warn_ungrouped_named_tokens_in_collection",
 */
        __pyx_t_9 = PyTuple_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2665, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_GIVEREF(__pyx_t_2);
        PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":2671
 *                         type(self).__name__,
 *                         e.resultsName),
 *                                   stacklevel=3)             # <<<<<<<<<<<<<<
 * 
 *         return super(ParseExpression, self)._setResultsName(name, listAllMatches)
 */
        __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2671, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_stacklevel, __pyx_int_3) < 0) __PYX_ERR(0, 2671, __pyx_L1_error)

        /* "Aspidites/_vendor/pyparsing_extension.py":2665
 *             for e in self.exprs:
 *                 if isinstance(e, ParserElement) and e.resultsName:
 *                     warnings.warn("{0}: setting results name {1!r} on {2} expression "             # <<<<<<<<<<<<<<
 *                                   "collides with {3!r} on contained expression".format(
 *                         "warn_ungrouped_named_tokens_in_collection",
 */
        __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_9, __pyx_t_2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2665, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":2664
 *         if __diag__.warn_ungrouped_named_tokens_in_collection:
 *             for e in self.exprs:
 *                 if isinstance(e, ParserElement) and e.resultsName:             # <<<<<<<<<<<<<<
 *                     warnings.warn("{0}: setting results name {1!r} on {2} expression "
 *                                   "collides with {3!r} on contained expression".format(
 */
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":2663
 *     def _setResultsName(self, name, listAllMatches=False):
 *         if __diag__.warn_ungrouped_named_tokens_in_collection:
 *             for e in self.exprs:             # <<<<<<<<<<<<<<
 *                 if isinstance(e, ParserElement) and e.resultsName:
 *                     warnings.warn("{0}: setting results name {1!r} on {2} expression "
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2662
 * 
 *     def _setResultsName(self, name, listAllMatches=False):
 *         if __diag__.warn_ungrouped_named_tokens_in_collection:             # <<<<<<<<<<<<<<
 *             for e in self.exprs:
 *                 if isinstance(e, ParserElement) and e.resultsName:
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2673
 *                                   stacklevel=3)
 * 
 *         return super(ParseExpression, self)._setResultsName(name, listAllMatches)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_ParseExpression); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_14);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_14);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_self);
  __pyx_t_14 = 0;
  __pyx_t_14 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_2, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_setResultsName); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __pyx_t_14 = NULL;
  __pyx_t_13 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_14)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_14);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_13 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_14, __pyx_v_name, __pyx_v_listAllMatches};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_13, 2+__pyx_t_13); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2673, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_14, __pyx_v_name, __pyx_v_listAllMatches};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_13, 2+__pyx_t_13); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2673, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_9 = PyTuple_New(2+__pyx_t_13); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2673, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (__pyx_t_14) {
      __Pyx_GIVEREF(__pyx_t_14); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_14); __pyx_t_14 = NULL;
    }
    __Pyx_INCREF(__pyx_v_name);
    __Pyx_GIVEREF(__pyx_v_name);
    PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_13, __pyx_v_name);
    __Pyx_INCREF(__pyx_v_listAllMatches);
    __Pyx_GIVEREF(__pyx_v_listAllMatches);
    PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_13, __pyx_v_listAllMatches);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2673, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2661
 *         return ret
 * 
 *     def _setResultsName(self, name, listAllMatches=False):             # <<<<<<<<<<<<<<
 *         if __diag__.warn_ungrouped_named_tokens_in_collection:
 *             for e in self.exprs:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseExpression._setResultsName", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2681
 *     """
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         super(Token, self).__init__(savelist=False)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5Token_1__init__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_5Token_1__init__ = {"__init__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5Token_1__init__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5Token_1__init__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5Token___init__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5Token___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2682
 * 
 *     def __init__(self):
 *         super(Token, self).__init__(savelist=False)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Token); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_self);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_savelist, Py_False) < 0) __PYX_ERR(0, 2682, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2681
 *     """
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         super(Token, self).__init__(savelist=False)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Token.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2689
 *     """
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         super(Empty, self).__init__()
 *         self.name = "Empty"
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5Empty_1__init__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_5Empty_1__init__ = {"__init__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5Empty_1__init__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5Empty_1__init__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5Empty___init__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5Empty___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2690
 * 
 *     def __init__(self):
 *         super(Empty, self).__init__()             # <<<<<<<<<<<<<<
 *         self.name = "Empty"
 *         self.mayReturnEmpty = True
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Empty); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2691
 *     def __init__(self):
 *         super(Empty, self).__init__()
 *         self.name = "Empty"             # <<<<<<<<<<<<<<
 *         self.mayReturnEmpty = True
 *         self.mayIndexError = False
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_name, __pyx_n_s_Empty) < 0) __PYX_ERR(0, 2691, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2692
 *         super(Empty, self).__init__()
 *         self.name = "Empty"
 *         self.mayReturnEmpty = True             # <<<<<<<<<<<<<<
 *         self.mayIndexError = False
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_mayReturnEmpty, Py_True) < 0) __PYX_ERR(0, 2692, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2693
 *         self.name = "Empty"
 *         self.mayReturnEmpty = True
 *         self.mayIndexError = False             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_mayIndexError, Py_False) < 0) __PYX_ERR(0, 2693, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2689
 *     """
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         super(Empty, self).__init__()
 *         self.name = "Empty"
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Empty.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2700
 *     """
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         super(NoMatch, self).__init__()
 *         self.name = "NoMatch"
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7NoMatch_1__init__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7NoMatch_1__init__ = {"__init__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7NoMatch_1__init__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7NoMatch_1__init__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7NoMatch___init__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7NoMatch___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2701
 * 
 *     def __init__(self):
 *         super(NoMatch, self).__init__()             # <<<<<<<<<<<<<<
 *         self.name = "NoMatch"
 *         self.mayReturnEmpty = True
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_NoMatch); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2701, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2701, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2701, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2701, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2701, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2702
 *     def __init__(self):
 *         super(NoMatch, self).__init__()
 *         self.name = "NoMatch"             # <<<<<<<<<<<<<<
 *         self.mayReturnEmpty = True
 *         self.mayIndexError = False
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_name, __pyx_n_s_NoMatch) < 0) __PYX_ERR(0, 2702, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2703
 *         super(NoMatch, self).__init__()
 *         self.name = "NoMatch"
 *         self.mayReturnEmpty = True             # <<<<<<<<<<<<<<
 *         self.mayIndexError = False
 *         self.errmsg = "Unmatchable token"
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_mayReturnEmpty, Py_True) < 0) __PYX_ERR(0, 2703, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2704
 *         self.name = "NoMatch"
 *         self.mayReturnEmpty = True
 *         self.mayIndexError = False             # <<<<<<<<<<<<<<
 *         self.errmsg = "Unmatchable token"
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_mayIndexError, Py_False) < 0) __PYX_ERR(0, 2704, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2705
 *         self.mayReturnEmpty = True
 *         self.mayIndexError = False
 *         self.errmsg = "Unmatchable token"             # <<<<<<<<<<<<<<
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_errmsg, __pyx_kp_s_Unmatchable_token) < 0) __PYX_ERR(0, 2705, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2700
 *     """
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         super(NoMatch, self).__init__()
 *         self.name = "NoMatch"
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.NoMatch.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2707
 *         self.errmsg = "Unmatchable token"
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         raise ParseException(instring, loc, self.errmsg, self)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7NoMatch_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7NoMatch_3parseImpl = {"parseImpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7NoMatch_3parseImpl, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7NoMatch_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  CYTHON_UNUSED PyObject *__pyx_v_doActions = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parseImpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_doActions,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 1); __PYX_ERR(0, 2707, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 2); __PYX_ERR(0, 2707, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_doActions);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "parseImpl") < 0)) __PYX_ERR(0, 2707, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_doActions = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2707, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.NoMatch.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7NoMatch_2parseImpl(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7NoMatch_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parseImpl", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2708
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         raise ParseException(instring, loc, self.errmsg, self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errmsg); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_4, __pyx_v_instring, __pyx_v_loc, __pyx_t_3, __pyx_v_self};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 4+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2708, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_4, __pyx_v_instring, __pyx_v_loc, __pyx_t_3, __pyx_v_self};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 4+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2708, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(4+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2708, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_instring);
    __Pyx_GIVEREF(__pyx_v_instring);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_instring);
    __Pyx_INCREF(__pyx_v_loc);
    __Pyx_GIVEREF(__pyx_v_loc);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_loc);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_5, __pyx_t_3);
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_6, 3+__pyx_t_5, __pyx_v_self);
    __pyx_t_3 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2708, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 2708, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2707
 *         self.errmsg = "Unmatchable token"
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         raise ParseException(instring, loc, self.errmsg, self)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.NoMatch.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2726
 *     """
 * 
 *     def __init__(self, matchString):             # <<<<<<<<<<<<<<
 *         super(Literal, self).__init__()
 *         self.match = matchString
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Literal_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7Literal_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Literal_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Literal_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_matchString = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_matchString,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_matchString)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(0, 2726, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 2726, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_matchString = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2726, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Literal.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Literal___init__(__pyx_self, __pyx_v_self, __pyx_v_matchString);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Literal___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_matchString) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  int __pyx_t_13;
  int __pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2727
 * 
 *     def __init__(self, matchString):
 *         super(Literal, self).__init__()             # <<<<<<<<<<<<<<
 *         self.match = matchString
 *         self.matchLen = len(matchString)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Literal); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2727, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2727, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2727, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2727, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2727, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2728
 *     def __init__(self, matchString):
 *         super(Literal, self).__init__()
 *         self.match = matchString             # <<<<<<<<<<<<<<
 *         self.matchLen = len(matchString)
 *         try:
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_match, __pyx_v_matchString) < 0) __PYX_ERR(0, 2728, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2729
 *         super(Literal, self).__init__()
 *         self.match = matchString
 *         self.matchLen = len(matchString)             # <<<<<<<<<<<<<<
 *         try:
 *             self.firstMatchChar = matchString[0]
 */
  __pyx_t_4 = PyObject_Length(__pyx_v_matchString); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2729, __pyx_L1_error)
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2729, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_matchLen, __pyx_t_1) < 0) __PYX_ERR(0, 2729, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2730
 *         self.match = matchString
 *         self.matchLen = len(matchString)
 *         try:             # <<<<<<<<<<<<<<
 *             self.firstMatchChar = matchString[0]
 *         except IndexError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_7);
    /*try:*/ {

      /* "Aspidites/_vendor/pyparsing_extension.py":2731
 *         self.matchLen = len(matchString)
 *         try:
 *             self.firstMatchChar = matchString[0]             # <<<<<<<<<<<<<<
 *         except IndexError:
 *             warnings.warn("null string passed to Literal; use Empty() instead",
 */
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_matchString, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2731, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_firstMatchChar, __pyx_t_1) < 0) __PYX_ERR(0, 2731, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":2730
 *         self.match = matchString
 *         self.matchLen = len(matchString)
 *         try:             # <<<<<<<<<<<<<<
 *             self.firstMatchChar = matchString[0]
 *         except IndexError:
 */
    }
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2732
 *         try:
 *             self.firstMatchChar = matchString[0]
 *         except IndexError:             # <<<<<<<<<<<<<<
 *             warnings.warn("null string passed to Literal; use Empty() instead",
 *                           SyntaxWarning, stacklevel=2)
 */
    __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_IndexError);
    if (__pyx_t_8) {
      __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Literal.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_3, &__pyx_t_2) < 0) __PYX_ERR(0, 2732, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GOTREF(__pyx_t_2);

      /* "Aspidites/_vendor/pyparsing_extension.py":2733
 *             self.firstMatchChar = matchString[0]
 *         except IndexError:
 *             warnings.warn("null string passed to Literal; use Empty() instead",             # <<<<<<<<<<<<<<
 *                           SyntaxWarning, stacklevel=2)
 *             self.__class__ = Empty
 */
      __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_warnings); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2733, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_warn); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2733, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":2734
 *         except IndexError:
 *             warnings.warn("null string passed to Literal; use Empty() instead",
 *                           SyntaxWarning, stacklevel=2)             # <<<<<<<<<<<<<<
 *             self.__class__ = Empty
 *         self.name = '"%s"' % str(self.match)
 */
      __pyx_t_9 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2734, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_stacklevel, __pyx_int_2) < 0) __PYX_ERR(0, 2734, __pyx_L5_except_error)

      /* "Aspidites/_vendor/pyparsing_extension.py":2733
 *             self.firstMatchChar = matchString[0]
 *         except IndexError:
 *             warnings.warn("null string passed to Literal; use Empty() instead",             # <<<<<<<<<<<<<<
 *                           SyntaxWarning, stacklevel=2)
 *             self.__class__ = Empty
 */
      __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_tuple__81, __pyx_t_9); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2733, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":2735
 *             warnings.warn("null string passed to Literal; use Empty() instead",
 *                           SyntaxWarning, stacklevel=2)
 *             self.__class__ = Empty             # <<<<<<<<<<<<<<
 *         self.name = '"%s"' % str(self.match)
 *         self.errmsg = "Expected " + self.name
 */
      __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_Empty); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2735, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_class, __pyx_t_11) < 0) __PYX_ERR(0, 2735, __pyx_L5_except_error)
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "Aspidites/_vendor/pyparsing_extension.py":2730
 *         self.match = matchString
 *         self.matchLen = len(matchString)
 *         try:             # <<<<<<<<<<<<<<
 *             self.firstMatchChar = matchString[0]
 *         except IndexError:
 */
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
    __pyx_L8_try_end:;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2736
 *                           SyntaxWarning, stacklevel=2)
 *             self.__class__ = Empty
 *         self.name = '"%s"' % str(self.match)             # <<<<<<<<<<<<<<
 *         self.errmsg = "Expected " + self.name
 *         self.mayReturnEmpty = False
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyString_FormatSafe(__pyx_kp_s_s, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_name, __pyx_t_2) < 0) __PYX_ERR(0, 2736, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2737
 *             self.__class__ = Empty
 *         self.name = '"%s"' % str(self.match)
 *         self.errmsg = "Expected " + self.name             # <<<<<<<<<<<<<<
 *         self.mayReturnEmpty = False
 *         self.mayIndexError = False
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2737, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Add(__pyx_kp_s_Expected, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2737, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_errmsg, __pyx_t_3) < 0) __PYX_ERR(0, 2737, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2738
 *         self.name = '"%s"' % str(self.match)
 *         self.errmsg = "Expected " + self.name
 *         self.mayReturnEmpty = False             # <<<<<<<<<<<<<<
 *         self.mayIndexError = False
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_mayReturnEmpty, Py_False) < 0) __PYX_ERR(0, 2738, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2739
 *         self.errmsg = "Expected " + self.name
 *         self.mayReturnEmpty = False
 *         self.mayIndexError = False             # <<<<<<<<<<<<<<
 * 
 *         # Performance tuning: modify __class__ to select
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_mayIndexError, Py_False) < 0) __PYX_ERR(0, 2739, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2743
 *         # Performance tuning: modify __class__ to select
 *         # a parseImpl optimized for single-character check
 *         if self.matchLen == 1 and type(self) is Literal:             # <<<<<<<<<<<<<<
 *             self.__class__ = _SingleCharLiteral
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_matchLen); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2743, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_t_3, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2743, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 2743, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_13) {
  } else {
    __pyx_t_12 = __pyx_t_13;
    goto __pyx_L12_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Literal); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2743, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_13 = (((PyObject *)Py_TYPE(__pyx_v_self)) == __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_14 = (__pyx_t_13 != 0);
  __pyx_t_12 = __pyx_t_14;
  __pyx_L12_bool_binop_done:;
  if (__pyx_t_12) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2744
 *         # a parseImpl optimized for single-character check
 *         if self.matchLen == 1 and type(self) is Literal:
 *             self.__class__ = _SingleCharLiteral             # <<<<<<<<<<<<<<
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SingleCharLiteral); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2744, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_class, __pyx_t_2) < 0) __PYX_ERR(0, 2744, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2743
 *         # Performance tuning: modify __class__ to select
 *         # a parseImpl optimized for single-character check
 *         if self.matchLen == 1 and type(self) is Literal:             # <<<<<<<<<<<<<<
 *             self.__class__ = _SingleCharLiteral
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2726
 *     """
 * 
 *     def __init__(self, matchString):             # <<<<<<<<<<<<<<
 *         super(Literal, self).__init__()
 *         self.match = matchString
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Literal.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2746
 *             self.__class__ = _SingleCharLiteral
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if instring[loc] == self.firstMatchChar and instring.startswith(self.match, loc):
 *             return loc + self.matchLen, self.match
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Literal_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7Literal_3parseImpl = {"parseImpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Literal_3parseImpl, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Literal_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  CYTHON_UNUSED PyObject *__pyx_v_doActions = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parseImpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_doActions,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 1); __PYX_ERR(0, 2746, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 2); __PYX_ERR(0, 2746, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_doActions);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "parseImpl") < 0)) __PYX_ERR(0, 2746, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_doActions = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2746, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Literal.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Literal_2parseImpl(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Literal_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parseImpl", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2747
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if instring[loc] == self.firstMatchChar and instring.startswith(self.match, loc):             # <<<<<<<<<<<<<<
 *             return loc + self.matchLen, self.match
 *         raise ParseException(instring, loc, self.errmsg, self)
 */
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_instring, __pyx_v_loc); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2747, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_firstMatchChar); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2747, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_2, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2747, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 2747, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_5) {
  } else {
    __pyx_t_1 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_instring, __pyx_n_s_startswith); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2747, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2747, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = NULL;
  __pyx_t_7 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_7 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_2, __pyx_v_loc};
    __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2747, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_2, __pyx_v_loc};
    __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2747, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  {
    __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2747, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (__pyx_t_6) {
      __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_2);
    __Pyx_INCREF(__pyx_v_loc);
    __Pyx_GIVEREF(__pyx_v_loc);
    PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_v_loc);
    __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2747, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 2747, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2748
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if instring[loc] == self.firstMatchChar and instring.startswith(self.match, loc):
 *             return loc + self.matchLen, self.match             # <<<<<<<<<<<<<<
 *         raise ParseException(instring, loc, self.errmsg, self)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_matchLen); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2748, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PyNumber_Add(__pyx_v_loc, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2748, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2748, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2748, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_4);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_8;
    __pyx_t_8 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2747
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if instring[loc] == self.firstMatchChar and instring.startswith(self.match, loc):             # <<<<<<<<<<<<<<
 *             return loc + self.matchLen, self.match
 *         raise ParseException(instring, loc, self.errmsg, self)
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2749
 *         if instring[loc] == self.firstMatchChar and instring.startswith(self.match, loc):
 *             return loc + self.matchLen, self.match
 *         raise ParseException(instring, loc, self.errmsg, self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2749, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errmsg); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2749, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  __pyx_t_7 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_7 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_4)) {
    PyObject *__pyx_temp[5] = {__pyx_t_2, __pyx_v_instring, __pyx_v_loc, __pyx_t_3, __pyx_v_self};
    __pyx_t_8 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2749, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
    PyObject *__pyx_temp[5] = {__pyx_t_2, __pyx_v_instring, __pyx_v_loc, __pyx_t_3, __pyx_v_self};
    __pyx_t_8 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2749, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(4+__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2749, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_instring);
    __Pyx_GIVEREF(__pyx_v_instring);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_7, __pyx_v_instring);
    __Pyx_INCREF(__pyx_v_loc);
    __Pyx_GIVEREF(__pyx_v_loc);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_7, __pyx_v_loc);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_7, __pyx_t_3);
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_6, 3+__pyx_t_7, __pyx_v_self);
    __pyx_t_3 = 0;
    __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2749, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_Raise(__pyx_t_8, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __PYX_ERR(0, 2749, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2746
 *             self.__class__ = _SingleCharLiteral
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if instring[loc] == self.firstMatchChar and instring.startswith(self.match, loc):
 *             return loc + self.matchLen, self.match
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Literal.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2753
 * 
 * class _SingleCharLiteral(Literal):
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if instring[loc] == self.firstMatchChar:
 *             return loc + 1, self.match
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_18_SingleCharLiteral_1parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_18_SingleCharLiteral_1parseImpl = {"parseImpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_18_SingleCharLiteral_1parseImpl, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_18_SingleCharLiteral_1parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  CYTHON_UNUSED PyObject *__pyx_v_doActions = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parseImpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_doActions,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 1); __PYX_ERR(0, 2753, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 2); __PYX_ERR(0, 2753, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_doActions);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "parseImpl") < 0)) __PYX_ERR(0, 2753, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_doActions = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2753, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._SingleCharLiteral.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_18_SingleCharLiteral_parseImpl(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_18_SingleCharLiteral_parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parseImpl", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2754
 * class _SingleCharLiteral(Literal):
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if instring[loc] == self.firstMatchChar:             # <<<<<<<<<<<<<<
 *             return loc + 1, self.match
 *         raise ParseException(instring, loc, self.errmsg, self)
 */
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_instring, __pyx_v_loc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2754, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_firstMatchChar); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2754, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 2754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2755
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if instring[loc] == self.firstMatchChar:
 *             return loc + 1, self.match             # <<<<<<<<<<<<<<
 *         raise ParseException(instring, loc, self.errmsg, self)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyInt_AddObjC(__pyx_v_loc, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2755, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2755, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2755, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
    __pyx_t_3 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2754
 * class _SingleCharLiteral(Literal):
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if instring[loc] == self.firstMatchChar:             # <<<<<<<<<<<<<<
 *             return loc + 1, self.match
 *         raise ParseException(instring, loc, self.errmsg, self)
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2756
 *         if instring[loc] == self.firstMatchChar:
 *             return loc + 1, self.match
 *         raise ParseException(instring, loc, self.errmsg, self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2756, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errmsg); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2756, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_5, __pyx_v_instring, __pyx_v_loc, __pyx_t_3, __pyx_v_self};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 4+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2756, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_5, __pyx_v_instring, __pyx_v_loc, __pyx_t_3, __pyx_v_self};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 4+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2756, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(4+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2756, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_instring);
    __Pyx_GIVEREF(__pyx_v_instring);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_instring);
    __Pyx_INCREF(__pyx_v_loc);
    __Pyx_GIVEREF(__pyx_v_loc);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_loc);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_3);
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 3+__pyx_t_6, __pyx_v_self);
    __pyx_t_3 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2756, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 2756, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2753
 * 
 * class _SingleCharLiteral(Literal):
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if instring[loc] == self.firstMatchChar:
 *             return loc + 1, self.match
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._SingleCharLiteral.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2790
 *     DEFAULT_KEYWORD_CHARS = alphanums + "_$"
 * 
 *     def __init__(self, matchString, identChars=None, caseless=False):             # <<<<<<<<<<<<<<
 *         super(Keyword, self).__init__()
 *         if identChars is None:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Keyword_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7Keyword_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Keyword_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Keyword_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_matchString = 0;
  PyObject *__pyx_v_identChars = 0;
  PyObject *__pyx_v_caseless = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_matchString,&__pyx_n_s_identChars,&__pyx_n_s_caseless,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)Py_False));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_matchString)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 2790, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_identChars);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_caseless);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 2790, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_matchString = values[1];
    __pyx_v_identChars = values[2];
    __pyx_v_caseless = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2790, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Keyword.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Keyword___init__(__pyx_self, __pyx_v_self, __pyx_v_matchString, __pyx_v_identChars, __pyx_v_caseless);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Keyword___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_matchString, PyObject *__pyx_v_identChars, PyObject *__pyx_v_caseless) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);
  __Pyx_INCREF(__pyx_v_identChars);

  /* "Aspidites/_vendor/pyparsing_extension.py":2791
 * 
 *     def __init__(self, matchString, identChars=None, caseless=False):
 *         super(Keyword, self).__init__()             # <<<<<<<<<<<<<<
 *         if identChars is None:
 *             identChars = Keyword.DEFAULT_KEYWORD_CHARS
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Keyword); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2791, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2791, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2791, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2791, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2791, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2792
 *     def __init__(self, matchString, identChars=None, caseless=False):
 *         super(Keyword, self).__init__()
 *         if identChars is None:             # <<<<<<<<<<<<<<
 *             identChars = Keyword.DEFAULT_KEYWORD_CHARS
 *         self.match = matchString
 */
  __pyx_t_4 = (__pyx_v_identChars == Py_None);
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2793
 *         super(Keyword, self).__init__()
 *         if identChars is None:
 *             identChars = Keyword.DEFAULT_KEYWORD_CHARS             # <<<<<<<<<<<<<<
 *         self.match = matchString
 *         self.matchLen = len(matchString)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Keyword); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2793, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_DEFAULT_KEYWORD_CHARS); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2793, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_identChars, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2792
 *     def __init__(self, matchString, identChars=None, caseless=False):
 *         super(Keyword, self).__init__()
 *         if identChars is None:             # <<<<<<<<<<<<<<
 *             identChars = Keyword.DEFAULT_KEYWORD_CHARS
 *         self.match = matchString
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2794
 *         if identChars is None:
 *             identChars = Keyword.DEFAULT_KEYWORD_CHARS
 *         self.match = matchString             # <<<<<<<<<<<<<<
 *         self.matchLen = len(matchString)
 *         try:
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_match, __pyx_v_matchString) < 0) __PYX_ERR(0, 2794, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2795
 *             identChars = Keyword.DEFAULT_KEYWORD_CHARS
 *         self.match = matchString
 *         self.matchLen = len(matchString)             # <<<<<<<<<<<<<<
 *         try:
 *             self.firstMatchChar = matchString[0]
 */
  __pyx_t_6 = PyObject_Length(__pyx_v_matchString); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2795, __pyx_L1_error)
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_matchLen, __pyx_t_3) < 0) __PYX_ERR(0, 2795, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2796
 *         self.match = matchString
 *         self.matchLen = len(matchString)
 *         try:             # <<<<<<<<<<<<<<
 *             self.firstMatchChar = matchString[0]
 *         except IndexError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
    __Pyx_XGOTREF(__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_8);
    __Pyx_XGOTREF(__pyx_t_9);
    /*try:*/ {

      /* "Aspidites/_vendor/pyparsing_extension.py":2797
 *         self.matchLen = len(matchString)
 *         try:
 *             self.firstMatchChar = matchString[0]             # <<<<<<<<<<<<<<
 *         except IndexError:
 *             warnings.warn("null string passed to Keyword; use Empty() instead",
 */
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_matchString, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2797, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_firstMatchChar, __pyx_t_3) < 0) __PYX_ERR(0, 2797, __pyx_L4_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":2796
 *         self.match = matchString
 *         self.matchLen = len(matchString)
 *         try:             # <<<<<<<<<<<<<<
 *             self.firstMatchChar = matchString[0]
 *         except IndexError:
 */
    }
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    goto __pyx_L9_try_end;
    __pyx_L4_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2798
 *         try:
 *             self.firstMatchChar = matchString[0]
 *         except IndexError:             # <<<<<<<<<<<<<<
 *             warnings.warn("null string passed to Keyword; use Empty() instead",
 *                           SyntaxWarning, stacklevel=2)
 */
    __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_IndexError);
    if (__pyx_t_10) {
      __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Keyword.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_1, &__pyx_t_2) < 0) __PYX_ERR(0, 2798, __pyx_L6_except_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_t_2);

      /* "Aspidites/_vendor/pyparsing_extension.py":2799
 *             self.firstMatchChar = matchString[0]
 *         except IndexError:
 *             warnings.warn("null string passed to Keyword; use Empty() instead",             # <<<<<<<<<<<<<<
 *                           SyntaxWarning, stacklevel=2)
 *         self.name = '"%s"' % self.match
 */
      __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_warnings); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2799, __pyx_L6_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_warn); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 2799, __pyx_L6_except_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":2800
 *         except IndexError:
 *             warnings.warn("null string passed to Keyword; use Empty() instead",
 *                           SyntaxWarning, stacklevel=2)             # <<<<<<<<<<<<<<
 *         self.name = '"%s"' % self.match
 *         self.errmsg = "Expected " + self.name
 */
      __pyx_t_11 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2800, __pyx_L6_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      if (PyDict_SetItem(__pyx_t_11, __pyx_n_s_stacklevel, __pyx_int_2) < 0) __PYX_ERR(0, 2800, __pyx_L6_except_error)

      /* "Aspidites/_vendor/pyparsing_extension.py":2799
 *             self.firstMatchChar = matchString[0]
 *         except IndexError:
 *             warnings.warn("null string passed to Keyword; use Empty() instead",             # <<<<<<<<<<<<<<
 *                           SyntaxWarning, stacklevel=2)
 *         self.name = '"%s"' % self.match
 */
      __pyx_t_13 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_tuple__82, __pyx_t_11); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2799, __pyx_L6_except_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L5_exception_handled;
    }
    goto __pyx_L6_except_error;
    __pyx_L6_except_error:;

    /* "Aspidites/_vendor/pyparsing_extension.py":2796
 *         self.match = matchString
 *         self.matchLen = len(matchString)
 *         try:             # <<<<<<<<<<<<<<
 *             self.firstMatchChar = matchString[0]
 *         except IndexError:
 */
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
    goto __pyx_L1_error;
    __pyx_L5_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
    __pyx_L9_try_end:;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2801
 *             warnings.warn("null string passed to Keyword; use Empty() instead",
 *                           SyntaxWarning, stacklevel=2)
 *         self.name = '"%s"' % self.match             # <<<<<<<<<<<<<<
 *         self.errmsg = "Expected " + self.name
 *         self.mayReturnEmpty = False
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyString_FormatSafe(__pyx_kp_s_s, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_name, __pyx_t_1) < 0) __PYX_ERR(0, 2801, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2802
 *                           SyntaxWarning, stacklevel=2)
 *         self.name = '"%s"' % self.match
 *         self.errmsg = "Expected " + self.name             # <<<<<<<<<<<<<<
 *         self.mayReturnEmpty = False
 *         self.mayIndexError = False
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2802, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_Add(__pyx_kp_s_Expected, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2802, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_errmsg, __pyx_t_2) < 0) __PYX_ERR(0, 2802, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2803
 *         self.name = '"%s"' % self.match
 *         self.errmsg = "Expected " + self.name
 *         self.mayReturnEmpty = False             # <<<<<<<<<<<<<<
 *         self.mayIndexError = False
 *         self.caseless = caseless
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_mayReturnEmpty, Py_False) < 0) __PYX_ERR(0, 2803, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2804
 *         self.errmsg = "Expected " + self.name
 *         self.mayReturnEmpty = False
 *         self.mayIndexError = False             # <<<<<<<<<<<<<<
 *         self.caseless = caseless
 *         if caseless:
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_mayIndexError, Py_False) < 0) __PYX_ERR(0, 2804, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2805
 *         self.mayReturnEmpty = False
 *         self.mayIndexError = False
 *         self.caseless = caseless             # <<<<<<<<<<<<<<
 *         if caseless:
 *             self.caselessmatch = matchString.upper()
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_caseless, __pyx_v_caseless) < 0) __PYX_ERR(0, 2805, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2806
 *         self.mayIndexError = False
 *         self.caseless = caseless
 *         if caseless:             # <<<<<<<<<<<<<<
 *             self.caselessmatch = matchString.upper()
 *             identChars = identChars.upper()
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_caseless); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 2806, __pyx_L1_error)
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2807
 *         self.caseless = caseless
 *         if caseless:
 *             self.caselessmatch = matchString.upper()             # <<<<<<<<<<<<<<
 *             identChars = identChars.upper()
 *         self.identChars = set(identChars)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_matchString, __pyx_n_s_upper); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2807, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2807, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_caselessmatch, __pyx_t_2) < 0) __PYX_ERR(0, 2807, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2808
 *         if caseless:
 *             self.caselessmatch = matchString.upper()
 *             identChars = identChars.upper()             # <<<<<<<<<<<<<<
 *         self.identChars = set(identChars)
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_identChars, __pyx_n_s_upper); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2808, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2808, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_identChars, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2806
 *         self.mayIndexError = False
 *         self.caseless = caseless
 *         if caseless:             # <<<<<<<<<<<<<<
 *             self.caselessmatch = matchString.upper()
 *             identChars = identChars.upper()
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2809
 *             self.caselessmatch = matchString.upper()
 *             identChars = identChars.upper()
 *         self.identChars = set(identChars)             # <<<<<<<<<<<<<<
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 */
  __pyx_t_2 = PySet_New(__pyx_v_identChars); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2809, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_identChars, __pyx_t_2) < 0) __PYX_ERR(0, 2809, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2790
 *     DEFAULT_KEYWORD_CHARS = alphanums + "_$"
 * 
 *     def __init__(self, matchString, identChars=None, caseless=False):             # <<<<<<<<<<<<<<
 *         super(Keyword, self).__init__()
 *         if identChars is None:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Keyword.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_identChars);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2811
 *         self.identChars = set(identChars)
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if self.caseless:
 *             if ((instring[loc:loc + self.matchLen].upper() == self.caselessmatch)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Keyword_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7Keyword_3parseImpl = {"parseImpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Keyword_3parseImpl, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Keyword_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  CYTHON_UNUSED PyObject *__pyx_v_doActions = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parseImpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_doActions,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 1); __PYX_ERR(0, 2811, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 2); __PYX_ERR(0, 2811, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_doActions);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "parseImpl") < 0)) __PYX_ERR(0, 2811, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_doActions = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2811, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Keyword.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Keyword_2parseImpl(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Keyword_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parseImpl", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2812
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if self.caseless:             # <<<<<<<<<<<<<<
 *             if ((instring[loc:loc + self.matchLen].upper() == self.caselessmatch)
 *                     and (loc >= len(instring) - self.matchLen
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_caseless); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2812, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 2812, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2813
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if self.caseless:
 *             if ((instring[loc:loc + self.matchLen].upper() == self.caselessmatch)             # <<<<<<<<<<<<<<
 *                     and (loc >= len(instring) - self.matchLen
 *                          or instring[loc + self.matchLen].upper() not in self.identChars)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_matchLen); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2813, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyNumber_Add(__pyx_v_loc, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2813, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_v_instring, 0, 0, &__pyx_v_loc, &__pyx_t_4, NULL, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2813, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_upper); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2813, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2813, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_caselessmatch); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2813, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2813, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 2813, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_5) {
    } else {
      __pyx_t_2 = __pyx_t_5;
      goto __pyx_L5_bool_binop_done;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":2814
 *         if self.caseless:
 *             if ((instring[loc:loc + self.matchLen].upper() == self.caselessmatch)
 *                     and (loc >= len(instring) - self.matchLen             # <<<<<<<<<<<<<<
 *                          or instring[loc + self.matchLen].upper() not in self.identChars)
 *                     and (loc == 0
 */
    __pyx_t_6 = PyObject_Length(__pyx_v_instring); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2814, __pyx_L1_error)
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2814, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_matchLen); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2814, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyNumber_Subtract(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2814, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyObject_RichCompare(__pyx_v_loc, __pyx_t_1, Py_GE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2814, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 2814, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (!__pyx_t_5) {
    } else {
      goto __pyx_L7_next_and;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":2815
 *             if ((instring[loc:loc + self.matchLen].upper() == self.caselessmatch)
 *                     and (loc >= len(instring) - self.matchLen
 *                          or instring[loc + self.matchLen].upper() not in self.identChars)             # <<<<<<<<<<<<<<
 *                     and (loc == 0
 *                          or instring[loc - 1].upper() not in self.identChars)):
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_matchLen); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2815, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyNumber_Add(__pyx_v_loc, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2815, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_instring, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2815, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_upper); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2815, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_4 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2815, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_identChars); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2815, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = (__Pyx_PySequence_ContainsTF(__pyx_t_4, __pyx_t_3, Py_NE)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 2815, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_7 = (__pyx_t_5 != 0);
    if (__pyx_t_7) {
    } else {
      __pyx_t_2 = __pyx_t_7;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_L7_next_and:;

    /* "Aspidites/_vendor/pyparsing_extension.py":2816
 *                     and (loc >= len(instring) - self.matchLen
 *                          or instring[loc + self.matchLen].upper() not in self.identChars)
 *                     and (loc == 0             # <<<<<<<<<<<<<<
 *                          or instring[loc - 1].upper() not in self.identChars)):
 *                 return loc + self.matchLen, self.match
 */
    __pyx_t_3 = __Pyx_PyInt_EqObjC(__pyx_v_loc, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2816, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 2816, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (!__pyx_t_7) {
    } else {
      __pyx_t_2 = __pyx_t_7;
      goto __pyx_L5_bool_binop_done;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":2817
 *                          or instring[loc + self.matchLen].upper() not in self.identChars)
 *                     and (loc == 0
 *                          or instring[loc - 1].upper() not in self.identChars)):             # <<<<<<<<<<<<<<
 *                 return loc + self.matchLen, self.match
 * 
 */
    __pyx_t_4 = __Pyx_PyInt_SubtractObjC(__pyx_v_loc, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2817, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_instring, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2817, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_upper); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2817, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2817, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_identChars); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2817, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = (__Pyx_PySequence_ContainsTF(__pyx_t_3, __pyx_t_4, Py_NE)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 2817, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = (__pyx_t_7 != 0);
    __pyx_t_2 = __pyx_t_5;
    __pyx_L5_bool_binop_done:;

    /* "Aspidites/_vendor/pyparsing_extension.py":2813
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if self.caseless:
 *             if ((instring[loc:loc + self.matchLen].upper() == self.caselessmatch)             # <<<<<<<<<<<<<<
 *                     and (loc >= len(instring) - self.matchLen
 *                          or instring[loc + self.matchLen].upper() not in self.identChars)
 */
    if (__pyx_t_2) {

      /* "Aspidites/_vendor/pyparsing_extension.py":2818
 *                     and (loc == 0
 *                          or instring[loc - 1].upper() not in self.identChars)):
 *                 return loc + self.matchLen, self.match             # <<<<<<<<<<<<<<
 * 
 *         else:
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_matchLen); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2818, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyNumber_Add(__pyx_v_loc, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2818, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2818, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2818, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
      __pyx_t_3 = 0;
      __pyx_t_4 = 0;
      __pyx_r = __pyx_t_1;
      __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "Aspidites/_vendor/pyparsing_extension.py":2813
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if self.caseless:
 *             if ((instring[loc:loc + self.matchLen].upper() == self.caselessmatch)             # <<<<<<<<<<<<<<
 *                     and (loc >= len(instring) - self.matchLen
 *                          or instring[loc + self.matchLen].upper() not in self.identChars)
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":2812
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if self.caseless:             # <<<<<<<<<<<<<<
 *             if ((instring[loc:loc + self.matchLen].upper() == self.caselessmatch)
 *                     and (loc >= len(instring) - self.matchLen
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2821
 * 
 *         else:
 *             if instring[loc] == self.firstMatchChar:             # <<<<<<<<<<<<<<
 *                 if ((self.matchLen == 1 or instring.startswith(self.match, loc))
 *                         and (loc >= len(instring) - self.matchLen
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_instring, __pyx_v_loc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2821, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_firstMatchChar); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2821, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2821, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 2821, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_2) {

      /* "Aspidites/_vendor/pyparsing_extension.py":2822
 *         else:
 *             if instring[loc] == self.firstMatchChar:
 *                 if ((self.matchLen == 1 or instring.startswith(self.match, loc))             # <<<<<<<<<<<<<<
 *                         and (loc >= len(instring) - self.matchLen
 *                              or instring[loc + self.matchLen] not in self.identChars)
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_matchLen); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2822, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyInt_EqObjC(__pyx_t_3, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2822, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 2822, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (!__pyx_t_5) {
      } else {
        goto __pyx_L13_next_and;
      }
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_instring, __pyx_n_s_startswith); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2822, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2822, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_8 = NULL;
      __pyx_t_9 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_9 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_1, __pyx_v_loc};
        __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2822, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_1, __pyx_v_loc};
        __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2822, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else
      #endif
      {
        __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2822, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        if (__pyx_t_8) {
          __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_1);
        PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_t_1);
        __Pyx_INCREF(__pyx_v_loc);
        __Pyx_GIVEREF(__pyx_v_loc);
        PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_v_loc);
        __pyx_t_1 = 0;
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2822, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 2822, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_5) {
      } else {
        __pyx_t_2 = __pyx_t_5;
        goto __pyx_L12_bool_binop_done;
      }
      __pyx_L13_next_and:;

      /* "Aspidites/_vendor/pyparsing_extension.py":2823
 *             if instring[loc] == self.firstMatchChar:
 *                 if ((self.matchLen == 1 or instring.startswith(self.match, loc))
 *                         and (loc >= len(instring) - self.matchLen             # <<<<<<<<<<<<<<
 *                              or instring[loc + self.matchLen] not in self.identChars)
 *                         and (loc == 0 or instring[loc - 1] not in self.identChars)):
 */
      __pyx_t_6 = PyObject_Length(__pyx_v_instring); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2823, __pyx_L1_error)
      __pyx_t_4 = PyInt_FromSsize_t(__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2823, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_matchLen); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2823, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_10 = PyNumber_Subtract(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2823, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_loc, __pyx_t_10, Py_GE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2823, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 2823, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (!__pyx_t_5) {
      } else {
        goto __pyx_L15_next_and;
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":2824
 *                 if ((self.matchLen == 1 or instring.startswith(self.match, loc))
 *                         and (loc >= len(instring) - self.matchLen
 *                              or instring[loc + self.matchLen] not in self.identChars)             # <<<<<<<<<<<<<<
 *                         and (loc == 0 or instring[loc - 1] not in self.identChars)):
 *                     return loc + self.matchLen, self.match
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_matchLen); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2824, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_10 = PyNumber_Add(__pyx_v_loc, __pyx_t_3); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2824, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_instring, __pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2824, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_identChars); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2824, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_5 = (__Pyx_PySequence_ContainsTF(__pyx_t_3, __pyx_t_10, Py_NE)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 2824, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_7 = (__pyx_t_5 != 0);
      if (__pyx_t_7) {
      } else {
        __pyx_t_2 = __pyx_t_7;
        goto __pyx_L12_bool_binop_done;
      }
      __pyx_L15_next_and:;

      /* "Aspidites/_vendor/pyparsing_extension.py":2825
 *                         and (loc >= len(instring) - self.matchLen
 *                              or instring[loc + self.matchLen] not in self.identChars)
 *                         and (loc == 0 or instring[loc - 1] not in self.identChars)):             # <<<<<<<<<<<<<<
 *                     return loc + self.matchLen, self.match
 * 
 */
      __pyx_t_10 = __Pyx_PyInt_EqObjC(__pyx_v_loc, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2825, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 2825, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (!__pyx_t_7) {
      } else {
        __pyx_t_2 = __pyx_t_7;
        goto __pyx_L12_bool_binop_done;
      }
      __pyx_t_10 = __Pyx_PyInt_SubtractObjC(__pyx_v_loc, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2825, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_instring, __pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2825, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_identChars); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2825, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_7 = (__Pyx_PySequence_ContainsTF(__pyx_t_3, __pyx_t_10, Py_NE)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 2825, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_5 = (__pyx_t_7 != 0);
      __pyx_t_2 = __pyx_t_5;
      __pyx_L12_bool_binop_done:;

      /* "Aspidites/_vendor/pyparsing_extension.py":2822
 *         else:
 *             if instring[loc] == self.firstMatchChar:
 *                 if ((self.matchLen == 1 or instring.startswith(self.match, loc))             # <<<<<<<<<<<<<<
 *                         and (loc >= len(instring) - self.matchLen
 *                              or instring[loc + self.matchLen] not in self.identChars)
 */
      if (__pyx_t_2) {

        /* "Aspidites/_vendor/pyparsing_extension.py":2826
 *                              or instring[loc + self.matchLen] not in self.identChars)
 *                         and (loc == 0 or instring[loc - 1] not in self.identChars)):
 *                     return loc + self.matchLen, self.match             # <<<<<<<<<<<<<<
 * 
 *         raise ParseException(instring, loc, self.errmsg, self)
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_matchLen); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2826, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_3 = PyNumber_Add(__pyx_v_loc, __pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2826, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2826, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2826, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_10);
        PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_10);
        __pyx_t_3 = 0;
        __pyx_t_10 = 0;
        __pyx_r = __pyx_t_4;
        __pyx_t_4 = 0;
        goto __pyx_L0;

        /* "Aspidites/_vendor/pyparsing_extension.py":2822
 *         else:
 *             if instring[loc] == self.firstMatchChar:
 *                 if ((self.matchLen == 1 or instring.startswith(self.match, loc))             # <<<<<<<<<<<<<<
 *                         and (loc >= len(instring) - self.matchLen
 *                              or instring[loc + self.matchLen] not in self.identChars)
 */
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":2821
 * 
 *         else:
 *             if instring[loc] == self.firstMatchChar:             # <<<<<<<<<<<<<<
 *                 if ((self.matchLen == 1 or instring.startswith(self.match, loc))
 *                         and (loc >= len(instring) - self.matchLen
 */
    }
  }
  __pyx_L3:;

  /* "Aspidites/_vendor/pyparsing_extension.py":2828
 *                     return loc + self.matchLen, self.match
 * 
 *         raise ParseException(instring, loc, self.errmsg, self)             # <<<<<<<<<<<<<<
 * 
 *     def copy(self):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errmsg); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = NULL;
  __pyx_t_9 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_10);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_10, function);
      __pyx_t_9 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_10)) {
    PyObject *__pyx_temp[5] = {__pyx_t_1, __pyx_v_instring, __pyx_v_loc, __pyx_t_3, __pyx_v_self};
    __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_9, 4+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2828, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
    PyObject *__pyx_temp[5] = {__pyx_t_1, __pyx_v_instring, __pyx_v_loc, __pyx_t_3, __pyx_v_self};
    __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_9, 4+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2828, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  {
    __pyx_t_8 = PyTuple_New(4+__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2828, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (__pyx_t_1) {
      __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_1); __pyx_t_1 = NULL;
    }
    __Pyx_INCREF(__pyx_v_instring);
    __Pyx_GIVEREF(__pyx_v_instring);
    PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_9, __pyx_v_instring);
    __Pyx_INCREF(__pyx_v_loc);
    __Pyx_GIVEREF(__pyx_v_loc);
    PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_9, __pyx_v_loc);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_9, __pyx_t_3);
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_8, 3+__pyx_t_9, __pyx_v_self);
    __pyx_t_3 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2828, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  }
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_Raise(__pyx_t_4, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __PYX_ERR(0, 2828, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2811
 *         self.identChars = set(identChars)
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if self.caseless:
 *             if ((instring[loc:loc + self.matchLen].upper() == self.caselessmatch)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Keyword.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2830
 *         raise ParseException(instring, loc, self.errmsg, self)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         c = super(Keyword, self).copy()
 *         c.identChars = Keyword.DEFAULT_KEYWORD_CHARS
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Keyword_5copy(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7Keyword_5copy = {"copy", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Keyword_5copy, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Keyword_5copy(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Keyword_4copy(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Keyword_4copy(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_c = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2831
 * 
 *     def copy(self):
 *         c = super(Keyword, self).copy()             # <<<<<<<<<<<<<<
 *         c.identChars = Keyword.DEFAULT_KEYWORD_CHARS
 *         return c
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Keyword); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_copy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_c = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2832
 *     def copy(self):
 *         c = super(Keyword, self).copy()
 *         c.identChars = Keyword.DEFAULT_KEYWORD_CHARS             # <<<<<<<<<<<<<<
 *         return c
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Keyword); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_DEFAULT_KEYWORD_CHARS); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_c, __pyx_n_s_identChars, __pyx_t_3) < 0) __PYX_ERR(0, 2832, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2833
 *         c = super(Keyword, self).copy()
 *         c.identChars = Keyword.DEFAULT_KEYWORD_CHARS
 *         return c             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_c);
  __pyx_r = __pyx_v_c;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2830
 *         raise ParseException(instring, loc, self.errmsg, self)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         c = super(Keyword, self).copy()
 *         c.identChars = Keyword.DEFAULT_KEYWORD_CHARS
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Keyword.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_c);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2836
 * 
 *     @staticmethod
 *     def setDefaultKeywordChars(chars):             # <<<<<<<<<<<<<<
 *         """Overrides the default Keyword chars
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Keyword_7setDefaultKeywordChars(PyObject *__pyx_self, PyObject *__pyx_v_chars); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_7Keyword_6setDefaultKeywordChars[] = "Overrides the default Keyword chars\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7Keyword_7setDefaultKeywordChars = {"setDefaultKeywordChars", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Keyword_7setDefaultKeywordChars, METH_O, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_7Keyword_6setDefaultKeywordChars};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Keyword_7setDefaultKeywordChars(PyObject *__pyx_self, PyObject *__pyx_v_chars) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setDefaultKeywordChars (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Keyword_6setDefaultKeywordChars(__pyx_self, ((PyObject *)__pyx_v_chars));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Keyword_6setDefaultKeywordChars(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_chars) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setDefaultKeywordChars", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2839
 *         """Overrides the default Keyword chars
 *         """
 *         Keyword.DEFAULT_KEYWORD_CHARS = chars             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Keyword); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2839, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_1, __pyx_n_s_DEFAULT_KEYWORD_CHARS, __pyx_v_chars) < 0) __PYX_ERR(0, 2839, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2836
 * 
 *     @staticmethod
 *     def setDefaultKeywordChars(chars):             # <<<<<<<<<<<<<<
 *         """Overrides the default Keyword chars
 *         """
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Keyword.setDefaultKeywordChars", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2854
 *     """
 * 
 *     def __init__(self, matchString):             # <<<<<<<<<<<<<<
 *         super(CaselessLiteral, self).__init__(matchString.upper())
 *         # Preserve the defining literal.
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15CaselessLiteral_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_15CaselessLiteral_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15CaselessLiteral_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15CaselessLiteral_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_matchString = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_matchString,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_matchString)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(0, 2854, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 2854, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_matchString = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2854, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.CaselessLiteral.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15CaselessLiteral___init__(__pyx_self, __pyx_v_self, __pyx_v_matchString);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15CaselessLiteral___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_matchString) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2855
 * 
 *     def __init__(self, matchString):
 *         super(CaselessLiteral, self).__init__(matchString.upper())             # <<<<<<<<<<<<<<
 *         # Preserve the defining literal.
 *         self.returnString = matchString
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_CaselessLiteral); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2855, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2855, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2855, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2855, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_matchString, __pyx_n_s_upper); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2855, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2855, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2855, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2857
 *         super(CaselessLiteral, self).__init__(matchString.upper())
 *         # Preserve the defining literal.
 *         self.returnString = matchString             # <<<<<<<<<<<<<<
 *         self.name = "'%s'" % self.returnString
 *         self.errmsg = "Expected " + self.name
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_returnString, __pyx_v_matchString) < 0) __PYX_ERR(0, 2857, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2858
 *         # Preserve the defining literal.
 *         self.returnString = matchString
 *         self.name = "'%s'" % self.returnString             # <<<<<<<<<<<<<<
 *         self.errmsg = "Expected " + self.name
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_returnString); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyString_FormatSafe(__pyx_kp_s_s_2, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_name, __pyx_t_3) < 0) __PYX_ERR(0, 2858, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2859
 *         self.returnString = matchString
 *         self.name = "'%s'" % self.returnString
 *         self.errmsg = "Expected " + self.name             # <<<<<<<<<<<<<<
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyNumber_Add(__pyx_kp_s_Expected, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_errmsg, __pyx_t_1) < 0) __PYX_ERR(0, 2859, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2854
 *     """
 * 
 *     def __init__(self, matchString):             # <<<<<<<<<<<<<<
 *         super(CaselessLiteral, self).__init__(matchString.upper())
 *         # Preserve the defining literal.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.CaselessLiteral.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2861
 *         self.errmsg = "Expected " + self.name
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if instring[loc:loc + self.matchLen].upper() == self.match:
 *             return loc + self.matchLen, self.returnString
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15CaselessLiteral_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_15CaselessLiteral_3parseImpl = {"parseImpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15CaselessLiteral_3parseImpl, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15CaselessLiteral_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  CYTHON_UNUSED PyObject *__pyx_v_doActions = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parseImpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_doActions,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 1); __PYX_ERR(0, 2861, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 2); __PYX_ERR(0, 2861, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_doActions);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "parseImpl") < 0)) __PYX_ERR(0, 2861, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_doActions = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2861, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.CaselessLiteral.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15CaselessLiteral_2parseImpl(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15CaselessLiteral_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parseImpl", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2862
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if instring[loc:loc + self.matchLen].upper() == self.match:             # <<<<<<<<<<<<<<
 *             return loc + self.matchLen, self.returnString
 *         raise ParseException(instring, loc, self.errmsg, self)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_matchLen); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2862, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Add(__pyx_v_loc, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2862, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetSlice(__pyx_v_instring, 0, 0, &__pyx_v_loc, &__pyx_t_3, NULL, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2862, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_upper); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2862, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2862, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2862, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2862, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 2862, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2863
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if instring[loc:loc + self.matchLen].upper() == self.match:
 *             return loc + self.matchLen, self.returnString             # <<<<<<<<<<<<<<
 *         raise ParseException(instring, loc, self.errmsg, self)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_matchLen); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2863, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyNumber_Add(__pyx_v_loc, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2863, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_returnString); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2863, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2863, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
    __pyx_t_3 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2862
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if instring[loc:loc + self.matchLen].upper() == self.match:             # <<<<<<<<<<<<<<
 *             return loc + self.matchLen, self.returnString
 *         raise ParseException(instring, loc, self.errmsg, self)
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2864
 *         if instring[loc:loc + self.matchLen].upper() == self.match:
 *             return loc + self.matchLen, self.returnString
 *         raise ParseException(instring, loc, self.errmsg, self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errmsg); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_5, __pyx_v_instring, __pyx_v_loc, __pyx_t_3, __pyx_v_self};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 4+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2864, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_5, __pyx_v_instring, __pyx_v_loc, __pyx_t_3, __pyx_v_self};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 4+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2864, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(4+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2864, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_instring);
    __Pyx_GIVEREF(__pyx_v_instring);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_instring);
    __Pyx_INCREF(__pyx_v_loc);
    __Pyx_GIVEREF(__pyx_v_loc);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_loc);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_3);
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 3+__pyx_t_6, __pyx_v_self);
    __pyx_t_3 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2864, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 2864, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2861
 *         self.errmsg = "Expected " + self.name
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if instring[loc:loc + self.matchLen].upper() == self.match:
 *             return loc + self.matchLen, self.returnString
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.CaselessLiteral.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2878
 *     """
 * 
 *     def __init__(self, matchString, identChars=None):             # <<<<<<<<<<<<<<
 *         super(CaselessKeyword, self).__init__(matchString, identChars, caseless=True)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15CaselessKeyword_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_15CaselessKeyword_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15CaselessKeyword_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_15CaselessKeyword_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_matchString = 0;
  PyObject *__pyx_v_identChars = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_matchString,&__pyx_n_s_identChars,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_matchString)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, 1); __PYX_ERR(0, 2878, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_identChars);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 2878, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_matchString = values[1];
    __pyx_v_identChars = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2878, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.CaselessKeyword.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15CaselessKeyword___init__(__pyx_self, __pyx_v_self, __pyx_v_matchString, __pyx_v_identChars);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_15CaselessKeyword___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_matchString, PyObject *__pyx_v_identChars) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2879
 * 
 *     def __init__(self, matchString, identChars=None):
 *         super(CaselessKeyword, self).__init__(matchString, identChars, caseless=True)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_CaselessKeyword); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_self);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_matchString);
  __Pyx_GIVEREF(__pyx_v_matchString);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_matchString);
  __Pyx_INCREF(__pyx_v_identChars);
  __Pyx_GIVEREF(__pyx_v_identChars);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_identChars);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_caseless, Py_True) < 0) __PYX_ERR(0, 2879, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2878
 *     """
 * 
 *     def __init__(self, matchString, identChars=None):             # <<<<<<<<<<<<<<
 *         super(CaselessKeyword, self).__init__(matchString, identChars, caseless=True)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.CaselessKeyword.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2916
 *     """
 * 
 *     def __init__(self, match_string, maxMismatches=1):             # <<<<<<<<<<<<<<
 *         super(CloseMatch, self).__init__()
 *         self.name = match_string
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10CloseMatch_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10CloseMatch_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10CloseMatch_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10CloseMatch_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_match_string = 0;
  PyObject *__pyx_v_maxMismatches = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_match_string,&__pyx_n_s_maxMismatches,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject *)__pyx_int_1));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_match_string)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, 1); __PYX_ERR(0, 2916, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_maxMismatches);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 2916, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_match_string = values[1];
    __pyx_v_maxMismatches = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2916, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.CloseMatch.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10CloseMatch___init__(__pyx_self, __pyx_v_self, __pyx_v_match_string, __pyx_v_maxMismatches);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10CloseMatch___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_match_string, PyObject *__pyx_v_maxMismatches) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":2917
 * 
 *     def __init__(self, match_string, maxMismatches=1):
 *         super(CloseMatch, self).__init__()             # <<<<<<<<<<<<<<
 *         self.name = match_string
 *         self.match_string = match_string
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_CloseMatch); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2918
 *     def __init__(self, match_string, maxMismatches=1):
 *         super(CloseMatch, self).__init__()
 *         self.name = match_string             # <<<<<<<<<<<<<<
 *         self.match_string = match_string
 *         self.maxMismatches = maxMismatches
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_name, __pyx_v_match_string) < 0) __PYX_ERR(0, 2918, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2919
 *         super(CloseMatch, self).__init__()
 *         self.name = match_string
 *         self.match_string = match_string             # <<<<<<<<<<<<<<
 *         self.maxMismatches = maxMismatches
 *         self.errmsg = "Expected %r (with up to %d mismatches)" % (self.match_string, self.maxMismatches)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_match_string, __pyx_v_match_string) < 0) __PYX_ERR(0, 2919, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2920
 *         self.name = match_string
 *         self.match_string = match_string
 *         self.maxMismatches = maxMismatches             # <<<<<<<<<<<<<<
 *         self.errmsg = "Expected %r (with up to %d mismatches)" % (self.match_string, self.maxMismatches)
 *         self.mayIndexError = False
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_maxMismatches, __pyx_v_maxMismatches) < 0) __PYX_ERR(0, 2920, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2921
 *         self.match_string = match_string
 *         self.maxMismatches = maxMismatches
 *         self.errmsg = "Expected %r (with up to %d mismatches)" % (self.match_string, self.maxMismatches)             # <<<<<<<<<<<<<<
 *         self.mayIndexError = False
 *         self.mayReturnEmpty = False
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match_string); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2921, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_maxMismatches); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2921, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2921, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_Expected_r_with_up_to_d_mismatch, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2921, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_errmsg, __pyx_t_3) < 0) __PYX_ERR(0, 2921, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2922
 *         self.maxMismatches = maxMismatches
 *         self.errmsg = "Expected %r (with up to %d mismatches)" % (self.match_string, self.maxMismatches)
 *         self.mayIndexError = False             # <<<<<<<<<<<<<<
 *         self.mayReturnEmpty = False
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_mayIndexError, Py_False) < 0) __PYX_ERR(0, 2922, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2923
 *         self.errmsg = "Expected %r (with up to %d mismatches)" % (self.match_string, self.maxMismatches)
 *         self.mayIndexError = False
 *         self.mayReturnEmpty = False             # <<<<<<<<<<<<<<
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_mayReturnEmpty, Py_False) < 0) __PYX_ERR(0, 2923, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2916
 *     """
 * 
 *     def __init__(self, match_string, maxMismatches=1):             # <<<<<<<<<<<<<<
 *         super(CloseMatch, self).__init__()
 *         self.name = match_string
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.CloseMatch.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":2925
 *         self.mayReturnEmpty = False
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         start = loc
 *         instrlen = len(instring)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10CloseMatch_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10CloseMatch_3parseImpl = {"parseImpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10CloseMatch_3parseImpl, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10CloseMatch_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  CYTHON_UNUSED PyObject *__pyx_v_doActions = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parseImpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_doActions,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 1); __PYX_ERR(0, 2925, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 2); __PYX_ERR(0, 2925, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_doActions);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "parseImpl") < 0)) __PYX_ERR(0, 2925, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_doActions = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2925, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.CloseMatch.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10CloseMatch_2parseImpl(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10CloseMatch_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions) {
  PyObject *__pyx_v_start = NULL;
  Py_ssize_t __pyx_v_instrlen;
  PyObject *__pyx_v_maxloc = NULL;
  PyObject *__pyx_v_match_string = NULL;
  PyObject *__pyx_v_match_stringloc = NULL;
  PyObject *__pyx_v_mismatches = NULL;
  PyObject *__pyx_v_maxMismatches = NULL;
  PyObject *__pyx_v_s_m = NULL;
  PyObject *__pyx_v_src = NULL;
  PyObject *__pyx_v_mat = NULL;
  PyObject *__pyx_v_results = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *(*__pyx_t_6)(PyObject *);
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *(*__pyx_t_9)(PyObject *);
  int __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parseImpl", 0);
  __Pyx_INCREF(__pyx_v_loc);

  /* "Aspidites/_vendor/pyparsing_extension.py":2926
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         start = loc             # <<<<<<<<<<<<<<
 *         instrlen = len(instring)
 *         maxloc = start + len(self.match_string)
 */
  __Pyx_INCREF(__pyx_v_loc);
  __pyx_v_start = __pyx_v_loc;

  /* "Aspidites/_vendor/pyparsing_extension.py":2927
 *     def parseImpl(self, instring, loc, doActions=True):
 *         start = loc
 *         instrlen = len(instring)             # <<<<<<<<<<<<<<
 *         maxloc = start + len(self.match_string)
 * 
 */
  __pyx_t_1 = PyObject_Length(__pyx_v_instring); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2927, __pyx_L1_error)
  __pyx_v_instrlen = __pyx_t_1;

  /* "Aspidites/_vendor/pyparsing_extension.py":2928
 *         start = loc
 *         instrlen = len(instring)
 *         maxloc = start + len(self.match_string)             # <<<<<<<<<<<<<<
 * 
 *         if maxloc <= instrlen:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match_string); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2928, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyObject_Length(__pyx_t_2); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2928, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2928, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Add(__pyx_v_start, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2928, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_maxloc = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2930
 *         maxloc = start + len(self.match_string)
 * 
 *         if maxloc <= instrlen:             # <<<<<<<<<<<<<<
 *             match_string = self.match_string
 *             match_stringloc = 0
 */
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_instrlen); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_maxloc, __pyx_t_3, Py_LE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2930, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 2930, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":2931
 * 
 *         if maxloc <= instrlen:
 *             match_string = self.match_string             # <<<<<<<<<<<<<<
 *             match_stringloc = 0
 *             mismatches = []
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_match_string); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2931, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_match_string = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2932
 *         if maxloc <= instrlen:
 *             match_string = self.match_string
 *             match_stringloc = 0             # <<<<<<<<<<<<<<
 *             mismatches = []
 *             maxMismatches = self.maxMismatches
 */
    __Pyx_INCREF(__pyx_int_0);
    __pyx_v_match_stringloc = __pyx_int_0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2933
 *             match_string = self.match_string
 *             match_stringloc = 0
 *             mismatches = []             # <<<<<<<<<<<<<<
 *             maxMismatches = self.maxMismatches
 * 
 */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2933, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_mismatches = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2934
 *             match_stringloc = 0
 *             mismatches = []
 *             maxMismatches = self.maxMismatches             # <<<<<<<<<<<<<<
 * 
 *             for match_stringloc, s_m in enumerate(zip(instring[loc:maxloc], match_string)):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_maxMismatches); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2934, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_maxMismatches = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2936
 *             maxMismatches = self.maxMismatches
 * 
 *             for match_stringloc, s_m in enumerate(zip(instring[loc:maxloc], match_string)):             # <<<<<<<<<<<<<<
 *                 src, mat = s_m
 *                 if src != mat:
 */
    __Pyx_INCREF(__pyx_int_0);
    __pyx_t_2 = __pyx_int_0;
    __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_v_instring, 0, 0, &__pyx_v_loc, &__pyx_v_maxloc, NULL, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2936, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2936, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3);
    __Pyx_INCREF(__pyx_v_match_string);
    __Pyx_GIVEREF(__pyx_v_match_string);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_match_string);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2936, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
      __pyx_t_5 = __pyx_t_3; __Pyx_INCREF(__pyx_t_5); __pyx_t_1 = 0;
      __pyx_t_6 = NULL;
    } else {
      __pyx_t_1 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2936, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2936, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    for (;;) {
      if (likely(!__pyx_t_6)) {
        if (likely(PyList_CheckExact(__pyx_t_5))) {
          if (__pyx_t_1 >= PyList_GET_SIZE(__pyx_t_5)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_1); __Pyx_INCREF(__pyx_t_3); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(0, 2936, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_5, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2936, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        } else {
          if (__pyx_t_1 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_1); __Pyx_INCREF(__pyx_t_3); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(0, 2936, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_5, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2936, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        }
      } else {
        __pyx_t_3 = __pyx_t_6(__pyx_t_5);
        if (unlikely(!__pyx_t_3)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 2936, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_XDECREF_SET(__pyx_v_s_m, __pyx_t_3);
      __pyx_t_3 = 0;
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_DECREF_SET(__pyx_v_match_stringloc, __pyx_t_2);
      __pyx_t_3 = __Pyx_PyInt_AddObjC(__pyx_t_2, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2936, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2);
      __pyx_t_2 = __pyx_t_3;
      __pyx_t_3 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":2937
 * 
 *             for match_stringloc, s_m in enumerate(zip(instring[loc:maxloc], match_string)):
 *                 src, mat = s_m             # <<<<<<<<<<<<<<
 *                 if src != mat:
 *                     mismatches.append(match_stringloc)
 */
      if ((likely(PyTuple_CheckExact(__pyx_v_s_m))) || (PyList_CheckExact(__pyx_v_s_m))) {
        PyObject* sequence = __pyx_v_s_m;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 2937, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_7 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_7 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_7);
        #else
        __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2937, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2937, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      } else {
        Py_ssize_t index = -1;
        __pyx_t_8 = PyObject_GetIter(__pyx_v_s_m); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2937, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_9 = Py_TYPE(__pyx_t_8)->tp_iternext;
        index = 0; __pyx_t_3 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_3)) goto __pyx_L6_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_3);
        index = 1; __pyx_t_7 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_7)) goto __pyx_L6_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_7);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_9(__pyx_t_8), 2) < 0) __PYX_ERR(0, 2937, __pyx_L1_error)
        __pyx_t_9 = NULL;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L7_unpacking_done;
        __pyx_L6_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_9 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 2937, __pyx_L1_error)
        __pyx_L7_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_src, __pyx_t_3);
      __pyx_t_3 = 0;
      __Pyx_XDECREF_SET(__pyx_v_mat, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":2938
 *             for match_stringloc, s_m in enumerate(zip(instring[loc:maxloc], match_string)):
 *                 src, mat = s_m
 *                 if src != mat:             # <<<<<<<<<<<<<<
 *                     mismatches.append(match_stringloc)
 *                     if len(mismatches) > maxMismatches:
 */
      __pyx_t_7 = PyObject_RichCompare(__pyx_v_src, __pyx_v_mat, Py_NE); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2938, __pyx_L1_error)
      __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 2938, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (__pyx_t_4) {

        /* "Aspidites/_vendor/pyparsing_extension.py":2939
 *                 src, mat = s_m
 *                 if src != mat:
 *                     mismatches.append(match_stringloc)             # <<<<<<<<<<<<<<
 *                     if len(mismatches) > maxMismatches:
 *                         break
 */
        __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_mismatches, __pyx_v_match_stringloc); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 2939, __pyx_L1_error)

        /* "Aspidites/_vendor/pyparsing_extension.py":2940
 *                 if src != mat:
 *                     mismatches.append(match_stringloc)
 *                     if len(mismatches) > maxMismatches:             # <<<<<<<<<<<<<<
 *                         break
 *             else:
 */
        __pyx_t_11 = PyList_GET_SIZE(__pyx_v_mismatches); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2940, __pyx_L1_error)
        __pyx_t_7 = PyInt_FromSsize_t(__pyx_t_11); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2940, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_3 = PyObject_RichCompare(__pyx_t_7, __pyx_v_maxMismatches, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2940, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 2940, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (__pyx_t_4) {

          /* "Aspidites/_vendor/pyparsing_extension.py":2941
 *                     mismatches.append(match_stringloc)
 *                     if len(mismatches) > maxMismatches:
 *                         break             # <<<<<<<<<<<<<<
 *             else:
 *                 loc = match_stringloc + 1
 */
          goto __pyx_L5_break;

          /* "Aspidites/_vendor/pyparsing_extension.py":2940
 *                 if src != mat:
 *                     mismatches.append(match_stringloc)
 *                     if len(mismatches) > maxMismatches:             # <<<<<<<<<<<<<<
 *                         break
 *             else:
 */
        }

        /* "Aspidites/_vendor/pyparsing_extension.py":2938
 *             for match_stringloc, s_m in enumerate(zip(instring[loc:maxloc], match_string)):
 *                 src, mat = s_m
 *                 if src != mat:             # <<<<<<<<<<<<<<
 *                     mismatches.append(match_stringloc)
 *                     if len(mismatches) > maxMismatches:
 */
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":2936
 *             maxMismatches = self.maxMismatches
 * 
 *             for match_stringloc, s_m in enumerate(zip(instring[loc:maxloc], match_string)):             # <<<<<<<<<<<<<<
 *                 src, mat = s_m
 *                 if src != mat:
 */
    }
    /*else*/ {

      /* "Aspidites/_vendor/pyparsing_extension.py":2943
 *                         break
 *             else:
 *                 loc = match_stringloc + 1             # <<<<<<<<<<<<<<
 *                 results = ParseResults([instring[start:loc]])
 *                 results['original'] = match_string
 */
      __pyx_t_3 = __Pyx_PyInt_AddObjC(__pyx_v_match_stringloc, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2943, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF_SET(__pyx_v_loc, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":2944
 *             else:
 *                 loc = match_stringloc + 1
 *                 results = ParseResults([instring[start:loc]])             # <<<<<<<<<<<<<<
 *                 results['original'] = match_string
 *                 results['mismatches'] = mismatches
 */
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_ParseResults); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2944, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = __Pyx_PyObject_GetSlice(__pyx_v_instring, 0, 0, &__pyx_v_start, &__pyx_v_loc, NULL, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2944, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_12 = PyList_New(1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 2944, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_GIVEREF(__pyx_t_8);
      PyList_SET_ITEM(__pyx_t_12, 0, __pyx_t_8);
      __pyx_t_8 = 0;
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      __pyx_t_3 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_12) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_12);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2944, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_v_results = __pyx_t_3;
      __pyx_t_3 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":2945
 *                 loc = match_stringloc + 1
 *                 results = ParseResults([instring[start:loc]])
 *                 results['original'] = match_string             # <<<<<<<<<<<<<<
 *                 results['mismatches'] = mismatches
 *                 return loc, results
 */
      if (unlikely(PyObject_SetItem(__pyx_v_results, __pyx_n_s_original, __pyx_v_match_string) < 0)) __PYX_ERR(0, 2945, __pyx_L1_error)

      /* "Aspidites/_vendor/pyparsing_extension.py":2946
 *                 results = ParseResults([instring[start:loc]])
 *                 results['original'] = match_string
 *                 results['mismatches'] = mismatches             # <<<<<<<<<<<<<<
 *                 return loc, results
 * 
 */
      if (unlikely(PyObject_SetItem(__pyx_v_results, __pyx_n_s_mismatches, __pyx_v_mismatches) < 0)) __PYX_ERR(0, 2946, __pyx_L1_error)

      /* "Aspidites/_vendor/pyparsing_extension.py":2947
 *                 results['original'] = match_string
 *                 results['mismatches'] = mismatches
 *                 return loc, results             # <<<<<<<<<<<<<<
 * 
 *         raise ParseException(instring, loc, self.errmsg, self)
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2947, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_v_loc);
      __Pyx_GIVEREF(__pyx_v_loc);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_loc);
      __Pyx_INCREF(__pyx_v_results);
      __Pyx_GIVEREF(__pyx_v_results);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_results);
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L0;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":2936
 *             maxMismatches = self.maxMismatches
 * 
 *             for match_stringloc, s_m in enumerate(zip(instring[loc:maxloc], match_string)):             # <<<<<<<<<<<<<<
 *                 src, mat = s_m
 *                 if src != mat:
 */
    __pyx_L5_break:;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":2930
 *         maxloc = start + len(self.match_string)
 * 
 *         if maxloc <= instrlen:             # <<<<<<<<<<<<<<
 *             match_string = self.match_string
 *             match_stringloc = 0
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":2949
 *                 return loc, results
 * 
 *         raise ParseException(instring, loc, self.errmsg, self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2949, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errmsg); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2949, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = NULL;
  __pyx_t_13 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_13 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[5] = {__pyx_t_7, __pyx_v_instring, __pyx_v_loc, __pyx_t_3, __pyx_v_self};
    __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_13, 4+__pyx_t_13); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2949, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[5] = {__pyx_t_7, __pyx_v_instring, __pyx_v_loc, __pyx_t_3, __pyx_v_self};
    __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_13, 4+__pyx_t_13); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2949, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  {
    __pyx_t_12 = PyTuple_New(4+__pyx_t_13); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 2949, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    if (__pyx_t_7) {
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_7); __pyx_t_7 = NULL;
    }
    __Pyx_INCREF(__pyx_v_instring);
    __Pyx_GIVEREF(__pyx_v_instring);
    PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_13, __pyx_v_instring);
    __Pyx_INCREF(__pyx_v_loc);
    __Pyx_GIVEREF(__pyx_v_loc);
    PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_13, __pyx_v_loc);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_12, 2+__pyx_t_13, __pyx_t_3);
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_12, 3+__pyx_t_13, __pyx_v_self);
    __pyx_t_3 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_12, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2949, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_Raise(__pyx_t_2, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __PYX_ERR(0, 2949, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2925
 *         self.mayReturnEmpty = False
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         start = loc
 *         instrlen = len(instring)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.CloseMatch.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_start);
  __Pyx_XDECREF(__pyx_v_maxloc);
  __Pyx_XDECREF(__pyx_v_match_string);
  __Pyx_XDECREF(__pyx_v_match_stringloc);
  __Pyx_XDECREF(__pyx_v_mismatches);
  __Pyx_XDECREF(__pyx_v_maxMismatches);
  __Pyx_XDECREF(__pyx_v_s_m);
  __Pyx_XDECREF(__pyx_v_src);
  __Pyx_XDECREF(__pyx_v_mat);
  __Pyx_XDECREF(__pyx_v_results);
  __Pyx_XDECREF(__pyx_v_loc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3006
 *     """
 * 
 *     def __init__(self, initChars, bodyChars=None, min=1, max=0, exact=0, asKeyword=False, excludeChars=None):             # <<<<<<<<<<<<<<
 *         super(Word, self).__init__()
 *         if excludeChars:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_4Word_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_4Word_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_4Word_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_4Word_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_initChars = 0;
  PyObject *__pyx_v_bodyChars = 0;
  PyObject *__pyx_v_min = 0;
  PyObject *__pyx_v_max = 0;
  PyObject *__pyx_v_exact = 0;
  PyObject *__pyx_v_asKeyword = 0;
  PyObject *__pyx_v_excludeChars = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_initChars,&__pyx_n_s_bodyChars,&__pyx_n_s_min,&__pyx_n_s_max,&__pyx_n_s_exact,&__pyx_n_s_asKeyword,&__pyx_n_s_excludeChars,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_1));
    values[4] = ((PyObject *)((PyObject *)__pyx_int_0));
    values[5] = ((PyObject *)((PyObject *)__pyx_int_0));
    values[6] = ((PyObject *)((PyObject *)Py_False));
    values[7] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_initChars)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 8, 1); __PYX_ERR(0, 3006, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_bodyChars);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_exact);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_asKeyword);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_excludeChars);
          if (value) { values[7] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 3006, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_initChars = values[1];
    __pyx_v_bodyChars = values[2];
    __pyx_v_min = values[3];
    __pyx_v_max = values[4];
    __pyx_v_exact = values[5];
    __pyx_v_asKeyword = values[6];
    __pyx_v_excludeChars = values[7];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 8, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3006, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Word.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Word___init__(__pyx_self, __pyx_v_self, __pyx_v_initChars, __pyx_v_bodyChars, __pyx_v_min, __pyx_v_max, __pyx_v_exact, __pyx_v_asKeyword, __pyx_v_excludeChars);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_4Word_8__init___2generator16(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/pyparsing_extension.py":3010
 *         if excludeChars:
 *             excludeChars = set(excludeChars)
 *             initChars = ''.join(c for c in initChars if c not in excludeChars)             # <<<<<<<<<<<<<<
 *             if bodyChars:
 *                 bodyChars = ''.join(c for c in bodyChars if c not in excludeChars)
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Word_8__init___genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 3010, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__ *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_4Word_8__init___2generator16, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_Word___init___locals_genexpr, __pyx_n_s_Aspidites__vendor_pyparsing_exte); if (unlikely(!gen)) __PYX_ERR(0, 3010, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Word.__init__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_4Word_8__init___2generator16(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L7_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 3010, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_initChars)) { __Pyx_RaiseClosureNameError("initChars"); __PYX_ERR(0, 3010, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_initChars)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_initChars)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_initChars; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_initChars); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3010, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3010, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 3010, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3010, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 3010, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3010, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 3010, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_c);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_c, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_excludeChars)) { __Pyx_RaiseClosureNameError("excludeChars"); __PYX_ERR(0, 3010, __pyx_L1_error) }
    __pyx_t_5 = (__Pyx_PySequence_ContainsTF(__pyx_cur_scope->__pyx_v_c, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_excludeChars, Py_NE)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 3010, __pyx_L1_error)
    __pyx_t_6 = (__pyx_t_5 != 0);
    if (__pyx_t_6) {
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_c);
      __pyx_r = __pyx_cur_scope->__pyx_v_c;
      __Pyx_XGIVEREF(__pyx_t_1);
      __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
      __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
      __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
      /* return from generator, yielding value */
      __pyx_generator->resume_label = 1;
      return __pyx_r;
      __pyx_L7_resume_from_yield:;
      __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
      __pyx_cur_scope->__pyx_t_0 = 0;
      __Pyx_XGOTREF(__pyx_t_1);
      __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
      __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 3010, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_4Word_8__init___5generator17(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/pyparsing_extension.py":3012
 *             initChars = ''.join(c for c in initChars if c not in excludeChars)
 *             if bodyChars:
 *                 bodyChars = ''.join(c for c in bodyChars if c not in excludeChars)             # <<<<<<<<<<<<<<
 *         self.initCharsOrig = initChars
 *         self.initChars = set(initChars)
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Word_8__init___3genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 3012, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__ *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_4Word_8__init___5generator17, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_Word___init___locals_genexpr, __pyx_n_s_Aspidites__vendor_pyparsing_exte); if (unlikely(!gen)) __PYX_ERR(0, 3012, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Word.__init__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_4Word_8__init___5generator17(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L7_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 3012, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_bodyChars)) { __Pyx_RaiseClosureNameError("bodyChars"); __PYX_ERR(0, 3012, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_bodyChars)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_bodyChars)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_bodyChars; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_bodyChars); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3012, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3012, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 3012, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3012, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 3012, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3012, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 3012, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_c);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_c, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_excludeChars)) { __Pyx_RaiseClosureNameError("excludeChars"); __PYX_ERR(0, 3012, __pyx_L1_error) }
    __pyx_t_5 = (__Pyx_PySequence_ContainsTF(__pyx_cur_scope->__pyx_v_c, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_excludeChars, Py_NE)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 3012, __pyx_L1_error)
    __pyx_t_6 = (__pyx_t_5 != 0);
    if (__pyx_t_6) {
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_c);
      __pyx_r = __pyx_cur_scope->__pyx_v_c;
      __Pyx_XGIVEREF(__pyx_t_1);
      __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
      __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
      __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
      /* return from generator, yielding value */
      __pyx_generator->resume_label = 1;
      return __pyx_r;
      __pyx_L7_resume_from_yield:;
      __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
      __pyx_cur_scope->__pyx_t_0 = 0;
      __Pyx_XGOTREF(__pyx_t_1);
      __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
      __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 3012, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3006
 *     """
 * 
 *     def __init__(self, initChars, bodyChars=None, min=1, max=0, exact=0, asKeyword=False, excludeChars=None):             # <<<<<<<<<<<<<<
 *         super(Word, self).__init__()
 *         if excludeChars:
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Word___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_initChars, PyObject *__pyx_v_bodyChars, PyObject *__pyx_v_min, PyObject *__pyx_v_max, PyObject *__pyx_v_exact, PyObject *__pyx_v_asKeyword, PyObject *__pyx_v_excludeChars) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__ *__pyx_cur_scope;
  PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_4Word_8__init___2generator16 = 0;
  PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_4Word_8__init___5generator17 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__ *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 3006, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_initChars = __pyx_v_initChars;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_initChars);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_initChars);
  __pyx_cur_scope->__pyx_v_bodyChars = __pyx_v_bodyChars;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_bodyChars);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_bodyChars);
  __pyx_cur_scope->__pyx_v_excludeChars = __pyx_v_excludeChars;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_excludeChars);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_excludeChars);

  /* "Aspidites/_vendor/pyparsing_extension.py":3007
 * 
 *     def __init__(self, initChars, bodyChars=None, min=1, max=0, exact=0, asKeyword=False, excludeChars=None):
 *         super(Word, self).__init__()             # <<<<<<<<<<<<<<
 *         if excludeChars:
 *             excludeChars = set(excludeChars)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Word); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3007, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3007, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3007, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3007, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3007, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3008
 *     def __init__(self, initChars, bodyChars=None, min=1, max=0, exact=0, asKeyword=False, excludeChars=None):
 *         super(Word, self).__init__()
 *         if excludeChars:             # <<<<<<<<<<<<<<
 *             excludeChars = set(excludeChars)
 *             initChars = ''.join(c for c in initChars if c not in excludeChars)
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_excludeChars); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 3008, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3009
 *         super(Word, self).__init__()
 *         if excludeChars:
 *             excludeChars = set(excludeChars)             # <<<<<<<<<<<<<<
 *             initChars = ''.join(c for c in initChars if c not in excludeChars)
 *             if bodyChars:
 */
    __pyx_t_1 = PySet_New(__pyx_cur_scope->__pyx_v_excludeChars); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3009, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_excludeChars);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_excludeChars, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3010
 *         if excludeChars:
 *             excludeChars = set(excludeChars)
 *             initChars = ''.join(c for c in initChars if c not in excludeChars)             # <<<<<<<<<<<<<<
 *             if bodyChars:
 *                 bodyChars = ''.join(c for c in bodyChars if c not in excludeChars)
 */
    __pyx_t_1 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Word_8__init___genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3010, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyString_Join(__pyx_kp_s__12, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3010, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_initChars);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_initChars, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3011
 *             excludeChars = set(excludeChars)
 *             initChars = ''.join(c for c in initChars if c not in excludeChars)
 *             if bodyChars:             # <<<<<<<<<<<<<<
 *                 bodyChars = ''.join(c for c in bodyChars if c not in excludeChars)
 *         self.initCharsOrig = initChars
 */
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_bodyChars); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 3011, __pyx_L1_error)
    if (__pyx_t_4) {

      /* "Aspidites/_vendor/pyparsing_extension.py":3012
 *             initChars = ''.join(c for c in initChars if c not in excludeChars)
 *             if bodyChars:
 *                 bodyChars = ''.join(c for c in bodyChars if c not in excludeChars)             # <<<<<<<<<<<<<<
 *         self.initCharsOrig = initChars
 *         self.initChars = set(initChars)
 */
      __pyx_t_3 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Word_8__init___3genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3012, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = __Pyx_PyString_Join(__pyx_kp_s__12, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3012, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_bodyChars);
      __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_bodyChars, __pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_1);
      __pyx_t_1 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":3011
 *             excludeChars = set(excludeChars)
 *             initChars = ''.join(c for c in initChars if c not in excludeChars)
 *             if bodyChars:             # <<<<<<<<<<<<<<
 *                 bodyChars = ''.join(c for c in bodyChars if c not in excludeChars)
 *         self.initCharsOrig = initChars
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":3008
 *     def __init__(self, initChars, bodyChars=None, min=1, max=0, exact=0, asKeyword=False, excludeChars=None):
 *         super(Word, self).__init__()
 *         if excludeChars:             # <<<<<<<<<<<<<<
 *             excludeChars = set(excludeChars)
 *             initChars = ''.join(c for c in initChars if c not in excludeChars)
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3013
 *             if bodyChars:
 *                 bodyChars = ''.join(c for c in bodyChars if c not in excludeChars)
 *         self.initCharsOrig = initChars             # <<<<<<<<<<<<<<
 *         self.initChars = set(initChars)
 *         if bodyChars:
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_initCharsOrig, __pyx_cur_scope->__pyx_v_initChars) < 0) __PYX_ERR(0, 3013, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3014
 *                 bodyChars = ''.join(c for c in bodyChars if c not in excludeChars)
 *         self.initCharsOrig = initChars
 *         self.initChars = set(initChars)             # <<<<<<<<<<<<<<
 *         if bodyChars:
 *             self.bodyCharsOrig = bodyChars
 */
  __pyx_t_1 = PySet_New(__pyx_cur_scope->__pyx_v_initChars); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_initChars, __pyx_t_1) < 0) __PYX_ERR(0, 3014, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3015
 *         self.initCharsOrig = initChars
 *         self.initChars = set(initChars)
 *         if bodyChars:             # <<<<<<<<<<<<<<
 *             self.bodyCharsOrig = bodyChars
 *             self.bodyChars = set(bodyChars)
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_bodyChars); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 3015, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3016
 *         self.initChars = set(initChars)
 *         if bodyChars:
 *             self.bodyCharsOrig = bodyChars             # <<<<<<<<<<<<<<
 *             self.bodyChars = set(bodyChars)
 *         else:
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_bodyCharsOrig, __pyx_cur_scope->__pyx_v_bodyChars) < 0) __PYX_ERR(0, 3016, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":3017
 *         if bodyChars:
 *             self.bodyCharsOrig = bodyChars
 *             self.bodyChars = set(bodyChars)             # <<<<<<<<<<<<<<
 *         else:
 *             self.bodyCharsOrig = initChars
 */
    __pyx_t_1 = PySet_New(__pyx_cur_scope->__pyx_v_bodyChars); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3017, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_bodyChars, __pyx_t_1) < 0) __PYX_ERR(0, 3017, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3015
 *         self.initCharsOrig = initChars
 *         self.initChars = set(initChars)
 *         if bodyChars:             # <<<<<<<<<<<<<<
 *             self.bodyCharsOrig = bodyChars
 *             self.bodyChars = set(bodyChars)
 */
    goto __pyx_L5;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3019
 *             self.bodyChars = set(bodyChars)
 *         else:
 *             self.bodyCharsOrig = initChars             # <<<<<<<<<<<<<<
 *             self.bodyChars = set(initChars)
 * 
 */
  /*else*/ {
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_bodyCharsOrig, __pyx_cur_scope->__pyx_v_initChars) < 0) __PYX_ERR(0, 3019, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":3020
 *         else:
 *             self.bodyCharsOrig = initChars
 *             self.bodyChars = set(initChars)             # <<<<<<<<<<<<<<
 * 
 *         self.maxSpecified = max > 0
 */
    __pyx_t_1 = PySet_New(__pyx_cur_scope->__pyx_v_initChars); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3020, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_bodyChars, __pyx_t_1) < 0) __PYX_ERR(0, 3020, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_L5:;

  /* "Aspidites/_vendor/pyparsing_extension.py":3022
 *             self.bodyChars = set(initChars)
 * 
 *         self.maxSpecified = max > 0             # <<<<<<<<<<<<<<
 * 
 *         if min < 1:
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_max, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3022, __pyx_L1_error)
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_maxSpecified, __pyx_t_1) < 0) __PYX_ERR(0, 3022, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3024
 *         self.maxSpecified = max > 0
 * 
 *         if min < 1:             # <<<<<<<<<<<<<<
 *             raise ValueError(
 *                 "cannot specify a minimum length < 1; use Optional(Word()) if zero-length word is permitted")
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_min, __pyx_int_1, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3024, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 3024, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_4)) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3025
 * 
 *         if min < 1:
 *             raise ValueError(             # <<<<<<<<<<<<<<
 *                 "cannot specify a minimum length < 1; use Optional(Word()) if zero-length word is permitted")
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__83, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3025, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 3025, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":3024
 *         self.maxSpecified = max > 0
 * 
 *         if min < 1:             # <<<<<<<<<<<<<<
 *             raise ValueError(
 *                 "cannot specify a minimum length < 1; use Optional(Word()) if zero-length word is permitted")
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3028
 *                 "cannot specify a minimum length < 1; use Optional(Word()) if zero-length word is permitted")
 * 
 *         self.minLen = min             # <<<<<<<<<<<<<<
 * 
 *         if max > 0:
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_minLen, __pyx_v_min) < 0) __PYX_ERR(0, 3028, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3030
 *         self.minLen = min
 * 
 *         if max > 0:             # <<<<<<<<<<<<<<
 *             self.maxLen = max
 *         else:
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_max, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3030, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 3030, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3031
 * 
 *         if max > 0:
 *             self.maxLen = max             # <<<<<<<<<<<<<<
 *         else:
 *             self.maxLen = sys.maxsize
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_maxLen, __pyx_v_max) < 0) __PYX_ERR(0, 3031, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":3030
 *         self.minLen = min
 * 
 *         if max > 0:             # <<<<<<<<<<<<<<
 *             self.maxLen = max
 *         else:
 */
    goto __pyx_L7;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3033
 *             self.maxLen = max
 *         else:
 *             self.maxLen = sys.maxsize             # <<<<<<<<<<<<<<
 * 
 *         if exact > 0:
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_sys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3033, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_maxsize); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3033, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_maxLen, __pyx_t_3) < 0) __PYX_ERR(0, 3033, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_L7:;

  /* "Aspidites/_vendor/pyparsing_extension.py":3035
 *             self.maxLen = sys.maxsize
 * 
 *         if exact > 0:             # <<<<<<<<<<<<<<
 *             self.maxLen = exact
 *             self.minLen = exact
 */
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_exact, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3035, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 3035, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3036
 * 
 *         if exact > 0:
 *             self.maxLen = exact             # <<<<<<<<<<<<<<
 *             self.minLen = exact
 * 
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_maxLen, __pyx_v_exact) < 0) __PYX_ERR(0, 3036, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":3037
 *         if exact > 0:
 *             self.maxLen = exact
 *             self.minLen = exact             # <<<<<<<<<<<<<<
 * 
 *         self.name = str(self)
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_minLen, __pyx_v_exact) < 0) __PYX_ERR(0, 3037, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":3035
 *             self.maxLen = sys.maxsize
 * 
 *         if exact > 0:             # <<<<<<<<<<<<<<
 *             self.maxLen = exact
 *             self.minLen = exact
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3039
 *             self.minLen = exact
 * 
 *         self.name = str(self)             # <<<<<<<<<<<<<<
 *         self.errmsg = "Expected " + self.name
 *         self.mayIndexError = False
 */
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3039, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_name, __pyx_t_3) < 0) __PYX_ERR(0, 3039, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3040
 * 
 *         self.name = str(self)
 *         self.errmsg = "Expected " + self.name             # <<<<<<<<<<<<<<
 *         self.mayIndexError = False
 *         self.asKeyword = asKeyword
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3040, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyNumber_Add(__pyx_kp_s_Expected, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3040, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_errmsg, __pyx_t_1) < 0) __PYX_ERR(0, 3040, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3041
 *         self.name = str(self)
 *         self.errmsg = "Expected " + self.name
 *         self.mayIndexError = False             # <<<<<<<<<<<<<<
 *         self.asKeyword = asKeyword
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_mayIndexError, Py_False) < 0) __PYX_ERR(0, 3041, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3042
 *         self.errmsg = "Expected " + self.name
 *         self.mayIndexError = False
 *         self.asKeyword = asKeyword             # <<<<<<<<<<<<<<
 * 
 *         if ' ' not in self.initCharsOrig + self.bodyCharsOrig and (min == 1 and max == 0 and exact == 0):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_asKeyword, __pyx_v_asKeyword) < 0) __PYX_ERR(0, 3042, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3044
 *         self.asKeyword = asKeyword
 * 
 *         if ' ' not in self.initCharsOrig + self.bodyCharsOrig and (min == 1 and max == 0 and exact == 0):             # <<<<<<<<<<<<<<
 *             if self.bodyCharsOrig == self.initCharsOrig:
 *                 self.reString = "[%s]+" % _escapeRegexRangeChars(self.initCharsOrig)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_initCharsOrig); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_bodyCharsOrig); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyNumber_Add(__pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = (__Pyx_PySequence_ContainsTF(__pyx_kp_s__29, __pyx_t_2, Py_NE)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 3044, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = (__pyx_t_5 != 0);
  if (__pyx_t_6) {
  } else {
    __pyx_t_4 = __pyx_t_6;
    goto __pyx_L10_bool_binop_done;
  }
  __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_v_min, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 3044, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_6) {
  } else {
    __pyx_t_4 = __pyx_t_6;
    goto __pyx_L10_bool_binop_done;
  }
  __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_v_max, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 3044, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_6) {
  } else {
    __pyx_t_4 = __pyx_t_6;
    goto __pyx_L10_bool_binop_done;
  }
  __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_v_exact, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 3044, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __pyx_t_6;
  __pyx_L10_bool_binop_done:;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3045
 * 
 *         if ' ' not in self.initCharsOrig + self.bodyCharsOrig and (min == 1 and max == 0 and exact == 0):
 *             if self.bodyCharsOrig == self.initCharsOrig:             # <<<<<<<<<<<<<<
 *                 self.reString = "[%s]+" % _escapeRegexRangeChars(self.initCharsOrig)
 *             elif len(self.initCharsOrig) == 1:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_bodyCharsOrig); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3045, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_initCharsOrig); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3045, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyObject_RichCompare(__pyx_t_2, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3045, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 3045, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_4) {

      /* "Aspidites/_vendor/pyparsing_extension.py":3046
 *         if ' ' not in self.initCharsOrig + self.bodyCharsOrig and (min == 1 and max == 0 and exact == 0):
 *             if self.bodyCharsOrig == self.initCharsOrig:
 *                 self.reString = "[%s]+" % _escapeRegexRangeChars(self.initCharsOrig)             # <<<<<<<<<<<<<<
 *             elif len(self.initCharsOrig) == 1:
 *                 self.reString = "%s[%s]*" % (re.escape(self.initCharsOrig),
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_escapeRegexRangeChars); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3046, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_initCharsOrig); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3046, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_7, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3046, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyString_FormatSafe(__pyx_kp_s_s_3, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3046, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_reString, __pyx_t_3) < 0) __PYX_ERR(0, 3046, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":3045
 * 
 *         if ' ' not in self.initCharsOrig + self.bodyCharsOrig and (min == 1 and max == 0 and exact == 0):
 *             if self.bodyCharsOrig == self.initCharsOrig:             # <<<<<<<<<<<<<<
 *                 self.reString = "[%s]+" % _escapeRegexRangeChars(self.initCharsOrig)
 *             elif len(self.initCharsOrig) == 1:
 */
      goto __pyx_L14;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":3047
 *             if self.bodyCharsOrig == self.initCharsOrig:
 *                 self.reString = "[%s]+" % _escapeRegexRangeChars(self.initCharsOrig)
 *             elif len(self.initCharsOrig) == 1:             # <<<<<<<<<<<<<<
 *                 self.reString = "%s[%s]*" % (re.escape(self.initCharsOrig),
 *                                              _escapeRegexRangeChars(self.bodyCharsOrig),)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_initCharsOrig); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3047, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = PyObject_Length(__pyx_t_3); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 3047, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = ((__pyx_t_8 == 1) != 0);
    if (__pyx_t_4) {

      /* "Aspidites/_vendor/pyparsing_extension.py":3048
 *                 self.reString = "[%s]+" % _escapeRegexRangeChars(self.initCharsOrig)
 *             elif len(self.initCharsOrig) == 1:
 *                 self.reString = "%s[%s]*" % (re.escape(self.initCharsOrig),             # <<<<<<<<<<<<<<
 *                                              _escapeRegexRangeChars(self.bodyCharsOrig),)
 *             else:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_re); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3048, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_escape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3048, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_initCharsOrig); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3048, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_3 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_7, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3048, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":3049
 *             elif len(self.initCharsOrig) == 1:
 *                 self.reString = "%s[%s]*" % (re.escape(self.initCharsOrig),
 *                                              _escapeRegexRangeChars(self.bodyCharsOrig),)             # <<<<<<<<<<<<<<
 *             else:
 *                 self.reString = "[%s][%s]*" % (_escapeRegexRangeChars(self.initCharsOrig),
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_escapeRegexRangeChars); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3049, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_bodyCharsOrig); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3049, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_9 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_2 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_9, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_7);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3049, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":3048
 *                 self.reString = "[%s]+" % _escapeRegexRangeChars(self.initCharsOrig)
 *             elif len(self.initCharsOrig) == 1:
 *                 self.reString = "%s[%s]*" % (re.escape(self.initCharsOrig),             # <<<<<<<<<<<<<<
 *                                              _escapeRegexRangeChars(self.bodyCharsOrig),)
 *             else:
 */
      __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3048, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
      __pyx_t_3 = 0;
      __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_s_s_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3048, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_reString, __pyx_t_2) < 0) __PYX_ERR(0, 3048, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":3047
 *             if self.bodyCharsOrig == self.initCharsOrig:
 *                 self.reString = "[%s]+" % _escapeRegexRangeChars(self.initCharsOrig)
 *             elif len(self.initCharsOrig) == 1:             # <<<<<<<<<<<<<<
 *                 self.reString = "%s[%s]*" % (re.escape(self.initCharsOrig),
 *                                              _escapeRegexRangeChars(self.bodyCharsOrig),)
 */
      goto __pyx_L14;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":3051
 *                                              _escapeRegexRangeChars(self.bodyCharsOrig),)
 *             else:
 *                 self.reString = "[%s][%s]*" % (_escapeRegexRangeChars(self.initCharsOrig),             # <<<<<<<<<<<<<<
 *                                                _escapeRegexRangeChars(self.bodyCharsOrig),)
 *             if self.asKeyword:
 */
    /*else*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_escapeRegexRangeChars); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3051, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_initCharsOrig); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3051, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_2 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_7, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3051, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":3052
 *             else:
 *                 self.reString = "[%s][%s]*" % (_escapeRegexRangeChars(self.initCharsOrig),
 *                                                _escapeRegexRangeChars(self.bodyCharsOrig),)             # <<<<<<<<<<<<<<
 *             if self.asKeyword:
 *                 self.reString = r"\b" + self.reString + r"\b"
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_escapeRegexRangeChars); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3052, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_bodyCharsOrig); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3052, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_9 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_9, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3052, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":3051
 *                                              _escapeRegexRangeChars(self.bodyCharsOrig),)
 *             else:
 *                 self.reString = "[%s][%s]*" % (_escapeRegexRangeChars(self.initCharsOrig),             # <<<<<<<<<<<<<<
 *                                                _escapeRegexRangeChars(self.bodyCharsOrig),)
 *             if self.asKeyword:
 */
      __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3051, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
      __pyx_t_2 = 0;
      __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_s_s_4, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3051, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_reString, __pyx_t_1) < 0) __PYX_ERR(0, 3051, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L14:;

    /* "Aspidites/_vendor/pyparsing_extension.py":3053
 *                 self.reString = "[%s][%s]*" % (_escapeRegexRangeChars(self.initCharsOrig),
 *                                                _escapeRegexRangeChars(self.bodyCharsOrig),)
 *             if self.asKeyword:             # <<<<<<<<<<<<<<
 *                 self.reString = r"\b" + self.reString + r"\b"
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_asKeyword); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3053, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 3053, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_4) {

      /* "Aspidites/_vendor/pyparsing_extension.py":3054
 *                                                _escapeRegexRangeChars(self.bodyCharsOrig),)
 *             if self.asKeyword:
 *                 self.reString = r"\b" + self.reString + r"\b"             # <<<<<<<<<<<<<<
 * 
 *             try:
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_reString); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3054, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = PyNumber_Add(__pyx_kp_s_b, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3054, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = PyNumber_Add(__pyx_t_3, __pyx_kp_s_b); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3054, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_reString, __pyx_t_1) < 0) __PYX_ERR(0, 3054, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":3053
 *                 self.reString = "[%s][%s]*" % (_escapeRegexRangeChars(self.initCharsOrig),
 *                                                _escapeRegexRangeChars(self.bodyCharsOrig),)
 *             if self.asKeyword:             # <<<<<<<<<<<<<<
 *                 self.reString = r"\b" + self.reString + r"\b"
 * 
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":3056
 *                 self.reString = r"\b" + self.reString + r"\b"
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 self.re = re.compile(self.reString)
 *             except Exception:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      /*try:*/ {

        /* "Aspidites/_vendor/pyparsing_extension.py":3057
 * 
 *             try:
 *                 self.re = re.compile(self.reString)             # <<<<<<<<<<<<<<
 *             except Exception:
 *                 self.re = None
 */
        __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_re); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3057, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_compile); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3057, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_reString); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3057, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_7, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3057, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_re, __pyx_t_1) < 0) __PYX_ERR(0, 3057, __pyx_L16_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":3056
 *                 self.reString = r"\b" + self.reString + r"\b"
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 self.re = re.compile(self.reString)
 *             except Exception:
 */
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":3061
 *                 self.re = None
 *             else:
 *                 self.re_match = self.re.match             # <<<<<<<<<<<<<<
 *                 self.__class__ = _WordRegex
 * 
 */
      /*else:*/ {
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_re); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3061, __pyx_L18_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3061, __pyx_L18_except_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_re_match, __pyx_t_2) < 0) __PYX_ERR(0, 3061, __pyx_L18_except_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":3062
 *             else:
 *                 self.re_match = self.re.match
 *                 self.__class__ = _WordRegex             # <<<<<<<<<<<<<<
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 */
        __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_WordRegex); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3062, __pyx_L18_except_error)
        __Pyx_GOTREF(__pyx_t_2);
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_class, __pyx_t_2) < 0) __PYX_ERR(0, 3062, __pyx_L18_except_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      goto __pyx_L21_try_end;
      __pyx_L16_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":3058
 *             try:
 *                 self.re = re.compile(self.reString)
 *             except Exception:             # <<<<<<<<<<<<<<
 *                 self.re = None
 *             else:
 */
      __pyx_t_13 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
      if (__pyx_t_13) {
        __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Word.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_1, &__pyx_t_3) < 0) __PYX_ERR(0, 3058, __pyx_L18_except_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_3);

        /* "Aspidites/_vendor/pyparsing_extension.py":3059
 *                 self.re = re.compile(self.reString)
 *             except Exception:
 *                 self.re = None             # <<<<<<<<<<<<<<
 *             else:
 *                 self.re_match = self.re.match
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_re, Py_None) < 0) __PYX_ERR(0, 3059, __pyx_L18_except_error)
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        goto __pyx_L17_exception_handled;
      }
      goto __pyx_L18_except_error;
      __pyx_L18_except_error:;

      /* "Aspidites/_vendor/pyparsing_extension.py":3056
 *                 self.reString = r"\b" + self.reString + r"\b"
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 self.re = re.compile(self.reString)
 *             except Exception:
 */
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      goto __pyx_L1_error;
      __pyx_L17_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      __pyx_L21_try_end:;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":3044
 *         self.asKeyword = asKeyword
 * 
 *         if ' ' not in self.initCharsOrig + self.bodyCharsOrig and (min == 1 and max == 0 and exact == 0):             # <<<<<<<<<<<<<<
 *             if self.bodyCharsOrig == self.initCharsOrig:
 *                 self.reString = "[%s]+" % _escapeRegexRangeChars(self.initCharsOrig)
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3006
 *     """
 * 
 *     def __init__(self, initChars, bodyChars=None, min=1, max=0, exact=0, asKeyword=False, excludeChars=None):             # <<<<<<<<<<<<<<
 *         super(Word, self).__init__()
 *         if excludeChars:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Word.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_4Word_8__init___2generator16);
  __Pyx_XDECREF(__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_4Word_8__init___5generator17);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3064
 *                 self.__class__ = _WordRegex
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if instring[loc] not in self.initChars:
 *             raise ParseException(instring, loc, self.errmsg, self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_4Word_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_4Word_3parseImpl = {"parseImpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_4Word_3parseImpl, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_4Word_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  CYTHON_UNUSED PyObject *__pyx_v_doActions = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parseImpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_doActions,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 1); __PYX_ERR(0, 3064, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 2); __PYX_ERR(0, 3064, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_doActions);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "parseImpl") < 0)) __PYX_ERR(0, 3064, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_doActions = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3064, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Word.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Word_2parseImpl(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Word_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions) {
  PyObject *__pyx_v_start = NULL;
  Py_ssize_t __pyx_v_instrlen;
  PyObject *__pyx_v_bodychars = NULL;
  PyObject *__pyx_v_maxloc = NULL;
  int __pyx_v_throwException;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parseImpl", 0);
  __Pyx_INCREF(__pyx_v_loc);

  /* "Aspidites/_vendor/pyparsing_extension.py":3065
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if instring[loc] not in self.initChars:             # <<<<<<<<<<<<<<
 *             raise ParseException(instring, loc, self.errmsg, self)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_instring, __pyx_v_loc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3065, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_initChars); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3065, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_t_2, Py_NE)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 3065, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (unlikely(__pyx_t_4)) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3066
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if instring[loc] not in self.initChars:
 *             raise ParseException(instring, loc, self.errmsg, self)             # <<<<<<<<<<<<<<
 * 
 *         start = loc
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3066, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errmsg); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3066, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[5] = {__pyx_t_6, __pyx_v_instring, __pyx_v_loc, __pyx_t_5, __pyx_v_self};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3066, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[5] = {__pyx_t_6, __pyx_v_instring, __pyx_v_loc, __pyx_t_5, __pyx_v_self};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3066, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(4+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3066, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_INCREF(__pyx_v_instring);
      __Pyx_GIVEREF(__pyx_v_instring);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_v_instring);
      __Pyx_INCREF(__pyx_v_loc);
      __Pyx_GIVEREF(__pyx_v_loc);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_v_loc);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_7, __pyx_t_5);
      __Pyx_INCREF(__pyx_v_self);
      __Pyx_GIVEREF(__pyx_v_self);
      PyTuple_SET_ITEM(__pyx_t_8, 3+__pyx_t_7, __pyx_v_self);
      __pyx_t_5 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3066, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 3066, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":3065
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if instring[loc] not in self.initChars:             # <<<<<<<<<<<<<<
 *             raise ParseException(instring, loc, self.errmsg, self)
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3068
 *             raise ParseException(instring, loc, self.errmsg, self)
 * 
 *         start = loc             # <<<<<<<<<<<<<<
 *         loc += 1
 *         instrlen = len(instring)
 */
  __Pyx_INCREF(__pyx_v_loc);
  __pyx_v_start = __pyx_v_loc;

  /* "Aspidites/_vendor/pyparsing_extension.py":3069
 * 
 *         start = loc
 *         loc += 1             # <<<<<<<<<<<<<<
 *         instrlen = len(instring)
 *         bodychars = self.bodyChars
 */
  __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_v_loc, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3069, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF_SET(__pyx_v_loc, __pyx_t_2);
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3070
 *         start = loc
 *         loc += 1
 *         instrlen = len(instring)             # <<<<<<<<<<<<<<
 *         bodychars = self.bodyChars
 *         maxloc = start + self.maxLen
 */
  __pyx_t_9 = PyObject_Length(__pyx_v_instring); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 3070, __pyx_L1_error)
  __pyx_v_instrlen = __pyx_t_9;

  /* "Aspidites/_vendor/pyparsing_extension.py":3071
 *         loc += 1
 *         instrlen = len(instring)
 *         bodychars = self.bodyChars             # <<<<<<<<<<<<<<
 *         maxloc = start + self.maxLen
 *         maxloc = min(maxloc, instrlen)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_bodyChars); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3071, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_bodychars = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3072
 *         instrlen = len(instring)
 *         bodychars = self.bodyChars
 *         maxloc = start + self.maxLen             # <<<<<<<<<<<<<<
 *         maxloc = min(maxloc, instrlen)
 *         while loc < maxloc and instring[loc] in bodychars:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_maxLen); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3072, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyNumber_Add(__pyx_v_start, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3072, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_maxloc = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3073
 *         bodychars = self.bodyChars
 *         maxloc = start + self.maxLen
 *         maxloc = min(maxloc, instrlen)             # <<<<<<<<<<<<<<
 *         while loc < maxloc and instring[loc] in bodychars:
 *             loc += 1
 */
  __pyx_t_9 = __pyx_v_instrlen;
  __Pyx_INCREF(__pyx_v_maxloc);
  __pyx_t_1 = __pyx_v_maxloc;
  __pyx_t_8 = PyInt_FromSsize_t(__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3073, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_5 = PyObject_RichCompare(__pyx_t_8, __pyx_t_1, Py_LT); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3073, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 3073, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_4) {
    __pyx_t_5 = PyInt_FromSsize_t(__pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3073, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = __pyx_t_5;
    __pyx_t_5 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = __pyx_t_1;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_t_2;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_maxloc, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3074
 *         maxloc = start + self.maxLen
 *         maxloc = min(maxloc, instrlen)
 *         while loc < maxloc and instring[loc] in bodychars:             # <<<<<<<<<<<<<<
 *             loc += 1
 * 
 */
  while (1) {
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_loc, __pyx_v_maxloc, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3074, __pyx_L1_error)
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 3074, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_3) {
    } else {
      __pyx_t_4 = __pyx_t_3;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_instring, __pyx_v_loc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3074, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_v_bodychars, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 3074, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_10 = (__pyx_t_3 != 0);
    __pyx_t_4 = __pyx_t_10;
    __pyx_L6_bool_binop_done:;
    if (!__pyx_t_4) break;

    /* "Aspidites/_vendor/pyparsing_extension.py":3075
 *         maxloc = min(maxloc, instrlen)
 *         while loc < maxloc and instring[loc] in bodychars:
 *             loc += 1             # <<<<<<<<<<<<<<
 * 
 *         throwException = False
 */
    __pyx_t_1 = __Pyx_PyInt_AddObjC(__pyx_v_loc, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3075, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_loc, __pyx_t_1);
    __pyx_t_1 = 0;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3077
 *             loc += 1
 * 
 *         throwException = False             # <<<<<<<<<<<<<<
 *         if loc - start < self.minLen:
 *             throwException = True
 */
  __pyx_v_throwException = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3078
 * 
 *         throwException = False
 *         if loc - start < self.minLen:             # <<<<<<<<<<<<<<
 *             throwException = True
 *         elif self.maxSpecified and loc < instrlen and instring[loc] in bodychars:
 */
  __pyx_t_1 = PyNumber_Subtract(__pyx_v_loc, __pyx_v_start); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_minLen); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_LT); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3078, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 3078, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3079
 *         throwException = False
 *         if loc - start < self.minLen:
 *             throwException = True             # <<<<<<<<<<<<<<
 *         elif self.maxSpecified and loc < instrlen and instring[loc] in bodychars:
 *             throwException = True
 */
    __pyx_v_throwException = 1;

    /* "Aspidites/_vendor/pyparsing_extension.py":3078
 * 
 *         throwException = False
 *         if loc - start < self.minLen:             # <<<<<<<<<<<<<<
 *             throwException = True
 *         elif self.maxSpecified and loc < instrlen and instring[loc] in bodychars:
 */
    goto __pyx_L8;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3080
 *         if loc - start < self.minLen:
 *             throwException = True
 *         elif self.maxSpecified and loc < instrlen and instring[loc] in bodychars:             # <<<<<<<<<<<<<<
 *             throwException = True
 *         elif self.asKeyword:
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_maxSpecified); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3080, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 3080, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_10) {
  } else {
    __pyx_t_4 = __pyx_t_10;
    goto __pyx_L9_bool_binop_done;
  }
  __pyx_t_5 = PyInt_FromSsize_t(__pyx_v_instrlen); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3080, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_loc, __pyx_t_5, Py_LT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3080, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 3080, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_10) {
  } else {
    __pyx_t_4 = __pyx_t_10;
    goto __pyx_L9_bool_binop_done;
  }
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_instring, __pyx_v_loc); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3080, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_10 = (__Pyx_PySequence_ContainsTF(__pyx_t_2, __pyx_v_bodychars, Py_EQ)); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 3080, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = (__pyx_t_10 != 0);
  __pyx_t_4 = __pyx_t_3;
  __pyx_L9_bool_binop_done:;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3081
 *             throwException = True
 *         elif self.maxSpecified and loc < instrlen and instring[loc] in bodychars:
 *             throwException = True             # <<<<<<<<<<<<<<
 *         elif self.asKeyword:
 *             if (start > 0 and instring[start - 1] in bodychars
 */
    __pyx_v_throwException = 1;

    /* "Aspidites/_vendor/pyparsing_extension.py":3080
 *         if loc - start < self.minLen:
 *             throwException = True
 *         elif self.maxSpecified and loc < instrlen and instring[loc] in bodychars:             # <<<<<<<<<<<<<<
 *             throwException = True
 *         elif self.asKeyword:
 */
    goto __pyx_L8;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3082
 *         elif self.maxSpecified and loc < instrlen and instring[loc] in bodychars:
 *             throwException = True
 *         elif self.asKeyword:             # <<<<<<<<<<<<<<
 *             if (start > 0 and instring[start - 1] in bodychars
 *                     or loc < instrlen and instring[loc] in bodychars):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_asKeyword); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3082, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 3082, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3083
 *             throwException = True
 *         elif self.asKeyword:
 *             if (start > 0 and instring[start - 1] in bodychars             # <<<<<<<<<<<<<<
 *                     or loc < instrlen and instring[loc] in bodychars):
 *                 throwException = True
 */
    __pyx_t_2 = PyObject_RichCompare(__pyx_v_start, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3083, __pyx_L1_error)
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 3083, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (!__pyx_t_3) {
      goto __pyx_L14_next_or;
    } else {
    }
    __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_v_start, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3083, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_instring, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3083, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_3 = (__Pyx_PySequence_ContainsTF(__pyx_t_5, __pyx_v_bodychars, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 3083, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_10 = (__pyx_t_3 != 0);
    if (!__pyx_t_10) {
    } else {
      __pyx_t_4 = __pyx_t_10;
      goto __pyx_L13_bool_binop_done;
    }
    __pyx_L14_next_or:;

    /* "Aspidites/_vendor/pyparsing_extension.py":3084
 *         elif self.asKeyword:
 *             if (start > 0 and instring[start - 1] in bodychars
 *                     or loc < instrlen and instring[loc] in bodychars):             # <<<<<<<<<<<<<<
 *                 throwException = True
 * 
 */
    __pyx_t_5 = PyInt_FromSsize_t(__pyx_v_instrlen); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3084, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = PyObject_RichCompare(__pyx_v_loc, __pyx_t_5, Py_LT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3084, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 3084, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_10) {
    } else {
      __pyx_t_4 = __pyx_t_10;
      goto __pyx_L13_bool_binop_done;
    }
    __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_instring, __pyx_v_loc); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3084, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_10 = (__Pyx_PySequence_ContainsTF(__pyx_t_2, __pyx_v_bodychars, Py_EQ)); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 3084, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_3 = (__pyx_t_10 != 0);
    __pyx_t_4 = __pyx_t_3;
    __pyx_L13_bool_binop_done:;

    /* "Aspidites/_vendor/pyparsing_extension.py":3083
 *             throwException = True
 *         elif self.asKeyword:
 *             if (start > 0 and instring[start - 1] in bodychars             # <<<<<<<<<<<<<<
 *                     or loc < instrlen and instring[loc] in bodychars):
 *                 throwException = True
 */
    if (__pyx_t_4) {

      /* "Aspidites/_vendor/pyparsing_extension.py":3085
 *             if (start > 0 and instring[start - 1] in bodychars
 *                     or loc < instrlen and instring[loc] in bodychars):
 *                 throwException = True             # <<<<<<<<<<<<<<
 * 
 *         if throwException:
 */
      __pyx_v_throwException = 1;

      /* "Aspidites/_vendor/pyparsing_extension.py":3083
 *             throwException = True
 *         elif self.asKeyword:
 *             if (start > 0 and instring[start - 1] in bodychars             # <<<<<<<<<<<<<<
 *                     or loc < instrlen and instring[loc] in bodychars):
 *                 throwException = True
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":3082
 *         elif self.maxSpecified and loc < instrlen and instring[loc] in bodychars:
 *             throwException = True
 *         elif self.asKeyword:             # <<<<<<<<<<<<<<
 *             if (start > 0 and instring[start - 1] in bodychars
 *                     or loc < instrlen and instring[loc] in bodychars):
 */
  }
  __pyx_L8:;

  /* "Aspidites/_vendor/pyparsing_extension.py":3087
 *                 throwException = True
 * 
 *         if throwException:             # <<<<<<<<<<<<<<
 *             raise ParseException(instring, loc, self.errmsg, self)
 * 
 */
  __pyx_t_4 = (__pyx_v_throwException != 0);
  if (unlikely(__pyx_t_4)) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3088
 * 
 *         if throwException:
 *             raise ParseException(instring, loc, self.errmsg, self)             # <<<<<<<<<<<<<<
 * 
 *         return loc, instring[start:loc]
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3088, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errmsg); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3088, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[5] = {__pyx_t_8, __pyx_v_instring, __pyx_v_loc, __pyx_t_1, __pyx_v_self};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3088, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[5] = {__pyx_t_8, __pyx_v_instring, __pyx_v_loc, __pyx_t_1, __pyx_v_self};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3088, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(4+__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3088, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_INCREF(__pyx_v_instring);
      __Pyx_GIVEREF(__pyx_v_instring);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_7, __pyx_v_instring);
      __Pyx_INCREF(__pyx_v_loc);
      __Pyx_GIVEREF(__pyx_v_loc);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_7, __pyx_v_loc);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_7, __pyx_t_1);
      __Pyx_INCREF(__pyx_v_self);
      __Pyx_GIVEREF(__pyx_v_self);
      PyTuple_SET_ITEM(__pyx_t_6, 3+__pyx_t_7, __pyx_v_self);
      __pyx_t_1 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3088, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 3088, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":3087
 *                 throwException = True
 * 
 *         if throwException:             # <<<<<<<<<<<<<<
 *             raise ParseException(instring, loc, self.errmsg, self)
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3090
 *             raise ParseException(instring, loc, self.errmsg, self)
 * 
 *         return loc, instring[start:loc]             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetSlice(__pyx_v_instring, 0, 0, &__pyx_v_start, &__pyx_v_loc, NULL, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3090, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3090, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_loc);
  __Pyx_GIVEREF(__pyx_v_loc);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_loc);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3064
 *                 self.__class__ = _WordRegex
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if instring[loc] not in self.initChars:
 *             raise ParseException(instring, loc, self.errmsg, self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Word.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_start);
  __Pyx_XDECREF(__pyx_v_bodychars);
  __Pyx_XDECREF(__pyx_v_maxloc);
  __Pyx_XDECREF(__pyx_v_loc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3092
 *         return loc, instring[start:loc]
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         try:
 *             return super(Word, self).__str__()
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_4Word_5__str__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_4Word_5__str__ = {"__str__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_4Word_5__str__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_4Word_5__str__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Word_4__str__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3100
 *         if self.strRepr is None:
 * 
 *             def charsAsStr(s):             # <<<<<<<<<<<<<<
 *                 if len(s) > 4:
 *                     return s[:4] + "..."
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_4Word_7__str___1charsAsStr(PyObject *__pyx_self, PyObject *__pyx_v_s); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_4Word_7__str___1charsAsStr = {"charsAsStr", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_4Word_7__str___1charsAsStr, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_4Word_7__str___1charsAsStr(PyObject *__pyx_self, PyObject *__pyx_v_s) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("charsAsStr (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Word_7__str___charsAsStr(__pyx_self, ((PyObject *)__pyx_v_s));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Word_7__str___charsAsStr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_s) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("charsAsStr", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":3101
 * 
 *             def charsAsStr(s):
 *                 if len(s) > 4:             # <<<<<<<<<<<<<<
 *                     return s[:4] + "..."
 *                 else:
 */
  __pyx_t_1 = PyObject_Length(__pyx_v_s); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 3101, __pyx_L1_error)
  __pyx_t_2 = ((__pyx_t_1 > 4) != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3102
 *             def charsAsStr(s):
 *                 if len(s) > 4:
 *                     return s[:4] + "..."             # <<<<<<<<<<<<<<
 *                 else:
 *                     return s
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_v_s, 0, 4, NULL, NULL, &__pyx_slice__84, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyNumber_Add(__pyx_t_3, __pyx_kp_s__85); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3101
 * 
 *             def charsAsStr(s):
 *                 if len(s) > 4:             # <<<<<<<<<<<<<<
 *                     return s[:4] + "..."
 *                 else:
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3104
 *                     return s[:4] + "..."
 *                 else:
 *                     return s             # <<<<<<<<<<<<<<
 * 
 *             if self.initCharsOrig != self.bodyCharsOrig:
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_s);
    __pyx_r = __pyx_v_s;
    goto __pyx_L0;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3100
 *         if self.strRepr is None:
 * 
 *             def charsAsStr(s):             # <<<<<<<<<<<<<<
 *                 if len(s) > 4:
 *                     return s[:4] + "..."
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Word.__str__.charsAsStr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3092
 *         return loc, instring[start:loc]
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         try:
 *             return super(Word, self).__str__()
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Word_4__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_charsAsStr = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":3093
 * 
 *     def __str__(self):
 *         try:             # <<<<<<<<<<<<<<
 *             return super(Word, self).__str__()
 *         except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "Aspidites/_vendor/pyparsing_extension.py":3094
 *     def __str__(self):
 *         try:
 *             return super(Word, self).__str__()             # <<<<<<<<<<<<<<
 *         except Exception:
 *             pass
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_Word); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3094, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3094, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5);
      __Pyx_INCREF(__pyx_v_self);
      __Pyx_GIVEREF(__pyx_v_self);
      PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_v_self);
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_6, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3094, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_str); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3094, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3094, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_r = __pyx_t_4;
      __pyx_t_4 = 0;
      goto __pyx_L7_try_return;

      /* "Aspidites/_vendor/pyparsing_extension.py":3093
 * 
 *     def __str__(self):
 *         try:             # <<<<<<<<<<<<<<
 *             return super(Word, self).__str__()
 *         except Exception:
 */
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3095
 *         try:
 *             return super(Word, self).__str__()
 *         except Exception:             # <<<<<<<<<<<<<<
 *             pass
 * 
 */
    __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_7) {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "Aspidites/_vendor/pyparsing_extension.py":3093
 * 
 *     def __str__(self):
 *         try:             # <<<<<<<<<<<<<<
 *             return super(Word, self).__str__()
 *         except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3098
 *             pass
 * 
 *         if self.strRepr is None:             # <<<<<<<<<<<<<<
 * 
 *             def charsAsStr(s):
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_strRepr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3098, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_8 = (__pyx_t_4 == Py_None);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_9 = (__pyx_t_8 != 0);
  if (__pyx_t_9) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3100
 *         if self.strRepr is None:
 * 
 *             def charsAsStr(s):             # <<<<<<<<<<<<<<
 *                 if len(s) > 4:
 *                     return s[:4] + "..."
 */
    __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_4Word_7__str___1charsAsStr, 0, __pyx_n_s_Word___str___locals_charsAsStr, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__87)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3100, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_v_charsAsStr = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3106
 *                     return s
 * 
 *             if self.initCharsOrig != self.bodyCharsOrig:             # <<<<<<<<<<<<<<
 *                 self.strRepr = "W:(%s, %s)" % (charsAsStr(self.initCharsOrig), charsAsStr(self.bodyCharsOrig))
 *             else:
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_initCharsOrig); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_bodyCharsOrig); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = PyObject_RichCompare(__pyx_t_4, __pyx_t_6, Py_NE); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3106, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 3106, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_9) {

      /* "Aspidites/_vendor/pyparsing_extension.py":3107
 * 
 *             if self.initCharsOrig != self.bodyCharsOrig:
 *                 self.strRepr = "W:(%s, %s)" % (charsAsStr(self.initCharsOrig), charsAsStr(self.bodyCharsOrig))             # <<<<<<<<<<<<<<
 *             else:
 *                 self.strRepr = "W:(%s)" % charsAsStr(self.initCharsOrig)
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_initCharsOrig); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3107, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Word_7__str___charsAsStr(__pyx_v_charsAsStr, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3107, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_bodyCharsOrig); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3107, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Word_7__str___charsAsStr(__pyx_v_charsAsStr, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3107, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3107, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
      __pyx_t_6 = 0;
      __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_W_s_s, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3107, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_strRepr, __pyx_t_4) < 0) __PYX_ERR(0, 3107, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":3106
 *                     return s
 * 
 *             if self.initCharsOrig != self.bodyCharsOrig:             # <<<<<<<<<<<<<<
 *                 self.strRepr = "W:(%s, %s)" % (charsAsStr(self.initCharsOrig), charsAsStr(self.bodyCharsOrig))
 *             else:
 */
      goto __pyx_L10;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":3109
 *                 self.strRepr = "W:(%s, %s)" % (charsAsStr(self.initCharsOrig), charsAsStr(self.bodyCharsOrig))
 *             else:
 *                 self.strRepr = "W:(%s)" % charsAsStr(self.initCharsOrig)             # <<<<<<<<<<<<<<
 * 
 *         return self.strRepr
 */
    /*else*/ {
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_initCharsOrig); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3109, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Word_7__str___charsAsStr(__pyx_v_charsAsStr, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3109, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyString_FormatSafe(__pyx_kp_s_W_s, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3109, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_strRepr, __pyx_t_4) < 0) __PYX_ERR(0, 3109, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __pyx_L10:;

    /* "Aspidites/_vendor/pyparsing_extension.py":3098
 *             pass
 * 
 *         if self.strRepr is None:             # <<<<<<<<<<<<<<
 * 
 *             def charsAsStr(s):
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3111
 *                 self.strRepr = "W:(%s)" % charsAsStr(self.initCharsOrig)
 * 
 *         return self.strRepr             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_strRepr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3092
 *         return loc, instring[start:loc]
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         try:
 *             return super(Word, self).__str__()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Word.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_charsAsStr);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3115
 * 
 * class _WordRegex(Word):
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         result = self.re_match(instring, loc)
 *         if not result:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_WordRegex_1parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10_WordRegex_1parseImpl = {"parseImpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_WordRegex_1parseImpl, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_WordRegex_1parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  CYTHON_UNUSED PyObject *__pyx_v_doActions = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parseImpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_doActions,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 1); __PYX_ERR(0, 3115, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 2); __PYX_ERR(0, 3115, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_doActions);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "parseImpl") < 0)) __PYX_ERR(0, 3115, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_doActions = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3115, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._WordRegex.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_WordRegex_parseImpl(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_WordRegex_parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions) {
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parseImpl", 0);
  __Pyx_INCREF(__pyx_v_loc);

  /* "Aspidites/_vendor/pyparsing_extension.py":3116
 * class _WordRegex(Word):
 *     def parseImpl(self, instring, loc, doActions=True):
 *         result = self.re_match(instring, loc)             # <<<<<<<<<<<<<<
 *         if not result:
 *             raise ParseException(instring, loc, self.errmsg, self)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_re_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_instring, __pyx_v_loc};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3116, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_instring, __pyx_v_loc};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3116, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3116, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_instring);
    __Pyx_GIVEREF(__pyx_v_instring);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_instring);
    __Pyx_INCREF(__pyx_v_loc);
    __Pyx_GIVEREF(__pyx_v_loc);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_loc);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3116, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3117
 *     def parseImpl(self, instring, loc, doActions=True):
 *         result = self.re_match(instring, loc)
 *         if not result:             # <<<<<<<<<<<<<<
 *             raise ParseException(instring, loc, self.errmsg, self)
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_result); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 3117, __pyx_L1_error)
  __pyx_t_7 = ((!__pyx_t_6) != 0);
  if (unlikely(__pyx_t_7)) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3118
 *         result = self.re_match(instring, loc)
 *         if not result:
 *             raise ParseException(instring, loc, self.errmsg, self)             # <<<<<<<<<<<<<<
 * 
 *         loc = result.end()
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3118, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errmsg); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3118, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = NULL;
    __pyx_t_4 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[5] = {__pyx_t_3, __pyx_v_instring, __pyx_v_loc, __pyx_t_5, __pyx_v_self};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 4+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3118, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[5] = {__pyx_t_3, __pyx_v_instring, __pyx_v_loc, __pyx_t_5, __pyx_v_self};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 4+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3118, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(4+__pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3118, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_3) {
        __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_3); __pyx_t_3 = NULL;
      }
      __Pyx_INCREF(__pyx_v_instring);
      __Pyx_GIVEREF(__pyx_v_instring);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_4, __pyx_v_instring);
      __Pyx_INCREF(__pyx_v_loc);
      __Pyx_GIVEREF(__pyx_v_loc);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_4, __pyx_v_loc);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_4, __pyx_t_5);
      __Pyx_INCREF(__pyx_v_self);
      __Pyx_GIVEREF(__pyx_v_self);
      PyTuple_SET_ITEM(__pyx_t_8, 3+__pyx_t_4, __pyx_v_self);
      __pyx_t_5 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3118, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 3118, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":3117
 *     def parseImpl(self, instring, loc, doActions=True):
 *         result = self.re_match(instring, loc)
 *         if not result:             # <<<<<<<<<<<<<<
 *             raise ParseException(instring, loc, self.errmsg, self)
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3120
 *             raise ParseException(instring, loc, self.errmsg, self)
 * 
 *         loc = result.end()             # <<<<<<<<<<<<<<
 *         return loc, result.group()
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_result, __pyx_n_s_end); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_loc, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3121
 * 
 *         loc = result.end()
 *         return loc, result.group()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_result, __pyx_n_s_group); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_loc);
  __Pyx_GIVEREF(__pyx_v_loc);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_loc);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3115
 * 
 * class _WordRegex(Word):
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         result = self.re_match(instring, loc)
 *         if not result:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._WordRegex.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_loc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3130
 *     """
 * 
 *     def __init__(self, charset, asKeyword=False, excludeChars=None):             # <<<<<<<<<<<<<<
 *         super(Char, self).__init__(charset, exact=1, asKeyword=asKeyword, excludeChars=excludeChars)
 *         self.reString = "[%s]" % _escapeRegexRangeChars(''.join(self.initChars))
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_4Char_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_4Char_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_4Char_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_4Char_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_charset = 0;
  PyObject *__pyx_v_asKeyword = 0;
  PyObject *__pyx_v_excludeChars = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_charset,&__pyx_n_s_asKeyword,&__pyx_n_s_excludeChars,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_False));
    values[3] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_charset)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 3130, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_asKeyword);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_excludeChars);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 3130, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_charset = values[1];
    __pyx_v_asKeyword = values[2];
    __pyx_v_excludeChars = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3130, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Char.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Char___init__(__pyx_self, __pyx_v_self, __pyx_v_charset, __pyx_v_asKeyword, __pyx_v_excludeChars);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Char___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_charset, PyObject *__pyx_v_asKeyword, PyObject *__pyx_v_excludeChars) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":3131
 * 
 *     def __init__(self, charset, asKeyword=False, excludeChars=None):
 *         super(Char, self).__init__(charset, exact=1, asKeyword=asKeyword, excludeChars=excludeChars)             # <<<<<<<<<<<<<<
 *         self.reString = "[%s]" % _escapeRegexRangeChars(''.join(self.initChars))
 *         if asKeyword:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Char); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_self);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_charset);
  __Pyx_GIVEREF(__pyx_v_charset);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_charset);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_exact, __pyx_int_1) < 0) __PYX_ERR(0, 3131, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_asKeyword, __pyx_v_asKeyword) < 0) __PYX_ERR(0, 3131, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_excludeChars, __pyx_v_excludeChars) < 0) __PYX_ERR(0, 3131, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3132
 *     def __init__(self, charset, asKeyword=False, excludeChars=None):
 *         super(Char, self).__init__(charset, exact=1, asKeyword=asKeyword, excludeChars=excludeChars)
 *         self.reString = "[%s]" % _escapeRegexRangeChars(''.join(self.initChars))             # <<<<<<<<<<<<<<
 *         if asKeyword:
 *             self.reString = r"\b%s\b" % self.reString
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_escapeRegexRangeChars); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_initChars); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyString_Join(__pyx_kp_s__12, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_4 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_1, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyString_FormatSafe(__pyx_kp_s_s_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_reString, __pyx_t_3) < 0) __PYX_ERR(0, 3132, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3133
 *         super(Char, self).__init__(charset, exact=1, asKeyword=asKeyword, excludeChars=excludeChars)
 *         self.reString = "[%s]" % _escapeRegexRangeChars(''.join(self.initChars))
 *         if asKeyword:             # <<<<<<<<<<<<<<
 *             self.reString = r"\b%s\b" % self.reString
 *         self.re = re.compile(self.reString)
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_asKeyword); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 3133, __pyx_L1_error)
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3134
 *         self.reString = "[%s]" % _escapeRegexRangeChars(''.join(self.initChars))
 *         if asKeyword:
 *             self.reString = r"\b%s\b" % self.reString             # <<<<<<<<<<<<<<
 *         self.re = re.compile(self.reString)
 *         self.re_match = self.re.match
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_reString); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyString_FormatSafe(__pyx_kp_s_b_s_b, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_reString, __pyx_t_4) < 0) __PYX_ERR(0, 3134, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3133
 *         super(Char, self).__init__(charset, exact=1, asKeyword=asKeyword, excludeChars=excludeChars)
 *         self.reString = "[%s]" % _escapeRegexRangeChars(''.join(self.initChars))
 *         if asKeyword:             # <<<<<<<<<<<<<<
 *             self.reString = r"\b%s\b" % self.reString
 *         self.re = re.compile(self.reString)
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3135
 *         if asKeyword:
 *             self.reString = r"\b%s\b" % self.reString
 *         self.re = re.compile(self.reString)             # <<<<<<<<<<<<<<
 *         self.re_match = self.re.match
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_re); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_compile); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_reString); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_4 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_1, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_re, __pyx_t_4) < 0) __PYX_ERR(0, 3135, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3136
 *             self.reString = r"\b%s\b" % self.reString
 *         self.re = re.compile(self.reString)
 *         self.re_match = self.re.match             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_re); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_re_match, __pyx_t_2) < 0) __PYX_ERR(0, 3136, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3130
 *     """
 * 
 *     def __init__(self, charset, asKeyword=False, excludeChars=None):             # <<<<<<<<<<<<<<
 *         super(Char, self).__init__(charset, exact=1, asKeyword=asKeyword, excludeChars=excludeChars)
 *         self.reString = "[%s]" % _escapeRegexRangeChars(''.join(self.initChars))
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Char.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3164
 *     """
 * 
 *     def __init__(self, pattern, flags=0, asGroupList=False, asMatch=False):             # <<<<<<<<<<<<<<
 *         """The parameters ``pattern`` and ``flags`` are passed
 *         to the ``re.compile()`` function as-is. See the Python
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5Regex_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_5Regex___init__[] = "The parameters ``pattern`` and ``flags`` are passed\n        to the ``re.compile()`` function as-is. See the Python\n        `re module <https://docs.python.org/3/library/re.html>`_ module for an\n        explanation of the acceptable patterns and flags.\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_5Regex_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5Regex_1__init__, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_5Regex___init__};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5Regex_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_pattern = 0;
  PyObject *__pyx_v_flags = 0;
  PyObject *__pyx_v_asGroupList = 0;
  PyObject *__pyx_v_asMatch = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_pattern,&__pyx_n_s_flags,&__pyx_n_s_asGroupList,&__pyx_n_s_asMatch,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)__pyx_int_0));
    values[3] = ((PyObject *)((PyObject *)Py_False));
    values[4] = ((PyObject *)((PyObject *)Py_False));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pattern)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 5, 1); __PYX_ERR(0, 3164, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_flags);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_asGroupList);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_asMatch);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 3164, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_pattern = values[1];
    __pyx_v_flags = values[2];
    __pyx_v_asGroupList = values[3];
    __pyx_v_asMatch = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3164, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Regex.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5Regex___init__(__pyx_self, __pyx_v_self, __pyx_v_pattern, __pyx_v_flags, __pyx_v_asGroupList, __pyx_v_asMatch);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5Regex___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_pattern, PyObject *__pyx_v_flags, PyObject *__pyx_v_asGroupList, PyObject *__pyx_v_asMatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":3170
 *         explanation of the acceptable patterns and flags.
 *         """
 *         super(Regex, self).__init__()             # <<<<<<<<<<<<<<
 * 
 *         if isinstance(pattern, str):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Regex); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3172
 *         super(Regex, self).__init__()
 * 
 *         if isinstance(pattern, str):             # <<<<<<<<<<<<<<
 *             if not pattern:
 *                 warnings.warn("null string passed to Regex; use Empty() instead",
 */
  __pyx_t_4 = PyString_Check(__pyx_v_pattern); 
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3173
 * 
 *         if isinstance(pattern, str):
 *             if not pattern:             # <<<<<<<<<<<<<<
 *                 warnings.warn("null string passed to Regex; use Empty() instead",
 *                               SyntaxWarning, stacklevel=2)
 */
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_pattern); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 3173, __pyx_L1_error)
    __pyx_t_4 = ((!__pyx_t_5) != 0);
    if (__pyx_t_4) {

      /* "Aspidites/_vendor/pyparsing_extension.py":3174
 *         if isinstance(pattern, str):
 *             if not pattern:
 *                 warnings.warn("null string passed to Regex; use Empty() instead",             # <<<<<<<<<<<<<<
 *                               SyntaxWarning, stacklevel=2)
 * 
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_warnings); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3174, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_warn); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3174, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":3175
 *             if not pattern:
 *                 warnings.warn("null string passed to Regex; use Empty() instead",
 *                               SyntaxWarning, stacklevel=2)             # <<<<<<<<<<<<<<
 * 
 *             self.pattern = pattern
 */
      __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3175, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_stacklevel, __pyx_int_2) < 0) __PYX_ERR(0, 3175, __pyx_L1_error)

      /* "Aspidites/_vendor/pyparsing_extension.py":3174
 *         if isinstance(pattern, str):
 *             if not pattern:
 *                 warnings.warn("null string passed to Regex; use Empty() instead",             # <<<<<<<<<<<<<<
 *                               SyntaxWarning, stacklevel=2)
 * 
 */
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__88, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3174, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":3173
 * 
 *         if isinstance(pattern, str):
 *             if not pattern:             # <<<<<<<<<<<<<<
 *                 warnings.warn("null string passed to Regex; use Empty() instead",
 *                               SyntaxWarning, stacklevel=2)
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":3177
 *                               SyntaxWarning, stacklevel=2)
 * 
 *             self.pattern = pattern             # <<<<<<<<<<<<<<
 *             self.flags = flags
 * 
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_pattern, __pyx_v_pattern) < 0) __PYX_ERR(0, 3177, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":3178
 * 
 *             self.pattern = pattern
 *             self.flags = flags             # <<<<<<<<<<<<<<
 * 
 *             try:
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_flags, __pyx_v_flags) < 0) __PYX_ERR(0, 3178, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":3180
 *             self.flags = flags
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 self.re = re.compile(self.pattern, self.flags)
 *                 self.reString = self.pattern
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      /*try:*/ {

        /* "Aspidites/_vendor/pyparsing_extension.py":3181
 * 
 *             try:
 *                 self.re = re.compile(self.pattern, self.flags)             # <<<<<<<<<<<<<<
 *                 self.reString = self.pattern
 *             except sre_constants.error:
 */
        __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_re); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3181, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_compile); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3181, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_pattern); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3181, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_flags); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3181, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_10 = NULL;
        __pyx_t_11 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_10)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_10);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_11 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_1, __pyx_t_9};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3181, __pyx_L5_error)
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_1, __pyx_t_9};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3181, __pyx_L5_error)
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        } else
        #endif
        {
          __pyx_t_12 = PyTuple_New(2+__pyx_t_11); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 3181, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_12);
          if (__pyx_t_10) {
            __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_10); __pyx_t_10 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_1);
          PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_11, __pyx_t_1);
          __Pyx_GIVEREF(__pyx_t_9);
          PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_11, __pyx_t_9);
          __pyx_t_1 = 0;
          __pyx_t_9 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_12, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3181, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_re, __pyx_t_2) < 0) __PYX_ERR(0, 3181, __pyx_L5_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":3182
 *             try:
 *                 self.re = re.compile(self.pattern, self.flags)
 *                 self.reString = self.pattern             # <<<<<<<<<<<<<<
 *             except sre_constants.error:
 *                 warnings.warn("invalid pattern (%s) passed to Regex" % pattern,
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_pattern); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3182, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_2);
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_reString, __pyx_t_2) < 0) __PYX_ERR(0, 3182, __pyx_L5_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":3180
 *             self.flags = flags
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 self.re = re.compile(self.pattern, self.flags)
 *                 self.reString = self.pattern
 */
      }
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L10_try_end;
      __pyx_L5_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":3183
 *                 self.re = re.compile(self.pattern, self.flags)
 *                 self.reString = self.pattern
 *             except sre_constants.error:             # <<<<<<<<<<<<<<
 *                 warnings.warn("invalid pattern (%s) passed to Regex" % pattern,
 *                               SyntaxWarning, stacklevel=2)
 */
      __Pyx_ErrFetch(&__pyx_t_2, &__pyx_t_3, &__pyx_t_12);
      __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_sre_constants); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3183, __pyx_L7_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_error); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3183, __pyx_L7_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_11 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_2, __pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_ErrRestore(__pyx_t_2, __pyx_t_3, __pyx_t_12);
      __pyx_t_2 = 0; __pyx_t_3 = 0; __pyx_t_12 = 0;
      if (__pyx_t_11) {
        __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Regex.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_12, &__pyx_t_3, &__pyx_t_2) < 0) __PYX_ERR(0, 3183, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_2);

        /* "Aspidites/_vendor/pyparsing_extension.py":3184
 *                 self.reString = self.pattern
 *             except sre_constants.error:
 *                 warnings.warn("invalid pattern (%s) passed to Regex" % pattern,             # <<<<<<<<<<<<<<
 *                               SyntaxWarning, stacklevel=2)
 *                 raise
 */
        __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_warnings); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3184, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_warn); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3184, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyString_FormatSafe(__pyx_kp_s_invalid_pattern_s_passed_to_Rege, __pyx_v_pattern); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3184, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_1);

        /* "Aspidites/_vendor/pyparsing_extension.py":3185
 *             except sre_constants.error:
 *                 warnings.warn("invalid pattern (%s) passed to Regex" % pattern,
 *                               SyntaxWarning, stacklevel=2)             # <<<<<<<<<<<<<<
 *                 raise
 * 
 */
        __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 3184, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_GIVEREF(__pyx_t_1);
        PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_1);
        __Pyx_INCREF(__pyx_builtin_SyntaxWarning);
        __Pyx_GIVEREF(__pyx_builtin_SyntaxWarning);
        PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_builtin_SyntaxWarning);
        __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3185, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_stacklevel, __pyx_int_2) < 0) __PYX_ERR(0, 3185, __pyx_L7_except_error)

        /* "Aspidites/_vendor/pyparsing_extension.py":3184
 *                 self.reString = self.pattern
 *             except sre_constants.error:
 *                 warnings.warn("invalid pattern (%s) passed to Regex" % pattern,             # <<<<<<<<<<<<<<
 *                               SyntaxWarning, stacklevel=2)
 *                 raise
 */
        __pyx_t_13 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_10, __pyx_t_1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3184, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":3186
 *                 warnings.warn("invalid pattern (%s) passed to Regex" % pattern,
 *                               SyntaxWarning, stacklevel=2)
 *                 raise             # <<<<<<<<<<<<<<
 * 
 *         elif hasattr(pattern, 'pattern') and hasattr(pattern, 'match'):
 */
        __Pyx_GIVEREF(__pyx_t_12);
        __Pyx_GIVEREF(__pyx_t_3);
        __Pyx_XGIVEREF(__pyx_t_2);
        __Pyx_ErrRestoreWithState(__pyx_t_12, __pyx_t_3, __pyx_t_2);
        __pyx_t_12 = 0; __pyx_t_3 = 0; __pyx_t_2 = 0; 
        __PYX_ERR(0, 3186, __pyx_L7_except_error)
      }
      goto __pyx_L7_except_error;
      __pyx_L7_except_error:;

      /* "Aspidites/_vendor/pyparsing_extension.py":3180
 *             self.flags = flags
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 self.re = re.compile(self.pattern, self.flags)
 *                 self.reString = self.pattern
 */
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
      goto __pyx_L1_error;
      __pyx_L10_try_end:;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":3172
 *         super(Regex, self).__init__()
 * 
 *         if isinstance(pattern, str):             # <<<<<<<<<<<<<<
 *             if not pattern:
 *                 warnings.warn("null string passed to Regex; use Empty() instead",
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3188
 *                 raise
 * 
 *         elif hasattr(pattern, 'pattern') and hasattr(pattern, 'match'):             # <<<<<<<<<<<<<<
 *             self.re = pattern
 *             self.pattern = self.reString = pattern.pattern
 */
  __pyx_t_5 = __Pyx_HasAttr(__pyx_v_pattern, __pyx_n_s_pattern); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 3188, __pyx_L1_error)
  __pyx_t_14 = (__pyx_t_5 != 0);
  if (__pyx_t_14) {
  } else {
    __pyx_t_4 = __pyx_t_14;
    goto __pyx_L13_bool_binop_done;
  }
  __pyx_t_14 = __Pyx_HasAttr(__pyx_v_pattern, __pyx_n_s_match); if (unlikely(__pyx_t_14 == ((int)-1))) __PYX_ERR(0, 3188, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_14 != 0);
  __pyx_t_4 = __pyx_t_5;
  __pyx_L13_bool_binop_done:;
  if (likely(__pyx_t_4)) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3189
 * 
 *         elif hasattr(pattern, 'pattern') and hasattr(pattern, 'match'):
 *             self.re = pattern             # <<<<<<<<<<<<<<
 *             self.pattern = self.reString = pattern.pattern
 *             self.flags = flags
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_re, __pyx_v_pattern) < 0) __PYX_ERR(0, 3189, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":3190
 *         elif hasattr(pattern, 'pattern') and hasattr(pattern, 'match'):
 *             self.re = pattern
 *             self.pattern = self.reString = pattern.pattern             # <<<<<<<<<<<<<<
 *             self.flags = flags
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_pattern, __pyx_n_s_pattern); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_pattern, __pyx_t_2) < 0) __PYX_ERR(0, 3190, __pyx_L1_error)
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_reString, __pyx_t_2) < 0) __PYX_ERR(0, 3190, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3191
 *             self.re = pattern
 *             self.pattern = self.reString = pattern.pattern
 *             self.flags = flags             # <<<<<<<<<<<<<<
 * 
 *         else:
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_flags, __pyx_v_flags) < 0) __PYX_ERR(0, 3191, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":3188
 *                 raise
 * 
 *         elif hasattr(pattern, 'pattern') and hasattr(pattern, 'match'):             # <<<<<<<<<<<<<<
 *             self.re = pattern
 *             self.pattern = self.reString = pattern.pattern
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3194
 * 
 *         else:
 *             raise TypeError("Regex may only be constructed with a string or a compiled RE object")             # <<<<<<<<<<<<<<
 * 
 *         self.re_match = self.re.match
 */
  /*else*/ {
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__89, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3194, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 3194, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "Aspidites/_vendor/pyparsing_extension.py":3196
 *             raise TypeError("Regex may only be constructed with a string or a compiled RE object")
 * 
 *         self.re_match = self.re.match             # <<<<<<<<<<<<<<
 * 
 *         self.name = str(self)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_re); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_re_match, __pyx_t_3) < 0) __PYX_ERR(0, 3196, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3198
 *         self.re_match = self.re.match
 * 
 *         self.name = str(self)             # <<<<<<<<<<<<<<
 *         self.errmsg = "Expected " + self.name
 *         self.mayIndexError = False
 */
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_name, __pyx_t_3) < 0) __PYX_ERR(0, 3198, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3199
 * 
 *         self.name = str(self)
 *         self.errmsg = "Expected " + self.name             # <<<<<<<<<<<<<<
 *         self.mayIndexError = False
 *         self.mayReturnEmpty = self.re_match("") is not None
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyNumber_Add(__pyx_kp_s_Expected, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_errmsg, __pyx_t_2) < 0) __PYX_ERR(0, 3199, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3200
 *         self.name = str(self)
 *         self.errmsg = "Expected " + self.name
 *         self.mayIndexError = False             # <<<<<<<<<<<<<<
 *         self.mayReturnEmpty = self.re_match("") is not None
 *         self.asGroupList = asGroupList
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_mayIndexError, Py_False) < 0) __PYX_ERR(0, 3200, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3201
 *         self.errmsg = "Expected " + self.name
 *         self.mayIndexError = False
 *         self.mayReturnEmpty = self.re_match("") is not None             # <<<<<<<<<<<<<<
 *         self.asGroupList = asGroupList
 *         self.asMatch = asMatch
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_re_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_12 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_12)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_12);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_12, __pyx_kp_s__12) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_kp_s__12);
  __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = (__pyx_t_2 != Py_None);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_mayReturnEmpty, __pyx_t_2) < 0) __PYX_ERR(0, 3201, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3202
 *         self.mayIndexError = False
 *         self.mayReturnEmpty = self.re_match("") is not None
 *         self.asGroupList = asGroupList             # <<<<<<<<<<<<<<
 *         self.asMatch = asMatch
 *         if self.asGroupList:
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_asGroupList, __pyx_v_asGroupList) < 0) __PYX_ERR(0, 3202, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3203
 *         self.mayReturnEmpty = self.re_match("") is not None
 *         self.asGroupList = asGroupList
 *         self.asMatch = asMatch             # <<<<<<<<<<<<<<
 *         if self.asGroupList:
 *             self.parseImpl = self.parseImplAsGroupList
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_asMatch, __pyx_v_asMatch) < 0) __PYX_ERR(0, 3203, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3204
 *         self.asGroupList = asGroupList
 *         self.asMatch = asMatch
 *         if self.asGroupList:             # <<<<<<<<<<<<<<
 *             self.parseImpl = self.parseImplAsGroupList
 *         if self.asMatch:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_asGroupList); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 3204, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3205
 *         self.asMatch = asMatch
 *         if self.asGroupList:
 *             self.parseImpl = self.parseImplAsGroupList             # <<<<<<<<<<<<<<
 *         if self.asMatch:
 *             self.parseImpl = self.parseImplAsMatch
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_parseImplAsGroupList); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3205, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parseImpl, __pyx_t_2) < 0) __PYX_ERR(0, 3205, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3204
 *         self.asGroupList = asGroupList
 *         self.asMatch = asMatch
 *         if self.asGroupList:             # <<<<<<<<<<<<<<
 *             self.parseImpl = self.parseImplAsGroupList
 *         if self.asMatch:
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3206
 *         if self.asGroupList:
 *             self.parseImpl = self.parseImplAsGroupList
 *         if self.asMatch:             # <<<<<<<<<<<<<<
 *             self.parseImpl = self.parseImplAsMatch
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_asMatch); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 3206, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3207
 *             self.parseImpl = self.parseImplAsGroupList
 *         if self.asMatch:
 *             self.parseImpl = self.parseImplAsMatch             # <<<<<<<<<<<<<<
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_parseImplAsMatch); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3207, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_parseImpl, __pyx_t_2) < 0) __PYX_ERR(0, 3207, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3206
 *         if self.asGroupList:
 *             self.parseImpl = self.parseImplAsGroupList
 *         if self.asMatch:             # <<<<<<<<<<<<<<
 *             self.parseImpl = self.parseImplAsMatch
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3164
 *     """
 * 
 *     def __init__(self, pattern, flags=0, asGroupList=False, asMatch=False):             # <<<<<<<<<<<<<<
 *         """The parameters ``pattern`` and ``flags`` are passed
 *         to the ``re.compile()`` function as-is. See the Python
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Regex.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3209
 *             self.parseImpl = self.parseImplAsMatch
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         result = self.re_match(instring, loc)
 *         if not result:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5Regex_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_5Regex_3parseImpl = {"parseImpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5Regex_3parseImpl, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5Regex_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  CYTHON_UNUSED PyObject *__pyx_v_doActions = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parseImpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_doActions,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 1); __PYX_ERR(0, 3209, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 2); __PYX_ERR(0, 3209, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_doActions);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "parseImpl") < 0)) __PYX_ERR(0, 3209, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_doActions = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3209, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Regex.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5Regex_2parseImpl(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5Regex_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions) {
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_v_ret = NULL;
  PyObject *__pyx_v_d = NULL;
  PyObject *__pyx_v_k = NULL;
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  PyObject *(*__pyx_t_10)(PyObject *);
  PyObject *(*__pyx_t_11)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parseImpl", 0);
  __Pyx_INCREF(__pyx_v_loc);

  /* "Aspidites/_vendor/pyparsing_extension.py":3210
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         result = self.re_match(instring, loc)             # <<<<<<<<<<<<<<
 *         if not result:
 *             raise ParseException(instring, loc, self.errmsg, self)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_re_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_instring, __pyx_v_loc};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3210, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_instring, __pyx_v_loc};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3210, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3210, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_instring);
    __Pyx_GIVEREF(__pyx_v_instring);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_instring);
    __Pyx_INCREF(__pyx_v_loc);
    __Pyx_GIVEREF(__pyx_v_loc);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_loc);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3210, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3211
 *     def parseImpl(self, instring, loc, doActions=True):
 *         result = self.re_match(instring, loc)
 *         if not result:             # <<<<<<<<<<<<<<
 *             raise ParseException(instring, loc, self.errmsg, self)
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_result); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 3211, __pyx_L1_error)
  __pyx_t_7 = ((!__pyx_t_6) != 0);
  if (unlikely(__pyx_t_7)) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3212
 *         result = self.re_match(instring, loc)
 *         if not result:
 *             raise ParseException(instring, loc, self.errmsg, self)             # <<<<<<<<<<<<<<
 * 
 *         loc = result.end()
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3212, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errmsg); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3212, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = NULL;
    __pyx_t_4 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[5] = {__pyx_t_3, __pyx_v_instring, __pyx_v_loc, __pyx_t_5, __pyx_v_self};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 4+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3212, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[5] = {__pyx_t_3, __pyx_v_instring, __pyx_v_loc, __pyx_t_5, __pyx_v_self};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 4+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3212, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(4+__pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3212, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_3) {
        __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_3); __pyx_t_3 = NULL;
      }
      __Pyx_INCREF(__pyx_v_instring);
      __Pyx_GIVEREF(__pyx_v_instring);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_4, __pyx_v_instring);
      __Pyx_INCREF(__pyx_v_loc);
      __Pyx_GIVEREF(__pyx_v_loc);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_4, __pyx_v_loc);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_4, __pyx_t_5);
      __Pyx_INCREF(__pyx_v_self);
      __Pyx_GIVEREF(__pyx_v_self);
      PyTuple_SET_ITEM(__pyx_t_8, 3+__pyx_t_4, __pyx_v_self);
      __pyx_t_5 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3212, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 3212, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":3211
 *     def parseImpl(self, instring, loc, doActions=True):
 *         result = self.re_match(instring, loc)
 *         if not result:             # <<<<<<<<<<<<<<
 *             raise ParseException(instring, loc, self.errmsg, self)
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3214
 *             raise ParseException(instring, loc, self.errmsg, self)
 * 
 *         loc = result.end()             # <<<<<<<<<<<<<<
 *         ret = ParseResults(result.group())
 *         d = result.groupdict()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_result, __pyx_n_s_end); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_loc, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3215
 * 
 *         loc = result.end()
 *         ret = ParseResults(result.group())             # <<<<<<<<<<<<<<
 *         d = result.groupdict()
 *         if d:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParseResults); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_result, __pyx_n_s_group); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_8 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_8);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_ret = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3216
 *         loc = result.end()
 *         ret = ParseResults(result.group())
 *         d = result.groupdict()             # <<<<<<<<<<<<<<
 *         if d:
 *             for k, v in d.items():
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_result, __pyx_n_s_groupdict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_d = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3217
 *         ret = ParseResults(result.group())
 *         d = result.groupdict()
 *         if d:             # <<<<<<<<<<<<<<
 *             for k, v in d.items():
 *                 ret[k] = v
 */
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_d); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 3217, __pyx_L1_error)
  if (__pyx_t_7) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3218
 *         d = result.groupdict()
 *         if d:
 *             for k, v in d.items():             # <<<<<<<<<<<<<<
 *                 ret[k] = v
 *         return loc, ret
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_d, __pyx_n_s_items); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3218, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3218, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
      __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_9 = 0;
      __pyx_t_10 = NULL;
    } else {
      __pyx_t_9 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3218, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_10 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 3218, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    for (;;) {
      if (likely(!__pyx_t_10)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_9); __Pyx_INCREF(__pyx_t_1); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 3218, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3218, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        } else {
          if (__pyx_t_9 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_9); __Pyx_INCREF(__pyx_t_1); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 3218, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3218, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        }
      } else {
        __pyx_t_1 = __pyx_t_10(__pyx_t_2);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 3218, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_1);
      }
      if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
        PyObject* sequence = __pyx_t_1;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 3218, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_8 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_8 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_5);
        #else
        __pyx_t_8 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3218, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3218, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3218, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_11 = Py_TYPE(__pyx_t_3)->tp_iternext;
        index = 0; __pyx_t_8 = __pyx_t_11(__pyx_t_3); if (unlikely(!__pyx_t_8)) goto __pyx_L7_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_8);
        index = 1; __pyx_t_5 = __pyx_t_11(__pyx_t_3); if (unlikely(!__pyx_t_5)) goto __pyx_L7_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_5);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_3), 2) < 0) __PYX_ERR(0, 3218, __pyx_L1_error)
        __pyx_t_11 = NULL;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        goto __pyx_L8_unpacking_done;
        __pyx_L7_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_11 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 3218, __pyx_L1_error)
        __pyx_L8_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_k, __pyx_t_8);
      __pyx_t_8 = 0;
      __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":3219
 *         if d:
 *             for k, v in d.items():
 *                 ret[k] = v             # <<<<<<<<<<<<<<
 *         return loc, ret
 * 
 */
      if (unlikely(PyObject_SetItem(__pyx_v_ret, __pyx_v_k, __pyx_v_v) < 0)) __PYX_ERR(0, 3219, __pyx_L1_error)

      /* "Aspidites/_vendor/pyparsing_extension.py":3218
 *         d = result.groupdict()
 *         if d:
 *             for k, v in d.items():             # <<<<<<<<<<<<<<
 *                 ret[k] = v
 *         return loc, ret
 */
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3217
 *         ret = ParseResults(result.group())
 *         d = result.groupdict()
 *         if d:             # <<<<<<<<<<<<<<
 *             for k, v in d.items():
 *                 ret[k] = v
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3220
 *             for k, v in d.items():
 *                 ret[k] = v
 *         return loc, ret             # <<<<<<<<<<<<<<
 * 
 *     def parseImplAsGroupList(self, instring, loc, doActions=True):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_loc);
  __Pyx_GIVEREF(__pyx_v_loc);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_loc);
  __Pyx_INCREF(__pyx_v_ret);
  __Pyx_GIVEREF(__pyx_v_ret);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_ret);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3209
 *             self.parseImpl = self.parseImplAsMatch
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         result = self.re_match(instring, loc)
 *         if not result:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Regex.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_XDECREF(__pyx_v_d);
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_loc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3222
 *         return loc, ret
 * 
 *     def parseImplAsGroupList(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         result = self.re_match(instring, loc)
 *         if not result:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5Regex_5parseImplAsGroupList(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_5Regex_5parseImplAsGroupList = {"parseImplAsGroupList", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5Regex_5parseImplAsGroupList, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5Regex_5parseImplAsGroupList(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  CYTHON_UNUSED PyObject *__pyx_v_doActions = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parseImplAsGroupList (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_doActions,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImplAsGroupList", 0, 3, 4, 1); __PYX_ERR(0, 3222, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImplAsGroupList", 0, 3, 4, 2); __PYX_ERR(0, 3222, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_doActions);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "parseImplAsGroupList") < 0)) __PYX_ERR(0, 3222, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_doActions = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("parseImplAsGroupList", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3222, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Regex.parseImplAsGroupList", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5Regex_4parseImplAsGroupList(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5Regex_4parseImplAsGroupList(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions) {
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_v_ret = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parseImplAsGroupList", 0);
  __Pyx_INCREF(__pyx_v_loc);

  /* "Aspidites/_vendor/pyparsing_extension.py":3223
 * 
 *     def parseImplAsGroupList(self, instring, loc, doActions=True):
 *         result = self.re_match(instring, loc)             # <<<<<<<<<<<<<<
 *         if not result:
 *             raise ParseException(instring, loc, self.errmsg, self)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_re_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_instring, __pyx_v_loc};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3223, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_instring, __pyx_v_loc};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3223, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3223, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_instring);
    __Pyx_GIVEREF(__pyx_v_instring);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_instring);
    __Pyx_INCREF(__pyx_v_loc);
    __Pyx_GIVEREF(__pyx_v_loc);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_loc);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3223, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3224
 *     def parseImplAsGroupList(self, instring, loc, doActions=True):
 *         result = self.re_match(instring, loc)
 *         if not result:             # <<<<<<<<<<<<<<
 *             raise ParseException(instring, loc, self.errmsg, self)
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_result); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 3224, __pyx_L1_error)
  __pyx_t_7 = ((!__pyx_t_6) != 0);
  if (unlikely(__pyx_t_7)) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3225
 *         result = self.re_match(instring, loc)
 *         if not result:
 *             raise ParseException(instring, loc, self.errmsg, self)             # <<<<<<<<<<<<<<
 * 
 *         loc = result.end()
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3225, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errmsg); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3225, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = NULL;
    __pyx_t_4 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[5] = {__pyx_t_3, __pyx_v_instring, __pyx_v_loc, __pyx_t_5, __pyx_v_self};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 4+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3225, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[5] = {__pyx_t_3, __pyx_v_instring, __pyx_v_loc, __pyx_t_5, __pyx_v_self};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 4+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3225, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(4+__pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3225, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_3) {
        __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_3); __pyx_t_3 = NULL;
      }
      __Pyx_INCREF(__pyx_v_instring);
      __Pyx_GIVEREF(__pyx_v_instring);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_4, __pyx_v_instring);
      __Pyx_INCREF(__pyx_v_loc);
      __Pyx_GIVEREF(__pyx_v_loc);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_4, __pyx_v_loc);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_4, __pyx_t_5);
      __Pyx_INCREF(__pyx_v_self);
      __Pyx_GIVEREF(__pyx_v_self);
      PyTuple_SET_ITEM(__pyx_t_8, 3+__pyx_t_4, __pyx_v_self);
      __pyx_t_5 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3225, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 3225, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":3224
 *     def parseImplAsGroupList(self, instring, loc, doActions=True):
 *         result = self.re_match(instring, loc)
 *         if not result:             # <<<<<<<<<<<<<<
 *             raise ParseException(instring, loc, self.errmsg, self)
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3227
 *             raise ParseException(instring, loc, self.errmsg, self)
 * 
 *         loc = result.end()             # <<<<<<<<<<<<<<
 *         ret = result.groups()
 *         return loc, ret
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_result, __pyx_n_s_end); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_loc, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3228
 * 
 *         loc = result.end()
 *         ret = result.groups()             # <<<<<<<<<<<<<<
 *         return loc, ret
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_result, __pyx_n_s_groups); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_ret = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3229
 *         loc = result.end()
 *         ret = result.groups()
 *         return loc, ret             # <<<<<<<<<<<<<<
 * 
 *     def parseImplAsMatch(self, instring, loc, doActions=True):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_loc);
  __Pyx_GIVEREF(__pyx_v_loc);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_loc);
  __Pyx_INCREF(__pyx_v_ret);
  __Pyx_GIVEREF(__pyx_v_ret);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_ret);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3222
 *         return loc, ret
 * 
 *     def parseImplAsGroupList(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         result = self.re_match(instring, loc)
 *         if not result:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Regex.parseImplAsGroupList", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_XDECREF(__pyx_v_loc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3231
 *         return loc, ret
 * 
 *     def parseImplAsMatch(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         result = self.re_match(instring, loc)
 *         if not result:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5Regex_7parseImplAsMatch(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_5Regex_7parseImplAsMatch = {"parseImplAsMatch", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5Regex_7parseImplAsMatch, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5Regex_7parseImplAsMatch(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  CYTHON_UNUSED PyObject *__pyx_v_doActions = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parseImplAsMatch (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_doActions,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImplAsMatch", 0, 3, 4, 1); __PYX_ERR(0, 3231, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImplAsMatch", 0, 3, 4, 2); __PYX_ERR(0, 3231, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_doActions);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "parseImplAsMatch") < 0)) __PYX_ERR(0, 3231, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_doActions = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("parseImplAsMatch", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3231, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Regex.parseImplAsMatch", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5Regex_6parseImplAsMatch(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5Regex_6parseImplAsMatch(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions) {
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_v_ret = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parseImplAsMatch", 0);
  __Pyx_INCREF(__pyx_v_loc);

  /* "Aspidites/_vendor/pyparsing_extension.py":3232
 * 
 *     def parseImplAsMatch(self, instring, loc, doActions=True):
 *         result = self.re_match(instring, loc)             # <<<<<<<<<<<<<<
 *         if not result:
 *             raise ParseException(instring, loc, self.errmsg, self)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_re_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_instring, __pyx_v_loc};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3232, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_instring, __pyx_v_loc};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3232, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_instring);
    __Pyx_GIVEREF(__pyx_v_instring);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_instring);
    __Pyx_INCREF(__pyx_v_loc);
    __Pyx_GIVEREF(__pyx_v_loc);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_loc);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3233
 *     def parseImplAsMatch(self, instring, loc, doActions=True):
 *         result = self.re_match(instring, loc)
 *         if not result:             # <<<<<<<<<<<<<<
 *             raise ParseException(instring, loc, self.errmsg, self)
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_result); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 3233, __pyx_L1_error)
  __pyx_t_7 = ((!__pyx_t_6) != 0);
  if (unlikely(__pyx_t_7)) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3234
 *         result = self.re_match(instring, loc)
 *         if not result:
 *             raise ParseException(instring, loc, self.errmsg, self)             # <<<<<<<<<<<<<<
 * 
 *         loc = result.end()
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3234, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errmsg); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3234, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = NULL;
    __pyx_t_4 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[5] = {__pyx_t_3, __pyx_v_instring, __pyx_v_loc, __pyx_t_5, __pyx_v_self};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 4+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3234, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[5] = {__pyx_t_3, __pyx_v_instring, __pyx_v_loc, __pyx_t_5, __pyx_v_self};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 4+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3234, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(4+__pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3234, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_3) {
        __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_3); __pyx_t_3 = NULL;
      }
      __Pyx_INCREF(__pyx_v_instring);
      __Pyx_GIVEREF(__pyx_v_instring);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_4, __pyx_v_instring);
      __Pyx_INCREF(__pyx_v_loc);
      __Pyx_GIVEREF(__pyx_v_loc);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_4, __pyx_v_loc);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_4, __pyx_t_5);
      __Pyx_INCREF(__pyx_v_self);
      __Pyx_GIVEREF(__pyx_v_self);
      PyTuple_SET_ITEM(__pyx_t_8, 3+__pyx_t_4, __pyx_v_self);
      __pyx_t_5 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3234, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 3234, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":3233
 *     def parseImplAsMatch(self, instring, loc, doActions=True):
 *         result = self.re_match(instring, loc)
 *         if not result:             # <<<<<<<<<<<<<<
 *             raise ParseException(instring, loc, self.errmsg, self)
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3236
 *             raise ParseException(instring, loc, self.errmsg, self)
 * 
 *         loc = result.end()             # <<<<<<<<<<<<<<
 *         ret = result
 *         return loc, ret
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_result, __pyx_n_s_end); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_loc, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3237
 * 
 *         loc = result.end()
 *         ret = result             # <<<<<<<<<<<<<<
 *         return loc, ret
 * 
 */
  __Pyx_INCREF(__pyx_v_result);
  __pyx_v_ret = __pyx_v_result;

  /* "Aspidites/_vendor/pyparsing_extension.py":3238
 *         loc = result.end()
 *         ret = result
 *         return loc, ret             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_loc);
  __Pyx_GIVEREF(__pyx_v_loc);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_loc);
  __Pyx_INCREF(__pyx_v_ret);
  __Pyx_GIVEREF(__pyx_v_ret);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_ret);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3231
 *         return loc, ret
 * 
 *     def parseImplAsMatch(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         result = self.re_match(instring, loc)
 *         if not result:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Regex.parseImplAsMatch", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_XDECREF(__pyx_v_loc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3240
 *         return loc, ret
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         try:
 *             return super(Regex, self).__str__()
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5Regex_9__str__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_5Regex_9__str__ = {"__str__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5Regex_9__str__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5Regex_9__str__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5Regex_8__str__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5Regex_8__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":3241
 * 
 *     def __str__(self):
 *         try:             # <<<<<<<<<<<<<<
 *             return super(Regex, self).__str__()
 *         except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "Aspidites/_vendor/pyparsing_extension.py":3242
 *     def __str__(self):
 *         try:
 *             return super(Regex, self).__str__()             # <<<<<<<<<<<<<<
 *         except Exception:
 *             pass
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_Regex); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3242, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3242, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5);
      __Pyx_INCREF(__pyx_v_self);
      __Pyx_GIVEREF(__pyx_v_self);
      PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_v_self);
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_6, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3242, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_str); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3242, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3242, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_r = __pyx_t_4;
      __pyx_t_4 = 0;
      goto __pyx_L7_try_return;

      /* "Aspidites/_vendor/pyparsing_extension.py":3241
 * 
 *     def __str__(self):
 *         try:             # <<<<<<<<<<<<<<
 *             return super(Regex, self).__str__()
 *         except Exception:
 */
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3243
 *         try:
 *             return super(Regex, self).__str__()
 *         except Exception:             # <<<<<<<<<<<<<<
 *             pass
 * 
 */
    __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_7) {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "Aspidites/_vendor/pyparsing_extension.py":3241
 * 
 *     def __str__(self):
 *         try:             # <<<<<<<<<<<<<<
 *             return super(Regex, self).__str__()
 *         except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3246
 *             pass
 * 
 *         if self.strRepr is None:             # <<<<<<<<<<<<<<
 *             self.strRepr = "Re:(%s)" % repr(self.pattern)
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_strRepr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_8 = (__pyx_t_4 == Py_None);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_9 = (__pyx_t_8 != 0);
  if (__pyx_t_9) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3247
 * 
 *         if self.strRepr is None:
 *             self.strRepr = "Re:(%s)" % repr(self.pattern)             # <<<<<<<<<<<<<<
 * 
 *         return self.strRepr
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_pattern); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = PyObject_Repr(__pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Re_s, __pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_strRepr, __pyx_t_4) < 0) __PYX_ERR(0, 3247, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3246
 *             pass
 * 
 *         if self.strRepr is None:             # <<<<<<<<<<<<<<
 *             self.strRepr = "Re:(%s)" % repr(self.pattern)
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3249
 *             self.strRepr = "Re:(%s)" % repr(self.pattern)
 * 
 *         return self.strRepr             # <<<<<<<<<<<<<<
 * 
 *     def sub(self, repl):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_strRepr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3240
 *         return loc, ret
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         try:
 *             return super(Regex, self).__str__()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Regex.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3251
 *         return self.strRepr
 * 
 *     def sub(self, repl):             # <<<<<<<<<<<<<<
 *         r"""
 *         Return Regex with an attached parse action to transform the parsed
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5Regex_11sub(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_5Regex_10sub[] = "\n        Return Regex with an attached parse action to transform the parsed\n        result as if called using `re.sub(expr, repl, string) <https://docs.python.org/3/library/re.html#re.sub>`_.\n\n        Example::\n\n            make_html = Regex(r\"(\\w+):(.*?):\").sub(r\"<\\1>\\2</\\1>\")\n            print(make_html.transformString(\"h1:main title:\"))\n            # prints \"<h1>main title</h1>\"\n        ";
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_5Regex_11sub = {"sub", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5Regex_11sub, METH_VARARGS|METH_KEYWORDS, __pyx_doc_9Aspidites_7_vendor_19pyparsing_extension_5Regex_10sub};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5Regex_11sub(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_repl = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sub (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_repl,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_repl)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sub", 1, 2, 2, 1); __PYX_ERR(0, 3251, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "sub") < 0)) __PYX_ERR(0, 3251, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_repl = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sub", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3251, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Regex.sub", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5Regex_10sub(__pyx_self, __pyx_v_self, __pyx_v_repl);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3273
 * 
 *         if self.asMatch:
 *             def pa(tokens):             # <<<<<<<<<<<<<<
 *                 return tokens[0].expand(repl)
 *         else:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5Regex_3sub_1pa(PyObject *__pyx_self, PyObject *__pyx_v_tokens); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_5Regex_3sub_1pa = {"pa", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5Regex_3sub_1pa, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5Regex_3sub_1pa(PyObject *__pyx_self, PyObject *__pyx_v_tokens) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pa (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5Regex_3sub_pa(__pyx_self, ((PyObject *)__pyx_v_tokens));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5Regex_3sub_pa(PyObject *__pyx_self, PyObject *__pyx_v_tokens) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub *__pyx_cur_scope;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pa", 0);
  __pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "Aspidites/_vendor/pyparsing_extension.py":3274
 *         if self.asMatch:
 *             def pa(tokens):
 *                 return tokens[0].expand(repl)             # <<<<<<<<<<<<<<
 *         else:
 *             def pa(tokens):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_tokens, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_expand); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_cur_scope->__pyx_v_repl)) { __Pyx_RaiseClosureNameError("repl"); __PYX_ERR(0, 3274, __pyx_L1_error) }
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_cur_scope->__pyx_v_repl) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_cur_scope->__pyx_v_repl);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3273
 * 
 *         if self.asMatch:
 *             def pa(tokens):             # <<<<<<<<<<<<<<
 *                 return tokens[0].expand(repl)
 *         else:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Regex.sub.pa", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3276
 *                 return tokens[0].expand(repl)
 *         else:
 *             def pa(tokens):             # <<<<<<<<<<<<<<
 *                 return self.re.sub(repl, tokens[0])
 *         return self.addParseAction(pa)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5Regex_3sub_3pa(PyObject *__pyx_self, PyObject *__pyx_v_tokens); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_5Regex_3sub_3pa = {"pa", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5Regex_3sub_3pa, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5Regex_3sub_3pa(PyObject *__pyx_self, PyObject *__pyx_v_tokens) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pa (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5Regex_3sub_2pa(__pyx_self, ((PyObject *)__pyx_v_tokens));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5Regex_3sub_2pa(PyObject *__pyx_self, PyObject *__pyx_v_tokens) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub *__pyx_cur_scope;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pa", 0);
  __pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "Aspidites/_vendor/pyparsing_extension.py":3277
 *         else:
 *             def pa(tokens):
 *                 return self.re.sub(repl, tokens[0])             # <<<<<<<<<<<<<<
 *         return self.addParseAction(pa)
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 3277, __pyx_L1_error) }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_re); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_sub); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_cur_scope->__pyx_v_repl)) { __Pyx_RaiseClosureNameError("repl"); __PYX_ERR(0, 3277, __pyx_L1_error) }
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_tokens, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_cur_scope->__pyx_v_repl, __pyx_t_2};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3277, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_cur_scope->__pyx_v_repl, __pyx_t_2};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3277, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3277, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_repl);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_repl);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_cur_scope->__pyx_v_repl);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3277, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3276
 *                 return tokens[0].expand(repl)
 *         else:
 *             def pa(tokens):             # <<<<<<<<<<<<<<
 *                 return self.re.sub(repl, tokens[0])
 *         return self.addParseAction(pa)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Regex.sub.pa", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3251
 *         return self.strRepr
 * 
 *     def sub(self, repl):             # <<<<<<<<<<<<<<
 *         r"""
 *         Return Regex with an attached parse action to transform the parsed
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5Regex_10sub(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_repl) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub *__pyx_cur_scope;
  PyObject *__pyx_v_pa = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("sub", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 3251, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_repl = __pyx_v_repl;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_repl);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_repl);

  /* "Aspidites/_vendor/pyparsing_extension.py":3262
 *             # prints "<h1>main title</h1>"
 *         """
 *         if self.asGroupList:             # <<<<<<<<<<<<<<
 *             warnings.warn("cannot use sub() with Regex(asGroupList=True)",
 *                           SyntaxWarning, stacklevel=2)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_asGroupList); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 3262, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_2)) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3263
 *         """
 *         if self.asGroupList:
 *             warnings.warn("cannot use sub() with Regex(asGroupList=True)",             # <<<<<<<<<<<<<<
 *                           SyntaxWarning, stacklevel=2)
 *             raise SyntaxError()
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_warnings); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_warn); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3264
 *         if self.asGroupList:
 *             warnings.warn("cannot use sub() with Regex(asGroupList=True)",
 *                           SyntaxWarning, stacklevel=2)             # <<<<<<<<<<<<<<
 *             raise SyntaxError()
 * 
 */
    __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3264, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_stacklevel, __pyx_int_2) < 0) __PYX_ERR(0, 3264, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":3263
 *         """
 *         if self.asGroupList:
 *             warnings.warn("cannot use sub() with Regex(asGroupList=True)",             # <<<<<<<<<<<<<<
 *                           SyntaxWarning, stacklevel=2)
 *             raise SyntaxError()
 */
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__90, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3265
 *             warnings.warn("cannot use sub() with Regex(asGroupList=True)",
 *                           SyntaxWarning, stacklevel=2)
 *             raise SyntaxError()             # <<<<<<<<<<<<<<
 * 
 *         if self.asMatch and callable(repl):
 */
    __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_builtin_SyntaxError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3265, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 3265, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":3262
 *             # prints "<h1>main title</h1>"
 *         """
 *         if self.asGroupList:             # <<<<<<<<<<<<<<
 *             warnings.warn("cannot use sub() with Regex(asGroupList=True)",
 *                           SyntaxWarning, stacklevel=2)
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3267
 *             raise SyntaxError()
 * 
 *         if self.asMatch and callable(repl):             # <<<<<<<<<<<<<<
 *             warnings.warn("cannot use sub() with a callable with Regex(asMatch=True)",
 *                           SyntaxWarning, stacklevel=2)
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_asMatch); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 3267, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_5) {
  } else {
    __pyx_t_2 = __pyx_t_5;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_4 = __pyx_cur_scope->__pyx_v_repl;
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyCallable_Check(__pyx_t_4); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 3267, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_6 = (__pyx_t_5 != 0);
  __pyx_t_2 = __pyx_t_6;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_2)) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3268
 * 
 *         if self.asMatch and callable(repl):
 *             warnings.warn("cannot use sub() with a callable with Regex(asMatch=True)",             # <<<<<<<<<<<<<<
 *                           SyntaxWarning, stacklevel=2)
 *             raise SyntaxError()
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_warnings); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3268, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_warn); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3268, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3269
 *         if self.asMatch and callable(repl):
 *             warnings.warn("cannot use sub() with a callable with Regex(asMatch=True)",
 *                           SyntaxWarning, stacklevel=2)             # <<<<<<<<<<<<<<
 *             raise SyntaxError()
 * 
 */
    __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3269, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_stacklevel, __pyx_int_2) < 0) __PYX_ERR(0, 3269, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":3268
 * 
 *         if self.asMatch and callable(repl):
 *             warnings.warn("cannot use sub() with a callable with Regex(asMatch=True)",             # <<<<<<<<<<<<<<
 *                           SyntaxWarning, stacklevel=2)
 *             raise SyntaxError()
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__91, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3268, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3270
 *             warnings.warn("cannot use sub() with a callable with Regex(asMatch=True)",
 *                           SyntaxWarning, stacklevel=2)
 *             raise SyntaxError()             # <<<<<<<<<<<<<<
 * 
 *         if self.asMatch:
 */
    __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_builtin_SyntaxError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3270, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 3270, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":3267
 *             raise SyntaxError()
 * 
 *         if self.asMatch and callable(repl):             # <<<<<<<<<<<<<<
 *             warnings.warn("cannot use sub() with a callable with Regex(asMatch=True)",
 *                           SyntaxWarning, stacklevel=2)
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3272
 *             raise SyntaxError()
 * 
 *         if self.asMatch:             # <<<<<<<<<<<<<<
 *             def pa(tokens):
 *                 return tokens[0].expand(repl)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_asMatch); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 3272, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3273
 * 
 *         if self.asMatch:
 *             def pa(tokens):             # <<<<<<<<<<<<<<
 *                 return tokens[0].expand(repl)
 *         else:
 */
    __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_5Regex_3sub_1pa, 0, __pyx_n_s_Regex_sub_locals_pa, ((PyObject*)__pyx_cur_scope), __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__93)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3273, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_pa = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3272
 *             raise SyntaxError()
 * 
 *         if self.asMatch:             # <<<<<<<<<<<<<<
 *             def pa(tokens):
 *                 return tokens[0].expand(repl)
 */
    goto __pyx_L7;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3276
 *                 return tokens[0].expand(repl)
 *         else:
 *             def pa(tokens):             # <<<<<<<<<<<<<<
 *                 return self.re.sub(repl, tokens[0])
 *         return self.addParseAction(pa)
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_5Regex_3sub_3pa, 0, __pyx_n_s_Regex_sub_locals_pa, ((PyObject*)__pyx_cur_scope), __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__95)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_pa = __pyx_t_3;
    __pyx_t_3 = 0;
  }
  __pyx_L7:;

  /* "Aspidites/_vendor/pyparsing_extension.py":3278
 *             def pa(tokens):
 *                 return self.re.sub(repl, tokens[0])
 *         return self.addParseAction(pa)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_addParseAction); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3278, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_1, __pyx_v_pa) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_pa);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3278, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3251
 *         return self.strRepr
 * 
 *     def sub(self, repl):             # <<<<<<<<<<<<<<
 *         r"""
 *         Return Regex with an attached parse action to transform the parsed
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Regex.sub", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pa);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3321
 *     """
 * 
 *     def __init__(self, quoteChar, escChar=None, escQuote=None, multiline=False,             # <<<<<<<<<<<<<<
 *                  unquoteResults=True, endQuoteChar=None, convertWhitespaceEscapes=True):
 *         super(QuotedString, self).__init__()
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12QuotedString_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12QuotedString_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12QuotedString_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12QuotedString_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_quoteChar = 0;
  PyObject *__pyx_v_escChar = 0;
  PyObject *__pyx_v_escQuote = 0;
  PyObject *__pyx_v_multiline = 0;
  PyObject *__pyx_v_unquoteResults = 0;
  PyObject *__pyx_v_endQuoteChar = 0;
  PyObject *__pyx_v_convertWhitespaceEscapes = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_quoteChar,&__pyx_n_s_escChar,&__pyx_n_s_escQuote,&__pyx_n_s_multiline,&__pyx_n_s_unquoteResults,&__pyx_n_s_endQuoteChar,&__pyx_n_s_convertWhitespaceEscapes,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    values[3] = ((PyObject *)((PyObject *)Py_None));
    values[4] = ((PyObject *)((PyObject *)Py_False));

    /* "Aspidites/_vendor/pyparsing_extension.py":3322
 * 
 *     def __init__(self, quoteChar, escChar=None, escQuote=None, multiline=False,
 *                  unquoteResults=True, endQuoteChar=None, convertWhitespaceEscapes=True):             # <<<<<<<<<<<<<<
 *         super(QuotedString, self).__init__()
 * 
 */
    values[5] = ((PyObject *)((PyObject *)Py_True));
    values[6] = ((PyObject *)((PyObject *)Py_None));
    values[7] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_quoteChar)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 8, 1); __PYX_ERR(0, 3321, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_escChar);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_escQuote);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_multiline);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_unquoteResults);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_endQuoteChar);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_convertWhitespaceEscapes);
          if (value) { values[7] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 3321, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_quoteChar = values[1];
    __pyx_v_escChar = values[2];
    __pyx_v_escQuote = values[3];
    __pyx_v_multiline = values[4];
    __pyx_v_unquoteResults = values[5];
    __pyx_v_endQuoteChar = values[6];
    __pyx_v_convertWhitespaceEscapes = values[7];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 8, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3321, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.QuotedString.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12QuotedString___init__(__pyx_self, __pyx_v_self, __pyx_v_quoteChar, __pyx_v_escChar, __pyx_v_escQuote, __pyx_v_multiline, __pyx_v_unquoteResults, __pyx_v_endQuoteChar, __pyx_v_convertWhitespaceEscapes);

  /* "Aspidites/_vendor/pyparsing_extension.py":3321
 *     """
 * 
 *     def __init__(self, quoteChar, escChar=None, escQuote=None, multiline=False,             # <<<<<<<<<<<<<<
 *                  unquoteResults=True, endQuoteChar=None, convertWhitespaceEscapes=True):
 *         super(QuotedString, self).__init__()
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_12QuotedString_8__init___2generator18(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/pyparsing_extension.py":3361
 *         if len(self.endQuoteChar) > 1:
 *             self.pattern += (
 *                     '|(?:' + ')|(?:'.join("%s[^%s]" % (re.escape(self.endQuoteChar[:i]),             # <<<<<<<<<<<<<<
 *                                                        _escapeRegexRangeChars(self.endQuoteChar[i]))
 *                                           for i in range(len(self.endQuoteChar) - 1, 0, -1)) + ')')
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12QuotedString_8__init___genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 3361, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__ *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_12QuotedString_8__init___2generator18, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_QuotedString___init___locals_gen, __pyx_n_s_Aspidites__vendor_pyparsing_exte); if (unlikely(!gen)) __PYX_ERR(0, 3361, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.QuotedString.__init__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_12QuotedString_8__init___2generator18(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 3361, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3363
 *                     '|(?:' + ')|(?:'.join("%s[^%s]" % (re.escape(self.endQuoteChar[:i]),
 *                                                        _escapeRegexRangeChars(self.endQuoteChar[i]))
 *                                           for i in range(len(self.endQuoteChar) - 1, 0, -1)) + ')')             # <<<<<<<<<<<<<<
 * 
 *         if escQuote:
 */
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 3363, __pyx_L1_error) }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_endQuoteChar); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 3363, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (__pyx_t_3 = (__pyx_t_2 - 1); __pyx_t_3 > 0; __pyx_t_3-=1) {
    __pyx_cur_scope->__pyx_v_i = __pyx_t_3;

    /* "Aspidites/_vendor/pyparsing_extension.py":3361
 *         if len(self.endQuoteChar) > 1:
 *             self.pattern += (
 *                     '|(?:' + ')|(?:'.join("%s[^%s]" % (re.escape(self.endQuoteChar[:i]),             # <<<<<<<<<<<<<<
 *                                                        _escapeRegexRangeChars(self.endQuoteChar[i]))
 *                                           for i in range(len(self.endQuoteChar) - 1, 0, -1)) + ')')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_re); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3361, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_escape); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3361, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 3361, __pyx_L1_error) }
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_endQuoteChar); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3361, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = __Pyx_PyObject_GetSlice(__pyx_t_4, 0, __pyx_cur_scope->__pyx_v_i, NULL, NULL, NULL, 0, 1, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3361, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3361, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3362
 *             self.pattern += (
 *                     '|(?:' + ')|(?:'.join("%s[^%s]" % (re.escape(self.endQuoteChar[:i]),
 *                                                        _escapeRegexRangeChars(self.endQuoteChar[i]))             # <<<<<<<<<<<<<<
 *                                           for i in range(len(self.endQuoteChar) - 1, 0, -1)) + ')')
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_escapeRegexRangeChars); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3362, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 3362, __pyx_L1_error) }
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_endQuoteChar); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3362, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_4, __pyx_cur_scope->__pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3362, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_5 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_4, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3362, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3361
 *         if len(self.endQuoteChar) > 1:
 *             self.pattern += (
 *                     '|(?:' + ')|(?:'.join("%s[^%s]" % (re.escape(self.endQuoteChar[:i]),             # <<<<<<<<<<<<<<
 *                                                        _escapeRegexRangeChars(self.endQuoteChar[i]))
 *                                           for i in range(len(self.endQuoteChar) - 1, 0, -1)) + ')')
 */
    __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3361, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_5);
    __pyx_t_1 = 0;
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_s_s_5, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3361, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_3;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_0;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_1;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 3361, __pyx_L1_error)
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3321
 *     """
 * 
 *     def __init__(self, quoteChar, escChar=None, escQuote=None, multiline=False,             # <<<<<<<<<<<<<<
 *                  unquoteResults=True, endQuoteChar=None, convertWhitespaceEscapes=True):
 *         super(QuotedString, self).__init__()
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12QuotedString___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_quoteChar, PyObject *__pyx_v_escChar, PyObject *__pyx_v_escQuote, PyObject *__pyx_v_multiline, PyObject *__pyx_v_unquoteResults, PyObject *__pyx_v_endQuoteChar, PyObject *__pyx_v_convertWhitespaceEscapes) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__ *__pyx_cur_scope;
  PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_12QuotedString_8__init___2generator18 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__ *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 3321, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_INCREF(__pyx_v_quoteChar);
  __Pyx_INCREF(__pyx_v_endQuoteChar);

  /* "Aspidites/_vendor/pyparsing_extension.py":3323
 *     def __init__(self, quoteChar, escChar=None, escQuote=None, multiline=False,
 *                  unquoteResults=True, endQuoteChar=None, convertWhitespaceEscapes=True):
 *         super(QuotedString, self).__init__()             # <<<<<<<<<<<<<<
 * 
 *         # remove white space from quote chars - wont work anyway
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_QuotedString); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_cur_scope->__pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3326
 * 
 *         # remove white space from quote chars - wont work anyway
 *         quoteChar = quoteChar.strip()             # <<<<<<<<<<<<<<
 *         if not quoteChar:
 *             warnings.warn("quoteChar cannot be the empty string", SyntaxWarning, stacklevel=2)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_quoteChar, __pyx_n_s_strip); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF_SET(__pyx_v_quoteChar, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3327
 *         # remove white space from quote chars - wont work anyway
 *         quoteChar = quoteChar.strip()
 *         if not quoteChar:             # <<<<<<<<<<<<<<
 *             warnings.warn("quoteChar cannot be the empty string", SyntaxWarning, stacklevel=2)
 *             raise SyntaxError()
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_quoteChar); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 3327, __pyx_L1_error)
  __pyx_t_5 = ((!__pyx_t_4) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3328
 *         quoteChar = quoteChar.strip()
 *         if not quoteChar:
 *             warnings.warn("quoteChar cannot be the empty string", SyntaxWarning, stacklevel=2)             # <<<<<<<<<<<<<<
 *             raise SyntaxError()
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_warnings); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3328, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_warn); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3328, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3328, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_stacklevel, __pyx_int_2) < 0) __PYX_ERR(0, 3328, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__96, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3328, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3329
 *         if not quoteChar:
 *             warnings.warn("quoteChar cannot be the empty string", SyntaxWarning, stacklevel=2)
 *             raise SyntaxError()             # <<<<<<<<<<<<<<
 * 
 *         if endQuoteChar is None:
 */
    __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_builtin_SyntaxError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3329, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 3329, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":3327
 *         # remove white space from quote chars - wont work anyway
 *         quoteChar = quoteChar.strip()
 *         if not quoteChar:             # <<<<<<<<<<<<<<
 *             warnings.warn("quoteChar cannot be the empty string", SyntaxWarning, stacklevel=2)
 *             raise SyntaxError()
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3331
 *             raise SyntaxError()
 * 
 *         if endQuoteChar is None:             # <<<<<<<<<<<<<<
 *             endQuoteChar = quoteChar
 *         else:
 */
  __pyx_t_5 = (__pyx_v_endQuoteChar == Py_None);
  __pyx_t_4 = (__pyx_t_5 != 0);
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3332
 * 
 *         if endQuoteChar is None:
 *             endQuoteChar = quoteChar             # <<<<<<<<<<<<<<
 *         else:
 *             endQuoteChar = endQuoteChar.strip()
 */
    __Pyx_INCREF(__pyx_v_quoteChar);
    __Pyx_DECREF_SET(__pyx_v_endQuoteChar, __pyx_v_quoteChar);

    /* "Aspidites/_vendor/pyparsing_extension.py":3331
 *             raise SyntaxError()
 * 
 *         if endQuoteChar is None:             # <<<<<<<<<<<<<<
 *             endQuoteChar = quoteChar
 *         else:
 */
    goto __pyx_L4;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3334
 *             endQuoteChar = quoteChar
 *         else:
 *             endQuoteChar = endQuoteChar.strip()             # <<<<<<<<<<<<<<
 *             if not endQuoteChar:
 *                 warnings.warn("endQuoteChar cannot be the empty string", SyntaxWarning, stacklevel=2)
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_endQuoteChar, __pyx_n_s_strip); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3334, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3334, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_endQuoteChar, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3335
 *         else:
 *             endQuoteChar = endQuoteChar.strip()
 *             if not endQuoteChar:             # <<<<<<<<<<<<<<
 *                 warnings.warn("endQuoteChar cannot be the empty string", SyntaxWarning, stacklevel=2)
 *                 raise SyntaxError()
 */
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_endQuoteChar); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 3335, __pyx_L1_error)
    __pyx_t_5 = ((!__pyx_t_4) != 0);
    if (unlikely(__pyx_t_5)) {

      /* "Aspidites/_vendor/pyparsing_extension.py":3336
 *             endQuoteChar = endQuoteChar.strip()
 *             if not endQuoteChar:
 *                 warnings.warn("endQuoteChar cannot be the empty string", SyntaxWarning, stacklevel=2)             # <<<<<<<<<<<<<<
 *                 raise SyntaxError()
 * 
 */
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_warnings); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3336, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_warn); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3336, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3336, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_stacklevel, __pyx_int_2) < 0) __PYX_ERR(0, 3336, __pyx_L1_error)
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__97, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3336, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":3337
 *             if not endQuoteChar:
 *                 warnings.warn("endQuoteChar cannot be the empty string", SyntaxWarning, stacklevel=2)
 *                 raise SyntaxError()             # <<<<<<<<<<<<<<
 * 
 *         self.quoteChar = quoteChar
 */
      __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_builtin_SyntaxError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3337, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 3337, __pyx_L1_error)

      /* "Aspidites/_vendor/pyparsing_extension.py":3335
 *         else:
 *             endQuoteChar = endQuoteChar.strip()
 *             if not endQuoteChar:             # <<<<<<<<<<<<<<
 *                 warnings.warn("endQuoteChar cannot be the empty string", SyntaxWarning, stacklevel=2)
 *                 raise SyntaxError()
 */
    }
  }
  __pyx_L4:;

  /* "Aspidites/_vendor/pyparsing_extension.py":3339
 *                 raise SyntaxError()
 * 
 *         self.quoteChar = quoteChar             # <<<<<<<<<<<<<<
 *         self.quoteCharLen = len(quoteChar)
 *         self.firstQuoteChar = quoteChar[0]
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_quoteChar, __pyx_v_quoteChar) < 0) __PYX_ERR(0, 3339, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3340
 * 
 *         self.quoteChar = quoteChar
 *         self.quoteCharLen = len(quoteChar)             # <<<<<<<<<<<<<<
 *         self.firstQuoteChar = quoteChar[0]
 *         self.endQuoteChar = endQuoteChar
 */
  __pyx_t_6 = PyObject_Length(__pyx_v_quoteChar); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 3340, __pyx_L1_error)
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3340, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_quoteCharLen, __pyx_t_3) < 0) __PYX_ERR(0, 3340, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3341
 *         self.quoteChar = quoteChar
 *         self.quoteCharLen = len(quoteChar)
 *         self.firstQuoteChar = quoteChar[0]             # <<<<<<<<<<<<<<
 *         self.endQuoteChar = endQuoteChar
 *         self.endQuoteCharLen = len(endQuoteChar)
 */
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_quoteChar, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_firstQuoteChar, __pyx_t_3) < 0) __PYX_ERR(0, 3341, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3342
 *         self.quoteCharLen = len(quoteChar)
 *         self.firstQuoteChar = quoteChar[0]
 *         self.endQuoteChar = endQuoteChar             # <<<<<<<<<<<<<<
 *         self.endQuoteCharLen = len(endQuoteChar)
 *         self.escChar = escChar
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_endQuoteChar, __pyx_v_endQuoteChar) < 0) __PYX_ERR(0, 3342, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3343
 *         self.firstQuoteChar = quoteChar[0]
 *         self.endQuoteChar = endQuoteChar
 *         self.endQuoteCharLen = len(endQuoteChar)             # <<<<<<<<<<<<<<
 *         self.escChar = escChar
 *         self.escQuote = escQuote
 */
  __pyx_t_6 = PyObject_Length(__pyx_v_endQuoteChar); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 3343, __pyx_L1_error)
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_endQuoteCharLen, __pyx_t_3) < 0) __PYX_ERR(0, 3343, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3344
 *         self.endQuoteChar = endQuoteChar
 *         self.endQuoteCharLen = len(endQuoteChar)
 *         self.escChar = escChar             # <<<<<<<<<<<<<<
 *         self.escQuote = escQuote
 *         self.unquoteResults = unquoteResults
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_escChar, __pyx_v_escChar) < 0) __PYX_ERR(0, 3344, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3345
 *         self.endQuoteCharLen = len(endQuoteChar)
 *         self.escChar = escChar
 *         self.escQuote = escQuote             # <<<<<<<<<<<<<<
 *         self.unquoteResults = unquoteResults
 *         self.convertWhitespaceEscapes = convertWhitespaceEscapes
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_escQuote, __pyx_v_escQuote) < 0) __PYX_ERR(0, 3345, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3346
 *         self.escChar = escChar
 *         self.escQuote = escQuote
 *         self.unquoteResults = unquoteResults             # <<<<<<<<<<<<<<
 *         self.convertWhitespaceEscapes = convertWhitespaceEscapes
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_unquoteResults, __pyx_v_unquoteResults) < 0) __PYX_ERR(0, 3346, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3347
 *         self.escQuote = escQuote
 *         self.unquoteResults = unquoteResults
 *         self.convertWhitespaceEscapes = convertWhitespaceEscapes             # <<<<<<<<<<<<<<
 * 
 *         if multiline:
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_convertWhitespaceEscapes, __pyx_v_convertWhitespaceEscapes) < 0) __PYX_ERR(0, 3347, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3349
 *         self.convertWhitespaceEscapes = convertWhitespaceEscapes
 * 
 *         if multiline:             # <<<<<<<<<<<<<<
 *             self.flags = re.MULTILINE | re.DOTALL
 *             self.pattern = r'%s(?:[^%s%s]' % (re.escape(self.quoteChar),
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_multiline); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 3349, __pyx_L1_error)
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3350
 * 
 *         if multiline:
 *             self.flags = re.MULTILINE | re.DOTALL             # <<<<<<<<<<<<<<
 *             self.pattern = r'%s(?:[^%s%s]' % (re.escape(self.quoteChar),
 *                                               _escapeRegexRangeChars(self.endQuoteChar[0]),
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_re); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3350, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_MULTILINE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3350, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_re); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3350, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_DOTALL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3350, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Or(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3350, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_flags, __pyx_t_3) < 0) __PYX_ERR(0, 3350, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3351
 *         if multiline:
 *             self.flags = re.MULTILINE | re.DOTALL
 *             self.pattern = r'%s(?:[^%s%s]' % (re.escape(self.quoteChar),             # <<<<<<<<<<<<<<
 *                                               _escapeRegexRangeChars(self.endQuoteChar[0]),
 *                                               (escChar is not None and _escapeRegexRangeChars(escChar) or ''))
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_re); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3351, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_escape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3351, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_quoteChar); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3351, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_7, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3351, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3352
 *             self.flags = re.MULTILINE | re.DOTALL
 *             self.pattern = r'%s(?:[^%s%s]' % (re.escape(self.quoteChar),
 *                                               _escapeRegexRangeChars(self.endQuoteChar[0]),             # <<<<<<<<<<<<<<
 *                                               (escChar is not None and _escapeRegexRangeChars(escChar) or ''))
 *         else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_escapeRegexRangeChars); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3352, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_endQuoteChar); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3352, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_GetItemInt(__pyx_t_7, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3352, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_7, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_8);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3352, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3353
 *             self.pattern = r'%s(?:[^%s%s]' % (re.escape(self.quoteChar),
 *                                               _escapeRegexRangeChars(self.endQuoteChar[0]),
 *                                               (escChar is not None and _escapeRegexRangeChars(escChar) or ''))             # <<<<<<<<<<<<<<
 *         else:
 *             self.flags = 0
 */
    __pyx_t_5 = (__pyx_v_escChar != Py_None);
    if (!__pyx_t_5) {
      goto __pyx_L8_next_or;
    } else {
    }
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_escapeRegexRangeChars); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3353, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_8 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_9, __pyx_v_escChar) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_escChar);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3353, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 3353, __pyx_L1_error)
    if (!__pyx_t_5) {
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else {
      __Pyx_INCREF(__pyx_t_8);
      __pyx_t_1 = __pyx_t_8;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_L8_next_or:;
    __Pyx_INCREF(__pyx_kp_s__12);
    __pyx_t_1 = __pyx_kp_s__12;
    __pyx_L7_bool_binop_done:;

    /* "Aspidites/_vendor/pyparsing_extension.py":3351
 *         if multiline:
 *             self.flags = re.MULTILINE | re.DOTALL
 *             self.pattern = r'%s(?:[^%s%s]' % (re.escape(self.quoteChar),             # <<<<<<<<<<<<<<
 *                                               _escapeRegexRangeChars(self.endQuoteChar[0]),
 *                                               (escChar is not None and _escapeRegexRangeChars(escChar) or ''))
 */
    __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3351, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_1);
    __pyx_t_3 = 0;
    __pyx_t_2 = 0;
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_s_s_s_2, __pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3351, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_pattern, __pyx_t_1) < 0) __PYX_ERR(0, 3351, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3349
 *         self.convertWhitespaceEscapes = convertWhitespaceEscapes
 * 
 *         if multiline:             # <<<<<<<<<<<<<<
 *             self.flags = re.MULTILINE | re.DOTALL
 *             self.pattern = r'%s(?:[^%s%s]' % (re.escape(self.quoteChar),
 */
    goto __pyx_L6;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3355
 *                                               (escChar is not None and _escapeRegexRangeChars(escChar) or ''))
 *         else:
 *             self.flags = 0             # <<<<<<<<<<<<<<
 *             self.pattern = r'%s(?:[^%s\n\r%s]' % (re.escape(self.quoteChar),
 *                                                   _escapeRegexRangeChars(self.endQuoteChar[0]),
 */
  /*else*/ {
    if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_flags, __pyx_int_0) < 0) __PYX_ERR(0, 3355, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":3356
 *         else:
 *             self.flags = 0
 *             self.pattern = r'%s(?:[^%s\n\r%s]' % (re.escape(self.quoteChar),             # <<<<<<<<<<<<<<
 *                                                   _escapeRegexRangeChars(self.endQuoteChar[0]),
 *                                                   (escChar is not None and _escapeRegexRangeChars(escChar) or ''))
 */
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_re); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3356, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_escape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3356, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_quoteChar); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3356, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_8);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3356, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3357
 *             self.flags = 0
 *             self.pattern = r'%s(?:[^%s\n\r%s]' % (re.escape(self.quoteChar),
 *                                                   _escapeRegexRangeChars(self.endQuoteChar[0]),             # <<<<<<<<<<<<<<
 *                                                   (escChar is not None and _escapeRegexRangeChars(escChar) or ''))
 *         if len(self.endQuoteChar) > 1:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_escapeRegexRangeChars); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3357, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_endQuoteChar); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3357, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3357, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_3, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3357, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3358
 *             self.pattern = r'%s(?:[^%s\n\r%s]' % (re.escape(self.quoteChar),
 *                                                   _escapeRegexRangeChars(self.endQuoteChar[0]),
 *                                                   (escChar is not None and _escapeRegexRangeChars(escChar) or ''))             # <<<<<<<<<<<<<<
 *         if len(self.endQuoteChar) > 1:
 *             self.pattern += (
 */
    __pyx_t_5 = (__pyx_v_escChar != Py_None);
    if (!__pyx_t_5) {
      goto __pyx_L11_next_or;
    } else {
    }
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_escapeRegexRangeChars); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3358, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_7 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_9, __pyx_v_escChar) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_escChar);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3358, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 3358, __pyx_L1_error)
    if (!__pyx_t_5) {
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else {
      __Pyx_INCREF(__pyx_t_7);
      __pyx_t_8 = __pyx_t_7;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L10_bool_binop_done;
    }
    __pyx_L11_next_or:;
    __Pyx_INCREF(__pyx_kp_s__12);
    __pyx_t_8 = __pyx_kp_s__12;
    __pyx_L10_bool_binop_done:;

    /* "Aspidites/_vendor/pyparsing_extension.py":3356
 *         else:
 *             self.flags = 0
 *             self.pattern = r'%s(?:[^%s\n\r%s]' % (re.escape(self.quoteChar),             # <<<<<<<<<<<<<<
 *                                                   _escapeRegexRangeChars(self.endQuoteChar[0]),
 *                                                   (escChar is not None and _escapeRegexRangeChars(escChar) or ''))
 */
    __pyx_t_7 = PyTuple_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3356, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_8);
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyString_Format(__pyx_kp_s_s_s_n_r_s, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3356, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_pattern, __pyx_t_8) < 0) __PYX_ERR(0, 3356, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  }
  __pyx_L6:;

  /* "Aspidites/_vendor/pyparsing_extension.py":3359
 *                                                   _escapeRegexRangeChars(self.endQuoteChar[0]),
 *                                                   (escChar is not None and _escapeRegexRangeChars(escChar) or ''))
 *         if len(self.endQuoteChar) > 1:             # <<<<<<<<<<<<<<
 *             self.pattern += (
 *                     '|(?:' + ')|(?:'.join("%s[^%s]" % (re.escape(self.endQuoteChar[:i]),
 */
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_endQuoteChar); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_6 = PyObject_Length(__pyx_t_8); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 3359, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_5 = ((__pyx_t_6 > 1) != 0);
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3360
 *                                                   (escChar is not None and _escapeRegexRangeChars(escChar) or ''))
 *         if len(self.endQuoteChar) > 1:
 *             self.pattern += (             # <<<<<<<<<<<<<<
 *                     '|(?:' + ')|(?:'.join("%s[^%s]" % (re.escape(self.endQuoteChar[:i]),
 *                                                        _escapeRegexRangeChars(self.endQuoteChar[i]))
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_pattern); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3360, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);

    /* "Aspidites/_vendor/pyparsing_extension.py":3361
 *         if len(self.endQuoteChar) > 1:
 *             self.pattern += (
 *                     '|(?:' + ')|(?:'.join("%s[^%s]" % (re.escape(self.endQuoteChar[:i]),             # <<<<<<<<<<<<<<
 *                                                        _escapeRegexRangeChars(self.endQuoteChar[i]))
 *                                           for i in range(len(self.endQuoteChar) - 1, 0, -1)) + ')')
 */
    __pyx_t_7 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12QuotedString_8__init___genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3361, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_2 = __Pyx_PyString_Join(__pyx_kp_s__99, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3361, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = PyNumber_Add(__pyx_kp_s__98, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3361, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3363
 *                     '|(?:' + ')|(?:'.join("%s[^%s]" % (re.escape(self.endQuoteChar[:i]),
 *                                                        _escapeRegexRangeChars(self.endQuoteChar[i]))
 *                                           for i in range(len(self.endQuoteChar) - 1, 0, -1)) + ')')             # <<<<<<<<<<<<<<
 * 
 *         if escQuote:
 */
    __pyx_t_2 = PyNumber_Add(__pyx_t_7, __pyx_kp_s__100); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3363, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3360
 *                                                   (escChar is not None and _escapeRegexRangeChars(escChar) or ''))
 *         if len(self.endQuoteChar) > 1:
 *             self.pattern += (             # <<<<<<<<<<<<<<
 *                     '|(?:' + ')|(?:'.join("%s[^%s]" % (re.escape(self.endQuoteChar[:i]),
 *                                                        _escapeRegexRangeChars(self.endQuoteChar[i]))
 */
    __pyx_t_7 = PyNumber_InPlaceAdd(__pyx_t_8, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3360, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_pattern, __pyx_t_7) < 0) __PYX_ERR(0, 3360, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3359
 *                                                   _escapeRegexRangeChars(self.endQuoteChar[0]),
 *                                                   (escChar is not None and _escapeRegexRangeChars(escChar) or ''))
 *         if len(self.endQuoteChar) > 1:             # <<<<<<<<<<<<<<
 *             self.pattern += (
 *                     '|(?:' + ')|(?:'.join("%s[^%s]" % (re.escape(self.endQuoteChar[:i]),
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3365
 *                                           for i in range(len(self.endQuoteChar) - 1, 0, -1)) + ')')
 * 
 *         if escQuote:             # <<<<<<<<<<<<<<
 *             self.pattern += (r'|(?:%s)' % re.escape(escQuote))
 *         if escChar:
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_escQuote); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 3365, __pyx_L1_error)
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3366
 * 
 *         if escQuote:
 *             self.pattern += (r'|(?:%s)' % re.escape(escQuote))             # <<<<<<<<<<<<<<
 *         if escChar:
 *             self.pattern += (r'|(?:%s.)' % re.escape(escChar))
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_pattern); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3366, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_re); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3366, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_escape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3366, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_8, __pyx_v_escQuote) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_escQuote);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3366, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyString_FormatSafe(__pyx_kp_s_s_6, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3366, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_t_7, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3366, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_pattern, __pyx_t_2) < 0) __PYX_ERR(0, 3366, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3365
 *                                           for i in range(len(self.endQuoteChar) - 1, 0, -1)) + ')')
 * 
 *         if escQuote:             # <<<<<<<<<<<<<<
 *             self.pattern += (r'|(?:%s)' % re.escape(escQuote))
 *         if escChar:
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3367
 *         if escQuote:
 *             self.pattern += (r'|(?:%s)' % re.escape(escQuote))
 *         if escChar:             # <<<<<<<<<<<<<<
 *             self.pattern += (r'|(?:%s.)' % re.escape(escChar))
 *             self.escCharReplacePattern = re.escape(self.escChar) + "(.)"
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_escChar); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 3367, __pyx_L1_error)
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3368
 *             self.pattern += (r'|(?:%s)' % re.escape(escQuote))
 *         if escChar:
 *             self.pattern += (r'|(?:%s.)' % re.escape(escChar))             # <<<<<<<<<<<<<<
 *             self.escCharReplacePattern = re.escape(self.escChar) + "(.)"
 *         self.pattern += (r')*%s' % re.escape(self.endQuoteChar))
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_pattern); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3368, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_re); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3368, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_escape); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3368, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_7, __pyx_v_escChar) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_v_escChar);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3368, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyString_FormatSafe(__pyx_kp_s_s_7, __pyx_t_1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3368, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_t_2, __pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3368, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_pattern, __pyx_t_1) < 0) __PYX_ERR(0, 3368, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3369
 *         if escChar:
 *             self.pattern += (r'|(?:%s.)' % re.escape(escChar))
 *             self.escCharReplacePattern = re.escape(self.escChar) + "(.)"             # <<<<<<<<<<<<<<
 *         self.pattern += (r')*%s' % re.escape(self.endQuoteChar))
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_re); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3369, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_escape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3369, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_escChar); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3369, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_7, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_8);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3369, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyNumber_Add(__pyx_t_1, __pyx_kp_s__101); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3369, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_escCharReplacePattern, __pyx_t_2) < 0) __PYX_ERR(0, 3369, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3367
 *         if escQuote:
 *             self.pattern += (r'|(?:%s)' % re.escape(escQuote))
 *         if escChar:             # <<<<<<<<<<<<<<
 *             self.pattern += (r'|(?:%s.)' % re.escape(escChar))
 *             self.escCharReplacePattern = re.escape(self.escChar) + "(.)"
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3370
 *             self.pattern += (r'|(?:%s.)' % re.escape(escChar))
 *             self.escCharReplacePattern = re.escape(self.escChar) + "(.)"
 *         self.pattern += (r')*%s' % re.escape(self.endQuoteChar))             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_pattern); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_re); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_escape); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_endQuoteChar); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_3, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_8);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyString_FormatSafe(__pyx_kp_s_s_8, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_t_2, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_pattern, __pyx_t_1) < 0) __PYX_ERR(0, 3370, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3372
 *         self.pattern += (r')*%s' % re.escape(self.endQuoteChar))
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             self.re = re.compile(self.pattern, self.flags)
 *             self.reString = self.pattern
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
    __Pyx_XGOTREF(__pyx_t_10);
    __Pyx_XGOTREF(__pyx_t_11);
    __Pyx_XGOTREF(__pyx_t_12);
    /*try:*/ {

      /* "Aspidites/_vendor/pyparsing_extension.py":3373
 * 
 *         try:
 *             self.re = re.compile(self.pattern, self.flags)             # <<<<<<<<<<<<<<
 *             self.reString = self.pattern
 *             self.re_match = self.re.match
 */
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_re); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3373, __pyx_L16_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_compile); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3373, __pyx_L16_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_pattern); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3373, __pyx_L16_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_flags); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3373, __pyx_L16_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_3 = NULL;
      __pyx_t_13 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_13 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_2)) {
        PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_7, __pyx_t_8};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_13, 2+__pyx_t_13); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3373, __pyx_L16_error)
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
        PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_7, __pyx_t_8};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_13, 2+__pyx_t_13); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3373, __pyx_L16_error)
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      } else
      #endif
      {
        __pyx_t_9 = PyTuple_New(2+__pyx_t_13); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3373, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (__pyx_t_3) {
          __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_3); __pyx_t_3 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_13, __pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_8);
        PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_13, __pyx_t_8);
        __pyx_t_7 = 0;
        __pyx_t_8 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3373, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_re, __pyx_t_1) < 0) __PYX_ERR(0, 3373, __pyx_L16_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":3374
 *         try:
 *             self.re = re.compile(self.pattern, self.flags)
 *             self.reString = self.pattern             # <<<<<<<<<<<<<<
 *             self.re_match = self.re.match
 *         except sre_constants.error:
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_pattern); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3374, __pyx_L16_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_reString, __pyx_t_1) < 0) __PYX_ERR(0, 3374, __pyx_L16_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":3375
 *             self.re = re.compile(self.pattern, self.flags)
 *             self.reString = self.pattern
 *             self.re_match = self.re.match             # <<<<<<<<<<<<<<
 *         except sre_constants.error:
 *             warnings.warn("invalid pattern (%s) passed to Regex" % self.pattern,
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_re); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3375, __pyx_L16_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_match); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3375, __pyx_L16_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_re_match, __pyx_t_2) < 0) __PYX_ERR(0, 3375, __pyx_L16_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":3372
 *         self.pattern += (r')*%s' % re.escape(self.endQuoteChar))
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             self.re = re.compile(self.pattern, self.flags)
 *             self.reString = self.pattern
 */
    }
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    goto __pyx_L21_try_end;
    __pyx_L16_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3376
 *             self.reString = self.pattern
 *             self.re_match = self.re.match
 *         except sre_constants.error:             # <<<<<<<<<<<<<<
 *             warnings.warn("invalid pattern (%s) passed to Regex" % self.pattern,
 *                           SyntaxWarning, stacklevel=2)
 */
    __Pyx_ErrFetch(&__pyx_t_2, &__pyx_t_1, &__pyx_t_9);
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_sre_constants); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3376, __pyx_L18_except_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_error); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3376, __pyx_L18_except_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_13 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_2, __pyx_t_7);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_ErrRestore(__pyx_t_2, __pyx_t_1, __pyx_t_9);
    __pyx_t_2 = 0; __pyx_t_1 = 0; __pyx_t_9 = 0;
    if (__pyx_t_13) {
      __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.QuotedString.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_9, &__pyx_t_1, &__pyx_t_2) < 0) __PYX_ERR(0, 3376, __pyx_L18_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_t_2);

      /* "Aspidites/_vendor/pyparsing_extension.py":3377
 *             self.re_match = self.re.match
 *         except sre_constants.error:
 *             warnings.warn("invalid pattern (%s) passed to Regex" % self.pattern,             # <<<<<<<<<<<<<<
 *                           SyntaxWarning, stacklevel=2)
 *             raise
 */
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_warnings); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3377, __pyx_L18_except_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_warn); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3377, __pyx_L18_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_pattern); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3377, __pyx_L18_except_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_3 = __Pyx_PyString_FormatSafe(__pyx_kp_s_invalid_pattern_s_passed_to_Rege, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3377, __pyx_L18_except_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":3378
 *         except sre_constants.error:
 *             warnings.warn("invalid pattern (%s) passed to Regex" % self.pattern,
 *                           SyntaxWarning, stacklevel=2)             # <<<<<<<<<<<<<<
 *             raise
 * 
 */
      __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3377, __pyx_L18_except_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3);
      __Pyx_INCREF(__pyx_builtin_SyntaxWarning);
      __Pyx_GIVEREF(__pyx_builtin_SyntaxWarning);
      PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_builtin_SyntaxWarning);
      __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3378, __pyx_L18_except_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_stacklevel, __pyx_int_2) < 0) __PYX_ERR(0, 3378, __pyx_L18_except_error)

      /* "Aspidites/_vendor/pyparsing_extension.py":3377
 *             self.re_match = self.re.match
 *         except sre_constants.error:
 *             warnings.warn("invalid pattern (%s) passed to Regex" % self.pattern,             # <<<<<<<<<<<<<<
 *                           SyntaxWarning, stacklevel=2)
 *             raise
 */
      __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_7, __pyx_t_3); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 3377, __pyx_L18_except_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":3379
 *             warnings.warn("invalid pattern (%s) passed to Regex" % self.pattern,
 *                           SyntaxWarning, stacklevel=2)
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         self.name = str(self)
 */
      __Pyx_GIVEREF(__pyx_t_9);
      __Pyx_GIVEREF(__pyx_t_1);
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_ErrRestoreWithState(__pyx_t_9, __pyx_t_1, __pyx_t_2);
      __pyx_t_9 = 0; __pyx_t_1 = 0; __pyx_t_2 = 0; 
      __PYX_ERR(0, 3379, __pyx_L18_except_error)
    }
    goto __pyx_L18_except_error;
    __pyx_L18_except_error:;

    /* "Aspidites/_vendor/pyparsing_extension.py":3372
 *         self.pattern += (r')*%s' % re.escape(self.endQuoteChar))
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             self.re = re.compile(self.pattern, self.flags)
 *             self.reString = self.pattern
 */
    __Pyx_XGIVEREF(__pyx_t_10);
    __Pyx_XGIVEREF(__pyx_t_11);
    __Pyx_XGIVEREF(__pyx_t_12);
    __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
    goto __pyx_L1_error;
    __pyx_L21_try_end:;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3381
 *             raise
 * 
 *         self.name = str(self)             # <<<<<<<<<<<<<<
 *         self.errmsg = "Expected " + self.name
 *         self.mayIndexError = False
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_cur_scope->__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_name, __pyx_t_2) < 0) __PYX_ERR(0, 3381, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3382
 * 
 *         self.name = str(self)
 *         self.errmsg = "Expected " + self.name             # <<<<<<<<<<<<<<
 *         self.mayIndexError = False
 *         self.mayReturnEmpty = True
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3382, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyNumber_Add(__pyx_kp_s_Expected, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3382, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_errmsg, __pyx_t_1) < 0) __PYX_ERR(0, 3382, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3383
 *         self.name = str(self)
 *         self.errmsg = "Expected " + self.name
 *         self.mayIndexError = False             # <<<<<<<<<<<<<<
 *         self.mayReturnEmpty = True
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_mayIndexError, Py_False) < 0) __PYX_ERR(0, 3383, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3384
 *         self.errmsg = "Expected " + self.name
 *         self.mayIndexError = False
 *         self.mayReturnEmpty = True             # <<<<<<<<<<<<<<
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_mayReturnEmpty, Py_True) < 0) __PYX_ERR(0, 3384, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3321
 *     """
 * 
 *     def __init__(self, quoteChar, escChar=None, escQuote=None, multiline=False,             # <<<<<<<<<<<<<<
 *                  unquoteResults=True, endQuoteChar=None, convertWhitespaceEscapes=True):
 *         super(QuotedString, self).__init__()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.QuotedString.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_12QuotedString_8__init___2generator18);
  __Pyx_XDECREF(__pyx_v_quoteChar);
  __Pyx_XDECREF(__pyx_v_endQuoteChar);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3386
 *         self.mayReturnEmpty = True
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         result = instring[loc] == self.firstQuoteChar and self.re_match(instring, loc) or None
 *         if not result:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12QuotedString_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12QuotedString_3parseImpl = {"parseImpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12QuotedString_3parseImpl, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12QuotedString_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  CYTHON_UNUSED PyObject *__pyx_v_doActions = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parseImpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_doActions,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 1); __PYX_ERR(0, 3386, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 2); __PYX_ERR(0, 3386, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_doActions);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "parseImpl") < 0)) __PYX_ERR(0, 3386, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_doActions = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3386, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.QuotedString.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12QuotedString_2parseImpl(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12QuotedString_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions) {
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_v_ret = NULL;
  PyObject *__pyx_v_ws_map = NULL;
  PyObject *__pyx_v_wslit = NULL;
  PyObject *__pyx_v_wschar = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  PyObject *(*__pyx_t_11)(PyObject *);
  PyObject *(*__pyx_t_12)(PyObject *);
  PyObject *__pyx_t_13 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parseImpl", 0);
  __Pyx_INCREF(__pyx_v_loc);

  /* "Aspidites/_vendor/pyparsing_extension.py":3387
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         result = instring[loc] == self.firstQuoteChar and self.re_match(instring, loc) or None             # <<<<<<<<<<<<<<
 *         if not result:
 *             raise ParseException(instring, loc, self.errmsg, self)
 */
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_instring, __pyx_v_loc); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_firstQuoteChar); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_2, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3387, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 3387, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (!__pyx_t_5) {
    goto __pyx_L4_next_or;
  } else {
  }
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_re_match); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_instring, __pyx_v_loc};
    __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3387, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_4);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_instring, __pyx_v_loc};
    __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3387, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_4);
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3387, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_instring);
    __Pyx_GIVEREF(__pyx_v_instring);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_instring);
    __Pyx_INCREF(__pyx_v_loc);
    __Pyx_GIVEREF(__pyx_v_loc);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_loc);
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3387, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 3387, __pyx_L1_error)
  if (!__pyx_t_5) {
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_1 = __pyx_t_4;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_L4_next_or:;
  __Pyx_INCREF(Py_None);
  __pyx_t_1 = Py_None;
  __pyx_L3_bool_binop_done:;
  __pyx_v_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3388
 *     def parseImpl(self, instring, loc, doActions=True):
 *         result = instring[loc] == self.firstQuoteChar and self.re_match(instring, loc) or None
 *         if not result:             # <<<<<<<<<<<<<<
 *             raise ParseException(instring, loc, self.errmsg, self)
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_result); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 3388, __pyx_L1_error)
  __pyx_t_8 = ((!__pyx_t_5) != 0);
  if (unlikely(__pyx_t_8)) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3389
 *         result = instring[loc] == self.firstQuoteChar and self.re_match(instring, loc) or None
 *         if not result:
 *             raise ParseException(instring, loc, self.errmsg, self)             # <<<<<<<<<<<<<<
 * 
 *         loc = result.end()
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3389, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errmsg); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3389, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[5] = {__pyx_t_7, __pyx_v_instring, __pyx_v_loc, __pyx_t_3, __pyx_v_self};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 4+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3389, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[5] = {__pyx_t_7, __pyx_v_instring, __pyx_v_loc, __pyx_t_3, __pyx_v_self};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 4+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3389, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(4+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3389, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_INCREF(__pyx_v_instring);
      __Pyx_GIVEREF(__pyx_v_instring);
      PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_6, __pyx_v_instring);
      __Pyx_INCREF(__pyx_v_loc);
      __Pyx_GIVEREF(__pyx_v_loc);
      PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_6, __pyx_v_loc);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_2, 2+__pyx_t_6, __pyx_t_3);
      __Pyx_INCREF(__pyx_v_self);
      __Pyx_GIVEREF(__pyx_v_self);
      PyTuple_SET_ITEM(__pyx_t_2, 3+__pyx_t_6, __pyx_v_self);
      __pyx_t_3 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3389, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 3389, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":3388
 *     def parseImpl(self, instring, loc, doActions=True):
 *         result = instring[loc] == self.firstQuoteChar and self.re_match(instring, loc) or None
 *         if not result:             # <<<<<<<<<<<<<<
 *             raise ParseException(instring, loc, self.errmsg, self)
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3391
 *             raise ParseException(instring, loc, self.errmsg, self)
 * 
 *         loc = result.end()             # <<<<<<<<<<<<<<
 *         ret = result.group()
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_result, __pyx_n_s_end); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF_SET(__pyx_v_loc, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3392
 * 
 *         loc = result.end()
 *         ret = result.group()             # <<<<<<<<<<<<<<
 * 
 *         if self.unquoteResults:
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_result, __pyx_n_s_group); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3392, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3392, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_ret = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3394
 *         ret = result.group()
 * 
 *         if self.unquoteResults:             # <<<<<<<<<<<<<<
 * 
 *             # strip off quotes
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_unquoteResults); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 3394, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_8) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3397
 * 
 *             # strip off quotes
 *             ret = ret[self.quoteCharLen: -self.endQuoteCharLen]             # <<<<<<<<<<<<<<
 * 
 *             if isinstance(ret, str):
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_quoteCharLen); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_endQuoteCharLen); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = PyNumber_Negative(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetSlice(__pyx_v_ret, 0, 0, &__pyx_t_1, &__pyx_t_2, NULL, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_ret, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3399
 *             ret = ret[self.quoteCharLen: -self.endQuoteCharLen]
 * 
 *             if isinstance(ret, str):             # <<<<<<<<<<<<<<
 *                 # replace escaped whitespace
 *                 if '\\' in ret and self.convertWhitespaceEscapes:
 */
    __pyx_t_8 = PyString_Check(__pyx_v_ret); 
    __pyx_t_5 = (__pyx_t_8 != 0);
    if (__pyx_t_5) {

      /* "Aspidites/_vendor/pyparsing_extension.py":3401
 *             if isinstance(ret, str):
 *                 # replace escaped whitespace
 *                 if '\\' in ret and self.convertWhitespaceEscapes:             # <<<<<<<<<<<<<<
 *                     ws_map = {
 *                         r'\t': '\t',
 */
      __pyx_t_8 = (__Pyx_PySequence_ContainsTF(__pyx_kp_s__26, __pyx_v_ret, Py_EQ)); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 3401, __pyx_L1_error)
      __pyx_t_9 = (__pyx_t_8 != 0);
      if (__pyx_t_9) {
      } else {
        __pyx_t_5 = __pyx_t_9;
        goto __pyx_L10_bool_binop_done;
      }
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_convertWhitespaceEscapes); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3401, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 3401, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_5 = __pyx_t_9;
      __pyx_L10_bool_binop_done:;
      if (__pyx_t_5) {

        /* "Aspidites/_vendor/pyparsing_extension.py":3403
 *                 if '\\' in ret and self.convertWhitespaceEscapes:
 *                     ws_map = {
 *                         r'\t': '\t',             # <<<<<<<<<<<<<<
 *                         r'\n': '\n',
 *                         r'\f': '\f',
 */
        __pyx_t_4 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3403, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        if (PyDict_SetItem(__pyx_t_4, __pyx_kp_s_t, __pyx_kp_s__6) < 0) __PYX_ERR(0, 3403, __pyx_L1_error)
        if (PyDict_SetItem(__pyx_t_4, __pyx_kp_s_n, __pyx_kp_s__4) < 0) __PYX_ERR(0, 3403, __pyx_L1_error)
        if (PyDict_SetItem(__pyx_t_4, __pyx_kp_s_f, __pyx_kp_s__102) < 0) __PYX_ERR(0, 3403, __pyx_L1_error)
        if (PyDict_SetItem(__pyx_t_4, __pyx_kp_s_r_2, __pyx_kp_s__103) < 0) __PYX_ERR(0, 3403, __pyx_L1_error)
        __pyx_v_ws_map = ((PyObject*)__pyx_t_4);
        __pyx_t_4 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":3408
 *                         r'\r': '\r',
 *                     }
 *                     for wslit, wschar in ws_map.items():             # <<<<<<<<<<<<<<
 *                         ret = ret.replace(wslit, wschar)
 * 
 */
        __pyx_t_4 = __Pyx_PyDict_Items(__pyx_v_ws_map); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3408, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
          __pyx_t_2 = __pyx_t_4; __Pyx_INCREF(__pyx_t_2); __pyx_t_10 = 0;
          __pyx_t_11 = NULL;
        } else {
          __pyx_t_10 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3408, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_11 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 3408, __pyx_L1_error)
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        for (;;) {
          if (likely(!__pyx_t_11)) {
            if (likely(PyList_CheckExact(__pyx_t_2))) {
              if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_2)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_4 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_10); __Pyx_INCREF(__pyx_t_4); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 3408, __pyx_L1_error)
              #else
              __pyx_t_4 = PySequence_ITEM(__pyx_t_2, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3408, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_4);
              #endif
            } else {
              if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_10); __Pyx_INCREF(__pyx_t_4); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 3408, __pyx_L1_error)
              #else
              __pyx_t_4 = PySequence_ITEM(__pyx_t_2, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3408, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_4);
              #endif
            }
          } else {
            __pyx_t_4 = __pyx_t_11(__pyx_t_2);
            if (unlikely(!__pyx_t_4)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 3408, __pyx_L1_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_4);
          }
          if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
            PyObject* sequence = __pyx_t_4;
            Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
            if (unlikely(size != 2)) {
              if (size > 2) __Pyx_RaiseTooManyValuesError(2);
              else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
              __PYX_ERR(0, 3408, __pyx_L1_error)
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            if (likely(PyTuple_CheckExact(sequence))) {
              __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
              __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
            } else {
              __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
              __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
            }
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_3);
            #else
            __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3408, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3408, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            #endif
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          } else {
            Py_ssize_t index = -1;
            __pyx_t_7 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3408, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_12 = Py_TYPE(__pyx_t_7)->tp_iternext;
            index = 0; __pyx_t_1 = __pyx_t_12(__pyx_t_7); if (unlikely(!__pyx_t_1)) goto __pyx_L14_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_1);
            index = 1; __pyx_t_3 = __pyx_t_12(__pyx_t_7); if (unlikely(!__pyx_t_3)) goto __pyx_L14_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_3);
            if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_7), 2) < 0) __PYX_ERR(0, 3408, __pyx_L1_error)
            __pyx_t_12 = NULL;
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            goto __pyx_L15_unpacking_done;
            __pyx_L14_unpacking_failed:;
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __pyx_t_12 = NULL;
            if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
            __PYX_ERR(0, 3408, __pyx_L1_error)
            __pyx_L15_unpacking_done:;
          }
          __Pyx_XDECREF_SET(__pyx_v_wslit, __pyx_t_1);
          __pyx_t_1 = 0;
          __Pyx_XDECREF_SET(__pyx_v_wschar, __pyx_t_3);
          __pyx_t_3 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":3409
 *                     }
 *                     for wslit, wschar in ws_map.items():
 *                         ret = ret.replace(wslit, wschar)             # <<<<<<<<<<<<<<
 * 
 *                 # replace escaped characters
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_ret, __pyx_n_s_replace); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3409, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_wslit, __pyx_v_wschar};
            __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3409, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_wslit, __pyx_v_wschar};
            __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3409, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          {
            __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3409, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (__pyx_t_1) {
              __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1); __pyx_t_1 = NULL;
            }
            __Pyx_INCREF(__pyx_v_wslit);
            __Pyx_GIVEREF(__pyx_v_wslit);
            PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_wslit);
            __Pyx_INCREF(__pyx_v_wschar);
            __Pyx_GIVEREF(__pyx_v_wschar);
            PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_wschar);
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3409, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF_SET(__pyx_v_ret, __pyx_t_4);
          __pyx_t_4 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":3408
 *                         r'\r': '\r',
 *                     }
 *                     for wslit, wschar in ws_map.items():             # <<<<<<<<<<<<<<
 *                         ret = ret.replace(wslit, wschar)
 * 
 */
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":3401
 *             if isinstance(ret, str):
 *                 # replace escaped whitespace
 *                 if '\\' in ret and self.convertWhitespaceEscapes:             # <<<<<<<<<<<<<<
 *                     ws_map = {
 *                         r'\t': '\t',
 */
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":3412
 * 
 *                 # replace escaped characters
 *                 if self.escChar:             # <<<<<<<<<<<<<<
 *                     ret = re.sub(self.escCharReplacePattern, r"\g<1>", ret)
 * 
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_escChar); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3412, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 3412, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (__pyx_t_5) {

        /* "Aspidites/_vendor/pyparsing_extension.py":3413
 *                 # replace escaped characters
 *                 if self.escChar:
 *                     ret = re.sub(self.escCharReplacePattern, r"\g<1>", ret)             # <<<<<<<<<<<<<<
 * 
 *                 # replace escaped quotes
 */
        __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_re); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3413, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_sub); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3413, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_escCharReplacePattern); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3413, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_7 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[4] = {__pyx_t_7, __pyx_t_4, __pyx_kp_s_g_1, __pyx_v_ret};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3413, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[4] = {__pyx_t_7, __pyx_t_4, __pyx_kp_s_g_1, __pyx_v_ret};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3413, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_1 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3413, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          if (__pyx_t_7) {
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_7); __pyx_t_7 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_6, __pyx_t_4);
          __Pyx_INCREF(__pyx_kp_s_g_1);
          __Pyx_GIVEREF(__pyx_kp_s_g_1);
          PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_6, __pyx_kp_s_g_1);
          __Pyx_INCREF(__pyx_v_ret);
          __Pyx_GIVEREF(__pyx_v_ret);
          PyTuple_SET_ITEM(__pyx_t_1, 2+__pyx_t_6, __pyx_v_ret);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3413, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF_SET(__pyx_v_ret, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":3412
 * 
 *                 # replace escaped characters
 *                 if self.escChar:             # <<<<<<<<<<<<<<
 *                     ret = re.sub(self.escCharReplacePattern, r"\g<1>", ret)
 * 
 */
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":3416
 * 
 *                 # replace escaped quotes
 *                 if self.escQuote:             # <<<<<<<<<<<<<<
 *                     ret = ret.replace(self.escQuote, self.endQuoteChar)
 * 
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_escQuote); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3416, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 3416, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (__pyx_t_5) {

        /* "Aspidites/_vendor/pyparsing_extension.py":3417
 *                 # replace escaped quotes
 *                 if self.escQuote:
 *                     ret = ret.replace(self.escQuote, self.endQuoteChar)             # <<<<<<<<<<<<<<
 * 
 *         return loc, ret
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_ret, __pyx_n_s_replace); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3417, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_escQuote); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3417, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_endQuoteChar); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3417, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_7 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_1, __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3417, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_1, __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3417, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_13 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3417, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_13);
          if (__pyx_t_7) {
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_7); __pyx_t_7 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_1);
          PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_6, __pyx_t_1);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_1 = 0;
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_13, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3417, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF_SET(__pyx_v_ret, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":3416
 * 
 *                 # replace escaped quotes
 *                 if self.escQuote:             # <<<<<<<<<<<<<<
 *                     ret = ret.replace(self.escQuote, self.endQuoteChar)
 * 
 */
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":3399
 *             ret = ret[self.quoteCharLen: -self.endQuoteCharLen]
 * 
 *             if isinstance(ret, str):             # <<<<<<<<<<<<<<
 *                 # replace escaped whitespace
 *                 if '\\' in ret and self.convertWhitespaceEscapes:
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":3394
 *         ret = result.group()
 * 
 *         if self.unquoteResults:             # <<<<<<<<<<<<<<
 * 
 *             # strip off quotes
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3419
 *                     ret = ret.replace(self.escQuote, self.endQuoteChar)
 * 
 *         return loc, ret             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3419, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_loc);
  __Pyx_GIVEREF(__pyx_v_loc);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_loc);
  __Pyx_INCREF(__pyx_v_ret);
  __Pyx_GIVEREF(__pyx_v_ret);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_ret);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3386
 *         self.mayReturnEmpty = True
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         result = instring[loc] == self.firstQuoteChar and self.re_match(instring, loc) or None
 *         if not result:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.QuotedString.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_XDECREF(__pyx_v_ws_map);
  __Pyx_XDECREF(__pyx_v_wslit);
  __Pyx_XDECREF(__pyx_v_wschar);
  __Pyx_XDECREF(__pyx_v_loc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3421
 *         return loc, ret
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         try:
 *             return super(QuotedString, self).__str__()
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12QuotedString_5__str__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12QuotedString_5__str__ = {"__str__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12QuotedString_5__str__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_12QuotedString_5__str__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12QuotedString_4__str__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_12QuotedString_4__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":3422
 * 
 *     def __str__(self):
 *         try:             # <<<<<<<<<<<<<<
 *             return super(QuotedString, self).__str__()
 *         except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "Aspidites/_vendor/pyparsing_extension.py":3423
 *     def __str__(self):
 *         try:
 *             return super(QuotedString, self).__str__()             # <<<<<<<<<<<<<<
 *         except Exception:
 *             pass
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_QuotedString); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3423, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3423, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5);
      __Pyx_INCREF(__pyx_v_self);
      __Pyx_GIVEREF(__pyx_v_self);
      PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_v_self);
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_6, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3423, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_str); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3423, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3423, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_r = __pyx_t_4;
      __pyx_t_4 = 0;
      goto __pyx_L7_try_return;

      /* "Aspidites/_vendor/pyparsing_extension.py":3422
 * 
 *     def __str__(self):
 *         try:             # <<<<<<<<<<<<<<
 *             return super(QuotedString, self).__str__()
 *         except Exception:
 */
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3424
 *         try:
 *             return super(QuotedString, self).__str__()
 *         except Exception:             # <<<<<<<<<<<<<<
 *             pass
 * 
 */
    __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_7) {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "Aspidites/_vendor/pyparsing_extension.py":3422
 * 
 *     def __str__(self):
 *         try:             # <<<<<<<<<<<<<<
 *             return super(QuotedString, self).__str__()
 *         except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3427
 *             pass
 * 
 *         if self.strRepr is None:             # <<<<<<<<<<<<<<
 *             self.strRepr = "quoted string, starting with %s ending with %s" % (self.quoteChar, self.endQuoteChar)
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_strRepr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_8 = (__pyx_t_4 == Py_None);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_9 = (__pyx_t_8 != 0);
  if (__pyx_t_9) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3428
 * 
 *         if self.strRepr is None:
 *             self.strRepr = "quoted string, starting with %s ending with %s" % (self.quoteChar, self.endQuoteChar)             # <<<<<<<<<<<<<<
 * 
 *         return self.strRepr
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_quoteChar); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3428, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_endQuoteChar); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3428, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3428, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_6);
    __pyx_t_4 = 0;
    __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyString_Format(__pyx_kp_s_quoted_string_starting_with_s_en, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3428, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_strRepr, __pyx_t_6) < 0) __PYX_ERR(0, 3428, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3427
 *             pass
 * 
 *         if self.strRepr is None:             # <<<<<<<<<<<<<<
 *             self.strRepr = "quoted string, starting with %s ending with %s" % (self.quoteChar, self.endQuoteChar)
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3430
 *             self.strRepr = "quoted string, starting with %s ending with %s" % (self.quoteChar, self.endQuoteChar)
 * 
 *         return self.strRepr             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_strRepr); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3430, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3421
 *         return loc, ret
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         try:
 *             return super(QuotedString, self).__str__()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.QuotedString.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3454
 *     """
 * 
 *     def __init__(self, notChars, min=1, max=0, exact=0):             # <<<<<<<<<<<<<<
 *         super(CharsNotIn, self).__init__()
 *         self.skipWhitespace = False
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10CharsNotIn_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10CharsNotIn_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10CharsNotIn_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10CharsNotIn_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_notChars = 0;
  PyObject *__pyx_v_min = 0;
  PyObject *__pyx_v_max = 0;
  PyObject *__pyx_v_exact = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_notChars,&__pyx_n_s_min,&__pyx_n_s_max,&__pyx_n_s_exact,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)__pyx_int_1));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_0));
    values[4] = ((PyObject *)((PyObject *)__pyx_int_0));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_notChars)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 5, 1); __PYX_ERR(0, 3454, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_exact);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 3454, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_notChars = values[1];
    __pyx_v_min = values[2];
    __pyx_v_max = values[3];
    __pyx_v_exact = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3454, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.CharsNotIn.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10CharsNotIn___init__(__pyx_self, __pyx_v_self, __pyx_v_notChars, __pyx_v_min, __pyx_v_max, __pyx_v_exact);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10CharsNotIn___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_notChars, PyObject *__pyx_v_min, PyObject *__pyx_v_max, PyObject *__pyx_v_exact) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":3455
 * 
 *     def __init__(self, notChars, min=1, max=0, exact=0):
 *         super(CharsNotIn, self).__init__()             # <<<<<<<<<<<<<<
 *         self.skipWhitespace = False
 *         self.notChars = notChars
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_CharsNotIn); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3456
 *     def __init__(self, notChars, min=1, max=0, exact=0):
 *         super(CharsNotIn, self).__init__()
 *         self.skipWhitespace = False             # <<<<<<<<<<<<<<
 *         self.notChars = notChars
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_skipWhitespace, Py_False) < 0) __PYX_ERR(0, 3456, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3457
 *         super(CharsNotIn, self).__init__()
 *         self.skipWhitespace = False
 *         self.notChars = notChars             # <<<<<<<<<<<<<<
 * 
 *         if min < 1:
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_notChars, __pyx_v_notChars) < 0) __PYX_ERR(0, 3457, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3459
 *         self.notChars = notChars
 * 
 *         if min < 1:             # <<<<<<<<<<<<<<
 *             raise ValueError("cannot specify a minimum length < 1; use "
 *                              "Optional(CharsNotIn()) if zero-length char group is permitted")
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_min, __pyx_int_1, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3459, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 3459, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_4)) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3460
 * 
 *         if min < 1:
 *             raise ValueError("cannot specify a minimum length < 1; use "             # <<<<<<<<<<<<<<
 *                              "Optional(CharsNotIn()) if zero-length char group is permitted")
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__104, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3460, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 3460, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":3459
 *         self.notChars = notChars
 * 
 *         if min < 1:             # <<<<<<<<<<<<<<
 *             raise ValueError("cannot specify a minimum length < 1; use "
 *                              "Optional(CharsNotIn()) if zero-length char group is permitted")
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3463
 *                              "Optional(CharsNotIn()) if zero-length char group is permitted")
 * 
 *         self.minLen = min             # <<<<<<<<<<<<<<
 * 
 *         if max > 0:
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_minLen, __pyx_v_min) < 0) __PYX_ERR(0, 3463, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3465
 *         self.minLen = min
 * 
 *         if max > 0:             # <<<<<<<<<<<<<<
 *             self.maxLen = max
 *         else:
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_max, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3465, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 3465, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3466
 * 
 *         if max > 0:
 *             self.maxLen = max             # <<<<<<<<<<<<<<
 *         else:
 *             self.maxLen = sys.maxsize
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_maxLen, __pyx_v_max) < 0) __PYX_ERR(0, 3466, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":3465
 *         self.minLen = min
 * 
 *         if max > 0:             # <<<<<<<<<<<<<<
 *             self.maxLen = max
 *         else:
 */
    goto __pyx_L4;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3468
 *             self.maxLen = max
 *         else:
 *             self.maxLen = sys.maxsize             # <<<<<<<<<<<<<<
 * 
 *         if exact > 0:
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_sys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3468, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_maxsize); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3468, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_maxLen, __pyx_t_3) < 0) __PYX_ERR(0, 3468, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_L4:;

  /* "Aspidites/_vendor/pyparsing_extension.py":3470
 *             self.maxLen = sys.maxsize
 * 
 *         if exact > 0:             # <<<<<<<<<<<<<<
 *             self.maxLen = exact
 *             self.minLen = exact
 */
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_exact, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3470, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 3470, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3471
 * 
 *         if exact > 0:
 *             self.maxLen = exact             # <<<<<<<<<<<<<<
 *             self.minLen = exact
 * 
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_maxLen, __pyx_v_exact) < 0) __PYX_ERR(0, 3471, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":3472
 *         if exact > 0:
 *             self.maxLen = exact
 *             self.minLen = exact             # <<<<<<<<<<<<<<
 * 
 *         self.name = str(self)
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_minLen, __pyx_v_exact) < 0) __PYX_ERR(0, 3472, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":3470
 *             self.maxLen = sys.maxsize
 * 
 *         if exact > 0:             # <<<<<<<<<<<<<<
 *             self.maxLen = exact
 *             self.minLen = exact
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3474
 *             self.minLen = exact
 * 
 *         self.name = str(self)             # <<<<<<<<<<<<<<
 *         self.errmsg = "Expected " + self.name
 *         self.mayReturnEmpty = (self.minLen == 0)
 */
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_name, __pyx_t_3) < 0) __PYX_ERR(0, 3474, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3475
 * 
 *         self.name = str(self)
 *         self.errmsg = "Expected " + self.name             # <<<<<<<<<<<<<<
 *         self.mayReturnEmpty = (self.minLen == 0)
 *         self.mayIndexError = False
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyNumber_Add(__pyx_kp_s_Expected, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_errmsg, __pyx_t_1) < 0) __PYX_ERR(0, 3475, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3476
 *         self.name = str(self)
 *         self.errmsg = "Expected " + self.name
 *         self.mayReturnEmpty = (self.minLen == 0)             # <<<<<<<<<<<<<<
 *         self.mayIndexError = False
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_minLen); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_EqObjC(__pyx_t_1, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_mayReturnEmpty, __pyx_t_3) < 0) __PYX_ERR(0, 3476, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3477
 *         self.errmsg = "Expected " + self.name
 *         self.mayReturnEmpty = (self.minLen == 0)
 *         self.mayIndexError = False             # <<<<<<<<<<<<<<
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_mayIndexError, Py_False) < 0) __PYX_ERR(0, 3477, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3454
 *     """
 * 
 *     def __init__(self, notChars, min=1, max=0, exact=0):             # <<<<<<<<<<<<<<
 *         super(CharsNotIn, self).__init__()
 *         self.skipWhitespace = False
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.CharsNotIn.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3479
 *         self.mayIndexError = False
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if instring[loc] in self.notChars:
 *             raise ParseException(instring, loc, self.errmsg, self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10CharsNotIn_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10CharsNotIn_3parseImpl = {"parseImpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10CharsNotIn_3parseImpl, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10CharsNotIn_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  CYTHON_UNUSED PyObject *__pyx_v_doActions = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parseImpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_doActions,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 1); __PYX_ERR(0, 3479, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 2); __PYX_ERR(0, 3479, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_doActions);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "parseImpl") < 0)) __PYX_ERR(0, 3479, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_doActions = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3479, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.CharsNotIn.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10CharsNotIn_2parseImpl(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10CharsNotIn_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions) {
  PyObject *__pyx_v_start = NULL;
  PyObject *__pyx_v_notchars = NULL;
  PyObject *__pyx_v_maxlen = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parseImpl", 0);
  __Pyx_INCREF(__pyx_v_loc);

  /* "Aspidites/_vendor/pyparsing_extension.py":3480
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if instring[loc] in self.notChars:             # <<<<<<<<<<<<<<
 *             raise ParseException(instring, loc, self.errmsg, self)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_instring, __pyx_v_loc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3480, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_notChars); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3480, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_t_2, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 3480, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (unlikely(__pyx_t_4)) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3481
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if instring[loc] in self.notChars:
 *             raise ParseException(instring, loc, self.errmsg, self)             # <<<<<<<<<<<<<<
 * 
 *         start = loc
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3481, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errmsg); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3481, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[5] = {__pyx_t_6, __pyx_v_instring, __pyx_v_loc, __pyx_t_5, __pyx_v_self};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3481, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[5] = {__pyx_t_6, __pyx_v_instring, __pyx_v_loc, __pyx_t_5, __pyx_v_self};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3481, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(4+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3481, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_INCREF(__pyx_v_instring);
      __Pyx_GIVEREF(__pyx_v_instring);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_v_instring);
      __Pyx_INCREF(__pyx_v_loc);
      __Pyx_GIVEREF(__pyx_v_loc);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_v_loc);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_7, __pyx_t_5);
      __Pyx_INCREF(__pyx_v_self);
      __Pyx_GIVEREF(__pyx_v_self);
      PyTuple_SET_ITEM(__pyx_t_8, 3+__pyx_t_7, __pyx_v_self);
      __pyx_t_5 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3481, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 3481, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":3480
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if instring[loc] in self.notChars:             # <<<<<<<<<<<<<<
 *             raise ParseException(instring, loc, self.errmsg, self)
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3483
 *             raise ParseException(instring, loc, self.errmsg, self)
 * 
 *         start = loc             # <<<<<<<<<<<<<<
 *         loc += 1
 *         notchars = self.notChars
 */
  __Pyx_INCREF(__pyx_v_loc);
  __pyx_v_start = __pyx_v_loc;

  /* "Aspidites/_vendor/pyparsing_extension.py":3484
 * 
 *         start = loc
 *         loc += 1             # <<<<<<<<<<<<<<
 *         notchars = self.notChars
 *         maxlen = min(start + self.maxLen, len(instring))
 */
  __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_v_loc, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF_SET(__pyx_v_loc, __pyx_t_2);
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3485
 *         start = loc
 *         loc += 1
 *         notchars = self.notChars             # <<<<<<<<<<<<<<
 *         maxlen = min(start + self.maxLen, len(instring))
 *         while loc < maxlen and instring[loc] not in notchars:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_notChars); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_notchars = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3486
 *         loc += 1
 *         notchars = self.notChars
 *         maxlen = min(start + self.maxLen, len(instring))             # <<<<<<<<<<<<<<
 *         while loc < maxlen and instring[loc] not in notchars:
 *             loc += 1
 */
  __pyx_t_9 = PyObject_Length(__pyx_v_instring); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 3486, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_maxLen); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyNumber_Add(__pyx_v_start, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_8 = PyInt_FromSsize_t(__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_5 = PyObject_RichCompare(__pyx_t_8, __pyx_t_1, Py_LT); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3486, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 3486, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_4) {
    __pyx_t_5 = PyInt_FromSsize_t(__pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3486, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = __pyx_t_5;
    __pyx_t_5 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = __pyx_t_1;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_t_2;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_maxlen = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3487
 *         notchars = self.notChars
 *         maxlen = min(start + self.maxLen, len(instring))
 *         while loc < maxlen and instring[loc] not in notchars:             # <<<<<<<<<<<<<<
 *             loc += 1
 * 
 */
  while (1) {
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_loc, __pyx_v_maxlen, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3487, __pyx_L1_error)
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 3487, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_3) {
    } else {
      __pyx_t_4 = __pyx_t_3;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_instring, __pyx_v_loc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3487, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_v_notchars, Py_NE)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 3487, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_10 = (__pyx_t_3 != 0);
    __pyx_t_4 = __pyx_t_10;
    __pyx_L6_bool_binop_done:;
    if (!__pyx_t_4) break;

    /* "Aspidites/_vendor/pyparsing_extension.py":3488
 *         maxlen = min(start + self.maxLen, len(instring))
 *         while loc < maxlen and instring[loc] not in notchars:
 *             loc += 1             # <<<<<<<<<<<<<<
 * 
 *         if loc - start < self.minLen:
 */
    __pyx_t_1 = __Pyx_PyInt_AddObjC(__pyx_v_loc, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3488, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_loc, __pyx_t_1);
    __pyx_t_1 = 0;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3490
 *             loc += 1
 * 
 *         if loc - start < self.minLen:             # <<<<<<<<<<<<<<
 *             raise ParseException(instring, loc, self.errmsg, self)
 * 
 */
  __pyx_t_1 = PyNumber_Subtract(__pyx_v_loc, __pyx_v_start); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_minLen); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_LT); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3490, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 3490, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(__pyx_t_4)) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3491
 * 
 *         if loc - start < self.minLen:
 *             raise ParseException(instring, loc, self.errmsg, self)             # <<<<<<<<<<<<<<
 * 
 *         return loc, instring[start:loc]
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3491, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errmsg); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3491, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[5] = {__pyx_t_8, __pyx_v_instring, __pyx_v_loc, __pyx_t_1, __pyx_v_self};
      __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3491, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[5] = {__pyx_t_8, __pyx_v_instring, __pyx_v_loc, __pyx_t_1, __pyx_v_self};
      __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3491, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(4+__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3491, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_INCREF(__pyx_v_instring);
      __Pyx_GIVEREF(__pyx_v_instring);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_7, __pyx_v_instring);
      __Pyx_INCREF(__pyx_v_loc);
      __Pyx_GIVEREF(__pyx_v_loc);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_7, __pyx_v_loc);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_7, __pyx_t_1);
      __Pyx_INCREF(__pyx_v_self);
      __Pyx_GIVEREF(__pyx_v_self);
      PyTuple_SET_ITEM(__pyx_t_6, 3+__pyx_t_7, __pyx_v_self);
      __pyx_t_1 = 0;
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3491, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 3491, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":3490
 *             loc += 1
 * 
 *         if loc - start < self.minLen:             # <<<<<<<<<<<<<<
 *             raise ParseException(instring, loc, self.errmsg, self)
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3493
 *             raise ParseException(instring, loc, self.errmsg, self)
 * 
 *         return loc, instring[start:loc]             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = __Pyx_PyObject_GetSlice(__pyx_v_instring, 0, 0, &__pyx_v_start, &__pyx_v_loc, NULL, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_loc);
  __Pyx_GIVEREF(__pyx_v_loc);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_loc);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3479
 *         self.mayIndexError = False
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if instring[loc] in self.notChars:
 *             raise ParseException(instring, loc, self.errmsg, self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.CharsNotIn.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_start);
  __Pyx_XDECREF(__pyx_v_notchars);
  __Pyx_XDECREF(__pyx_v_maxlen);
  __Pyx_XDECREF(__pyx_v_loc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3495
 *         return loc, instring[start:loc]
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         try:
 *             return super(CharsNotIn, self).__str__()
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10CharsNotIn_5__str__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10CharsNotIn_5__str__ = {"__str__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10CharsNotIn_5__str__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10CharsNotIn_5__str__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10CharsNotIn_4__str__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10CharsNotIn_4__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":3496
 * 
 *     def __str__(self):
 *         try:             # <<<<<<<<<<<<<<
 *             return super(CharsNotIn, self).__str__()
 *         except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "Aspidites/_vendor/pyparsing_extension.py":3497
 *     def __str__(self):
 *         try:
 *             return super(CharsNotIn, self).__str__()             # <<<<<<<<<<<<<<
 *         except Exception:
 *             pass
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_CharsNotIn); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3497, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3497, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5);
      __Pyx_INCREF(__pyx_v_self);
      __Pyx_GIVEREF(__pyx_v_self);
      PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_v_self);
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_6, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3497, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_str); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3497, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3497, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_r = __pyx_t_4;
      __pyx_t_4 = 0;
      goto __pyx_L7_try_return;

      /* "Aspidites/_vendor/pyparsing_extension.py":3496
 * 
 *     def __str__(self):
 *         try:             # <<<<<<<<<<<<<<
 *             return super(CharsNotIn, self).__str__()
 *         except Exception:
 */
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3498
 *         try:
 *             return super(CharsNotIn, self).__str__()
 *         except Exception:             # <<<<<<<<<<<<<<
 *             pass
 * 
 */
    __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_7) {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "Aspidites/_vendor/pyparsing_extension.py":3496
 * 
 *     def __str__(self):
 *         try:             # <<<<<<<<<<<<<<
 *             return super(CharsNotIn, self).__str__()
 *         except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3501
 *             pass
 * 
 *         if self.strRepr is None:             # <<<<<<<<<<<<<<
 *             if len(self.notChars) > 4:
 *                 self.strRepr = "!W:(%s...)" % self.notChars[:4]
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_strRepr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_8 = (__pyx_t_4 == Py_None);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_9 = (__pyx_t_8 != 0);
  if (__pyx_t_9) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3502
 * 
 *         if self.strRepr is None:
 *             if len(self.notChars) > 4:             # <<<<<<<<<<<<<<
 *                 self.strRepr = "!W:(%s...)" % self.notChars[:4]
 *             else:
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_notChars); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3502, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_10 = PyObject_Length(__pyx_t_4); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(0, 3502, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = ((__pyx_t_10 > 4) != 0);
    if (__pyx_t_9) {

      /* "Aspidites/_vendor/pyparsing_extension.py":3503
 *         if self.strRepr is None:
 *             if len(self.notChars) > 4:
 *                 self.strRepr = "!W:(%s...)" % self.notChars[:4]             # <<<<<<<<<<<<<<
 *             else:
 *                 self.strRepr = "!W:(%s)" % self.notChars
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_notChars); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3503, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = __Pyx_PyObject_GetSlice(__pyx_t_4, 0, 4, NULL, NULL, &__pyx_slice__84, 0, 1, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3503, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyString_FormatSafe(__pyx_kp_s_W_s_2, __pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3503, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_strRepr, __pyx_t_4) < 0) __PYX_ERR(0, 3503, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":3502
 * 
 *         if self.strRepr is None:
 *             if len(self.notChars) > 4:             # <<<<<<<<<<<<<<
 *                 self.strRepr = "!W:(%s...)" % self.notChars[:4]
 *             else:
 */
      goto __pyx_L10;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":3505
 *                 self.strRepr = "!W:(%s...)" % self.notChars[:4]
 *             else:
 *                 self.strRepr = "!W:(%s)" % self.notChars             # <<<<<<<<<<<<<<
 * 
 *         return self.strRepr
 */
    /*else*/ {
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_notChars); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3505, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = __Pyx_PyString_FormatSafe(__pyx_kp_s_W_s_3, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3505, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_strRepr, __pyx_t_6) < 0) __PYX_ERR(0, 3505, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __pyx_L10:;

    /* "Aspidites/_vendor/pyparsing_extension.py":3501
 *             pass
 * 
 *         if self.strRepr is None:             # <<<<<<<<<<<<<<
 *             if len(self.notChars) > 4:
 *                 self.strRepr = "!W:(%s...)" % self.notChars[:4]
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3507
 *                 self.strRepr = "!W:(%s)" % self.notChars
 * 
 *         return self.strRepr             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_strRepr); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3507, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3495
 *         return loc, instring[start:loc]
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         try:
 *             return super(CharsNotIn, self).__str__()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.CharsNotIn.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3545
 *     }
 * 
 *     def __init__(self, ws=" \t\r\n", min=1, max=0, exact=0):             # <<<<<<<<<<<<<<
 *         super(White, self).__init__()
 *         self.matchWhite = ws
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5White_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_5White_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5White_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5White_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_ws = 0;
  PyObject *__pyx_v_min = 0;
  PyObject *__pyx_v_max = 0;
  PyObject *__pyx_v_exact = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ws,&__pyx_n_s_min,&__pyx_n_s_max,&__pyx_n_s_exact,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[1] = ((PyObject *)((PyObject*)__pyx_kp_s__105));
    values[2] = ((PyObject *)((PyObject *)__pyx_int_1));
    values[3] = ((PyObject *)((PyObject *)__pyx_int_0));
    values[4] = ((PyObject *)((PyObject *)__pyx_int_0));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ws);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_exact);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 3545, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_ws = values[1];
    __pyx_v_min = values[2];
    __pyx_v_max = values[3];
    __pyx_v_exact = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 1, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3545, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.White.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5White___init__(__pyx_self, __pyx_v_self, __pyx_v_ws, __pyx_v_min, __pyx_v_max, __pyx_v_exact);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_5White_8__init___2generator19(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/pyparsing_extension.py":3548
 *         super(White, self).__init__()
 *         self.matchWhite = ws
 *         self.setWhitespaceChars("".join(c for c in self.whiteChars if c not in self.matchWhite))             # <<<<<<<<<<<<<<
 *         # ~ self.leaveWhitespace()
 *         self.name = ("".join(White.whiteStrs[c] for c in self.matchWhite))
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5White_8__init___genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 3548, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__ *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_5White_8__init___2generator19, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_White___init___locals_genexpr, __pyx_n_s_Aspidites__vendor_pyparsing_exte); if (unlikely(!gen)) __PYX_ERR(0, 3548, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.White.__init__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_5White_8__init___2generator19(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L7_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 3548, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 3548, __pyx_L1_error) }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_whiteChars); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3548, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3548, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 3548, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3548, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 3548, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3548, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 3548, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_c);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_c, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 3548, __pyx_L1_error) }
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_matchWhite); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3548, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = (__Pyx_PySequence_ContainsTF(__pyx_cur_scope->__pyx_v_c, __pyx_t_1, Py_NE)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 3548, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = (__pyx_t_5 != 0);
    if (__pyx_t_6) {
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_c);
      __pyx_r = __pyx_cur_scope->__pyx_v_c;
      __Pyx_XGIVEREF(__pyx_t_2);
      __pyx_cur_scope->__pyx_t_0 = __pyx_t_2;
      __pyx_cur_scope->__pyx_t_1 = __pyx_t_3;
      __pyx_cur_scope->__pyx_t_2 = __pyx_t_4;
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
      /* return from generator, yielding value */
      __pyx_generator->resume_label = 1;
      return __pyx_r;
      __pyx_L7_resume_from_yield:;
      __pyx_t_2 = __pyx_cur_scope->__pyx_t_0;
      __pyx_cur_scope->__pyx_t_0 = 0;
      __Pyx_XGOTREF(__pyx_t_2);
      __pyx_t_3 = __pyx_cur_scope->__pyx_t_1;
      __pyx_t_4 = __pyx_cur_scope->__pyx_t_2;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 3548, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_5White_8__init___5generator20(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/pyparsing_extension.py":3550
 *         self.setWhitespaceChars("".join(c for c in self.whiteChars if c not in self.matchWhite))
 *         # ~ self.leaveWhitespace()
 *         self.name = ("".join(White.whiteStrs[c] for c in self.matchWhite))             # <<<<<<<<<<<<<<
 *         self.mayReturnEmpty = True
 *         self.errmsg = "Expected " + self.name
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5White_8__init___3genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 3550, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__ *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_5White_8__init___5generator20, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_White___init___locals_genexpr, __pyx_n_s_Aspidites__vendor_pyparsing_exte); if (unlikely(!gen)) __PYX_ERR(0, 3550, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.White.__init__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_5White_8__init___5generator20(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 3550, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 3550, __pyx_L1_error) }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_matchWhite); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3550, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3550, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 3550, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3550, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 3550, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3550, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 3550, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_c);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_c, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_White); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3550, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_whiteStrs); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3550, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_t_5, __pyx_cur_scope->__pyx_v_c); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3550, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    __Pyx_XGIVEREF(__pyx_t_2);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_3;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_4;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_2);
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 3550, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3545
 *     }
 * 
 *     def __init__(self, ws=" \t\r\n", min=1, max=0, exact=0):             # <<<<<<<<<<<<<<
 *         super(White, self).__init__()
 *         self.matchWhite = ws
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5White___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_ws, PyObject *__pyx_v_min, PyObject *__pyx_v_max, PyObject *__pyx_v_exact) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__ *__pyx_cur_scope;
  PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_5White_8__init___2generator19 = 0;
  PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_5White_8__init___5generator20 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__ *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 3545, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);

  /* "Aspidites/_vendor/pyparsing_extension.py":3546
 * 
 *     def __init__(self, ws=" \t\r\n", min=1, max=0, exact=0):
 *         super(White, self).__init__()             # <<<<<<<<<<<<<<
 *         self.matchWhite = ws
 *         self.setWhitespaceChars("".join(c for c in self.whiteChars if c not in self.matchWhite))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_White); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_cur_scope->__pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3547
 *     def __init__(self, ws=" \t\r\n", min=1, max=0, exact=0):
 *         super(White, self).__init__()
 *         self.matchWhite = ws             # <<<<<<<<<<<<<<
 *         self.setWhitespaceChars("".join(c for c in self.whiteChars if c not in self.matchWhite))
 *         # ~ self.leaveWhitespace()
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_matchWhite, __pyx_v_ws) < 0) __PYX_ERR(0, 3547, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3548
 *         super(White, self).__init__()
 *         self.matchWhite = ws
 *         self.setWhitespaceChars("".join(c for c in self.whiteChars if c not in self.matchWhite))             # <<<<<<<<<<<<<<
 *         # ~ self.leaveWhitespace()
 *         self.name = ("".join(White.whiteStrs[c] for c in self.matchWhite))
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_setWhitespaceChars); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5White_8__init___genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyString_Join(__pyx_kp_s__12, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3550
 *         self.setWhitespaceChars("".join(c for c in self.whiteChars if c not in self.matchWhite))
 *         # ~ self.leaveWhitespace()
 *         self.name = ("".join(White.whiteStrs[c] for c in self.matchWhite))             # <<<<<<<<<<<<<<
 *         self.mayReturnEmpty = True
 *         self.errmsg = "Expected " + self.name
 */
  __pyx_t_1 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5White_8__init___3genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyString_Join(__pyx_kp_s__12, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_name, __pyx_t_3) < 0) __PYX_ERR(0, 3550, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3551
 *         # ~ self.leaveWhitespace()
 *         self.name = ("".join(White.whiteStrs[c] for c in self.matchWhite))
 *         self.mayReturnEmpty = True             # <<<<<<<<<<<<<<
 *         self.errmsg = "Expected " + self.name
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_mayReturnEmpty, Py_True) < 0) __PYX_ERR(0, 3551, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3552
 *         self.name = ("".join(White.whiteStrs[c] for c in self.matchWhite))
 *         self.mayReturnEmpty = True
 *         self.errmsg = "Expected " + self.name             # <<<<<<<<<<<<<<
 * 
 *         self.minLen = min
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyNumber_Add(__pyx_kp_s_Expected, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_errmsg, __pyx_t_1) < 0) __PYX_ERR(0, 3552, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3554
 *         self.errmsg = "Expected " + self.name
 * 
 *         self.minLen = min             # <<<<<<<<<<<<<<
 * 
 *         if max > 0:
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_minLen, __pyx_v_min) < 0) __PYX_ERR(0, 3554, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3556
 *         self.minLen = min
 * 
 *         if max > 0:             # <<<<<<<<<<<<<<
 *             self.maxLen = max
 *         else:
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_max, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3556, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 3556, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3557
 * 
 *         if max > 0:
 *             self.maxLen = max             # <<<<<<<<<<<<<<
 *         else:
 *             self.maxLen = sys.maxsize
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_maxLen, __pyx_v_max) < 0) __PYX_ERR(0, 3557, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":3556
 *         self.minLen = min
 * 
 *         if max > 0:             # <<<<<<<<<<<<<<
 *             self.maxLen = max
 *         else:
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3559
 *             self.maxLen = max
 *         else:
 *             self.maxLen = sys.maxsize             # <<<<<<<<<<<<<<
 * 
 *         if exact > 0:
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_sys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3559, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_maxsize); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3559, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_maxLen, __pyx_t_3) < 0) __PYX_ERR(0, 3559, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_L3:;

  /* "Aspidites/_vendor/pyparsing_extension.py":3561
 *             self.maxLen = sys.maxsize
 * 
 *         if exact > 0:             # <<<<<<<<<<<<<<
 *             self.maxLen = exact
 *             self.minLen = exact
 */
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_exact, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3561, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 3561, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3562
 * 
 *         if exact > 0:
 *             self.maxLen = exact             # <<<<<<<<<<<<<<
 *             self.minLen = exact
 * 
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_maxLen, __pyx_v_exact) < 0) __PYX_ERR(0, 3562, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":3563
 *         if exact > 0:
 *             self.maxLen = exact
 *             self.minLen = exact             # <<<<<<<<<<<<<<
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_minLen, __pyx_v_exact) < 0) __PYX_ERR(0, 3563, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":3561
 *             self.maxLen = sys.maxsize
 * 
 *         if exact > 0:             # <<<<<<<<<<<<<<
 *             self.maxLen = exact
 *             self.minLen = exact
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3545
 *     }
 * 
 *     def __init__(self, ws=" \t\r\n", min=1, max=0, exact=0):             # <<<<<<<<<<<<<<
 *         super(White, self).__init__()
 *         self.matchWhite = ws
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.White.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_5White_8__init___2generator19);
  __Pyx_XDECREF(__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_5White_8__init___5generator20);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3565
 *             self.minLen = exact
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if instring[loc] not in self.matchWhite:
 *             raise ParseException(instring, loc, self.errmsg, self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5White_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_5White_3parseImpl = {"parseImpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5White_3parseImpl, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5White_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  CYTHON_UNUSED PyObject *__pyx_v_doActions = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parseImpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_doActions,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 1); __PYX_ERR(0, 3565, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 2); __PYX_ERR(0, 3565, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_doActions);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "parseImpl") < 0)) __PYX_ERR(0, 3565, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_doActions = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3565, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.White.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5White_2parseImpl(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5White_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions) {
  PyObject *__pyx_v_start = NULL;
  PyObject *__pyx_v_maxloc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parseImpl", 0);
  __Pyx_INCREF(__pyx_v_loc);

  /* "Aspidites/_vendor/pyparsing_extension.py":3566
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if instring[loc] not in self.matchWhite:             # <<<<<<<<<<<<<<
 *             raise ParseException(instring, loc, self.errmsg, self)
 *         start = loc
 */
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_instring, __pyx_v_loc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_matchWhite); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_t_2, Py_NE)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 3566, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (unlikely(__pyx_t_4)) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3567
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if instring[loc] not in self.matchWhite:
 *             raise ParseException(instring, loc, self.errmsg, self)             # <<<<<<<<<<<<<<
 *         start = loc
 *         loc += 1
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3567, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errmsg); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3567, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[5] = {__pyx_t_6, __pyx_v_instring, __pyx_v_loc, __pyx_t_5, __pyx_v_self};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3567, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[5] = {__pyx_t_6, __pyx_v_instring, __pyx_v_loc, __pyx_t_5, __pyx_v_self};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3567, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(4+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3567, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_INCREF(__pyx_v_instring);
      __Pyx_GIVEREF(__pyx_v_instring);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_v_instring);
      __Pyx_INCREF(__pyx_v_loc);
      __Pyx_GIVEREF(__pyx_v_loc);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_v_loc);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_7, __pyx_t_5);
      __Pyx_INCREF(__pyx_v_self);
      __Pyx_GIVEREF(__pyx_v_self);
      PyTuple_SET_ITEM(__pyx_t_8, 3+__pyx_t_7, __pyx_v_self);
      __pyx_t_5 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3567, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 3567, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":3566
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if instring[loc] not in self.matchWhite:             # <<<<<<<<<<<<<<
 *             raise ParseException(instring, loc, self.errmsg, self)
 *         start = loc
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3568
 *         if instring[loc] not in self.matchWhite:
 *             raise ParseException(instring, loc, self.errmsg, self)
 *         start = loc             # <<<<<<<<<<<<<<
 *         loc += 1
 *         maxloc = start + self.maxLen
 */
  __Pyx_INCREF(__pyx_v_loc);
  __pyx_v_start = __pyx_v_loc;

  /* "Aspidites/_vendor/pyparsing_extension.py":3569
 *             raise ParseException(instring, loc, self.errmsg, self)
 *         start = loc
 *         loc += 1             # <<<<<<<<<<<<<<
 *         maxloc = start + self.maxLen
 *         maxloc = min(maxloc, len(instring))
 */
  __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_v_loc, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF_SET(__pyx_v_loc, __pyx_t_2);
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3570
 *         start = loc
 *         loc += 1
 *         maxloc = start + self.maxLen             # <<<<<<<<<<<<<<
 *         maxloc = min(maxloc, len(instring))
 *         while loc < maxloc and instring[loc] in self.matchWhite:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_maxLen); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyNumber_Add(__pyx_v_start, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_maxloc = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3571
 *         loc += 1
 *         maxloc = start + self.maxLen
 *         maxloc = min(maxloc, len(instring))             # <<<<<<<<<<<<<<
 *         while loc < maxloc and instring[loc] in self.matchWhite:
 *             loc += 1
 */
  __pyx_t_9 = PyObject_Length(__pyx_v_instring); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 3571, __pyx_L1_error)
  __Pyx_INCREF(__pyx_v_maxloc);
  __pyx_t_1 = __pyx_v_maxloc;
  __pyx_t_8 = PyInt_FromSsize_t(__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3571, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_5 = PyObject_RichCompare(__pyx_t_8, __pyx_t_1, Py_LT); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3571, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 3571, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_4) {
    __pyx_t_5 = PyInt_FromSsize_t(__pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3571, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = __pyx_t_5;
    __pyx_t_5 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = __pyx_t_1;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_t_2;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_maxloc, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3572
 *         maxloc = start + self.maxLen
 *         maxloc = min(maxloc, len(instring))
 *         while loc < maxloc and instring[loc] in self.matchWhite:             # <<<<<<<<<<<<<<
 *             loc += 1
 * 
 */
  while (1) {
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_loc, __pyx_v_maxloc, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3572, __pyx_L1_error)
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 3572, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_3) {
    } else {
      __pyx_t_4 = __pyx_t_3;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_instring, __pyx_v_loc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3572, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_matchWhite); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3572, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_t_2, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 3572, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_10 = (__pyx_t_3 != 0);
    __pyx_t_4 = __pyx_t_10;
    __pyx_L6_bool_binop_done:;
    if (!__pyx_t_4) break;

    /* "Aspidites/_vendor/pyparsing_extension.py":3573
 *         maxloc = min(maxloc, len(instring))
 *         while loc < maxloc and instring[loc] in self.matchWhite:
 *             loc += 1             # <<<<<<<<<<<<<<
 * 
 *         if loc - start < self.minLen:
 */
    __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_v_loc, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3573, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_loc, __pyx_t_2);
    __pyx_t_2 = 0;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3575
 *             loc += 1
 * 
 *         if loc - start < self.minLen:             # <<<<<<<<<<<<<<
 *             raise ParseException(instring, loc, self.errmsg, self)
 * 
 */
  __pyx_t_2 = PyNumber_Subtract(__pyx_v_loc, __pyx_v_start); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3575, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_minLen); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3575, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyObject_RichCompare(__pyx_t_2, __pyx_t_1, Py_LT); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3575, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 3575, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(__pyx_t_4)) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3576
 * 
 *         if loc - start < self.minLen:
 *             raise ParseException(instring, loc, self.errmsg, self)             # <<<<<<<<<<<<<<
 * 
 *         return loc, instring[start:loc]
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3576, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errmsg); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3576, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_8 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[5] = {__pyx_t_8, __pyx_v_instring, __pyx_v_loc, __pyx_t_2, __pyx_v_self};
      __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3576, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[5] = {__pyx_t_8, __pyx_v_instring, __pyx_v_loc, __pyx_t_2, __pyx_v_self};
      __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3576, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(4+__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3576, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_INCREF(__pyx_v_instring);
      __Pyx_GIVEREF(__pyx_v_instring);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_7, __pyx_v_instring);
      __Pyx_INCREF(__pyx_v_loc);
      __Pyx_GIVEREF(__pyx_v_loc);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_7, __pyx_v_loc);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_7, __pyx_t_2);
      __Pyx_INCREF(__pyx_v_self);
      __Pyx_GIVEREF(__pyx_v_self);
      PyTuple_SET_ITEM(__pyx_t_6, 3+__pyx_t_7, __pyx_v_self);
      __pyx_t_2 = 0;
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_6, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3576, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 3576, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":3575
 *             loc += 1
 * 
 *         if loc - start < self.minLen:             # <<<<<<<<<<<<<<
 *             raise ParseException(instring, loc, self.errmsg, self)
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3578
 *             raise ParseException(instring, loc, self.errmsg, self)
 * 
 *         return loc, instring[start:loc]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = __Pyx_PyObject_GetSlice(__pyx_v_instring, 0, 0, &__pyx_v_start, &__pyx_v_loc, NULL, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3578, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3578, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_loc);
  __Pyx_GIVEREF(__pyx_v_loc);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_loc);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3565
 *             self.minLen = exact
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if instring[loc] not in self.matchWhite:
 *             raise ParseException(instring, loc, self.errmsg, self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.White.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_start);
  __Pyx_XDECREF(__pyx_v_maxloc);
  __Pyx_XDECREF(__pyx_v_loc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3582
 * 
 * class _PositionToken(Token):
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         super(_PositionToken, self).__init__()
 *         self.name = self.__class__.__name__
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_14_PositionToken_1__init__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_14_PositionToken_1__init__ = {"__init__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_14_PositionToken_1__init__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_14_PositionToken_1__init__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_14_PositionToken___init__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_14_PositionToken___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":3583
 * class _PositionToken(Token):
 *     def __init__(self):
 *         super(_PositionToken, self).__init__()             # <<<<<<<<<<<<<<
 *         self.name = self.__class__.__name__
 *         self.mayReturnEmpty = True
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_PositionToken); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3584
 *     def __init__(self):
 *         super(_PositionToken, self).__init__()
 *         self.name = self.__class__.__name__             # <<<<<<<<<<<<<<
 *         self.mayReturnEmpty = True
 *         self.mayIndexError = False
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_name_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_name, __pyx_t_3) < 0) __PYX_ERR(0, 3584, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3585
 *         super(_PositionToken, self).__init__()
 *         self.name = self.__class__.__name__
 *         self.mayReturnEmpty = True             # <<<<<<<<<<<<<<
 *         self.mayIndexError = False
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_mayReturnEmpty, Py_True) < 0) __PYX_ERR(0, 3585, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3586
 *         self.name = self.__class__.__name__
 *         self.mayReturnEmpty = True
 *         self.mayIndexError = False             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_mayIndexError, Py_False) < 0) __PYX_ERR(0, 3586, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3582
 * 
 * class _PositionToken(Token):
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         super(_PositionToken, self).__init__()
 *         self.name = self.__class__.__name__
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._PositionToken.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3594
 *     """
 * 
 *     def __init__(self, colno):             # <<<<<<<<<<<<<<
 *         super(GoToColumn, self).__init__()
 *         self.col = colno
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10GoToColumn_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10GoToColumn_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10GoToColumn_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10GoToColumn_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_colno = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_colno,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_colno)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(0, 3594, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 3594, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_colno = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3594, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.GoToColumn.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10GoToColumn___init__(__pyx_self, __pyx_v_self, __pyx_v_colno);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10GoToColumn___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_colno) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":3595
 * 
 *     def __init__(self, colno):
 *         super(GoToColumn, self).__init__()             # <<<<<<<<<<<<<<
 *         self.col = colno
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_GoToColumn); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3596
 *     def __init__(self, colno):
 *         super(GoToColumn, self).__init__()
 *         self.col = colno             # <<<<<<<<<<<<<<
 * 
 *     def preParse(self, instring, loc):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_col, __pyx_v_colno) < 0) __PYX_ERR(0, 3596, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3594
 *     """
 * 
 *     def __init__(self, colno):             # <<<<<<<<<<<<<<
 *         super(GoToColumn, self).__init__()
 *         self.col = colno
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.GoToColumn.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3598
 *         self.col = colno
 * 
 *     def preParse(self, instring, loc):             # <<<<<<<<<<<<<<
 *         if col(loc, instring) != self.col:
 *             instrlen = len(instring)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10GoToColumn_3preParse(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10GoToColumn_3preParse = {"preParse", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10GoToColumn_3preParse, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10GoToColumn_3preParse(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("preParse (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("preParse", 1, 3, 3, 1); __PYX_ERR(0, 3598, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("preParse", 1, 3, 3, 2); __PYX_ERR(0, 3598, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "preParse") < 0)) __PYX_ERR(0, 3598, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("preParse", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3598, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.GoToColumn.preParse", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10GoToColumn_2preParse(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10GoToColumn_2preParse(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc) {
  Py_ssize_t __pyx_v_instrlen;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("preParse", 0);
  __Pyx_INCREF(__pyx_v_loc);

  /* "Aspidites/_vendor/pyparsing_extension.py":3599
 * 
 *     def preParse(self, instring, loc):
 *         if col(loc, instring) != self.col:             # <<<<<<<<<<<<<<
 *             instrlen = len(instring)
 *             if self.ignoreExprs:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_col); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_loc, __pyx_v_instring};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3599, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_loc, __pyx_v_instring};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3599, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3599, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_loc);
    __Pyx_GIVEREF(__pyx_v_loc);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_loc);
    __Pyx_INCREF(__pyx_v_instring);
    __Pyx_GIVEREF(__pyx_v_instring);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_instring);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3599, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_col); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_NE); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3599, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 3599, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_6) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3600
 *     def preParse(self, instring, loc):
 *         if col(loc, instring) != self.col:
 *             instrlen = len(instring)             # <<<<<<<<<<<<<<
 *             if self.ignoreExprs:
 *                 loc = self._skipIgnorables(instring, loc)
 */
    __pyx_t_7 = PyObject_Length(__pyx_v_instring); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 3600, __pyx_L1_error)
    __pyx_v_instrlen = __pyx_t_7;

    /* "Aspidites/_vendor/pyparsing_extension.py":3601
 *         if col(loc, instring) != self.col:
 *             instrlen = len(instring)
 *             if self.ignoreExprs:             # <<<<<<<<<<<<<<
 *                 loc = self._skipIgnorables(instring, loc)
 *             while loc < instrlen and instring[loc].isspace() and col(loc, instring) != self.col:
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ignoreExprs); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3601, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 3601, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_6) {

      /* "Aspidites/_vendor/pyparsing_extension.py":3602
 *             instrlen = len(instring)
 *             if self.ignoreExprs:
 *                 loc = self._skipIgnorables(instring, loc)             # <<<<<<<<<<<<<<
 *             while loc < instrlen and instring[loc].isspace() and col(loc, instring) != self.col:
 *                 loc += 1
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_skipIgnorables); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3602, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = NULL;
      __pyx_t_4 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_4 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_2)) {
        PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_instring, __pyx_v_loc};
        __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3602, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_GOTREF(__pyx_t_5);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
        PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_instring, __pyx_v_loc};
        __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3602, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_GOTREF(__pyx_t_5);
      } else
      #endif
      {
        __pyx_t_3 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3602, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (__pyx_t_1) {
          __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1); __pyx_t_1 = NULL;
        }
        __Pyx_INCREF(__pyx_v_instring);
        __Pyx_GIVEREF(__pyx_v_instring);
        PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_4, __pyx_v_instring);
        __Pyx_INCREF(__pyx_v_loc);
        __Pyx_GIVEREF(__pyx_v_loc);
        PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_4, __pyx_v_loc);
        __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3602, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF_SET(__pyx_v_loc, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":3601
 *         if col(loc, instring) != self.col:
 *             instrlen = len(instring)
 *             if self.ignoreExprs:             # <<<<<<<<<<<<<<
 *                 loc = self._skipIgnorables(instring, loc)
 *             while loc < instrlen and instring[loc].isspace() and col(loc, instring) != self.col:
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":3603
 *             if self.ignoreExprs:
 *                 loc = self._skipIgnorables(instring, loc)
 *             while loc < instrlen and instring[loc].isspace() and col(loc, instring) != self.col:             # <<<<<<<<<<<<<<
 *                 loc += 1
 *         return loc
 */
    while (1) {
      __pyx_t_5 = PyInt_FromSsize_t(__pyx_v_instrlen); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3603, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = PyObject_RichCompare(__pyx_v_loc, __pyx_t_5, Py_LT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3603, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 3603, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (__pyx_t_8) {
      } else {
        __pyx_t_6 = __pyx_t_8;
        goto __pyx_L7_bool_binop_done;
      }
      __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_instring, __pyx_v_loc); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3603, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_isspace); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3603, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3603, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 3603, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (__pyx_t_8) {
      } else {
        __pyx_t_6 = __pyx_t_8;
        goto __pyx_L7_bool_binop_done;
      }
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_col); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3603, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = NULL;
      __pyx_t_4 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_4 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_loc, __pyx_v_instring};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3603, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_loc, __pyx_v_instring};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3603, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else
      #endif
      {
        __pyx_t_1 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3603, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (__pyx_t_5) {
          __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5); __pyx_t_5 = NULL;
        }
        __Pyx_INCREF(__pyx_v_loc);
        __Pyx_GIVEREF(__pyx_v_loc);
        PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_4, __pyx_v_loc);
        __Pyx_INCREF(__pyx_v_instring);
        __Pyx_GIVEREF(__pyx_v_instring);
        PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_4, __pyx_v_instring);
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3603, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_col); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3603, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = PyObject_RichCompare(__pyx_t_2, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3603, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 3603, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_6 = __pyx_t_8;
      __pyx_L7_bool_binop_done:;
      if (!__pyx_t_6) break;

      /* "Aspidites/_vendor/pyparsing_extension.py":3604
 *                 loc = self._skipIgnorables(instring, loc)
 *             while loc < instrlen and instring[loc].isspace() and col(loc, instring) != self.col:
 *                 loc += 1             # <<<<<<<<<<<<<<
 *         return loc
 * 
 */
      __pyx_t_1 = __Pyx_PyInt_AddObjC(__pyx_v_loc, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3604, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_loc, __pyx_t_1);
      __pyx_t_1 = 0;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":3599
 * 
 *     def preParse(self, instring, loc):
 *         if col(loc, instring) != self.col:             # <<<<<<<<<<<<<<
 *             instrlen = len(instring)
 *             if self.ignoreExprs:
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3605
 *             while loc < instrlen and instring[loc].isspace() and col(loc, instring) != self.col:
 *                 loc += 1
 *         return loc             # <<<<<<<<<<<<<<
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_loc);
  __pyx_r = __pyx_v_loc;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3598
 *         self.col = colno
 * 
 *     def preParse(self, instring, loc):             # <<<<<<<<<<<<<<
 *         if col(loc, instring) != self.col:
 *             instrlen = len(instring)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.GoToColumn.preParse", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_loc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3607
 *         return loc
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         thiscol = col(loc, instring)
 *         if thiscol > self.col:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10GoToColumn_5parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10GoToColumn_5parseImpl = {"parseImpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10GoToColumn_5parseImpl, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10GoToColumn_5parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  CYTHON_UNUSED PyObject *__pyx_v_doActions = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parseImpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_doActions,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 1); __PYX_ERR(0, 3607, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 2); __PYX_ERR(0, 3607, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_doActions);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "parseImpl") < 0)) __PYX_ERR(0, 3607, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_doActions = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3607, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.GoToColumn.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10GoToColumn_4parseImpl(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10GoToColumn_4parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions) {
  PyObject *__pyx_v_thiscol = NULL;
  PyObject *__pyx_v_newloc = NULL;
  PyObject *__pyx_v_ret = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parseImpl", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":3608
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         thiscol = col(loc, instring)             # <<<<<<<<<<<<<<
 *         if thiscol > self.col:
 *             raise ParseException(instring, loc, "Text not in expected column", self)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_col); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3608, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_loc, __pyx_v_instring};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3608, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_loc, __pyx_v_instring};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3608, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3608, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_loc);
    __Pyx_GIVEREF(__pyx_v_loc);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_loc);
    __Pyx_INCREF(__pyx_v_instring);
    __Pyx_GIVEREF(__pyx_v_instring);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_instring);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3608, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_thiscol = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3609
 *     def parseImpl(self, instring, loc, doActions=True):
 *         thiscol = col(loc, instring)
 *         if thiscol > self.col:             # <<<<<<<<<<<<<<
 *             raise ParseException(instring, loc, "Text not in expected column", self)
 *         newloc = loc + self.col - thiscol
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_col); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3609, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_thiscol, __pyx_t_1, Py_GT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3609, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 3609, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_t_6)) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3610
 *         thiscol = col(loc, instring)
 *         if thiscol > self.col:
 *             raise ParseException(instring, loc, "Text not in expected column", self)             # <<<<<<<<<<<<<<
 *         newloc = loc + self.col - thiscol
 *         ret = instring[loc: newloc]
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3610, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = NULL;
    __pyx_t_4 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_4 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[5] = {__pyx_t_5, __pyx_v_instring, __pyx_v_loc, __pyx_kp_s_Text_not_in_expected_column, __pyx_v_self};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_4, 4+__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3610, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[5] = {__pyx_t_5, __pyx_v_instring, __pyx_v_loc, __pyx_kp_s_Text_not_in_expected_column, __pyx_v_self};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_4, 4+__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3610, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else
    #endif
    {
      __pyx_t_3 = PyTuple_New(4+__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3610, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_INCREF(__pyx_v_instring);
      __Pyx_GIVEREF(__pyx_v_instring);
      PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_4, __pyx_v_instring);
      __Pyx_INCREF(__pyx_v_loc);
      __Pyx_GIVEREF(__pyx_v_loc);
      PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_4, __pyx_v_loc);
      __Pyx_INCREF(__pyx_kp_s_Text_not_in_expected_column);
      __Pyx_GIVEREF(__pyx_kp_s_Text_not_in_expected_column);
      PyTuple_SET_ITEM(__pyx_t_3, 2+__pyx_t_4, __pyx_kp_s_Text_not_in_expected_column);
      __Pyx_INCREF(__pyx_v_self);
      __Pyx_GIVEREF(__pyx_v_self);
      PyTuple_SET_ITEM(__pyx_t_3, 3+__pyx_t_4, __pyx_v_self);
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3610, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 3610, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":3609
 *     def parseImpl(self, instring, loc, doActions=True):
 *         thiscol = col(loc, instring)
 *         if thiscol > self.col:             # <<<<<<<<<<<<<<
 *             raise ParseException(instring, loc, "Text not in expected column", self)
 *         newloc = loc + self.col - thiscol
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3611
 *         if thiscol > self.col:
 *             raise ParseException(instring, loc, "Text not in expected column", self)
 *         newloc = loc + self.col - thiscol             # <<<<<<<<<<<<<<
 *         ret = instring[loc: newloc]
 *         return newloc, ret
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_col); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3611, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyNumber_Add(__pyx_v_loc, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3611, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_Subtract(__pyx_t_1, __pyx_v_thiscol); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3611, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_newloc = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3612
 *             raise ParseException(instring, loc, "Text not in expected column", self)
 *         newloc = loc + self.col - thiscol
 *         ret = instring[loc: newloc]             # <<<<<<<<<<<<<<
 *         return newloc, ret
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetSlice(__pyx_v_instring, 0, 0, &__pyx_v_loc, &__pyx_v_newloc, NULL, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_ret = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3613
 *         newloc = loc + self.col - thiscol
 *         ret = instring[loc: newloc]
 *         return newloc, ret             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_newloc);
  __Pyx_GIVEREF(__pyx_v_newloc);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_newloc);
  __Pyx_INCREF(__pyx_v_ret);
  __Pyx_GIVEREF(__pyx_v_ret);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_ret);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3607
 *         return loc
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         thiscol = col(loc, instring)
 *         if thiscol > self.col:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.GoToColumn.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_thiscol);
  __Pyx_XDECREF(__pyx_v_newloc);
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3639
 *     """
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         super(LineStart, self).__init__()
 *         self.errmsg = "Expected start of line"
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_9LineStart_1__init__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_9LineStart_1__init__ = {"__init__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_9LineStart_1__init__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_9LineStart_1__init__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_9LineStart___init__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_9LineStart___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":3640
 * 
 *     def __init__(self):
 *         super(LineStart, self).__init__()             # <<<<<<<<<<<<<<
 *         self.errmsg = "Expected start of line"
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_LineStart); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3640, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3640, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3640, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3640, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3640, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3641
 *     def __init__(self):
 *         super(LineStart, self).__init__()
 *         self.errmsg = "Expected start of line"             # <<<<<<<<<<<<<<
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_errmsg, __pyx_kp_s_Expected_start_of_line) < 0) __PYX_ERR(0, 3641, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3639
 *     """
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         super(LineStart, self).__init__()
 *         self.errmsg = "Expected start of line"
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.LineStart.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3643
 *         self.errmsg = "Expected start of line"
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if col(loc, instring) == 1:
 *             return loc, []
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_9LineStart_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_9LineStart_3parseImpl = {"parseImpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_9LineStart_3parseImpl, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_9LineStart_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  CYTHON_UNUSED PyObject *__pyx_v_doActions = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parseImpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_doActions,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 1); __PYX_ERR(0, 3643, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 2); __PYX_ERR(0, 3643, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_doActions);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "parseImpl") < 0)) __PYX_ERR(0, 3643, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_doActions = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3643, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.LineStart.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_9LineStart_2parseImpl(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_9LineStart_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parseImpl", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":3644
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if col(loc, instring) == 1:             # <<<<<<<<<<<<<<
 *             return loc, []
 *         raise ParseException(instring, loc, self.errmsg, self)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_col); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_loc, __pyx_v_instring};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3644, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_loc, __pyx_v_instring};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3644, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3644, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_loc);
    __Pyx_GIVEREF(__pyx_v_loc);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_loc);
    __Pyx_INCREF(__pyx_v_instring);
    __Pyx_GIVEREF(__pyx_v_instring);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_instring);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3644, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_t_1, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 3644, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_6) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3645
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if col(loc, instring) == 1:
 *             return loc, []             # <<<<<<<<<<<<<<
 *         raise ParseException(instring, loc, self.errmsg, self)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3645, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3645, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_loc);
    __Pyx_GIVEREF(__pyx_v_loc);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_loc);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3644
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if col(loc, instring) == 1:             # <<<<<<<<<<<<<<
 *             return loc, []
 *         raise ParseException(instring, loc, self.errmsg, self)
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3646
 *         if col(loc, instring) == 1:
 *             return loc, []
 *         raise ParseException(instring, loc, self.errmsg, self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errmsg); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_3, __pyx_v_instring, __pyx_v_loc, __pyx_t_5, __pyx_v_self};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 4+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3646, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_3, __pyx_v_instring, __pyx_v_loc, __pyx_t_5, __pyx_v_self};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 4+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3646, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(4+__pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3646, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_instring);
    __Pyx_GIVEREF(__pyx_v_instring);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_4, __pyx_v_instring);
    __Pyx_INCREF(__pyx_v_loc);
    __Pyx_GIVEREF(__pyx_v_loc);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_4, __pyx_v_loc);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_4, __pyx_t_5);
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_7, 3+__pyx_t_4, __pyx_v_self);
    __pyx_t_5 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3646, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 3646, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3643
 *         self.errmsg = "Expected start of line"
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if col(loc, instring) == 1:
 *             return loc, []
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.LineStart.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3654
 *     """
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         super(LineEnd, self).__init__()
 *         self.setWhitespaceChars(ParserElement.DEFAULT_WHITE_CHARS.replace("\n", ""))
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7LineEnd_1__init__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7LineEnd_1__init__ = {"__init__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7LineEnd_1__init__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7LineEnd_1__init__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7LineEnd___init__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7LineEnd___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":3655
 * 
 *     def __init__(self):
 *         super(LineEnd, self).__init__()             # <<<<<<<<<<<<<<
 *         self.setWhitespaceChars(ParserElement.DEFAULT_WHITE_CHARS.replace("\n", ""))
 *         self.errmsg = "Expected end of line"
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_LineEnd); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3656
 *     def __init__(self):
 *         super(LineEnd, self).__init__()
 *         self.setWhitespaceChars(ParserElement.DEFAULT_WHITE_CHARS.replace("\n", ""))             # <<<<<<<<<<<<<<
 *         self.errmsg = "Expected end of line"
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_setWhitespaceChars); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_DEFAULT_WHITE_CHARS); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_replace); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__106, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3657
 *         super(LineEnd, self).__init__()
 *         self.setWhitespaceChars(ParserElement.DEFAULT_WHITE_CHARS.replace("\n", ""))
 *         self.errmsg = "Expected end of line"             # <<<<<<<<<<<<<<
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_errmsg, __pyx_kp_s_Expected_end_of_line) < 0) __PYX_ERR(0, 3657, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3654
 *     """
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         super(LineEnd, self).__init__()
 *         self.setWhitespaceChars(ParserElement.DEFAULT_WHITE_CHARS.replace("\n", ""))
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.LineEnd.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3659
 *         self.errmsg = "Expected end of line"
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if loc < len(instring):
 *             if instring[loc] == "\n":
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7LineEnd_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7LineEnd_3parseImpl = {"parseImpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7LineEnd_3parseImpl, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7LineEnd_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  CYTHON_UNUSED PyObject *__pyx_v_doActions = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parseImpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_doActions,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 1); __PYX_ERR(0, 3659, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 2); __PYX_ERR(0, 3659, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_doActions);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "parseImpl") < 0)) __PYX_ERR(0, 3659, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_doActions = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3659, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.LineEnd.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7LineEnd_2parseImpl(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7LineEnd_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parseImpl", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":3660
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if loc < len(instring):             # <<<<<<<<<<<<<<
 *             if instring[loc] == "\n":
 *                 return loc + 1, "\n"
 */
  __pyx_t_1 = PyObject_Length(__pyx_v_instring); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 3660, __pyx_L1_error)
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3660, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_loc, __pyx_t_2, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3660, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 3660, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3661
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if loc < len(instring):
 *             if instring[loc] == "\n":             # <<<<<<<<<<<<<<
 *                 return loc + 1, "\n"
 *             else:
 */
    __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_instring, __pyx_v_loc); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3661, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_t_3, __pyx_kp_s__4, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 3661, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (likely(__pyx_t_4)) {

      /* "Aspidites/_vendor/pyparsing_extension.py":3662
 *         if loc < len(instring):
 *             if instring[loc] == "\n":
 *                 return loc + 1, "\n"             # <<<<<<<<<<<<<<
 *             else:
 *                 raise ParseException(instring, loc, self.errmsg, self)
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = __Pyx_PyInt_AddObjC(__pyx_v_loc, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3662, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3662, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
      __Pyx_INCREF(__pyx_kp_s__4);
      __Pyx_GIVEREF(__pyx_kp_s__4);
      PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_kp_s__4);
      __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      goto __pyx_L0;

      /* "Aspidites/_vendor/pyparsing_extension.py":3661
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if loc < len(instring):
 *             if instring[loc] == "\n":             # <<<<<<<<<<<<<<
 *                 return loc + 1, "\n"
 *             else:
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":3664
 *                 return loc + 1, "\n"
 *             else:
 *                 raise ParseException(instring, loc, self.errmsg, self)             # <<<<<<<<<<<<<<
 *         elif loc == len(instring):
 *             return loc + 1, []
 */
    /*else*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3664, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errmsg); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3664, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_7 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[5] = {__pyx_t_6, __pyx_v_instring, __pyx_v_loc, __pyx_t_5, __pyx_v_self};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3664, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[5] = {__pyx_t_6, __pyx_v_instring, __pyx_v_loc, __pyx_t_5, __pyx_v_self};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3664, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      } else
      #endif
      {
        __pyx_t_8 = PyTuple_New(4+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3664, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        if (__pyx_t_6) {
          __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
        }
        __Pyx_INCREF(__pyx_v_instring);
        __Pyx_GIVEREF(__pyx_v_instring);
        PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_v_instring);
        __Pyx_INCREF(__pyx_v_loc);
        __Pyx_GIVEREF(__pyx_v_loc);
        PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_v_loc);
        __Pyx_GIVEREF(__pyx_t_5);
        PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_7, __pyx_t_5);
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_8, 3+__pyx_t_7, __pyx_v_self);
        __pyx_t_5 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3664, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 3664, __pyx_L1_error)
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":3660
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if loc < len(instring):             # <<<<<<<<<<<<<<
 *             if instring[loc] == "\n":
 *                 return loc + 1, "\n"
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3665
 *             else:
 *                 raise ParseException(instring, loc, self.errmsg, self)
 *         elif loc == len(instring):             # <<<<<<<<<<<<<<
 *             return loc + 1, []
 *         else:
 */
  __pyx_t_1 = PyObject_Length(__pyx_v_instring); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 3665, __pyx_L1_error)
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3665, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_loc, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3665, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 3665, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (likely(__pyx_t_4)) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3666
 *                 raise ParseException(instring, loc, self.errmsg, self)
 *         elif loc == len(instring):
 *             return loc + 1, []             # <<<<<<<<<<<<<<
 *         else:
 *             raise ParseException(instring, loc, self.errmsg, self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyInt_AddObjC(__pyx_v_loc, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3666, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3666, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3666, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_2);
    __pyx_t_3 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_8;
    __pyx_t_8 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3665
 *             else:
 *                 raise ParseException(instring, loc, self.errmsg, self)
 *         elif loc == len(instring):             # <<<<<<<<<<<<<<
 *             return loc + 1, []
 *         else:
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3668
 *             return loc + 1, []
 *         else:
 *             raise ParseException(instring, loc, self.errmsg, self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3668, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errmsg); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3668, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[5] = {__pyx_t_5, __pyx_v_instring, __pyx_v_loc, __pyx_t_3, __pyx_v_self};
      __pyx_t_8 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3668, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[5] = {__pyx_t_5, __pyx_v_instring, __pyx_v_loc, __pyx_t_3, __pyx_v_self};
      __pyx_t_8 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3668, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(4+__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3668, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_INCREF(__pyx_v_instring);
      __Pyx_GIVEREF(__pyx_v_instring);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_7, __pyx_v_instring);
      __Pyx_INCREF(__pyx_v_loc);
      __Pyx_GIVEREF(__pyx_v_loc);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_7, __pyx_v_loc);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_7, __pyx_t_3);
      __Pyx_INCREF(__pyx_v_self);
      __Pyx_GIVEREF(__pyx_v_self);
      PyTuple_SET_ITEM(__pyx_t_6, 3+__pyx_t_7, __pyx_v_self);
      __pyx_t_3 = 0;
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3668, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_8, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __PYX_ERR(0, 3668, __pyx_L1_error)
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3659
 *         self.errmsg = "Expected end of line"
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if loc < len(instring):
 *             if instring[loc] == "\n":
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.LineEnd.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3676
 *     """
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         super(StringStart, self).__init__()
 *         self.errmsg = "Expected start of text"
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_11StringStart_1__init__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_11StringStart_1__init__ = {"__init__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_11StringStart_1__init__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_11StringStart_1__init__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_11StringStart___init__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_11StringStart___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":3677
 * 
 *     def __init__(self):
 *         super(StringStart, self).__init__()             # <<<<<<<<<<<<<<
 *         self.errmsg = "Expected start of text"
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_StringStart); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3677, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3677, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3677, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3677, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3677, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3678
 *     def __init__(self):
 *         super(StringStart, self).__init__()
 *         self.errmsg = "Expected start of text"             # <<<<<<<<<<<<<<
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_errmsg, __pyx_kp_s_Expected_start_of_text) < 0) __PYX_ERR(0, 3678, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3676
 *     """
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         super(StringStart, self).__init__()
 *         self.errmsg = "Expected start of text"
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.StringStart.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3680
 *         self.errmsg = "Expected start of text"
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if loc != 0:
 *             # see if entire string up to here is just whitespace and ignoreables
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_11StringStart_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_11StringStart_3parseImpl = {"parseImpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_11StringStart_3parseImpl, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_11StringStart_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  CYTHON_UNUSED PyObject *__pyx_v_doActions = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parseImpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_doActions,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 1); __PYX_ERR(0, 3680, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 2); __PYX_ERR(0, 3680, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_doActions);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "parseImpl") < 0)) __PYX_ERR(0, 3680, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_doActions = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3680, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.StringStart.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_11StringStart_2parseImpl(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_11StringStart_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parseImpl", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":3681
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if loc != 0:             # <<<<<<<<<<<<<<
 *             # see if entire string up to here is just whitespace and ignoreables
 *             if loc != self.preParse(instring, 0):
 */
  __pyx_t_1 = __Pyx_PyInt_NeObjC(__pyx_v_loc, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 3681, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3683
 *         if loc != 0:
 *             # see if entire string up to here is just whitespace and ignoreables
 *             if loc != self.preParse(instring, 0):             # <<<<<<<<<<<<<<
 *                 raise ParseException(instring, loc, self.errmsg, self)
 *         return loc, []
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_preParse); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3683, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_instring, __pyx_int_0};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3683, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_instring, __pyx_int_0};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3683, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3683, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_4) {
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
      }
      __Pyx_INCREF(__pyx_v_instring);
      __Pyx_GIVEREF(__pyx_v_instring);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_instring);
      __Pyx_INCREF(__pyx_int_0);
      __Pyx_GIVEREF(__pyx_int_0);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_int_0);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3683, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyObject_RichCompare(__pyx_v_loc, __pyx_t_1, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3683, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 3683, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(__pyx_t_2)) {

      /* "Aspidites/_vendor/pyparsing_extension.py":3684
 *             # see if entire string up to here is just whitespace and ignoreables
 *             if loc != self.preParse(instring, 0):
 *                 raise ParseException(instring, loc, self.errmsg, self)             # <<<<<<<<<<<<<<
 *         return loc, []
 * 
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3684, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errmsg); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3684, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_4 = NULL;
      __pyx_t_5 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_5 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[5] = {__pyx_t_4, __pyx_v_instring, __pyx_v_loc, __pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_5, 4+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3684, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[5] = {__pyx_t_4, __pyx_v_instring, __pyx_v_loc, __pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_5, 4+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3684, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(4+__pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3684, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        if (__pyx_t_4) {
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4); __pyx_t_4 = NULL;
        }
        __Pyx_INCREF(__pyx_v_instring);
        __Pyx_GIVEREF(__pyx_v_instring);
        PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_5, __pyx_v_instring);
        __Pyx_INCREF(__pyx_v_loc);
        __Pyx_GIVEREF(__pyx_v_loc);
        PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_5, __pyx_v_loc);
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_5, __pyx_t_6);
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_7, 3+__pyx_t_5, __pyx_v_self);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3684, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 3684, __pyx_L1_error)

      /* "Aspidites/_vendor/pyparsing_extension.py":3683
 *         if loc != 0:
 *             # see if entire string up to here is just whitespace and ignoreables
 *             if loc != self.preParse(instring, 0):             # <<<<<<<<<<<<<<
 *                 raise ParseException(instring, loc, self.errmsg, self)
 *         return loc, []
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":3681
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if loc != 0:             # <<<<<<<<<<<<<<
 *             # see if entire string up to here is just whitespace and ignoreables
 *             if loc != self.preParse(instring, 0):
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3685
 *             if loc != self.preParse(instring, 0):
 *                 raise ParseException(instring, loc, self.errmsg, self)
 *         return loc, []             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3685, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3685, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_loc);
  __Pyx_GIVEREF(__pyx_v_loc);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_loc);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3680
 *         self.errmsg = "Expected start of text"
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if loc != 0:
 *             # see if entire string up to here is just whitespace and ignoreables
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.StringStart.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3692
 *     """
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         super(StringEnd, self).__init__()
 *         self.errmsg = "Expected end of text"
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_9StringEnd_1__init__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_9StringEnd_1__init__ = {"__init__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_9StringEnd_1__init__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_9StringEnd_1__init__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_9StringEnd___init__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_9StringEnd___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":3693
 * 
 *     def __init__(self):
 *         super(StringEnd, self).__init__()             # <<<<<<<<<<<<<<
 *         self.errmsg = "Expected end of text"
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_StringEnd); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3693, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3693, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3693, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3693, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3693, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3694
 *     def __init__(self):
 *         super(StringEnd, self).__init__()
 *         self.errmsg = "Expected end of text"             # <<<<<<<<<<<<<<
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_errmsg, __pyx_kp_s_Expected_end_of_text) < 0) __PYX_ERR(0, 3694, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3692
 *     """
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         super(StringEnd, self).__init__()
 *         self.errmsg = "Expected end of text"
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.StringEnd.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3696
 *         self.errmsg = "Expected end of text"
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if loc < len(instring):
 *             raise ParseException(instring, loc, self.errmsg, self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_9StringEnd_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_9StringEnd_3parseImpl = {"parseImpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_9StringEnd_3parseImpl, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_9StringEnd_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  CYTHON_UNUSED PyObject *__pyx_v_doActions = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parseImpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_doActions,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 1); __PYX_ERR(0, 3696, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 2); __PYX_ERR(0, 3696, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_doActions);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "parseImpl") < 0)) __PYX_ERR(0, 3696, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_doActions = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3696, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.StringEnd.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_9StringEnd_2parseImpl(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_9StringEnd_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parseImpl", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":3697
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if loc < len(instring):             # <<<<<<<<<<<<<<
 *             raise ParseException(instring, loc, self.errmsg, self)
 *         elif loc == len(instring):
 */
  __pyx_t_1 = PyObject_Length(__pyx_v_instring); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 3697, __pyx_L1_error)
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_loc, __pyx_t_2, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3697, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 3697, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(__pyx_t_4)) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3698
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if loc < len(instring):
 *             raise ParseException(instring, loc, self.errmsg, self)             # <<<<<<<<<<<<<<
 *         elif loc == len(instring):
 *             return loc + 1, []
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3698, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errmsg); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3698, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[5] = {__pyx_t_6, __pyx_v_instring, __pyx_v_loc, __pyx_t_5, __pyx_v_self};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3698, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[5] = {__pyx_t_6, __pyx_v_instring, __pyx_v_loc, __pyx_t_5, __pyx_v_self};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3698, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(4+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3698, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_INCREF(__pyx_v_instring);
      __Pyx_GIVEREF(__pyx_v_instring);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_v_instring);
      __Pyx_INCREF(__pyx_v_loc);
      __Pyx_GIVEREF(__pyx_v_loc);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_v_loc);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_7, __pyx_t_5);
      __Pyx_INCREF(__pyx_v_self);
      __Pyx_GIVEREF(__pyx_v_self);
      PyTuple_SET_ITEM(__pyx_t_8, 3+__pyx_t_7, __pyx_v_self);
      __pyx_t_5 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_8, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3698, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 3698, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":3697
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if loc < len(instring):             # <<<<<<<<<<<<<<
 *             raise ParseException(instring, loc, self.errmsg, self)
 *         elif loc == len(instring):
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3699
 *         if loc < len(instring):
 *             raise ParseException(instring, loc, self.errmsg, self)
 *         elif loc == len(instring):             # <<<<<<<<<<<<<<
 *             return loc + 1, []
 *         elif loc > len(instring):
 */
  __pyx_t_1 = PyObject_Length(__pyx_v_instring); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 3699, __pyx_L1_error)
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3699, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_loc, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3699, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 3699, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3700
 *             raise ParseException(instring, loc, self.errmsg, self)
 *         elif loc == len(instring):
 *             return loc + 1, []             # <<<<<<<<<<<<<<
 *         elif loc > len(instring):
 *             return loc, []
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_v_loc, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3700, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3700, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3700, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_3);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_8;
    __pyx_t_8 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3699
 *         if loc < len(instring):
 *             raise ParseException(instring, loc, self.errmsg, self)
 *         elif loc == len(instring):             # <<<<<<<<<<<<<<
 *             return loc + 1, []
 *         elif loc > len(instring):
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3701
 *         elif loc == len(instring):
 *             return loc + 1, []
 *         elif loc > len(instring):             # <<<<<<<<<<<<<<
 *             return loc, []
 *         else:
 */
  __pyx_t_1 = PyObject_Length(__pyx_v_instring); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 3701, __pyx_L1_error)
  __pyx_t_8 = PyInt_FromSsize_t(__pyx_t_1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3701, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_loc, __pyx_t_8, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3701, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 3701, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (likely(__pyx_t_4)) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3702
 *             return loc + 1, []
 *         elif loc > len(instring):
 *             return loc, []             # <<<<<<<<<<<<<<
 *         else:
 *             raise ParseException(instring, loc, self.errmsg, self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3702, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3702, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_INCREF(__pyx_v_loc);
    __Pyx_GIVEREF(__pyx_v_loc);
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_loc);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_8;
    __pyx_t_8 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3701
 *         elif loc == len(instring):
 *             return loc + 1, []
 *         elif loc > len(instring):             # <<<<<<<<<<<<<<
 *             return loc, []
 *         else:
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3704
 *             return loc, []
 *         else:
 *             raise ParseException(instring, loc, self.errmsg, self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3704, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errmsg); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3704, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[5] = {__pyx_t_5, __pyx_v_instring, __pyx_v_loc, __pyx_t_2, __pyx_v_self};
      __pyx_t_8 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3704, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[5] = {__pyx_t_5, __pyx_v_instring, __pyx_v_loc, __pyx_t_2, __pyx_v_self};
      __pyx_t_8 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3704, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(4+__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3704, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_INCREF(__pyx_v_instring);
      __Pyx_GIVEREF(__pyx_v_instring);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_7, __pyx_v_instring);
      __Pyx_INCREF(__pyx_v_loc);
      __Pyx_GIVEREF(__pyx_v_loc);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_7, __pyx_v_loc);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_7, __pyx_t_2);
      __Pyx_INCREF(__pyx_v_self);
      __Pyx_GIVEREF(__pyx_v_self);
      PyTuple_SET_ITEM(__pyx_t_6, 3+__pyx_t_7, __pyx_v_self);
      __pyx_t_2 = 0;
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3704, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_8, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __PYX_ERR(0, 3704, __pyx_L1_error)
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3696
 *         self.errmsg = "Expected end of text"
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if loc < len(instring):
 *             raise ParseException(instring, loc, self.errmsg, self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.StringEnd.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3717
 *     """
 * 
 *     def __init__(self, wordChars=printables):             # <<<<<<<<<<<<<<
 *         super(WordStart, self).__init__()
 *         self.wordChars = set(wordChars)
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_38__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__defaults__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3717, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults5, __pyx_self)->__pyx_arg_wordChars);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults5, __pyx_self)->__pyx_arg_wordChars);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __Pyx_CyFunction_Defaults(__pyx_defaults5, __pyx_self)->__pyx_arg_wordChars);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3717, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_2, 1, Py_None);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_9WordStart_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_9WordStart_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_9WordStart_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_9WordStart_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_wordChars = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_wordChars,0};
    PyObject* values[2] = {0,0};
    __pyx_defaults5 *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults5, __pyx_self);
    values[1] = __pyx_dynamic_args->__pyx_arg_wordChars;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_wordChars);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 3717, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_wordChars = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3717, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.WordStart.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_9WordStart___init__(__pyx_self, __pyx_v_self, __pyx_v_wordChars);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_9WordStart___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_wordChars) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":3718
 * 
 *     def __init__(self, wordChars=printables):
 *         super(WordStart, self).__init__()             # <<<<<<<<<<<<<<
 *         self.wordChars = set(wordChars)
 *         self.errmsg = "Not at the start of a word"
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_WordStart); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3719
 *     def __init__(self, wordChars=printables):
 *         super(WordStart, self).__init__()
 *         self.wordChars = set(wordChars)             # <<<<<<<<<<<<<<
 *         self.errmsg = "Not at the start of a word"
 * 
 */
  __pyx_t_1 = PySet_New(__pyx_v_wordChars); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3719, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_wordChars, __pyx_t_1) < 0) __PYX_ERR(0, 3719, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3720
 *         super(WordStart, self).__init__()
 *         self.wordChars = set(wordChars)
 *         self.errmsg = "Not at the start of a word"             # <<<<<<<<<<<<<<
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_errmsg, __pyx_kp_s_Not_at_the_start_of_a_word) < 0) __PYX_ERR(0, 3720, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3717
 *     """
 * 
 *     def __init__(self, wordChars=printables):             # <<<<<<<<<<<<<<
 *         super(WordStart, self).__init__()
 *         self.wordChars = set(wordChars)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.WordStart.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3722
 *         self.errmsg = "Not at the start of a word"
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if loc != 0:
 *             if (instring[loc - 1] in self.wordChars
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_9WordStart_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_9WordStart_3parseImpl = {"parseImpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_9WordStart_3parseImpl, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_9WordStart_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  CYTHON_UNUSED PyObject *__pyx_v_doActions = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parseImpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_doActions,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 1); __PYX_ERR(0, 3722, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 2); __PYX_ERR(0, 3722, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_doActions);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "parseImpl") < 0)) __PYX_ERR(0, 3722, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_doActions = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3722, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.WordStart.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_9WordStart_2parseImpl(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_9WordStart_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parseImpl", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":3723
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if loc != 0:             # <<<<<<<<<<<<<<
 *             if (instring[loc - 1] in self.wordChars
 *                     or instring[loc] not in self.wordChars):
 */
  __pyx_t_1 = __Pyx_PyInt_NeObjC(__pyx_v_loc, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3723, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 3723, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3724
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if loc != 0:
 *             if (instring[loc - 1] in self.wordChars             # <<<<<<<<<<<<<<
 *                     or instring[loc] not in self.wordChars):
 *                 raise ParseException(instring, loc, self.errmsg, self)
 */
    __pyx_t_1 = __Pyx_PyInt_SubtractObjC(__pyx_v_loc, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3724, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_instring, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3724, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_wordChars); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3724, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = (__Pyx_PySequence_ContainsTF(__pyx_t_3, __pyx_t_1, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 3724, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = (__pyx_t_4 != 0);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_2 = __pyx_t_5;
      goto __pyx_L5_bool_binop_done;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":3725
 *         if loc != 0:
 *             if (instring[loc - 1] in self.wordChars
 *                     or instring[loc] not in self.wordChars):             # <<<<<<<<<<<<<<
 *                 raise ParseException(instring, loc, self.errmsg, self)
 *         return loc, []
 */
    __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_instring, __pyx_v_loc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3725, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_wordChars); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3725, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_t_3, Py_NE)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 3725, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = (__pyx_t_5 != 0);
    __pyx_t_2 = __pyx_t_4;
    __pyx_L5_bool_binop_done:;

    /* "Aspidites/_vendor/pyparsing_extension.py":3724
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if loc != 0:
 *             if (instring[loc - 1] in self.wordChars             # <<<<<<<<<<<<<<
 *                     or instring[loc] not in self.wordChars):
 *                 raise ParseException(instring, loc, self.errmsg, self)
 */
    if (unlikely(__pyx_t_2)) {

      /* "Aspidites/_vendor/pyparsing_extension.py":3726
 *             if (instring[loc - 1] in self.wordChars
 *                     or instring[loc] not in self.wordChars):
 *                 raise ParseException(instring, loc, self.errmsg, self)             # <<<<<<<<<<<<<<
 *         return loc, []
 * 
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3726, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errmsg); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3726, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = NULL;
      __pyx_t_8 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_8 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[5] = {__pyx_t_7, __pyx_v_instring, __pyx_v_loc, __pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_8, 4+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3726, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[5] = {__pyx_t_7, __pyx_v_instring, __pyx_v_loc, __pyx_t_6, __pyx_v_self};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_8, 4+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3726, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_9 = PyTuple_New(4+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3726, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (__pyx_t_7) {
          __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;
        }
        __Pyx_INCREF(__pyx_v_instring);
        __Pyx_GIVEREF(__pyx_v_instring);
        PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_v_instring);
        __Pyx_INCREF(__pyx_v_loc);
        __Pyx_GIVEREF(__pyx_v_loc);
        PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_v_loc);
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_9, 2+__pyx_t_8, __pyx_t_6);
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_9, 3+__pyx_t_8, __pyx_v_self);
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_9, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3726, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 3726, __pyx_L1_error)

      /* "Aspidites/_vendor/pyparsing_extension.py":3724
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if loc != 0:
 *             if (instring[loc - 1] in self.wordChars             # <<<<<<<<<<<<<<
 *                     or instring[loc] not in self.wordChars):
 *                 raise ParseException(instring, loc, self.errmsg, self)
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":3723
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if loc != 0:             # <<<<<<<<<<<<<<
 *             if (instring[loc - 1] in self.wordChars
 *                     or instring[loc] not in self.wordChars):
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3727
 *                     or instring[loc] not in self.wordChars):
 *                 raise ParseException(instring, loc, self.errmsg, self)
 *         return loc, []             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3727, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3727, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_loc);
  __Pyx_GIVEREF(__pyx_v_loc);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_loc);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3722
 *         self.errmsg = "Not at the start of a word"
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if loc != 0:
 *             if (instring[loc - 1] in self.wordChars
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.WordStart.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3739
 *     """
 * 
 *     def __init__(self, wordChars=printables):             # <<<<<<<<<<<<<<
 *         super(WordEnd, self).__init__()
 *         self.wordChars = set(wordChars)
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_40__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__defaults__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3739, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults6, __pyx_self)->__pyx_arg_wordChars);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults6, __pyx_self)->__pyx_arg_wordChars);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __Pyx_CyFunction_Defaults(__pyx_defaults6, __pyx_self)->__pyx_arg_wordChars);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3739, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_2, 1, Py_None);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7WordEnd_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7WordEnd_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7WordEnd_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7WordEnd_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_wordChars = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_wordChars,0};
    PyObject* values[2] = {0,0};
    __pyx_defaults6 *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults6, __pyx_self);
    values[1] = __pyx_dynamic_args->__pyx_arg_wordChars;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_wordChars);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 3739, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_wordChars = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3739, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.WordEnd.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7WordEnd___init__(__pyx_self, __pyx_v_self, __pyx_v_wordChars);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7WordEnd___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_wordChars) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":3740
 * 
 *     def __init__(self, wordChars=printables):
 *         super(WordEnd, self).__init__()             # <<<<<<<<<<<<<<
 *         self.wordChars = set(wordChars)
 *         self.skipWhitespace = False
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_WordEnd); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3740, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3740, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3740, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3740, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3740, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3741
 *     def __init__(self, wordChars=printables):
 *         super(WordEnd, self).__init__()
 *         self.wordChars = set(wordChars)             # <<<<<<<<<<<<<<
 *         self.skipWhitespace = False
 *         self.errmsg = "Not at the end of a word"
 */
  __pyx_t_1 = PySet_New(__pyx_v_wordChars); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3741, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_wordChars, __pyx_t_1) < 0) __PYX_ERR(0, 3741, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3742
 *         super(WordEnd, self).__init__()
 *         self.wordChars = set(wordChars)
 *         self.skipWhitespace = False             # <<<<<<<<<<<<<<
 *         self.errmsg = "Not at the end of a word"
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_skipWhitespace, Py_False) < 0) __PYX_ERR(0, 3742, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3743
 *         self.wordChars = set(wordChars)
 *         self.skipWhitespace = False
 *         self.errmsg = "Not at the end of a word"             # <<<<<<<<<<<<<<
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_errmsg, __pyx_kp_s_Not_at_the_end_of_a_word) < 0) __PYX_ERR(0, 3743, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3739
 *     """
 * 
 *     def __init__(self, wordChars=printables):             # <<<<<<<<<<<<<<
 *         super(WordEnd, self).__init__()
 *         self.wordChars = set(wordChars)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.WordEnd.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3745
 *         self.errmsg = "Not at the end of a word"
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         instrlen = len(instring)
 *         if instrlen > 0 and loc < instrlen:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7WordEnd_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7WordEnd_3parseImpl = {"parseImpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7WordEnd_3parseImpl, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7WordEnd_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  CYTHON_UNUSED PyObject *__pyx_v_doActions = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parseImpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_doActions,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 1); __PYX_ERR(0, 3745, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 2); __PYX_ERR(0, 3745, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_doActions);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "parseImpl") < 0)) __PYX_ERR(0, 3745, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_doActions = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3745, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.WordEnd.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7WordEnd_2parseImpl(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7WordEnd_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions) {
  Py_ssize_t __pyx_v_instrlen;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parseImpl", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":3746
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         instrlen = len(instring)             # <<<<<<<<<<<<<<
 *         if instrlen > 0 and loc < instrlen:
 *             if (instring[loc] in self.wordChars or
 */
  __pyx_t_1 = PyObject_Length(__pyx_v_instring); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 3746, __pyx_L1_error)
  __pyx_v_instrlen = __pyx_t_1;

  /* "Aspidites/_vendor/pyparsing_extension.py":3747
 *     def parseImpl(self, instring, loc, doActions=True):
 *         instrlen = len(instring)
 *         if instrlen > 0 and loc < instrlen:             # <<<<<<<<<<<<<<
 *             if (instring[loc] in self.wordChars or
 *                     instring[loc - 1] not in self.wordChars):
 */
  __pyx_t_3 = ((__pyx_v_instrlen > 0) != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_instrlen); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3747, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyObject_RichCompare(__pyx_v_loc, __pyx_t_4, Py_LT); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3747, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 3747, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3748
 *         instrlen = len(instring)
 *         if instrlen > 0 and loc < instrlen:
 *             if (instring[loc] in self.wordChars or             # <<<<<<<<<<<<<<
 *                     instring[loc - 1] not in self.wordChars):
 *                 raise ParseException(instring, loc, self.errmsg, self)
 */
    __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_instring, __pyx_v_loc); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3748, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_wordChars); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3748, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = (__Pyx_PySequence_ContainsTF(__pyx_t_5, __pyx_t_4, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 3748, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_6 = (__pyx_t_3 != 0);
    if (!__pyx_t_6) {
    } else {
      __pyx_t_2 = __pyx_t_6;
      goto __pyx_L7_bool_binop_done;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":3749
 *         if instrlen > 0 and loc < instrlen:
 *             if (instring[loc] in self.wordChars or
 *                     instring[loc - 1] not in self.wordChars):             # <<<<<<<<<<<<<<
 *                 raise ParseException(instring, loc, self.errmsg, self)
 *         return loc, []
 */
    __pyx_t_4 = __Pyx_PyInt_SubtractObjC(__pyx_v_loc, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3749, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_instring, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3749, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_wordChars); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3749, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = (__Pyx_PySequence_ContainsTF(__pyx_t_5, __pyx_t_4, Py_NE)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 3749, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_3 = (__pyx_t_6 != 0);
    __pyx_t_2 = __pyx_t_3;
    __pyx_L7_bool_binop_done:;

    /* "Aspidites/_vendor/pyparsing_extension.py":3748
 *         instrlen = len(instring)
 *         if instrlen > 0 and loc < instrlen:
 *             if (instring[loc] in self.wordChars or             # <<<<<<<<<<<<<<
 *                     instring[loc - 1] not in self.wordChars):
 *                 raise ParseException(instring, loc, self.errmsg, self)
 */
    if (unlikely(__pyx_t_2)) {

      /* "Aspidites/_vendor/pyparsing_extension.py":3750
 *             if (instring[loc] in self.wordChars or
 *                     instring[loc - 1] not in self.wordChars):
 *                 raise ParseException(instring, loc, self.errmsg, self)             # <<<<<<<<<<<<<<
 *         return loc, []
 * 
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3750, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errmsg); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3750, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = NULL;
      __pyx_t_9 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_9 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[5] = {__pyx_t_8, __pyx_v_instring, __pyx_v_loc, __pyx_t_7, __pyx_v_self};
        __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 4+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3750, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[5] = {__pyx_t_8, __pyx_v_instring, __pyx_v_loc, __pyx_t_7, __pyx_v_self};
        __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 4+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3750, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else
      #endif
      {
        __pyx_t_10 = PyTuple_New(4+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 3750, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        if (__pyx_t_8) {
          __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
        }
        __Pyx_INCREF(__pyx_v_instring);
        __Pyx_GIVEREF(__pyx_v_instring);
        PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_v_instring);
        __Pyx_INCREF(__pyx_v_loc);
        __Pyx_GIVEREF(__pyx_v_loc);
        PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_v_loc);
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_10, 2+__pyx_t_9, __pyx_t_7);
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_10, 3+__pyx_t_9, __pyx_v_self);
        __pyx_t_7 = 0;
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3750, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __PYX_ERR(0, 3750, __pyx_L1_error)

      /* "Aspidites/_vendor/pyparsing_extension.py":3748
 *         instrlen = len(instring)
 *         if instrlen > 0 and loc < instrlen:
 *             if (instring[loc] in self.wordChars or             # <<<<<<<<<<<<<<
 *                     instring[loc - 1] not in self.wordChars):
 *                 raise ParseException(instring, loc, self.errmsg, self)
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":3747
 *     def parseImpl(self, instring, loc, doActions=True):
 *         instrlen = len(instring)
 *         if instrlen > 0 and loc < instrlen:             # <<<<<<<<<<<<<<
 *             if (instring[loc] in self.wordChars or
 *                     instring[loc - 1] not in self.wordChars):
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3751
 *                     instring[loc - 1] not in self.wordChars):
 *                 raise ParseException(instring, loc, self.errmsg, self)
 *         return loc, []             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3751, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3751, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_loc);
  __Pyx_GIVEREF(__pyx_v_loc);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_loc);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3745
 *         self.errmsg = "Not at the end of a word"
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         instrlen = len(instring)
 *         if instrlen > 0 and loc < instrlen:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.WordEnd.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3773
 * 
 *     class _ErrorStop(Empty):
 *         def __init__(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *             super(And._ErrorStop, self).__init__(*args, **kwargs)
 *             self.name = '-'
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_ErrorStop_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10_ErrorStop_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_ErrorStop_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_ErrorStop_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  if (PyTuple_GET_SIZE(__pyx_args) > 1) {
    __pyx_v_args = PyTuple_GetSlice(__pyx_args, 1, PyTuple_GET_SIZE(__pyx_args));
    if (unlikely(!__pyx_v_args)) {
      __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
      __Pyx_RefNannyFinishContext();
      return NULL;
    }
    __Pyx_GOTREF(__pyx_v_args);
  } else {
    __pyx_v_args = __pyx_empty_tuple; __Pyx_INCREF(__pyx_empty_tuple);
  }
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        default:
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t used_pos_args = (pos_args < 1) ? pos_args : 1;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwargs, values, used_pos_args, "__init__") < 0)) __PYX_ERR(0, 3773, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) < 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_self = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3773, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_args); __pyx_v_args = 0;
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._ErrorStop.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_ErrorStop___init__(__pyx_self, __pyx_v_self, __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_ErrorStop___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":3774
 *     class _ErrorStop(Empty):
 *         def __init__(self, *args, **kwargs):
 *             super(And._ErrorStop, self).__init__(*args, **kwargs)             # <<<<<<<<<<<<<<
 *             self.name = '-'
 *             self.leaveWhitespace()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_And); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3774, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ErrorStop); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3774, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3774, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3774, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3774, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3774, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_v_args, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3774, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3775
 *         def __init__(self, *args, **kwargs):
 *             super(And._ErrorStop, self).__init__(*args, **kwargs)
 *             self.name = '-'             # <<<<<<<<<<<<<<
 *             self.leaveWhitespace()
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_name, __pyx_kp_s__107) < 0) __PYX_ERR(0, 3775, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3776
 *             super(And._ErrorStop, self).__init__(*args, **kwargs)
 *             self.name = '-'
 *             self.leaveWhitespace()             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, exprs, savelist=True):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_leaveWhitespace); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3776, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3776, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3773
 * 
 *     class _ErrorStop(Empty):
 *         def __init__(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *             super(And._ErrorStop, self).__init__(*args, **kwargs)
 *             self.name = '-'
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._ErrorStop.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3778
 *             self.leaveWhitespace()
 * 
 *     def __init__(self, exprs, savelist=True):             # <<<<<<<<<<<<<<
 *         exprs = list(exprs)
 *         if exprs and Ellipsis in exprs:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_3And_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_3And_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_3And_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_3And_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_exprs = 0;
  PyObject *__pyx_v_savelist = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_exprs,&__pyx_n_s_savelist,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_exprs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, 1); __PYX_ERR(0, 3778, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_savelist);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 3778, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_exprs = values[1];
    __pyx_v_savelist = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3778, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.And.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_3And___init__(__pyx_self, __pyx_v_self, __pyx_v_exprs, __pyx_v_savelist);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_3And_8__init___2generator21(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/pyparsing_extension.py":3793
 *             exprs[:] = tmp
 *         super(And, self).__init__(exprs, savelist)
 *         self.mayReturnEmpty = all(e.mayReturnEmpty for e in self.exprs)             # <<<<<<<<<<<<<<
 *         self.setWhitespaceChars(self.exprs[0].whiteChars)
 *         self.skipWhitespace = self.exprs[0].skipWhitespace
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_3And_8__init___genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 3793, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__ *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_3And_8__init___2generator21, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_And___init___locals_genexpr, __pyx_n_s_Aspidites__vendor_pyparsing_exte); if (unlikely(!gen)) __PYX_ERR(0, 3793, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.And.__init__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_3And_8__init___2generator21(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 3793, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 3793, __pyx_L1_error) }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3793, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3793, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3793, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 3793, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3793, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 3793, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3793, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 3793, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_e);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_e, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_e, __pyx_n_s_mayReturnEmpty); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3793, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 3793, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = ((!__pyx_t_5) != 0);
    if (__pyx_t_6) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L0;
    }
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_True);
    __pyx_r = Py_True;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3778
 *             self.leaveWhitespace()
 * 
 *     def __init__(self, exprs, savelist=True):             # <<<<<<<<<<<<<<
 *         exprs = list(exprs)
 *         if exprs and Ellipsis in exprs:
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_3And___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_exprs, PyObject *__pyx_v_savelist) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__ *__pyx_cur_scope;
  PyObject *__pyx_v_tmp = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_expr = NULL;
  PyObject *__pyx_v_skipto_arg = NULL;
  PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_3And_8__init___2generator21 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  int __pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__ *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 3778, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_INCREF(__pyx_v_exprs);

  /* "Aspidites/_vendor/pyparsing_extension.py":3779
 * 
 *     def __init__(self, exprs, savelist=True):
 *         exprs = list(exprs)             # <<<<<<<<<<<<<<
 *         if exprs and Ellipsis in exprs:
 *             tmp = []
 */
  __pyx_t_1 = PySequence_List(__pyx_v_exprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3779, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_exprs, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3780
 *     def __init__(self, exprs, savelist=True):
 *         exprs = list(exprs)
 *         if exprs and Ellipsis in exprs:             # <<<<<<<<<<<<<<
 *             tmp = []
 *             for i, expr in enumerate(exprs):
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_exprs); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 3780, __pyx_L1_error)
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (__Pyx_PySequence_ContainsTF(__pyx_builtin_Ellipsis, __pyx_v_exprs, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 3780, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 != 0);
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3781
 *         exprs = list(exprs)
 *         if exprs and Ellipsis in exprs:
 *             tmp = []             # <<<<<<<<<<<<<<
 *             for i, expr in enumerate(exprs):
 *                 if expr is Ellipsis:
 */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3781, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_tmp = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3782
 *         if exprs and Ellipsis in exprs:
 *             tmp = []
 *             for i, expr in enumerate(exprs):             # <<<<<<<<<<<<<<
 *                 if expr is Ellipsis:
 *                     if i < len(exprs) - 1:
 */
    __Pyx_INCREF(__pyx_int_0);
    __pyx_t_1 = __pyx_int_0;
    if (likely(PyList_CheckExact(__pyx_v_exprs)) || PyTuple_CheckExact(__pyx_v_exprs)) {
      __pyx_t_5 = __pyx_v_exprs; __Pyx_INCREF(__pyx_t_5); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_v_exprs); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3782, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_7 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3782, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_5))) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_5)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_8 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_6); __Pyx_INCREF(__pyx_t_8); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 3782, __pyx_L1_error)
          #else
          __pyx_t_8 = PySequence_ITEM(__pyx_t_5, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3782, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          #endif
        } else {
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_6); __Pyx_INCREF(__pyx_t_8); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 3782, __pyx_L1_error)
          #else
          __pyx_t_8 = PySequence_ITEM(__pyx_t_5, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3782, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          #endif
        }
      } else {
        __pyx_t_8 = __pyx_t_7(__pyx_t_5);
        if (unlikely(!__pyx_t_8)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 3782, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_8);
      }
      __Pyx_XDECREF_SET(__pyx_v_expr, __pyx_t_8);
      __pyx_t_8 = 0;
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_1);
      __pyx_t_8 = __Pyx_PyInt_AddObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3782, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_1);
      __pyx_t_1 = __pyx_t_8;
      __pyx_t_8 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":3783
 *             tmp = []
 *             for i, expr in enumerate(exprs):
 *                 if expr is Ellipsis:             # <<<<<<<<<<<<<<
 *                     if i < len(exprs) - 1:
 *                         skipto_arg = (Empty() + exprs[i + 1]).exprs[-1]
 */
      __pyx_t_2 = (__pyx_v_expr == __pyx_builtin_Ellipsis);
      __pyx_t_4 = (__pyx_t_2 != 0);
      if (__pyx_t_4) {

        /* "Aspidites/_vendor/pyparsing_extension.py":3784
 *             for i, expr in enumerate(exprs):
 *                 if expr is Ellipsis:
 *                     if i < len(exprs) - 1:             # <<<<<<<<<<<<<<
 *                         skipto_arg = (Empty() + exprs[i + 1]).exprs[-1]
 *                         tmp.append(SkipTo(skipto_arg)("_skipped*"))
 */
        __pyx_t_9 = PyObject_Length(__pyx_v_exprs); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 3784, __pyx_L1_error)
        __pyx_t_8 = PyInt_FromSsize_t((__pyx_t_9 - 1)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3784, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_10 = PyObject_RichCompare(__pyx_v_i, __pyx_t_8, Py_LT); __Pyx_XGOTREF(__pyx_t_10); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 3784, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 3784, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (likely(__pyx_t_4)) {

          /* "Aspidites/_vendor/pyparsing_extension.py":3785
 *                 if expr is Ellipsis:
 *                     if i < len(exprs) - 1:
 *                         skipto_arg = (Empty() + exprs[i + 1]).exprs[-1]             # <<<<<<<<<<<<<<
 *                         tmp.append(SkipTo(skipto_arg)("_skipped*"))
 *                     else:
 */
          __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_Empty); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3785, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_11 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
            __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_8);
            if (likely(__pyx_t_11)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
              __Pyx_INCREF(__pyx_t_11);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_8, function);
            }
          }
          __pyx_t_10 = (__pyx_t_11) ? __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_11) : __Pyx_PyObject_CallNoArg(__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 3785, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __pyx_t_8 = __Pyx_PyInt_AddObjC(__pyx_v_i, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3785, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_11 = __Pyx_PyObject_GetItem(__pyx_v_exprs, __pyx_t_8); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 3785, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __pyx_t_8 = PyNumber_Add(__pyx_t_10, __pyx_t_11); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3785, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_exprs); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 3785, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __pyx_t_8 = __Pyx_GetItemInt(__pyx_t_11, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3785, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_XDECREF_SET(__pyx_v_skipto_arg, __pyx_t_8);
          __pyx_t_8 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":3786
 *                     if i < len(exprs) - 1:
 *                         skipto_arg = (Empty() + exprs[i + 1]).exprs[-1]
 *                         tmp.append(SkipTo(skipto_arg)("_skipped*"))             # <<<<<<<<<<<<<<
 *                     else:
 *                         raise Exception("cannot construct And with sequence ending in ...")
 */
          __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_SkipTo); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 3786, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_12 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
            __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_10);
            if (likely(__pyx_t_12)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
              __Pyx_INCREF(__pyx_t_12);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_10, function);
            }
          }
          __pyx_t_11 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_12, __pyx_v_skipto_arg) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_v_skipto_arg);
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 3786, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_t_10 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_10)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_10);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
            }
          }
          __pyx_t_8 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_10, __pyx_kp_s_skipped) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s_skipped);
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3786, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_tmp, __pyx_t_8); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 3786, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":3784
 *             for i, expr in enumerate(exprs):
 *                 if expr is Ellipsis:
 *                     if i < len(exprs) - 1:             # <<<<<<<<<<<<<<
 *                         skipto_arg = (Empty() + exprs[i + 1]).exprs[-1]
 *                         tmp.append(SkipTo(skipto_arg)("_skipped*"))
 */
          goto __pyx_L9;
        }

        /* "Aspidites/_vendor/pyparsing_extension.py":3788
 *                         tmp.append(SkipTo(skipto_arg)("_skipped*"))
 *                     else:
 *                         raise Exception("cannot construct And with sequence ending in ...")             # <<<<<<<<<<<<<<
 *                 else:
 *                     tmp.append(expr)
 */
        /*else*/ {
          __pyx_t_8 = __Pyx_PyObject_Call(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])), __pyx_tuple__108, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3788, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_Raise(__pyx_t_8, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __PYX_ERR(0, 3788, __pyx_L1_error)
        }
        __pyx_L9:;

        /* "Aspidites/_vendor/pyparsing_extension.py":3783
 *             tmp = []
 *             for i, expr in enumerate(exprs):
 *                 if expr is Ellipsis:             # <<<<<<<<<<<<<<
 *                     if i < len(exprs) - 1:
 *                         skipto_arg = (Empty() + exprs[i + 1]).exprs[-1]
 */
        goto __pyx_L8;
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":3790
 *                         raise Exception("cannot construct And with sequence ending in ...")
 *                 else:
 *                     tmp.append(expr)             # <<<<<<<<<<<<<<
 *             exprs[:] = tmp
 *         super(And, self).__init__(exprs, savelist)
 */
      /*else*/ {
        __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_tmp, __pyx_v_expr); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 3790, __pyx_L1_error)
      }
      __pyx_L8:;

      /* "Aspidites/_vendor/pyparsing_extension.py":3782
 *         if exprs and Ellipsis in exprs:
 *             tmp = []
 *             for i, expr in enumerate(exprs):             # <<<<<<<<<<<<<<
 *                 if expr is Ellipsis:
 *                     if i < len(exprs) - 1:
 */
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3791
 *                 else:
 *                     tmp.append(expr)
 *             exprs[:] = tmp             # <<<<<<<<<<<<<<
 *         super(And, self).__init__(exprs, savelist)
 *         self.mayReturnEmpty = all(e.mayReturnEmpty for e in self.exprs)
 */
    if (__Pyx_PyObject_SetSlice(__pyx_v_exprs, __pyx_v_tmp, 0, 0, NULL, NULL, &__pyx_slice__11, 0, 0, 1) < 0) __PYX_ERR(0, 3791, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":3780
 *     def __init__(self, exprs, savelist=True):
 *         exprs = list(exprs)
 *         if exprs and Ellipsis in exprs:             # <<<<<<<<<<<<<<
 *             tmp = []
 *             for i, expr in enumerate(exprs):
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3792
 *                     tmp.append(expr)
 *             exprs[:] = tmp
 *         super(And, self).__init__(exprs, savelist)             # <<<<<<<<<<<<<<
 *         self.mayReturnEmpty = all(e.mayReturnEmpty for e in self.exprs)
 *         self.setWhitespaceChars(self.exprs[0].whiteChars)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_And); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_5);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_cur_scope->__pyx_v_self);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_8, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_init); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  __pyx_t_14 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
      __pyx_t_14 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_8)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_exprs, __pyx_v_savelist};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_14, 2+__pyx_t_14); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3792, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_exprs, __pyx_v_savelist};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_14, 2+__pyx_t_14); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3792, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_11 = PyTuple_New(2+__pyx_t_14); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 3792, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_exprs);
    __Pyx_GIVEREF(__pyx_v_exprs);
    PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_14, __pyx_v_exprs);
    __Pyx_INCREF(__pyx_v_savelist);
    __Pyx_GIVEREF(__pyx_v_savelist);
    PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_14, __pyx_v_savelist);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_11, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3792, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  }
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3793
 *             exprs[:] = tmp
 *         super(And, self).__init__(exprs, savelist)
 *         self.mayReturnEmpty = all(e.mayReturnEmpty for e in self.exprs)             # <<<<<<<<<<<<<<
 *         self.setWhitespaceChars(self.exprs[0].whiteChars)
 *         self.skipWhitespace = self.exprs[0].skipWhitespace
 */
  __pyx_t_1 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_3And_8__init___genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3793, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = __Pyx_Generator_Next(__pyx_t_1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3793, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_mayReturnEmpty, __pyx_t_8) < 0) __PYX_ERR(0, 3793, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3794
 *         super(And, self).__init__(exprs, savelist)
 *         self.mayReturnEmpty = all(e.mayReturnEmpty for e in self.exprs)
 *         self.setWhitespaceChars(self.exprs[0].whiteChars)             # <<<<<<<<<<<<<<
 *         self.skipWhitespace = self.exprs[0].skipWhitespace
 *         self.callPreparse = True
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_setWhitespaceChars); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3794, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 3794, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_11, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3794, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_whiteChars); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 3794, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_8 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_t_11) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_11);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3794, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3795
 *         self.mayReturnEmpty = all(e.mayReturnEmpty for e in self.exprs)
 *         self.setWhitespaceChars(self.exprs[0].whiteChars)
 *         self.skipWhitespace = self.exprs[0].skipWhitespace             # <<<<<<<<<<<<<<
 *         self.callPreparse = True
 * 
 */
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_8, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_skipWhitespace); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_skipWhitespace, __pyx_t_8) < 0) __PYX_ERR(0, 3795, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3796
 *         self.setWhitespaceChars(self.exprs[0].whiteChars)
 *         self.skipWhitespace = self.exprs[0].skipWhitespace
 *         self.callPreparse = True             # <<<<<<<<<<<<<<
 * 
 *     def streamline(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_callPreparse, Py_True) < 0) __PYX_ERR(0, 3796, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3778
 *             self.leaveWhitespace()
 * 
 *     def __init__(self, exprs, savelist=True):             # <<<<<<<<<<<<<<
 *         exprs = list(exprs)
 *         if exprs and Ellipsis in exprs:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.And.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tmp);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_expr);
  __Pyx_XDECREF(__pyx_v_skipto_arg);
  __Pyx_XDECREF(__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_3And_8__init___2generator21);
  __Pyx_XDECREF(__pyx_v_exprs);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3798
 *         self.callPreparse = True
 * 
 *     def streamline(self):             # <<<<<<<<<<<<<<
 *         # collapse any _PendingSkip's
 *         from .pyparsing import _PendingSkip
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_3And_3streamline(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_3And_3streamline = {"streamline", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_3And_3streamline, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_3And_3streamline(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("streamline (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_3And_2streamline(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_3And_10streamline_2generator22(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/pyparsing_extension.py":3802
 *         from .pyparsing import _PendingSkip
 *         if self.exprs:
 *             if any(isinstance(e, ParseExpression) and e.exprs and isinstance(e.exprs[-1], _PendingSkip)             # <<<<<<<<<<<<<<
 *                    for e in self.exprs[:-1]):
 *                 for i, e in enumerate(self.exprs[:-1]):
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_3And_10streamline_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 3802, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_3And_10streamline_2generator22, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_And_streamline_locals_genexpr, __pyx_n_s_Aspidites__vendor_pyparsing_exte); if (unlikely(!gen)) __PYX_ERR(0, 3802, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.And.streamline.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_3And_10streamline_2generator22(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 3802, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3803
 *         if self.exprs:
 *             if any(isinstance(e, ParseExpression) and e.exprs and isinstance(e.exprs[-1], _PendingSkip)
 *                    for e in self.exprs[:-1]):             # <<<<<<<<<<<<<<
 *                 for i, e in enumerate(self.exprs[:-1]):
 *                     if e is None:
 */
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 3803, __pyx_L1_error) }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3803, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetSlice(__pyx_t_1, 0, -1L, NULL, NULL, &__pyx_slice__34, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3803, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3803, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3803, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 3803, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3803, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 3803, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3803, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_4(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 3803, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_e);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_e, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3802
 *         from .pyparsing import _PendingSkip
 *         if self.exprs:
 *             if any(isinstance(e, ParseExpression) and e.exprs and isinstance(e.exprs[-1], _PendingSkip)             # <<<<<<<<<<<<<<
 *                    for e in self.exprs[:-1]):
 *                 for i, e in enumerate(self.exprs[:-1]):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParseExpression); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3802, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = PyObject_IsInstance(__pyx_cur_scope->__pyx_v_e, __pyx_t_2); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 3802, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_7 = (__pyx_t_6 != 0);
    if (__pyx_t_7) {
    } else {
      __pyx_t_5 = __pyx_t_7;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_e, __pyx_n_s_exprs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3802, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 3802, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_7) {
    } else {
      __pyx_t_5 = __pyx_t_7;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_e, __pyx_n_s_exprs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3802, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_8 = __Pyx_GetItemInt(__pyx_t_2, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3802, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v__PendingSkip)) { __Pyx_RaiseClosureNameError("_PendingSkip"); __PYX_ERR(0, 3802, __pyx_L1_error) }
    __pyx_t_2 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v__PendingSkip;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_7 = PyObject_IsInstance(__pyx_t_8, __pyx_t_2); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 3802, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = (__pyx_t_7 != 0);
    __pyx_t_5 = __pyx_t_6;
    __pyx_L7_bool_binop_done:;
    if (__pyx_t_5) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_True);
      __pyx_r = Py_True;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":3803
 *         if self.exprs:
 *             if any(isinstance(e, ParseExpression) and e.exprs and isinstance(e.exprs[-1], _PendingSkip)
 *                    for e in self.exprs[:-1]):             # <<<<<<<<<<<<<<
 *                 for i, e in enumerate(self.exprs[:-1]):
 *                     if e is None:
 */
  }
  /*else*/ {

    /* "Aspidites/_vendor/pyparsing_extension.py":3802
 *         from .pyparsing import _PendingSkip
 *         if self.exprs:
 *             if any(isinstance(e, ParseExpression) and e.exprs and isinstance(e.exprs[-1], _PendingSkip)             # <<<<<<<<<<<<<<
 *                    for e in self.exprs[:-1]):
 *                 for i, e in enumerate(self.exprs[:-1]):
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3803
 *         if self.exprs:
 *             if any(isinstance(e, ParseExpression) and e.exprs and isinstance(e.exprs[-1], _PendingSkip)
 *                    for e in self.exprs[:-1]):             # <<<<<<<<<<<<<<
 *                 for i, e in enumerate(self.exprs[:-1]):
 *                     if e is None:
 */
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "Aspidites/_vendor/pyparsing_extension.py":3802
 *         from .pyparsing import _PendingSkip
 *         if self.exprs:
 *             if any(isinstance(e, ParseExpression) and e.exprs and isinstance(e.exprs[-1], _PendingSkip)             # <<<<<<<<<<<<<<
 *                    for e in self.exprs[:-1]):
 *                 for i, e in enumerate(self.exprs[:-1]):
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_3And_10streamline_5generator23(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/pyparsing_extension.py":3814
 * 
 *         super(And, self).streamline()
 *         self.mayReturnEmpty = all(e.mayReturnEmpty for e in self.exprs)             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_3And_10streamline_3genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 3814, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_3And_10streamline_5generator23, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_And_streamline_locals_genexpr, __pyx_n_s_Aspidites__vendor_pyparsing_exte); if (unlikely(!gen)) __PYX_ERR(0, 3814, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.And.streamline.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_3And_10streamline_5generator23(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 3814, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 3814, __pyx_L1_error) }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3814, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3814, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3814, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 3814, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3814, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 3814, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3814, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 3814, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_e);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_e, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_e, __pyx_n_s_mayReturnEmpty); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3814, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 3814, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = ((!__pyx_t_5) != 0);
    if (__pyx_t_6) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L0;
    }
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_True);
    __pyx_r = Py_True;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3798
 *         self.callPreparse = True
 * 
 *     def streamline(self):             # <<<<<<<<<<<<<<
 *         # collapse any _PendingSkip's
 *         from .pyparsing import _PendingSkip
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_3And_2streamline(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline *__pyx_cur_scope;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_3And_10streamline_2generator22 = 0;
  PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_3And_10streamline_5generator23 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("streamline", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 3798, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);

  /* "Aspidites/_vendor/pyparsing_extension.py":3800
 *     def streamline(self):
 *         # collapse any _PendingSkip's
 *         from .pyparsing import _PendingSkip             # <<<<<<<<<<<<<<
 *         if self.exprs:
 *             if any(isinstance(e, ParseExpression) and e.exprs and isinstance(e.exprs[-1], _PendingSkip)
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3800, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_PendingSkip);
  __Pyx_GIVEREF(__pyx_n_s_PendingSkip);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PendingSkip);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_pyparsing, __pyx_t_1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3800, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_PendingSkip); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3800, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v__PendingSkip = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3801
 *         # collapse any _PendingSkip's
 *         from .pyparsing import _PendingSkip
 *         if self.exprs:             # <<<<<<<<<<<<<<
 *             if any(isinstance(e, ParseExpression) and e.exprs and isinstance(e.exprs[-1], _PendingSkip)
 *                    for e in self.exprs[:-1]):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 3801, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3802
 *         from .pyparsing import _PendingSkip
 *         if self.exprs:
 *             if any(isinstance(e, ParseExpression) and e.exprs and isinstance(e.exprs[-1], _PendingSkip)             # <<<<<<<<<<<<<<
 *                    for e in self.exprs[:-1]):
 *                 for i, e in enumerate(self.exprs[:-1]):
 */
    __pyx_t_2 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_3And_10streamline_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3802, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_Generator_Next(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3802, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 3802, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_3) {

      /* "Aspidites/_vendor/pyparsing_extension.py":3804
 *             if any(isinstance(e, ParseExpression) and e.exprs and isinstance(e.exprs[-1], _PendingSkip)
 *                    for e in self.exprs[:-1]):
 *                 for i, e in enumerate(self.exprs[:-1]):             # <<<<<<<<<<<<<<
 *                     if e is None:
 *                         continue
 */
      __Pyx_INCREF(__pyx_int_0);
      __pyx_t_1 = __pyx_int_0;
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3804, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = __Pyx_PyObject_GetSlice(__pyx_t_2, 0, -1L, NULL, NULL, &__pyx_slice__34, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3804, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
        __pyx_t_2 = __pyx_t_4; __Pyx_INCREF(__pyx_t_2); __pyx_t_5 = 0;
        __pyx_t_6 = NULL;
      } else {
        __pyx_t_5 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3804, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_6 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3804, __pyx_L1_error)
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      for (;;) {
        if (likely(!__pyx_t_6)) {
          if (likely(PyList_CheckExact(__pyx_t_2))) {
            if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_2)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_4 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_4); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 3804, __pyx_L1_error)
            #else
            __pyx_t_4 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3804, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            #endif
          } else {
            if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_4); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 3804, __pyx_L1_error)
            #else
            __pyx_t_4 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3804, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            #endif
          }
        } else {
          __pyx_t_4 = __pyx_t_6(__pyx_t_2);
          if (unlikely(!__pyx_t_4)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 3804, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_4);
        }
        __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_4);
        __pyx_t_4 = 0;
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_1);
        __pyx_t_4 = __Pyx_PyInt_AddObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3804, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1);
        __pyx_t_1 = __pyx_t_4;
        __pyx_t_4 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":3805
 *                    for e in self.exprs[:-1]):
 *                 for i, e in enumerate(self.exprs[:-1]):
 *                     if e is None:             # <<<<<<<<<<<<<<
 *                         continue
 *                     if (isinstance(e, ParseExpression)
 */
        __pyx_t_3 = (__pyx_v_e == Py_None);
        __pyx_t_7 = (__pyx_t_3 != 0);
        if (__pyx_t_7) {

          /* "Aspidites/_vendor/pyparsing_extension.py":3806
 *                 for i, e in enumerate(self.exprs[:-1]):
 *                     if e is None:
 *                         continue             # <<<<<<<<<<<<<<
 *                     if (isinstance(e, ParseExpression)
 *                             and e.exprs and isinstance(e.exprs[-1], _PendingSkip)):
 */
          goto __pyx_L5_continue;

          /* "Aspidites/_vendor/pyparsing_extension.py":3805
 *                    for e in self.exprs[:-1]):
 *                 for i, e in enumerate(self.exprs[:-1]):
 *                     if e is None:             # <<<<<<<<<<<<<<
 *                         continue
 *                     if (isinstance(e, ParseExpression)
 */
        }

        /* "Aspidites/_vendor/pyparsing_extension.py":3807
 *                     if e is None:
 *                         continue
 *                     if (isinstance(e, ParseExpression)             # <<<<<<<<<<<<<<
 *                             and e.exprs and isinstance(e.exprs[-1], _PendingSkip)):
 *                         e.exprs[-1] = e.exprs[-1] + self.exprs[i + 1]
 */
        __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ParseExpression); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3807, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_3 = PyObject_IsInstance(__pyx_v_e, __pyx_t_4); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 3807, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_8 = (__pyx_t_3 != 0);
        if (__pyx_t_8) {
        } else {
          __pyx_t_7 = __pyx_t_8;
          goto __pyx_L9_bool_binop_done;
        }

        /* "Aspidites/_vendor/pyparsing_extension.py":3808
 *                         continue
 *                     if (isinstance(e, ParseExpression)
 *                             and e.exprs and isinstance(e.exprs[-1], _PendingSkip)):             # <<<<<<<<<<<<<<
 *                         e.exprs[-1] = e.exprs[-1] + self.exprs[i + 1]
 *                         self.exprs[i + 1] = None
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_exprs); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3808, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 3808, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (__pyx_t_8) {
        } else {
          __pyx_t_7 = __pyx_t_8;
          goto __pyx_L9_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_exprs); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3808, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_9 = __Pyx_GetItemInt(__pyx_t_4, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3808, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __pyx_cur_scope->__pyx_v__PendingSkip;
        __Pyx_INCREF(__pyx_t_4);
        __pyx_t_8 = PyObject_IsInstance(__pyx_t_9, __pyx_t_4); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 3808, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_3 = (__pyx_t_8 != 0);
        __pyx_t_7 = __pyx_t_3;
        __pyx_L9_bool_binop_done:;

        /* "Aspidites/_vendor/pyparsing_extension.py":3807
 *                     if e is None:
 *                         continue
 *                     if (isinstance(e, ParseExpression)             # <<<<<<<<<<<<<<
 *                             and e.exprs and isinstance(e.exprs[-1], _PendingSkip)):
 *                         e.exprs[-1] = e.exprs[-1] + self.exprs[i + 1]
 */
        if (__pyx_t_7) {

          /* "Aspidites/_vendor/pyparsing_extension.py":3809
 *                     if (isinstance(e, ParseExpression)
 *                             and e.exprs and isinstance(e.exprs[-1], _PendingSkip)):
 *                         e.exprs[-1] = e.exprs[-1] + self.exprs[i + 1]             # <<<<<<<<<<<<<<
 *                         self.exprs[i + 1] = None
 *                 self.exprs = [e for e in self.exprs if e is not None]
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_exprs); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3809, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_9 = __Pyx_GetItemInt(__pyx_t_4, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3809, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3809, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_10 = __Pyx_PyInt_AddObjC(__pyx_v_i, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 3809, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_11 = __Pyx_PyObject_GetItem(__pyx_t_4, __pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 3809, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_t_10 = PyNumber_Add(__pyx_t_9, __pyx_t_11); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 3809, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_exprs); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 3809, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          if (unlikely(__Pyx_SetItemInt(__pyx_t_11, -1L, __pyx_t_10, long, 1, __Pyx_PyInt_From_long, 0, 1, 1) < 0)) __PYX_ERR(0, 3809, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":3810
 *                             and e.exprs and isinstance(e.exprs[-1], _PendingSkip)):
 *                         e.exprs[-1] = e.exprs[-1] + self.exprs[i + 1]
 *                         self.exprs[i + 1] = None             # <<<<<<<<<<<<<<
 *                 self.exprs = [e for e in self.exprs if e is not None]
 * 
 */
          __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 3810, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_11 = __Pyx_PyInt_AddObjC(__pyx_v_i, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 3810, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          if (unlikely(PyObject_SetItem(__pyx_t_10, __pyx_t_11, Py_None) < 0)) __PYX_ERR(0, 3810, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":3807
 *                     if e is None:
 *                         continue
 *                     if (isinstance(e, ParseExpression)             # <<<<<<<<<<<<<<
 *                             and e.exprs and isinstance(e.exprs[-1], _PendingSkip)):
 *                         e.exprs[-1] = e.exprs[-1] + self.exprs[i + 1]
 */
        }

        /* "Aspidites/_vendor/pyparsing_extension.py":3804
 *             if any(isinstance(e, ParseExpression) and e.exprs and isinstance(e.exprs[-1], _PendingSkip)
 *                    for e in self.exprs[:-1]):
 *                 for i, e in enumerate(self.exprs[:-1]):             # <<<<<<<<<<<<<<
 *                     if e is None:
 *                         continue
 */
        __pyx_L5_continue:;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":3811
 *                         e.exprs[-1] = e.exprs[-1] + self.exprs[i + 1]
 *                         self.exprs[i + 1] = None
 *                 self.exprs = [e for e in self.exprs if e is not None]             # <<<<<<<<<<<<<<
 * 
 *         super(And, self).streamline()
 */
      __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3811, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3811, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
        __pyx_t_11 = __pyx_t_2; __Pyx_INCREF(__pyx_t_11); __pyx_t_5 = 0;
        __pyx_t_6 = NULL;
      } else {
        __pyx_t_5 = -1; __pyx_t_11 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 3811, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_6 = Py_TYPE(__pyx_t_11)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3811, __pyx_L1_error)
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      for (;;) {
        if (likely(!__pyx_t_6)) {
          if (likely(PyList_CheckExact(__pyx_t_11))) {
            if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_11)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = PyList_GET_ITEM(__pyx_t_11, __pyx_t_5); __Pyx_INCREF(__pyx_t_2); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 3811, __pyx_L1_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_11, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3811, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
          } else {
            if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_11)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_11, __pyx_t_5); __Pyx_INCREF(__pyx_t_2); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 3811, __pyx_L1_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_11, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3811, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
          }
        } else {
          __pyx_t_2 = __pyx_t_6(__pyx_t_11);
          if (unlikely(!__pyx_t_2)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 3811, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_2);
        }
        __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_2);
        __pyx_t_2 = 0;
        __pyx_t_7 = (__pyx_v_e != Py_None);
        __pyx_t_3 = (__pyx_t_7 != 0);
        if (__pyx_t_3) {
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_v_e))) __PYX_ERR(0, 3811, __pyx_L1_error)
        }
      }
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_exprs, __pyx_t_1) < 0) __PYX_ERR(0, 3811, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":3802
 *         from .pyparsing import _PendingSkip
 *         if self.exprs:
 *             if any(isinstance(e, ParseExpression) and e.exprs and isinstance(e.exprs[-1], _PendingSkip)             # <<<<<<<<<<<<<<
 *                    for e in self.exprs[:-1]):
 *                 for i, e in enumerate(self.exprs[:-1]):
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":3801
 *         # collapse any _PendingSkip's
 *         from .pyparsing import _PendingSkip
 *         if self.exprs:             # <<<<<<<<<<<<<<
 *             if any(isinstance(e, ParseExpression) and e.exprs and isinstance(e.exprs[-1], _PendingSkip)
 *                    for e in self.exprs[:-1]):
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3813
 *                 self.exprs = [e for e in self.exprs if e is not None]
 * 
 *         super(And, self).streamline()             # <<<<<<<<<<<<<<
 *         self.mayReturnEmpty = all(e.mayReturnEmpty for e in self.exprs)
 *         return self
 */
  __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_And); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 3813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_11);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_cur_scope->__pyx_v_self);
  __pyx_t_11 = 0;
  __pyx_t_11 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_2, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 3813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_streamline); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_11 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_11)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_11);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_11) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_11) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3814
 * 
 *         super(And, self).streamline()
 *         self.mayReturnEmpty = all(e.mayReturnEmpty for e in self.exprs)             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __pyx_t_1 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_3And_10streamline_3genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3814, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_Generator_Next(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3814, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_mayReturnEmpty, __pyx_t_2) < 0) __PYX_ERR(0, 3814, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3815
 *         super(And, self).streamline()
 *         self.mayReturnEmpty = all(e.mayReturnEmpty for e in self.exprs)
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __pyx_r = __pyx_cur_scope->__pyx_v_self;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3798
 *         self.callPreparse = True
 * 
 *     def streamline(self):             # <<<<<<<<<<<<<<
 *         # collapse any _PendingSkip's
 *         from .pyparsing import _PendingSkip
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.And.streamline", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XDECREF(__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_3And_10streamline_2generator22);
  __Pyx_XDECREF(__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_3And_10streamline_5generator23);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3817
 *         return self
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         # pass False as last arg to _parse for first element, since we already
 *         # pre-parsed the string as part of our And pre-parsing
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_3And_5parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_3And_5parseImpl = {"parseImpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_3And_5parseImpl, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_3And_5parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  PyObject *__pyx_v_doActions = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parseImpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_doActions,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 1); __PYX_ERR(0, 3817, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 2); __PYX_ERR(0, 3817, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_doActions);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "parseImpl") < 0)) __PYX_ERR(0, 3817, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_doActions = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3817, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.And.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_3And_4parseImpl(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_3And_4parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, PyObject *__pyx_v_doActions) {
  PyObject *__pyx_v_resultlist = NULL;
  int __pyx_v_errorStop;
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_v_exprtokens = NULL;
  PyObject *__pyx_v_pe = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  Py_ssize_t __pyx_t_17;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parseImpl", 0);
  __Pyx_INCREF(__pyx_v_loc);

  /* "Aspidites/_vendor/pyparsing_extension.py":3820
 *         # pass False as last arg to _parse for first element, since we already
 *         # pre-parsed the string as part of our And pre-parsing
 *         loc, resultlist = self.exprs[0]._parse(instring, loc, doActions, callPreParse=False)             # <<<<<<<<<<<<<<
 *         errorStop = False
 *         for e in self.exprs[1:]:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_parse); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_instring);
  __Pyx_GIVEREF(__pyx_v_instring);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_instring);
  __Pyx_INCREF(__pyx_v_loc);
  __Pyx_GIVEREF(__pyx_v_loc);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_loc);
  __Pyx_INCREF(__pyx_v_doActions);
  __Pyx_GIVEREF(__pyx_v_doActions);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_doActions);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_callPreParse, Py_False) < 0) __PYX_ERR(0, 3820, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
    PyObject* sequence = __pyx_t_4;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 3820, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_2);
    #else
    __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3820, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3820, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_1 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3820, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = Py_TYPE(__pyx_t_1)->tp_iternext;
    index = 0; __pyx_t_3 = __pyx_t_5(__pyx_t_1); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    index = 1; __pyx_t_2 = __pyx_t_5(__pyx_t_1); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_1), 2) < 0) __PYX_ERR(0, 3820, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 3820, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __Pyx_DECREF_SET(__pyx_v_loc, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_v_resultlist = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3821
 *         # pre-parsed the string as part of our And pre-parsing
 *         loc, resultlist = self.exprs[0]._parse(instring, loc, doActions, callPreParse=False)
 *         errorStop = False             # <<<<<<<<<<<<<<
 *         for e in self.exprs[1:]:
 *             if isinstance(e, And._ErrorStop):
 */
  __pyx_v_errorStop = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3822
 *         loc, resultlist = self.exprs[0]._parse(instring, loc, doActions, callPreParse=False)
 *         errorStop = False
 *         for e in self.exprs[1:]:             # <<<<<<<<<<<<<<
 *             if isinstance(e, And._ErrorStop):
 *                 errorStop = True
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyObject_GetSlice(__pyx_t_4, 1, 0, NULL, NULL, &__pyx_slice__69, 1, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_4 = __pyx_t_2; __Pyx_INCREF(__pyx_t_4); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3822, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3822, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_4))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 3822, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3822, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 3822, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3822, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_7(__pyx_t_4);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 3822, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3823
 *         errorStop = False
 *         for e in self.exprs[1:]:
 *             if isinstance(e, And._ErrorStop):             # <<<<<<<<<<<<<<
 *                 errorStop = True
 *                 continue
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_And); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3823, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ErrorStop); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3823, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_8 = PyObject_IsInstance(__pyx_v_e, __pyx_t_3); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 3823, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_9 = (__pyx_t_8 != 0);
    if (__pyx_t_9) {

      /* "Aspidites/_vendor/pyparsing_extension.py":3824
 *         for e in self.exprs[1:]:
 *             if isinstance(e, And._ErrorStop):
 *                 errorStop = True             # <<<<<<<<<<<<<<
 *                 continue
 *             if errorStop:
 */
      __pyx_v_errorStop = 1;

      /* "Aspidites/_vendor/pyparsing_extension.py":3825
 *             if isinstance(e, And._ErrorStop):
 *                 errorStop = True
 *                 continue             # <<<<<<<<<<<<<<
 *             if errorStop:
 *                 try:
 */
      goto __pyx_L5_continue;

      /* "Aspidites/_vendor/pyparsing_extension.py":3823
 *         errorStop = False
 *         for e in self.exprs[1:]:
 *             if isinstance(e, And._ErrorStop):             # <<<<<<<<<<<<<<
 *                 errorStop = True
 *                 continue
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":3826
 *                 errorStop = True
 *                 continue
 *             if errorStop:             # <<<<<<<<<<<<<<
 *                 try:
 *                     loc, exprtokens = e._parse(instring, loc, doActions)
 */
    __pyx_t_9 = (__pyx_v_errorStop != 0);
    if (__pyx_t_9) {

      /* "Aspidites/_vendor/pyparsing_extension.py":3827
 *                 continue
 *             if errorStop:
 *                 try:             # <<<<<<<<<<<<<<
 *                     loc, exprtokens = e._parse(instring, loc, doActions)
 *                 except ParseSyntaxException:
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
        __Pyx_XGOTREF(__pyx_t_10);
        __Pyx_XGOTREF(__pyx_t_11);
        __Pyx_XGOTREF(__pyx_t_12);
        /*try:*/ {

          /* "Aspidites/_vendor/pyparsing_extension.py":3828
 *             if errorStop:
 *                 try:
 *                     loc, exprtokens = e._parse(instring, loc, doActions)             # <<<<<<<<<<<<<<
 *                 except ParseSyntaxException:
 *                     raise
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_parse); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3828, __pyx_L9_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_1 = NULL;
          __pyx_t_13 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_13 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[4] = {__pyx_t_1, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_13, 3+__pyx_t_13); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3828, __pyx_L9_error)
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
            PyObject *__pyx_temp[4] = {__pyx_t_1, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_13, 3+__pyx_t_13); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3828, __pyx_L9_error)
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_14 = PyTuple_New(3+__pyx_t_13); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 3828, __pyx_L9_error)
            __Pyx_GOTREF(__pyx_t_14);
            if (__pyx_t_1) {
              __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_1); __pyx_t_1 = NULL;
            }
            __Pyx_INCREF(__pyx_v_instring);
            __Pyx_GIVEREF(__pyx_v_instring);
            PyTuple_SET_ITEM(__pyx_t_14, 0+__pyx_t_13, __pyx_v_instring);
            __Pyx_INCREF(__pyx_v_loc);
            __Pyx_GIVEREF(__pyx_v_loc);
            PyTuple_SET_ITEM(__pyx_t_14, 1+__pyx_t_13, __pyx_v_loc);
            __Pyx_INCREF(__pyx_v_doActions);
            __Pyx_GIVEREF(__pyx_v_doActions);
            PyTuple_SET_ITEM(__pyx_t_14, 2+__pyx_t_13, __pyx_v_doActions);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_14, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3828, __pyx_L9_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
            PyObject* sequence = __pyx_t_3;
            Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
            if (unlikely(size != 2)) {
              if (size > 2) __Pyx_RaiseTooManyValuesError(2);
              else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
              __PYX_ERR(0, 3828, __pyx_L9_error)
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            if (likely(PyTuple_CheckExact(sequence))) {
              __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
              __pyx_t_14 = PyTuple_GET_ITEM(sequence, 1); 
            } else {
              __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
              __pyx_t_14 = PyList_GET_ITEM(sequence, 1); 
            }
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_14);
            #else
            __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3828, __pyx_L9_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_14 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 3828, __pyx_L9_error)
            __Pyx_GOTREF(__pyx_t_14);
            #endif
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          } else {
            Py_ssize_t index = -1;
            __pyx_t_1 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3828, __pyx_L9_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_5 = Py_TYPE(__pyx_t_1)->tp_iternext;
            index = 0; __pyx_t_2 = __pyx_t_5(__pyx_t_1); if (unlikely(!__pyx_t_2)) goto __pyx_L17_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_2);
            index = 1; __pyx_t_14 = __pyx_t_5(__pyx_t_1); if (unlikely(!__pyx_t_14)) goto __pyx_L17_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_14);
            if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_1), 2) < 0) __PYX_ERR(0, 3828, __pyx_L9_error)
            __pyx_t_5 = NULL;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            goto __pyx_L18_unpacking_done;
            __pyx_L17_unpacking_failed:;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_5 = NULL;
            if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
            __PYX_ERR(0, 3828, __pyx_L9_error)
            __pyx_L18_unpacking_done:;
          }
          __Pyx_DECREF_SET(__pyx_v_loc, __pyx_t_2);
          __pyx_t_2 = 0;
          __Pyx_XDECREF_SET(__pyx_v_exprtokens, __pyx_t_14);
          __pyx_t_14 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":3827
 *                 continue
 *             if errorStop:
 *                 try:             # <<<<<<<<<<<<<<
 *                     loc, exprtokens = e._parse(instring, loc, doActions)
 *                 except ParseSyntaxException:
 */
        }
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        goto __pyx_L16_try_end;
        __pyx_L9_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":3829
 *                 try:
 *                     loc, exprtokens = e._parse(instring, loc, doActions)
 *                 except ParseSyntaxException:             # <<<<<<<<<<<<<<
 *                     raise
 *                 except ParseBaseException as pe:
 */
        __Pyx_ErrFetch(&__pyx_t_3, &__pyx_t_14, &__pyx_t_2);
        __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParseSyntaxException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3829, __pyx_L11_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_13 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_3, __pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_ErrRestore(__pyx_t_3, __pyx_t_14, __pyx_t_2);
        __pyx_t_3 = 0; __pyx_t_14 = 0; __pyx_t_2 = 0;
        if (__pyx_t_13) {
          __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.And.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_14, &__pyx_t_3) < 0) __PYX_ERR(0, 3829, __pyx_L11_except_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_GOTREF(__pyx_t_3);

          /* "Aspidites/_vendor/pyparsing_extension.py":3830
 *                     loc, exprtokens = e._parse(instring, loc, doActions)
 *                 except ParseSyntaxException:
 *                     raise             # <<<<<<<<<<<<<<
 *                 except ParseBaseException as pe:
 *                     pe.__traceback__ = None
 */
          __Pyx_GIVEREF(__pyx_t_2);
          __Pyx_GIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_3);
          __Pyx_ErrRestoreWithState(__pyx_t_2, __pyx_t_14, __pyx_t_3);
          __pyx_t_2 = 0; __pyx_t_14 = 0; __pyx_t_3 = 0; 
          __PYX_ERR(0, 3830, __pyx_L11_except_error)
        }

        /* "Aspidites/_vendor/pyparsing_extension.py":3831
 *                 except ParseSyntaxException:
 *                     raise
 *                 except ParseBaseException as pe:             # <<<<<<<<<<<<<<
 *                     pe.__traceback__ = None
 *                     raise ParseSyntaxException._from_exception(pe)
 */
        __Pyx_ErrFetch(&__pyx_t_3, &__pyx_t_14, &__pyx_t_2);
        __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParseBaseException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3831, __pyx_L11_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_13 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_3, __pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_ErrRestore(__pyx_t_3, __pyx_t_14, __pyx_t_2);
        __pyx_t_3 = 0; __pyx_t_14 = 0; __pyx_t_2 = 0;
        if (__pyx_t_13) {
          __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.And.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_14, &__pyx_t_3) < 0) __PYX_ERR(0, 3831, __pyx_L11_except_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_14);
          __pyx_v_pe = __pyx_t_14;

          /* "Aspidites/_vendor/pyparsing_extension.py":3832
 *                     raise
 *                 except ParseBaseException as pe:
 *                     pe.__traceback__ = None             # <<<<<<<<<<<<<<
 *                     raise ParseSyntaxException._from_exception(pe)
 *                 except IndexError:
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_pe, __pyx_n_s_traceback, Py_None) < 0) __PYX_ERR(0, 3832, __pyx_L11_except_error)

          /* "Aspidites/_vendor/pyparsing_extension.py":3833
 *                 except ParseBaseException as pe:
 *                     pe.__traceback__ = None
 *                     raise ParseSyntaxException._from_exception(pe)             # <<<<<<<<<<<<<<
 *                 except IndexError:
 *                     raise ParseSyntaxException(instring, len(instring), self.errmsg, self)
 */
          __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_ParseSyntaxException); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 3833, __pyx_L11_except_error)
          __Pyx_GOTREF(__pyx_t_15);
          __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_from_exception); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 3833, __pyx_L11_except_error)
          __Pyx_GOTREF(__pyx_t_16);
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          __pyx_t_15 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_16))) {
            __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_16);
            if (likely(__pyx_t_15)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_16);
              __Pyx_INCREF(__pyx_t_15);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_16, function);
            }
          }
          __pyx_t_1 = (__pyx_t_15) ? __Pyx_PyObject_Call2Args(__pyx_t_16, __pyx_t_15, __pyx_v_pe) : __Pyx_PyObject_CallOneArg(__pyx_t_16, __pyx_v_pe);
          __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3833, __pyx_L11_except_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          __Pyx_Raise(__pyx_t_1, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __PYX_ERR(0, 3833, __pyx_L11_except_error)
        }

        /* "Aspidites/_vendor/pyparsing_extension.py":3834
 *                     pe.__traceback__ = None
 *                     raise ParseSyntaxException._from_exception(pe)
 *                 except IndexError:             # <<<<<<<<<<<<<<
 *                     raise ParseSyntaxException(instring, len(instring), self.errmsg, self)
 *             else:
 */
        __pyx_t_13 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_IndexError);
        if (__pyx_t_13) {
          __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.And.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_14, &__pyx_t_2) < 0) __PYX_ERR(0, 3834, __pyx_L11_except_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_GOTREF(__pyx_t_2);

          /* "Aspidites/_vendor/pyparsing_extension.py":3835
 *                     raise ParseSyntaxException._from_exception(pe)
 *                 except IndexError:
 *                     raise ParseSyntaxException(instring, len(instring), self.errmsg, self)             # <<<<<<<<<<<<<<
 *             else:
 *                 loc, exprtokens = e._parse(instring, loc, doActions)
 */
          __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_n_s_ParseSyntaxException); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 3835, __pyx_L11_except_error)
          __Pyx_GOTREF(__pyx_t_16);
          __pyx_t_17 = PyObject_Length(__pyx_v_instring); if (unlikely(__pyx_t_17 == ((Py_ssize_t)-1))) __PYX_ERR(0, 3835, __pyx_L11_except_error)
          __pyx_t_15 = PyInt_FromSsize_t(__pyx_t_17); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 3835, __pyx_L11_except_error)
          __Pyx_GOTREF(__pyx_t_15);
          __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errmsg); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 3835, __pyx_L11_except_error)
          __Pyx_GOTREF(__pyx_t_18);
          __pyx_t_19 = NULL;
          __pyx_t_13 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_16))) {
            __pyx_t_19 = PyMethod_GET_SELF(__pyx_t_16);
            if (likely(__pyx_t_19)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_16);
              __Pyx_INCREF(__pyx_t_19);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_16, function);
              __pyx_t_13 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_16)) {
            PyObject *__pyx_temp[5] = {__pyx_t_19, __pyx_v_instring, __pyx_t_15, __pyx_t_18, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_16, __pyx_temp+1-__pyx_t_13, 4+__pyx_t_13); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3835, __pyx_L11_except_error)
            __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_16)) {
            PyObject *__pyx_temp[5] = {__pyx_t_19, __pyx_v_instring, __pyx_t_15, __pyx_t_18, __pyx_v_self};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_16, __pyx_temp+1-__pyx_t_13, 4+__pyx_t_13); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3835, __pyx_L11_except_error)
            __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
          } else
          #endif
          {
            __pyx_t_20 = PyTuple_New(4+__pyx_t_13); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 3835, __pyx_L11_except_error)
            __Pyx_GOTREF(__pyx_t_20);
            if (__pyx_t_19) {
              __Pyx_GIVEREF(__pyx_t_19); PyTuple_SET_ITEM(__pyx_t_20, 0, __pyx_t_19); __pyx_t_19 = NULL;
            }
            __Pyx_INCREF(__pyx_v_instring);
            __Pyx_GIVEREF(__pyx_v_instring);
            PyTuple_SET_ITEM(__pyx_t_20, 0+__pyx_t_13, __pyx_v_instring);
            __Pyx_GIVEREF(__pyx_t_15);
            PyTuple_SET_ITEM(__pyx_t_20, 1+__pyx_t_13, __pyx_t_15);
            __Pyx_GIVEREF(__pyx_t_18);
            PyTuple_SET_ITEM(__pyx_t_20, 2+__pyx_t_13, __pyx_t_18);
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_20, 3+__pyx_t_13, __pyx_v_self);
            __pyx_t_15 = 0;
            __pyx_t_18 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_16, __pyx_t_20, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3835, __pyx_L11_except_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
          }
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          __Pyx_Raise(__pyx_t_1, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __PYX_ERR(0, 3835, __pyx_L11_except_error)
        }
        goto __pyx_L11_except_error;
        __pyx_L11_except_error:;

        /* "Aspidites/_vendor/pyparsing_extension.py":3827
 *                 continue
 *             if errorStop:
 *                 try:             # <<<<<<<<<<<<<<
 *                     loc, exprtokens = e._parse(instring, loc, doActions)
 *                 except ParseSyntaxException:
 */
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
        goto __pyx_L1_error;
        __pyx_L16_try_end:;
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":3826
 *                 errorStop = True
 *                 continue
 *             if errorStop:             # <<<<<<<<<<<<<<
 *                 try:
 *                     loc, exprtokens = e._parse(instring, loc, doActions)
 */
      goto __pyx_L8;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":3837
 *                     raise ParseSyntaxException(instring, len(instring), self.errmsg, self)
 *             else:
 *                 loc, exprtokens = e._parse(instring, loc, doActions)             # <<<<<<<<<<<<<<
 *             if exprtokens or exprtokens.haskeys():
 *                 resultlist += exprtokens
 */
    /*else*/ {
      __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_parse); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 3837, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_3 = NULL;
      __pyx_t_13 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_14))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_14);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_14, function);
          __pyx_t_13 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_14)) {
        PyObject *__pyx_temp[4] = {__pyx_t_3, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_14, __pyx_temp+1-__pyx_t_13, 3+__pyx_t_13); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3837, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_14)) {
        PyObject *__pyx_temp[4] = {__pyx_t_3, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_14, __pyx_temp+1-__pyx_t_13, 3+__pyx_t_13); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3837, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else
      #endif
      {
        __pyx_t_1 = PyTuple_New(3+__pyx_t_13); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3837, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (__pyx_t_3) {
          __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3); __pyx_t_3 = NULL;
        }
        __Pyx_INCREF(__pyx_v_instring);
        __Pyx_GIVEREF(__pyx_v_instring);
        PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_13, __pyx_v_instring);
        __Pyx_INCREF(__pyx_v_loc);
        __Pyx_GIVEREF(__pyx_v_loc);
        PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_13, __pyx_v_loc);
        __Pyx_INCREF(__pyx_v_doActions);
        __Pyx_GIVEREF(__pyx_v_doActions);
        PyTuple_SET_ITEM(__pyx_t_1, 2+__pyx_t_13, __pyx_v_doActions);
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_14, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3837, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
        PyObject* sequence = __pyx_t_2;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 3837, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_14 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_1 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_14 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_1 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_14);
        __Pyx_INCREF(__pyx_t_1);
        #else
        __pyx_t_14 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 3837, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
        __pyx_t_1 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3837, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3837, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_5 = Py_TYPE(__pyx_t_3)->tp_iternext;
        index = 0; __pyx_t_14 = __pyx_t_5(__pyx_t_3); if (unlikely(!__pyx_t_14)) goto __pyx_L25_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_14);
        index = 1; __pyx_t_1 = __pyx_t_5(__pyx_t_3); if (unlikely(!__pyx_t_1)) goto __pyx_L25_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_1);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_3), 2) < 0) __PYX_ERR(0, 3837, __pyx_L1_error)
        __pyx_t_5 = NULL;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        goto __pyx_L26_unpacking_done;
        __pyx_L25_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 3837, __pyx_L1_error)
        __pyx_L26_unpacking_done:;
      }
      __Pyx_DECREF_SET(__pyx_v_loc, __pyx_t_14);
      __pyx_t_14 = 0;
      __Pyx_XDECREF_SET(__pyx_v_exprtokens, __pyx_t_1);
      __pyx_t_1 = 0;
    }
    __pyx_L8:;

    /* "Aspidites/_vendor/pyparsing_extension.py":3838
 *             else:
 *                 loc, exprtokens = e._parse(instring, loc, doActions)
 *             if exprtokens or exprtokens.haskeys():             # <<<<<<<<<<<<<<
 *                 resultlist += exprtokens
 *         return loc, resultlist
 */
    __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_v_exprtokens); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 3838, __pyx_L1_error)
    if (!__pyx_t_8) {
    } else {
      __pyx_t_9 = __pyx_t_8;
      goto __pyx_L28_bool_binop_done;
    }
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_exprtokens, __pyx_n_s_haskeys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3838, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_14 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_14)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_14);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_14) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_14) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3838, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 3838, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_9 = __pyx_t_8;
    __pyx_L28_bool_binop_done:;
    if (__pyx_t_9) {

      /* "Aspidites/_vendor/pyparsing_extension.py":3839
 *                 loc, exprtokens = e._parse(instring, loc, doActions)
 *             if exprtokens or exprtokens.haskeys():
 *                 resultlist += exprtokens             # <<<<<<<<<<<<<<
 *         return loc, resultlist
 * 
 */
      __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_resultlist, __pyx_v_exprtokens); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3839, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF_SET(__pyx_v_resultlist, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":3838
 *             else:
 *                 loc, exprtokens = e._parse(instring, loc, doActions)
 *             if exprtokens or exprtokens.haskeys():             # <<<<<<<<<<<<<<
 *                 resultlist += exprtokens
 *         return loc, resultlist
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":3822
 *         loc, resultlist = self.exprs[0]._parse(instring, loc, doActions, callPreParse=False)
 *         errorStop = False
 *         for e in self.exprs[1:]:             # <<<<<<<<<<<<<<
 *             if isinstance(e, And._ErrorStop):
 *                 errorStop = True
 */
    __pyx_L5_continue:;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3840
 *             if exprtokens or exprtokens.haskeys():
 *                 resultlist += exprtokens
 *         return loc, resultlist             # <<<<<<<<<<<<<<
 * 
 *     def __iadd__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_loc);
  __Pyx_GIVEREF(__pyx_v_loc);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_loc);
  __Pyx_INCREF(__pyx_v_resultlist);
  __Pyx_GIVEREF(__pyx_v_resultlist);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_resultlist);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3817
 *         return self
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         # pass False as last arg to _parse for first element, since we already
 *         # pre-parsed the string as part of our And pre-parsing
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.And.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_resultlist);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XDECREF(__pyx_v_exprtokens);
  __Pyx_XDECREF(__pyx_v_pe);
  __Pyx_XDECREF(__pyx_v_loc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3842
 *         return loc, resultlist
 * 
 *     def __iadd__(self, other):             # <<<<<<<<<<<<<<
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_3And_7__iadd__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_3And_7__iadd__ = {"__iadd__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_3And_7__iadd__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_3And_7__iadd__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_other = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iadd__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_other,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_other)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__iadd__", 1, 2, 2, 1); __PYX_ERR(0, 3842, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__iadd__") < 0)) __PYX_ERR(0, 3842, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_other = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__iadd__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3842, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.And.__iadd__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_3And_6__iadd__(__pyx_self, __pyx_v_self, __pyx_v_other);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_3And_6__iadd__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__iadd__", 0);
  __Pyx_INCREF(__pyx_v_other);

  /* "Aspidites/_vendor/pyparsing_extension.py":3843
 * 
 *     def __iadd__(self, other):
 *         if isinstance(other, str):             # <<<<<<<<<<<<<<
 *             other = self._literalStringClass(other)
 *         return self.append(other)  # And([self, other])
 */
  __pyx_t_1 = PyString_Check(__pyx_v_other); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3844
 *     def __iadd__(self, other):
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)             # <<<<<<<<<<<<<<
 *         return self.append(other)  # And([self, other])
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_literalStringClass); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3844, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_other) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_other);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3844, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_other, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3843
 * 
 *     def __iadd__(self, other):
 *         if isinstance(other, str):             # <<<<<<<<<<<<<<
 *             other = self._literalStringClass(other)
 *         return self.append(other)  # And([self, other])
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3845
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 *         return self.append(other)  # And([self, other])             # <<<<<<<<<<<<<<
 * 
 *     def checkRecursion(self, parseElementList):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_append); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3845, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_other) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_other);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3845, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3842
 *         return loc, resultlist
 * 
 *     def __iadd__(self, other):             # <<<<<<<<<<<<<<
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.And.__iadd__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_other);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3847
 *         return self.append(other)  # And([self, other])
 * 
 *     def checkRecursion(self, parseElementList):             # <<<<<<<<<<<<<<
 *         subRecCheckList = parseElementList[:] + [self]
 *         for e in self.exprs:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_3And_9checkRecursion(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_3And_9checkRecursion = {"checkRecursion", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_3And_9checkRecursion, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_3And_9checkRecursion(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parseElementList = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("checkRecursion (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parseElementList,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_parseElementList)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("checkRecursion", 1, 2, 2, 1); __PYX_ERR(0, 3847, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "checkRecursion") < 0)) __PYX_ERR(0, 3847, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_parseElementList = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("checkRecursion", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3847, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.And.checkRecursion", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_3And_8checkRecursion(__pyx_self, __pyx_v_self, __pyx_v_parseElementList);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_3And_8checkRecursion(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parseElementList) {
  PyObject *__pyx_v_subRecCheckList = NULL;
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("checkRecursion", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":3848
 * 
 *     def checkRecursion(self, parseElementList):
 *         subRecCheckList = parseElementList[:] + [self]             # <<<<<<<<<<<<<<
 *         for e in self.exprs:
 *             e.checkRecursion(subRecCheckList)
 */
  __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_parseElementList, 0, 0, NULL, NULL, &__pyx_slice__11, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3848, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3848, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_v_self);
  __pyx_t_3 = PyNumber_Add(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3848, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_subRecCheckList = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3849
 *     def checkRecursion(self, parseElementList):
 *         subRecCheckList = parseElementList[:] + [self]
 *         for e in self.exprs:             # <<<<<<<<<<<<<<
 *             e.checkRecursion(subRecCheckList)
 *             if not e.mayReturnEmpty:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3849, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_2 = __pyx_t_3; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3849, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3849, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 3849, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3849, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 3849, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3849, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_5(__pyx_t_2);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 3849, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3850
 *         subRecCheckList = parseElementList[:] + [self]
 *         for e in self.exprs:
 *             e.checkRecursion(subRecCheckList)             # <<<<<<<<<<<<<<
 *             if not e.mayReturnEmpty:
 *                 break
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_checkRecursion); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3850, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_v_subRecCheckList) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_subRecCheckList);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3850, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3851
 *         for e in self.exprs:
 *             e.checkRecursion(subRecCheckList)
 *             if not e.mayReturnEmpty:             # <<<<<<<<<<<<<<
 *                 break
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_mayReturnEmpty); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3851, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 3851, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_8 = ((!__pyx_t_7) != 0);
    if (__pyx_t_8) {

      /* "Aspidites/_vendor/pyparsing_extension.py":3852
 *             e.checkRecursion(subRecCheckList)
 *             if not e.mayReturnEmpty:
 *                 break             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
      goto __pyx_L4_break;

      /* "Aspidites/_vendor/pyparsing_extension.py":3851
 *         for e in self.exprs:
 *             e.checkRecursion(subRecCheckList)
 *             if not e.mayReturnEmpty:             # <<<<<<<<<<<<<<
 *                 break
 * 
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":3849
 *     def checkRecursion(self, parseElementList):
 *         subRecCheckList = parseElementList[:] + [self]
 *         for e in self.exprs:             # <<<<<<<<<<<<<<
 *             e.checkRecursion(subRecCheckList)
 *             if not e.mayReturnEmpty:
 */
  }
  __pyx_L4_break:;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3847
 *         return self.append(other)  # And([self, other])
 * 
 *     def checkRecursion(self, parseElementList):             # <<<<<<<<<<<<<<
 *         subRecCheckList = parseElementList[:] + [self]
 *         for e in self.exprs:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.And.checkRecursion", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_subRecCheckList);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3854
 *                 break
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_3And_11__str__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_3And_11__str__ = {"__str__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_3And_11__str__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_3And_11__str__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_3And_10__str__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_3And_7__str___2generator24(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/pyparsing_extension.py":3859
 * 
 *         if self.strRepr is None:
 *             self.strRepr = "{" + " ".join(str(e) for e in self.exprs) + "}"             # <<<<<<<<<<<<<<
 * 
 *         return self.strRepr
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_3And_7__str___genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 3859, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__ *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_3And_7__str___2generator24, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_And___str___locals_genexpr, __pyx_n_s_Aspidites__vendor_pyparsing_exte); if (unlikely(!gen)) __PYX_ERR(0, 3859, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.And.__str__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_3And_7__str___2generator24(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 3859, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 3859, __pyx_L1_error) }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3859, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3859, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 3859, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3859, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 3859, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3859, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 3859, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_e);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_e, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_cur_scope->__pyx_v_e); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3859, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    __Pyx_XGIVEREF(__pyx_t_2);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_3;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_4;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_2);
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 3859, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3854
 *                 break
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_3And_10__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__ *__pyx_cur_scope;
  PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_3And_7__str___2generator24 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__ *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 3854, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);

  /* "Aspidites/_vendor/pyparsing_extension.py":3855
 * 
 *     def __str__(self):
 *         if hasattr(self, "name"):             # <<<<<<<<<<<<<<
 *             return self.name
 * 
 */
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_self;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_HasAttr(__pyx_t_1, __pyx_n_s_name); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 3855, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3856
 *     def __str__(self):
 *         if hasattr(self, "name"):
 *             return self.name             # <<<<<<<<<<<<<<
 * 
 *         if self.strRepr is None:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3856, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3855
 * 
 *     def __str__(self):
 *         if hasattr(self, "name"):             # <<<<<<<<<<<<<<
 *             return self.name
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3858
 *             return self.name
 * 
 *         if self.strRepr is None:             # <<<<<<<<<<<<<<
 *             self.strRepr = "{" + " ".join(str(e) for e in self.exprs) + "}"
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_strRepr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = (__pyx_t_1 == Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3859
 * 
 *         if self.strRepr is None:
 *             self.strRepr = "{" + " ".join(str(e) for e in self.exprs) + "}"             # <<<<<<<<<<<<<<
 * 
 *         return self.strRepr
 */
    __pyx_t_1 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_3And_7__str___genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3859, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyString_Join(__pyx_kp_s__29, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3859, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_Add(__pyx_kp_s__109, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3859, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyNumber_Add(__pyx_t_1, __pyx_kp_s__110); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3859, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_strRepr, __pyx_t_4) < 0) __PYX_ERR(0, 3859, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3858
 *             return self.name
 * 
 *         if self.strRepr is None:             # <<<<<<<<<<<<<<
 *             self.strRepr = "{" + " ".join(str(e) for e in self.exprs) + "}"
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3861
 *             self.strRepr = "{" + " ".join(str(e) for e in self.exprs) + "}"
 * 
 *         return self.strRepr             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_strRepr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3854
 *                 break
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.And.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_3And_7__str___2generator24);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3882
 *     """
 * 
 *     def __init__(self, exprs, savelist=False):             # <<<<<<<<<<<<<<
 *         super(Or, self).__init__(exprs, savelist)
 *         if self.exprs:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_2Or_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_2Or_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_2Or_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_2Or_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_exprs = 0;
  PyObject *__pyx_v_savelist = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_exprs,&__pyx_n_s_savelist,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_False));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_exprs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, 1); __PYX_ERR(0, 3882, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_savelist);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 3882, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_exprs = values[1];
    __pyx_v_savelist = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3882, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Or.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_2Or___init__(__pyx_self, __pyx_v_self, __pyx_v_exprs, __pyx_v_savelist);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_2Or_8__init___2generator25(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/pyparsing_extension.py":3885
 *         super(Or, self).__init__(exprs, savelist)
 *         if self.exprs:
 *             self.mayReturnEmpty = any(e.mayReturnEmpty for e in self.exprs)             # <<<<<<<<<<<<<<
 *         else:
 *             self.mayReturnEmpty = True
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_2Or_8__init___genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 3885, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__ *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_2Or_8__init___2generator25, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_Or___init___locals_genexpr, __pyx_n_s_Aspidites__vendor_pyparsing_exte); if (unlikely(!gen)) __PYX_ERR(0, 3885, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Or.__init__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_2Or_8__init___2generator25(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 3885, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 3885, __pyx_L1_error) }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3885, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3885, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 3885, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3885, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 3885, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3885, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 3885, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_e);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_e, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_e, __pyx_n_s_mayReturnEmpty); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3885, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 3885, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_5) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_True);
      __pyx_r = Py_True;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L0;
    }
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3882
 *     """
 * 
 *     def __init__(self, exprs, savelist=False):             # <<<<<<<<<<<<<<
 *         super(Or, self).__init__(exprs, savelist)
 *         if self.exprs:
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_2Or___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_exprs, PyObject *__pyx_v_savelist) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__ *__pyx_cur_scope;
  PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_2Or_8__init___2generator25 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__ *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 3882, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);

  /* "Aspidites/_vendor/pyparsing_extension.py":3883
 * 
 *     def __init__(self, exprs, savelist=False):
 *         super(Or, self).__init__(exprs, savelist)             # <<<<<<<<<<<<<<
 *         if self.exprs:
 *             self.mayReturnEmpty = any(e.mayReturnEmpty for e in self.exprs)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Or); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3883, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3883, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_cur_scope->__pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3883, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3883, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_exprs, __pyx_v_savelist};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3883, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_exprs, __pyx_v_savelist};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3883, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3883, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_exprs);
    __Pyx_GIVEREF(__pyx_v_exprs);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_exprs);
    __Pyx_INCREF(__pyx_v_savelist);
    __Pyx_GIVEREF(__pyx_v_savelist);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_savelist);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3883, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3884
 *     def __init__(self, exprs, savelist=False):
 *         super(Or, self).__init__(exprs, savelist)
 *         if self.exprs:             # <<<<<<<<<<<<<<
 *             self.mayReturnEmpty = any(e.mayReturnEmpty for e in self.exprs)
 *         else:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 3884, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_6) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3885
 *         super(Or, self).__init__(exprs, savelist)
 *         if self.exprs:
 *             self.mayReturnEmpty = any(e.mayReturnEmpty for e in self.exprs)             # <<<<<<<<<<<<<<
 *         else:
 *             self.mayReturnEmpty = True
 */
    __pyx_t_1 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_2Or_8__init___genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3885, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_Generator_Next(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3885, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_mayReturnEmpty, __pyx_t_3) < 0) __PYX_ERR(0, 3885, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3884
 *     def __init__(self, exprs, savelist=False):
 *         super(Or, self).__init__(exprs, savelist)
 *         if self.exprs:             # <<<<<<<<<<<<<<
 *             self.mayReturnEmpty = any(e.mayReturnEmpty for e in self.exprs)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3887
 *             self.mayReturnEmpty = any(e.mayReturnEmpty for e in self.exprs)
 *         else:
 *             self.mayReturnEmpty = True             # <<<<<<<<<<<<<<
 * 
 *     def streamline(self):
 */
  /*else*/ {
    if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_mayReturnEmpty, Py_True) < 0) __PYX_ERR(0, 3887, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "Aspidites/_vendor/pyparsing_extension.py":3882
 *     """
 * 
 *     def __init__(self, exprs, savelist=False):             # <<<<<<<<<<<<<<
 *         super(Or, self).__init__(exprs, savelist)
 *         if self.exprs:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Or.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_2Or_8__init___2generator25);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3889
 *             self.mayReturnEmpty = True
 * 
 *     def streamline(self):             # <<<<<<<<<<<<<<
 *         super(Or, self).streamline()
 *         if __diag__.collect_all_And_tokens:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_2Or_3streamline(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_2Or_3streamline = {"streamline", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_2Or_3streamline, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_2Or_3streamline(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("streamline (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_2Or_2streamline(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_2Or_10streamline_2generator26(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/pyparsing_extension.py":3892
 *         super(Or, self).streamline()
 *         if __diag__.collect_all_And_tokens:
 *             self.saveAsList = any(e.saveAsList for e in self.exprs)             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_2Or_10streamline_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 3892, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_2Or_10streamline_2generator26, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_Or_streamline_locals_genexpr, __pyx_n_s_Aspidites__vendor_pyparsing_exte); if (unlikely(!gen)) __PYX_ERR(0, 3892, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Or.streamline.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_2Or_10streamline_2generator26(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 3892, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 3892, __pyx_L1_error) }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3892, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3892, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3892, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 3892, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3892, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 3892, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3892, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 3892, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_e);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_e, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_e, __pyx_n_s_saveAsList); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3892, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 3892, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_5) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_True);
      __pyx_r = Py_True;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L0;
    }
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3889
 *             self.mayReturnEmpty = True
 * 
 *     def streamline(self):             # <<<<<<<<<<<<<<
 *         super(Or, self).streamline()
 *         if __diag__.collect_all_And_tokens:
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_2Or_2streamline(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline *__pyx_cur_scope;
  PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_2Or_10streamline_2generator26 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("streamline", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 3889, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);

  /* "Aspidites/_vendor/pyparsing_extension.py":3890
 * 
 *     def streamline(self):
 *         super(Or, self).streamline()             # <<<<<<<<<<<<<<
 *         if __diag__.collect_all_And_tokens:
 *             self.saveAsList = any(e.saveAsList for e in self.exprs)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Or); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3890, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3890, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_cur_scope->__pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3890, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_streamline); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3890, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3890, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3891
 *     def streamline(self):
 *         super(Or, self).streamline()
 *         if __diag__.collect_all_And_tokens:             # <<<<<<<<<<<<<<
 *             self.saveAsList = any(e.saveAsList for e in self.exprs)
 *         return self
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_diag); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_collect_all_And_tokens); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 3891, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3892
 *         super(Or, self).streamline()
 *         if __diag__.collect_all_And_tokens:
 *             self.saveAsList = any(e.saveAsList for e in self.exprs)             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
    __pyx_t_3 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_2Or_10streamline_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3892, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_Generator_Next(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3892, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_saveAsList, __pyx_t_1) < 0) __PYX_ERR(0, 3892, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3891
 *     def streamline(self):
 *         super(Or, self).streamline()
 *         if __diag__.collect_all_And_tokens:             # <<<<<<<<<<<<<<
 *             self.saveAsList = any(e.saveAsList for e in self.exprs)
 *         return self
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3893
 *         if __diag__.collect_all_And_tokens:
 *             self.saveAsList = any(e.saveAsList for e in self.exprs)
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __pyx_r = __pyx_cur_scope->__pyx_v_self;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3889
 *             self.mayReturnEmpty = True
 * 
 *     def streamline(self):             # <<<<<<<<<<<<<<
 *         super(Or, self).streamline()
 *         if __diag__.collect_all_And_tokens:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Or.streamline", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_2Or_10streamline_2generator26);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3895
 *         return self
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         maxExcLoc = -1
 *         maxException = None
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_2Or_5parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_2Or_5parseImpl = {"parseImpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_2Or_5parseImpl, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_2Or_5parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  PyObject *__pyx_v_doActions = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parseImpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_doActions,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 1); __PYX_ERR(0, 3895, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 2); __PYX_ERR(0, 3895, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_doActions);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "parseImpl") < 0)) __PYX_ERR(0, 3895, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_doActions = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3895, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Or.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_2Or_4parseImpl(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_2Or_4parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, PyObject *__pyx_v_doActions) {
  PyObject *__pyx_v_maxExcLoc = NULL;
  PyObject *__pyx_v_maxException = NULL;
  PyObject *__pyx_v_matches = NULL;
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_v_loc2 = NULL;
  PyObject *__pyx_v_err = NULL;
  PyObject *__pyx_v_best_expr = NULL;
  PyObject *__pyx_v_longest = NULL;
  PyObject *__pyx_v_loc1 = NULL;
  PyObject *__pyx_v_expr1 = NULL;
  PyObject *__pyx_v_toks = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  int __pyx_t_20;
  PyObject *(*__pyx_t_21)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parseImpl", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":3896
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         maxExcLoc = -1             # <<<<<<<<<<<<<<
 *         maxException = None
 *         matches = []
 */
  __Pyx_INCREF(__pyx_int_neg_1);
  __pyx_v_maxExcLoc = __pyx_int_neg_1;

  /* "Aspidites/_vendor/pyparsing_extension.py":3897
 *     def parseImpl(self, instring, loc, doActions=True):
 *         maxExcLoc = -1
 *         maxException = None             # <<<<<<<<<<<<<<
 *         matches = []
 *         for e in self.exprs:
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_maxException = Py_None;

  /* "Aspidites/_vendor/pyparsing_extension.py":3898
 *         maxExcLoc = -1
 *         maxException = None
 *         matches = []             # <<<<<<<<<<<<<<
 *         for e in self.exprs:
 *             try:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3898, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_matches = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3899
 *         maxException = None
 *         matches = []
 *         for e in self.exprs:             # <<<<<<<<<<<<<<
 *             try:
 *                 loc2 = e.tryParse(instring, loc)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3899, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3899, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3899, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 3899, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3899, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 3899, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3899, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 3899, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3900
 *         matches = []
 *         for e in self.exprs:
 *             try:             # <<<<<<<<<<<<<<
 *                 loc2 = e.tryParse(instring, loc)
 *             except ParseException as err:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      /*try:*/ {

        /* "Aspidites/_vendor/pyparsing_extension.py":3901
 *         for e in self.exprs:
 *             try:
 *                 loc2 = e.tryParse(instring, loc)             # <<<<<<<<<<<<<<
 *             except ParseException as err:
 *                 err.__traceback__ = None
 */
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_tryParse); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3901, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_9 = NULL;
        __pyx_t_10 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
          __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
          if (likely(__pyx_t_9)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
            __Pyx_INCREF(__pyx_t_9);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_8, function);
            __pyx_t_10 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_8)) {
          PyObject *__pyx_temp[3] = {__pyx_t_9, __pyx_v_instring, __pyx_v_loc};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3901, __pyx_L5_error)
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
          PyObject *__pyx_temp[3] = {__pyx_t_9, __pyx_v_instring, __pyx_v_loc};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3901, __pyx_L5_error)
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_11 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 3901, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_11);
          if (__pyx_t_9) {
            __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_9); __pyx_t_9 = NULL;
          }
          __Pyx_INCREF(__pyx_v_instring);
          __Pyx_GIVEREF(__pyx_v_instring);
          PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_10, __pyx_v_instring);
          __Pyx_INCREF(__pyx_v_loc);
          __Pyx_GIVEREF(__pyx_v_loc);
          PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_10, __pyx_v_loc);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_11, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3901, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        }
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF_SET(__pyx_v_loc2, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":3900
 *         matches = []
 *         for e in self.exprs:
 *             try:             # <<<<<<<<<<<<<<
 *                 loc2 = e.tryParse(instring, loc)
 *             except ParseException as err:
 */
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":3913
 *             else:
 *                 # save match among all matches, to retry longest to shortest
 *                 matches.append((loc2, e))             # <<<<<<<<<<<<<<
 * 
 *         if matches:
 */
      /*else:*/ {
        __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3913, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_v_loc2);
        __Pyx_GIVEREF(__pyx_v_loc2);
        PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_loc2);
        __Pyx_INCREF(__pyx_v_e);
        __Pyx_GIVEREF(__pyx_v_e);
        PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_e);
        __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_matches, __pyx_t_1); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 3913, __pyx_L7_except_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L12_try_end;
      __pyx_L5_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":3902
 *             try:
 *                 loc2 = e.tryParse(instring, loc)
 *             except ParseException as err:             # <<<<<<<<<<<<<<
 *                 err.__traceback__ = None
 *                 if err.loc > maxExcLoc:
 */
      __Pyx_ErrFetch(&__pyx_t_1, &__pyx_t_8, &__pyx_t_11);
      __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3902, __pyx_L7_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_1, __pyx_t_9);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_ErrRestore(__pyx_t_1, __pyx_t_8, __pyx_t_11);
      __pyx_t_1 = 0; __pyx_t_8 = 0; __pyx_t_11 = 0;
      if (__pyx_t_10) {
        __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Or.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_11, &__pyx_t_8, &__pyx_t_1) < 0) __PYX_ERR(0, 3902, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_XDECREF_SET(__pyx_v_err, __pyx_t_8);

        /* "Aspidites/_vendor/pyparsing_extension.py":3903
 *                 loc2 = e.tryParse(instring, loc)
 *             except ParseException as err:
 *                 err.__traceback__ = None             # <<<<<<<<<<<<<<
 *                 if err.loc > maxExcLoc:
 *                     maxException = err
 */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_err, __pyx_n_s_traceback, Py_None) < 0) __PYX_ERR(0, 3903, __pyx_L7_except_error)

        /* "Aspidites/_vendor/pyparsing_extension.py":3904
 *             except ParseException as err:
 *                 err.__traceback__ = None
 *                 if err.loc > maxExcLoc:             # <<<<<<<<<<<<<<
 *                     maxException = err
 *                     maxExcLoc = err.loc
 */
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_err, __pyx_n_s_loc); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3904, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_13 = PyObject_RichCompare(__pyx_t_9, __pyx_v_maxExcLoc, Py_GT); __Pyx_XGOTREF(__pyx_t_13); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3904, __pyx_L7_except_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_14 = __Pyx_PyObject_IsTrue(__pyx_t_13); if (unlikely(__pyx_t_14 < 0)) __PYX_ERR(0, 3904, __pyx_L7_except_error)
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        if (__pyx_t_14) {

          /* "Aspidites/_vendor/pyparsing_extension.py":3905
 *                 err.__traceback__ = None
 *                 if err.loc > maxExcLoc:
 *                     maxException = err             # <<<<<<<<<<<<<<
 *                     maxExcLoc = err.loc
 *             except IndexError:
 */
          __Pyx_INCREF(__pyx_v_err);
          __Pyx_DECREF_SET(__pyx_v_maxException, __pyx_v_err);

          /* "Aspidites/_vendor/pyparsing_extension.py":3906
 *                 if err.loc > maxExcLoc:
 *                     maxException = err
 *                     maxExcLoc = err.loc             # <<<<<<<<<<<<<<
 *             except IndexError:
 *                 if len(instring) > maxExcLoc:
 */
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_err, __pyx_n_s_loc); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3906, __pyx_L7_except_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF_SET(__pyx_v_maxExcLoc, __pyx_t_13);
          __pyx_t_13 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":3904
 *             except ParseException as err:
 *                 err.__traceback__ = None
 *                 if err.loc > maxExcLoc:             # <<<<<<<<<<<<<<
 *                     maxException = err
 *                     maxExcLoc = err.loc
 */
        }
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L6_exception_handled;
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":3907
 *                     maxException = err
 *                     maxExcLoc = err.loc
 *             except IndexError:             # <<<<<<<<<<<<<<
 *                 if len(instring) > maxExcLoc:
 *                     maxException = ParseException(instring, len(instring), e.errmsg, self)
 */
      __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_IndexError);
      if (__pyx_t_10) {
        __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Or.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_8, &__pyx_t_11) < 0) __PYX_ERR(0, 3907, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_GOTREF(__pyx_t_11);

        /* "Aspidites/_vendor/pyparsing_extension.py":3908
 *                     maxExcLoc = err.loc
 *             except IndexError:
 *                 if len(instring) > maxExcLoc:             # <<<<<<<<<<<<<<
 *                     maxException = ParseException(instring, len(instring), e.errmsg, self)
 *                     maxExcLoc = len(instring)
 */
        __pyx_t_15 = PyObject_Length(__pyx_v_instring); if (unlikely(__pyx_t_15 == ((Py_ssize_t)-1))) __PYX_ERR(0, 3908, __pyx_L7_except_error)
        __pyx_t_13 = PyInt_FromSsize_t(__pyx_t_15); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3908, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_13);
        __pyx_t_9 = PyObject_RichCompare(__pyx_t_13, __pyx_v_maxExcLoc, Py_GT); __Pyx_XGOTREF(__pyx_t_9); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3908, __pyx_L7_except_error)
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __pyx_t_14 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_14 < 0)) __PYX_ERR(0, 3908, __pyx_L7_except_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (__pyx_t_14) {

          /* "Aspidites/_vendor/pyparsing_extension.py":3909
 *             except IndexError:
 *                 if len(instring) > maxExcLoc:
 *                     maxException = ParseException(instring, len(instring), e.errmsg, self)             # <<<<<<<<<<<<<<
 *                     maxExcLoc = len(instring)
 *             else:
 */
          __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3909, __pyx_L7_except_error)
          __Pyx_GOTREF(__pyx_t_13);
          __pyx_t_15 = PyObject_Length(__pyx_v_instring); if (unlikely(__pyx_t_15 == ((Py_ssize_t)-1))) __PYX_ERR(0, 3909, __pyx_L7_except_error)
          __pyx_t_16 = PyInt_FromSsize_t(__pyx_t_15); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 3909, __pyx_L7_except_error)
          __Pyx_GOTREF(__pyx_t_16);
          __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_errmsg); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 3909, __pyx_L7_except_error)
          __Pyx_GOTREF(__pyx_t_17);
          __pyx_t_18 = NULL;
          __pyx_t_10 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_18 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_18)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_18);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_10 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[5] = {__pyx_t_18, __pyx_v_instring, __pyx_t_16, __pyx_t_17, __pyx_v_self};
            __pyx_t_9 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_10, 4+__pyx_t_10); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3909, __pyx_L7_except_error)
            __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
            __Pyx_GOTREF(__pyx_t_9);
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
            PyObject *__pyx_temp[5] = {__pyx_t_18, __pyx_v_instring, __pyx_t_16, __pyx_t_17, __pyx_v_self};
            __pyx_t_9 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_10, 4+__pyx_t_10); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3909, __pyx_L7_except_error)
            __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
            __Pyx_GOTREF(__pyx_t_9);
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
          } else
          #endif
          {
            __pyx_t_19 = PyTuple_New(4+__pyx_t_10); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 3909, __pyx_L7_except_error)
            __Pyx_GOTREF(__pyx_t_19);
            if (__pyx_t_18) {
              __Pyx_GIVEREF(__pyx_t_18); PyTuple_SET_ITEM(__pyx_t_19, 0, __pyx_t_18); __pyx_t_18 = NULL;
            }
            __Pyx_INCREF(__pyx_v_instring);
            __Pyx_GIVEREF(__pyx_v_instring);
            PyTuple_SET_ITEM(__pyx_t_19, 0+__pyx_t_10, __pyx_v_instring);
            __Pyx_GIVEREF(__pyx_t_16);
            PyTuple_SET_ITEM(__pyx_t_19, 1+__pyx_t_10, __pyx_t_16);
            __Pyx_GIVEREF(__pyx_t_17);
            PyTuple_SET_ITEM(__pyx_t_19, 2+__pyx_t_10, __pyx_t_17);
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_19, 3+__pyx_t_10, __pyx_v_self);
            __pyx_t_16 = 0;
            __pyx_t_17 = 0;
            __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_19, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3909, __pyx_L7_except_error)
            __Pyx_GOTREF(__pyx_t_9);
            __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF_SET(__pyx_v_maxException, __pyx_t_9);
          __pyx_t_9 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":3910
 *                 if len(instring) > maxExcLoc:
 *                     maxException = ParseException(instring, len(instring), e.errmsg, self)
 *                     maxExcLoc = len(instring)             # <<<<<<<<<<<<<<
 *             else:
 *                 # save match among all matches, to retry longest to shortest
 */
          __pyx_t_15 = PyObject_Length(__pyx_v_instring); if (unlikely(__pyx_t_15 == ((Py_ssize_t)-1))) __PYX_ERR(0, 3910, __pyx_L7_except_error)
          __pyx_t_9 = PyInt_FromSsize_t(__pyx_t_15); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3910, __pyx_L7_except_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF_SET(__pyx_v_maxExcLoc, __pyx_t_9);
          __pyx_t_9 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":3908
 *                     maxExcLoc = err.loc
 *             except IndexError:
 *                 if len(instring) > maxExcLoc:             # <<<<<<<<<<<<<<
 *                     maxException = ParseException(instring, len(instring), e.errmsg, self)
 *                     maxExcLoc = len(instring)
 */
        }
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        goto __pyx_L6_exception_handled;
      }
      goto __pyx_L7_except_error;
      __pyx_L7_except_error:;

      /* "Aspidites/_vendor/pyparsing_extension.py":3900
 *         matches = []
 *         for e in self.exprs:
 *             try:             # <<<<<<<<<<<<<<
 *                 loc2 = e.tryParse(instring, loc)
 *             except ParseException as err:
 */
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
      goto __pyx_L1_error;
      __pyx_L6_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
      __pyx_L12_try_end:;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":3899
 *         maxException = None
 *         matches = []
 *         for e in self.exprs:             # <<<<<<<<<<<<<<
 *             try:
 *                 loc2 = e.tryParse(instring, loc)
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3915
 *                 matches.append((loc2, e))
 * 
 *         if matches:             # <<<<<<<<<<<<<<
 *             # re-evaluate all matches in descending order of length of match, in case attached actions
 *             # might change whether or how much they match of the input.
 */
  __pyx_t_14 = (PyList_GET_SIZE(__pyx_v_matches) != 0);
  if (__pyx_t_14) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3918
 *             # re-evaluate all matches in descending order of length of match, in case attached actions
 *             # might change whether or how much they match of the input.
 *             matches.sort(key=itemgetter(0), reverse=True)             # <<<<<<<<<<<<<<
 * 
 *             if not doActions:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_matches, __pyx_n_s_sort); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3918, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_11 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 3918, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_itemgetter); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3918, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_8 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_9, __pyx_int_0) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_int_0);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3918, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (PyDict_SetItem(__pyx_t_11, __pyx_n_s_key, __pyx_t_8) < 0) __PYX_ERR(0, 3918, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (PyDict_SetItem(__pyx_t_11, __pyx_n_s_reverse, Py_True) < 0) __PYX_ERR(0, 3918, __pyx_L1_error)
    __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_11); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3918, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3920
 *             matches.sort(key=itemgetter(0), reverse=True)
 * 
 *             if not doActions:             # <<<<<<<<<<<<<<
 *                 # no further conditions or parse actions to change the selection of
 *                 # alternative, so the first match will be the best match
 */
    __pyx_t_14 = __Pyx_PyObject_IsTrue(__pyx_v_doActions); if (unlikely(__pyx_t_14 < 0)) __PYX_ERR(0, 3920, __pyx_L1_error)
    __pyx_t_20 = ((!__pyx_t_14) != 0);
    if (__pyx_t_20) {

      /* "Aspidites/_vendor/pyparsing_extension.py":3923
 *                 # no further conditions or parse actions to change the selection of
 *                 # alternative, so the first match will be the best match
 *                 best_expr = matches[0][1]             # <<<<<<<<<<<<<<
 *                 return best_expr._parse(instring, loc, doActions)
 * 
 */
      __pyx_t_8 = __Pyx_GetItemInt_List(__pyx_v_matches, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3923, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_11 = __Pyx_GetItemInt(__pyx_t_8, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 3923, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_v_best_expr = __pyx_t_11;
      __pyx_t_11 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":3924
 *                 # alternative, so the first match will be the best match
 *                 best_expr = matches[0][1]
 *                 return best_expr._parse(instring, loc, doActions)             # <<<<<<<<<<<<<<
 * 
 *             longest = -1, None
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_best_expr, __pyx_n_s_parse); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3924, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_2 = NULL;
      __pyx_t_10 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
          __pyx_t_10 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_8)) {
        PyObject *__pyx_temp[4] = {__pyx_t_2, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions};
        __pyx_t_11 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_10, 3+__pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 3924, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_GOTREF(__pyx_t_11);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
        PyObject *__pyx_temp[4] = {__pyx_t_2, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions};
        __pyx_t_11 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_10, 3+__pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 3924, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_GOTREF(__pyx_t_11);
      } else
      #endif
      {
        __pyx_t_1 = PyTuple_New(3+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3924, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (__pyx_t_2) {
          __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2); __pyx_t_2 = NULL;
        }
        __Pyx_INCREF(__pyx_v_instring);
        __Pyx_GIVEREF(__pyx_v_instring);
        PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_10, __pyx_v_instring);
        __Pyx_INCREF(__pyx_v_loc);
        __Pyx_GIVEREF(__pyx_v_loc);
        PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_10, __pyx_v_loc);
        __Pyx_INCREF(__pyx_v_doActions);
        __Pyx_GIVEREF(__pyx_v_doActions);
        PyTuple_SET_ITEM(__pyx_t_1, 2+__pyx_t_10, __pyx_v_doActions);
        __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_1, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 3924, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_r = __pyx_t_11;
      __pyx_t_11 = 0;
      goto __pyx_L0;

      /* "Aspidites/_vendor/pyparsing_extension.py":3920
 *             matches.sort(key=itemgetter(0), reverse=True)
 * 
 *             if not doActions:             # <<<<<<<<<<<<<<
 *                 # no further conditions or parse actions to change the selection of
 *                 # alternative, so the first match will be the best match
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":3926
 *                 return best_expr._parse(instring, loc, doActions)
 * 
 *             longest = -1, None             # <<<<<<<<<<<<<<
 *             for loc1, expr1 in matches:
 *                 if loc1 <= longest[0]:
 */
    __Pyx_INCREF(__pyx_tuple__111);
    __pyx_v_longest = __pyx_tuple__111;

    /* "Aspidites/_vendor/pyparsing_extension.py":3927
 * 
 *             longest = -1, None
 *             for loc1, expr1 in matches:             # <<<<<<<<<<<<<<
 *                 if loc1 <= longest[0]:
 *                     # already have a longer match than this one will deliver, we are done
 */
    __pyx_t_11 = __pyx_v_matches; __Pyx_INCREF(__pyx_t_11); __pyx_t_3 = 0;
    for (;;) {
      if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_11)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_8 = PyList_GET_ITEM(__pyx_t_11, __pyx_t_3); __Pyx_INCREF(__pyx_t_8); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 3927, __pyx_L1_error)
      #else
      __pyx_t_8 = PySequence_ITEM(__pyx_t_11, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3927, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      #endif
      if ((likely(PyTuple_CheckExact(__pyx_t_8))) || (PyList_CheckExact(__pyx_t_8))) {
        PyObject* sequence = __pyx_t_8;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 3927, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_2);
        #else
        __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3927, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3927, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_9 = PyObject_GetIter(__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3927, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_21 = Py_TYPE(__pyx_t_9)->tp_iternext;
        index = 0; __pyx_t_1 = __pyx_t_21(__pyx_t_9); if (unlikely(!__pyx_t_1)) goto __pyx_L23_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_1);
        index = 1; __pyx_t_2 = __pyx_t_21(__pyx_t_9); if (unlikely(!__pyx_t_2)) goto __pyx_L23_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_2);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_21(__pyx_t_9), 2) < 0) __PYX_ERR(0, 3927, __pyx_L1_error)
        __pyx_t_21 = NULL;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L24_unpacking_done;
        __pyx_L23_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_21 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 3927, __pyx_L1_error)
        __pyx_L24_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_loc1, __pyx_t_1);
      __pyx_t_1 = 0;
      __Pyx_XDECREF_SET(__pyx_v_expr1, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":3928
 *             longest = -1, None
 *             for loc1, expr1 in matches:
 *                 if loc1 <= longest[0]:             # <<<<<<<<<<<<<<
 *                     # already have a longer match than this one will deliver, we are done
 *                     return longest
 */
      __pyx_t_8 = __Pyx_GetItemInt_Tuple(__pyx_v_longest, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3928, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_2 = PyObject_RichCompare(__pyx_v_loc1, __pyx_t_8, Py_LE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3928, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_20 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_20 < 0)) __PYX_ERR(0, 3928, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (__pyx_t_20) {

        /* "Aspidites/_vendor/pyparsing_extension.py":3930
 *                 if loc1 <= longest[0]:
 *                     # already have a longer match than this one will deliver, we are done
 *                     return longest             # <<<<<<<<<<<<<<
 * 
 *                 try:
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_v_longest);
        __pyx_r = __pyx_v_longest;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        goto __pyx_L0;

        /* "Aspidites/_vendor/pyparsing_extension.py":3928
 *             longest = -1, None
 *             for loc1, expr1 in matches:
 *                 if loc1 <= longest[0]:             # <<<<<<<<<<<<<<
 *                     # already have a longer match than this one will deliver, we are done
 *                     return longest
 */
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":3932
 *                     return longest
 * 
 *                 try:             # <<<<<<<<<<<<<<
 *                     loc2, toks = expr1._parse(instring, loc, doActions)
 *                 except ParseException as err:
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_6, &__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_5);
        /*try:*/ {

          /* "Aspidites/_vendor/pyparsing_extension.py":3933
 * 
 *                 try:
 *                     loc2, toks = expr1._parse(instring, loc, doActions)             # <<<<<<<<<<<<<<
 *                 except ParseException as err:
 *                     err.__traceback__ = None
 */
          __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_expr1, __pyx_n_s_parse); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3933, __pyx_L26_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_1 = NULL;
          __pyx_t_10 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_8);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_8, function);
              __pyx_t_10 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_8)) {
            PyObject *__pyx_temp[4] = {__pyx_t_1, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions};
            __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_10, 3+__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3933, __pyx_L26_error)
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
            PyObject *__pyx_temp[4] = {__pyx_t_1, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions};
            __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_10, 3+__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3933, __pyx_L26_error)
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_2);
          } else
          #endif
          {
            __pyx_t_9 = PyTuple_New(3+__pyx_t_10); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3933, __pyx_L26_error)
            __Pyx_GOTREF(__pyx_t_9);
            if (__pyx_t_1) {
              __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_1); __pyx_t_1 = NULL;
            }
            __Pyx_INCREF(__pyx_v_instring);
            __Pyx_GIVEREF(__pyx_v_instring);
            PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_10, __pyx_v_instring);
            __Pyx_INCREF(__pyx_v_loc);
            __Pyx_GIVEREF(__pyx_v_loc);
            PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_10, __pyx_v_loc);
            __Pyx_INCREF(__pyx_v_doActions);
            __Pyx_GIVEREF(__pyx_v_doActions);
            PyTuple_SET_ITEM(__pyx_t_9, 2+__pyx_t_10, __pyx_v_doActions);
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_9, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3933, __pyx_L26_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          }
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
            PyObject* sequence = __pyx_t_2;
            Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
            if (unlikely(size != 2)) {
              if (size > 2) __Pyx_RaiseTooManyValuesError(2);
              else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
              __PYX_ERR(0, 3933, __pyx_L26_error)
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            if (likely(PyTuple_CheckExact(sequence))) {
              __pyx_t_8 = PyTuple_GET_ITEM(sequence, 0); 
              __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
            } else {
              __pyx_t_8 = PyList_GET_ITEM(sequence, 0); 
              __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
            }
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(__pyx_t_9);
            #else
            __pyx_t_8 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3933, __pyx_L26_error)
            __Pyx_GOTREF(__pyx_t_8);
            __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3933, __pyx_L26_error)
            __Pyx_GOTREF(__pyx_t_9);
            #endif
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else {
            Py_ssize_t index = -1;
            __pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3933, __pyx_L26_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_21 = Py_TYPE(__pyx_t_1)->tp_iternext;
            index = 0; __pyx_t_8 = __pyx_t_21(__pyx_t_1); if (unlikely(!__pyx_t_8)) goto __pyx_L34_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_8);
            index = 1; __pyx_t_9 = __pyx_t_21(__pyx_t_1); if (unlikely(!__pyx_t_9)) goto __pyx_L34_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_9);
            if (__Pyx_IternextUnpackEndCheck(__pyx_t_21(__pyx_t_1), 2) < 0) __PYX_ERR(0, 3933, __pyx_L26_error)
            __pyx_t_21 = NULL;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            goto __pyx_L35_unpacking_done;
            __pyx_L34_unpacking_failed:;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_21 = NULL;
            if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
            __PYX_ERR(0, 3933, __pyx_L26_error)
            __pyx_L35_unpacking_done:;
          }
          __Pyx_XDECREF_SET(__pyx_v_loc2, __pyx_t_8);
          __pyx_t_8 = 0;
          __Pyx_XDECREF_SET(__pyx_v_toks, __pyx_t_9);
          __pyx_t_9 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":3932
 *                     return longest
 * 
 *                 try:             # <<<<<<<<<<<<<<
 *                     loc2, toks = expr1._parse(instring, loc, doActions)
 *                 except ParseException as err:
 */
        }

        /* "Aspidites/_vendor/pyparsing_extension.py":3940
 *                         maxExcLoc = err.loc
 *                 else:
 *                     if loc2 >= loc1:             # <<<<<<<<<<<<<<
 *                         return loc2, toks
 *                     # didn't match as much as before
 */
        /*else:*/ {
          __pyx_t_2 = PyObject_RichCompare(__pyx_v_loc2, __pyx_v_loc1, Py_GE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3940, __pyx_L28_except_error)
          __pyx_t_20 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_20 < 0)) __PYX_ERR(0, 3940, __pyx_L28_except_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          if (__pyx_t_20) {

            /* "Aspidites/_vendor/pyparsing_extension.py":3941
 *                 else:
 *                     if loc2 >= loc1:
 *                         return loc2, toks             # <<<<<<<<<<<<<<
 *                     # didn't match as much as before
 *                     elif loc2 > longest[0]:
 */
            __Pyx_XDECREF(__pyx_r);
            __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3941, __pyx_L28_except_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_INCREF(__pyx_v_loc2);
            __Pyx_GIVEREF(__pyx_v_loc2);
            PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_loc2);
            __Pyx_INCREF(__pyx_v_toks);
            __Pyx_GIVEREF(__pyx_v_toks);
            PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_toks);
            __pyx_r = __pyx_t_2;
            __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            goto __pyx_L29_except_return;

            /* "Aspidites/_vendor/pyparsing_extension.py":3940
 *                         maxExcLoc = err.loc
 *                 else:
 *                     if loc2 >= loc1:             # <<<<<<<<<<<<<<
 *                         return loc2, toks
 *                     # didn't match as much as before
 */
          }

          /* "Aspidites/_vendor/pyparsing_extension.py":3943
 *                         return loc2, toks
 *                     # didn't match as much as before
 *                     elif loc2 > longest[0]:             # <<<<<<<<<<<<<<
 *                         longest = loc2, toks
 * 
 */
          __pyx_t_2 = __Pyx_GetItemInt_Tuple(__pyx_v_longest, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3943, __pyx_L28_except_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_9 = PyObject_RichCompare(__pyx_v_loc2, __pyx_t_2, Py_GT); __Pyx_XGOTREF(__pyx_t_9); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3943, __pyx_L28_except_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_20 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_20 < 0)) __PYX_ERR(0, 3943, __pyx_L28_except_error)
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (__pyx_t_20) {

            /* "Aspidites/_vendor/pyparsing_extension.py":3944
 *                     # didn't match as much as before
 *                     elif loc2 > longest[0]:
 *                         longest = loc2, toks             # <<<<<<<<<<<<<<
 * 
 *             if longest != (-1, None):
 */
            __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3944, __pyx_L28_except_error)
            __Pyx_GOTREF(__pyx_t_9);
            __Pyx_INCREF(__pyx_v_loc2);
            __Pyx_GIVEREF(__pyx_v_loc2);
            PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v_loc2);
            __Pyx_INCREF(__pyx_v_toks);
            __Pyx_GIVEREF(__pyx_v_toks);
            PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_v_toks);
            __Pyx_DECREF_SET(__pyx_v_longest, ((PyObject*)__pyx_t_9));
            __pyx_t_9 = 0;

            /* "Aspidites/_vendor/pyparsing_extension.py":3943
 *                         return loc2, toks
 *                     # didn't match as much as before
 *                     elif loc2 > longest[0]:             # <<<<<<<<<<<<<<
 *                         longest = loc2, toks
 * 
 */
          }
        }
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        goto __pyx_L33_try_end;
        __pyx_L26_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
        __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
        __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":3934
 *                 try:
 *                     loc2, toks = expr1._parse(instring, loc, doActions)
 *                 except ParseException as err:             # <<<<<<<<<<<<<<
 *                     err.__traceback__ = None
 *                     if err.loc > maxExcLoc:
 */
        __Pyx_ErrFetch(&__pyx_t_9, &__pyx_t_2, &__pyx_t_8);
        __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3934, __pyx_L28_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_10 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_9, __pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_ErrRestore(__pyx_t_9, __pyx_t_2, __pyx_t_8);
        __pyx_t_9 = 0; __pyx_t_2 = 0; __pyx_t_8 = 0;
        if (__pyx_t_10) {
          __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Or.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_2, &__pyx_t_9) < 0) __PYX_ERR(0, 3934, __pyx_L28_except_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_XDECREF_SET(__pyx_v_err, __pyx_t_2);

          /* "Aspidites/_vendor/pyparsing_extension.py":3935
 *                     loc2, toks = expr1._parse(instring, loc, doActions)
 *                 except ParseException as err:
 *                     err.__traceback__ = None             # <<<<<<<<<<<<<<
 *                     if err.loc > maxExcLoc:
 *                         maxException = err
 */
          if (__Pyx_PyObject_SetAttrStr(__pyx_v_err, __pyx_n_s_traceback, Py_None) < 0) __PYX_ERR(0, 3935, __pyx_L28_except_error)

          /* "Aspidites/_vendor/pyparsing_extension.py":3936
 *                 except ParseException as err:
 *                     err.__traceback__ = None
 *                     if err.loc > maxExcLoc:             # <<<<<<<<<<<<<<
 *                         maxException = err
 *                         maxExcLoc = err.loc
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_err, __pyx_n_s_loc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3936, __pyx_L28_except_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_13 = PyObject_RichCompare(__pyx_t_1, __pyx_v_maxExcLoc, Py_GT); __Pyx_XGOTREF(__pyx_t_13); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3936, __pyx_L28_except_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_20 = __Pyx_PyObject_IsTrue(__pyx_t_13); if (unlikely(__pyx_t_20 < 0)) __PYX_ERR(0, 3936, __pyx_L28_except_error)
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          if (__pyx_t_20) {

            /* "Aspidites/_vendor/pyparsing_extension.py":3937
 *                     err.__traceback__ = None
 *                     if err.loc > maxExcLoc:
 *                         maxException = err             # <<<<<<<<<<<<<<
 *                         maxExcLoc = err.loc
 *                 else:
 */
            __Pyx_INCREF(__pyx_v_err);
            __Pyx_DECREF_SET(__pyx_v_maxException, __pyx_v_err);

            /* "Aspidites/_vendor/pyparsing_extension.py":3938
 *                     if err.loc > maxExcLoc:
 *                         maxException = err
 *                         maxExcLoc = err.loc             # <<<<<<<<<<<<<<
 *                 else:
 *                     if loc2 >= loc1:
 */
            __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_err, __pyx_n_s_loc); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3938, __pyx_L28_except_error)
            __Pyx_GOTREF(__pyx_t_13);
            __Pyx_DECREF_SET(__pyx_v_maxExcLoc, __pyx_t_13);
            __pyx_t_13 = 0;

            /* "Aspidites/_vendor/pyparsing_extension.py":3936
 *                 except ParseException as err:
 *                     err.__traceback__ = None
 *                     if err.loc > maxExcLoc:             # <<<<<<<<<<<<<<
 *                         maxException = err
 *                         maxExcLoc = err.loc
 */
          }
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          goto __pyx_L27_exception_handled;
        }
        goto __pyx_L28_except_error;
        __pyx_L28_except_error:;

        /* "Aspidites/_vendor/pyparsing_extension.py":3932
 *                     return longest
 * 
 *                 try:             # <<<<<<<<<<<<<<
 *                     loc2, toks = expr1._parse(instring, loc, doActions)
 *                 except ParseException as err:
 */
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_5);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_6, __pyx_t_5);
        goto __pyx_L1_error;
        __pyx_L29_except_return:;
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_5);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_6, __pyx_t_5);
        goto __pyx_L0;
        __pyx_L27_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_5);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_6, __pyx_t_5);
        __pyx_L33_try_end:;
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":3927
 * 
 *             longest = -1, None
 *             for loc1, expr1 in matches:             # <<<<<<<<<<<<<<
 *                 if loc1 <= longest[0]:
 *                     # already have a longer match than this one will deliver, we are done
 */
    }
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3946
 *                         longest = loc2, toks
 * 
 *             if longest != (-1, None):             # <<<<<<<<<<<<<<
 *                 return longest
 * 
 */
    __pyx_t_11 = PyObject_RichCompare(__pyx_v_longest, __pyx_tuple__111, Py_NE); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 3946, __pyx_L1_error)
    __pyx_t_20 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely(__pyx_t_20 < 0)) __PYX_ERR(0, 3946, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (__pyx_t_20) {

      /* "Aspidites/_vendor/pyparsing_extension.py":3947
 * 
 *             if longest != (-1, None):
 *                 return longest             # <<<<<<<<<<<<<<
 * 
 *         if maxException is not None:
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_longest);
      __pyx_r = __pyx_v_longest;
      goto __pyx_L0;

      /* "Aspidites/_vendor/pyparsing_extension.py":3946
 *                         longest = loc2, toks
 * 
 *             if longest != (-1, None):             # <<<<<<<<<<<<<<
 *                 return longest
 * 
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":3915
 *                 matches.append((loc2, e))
 * 
 *         if matches:             # <<<<<<<<<<<<<<
 *             # re-evaluate all matches in descending order of length of match, in case attached actions
 *             # might change whether or how much they match of the input.
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3949
 *                 return longest
 * 
 *         if maxException is not None:             # <<<<<<<<<<<<<<
 *             maxException.msg = self.errmsg
 *             raise maxException
 */
  __pyx_t_20 = (__pyx_v_maxException != Py_None);
  __pyx_t_14 = (__pyx_t_20 != 0);
  if (unlikely(__pyx_t_14)) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3950
 * 
 *         if maxException is not None:
 *             maxException.msg = self.errmsg             # <<<<<<<<<<<<<<
 *             raise maxException
 *         else:
 */
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errmsg); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 3950, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_maxException, __pyx_n_s_msg, __pyx_t_11) < 0) __PYX_ERR(0, 3950, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3951
 *         if maxException is not None:
 *             maxException.msg = self.errmsg
 *             raise maxException             # <<<<<<<<<<<<<<
 *         else:
 *             raise ParseException(instring, loc, "no defined alternatives to match", self)
 */
    __Pyx_Raise(__pyx_v_maxException, 0, 0, 0);
    __PYX_ERR(0, 3951, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":3949
 *                 return longest
 * 
 *         if maxException is not None:             # <<<<<<<<<<<<<<
 *             maxException.msg = self.errmsg
 *             raise maxException
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3953
 *             raise maxException
 *         else:
 *             raise ParseException(instring, loc, "no defined alternatives to match", self)             # <<<<<<<<<<<<<<
 * 
 *     def __ixor__(self, other):
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3953, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_2 = NULL;
    __pyx_t_10 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
        __pyx_t_10 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_9)) {
      PyObject *__pyx_temp[5] = {__pyx_t_2, __pyx_v_instring, __pyx_v_loc, __pyx_kp_s_no_defined_alternatives_to_match, __pyx_v_self};
      __pyx_t_11 = __Pyx_PyFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_10, 4+__pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 3953, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_11);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_9)) {
      PyObject *__pyx_temp[5] = {__pyx_t_2, __pyx_v_instring, __pyx_v_loc, __pyx_kp_s_no_defined_alternatives_to_match, __pyx_v_self};
      __pyx_t_11 = __Pyx_PyCFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_10, 4+__pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 3953, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_11);
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(4+__pyx_t_10); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3953, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_2) {
        __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_2); __pyx_t_2 = NULL;
      }
      __Pyx_INCREF(__pyx_v_instring);
      __Pyx_GIVEREF(__pyx_v_instring);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_10, __pyx_v_instring);
      __Pyx_INCREF(__pyx_v_loc);
      __Pyx_GIVEREF(__pyx_v_loc);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_10, __pyx_v_loc);
      __Pyx_INCREF(__pyx_kp_s_no_defined_alternatives_to_match);
      __Pyx_GIVEREF(__pyx_kp_s_no_defined_alternatives_to_match);
      PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_10, __pyx_kp_s_no_defined_alternatives_to_match);
      __Pyx_INCREF(__pyx_v_self);
      __Pyx_GIVEREF(__pyx_v_self);
      PyTuple_SET_ITEM(__pyx_t_8, 3+__pyx_t_10, __pyx_v_self);
      __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_8, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 3953, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_Raise(__pyx_t_11, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __PYX_ERR(0, 3953, __pyx_L1_error)
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3895
 *         return self
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         maxExcLoc = -1
 *         maxException = None
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Or.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_maxExcLoc);
  __Pyx_XDECREF(__pyx_v_maxException);
  __Pyx_XDECREF(__pyx_v_matches);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XDECREF(__pyx_v_loc2);
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XDECREF(__pyx_v_best_expr);
  __Pyx_XDECREF(__pyx_v_longest);
  __Pyx_XDECREF(__pyx_v_loc1);
  __Pyx_XDECREF(__pyx_v_expr1);
  __Pyx_XDECREF(__pyx_v_toks);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3955
 *             raise ParseException(instring, loc, "no defined alternatives to match", self)
 * 
 *     def __ixor__(self, other):             # <<<<<<<<<<<<<<
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_2Or_7__ixor__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_2Or_7__ixor__ = {"__ixor__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_2Or_7__ixor__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_2Or_7__ixor__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_other = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__ixor__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_other,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_other)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__ixor__", 1, 2, 2, 1); __PYX_ERR(0, 3955, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__ixor__") < 0)) __PYX_ERR(0, 3955, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_other = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__ixor__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3955, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Or.__ixor__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_2Or_6__ixor__(__pyx_self, __pyx_v_self, __pyx_v_other);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_2Or_6__ixor__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__ixor__", 0);
  __Pyx_INCREF(__pyx_v_other);

  /* "Aspidites/_vendor/pyparsing_extension.py":3956
 * 
 *     def __ixor__(self, other):
 *         if isinstance(other, str):             # <<<<<<<<<<<<<<
 *             other = self._literalStringClass(other)
 *         return self.append(other)  # Or([self, other])
 */
  __pyx_t_1 = PyString_Check(__pyx_v_other); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3957
 *     def __ixor__(self, other):
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)             # <<<<<<<<<<<<<<
 *         return self.append(other)  # Or([self, other])
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_literalStringClass); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3957, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_other) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_other);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3957, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_other, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3956
 * 
 *     def __ixor__(self, other):
 *         if isinstance(other, str):             # <<<<<<<<<<<<<<
 *             other = self._literalStringClass(other)
 *         return self.append(other)  # Or([self, other])
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3958
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 *         return self.append(other)  # Or([self, other])             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_append); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3958, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_other) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_other);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3958, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3955
 *             raise ParseException(instring, loc, "no defined alternatives to match", self)
 * 
 *     def __ixor__(self, other):             # <<<<<<<<<<<<<<
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Or.__ixor__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_other);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3960
 *         return self.append(other)  # Or([self, other])
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_2Or_9__str__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_2Or_9__str__ = {"__str__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_2Or_9__str__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_2Or_9__str__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_2Or_8__str__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_2Or_7__str___2generator27(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/pyparsing_extension.py":3965
 * 
 *         if self.strRepr is None:
 *             self.strRepr = "{" + " ^ ".join(str(e) for e in self.exprs) + "}"             # <<<<<<<<<<<<<<
 * 
 *         return self.strRepr
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_2Or_7__str___genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 3965, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__ *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_2Or_7__str___2generator27, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_Or___str___locals_genexpr, __pyx_n_s_Aspidites__vendor_pyparsing_exte); if (unlikely(!gen)) __PYX_ERR(0, 3965, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Or.__str__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_2Or_7__str___2generator27(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 3965, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 3965, __pyx_L1_error) }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3965, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3965, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3965, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 3965, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3965, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 3965, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3965, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 3965, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_e);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_e, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_cur_scope->__pyx_v_e); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3965, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    __Pyx_XGIVEREF(__pyx_t_2);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_3;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_4;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_2);
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 3965, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3960
 *         return self.append(other)  # Or([self, other])
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_2Or_8__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__ *__pyx_cur_scope;
  PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_2Or_7__str___2generator27 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__ *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 3960, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);

  /* "Aspidites/_vendor/pyparsing_extension.py":3961
 * 
 *     def __str__(self):
 *         if hasattr(self, "name"):             # <<<<<<<<<<<<<<
 *             return self.name
 * 
 */
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_self;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_HasAttr(__pyx_t_1, __pyx_n_s_name); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 3961, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3962
 *     def __str__(self):
 *         if hasattr(self, "name"):
 *             return self.name             # <<<<<<<<<<<<<<
 * 
 *         if self.strRepr is None:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3962, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3961
 * 
 *     def __str__(self):
 *         if hasattr(self, "name"):             # <<<<<<<<<<<<<<
 *             return self.name
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3964
 *             return self.name
 * 
 *         if self.strRepr is None:             # <<<<<<<<<<<<<<
 *             self.strRepr = "{" + " ^ ".join(str(e) for e in self.exprs) + "}"
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_strRepr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3964, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = (__pyx_t_1 == Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3965
 * 
 *         if self.strRepr is None:
 *             self.strRepr = "{" + " ^ ".join(str(e) for e in self.exprs) + "}"             # <<<<<<<<<<<<<<
 * 
 *         return self.strRepr
 */
    __pyx_t_1 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_2Or_7__str___genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3965, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyString_Join(__pyx_kp_s__112, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3965, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_Add(__pyx_kp_s__109, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3965, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyNumber_Add(__pyx_t_1, __pyx_kp_s__110); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3965, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_strRepr, __pyx_t_4) < 0) __PYX_ERR(0, 3965, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3964
 *             return self.name
 * 
 *         if self.strRepr is None:             # <<<<<<<<<<<<<<
 *             self.strRepr = "{" + " ^ ".join(str(e) for e in self.exprs) + "}"
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3967
 *             self.strRepr = "{" + " ^ ".join(str(e) for e in self.exprs) + "}"
 * 
 *         return self.strRepr             # <<<<<<<<<<<<<<
 * 
 *     def checkRecursion(self, parseElementList):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_strRepr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3967, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3960
 *         return self.append(other)  # Or([self, other])
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Or.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_2Or_7__str___2generator27);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3969
 *         return self.strRepr
 * 
 *     def checkRecursion(self, parseElementList):             # <<<<<<<<<<<<<<
 *         subRecCheckList = parseElementList[:] + [self]
 *         for e in self.exprs:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_2Or_11checkRecursion(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_2Or_11checkRecursion = {"checkRecursion", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_2Or_11checkRecursion, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_2Or_11checkRecursion(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parseElementList = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("checkRecursion (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parseElementList,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_parseElementList)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("checkRecursion", 1, 2, 2, 1); __PYX_ERR(0, 3969, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "checkRecursion") < 0)) __PYX_ERR(0, 3969, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_parseElementList = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("checkRecursion", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3969, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Or.checkRecursion", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_2Or_10checkRecursion(__pyx_self, __pyx_v_self, __pyx_v_parseElementList);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_2Or_10checkRecursion(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parseElementList) {
  PyObject *__pyx_v_subRecCheckList = NULL;
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("checkRecursion", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":3970
 * 
 *     def checkRecursion(self, parseElementList):
 *         subRecCheckList = parseElementList[:] + [self]             # <<<<<<<<<<<<<<
 *         for e in self.exprs:
 *             e.checkRecursion(subRecCheckList)
 */
  __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_parseElementList, 0, 0, NULL, NULL, &__pyx_slice__11, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3970, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3970, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_v_self);
  __pyx_t_3 = PyNumber_Add(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3970, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_subRecCheckList = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3971
 *     def checkRecursion(self, parseElementList):
 *         subRecCheckList = parseElementList[:] + [self]
 *         for e in self.exprs:             # <<<<<<<<<<<<<<
 *             e.checkRecursion(subRecCheckList)
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3971, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_2 = __pyx_t_3; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3971, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3971, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 3971, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3971, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 3971, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3971, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_5(__pyx_t_2);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 3971, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3972
 *         subRecCheckList = parseElementList[:] + [self]
 *         for e in self.exprs:
 *             e.checkRecursion(subRecCheckList)             # <<<<<<<<<<<<<<
 * 
 *     def _setResultsName(self, name, listAllMatches=False):
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_checkRecursion); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3972, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_v_subRecCheckList) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_subRecCheckList);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3972, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":3971
 *     def checkRecursion(self, parseElementList):
 *         subRecCheckList = parseElementList[:] + [self]
 *         for e in self.exprs:             # <<<<<<<<<<<<<<
 *             e.checkRecursion(subRecCheckList)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3969
 *         return self.strRepr
 * 
 *     def checkRecursion(self, parseElementList):             # <<<<<<<<<<<<<<
 *         subRecCheckList = parseElementList[:] + [self]
 *         for e in self.exprs:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Or.checkRecursion", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_subRecCheckList);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3974
 *             e.checkRecursion(subRecCheckList)
 * 
 *     def _setResultsName(self, name, listAllMatches=False):             # <<<<<<<<<<<<<<
 *         if (not __diag__.collect_all_And_tokens
 *                 and __diag__.warn_multiple_tokens_in_named_alternation):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_2Or_13_setResultsName(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_2Or_13_setResultsName = {"_setResultsName", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_2Or_13_setResultsName, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_2Or_13_setResultsName(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_listAllMatches = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_setResultsName (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_name,&__pyx_n_s_listAllMatches,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_False));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_setResultsName", 0, 2, 3, 1); __PYX_ERR(0, 3974, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_listAllMatches);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_setResultsName") < 0)) __PYX_ERR(0, 3974, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_name = values[1];
    __pyx_v_listAllMatches = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_setResultsName", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 3974, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Or._setResultsName", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_2Or_12_setResultsName(__pyx_self, __pyx_v_self, __pyx_v_name, __pyx_v_listAllMatches);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_2Or_15_setResultsName_2generator28(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/pyparsing_extension.py":3977
 *         if (not __diag__.collect_all_And_tokens
 *                 and __diag__.warn_multiple_tokens_in_named_alternation):
 *             if any(isinstance(e, And) for e in self.exprs):             # <<<<<<<<<<<<<<
 *                 warnings.warn("{0}: setting results name {1!r} on {2} expression "
 *                               "may only return a single token for an And alternative, "
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_2Or_15_setResultsName_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 3977, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_2Or_15_setResultsName_2generator28, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_Or__setResultsName_locals_genexp, __pyx_n_s_Aspidites__vendor_pyparsing_exte); if (unlikely(!gen)) __PYX_ERR(0, 3977, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Or._setResultsName.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_2Or_15_setResultsName_2generator28(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 3977, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 3977, __pyx_L1_error) }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3977, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3977, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3977, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 3977, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3977, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 3977, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3977, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 3977, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_e);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_e, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_And); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3977, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = PyObject_IsInstance(__pyx_cur_scope->__pyx_v_e, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 3977, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = (__pyx_t_5 != 0);
    if (__pyx_t_6) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_True);
      __pyx_r = Py_True;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L0;
    }
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":3974
 *             e.checkRecursion(subRecCheckList)
 * 
 *     def _setResultsName(self, name, listAllMatches=False):             # <<<<<<<<<<<<<<
 *         if (not __diag__.collect_all_And_tokens
 *                 and __diag__.warn_multiple_tokens_in_named_alternation):
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_2Or_12_setResultsName(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_listAllMatches) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName *__pyx_cur_scope;
  PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_2Or_15_setResultsName_2generator28 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_setResultsName", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 3974, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);

  /* "Aspidites/_vendor/pyparsing_extension.py":3975
 * 
 *     def _setResultsName(self, name, listAllMatches=False):
 *         if (not __diag__.collect_all_And_tokens             # <<<<<<<<<<<<<<
 *                 and __diag__.warn_multiple_tokens_in_named_alternation):
 *             if any(isinstance(e, And) for e in self.exprs):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_diag); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3975, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_collect_all_And_tokens); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3975, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 3975, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = ((!__pyx_t_4) != 0);
  if (__pyx_t_5) {
  } else {
    __pyx_t_1 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3976
 *     def _setResultsName(self, name, listAllMatches=False):
 *         if (not __diag__.collect_all_And_tokens
 *                 and __diag__.warn_multiple_tokens_in_named_alternation):             # <<<<<<<<<<<<<<
 *             if any(isinstance(e, And) for e in self.exprs):
 *                 warnings.warn("{0}: setting results name {1!r} on {2} expression "
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_diag); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3976, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_warn_multiple_tokens_in_named_al); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3976, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 3976, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;

  /* "Aspidites/_vendor/pyparsing_extension.py":3975
 * 
 *     def _setResultsName(self, name, listAllMatches=False):
 *         if (not __diag__.collect_all_And_tokens             # <<<<<<<<<<<<<<
 *                 and __diag__.warn_multiple_tokens_in_named_alternation):
 *             if any(isinstance(e, And) for e in self.exprs):
 */
  if (__pyx_t_1) {

    /* "Aspidites/_vendor/pyparsing_extension.py":3977
 *         if (not __diag__.collect_all_And_tokens
 *                 and __diag__.warn_multiple_tokens_in_named_alternation):
 *             if any(isinstance(e, And) for e in self.exprs):             # <<<<<<<<<<<<<<
 *                 warnings.warn("{0}: setting results name {1!r} on {2} expression "
 *                               "may only return a single token for an And alternative, "
 */
    __pyx_t_2 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_2Or_15_setResultsName_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3977, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_Generator_Next(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3977, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 3977, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_1) {

      /* "Aspidites/_vendor/pyparsing_extension.py":3978
 *                 and __diag__.warn_multiple_tokens_in_named_alternation):
 *             if any(isinstance(e, And) for e in self.exprs):
 *                 warnings.warn("{0}: setting results name {1!r} on {2} expression "             # <<<<<<<<<<<<<<
 *                               "may only return a single token for an And alternative, "
 *                               "in future will return the full list of tokens".format(
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_warnings); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3978, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_warn); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3978, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":3980
 *                 warnings.warn("{0}: setting results name {1!r} on {2} expression "
 *                               "may only return a single token for an And alternative, "
 *                               "in future will return the full list of tokens".format(             # <<<<<<<<<<<<<<
 *                     "warn_multiple_tokens_in_named_alternation", name, type(self).__name__),
 *                     stacklevel=3)
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_0_setting_results_name_1_r_on_2_2, __pyx_n_s_format); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3980, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);

      /* "Aspidites/_vendor/pyparsing_extension.py":3981
 *                               "may only return a single token for an And alternative, "
 *                               "in future will return the full list of tokens".format(
 *                     "warn_multiple_tokens_in_named_alternation", name, type(self).__name__),             # <<<<<<<<<<<<<<
 *                     stacklevel=3)
 * 
 */
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(__pyx_cur_scope->__pyx_v_self)), __pyx_n_s_name_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3981, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = NULL;
      __pyx_t_9 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
          __pyx_t_9 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_6)) {
        PyObject *__pyx_temp[4] = {__pyx_t_8, __pyx_n_s_warn_multiple_tokens_in_named_al, __pyx_v_name, __pyx_t_7};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3980, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
        PyObject *__pyx_temp[4] = {__pyx_t_8, __pyx_n_s_warn_multiple_tokens_in_named_al, __pyx_v_name, __pyx_t_7};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3980, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else
      #endif
      {
        __pyx_t_10 = PyTuple_New(3+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 3980, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        if (__pyx_t_8) {
          __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
        }
        __Pyx_INCREF(__pyx_n_s_warn_multiple_tokens_in_named_al);
        __Pyx_GIVEREF(__pyx_n_s_warn_multiple_tokens_in_named_al);
        PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_n_s_warn_multiple_tokens_in_named_al);
        __Pyx_INCREF(__pyx_v_name);
        __Pyx_GIVEREF(__pyx_v_name);
        PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_v_name);
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_10, 2+__pyx_t_9, __pyx_t_7);
        __pyx_t_7 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_10, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3980, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":3978
 *                 and __diag__.warn_multiple_tokens_in_named_alternation):
 *             if any(isinstance(e, And) for e in self.exprs):
 *                 warnings.warn("{0}: setting results name {1!r} on {2} expression "             # <<<<<<<<<<<<<<
 *                               "may only return a single token for an And alternative, "
 *                               "in future will return the full list of tokens".format(
 */
      __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3978, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":3982
 *                               "in future will return the full list of tokens".format(
 *                     "warn_multiple_tokens_in_named_alternation", name, type(self).__name__),
 *                     stacklevel=3)             # <<<<<<<<<<<<<<
 * 
 *         return super(Or, self)._setResultsName(name, listAllMatches)
 */
      __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3982, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_stacklevel, __pyx_int_3) < 0) __PYX_ERR(0, 3982, __pyx_L1_error)

      /* "Aspidites/_vendor/pyparsing_extension.py":3978
 *                 and __diag__.warn_multiple_tokens_in_named_alternation):
 *             if any(isinstance(e, And) for e in self.exprs):
 *                 warnings.warn("{0}: setting results name {1!r} on {2} expression "             # <<<<<<<<<<<<<<
 *                               "may only return a single token for an And alternative, "
 *                               "in future will return the full list of tokens".format(
 */
      __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, __pyx_t_3); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 3978, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":3977
 *         if (not __diag__.collect_all_And_tokens
 *                 and __diag__.warn_multiple_tokens_in_named_alternation):
 *             if any(isinstance(e, And) for e in self.exprs):             # <<<<<<<<<<<<<<
 *                 warnings.warn("{0}: setting results name {1!r} on {2} expression "
 *                               "may only return a single token for an And alternative, "
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":3975
 * 
 *     def _setResultsName(self, name, listAllMatches=False):
 *         if (not __diag__.collect_all_And_tokens             # <<<<<<<<<<<<<<
 *                 and __diag__.warn_multiple_tokens_in_named_alternation):
 *             if any(isinstance(e, And) for e in self.exprs):
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":3984
 *                     stacklevel=3)
 * 
 *         return super(Or, self)._setResultsName(name, listAllMatches)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Or); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3984, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3984, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_cur_scope->__pyx_v_self);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3984, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_setResultsName); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3984, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_9 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
      __pyx_t_9 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_name, __pyx_v_listAllMatches};
    __pyx_t_10 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 3984, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_10);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_name, __pyx_v_listAllMatches};
    __pyx_t_10 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 3984, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_10);
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3984, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_name);
    __Pyx_GIVEREF(__pyx_v_name);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_9, __pyx_v_name);
    __Pyx_INCREF(__pyx_v_listAllMatches);
    __Pyx_GIVEREF(__pyx_v_listAllMatches);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_9, __pyx_v_listAllMatches);
    __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_2, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 3984, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_r = __pyx_t_10;
  __pyx_t_10 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3974
 *             e.checkRecursion(subRecCheckList)
 * 
 *     def _setResultsName(self, name, listAllMatches=False):             # <<<<<<<<<<<<<<
 *         if (not __diag__.collect_all_And_tokens
 *                 and __diag__.warn_multiple_tokens_in_named_alternation):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Or._setResultsName", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_2Or_15_setResultsName_2generator28);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4005
 *     """
 * 
 *     def __init__(self, exprs, savelist=False):             # <<<<<<<<<<<<<<
 *         super(MatchFirst, self).__init__(exprs, savelist)
 *         if self.exprs:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_exprs = 0;
  PyObject *__pyx_v_savelist = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_exprs,&__pyx_n_s_savelist,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_False));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_exprs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, 1); __PYX_ERR(0, 4005, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_savelist);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 4005, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_exprs = values[1];
    __pyx_v_savelist = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4005, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.MatchFirst.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst___init__(__pyx_self, __pyx_v_self, __pyx_v_exprs, __pyx_v_savelist);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_8__init___2generator29(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/pyparsing_extension.py":4008
 *         super(MatchFirst, self).__init__(exprs, savelist)
 *         if self.exprs:
 *             self.mayReturnEmpty = any(e.mayReturnEmpty for e in self.exprs)             # <<<<<<<<<<<<<<
 *         else:
 *             self.mayReturnEmpty = True
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_8__init___genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 4008, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__ *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_8__init___2generator29, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_MatchFirst___init___locals_genex, __pyx_n_s_Aspidites__vendor_pyparsing_exte); if (unlikely(!gen)) __PYX_ERR(0, 4008, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.MatchFirst.__init__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_8__init___2generator29(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 4008, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 4008, __pyx_L1_error) }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4008, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4008, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4008, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 4008, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4008, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 4008, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4008, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 4008, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_e);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_e, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_e, __pyx_n_s_mayReturnEmpty); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4008, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 4008, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_5) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_True);
      __pyx_r = Py_True;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L0;
    }
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4005
 *     """
 * 
 *     def __init__(self, exprs, savelist=False):             # <<<<<<<<<<<<<<
 *         super(MatchFirst, self).__init__(exprs, savelist)
 *         if self.exprs:
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_exprs, PyObject *__pyx_v_savelist) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__ *__pyx_cur_scope;
  PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_8__init___2generator29 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__ *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 4005, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);

  /* "Aspidites/_vendor/pyparsing_extension.py":4006
 * 
 *     def __init__(self, exprs, savelist=False):
 *         super(MatchFirst, self).__init__(exprs, savelist)             # <<<<<<<<<<<<<<
 *         if self.exprs:
 *             self.mayReturnEmpty = any(e.mayReturnEmpty for e in self.exprs)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_MatchFirst); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_cur_scope->__pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_exprs, __pyx_v_savelist};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4006, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_exprs, __pyx_v_savelist};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4006, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4006, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_exprs);
    __Pyx_GIVEREF(__pyx_v_exprs);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_exprs);
    __Pyx_INCREF(__pyx_v_savelist);
    __Pyx_GIVEREF(__pyx_v_savelist);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_savelist);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4006, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4007
 *     def __init__(self, exprs, savelist=False):
 *         super(MatchFirst, self).__init__(exprs, savelist)
 *         if self.exprs:             # <<<<<<<<<<<<<<
 *             self.mayReturnEmpty = any(e.mayReturnEmpty for e in self.exprs)
 *         else:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4007, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 4007, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_6) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4008
 *         super(MatchFirst, self).__init__(exprs, savelist)
 *         if self.exprs:
 *             self.mayReturnEmpty = any(e.mayReturnEmpty for e in self.exprs)             # <<<<<<<<<<<<<<
 *         else:
 *             self.mayReturnEmpty = True
 */
    __pyx_t_1 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_8__init___genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4008, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_Generator_Next(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4008, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_mayReturnEmpty, __pyx_t_3) < 0) __PYX_ERR(0, 4008, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4007
 *     def __init__(self, exprs, savelist=False):
 *         super(MatchFirst, self).__init__(exprs, savelist)
 *         if self.exprs:             # <<<<<<<<<<<<<<
 *             self.mayReturnEmpty = any(e.mayReturnEmpty for e in self.exprs)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4010
 *             self.mayReturnEmpty = any(e.mayReturnEmpty for e in self.exprs)
 *         else:
 *             self.mayReturnEmpty = True             # <<<<<<<<<<<<<<
 * 
 *     def streamline(self):
 */
  /*else*/ {
    if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_mayReturnEmpty, Py_True) < 0) __PYX_ERR(0, 4010, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "Aspidites/_vendor/pyparsing_extension.py":4005
 *     """
 * 
 *     def __init__(self, exprs, savelist=False):             # <<<<<<<<<<<<<<
 *         super(MatchFirst, self).__init__(exprs, savelist)
 *         if self.exprs:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.MatchFirst.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_8__init___2generator29);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4012
 *             self.mayReturnEmpty = True
 * 
 *     def streamline(self):             # <<<<<<<<<<<<<<
 *         super(MatchFirst, self).streamline()
 *         if __diag__.collect_all_And_tokens:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_3streamline(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_3streamline = {"streamline", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_3streamline, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_3streamline(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("streamline (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_2streamline(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_10streamline_2generator30(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/pyparsing_extension.py":4015
 *         super(MatchFirst, self).streamline()
 *         if __diag__.collect_all_And_tokens:
 *             self.saveAsList = any(e.saveAsList for e in self.exprs)             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_10streamline_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 4015, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_10streamline_2generator30, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_MatchFirst_streamline_locals_gen, __pyx_n_s_Aspidites__vendor_pyparsing_exte); if (unlikely(!gen)) __PYX_ERR(0, 4015, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.MatchFirst.streamline.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_10streamline_2generator30(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 4015, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 4015, __pyx_L1_error) }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4015, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4015, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 4015, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4015, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 4015, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4015, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 4015, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_e);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_e, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_e, __pyx_n_s_saveAsList); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4015, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 4015, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_5) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_True);
      __pyx_r = Py_True;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L0;
    }
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4012
 *             self.mayReturnEmpty = True
 * 
 *     def streamline(self):             # <<<<<<<<<<<<<<
 *         super(MatchFirst, self).streamline()
 *         if __diag__.collect_all_And_tokens:
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_2streamline(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline *__pyx_cur_scope;
  PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_10streamline_2generator30 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("streamline", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 4012, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);

  /* "Aspidites/_vendor/pyparsing_extension.py":4013
 * 
 *     def streamline(self):
 *         super(MatchFirst, self).streamline()             # <<<<<<<<<<<<<<
 *         if __diag__.collect_all_And_tokens:
 *             self.saveAsList = any(e.saveAsList for e in self.exprs)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_MatchFirst); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_cur_scope->__pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_streamline); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4014
 *     def streamline(self):
 *         super(MatchFirst, self).streamline()
 *         if __diag__.collect_all_And_tokens:             # <<<<<<<<<<<<<<
 *             self.saveAsList = any(e.saveAsList for e in self.exprs)
 *         return self
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_diag); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_collect_all_And_tokens); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 4014, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4015
 *         super(MatchFirst, self).streamline()
 *         if __diag__.collect_all_And_tokens:
 *             self.saveAsList = any(e.saveAsList for e in self.exprs)             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
    __pyx_t_3 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_10streamline_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4015, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_Generator_Next(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4015, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_saveAsList, __pyx_t_1) < 0) __PYX_ERR(0, 4015, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4014
 *     def streamline(self):
 *         super(MatchFirst, self).streamline()
 *         if __diag__.collect_all_And_tokens:             # <<<<<<<<<<<<<<
 *             self.saveAsList = any(e.saveAsList for e in self.exprs)
 *         return self
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4016
 *         if __diag__.collect_all_And_tokens:
 *             self.saveAsList = any(e.saveAsList for e in self.exprs)
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __pyx_r = __pyx_cur_scope->__pyx_v_self;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4012
 *             self.mayReturnEmpty = True
 * 
 *     def streamline(self):             # <<<<<<<<<<<<<<
 *         super(MatchFirst, self).streamline()
 *         if __diag__.collect_all_And_tokens:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.MatchFirst.streamline", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_10streamline_2generator30);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4018
 *         return self
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         maxExcLoc = -1
 *         maxException = None
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_5parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_5parseImpl = {"parseImpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_5parseImpl, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_5parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  PyObject *__pyx_v_doActions = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parseImpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_doActions,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 1); __PYX_ERR(0, 4018, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 2); __PYX_ERR(0, 4018, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_doActions);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "parseImpl") < 0)) __PYX_ERR(0, 4018, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_doActions = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4018, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.MatchFirst.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_4parseImpl(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_4parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, PyObject *__pyx_v_doActions) {
  PyObject *__pyx_v_maxExcLoc = NULL;
  PyObject *__pyx_v_maxException = NULL;
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_v_ret = NULL;
  PyObject *__pyx_v_err = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  int __pyx_t_19;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parseImpl", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":4019
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         maxExcLoc = -1             # <<<<<<<<<<<<<<
 *         maxException = None
 *         for e in self.exprs:
 */
  __Pyx_INCREF(__pyx_int_neg_1);
  __pyx_v_maxExcLoc = __pyx_int_neg_1;

  /* "Aspidites/_vendor/pyparsing_extension.py":4020
 *     def parseImpl(self, instring, loc, doActions=True):
 *         maxExcLoc = -1
 *         maxException = None             # <<<<<<<<<<<<<<
 *         for e in self.exprs:
 *             try:
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_maxException = Py_None;

  /* "Aspidites/_vendor/pyparsing_extension.py":4021
 *         maxExcLoc = -1
 *         maxException = None
 *         for e in self.exprs:             # <<<<<<<<<<<<<<
 *             try:
 *                 ret = e._parse(instring, loc, doActions)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4021, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4021, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4021, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 4021, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4021, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 4021, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4021, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 4021, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4022
 *         maxException = None
 *         for e in self.exprs:
 *             try:             # <<<<<<<<<<<<<<
 *                 ret = e._parse(instring, loc, doActions)
 *                 return ret
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      /*try:*/ {

        /* "Aspidites/_vendor/pyparsing_extension.py":4023
 *         for e in self.exprs:
 *             try:
 *                 ret = e._parse(instring, loc, doActions)             # <<<<<<<<<<<<<<
 *                 return ret
 *             except ParseException as err:
 */
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_parse); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 4023, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_9 = NULL;
        __pyx_t_10 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
          __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
          if (likely(__pyx_t_9)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
            __Pyx_INCREF(__pyx_t_9);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_8, function);
            __pyx_t_10 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_8)) {
          PyObject *__pyx_temp[4] = {__pyx_t_9, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_10, 3+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4023, __pyx_L5_error)
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
          PyObject *__pyx_temp[4] = {__pyx_t_9, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_10, 3+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4023, __pyx_L5_error)
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_11 = PyTuple_New(3+__pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 4023, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_11);
          if (__pyx_t_9) {
            __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_9); __pyx_t_9 = NULL;
          }
          __Pyx_INCREF(__pyx_v_instring);
          __Pyx_GIVEREF(__pyx_v_instring);
          PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_10, __pyx_v_instring);
          __Pyx_INCREF(__pyx_v_loc);
          __Pyx_GIVEREF(__pyx_v_loc);
          PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_10, __pyx_v_loc);
          __Pyx_INCREF(__pyx_v_doActions);
          __Pyx_GIVEREF(__pyx_v_doActions);
          PyTuple_SET_ITEM(__pyx_t_11, 2+__pyx_t_10, __pyx_v_doActions);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_11, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4023, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        }
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF_SET(__pyx_v_ret, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":4024
 *             try:
 *                 ret = e._parse(instring, loc, doActions)
 *                 return ret             # <<<<<<<<<<<<<<
 *             except ParseException as err:
 *                 if err.loc > maxExcLoc:
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_v_ret);
        __pyx_r = __pyx_v_ret;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        goto __pyx_L9_try_return;

        /* "Aspidites/_vendor/pyparsing_extension.py":4022
 *         maxException = None
 *         for e in self.exprs:
 *             try:             # <<<<<<<<<<<<<<
 *                 ret = e._parse(instring, loc, doActions)
 *                 return ret
 */
      }
      __pyx_L5_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":4025
 *                 ret = e._parse(instring, loc, doActions)
 *                 return ret
 *             except ParseException as err:             # <<<<<<<<<<<<<<
 *                 if err.loc > maxExcLoc:
 *                     maxException = err
 */
      __Pyx_ErrFetch(&__pyx_t_1, &__pyx_t_8, &__pyx_t_11);
      __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 4025, __pyx_L7_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_1, __pyx_t_9);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_ErrRestore(__pyx_t_1, __pyx_t_8, __pyx_t_11);
      __pyx_t_1 = 0; __pyx_t_8 = 0; __pyx_t_11 = 0;
      if (__pyx_t_10) {
        __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.MatchFirst.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_11, &__pyx_t_8, &__pyx_t_1) < 0) __PYX_ERR(0, 4025, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_XDECREF_SET(__pyx_v_err, __pyx_t_8);

        /* "Aspidites/_vendor/pyparsing_extension.py":4026
 *                 return ret
 *             except ParseException as err:
 *                 if err.loc > maxExcLoc:             # <<<<<<<<<<<<<<
 *                     maxException = err
 *                     maxExcLoc = err.loc
 */
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_err, __pyx_n_s_loc); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 4026, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_12 = PyObject_RichCompare(__pyx_t_9, __pyx_v_maxExcLoc, Py_GT); __Pyx_XGOTREF(__pyx_t_12); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 4026, __pyx_L7_except_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_12); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 4026, __pyx_L7_except_error)
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        if (__pyx_t_13) {

          /* "Aspidites/_vendor/pyparsing_extension.py":4027
 *             except ParseException as err:
 *                 if err.loc > maxExcLoc:
 *                     maxException = err             # <<<<<<<<<<<<<<
 *                     maxExcLoc = err.loc
 *             except IndexError:
 */
          __Pyx_INCREF(__pyx_v_err);
          __Pyx_DECREF_SET(__pyx_v_maxException, __pyx_v_err);

          /* "Aspidites/_vendor/pyparsing_extension.py":4028
 *                 if err.loc > maxExcLoc:
 *                     maxException = err
 *                     maxExcLoc = err.loc             # <<<<<<<<<<<<<<
 *             except IndexError:
 *                 if len(instring) > maxExcLoc:
 */
          __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_err, __pyx_n_s_loc); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 4028, __pyx_L7_except_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF_SET(__pyx_v_maxExcLoc, __pyx_t_12);
          __pyx_t_12 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":4026
 *                 return ret
 *             except ParseException as err:
 *                 if err.loc > maxExcLoc:             # <<<<<<<<<<<<<<
 *                     maxException = err
 *                     maxExcLoc = err.loc
 */
        }
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L6_exception_handled;
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":4029
 *                     maxException = err
 *                     maxExcLoc = err.loc
 *             except IndexError:             # <<<<<<<<<<<<<<
 *                 if len(instring) > maxExcLoc:
 *                     maxException = ParseException(instring, len(instring), e.errmsg, self)
 */
      __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_IndexError);
      if (__pyx_t_10) {
        __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.MatchFirst.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_8, &__pyx_t_11) < 0) __PYX_ERR(0, 4029, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_GOTREF(__pyx_t_11);

        /* "Aspidites/_vendor/pyparsing_extension.py":4030
 *                     maxExcLoc = err.loc
 *             except IndexError:
 *                 if len(instring) > maxExcLoc:             # <<<<<<<<<<<<<<
 *                     maxException = ParseException(instring, len(instring), e.errmsg, self)
 *                     maxExcLoc = len(instring)
 */
        __pyx_t_14 = PyObject_Length(__pyx_v_instring); if (unlikely(__pyx_t_14 == ((Py_ssize_t)-1))) __PYX_ERR(0, 4030, __pyx_L7_except_error)
        __pyx_t_12 = PyInt_FromSsize_t(__pyx_t_14); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 4030, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_9 = PyObject_RichCompare(__pyx_t_12, __pyx_v_maxExcLoc, Py_GT); __Pyx_XGOTREF(__pyx_t_9); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 4030, __pyx_L7_except_error)
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 4030, __pyx_L7_except_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (__pyx_t_13) {

          /* "Aspidites/_vendor/pyparsing_extension.py":4031
 *             except IndexError:
 *                 if len(instring) > maxExcLoc:
 *                     maxException = ParseException(instring, len(instring), e.errmsg, self)             # <<<<<<<<<<<<<<
 *                     maxExcLoc = len(instring)
 * 
 */
          __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 4031, __pyx_L7_except_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_14 = PyObject_Length(__pyx_v_instring); if (unlikely(__pyx_t_14 == ((Py_ssize_t)-1))) __PYX_ERR(0, 4031, __pyx_L7_except_error)
          __pyx_t_15 = PyInt_FromSsize_t(__pyx_t_14); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 4031, __pyx_L7_except_error)
          __Pyx_GOTREF(__pyx_t_15);
          __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_errmsg); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 4031, __pyx_L7_except_error)
          __Pyx_GOTREF(__pyx_t_16);
          __pyx_t_17 = NULL;
          __pyx_t_10 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_12))) {
            __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_12);
            if (likely(__pyx_t_17)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
              __Pyx_INCREF(__pyx_t_17);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_12, function);
              __pyx_t_10 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[5] = {__pyx_t_17, __pyx_v_instring, __pyx_t_15, __pyx_t_16, __pyx_v_self};
            __pyx_t_9 = __Pyx_PyFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_10, 4+__pyx_t_10); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 4031, __pyx_L7_except_error)
            __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
            __Pyx_GOTREF(__pyx_t_9);
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_12)) {
            PyObject *__pyx_temp[5] = {__pyx_t_17, __pyx_v_instring, __pyx_t_15, __pyx_t_16, __pyx_v_self};
            __pyx_t_9 = __Pyx_PyCFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_10, 4+__pyx_t_10); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 4031, __pyx_L7_except_error)
            __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
            __Pyx_GOTREF(__pyx_t_9);
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          } else
          #endif
          {
            __pyx_t_18 = PyTuple_New(4+__pyx_t_10); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 4031, __pyx_L7_except_error)
            __Pyx_GOTREF(__pyx_t_18);
            if (__pyx_t_17) {
              __Pyx_GIVEREF(__pyx_t_17); PyTuple_SET_ITEM(__pyx_t_18, 0, __pyx_t_17); __pyx_t_17 = NULL;
            }
            __Pyx_INCREF(__pyx_v_instring);
            __Pyx_GIVEREF(__pyx_v_instring);
            PyTuple_SET_ITEM(__pyx_t_18, 0+__pyx_t_10, __pyx_v_instring);
            __Pyx_GIVEREF(__pyx_t_15);
            PyTuple_SET_ITEM(__pyx_t_18, 1+__pyx_t_10, __pyx_t_15);
            __Pyx_GIVEREF(__pyx_t_16);
            PyTuple_SET_ITEM(__pyx_t_18, 2+__pyx_t_10, __pyx_t_16);
            __Pyx_INCREF(__pyx_v_self);
            __Pyx_GIVEREF(__pyx_v_self);
            PyTuple_SET_ITEM(__pyx_t_18, 3+__pyx_t_10, __pyx_v_self);
            __pyx_t_15 = 0;
            __pyx_t_16 = 0;
            __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_18, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 4031, __pyx_L7_except_error)
            __Pyx_GOTREF(__pyx_t_9);
            __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
          }
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF_SET(__pyx_v_maxException, __pyx_t_9);
          __pyx_t_9 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":4032
 *                 if len(instring) > maxExcLoc:
 *                     maxException = ParseException(instring, len(instring), e.errmsg, self)
 *                     maxExcLoc = len(instring)             # <<<<<<<<<<<<<<
 * 
 *         # only got here if no expression matched, raise exception for match that made it the furthest
 */
          __pyx_t_14 = PyObject_Length(__pyx_v_instring); if (unlikely(__pyx_t_14 == ((Py_ssize_t)-1))) __PYX_ERR(0, 4032, __pyx_L7_except_error)
          __pyx_t_9 = PyInt_FromSsize_t(__pyx_t_14); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 4032, __pyx_L7_except_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF_SET(__pyx_v_maxExcLoc, __pyx_t_9);
          __pyx_t_9 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":4030
 *                     maxExcLoc = err.loc
 *             except IndexError:
 *                 if len(instring) > maxExcLoc:             # <<<<<<<<<<<<<<
 *                     maxException = ParseException(instring, len(instring), e.errmsg, self)
 *                     maxExcLoc = len(instring)
 */
        }
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        goto __pyx_L6_exception_handled;
      }
      goto __pyx_L7_except_error;
      __pyx_L7_except_error:;

      /* "Aspidites/_vendor/pyparsing_extension.py":4022
 *         maxException = None
 *         for e in self.exprs:
 *             try:             # <<<<<<<<<<<<<<
 *                 ret = e._parse(instring, loc, doActions)
 *                 return ret
 */
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
      goto __pyx_L1_error;
      __pyx_L9_try_return:;
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
      goto __pyx_L0;
      __pyx_L6_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":4021
 *         maxExcLoc = -1
 *         maxException = None
 *         for e in self.exprs:             # <<<<<<<<<<<<<<
 *             try:
 *                 ret = e._parse(instring, loc, doActions)
 */
  }
  /*else*/ {

    /* "Aspidites/_vendor/pyparsing_extension.py":4036
 *         # only got here if no expression matched, raise exception for match that made it the furthest
 *         else:
 *             if maxException is not None:             # <<<<<<<<<<<<<<
 *                 maxException.msg = self.errmsg
 *                 raise maxException
 */
    __pyx_t_13 = (__pyx_v_maxException != Py_None);
    __pyx_t_19 = (__pyx_t_13 != 0);
    if (unlikely(__pyx_t_19)) {

      /* "Aspidites/_vendor/pyparsing_extension.py":4037
 *         else:
 *             if maxException is not None:
 *                 maxException.msg = self.errmsg             # <<<<<<<<<<<<<<
 *                 raise maxException
 *             else:
 */
      __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errmsg); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 4037, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_maxException, __pyx_n_s_msg, __pyx_t_11) < 0) __PYX_ERR(0, 4037, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":4038
 *             if maxException is not None:
 *                 maxException.msg = self.errmsg
 *                 raise maxException             # <<<<<<<<<<<<<<
 *             else:
 *                 raise ParseException(instring, loc, "no defined alternatives to match", self)
 */
      __Pyx_Raise(__pyx_v_maxException, 0, 0, 0);
      __PYX_ERR(0, 4038, __pyx_L1_error)

      /* "Aspidites/_vendor/pyparsing_extension.py":4036
 *         # only got here if no expression matched, raise exception for match that made it the furthest
 *         else:
 *             if maxException is not None:             # <<<<<<<<<<<<<<
 *                 maxException.msg = self.errmsg
 *                 raise maxException
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":4040
 *                 raise maxException
 *             else:
 *                 raise ParseException(instring, loc, "no defined alternatives to match", self)             # <<<<<<<<<<<<<<
 * 
 *     def __ior__(self, other):
 */
    /*else*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 4040, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_1 = NULL;
      __pyx_t_10 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
          __pyx_t_10 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_8)) {
        PyObject *__pyx_temp[5] = {__pyx_t_1, __pyx_v_instring, __pyx_v_loc, __pyx_kp_s_no_defined_alternatives_to_match, __pyx_v_self};
        __pyx_t_11 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_10, 4+__pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 4040, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_GOTREF(__pyx_t_11);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
        PyObject *__pyx_temp[5] = {__pyx_t_1, __pyx_v_instring, __pyx_v_loc, __pyx_kp_s_no_defined_alternatives_to_match, __pyx_v_self};
        __pyx_t_11 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_10, 4+__pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 4040, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_GOTREF(__pyx_t_11);
      } else
      #endif
      {
        __pyx_t_9 = PyTuple_New(4+__pyx_t_10); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 4040, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (__pyx_t_1) {
          __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_1); __pyx_t_1 = NULL;
        }
        __Pyx_INCREF(__pyx_v_instring);
        __Pyx_GIVEREF(__pyx_v_instring);
        PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_10, __pyx_v_instring);
        __Pyx_INCREF(__pyx_v_loc);
        __Pyx_GIVEREF(__pyx_v_loc);
        PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_10, __pyx_v_loc);
        __Pyx_INCREF(__pyx_kp_s_no_defined_alternatives_to_match);
        __Pyx_GIVEREF(__pyx_kp_s_no_defined_alternatives_to_match);
        PyTuple_SET_ITEM(__pyx_t_9, 2+__pyx_t_10, __pyx_kp_s_no_defined_alternatives_to_match);
        __Pyx_INCREF(__pyx_v_self);
        __Pyx_GIVEREF(__pyx_v_self);
        PyTuple_SET_ITEM(__pyx_t_9, 3+__pyx_t_10, __pyx_v_self);
        __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_9, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 4040, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_Raise(__pyx_t_11, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __PYX_ERR(0, 4040, __pyx_L1_error)
    }
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4021
 *         maxExcLoc = -1
 *         maxException = None
 *         for e in self.exprs:             # <<<<<<<<<<<<<<
 *             try:
 *                 ret = e._parse(instring, loc, doActions)
 */
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4018
 *         return self
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         maxExcLoc = -1
 *         maxException = None
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.MatchFirst.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_maxExcLoc);
  __Pyx_XDECREF(__pyx_v_maxException);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4042
 *                 raise ParseException(instring, loc, "no defined alternatives to match", self)
 * 
 *     def __ior__(self, other):             # <<<<<<<<<<<<<<
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_7__ior__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_7__ior__ = {"__ior__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_7__ior__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_7__ior__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_other = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__ior__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_other,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_other)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__ior__", 1, 2, 2, 1); __PYX_ERR(0, 4042, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__ior__") < 0)) __PYX_ERR(0, 4042, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_other = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__ior__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4042, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.MatchFirst.__ior__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_6__ior__(__pyx_self, __pyx_v_self, __pyx_v_other);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_6__ior__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__ior__", 0);
  __Pyx_INCREF(__pyx_v_other);

  /* "Aspidites/_vendor/pyparsing_extension.py":4043
 * 
 *     def __ior__(self, other):
 *         if isinstance(other, str):             # <<<<<<<<<<<<<<
 *             other = self._literalStringClass(other)
 *         return self.append(other)  # MatchFirst([self, other])
 */
  __pyx_t_1 = PyString_Check(__pyx_v_other); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4044
 *     def __ior__(self, other):
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)             # <<<<<<<<<<<<<<
 *         return self.append(other)  # MatchFirst([self, other])
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_literalStringClass); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4044, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_other) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_other);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4044, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_other, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4043
 * 
 *     def __ior__(self, other):
 *         if isinstance(other, str):             # <<<<<<<<<<<<<<
 *             other = self._literalStringClass(other)
 *         return self.append(other)  # MatchFirst([self, other])
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4045
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 *         return self.append(other)  # MatchFirst([self, other])             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_append); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4045, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_other) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_other);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4045, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4042
 *                 raise ParseException(instring, loc, "no defined alternatives to match", self)
 * 
 *     def __ior__(self, other):             # <<<<<<<<<<<<<<
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.MatchFirst.__ior__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_other);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4047
 *         return self.append(other)  # MatchFirst([self, other])
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_9__str__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_9__str__ = {"__str__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_9__str__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_9__str__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_8__str__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_7__str___2generator31(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/pyparsing_extension.py":4052
 * 
 *         if self.strRepr is None:
 *             self.strRepr = "{" + " | ".join(str(e) for e in self.exprs) + "}"             # <<<<<<<<<<<<<<
 * 
 *         return self.strRepr
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_7__str___genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 4052, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__ *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_7__str___2generator31, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_MatchFirst___str___locals_genexp, __pyx_n_s_Aspidites__vendor_pyparsing_exte); if (unlikely(!gen)) __PYX_ERR(0, 4052, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.MatchFirst.__str__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_7__str___2generator31(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 4052, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 4052, __pyx_L1_error) }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4052, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4052, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4052, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 4052, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4052, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 4052, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4052, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 4052, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_e);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_e, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_cur_scope->__pyx_v_e); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4052, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    __Pyx_XGIVEREF(__pyx_t_2);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_3;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_4;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_2);
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 4052, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4047
 *         return self.append(other)  # MatchFirst([self, other])
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_8__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__ *__pyx_cur_scope;
  PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_7__str___2generator31 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__ *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 4047, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);

  /* "Aspidites/_vendor/pyparsing_extension.py":4048
 * 
 *     def __str__(self):
 *         if hasattr(self, "name"):             # <<<<<<<<<<<<<<
 *             return self.name
 * 
 */
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_self;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_HasAttr(__pyx_t_1, __pyx_n_s_name); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 4048, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4049
 *     def __str__(self):
 *         if hasattr(self, "name"):
 *             return self.name             # <<<<<<<<<<<<<<
 * 
 *         if self.strRepr is None:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4049, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4048
 * 
 *     def __str__(self):
 *         if hasattr(self, "name"):             # <<<<<<<<<<<<<<
 *             return self.name
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4051
 *             return self.name
 * 
 *         if self.strRepr is None:             # <<<<<<<<<<<<<<
 *             self.strRepr = "{" + " | ".join(str(e) for e in self.exprs) + "}"
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_strRepr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4051, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = (__pyx_t_1 == Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4052
 * 
 *         if self.strRepr is None:
 *             self.strRepr = "{" + " | ".join(str(e) for e in self.exprs) + "}"             # <<<<<<<<<<<<<<
 * 
 *         return self.strRepr
 */
    __pyx_t_1 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_7__str___genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4052, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyString_Join(__pyx_kp_s__113, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4052, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_Add(__pyx_kp_s__109, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4052, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyNumber_Add(__pyx_t_1, __pyx_kp_s__110); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4052, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_strRepr, __pyx_t_4) < 0) __PYX_ERR(0, 4052, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4051
 *             return self.name
 * 
 *         if self.strRepr is None:             # <<<<<<<<<<<<<<
 *             self.strRepr = "{" + " | ".join(str(e) for e in self.exprs) + "}"
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4054
 *             self.strRepr = "{" + " | ".join(str(e) for e in self.exprs) + "}"
 * 
 *         return self.strRepr             # <<<<<<<<<<<<<<
 * 
 *     def checkRecursion(self, parseElementList):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_strRepr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4054, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4047
 *         return self.append(other)  # MatchFirst([self, other])
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.MatchFirst.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_7__str___2generator31);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4056
 *         return self.strRepr
 * 
 *     def checkRecursion(self, parseElementList):             # <<<<<<<<<<<<<<
 *         subRecCheckList = parseElementList[:] + [self]
 *         for e in self.exprs:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_11checkRecursion(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_11checkRecursion = {"checkRecursion", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_11checkRecursion, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_11checkRecursion(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parseElementList = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("checkRecursion (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parseElementList,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_parseElementList)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("checkRecursion", 1, 2, 2, 1); __PYX_ERR(0, 4056, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "checkRecursion") < 0)) __PYX_ERR(0, 4056, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_parseElementList = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("checkRecursion", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4056, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.MatchFirst.checkRecursion", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_10checkRecursion(__pyx_self, __pyx_v_self, __pyx_v_parseElementList);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_10checkRecursion(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parseElementList) {
  PyObject *__pyx_v_subRecCheckList = NULL;
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("checkRecursion", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":4057
 * 
 *     def checkRecursion(self, parseElementList):
 *         subRecCheckList = parseElementList[:] + [self]             # <<<<<<<<<<<<<<
 *         for e in self.exprs:
 *             e.checkRecursion(subRecCheckList)
 */
  __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_parseElementList, 0, 0, NULL, NULL, &__pyx_slice__11, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4057, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4057, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_v_self);
  __pyx_t_3 = PyNumber_Add(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4057, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_subRecCheckList = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4058
 *     def checkRecursion(self, parseElementList):
 *         subRecCheckList = parseElementList[:] + [self]
 *         for e in self.exprs:             # <<<<<<<<<<<<<<
 *             e.checkRecursion(subRecCheckList)
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4058, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_2 = __pyx_t_3; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4058, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4058, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 4058, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4058, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 4058, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4058, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_5(__pyx_t_2);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 4058, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4059
 *         subRecCheckList = parseElementList[:] + [self]
 *         for e in self.exprs:
 *             e.checkRecursion(subRecCheckList)             # <<<<<<<<<<<<<<
 * 
 *     def _setResultsName(self, name, listAllMatches=False):
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_checkRecursion); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4059, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_v_subRecCheckList) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_subRecCheckList);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4059, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4058
 *     def checkRecursion(self, parseElementList):
 *         subRecCheckList = parseElementList[:] + [self]
 *         for e in self.exprs:             # <<<<<<<<<<<<<<
 *             e.checkRecursion(subRecCheckList)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4056
 *         return self.strRepr
 * 
 *     def checkRecursion(self, parseElementList):             # <<<<<<<<<<<<<<
 *         subRecCheckList = parseElementList[:] + [self]
 *         for e in self.exprs:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.MatchFirst.checkRecursion", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_subRecCheckList);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4061
 *             e.checkRecursion(subRecCheckList)
 * 
 *     def _setResultsName(self, name, listAllMatches=False):             # <<<<<<<<<<<<<<
 *         if (not __diag__.collect_all_And_tokens
 *                 and __diag__.warn_multiple_tokens_in_named_alternation):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_13_setResultsName(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_13_setResultsName = {"_setResultsName", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_13_setResultsName, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_13_setResultsName(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_listAllMatches = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_setResultsName (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_name,&__pyx_n_s_listAllMatches,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_False));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_setResultsName", 0, 2, 3, 1); __PYX_ERR(0, 4061, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_listAllMatches);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_setResultsName") < 0)) __PYX_ERR(0, 4061, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_name = values[1];
    __pyx_v_listAllMatches = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_setResultsName", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4061, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.MatchFirst._setResultsName", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_12_setResultsName(__pyx_self, __pyx_v_self, __pyx_v_name, __pyx_v_listAllMatches);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_15_setResultsName_2generator32(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/pyparsing_extension.py":4064
 *         if (not __diag__.collect_all_And_tokens
 *                 and __diag__.warn_multiple_tokens_in_named_alternation):
 *             if any(isinstance(e, And) for e in self.exprs):             # <<<<<<<<<<<<<<
 *                 warnings.warn("{0}: setting results name {1!r} on {2} expression "
 *                               "may only return a single token for an And alternative, "
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_15_setResultsName_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 4064, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_15_setResultsName_2generator32, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_MatchFirst__setResultsName_local, __pyx_n_s_Aspidites__vendor_pyparsing_exte); if (unlikely(!gen)) __PYX_ERR(0, 4064, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.MatchFirst._setResultsName.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_15_setResultsName_2generator32(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 4064, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 4064, __pyx_L1_error) }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4064, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4064, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4064, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 4064, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4064, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 4064, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4064, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 4064, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_e);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_e, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_And); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4064, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = PyObject_IsInstance(__pyx_cur_scope->__pyx_v_e, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 4064, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = (__pyx_t_5 != 0);
    if (__pyx_t_6) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_True);
      __pyx_r = Py_True;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L0;
    }
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4061
 *             e.checkRecursion(subRecCheckList)
 * 
 *     def _setResultsName(self, name, listAllMatches=False):             # <<<<<<<<<<<<<<
 *         if (not __diag__.collect_all_And_tokens
 *                 and __diag__.warn_multiple_tokens_in_named_alternation):
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_12_setResultsName(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_listAllMatches) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName *__pyx_cur_scope;
  PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_15_setResultsName_2generator32 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_setResultsName", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 4061, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);

  /* "Aspidites/_vendor/pyparsing_extension.py":4062
 * 
 *     def _setResultsName(self, name, listAllMatches=False):
 *         if (not __diag__.collect_all_And_tokens             # <<<<<<<<<<<<<<
 *                 and __diag__.warn_multiple_tokens_in_named_alternation):
 *             if any(isinstance(e, And) for e in self.exprs):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_diag); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4062, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_collect_all_And_tokens); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4062, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 4062, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = ((!__pyx_t_4) != 0);
  if (__pyx_t_5) {
  } else {
    __pyx_t_1 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4063
 *     def _setResultsName(self, name, listAllMatches=False):
 *         if (not __diag__.collect_all_And_tokens
 *                 and __diag__.warn_multiple_tokens_in_named_alternation):             # <<<<<<<<<<<<<<
 *             if any(isinstance(e, And) for e in self.exprs):
 *                 warnings.warn("{0}: setting results name {1!r} on {2} expression "
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_diag); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4063, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_warn_multiple_tokens_in_named_al); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4063, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 4063, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;

  /* "Aspidites/_vendor/pyparsing_extension.py":4062
 * 
 *     def _setResultsName(self, name, listAllMatches=False):
 *         if (not __diag__.collect_all_And_tokens             # <<<<<<<<<<<<<<
 *                 and __diag__.warn_multiple_tokens_in_named_alternation):
 *             if any(isinstance(e, And) for e in self.exprs):
 */
  if (__pyx_t_1) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4064
 *         if (not __diag__.collect_all_And_tokens
 *                 and __diag__.warn_multiple_tokens_in_named_alternation):
 *             if any(isinstance(e, And) for e in self.exprs):             # <<<<<<<<<<<<<<
 *                 warnings.warn("{0}: setting results name {1!r} on {2} expression "
 *                               "may only return a single token for an And alternative, "
 */
    __pyx_t_2 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_15_setResultsName_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4064, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_Generator_Next(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4064, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 4064, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_1) {

      /* "Aspidites/_vendor/pyparsing_extension.py":4065
 *                 and __diag__.warn_multiple_tokens_in_named_alternation):
 *             if any(isinstance(e, And) for e in self.exprs):
 *                 warnings.warn("{0}: setting results name {1!r} on {2} expression "             # <<<<<<<<<<<<<<
 *                               "may only return a single token for an And alternative, "
 *                               "in future will return the full list of tokens".format(
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_warnings); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4065, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_warn); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4065, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":4067
 *                 warnings.warn("{0}: setting results name {1!r} on {2} expression "
 *                               "may only return a single token for an And alternative, "
 *                               "in future will return the full list of tokens".format(             # <<<<<<<<<<<<<<
 *                     "warn_multiple_tokens_in_named_alternation", name, type(self).__name__),
 *                     stacklevel=3)
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_0_setting_results_name_1_r_on_2_2, __pyx_n_s_format); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4067, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);

      /* "Aspidites/_vendor/pyparsing_extension.py":4068
 *                               "may only return a single token for an And alternative, "
 *                               "in future will return the full list of tokens".format(
 *                     "warn_multiple_tokens_in_named_alternation", name, type(self).__name__),             # <<<<<<<<<<<<<<
 *                     stacklevel=3)
 * 
 */
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(__pyx_cur_scope->__pyx_v_self)), __pyx_n_s_name_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4068, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = NULL;
      __pyx_t_9 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
          __pyx_t_9 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_6)) {
        PyObject *__pyx_temp[4] = {__pyx_t_8, __pyx_n_s_warn_multiple_tokens_in_named_al, __pyx_v_name, __pyx_t_7};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4067, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
        PyObject *__pyx_temp[4] = {__pyx_t_8, __pyx_n_s_warn_multiple_tokens_in_named_al, __pyx_v_name, __pyx_t_7};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4067, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else
      #endif
      {
        __pyx_t_10 = PyTuple_New(3+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 4067, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        if (__pyx_t_8) {
          __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
        }
        __Pyx_INCREF(__pyx_n_s_warn_multiple_tokens_in_named_al);
        __Pyx_GIVEREF(__pyx_n_s_warn_multiple_tokens_in_named_al);
        PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_n_s_warn_multiple_tokens_in_named_al);
        __Pyx_INCREF(__pyx_v_name);
        __Pyx_GIVEREF(__pyx_v_name);
        PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_v_name);
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_10, 2+__pyx_t_9, __pyx_t_7);
        __pyx_t_7 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_10, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4067, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":4065
 *                 and __diag__.warn_multiple_tokens_in_named_alternation):
 *             if any(isinstance(e, And) for e in self.exprs):
 *                 warnings.warn("{0}: setting results name {1!r} on {2} expression "             # <<<<<<<<<<<<<<
 *                               "may only return a single token for an And alternative, "
 *                               "in future will return the full list of tokens".format(
 */
      __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4065, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":4069
 *                               "in future will return the full list of tokens".format(
 *                     "warn_multiple_tokens_in_named_alternation", name, type(self).__name__),
 *                     stacklevel=3)             # <<<<<<<<<<<<<<
 * 
 *         return super(MatchFirst, self)._setResultsName(name, listAllMatches)
 */
      __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4069, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_stacklevel, __pyx_int_3) < 0) __PYX_ERR(0, 4069, __pyx_L1_error)

      /* "Aspidites/_vendor/pyparsing_extension.py":4065
 *                 and __diag__.warn_multiple_tokens_in_named_alternation):
 *             if any(isinstance(e, And) for e in self.exprs):
 *                 warnings.warn("{0}: setting results name {1!r} on {2} expression "             # <<<<<<<<<<<<<<
 *                               "may only return a single token for an And alternative, "
 *                               "in future will return the full list of tokens".format(
 */
      __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, __pyx_t_3); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 4065, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":4064
 *         if (not __diag__.collect_all_And_tokens
 *                 and __diag__.warn_multiple_tokens_in_named_alternation):
 *             if any(isinstance(e, And) for e in self.exprs):             # <<<<<<<<<<<<<<
 *                 warnings.warn("{0}: setting results name {1!r} on {2} expression "
 *                               "may only return a single token for an And alternative, "
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":4062
 * 
 *     def _setResultsName(self, name, listAllMatches=False):
 *         if (not __diag__.collect_all_And_tokens             # <<<<<<<<<<<<<<
 *                 and __diag__.warn_multiple_tokens_in_named_alternation):
 *             if any(isinstance(e, And) for e in self.exprs):
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4071
 *                     stacklevel=3)
 * 
 *         return super(MatchFirst, self)._setResultsName(name, listAllMatches)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_MatchFirst); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4071, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4071, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_cur_scope->__pyx_v_self);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4071, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_setResultsName); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4071, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_9 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
      __pyx_t_9 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_name, __pyx_v_listAllMatches};
    __pyx_t_10 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 4071, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_10);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_name, __pyx_v_listAllMatches};
    __pyx_t_10 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 4071, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_10);
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4071, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_name);
    __Pyx_GIVEREF(__pyx_v_name);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_9, __pyx_v_name);
    __Pyx_INCREF(__pyx_v_listAllMatches);
    __Pyx_GIVEREF(__pyx_v_listAllMatches);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_9, __pyx_v_listAllMatches);
    __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_2, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 4071, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_r = __pyx_t_10;
  __pyx_t_10 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4061
 *             e.checkRecursion(subRecCheckList)
 * 
 *     def _setResultsName(self, name, listAllMatches=False):             # <<<<<<<<<<<<<<
 *         if (not __diag__.collect_all_And_tokens
 *                 and __diag__.warn_multiple_tokens_in_named_alternation):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.MatchFirst._setResultsName", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_15_setResultsName_2generator32);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4132
 *     """
 * 
 *     def __init__(self, exprs, savelist=True):             # <<<<<<<<<<<<<<
 *         super(Each, self).__init__(exprs, savelist)
 *         self.mayReturnEmpty = all(e.mayReturnEmpty for e in self.exprs)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_4Each_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_4Each_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_4Each_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_4Each_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_exprs = 0;
  PyObject *__pyx_v_savelist = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_exprs,&__pyx_n_s_savelist,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_exprs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, 1); __PYX_ERR(0, 4132, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_savelist);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 4132, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_exprs = values[1];
    __pyx_v_savelist = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4132, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Each.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Each___init__(__pyx_self, __pyx_v_self, __pyx_v_exprs, __pyx_v_savelist);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_4Each_8__init___2generator33(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/pyparsing_extension.py":4134
 *     def __init__(self, exprs, savelist=True):
 *         super(Each, self).__init__(exprs, savelist)
 *         self.mayReturnEmpty = all(e.mayReturnEmpty for e in self.exprs)             # <<<<<<<<<<<<<<
 *         self.skipWhitespace = True
 *         self.initExprGroups = True
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Each_8__init___genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 4134, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__ *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_4Each_8__init___2generator33, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_Each___init___locals_genexpr, __pyx_n_s_Aspidites__vendor_pyparsing_exte); if (unlikely(!gen)) __PYX_ERR(0, 4134, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Each.__init__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_4Each_8__init___2generator33(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 4134, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 4134, __pyx_L1_error) }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4134, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 4134, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4134, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 4134, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4134, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 4134, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_e);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_e, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_e, __pyx_n_s_mayReturnEmpty); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 4134, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = ((!__pyx_t_5) != 0);
    if (__pyx_t_6) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L0;
    }
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_True);
    __pyx_r = Py_True;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4132
 *     """
 * 
 *     def __init__(self, exprs, savelist=True):             # <<<<<<<<<<<<<<
 *         super(Each, self).__init__(exprs, savelist)
 *         self.mayReturnEmpty = all(e.mayReturnEmpty for e in self.exprs)
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Each___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_exprs, PyObject *__pyx_v_savelist) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__ *__pyx_cur_scope;
  PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_4Each_8__init___2generator33 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__ *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 4132, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);

  /* "Aspidites/_vendor/pyparsing_extension.py":4133
 * 
 *     def __init__(self, exprs, savelist=True):
 *         super(Each, self).__init__(exprs, savelist)             # <<<<<<<<<<<<<<
 *         self.mayReturnEmpty = all(e.mayReturnEmpty for e in self.exprs)
 *         self.skipWhitespace = True
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Each); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_cur_scope->__pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_exprs, __pyx_v_savelist};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4133, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_exprs, __pyx_v_savelist};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4133, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4133, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_exprs);
    __Pyx_GIVEREF(__pyx_v_exprs);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_exprs);
    __Pyx_INCREF(__pyx_v_savelist);
    __Pyx_GIVEREF(__pyx_v_savelist);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_savelist);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4133, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4134
 *     def __init__(self, exprs, savelist=True):
 *         super(Each, self).__init__(exprs, savelist)
 *         self.mayReturnEmpty = all(e.mayReturnEmpty for e in self.exprs)             # <<<<<<<<<<<<<<
 *         self.skipWhitespace = True
 *         self.initExprGroups = True
 */
  __pyx_t_1 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Each_8__init___genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_Generator_Next(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_mayReturnEmpty, __pyx_t_3) < 0) __PYX_ERR(0, 4134, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4135
 *         super(Each, self).__init__(exprs, savelist)
 *         self.mayReturnEmpty = all(e.mayReturnEmpty for e in self.exprs)
 *         self.skipWhitespace = True             # <<<<<<<<<<<<<<
 *         self.initExprGroups = True
 *         self.saveAsList = True
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_skipWhitespace, Py_True) < 0) __PYX_ERR(0, 4135, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4136
 *         self.mayReturnEmpty = all(e.mayReturnEmpty for e in self.exprs)
 *         self.skipWhitespace = True
 *         self.initExprGroups = True             # <<<<<<<<<<<<<<
 *         self.saveAsList = True
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_initExprGroups, Py_True) < 0) __PYX_ERR(0, 4136, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4137
 *         self.skipWhitespace = True
 *         self.initExprGroups = True
 *         self.saveAsList = True             # <<<<<<<<<<<<<<
 * 
 *     def streamline(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_saveAsList, Py_True) < 0) __PYX_ERR(0, 4137, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4132
 *     """
 * 
 *     def __init__(self, exprs, savelist=True):             # <<<<<<<<<<<<<<
 *         super(Each, self).__init__(exprs, savelist)
 *         self.mayReturnEmpty = all(e.mayReturnEmpty for e in self.exprs)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Each.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_4Each_8__init___2generator33);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4139
 *         self.saveAsList = True
 * 
 *     def streamline(self):             # <<<<<<<<<<<<<<
 *         super(Each, self).streamline()
 *         self.mayReturnEmpty = all(e.mayReturnEmpty for e in self.exprs)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_4Each_3streamline(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_4Each_3streamline = {"streamline", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_4Each_3streamline, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_4Each_3streamline(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("streamline (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Each_2streamline(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_4Each_10streamline_2generator34(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/pyparsing_extension.py":4141
 *     def streamline(self):
 *         super(Each, self).streamline()
 *         self.mayReturnEmpty = all(e.mayReturnEmpty for e in self.exprs)             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Each_10streamline_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 4141, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_4Each_10streamline_2generator34, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_Each_streamline_locals_genexpr, __pyx_n_s_Aspidites__vendor_pyparsing_exte); if (unlikely(!gen)) __PYX_ERR(0, 4141, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Each.streamline.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_4Each_10streamline_2generator34(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 4141, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 4141, __pyx_L1_error) }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4141, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4141, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 4141, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4141, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 4141, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4141, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 4141, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_e);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_e, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_e, __pyx_n_s_mayReturnEmpty); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4141, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 4141, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = ((!__pyx_t_5) != 0);
    if (__pyx_t_6) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L0;
    }
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_True);
    __pyx_r = Py_True;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4139
 *         self.saveAsList = True
 * 
 *     def streamline(self):             # <<<<<<<<<<<<<<
 *         super(Each, self).streamline()
 *         self.mayReturnEmpty = all(e.mayReturnEmpty for e in self.exprs)
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Each_2streamline(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline *__pyx_cur_scope;
  PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_4Each_10streamline_2generator34 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("streamline", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 4139, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);

  /* "Aspidites/_vendor/pyparsing_extension.py":4140
 * 
 *     def streamline(self):
 *         super(Each, self).streamline()             # <<<<<<<<<<<<<<
 *         self.mayReturnEmpty = all(e.mayReturnEmpty for e in self.exprs)
 *         return self
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Each); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_cur_scope->__pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_streamline); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4141
 *     def streamline(self):
 *         super(Each, self).streamline()
 *         self.mayReturnEmpty = all(e.mayReturnEmpty for e in self.exprs)             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __pyx_t_1 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Each_10streamline_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_Generator_Next(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_mayReturnEmpty, __pyx_t_3) < 0) __PYX_ERR(0, 4141, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4142
 *         super(Each, self).streamline()
 *         self.mayReturnEmpty = all(e.mayReturnEmpty for e in self.exprs)
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __pyx_r = __pyx_cur_scope->__pyx_v_self;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4139
 *         self.saveAsList = True
 * 
 *     def streamline(self):             # <<<<<<<<<<<<<<
 *         super(Each, self).streamline()
 *         self.mayReturnEmpty = all(e.mayReturnEmpty for e in self.exprs)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Each.streamline", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_4Each_10streamline_2generator34);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4144
 *         return self
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if self.initExprGroups:
 *             self.opt1map = dict((id(e.expr), e) for e in self.exprs if isinstance(e, Optional))
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_4Each_5parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_4Each_5parseImpl = {"parseImpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_4Each_5parseImpl, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_4Each_5parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  PyObject *__pyx_v_doActions = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parseImpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_doActions,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 1); __PYX_ERR(0, 4144, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 2); __PYX_ERR(0, 4144, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_doActions);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "parseImpl") < 0)) __PYX_ERR(0, 4144, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_doActions = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4144, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Each.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Each_4parseImpl(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_4Each_9parseImpl_2generator35(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/pyparsing_extension.py":4146
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if self.initExprGroups:
 *             self.opt1map = dict((id(e.expr), e) for e in self.exprs if isinstance(e, Optional))             # <<<<<<<<<<<<<<
 *             opt1 = [e.expr for e in self.exprs if isinstance(e, Optional)]
 *             opt2 = [e for e in self.exprs if e.mayReturnEmpty and not isinstance(e, (Optional, Regex))]
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Each_9parseImpl_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 4146, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_4Each_9parseImpl_2generator35, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_Each_parseImpl_locals_genexpr, __pyx_n_s_Aspidites__vendor_pyparsing_exte); if (unlikely(!gen)) __PYX_ERR(0, 4146, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Each.parseImpl.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_4Each_9parseImpl_2generator35(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 4146, __pyx_L1_error)
  __pyx_r = PyDict_New(); if (unlikely(!__pyx_r)) __PYX_ERR(0, 4146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 4146, __pyx_L1_error) }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4146, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4146, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 4146, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4146, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 4146, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4146, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 4146, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_e);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_e, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Optional); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4146, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = PyObject_IsInstance(__pyx_cur_scope->__pyx_v_e, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 4146, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = (__pyx_t_5 != 0);
    if (__pyx_t_6) {
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_e, __pyx_n_s_expr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4146, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_builtin_id, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4146, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(PyDict_SetItem(__pyx_r, (PyObject*)__pyx_t_7, (PyObject*)__pyx_cur_scope->__pyx_v_e))) __PYX_ERR(0, 4146, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_4Each_9parseImpl_5generator36(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/pyparsing_extension.py":4179
 * 
 *         if tmpReqd:
 *             missing = ", ".join(str(e) for e in tmpReqd)             # <<<<<<<<<<<<<<
 *             raise ParseException(instring, loc, "Missing one or more required elements (%s)" % missing)
 * 
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Each_9parseImpl_3genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 4179, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_4Each_9parseImpl_5generator36, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_Each_parseImpl_locals_genexpr, __pyx_n_s_Aspidites__vendor_pyparsing_exte); if (unlikely(!gen)) __PYX_ERR(0, 4179, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Each.parseImpl.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_4Each_9parseImpl_5generator36(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 4179, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_tmpReqd)) { __Pyx_RaiseClosureNameError("tmpReqd"); __PYX_ERR(0, 4179, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_tmpReqd)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_tmpReqd)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_tmpReqd; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_tmpReqd); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4179, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4179, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 4179, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4179, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 4179, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4179, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 4179, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_e);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_e, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_cur_scope->__pyx_v_e); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4179, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 4179, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4144
 *         return self
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if self.initExprGroups:
 *             self.opt1map = dict((id(e.expr), e) for e in self.exprs if isinstance(e, Optional))
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Each_4parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, PyObject *__pyx_v_doActions) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl *__pyx_cur_scope;
  PyObject *__pyx_v_opt1 = NULL;
  PyObject *__pyx_v_opt2 = NULL;
  PyObject *__pyx_v_tmpLoc = NULL;
  PyObject *__pyx_v_tmpOpt = NULL;
  PyObject *__pyx_v_matchOrder = NULL;
  int __pyx_v_keepMatching;
  PyObject *__pyx_v_tmpExprs = NULL;
  PyObject *__pyx_v_failed = NULL;
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_v_missing = NULL;
  PyObject *__pyx_v_resultlist = NULL;
  PyObject *__pyx_v_results = NULL;
  PyObject *__pyx_v_finalResults = NULL;
  PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_4Each_9parseImpl_2generator35 = 0;
  PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_4Each_9parseImpl_5generator36 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  int __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  PyObject *(*__pyx_t_19)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parseImpl", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 4144, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_INCREF(__pyx_v_loc);

  /* "Aspidites/_vendor/pyparsing_extension.py":4145
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if self.initExprGroups:             # <<<<<<<<<<<<<<
 *             self.opt1map = dict((id(e.expr), e) for e in self.exprs if isinstance(e, Optional))
 *             opt1 = [e.expr for e in self.exprs if isinstance(e, Optional)]
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_initExprGroups); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 4145, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4146
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if self.initExprGroups:
 *             self.opt1map = dict((id(e.expr), e) for e in self.exprs if isinstance(e, Optional))             # <<<<<<<<<<<<<<
 *             opt1 = [e.expr for e in self.exprs if isinstance(e, Optional)]
 *             opt2 = [e for e in self.exprs if e.mayReturnEmpty and not isinstance(e, (Optional, Regex))]
 */
    __pyx_t_1 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Each_9parseImpl_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4146, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_Generator_Next(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4146, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_opt1map, __pyx_t_3) < 0) __PYX_ERR(0, 4146, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4147
 *         if self.initExprGroups:
 *             self.opt1map = dict((id(e.expr), e) for e in self.exprs if isinstance(e, Optional))
 *             opt1 = [e.expr for e in self.exprs if isinstance(e, Optional)]             # <<<<<<<<<<<<<<
 *             opt2 = [e for e in self.exprs if e.mayReturnEmpty and not isinstance(e, (Optional, Regex))]
 *             self.optionals = opt1 + opt2
 */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
      __pyx_t_4 = __pyx_t_1; __Pyx_INCREF(__pyx_t_4); __pyx_t_5 = 0;
      __pyx_t_6 = NULL;
    } else {
      __pyx_t_5 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4147, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4147, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    for (;;) {
      if (likely(!__pyx_t_6)) {
        if (likely(PyList_CheckExact(__pyx_t_4))) {
          if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 4147, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4147, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        } else {
          if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 4147, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4147, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        }
      } else {
        __pyx_t_1 = __pyx_t_6(__pyx_t_4);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 4147, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_1);
      __pyx_t_1 = 0;
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Optional); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4147, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = PyObject_IsInstance(__pyx_v_e, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 4147, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_7 = (__pyx_t_2 != 0);
      if (__pyx_t_7) {
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_expr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4147, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_1))) __PYX_ERR(0, 4147, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_opt1 = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4148
 *             self.opt1map = dict((id(e.expr), e) for e in self.exprs if isinstance(e, Optional))
 *             opt1 = [e.expr for e in self.exprs if isinstance(e, Optional)]
 *             opt2 = [e for e in self.exprs if e.mayReturnEmpty and not isinstance(e, (Optional, Regex))]             # <<<<<<<<<<<<<<
 *             self.optionals = opt1 + opt2
 *             self.multioptionals = [e.expr for e in self.exprs if isinstance(e, ZeroOrMore)]
 */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
      __pyx_t_1 = __pyx_t_4; __Pyx_INCREF(__pyx_t_1); __pyx_t_5 = 0;
      __pyx_t_6 = NULL;
    } else {
      __pyx_t_5 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4148, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4148, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    for (;;) {
      if (likely(!__pyx_t_6)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_5); __Pyx_INCREF(__pyx_t_4); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 4148, __pyx_L1_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4148, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        } else {
          if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_5); __Pyx_INCREF(__pyx_t_4); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 4148, __pyx_L1_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4148, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        }
      } else {
        __pyx_t_4 = __pyx_t_6(__pyx_t_1);
        if (unlikely(!__pyx_t_4)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 4148, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_4);
      }
      __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_mayReturnEmpty); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4148, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 4148, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_2) {
      } else {
        __pyx_t_7 = __pyx_t_2;
        goto __pyx_L10_bool_binop_done;
      }
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_Optional); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4148, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_Regex); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 4148, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = PyObject_IsInstance(__pyx_v_e, __pyx_t_4); 
      __pyx_t_10 = (__pyx_t_9 != 0);
      if (!__pyx_t_10) {
      } else {
        __pyx_t_2 = __pyx_t_10;
        goto __pyx_L12_bool_binop_done;
      }
      __pyx_t_10 = PyObject_IsInstance(__pyx_v_e, __pyx_t_8); 
      __pyx_t_9 = (__pyx_t_10 != 0);
      __pyx_t_2 = __pyx_t_9;
      __pyx_L12_bool_binop_done:;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_9 = ((!(__pyx_t_2 != 0)) != 0);
      __pyx_t_7 = __pyx_t_9;
      __pyx_L10_bool_binop_done:;
      if (__pyx_t_7) {
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_v_e))) __PYX_ERR(0, 4148, __pyx_L1_error)
      }
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_opt2 = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4149
 *             opt1 = [e.expr for e in self.exprs if isinstance(e, Optional)]
 *             opt2 = [e for e in self.exprs if e.mayReturnEmpty and not isinstance(e, (Optional, Regex))]
 *             self.optionals = opt1 + opt2             # <<<<<<<<<<<<<<
 *             self.multioptionals = [e.expr for e in self.exprs if isinstance(e, ZeroOrMore)]
 *             self.multirequired = [e.expr for e in self.exprs if isinstance(e, OneOrMore)]
 */
    __pyx_t_3 = PyNumber_Add(__pyx_v_opt1, __pyx_v_opt2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_optionals, __pyx_t_3) < 0) __PYX_ERR(0, 4149, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4150
 *             opt2 = [e for e in self.exprs if e.mayReturnEmpty and not isinstance(e, (Optional, Regex))]
 *             self.optionals = opt1 + opt2
 *             self.multioptionals = [e.expr for e in self.exprs if isinstance(e, ZeroOrMore)]             # <<<<<<<<<<<<<<
 *             self.multirequired = [e.expr for e in self.exprs if isinstance(e, OneOrMore)]
 *             self.required = [e for e in self.exprs if not isinstance(e, (Optional, ZeroOrMore, OneOrMore))]
 */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
      __pyx_t_4 = __pyx_t_1; __Pyx_INCREF(__pyx_t_4); __pyx_t_5 = 0;
      __pyx_t_6 = NULL;
    } else {
      __pyx_t_5 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4150, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4150, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    for (;;) {
      if (likely(!__pyx_t_6)) {
        if (likely(PyList_CheckExact(__pyx_t_4))) {
          if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 4150, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4150, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        } else {
          if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 4150, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4150, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        }
      } else {
        __pyx_t_1 = __pyx_t_6(__pyx_t_4);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 4150, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_1);
      __pyx_t_1 = 0;
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ZeroOrMore); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4150, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = PyObject_IsInstance(__pyx_v_e, __pyx_t_1); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 4150, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_9 = (__pyx_t_7 != 0);
      if (__pyx_t_9) {
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_expr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4150, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_1))) __PYX_ERR(0, 4150, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_multioptionals, __pyx_t_3) < 0) __PYX_ERR(0, 4150, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4151
 *             self.optionals = opt1 + opt2
 *             self.multioptionals = [e.expr for e in self.exprs if isinstance(e, ZeroOrMore)]
 *             self.multirequired = [e.expr for e in self.exprs if isinstance(e, OneOrMore)]             # <<<<<<<<<<<<<<
 *             self.required = [e for e in self.exprs if not isinstance(e, (Optional, ZeroOrMore, OneOrMore))]
 *             self.required += self.multirequired
 */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
      __pyx_t_1 = __pyx_t_4; __Pyx_INCREF(__pyx_t_1); __pyx_t_5 = 0;
      __pyx_t_6 = NULL;
    } else {
      __pyx_t_5 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4151, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4151, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    for (;;) {
      if (likely(!__pyx_t_6)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_5); __Pyx_INCREF(__pyx_t_4); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 4151, __pyx_L1_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4151, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        } else {
          if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_5); __Pyx_INCREF(__pyx_t_4); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 4151, __pyx_L1_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4151, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        }
      } else {
        __pyx_t_4 = __pyx_t_6(__pyx_t_1);
        if (unlikely(!__pyx_t_4)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 4151, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_4);
      }
      __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_4);
      __pyx_t_4 = 0;
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_OneOrMore); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4151, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_9 = PyObject_IsInstance(__pyx_v_e, __pyx_t_4); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 4151, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_7 = (__pyx_t_9 != 0);
      if (__pyx_t_7) {
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_expr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4151, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 4151, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_multirequired, __pyx_t_3) < 0) __PYX_ERR(0, 4151, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4152
 *             self.multioptionals = [e.expr for e in self.exprs if isinstance(e, ZeroOrMore)]
 *             self.multirequired = [e.expr for e in self.exprs if isinstance(e, OneOrMore)]
 *             self.required = [e for e in self.exprs if not isinstance(e, (Optional, ZeroOrMore, OneOrMore))]             # <<<<<<<<<<<<<<
 *             self.required += self.multirequired
 *             self.initExprGroups = False
 */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
      __pyx_t_4 = __pyx_t_1; __Pyx_INCREF(__pyx_t_4); __pyx_t_5 = 0;
      __pyx_t_6 = NULL;
    } else {
      __pyx_t_5 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4152, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4152, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    for (;;) {
      if (likely(!__pyx_t_6)) {
        if (likely(PyList_CheckExact(__pyx_t_4))) {
          if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 4152, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4152, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        } else {
          if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 4152, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4152, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        }
      } else {
        __pyx_t_1 = __pyx_t_6(__pyx_t_4);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 4152, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_1);
      __pyx_t_1 = 0;
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Optional); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4152, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_ZeroOrMore); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 4152, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_OneOrMore); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 4152, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_9 = PyObject_IsInstance(__pyx_v_e, __pyx_t_1); 
      __pyx_t_2 = (__pyx_t_9 != 0);
      if (!__pyx_t_2) {
      } else {
        __pyx_t_7 = __pyx_t_2;
        goto __pyx_L23_bool_binop_done;
      }
      __pyx_t_2 = PyObject_IsInstance(__pyx_v_e, __pyx_t_8); 
      __pyx_t_9 = (__pyx_t_2 != 0);
      if (!__pyx_t_9) {
      } else {
        __pyx_t_7 = __pyx_t_9;
        goto __pyx_L23_bool_binop_done;
      }
      __pyx_t_9 = PyObject_IsInstance(__pyx_v_e, __pyx_t_11); 
      __pyx_t_2 = (__pyx_t_9 != 0);
      __pyx_t_7 = __pyx_t_2;
      __pyx_L23_bool_binop_done:;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_2 = ((!(__pyx_t_7 != 0)) != 0);
      if (__pyx_t_2) {
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_v_e))) __PYX_ERR(0, 4152, __pyx_L1_error)
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_required, __pyx_t_3) < 0) __PYX_ERR(0, 4152, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4153
 *             self.multirequired = [e.expr for e in self.exprs if isinstance(e, OneOrMore)]
 *             self.required = [e for e in self.exprs if not isinstance(e, (Optional, ZeroOrMore, OneOrMore))]
 *             self.required += self.multirequired             # <<<<<<<<<<<<<<
 *             self.initExprGroups = False
 *         tmpLoc = loc
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_required); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4153, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_multirequired); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4153, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4153, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_required, __pyx_t_1) < 0) __PYX_ERR(0, 4153, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4154
 *             self.required = [e for e in self.exprs if not isinstance(e, (Optional, ZeroOrMore, OneOrMore))]
 *             self.required += self.multirequired
 *             self.initExprGroups = False             # <<<<<<<<<<<<<<
 *         tmpLoc = loc
 *         tmpReqd = self.required[:]
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_initExprGroups, Py_False) < 0) __PYX_ERR(0, 4154, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":4145
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if self.initExprGroups:             # <<<<<<<<<<<<<<
 *             self.opt1map = dict((id(e.expr), e) for e in self.exprs if isinstance(e, Optional))
 *             opt1 = [e.expr for e in self.exprs if isinstance(e, Optional)]
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4155
 *             self.required += self.multirequired
 *             self.initExprGroups = False
 *         tmpLoc = loc             # <<<<<<<<<<<<<<
 *         tmpReqd = self.required[:]
 *         tmpOpt = self.optionals[:]
 */
  __Pyx_INCREF(__pyx_v_loc);
  __pyx_v_tmpLoc = __pyx_v_loc;

  /* "Aspidites/_vendor/pyparsing_extension.py":4156
 *             self.initExprGroups = False
 *         tmpLoc = loc
 *         tmpReqd = self.required[:]             # <<<<<<<<<<<<<<
 *         tmpOpt = self.optionals[:]
 *         matchOrder = []
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_required); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetSlice(__pyx_t_1, 0, 0, NULL, NULL, &__pyx_slice__11, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_cur_scope->__pyx_v_tmpReqd = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4157
 *         tmpLoc = loc
 *         tmpReqd = self.required[:]
 *         tmpOpt = self.optionals[:]             # <<<<<<<<<<<<<<
 *         matchOrder = []
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_optionals); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_t_4, 0, 0, NULL, NULL, &__pyx_slice__11, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_tmpOpt = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4158
 *         tmpReqd = self.required[:]
 *         tmpOpt = self.optionals[:]
 *         matchOrder = []             # <<<<<<<<<<<<<<
 * 
 *         keepMatching = True
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_matchOrder = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4160
 *         matchOrder = []
 * 
 *         keepMatching = True             # <<<<<<<<<<<<<<
 *         while keepMatching:
 *             tmpExprs = tmpReqd + tmpOpt + self.multioptionals + self.multirequired
 */
  __pyx_v_keepMatching = 1;

  /* "Aspidites/_vendor/pyparsing_extension.py":4161
 * 
 *         keepMatching = True
 *         while keepMatching:             # <<<<<<<<<<<<<<
 *             tmpExprs = tmpReqd + tmpOpt + self.multioptionals + self.multirequired
 *             failed = []
 */
  while (1) {
    __pyx_t_2 = (__pyx_v_keepMatching != 0);
    if (!__pyx_t_2) break;

    /* "Aspidites/_vendor/pyparsing_extension.py":4162
 *         keepMatching = True
 *         while keepMatching:
 *             tmpExprs = tmpReqd + tmpOpt + self.multioptionals + self.multirequired             # <<<<<<<<<<<<<<
 *             failed = []
 *             for e in tmpExprs:
 */
    __pyx_t_1 = PyNumber_Add(__pyx_cur_scope->__pyx_v_tmpReqd, __pyx_v_tmpOpt); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4162, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_multioptionals); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4162, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PyNumber_Add(__pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4162, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_multirequired); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4162, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyNumber_Add(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4162, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF_SET(__pyx_v_tmpExprs, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4163
 *         while keepMatching:
 *             tmpExprs = tmpReqd + tmpOpt + self.multioptionals + self.multirequired
 *             failed = []             # <<<<<<<<<<<<<<
 *             for e in tmpExprs:
 *                 try:
 */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4163, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_failed, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4164
 *             tmpExprs = tmpReqd + tmpOpt + self.multioptionals + self.multirequired
 *             failed = []
 *             for e in tmpExprs:             # <<<<<<<<<<<<<<
 *                 try:
 *                     tmpLoc = e.tryParse(instring, tmpLoc)
 */
    if (likely(PyList_CheckExact(__pyx_v_tmpExprs)) || PyTuple_CheckExact(__pyx_v_tmpExprs)) {
      __pyx_t_1 = __pyx_v_tmpExprs; __Pyx_INCREF(__pyx_t_1); __pyx_t_5 = 0;
      __pyx_t_6 = NULL;
    } else {
      __pyx_t_5 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_tmpExprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4164, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4164, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_6)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_5); __Pyx_INCREF(__pyx_t_4); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 4164, __pyx_L1_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4164, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        } else {
          if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_5); __Pyx_INCREF(__pyx_t_4); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 4164, __pyx_L1_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4164, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        }
      } else {
        __pyx_t_4 = __pyx_t_6(__pyx_t_1);
        if (unlikely(!__pyx_t_4)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 4164, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_4);
      }
      __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":4165
 *             failed = []
 *             for e in tmpExprs:
 *                 try:             # <<<<<<<<<<<<<<
 *                     tmpLoc = e.tryParse(instring, tmpLoc)
 *                 except ParseException:
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
        __Pyx_XGOTREF(__pyx_t_12);
        __Pyx_XGOTREF(__pyx_t_13);
        __Pyx_XGOTREF(__pyx_t_14);
        /*try:*/ {

          /* "Aspidites/_vendor/pyparsing_extension.py":4166
 *             for e in tmpExprs:
 *                 try:
 *                     tmpLoc = e.tryParse(instring, tmpLoc)             # <<<<<<<<<<<<<<
 *                 except ParseException:
 *                     failed.append(e)
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_tryParse); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4166, __pyx_L30_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_8 = NULL;
          __pyx_t_15 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_8)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_8);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
              __pyx_t_15 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_instring, __pyx_v_tmpLoc};
            __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_15, 2+__pyx_t_15); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4166, __pyx_L30_error)
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_instring, __pyx_v_tmpLoc};
            __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_15, 2+__pyx_t_15); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4166, __pyx_L30_error)
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else
          #endif
          {
            __pyx_t_11 = PyTuple_New(2+__pyx_t_15); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 4166, __pyx_L30_error)
            __Pyx_GOTREF(__pyx_t_11);
            if (__pyx_t_8) {
              __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_8); __pyx_t_8 = NULL;
            }
            __Pyx_INCREF(__pyx_v_instring);
            __Pyx_GIVEREF(__pyx_v_instring);
            PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_15, __pyx_v_instring);
            __Pyx_INCREF(__pyx_v_tmpLoc);
            __Pyx_GIVEREF(__pyx_v_tmpLoc);
            PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_15, __pyx_v_tmpLoc);
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_11, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4166, __pyx_L30_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF_SET(__pyx_v_tmpLoc, __pyx_t_4);
          __pyx_t_4 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":4165
 *             failed = []
 *             for e in tmpExprs:
 *                 try:             # <<<<<<<<<<<<<<
 *                     tmpLoc = e.tryParse(instring, tmpLoc)
 *                 except ParseException:
 */
        }

        /* "Aspidites/_vendor/pyparsing_extension.py":4170
 *                     failed.append(e)
 *                 else:
 *                     matchOrder.append(self.opt1map.get(id(e), e))             # <<<<<<<<<<<<<<
 *                     if e in tmpReqd:
 *                         tmpReqd.remove(e)
 */
        /*else:*/ {
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_opt1map); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4170, __pyx_L32_except_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_get); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 4170, __pyx_L32_except_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_id, __pyx_v_e); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4170, __pyx_L32_except_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_8 = NULL;
          __pyx_t_15 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_8)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_8);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
              __pyx_t_15 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_11)) {
            PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_3, __pyx_v_e};
            __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_15, 2+__pyx_t_15); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4170, __pyx_L32_except_error)
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_11)) {
            PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_3, __pyx_v_e};
            __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_15, 2+__pyx_t_15); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4170, __pyx_L32_except_error)
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          } else
          #endif
          {
            __pyx_t_16 = PyTuple_New(2+__pyx_t_15); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 4170, __pyx_L32_except_error)
            __Pyx_GOTREF(__pyx_t_16);
            if (__pyx_t_8) {
              __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_t_8); __pyx_t_8 = NULL;
            }
            __Pyx_GIVEREF(__pyx_t_3);
            PyTuple_SET_ITEM(__pyx_t_16, 0+__pyx_t_15, __pyx_t_3);
            __Pyx_INCREF(__pyx_v_e);
            __Pyx_GIVEREF(__pyx_v_e);
            PyTuple_SET_ITEM(__pyx_t_16, 1+__pyx_t_15, __pyx_v_e);
            __pyx_t_3 = 0;
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_16, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4170, __pyx_L32_except_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          }
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_17 = __Pyx_PyList_Append(__pyx_v_matchOrder, __pyx_t_4); if (unlikely(__pyx_t_17 == ((int)-1))) __PYX_ERR(0, 4170, __pyx_L32_except_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":4171
 *                 else:
 *                     matchOrder.append(self.opt1map.get(id(e), e))
 *                     if e in tmpReqd:             # <<<<<<<<<<<<<<
 *                         tmpReqd.remove(e)
 *                     elif e in tmpOpt:
 */
          __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_v_e, __pyx_cur_scope->__pyx_v_tmpReqd, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 4171, __pyx_L32_except_error)
          __pyx_t_7 = (__pyx_t_2 != 0);
          if (__pyx_t_7) {

            /* "Aspidites/_vendor/pyparsing_extension.py":4172
 *                     matchOrder.append(self.opt1map.get(id(e), e))
 *                     if e in tmpReqd:
 *                         tmpReqd.remove(e)             # <<<<<<<<<<<<<<
 *                     elif e in tmpOpt:
 *                         tmpOpt.remove(e)
 */
            __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_tmpReqd, __pyx_n_s_remove); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 4172, __pyx_L32_except_error)
            __Pyx_GOTREF(__pyx_t_11);
            __pyx_t_16 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
              __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_11);
              if (likely(__pyx_t_16)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
                __Pyx_INCREF(__pyx_t_16);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_11, function);
              }
            }
            __pyx_t_4 = (__pyx_t_16) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_16, __pyx_v_e) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_v_e);
            __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
            if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4172, __pyx_L32_except_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

            /* "Aspidites/_vendor/pyparsing_extension.py":4171
 *                 else:
 *                     matchOrder.append(self.opt1map.get(id(e), e))
 *                     if e in tmpReqd:             # <<<<<<<<<<<<<<
 *                         tmpReqd.remove(e)
 *                     elif e in tmpOpt:
 */
            goto __pyx_L38;
          }

          /* "Aspidites/_vendor/pyparsing_extension.py":4173
 *                     if e in tmpReqd:
 *                         tmpReqd.remove(e)
 *                     elif e in tmpOpt:             # <<<<<<<<<<<<<<
 *                         tmpOpt.remove(e)
 *             if len(failed) == len(tmpExprs):
 */
          __pyx_t_7 = (__Pyx_PySequence_ContainsTF(__pyx_v_e, __pyx_v_tmpOpt, Py_EQ)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 4173, __pyx_L32_except_error)
          __pyx_t_2 = (__pyx_t_7 != 0);
          if (__pyx_t_2) {

            /* "Aspidites/_vendor/pyparsing_extension.py":4174
 *                         tmpReqd.remove(e)
 *                     elif e in tmpOpt:
 *                         tmpOpt.remove(e)             # <<<<<<<<<<<<<<
 *             if len(failed) == len(tmpExprs):
 *                 keepMatching = False
 */
            __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_tmpOpt, __pyx_n_s_remove); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 4174, __pyx_L32_except_error)
            __Pyx_GOTREF(__pyx_t_11);
            __pyx_t_16 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
              __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_11);
              if (likely(__pyx_t_16)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
                __Pyx_INCREF(__pyx_t_16);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_11, function);
              }
            }
            __pyx_t_4 = (__pyx_t_16) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_16, __pyx_v_e) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_v_e);
            __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
            if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4174, __pyx_L32_except_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

            /* "Aspidites/_vendor/pyparsing_extension.py":4173
 *                     if e in tmpReqd:
 *                         tmpReqd.remove(e)
 *                     elif e in tmpOpt:             # <<<<<<<<<<<<<<
 *                         tmpOpt.remove(e)
 *             if len(failed) == len(tmpExprs):
 */
          }
          __pyx_L38:;
        }
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        goto __pyx_L37_try_end;
        __pyx_L30_error:;
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":4167
 *                 try:
 *                     tmpLoc = e.tryParse(instring, tmpLoc)
 *                 except ParseException:             # <<<<<<<<<<<<<<
 *                     failed.append(e)
 *                 else:
 */
        __Pyx_ErrFetch(&__pyx_t_4, &__pyx_t_11, &__pyx_t_16);
        __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4167, __pyx_L32_except_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_15 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_4, __pyx_t_3);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_ErrRestore(__pyx_t_4, __pyx_t_11, __pyx_t_16);
        __pyx_t_4 = 0; __pyx_t_11 = 0; __pyx_t_16 = 0;
        if (__pyx_t_15) {
          __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Each.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_16, &__pyx_t_11, &__pyx_t_4) < 0) __PYX_ERR(0, 4167, __pyx_L32_except_error)
          __Pyx_GOTREF(__pyx_t_16);
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_GOTREF(__pyx_t_4);

          /* "Aspidites/_vendor/pyparsing_extension.py":4168
 *                     tmpLoc = e.tryParse(instring, tmpLoc)
 *                 except ParseException:
 *                     failed.append(e)             # <<<<<<<<<<<<<<
 *                 else:
 *                     matchOrder.append(self.opt1map.get(id(e), e))
 */
          __pyx_t_17 = __Pyx_PyList_Append(__pyx_v_failed, __pyx_v_e); if (unlikely(__pyx_t_17 == ((int)-1))) __PYX_ERR(0, 4168, __pyx_L32_except_error)
          __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          goto __pyx_L31_exception_handled;
        }
        goto __pyx_L32_except_error;
        __pyx_L32_except_error:;

        /* "Aspidites/_vendor/pyparsing_extension.py":4165
 *             failed = []
 *             for e in tmpExprs:
 *                 try:             # <<<<<<<<<<<<<<
 *                     tmpLoc = e.tryParse(instring, tmpLoc)
 *                 except ParseException:
 */
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
        goto __pyx_L1_error;
        __pyx_L31_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
        __pyx_L37_try_end:;
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":4164
 *             tmpExprs = tmpReqd + tmpOpt + self.multioptionals + self.multirequired
 *             failed = []
 *             for e in tmpExprs:             # <<<<<<<<<<<<<<
 *                 try:
 *                     tmpLoc = e.tryParse(instring, tmpLoc)
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4175
 *                     elif e in tmpOpt:
 *                         tmpOpt.remove(e)
 *             if len(failed) == len(tmpExprs):             # <<<<<<<<<<<<<<
 *                 keepMatching = False
 * 
 */
    __pyx_t_5 = PyList_GET_SIZE(__pyx_v_failed); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 4175, __pyx_L1_error)
    __pyx_t_18 = PyObject_Length(__pyx_v_tmpExprs); if (unlikely(__pyx_t_18 == ((Py_ssize_t)-1))) __PYX_ERR(0, 4175, __pyx_L1_error)
    __pyx_t_2 = ((__pyx_t_5 == __pyx_t_18) != 0);
    if (__pyx_t_2) {

      /* "Aspidites/_vendor/pyparsing_extension.py":4176
 *                         tmpOpt.remove(e)
 *             if len(failed) == len(tmpExprs):
 *                 keepMatching = False             # <<<<<<<<<<<<<<
 * 
 *         if tmpReqd:
 */
      __pyx_v_keepMatching = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":4175
 *                     elif e in tmpOpt:
 *                         tmpOpt.remove(e)
 *             if len(failed) == len(tmpExprs):             # <<<<<<<<<<<<<<
 *                 keepMatching = False
 * 
 */
    }
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4178
 *                 keepMatching = False
 * 
 *         if tmpReqd:             # <<<<<<<<<<<<<<
 *             missing = ", ".join(str(e) for e in tmpReqd)
 *             raise ParseException(instring, loc, "Missing one or more required elements (%s)" % missing)
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_tmpReqd); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 4178, __pyx_L1_error)
  if (unlikely(__pyx_t_2)) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4179
 * 
 *         if tmpReqd:
 *             missing = ", ".join(str(e) for e in tmpReqd)             # <<<<<<<<<<<<<<
 *             raise ParseException(instring, loc, "Missing one or more required elements (%s)" % missing)
 * 
 */
    __pyx_t_1 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Each_9parseImpl_3genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4179, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyString_Join(__pyx_kp_s__15, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4179, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_missing = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4180
 *         if tmpReqd:
 *             missing = ", ".join(str(e) for e in tmpReqd)
 *             raise ParseException(instring, loc, "Missing one or more required elements (%s)" % missing)             # <<<<<<<<<<<<<<
 * 
 *         # add any unmatched Optionals, in case they have default values defined
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4180, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_11 = __Pyx_PyString_Format(__pyx_kp_s_Missing_one_or_more_required_ele, __pyx_v_missing); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 4180, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_16 = NULL;
    __pyx_t_15 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_16)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_16);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_15 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[4] = {__pyx_t_16, __pyx_v_instring, __pyx_v_loc, __pyx_t_11};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_15, 3+__pyx_t_15); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4180, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[4] = {__pyx_t_16, __pyx_v_instring, __pyx_v_loc, __pyx_t_11};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_15, 3+__pyx_t_15); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4180, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    } else
    #endif
    {
      __pyx_t_3 = PyTuple_New(3+__pyx_t_15); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4180, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (__pyx_t_16) {
        __Pyx_GIVEREF(__pyx_t_16); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_16); __pyx_t_16 = NULL;
      }
      __Pyx_INCREF(__pyx_v_instring);
      __Pyx_GIVEREF(__pyx_v_instring);
      PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_15, __pyx_v_instring);
      __Pyx_INCREF(__pyx_v_loc);
      __Pyx_GIVEREF(__pyx_v_loc);
      PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_15, __pyx_v_loc);
      __Pyx_GIVEREF(__pyx_t_11);
      PyTuple_SET_ITEM(__pyx_t_3, 2+__pyx_t_15, __pyx_t_11);
      __pyx_t_11 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4180, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 4180, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":4178
 *                 keepMatching = False
 * 
 *         if tmpReqd:             # <<<<<<<<<<<<<<
 *             missing = ", ".join(str(e) for e in tmpReqd)
 *             raise ParseException(instring, loc, "Missing one or more required elements (%s)" % missing)
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4183
 * 
 *         # add any unmatched Optionals, in case they have default values defined
 *         matchOrder += [e for e in self.exprs if isinstance(e, Optional) and e.expr in tmpOpt]             # <<<<<<<<<<<<<<
 * 
 *         resultlist = []
 */
  __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_3 = __pyx_t_1; __Pyx_INCREF(__pyx_t_3); __pyx_t_18 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_18 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4183, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4183, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_18 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_18); __Pyx_INCREF(__pyx_t_1); __pyx_t_18++; if (unlikely(0 < 0)) __PYX_ERR(0, 4183, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_18); __pyx_t_18++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4183, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_18 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_18); __Pyx_INCREF(__pyx_t_1); __pyx_t_18++; if (unlikely(0 < 0)) __PYX_ERR(0, 4183, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_18); __pyx_t_18++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4183, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_6(__pyx_t_3);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 4183, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Optional); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4183, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = PyObject_IsInstance(__pyx_v_e, __pyx_t_1); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 4183, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_9 = (__pyx_t_7 != 0);
    if (__pyx_t_9) {
    } else {
      __pyx_t_2 = __pyx_t_9;
      goto __pyx_L46_bool_binop_done;
    }
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_expr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4183, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_v_tmpOpt, Py_EQ)); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 4183, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_7 = (__pyx_t_9 != 0);
    __pyx_t_2 = __pyx_t_7;
    __pyx_L46_bool_binop_done:;
    if (__pyx_t_2) {
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_v_e))) __PYX_ERR(0, 4183, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_matchOrder, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF_SET(__pyx_v_matchOrder, ((PyObject*)__pyx_t_3));
  __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4185
 *         matchOrder += [e for e in self.exprs if isinstance(e, Optional) and e.expr in tmpOpt]
 * 
 *         resultlist = []             # <<<<<<<<<<<<<<
 *         for e in matchOrder:
 *             loc, results = e._parse(instring, loc, doActions)
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_resultlist = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4186
 * 
 *         resultlist = []
 *         for e in matchOrder:             # <<<<<<<<<<<<<<
 *             loc, results = e._parse(instring, loc, doActions)
 *             resultlist.append(results)
 */
  __pyx_t_3 = __pyx_v_matchOrder; __Pyx_INCREF(__pyx_t_3); __pyx_t_18 = 0;
  for (;;) {
    if (__pyx_t_18 >= PyList_GET_SIZE(__pyx_t_3)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_4 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_18); __Pyx_INCREF(__pyx_t_4); __pyx_t_18++; if (unlikely(0 < 0)) __PYX_ERR(0, 4186, __pyx_L1_error)
    #else
    __pyx_t_4 = PySequence_ITEM(__pyx_t_3, __pyx_t_18); __pyx_t_18++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4186, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4187
 *         resultlist = []
 *         for e in matchOrder:
 *             loc, results = e._parse(instring, loc, doActions)             # <<<<<<<<<<<<<<
 *             resultlist.append(results)
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_parse); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4187, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_11 = NULL;
    __pyx_t_15 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_11)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_15 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[4] = {__pyx_t_11, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_15, 3+__pyx_t_15); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4187, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_GOTREF(__pyx_t_4);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[4] = {__pyx_t_11, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_15, 3+__pyx_t_15); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4187, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_GOTREF(__pyx_t_4);
    } else
    #endif
    {
      __pyx_t_16 = PyTuple_New(3+__pyx_t_15); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 4187, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_16);
      if (__pyx_t_11) {
        __Pyx_GIVEREF(__pyx_t_11); PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_t_11); __pyx_t_11 = NULL;
      }
      __Pyx_INCREF(__pyx_v_instring);
      __Pyx_GIVEREF(__pyx_v_instring);
      PyTuple_SET_ITEM(__pyx_t_16, 0+__pyx_t_15, __pyx_v_instring);
      __Pyx_INCREF(__pyx_v_loc);
      __Pyx_GIVEREF(__pyx_v_loc);
      PyTuple_SET_ITEM(__pyx_t_16, 1+__pyx_t_15, __pyx_v_loc);
      __Pyx_INCREF(__pyx_v_doActions);
      __Pyx_GIVEREF(__pyx_v_doActions);
      PyTuple_SET_ITEM(__pyx_t_16, 2+__pyx_t_15, __pyx_v_doActions);
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_16, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4187, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
      PyObject* sequence = __pyx_t_4;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 4187, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_16 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_16 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_16);
      #else
      __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4187, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_16 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 4187, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_16);
      #endif
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_11 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 4187, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_19 = Py_TYPE(__pyx_t_11)->tp_iternext;
      index = 0; __pyx_t_1 = __pyx_t_19(__pyx_t_11); if (unlikely(!__pyx_t_1)) goto __pyx_L50_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_1);
      index = 1; __pyx_t_16 = __pyx_t_19(__pyx_t_11); if (unlikely(!__pyx_t_16)) goto __pyx_L50_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_16);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_19(__pyx_t_11), 2) < 0) __PYX_ERR(0, 4187, __pyx_L1_error)
      __pyx_t_19 = NULL;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      goto __pyx_L51_unpacking_done;
      __pyx_L50_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_t_19 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 4187, __pyx_L1_error)
      __pyx_L51_unpacking_done:;
    }
    __Pyx_DECREF_SET(__pyx_v_loc, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_results, __pyx_t_16);
    __pyx_t_16 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4188
 *         for e in matchOrder:
 *             loc, results = e._parse(instring, loc, doActions)
 *             resultlist.append(results)             # <<<<<<<<<<<<<<
 * 
 *         finalResults = sum(resultlist, ParseResults([]))
 */
    __pyx_t_17 = __Pyx_PyList_Append(__pyx_v_resultlist, __pyx_v_results); if (unlikely(__pyx_t_17 == ((int)-1))) __PYX_ERR(0, 4188, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":4186
 * 
 *         resultlist = []
 *         for e in matchOrder:             # <<<<<<<<<<<<<<
 *             loc, results = e._parse(instring, loc, doActions)
 *             resultlist.append(results)
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4190
 *             resultlist.append(results)
 * 
 *         finalResults = sum(resultlist, ParseResults([]))             # <<<<<<<<<<<<<<
 *         return loc, finalResults
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ParseResults); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_16 = PyList_New(0); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 4190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_1, __pyx_t_16) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_16);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_resultlist);
  __Pyx_GIVEREF(__pyx_v_resultlist);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_resultlist);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_sum, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_finalResults = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4191
 * 
 *         finalResults = sum(resultlist, ParseResults([]))
 *         return loc, finalResults             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_loc);
  __Pyx_GIVEREF(__pyx_v_loc);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_loc);
  __Pyx_INCREF(__pyx_v_finalResults);
  __Pyx_GIVEREF(__pyx_v_finalResults);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_finalResults);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4144
 *         return self
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if self.initExprGroups:
 *             self.opt1map = dict((id(e.expr), e) for e in self.exprs if isinstance(e, Optional))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Each.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_opt1);
  __Pyx_XDECREF(__pyx_v_opt2);
  __Pyx_XDECREF(__pyx_v_tmpLoc);
  __Pyx_XDECREF(__pyx_v_tmpOpt);
  __Pyx_XDECREF(__pyx_v_matchOrder);
  __Pyx_XDECREF(__pyx_v_tmpExprs);
  __Pyx_XDECREF(__pyx_v_failed);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XDECREF(__pyx_v_missing);
  __Pyx_XDECREF(__pyx_v_resultlist);
  __Pyx_XDECREF(__pyx_v_results);
  __Pyx_XDECREF(__pyx_v_finalResults);
  __Pyx_XDECREF(__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_4Each_9parseImpl_2generator35);
  __Pyx_XDECREF(__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_4Each_9parseImpl_5generator36);
  __Pyx_XDECREF(__pyx_v_loc);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4193
 *         return loc, finalResults
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_4Each_7__str__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_4Each_7__str__ = {"__str__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_4Each_7__str__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_4Each_7__str__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Each_6__str__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_4Each_7__str___2generator37(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "Aspidites/_vendor/pyparsing_extension.py":4198
 * 
 *         if self.strRepr is None:
 *             self.strRepr = "{" + " & ".join(str(e) for e in self.exprs) + "}"             # <<<<<<<<<<<<<<
 * 
 *         return self.strRepr
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Each_7__str___genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 4198, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__ *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_4Each_7__str___2generator37, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_Each___str___locals_genexpr, __pyx_n_s_Aspidites__vendor_pyparsing_exte); if (unlikely(!gen)) __PYX_ERR(0, 4198, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Each.__str__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_4Each_7__str___2generator37(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 4198, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 4198, __pyx_L1_error) }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4198, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4198, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 4198, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4198, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 4198, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4198, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 4198, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_e);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_e, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_cur_scope->__pyx_v_e); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4198, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    __Pyx_XGIVEREF(__pyx_t_2);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_3;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_4;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_2);
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 4198, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4193
 *         return loc, finalResults
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Each_6__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__ *__pyx_cur_scope;
  PyObject *__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_4Each_7__str___2generator37 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);
  __pyx_cur_scope = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__ *)__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__(__pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 4193, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);

  /* "Aspidites/_vendor/pyparsing_extension.py":4194
 * 
 *     def __str__(self):
 *         if hasattr(self, "name"):             # <<<<<<<<<<<<<<
 *             return self.name
 * 
 */
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_self;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_HasAttr(__pyx_t_1, __pyx_n_s_name); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 4194, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4195
 *     def __str__(self):
 *         if hasattr(self, "name"):
 *             return self.name             # <<<<<<<<<<<<<<
 * 
 *         if self.strRepr is None:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4195, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4194
 * 
 *     def __str__(self):
 *         if hasattr(self, "name"):             # <<<<<<<<<<<<<<
 *             return self.name
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4197
 *             return self.name
 * 
 *         if self.strRepr is None:             # <<<<<<<<<<<<<<
 *             self.strRepr = "{" + " & ".join(str(e) for e in self.exprs) + "}"
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_strRepr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = (__pyx_t_1 == Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4198
 * 
 *         if self.strRepr is None:
 *             self.strRepr = "{" + " & ".join(str(e) for e in self.exprs) + "}"             # <<<<<<<<<<<<<<
 * 
 *         return self.strRepr
 */
    __pyx_t_1 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Each_7__str___genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4198, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyString_Join(__pyx_kp_s__114, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4198, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_Add(__pyx_kp_s__109, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4198, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyNumber_Add(__pyx_t_1, __pyx_kp_s__110); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4198, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_strRepr, __pyx_t_4) < 0) __PYX_ERR(0, 4198, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4197
 *             return self.name
 * 
 *         if self.strRepr is None:             # <<<<<<<<<<<<<<
 *             self.strRepr = "{" + " & ".join(str(e) for e in self.exprs) + "}"
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4200
 *             self.strRepr = "{" + " & ".join(str(e) for e in self.exprs) + "}"
 * 
 *         return self.strRepr             # <<<<<<<<<<<<<<
 * 
 *     def checkRecursion(self, parseElementList):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_strRepr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4193
 *         return loc, finalResults
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Each.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_gb_9Aspidites_7_vendor_19pyparsing_extension_4Each_7__str___2generator37);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4202
 *         return self.strRepr
 * 
 *     def checkRecursion(self, parseElementList):             # <<<<<<<<<<<<<<
 *         subRecCheckList = parseElementList[:] + [self]
 *         for e in self.exprs:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_4Each_9checkRecursion(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_4Each_9checkRecursion = {"checkRecursion", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_4Each_9checkRecursion, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_4Each_9checkRecursion(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parseElementList = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("checkRecursion (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parseElementList,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_parseElementList)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("checkRecursion", 1, 2, 2, 1); __PYX_ERR(0, 4202, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "checkRecursion") < 0)) __PYX_ERR(0, 4202, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_parseElementList = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("checkRecursion", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4202, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Each.checkRecursion", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Each_8checkRecursion(__pyx_self, __pyx_v_self, __pyx_v_parseElementList);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Each_8checkRecursion(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parseElementList) {
  PyObject *__pyx_v_subRecCheckList = NULL;
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("checkRecursion", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":4203
 * 
 *     def checkRecursion(self, parseElementList):
 *         subRecCheckList = parseElementList[:] + [self]             # <<<<<<<<<<<<<<
 *         for e in self.exprs:
 *             e.checkRecursion(subRecCheckList)
 */
  __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_parseElementList, 0, 0, NULL, NULL, &__pyx_slice__11, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_v_self);
  __pyx_t_3 = PyNumber_Add(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_subRecCheckList = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4204
 *     def checkRecursion(self, parseElementList):
 *         subRecCheckList = parseElementList[:] + [self]
 *         for e in self.exprs:             # <<<<<<<<<<<<<<
 *             e.checkRecursion(subRecCheckList)
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exprs); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_2 = __pyx_t_3; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4204, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 4204, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4204, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 4204, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4204, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_5(__pyx_t_2);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 4204, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4205
 *         subRecCheckList = parseElementList[:] + [self]
 *         for e in self.exprs:
 *             e.checkRecursion(subRecCheckList)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_checkRecursion); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4205, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_v_subRecCheckList) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_subRecCheckList);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4205, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4204
 *     def checkRecursion(self, parseElementList):
 *         subRecCheckList = parseElementList[:] + [self]
 *         for e in self.exprs:             # <<<<<<<<<<<<<<
 *             e.checkRecursion(subRecCheckList)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4202
 *         return self.strRepr
 * 
 *     def checkRecursion(self, parseElementList):             # <<<<<<<<<<<<<<
 *         subRecCheckList = parseElementList[:] + [self]
 *         for e in self.exprs:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Each.checkRecursion", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_subRecCheckList);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4213
 *     """
 * 
 *     def __init__(self, expr, savelist=False):             # <<<<<<<<<<<<<<
 *         super(ParseElementEnhance, self).__init__(savelist)
 *         if isinstance(expr, str):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_expr = 0;
  PyObject *__pyx_v_savelist = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_expr,&__pyx_n_s_savelist,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_False));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_expr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, 1); __PYX_ERR(0, 4213, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_savelist);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 4213, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_expr = values[1];
    __pyx_v_savelist = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4213, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseElementEnhance.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance___init__(__pyx_self, __pyx_v_self, __pyx_v_expr, __pyx_v_savelist);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_expr, PyObject *__pyx_v_savelist) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);
  __Pyx_INCREF(__pyx_v_expr);

  /* "Aspidites/_vendor/pyparsing_extension.py":4214
 * 
 *     def __init__(self, expr, savelist=False):
 *         super(ParseElementEnhance, self).__init__(savelist)             # <<<<<<<<<<<<<<
 *         if isinstance(expr, str):
 *             if issubclass(self._literalStringClass, Token):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParseElementEnhance); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_savelist) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_savelist);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4215
 *     def __init__(self, expr, savelist=False):
 *         super(ParseElementEnhance, self).__init__(savelist)
 *         if isinstance(expr, str):             # <<<<<<<<<<<<<<
 *             if issubclass(self._literalStringClass, Token):
 *                 expr = self._literalStringClass(expr)
 */
  __pyx_t_4 = PyString_Check(__pyx_v_expr); 
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4216
 *         super(ParseElementEnhance, self).__init__(savelist)
 *         if isinstance(expr, str):
 *             if issubclass(self._literalStringClass, Token):             # <<<<<<<<<<<<<<
 *                 expr = self._literalStringClass(expr)
 *             else:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_literalStringClass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4216, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Token); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4216, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = PyObject_IsSubclass(__pyx_t_1, __pyx_t_3); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 4216, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = (__pyx_t_5 != 0);
    if (__pyx_t_4) {

      /* "Aspidites/_vendor/pyparsing_extension.py":4217
 *         if isinstance(expr, str):
 *             if issubclass(self._literalStringClass, Token):
 *                 expr = self._literalStringClass(expr)             # <<<<<<<<<<<<<<
 *             else:
 *                 expr = self._literalStringClass(Literal(expr))
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_literalStringClass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4217, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_v_expr) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_expr);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4217, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF_SET(__pyx_v_expr, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":4216
 *         super(ParseElementEnhance, self).__init__(savelist)
 *         if isinstance(expr, str):
 *             if issubclass(self._literalStringClass, Token):             # <<<<<<<<<<<<<<
 *                 expr = self._literalStringClass(expr)
 *             else:
 */
      goto __pyx_L4;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":4219
 *                 expr = self._literalStringClass(expr)
 *             else:
 *                 expr = self._literalStringClass(Literal(expr))             # <<<<<<<<<<<<<<
 *         self.expr = expr
 *         self.strRepr = None
 */
    /*else*/ {
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_literalStringClass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4219, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_Literal); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4219, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_2 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_7, __pyx_v_expr) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_expr);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4219, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4219, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF_SET(__pyx_v_expr, __pyx_t_3);
      __pyx_t_3 = 0;
    }
    __pyx_L4:;

    /* "Aspidites/_vendor/pyparsing_extension.py":4215
 *     def __init__(self, expr, savelist=False):
 *         super(ParseElementEnhance, self).__init__(savelist)
 *         if isinstance(expr, str):             # <<<<<<<<<<<<<<
 *             if issubclass(self._literalStringClass, Token):
 *                 expr = self._literalStringClass(expr)
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4220
 *             else:
 *                 expr = self._literalStringClass(Literal(expr))
 *         self.expr = expr             # <<<<<<<<<<<<<<
 *         self.strRepr = None
 *         if expr is not None:
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_expr, __pyx_v_expr) < 0) __PYX_ERR(0, 4220, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4221
 *                 expr = self._literalStringClass(Literal(expr))
 *         self.expr = expr
 *         self.strRepr = None             # <<<<<<<<<<<<<<
 *         if expr is not None:
 *             self.mayIndexError = expr.mayIndexError
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_strRepr, Py_None) < 0) __PYX_ERR(0, 4221, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4222
 *         self.expr = expr
 *         self.strRepr = None
 *         if expr is not None:             # <<<<<<<<<<<<<<
 *             self.mayIndexError = expr.mayIndexError
 *             self.mayReturnEmpty = expr.mayReturnEmpty
 */
  __pyx_t_4 = (__pyx_v_expr != Py_None);
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4223
 *         self.strRepr = None
 *         if expr is not None:
 *             self.mayIndexError = expr.mayIndexError             # <<<<<<<<<<<<<<
 *             self.mayReturnEmpty = expr.mayReturnEmpty
 *             self.setWhitespaceChars(expr.whiteChars)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_expr, __pyx_n_s_mayIndexError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4223, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_mayIndexError, __pyx_t_3) < 0) __PYX_ERR(0, 4223, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4224
 *         if expr is not None:
 *             self.mayIndexError = expr.mayIndexError
 *             self.mayReturnEmpty = expr.mayReturnEmpty             # <<<<<<<<<<<<<<
 *             self.setWhitespaceChars(expr.whiteChars)
 *             self.skipWhitespace = expr.skipWhitespace
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_expr, __pyx_n_s_mayReturnEmpty); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4224, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_mayReturnEmpty, __pyx_t_3) < 0) __PYX_ERR(0, 4224, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4225
 *             self.mayIndexError = expr.mayIndexError
 *             self.mayReturnEmpty = expr.mayReturnEmpty
 *             self.setWhitespaceChars(expr.whiteChars)             # <<<<<<<<<<<<<<
 *             self.skipWhitespace = expr.skipWhitespace
 *             self.saveAsList = expr.saveAsList
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_setWhitespaceChars); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4225, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_expr, __pyx_n_s_whiteChars); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4225, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4225, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4226
 *             self.mayReturnEmpty = expr.mayReturnEmpty
 *             self.setWhitespaceChars(expr.whiteChars)
 *             self.skipWhitespace = expr.skipWhitespace             # <<<<<<<<<<<<<<
 *             self.saveAsList = expr.saveAsList
 *             self.callPreparse = expr.callPreparse
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_expr, __pyx_n_s_skipWhitespace); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4226, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_skipWhitespace, __pyx_t_3) < 0) __PYX_ERR(0, 4226, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4227
 *             self.setWhitespaceChars(expr.whiteChars)
 *             self.skipWhitespace = expr.skipWhitespace
 *             self.saveAsList = expr.saveAsList             # <<<<<<<<<<<<<<
 *             self.callPreparse = expr.callPreparse
 *             self.ignoreExprs.extend(expr.ignoreExprs)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_expr, __pyx_n_s_saveAsList); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4227, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_saveAsList, __pyx_t_3) < 0) __PYX_ERR(0, 4227, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4228
 *             self.skipWhitespace = expr.skipWhitespace
 *             self.saveAsList = expr.saveAsList
 *             self.callPreparse = expr.callPreparse             # <<<<<<<<<<<<<<
 *             self.ignoreExprs.extend(expr.ignoreExprs)
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_expr, __pyx_n_s_callPreparse); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4228, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_callPreparse, __pyx_t_3) < 0) __PYX_ERR(0, 4228, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4229
 *             self.saveAsList = expr.saveAsList
 *             self.callPreparse = expr.callPreparse
 *             self.ignoreExprs.extend(expr.ignoreExprs)             # <<<<<<<<<<<<<<
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ignoreExprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4229, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_extend); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4229, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_expr, __pyx_n_s_ignoreExprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4229, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_6, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4229, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4222
 *         self.expr = expr
 *         self.strRepr = None
 *         if expr is not None:             # <<<<<<<<<<<<<<
 *             self.mayIndexError = expr.mayIndexError
 *             self.mayReturnEmpty = expr.mayReturnEmpty
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4213
 *     """
 * 
 *     def __init__(self, expr, savelist=False):             # <<<<<<<<<<<<<<
 *         super(ParseElementEnhance, self).__init__(savelist)
 *         if isinstance(expr, str):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseElementEnhance.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_expr);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4231
 *             self.ignoreExprs.extend(expr.ignoreExprs)
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if self.expr is not None:
 *             return self.expr._parse(instring, loc, doActions, callPreParse=False)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_3parseImpl = {"parseImpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_3parseImpl, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  PyObject *__pyx_v_doActions = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parseImpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_doActions,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 1); __PYX_ERR(0, 4231, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 2); __PYX_ERR(0, 4231, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_doActions);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "parseImpl") < 0)) __PYX_ERR(0, 4231, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_doActions = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4231, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseElementEnhance.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_2parseImpl(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, PyObject *__pyx_v_doActions) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parseImpl", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":4232
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if self.expr is not None:             # <<<<<<<<<<<<<<
 *             return self.expr._parse(instring, loc, doActions, callPreParse=False)
 *         else:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__pyx_t_1 != Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (likely(__pyx_t_3)) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4233
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if self.expr is not None:
 *             return self.expr._parse(instring, loc, doActions, callPreParse=False)             # <<<<<<<<<<<<<<
 *         else:
 *             raise ParseException("", loc, self.errmsg, self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4233, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_parse); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4233, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4233, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_instring);
    __Pyx_GIVEREF(__pyx_v_instring);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_instring);
    __Pyx_INCREF(__pyx_v_loc);
    __Pyx_GIVEREF(__pyx_v_loc);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_loc);
    __Pyx_INCREF(__pyx_v_doActions);
    __Pyx_GIVEREF(__pyx_v_doActions);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_doActions);
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4233, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_callPreParse, Py_False) < 0) __PYX_ERR(0, 4233, __pyx_L1_error)
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_1, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4233, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4232
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if self.expr is not None:             # <<<<<<<<<<<<<<
 *             return self.expr._parse(instring, loc, doActions, callPreParse=False)
 *         else:
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4235
 *             return self.expr._parse(instring, loc, doActions, callPreParse=False)
 *         else:
 *             raise ParseException("", loc, self.errmsg, self)             # <<<<<<<<<<<<<<
 * 
 *     def leaveWhitespace(self):
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4235, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errmsg); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4235, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[5] = {__pyx_t_4, __pyx_kp_s__12, __pyx_v_loc, __pyx_t_1, __pyx_v_self};
      __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4235, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[5] = {__pyx_t_4, __pyx_kp_s__12, __pyx_v_loc, __pyx_t_1, __pyx_v_self};
      __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4235, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(4+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 4235, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_4) {
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_4); __pyx_t_4 = NULL;
      }
      __Pyx_INCREF(__pyx_kp_s__12);
      __Pyx_GIVEREF(__pyx_kp_s__12);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_kp_s__12);
      __Pyx_INCREF(__pyx_v_loc);
      __Pyx_GIVEREF(__pyx_v_loc);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_v_loc);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_7, __pyx_t_1);
      __Pyx_INCREF(__pyx_v_self);
      __Pyx_GIVEREF(__pyx_v_self);
      PyTuple_SET_ITEM(__pyx_t_8, 3+__pyx_t_7, __pyx_v_self);
      __pyx_t_1 = 0;
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4235, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 4235, __pyx_L1_error)
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4231
 *             self.ignoreExprs.extend(expr.ignoreExprs)
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if self.expr is not None:
 *             return self.expr._parse(instring, loc, doActions, callPreParse=False)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseElementEnhance.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4237
 *             raise ParseException("", loc, self.errmsg, self)
 * 
 *     def leaveWhitespace(self):             # <<<<<<<<<<<<<<
 *         self.skipWhitespace = False
 *         self.expr = self.expr.copy()
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_5leaveWhitespace(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_5leaveWhitespace = {"leaveWhitespace", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_5leaveWhitespace, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_5leaveWhitespace(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("leaveWhitespace (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_4leaveWhitespace(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_4leaveWhitespace(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("leaveWhitespace", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":4238
 * 
 *     def leaveWhitespace(self):
 *         self.skipWhitespace = False             # <<<<<<<<<<<<<<
 *         self.expr = self.expr.copy()
 *         if self.expr is not None:
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_skipWhitespace, Py_False) < 0) __PYX_ERR(0, 4238, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4239
 *     def leaveWhitespace(self):
 *         self.skipWhitespace = False
 *         self.expr = self.expr.copy()             # <<<<<<<<<<<<<<
 *         if self.expr is not None:
 *             self.expr.leaveWhitespace()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_copy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_expr, __pyx_t_1) < 0) __PYX_ERR(0, 4239, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4240
 *         self.skipWhitespace = False
 *         self.expr = self.expr.copy()
 *         if self.expr is not None:             # <<<<<<<<<<<<<<
 *             self.expr.leaveWhitespace()
 *         return self
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = (__pyx_t_1 != Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4241
 *         self.expr = self.expr.copy()
 *         if self.expr is not None:
 *             self.expr.leaveWhitespace()             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4241, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_leaveWhitespace); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4241, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4241, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4240
 *         self.skipWhitespace = False
 *         self.expr = self.expr.copy()
 *         if self.expr is not None:             # <<<<<<<<<<<<<<
 *             self.expr.leaveWhitespace()
 *         return self
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4242
 *         if self.expr is not None:
 *             self.expr.leaveWhitespace()
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def ignore(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4237
 *             raise ParseException("", loc, self.errmsg, self)
 * 
 *     def leaveWhitespace(self):             # <<<<<<<<<<<<<<
 *         self.skipWhitespace = False
 *         self.expr = self.expr.copy()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseElementEnhance.leaveWhitespace", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4244
 *         return self
 * 
 *     def ignore(self, other):             # <<<<<<<<<<<<<<
 *         if isinstance(other, Suppress):
 *             if other not in self.ignoreExprs:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_7ignore(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_7ignore = {"ignore", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_7ignore, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_7ignore(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_other = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ignore (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_other,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_other)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ignore", 1, 2, 2, 1); __PYX_ERR(0, 4244, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "ignore") < 0)) __PYX_ERR(0, 4244, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_other = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ignore", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4244, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseElementEnhance.ignore", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_6ignore(__pyx_self, __pyx_v_self, __pyx_v_other);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_6ignore(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ignore", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":4245
 * 
 *     def ignore(self, other):
 *         if isinstance(other, Suppress):             # <<<<<<<<<<<<<<
 *             if other not in self.ignoreExprs:
 *                 super(ParseElementEnhance, self).ignore(other)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Suppress); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_IsInstance(__pyx_v_other, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 4245, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4246
 *     def ignore(self, other):
 *         if isinstance(other, Suppress):
 *             if other not in self.ignoreExprs:             # <<<<<<<<<<<<<<
 *                 super(ParseElementEnhance, self).ignore(other)
 *                 if self.expr is not None:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ignoreExprs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4246, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = (__Pyx_PySequence_ContainsTF(__pyx_v_other, __pyx_t_1, Py_NE)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 4246, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_2 = (__pyx_t_3 != 0);
    if (__pyx_t_2) {

      /* "Aspidites/_vendor/pyparsing_extension.py":4247
 *         if isinstance(other, Suppress):
 *             if other not in self.ignoreExprs:
 *                 super(ParseElementEnhance, self).ignore(other)             # <<<<<<<<<<<<<<
 *                 if self.expr is not None:
 *                     self.expr.ignore(self.ignoreExprs[-1])
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ParseElementEnhance); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4247, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4247, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
      __Pyx_INCREF(__pyx_v_self);
      __Pyx_GIVEREF(__pyx_v_self);
      PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_self);
      __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_5, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4247, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_ignore); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4247, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_v_other) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_other);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4247, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":4248
 *             if other not in self.ignoreExprs:
 *                 super(ParseElementEnhance, self).ignore(other)
 *                 if self.expr is not None:             # <<<<<<<<<<<<<<
 *                     self.expr.ignore(self.ignoreExprs[-1])
 *         else:
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4248, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = (__pyx_t_1 != Py_None);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {

        /* "Aspidites/_vendor/pyparsing_extension.py":4249
 *                 super(ParseElementEnhance, self).ignore(other)
 *                 if self.expr is not None:
 *                     self.expr.ignore(self.ignoreExprs[-1])             # <<<<<<<<<<<<<<
 *         else:
 *             super(ParseElementEnhance, self).ignore(other)
 */
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4249, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_ignore); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4249, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ignoreExprs); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4249, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_5, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4249, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4249, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":4248
 *             if other not in self.ignoreExprs:
 *                 super(ParseElementEnhance, self).ignore(other)
 *                 if self.expr is not None:             # <<<<<<<<<<<<<<
 *                     self.expr.ignore(self.ignoreExprs[-1])
 *         else:
 */
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":4246
 *     def ignore(self, other):
 *         if isinstance(other, Suppress):
 *             if other not in self.ignoreExprs:             # <<<<<<<<<<<<<<
 *                 super(ParseElementEnhance, self).ignore(other)
 *                 if self.expr is not None:
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":4245
 * 
 *     def ignore(self, other):
 *         if isinstance(other, Suppress):             # <<<<<<<<<<<<<<
 *             if other not in self.ignoreExprs:
 *                 super(ParseElementEnhance, self).ignore(other)
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4251
 *                     self.expr.ignore(self.ignoreExprs[-1])
 *         else:
 *             super(ParseElementEnhance, self).ignore(other)             # <<<<<<<<<<<<<<
 *             if self.expr is not None:
 *                 self.expr.ignore(self.ignoreExprs[-1])
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ParseElementEnhance); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4);
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_v_self);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_ignore); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_4, __pyx_v_other) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_other);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4252
 *         else:
 *             super(ParseElementEnhance, self).ignore(other)
 *             if self.expr is not None:             # <<<<<<<<<<<<<<
 *                 self.expr.ignore(self.ignoreExprs[-1])
 *         return self
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4252, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = (__pyx_t_1 != Py_None);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_2 = (__pyx_t_3 != 0);
    if (__pyx_t_2) {

      /* "Aspidites/_vendor/pyparsing_extension.py":4253
 *             super(ParseElementEnhance, self).ignore(other)
 *             if self.expr is not None:
 *                 self.expr.ignore(self.ignoreExprs[-1])             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4253, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_ignore); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4253, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ignoreExprs); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4253, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_6, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4253, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4253, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":4252
 *         else:
 *             super(ParseElementEnhance, self).ignore(other)
 *             if self.expr is not None:             # <<<<<<<<<<<<<<
 *                 self.expr.ignore(self.ignoreExprs[-1])
 *         return self
 */
    }
  }
  __pyx_L3:;

  /* "Aspidites/_vendor/pyparsing_extension.py":4254
 *             if self.expr is not None:
 *                 self.expr.ignore(self.ignoreExprs[-1])
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def streamline(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4244
 *         return self
 * 
 *     def ignore(self, other):             # <<<<<<<<<<<<<<
 *         if isinstance(other, Suppress):
 *             if other not in self.ignoreExprs:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseElementEnhance.ignore", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4256
 *         return self
 * 
 *     def streamline(self):             # <<<<<<<<<<<<<<
 *         super(ParseElementEnhance, self).streamline()
 *         if self.expr is not None:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_9streamline(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_9streamline = {"streamline", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_9streamline, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_9streamline(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("streamline (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_8streamline(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_8streamline(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("streamline", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":4257
 * 
 *     def streamline(self):
 *         super(ParseElementEnhance, self).streamline()             # <<<<<<<<<<<<<<
 *         if self.expr is not None:
 *             self.expr.streamline()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParseElementEnhance); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_streamline); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4258
 *     def streamline(self):
 *         super(ParseElementEnhance, self).streamline()
 *         if self.expr is not None:             # <<<<<<<<<<<<<<
 *             self.expr.streamline()
 *         return self
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = (__pyx_t_1 != Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4259
 *         super(ParseElementEnhance, self).streamline()
 *         if self.expr is not None:
 *             self.expr.streamline()             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4259, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_streamline); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4259, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4259, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4258
 *     def streamline(self):
 *         super(ParseElementEnhance, self).streamline()
 *         if self.expr is not None:             # <<<<<<<<<<<<<<
 *             self.expr.streamline()
 *         return self
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4260
 *         if self.expr is not None:
 *             self.expr.streamline()
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def checkRecursion(self, parseElementList):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4256
 *         return self
 * 
 *     def streamline(self):             # <<<<<<<<<<<<<<
 *         super(ParseElementEnhance, self).streamline()
 *         if self.expr is not None:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseElementEnhance.streamline", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4262
 *         return self
 * 
 *     def checkRecursion(self, parseElementList):             # <<<<<<<<<<<<<<
 *         if self in parseElementList:
 *             raise RecursiveGrammarException(parseElementList + [self])
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_11checkRecursion(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_11checkRecursion = {"checkRecursion", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_11checkRecursion, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_11checkRecursion(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_parseElementList = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("checkRecursion (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_parseElementList,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_parseElementList)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("checkRecursion", 1, 2, 2, 1); __PYX_ERR(0, 4262, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "checkRecursion") < 0)) __PYX_ERR(0, 4262, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_parseElementList = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("checkRecursion", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4262, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseElementEnhance.checkRecursion", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_10checkRecursion(__pyx_self, __pyx_v_self, __pyx_v_parseElementList);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_10checkRecursion(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_parseElementList) {
  PyObject *__pyx_v_subRecCheckList = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("checkRecursion", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":4263
 * 
 *     def checkRecursion(self, parseElementList):
 *         if self in parseElementList:             # <<<<<<<<<<<<<<
 *             raise RecursiveGrammarException(parseElementList + [self])
 *         subRecCheckList = parseElementList[:] + [self]
 */
  __pyx_t_1 = (__Pyx_PySequence_ContainsTF(__pyx_v_self, __pyx_v_parseElementList, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 4263, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (unlikely(__pyx_t_2)) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4264
 *     def checkRecursion(self, parseElementList):
 *         if self in parseElementList:
 *             raise RecursiveGrammarException(parseElementList + [self])             # <<<<<<<<<<<<<<
 *         subRecCheckList = parseElementList[:] + [self]
 *         if self.expr is not None:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_RecursiveGrammarException); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4264, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyList_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4264, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyList_SET_ITEM(__pyx_t_5, 0, __pyx_v_self);
    __pyx_t_6 = PyNumber_Add(__pyx_v_parseElementList, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4264, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4264, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 4264, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":4263
 * 
 *     def checkRecursion(self, parseElementList):
 *         if self in parseElementList:             # <<<<<<<<<<<<<<
 *             raise RecursiveGrammarException(parseElementList + [self])
 *         subRecCheckList = parseElementList[:] + [self]
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4265
 *         if self in parseElementList:
 *             raise RecursiveGrammarException(parseElementList + [self])
 *         subRecCheckList = parseElementList[:] + [self]             # <<<<<<<<<<<<<<
 *         if self.expr is not None:
 *             self.expr.checkRecursion(subRecCheckList)
 */
  __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_v_parseElementList, 0, 0, NULL, NULL, &__pyx_slice__11, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyList_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyList_SET_ITEM(__pyx_t_4, 0, __pyx_v_self);
  __pyx_t_6 = PyNumber_Add(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_subRecCheckList = __pyx_t_6;
  __pyx_t_6 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4266
 *             raise RecursiveGrammarException(parseElementList + [self])
 *         subRecCheckList = parseElementList[:] + [self]
 *         if self.expr is not None:             # <<<<<<<<<<<<<<
 *             self.expr.checkRecursion(subRecCheckList)
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_2 = (__pyx_t_6 != Py_None);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4267
 *         subRecCheckList = parseElementList[:] + [self]
 *         if self.expr is not None:
 *             self.expr.checkRecursion(subRecCheckList)             # <<<<<<<<<<<<<<
 * 
 *     def validate(self, validateTrace=None):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4267, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_checkRecursion); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4267, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_6 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_subRecCheckList) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_subRecCheckList);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4267, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4266
 *             raise RecursiveGrammarException(parseElementList + [self])
 *         subRecCheckList = parseElementList[:] + [self]
 *         if self.expr is not None:             # <<<<<<<<<<<<<<
 *             self.expr.checkRecursion(subRecCheckList)
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4262
 *         return self
 * 
 *     def checkRecursion(self, parseElementList):             # <<<<<<<<<<<<<<
 *         if self in parseElementList:
 *             raise RecursiveGrammarException(parseElementList + [self])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseElementEnhance.checkRecursion", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_subRecCheckList);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4269
 *             self.expr.checkRecursion(subRecCheckList)
 * 
 *     def validate(self, validateTrace=None):             # <<<<<<<<<<<<<<
 *         if validateTrace is None:
 *             validateTrace = []
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_13validate(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_13validate = {"validate", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_13validate, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_13validate(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_validateTrace = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("validate (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_validateTrace,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_validateTrace);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "validate") < 0)) __PYX_ERR(0, 4269, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_validateTrace = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("validate", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4269, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseElementEnhance.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_12validate(__pyx_self, __pyx_v_self, __pyx_v_validateTrace);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_12validate(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_validateTrace) {
  PyObject *__pyx_v_tmp = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("validate", 0);
  __Pyx_INCREF(__pyx_v_validateTrace);

  /* "Aspidites/_vendor/pyparsing_extension.py":4270
 * 
 *     def validate(self, validateTrace=None):
 *         if validateTrace is None:             # <<<<<<<<<<<<<<
 *             validateTrace = []
 *         tmp = validateTrace[:] + [self]
 */
  __pyx_t_1 = (__pyx_v_validateTrace == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4271
 *     def validate(self, validateTrace=None):
 *         if validateTrace is None:
 *             validateTrace = []             # <<<<<<<<<<<<<<
 *         tmp = validateTrace[:] + [self]
 *         if self.expr is not None:
 */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4271, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF_SET(__pyx_v_validateTrace, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4270
 * 
 *     def validate(self, validateTrace=None):
 *         if validateTrace is None:             # <<<<<<<<<<<<<<
 *             validateTrace = []
 *         tmp = validateTrace[:] + [self]
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4272
 *         if validateTrace is None:
 *             validateTrace = []
 *         tmp = validateTrace[:] + [self]             # <<<<<<<<<<<<<<
 *         if self.expr is not None:
 *             self.expr.validate(tmp)
 */
  __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_v_validateTrace, 0, 0, NULL, NULL, &__pyx_slice__11, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyList_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyList_SET_ITEM(__pyx_t_4, 0, __pyx_v_self);
  __pyx_t_5 = PyNumber_Add(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_tmp = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4273
 *             validateTrace = []
 *         tmp = validateTrace[:] + [self]
 *         if self.expr is not None:             # <<<<<<<<<<<<<<
 *             self.expr.validate(tmp)
 *         self.checkRecursion([])
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = (__pyx_t_5 != Py_None);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4274
 *         tmp = validateTrace[:] + [self]
 *         if self.expr is not None:
 *             self.expr.validate(tmp)             # <<<<<<<<<<<<<<
 *         self.checkRecursion([])
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_validate); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_5 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_tmp) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_tmp);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4273
 *             validateTrace = []
 *         tmp = validateTrace[:] + [self]
 *         if self.expr is not None:             # <<<<<<<<<<<<<<
 *             self.expr.validate(tmp)
 *         self.checkRecursion([])
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4275
 *         if self.expr is not None:
 *             self.expr.validate(tmp)
 *         self.checkRecursion([])             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_checkRecursion); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_5 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_6, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4269
 *             self.expr.checkRecursion(subRecCheckList)
 * 
 *     def validate(self, validateTrace=None):             # <<<<<<<<<<<<<<
 *         if validateTrace is None:
 *             validateTrace = []
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseElementEnhance.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tmp);
  __Pyx_XDECREF(__pyx_v_validateTrace);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4277
 *         self.checkRecursion([])
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         try:
 *             return super(ParseElementEnhance, self).__str__()
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_15__str__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_15__str__ = {"__str__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_15__str__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_15__str__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_14__str__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_14__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":4278
 * 
 *     def __str__(self):
 *         try:             # <<<<<<<<<<<<<<
 *             return super(ParseElementEnhance, self).__str__()
 *         except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "Aspidites/_vendor/pyparsing_extension.py":4279
 *     def __str__(self):
 *         try:
 *             return super(ParseElementEnhance, self).__str__()             # <<<<<<<<<<<<<<
 *         except Exception:
 *             pass
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_ParseElementEnhance); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4279, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4279, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5);
      __Pyx_INCREF(__pyx_v_self);
      __Pyx_GIVEREF(__pyx_v_self);
      PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_v_self);
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_6, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4279, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_str); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4279, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4279, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_r = __pyx_t_4;
      __pyx_t_4 = 0;
      goto __pyx_L7_try_return;

      /* "Aspidites/_vendor/pyparsing_extension.py":4278
 * 
 *     def __str__(self):
 *         try:             # <<<<<<<<<<<<<<
 *             return super(ParseElementEnhance, self).__str__()
 *         except Exception:
 */
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4280
 *         try:
 *             return super(ParseElementEnhance, self).__str__()
 *         except Exception:             # <<<<<<<<<<<<<<
 *             pass
 * 
 */
    __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_7) {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "Aspidites/_vendor/pyparsing_extension.py":4278
 * 
 *     def __str__(self):
 *         try:             # <<<<<<<<<<<<<<
 *             return super(ParseElementEnhance, self).__str__()
 *         except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4283
 *             pass
 * 
 *         if self.strRepr is None and self.expr is not None:             # <<<<<<<<<<<<<<
 *             self.strRepr = "%s:(%s)" % (self.__class__.__name__, str(self.expr))
 *         return self.strRepr
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_strRepr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_9 = (__pyx_t_4 == Py_None);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_10 = (__pyx_t_9 != 0);
  if (__pyx_t_10) {
  } else {
    __pyx_t_8 = __pyx_t_10;
    goto __pyx_L10_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_10 = (__pyx_t_4 != Py_None);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_9 = (__pyx_t_10 != 0);
  __pyx_t_8 = __pyx_t_9;
  __pyx_L10_bool_binop_done:;
  if (__pyx_t_8) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4284
 * 
 *         if self.strRepr is None and self.expr is not None:
 *             self.strRepr = "%s:(%s)" % (self.__class__.__name__, str(self.expr))             # <<<<<<<<<<<<<<
 *         return self.strRepr
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_class); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4284, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_name_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4284, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4284, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4284, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4284, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_5);
    __pyx_t_6 = 0;
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_s_s_2, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4284, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_strRepr, __pyx_t_5) < 0) __PYX_ERR(0, 4284, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4283
 *             pass
 * 
 *         if self.strRepr is None and self.expr is not None:             # <<<<<<<<<<<<<<
 *             self.strRepr = "%s:(%s)" % (self.__class__.__name__, str(self.expr))
 *         return self.strRepr
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4285
 *         if self.strRepr is None and self.expr is not None:
 *             self.strRepr = "%s:(%s)" % (self.__class__.__name__, str(self.expr))
 *         return self.strRepr             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_strRepr); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4277
 *         self.checkRecursion([])
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         try:
 *             return super(ParseElementEnhance, self).__str__()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ParseElementEnhance.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4311
 *     """
 * 
 *     def __init__(self, expr):             # <<<<<<<<<<<<<<
 *         super(FollowedBy, self).__init__(expr)
 *         self.mayReturnEmpty = True
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10FollowedBy_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10FollowedBy_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10FollowedBy_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10FollowedBy_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_expr = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_expr,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_expr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(0, 4311, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 4311, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_expr = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4311, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.FollowedBy.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10FollowedBy___init__(__pyx_self, __pyx_v_self, __pyx_v_expr);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10FollowedBy___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_expr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":4312
 * 
 *     def __init__(self, expr):
 *         super(FollowedBy, self).__init__(expr)             # <<<<<<<<<<<<<<
 *         self.mayReturnEmpty = True
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_FollowedBy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4312, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4312, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4312, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4312, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_expr) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_expr);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4312, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4313
 *     def __init__(self, expr):
 *         super(FollowedBy, self).__init__(expr)
 *         self.mayReturnEmpty = True             # <<<<<<<<<<<<<<
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_mayReturnEmpty, Py_True) < 0) __PYX_ERR(0, 4313, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4311
 *     """
 * 
 *     def __init__(self, expr):             # <<<<<<<<<<<<<<
 *         super(FollowedBy, self).__init__(expr)
 *         self.mayReturnEmpty = True
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.FollowedBy.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4315
 *         self.mayReturnEmpty = True
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         # by using self._expr.parse and deleting the contents of the returned ParseResults list
 *         # we keep any named results that were defined in the FollowedBy expression
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10FollowedBy_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10FollowedBy_3parseImpl = {"parseImpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10FollowedBy_3parseImpl, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10FollowedBy_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  PyObject *__pyx_v_doActions = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parseImpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_doActions,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 1); __PYX_ERR(0, 4315, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 2); __PYX_ERR(0, 4315, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_doActions);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "parseImpl") < 0)) __PYX_ERR(0, 4315, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_doActions = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4315, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.FollowedBy.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10FollowedBy_2parseImpl(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10FollowedBy_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, PyObject *__pyx_v_doActions) {
  CYTHON_UNUSED PyObject *__pyx_v__ = NULL;
  PyObject *__pyx_v_ret = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parseImpl", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":4318
 *         # by using self._expr.parse and deleting the contents of the returned ParseResults list
 *         # we keep any named results that were defined in the FollowedBy expression
 *         _, ret = self.expr._parse(instring, loc, doActions=doActions)             # <<<<<<<<<<<<<<
 *         del ret[:]
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_parse); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_instring);
  __Pyx_GIVEREF(__pyx_v_instring);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_instring);
  __Pyx_INCREF(__pyx_v_loc);
  __Pyx_GIVEREF(__pyx_v_loc);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_loc);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_doActions, __pyx_v_doActions) < 0) __PYX_ERR(0, 4318, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
    PyObject* sequence = __pyx_t_4;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 4318, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_1 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_1);
    #else
    __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4318, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4318, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    #endif
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_2 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4318, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext;
    index = 0; __pyx_t_3 = __pyx_t_5(__pyx_t_2); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    index = 1; __pyx_t_1 = __pyx_t_5(__pyx_t_2); if (unlikely(!__pyx_t_1)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_1);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_2), 2) < 0) __PYX_ERR(0, 4318, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 4318, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v__ = __pyx_t_3;
  __pyx_t_3 = 0;
  __pyx_v_ret = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4319
 *         # we keep any named results that were defined in the FollowedBy expression
 *         _, ret = self.expr._parse(instring, loc, doActions=doActions)
 *         del ret[:]             # <<<<<<<<<<<<<<
 * 
 *         return loc, ret
 */
  if (__Pyx_PyObject_DelSlice(__pyx_v_ret, 0, 0, NULL, NULL, &__pyx_slice__11, 0, 0, 1) < 0) __PYX_ERR(0, 4319, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4321
 *         del ret[:]
 * 
 *         return loc, ret             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_loc);
  __Pyx_GIVEREF(__pyx_v_loc);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_loc);
  __Pyx_INCREF(__pyx_v_ret);
  __Pyx_GIVEREF(__pyx_v_ret);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_ret);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4315
 *         self.mayReturnEmpty = True
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         # by using self._expr.parse and deleting the contents of the returned ParseResults list
 *         # we keep any named results that were defined in the FollowedBy expression
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.FollowedBy.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v__);
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4353
 *     """
 * 
 *     def __init__(self, expr, retreat=None):             # <<<<<<<<<<<<<<
 *         super(PrecededBy, self).__init__(expr)
 *         self.expr = self.expr().leaveWhitespace()
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10PrecededBy_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10PrecededBy_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10PrecededBy_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10PrecededBy_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_expr = 0;
  PyObject *__pyx_v_retreat = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_expr,&__pyx_n_s_retreat,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_expr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, 1); __PYX_ERR(0, 4353, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_retreat);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 4353, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_expr = values[1];
    __pyx_v_retreat = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4353, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.PrecededBy.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10PrecededBy___init__(__pyx_self, __pyx_v_self, __pyx_v_expr, __pyx_v_retreat);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4374
 *         self.errmsg = "not preceded by " + str(expr)
 *         self.skipWhitespace = False
 *         self.parseAction.append(lambda s, l, t: t.__delitem__(slice(None, None)))             # <<<<<<<<<<<<<<
 * 
 *     def parseImpl(self, instring, loc=0, doActions=True):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10PrecededBy_8__init___lambda37(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10PrecededBy_8__init___lambda37 = {"lambda37", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10PrecededBy_8__init___lambda37, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10PrecededBy_8__init___lambda37(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_s = 0;
  CYTHON_UNUSED PyObject *__pyx_v_l = 0;
  PyObject *__pyx_v_t = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda37 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_s_4,&__pyx_n_s_l,&__pyx_n_s_t_2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_s_4)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_l)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("lambda37", 1, 3, 3, 1); __PYX_ERR(0, 4374, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_t_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("lambda37", 1, 3, 3, 2); __PYX_ERR(0, 4374, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "lambda37") < 0)) __PYX_ERR(0, 4374, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_s = values[0];
    __pyx_v_l = values[1];
    __pyx_v_t = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("lambda37", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4374, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.PrecededBy.__init__.lambda37", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_lambda_funcdef_lambda37(__pyx_self, __pyx_v_s, __pyx_v_l, __pyx_v_t);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda37(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_s, CYTHON_UNUSED PyObject *__pyx_v_l, PyObject *__pyx_v_t) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda37", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_t, __pyx_n_s_delitem); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_slice__11) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_slice__11);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.PrecededBy.__init__.lambda37", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4353
 *     """
 * 
 *     def __init__(self, expr, retreat=None):             # <<<<<<<<<<<<<<
 *         super(PrecededBy, self).__init__(expr)
 *         self.expr = self.expr().leaveWhitespace()
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10PrecededBy___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_expr, PyObject *__pyx_v_retreat) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);
  __Pyx_INCREF(__pyx_v_retreat);

  /* "Aspidites/_vendor/pyparsing_extension.py":4354
 * 
 *     def __init__(self, expr, retreat=None):
 *         super(PrecededBy, self).__init__(expr)             # <<<<<<<<<<<<<<
 *         self.expr = self.expr().leaveWhitespace()
 *         self.mayReturnEmpty = True
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_PrecededBy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_expr) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_expr);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4355
 *     def __init__(self, expr, retreat=None):
 *         super(PrecededBy, self).__init__(expr)
 *         self.expr = self.expr().leaveWhitespace()             # <<<<<<<<<<<<<<
 *         self.mayReturnEmpty = True
 *         self.mayIndexError = False
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4355, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4355, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_leaveWhitespace); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4355, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4355, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_expr, __pyx_t_1) < 0) __PYX_ERR(0, 4355, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4356
 *         super(PrecededBy, self).__init__(expr)
 *         self.expr = self.expr().leaveWhitespace()
 *         self.mayReturnEmpty = True             # <<<<<<<<<<<<<<
 *         self.mayIndexError = False
 *         self.exact = False
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_mayReturnEmpty, Py_True) < 0) __PYX_ERR(0, 4356, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4357
 *         self.expr = self.expr().leaveWhitespace()
 *         self.mayReturnEmpty = True
 *         self.mayIndexError = False             # <<<<<<<<<<<<<<
 *         self.exact = False
 *         if isinstance(expr, str):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_mayIndexError, Py_False) < 0) __PYX_ERR(0, 4357, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4358
 *         self.mayReturnEmpty = True
 *         self.mayIndexError = False
 *         self.exact = False             # <<<<<<<<<<<<<<
 *         if isinstance(expr, str):
 *             retreat = len(expr)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_exact, Py_False) < 0) __PYX_ERR(0, 4358, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4359
 *         self.mayIndexError = False
 *         self.exact = False
 *         if isinstance(expr, str):             # <<<<<<<<<<<<<<
 *             retreat = len(expr)
 *             self.exact = True
 */
  __pyx_t_5 = PyString_Check(__pyx_v_expr); 
  __pyx_t_6 = (__pyx_t_5 != 0);
  if (__pyx_t_6) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4360
 *         self.exact = False
 *         if isinstance(expr, str):
 *             retreat = len(expr)             # <<<<<<<<<<<<<<
 *             self.exact = True
 *         elif isinstance(expr, (Literal, Keyword)):
 */
    __pyx_t_7 = PyObject_Length(__pyx_v_expr); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 4360, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4360, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_retreat, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4361
 *         if isinstance(expr, str):
 *             retreat = len(expr)
 *             self.exact = True             # <<<<<<<<<<<<<<
 *         elif isinstance(expr, (Literal, Keyword)):
 *             retreat = expr.matchLen
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_exact, Py_True) < 0) __PYX_ERR(0, 4361, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":4359
 *         self.mayIndexError = False
 *         self.exact = False
 *         if isinstance(expr, str):             # <<<<<<<<<<<<<<
 *             retreat = len(expr)
 *             self.exact = True
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4362
 *             retreat = len(expr)
 *             self.exact = True
 *         elif isinstance(expr, (Literal, Keyword)):             # <<<<<<<<<<<<<<
 *             retreat = expr.matchLen
 *             self.exact = True
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Literal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4362, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Keyword); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4362, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyObject_IsInstance(__pyx_v_expr, __pyx_t_1); 
  __pyx_t_8 = (__pyx_t_5 != 0);
  if (!__pyx_t_8) {
  } else {
    __pyx_t_6 = __pyx_t_8;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_8 = PyObject_IsInstance(__pyx_v_expr, __pyx_t_2); 
  __pyx_t_5 = (__pyx_t_8 != 0);
  __pyx_t_6 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = (__pyx_t_6 != 0);
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4363
 *             self.exact = True
 *         elif isinstance(expr, (Literal, Keyword)):
 *             retreat = expr.matchLen             # <<<<<<<<<<<<<<
 *             self.exact = True
 *         elif isinstance(expr, (Word, CharsNotIn)) and expr.maxLen != sys.maxsize:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_expr, __pyx_n_s_matchLen); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4363, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_retreat, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4364
 *         elif isinstance(expr, (Literal, Keyword)):
 *             retreat = expr.matchLen
 *             self.exact = True             # <<<<<<<<<<<<<<
 *         elif isinstance(expr, (Word, CharsNotIn)) and expr.maxLen != sys.maxsize:
 *             retreat = expr.maxLen
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_exact, Py_True) < 0) __PYX_ERR(0, 4364, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":4362
 *             retreat = len(expr)
 *             self.exact = True
 *         elif isinstance(expr, (Literal, Keyword)):             # <<<<<<<<<<<<<<
 *             retreat = expr.matchLen
 *             self.exact = True
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4365
 *             retreat = expr.matchLen
 *             self.exact = True
 *         elif isinstance(expr, (Word, CharsNotIn)) and expr.maxLen != sys.maxsize:             # <<<<<<<<<<<<<<
 *             retreat = expr.maxLen
 *             self.exact = True
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Word); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4365, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_CharsNotIn); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4365, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = PyObject_IsInstance(__pyx_v_expr, __pyx_t_1); 
  __pyx_t_9 = (__pyx_t_8 != 0);
  if (!__pyx_t_9) {
  } else {
    __pyx_t_6 = __pyx_t_9;
    goto __pyx_L8_bool_binop_done;
  }
  __pyx_t_9 = PyObject_IsInstance(__pyx_v_expr, __pyx_t_2); 
  __pyx_t_8 = (__pyx_t_9 != 0);
  __pyx_t_6 = __pyx_t_8;
  __pyx_L8_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_8 = (__pyx_t_6 != 0);
  if (__pyx_t_8) {
  } else {
    __pyx_t_5 = __pyx_t_8;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_expr, __pyx_n_s_maxLen); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4365, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_sys); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4365, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_maxsize); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4365, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4365, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 4365, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __pyx_t_8;
  __pyx_L6_bool_binop_done:;
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4366
 *             self.exact = True
 *         elif isinstance(expr, (Word, CharsNotIn)) and expr.maxLen != sys.maxsize:
 *             retreat = expr.maxLen             # <<<<<<<<<<<<<<
 *             self.exact = True
 *         elif isinstance(expr, _PositionToken):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_expr, __pyx_n_s_maxLen); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4366, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_retreat, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4367
 *         elif isinstance(expr, (Word, CharsNotIn)) and expr.maxLen != sys.maxsize:
 *             retreat = expr.maxLen
 *             self.exact = True             # <<<<<<<<<<<<<<
 *         elif isinstance(expr, _PositionToken):
 *             retreat = 0
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_exact, Py_True) < 0) __PYX_ERR(0, 4367, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":4365
 *             retreat = expr.matchLen
 *             self.exact = True
 *         elif isinstance(expr, (Word, CharsNotIn)) and expr.maxLen != sys.maxsize:             # <<<<<<<<<<<<<<
 *             retreat = expr.maxLen
 *             self.exact = True
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4368
 *             retreat = expr.maxLen
 *             self.exact = True
 *         elif isinstance(expr, _PositionToken):             # <<<<<<<<<<<<<<
 *             retreat = 0
 *             self.exact = True
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_PositionToken); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyObject_IsInstance(__pyx_v_expr, __pyx_t_2); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 4368, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_8 = (__pyx_t_5 != 0);
  if (__pyx_t_8) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4369
 *             self.exact = True
 *         elif isinstance(expr, _PositionToken):
 *             retreat = 0             # <<<<<<<<<<<<<<
 *             self.exact = True
 *         self.retreat = retreat
 */
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_DECREF_SET(__pyx_v_retreat, __pyx_int_0);

    /* "Aspidites/_vendor/pyparsing_extension.py":4370
 *         elif isinstance(expr, _PositionToken):
 *             retreat = 0
 *             self.exact = True             # <<<<<<<<<<<<<<
 *         self.retreat = retreat
 *         self.errmsg = "not preceded by " + str(expr)
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_exact, Py_True) < 0) __PYX_ERR(0, 4370, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":4368
 *             retreat = expr.maxLen
 *             self.exact = True
 *         elif isinstance(expr, _PositionToken):             # <<<<<<<<<<<<<<
 *             retreat = 0
 *             self.exact = True
 */
  }
  __pyx_L3:;

  /* "Aspidites/_vendor/pyparsing_extension.py":4371
 *             retreat = 0
 *             self.exact = True
 *         self.retreat = retreat             # <<<<<<<<<<<<<<
 *         self.errmsg = "not preceded by " + str(expr)
 *         self.skipWhitespace = False
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_retreat, __pyx_v_retreat) < 0) __PYX_ERR(0, 4371, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4372
 *             self.exact = True
 *         self.retreat = retreat
 *         self.errmsg = "not preceded by " + str(expr)             # <<<<<<<<<<<<<<
 *         self.skipWhitespace = False
 *         self.parseAction.append(lambda s, l, t: t.__delitem__(slice(None, None)))
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_v_expr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Add(__pyx_kp_s_not_preceded_by, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_errmsg, __pyx_t_3) < 0) __PYX_ERR(0, 4372, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4373
 *         self.retreat = retreat
 *         self.errmsg = "not preceded by " + str(expr)
 *         self.skipWhitespace = False             # <<<<<<<<<<<<<<
 *         self.parseAction.append(lambda s, l, t: t.__delitem__(slice(None, None)))
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_skipWhitespace, Py_False) < 0) __PYX_ERR(0, 4373, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4374
 *         self.errmsg = "not preceded by " + str(expr)
 *         self.skipWhitespace = False
 *         self.parseAction.append(lambda s, l, t: t.__delitem__(slice(None, None)))             # <<<<<<<<<<<<<<
 * 
 *     def parseImpl(self, instring, loc=0, doActions=True):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_parseAction); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10PrecededBy_8__init___lambda37, 0, __pyx_n_s_PrecededBy___init___locals_lambd, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_10 = __Pyx_PyObject_Append(__pyx_t_3, __pyx_t_2); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 4374, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4353
 *     """
 * 
 *     def __init__(self, expr, retreat=None):             # <<<<<<<<<<<<<<
 *         super(PrecededBy, self).__init__(expr)
 *         self.expr = self.expr().leaveWhitespace()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.PrecededBy.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_retreat);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4376
 *         self.parseAction.append(lambda s, l, t: t.__delitem__(slice(None, None)))
 * 
 *     def parseImpl(self, instring, loc=0, doActions=True):             # <<<<<<<<<<<<<<
 *         if self.exact:
 *             if loc < self.retreat:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10PrecededBy_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10PrecededBy_3parseImpl = {"parseImpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10PrecededBy_3parseImpl, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10PrecededBy_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  CYTHON_UNUSED PyObject *__pyx_v_doActions = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parseImpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_doActions,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)__pyx_int_0));
    values[3] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 2, 4, 1); __PYX_ERR(0, 4376, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_doActions);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "parseImpl") < 0)) __PYX_ERR(0, 4376, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_doActions = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4376, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.PrecededBy.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10PrecededBy_2parseImpl(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10PrecededBy_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions) {
  PyObject *__pyx_v_start = NULL;
  CYTHON_UNUSED PyObject *__pyx_v__ = NULL;
  PyObject *__pyx_v_ret = NULL;
  PyObject *__pyx_v_test_expr = NULL;
  PyObject *__pyx_v_instring_slice = NULL;
  PyObject *__pyx_v_last_expr = NULL;
  PyObject *__pyx_v_offset = NULL;
  PyObject *__pyx_v_pbe = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  long __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  PyObject *(*__pyx_t_11)(PyObject *);
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  Py_ssize_t __pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parseImpl", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":4377
 * 
 *     def parseImpl(self, instring, loc=0, doActions=True):
 *         if self.exact:             # <<<<<<<<<<<<<<
 *             if loc < self.retreat:
 *                 raise ParseException(instring, loc, self.errmsg)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_exact); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 4377, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4378
 *     def parseImpl(self, instring, loc=0, doActions=True):
 *         if self.exact:
 *             if loc < self.retreat:             # <<<<<<<<<<<<<<
 *                 raise ParseException(instring, loc, self.errmsg)
 *             start = loc - self.retreat
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_retreat); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4378, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyObject_RichCompare(__pyx_v_loc, __pyx_t_1, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4378, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 4378, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(__pyx_t_2)) {

      /* "Aspidites/_vendor/pyparsing_extension.py":4379
 *         if self.exact:
 *             if loc < self.retreat:
 *                 raise ParseException(instring, loc, self.errmsg)             # <<<<<<<<<<<<<<
 *             start = loc - self.retreat
 *             _, ret = self.expr._parse(instring, start)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4379, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errmsg); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4379, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      __pyx_t_6 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_6 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_instring, __pyx_v_loc, __pyx_t_4};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4379, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_instring, __pyx_v_loc, __pyx_t_4};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4379, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4379, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        if (__pyx_t_5) {
          __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        }
        __Pyx_INCREF(__pyx_v_instring);
        __Pyx_GIVEREF(__pyx_v_instring);
        PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_instring);
        __Pyx_INCREF(__pyx_v_loc);
        __Pyx_GIVEREF(__pyx_v_loc);
        PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_loc);
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
        __pyx_t_4 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4379, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 4379, __pyx_L1_error)

      /* "Aspidites/_vendor/pyparsing_extension.py":4378
 *     def parseImpl(self, instring, loc=0, doActions=True):
 *         if self.exact:
 *             if loc < self.retreat:             # <<<<<<<<<<<<<<
 *                 raise ParseException(instring, loc, self.errmsg)
 *             start = loc - self.retreat
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":4380
 *             if loc < self.retreat:
 *                 raise ParseException(instring, loc, self.errmsg)
 *             start = loc - self.retreat             # <<<<<<<<<<<<<<
 *             _, ret = self.expr._parse(instring, start)
 *         else:
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_retreat); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4380, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyNumber_Subtract(__pyx_v_loc, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4380, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_start = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4381
 *                 raise ParseException(instring, loc, self.errmsg)
 *             start = loc - self.retreat
 *             _, ret = self.expr._parse(instring, start)             # <<<<<<<<<<<<<<
 *         else:
 *             # retreat specified a maximum lookbehind window, iterate
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4381, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_parse); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4381, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_7)) {
      PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_instring, __pyx_v_start};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4381, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
      PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_instring, __pyx_v_start};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4381, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4381, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (__pyx_t_3) {
        __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __pyx_t_3 = NULL;
      }
      __Pyx_INCREF(__pyx_v_instring);
      __Pyx_GIVEREF(__pyx_v_instring);
      PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_v_instring);
      __Pyx_INCREF(__pyx_v_start);
      __Pyx_GIVEREF(__pyx_v_start);
      PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_v_start);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4381, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
      PyObject* sequence = __pyx_t_1;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 4381, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_7 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_7 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_7 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4381, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4381, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4381, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_8 = Py_TYPE(__pyx_t_3)->tp_iternext;
      index = 0; __pyx_t_7 = __pyx_t_8(__pyx_t_3); if (unlikely(!__pyx_t_7)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_7);
      index = 1; __pyx_t_4 = __pyx_t_8(__pyx_t_3); if (unlikely(!__pyx_t_4)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_3), 2) < 0) __PYX_ERR(0, 4381, __pyx_L1_error)
      __pyx_t_8 = NULL;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_8 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 4381, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __pyx_v__ = __pyx_t_7;
    __pyx_t_7 = 0;
    __pyx_v_ret = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4377
 * 
 *     def parseImpl(self, instring, loc=0, doActions=True):
 *         if self.exact:             # <<<<<<<<<<<<<<
 *             if loc < self.retreat:
 *                 raise ParseException(instring, loc, self.errmsg)
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4384
 *         else:
 *             # retreat specified a maximum lookbehind window, iterate
 *             test_expr = self.expr + StringEnd()             # <<<<<<<<<<<<<<
 *             instring_slice = instring[max(0, loc - self.retreat):loc]
 *             last_expr = ParseException(instring, loc, self.errmsg)
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4384, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_StringEnd); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4384, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_4 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4384, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = PyNumber_Add(__pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4384, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_test_expr = __pyx_t_7;
    __pyx_t_7 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4385
 *             # retreat specified a maximum lookbehind window, iterate
 *             test_expr = self.expr + StringEnd()
 *             instring_slice = instring[max(0, loc - self.retreat):loc]             # <<<<<<<<<<<<<<
 *             last_expr = ParseException(instring, loc, self.errmsg)
 *             for offset in range(1, min(loc, self.retreat + 1) + 1):
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_retreat); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4385, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_4 = PyNumber_Subtract(__pyx_v_loc, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4385, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_9 = 0;
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4385, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyObject_RichCompare(__pyx_t_4, __pyx_t_1, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4385, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 4385, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_2) {
      __Pyx_INCREF(__pyx_t_4);
      __pyx_t_7 = __pyx_t_4;
    } else {
      __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_t_9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4385, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = __pyx_t_3;
      __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetSlice(__pyx_v_instring, 0, 0, &__pyx_t_7, &__pyx_v_loc, NULL, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4385, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_v_instring_slice = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4386
 *             test_expr = self.expr + StringEnd()
 *             instring_slice = instring[max(0, loc - self.retreat):loc]
 *             last_expr = ParseException(instring, loc, self.errmsg)             # <<<<<<<<<<<<<<
 *             for offset in range(1, min(loc, self.retreat + 1) + 1):
 *                 try:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4386, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errmsg); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4386, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_7)) {
      PyObject *__pyx_temp[4] = {__pyx_t_1, __pyx_v_instring, __pyx_v_loc, __pyx_t_3};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4386, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
      PyObject *__pyx_temp[4] = {__pyx_t_1, __pyx_v_instring, __pyx_v_loc, __pyx_t_3};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4386, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4386, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_1) {
        __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1); __pyx_t_1 = NULL;
      }
      __Pyx_INCREF(__pyx_v_instring);
      __Pyx_GIVEREF(__pyx_v_instring);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_v_instring);
      __Pyx_INCREF(__pyx_v_loc);
      __Pyx_GIVEREF(__pyx_v_loc);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_v_loc);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_6, __pyx_t_3);
      __pyx_t_3 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4386, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_v_last_expr = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4387
 *             instring_slice = instring[max(0, loc - self.retreat):loc]
 *             last_expr = ParseException(instring, loc, self.errmsg)
 *             for offset in range(1, min(loc, self.retreat + 1) + 1):             # <<<<<<<<<<<<<<
 *                 try:
 *                     # print('trying', offset, instring_slice, repr(instring_slice[loc - offset:]))
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_retreat); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4387, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_PyInt_AddObjC(__pyx_t_4, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4387, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_INCREF(__pyx_v_loc);
    __pyx_t_4 = __pyx_v_loc;
    __pyx_t_3 = PyObject_RichCompare(__pyx_t_7, __pyx_t_4, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4387, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 4387, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_2) {
      __Pyx_INCREF(__pyx_t_7);
      __pyx_t_5 = __pyx_t_7;
    } else {
      __Pyx_INCREF(__pyx_t_4);
      __pyx_t_5 = __pyx_t_4;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyInt_AddObjC(__pyx_t_5, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4387, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4387, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_int_1);
    __Pyx_GIVEREF(__pyx_int_1);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_int_1);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_5, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4387, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (likely(PyList_CheckExact(__pyx_t_7)) || PyTuple_CheckExact(__pyx_t_7)) {
      __pyx_t_5 = __pyx_t_7; __Pyx_INCREF(__pyx_t_5); __pyx_t_10 = 0;
      __pyx_t_11 = NULL;
    } else {
      __pyx_t_10 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4387, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_11 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 4387, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    for (;;) {
      if (likely(!__pyx_t_11)) {
        if (likely(PyList_CheckExact(__pyx_t_5))) {
          if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_5)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_7 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_10); __Pyx_INCREF(__pyx_t_7); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 4387, __pyx_L1_error)
          #else
          __pyx_t_7 = PySequence_ITEM(__pyx_t_5, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4387, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          #endif
        } else {
          if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_10); __Pyx_INCREF(__pyx_t_7); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 4387, __pyx_L1_error)
          #else
          __pyx_t_7 = PySequence_ITEM(__pyx_t_5, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4387, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          #endif
        }
      } else {
        __pyx_t_7 = __pyx_t_11(__pyx_t_5);
        if (unlikely(!__pyx_t_7)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 4387, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_7);
      }
      __Pyx_XDECREF_SET(__pyx_v_offset, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":4388
 *             last_expr = ParseException(instring, loc, self.errmsg)
 *             for offset in range(1, min(loc, self.retreat + 1) + 1):
 *                 try:             # <<<<<<<<<<<<<<
 *                     # print('trying', offset, instring_slice, repr(instring_slice[loc - offset:]))
 *                     _, ret = test_expr._parse(instring_slice, len(instring_slice) - offset)
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
        __Pyx_XGOTREF(__pyx_t_12);
        __Pyx_XGOTREF(__pyx_t_13);
        __Pyx_XGOTREF(__pyx_t_14);
        /*try:*/ {

          /* "Aspidites/_vendor/pyparsing_extension.py":4390
 *                 try:
 *                     # print('trying', offset, instring_slice, repr(instring_slice[loc - offset:]))
 *                     _, ret = test_expr._parse(instring_slice, len(instring_slice) - offset)             # <<<<<<<<<<<<<<
 *                 except ParseBaseException as pbe:
 *                     last_expr = pbe
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_test_expr, __pyx_n_s_parse); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4390, __pyx_L9_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_15 = PyObject_Length(__pyx_v_instring_slice); if (unlikely(__pyx_t_15 == ((Py_ssize_t)-1))) __PYX_ERR(0, 4390, __pyx_L9_error)
          __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_15); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4390, __pyx_L9_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = PyNumber_Subtract(__pyx_t_3, __pyx_v_offset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4390, __pyx_L9_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_instring_slice, __pyx_t_1};
            __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4390, __pyx_L9_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
            PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_instring_slice, __pyx_t_1};
            __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4390, __pyx_L9_error)
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          } else
          #endif
          {
            __pyx_t_16 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 4390, __pyx_L9_error)
            __Pyx_GOTREF(__pyx_t_16);
            if (__pyx_t_3) {
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_t_3); __pyx_t_3 = NULL;
            }
            __Pyx_INCREF(__pyx_v_instring_slice);
            __Pyx_GIVEREF(__pyx_v_instring_slice);
            PyTuple_SET_ITEM(__pyx_t_16, 0+__pyx_t_6, __pyx_v_instring_slice);
            __Pyx_GIVEREF(__pyx_t_1);
            PyTuple_SET_ITEM(__pyx_t_16, 1+__pyx_t_6, __pyx_t_1);
            __pyx_t_1 = 0;
            __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_16, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4390, __pyx_L9_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if ((likely(PyTuple_CheckExact(__pyx_t_7))) || (PyList_CheckExact(__pyx_t_7))) {
            PyObject* sequence = __pyx_t_7;
            Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
            if (unlikely(size != 2)) {
              if (size > 2) __Pyx_RaiseTooManyValuesError(2);
              else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
              __PYX_ERR(0, 4390, __pyx_L9_error)
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            if (likely(PyTuple_CheckExact(sequence))) {
              __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
              __pyx_t_16 = PyTuple_GET_ITEM(sequence, 1); 
            } else {
              __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
              __pyx_t_16 = PyList_GET_ITEM(sequence, 1); 
            }
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_16);
            #else
            __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4390, __pyx_L9_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_16 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 4390, __pyx_L9_error)
            __Pyx_GOTREF(__pyx_t_16);
            #endif
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          } else {
            Py_ssize_t index = -1;
            __pyx_t_1 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4390, __pyx_L9_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __pyx_t_8 = Py_TYPE(__pyx_t_1)->tp_iternext;
            index = 0; __pyx_t_4 = __pyx_t_8(__pyx_t_1); if (unlikely(!__pyx_t_4)) goto __pyx_L17_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_4);
            index = 1; __pyx_t_16 = __pyx_t_8(__pyx_t_1); if (unlikely(!__pyx_t_16)) goto __pyx_L17_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_16);
            if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_1), 2) < 0) __PYX_ERR(0, 4390, __pyx_L9_error)
            __pyx_t_8 = NULL;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            goto __pyx_L18_unpacking_done;
            __pyx_L17_unpacking_failed:;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_8 = NULL;
            if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
            __PYX_ERR(0, 4390, __pyx_L9_error)
            __pyx_L18_unpacking_done:;
          }
          __Pyx_XDECREF_SET(__pyx_v__, __pyx_t_4);
          __pyx_t_4 = 0;
          __Pyx_XDECREF_SET(__pyx_v_ret, __pyx_t_16);
          __pyx_t_16 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":4388
 *             last_expr = ParseException(instring, loc, self.errmsg)
 *             for offset in range(1, min(loc, self.retreat + 1) + 1):
 *                 try:             # <<<<<<<<<<<<<<
 *                     # print('trying', offset, instring_slice, repr(instring_slice[loc - offset:]))
 *                     _, ret = test_expr._parse(instring_slice, len(instring_slice) - offset)
 */
        }

        /* "Aspidites/_vendor/pyparsing_extension.py":4394
 *                     last_expr = pbe
 *                 else:
 *                     break             # <<<<<<<<<<<<<<
 *             else:
 *                 raise last_expr
 */
        /*else:*/ {
          goto __pyx_L14_try_break;
        }
        __pyx_L9_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":4391
 *                     # print('trying', offset, instring_slice, repr(instring_slice[loc - offset:]))
 *                     _, ret = test_expr._parse(instring_slice, len(instring_slice) - offset)
 *                 except ParseBaseException as pbe:             # <<<<<<<<<<<<<<
 *                     last_expr = pbe
 *                 else:
 */
        __Pyx_ErrFetch(&__pyx_t_7, &__pyx_t_16, &__pyx_t_4);
        __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParseBaseException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4391, __pyx_L11_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_6 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_7, __pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_ErrRestore(__pyx_t_7, __pyx_t_16, __pyx_t_4);
        __pyx_t_7 = 0; __pyx_t_16 = 0; __pyx_t_4 = 0;
        if (__pyx_t_6) {
          __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.PrecededBy.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_16, &__pyx_t_7) < 0) __PYX_ERR(0, 4391, __pyx_L11_except_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_GOTREF(__pyx_t_16);
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_16);
          __Pyx_XDECREF_SET(__pyx_v_pbe, __pyx_t_16);

          /* "Aspidites/_vendor/pyparsing_extension.py":4392
 *                     _, ret = test_expr._parse(instring_slice, len(instring_slice) - offset)
 *                 except ParseBaseException as pbe:
 *                     last_expr = pbe             # <<<<<<<<<<<<<<
 *                 else:
 *                     break
 */
          __Pyx_INCREF(__pyx_v_pbe);
          __Pyx_DECREF_SET(__pyx_v_last_expr, __pyx_v_pbe);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          goto __pyx_L10_exception_handled;
        }
        goto __pyx_L11_except_error;
        __pyx_L11_except_error:;

        /* "Aspidites/_vendor/pyparsing_extension.py":4388
 *             last_expr = ParseException(instring, loc, self.errmsg)
 *             for offset in range(1, min(loc, self.retreat + 1) + 1):
 *                 try:             # <<<<<<<<<<<<<<
 *                     # print('trying', offset, instring_slice, repr(instring_slice[loc - offset:]))
 *                     _, ret = test_expr._parse(instring_slice, len(instring_slice) - offset)
 */
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
        goto __pyx_L1_error;
        __pyx_L14_try_break:;
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
        goto __pyx_L8_break;
        __pyx_L10_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":4387
 *             instring_slice = instring[max(0, loc - self.retreat):loc]
 *             last_expr = ParseException(instring, loc, self.errmsg)
 *             for offset in range(1, min(loc, self.retreat + 1) + 1):             # <<<<<<<<<<<<<<
 *                 try:
 *                     # print('trying', offset, instring_slice, repr(instring_slice[loc - offset:]))
 */
    }
    /*else*/ {

      /* "Aspidites/_vendor/pyparsing_extension.py":4396
 *                     break
 *             else:
 *                 raise last_expr             # <<<<<<<<<<<<<<
 *         return loc, ret
 * 
 */
      __Pyx_Raise(__pyx_v_last_expr, 0, 0, 0);
      __PYX_ERR(0, 4396, __pyx_L1_error)
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":4387
 *             instring_slice = instring[max(0, loc - self.retreat):loc]
 *             last_expr = ParseException(instring, loc, self.errmsg)
 *             for offset in range(1, min(loc, self.retreat + 1) + 1):             # <<<<<<<<<<<<<<
 *                 try:
 *                     # print('trying', offset, instring_slice, repr(instring_slice[loc - offset:]))
 */
    __pyx_L8_break:;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_L3:;

  /* "Aspidites/_vendor/pyparsing_extension.py":4397
 *             else:
 *                 raise last_expr
 *         return loc, ret             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_loc);
  __Pyx_GIVEREF(__pyx_v_loc);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_loc);
  __Pyx_INCREF(__pyx_v_ret);
  __Pyx_GIVEREF(__pyx_v_ret);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_ret);
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4376
 *         self.parseAction.append(lambda s, l, t: t.__delitem__(slice(None, None)))
 * 
 *     def parseImpl(self, instring, loc=0, doActions=True):             # <<<<<<<<<<<<<<
 *         if self.exact:
 *             if loc < self.retreat:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.PrecededBy.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_start);
  __Pyx_XDECREF(__pyx_v__);
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_XDECREF(__pyx_v_test_expr);
  __Pyx_XDECREF(__pyx_v_instring_slice);
  __Pyx_XDECREF(__pyx_v_last_expr);
  __Pyx_XDECREF(__pyx_v_offset);
  __Pyx_XDECREF(__pyx_v_pbe);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4424
 *     """
 * 
 *     def __init__(self, expr):             # <<<<<<<<<<<<<<
 *         super(NotAny, self).__init__(expr)
 *         # ~ self.leaveWhitespace()
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_6NotAny_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_6NotAny_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_6NotAny_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_6NotAny_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_expr = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_expr,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_expr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(0, 4424, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 4424, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_expr = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4424, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.NotAny.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_6NotAny___init__(__pyx_self, __pyx_v_self, __pyx_v_expr);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_6NotAny___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_expr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":4425
 * 
 *     def __init__(self, expr):
 *         super(NotAny, self).__init__(expr)             # <<<<<<<<<<<<<<
 *         # ~ self.leaveWhitespace()
 *         self.skipWhitespace = False  # do NOT use self.leaveWhitespace(), don't want to propagate to exprs
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_NotAny); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_expr) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_expr);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4427
 *         super(NotAny, self).__init__(expr)
 *         # ~ self.leaveWhitespace()
 *         self.skipWhitespace = False  # do NOT use self.leaveWhitespace(), don't want to propagate to exprs             # <<<<<<<<<<<<<<
 *         self.mayReturnEmpty = True
 *         self.errmsg = "Found unwanted token, " + str(self.expr)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_skipWhitespace, Py_False) < 0) __PYX_ERR(0, 4427, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4428
 *         # ~ self.leaveWhitespace()
 *         self.skipWhitespace = False  # do NOT use self.leaveWhitespace(), don't want to propagate to exprs
 *         self.mayReturnEmpty = True             # <<<<<<<<<<<<<<
 *         self.errmsg = "Found unwanted token, " + str(self.expr)
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_mayReturnEmpty, Py_True) < 0) __PYX_ERR(0, 4428, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4429
 *         self.skipWhitespace = False  # do NOT use self.leaveWhitespace(), don't want to propagate to exprs
 *         self.mayReturnEmpty = True
 *         self.errmsg = "Found unwanted token, " + str(self.expr)             # <<<<<<<<<<<<<<
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Add(__pyx_kp_s_Found_unwanted_token, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_errmsg, __pyx_t_1) < 0) __PYX_ERR(0, 4429, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4424
 *     """
 * 
 *     def __init__(self, expr):             # <<<<<<<<<<<<<<
 *         super(NotAny, self).__init__(expr)
 *         # ~ self.leaveWhitespace()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.NotAny.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4431
 *         self.errmsg = "Found unwanted token, " + str(self.expr)
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if self.expr.canParseNext(instring, loc):
 *             raise ParseException(instring, loc, self.errmsg, self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_6NotAny_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_6NotAny_3parseImpl = {"parseImpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_6NotAny_3parseImpl, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_6NotAny_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  CYTHON_UNUSED PyObject *__pyx_v_doActions = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parseImpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_doActions,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 1); __PYX_ERR(0, 4431, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 2); __PYX_ERR(0, 4431, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_doActions);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "parseImpl") < 0)) __PYX_ERR(0, 4431, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_doActions = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4431, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.NotAny.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_6NotAny_2parseImpl(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_6NotAny_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_doActions) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parseImpl", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":4432
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if self.expr.canParseNext(instring, loc):             # <<<<<<<<<<<<<<
 *             raise ParseException(instring, loc, self.errmsg, self)
 *         return loc, []
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4432, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_canParseNext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4432, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_instring, __pyx_v_loc};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4432, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_instring, __pyx_v_loc};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4432, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4432, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_instring);
    __Pyx_GIVEREF(__pyx_v_instring);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_instring);
    __Pyx_INCREF(__pyx_v_loc);
    __Pyx_GIVEREF(__pyx_v_loc);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_loc);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4432, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 4432, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_6)) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4433
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if self.expr.canParseNext(instring, loc):
 *             raise ParseException(instring, loc, self.errmsg, self)             # <<<<<<<<<<<<<<
 *         return loc, []
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4433, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errmsg); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4433, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = NULL;
    __pyx_t_4 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_4 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[5] = {__pyx_t_2, __pyx_v_instring, __pyx_v_loc, __pyx_t_5, __pyx_v_self};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 4+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4433, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[5] = {__pyx_t_2, __pyx_v_instring, __pyx_v_loc, __pyx_t_5, __pyx_v_self};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 4+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4433, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    {
      __pyx_t_7 = PyTuple_New(4+__pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4433, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (__pyx_t_2) {
        __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2); __pyx_t_2 = NULL;
      }
      __Pyx_INCREF(__pyx_v_instring);
      __Pyx_GIVEREF(__pyx_v_instring);
      PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_4, __pyx_v_instring);
      __Pyx_INCREF(__pyx_v_loc);
      __Pyx_GIVEREF(__pyx_v_loc);
      PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_4, __pyx_v_loc);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_4, __pyx_t_5);
      __Pyx_INCREF(__pyx_v_self);
      __Pyx_GIVEREF(__pyx_v_self);
      PyTuple_SET_ITEM(__pyx_t_7, 3+__pyx_t_4, __pyx_v_self);
      __pyx_t_5 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4433, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 4433, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":4432
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if self.expr.canParseNext(instring, loc):             # <<<<<<<<<<<<<<
 *             raise ParseException(instring, loc, self.errmsg, self)
 *         return loc, []
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4434
 *         if self.expr.canParseNext(instring, loc):
 *             raise ParseException(instring, loc, self.errmsg, self)
 *         return loc, []             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_loc);
  __Pyx_GIVEREF(__pyx_v_loc);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_loc);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4431
 *         self.errmsg = "Found unwanted token, " + str(self.expr)
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if self.expr.canParseNext(instring, loc):
 *             raise ParseException(instring, loc, self.errmsg, self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.NotAny.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4436
 *         return loc, []
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_6NotAny_5__str__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_6NotAny_5__str__ = {"__str__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_6NotAny_5__str__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_6NotAny_5__str__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_6NotAny_4__str__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_6NotAny_4__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":4437
 * 
 *     def __str__(self):
 *         if hasattr(self, "name"):             # <<<<<<<<<<<<<<
 *             return self.name
 * 
 */
  __pyx_t_1 = __Pyx_HasAttr(__pyx_v_self, __pyx_n_s_name); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4437, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4438
 *     def __str__(self):
 *         if hasattr(self, "name"):
 *             return self.name             # <<<<<<<<<<<<<<
 * 
 *         if self.strRepr is None:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4438, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4437
 * 
 *     def __str__(self):
 *         if hasattr(self, "name"):             # <<<<<<<<<<<<<<
 *             return self.name
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4440
 *             return self.name
 * 
 *         if self.strRepr is None:             # <<<<<<<<<<<<<<
 *             self.strRepr = "~{" + str(self.expr) + "}"
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_strRepr); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = (__pyx_t_3 == Py_None);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4441
 * 
 *         if self.strRepr is None:
 *             self.strRepr = "~{" + str(self.expr) + "}"             # <<<<<<<<<<<<<<
 * 
 *         return self.strRepr
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4441, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4441, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Add(__pyx_kp_s__115, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4441, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyNumber_Add(__pyx_t_3, __pyx_kp_s__110); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4441, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_strRepr, __pyx_t_4) < 0) __PYX_ERR(0, 4441, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4440
 *             return self.name
 * 
 *         if self.strRepr is None:             # <<<<<<<<<<<<<<
 *             self.strRepr = "~{" + str(self.expr) + "}"
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4443
 *             self.strRepr = "~{" + str(self.expr) + "}"
 * 
 *         return self.strRepr             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_strRepr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4443, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4436
 *         return loc, []
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.NotAny.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4447
 * 
 * class _MultipleMatch(ParseElementEnhance):
 *     def __init__(self, expr, stopOn=None):             # <<<<<<<<<<<<<<
 *         super(_MultipleMatch, self).__init__(expr)
 *         self.saveAsList = True
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_14_MultipleMatch_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_14_MultipleMatch_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_14_MultipleMatch_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_14_MultipleMatch_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_expr = 0;
  PyObject *__pyx_v_stopOn = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_expr,&__pyx_n_s_stopOn,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_expr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, 1); __PYX_ERR(0, 4447, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_stopOn);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 4447, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_expr = values[1];
    __pyx_v_stopOn = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4447, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._MultipleMatch.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_14_MultipleMatch___init__(__pyx_self, __pyx_v_self, __pyx_v_expr, __pyx_v_stopOn);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_14_MultipleMatch___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_expr, PyObject *__pyx_v_stopOn) {
  PyObject *__pyx_v_ender = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":4448
 * class _MultipleMatch(ParseElementEnhance):
 *     def __init__(self, expr, stopOn=None):
 *         super(_MultipleMatch, self).__init__(expr)             # <<<<<<<<<<<<<<
 *         self.saveAsList = True
 *         ender = stopOn
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_MultipleMatch); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_expr) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_expr);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4449
 *     def __init__(self, expr, stopOn=None):
 *         super(_MultipleMatch, self).__init__(expr)
 *         self.saveAsList = True             # <<<<<<<<<<<<<<
 *         ender = stopOn
 *         if isinstance(ender, str):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_saveAsList, Py_True) < 0) __PYX_ERR(0, 4449, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4450
 *         super(_MultipleMatch, self).__init__(expr)
 *         self.saveAsList = True
 *         ender = stopOn             # <<<<<<<<<<<<<<
 *         if isinstance(ender, str):
 *             ender = self._literalStringClass(ender)
 */
  __Pyx_INCREF(__pyx_v_stopOn);
  __pyx_v_ender = __pyx_v_stopOn;

  /* "Aspidites/_vendor/pyparsing_extension.py":4451
 *         self.saveAsList = True
 *         ender = stopOn
 *         if isinstance(ender, str):             # <<<<<<<<<<<<<<
 *             ender = self._literalStringClass(ender)
 *         self.stopOn(ender)
 */
  __pyx_t_4 = PyString_Check(__pyx_v_ender); 
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4452
 *         ender = stopOn
 *         if isinstance(ender, str):
 *             ender = self._literalStringClass(ender)             # <<<<<<<<<<<<<<
 *         self.stopOn(ender)
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_literalStringClass); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4452, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_ender) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ender);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4452, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_ender, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4451
 *         self.saveAsList = True
 *         ender = stopOn
 *         if isinstance(ender, str):             # <<<<<<<<<<<<<<
 *             ender = self._literalStringClass(ender)
 *         self.stopOn(ender)
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4453
 *         if isinstance(ender, str):
 *             ender = self._literalStringClass(ender)
 *         self.stopOn(ender)             # <<<<<<<<<<<<<<
 * 
 *     def stopOn(self, ender):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_stopOn); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_ender) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_ender);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4447
 * 
 * class _MultipleMatch(ParseElementEnhance):
 *     def __init__(self, expr, stopOn=None):             # <<<<<<<<<<<<<<
 *         super(_MultipleMatch, self).__init__(expr)
 *         self.saveAsList = True
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._MultipleMatch.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ender);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4455
 *         self.stopOn(ender)
 * 
 *     def stopOn(self, ender):             # <<<<<<<<<<<<<<
 *         if isinstance(ender, str):
 *             ender = self._literalStringClass(ender)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_14_MultipleMatch_3stopOn(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_14_MultipleMatch_3stopOn = {"stopOn", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_14_MultipleMatch_3stopOn, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_14_MultipleMatch_3stopOn(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_ender = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("stopOn (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ender,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ender)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("stopOn", 1, 2, 2, 1); __PYX_ERR(0, 4455, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "stopOn") < 0)) __PYX_ERR(0, 4455, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ender = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("stopOn", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4455, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._MultipleMatch.stopOn", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_14_MultipleMatch_2stopOn(__pyx_self, __pyx_v_self, __pyx_v_ender);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_14_MultipleMatch_2stopOn(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_ender) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("stopOn", 0);
  __Pyx_INCREF(__pyx_v_ender);

  /* "Aspidites/_vendor/pyparsing_extension.py":4456
 * 
 *     def stopOn(self, ender):
 *         if isinstance(ender, str):             # <<<<<<<<<<<<<<
 *             ender = self._literalStringClass(ender)
 *         self.not_ender = ~ender if ender is not None else None
 */
  __pyx_t_1 = PyString_Check(__pyx_v_ender); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4457
 *     def stopOn(self, ender):
 *         if isinstance(ender, str):
 *             ender = self._literalStringClass(ender)             # <<<<<<<<<<<<<<
 *         self.not_ender = ~ender if ender is not None else None
 *         return self
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_literalStringClass); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4457, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_ender) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_ender);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4457, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_ender, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4456
 * 
 *     def stopOn(self, ender):
 *         if isinstance(ender, str):             # <<<<<<<<<<<<<<
 *             ender = self._literalStringClass(ender)
 *         self.not_ender = ~ender if ender is not None else None
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4458
 *         if isinstance(ender, str):
 *             ender = self._literalStringClass(ender)
 *         self.not_ender = ~ender if ender is not None else None             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __pyx_t_2 = (__pyx_v_ender != Py_None);
  if ((__pyx_t_2 != 0)) {
    __pyx_t_4 = PyNumber_Invert(__pyx_v_ender); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4458, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
  } else {
    __Pyx_INCREF(Py_None);
    __pyx_t_3 = Py_None;
  }
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_not_ender, __pyx_t_3) < 0) __PYX_ERR(0, 4458, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4459
 *             ender = self._literalStringClass(ender)
 *         self.not_ender = ~ender if ender is not None else None
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4455
 *         self.stopOn(ender)
 * 
 *     def stopOn(self, ender):             # <<<<<<<<<<<<<<
 *         if isinstance(ender, str):
 *             ender = self._literalStringClass(ender)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._MultipleMatch.stopOn", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ender);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4461
 *         return self
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         self_expr_parse = self.expr._parse
 *         self_skip_ignorables = self._skipIgnorables
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_14_MultipleMatch_5parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_14_MultipleMatch_5parseImpl = {"parseImpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_14_MultipleMatch_5parseImpl, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_14_MultipleMatch_5parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  PyObject *__pyx_v_doActions = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parseImpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_doActions,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 1); __PYX_ERR(0, 4461, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 2); __PYX_ERR(0, 4461, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_doActions);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "parseImpl") < 0)) __PYX_ERR(0, 4461, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_doActions = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4461, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._MultipleMatch.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_14_MultipleMatch_4parseImpl(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_14_MultipleMatch_4parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, PyObject *__pyx_v_doActions) {
  PyObject *__pyx_v_self_expr_parse = NULL;
  PyObject *__pyx_v_self_skip_ignorables = NULL;
  PyObject *__pyx_v_check_ender = NULL;
  PyObject *__pyx_v_try_not_ender = NULL;
  PyObject *__pyx_v_tokens = NULL;
  int __pyx_v_hasIgnoreExprs;
  PyObject *__pyx_v_preloc = NULL;
  PyObject *__pyx_v_tmptokens = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parseImpl", 0);
  __Pyx_INCREF(__pyx_v_loc);

  /* "Aspidites/_vendor/pyparsing_extension.py":4462
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         self_expr_parse = self.expr._parse             # <<<<<<<<<<<<<<
 *         self_skip_ignorables = self._skipIgnorables
 *         check_ender = self.not_ender is not None
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_parse); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self_expr_parse = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4463
 *     def parseImpl(self, instring, loc, doActions=True):
 *         self_expr_parse = self.expr._parse
 *         self_skip_ignorables = self._skipIgnorables             # <<<<<<<<<<<<<<
 *         check_ender = self.not_ender is not None
 *         if check_ender:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_skipIgnorables); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4463, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_self_skip_ignorables = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4464
 *         self_expr_parse = self.expr._parse
 *         self_skip_ignorables = self._skipIgnorables
 *         check_ender = self.not_ender is not None             # <<<<<<<<<<<<<<
 *         if check_ender:
 *             try_not_ender = self.not_ender.tryParse
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_not_ender); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = (__pyx_t_2 != Py_None);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_check_ender = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4465
 *         self_skip_ignorables = self._skipIgnorables
 *         check_ender = self.not_ender is not None
 *         if check_ender:             # <<<<<<<<<<<<<<
 *             try_not_ender = self.not_ender.tryParse
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_check_ender); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 4465, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4466
 *         check_ender = self.not_ender is not None
 *         if check_ender:
 *             try_not_ender = self.not_ender.tryParse             # <<<<<<<<<<<<<<
 * 
 *         # must be at least one (but first see if we are the stopOn sentinel;
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_not_ender); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4466, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_tryParse); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4466, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_try_not_ender = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4465
 *         self_skip_ignorables = self._skipIgnorables
 *         check_ender = self.not_ender is not None
 *         if check_ender:             # <<<<<<<<<<<<<<
 *             try_not_ender = self.not_ender.tryParse
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4470
 *         # must be at least one (but first see if we are the stopOn sentinel;
 *         # if so, fail)
 *         if check_ender:             # <<<<<<<<<<<<<<
 *             try_not_ender(instring, loc)
 *         loc, tokens = self_expr_parse(instring, loc, doActions, callPreParse=False)
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_check_ender); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 4470, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4471
 *         # if so, fail)
 *         if check_ender:
 *             try_not_ender(instring, loc)             # <<<<<<<<<<<<<<
 *         loc, tokens = self_expr_parse(instring, loc, doActions, callPreParse=False)
 *         try:
 */
    if (unlikely(!__pyx_v_try_not_ender)) { __Pyx_RaiseUnboundLocalError("try_not_ender"); __PYX_ERR(0, 4471, __pyx_L1_error) }
    __Pyx_INCREF(__pyx_v_try_not_ender);
    __pyx_t_2 = __pyx_v_try_not_ender; __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_5 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_instring, __pyx_v_loc};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4471, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_instring, __pyx_v_loc};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4471, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4471, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_4) {
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
      }
      __Pyx_INCREF(__pyx_v_instring);
      __Pyx_GIVEREF(__pyx_v_instring);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_instring);
      __Pyx_INCREF(__pyx_v_loc);
      __Pyx_GIVEREF(__pyx_v_loc);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_loc);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4471, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4470
 *         # must be at least one (but first see if we are the stopOn sentinel;
 *         # if so, fail)
 *         if check_ender:             # <<<<<<<<<<<<<<
 *             try_not_ender(instring, loc)
 *         loc, tokens = self_expr_parse(instring, loc, doActions, callPreParse=False)
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4472
 *         if check_ender:
 *             try_not_ender(instring, loc)
 *         loc, tokens = self_expr_parse(instring, loc, doActions, callPreParse=False)             # <<<<<<<<<<<<<<
 *         try:
 *             hasIgnoreExprs = (not not self.ignoreExprs)
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4472, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_instring);
  __Pyx_GIVEREF(__pyx_v_instring);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_instring);
  __Pyx_INCREF(__pyx_v_loc);
  __Pyx_GIVEREF(__pyx_v_loc);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_loc);
  __Pyx_INCREF(__pyx_v_doActions);
  __Pyx_GIVEREF(__pyx_v_doActions);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_doActions);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4472, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_callPreParse, Py_False) < 0) __PYX_ERR(0, 4472, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_v_self_expr_parse, __pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4472, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_6))) || (PyList_CheckExact(__pyx_t_6))) {
    PyObject* sequence = __pyx_t_6;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 4472, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_1 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_1);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4472, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4472, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    #endif
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_4 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4472, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_7 = Py_TYPE(__pyx_t_4)->tp_iternext;
    index = 0; __pyx_t_2 = __pyx_t_7(__pyx_t_4); if (unlikely(!__pyx_t_2)) goto __pyx_L5_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_1 = __pyx_t_7(__pyx_t_4); if (unlikely(!__pyx_t_1)) goto __pyx_L5_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_1);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_4), 2) < 0) __PYX_ERR(0, 4472, __pyx_L1_error)
    __pyx_t_7 = NULL;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L6_unpacking_done;
    __pyx_L5_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_7 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 4472, __pyx_L1_error)
    __pyx_L6_unpacking_done:;
  }
  __Pyx_DECREF_SET(__pyx_v_loc, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_v_tokens = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4473
 *             try_not_ender(instring, loc)
 *         loc, tokens = self_expr_parse(instring, loc, doActions, callPreParse=False)
 *         try:             # <<<<<<<<<<<<<<
 *             hasIgnoreExprs = (not not self.ignoreExprs)
 *             while 1:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
    __Pyx_XGOTREF(__pyx_t_8);
    __Pyx_XGOTREF(__pyx_t_9);
    __Pyx_XGOTREF(__pyx_t_10);
    /*try:*/ {

      /* "Aspidites/_vendor/pyparsing_extension.py":4474
 *         loc, tokens = self_expr_parse(instring, loc, doActions, callPreParse=False)
 *         try:
 *             hasIgnoreExprs = (not not self.ignoreExprs)             # <<<<<<<<<<<<<<
 *             while 1:
 *                 if check_ender:
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ignoreExprs); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4474, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 4474, __pyx_L7_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_v_hasIgnoreExprs = (!((!__pyx_t_3) != 0));

      /* "Aspidites/_vendor/pyparsing_extension.py":4475
 *         try:
 *             hasIgnoreExprs = (not not self.ignoreExprs)
 *             while 1:             # <<<<<<<<<<<<<<
 *                 if check_ender:
 *                     try_not_ender(instring, loc)
 */
      while (1) {

        /* "Aspidites/_vendor/pyparsing_extension.py":4476
 *             hasIgnoreExprs = (not not self.ignoreExprs)
 *             while 1:
 *                 if check_ender:             # <<<<<<<<<<<<<<
 *                     try_not_ender(instring, loc)
 *                 if hasIgnoreExprs:
 */
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_check_ender); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 4476, __pyx_L7_error)
        if (__pyx_t_3) {

          /* "Aspidites/_vendor/pyparsing_extension.py":4477
 *             while 1:
 *                 if check_ender:
 *                     try_not_ender(instring, loc)             # <<<<<<<<<<<<<<
 *                 if hasIgnoreExprs:
 *                     preloc = self_skip_ignorables(instring, loc)
 */
          if (unlikely(!__pyx_v_try_not_ender)) { __Pyx_RaiseUnboundLocalError("try_not_ender"); __PYX_ERR(0, 4477, __pyx_L7_error) }
          __Pyx_INCREF(__pyx_v_try_not_ender);
          __pyx_t_1 = __pyx_v_try_not_ender; __pyx_t_2 = NULL;
          __pyx_t_5 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_1, function);
              __pyx_t_5 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_1)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_instring, __pyx_v_loc};
            __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4477, __pyx_L7_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_6);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
            PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_instring, __pyx_v_loc};
            __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4477, __pyx_L7_error)
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_GOTREF(__pyx_t_6);
          } else
          #endif
          {
            __pyx_t_4 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4477, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (__pyx_t_2) {
              __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
            }
            __Pyx_INCREF(__pyx_v_instring);
            __Pyx_GIVEREF(__pyx_v_instring);
            PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_5, __pyx_v_instring);
            __Pyx_INCREF(__pyx_v_loc);
            __Pyx_GIVEREF(__pyx_v_loc);
            PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_5, __pyx_v_loc);
            __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4477, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":4476
 *             hasIgnoreExprs = (not not self.ignoreExprs)
 *             while 1:
 *                 if check_ender:             # <<<<<<<<<<<<<<
 *                     try_not_ender(instring, loc)
 *                 if hasIgnoreExprs:
 */
        }

        /* "Aspidites/_vendor/pyparsing_extension.py":4478
 *                 if check_ender:
 *                     try_not_ender(instring, loc)
 *                 if hasIgnoreExprs:             # <<<<<<<<<<<<<<
 *                     preloc = self_skip_ignorables(instring, loc)
 *                 else:
 */
        __pyx_t_3 = (__pyx_v_hasIgnoreExprs != 0);
        if (__pyx_t_3) {

          /* "Aspidites/_vendor/pyparsing_extension.py":4479
 *                     try_not_ender(instring, loc)
 *                 if hasIgnoreExprs:
 *                     preloc = self_skip_ignorables(instring, loc)             # <<<<<<<<<<<<<<
 *                 else:
 *                     preloc = loc
 */
          __Pyx_INCREF(__pyx_v_self_skip_ignorables);
          __pyx_t_1 = __pyx_v_self_skip_ignorables; __pyx_t_4 = NULL;
          __pyx_t_5 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_1, function);
              __pyx_t_5 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_1)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_instring, __pyx_v_loc};
            __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4479, __pyx_L7_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_6);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_instring, __pyx_v_loc};
            __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4479, __pyx_L7_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_6);
          } else
          #endif
          {
            __pyx_t_2 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4479, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_2);
            if (__pyx_t_4) {
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __pyx_t_4 = NULL;
            }
            __Pyx_INCREF(__pyx_v_instring);
            __Pyx_GIVEREF(__pyx_v_instring);
            PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_5, __pyx_v_instring);
            __Pyx_INCREF(__pyx_v_loc);
            __Pyx_GIVEREF(__pyx_v_loc);
            PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_5, __pyx_v_loc);
            __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4479, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF_SET(__pyx_v_preloc, __pyx_t_6);
          __pyx_t_6 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":4478
 *                 if check_ender:
 *                     try_not_ender(instring, loc)
 *                 if hasIgnoreExprs:             # <<<<<<<<<<<<<<
 *                     preloc = self_skip_ignorables(instring, loc)
 *                 else:
 */
          goto __pyx_L16;
        }

        /* "Aspidites/_vendor/pyparsing_extension.py":4481
 *                     preloc = self_skip_ignorables(instring, loc)
 *                 else:
 *                     preloc = loc             # <<<<<<<<<<<<<<
 *                 loc, tmptokens = self_expr_parse(instring, preloc, doActions)
 *                 if tmptokens or tmptokens.haskeys():
 */
        /*else*/ {
          __Pyx_INCREF(__pyx_v_loc);
          __Pyx_XDECREF_SET(__pyx_v_preloc, __pyx_v_loc);
        }
        __pyx_L16:;

        /* "Aspidites/_vendor/pyparsing_extension.py":4482
 *                 else:
 *                     preloc = loc
 *                 loc, tmptokens = self_expr_parse(instring, preloc, doActions)             # <<<<<<<<<<<<<<
 *                 if tmptokens or tmptokens.haskeys():
 *                     tokens += tmptokens
 */
        __Pyx_INCREF(__pyx_v_self_expr_parse);
        __pyx_t_1 = __pyx_v_self_expr_parse; __pyx_t_2 = NULL;
        __pyx_t_5 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_1, function);
            __pyx_t_5 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_1)) {
          PyObject *__pyx_temp[4] = {__pyx_t_2, __pyx_v_instring, __pyx_v_preloc, __pyx_v_doActions};
          __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4482, __pyx_L7_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_6);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
          PyObject *__pyx_temp[4] = {__pyx_t_2, __pyx_v_instring, __pyx_v_preloc, __pyx_v_doActions};
          __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_5, 3+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4482, __pyx_L7_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_6);
        } else
        #endif
        {
          __pyx_t_4 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4482, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_4);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_INCREF(__pyx_v_instring);
          __Pyx_GIVEREF(__pyx_v_instring);
          PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_5, __pyx_v_instring);
          __Pyx_INCREF(__pyx_v_preloc);
          __Pyx_GIVEREF(__pyx_v_preloc);
          PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_5, __pyx_v_preloc);
          __Pyx_INCREF(__pyx_v_doActions);
          __Pyx_GIVEREF(__pyx_v_doActions);
          PyTuple_SET_ITEM(__pyx_t_4, 2+__pyx_t_5, __pyx_v_doActions);
          __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4482, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if ((likely(PyTuple_CheckExact(__pyx_t_6))) || (PyList_CheckExact(__pyx_t_6))) {
          PyObject* sequence = __pyx_t_6;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 4482, __pyx_L7_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
          } else {
            __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_4 = PyList_GET_ITEM(sequence, 1); 
          }
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_4);
          #else
          __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4482, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4482, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        } else {
          Py_ssize_t index = -1;
          __pyx_t_2 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4482, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_7 = Py_TYPE(__pyx_t_2)->tp_iternext;
          index = 0; __pyx_t_1 = __pyx_t_7(__pyx_t_2); if (unlikely(!__pyx_t_1)) goto __pyx_L17_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_1);
          index = 1; __pyx_t_4 = __pyx_t_7(__pyx_t_2); if (unlikely(!__pyx_t_4)) goto __pyx_L17_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_4);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_2), 2) < 0) __PYX_ERR(0, 4482, __pyx_L7_error)
          __pyx_t_7 = NULL;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          goto __pyx_L18_unpacking_done;
          __pyx_L17_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_7 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 4482, __pyx_L7_error)
          __pyx_L18_unpacking_done:;
        }
        __Pyx_DECREF_SET(__pyx_v_loc, __pyx_t_1);
        __pyx_t_1 = 0;
        __Pyx_XDECREF_SET(__pyx_v_tmptokens, __pyx_t_4);
        __pyx_t_4 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":4483
 *                     preloc = loc
 *                 loc, tmptokens = self_expr_parse(instring, preloc, doActions)
 *                 if tmptokens or tmptokens.haskeys():             # <<<<<<<<<<<<<<
 *                     tokens += tmptokens
 *         except (ParseException, IndexError):
 */
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_v_tmptokens); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 4483, __pyx_L7_error)
        if (!__pyx_t_11) {
        } else {
          __pyx_t_3 = __pyx_t_11;
          goto __pyx_L20_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_tmptokens, __pyx_n_s_haskeys); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4483, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_6 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4483, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 4483, __pyx_L7_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_3 = __pyx_t_11;
        __pyx_L20_bool_binop_done:;
        if (__pyx_t_3) {

          /* "Aspidites/_vendor/pyparsing_extension.py":4484
 *                 loc, tmptokens = self_expr_parse(instring, preloc, doActions)
 *                 if tmptokens or tmptokens.haskeys():
 *                     tokens += tmptokens             # <<<<<<<<<<<<<<
 *         except (ParseException, IndexError):
 *             pass
 */
          __pyx_t_6 = PyNumber_InPlaceAdd(__pyx_v_tokens, __pyx_v_tmptokens); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4484, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF_SET(__pyx_v_tokens, __pyx_t_6);
          __pyx_t_6 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":4483
 *                     preloc = loc
 *                 loc, tmptokens = self_expr_parse(instring, preloc, doActions)
 *                 if tmptokens or tmptokens.haskeys():             # <<<<<<<<<<<<<<
 *                     tokens += tmptokens
 *         except (ParseException, IndexError):
 */
        }
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":4473
 *             try_not_ender(instring, loc)
 *         loc, tokens = self_expr_parse(instring, loc, doActions, callPreParse=False)
 *         try:             # <<<<<<<<<<<<<<
 *             hasIgnoreExprs = (not not self.ignoreExprs)
 *             while 1:
 */
    }
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    goto __pyx_L12_try_end;
    __pyx_L7_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4485
 *                 if tmptokens or tmptokens.haskeys():
 *                     tokens += tmptokens
 *         except (ParseException, IndexError):             # <<<<<<<<<<<<<<
 *             pass
 * 
 */
    __Pyx_ErrFetch(&__pyx_t_6, &__pyx_t_4, &__pyx_t_1);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4485, __pyx_L9_except_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_6, __pyx_t_2) || __Pyx_PyErr_GivenExceptionMatches(__pyx_t_6, __pyx_builtin_IndexError);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_ErrRestore(__pyx_t_6, __pyx_t_4, __pyx_t_1);
    __pyx_t_6 = 0; __pyx_t_4 = 0; __pyx_t_1 = 0;
    if (__pyx_t_5) {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L8_exception_handled;
    }
    goto __pyx_L9_except_error;
    __pyx_L9_except_error:;

    /* "Aspidites/_vendor/pyparsing_extension.py":4473
 *             try_not_ender(instring, loc)
 *         loc, tokens = self_expr_parse(instring, loc, doActions, callPreParse=False)
 *         try:             # <<<<<<<<<<<<<<
 *             hasIgnoreExprs = (not not self.ignoreExprs)
 *             while 1:
 */
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_XGIVEREF(__pyx_t_10);
    __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
    goto __pyx_L1_error;
    __pyx_L8_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_XGIVEREF(__pyx_t_10);
    __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
    __pyx_L12_try_end:;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4488
 *             pass
 * 
 *         return loc, tokens             # <<<<<<<<<<<<<<
 * 
 *     def _setResultsName(self, name, listAllMatches=False):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_loc);
  __Pyx_GIVEREF(__pyx_v_loc);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_loc);
  __Pyx_INCREF(__pyx_v_tokens);
  __Pyx_GIVEREF(__pyx_v_tokens);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_tokens);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4461
 *         return self
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         self_expr_parse = self.expr._parse
 *         self_skip_ignorables = self._skipIgnorables
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._MultipleMatch.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_self_expr_parse);
  __Pyx_XDECREF(__pyx_v_self_skip_ignorables);
  __Pyx_XDECREF(__pyx_v_check_ender);
  __Pyx_XDECREF(__pyx_v_try_not_ender);
  __Pyx_XDECREF(__pyx_v_tokens);
  __Pyx_XDECREF(__pyx_v_preloc);
  __Pyx_XDECREF(__pyx_v_tmptokens);
  __Pyx_XDECREF(__pyx_v_loc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4490
 *         return loc, tokens
 * 
 *     def _setResultsName(self, name, listAllMatches=False):             # <<<<<<<<<<<<<<
 *         if __diag__.warn_ungrouped_named_tokens_in_collection:
 *             for e in [self.expr] + getattr(self.expr, 'exprs', []):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_14_MultipleMatch_7_setResultsName(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_14_MultipleMatch_7_setResultsName = {"_setResultsName", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_14_MultipleMatch_7_setResultsName, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_14_MultipleMatch_7_setResultsName(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_listAllMatches = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_setResultsName (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_name,&__pyx_n_s_listAllMatches,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_False));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_setResultsName", 0, 2, 3, 1); __PYX_ERR(0, 4490, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_listAllMatches);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_setResultsName") < 0)) __PYX_ERR(0, 4490, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_name = values[1];
    __pyx_v_listAllMatches = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_setResultsName", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4490, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._MultipleMatch._setResultsName", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_14_MultipleMatch_6_setResultsName(__pyx_self, __pyx_v_self, __pyx_v_name, __pyx_v_listAllMatches);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_14_MultipleMatch_6_setResultsName(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_listAllMatches) {
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_setResultsName", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":4491
 * 
 *     def _setResultsName(self, name, listAllMatches=False):
 *         if __diag__.warn_ungrouped_named_tokens_in_collection:             # <<<<<<<<<<<<<<
 *             for e in [self.expr] + getattr(self.expr, 'exprs', []):
 *                 if isinstance(e, ParserElement) and e.resultsName:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_diag); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_warn_ungrouped_named_tokens_in_c); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 4491, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4492
 *     def _setResultsName(self, name, listAllMatches=False):
 *         if __diag__.warn_ungrouped_named_tokens_in_collection:
 *             for e in [self.expr] + getattr(self.expr, 'exprs', []):             # <<<<<<<<<<<<<<
 *                 if isinstance(e, ParserElement) and e.resultsName:
 *                     warnings.warn("{0}: setting results name {1!r} on {2} expression "
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4492, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4492, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4492, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4492, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetAttr3(__pyx_t_2, __pyx_n_s_exprs, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4492, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyNumber_Add(__pyx_t_1, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4492, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
      __pyx_t_5 = __pyx_t_4; __Pyx_INCREF(__pyx_t_5); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4492, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_7 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4492, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_5))) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_5)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_6); __Pyx_INCREF(__pyx_t_4); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 4492, __pyx_L1_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_5, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4492, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        } else {
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_6); __Pyx_INCREF(__pyx_t_4); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 4492, __pyx_L1_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_5, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4492, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        }
      } else {
        __pyx_t_4 = __pyx_t_7(__pyx_t_5);
        if (unlikely(!__pyx_t_4)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 4492, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_4);
      }
      __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":4493
 *         if __diag__.warn_ungrouped_named_tokens_in_collection:
 *             for e in [self.expr] + getattr(self.expr, 'exprs', []):
 *                 if isinstance(e, ParserElement) and e.resultsName:             # <<<<<<<<<<<<<<
 *                     warnings.warn("{0}: setting results name {1!r} on {2} expression "
 *                                   "collides with {3!r} on contained expression".format(
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4493, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = PyObject_IsInstance(__pyx_v_e, __pyx_t_4); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 4493, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_9 = (__pyx_t_8 != 0);
      if (__pyx_t_9) {
      } else {
        __pyx_t_3 = __pyx_t_9;
        goto __pyx_L7_bool_binop_done;
      }
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_resultsName); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4493, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 4493, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_3 = __pyx_t_9;
      __pyx_L7_bool_binop_done:;
      if (__pyx_t_3) {

        /* "Aspidites/_vendor/pyparsing_extension.py":4494
 *             for e in [self.expr] + getattr(self.expr, 'exprs', []):
 *                 if isinstance(e, ParserElement) and e.resultsName:
 *                     warnings.warn("{0}: setting results name {1!r} on {2} expression "             # <<<<<<<<<<<<<<
 *                                   "collides with {3!r} on contained expression".format(
 *                         "warn_ungrouped_named_tokens_in_collection",
 */
        __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_warnings); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4494, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_warn); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4494, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":4495
 *                 if isinstance(e, ParserElement) and e.resultsName:
 *                     warnings.warn("{0}: setting results name {1!r} on {2} expression "
 *                                   "collides with {3!r} on contained expression".format(             # <<<<<<<<<<<<<<
 *                         "warn_ungrouped_named_tokens_in_collection",
 *                         name,
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_0_setting_results_name_1_r_on_2, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4495, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);

        /* "Aspidites/_vendor/pyparsing_extension.py":4498
 *                         "warn_ungrouped_named_tokens_in_collection",
 *                         name,
 *                         type(self).__name__,             # <<<<<<<<<<<<<<
 *                         e.resultsName),
 *                                   stacklevel=3)
 */
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(__pyx_v_self)), __pyx_n_s_name_2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 4498, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);

        /* "Aspidites/_vendor/pyparsing_extension.py":4499
 *                         name,
 *                         type(self).__name__,
 *                         e.resultsName),             # <<<<<<<<<<<<<<
 *                                   stacklevel=3)
 * 
 */
        __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_resultsName); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 4499, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_12 = NULL;
        __pyx_t_13 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_12)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_12);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
            __pyx_t_13 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_2)) {
          PyObject *__pyx_temp[5] = {__pyx_t_12, __pyx_n_s_warn_ungrouped_named_tokens_in_c, __pyx_v_name, __pyx_t_10, __pyx_t_11};
          __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_13, 4+__pyx_t_13); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4495, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
          PyObject *__pyx_temp[5] = {__pyx_t_12, __pyx_n_s_warn_ungrouped_named_tokens_in_c, __pyx_v_name, __pyx_t_10, __pyx_t_11};
          __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_13, 4+__pyx_t_13); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4495, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        } else
        #endif
        {
          __pyx_t_14 = PyTuple_New(4+__pyx_t_13); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 4495, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_14);
          if (__pyx_t_12) {
            __Pyx_GIVEREF(__pyx_t_12); PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_12); __pyx_t_12 = NULL;
          }
          __Pyx_INCREF(__pyx_n_s_warn_ungrouped_named_tokens_in_c);
          __Pyx_GIVEREF(__pyx_n_s_warn_ungrouped_named_tokens_in_c);
          PyTuple_SET_ITEM(__pyx_t_14, 0+__pyx_t_13, __pyx_n_s_warn_ungrouped_named_tokens_in_c);
          __Pyx_INCREF(__pyx_v_name);
          __Pyx_GIVEREF(__pyx_v_name);
          PyTuple_SET_ITEM(__pyx_t_14, 1+__pyx_t_13, __pyx_v_name);
          __Pyx_GIVEREF(__pyx_t_10);
          PyTuple_SET_ITEM(__pyx_t_14, 2+__pyx_t_13, __pyx_t_10);
          __Pyx_GIVEREF(__pyx_t_11);
          PyTuple_SET_ITEM(__pyx_t_14, 3+__pyx_t_13, __pyx_t_11);
          __pyx_t_10 = 0;
          __pyx_t_11 = 0;
          __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_14, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4495, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":4494
 *             for e in [self.expr] + getattr(self.expr, 'exprs', []):
 *                 if isinstance(e, ParserElement) and e.resultsName:
 *                     warnings.warn("{0}: setting results name {1!r} on {2} expression "             # <<<<<<<<<<<<<<
 *                                   "collides with {3!r} on contained expression".format(
 *                         "warn_ungrouped_named_tokens_in_collection",
 */
        __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4494, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4);
        __pyx_t_4 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":4500
 *                         type(self).__name__,
 *                         e.resultsName),
 *                                   stacklevel=3)             # <<<<<<<<<<<<<<
 * 
 *         return super(_MultipleMatch, self)._setResultsName(name, listAllMatches)
 */
        __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4500, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_stacklevel, __pyx_int_3) < 0) __PYX_ERR(0, 4500, __pyx_L1_error)

        /* "Aspidites/_vendor/pyparsing_extension.py":4494
 *             for e in [self.expr] + getattr(self.expr, 'exprs', []):
 *                 if isinstance(e, ParserElement) and e.resultsName:
 *                     warnings.warn("{0}: setting results name {1!r} on {2} expression "             # <<<<<<<<<<<<<<
 *                                   "collides with {3!r} on contained expression".format(
 *                         "warn_ungrouped_named_tokens_in_collection",
 */
        __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 4494, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":4493
 *         if __diag__.warn_ungrouped_named_tokens_in_collection:
 *             for e in [self.expr] + getattr(self.expr, 'exprs', []):
 *                 if isinstance(e, ParserElement) and e.resultsName:             # <<<<<<<<<<<<<<
 *                     warnings.warn("{0}: setting results name {1!r} on {2} expression "
 *                                   "collides with {3!r} on contained expression".format(
 */
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":4492
 *     def _setResultsName(self, name, listAllMatches=False):
 *         if __diag__.warn_ungrouped_named_tokens_in_collection:
 *             for e in [self.expr] + getattr(self.expr, 'exprs', []):             # <<<<<<<<<<<<<<
 *                 if isinstance(e, ParserElement) and e.resultsName:
 *                     warnings.warn("{0}: setting results name {1!r} on {2} expression "
 */
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4491
 * 
 *     def _setResultsName(self, name, listAllMatches=False):
 *         if __diag__.warn_ungrouped_named_tokens_in_collection:             # <<<<<<<<<<<<<<
 *             for e in [self.expr] + getattr(self.expr, 'exprs', []):
 *                 if isinstance(e, ParserElement) and e.resultsName:
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4502
 *                                   stacklevel=3)
 * 
 *         return super(_MultipleMatch, self)._setResultsName(name, listAllMatches)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_MultipleMatch); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 4502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_14);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_14);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_self);
  __pyx_t_14 = 0;
  __pyx_t_14 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_4, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 4502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_setResultsName); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __pyx_t_14 = NULL;
  __pyx_t_13 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_14)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_14);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_13 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_4)) {
    PyObject *__pyx_temp[3] = {__pyx_t_14, __pyx_v_name, __pyx_v_listAllMatches};
    __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_13, 2+__pyx_t_13); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4502, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_GOTREF(__pyx_t_5);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
    PyObject *__pyx_temp[3] = {__pyx_t_14, __pyx_v_name, __pyx_v_listAllMatches};
    __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_13, 2+__pyx_t_13); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4502, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_GOTREF(__pyx_t_5);
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(2+__pyx_t_13); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4502, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_14) {
      __Pyx_GIVEREF(__pyx_t_14); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_14); __pyx_t_14 = NULL;
    }
    __Pyx_INCREF(__pyx_v_name);
    __Pyx_GIVEREF(__pyx_v_name);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_13, __pyx_v_name);
    __Pyx_INCREF(__pyx_v_listAllMatches);
    __Pyx_GIVEREF(__pyx_v_listAllMatches);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_13, __pyx_v_listAllMatches);
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4502, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4490
 *         return loc, tokens
 * 
 *     def _setResultsName(self, name, listAllMatches=False):             # <<<<<<<<<<<<<<
 *         if __diag__.warn_ungrouped_named_tokens_in_collection:
 *             for e in [self.expr] + getattr(self.expr, 'exprs', []):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension._MultipleMatch._setResultsName", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4531
 *     """
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_9OneOrMore_1__str__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_9OneOrMore_1__str__ = {"__str__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_9OneOrMore_1__str__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_9OneOrMore_1__str__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_9OneOrMore___str__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_9OneOrMore___str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":4532
 * 
 *     def __str__(self):
 *         if hasattr(self, "name"):             # <<<<<<<<<<<<<<
 *             return self.name
 * 
 */
  __pyx_t_1 = __Pyx_HasAttr(__pyx_v_self, __pyx_n_s_name); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4532, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4533
 *     def __str__(self):
 *         if hasattr(self, "name"):
 *             return self.name             # <<<<<<<<<<<<<<
 * 
 *         if self.strRepr is None:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4533, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4532
 * 
 *     def __str__(self):
 *         if hasattr(self, "name"):             # <<<<<<<<<<<<<<
 *             return self.name
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4535
 *             return self.name
 * 
 *         if self.strRepr is None:             # <<<<<<<<<<<<<<
 *             self.strRepr = "{" + str(self.expr) + "}..."
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_strRepr); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = (__pyx_t_3 == Py_None);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4536
 * 
 *         if self.strRepr is None:
 *             self.strRepr = "{" + str(self.expr) + "}..."             # <<<<<<<<<<<<<<
 * 
 *         return self.strRepr
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4536, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4536, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Add(__pyx_kp_s__109, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4536, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyNumber_Add(__pyx_t_3, __pyx_kp_s__116); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4536, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_strRepr, __pyx_t_4) < 0) __PYX_ERR(0, 4536, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4535
 *             return self.name
 * 
 *         if self.strRepr is None:             # <<<<<<<<<<<<<<
 *             self.strRepr = "{" + str(self.expr) + "}..."
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4538
 *             self.strRepr = "{" + str(self.expr) + "}..."
 * 
 *         return self.strRepr             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_strRepr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4531
 *     """
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.OneOrMore.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4553
 *     """
 * 
 *     def __init__(self, expr, stopOn=None):             # <<<<<<<<<<<<<<
 *         super(ZeroOrMore, self).__init__(expr, stopOn=stopOn)
 *         self.mayReturnEmpty = True
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10ZeroOrMore_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10ZeroOrMore_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10ZeroOrMore_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10ZeroOrMore_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_expr = 0;
  PyObject *__pyx_v_stopOn = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_expr,&__pyx_n_s_stopOn,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_expr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, 1); __PYX_ERR(0, 4553, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_stopOn);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 4553, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_expr = values[1];
    __pyx_v_stopOn = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4553, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ZeroOrMore.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10ZeroOrMore___init__(__pyx_self, __pyx_v_self, __pyx_v_expr, __pyx_v_stopOn);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10ZeroOrMore___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_expr, PyObject *__pyx_v_stopOn) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":4554
 * 
 *     def __init__(self, expr, stopOn=None):
 *         super(ZeroOrMore, self).__init__(expr, stopOn=stopOn)             # <<<<<<<<<<<<<<
 *         self.mayReturnEmpty = True
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ZeroOrMore); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_self);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_expr);
  __Pyx_GIVEREF(__pyx_v_expr);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_expr);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_stopOn, __pyx_v_stopOn) < 0) __PYX_ERR(0, 4554, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4555
 *     def __init__(self, expr, stopOn=None):
 *         super(ZeroOrMore, self).__init__(expr, stopOn=stopOn)
 *         self.mayReturnEmpty = True             # <<<<<<<<<<<<<<
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_mayReturnEmpty, Py_True) < 0) __PYX_ERR(0, 4555, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4553
 *     """
 * 
 *     def __init__(self, expr, stopOn=None):             # <<<<<<<<<<<<<<
 *         super(ZeroOrMore, self).__init__(expr, stopOn=stopOn)
 *         self.mayReturnEmpty = True
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ZeroOrMore.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4557
 *         self.mayReturnEmpty = True
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         try:
 *             return super(ZeroOrMore, self).parseImpl(instring, loc, doActions)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10ZeroOrMore_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10ZeroOrMore_3parseImpl = {"parseImpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10ZeroOrMore_3parseImpl, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10ZeroOrMore_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  PyObject *__pyx_v_doActions = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parseImpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_doActions,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 1); __PYX_ERR(0, 4557, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 2); __PYX_ERR(0, 4557, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_doActions);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "parseImpl") < 0)) __PYX_ERR(0, 4557, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_doActions = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4557, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ZeroOrMore.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10ZeroOrMore_2parseImpl(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10ZeroOrMore_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, PyObject *__pyx_v_doActions) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parseImpl", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":4558
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         try:             # <<<<<<<<<<<<<<
 *             return super(ZeroOrMore, self).parseImpl(instring, loc, doActions)
 *         except (ParseException, IndexError):
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "Aspidites/_vendor/pyparsing_extension.py":4559
 *     def parseImpl(self, instring, loc, doActions=True):
 *         try:
 *             return super(ZeroOrMore, self).parseImpl(instring, loc, doActions)             # <<<<<<<<<<<<<<
 *         except (ParseException, IndexError):
 *             return loc, []
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_ZeroOrMore); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4559, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4559, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5);
      __Pyx_INCREF(__pyx_v_self);
      __Pyx_GIVEREF(__pyx_v_self);
      PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_v_self);
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_6, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4559, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_parseImpl); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4559, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = NULL;
      __pyx_t_7 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
          __pyx_t_7 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_6)) {
        PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions};
        __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 3+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4559, __pyx_L3_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_4);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
        PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions};
        __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 3+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4559, __pyx_L3_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_4);
      } else
      #endif
      {
        __pyx_t_8 = PyTuple_New(3+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 4559, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_8);
        if (__pyx_t_5) {
          __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_5); __pyx_t_5 = NULL;
        }
        __Pyx_INCREF(__pyx_v_instring);
        __Pyx_GIVEREF(__pyx_v_instring);
        PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_v_instring);
        __Pyx_INCREF(__pyx_v_loc);
        __Pyx_GIVEREF(__pyx_v_loc);
        PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_v_loc);
        __Pyx_INCREF(__pyx_v_doActions);
        __Pyx_GIVEREF(__pyx_v_doActions);
        PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_7, __pyx_v_doActions);
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4559, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_r = __pyx_t_4;
      __pyx_t_4 = 0;
      goto __pyx_L7_try_return;

      /* "Aspidites/_vendor/pyparsing_extension.py":4558
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         try:             # <<<<<<<<<<<<<<
 *             return super(ZeroOrMore, self).parseImpl(instring, loc, doActions)
 *         except (ParseException, IndexError):
 */
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4560
 *         try:
 *             return super(ZeroOrMore, self).parseImpl(instring, loc, doActions)
 *         except (ParseException, IndexError):             # <<<<<<<<<<<<<<
 *             return loc, []
 * 
 */
    __Pyx_ErrFetch(&__pyx_t_4, &__pyx_t_6, &__pyx_t_8);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4560, __pyx_L5_except_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_4, __pyx_t_5) || __Pyx_PyErr_GivenExceptionMatches(__pyx_t_4, __pyx_builtin_IndexError);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_ErrRestore(__pyx_t_4, __pyx_t_6, __pyx_t_8);
    __pyx_t_4 = 0; __pyx_t_6 = 0; __pyx_t_8 = 0;
    if (__pyx_t_7) {
      __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ZeroOrMore.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_6, &__pyx_t_4) < 0) __PYX_ERR(0, 4560, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_4);

      /* "Aspidites/_vendor/pyparsing_extension.py":4561
 *             return super(ZeroOrMore, self).parseImpl(instring, loc, doActions)
 *         except (ParseException, IndexError):
 *             return loc, []             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4561, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 4561, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_v_loc);
      __Pyx_GIVEREF(__pyx_v_loc);
      PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v_loc);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_r = __pyx_t_9;
      __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L6_except_return;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "Aspidites/_vendor/pyparsing_extension.py":4558
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         try:             # <<<<<<<<<<<<<<
 *             return super(ZeroOrMore, self).parseImpl(instring, loc, doActions)
 *         except (ParseException, IndexError):
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4557
 *         self.mayReturnEmpty = True
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         try:
 *             return super(ZeroOrMore, self).parseImpl(instring, loc, doActions)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ZeroOrMore.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4563
 *             return loc, []
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10ZeroOrMore_5__str__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10ZeroOrMore_5__str__ = {"__str__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10ZeroOrMore_5__str__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10ZeroOrMore_5__str__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10ZeroOrMore_4__str__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10ZeroOrMore_4__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":4564
 * 
 *     def __str__(self):
 *         if hasattr(self, "name"):             # <<<<<<<<<<<<<<
 *             return self.name
 * 
 */
  __pyx_t_1 = __Pyx_HasAttr(__pyx_v_self, __pyx_n_s_name); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4564, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4565
 *     def __str__(self):
 *         if hasattr(self, "name"):
 *             return self.name             # <<<<<<<<<<<<<<
 * 
 *         if self.strRepr is None:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4565, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4564
 * 
 *     def __str__(self):
 *         if hasattr(self, "name"):             # <<<<<<<<<<<<<<
 *             return self.name
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4567
 *             return self.name
 * 
 *         if self.strRepr is None:             # <<<<<<<<<<<<<<
 *             self.strRepr = "[" + str(self.expr) + "]..."
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_strRepr); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = (__pyx_t_3 == Py_None);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4568
 * 
 *         if self.strRepr is None:
 *             self.strRepr = "[" + str(self.expr) + "]..."             # <<<<<<<<<<<<<<
 * 
 *         return self.strRepr
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4568, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4568, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Add(__pyx_kp_s__14, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4568, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyNumber_Add(__pyx_t_3, __pyx_kp_s__117); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4568, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_strRepr, __pyx_t_4) < 0) __PYX_ERR(0, 4568, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4567
 *             return self.name
 * 
 *         if self.strRepr is None:             # <<<<<<<<<<<<<<
 *             self.strRepr = "[" + str(self.expr) + "]..."
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4570
 *             self.strRepr = "[" + str(self.expr) + "]..."
 * 
 *         return self.strRepr             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_strRepr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4563
 *             return loc, []
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.ZeroOrMore.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4574
 * 
 * class _NullToken(object):
 *     def __bool__(self):             # <<<<<<<<<<<<<<
 *         return False
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_NullToken_1__bool__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10_NullToken_1__bool__ = {"__bool__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_NullToken_1__bool__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_NullToken_1__bool__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__bool__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_NullToken___bool__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_NullToken___bool__(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__bool__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":4575
 * class _NullToken(object):
 *     def __bool__(self):
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     __nonzero__ = __bool__
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_False);
  __pyx_r = Py_False;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4574
 * 
 * class _NullToken(object):
 *     def __bool__(self):             # <<<<<<<<<<<<<<
 *         return False
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4579
 *     __nonzero__ = __bool__
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return ""
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_NullToken_3__str__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10_NullToken_3__str__ = {"__str__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_NullToken_3__str__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_10_NullToken_3__str__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_NullToken_2__str__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_10_NullToken_2__str__(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":4580
 * 
 *     def __str__(self):
 *         return ""             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_kp_s__12);
  __pyx_r = __pyx_kp_s__12;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4579
 *     __nonzero__ = __bool__
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return ""
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4622
 *     __optionalNotMatched = _NullToken()
 * 
 *     def __init__(self, expr, default=__optionalNotMatched):             # <<<<<<<<<<<<<<
 *         super(Optional, self).__init__(expr, savelist=False)
 *         self.saveAsList = self.expr.saveAsList
 */

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_42__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__defaults__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults7, __pyx_self)->__pyx_arg_default);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults7, __pyx_self)->__pyx_arg_default);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __Pyx_CyFunction_Defaults(__pyx_defaults7, __pyx_self)->__pyx_arg_default);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_2, 1, Py_None);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_8Optional_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_8Optional_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_8Optional_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_8Optional_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_expr = 0;
  PyObject *__pyx_v_default = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_expr,&__pyx_n_s_default,0};
    PyObject* values[3] = {0,0,0};
    __pyx_defaults7 *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults7, __pyx_self);
    values[2] = __pyx_dynamic_args->__pyx_arg_default;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_expr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, 1); __PYX_ERR(0, 4622, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_default);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 4622, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_expr = values[1];
    __pyx_v_default = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4622, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Optional.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_8Optional___init__(__pyx_self, __pyx_v_self, __pyx_v_expr, __pyx_v_default);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_8Optional___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_expr, PyObject *__pyx_v_default) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":4623
 * 
 *     def __init__(self, expr, default=__optionalNotMatched):
 *         super(Optional, self).__init__(expr, savelist=False)             # <<<<<<<<<<<<<<
 *         self.saveAsList = self.expr.saveAsList
 *         self.defaultValue = default
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Optional); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_self);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_expr);
  __Pyx_GIVEREF(__pyx_v_expr);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_expr);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_savelist, Py_False) < 0) __PYX_ERR(0, 4623, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4624
 *     def __init__(self, expr, default=__optionalNotMatched):
 *         super(Optional, self).__init__(expr, savelist=False)
 *         self.saveAsList = self.expr.saveAsList             # <<<<<<<<<<<<<<
 *         self.defaultValue = default
 *         self.mayReturnEmpty = True
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_saveAsList); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_saveAsList, __pyx_t_3) < 0) __PYX_ERR(0, 4624, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4625
 *         super(Optional, self).__init__(expr, savelist=False)
 *         self.saveAsList = self.expr.saveAsList
 *         self.defaultValue = default             # <<<<<<<<<<<<<<
 *         self.mayReturnEmpty = True
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_defaultValue, __pyx_v_default) < 0) __PYX_ERR(0, 4625, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4626
 *         self.saveAsList = self.expr.saveAsList
 *         self.defaultValue = default
 *         self.mayReturnEmpty = True             # <<<<<<<<<<<<<<
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_mayReturnEmpty, Py_True) < 0) __PYX_ERR(0, 4626, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4622
 *     __optionalNotMatched = _NullToken()
 * 
 *     def __init__(self, expr, default=__optionalNotMatched):             # <<<<<<<<<<<<<<
 *         super(Optional, self).__init__(expr, savelist=False)
 *         self.saveAsList = self.expr.saveAsList
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Optional.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4628
 *         self.mayReturnEmpty = True
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         try:
 *             loc, tokens = self.expr._parse(instring, loc, doActions, callPreParse=False)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_8Optional_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_8Optional_3parseImpl = {"parseImpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_8Optional_3parseImpl, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_8Optional_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  PyObject *__pyx_v_doActions = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parseImpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_doActions,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 1); __PYX_ERR(0, 4628, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 2); __PYX_ERR(0, 4628, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_doActions);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "parseImpl") < 0)) __PYX_ERR(0, 4628, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_doActions = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4628, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Optional.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_8Optional_2parseImpl(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_8Optional_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, PyObject *__pyx_v_doActions) {
  PyObject *__pyx_v_tokens = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parseImpl", 0);
  __Pyx_INCREF(__pyx_v_loc);

  /* "Aspidites/_vendor/pyparsing_extension.py":4629
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         try:             # <<<<<<<<<<<<<<
 *             loc, tokens = self.expr._parse(instring, loc, doActions, callPreParse=False)
 *         except (ParseException, IndexError):
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "Aspidites/_vendor/pyparsing_extension.py":4630
 *     def parseImpl(self, instring, loc, doActions=True):
 *         try:
 *             loc, tokens = self.expr._parse(instring, loc, doActions, callPreParse=False)             # <<<<<<<<<<<<<<
 *         except (ParseException, IndexError):
 *             if self.defaultValue is not self.__optionalNotMatched:
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4630, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_parse); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4630, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4630, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_v_instring);
      __Pyx_GIVEREF(__pyx_v_instring);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_instring);
      __Pyx_INCREF(__pyx_v_loc);
      __Pyx_GIVEREF(__pyx_v_loc);
      PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_loc);
      __Pyx_INCREF(__pyx_v_doActions);
      __Pyx_GIVEREF(__pyx_v_doActions);
      PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_doActions);
      __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4630, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_callPreParse, Py_False) < 0) __PYX_ERR(0, 4630, __pyx_L3_error)
      __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4630, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if ((likely(PyTuple_CheckExact(__pyx_t_7))) || (PyList_CheckExact(__pyx_t_7))) {
        PyObject* sequence = __pyx_t_7;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 4630, __pyx_L3_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_6 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_6 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_4 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_4);
        #else
        __pyx_t_6 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4630, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4630, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_5 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4630, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_8 = Py_TYPE(__pyx_t_5)->tp_iternext;
        index = 0; __pyx_t_6 = __pyx_t_8(__pyx_t_5); if (unlikely(!__pyx_t_6)) goto __pyx_L9_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_6);
        index = 1; __pyx_t_4 = __pyx_t_8(__pyx_t_5); if (unlikely(!__pyx_t_4)) goto __pyx_L9_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_4);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_5), 2) < 0) __PYX_ERR(0, 4630, __pyx_L3_error)
        __pyx_t_8 = NULL;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        goto __pyx_L10_unpacking_done;
        __pyx_L9_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_8 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 4630, __pyx_L3_error)
        __pyx_L10_unpacking_done:;
      }
      __Pyx_DECREF_SET(__pyx_v_loc, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_v_tokens = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":4629
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         try:             # <<<<<<<<<<<<<<
 *             loc, tokens = self.expr._parse(instring, loc, doActions, callPreParse=False)
 *         except (ParseException, IndexError):
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4631
 *         try:
 *             loc, tokens = self.expr._parse(instring, loc, doActions, callPreParse=False)
 *         except (ParseException, IndexError):             # <<<<<<<<<<<<<<
 *             if self.defaultValue is not self.__optionalNotMatched:
 *                 if self.expr.resultsName:
 */
    __Pyx_ErrFetch(&__pyx_t_7, &__pyx_t_4, &__pyx_t_6);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4631, __pyx_L5_except_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_9 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_7, __pyx_t_5) || __Pyx_PyErr_GivenExceptionMatches(__pyx_t_7, __pyx_builtin_IndexError);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_ErrRestore(__pyx_t_7, __pyx_t_4, __pyx_t_6);
    __pyx_t_7 = 0; __pyx_t_4 = 0; __pyx_t_6 = 0;
    if (__pyx_t_9) {
      __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Optional.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_4, &__pyx_t_7) < 0) __PYX_ERR(0, 4631, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_7);

      /* "Aspidites/_vendor/pyparsing_extension.py":4632
 *             loc, tokens = self.expr._parse(instring, loc, doActions, callPreParse=False)
 *         except (ParseException, IndexError):
 *             if self.defaultValue is not self.__optionalNotMatched:             # <<<<<<<<<<<<<<
 *                 if self.expr.resultsName:
 *                     tokens = ParseResults([self.defaultValue])
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_defaultValue); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4632, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_Optional__optionalNotMatched); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 4632, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_11 = (__pyx_t_5 != __pyx_t_10);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_12 = (__pyx_t_11 != 0);
      if (__pyx_t_12) {

        /* "Aspidites/_vendor/pyparsing_extension.py":4633
 *         except (ParseException, IndexError):
 *             if self.defaultValue is not self.__optionalNotMatched:
 *                 if self.expr.resultsName:             # <<<<<<<<<<<<<<
 *                     tokens = ParseResults([self.defaultValue])
 *                     tokens[self.expr.resultsName] = self.defaultValue
 */
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 4633, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_resultsName); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4633, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 4633, __pyx_L5_except_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (__pyx_t_12) {

          /* "Aspidites/_vendor/pyparsing_extension.py":4634
 *             if self.defaultValue is not self.__optionalNotMatched:
 *                 if self.expr.resultsName:
 *                     tokens = ParseResults([self.defaultValue])             # <<<<<<<<<<<<<<
 *                     tokens[self.expr.resultsName] = self.defaultValue
 *                 else:
 */
          __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_ParseResults); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 4634, __pyx_L5_except_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_defaultValue); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4634, __pyx_L5_except_error)
          __Pyx_GOTREF(__pyx_t_13);
          __pyx_t_14 = PyList_New(1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 4634, __pyx_L5_except_error)
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_GIVEREF(__pyx_t_13);
          PyList_SET_ITEM(__pyx_t_14, 0, __pyx_t_13);
          __pyx_t_13 = 0;
          __pyx_t_13 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
            __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_10);
            if (likely(__pyx_t_13)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
              __Pyx_INCREF(__pyx_t_13);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_10, function);
            }
          }
          __pyx_t_5 = (__pyx_t_13) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_13, __pyx_t_14) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_14);
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4634, __pyx_L5_except_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF_SET(__pyx_v_tokens, __pyx_t_5);
          __pyx_t_5 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":4635
 *                 if self.expr.resultsName:
 *                     tokens = ParseResults([self.defaultValue])
 *                     tokens[self.expr.resultsName] = self.defaultValue             # <<<<<<<<<<<<<<
 *                 else:
 *                     tokens = [self.defaultValue]
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_defaultValue); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4635, __pyx_L5_except_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 4635, __pyx_L5_except_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_resultsName); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 4635, __pyx_L5_except_error)
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          if (unlikely(PyObject_SetItem(__pyx_v_tokens, __pyx_t_14, __pyx_t_5) < 0)) __PYX_ERR(0, 4635, __pyx_L5_except_error)
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":4633
 *         except (ParseException, IndexError):
 *             if self.defaultValue is not self.__optionalNotMatched:
 *                 if self.expr.resultsName:             # <<<<<<<<<<<<<<
 *                     tokens = ParseResults([self.defaultValue])
 *                     tokens[self.expr.resultsName] = self.defaultValue
 */
          goto __pyx_L14;
        }

        /* "Aspidites/_vendor/pyparsing_extension.py":4637
 *                     tokens[self.expr.resultsName] = self.defaultValue
 *                 else:
 *                     tokens = [self.defaultValue]             # <<<<<<<<<<<<<<
 *             else:
 *                 tokens = []
 */
        /*else*/ {
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_defaultValue); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4637, __pyx_L5_except_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_14 = PyList_New(1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 4637, __pyx_L5_except_error)
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_GIVEREF(__pyx_t_5);
          PyList_SET_ITEM(__pyx_t_14, 0, __pyx_t_5);
          __pyx_t_5 = 0;
          __Pyx_XDECREF_SET(__pyx_v_tokens, __pyx_t_14);
          __pyx_t_14 = 0;
        }
        __pyx_L14:;

        /* "Aspidites/_vendor/pyparsing_extension.py":4632
 *             loc, tokens = self.expr._parse(instring, loc, doActions, callPreParse=False)
 *         except (ParseException, IndexError):
 *             if self.defaultValue is not self.__optionalNotMatched:             # <<<<<<<<<<<<<<
 *                 if self.expr.resultsName:
 *                     tokens = ParseResults([self.defaultValue])
 */
        goto __pyx_L13;
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":4639
 *                     tokens = [self.defaultValue]
 *             else:
 *                 tokens = []             # <<<<<<<<<<<<<<
 *         return loc, tokens
 * 
 */
      /*else*/ {
        __pyx_t_14 = PyList_New(0); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 4639, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_14);
        __Pyx_XDECREF_SET(__pyx_v_tokens, __pyx_t_14);
        __pyx_t_14 = 0;
      }
      __pyx_L13:;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "Aspidites/_vendor/pyparsing_extension.py":4629
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         try:             # <<<<<<<<<<<<<<
 *             loc, tokens = self.expr._parse(instring, loc, doActions, callPreParse=False)
 *         except (ParseException, IndexError):
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L8_try_end:;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4640
 *             else:
 *                 tokens = []
 *         return loc, tokens             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4640, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_v_loc);
  __Pyx_GIVEREF(__pyx_v_loc);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_loc);
  __Pyx_INCREF(__pyx_v_tokens);
  __Pyx_GIVEREF(__pyx_v_tokens);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_v_tokens);
  __pyx_r = __pyx_t_7;
  __pyx_t_7 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4628
 *         self.mayReturnEmpty = True
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         try:
 *             loc, tokens = self.expr._parse(instring, loc, doActions, callPreParse=False)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Optional.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tokens);
  __Pyx_XDECREF(__pyx_v_loc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4642
 *         return loc, tokens
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_8Optional_5__str__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_8Optional_5__str__ = {"__str__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_8Optional_5__str__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_8Optional_5__str__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_8Optional_4__str__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_8Optional_4__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":4643
 * 
 *     def __str__(self):
 *         if hasattr(self, "name"):             # <<<<<<<<<<<<<<
 *             return self.name
 * 
 */
  __pyx_t_1 = __Pyx_HasAttr(__pyx_v_self, __pyx_n_s_name); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4643, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4644
 *     def __str__(self):
 *         if hasattr(self, "name"):
 *             return self.name             # <<<<<<<<<<<<<<
 * 
 *         if self.strRepr is None:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4644, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4643
 * 
 *     def __str__(self):
 *         if hasattr(self, "name"):             # <<<<<<<<<<<<<<
 *             return self.name
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4646
 *             return self.name
 * 
 *         if self.strRepr is None:             # <<<<<<<<<<<<<<
 *             self.strRepr = "[" + str(self.expr) + "]"
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_strRepr); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = (__pyx_t_3 == Py_None);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4647
 * 
 *         if self.strRepr is None:
 *             self.strRepr = "[" + str(self.expr) + "]"             # <<<<<<<<<<<<<<
 * 
 *         return self.strRepr
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4647, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4647, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Add(__pyx_kp_s__14, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4647, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyNumber_Add(__pyx_t_3, __pyx_kp_s__16); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4647, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_strRepr, __pyx_t_4) < 0) __PYX_ERR(0, 4647, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4646
 *             return self.name
 * 
 *         if self.strRepr is None:             # <<<<<<<<<<<<<<
 *             self.strRepr = "[" + str(self.expr) + "]"
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4649
 *             self.strRepr = "[" + str(self.expr) + "]"
 * 
 *         return self.strRepr             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_strRepr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4642
 *         return loc, tokens
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Optional.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4711
 *     """
 * 
 *     def __init__(self, other, include=False, ignore=None, failOn=None):             # <<<<<<<<<<<<<<
 *         super(SkipTo, self).__init__(other)
 *         self.ignoreExpr = ignore
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_6SkipTo_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_6SkipTo_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_6SkipTo_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_6SkipTo_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_other = 0;
  PyObject *__pyx_v_include = 0;
  PyObject *__pyx_v_ignore = 0;
  PyObject *__pyx_v_failOn = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_other,&__pyx_n_s_include,&__pyx_n_s_ignore,&__pyx_n_s_failOn,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_False));
    values[3] = ((PyObject *)((PyObject *)Py_None));
    values[4] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_other)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 5, 1); __PYX_ERR(0, 4711, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_include);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ignore);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_failOn);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 4711, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_other = values[1];
    __pyx_v_include = values[2];
    __pyx_v_ignore = values[3];
    __pyx_v_failOn = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4711, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.SkipTo.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_6SkipTo___init__(__pyx_self, __pyx_v_self, __pyx_v_other, __pyx_v_include, __pyx_v_ignore, __pyx_v_failOn);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_6SkipTo___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other, PyObject *__pyx_v_include, PyObject *__pyx_v_ignore, PyObject *__pyx_v_failOn) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":4712
 * 
 *     def __init__(self, other, include=False, ignore=None, failOn=None):
 *         super(SkipTo, self).__init__(other)             # <<<<<<<<<<<<<<
 *         self.ignoreExpr = ignore
 *         self.mayReturnEmpty = True
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SkipTo); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4712, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4712, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4712, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4712, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_other) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_other);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4712, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4713
 *     def __init__(self, other, include=False, ignore=None, failOn=None):
 *         super(SkipTo, self).__init__(other)
 *         self.ignoreExpr = ignore             # <<<<<<<<<<<<<<
 *         self.mayReturnEmpty = True
 *         self.mayIndexError = False
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_ignoreExpr, __pyx_v_ignore) < 0) __PYX_ERR(0, 4713, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4714
 *         super(SkipTo, self).__init__(other)
 *         self.ignoreExpr = ignore
 *         self.mayReturnEmpty = True             # <<<<<<<<<<<<<<
 *         self.mayIndexError = False
 *         self.includeMatch = include
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_mayReturnEmpty, Py_True) < 0) __PYX_ERR(0, 4714, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4715
 *         self.ignoreExpr = ignore
 *         self.mayReturnEmpty = True
 *         self.mayIndexError = False             # <<<<<<<<<<<<<<
 *         self.includeMatch = include
 *         self.saveAsList = False
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_mayIndexError, Py_False) < 0) __PYX_ERR(0, 4715, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4716
 *         self.mayReturnEmpty = True
 *         self.mayIndexError = False
 *         self.includeMatch = include             # <<<<<<<<<<<<<<
 *         self.saveAsList = False
 *         if isinstance(failOn, str):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_includeMatch, __pyx_v_include) < 0) __PYX_ERR(0, 4716, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4717
 *         self.mayIndexError = False
 *         self.includeMatch = include
 *         self.saveAsList = False             # <<<<<<<<<<<<<<
 *         if isinstance(failOn, str):
 *             self.failOn = self._literalStringClass(failOn)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_saveAsList, Py_False) < 0) __PYX_ERR(0, 4717, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4718
 *         self.includeMatch = include
 *         self.saveAsList = False
 *         if isinstance(failOn, str):             # <<<<<<<<<<<<<<
 *             self.failOn = self._literalStringClass(failOn)
 *         else:
 */
  __pyx_t_4 = PyString_Check(__pyx_v_failOn); 
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4719
 *         self.saveAsList = False
 *         if isinstance(failOn, str):
 *             self.failOn = self._literalStringClass(failOn)             # <<<<<<<<<<<<<<
 *         else:
 *             self.failOn = failOn
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_literalStringClass); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4719, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_failOn) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_failOn);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4719, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_failOn, __pyx_t_1) < 0) __PYX_ERR(0, 4719, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4718
 *         self.includeMatch = include
 *         self.saveAsList = False
 *         if isinstance(failOn, str):             # <<<<<<<<<<<<<<
 *             self.failOn = self._literalStringClass(failOn)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4721
 *             self.failOn = self._literalStringClass(failOn)
 *         else:
 *             self.failOn = failOn             # <<<<<<<<<<<<<<
 *         self.errmsg = "No match found for " + str(self.expr)
 * 
 */
  /*else*/ {
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_failOn, __pyx_v_failOn) < 0) __PYX_ERR(0, 4721, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "Aspidites/_vendor/pyparsing_extension.py":4722
 *         else:
 *             self.failOn = failOn
 *         self.errmsg = "No match found for " + str(self.expr)             # <<<<<<<<<<<<<<
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4722, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4722, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Add(__pyx_kp_s_No_match_found_for, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4722, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_errmsg, __pyx_t_1) < 0) __PYX_ERR(0, 4722, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4711
 *     """
 * 
 *     def __init__(self, other, include=False, ignore=None, failOn=None):             # <<<<<<<<<<<<<<
 *         super(SkipTo, self).__init__(other)
 *         self.ignoreExpr = ignore
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.SkipTo.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4724
 *         self.errmsg = "No match found for " + str(self.expr)
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         startloc = loc
 *         instrlen = len(instring)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_6SkipTo_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_6SkipTo_3parseImpl = {"parseImpl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_6SkipTo_3parseImpl, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_6SkipTo_3parseImpl(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_instring = 0;
  PyObject *__pyx_v_loc = 0;
  PyObject *__pyx_v_doActions = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parseImpl (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_doActions,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 1); __PYX_ERR(0, 4724, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, 2); __PYX_ERR(0, 4724, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_doActions);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "parseImpl") < 0)) __PYX_ERR(0, 4724, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_doActions = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("parseImpl", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4724, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.SkipTo.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_6SkipTo_2parseImpl(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_doActions);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_6SkipTo_2parseImpl(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_instring, PyObject *__pyx_v_loc, PyObject *__pyx_v_doActions) {
  PyObject *__pyx_v_startloc = NULL;
  Py_ssize_t __pyx_v_instrlen;
  CYTHON_UNUSED PyObject *__pyx_v_expr = NULL;
  PyObject *__pyx_v_expr_parse = NULL;
  PyObject *__pyx_v_self_failOn_canParseNext = NULL;
  PyObject *__pyx_v_self_ignoreExpr_tryParse = NULL;
  PyObject *__pyx_v_tmploc = NULL;
  PyObject *__pyx_v_skiptext = NULL;
  PyObject *__pyx_v_skipresult = NULL;
  PyObject *__pyx_v_mat = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *(*__pyx_t_13)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parseImpl", 0);
  __Pyx_INCREF(__pyx_v_loc);

  /* "Aspidites/_vendor/pyparsing_extension.py":4725
 * 
 *     def parseImpl(self, instring, loc, doActions=True):
 *         startloc = loc             # <<<<<<<<<<<<<<
 *         instrlen = len(instring)
 *         expr = self.expr
 */
  __Pyx_INCREF(__pyx_v_loc);
  __pyx_v_startloc = __pyx_v_loc;

  /* "Aspidites/_vendor/pyparsing_extension.py":4726
 *     def parseImpl(self, instring, loc, doActions=True):
 *         startloc = loc
 *         instrlen = len(instring)             # <<<<<<<<<<<<<<
 *         expr = self.expr
 *         expr_parse = self.expr._parse
 */
  __pyx_t_1 = PyObject_Length(__pyx_v_instring); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 4726, __pyx_L1_error)
  __pyx_v_instrlen = __pyx_t_1;

  /* "Aspidites/_vendor/pyparsing_extension.py":4727
 *         startloc = loc
 *         instrlen = len(instring)
 *         expr = self.expr             # <<<<<<<<<<<<<<
 *         expr_parse = self.expr._parse
 *         self_failOn_canParseNext = self.failOn.canParseNext if self.failOn is not None else None
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4727, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_expr = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4728
 *         instrlen = len(instring)
 *         expr = self.expr
 *         expr_parse = self.expr._parse             # <<<<<<<<<<<<<<
 *         self_failOn_canParseNext = self.failOn.canParseNext if self.failOn is not None else None
 *         self_ignoreExpr_tryParse = self.ignoreExpr.tryParse if self.ignoreExpr is not None else None
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4728, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_parse); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4728, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_expr_parse = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4729
 *         expr = self.expr
 *         expr_parse = self.expr._parse
 *         self_failOn_canParseNext = self.failOn.canParseNext if self.failOn is not None else None             # <<<<<<<<<<<<<<
 *         self_ignoreExpr_tryParse = self.ignoreExpr.tryParse if self.ignoreExpr is not None else None
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_failOn); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4729, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = (__pyx_t_2 != Py_None);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if ((__pyx_t_4 != 0)) {
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_failOn); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4729, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_canParseNext); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4729, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_3 = __pyx_t_5;
    __pyx_t_5 = 0;
  } else {
    __Pyx_INCREF(Py_None);
    __pyx_t_3 = Py_None;
  }
  __pyx_v_self_failOn_canParseNext = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4730
 *         expr_parse = self.expr._parse
 *         self_failOn_canParseNext = self.failOn.canParseNext if self.failOn is not None else None
 *         self_ignoreExpr_tryParse = self.ignoreExpr.tryParse if self.ignoreExpr is not None else None             # <<<<<<<<<<<<<<
 * 
 *         tmploc = loc
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ignoreExpr); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = (__pyx_t_5 != Py_None);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if ((__pyx_t_4 != 0)) {
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ignoreExpr); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_tryParse); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_3 = __pyx_t_2;
    __pyx_t_2 = 0;
  } else {
    __Pyx_INCREF(Py_None);
    __pyx_t_3 = Py_None;
  }
  __pyx_v_self_ignoreExpr_tryParse = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4732
 *         self_ignoreExpr_tryParse = self.ignoreExpr.tryParse if self.ignoreExpr is not None else None
 * 
 *         tmploc = loc             # <<<<<<<<<<<<<<
 *         while tmploc <= instrlen:
 *             if self_failOn_canParseNext is not None:
 */
  __Pyx_INCREF(__pyx_v_loc);
  __pyx_v_tmploc = __pyx_v_loc;

  /* "Aspidites/_vendor/pyparsing_extension.py":4733
 * 
 *         tmploc = loc
 *         while tmploc <= instrlen:             # <<<<<<<<<<<<<<
 *             if self_failOn_canParseNext is not None:
 *                 # break if failOn expression matches
 */
  while (1) {
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_instrlen); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4733, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyObject_RichCompare(__pyx_v_tmploc, __pyx_t_3, Py_LE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4733, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 4733, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (!__pyx_t_4) break;

    /* "Aspidites/_vendor/pyparsing_extension.py":4734
 *         tmploc = loc
 *         while tmploc <= instrlen:
 *             if self_failOn_canParseNext is not None:             # <<<<<<<<<<<<<<
 *                 # break if failOn expression matches
 *                 if self_failOn_canParseNext(instring, tmploc):
 */
    __pyx_t_4 = (__pyx_v_self_failOn_canParseNext != Py_None);
    __pyx_t_6 = (__pyx_t_4 != 0);
    if (__pyx_t_6) {

      /* "Aspidites/_vendor/pyparsing_extension.py":4736
 *             if self_failOn_canParseNext is not None:
 *                 # break if failOn expression matches
 *                 if self_failOn_canParseNext(instring, tmploc):             # <<<<<<<<<<<<<<
 *                     break
 * 
 */
      __Pyx_INCREF(__pyx_v_self_failOn_canParseNext);
      __pyx_t_3 = __pyx_v_self_failOn_canParseNext; __pyx_t_5 = NULL;
      __pyx_t_7 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_7 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_instring, __pyx_v_tmploc};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4736, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_instring, __pyx_v_tmploc};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4736, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else
      #endif
      {
        __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 4736, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        if (__pyx_t_5) {
          __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_5); __pyx_t_5 = NULL;
        }
        __Pyx_INCREF(__pyx_v_instring);
        __Pyx_GIVEREF(__pyx_v_instring);
        PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_v_instring);
        __Pyx_INCREF(__pyx_v_tmploc);
        __Pyx_GIVEREF(__pyx_v_tmploc);
        PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_v_tmploc);
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4736, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 4736, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (__pyx_t_6) {

        /* "Aspidites/_vendor/pyparsing_extension.py":4737
 *                 # break if failOn expression matches
 *                 if self_failOn_canParseNext(instring, tmploc):
 *                     break             # <<<<<<<<<<<<<<
 * 
 *             if self_ignoreExpr_tryParse is not None:
 */
        goto __pyx_L4_break;

        /* "Aspidites/_vendor/pyparsing_extension.py":4736
 *             if self_failOn_canParseNext is not None:
 *                 # break if failOn expression matches
 *                 if self_failOn_canParseNext(instring, tmploc):             # <<<<<<<<<<<<<<
 *                     break
 * 
 */
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":4734
 *         tmploc = loc
 *         while tmploc <= instrlen:
 *             if self_failOn_canParseNext is not None:             # <<<<<<<<<<<<<<
 *                 # break if failOn expression matches
 *                 if self_failOn_canParseNext(instring, tmploc):
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":4739
 *                     break
 * 
 *             if self_ignoreExpr_tryParse is not None:             # <<<<<<<<<<<<<<
 *                 # advance past ignore expressions
 *                 while 1:
 */
    __pyx_t_6 = (__pyx_v_self_ignoreExpr_tryParse != Py_None);
    __pyx_t_4 = (__pyx_t_6 != 0);
    if (__pyx_t_4) {

      /* "Aspidites/_vendor/pyparsing_extension.py":4741
 *             if self_ignoreExpr_tryParse is not None:
 *                 # advance past ignore expressions
 *                 while 1:             # <<<<<<<<<<<<<<
 *                     try:
 *                         tmploc = self_ignoreExpr_tryParse(instring, tmploc)
 */
      while (1) {

        /* "Aspidites/_vendor/pyparsing_extension.py":4742
 *                 # advance past ignore expressions
 *                 while 1:
 *                     try:             # <<<<<<<<<<<<<<
 *                         tmploc = self_ignoreExpr_tryParse(instring, tmploc)
 *                     except ParseBaseException:
 */
        {
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
          __Pyx_XGOTREF(__pyx_t_9);
          __Pyx_XGOTREF(__pyx_t_10);
          __Pyx_XGOTREF(__pyx_t_11);
          /*try:*/ {

            /* "Aspidites/_vendor/pyparsing_extension.py":4743
 *                 while 1:
 *                     try:
 *                         tmploc = self_ignoreExpr_tryParse(instring, tmploc)             # <<<<<<<<<<<<<<
 *                     except ParseBaseException:
 *                         break
 */
            __Pyx_INCREF(__pyx_v_self_ignoreExpr_tryParse);
            __pyx_t_3 = __pyx_v_self_ignoreExpr_tryParse; __pyx_t_8 = NULL;
            __pyx_t_7 = 0;
            if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
              __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
              if (likely(__pyx_t_8)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                __Pyx_INCREF(__pyx_t_8);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_3, function);
                __pyx_t_7 = 1;
              }
            }
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_instring, __pyx_v_tmploc};
              __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4743, __pyx_L10_error)
              __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
              __Pyx_GOTREF(__pyx_t_2);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_instring, __pyx_v_tmploc};
              __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4743, __pyx_L10_error)
              __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
              __Pyx_GOTREF(__pyx_t_2);
            } else
            #endif
            {
              __pyx_t_5 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4743, __pyx_L10_error)
              __Pyx_GOTREF(__pyx_t_5);
              if (__pyx_t_8) {
                __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_8); __pyx_t_8 = NULL;
              }
              __Pyx_INCREF(__pyx_v_instring);
              __Pyx_GIVEREF(__pyx_v_instring);
              PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_7, __pyx_v_instring);
              __Pyx_INCREF(__pyx_v_tmploc);
              __Pyx_GIVEREF(__pyx_v_tmploc);
              PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_7, __pyx_v_tmploc);
              __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4743, __pyx_L10_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            }
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF_SET(__pyx_v_tmploc, __pyx_t_2);
            __pyx_t_2 = 0;

            /* "Aspidites/_vendor/pyparsing_extension.py":4742
 *                 # advance past ignore expressions
 *                 while 1:
 *                     try:             # <<<<<<<<<<<<<<
 *                         tmploc = self_ignoreExpr_tryParse(instring, tmploc)
 *                     except ParseBaseException:
 */
          }
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          goto __pyx_L17_try_end;
          __pyx_L10_error:;
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":4744
 *                     try:
 *                         tmploc = self_ignoreExpr_tryParse(instring, tmploc)
 *                     except ParseBaseException:             # <<<<<<<<<<<<<<
 *                         break
 * 
 */
          __Pyx_ErrFetch(&__pyx_t_2, &__pyx_t_3, &__pyx_t_5);
          __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_ParseBaseException); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 4744, __pyx_L12_except_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_7 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_2, __pyx_t_8);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_ErrRestore(__pyx_t_2, __pyx_t_3, __pyx_t_5);
          __pyx_t_2 = 0; __pyx_t_3 = 0; __pyx_t_5 = 0;
          if (__pyx_t_7) {
            __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.SkipTo.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
            if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_3, &__pyx_t_2) < 0) __PYX_ERR(0, 4744, __pyx_L12_except_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_GOTREF(__pyx_t_2);

            /* "Aspidites/_vendor/pyparsing_extension.py":4745
 *                         tmploc = self_ignoreExpr_tryParse(instring, tmploc)
 *                     except ParseBaseException:
 *                         break             # <<<<<<<<<<<<<<
 * 
 *             try:
 */
            goto __pyx_L18_except_break;
            __pyx_L18_except_break:;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            goto __pyx_L15_try_break;
          }
          goto __pyx_L12_except_error;
          __pyx_L12_except_error:;

          /* "Aspidites/_vendor/pyparsing_extension.py":4742
 *                 # advance past ignore expressions
 *                 while 1:
 *                     try:             # <<<<<<<<<<<<<<
 *                         tmploc = self_ignoreExpr_tryParse(instring, tmploc)
 *                     except ParseBaseException:
 */
          __Pyx_XGIVEREF(__pyx_t_9);
          __Pyx_XGIVEREF(__pyx_t_10);
          __Pyx_XGIVEREF(__pyx_t_11);
          __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
          goto __pyx_L1_error;
          __pyx_L15_try_break:;
          __Pyx_XGIVEREF(__pyx_t_9);
          __Pyx_XGIVEREF(__pyx_t_10);
          __Pyx_XGIVEREF(__pyx_t_11);
          __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
          goto __pyx_L9_break;
          __pyx_L17_try_end:;
        }
      }
      __pyx_L9_break:;

      /* "Aspidites/_vendor/pyparsing_extension.py":4739
 *                     break
 * 
 *             if self_ignoreExpr_tryParse is not None:             # <<<<<<<<<<<<<<
 *                 # advance past ignore expressions
 *                 while 1:
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":4747
 *                         break
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 expr_parse(instring, tmploc, doActions=False, callPreParse=False)
 *             except (ParseException, IndexError):
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_11, &__pyx_t_10, &__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      /*try:*/ {

        /* "Aspidites/_vendor/pyparsing_extension.py":4748
 * 
 *             try:
 *                 expr_parse(instring, tmploc, doActions=False, callPreParse=False)             # <<<<<<<<<<<<<<
 *             except (ParseException, IndexError):
 *                 # no match, advance loc in string
 */
        __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4748, __pyx_L20_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_v_instring);
        __Pyx_GIVEREF(__pyx_v_instring);
        PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_instring);
        __Pyx_INCREF(__pyx_v_tmploc);
        __Pyx_GIVEREF(__pyx_v_tmploc);
        PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_tmploc);
        __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4748, __pyx_L20_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_doActions, Py_False) < 0) __PYX_ERR(0, 4748, __pyx_L20_error)
        if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_callPreParse, Py_False) < 0) __PYX_ERR(0, 4748, __pyx_L20_error)
        __pyx_t_5 = __Pyx_PyObject_Call(__pyx_v_expr_parse, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4748, __pyx_L20_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":4747
 *                         break
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 expr_parse(instring, tmploc, doActions=False, callPreParse=False)
 *             except (ParseException, IndexError):
 */
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":4754
 *             else:
 *                 # matched skipto expr, done
 *                 break             # <<<<<<<<<<<<<<
 * 
 *         else:
 */
      /*else:*/ {
        goto __pyx_L25_try_break;
      }
      __pyx_L20_error:;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":4749
 *             try:
 *                 expr_parse(instring, tmploc, doActions=False, callPreParse=False)
 *             except (ParseException, IndexError):             # <<<<<<<<<<<<<<
 *                 # no match, advance loc in string
 *                 tmploc += 1
 */
      __Pyx_ErrFetch(&__pyx_t_5, &__pyx_t_3, &__pyx_t_2);
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 4749, __pyx_L22_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_7 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_5, __pyx_t_8) || __Pyx_PyErr_GivenExceptionMatches(__pyx_t_5, __pyx_builtin_IndexError);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_ErrRestore(__pyx_t_5, __pyx_t_3, __pyx_t_2);
      __pyx_t_5 = 0; __pyx_t_3 = 0; __pyx_t_2 = 0;
      if (__pyx_t_7) {
        __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.SkipTo.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_3, &__pyx_t_5) < 0) __PYX_ERR(0, 4749, __pyx_L22_except_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_5);

        /* "Aspidites/_vendor/pyparsing_extension.py":4751
 *             except (ParseException, IndexError):
 *                 # no match, advance loc in string
 *                 tmploc += 1             # <<<<<<<<<<<<<<
 *             else:
 *                 # matched skipto expr, done
 */
        __pyx_t_8 = __Pyx_PyInt_AddObjC(__pyx_v_tmploc, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 4751, __pyx_L22_except_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF_SET(__pyx_v_tmploc, __pyx_t_8);
        __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        goto __pyx_L21_exception_handled;
      }
      goto __pyx_L22_except_error;
      __pyx_L22_except_error:;

      /* "Aspidites/_vendor/pyparsing_extension.py":4747
 *                         break
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 expr_parse(instring, tmploc, doActions=False, callPreParse=False)
 *             except (ParseException, IndexError):
 */
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_10, __pyx_t_9);
      goto __pyx_L1_error;
      __pyx_L25_try_break:;
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_10, __pyx_t_9);
      goto __pyx_L4_break;
      __pyx_L21_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_10, __pyx_t_9);
    }
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4758
 *         else:
 *             # ran off the end of the input string without matching skipto expr, fail
 *             raise ParseException(instring, loc, self.errmsg, self)             # <<<<<<<<<<<<<<
 * 
 *         # build up return values
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4758, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_errmsg); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4758, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_8 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[5] = {__pyx_t_8, __pyx_v_instring, __pyx_v_loc, __pyx_t_2, __pyx_v_self};
      __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4758, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[5] = {__pyx_t_8, __pyx_v_instring, __pyx_v_loc, __pyx_t_2, __pyx_v_self};
      __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4758, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else
    #endif
    {
      __pyx_t_12 = PyTuple_New(4+__pyx_t_7); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 4758, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_INCREF(__pyx_v_instring);
      __Pyx_GIVEREF(__pyx_v_instring);
      PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_7, __pyx_v_instring);
      __Pyx_INCREF(__pyx_v_loc);
      __Pyx_GIVEREF(__pyx_v_loc);
      PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_7, __pyx_v_loc);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_12, 2+__pyx_t_7, __pyx_t_2);
      __Pyx_INCREF(__pyx_v_self);
      __Pyx_GIVEREF(__pyx_v_self);
      PyTuple_SET_ITEM(__pyx_t_12, 3+__pyx_t_7, __pyx_v_self);
      __pyx_t_2 = 0;
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_12, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4758, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 4758, __pyx_L1_error)
  }
  __pyx_L4_break:;

  /* "Aspidites/_vendor/pyparsing_extension.py":4761
 * 
 *         # build up return values
 *         loc = tmploc             # <<<<<<<<<<<<<<
 *         skiptext = instring[startloc:loc]
 *         skipresult = ParseResults(skiptext)
 */
  __Pyx_INCREF(__pyx_v_tmploc);
  __Pyx_DECREF_SET(__pyx_v_loc, __pyx_v_tmploc);

  /* "Aspidites/_vendor/pyparsing_extension.py":4762
 *         # build up return values
 *         loc = tmploc
 *         skiptext = instring[startloc:loc]             # <<<<<<<<<<<<<<
 *         skipresult = ParseResults(skiptext)
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_GetSlice(__pyx_v_instring, 0, 0, &__pyx_v_startloc, &__pyx_v_loc, NULL, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4762, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v_skiptext = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4763
 *         loc = tmploc
 *         skiptext = instring[startloc:loc]
 *         skipresult = ParseResults(skiptext)             # <<<<<<<<<<<<<<
 * 
 *         if self.includeMatch:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ParseResults); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4763, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_12 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_12)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_12);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_5 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_12, __pyx_v_skiptext) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_skiptext);
  __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4763, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_skipresult = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4765
 *         skipresult = ParseResults(skiptext)
 * 
 *         if self.includeMatch:             # <<<<<<<<<<<<<<
 *             loc, mat = expr_parse(instring, loc, doActions, callPreParse=False)
 *             skipresult += mat
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_includeMatch); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4765, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 4765, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4766
 * 
 *         if self.includeMatch:
 *             loc, mat = expr_parse(instring, loc, doActions, callPreParse=False)             # <<<<<<<<<<<<<<
 *             skipresult += mat
 * 
 */
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4766, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_instring);
    __Pyx_GIVEREF(__pyx_v_instring);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_instring);
    __Pyx_INCREF(__pyx_v_loc);
    __Pyx_GIVEREF(__pyx_v_loc);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_loc);
    __Pyx_INCREF(__pyx_v_doActions);
    __Pyx_GIVEREF(__pyx_v_doActions);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_doActions);
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4766, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_callPreParse, Py_False) < 0) __PYX_ERR(0, 4766, __pyx_L1_error)
    __pyx_t_12 = __Pyx_PyObject_Call(__pyx_v_expr_parse, __pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 4766, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if ((likely(PyTuple_CheckExact(__pyx_t_12))) || (PyList_CheckExact(__pyx_t_12))) {
      PyObject* sequence = __pyx_t_12;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 4766, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      #else
      __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4766, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4766, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_2 = PyObject_GetIter(__pyx_t_12); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4766, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __pyx_t_13 = Py_TYPE(__pyx_t_2)->tp_iternext;
      index = 0; __pyx_t_3 = __pyx_t_13(__pyx_t_2); if (unlikely(!__pyx_t_3)) goto __pyx_L31_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_3);
      index = 1; __pyx_t_5 = __pyx_t_13(__pyx_t_2); if (unlikely(!__pyx_t_5)) goto __pyx_L31_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_13(__pyx_t_2), 2) < 0) __PYX_ERR(0, 4766, __pyx_L1_error)
      __pyx_t_13 = NULL;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L32_unpacking_done;
      __pyx_L31_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_13 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 4766, __pyx_L1_error)
      __pyx_L32_unpacking_done:;
    }
    __Pyx_DECREF_SET(__pyx_v_loc, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_v_mat = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4767
 *         if self.includeMatch:
 *             loc, mat = expr_parse(instring, loc, doActions, callPreParse=False)
 *             skipresult += mat             # <<<<<<<<<<<<<<
 * 
 *         return loc, skipresult
 */
    __pyx_t_12 = PyNumber_InPlaceAdd(__pyx_v_skipresult, __pyx_v_mat); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 4767, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF_SET(__pyx_v_skipresult, __pyx_t_12);
    __pyx_t_12 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4765
 *         skipresult = ParseResults(skiptext)
 * 
 *         if self.includeMatch:             # <<<<<<<<<<<<<<
 *             loc, mat = expr_parse(instring, loc, doActions, callPreParse=False)
 *             skipresult += mat
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4769
 *             skipresult += mat
 * 
 *         return loc, skipresult             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 4769, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_INCREF(__pyx_v_loc);
  __Pyx_GIVEREF(__pyx_v_loc);
  PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_v_loc);
  __Pyx_INCREF(__pyx_v_skipresult);
  __Pyx_GIVEREF(__pyx_v_skipresult);
  PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_v_skipresult);
  __pyx_r = __pyx_t_12;
  __pyx_t_12 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4724
 *         self.errmsg = "No match found for " + str(self.expr)
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         startloc = loc
 *         instrlen = len(instring)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.SkipTo.parseImpl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_startloc);
  __Pyx_XDECREF(__pyx_v_expr);
  __Pyx_XDECREF(__pyx_v_expr_parse);
  __Pyx_XDECREF(__pyx_v_self_failOn_canParseNext);
  __Pyx_XDECREF(__pyx_v_self_ignoreExpr_tryParse);
  __Pyx_XDECREF(__pyx_v_tmploc);
  __Pyx_XDECREF(__pyx_v_skiptext);
  __Pyx_XDECREF(__pyx_v_skipresult);
  __Pyx_XDECREF(__pyx_v_mat);
  __Pyx_XDECREF(__pyx_v_loc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4800
 *     """
 * 
 *     def __init__(self, other=None):             # <<<<<<<<<<<<<<
 *         super(Forward, self).__init__(other, savelist=False)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Forward_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7Forward_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Forward_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Forward_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_other = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_other,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_other);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 4800, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_other = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4800, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Forward.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Forward___init__(__pyx_self, __pyx_v_self, __pyx_v_other);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Forward___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":4801
 * 
 *     def __init__(self, other=None):
 *         super(Forward, self).__init__(other, savelist=False)             # <<<<<<<<<<<<<<
 * 
 *     def __lshift__(self, other):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Forward); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_self);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_other);
  __Pyx_GIVEREF(__pyx_v_other);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_other);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_savelist, Py_False) < 0) __PYX_ERR(0, 4801, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4800
 *     """
 * 
 *     def __init__(self, other=None):             # <<<<<<<<<<<<<<
 *         super(Forward, self).__init__(other, savelist=False)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Forward.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4803
 *         super(Forward, self).__init__(other, savelist=False)
 * 
 *     def __lshift__(self, other):             # <<<<<<<<<<<<<<
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Forward_3__lshift__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7Forward_3__lshift__ = {"__lshift__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Forward_3__lshift__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Forward_3__lshift__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_other = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__lshift__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_other,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_other)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__lshift__", 1, 2, 2, 1); __PYX_ERR(0, 4803, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__lshift__") < 0)) __PYX_ERR(0, 4803, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_other = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__lshift__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4803, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Forward.__lshift__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Forward_2__lshift__(__pyx_self, __pyx_v_self, __pyx_v_other);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Forward_2__lshift__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__lshift__", 0);
  __Pyx_INCREF(__pyx_v_other);

  /* "Aspidites/_vendor/pyparsing_extension.py":4804
 * 
 *     def __lshift__(self, other):
 *         if isinstance(other, str):             # <<<<<<<<<<<<<<
 *             other = self._literalStringClass(other)
 *         self.expr = other
 */
  __pyx_t_1 = PyString_Check(__pyx_v_other); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4805
 *     def __lshift__(self, other):
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)             # <<<<<<<<<<<<<<
 *         self.expr = other
 *         self.strRepr = None
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_literalStringClass); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4805, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_other) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_other);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4805, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_other, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4804
 * 
 *     def __lshift__(self, other):
 *         if isinstance(other, str):             # <<<<<<<<<<<<<<
 *             other = self._literalStringClass(other)
 *         self.expr = other
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4806
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 *         self.expr = other             # <<<<<<<<<<<<<<
 *         self.strRepr = None
 *         self.mayIndexError = self.expr.mayIndexError
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_expr, __pyx_v_other) < 0) __PYX_ERR(0, 4806, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4807
 *             other = self._literalStringClass(other)
 *         self.expr = other
 *         self.strRepr = None             # <<<<<<<<<<<<<<
 *         self.mayIndexError = self.expr.mayIndexError
 *         self.mayReturnEmpty = self.expr.mayReturnEmpty
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_strRepr, Py_None) < 0) __PYX_ERR(0, 4807, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4808
 *         self.expr = other
 *         self.strRepr = None
 *         self.mayIndexError = self.expr.mayIndexError             # <<<<<<<<<<<<<<
 *         self.mayReturnEmpty = self.expr.mayReturnEmpty
 *         self.setWhitespaceChars(self.expr.whiteChars)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_mayIndexError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_mayIndexError, __pyx_t_4) < 0) __PYX_ERR(0, 4808, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4809
 *         self.strRepr = None
 *         self.mayIndexError = self.expr.mayIndexError
 *         self.mayReturnEmpty = self.expr.mayReturnEmpty             # <<<<<<<<<<<<<<
 *         self.setWhitespaceChars(self.expr.whiteChars)
 *         self.skipWhitespace = self.expr.skipWhitespace
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4809, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_mayReturnEmpty); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4809, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_mayReturnEmpty, __pyx_t_3) < 0) __PYX_ERR(0, 4809, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4810
 *         self.mayIndexError = self.expr.mayIndexError
 *         self.mayReturnEmpty = self.expr.mayReturnEmpty
 *         self.setWhitespaceChars(self.expr.whiteChars)             # <<<<<<<<<<<<<<
 *         self.skipWhitespace = self.expr.skipWhitespace
 *         self.saveAsList = self.expr.saveAsList
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_setWhitespaceChars); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_whiteChars); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4811
 *         self.mayReturnEmpty = self.expr.mayReturnEmpty
 *         self.setWhitespaceChars(self.expr.whiteChars)
 *         self.skipWhitespace = self.expr.skipWhitespace             # <<<<<<<<<<<<<<
 *         self.saveAsList = self.expr.saveAsList
 *         self.ignoreExprs.extend(self.expr.ignoreExprs)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_skipWhitespace); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_skipWhitespace, __pyx_t_4) < 0) __PYX_ERR(0, 4811, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4812
 *         self.setWhitespaceChars(self.expr.whiteChars)
 *         self.skipWhitespace = self.expr.skipWhitespace
 *         self.saveAsList = self.expr.saveAsList             # <<<<<<<<<<<<<<
 *         self.ignoreExprs.extend(self.expr.ignoreExprs)
 *         return self
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4812, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_saveAsList); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4812, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_saveAsList, __pyx_t_3) < 0) __PYX_ERR(0, 4812, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4813
 *         self.skipWhitespace = self.expr.skipWhitespace
 *         self.saveAsList = self.expr.saveAsList
 *         self.ignoreExprs.extend(self.expr.ignoreExprs)             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ignoreExprs); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_extend); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_ignoreExprs); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4814
 *         self.saveAsList = self.expr.saveAsList
 *         self.ignoreExprs.extend(self.expr.ignoreExprs)
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __ilshift__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4803
 *         super(Forward, self).__init__(other, savelist=False)
 * 
 *     def __lshift__(self, other):             # <<<<<<<<<<<<<<
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Forward.__lshift__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_other);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4816
 *         return self
 * 
 *     def __ilshift__(self, other):             # <<<<<<<<<<<<<<
 *         return self << other
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Forward_5__ilshift__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7Forward_5__ilshift__ = {"__ilshift__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Forward_5__ilshift__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Forward_5__ilshift__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_other = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__ilshift__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_other,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_other)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__ilshift__", 1, 2, 2, 1); __PYX_ERR(0, 4816, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__ilshift__") < 0)) __PYX_ERR(0, 4816, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_other = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__ilshift__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4816, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Forward.__ilshift__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Forward_4__ilshift__(__pyx_self, __pyx_v_self, __pyx_v_other);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Forward_4__ilshift__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__ilshift__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":4817
 * 
 *     def __ilshift__(self, other):
 *         return self << other             # <<<<<<<<<<<<<<
 * 
 *     def leaveWhitespace(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyNumber_Lshift(__pyx_v_self, __pyx_v_other); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4817, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4816
 *         return self
 * 
 *     def __ilshift__(self, other):             # <<<<<<<<<<<<<<
 *         return self << other
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Forward.__ilshift__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4819
 *         return self << other
 * 
 *     def leaveWhitespace(self):             # <<<<<<<<<<<<<<
 *         self.skipWhitespace = False
 *         return self
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Forward_7leaveWhitespace(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7Forward_7leaveWhitespace = {"leaveWhitespace", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Forward_7leaveWhitespace, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Forward_7leaveWhitespace(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("leaveWhitespace (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Forward_6leaveWhitespace(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Forward_6leaveWhitespace(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("leaveWhitespace", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":4820
 * 
 *     def leaveWhitespace(self):
 *         self.skipWhitespace = False             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_skipWhitespace, Py_False) < 0) __PYX_ERR(0, 4820, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4821
 *     def leaveWhitespace(self):
 *         self.skipWhitespace = False
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def streamline(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4819
 *         return self << other
 * 
 *     def leaveWhitespace(self):             # <<<<<<<<<<<<<<
 *         self.skipWhitespace = False
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Forward.leaveWhitespace", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4823
 *         return self
 * 
 *     def streamline(self):             # <<<<<<<<<<<<<<
 *         if not self.streamlined:
 *             self.streamlined = True
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Forward_9streamline(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7Forward_9streamline = {"streamline", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Forward_9streamline, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Forward_9streamline(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("streamline (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Forward_8streamline(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Forward_8streamline(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("streamline", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":4824
 * 
 *     def streamline(self):
 *         if not self.streamlined:             # <<<<<<<<<<<<<<
 *             self.streamlined = True
 *             if self.expr is not None:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_streamlined); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4824, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 4824, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = ((!__pyx_t_2) != 0);
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4825
 *     def streamline(self):
 *         if not self.streamlined:
 *             self.streamlined = True             # <<<<<<<<<<<<<<
 *             if self.expr is not None:
 *                 self.expr.streamline()
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_streamlined, Py_True) < 0) __PYX_ERR(0, 4825, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":4826
 *         if not self.streamlined:
 *             self.streamlined = True
 *             if self.expr is not None:             # <<<<<<<<<<<<<<
 *                 self.expr.streamline()
 *         return self
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4826, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = (__pyx_t_1 != Py_None);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_2 = (__pyx_t_3 != 0);
    if (__pyx_t_2) {

      /* "Aspidites/_vendor/pyparsing_extension.py":4827
 *             self.streamlined = True
 *             if self.expr is not None:
 *                 self.expr.streamline()             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4827, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_streamline); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4827, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4827, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":4826
 *         if not self.streamlined:
 *             self.streamlined = True
 *             if self.expr is not None:             # <<<<<<<<<<<<<<
 *                 self.expr.streamline()
 *         return self
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":4824
 * 
 *     def streamline(self):
 *         if not self.streamlined:             # <<<<<<<<<<<<<<
 *             self.streamlined = True
 *             if self.expr is not None:
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4828
 *             if self.expr is not None:
 *                 self.expr.streamline()
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def validate(self, validateTrace=None):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4823
 *         return self
 * 
 *     def streamline(self):             # <<<<<<<<<<<<<<
 *         if not self.streamlined:
 *             self.streamlined = True
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Forward.streamline", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4830
 *         return self
 * 
 *     def validate(self, validateTrace=None):             # <<<<<<<<<<<<<<
 *         if validateTrace is None:
 *             validateTrace = []
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Forward_11validate(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7Forward_11validate = {"validate", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Forward_11validate, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Forward_11validate(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_validateTrace = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("validate (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_validateTrace,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_validateTrace);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "validate") < 0)) __PYX_ERR(0, 4830, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_validateTrace = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("validate", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4830, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Forward.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Forward_10validate(__pyx_self, __pyx_v_self, __pyx_v_validateTrace);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Forward_10validate(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_validateTrace) {
  PyObject *__pyx_v_tmp = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("validate", 0);
  __Pyx_INCREF(__pyx_v_validateTrace);

  /* "Aspidites/_vendor/pyparsing_extension.py":4831
 * 
 *     def validate(self, validateTrace=None):
 *         if validateTrace is None:             # <<<<<<<<<<<<<<
 *             validateTrace = []
 * 
 */
  __pyx_t_1 = (__pyx_v_validateTrace == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4832
 *     def validate(self, validateTrace=None):
 *         if validateTrace is None:
 *             validateTrace = []             # <<<<<<<<<<<<<<
 * 
 *         if self not in validateTrace:
 */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4832, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF_SET(__pyx_v_validateTrace, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4831
 * 
 *     def validate(self, validateTrace=None):
 *         if validateTrace is None:             # <<<<<<<<<<<<<<
 *             validateTrace = []
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4834
 *             validateTrace = []
 * 
 *         if self not in validateTrace:             # <<<<<<<<<<<<<<
 *             tmp = validateTrace[:] + [self]
 *             if self.expr is not None:
 */
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_v_self, __pyx_v_validateTrace, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 4834, __pyx_L1_error)
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4835
 * 
 *         if self not in validateTrace:
 *             tmp = validateTrace[:] + [self]             # <<<<<<<<<<<<<<
 *             if self.expr is not None:
 *                 self.expr.validate(tmp)
 */
    __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_v_validateTrace, 0, 0, NULL, NULL, &__pyx_slice__11, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4835, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyList_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4835, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyList_SET_ITEM(__pyx_t_4, 0, __pyx_v_self);
    __pyx_t_5 = PyNumber_Add(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4835, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_tmp = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4836
 *         if self not in validateTrace:
 *             tmp = validateTrace[:] + [self]
 *             if self.expr is not None:             # <<<<<<<<<<<<<<
 *                 self.expr.validate(tmp)
 *         self.checkRecursion([])
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4836, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = (__pyx_t_5 != Py_None);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_2 = (__pyx_t_1 != 0);
    if (__pyx_t_2) {

      /* "Aspidites/_vendor/pyparsing_extension.py":4837
 *             tmp = validateTrace[:] + [self]
 *             if self.expr is not None:
 *                 self.expr.validate(tmp)             # <<<<<<<<<<<<<<
 *         self.checkRecursion([])
 * 
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4837, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_validate); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4837, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      __pyx_t_5 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_tmp) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_tmp);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4837, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":4836
 *         if self not in validateTrace:
 *             tmp = validateTrace[:] + [self]
 *             if self.expr is not None:             # <<<<<<<<<<<<<<
 *                 self.expr.validate(tmp)
 *         self.checkRecursion([])
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":4834
 *             validateTrace = []
 * 
 *         if self not in validateTrace:             # <<<<<<<<<<<<<<
 *             tmp = validateTrace[:] + [self]
 *             if self.expr is not None:
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4838
 *             if self.expr is not None:
 *                 self.expr.validate(tmp)
 *         self.checkRecursion([])             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_checkRecursion); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4838, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4838, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_5 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_6, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4838, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4830
 *         return self
 * 
 *     def validate(self, validateTrace=None):             # <<<<<<<<<<<<<<
 *         if validateTrace is None:
 *             validateTrace = []
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Forward.validate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tmp);
  __Pyx_XDECREF(__pyx_v_validateTrace);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4840
 *         self.checkRecursion([])
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Forward_13__str__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7Forward_13__str__ = {"__str__", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Forward_13__str__, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Forward_13__str__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Forward_12__str__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Forward_12__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_retString = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  char const *__pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":4841
 * 
 *     def __str__(self):
 *         if hasattr(self, "name"):             # <<<<<<<<<<<<<<
 *             return self.name
 *         if self.strRepr is not None:
 */
  __pyx_t_1 = __Pyx_HasAttr(__pyx_v_self, __pyx_n_s_name); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 4841, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4842
 *     def __str__(self):
 *         if hasattr(self, "name"):
 *             return self.name             # <<<<<<<<<<<<<<
 *         if self.strRepr is not None:
 *             return self.strRepr
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4842, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4841
 * 
 *     def __str__(self):
 *         if hasattr(self, "name"):             # <<<<<<<<<<<<<<
 *             return self.name
 *         if self.strRepr is not None:
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4843
 *         if hasattr(self, "name"):
 *             return self.name
 *         if self.strRepr is not None:             # <<<<<<<<<<<<<<
 *             return self.strRepr
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_strRepr); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4843, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = (__pyx_t_3 != Py_None);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4844
 *             return self.name
 *         if self.strRepr is not None:
 *             return self.strRepr             # <<<<<<<<<<<<<<
 * 
 *         # Avoid infinite recursion by setting a temporary strRepr
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_strRepr); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4844, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4843
 *         if hasattr(self, "name"):
 *             return self.name
 *         if self.strRepr is not None:             # <<<<<<<<<<<<<<
 *             return self.strRepr
 * 
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4847
 * 
 *         # Avoid infinite recursion by setting a temporary strRepr
 *         self.strRepr = ": ..."             # <<<<<<<<<<<<<<
 * 
 *         # Use the string representation of main expression.
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_strRepr, __pyx_kp_s__118) < 0) __PYX_ERR(0, 4847, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4850
 * 
 *         # Use the string representation of main expression.
 *         retString = '...'             # <<<<<<<<<<<<<<
 *         try:
 *             if self.expr is not None:
 */
  __Pyx_INCREF(__pyx_kp_s__85);
  __pyx_v_retString = __pyx_kp_s__85;

  /* "Aspidites/_vendor/pyparsing_extension.py":4851
 *         # Use the string representation of main expression.
 *         retString = '...'
 *         try:             # <<<<<<<<<<<<<<
 *             if self.expr is not None:
 *                 retString = str(self.expr)[:1000]
 */
  /*try:*/ {

    /* "Aspidites/_vendor/pyparsing_extension.py":4852
 *         retString = '...'
 *         try:
 *             if self.expr is not None:             # <<<<<<<<<<<<<<
 *                 retString = str(self.expr)[:1000]
 *             else:
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4852, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = (__pyx_t_3 != Py_None);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_2 = (__pyx_t_1 != 0);
    if (__pyx_t_2) {

      /* "Aspidites/_vendor/pyparsing_extension.py":4853
 *         try:
 *             if self.expr is not None:
 *                 retString = str(self.expr)[:1000]             # <<<<<<<<<<<<<<
 *             else:
 *                 retString = "None"
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4853, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4853, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_t_4, 0, 0x3E8, NULL, NULL, &__pyx_slice__119, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4853, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF_SET(__pyx_v_retString, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":4852
 *         retString = '...'
 *         try:
 *             if self.expr is not None:             # <<<<<<<<<<<<<<
 *                 retString = str(self.expr)[:1000]
 *             else:
 */
      goto __pyx_L8;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":4855
 *                 retString = str(self.expr)[:1000]
 *             else:
 *                 retString = "None"             # <<<<<<<<<<<<<<
 *         finally:
 *             self.strRepr = self.__class__.__name__ + ": " + retString
 */
    /*else*/ {
      __Pyx_INCREF(__pyx_n_s_None);
      __Pyx_DECREF_SET(__pyx_v_retString, __pyx_n_s_None);
    }
    __pyx_L8:;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4857
 *                 retString = "None"
 *         finally:
 *             self.strRepr = self.__class__.__name__ + ": " + retString             # <<<<<<<<<<<<<<
 *         return self.strRepr
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_class); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4857, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_name_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4857, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = PyNumber_Add(__pyx_t_4, __pyx_kp_s__120); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4857, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PyNumber_Add(__pyx_t_3, __pyx_v_retString); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4857, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_strRepr, __pyx_t_4) < 0) __PYX_ERR(0, 4857, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L7;
    }
    __pyx_L6_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10) < 0)) __Pyx_ErrFetch(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __pyx_t_5 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_7 = __pyx_filename;
      {
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_class); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4857, __pyx_L10_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_name_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4857, __pyx_L10_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyNumber_Add(__pyx_t_3, __pyx_kp_s__120); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4857, __pyx_L10_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_Add(__pyx_t_4, __pyx_v_retString); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4857, __pyx_L10_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_strRepr, __pyx_t_3) < 0) __PYX_ERR(0, 4857, __pyx_L10_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
      }
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ErrRestore(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0;
      __pyx_lineno = __pyx_t_5; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_7;
      goto __pyx_L1_error;
      __pyx_L10_error:;
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L7:;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4858
 *         finally:
 *             self.strRepr = self.__class__.__name__ + ": " + retString
 *         return self.strRepr             # <<<<<<<<<<<<<<
 * 
 *     def copy(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_strRepr); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4840
 *         self.checkRecursion([])
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Forward.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_retString);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4860
 *         return self.strRepr
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         if self.expr is not None:
 *             return super(Forward, self).copy()
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Forward_15copy(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7Forward_15copy = {"copy", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Forward_15copy, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Forward_15copy(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Forward_14copy(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Forward_14copy(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_v_ret = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":4861
 * 
 *     def copy(self):
 *         if self.expr is not None:             # <<<<<<<<<<<<<<
 *             return super(Forward, self).copy()
 *         else:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__pyx_t_1 != Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4862
 *     def copy(self):
 *         if self.expr is not None:
 *             return super(Forward, self).copy()             # <<<<<<<<<<<<<<
 *         else:
 *             ret = Forward()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_Forward); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4862, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4862, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_self);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_5, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4862, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_copy); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4862, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4862, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4861
 * 
 *     def copy(self):
 *         if self.expr is not None:             # <<<<<<<<<<<<<<
 *             return super(Forward, self).copy()
 *         else:
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4864
 *             return super(Forward, self).copy()
 *         else:
 *             ret = Forward()             # <<<<<<<<<<<<<<
 *             ret <<= self
 *             return ret
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_Forward); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4864, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4864, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_ret = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4865
 *         else:
 *             ret = Forward()
 *             ret <<= self             # <<<<<<<<<<<<<<
 *             return ret
 * 
 */
    __pyx_t_1 = PyNumber_InPlaceLshift(__pyx_v_ret, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4865, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_ret, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4866
 *             ret = Forward()
 *             ret <<= self
 *             return ret             # <<<<<<<<<<<<<<
 * 
 *     def _setResultsName(self, name, listAllMatches=False):
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_ret);
    __pyx_r = __pyx_v_ret;
    goto __pyx_L0;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4860
 *         return self.strRepr
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         if self.expr is not None:
 *             return super(Forward, self).copy()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Forward.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4868
 *             return ret
 * 
 *     def _setResultsName(self, name, listAllMatches=False):             # <<<<<<<<<<<<<<
 * 
 *         if __diag__.warn_name_set_on_empty_Forward:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Forward_17_setResultsName(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7Forward_17_setResultsName = {"_setResultsName", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Forward_17_setResultsName, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Forward_17_setResultsName(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_listAllMatches = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_setResultsName (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_name,&__pyx_n_s_listAllMatches,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_False));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_setResultsName", 0, 2, 3, 1); __PYX_ERR(0, 4868, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_listAllMatches);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_setResultsName") < 0)) __PYX_ERR(0, 4868, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_name = values[1];
    __pyx_v_listAllMatches = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_setResultsName", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4868, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Forward._setResultsName", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Forward_16_setResultsName(__pyx_self, __pyx_v_self, __pyx_v_name, __pyx_v_listAllMatches);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Forward_16_setResultsName(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_listAllMatches) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_setResultsName", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":4870
 *     def _setResultsName(self, name, listAllMatches=False):
 * 
 *         if __diag__.warn_name_set_on_empty_Forward:             # <<<<<<<<<<<<<<
 *             if self.expr is None:
 *                 warnings.warn("{0}: setting results name {0!r} on {1} expression "
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_diag); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4870, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_warn_name_set_on_empty_Forward); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4870, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 4870, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4871
 * 
 *         if __diag__.warn_name_set_on_empty_Forward:
 *             if self.expr is None:             # <<<<<<<<<<<<<<
 *                 warnings.warn("{0}: setting results name {0!r} on {1} expression "
 *                               "that has no contained expression".format("warn_name_set_on_empty_Forward",
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_expr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4871, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = (__pyx_t_2 == Py_None);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = (__pyx_t_3 != 0);
    if (__pyx_t_4) {

      /* "Aspidites/_vendor/pyparsing_extension.py":4872
 *         if __diag__.warn_name_set_on_empty_Forward:
 *             if self.expr is None:
 *                 warnings.warn("{0}: setting results name {0!r} on {1} expression "             # <<<<<<<<<<<<<<
 *                               "that has no contained expression".format("warn_name_set_on_empty_Forward",
 *                                                                         name,
 */
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_warnings); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4872, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_warn); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4872, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":4873
 *             if self.expr is None:
 *                 warnings.warn("{0}: setting results name {0!r} on {1} expression "
 *                               "that has no contained expression".format("warn_name_set_on_empty_Forward",             # <<<<<<<<<<<<<<
 *                                                                         name,
 *                                                                         type(self).__name__),
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_0_setting_results_name_0_r_on_1, __pyx_n_s_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4873, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);

      /* "Aspidites/_vendor/pyparsing_extension.py":4875
 *                               "that has no contained expression".format("warn_name_set_on_empty_Forward",
 *                                                                         name,
 *                                                                         type(self).__name__),             # <<<<<<<<<<<<<<
 *                               stacklevel=3)
 * 
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(__pyx_v_self)), __pyx_n_s_name_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4875, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = NULL;
      __pyx_t_8 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_8 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[4] = {__pyx_t_7, __pyx_n_s_warn_name_set_on_empty_Forward, __pyx_v_name, __pyx_t_6};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_8, 3+__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4873, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[4] = {__pyx_t_7, __pyx_n_s_warn_name_set_on_empty_Forward, __pyx_v_name, __pyx_t_6};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_8, 3+__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4873, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_9 = PyTuple_New(3+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 4873, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (__pyx_t_7) {
          __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;
        }
        __Pyx_INCREF(__pyx_n_s_warn_name_set_on_empty_Forward);
        __Pyx_GIVEREF(__pyx_n_s_warn_name_set_on_empty_Forward);
        PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_n_s_warn_name_set_on_empty_Forward);
        __Pyx_INCREF(__pyx_v_name);
        __Pyx_GIVEREF(__pyx_v_name);
        PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_v_name);
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_9, 2+__pyx_t_8, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_9, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4873, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":4872
 *         if __diag__.warn_name_set_on_empty_Forward:
 *             if self.expr is None:
 *                 warnings.warn("{0}: setting results name {0!r} on {1} expression "             # <<<<<<<<<<<<<<
 *                               "that has no contained expression".format("warn_name_set_on_empty_Forward",
 *                                                                         name,
 */
      __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4872, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":4876
 *                                                                         name,
 *                                                                         type(self).__name__),
 *                               stacklevel=3)             # <<<<<<<<<<<<<<
 * 
 *         return super(Forward, self)._setResultsName(name, listAllMatches)
 */
      __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4876, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_stacklevel, __pyx_int_3) < 0) __PYX_ERR(0, 4876, __pyx_L1_error)

      /* "Aspidites/_vendor/pyparsing_extension.py":4872
 *         if __diag__.warn_name_set_on_empty_Forward:
 *             if self.expr is None:
 *                 warnings.warn("{0}: setting results name {0!r} on {1} expression "             # <<<<<<<<<<<<<<
 *                               "that has no contained expression".format("warn_name_set_on_empty_Forward",
 *                                                                         name,
 */
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_5, __pyx_t_2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 4872, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":4871
 * 
 *         if __diag__.warn_name_set_on_empty_Forward:
 *             if self.expr is None:             # <<<<<<<<<<<<<<
 *                 warnings.warn("{0}: setting results name {0!r} on {1} expression "
 *                               "that has no contained expression".format("warn_name_set_on_empty_Forward",
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":4870
 *     def _setResultsName(self, name, listAllMatches=False):
 * 
 *         if __diag__.warn_name_set_on_empty_Forward:             # <<<<<<<<<<<<<<
 *             if self.expr is None:
 *                 warnings.warn("{0}: setting results name {0!r} on {1} expression "
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4878
 *                               stacklevel=3)
 * 
 *         return super(Forward, self)._setResultsName(name, listAllMatches)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Forward); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4878, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4878, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4878, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_setResultsName); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 4878, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_8 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_name, __pyx_v_listAllMatches};
    __pyx_t_9 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 4878, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_9);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_name, __pyx_v_listAllMatches};
    __pyx_t_9 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 4878, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_9);
  } else
  #endif
  {
    __pyx_t_1 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4878, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_name);
    __Pyx_GIVEREF(__pyx_v_name);
    PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_8, __pyx_v_name);
    __Pyx_INCREF(__pyx_v_listAllMatches);
    __Pyx_GIVEREF(__pyx_v_listAllMatches);
    PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_8, __pyx_v_listAllMatches);
    __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_1, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 4878, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_r = __pyx_t_9;
  __pyx_t_9 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4868
 *             return ret
 * 
 *     def _setResultsName(self, name, listAllMatches=False):             # <<<<<<<<<<<<<<
 * 
 *         if __diag__.warn_name_set_on_empty_Forward:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Forward._setResultsName", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4886
 *     """
 * 
 *     def __init__(self, expr, savelist=False):             # <<<<<<<<<<<<<<
 *         super(TokenConverter, self).__init__(expr)  # , savelist)
 *         self.saveAsList = False
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_14TokenConverter_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_14TokenConverter_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_14TokenConverter_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_14TokenConverter_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_expr = 0;
  CYTHON_UNUSED PyObject *__pyx_v_savelist = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_expr,&__pyx_n_s_savelist,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_False));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_expr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, 1); __PYX_ERR(0, 4886, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_savelist);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 4886, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_expr = values[1];
    __pyx_v_savelist = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4886, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.TokenConverter.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_14TokenConverter___init__(__pyx_self, __pyx_v_self, __pyx_v_expr, __pyx_v_savelist);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_14TokenConverter___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_expr, CYTHON_UNUSED PyObject *__pyx_v_savelist) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":4887
 * 
 *     def __init__(self, expr, savelist=False):
 *         super(TokenConverter, self).__init__(expr)  # , savelist)             # <<<<<<<<<<<<<<
 *         self.saveAsList = False
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_TokenConverter); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4887, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4887, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4887, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4887, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_expr) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_expr);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4887, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4888
 *     def __init__(self, expr, savelist=False):
 *         super(TokenConverter, self).__init__(expr)  # , savelist)
 *         self.saveAsList = False             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_saveAsList, Py_False) < 0) __PYX_ERR(0, 4888, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4886
 *     """
 * 
 *     def __init__(self, expr, savelist=False):             # <<<<<<<<<<<<<<
 *         super(TokenConverter, self).__init__(expr)  # , savelist)
 *         self.saveAsList = False
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.TokenConverter.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4910
 *     """
 * 
 *     def __init__(self, expr, joinString="", adjacent=True):             # <<<<<<<<<<<<<<
 *         super(Combine, self).__init__(expr)
 *         # suppress whitespace-stripping in contained parse expressions, but re-enable it on the Combine itself
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Combine_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7Combine_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Combine_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Combine_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_expr = 0;
  PyObject *__pyx_v_joinString = 0;
  PyObject *__pyx_v_adjacent = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_expr,&__pyx_n_s_joinString,&__pyx_n_s_adjacent,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)((PyObject*)__pyx_kp_s__12));
    values[3] = ((PyObject *)((PyObject *)Py_True));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_expr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(0, 4910, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_joinString);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_adjacent);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 4910, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_expr = values[1];
    __pyx_v_joinString = values[2];
    __pyx_v_adjacent = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4910, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Combine.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Combine___init__(__pyx_self, __pyx_v_self, __pyx_v_expr, __pyx_v_joinString, __pyx_v_adjacent);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Combine___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_expr, PyObject *__pyx_v_joinString, PyObject *__pyx_v_adjacent) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":4911
 * 
 *     def __init__(self, expr, joinString="", adjacent=True):
 *         super(Combine, self).__init__(expr)             # <<<<<<<<<<<<<<
 *         # suppress whitespace-stripping in contained parse expressions, but re-enable it on the Combine itself
 *         if adjacent:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Combine); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4911, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4911, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4911, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4911, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_expr) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_expr);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4911, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4913
 *         super(Combine, self).__init__(expr)
 *         # suppress whitespace-stripping in contained parse expressions, but re-enable it on the Combine itself
 *         if adjacent:             # <<<<<<<<<<<<<<
 *             self.leaveWhitespace()
 *         self.adjacent = adjacent
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_adjacent); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 4913, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4914
 *         # suppress whitespace-stripping in contained parse expressions, but re-enable it on the Combine itself
 *         if adjacent:
 *             self.leaveWhitespace()             # <<<<<<<<<<<<<<
 *         self.adjacent = adjacent
 *         self.skipWhitespace = True
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_leaveWhitespace); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4914, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4914, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4913
 *         super(Combine, self).__init__(expr)
 *         # suppress whitespace-stripping in contained parse expressions, but re-enable it on the Combine itself
 *         if adjacent:             # <<<<<<<<<<<<<<
 *             self.leaveWhitespace()
 *         self.adjacent = adjacent
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4915
 *         if adjacent:
 *             self.leaveWhitespace()
 *         self.adjacent = adjacent             # <<<<<<<<<<<<<<
 *         self.skipWhitespace = True
 *         self.joinString = joinString
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_adjacent, __pyx_v_adjacent) < 0) __PYX_ERR(0, 4915, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4916
 *             self.leaveWhitespace()
 *         self.adjacent = adjacent
 *         self.skipWhitespace = True             # <<<<<<<<<<<<<<
 *         self.joinString = joinString
 *         self.callPreparse = True
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_skipWhitespace, Py_True) < 0) __PYX_ERR(0, 4916, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4917
 *         self.adjacent = adjacent
 *         self.skipWhitespace = True
 *         self.joinString = joinString             # <<<<<<<<<<<<<<
 *         self.callPreparse = True
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_joinString, __pyx_v_joinString) < 0) __PYX_ERR(0, 4917, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4918
 *         self.skipWhitespace = True
 *         self.joinString = joinString
 *         self.callPreparse = True             # <<<<<<<<<<<<<<
 * 
 *     def ignore(self, other):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_callPreparse, Py_True) < 0) __PYX_ERR(0, 4918, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4910
 *     """
 * 
 *     def __init__(self, expr, joinString="", adjacent=True):             # <<<<<<<<<<<<<<
 *         super(Combine, self).__init__(expr)
 *         # suppress whitespace-stripping in contained parse expressions, but re-enable it on the Combine itself
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Combine.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4920
 *         self.callPreparse = True
 * 
 *     def ignore(self, other):             # <<<<<<<<<<<<<<
 *         if self.adjacent:
 *             ParserElement.ignore(self, other)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Combine_3ignore(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7Combine_3ignore = {"ignore", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Combine_3ignore, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Combine_3ignore(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_other = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ignore (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_other,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_other)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ignore", 1, 2, 2, 1); __PYX_ERR(0, 4920, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "ignore") < 0)) __PYX_ERR(0, 4920, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_other = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ignore", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4920, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Combine.ignore", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Combine_2ignore(__pyx_self, __pyx_v_self, __pyx_v_other);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Combine_2ignore(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ignore", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":4921
 * 
 *     def ignore(self, other):
 *         if self.adjacent:             # <<<<<<<<<<<<<<
 *             ParserElement.ignore(self, other)
 *         else:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_adjacent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4921, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 4921, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4922
 *     def ignore(self, other):
 *         if self.adjacent:
 *             ParserElement.ignore(self, other)             # <<<<<<<<<<<<<<
 *         else:
 *             super(Combine, self).ignore(other)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4922, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ignore); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4922, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_5 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_self, __pyx_v_other};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4922, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_self, __pyx_v_other};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4922, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4922, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_3) {
        __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
      }
      __Pyx_INCREF(__pyx_v_self);
      __Pyx_GIVEREF(__pyx_v_self);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_self);
      __Pyx_INCREF(__pyx_v_other);
      __Pyx_GIVEREF(__pyx_v_other);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_other);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4922, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4921
 * 
 *     def ignore(self, other):
 *         if self.adjacent:             # <<<<<<<<<<<<<<
 *             ParserElement.ignore(self, other)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4924
 *             ParserElement.ignore(self, other)
 *         else:
 *             super(Combine, self).ignore(other)             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_Combine); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4924, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4924, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4);
    __Pyx_INCREF(__pyx_v_self);
    __Pyx_GIVEREF(__pyx_v_self);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_v_self);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4924, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_ignore); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 4924, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_4, __pyx_v_other) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_other);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4924, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_L3:;

  /* "Aspidites/_vendor/pyparsing_extension.py":4925
 *         else:
 *             super(Combine, self).ignore(other)
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def postParse(self, instring, loc, tokenlist):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4920
 *         self.callPreparse = True
 * 
 *     def ignore(self, other):             # <<<<<<<<<<<<<<
 *         if self.adjacent:
 *             ParserElement.ignore(self, other)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Combine.ignore", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4927
 *         return self
 * 
 *     def postParse(self, instring, loc, tokenlist):             # <<<<<<<<<<<<<<
 *         retToks = tokenlist.copy()
 *         del retToks[:]
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Combine_5postParse(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7Combine_5postParse = {"postParse", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Combine_5postParse, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_7Combine_5postParse(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_instring = 0;
  CYTHON_UNUSED PyObject *__pyx_v_loc = 0;
  PyObject *__pyx_v_tokenlist = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("postParse (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_tokenlist,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("postParse", 1, 4, 4, 1); __PYX_ERR(0, 4927, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("postParse", 1, 4, 4, 2); __PYX_ERR(0, 4927, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tokenlist)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("postParse", 1, 4, 4, 3); __PYX_ERR(0, 4927, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "postParse") < 0)) __PYX_ERR(0, 4927, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_tokenlist = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("postParse", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4927, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Combine.postParse", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Combine_4postParse(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_tokenlist);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_7Combine_4postParse(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_instring, CYTHON_UNUSED PyObject *__pyx_v_loc, PyObject *__pyx_v_tokenlist) {
  PyObject *__pyx_v_retToks = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("postParse", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":4928
 * 
 *     def postParse(self, instring, loc, tokenlist):
 *         retToks = tokenlist.copy()             # <<<<<<<<<<<<<<
 *         del retToks[:]
 *         retToks += ParseResults(["".join(tokenlist._asStringList(self.joinString))], modal=self.modalResults)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_tokenlist, __pyx_n_s_copy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4928, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4928, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_retToks = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4929
 *     def postParse(self, instring, loc, tokenlist):
 *         retToks = tokenlist.copy()
 *         del retToks[:]             # <<<<<<<<<<<<<<
 *         retToks += ParseResults(["".join(tokenlist._asStringList(self.joinString))], modal=self.modalResults)
 * 
 */
  if (__Pyx_PyObject_DelSlice(__pyx_v_retToks, 0, 0, NULL, NULL, &__pyx_slice__11, 0, 0, 1) < 0) __PYX_ERR(0, 4929, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4930
 *         retToks = tokenlist.copy()
 *         del retToks[:]
 *         retToks += ParseResults(["".join(tokenlist._asStringList(self.joinString))], modal=self.modalResults)             # <<<<<<<<<<<<<<
 * 
 *         if self.resultsName and retToks.haskeys():
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParseResults); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_tokenlist, __pyx_n_s_asStringList); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_joinString); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyString_Join(__pyx_kp_s__12, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_modalResults); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_modal, __pyx_t_4) < 0) __PYX_ERR(0, 4930, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_retToks, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF_SET(__pyx_v_retToks, __pyx_t_2);
  __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4932
 *         retToks += ParseResults(["".join(tokenlist._asStringList(self.joinString))], modal=self.modalResults)
 * 
 *         if self.resultsName and retToks.haskeys():             # <<<<<<<<<<<<<<
 *             return [retToks]
 *         else:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_resultsName); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4932, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 4932, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_7) {
  } else {
    __pyx_t_6 = __pyx_t_7;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_retToks, __pyx_n_s_haskeys); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 4932, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4932, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 4932, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_6) {

    /* "Aspidites/_vendor/pyparsing_extension.py":4933
 * 
 *         if self.resultsName and retToks.haskeys():
 *             return [retToks]             # <<<<<<<<<<<<<<
 *         else:
 *             return retToks
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4933, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_retToks);
    __Pyx_GIVEREF(__pyx_v_retToks);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_v_retToks);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":4932
 *         retToks += ParseResults(["".join(tokenlist._asStringList(self.joinString))], modal=self.modalResults)
 * 
 *         if self.resultsName and retToks.haskeys():             # <<<<<<<<<<<<<<
 *             return [retToks]
 *         else:
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4935
 *             return [retToks]
 *         else:
 *             return retToks             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_retToks);
    __pyx_r = __pyx_v_retToks;
    goto __pyx_L0;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":4927
 *         return self
 * 
 *     def postParse(self, instring, loc, tokenlist):             # <<<<<<<<<<<<<<
 *         retToks = tokenlist.copy()
 *         del retToks[:]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Combine.postParse", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_retToks);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4954
 *     """
 * 
 *     def __init__(self, expr):             # <<<<<<<<<<<<<<
 *         super(Group, self).__init__(expr)
 *         self.saveAsList = True
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5Group_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_5Group_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5Group_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5Group_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_expr = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_expr,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_expr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(0, 4954, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 4954, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_expr = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4954, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Group.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5Group___init__(__pyx_self, __pyx_v_self, __pyx_v_expr);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5Group___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_expr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":4955
 * 
 *     def __init__(self, expr):
 *         super(Group, self).__init__(expr)             # <<<<<<<<<<<<<<
 *         self.saveAsList = True
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Group); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4955, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4955, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4955, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 4955, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_expr) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_expr);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4955, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4956
 *     def __init__(self, expr):
 *         super(Group, self).__init__(expr)
 *         self.saveAsList = True             # <<<<<<<<<<<<<<
 * 
 *     def postParse(self, instring, loc, tokenlist):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_saveAsList, Py_True) < 0) __PYX_ERR(0, 4956, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4954
 *     """
 * 
 *     def __init__(self, expr):             # <<<<<<<<<<<<<<
 *         super(Group, self).__init__(expr)
 *         self.saveAsList = True
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Group.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":4958
 *         self.saveAsList = True
 * 
 *     def postParse(self, instring, loc, tokenlist):             # <<<<<<<<<<<<<<
 *         return [tokenlist]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5Group_3postParse(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_5Group_3postParse = {"postParse", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5Group_3postParse, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_5Group_3postParse(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_instring = 0;
  CYTHON_UNUSED PyObject *__pyx_v_loc = 0;
  PyObject *__pyx_v_tokenlist = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("postParse (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_tokenlist,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("postParse", 1, 4, 4, 1); __PYX_ERR(0, 4958, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("postParse", 1, 4, 4, 2); __PYX_ERR(0, 4958, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tokenlist)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("postParse", 1, 4, 4, 3); __PYX_ERR(0, 4958, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "postParse") < 0)) __PYX_ERR(0, 4958, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_tokenlist = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("postParse", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 4958, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Group.postParse", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5Group_2postParse(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_tokenlist);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_5Group_2postParse(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_instring, CYTHON_UNUSED PyObject *__pyx_v_loc, PyObject *__pyx_v_tokenlist) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("postParse", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":4959
 * 
 *     def postParse(self, instring, loc, tokenlist):
 *         return [tokenlist]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4959, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_tokenlist);
  __Pyx_GIVEREF(__pyx_v_tokenlist);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_v_tokenlist);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4958
 *         self.saveAsList = True
 * 
 *     def postParse(self, instring, loc, tokenlist):             # <<<<<<<<<<<<<<
 *         return [tokenlist]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Group.postParse", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":5002
 *     """
 * 
 *     def __init__(self, expr):             # <<<<<<<<<<<<<<
 *         super(Dict, self).__init__(expr)
 *         self.saveAsList = True
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_4Dict_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_4Dict_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_4Dict_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_4Dict_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_expr = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_expr,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_expr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(0, 5002, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 5002, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_expr = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5002, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Dict.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Dict___init__(__pyx_self, __pyx_v_self, __pyx_v_expr);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Dict___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_expr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":5003
 * 
 *     def __init__(self, expr):
 *         super(Dict, self).__init__(expr)             # <<<<<<<<<<<<<<
 *         self.saveAsList = True
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Dict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_expr) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_expr);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":5004
 *     def __init__(self, expr):
 *         super(Dict, self).__init__(expr)
 *         self.saveAsList = True             # <<<<<<<<<<<<<<
 * 
 *     def postParse(self, instring, loc, tokenlist):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_saveAsList, Py_True) < 0) __PYX_ERR(0, 5004, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":5002
 *     """
 * 
 *     def __init__(self, expr):             # <<<<<<<<<<<<<<
 *         super(Dict, self).__init__(expr)
 *         self.saveAsList = True
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Dict.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":5006
 *         self.saveAsList = True
 * 
 *     def postParse(self, instring, loc, tokenlist):             # <<<<<<<<<<<<<<
 *         for i, tok in enumerate(tokenlist):
 *             if len(tok) == 0:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_4Dict_3postParse(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_4Dict_3postParse = {"postParse", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_4Dict_3postParse, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_4Dict_3postParse(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_instring = 0;
  CYTHON_UNUSED PyObject *__pyx_v_loc = 0;
  PyObject *__pyx_v_tokenlist = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("postParse (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_tokenlist,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("postParse", 1, 4, 4, 1); __PYX_ERR(0, 5006, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("postParse", 1, 4, 4, 2); __PYX_ERR(0, 5006, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tokenlist)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("postParse", 1, 4, 4, 3); __PYX_ERR(0, 5006, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "postParse") < 0)) __PYX_ERR(0, 5006, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_tokenlist = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("postParse", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5006, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Dict.postParse", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Dict_2postParse(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_tokenlist);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_4Dict_2postParse(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_instring, CYTHON_UNUSED PyObject *__pyx_v_loc, PyObject *__pyx_v_tokenlist) {
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_tok = NULL;
  PyObject *__pyx_v_ikey = NULL;
  PyObject *__pyx_v_dictvalue = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("postParse", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":5007
 * 
 *     def postParse(self, instring, loc, tokenlist):
 *         for i, tok in enumerate(tokenlist):             # <<<<<<<<<<<<<<
 *             if len(tok) == 0:
 *                 continue
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_t_1 = __pyx_int_0;
  if (likely(PyList_CheckExact(__pyx_v_tokenlist)) || PyTuple_CheckExact(__pyx_v_tokenlist)) {
    __pyx_t_2 = __pyx_v_tokenlist; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_tokenlist); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5007, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5007, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 5007, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5007, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 5007, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5007, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      }
    } else {
      __pyx_t_5 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 5007, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_XDECREF_SET(__pyx_v_tok, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_1);
    __pyx_t_5 = __Pyx_PyInt_AddObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5007, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1);
    __pyx_t_1 = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":5008
 *     def postParse(self, instring, loc, tokenlist):
 *         for i, tok in enumerate(tokenlist):
 *             if len(tok) == 0:             # <<<<<<<<<<<<<<
 *                 continue
 *             ikey = tok[0]
 */
    __pyx_t_6 = PyObject_Length(__pyx_v_tok); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 5008, __pyx_L1_error)
    __pyx_t_7 = ((__pyx_t_6 == 0) != 0);
    if (__pyx_t_7) {

      /* "Aspidites/_vendor/pyparsing_extension.py":5009
 *         for i, tok in enumerate(tokenlist):
 *             if len(tok) == 0:
 *                 continue             # <<<<<<<<<<<<<<
 *             ikey = tok[0]
 *             if isinstance(ikey, int):
 */
      goto __pyx_L3_continue;

      /* "Aspidites/_vendor/pyparsing_extension.py":5008
 *     def postParse(self, instring, loc, tokenlist):
 *         for i, tok in enumerate(tokenlist):
 *             if len(tok) == 0:             # <<<<<<<<<<<<<<
 *                 continue
 *             ikey = tok[0]
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":5010
 *             if len(tok) == 0:
 *                 continue
 *             ikey = tok[0]             # <<<<<<<<<<<<<<
 *             if isinstance(ikey, int):
 *                 ikey = str(tok[0]).strip()
 */
    __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_tok, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5010, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XDECREF_SET(__pyx_v_ikey, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":5011
 *                 continue
 *             ikey = tok[0]
 *             if isinstance(ikey, int):             # <<<<<<<<<<<<<<
 *                 ikey = str(tok[0]).strip()
 *             if len(tok) == 1:
 */
    __pyx_t_7 = PyInt_Check(__pyx_v_ikey); 
    __pyx_t_8 = (__pyx_t_7 != 0);
    if (__pyx_t_8) {

      /* "Aspidites/_vendor/pyparsing_extension.py":5012
 *             ikey = tok[0]
 *             if isinstance(ikey, int):
 *                 ikey = str(tok[0]).strip()             # <<<<<<<<<<<<<<
 *             if len(tok) == 1:
 *                 tokenlist[ikey] = _ParseResultsWithOffset("", i)
 */
      __pyx_t_9 = __Pyx_GetItemInt(__pyx_v_tok, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5012, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 5012, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_strip); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5012, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_9, function);
        }
      }
      __pyx_t_5 = (__pyx_t_10) ? __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_10) : __Pyx_PyObject_CallNoArg(__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5012, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF_SET(__pyx_v_ikey, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":5011
 *                 continue
 *             ikey = tok[0]
 *             if isinstance(ikey, int):             # <<<<<<<<<<<<<<
 *                 ikey = str(tok[0]).strip()
 *             if len(tok) == 1:
 */
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":5013
 *             if isinstance(ikey, int):
 *                 ikey = str(tok[0]).strip()
 *             if len(tok) == 1:             # <<<<<<<<<<<<<<
 *                 tokenlist[ikey] = _ParseResultsWithOffset("", i)
 *             elif len(tok) == 2 and not isinstance(tok[1], ParseResults):
 */
    __pyx_t_6 = PyObject_Length(__pyx_v_tok); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 5013, __pyx_L1_error)
    __pyx_t_8 = ((__pyx_t_6 == 1) != 0);
    if (__pyx_t_8) {

      /* "Aspidites/_vendor/pyparsing_extension.py":5014
 *                 ikey = str(tok[0]).strip()
 *             if len(tok) == 1:
 *                 tokenlist[ikey] = _ParseResultsWithOffset("", i)             # <<<<<<<<<<<<<<
 *             elif len(tok) == 2 and not isinstance(tok[1], ParseResults):
 *                 tokenlist[ikey] = _ParseResultsWithOffset(tok[1], i)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_ParseResultsWithOffset); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5014, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = NULL;
      __pyx_t_11 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_9, function);
          __pyx_t_11 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_9)) {
        PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_kp_s__12, __pyx_v_i};
        __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5014, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_GOTREF(__pyx_t_5);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_9)) {
        PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_kp_s__12, __pyx_v_i};
        __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5014, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_GOTREF(__pyx_t_5);
      } else
      #endif
      {
        __pyx_t_12 = PyTuple_New(2+__pyx_t_11); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 5014, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        if (__pyx_t_10) {
          __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_10); __pyx_t_10 = NULL;
        }
        __Pyx_INCREF(__pyx_kp_s__12);
        __Pyx_GIVEREF(__pyx_kp_s__12);
        PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_11, __pyx_kp_s__12);
        __Pyx_INCREF(__pyx_v_i);
        __Pyx_GIVEREF(__pyx_v_i);
        PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_11, __pyx_v_i);
        __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_12, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5014, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      }
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(PyObject_SetItem(__pyx_v_tokenlist, __pyx_v_ikey, __pyx_t_5) < 0)) __PYX_ERR(0, 5014, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":5013
 *             if isinstance(ikey, int):
 *                 ikey = str(tok[0]).strip()
 *             if len(tok) == 1:             # <<<<<<<<<<<<<<
 *                 tokenlist[ikey] = _ParseResultsWithOffset("", i)
 *             elif len(tok) == 2 and not isinstance(tok[1], ParseResults):
 */
      goto __pyx_L7;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":5015
 *             if len(tok) == 1:
 *                 tokenlist[ikey] = _ParseResultsWithOffset("", i)
 *             elif len(tok) == 2 and not isinstance(tok[1], ParseResults):             # <<<<<<<<<<<<<<
 *                 tokenlist[ikey] = _ParseResultsWithOffset(tok[1], i)
 *             else:
 */
    __pyx_t_6 = PyObject_Length(__pyx_v_tok); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 5015, __pyx_L1_error)
    __pyx_t_7 = ((__pyx_t_6 == 2) != 0);
    if (__pyx_t_7) {
    } else {
      __pyx_t_8 = __pyx_t_7;
      goto __pyx_L8_bool_binop_done;
    }
    __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_tok, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5015, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_ParseResults); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5015, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_7 = PyObject_IsInstance(__pyx_t_5, __pyx_t_9); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 5015, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_13 = ((!(__pyx_t_7 != 0)) != 0);
    __pyx_t_8 = __pyx_t_13;
    __pyx_L8_bool_binop_done:;
    if (__pyx_t_8) {

      /* "Aspidites/_vendor/pyparsing_extension.py":5016
 *                 tokenlist[ikey] = _ParseResultsWithOffset("", i)
 *             elif len(tok) == 2 and not isinstance(tok[1], ParseResults):
 *                 tokenlist[ikey] = _ParseResultsWithOffset(tok[1], i)             # <<<<<<<<<<<<<<
 *             else:
 *                 dictvalue = tok.copy()  # ParseResults(i)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_ParseResultsWithOffset); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5016, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_12 = __Pyx_GetItemInt(__pyx_v_tok, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 5016, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_t_10 = NULL;
      __pyx_t_11 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_11 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_12, __pyx_v_i};
        __pyx_t_9 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5016, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_12, __pyx_v_i};
        __pyx_t_9 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5016, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      } else
      #endif
      {
        __pyx_t_14 = PyTuple_New(2+__pyx_t_11); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 5016, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
        if (__pyx_t_10) {
          __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_10); __pyx_t_10 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_12);
        PyTuple_SET_ITEM(__pyx_t_14, 0+__pyx_t_11, __pyx_t_12);
        __Pyx_INCREF(__pyx_v_i);
        __Pyx_GIVEREF(__pyx_v_i);
        PyTuple_SET_ITEM(__pyx_t_14, 1+__pyx_t_11, __pyx_v_i);
        __pyx_t_12 = 0;
        __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_14, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5016, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(PyObject_SetItem(__pyx_v_tokenlist, __pyx_v_ikey, __pyx_t_9) < 0)) __PYX_ERR(0, 5016, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":5015
 *             if len(tok) == 1:
 *                 tokenlist[ikey] = _ParseResultsWithOffset("", i)
 *             elif len(tok) == 2 and not isinstance(tok[1], ParseResults):             # <<<<<<<<<<<<<<
 *                 tokenlist[ikey] = _ParseResultsWithOffset(tok[1], i)
 *             else:
 */
      goto __pyx_L7;
    }

    /* "Aspidites/_vendor/pyparsing_extension.py":5018
 *                 tokenlist[ikey] = _ParseResultsWithOffset(tok[1], i)
 *             else:
 *                 dictvalue = tok.copy()  # ParseResults(i)             # <<<<<<<<<<<<<<
 *                 del dictvalue[0]
 *                 if len(dictvalue) != 1 or (isinstance(dictvalue, ParseResults) and dictvalue.haskeys()):
 */
    /*else*/ {
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_tok, __pyx_n_s_copy); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5018, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_14 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_14)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_14);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_9 = (__pyx_t_14) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_14) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5018, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF_SET(__pyx_v_dictvalue, __pyx_t_9);
      __pyx_t_9 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":5019
 *             else:
 *                 dictvalue = tok.copy()  # ParseResults(i)
 *                 del dictvalue[0]             # <<<<<<<<<<<<<<
 *                 if len(dictvalue) != 1 or (isinstance(dictvalue, ParseResults) and dictvalue.haskeys()):
 *                     tokenlist[ikey] = _ParseResultsWithOffset(dictvalue, i)
 */
      if (unlikely(__Pyx_DelItemInt(__pyx_v_dictvalue, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1) < 0)) __PYX_ERR(0, 5019, __pyx_L1_error)

      /* "Aspidites/_vendor/pyparsing_extension.py":5020
 *                 dictvalue = tok.copy()  # ParseResults(i)
 *                 del dictvalue[0]
 *                 if len(dictvalue) != 1 or (isinstance(dictvalue, ParseResults) and dictvalue.haskeys()):             # <<<<<<<<<<<<<<
 *                     tokenlist[ikey] = _ParseResultsWithOffset(dictvalue, i)
 *                 else:
 */
      __pyx_t_6 = PyObject_Length(__pyx_v_dictvalue); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 5020, __pyx_L1_error)
      __pyx_t_13 = ((__pyx_t_6 != 1) != 0);
      if (!__pyx_t_13) {
      } else {
        __pyx_t_8 = __pyx_t_13;
        goto __pyx_L11_bool_binop_done;
      }
      __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_ParseResults); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5020, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_13 = PyObject_IsInstance(__pyx_v_dictvalue, __pyx_t_9); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 5020, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_7 = (__pyx_t_13 != 0);
      if (__pyx_t_7) {
      } else {
        __pyx_t_8 = __pyx_t_7;
        goto __pyx_L11_bool_binop_done;
      }
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_dictvalue, __pyx_n_s_haskeys); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5020, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_14 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_14)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_14);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_9 = (__pyx_t_14) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_14) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5020, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 5020, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_8 = __pyx_t_7;
      __pyx_L11_bool_binop_done:;
      if (__pyx_t_8) {

        /* "Aspidites/_vendor/pyparsing_extension.py":5021
 *                 del dictvalue[0]
 *                 if len(dictvalue) != 1 or (isinstance(dictvalue, ParseResults) and dictvalue.haskeys()):
 *                     tokenlist[ikey] = _ParseResultsWithOffset(dictvalue, i)             # <<<<<<<<<<<<<<
 *                 else:
 *                     tokenlist[ikey] = _ParseResultsWithOffset(dictvalue[0], i)
 */
        __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_ParseResultsWithOffset); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5021, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_14 = NULL;
        __pyx_t_11 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_14)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_14);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_11 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[3] = {__pyx_t_14, __pyx_v_dictvalue, __pyx_v_i};
          __pyx_t_9 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5021, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_GOTREF(__pyx_t_9);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[3] = {__pyx_t_14, __pyx_v_dictvalue, __pyx_v_i};
          __pyx_t_9 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5021, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_GOTREF(__pyx_t_9);
        } else
        #endif
        {
          __pyx_t_12 = PyTuple_New(2+__pyx_t_11); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 5021, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_12);
          if (__pyx_t_14) {
            __Pyx_GIVEREF(__pyx_t_14); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_14); __pyx_t_14 = NULL;
          }
          __Pyx_INCREF(__pyx_v_dictvalue);
          __Pyx_GIVEREF(__pyx_v_dictvalue);
          PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_11, __pyx_v_dictvalue);
          __Pyx_INCREF(__pyx_v_i);
          __Pyx_GIVEREF(__pyx_v_i);
          PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_11, __pyx_v_i);
          __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_12, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5021, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(PyObject_SetItem(__pyx_v_tokenlist, __pyx_v_ikey, __pyx_t_9) < 0)) __PYX_ERR(0, 5021, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":5020
 *                 dictvalue = tok.copy()  # ParseResults(i)
 *                 del dictvalue[0]
 *                 if len(dictvalue) != 1 or (isinstance(dictvalue, ParseResults) and dictvalue.haskeys()):             # <<<<<<<<<<<<<<
 *                     tokenlist[ikey] = _ParseResultsWithOffset(dictvalue, i)
 *                 else:
 */
        goto __pyx_L10;
      }

      /* "Aspidites/_vendor/pyparsing_extension.py":5023
 *                     tokenlist[ikey] = _ParseResultsWithOffset(dictvalue, i)
 *                 else:
 *                     tokenlist[ikey] = _ParseResultsWithOffset(dictvalue[0], i)             # <<<<<<<<<<<<<<
 * 
 *         if self.resultsName:
 */
      /*else*/ {
        __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_ParseResultsWithOffset); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5023, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_12 = __Pyx_GetItemInt(__pyx_v_dictvalue, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 5023, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_14 = NULL;
        __pyx_t_11 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_14)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_14);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_11 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[3] = {__pyx_t_14, __pyx_t_12, __pyx_v_i};
          __pyx_t_9 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5023, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[3] = {__pyx_t_14, __pyx_t_12, __pyx_v_i};
          __pyx_t_9 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5023, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        } else
        #endif
        {
          __pyx_t_10 = PyTuple_New(2+__pyx_t_11); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 5023, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          if (__pyx_t_14) {
            __Pyx_GIVEREF(__pyx_t_14); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_14); __pyx_t_14 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_12);
          PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_11, __pyx_t_12);
          __Pyx_INCREF(__pyx_v_i);
          __Pyx_GIVEREF(__pyx_v_i);
          PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_11, __pyx_v_i);
          __pyx_t_12 = 0;
          __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_10, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 5023, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(PyObject_SetItem(__pyx_v_tokenlist, __pyx_v_ikey, __pyx_t_9) < 0)) __PYX_ERR(0, 5023, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __pyx_L10:;
    }
    __pyx_L7:;

    /* "Aspidites/_vendor/pyparsing_extension.py":5007
 * 
 *     def postParse(self, instring, loc, tokenlist):
 *         for i, tok in enumerate(tokenlist):             # <<<<<<<<<<<<<<
 *             if len(tok) == 0:
 *                 continue
 */
    __pyx_L3_continue:;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":5025
 *                     tokenlist[ikey] = _ParseResultsWithOffset(dictvalue[0], i)
 * 
 *         if self.resultsName:             # <<<<<<<<<<<<<<
 *             return [tokenlist]
 *         else:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_resultsName); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5025, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 5025, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_8) {

    /* "Aspidites/_vendor/pyparsing_extension.py":5026
 * 
 *         if self.resultsName:
 *             return [tokenlist]             # <<<<<<<<<<<<<<
 *         else:
 *             return tokenlist
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5026, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_tokenlist);
    __Pyx_GIVEREF(__pyx_v_tokenlist);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_v_tokenlist);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":5025
 *                     tokenlist[ikey] = _ParseResultsWithOffset(dictvalue[0], i)
 * 
 *         if self.resultsName:             # <<<<<<<<<<<<<<
 *             return [tokenlist]
 *         else:
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":5028
 *             return [tokenlist]
 *         else:
 *             return tokenlist             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_tokenlist);
    __pyx_r = __pyx_v_tokenlist;
    goto __pyx_L0;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":5006
 *         self.saveAsList = True
 * 
 *     def postParse(self, instring, loc, tokenlist):             # <<<<<<<<<<<<<<
 *         for i, tok in enumerate(tokenlist):
 *             if len(tok) == 0:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Dict.postParse", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_tok);
  __Pyx_XDECREF(__pyx_v_ikey);
  __Pyx_XDECREF(__pyx_v_dictvalue);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":5054
 *     """
 * 
 *     def postParse(self, instring, loc, tokenlist):             # <<<<<<<<<<<<<<
 *         return []
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_8Suppress_1postParse(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_8Suppress_1postParse = {"postParse", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_8Suppress_1postParse, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_8Suppress_1postParse(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_instring = 0;
  CYTHON_UNUSED PyObject *__pyx_v_loc = 0;
  CYTHON_UNUSED PyObject *__pyx_v_tokenlist = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("postParse (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_instring,&__pyx_n_s_loc,&__pyx_n_s_tokenlist,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instring)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("postParse", 1, 4, 4, 1); __PYX_ERR(0, 5054, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_loc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("postParse", 1, 4, 4, 2); __PYX_ERR(0, 5054, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tokenlist)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("postParse", 1, 4, 4, 3); __PYX_ERR(0, 5054, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "postParse") < 0)) __PYX_ERR(0, 5054, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_self = values[0];
    __pyx_v_instring = values[1];
    __pyx_v_loc = values[2];
    __pyx_v_tokenlist = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("postParse", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5054, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Suppress.postParse", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_8Suppress_postParse(__pyx_self, __pyx_v_self, __pyx_v_instring, __pyx_v_loc, __pyx_v_tokenlist);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_8Suppress_postParse(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_instring, CYTHON_UNUSED PyObject *__pyx_v_loc, CYTHON_UNUSED PyObject *__pyx_v_tokenlist) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("postParse", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":5055
 * 
 *     def postParse(self, instring, loc, tokenlist):
 *         return []             # <<<<<<<<<<<<<<
 * 
 *     def suppress(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5055, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":5054
 *     """
 * 
 *     def postParse(self, instring, loc, tokenlist):             # <<<<<<<<<<<<<<
 *         return []
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.Suppress.postParse", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":5057
 *         return []
 * 
 *     def suppress(self):             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_8Suppress_3suppress(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_8Suppress_3suppress = {"suppress", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_8Suppress_3suppress, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_8Suppress_3suppress(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("suppress (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_8Suppress_2suppress(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_8Suppress_2suppress(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("suppress", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":5058
 * 
 *     def suppress(self):
 *         return self             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self);
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "Aspidites/_vendor/pyparsing_extension.py":5057
 *         return []
 * 
 *     def suppress(self):             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":5064
 *     """Wrapper for parse actions, to ensure they are only called once.
 *     """
 *     def __init__(self, methodCall):             # <<<<<<<<<<<<<<
 *         from .pyparsing import _trim_arity
 *         self.callable = _trim_arity(methodCall)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_8OnlyOnce_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_8OnlyOnce_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_8OnlyOnce_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_8OnlyOnce_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_methodCall = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_methodCall,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_methodCall)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(0, 5064, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 5064, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_methodCall = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5064, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.OnlyOnce.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_8OnlyOnce___init__(__pyx_self, __pyx_v_self, __pyx_v_methodCall);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_8OnlyOnce___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_methodCall) {
  PyObject *__pyx_v__trim_arity = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":5065
 *     """
 *     def __init__(self, methodCall):
 *         from .pyparsing import _trim_arity             # <<<<<<<<<<<<<<
 *         self.callable = _trim_arity(methodCall)
 *         self.called = False
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5065, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_trim_arity);
  __Pyx_GIVEREF(__pyx_n_s_trim_arity);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_trim_arity);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_pyparsing, __pyx_t_1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5065, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_trim_arity); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5065, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v__trim_arity = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":5066
 *     def __init__(self, methodCall):
 *         from .pyparsing import _trim_arity
 *         self.callable = _trim_arity(methodCall)             # <<<<<<<<<<<<<<
 *         self.called = False
 * 
 */
  __Pyx_INCREF(__pyx_v__trim_arity);
  __pyx_t_1 = __pyx_v__trim_arity; __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_v_methodCall) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_methodCall);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5066, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_callable, __pyx_t_2) < 0) __PYX_ERR(0, 5066, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":5067
 *         from .pyparsing import _trim_arity
 *         self.callable = _trim_arity(methodCall)
 *         self.called = False             # <<<<<<<<<<<<<<
 * 
 *     def __call__(self, s, l, t):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_called, Py_False) < 0) __PYX_ERR(0, 5067, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":5064
 *     """Wrapper for parse actions, to ensure they are only called once.
 *     """
 *     def __init__(self, methodCall):             # <<<<<<<<<<<<<<
 *         from .pyparsing import _trim_arity
 *         self.callable = _trim_arity(methodCall)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.OnlyOnce.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v__trim_arity);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":5069
 *         self.called = False
 * 
 *     def __call__(self, s, l, t):             # <<<<<<<<<<<<<<
 *         if not self.called:
 *             results = self.callable(s, l, t)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_8OnlyOnce_3__call__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_8OnlyOnce_3__call__ = {"__call__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_8OnlyOnce_3__call__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_8OnlyOnce_3__call__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_s = 0;
  PyObject *__pyx_v_l = 0;
  PyObject *__pyx_v_t = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__call__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_s_4,&__pyx_n_s_l,&__pyx_n_s_t_2,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_s_4)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__call__", 1, 4, 4, 1); __PYX_ERR(0, 5069, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_l)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__call__", 1, 4, 4, 2); __PYX_ERR(0, 5069, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_t_2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__call__", 1, 4, 4, 3); __PYX_ERR(0, 5069, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__call__") < 0)) __PYX_ERR(0, 5069, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_self = values[0];
    __pyx_v_s = values[1];
    __pyx_v_l = values[2];
    __pyx_v_t = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__call__", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 5069, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.OnlyOnce.__call__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_8OnlyOnce_2__call__(__pyx_self, __pyx_v_self, __pyx_v_s, __pyx_v_l, __pyx_v_t);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_8OnlyOnce_2__call__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_s, PyObject *__pyx_v_l, PyObject *__pyx_v_t) {
  PyObject *__pyx_v_results = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__call__", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":5070
 * 
 *     def __call__(self, s, l, t):
 *         if not self.called:             # <<<<<<<<<<<<<<
 *             results = self.callable(s, l, t)
 *             self.called = True
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_called); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5070, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 5070, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = ((!__pyx_t_2) != 0);
  if (__pyx_t_3) {

    /* "Aspidites/_vendor/pyparsing_extension.py":5071
 *     def __call__(self, s, l, t):
 *         if not self.called:
 *             results = self.callable(s, l, t)             # <<<<<<<<<<<<<<
 *             self.called = True
 *             return results
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_callable); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5071, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_s, __pyx_v_l, __pyx_v_t};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5071, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_s, __pyx_v_l, __pyx_v_t};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5071, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5071, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_INCREF(__pyx_v_s);
      __Pyx_GIVEREF(__pyx_v_s);
      PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_s);
      __Pyx_INCREF(__pyx_v_l);
      __Pyx_GIVEREF(__pyx_v_l);
      PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_l);
      __Pyx_INCREF(__pyx_v_t);
      __Pyx_GIVEREF(__pyx_v_t);
      PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_v_t);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5071, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_results = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":5072
 *         if not self.called:
 *             results = self.callable(s, l, t)
 *             self.called = True             # <<<<<<<<<<<<<<
 *             return results
 *         raise ParseException(s, l, "")
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_called, Py_True) < 0) __PYX_ERR(0, 5072, __pyx_L1_error)

    /* "Aspidites/_vendor/pyparsing_extension.py":5073
 *             results = self.callable(s, l, t)
 *             self.called = True
 *             return results             # <<<<<<<<<<<<<<
 *         raise ParseException(s, l, "")
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_results);
    __pyx_r = __pyx_v_results;
    goto __pyx_L0;

    /* "Aspidites/_vendor/pyparsing_extension.py":5070
 * 
 *     def __call__(self, s, l, t):
 *         if not self.called:             # <<<<<<<<<<<<<<
 *             results = self.callable(s, l, t)
 *             self.called = True
 */
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":5074
 *             self.called = True
 *             return results
 *         raise ParseException(s, l, "")             # <<<<<<<<<<<<<<
 * 
 *     def reset(self):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ParseException); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5074, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_4)) {
    PyObject *__pyx_temp[4] = {__pyx_t_7, __pyx_v_s, __pyx_v_l, __pyx_kp_s__12};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5074, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
    PyObject *__pyx_temp[4] = {__pyx_t_7, __pyx_v_s, __pyx_v_l, __pyx_kp_s__12};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5074, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5074, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_7) {
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
    }
    __Pyx_INCREF(__pyx_v_s);
    __Pyx_GIVEREF(__pyx_v_s);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_v_s);
    __Pyx_INCREF(__pyx_v_l);
    __Pyx_GIVEREF(__pyx_v_l);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_v_l);
    __Pyx_INCREF(__pyx_kp_s__12);
    __Pyx_GIVEREF(__pyx_kp_s__12);
    PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_6, __pyx_kp_s__12);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5074, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 5074, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":5069
 *         self.called = False
 * 
 *     def __call__(self, s, l, t):             # <<<<<<<<<<<<<<
 *         if not self.called:
 *             results = self.callable(s, l, t)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.OnlyOnce.__call__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_results);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "Aspidites/_vendor/pyparsing_extension.py":5076
 *         raise ParseException(s, l, "")
 * 
 *     def reset(self):             # <<<<<<<<<<<<<<
 *         self.called = False
 */

/* Python wrapper */
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_8OnlyOnce_5reset(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_8OnlyOnce_5reset = {"reset", (PyCFunction)__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_8OnlyOnce_5reset, METH_O, 0};
static PyObject *__pyx_pw_9Aspidites_7_vendor_19pyparsing_extension_8OnlyOnce_5reset(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("reset (wrapper)", 0);
  __pyx_r = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_8OnlyOnce_4reset(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_8OnlyOnce_4reset(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("reset", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":5077
 * 
 *     def reset(self):
 *         self.called = False             # <<<<<<<<<<<<<<
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_called, Py_False) < 0) __PYX_ERR(0, 5077, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":5076
 *         raise ParseException(s, l, "")
 * 
 *     def reset(self):             # <<<<<<<<<<<<<<
 *         self.called = False
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension.OnlyOnce.reset", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct__genexpr *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct__genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct__genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct__genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct__genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct__genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct__genexpr[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct__genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct__genexpr));
    (void) PyObject_INIT(o, t);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct__genexpr(PyObject *o) {
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct__genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct__genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct__genexpr[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct__genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct__genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct__genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct__genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct__genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct__genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct__genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_c);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr *)o;
  if (p->__pyx_v_c) {
    e = (*v)(p->__pyx_v_c, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_1_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_s);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr *)o;
  if (p->__pyx_v_s) {
    e = (*v)(p->__pyx_v_s, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_2_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues *)o;
  if (p->__pyx_v_self) {
    e = (*v)(p->__pyx_v_self, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_3__itervalues", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_k);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_k) {
    e = (*v)(p->__pyx_v_k, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_4_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems *)o;
  if (p->__pyx_v_self) {
    e = (*v)(p->__pyx_v_self, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_5__iteritems", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_k);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_k) {
    e = (*v)(p->__pyx_v_k, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_6_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_7___iadd__ *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_7___iadd__[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_7___iadd__ = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_7___iadd__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_7___iadd__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_7___iadd__)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_7___iadd__[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_7___iadd__];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_7___iadd__));
    (void) PyObject_INIT(o, t);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_7___iadd__(PyObject *o) {
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_7___iadd__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_7___iadd__)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_7___iadd__[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_7___iadd__++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_7___iadd__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_7___iadd__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_7___iadd__", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_7___iadd__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_7___iadd__, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_7___iadd__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__ *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__ = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__ *)o;
  if (p->__pyx_v_self) {
    e = (*v)(p->__pyx_v_self, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__ *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_8___str__", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_i);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_i) {
    e = (*v)(p->__pyx_v_i, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_9_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_item_fn);
  Py_CLEAR(p->__pyx_v_toItem);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict *)o;
  if (p->__pyx_v_item_fn) {
    e = (*v)(p->__pyx_v_item_fn, a); if (e) return e;
  }
  if (p->__pyx_v_toItem) {
    e = (*v)(p->__pyx_v_toItem, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict *)o;
  tmp = ((PyObject*)p->__pyx_v_item_fn);
  p->__pyx_v_item_fn = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_toItem);
  p->__pyx_v_toItem = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_10_asDict", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_k);
  Py_CLEAR(p->__pyx_v_v);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_k) {
    e = (*v)(p->__pyx_v_k, a); if (e) return e;
  }
  if (p->__pyx_v_v) {
    e = (*v)(p->__pyx_v_v, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_11_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML *)o;
  if (p->__pyx_v_self) {
    e = (*v)(p->__pyx_v_self, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_12_asXML", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_k);
  Py_CLEAR(p->__pyx_v_v);
  Py_CLEAR(p->__pyx_v_vlist);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_k) {
    e = (*v)(p->__pyx_v_k, a); if (e) return e;
  }
  if (p->__pyx_v_v) {
    e = (*v)(p->__pyx_v_v, a); if (e) return e;
  }
  if (p->__pyx_v_vlist) {
    e = (*v)(p->__pyx_v_vlist, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_13_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump *)o;
  if (p->__pyx_v_self) {
    e = (*v)(p->__pyx_v_self, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_14_dump", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_k);
  Py_CLEAR(p->__pyx_v_v);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_k) {
    e = (*v)(p->__pyx_v_k, a); if (e) return e;
  }
  if (p->__pyx_v_v) {
    e = (*v)(p->__pyx_v_v, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_15_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_vv);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_vv) {
    e = (*v)(p->__pyx_v_vv, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_16_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v__parseMethod);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak *)o;
  if (p->__pyx_v__parseMethod) {
    e = (*v)(p->__pyx_v__parseMethod, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak *)o;
  tmp = ((PyObject*)p->__pyx_v__parseMethod);
  p->__pyx_v__parseMethod = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_17_setBreak", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_fns);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction *)o;
  if (p->__pyx_v_fns) {
    e = (*v)(p->__pyx_v_fns, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction *)o;
  tmp = ((PyObject*)p->__pyx_v_fns);
  p->__pyx_v_fns = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_18_setParseAction", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_fn);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_fn) {
    e = (*v)(p->__pyx_v_fn, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_19_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__ *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__ = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_cache);
  Py_CLEAR(p->__pyx_v_not_in_cache);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__ *)o;
  if (p->__pyx_v_cache) {
    e = (*v)(p->__pyx_v_cache, a); if (e) return e;
  }
  if (p->__pyx_v_not_in_cache) {
    e = (*v)(p->__pyx_v_not_in_cache, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__ *)o;
  tmp = ((PyObject*)p->__pyx_v_cache);
  p->__pyx_v_cache = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_not_in_cache);
  p->__pyx_v_not_in_cache = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_20___init__", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__ *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__ = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_cache);
  Py_CLEAR(p->__pyx_v_not_in_cache);
  Py_CLEAR(p->__pyx_v_size);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__ *)o;
  if (p->__pyx_v_cache) {
    e = (*v)(p->__pyx_v_cache, a); if (e) return e;
  }
  if (p->__pyx_v_not_in_cache) {
    e = (*v)(p->__pyx_v_not_in_cache, a); if (e) return e;
  }
  if (p->__pyx_v_size) {
    e = (*v)(p->__pyx_v_size, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__ *)o;
  tmp = ((PyObject*)p->__pyx_v_cache);
  p->__pyx_v_cache = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_not_in_cache);
  p->__pyx_v_not_in_cache = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_size);
  p->__pyx_v_size = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_21___init__", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__ *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__ = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_cache);
  Py_CLEAR(p->__pyx_v_key_fifo);
  Py_CLEAR(p->__pyx_v_not_in_cache);
  Py_CLEAR(p->__pyx_v_size);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__ *)o;
  if (p->__pyx_v_cache) {
    e = (*v)(p->__pyx_v_cache, a); if (e) return e;
  }
  if (p->__pyx_v_key_fifo) {
    e = (*v)(p->__pyx_v_key_fifo, a); if (e) return e;
  }
  if (p->__pyx_v_not_in_cache) {
    e = (*v)(p->__pyx_v_not_in_cache, a); if (e) return e;
  }
  if (p->__pyx_v_size) {
    e = (*v)(p->__pyx_v_size, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__ *)o;
  tmp = ((PyObject*)p->__pyx_v_cache);
  p->__pyx_v_cache = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_key_fifo);
  p->__pyx_v_key_fifo = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_not_in_cache);
  p->__pyx_v_not_in_cache = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_size);
  p->__pyx_v_size = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_22___init__", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_e);
  Py_CLEAR(p->__pyx_v_exc);
  Py_CLEAR(p->__pyx_v_instring);
  Py_CLEAR(p->__pyx_v_loc);
  Py_CLEAR(p->__pyx_v_matches);
  Py_CLEAR(p->__pyx_v_maxMatches);
  Py_CLEAR(p->__pyx_v_nextLoc);
  Py_CLEAR(p->__pyx_v_nextloc);
  Py_CLEAR(p->__pyx_v_overlap);
  Py_CLEAR(p->__pyx_v_parseFn);
  Py_CLEAR(p->__pyx_v_preloc);
  Py_CLEAR(p->__pyx_v_preparseFn);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_tokens);
  Py_CLEAR(p->__pyx_t_0);
  Py_CLEAR(p->__pyx_t_1);
  Py_CLEAR(p->__pyx_t_2);
  Py_CLEAR(p->__pyx_t_3);
  Py_CLEAR(p->__pyx_t_4);
  Py_CLEAR(p->__pyx_t_5);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString *)o;
  if (p->__pyx_v_e) {
    e = (*v)(p->__pyx_v_e, a); if (e) return e;
  }
  if (p->__pyx_v_exc) {
    e = (*v)(p->__pyx_v_exc, a); if (e) return e;
  }
  if (p->__pyx_v_instring) {
    e = (*v)(p->__pyx_v_instring, a); if (e) return e;
  }
  if (p->__pyx_v_loc) {
    e = (*v)(p->__pyx_v_loc, a); if (e) return e;
  }
  if (p->__pyx_v_matches) {
    e = (*v)(p->__pyx_v_matches, a); if (e) return e;
  }
  if (p->__pyx_v_maxMatches) {
    e = (*v)(p->__pyx_v_maxMatches, a); if (e) return e;
  }
  if (p->__pyx_v_nextLoc) {
    e = (*v)(p->__pyx_v_nextLoc, a); if (e) return e;
  }
  if (p->__pyx_v_nextloc) {
    e = (*v)(p->__pyx_v_nextloc, a); if (e) return e;
  }
  if (p->__pyx_v_overlap) {
    e = (*v)(p->__pyx_v_overlap, a); if (e) return e;
  }
  if (p->__pyx_v_parseFn) {
    e = (*v)(p->__pyx_v_parseFn, a); if (e) return e;
  }
  if (p->__pyx_v_preloc) {
    e = (*v)(p->__pyx_v_preloc, a); if (e) return e;
  }
  if (p->__pyx_v_preparseFn) {
    e = (*v)(p->__pyx_v_preparseFn, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(p->__pyx_v_self, a); if (e) return e;
  }
  if (p->__pyx_v_tokens) {
    e = (*v)(p->__pyx_v_tokens, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  if (p->__pyx_t_1) {
    e = (*v)(p->__pyx_t_1, a); if (e) return e;
  }
  if (p->__pyx_t_2) {
    e = (*v)(p->__pyx_t_2, a); if (e) return e;
  }
  if (p->__pyx_t_3) {
    e = (*v)(p->__pyx_t_3, a); if (e) return e;
  }
  if (p->__pyx_t_4) {
    e = (*v)(p->__pyx_t_4, a); if (e) return e;
  }
  if (p->__pyx_t_5) {
    e = (*v)(p->__pyx_t_5, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_23_scanString", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_e);
  Py_CLEAR(p->__pyx_v_includeSeparators);
  Py_CLEAR(p->__pyx_v_instring);
  Py_CLEAR(p->__pyx_v_last);
  Py_CLEAR(p->__pyx_v_maxsplit);
  Py_CLEAR(p->__pyx_v_s);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_t);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split *)o;
  if (p->__pyx_v_e) {
    e = (*v)(p->__pyx_v_e, a); if (e) return e;
  }
  if (p->__pyx_v_includeSeparators) {
    e = (*v)(p->__pyx_v_includeSeparators, a); if (e) return e;
  }
  if (p->__pyx_v_instring) {
    e = (*v)(p->__pyx_v_instring, a); if (e) return e;
  }
  if (p->__pyx_v_last) {
    e = (*v)(p->__pyx_v_last, a); if (e) return e;
  }
  if (p->__pyx_v_maxsplit) {
    e = (*v)(p->__pyx_v_maxsplit, a); if (e) return e;
  }
  if (p->__pyx_v_s) {
    e = (*v)(p->__pyx_v_s, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(p->__pyx_v_self, a); if (e) return e;
  }
  if (p->__pyx_v_t) {
    e = (*v)(p->__pyx_v_t, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_24_split", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__ *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__ = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_makeOptionalList);
  Py_CLEAR(p->__pyx_v_other);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__ *)o;
  if (p->__pyx_v_makeOptionalList) {
    e = (*v)(p->__pyx_v_makeOptionalList, a); if (e) return e;
  }
  if (p->__pyx_v_other) {
    e = (*v)(p->__pyx_v_other, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(p->__pyx_v_self, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__ *)o;
  tmp = ((PyObject*)p->__pyx_v_makeOptionalList);
  p->__pyx_v_makeOptionalList = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_other);
  p->__pyx_v_other = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_25___mul__", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_o);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_o) {
    e = (*v)(p->__pyx_v_o, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_26_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__ *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__ = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_exprs);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__ *)o;
  if (p->__pyx_v_exprs) {
    e = (*v)(p->__pyx_v_exprs, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(p->__pyx_v_self, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__ *)o;
  tmp = ((PyObject*)p->__pyx_v_exprs);
  p->__pyx_v_exprs = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_27___init__", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_expr);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_expr) {
    e = (*v)(p->__pyx_v_expr, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_28_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_e);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_e) {
    e = (*v)(p->__pyx_v_e, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_29_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__ *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__ = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_bodyChars);
  Py_CLEAR(p->__pyx_v_excludeChars);
  Py_CLEAR(p->__pyx_v_initChars);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__ *)o;
  if (p->__pyx_v_bodyChars) {
    e = (*v)(p->__pyx_v_bodyChars, a); if (e) return e;
  }
  if (p->__pyx_v_excludeChars) {
    e = (*v)(p->__pyx_v_excludeChars, a); if (e) return e;
  }
  if (p->__pyx_v_initChars) {
    e = (*v)(p->__pyx_v_initChars, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__ *)o;
  tmp = ((PyObject*)p->__pyx_v_bodyChars);
  p->__pyx_v_bodyChars = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_excludeChars);
  p->__pyx_v_excludeChars = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_initChars);
  p->__pyx_v_initChars = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_30___init__", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_c);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_c) {
    e = (*v)(p->__pyx_v_c, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_31_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_c);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_c) {
    e = (*v)(p->__pyx_v_c, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_32_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_repl);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub *)o;
  if (p->__pyx_v_repl) {
    e = (*v)(p->__pyx_v_repl, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(p->__pyx_v_self, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub *)o;
  tmp = ((PyObject*)p->__pyx_v_repl);
  p->__pyx_v_repl = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_33_sub", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__ *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__ = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__ *)o;
  if (p->__pyx_v_self) {
    e = (*v)(p->__pyx_v_self, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__ *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_34___init__", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_35_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__ *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__ = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__ *)o;
  if (p->__pyx_v_self) {
    e = (*v)(p->__pyx_v_self, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__ *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_36___init__", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_c);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_c) {
    e = (*v)(p->__pyx_v_c, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_37_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_c);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_c) {
    e = (*v)(p->__pyx_v_c, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_38_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__ *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__ = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__ *)o;
  if (p->__pyx_v_self) {
    e = (*v)(p->__pyx_v_self, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__ *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_39___init__", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_e);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_e) {
    e = (*v)(p->__pyx_v_e, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_40_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v__PendingSkip);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline *)o;
  if (p->__pyx_v__PendingSkip) {
    e = (*v)(p->__pyx_v__PendingSkip, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(p->__pyx_v_self, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline *)o;
  tmp = ((PyObject*)p->__pyx_v__PendingSkip);
  p->__pyx_v__PendingSkip = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_41_streamline", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_e);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_e) {
    e = (*v)(p->__pyx_v_e, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_42_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_e);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_e) {
    e = (*v)(p->__pyx_v_e, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_43_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__ *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__ = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__ *)o;
  if (p->__pyx_v_self) {
    e = (*v)(p->__pyx_v_self, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__ *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_44___str__", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_e);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_e) {
    e = (*v)(p->__pyx_v_e, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_45_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__ *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__ = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__ *)o;
  if (p->__pyx_v_self) {
    e = (*v)(p->__pyx_v_self, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__ *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_46___init__", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_e);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_e) {
    e = (*v)(p->__pyx_v_e, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_47_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline *)o;
  if (p->__pyx_v_self) {
    e = (*v)(p->__pyx_v_self, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_48_streamline", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_e);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_e) {
    e = (*v)(p->__pyx_v_e, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_49_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__ *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__ = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__ *)o;
  if (p->__pyx_v_self) {
    e = (*v)(p->__pyx_v_self, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__ *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_50___str__", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_e);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_e) {
    e = (*v)(p->__pyx_v_e, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_51_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName *)o;
  if (p->__pyx_v_self) {
    e = (*v)(p->__pyx_v_self, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_52__setResultsName", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_e);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_e) {
    e = (*v)(p->__pyx_v_e, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_53_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__ *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__ = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__ *)o;
  if (p->__pyx_v_self) {
    e = (*v)(p->__pyx_v_self, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__ *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_54___init__", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_e);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_e) {
    e = (*v)(p->__pyx_v_e, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_55_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline *)o;
  if (p->__pyx_v_self) {
    e = (*v)(p->__pyx_v_self, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_56_streamline", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_e);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_e) {
    e = (*v)(p->__pyx_v_e, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_57_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__ *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__ = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__ *)o;
  if (p->__pyx_v_self) {
    e = (*v)(p->__pyx_v_self, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__ *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_58___str__", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_e);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_e) {
    e = (*v)(p->__pyx_v_e, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_59_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName *)o;
  if (p->__pyx_v_self) {
    e = (*v)(p->__pyx_v_self, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_60__setResultsName", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_e);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_e) {
    e = (*v)(p->__pyx_v_e, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_61_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__ *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__ = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__ *)o;
  if (p->__pyx_v_self) {
    e = (*v)(p->__pyx_v_self, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__ *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_62___init__", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_e);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_e) {
    e = (*v)(p->__pyx_v_e, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_63_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline *)o;
  if (p->__pyx_v_self) {
    e = (*v)(p->__pyx_v_self, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_64_streamline", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_e);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_e) {
    e = (*v)(p->__pyx_v_e, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_65_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_tmpReqd);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl *)o;
  if (p->__pyx_v_self) {
    e = (*v)(p->__pyx_v_self, a); if (e) return e;
  }
  if (p->__pyx_v_tmpReqd) {
    e = (*v)(p->__pyx_v_tmpReqd, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_tmpReqd);
  p->__pyx_v_tmpReqd = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_66_parseImpl", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_e);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_e) {
    e = (*v)(p->__pyx_v_e, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_67_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_e);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_e) {
    e = (*v)(p->__pyx_v_e, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_68_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__ *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__ = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__ *)o;
  if (p->__pyx_v_self) {
    e = (*v)(p->__pyx_v_self, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__ *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__ *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_69___str__", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__, /*tp_traverse*/
  __pyx_tp_clear_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr *__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr[8];
static int __pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr = 0;

static PyObject *__pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr[--__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr(PyObject *o) {
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_e);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr)))) {
    __pyx_freelist_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr[__pyx_freecount_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr++] = ((struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr *p = (struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_e) {
    e = (*v)(p->__pyx_v_e, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "Aspidites._vendor.pyparsing_extension.__pyx_scope_struct_70_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_pyparsing_extension(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_pyparsing_extension},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "pyparsing_extension",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_s_, __pyx_k_, sizeof(__pyx_k_), 0, 0, 1, 0},
  {&__pyx_kp_s_0, __pyx_k_0, sizeof(__pyx_k_0), 0, 0, 1, 0},
  {&__pyx_kp_s_0123456789, __pyx_k_0123456789, sizeof(__pyx_k_0123456789), 0, 0, 1, 0},
  {&__pyx_kp_s_0_1, __pyx_k_0_1, sizeof(__pyx_k_0_1), 0, 0, 1, 0},
  {&__pyx_kp_s_0_1_2, __pyx_k_0_1_2, sizeof(__pyx_k_0_1_2), 0, 0, 1, 0},
  {&__pyx_kp_s_0_1_3, __pyx_k_0_1_3, sizeof(__pyx_k_0_1_3), 0, 0, 1, 0},
  {&__pyx_kp_s_0_2, __pyx_k_0_2, sizeof(__pyx_k_0_2), 0, 0, 1, 0},
  {&__pyx_kp_s_0_failed_1_2, __pyx_k_0_failed_1_2, sizeof(__pyx_k_0_failed_1_2), 0, 0, 1, 0},
  {&__pyx_kp_s_0_setting_results_name_0_r_on_1, __pyx_k_0_setting_results_name_0_r_on_1, sizeof(__pyx_k_0_setting_results_name_0_r_on_1), 0, 0, 1, 0},
  {&__pyx_kp_s_0_setting_results_name_1_r_on_2, __pyx_k_0_setting_results_name_1_r_on_2, sizeof(__pyx_k_0_setting_results_name_1_r_on_2), 0, 0, 1, 0},
  {&__pyx_kp_s_0_setting_results_name_1_r_on_2_2, __pyx_k_0_setting_results_name_1_r_on_2_2, sizeof(__pyx_k_0_setting_results_name_1_r_on_2_2), 0, 0, 1, 0},
  {&__pyx_n_s_ABCDEFabcdef, __pyx_k_ABCDEFabcdef, sizeof(__pyx_k_ABCDEFabcdef), 0, 0, 1, 1},
  {&__pyx_kp_s_A_short_cut_class_for_defining_W, __pyx_k_A_short_cut_class_for_defining_W, sizeof(__pyx_k_A_short_cut_class_for_defining_W), 0, 0, 1, 0},
  {&__pyx_kp_s_A_token_that_will_never_match, __pyx_k_A_token_that_will_never_match, sizeof(__pyx_k_A_token_that_will_never_match), 0, 0, 1, 0},
  {&__pyx_kp_s_A_variation_on_class_Literal_whi, __pyx_k_A_variation_on_class_Literal_whi, sizeof(__pyx_k_A_variation_on_class_Literal_whi), 0, 0, 1, 0},
  {&__pyx_kp_s_Abstract_base_level_parser_eleme, __pyx_k_Abstract_base_level_parser_eleme, sizeof(__pyx_k_Abstract_base_level_parser_eleme), 0, 0, 1, 0},
  {&__pyx_kp_s_Abstract_class_ParserElement_sub, __pyx_k_Abstract_class_ParserElement_sub, sizeof(__pyx_k_Abstract_class_ParserElement_sub), 0, 0, 1, 0},
  {&__pyx_kp_s_Abstract_subclass_of_ParserEleme, __pyx_k_Abstract_subclass_of_ParserEleme, sizeof(__pyx_k_Abstract_subclass_of_ParserEleme), 0, 0, 1, 0},
  {&__pyx_kp_s_Abstract_subclass_of_class_Pars, __pyx_k_Abstract_subclass_of_class_Pars, sizeof(__pyx_k_Abstract_subclass_of_class_Pars), 0, 0, 1, 0},
  {&__pyx_kp_s_Abstract_subclass_of_class_Parse, __pyx_k_Abstract_subclass_of_class_Parse, sizeof(__pyx_k_Abstract_subclass_of_class_Parse), 0, 0, 1, 0},
  {&__pyx_kp_s_An_empty_token_will_always_match, __pyx_k_An_empty_token_will_always_match, sizeof(__pyx_k_An_empty_token_will_always_match), 0, 0, 1, 0},
  {&__pyx_n_s_And, __pyx_k_And, sizeof(__pyx_k_And), 0, 0, 1, 1},
  {&__pyx_n_s_And__ErrorStop, __pyx_k_And__ErrorStop, sizeof(__pyx_k_And__ErrorStop), 0, 0, 1, 1},
  {&__pyx_n_s_And__ErrorStop___init, __pyx_k_And__ErrorStop___init, sizeof(__pyx_k_And__ErrorStop___init), 0, 0, 1, 1},
  {&__pyx_n_s_And___iadd, __pyx_k_And___iadd, sizeof(__pyx_k_And___iadd), 0, 0, 1, 1},
  {&__pyx_n_s_And___init, __pyx_k_And___init, sizeof(__pyx_k_And___init), 0, 0, 1, 1},
  {&__pyx_n_s_And___init___locals_genexpr, __pyx_k_And___init___locals_genexpr, sizeof(__pyx_k_And___init___locals_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_And___str, __pyx_k_And___str, sizeof(__pyx_k_And___str), 0, 0, 1, 1},
  {&__pyx_n_s_And___str___locals_genexpr, __pyx_k_And___str___locals_genexpr, sizeof(__pyx_k_And___str___locals_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_And_checkRecursion, __pyx_k_And_checkRecursion, sizeof(__pyx_k_And_checkRecursion), 0, 0, 1, 1},
  {&__pyx_n_s_And_parseImpl, __pyx_k_And_parseImpl, sizeof(__pyx_k_And_parseImpl), 0, 0, 1, 1},
  {&__pyx_n_s_And_streamline, __pyx_k_And_streamline, sizeof(__pyx_k_And_streamline), 0, 0, 1, 1},
  {&__pyx_n_s_And_streamline_locals_genexpr, __pyx_k_And_streamline_locals_genexpr, sizeof(__pyx_k_And_streamline_locals_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_k_Aspidites__vendor_pyparsing_exte, sizeof(__pyx_k_Aspidites__vendor_pyparsing_exte), 0, 0, 1, 1},
  {&__pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_k_Aspidites__vendor_pyparsing_exte_2, sizeof(__pyx_k_Aspidites__vendor_pyparsing_exte_2), 0, 0, 1, 0},
  {&__pyx_n_s_AttributeError, __pyx_k_AttributeError, sizeof(__pyx_k_AttributeError), 0, 0, 1, 1},
  {&__pyx_n_s_BOM, __pyx_k_BOM, sizeof(__pyx_k_BOM), 0, 0, 1, 1},
  {&__pyx_kp_s_CR, __pyx_k_CR, sizeof(__pyx_k_CR), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_combine_element_of_type_s, __pyx_k_Cannot_combine_element_of_type_s, sizeof(__pyx_k_Cannot_combine_element_of_type_s), 0, 0, 1, 0},
  {&__pyx_n_s_CaselessKeyword, __pyx_k_CaselessKeyword, sizeof(__pyx_k_CaselessKeyword), 0, 0, 1, 1},
  {&__pyx_n_s_CaselessKeyword___init, __pyx_k_CaselessKeyword___init, sizeof(__pyx_k_CaselessKeyword___init), 0, 0, 1, 1},
  {&__pyx_n_s_CaselessLiteral, __pyx_k_CaselessLiteral, sizeof(__pyx_k_CaselessLiteral), 0, 0, 1, 1},
  {&__pyx_n_s_CaselessLiteral___init, __pyx_k_CaselessLiteral___init, sizeof(__pyx_k_CaselessLiteral___init), 0, 0, 1, 1},
  {&__pyx_n_s_CaselessLiteral_parseImpl, __pyx_k_CaselessLiteral_parseImpl, sizeof(__pyx_k_CaselessLiteral_parseImpl), 0, 0, 1, 1},
  {&__pyx_kp_s_Caseless_version_of_class_Keywo, __pyx_k_Caseless_version_of_class_Keywo, sizeof(__pyx_k_Caseless_version_of_class_Keywo), 0, 0, 1, 0},
  {&__pyx_n_s_Char, __pyx_k_Char, sizeof(__pyx_k_Char), 0, 0, 1, 1},
  {&__pyx_n_s_Char___init, __pyx_k_Char___init, sizeof(__pyx_k_Char___init), 0, 0, 1, 1},
  {&__pyx_n_s_CharsNotIn, __pyx_k_CharsNotIn, sizeof(__pyx_k_CharsNotIn), 0, 0, 1, 1},
  {&__pyx_n_s_CharsNotIn___init, __pyx_k_CharsNotIn___init, sizeof(__pyx_k_CharsNotIn___init), 0, 0, 1, 1},
  {&__pyx_n_s_CharsNotIn___str, __pyx_k_CharsNotIn___str, sizeof(__pyx_k_CharsNotIn___str), 0, 0, 1, 1},
  {&__pyx_n_s_CharsNotIn_parseImpl, __pyx_k_CharsNotIn_parseImpl, sizeof(__pyx_k_CharsNotIn_parseImpl), 0, 0, 1, 1},
  {&__pyx_n_s_CloseMatch, __pyx_k_CloseMatch, sizeof(__pyx_k_CloseMatch), 0, 0, 1, 1},
  {&__pyx_n_s_CloseMatch___init, __pyx_k_CloseMatch___init, sizeof(__pyx_k_CloseMatch___init), 0, 0, 1, 1},
  {&__pyx_n_s_CloseMatch_parseImpl, __pyx_k_CloseMatch_parseImpl, sizeof(__pyx_k_CloseMatch_parseImpl), 0, 0, 1, 1},
  {&__pyx_n_s_Combine, __pyx_k_Combine, sizeof(__pyx_k_Combine), 0, 0, 1, 1},
  {&__pyx_n_s_Combine___init, __pyx_k_Combine___init, sizeof(__pyx_k_Combine___init), 0, 0, 1, 1},
  {&__pyx_n_s_Combine_ignore, __pyx_k_Combine_ignore, sizeof(__pyx_k_Combine_ignore), 0, 0, 1, 1},
  {&__pyx_n_s_Combine_postParse, __pyx_k_Combine_postParse, sizeof(__pyx_k_Combine_postParse), 0, 0, 1, 1},
  {&__pyx_kp_s_Converter_for_ignoring_the_resul, __pyx_k_Converter_for_ignoring_the_resul, sizeof(__pyx_k_Converter_for_ignoring_the_resul), 0, 0, 1, 0},
  {&__pyx_kp_s_Converter_to_concatenate_all_mat, __pyx_k_Converter_to_concatenate_all_mat, sizeof(__pyx_k_Converter_to_concatenate_all_mat), 0, 0, 1, 0},
  {&__pyx_kp_s_Converter_to_return_a_repetitive, __pyx_k_Converter_to_return_a_repetitive, sizeof(__pyx_k_Converter_to_return_a_repetitive), 0, 0, 1, 0},
  {&__pyx_kp_s_Converter_to_return_the_matched, __pyx_k_Converter_to_return_the_matched, sizeof(__pyx_k_Converter_to_return_the_matched), 0, 0, 1, 0},
  {&__pyx_n_s_DEFAULT_KEYWORD_CHARS, __pyx_k_DEFAULT_KEYWORD_CHARS, sizeof(__pyx_k_DEFAULT_KEYWORD_CHARS), 0, 0, 1, 1},
  {&__pyx_n_s_DEFAULT_WHITE_CHARS, __pyx_k_DEFAULT_WHITE_CHARS, sizeof(__pyx_k_DEFAULT_WHITE_CHARS), 0, 0, 1, 1},
  {&__pyx_n_s_DOTALL, __pyx_k_DOTALL, sizeof(__pyx_k_DOTALL), 0, 0, 1, 1},
  {&__pyx_kp_s_Diagnostic_configuration_all_de, __pyx_k_Diagnostic_configuration_all_de, sizeof(__pyx_k_Diagnostic_configuration_all_de), 0, 0, 1, 0},
  {&__pyx_n_s_Dict, __pyx_k_Dict, sizeof(__pyx_k_Dict), 0, 0, 1, 1},
  {&__pyx_n_s_Dict___init, __pyx_k_Dict___init, sizeof(__pyx_k_Dict___init), 0, 0, 1, 1},
  {&__pyx_n_s_Dict_postParse, __pyx_k_Dict_postParse, sizeof(__pyx_k_Dict_postParse), 0, 0, 1, 1},
  {&__pyx_kp_s_EM_QUAD, __pyx_k_EM_QUAD, sizeof(__pyx_k_EM_QUAD), 0, 0, 1, 0},
  {&__pyx_kp_s_EM_SPACE, __pyx_k_EM_SPACE, sizeof(__pyx_k_EM_SPACE), 0, 0, 1, 0},
  {&__pyx_kp_s_EN_QUAD, __pyx_k_EN_QUAD, sizeof(__pyx_k_EN_QUAD), 0, 0, 1, 0},
  {&__pyx_kp_s_EN_SPACE, __pyx_k_EN_SPACE, sizeof(__pyx_k_EN_SPACE), 0, 0, 1, 0},
  {&__pyx_n_s_Each, __pyx_k_Each, sizeof(__pyx_k_Each), 0, 0, 1, 1},
  {&__pyx_n_s_Each___init, __pyx_k_Each___init, sizeof(__pyx_k_Each___init), 0, 0, 1, 1},
  {&__pyx_n_s_Each___init___locals_genexpr, __pyx_k_Each___init___locals_genexpr, sizeof(__pyx_k_Each___init___locals_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_Each___str, __pyx_k_Each___str, sizeof(__pyx_k_Each___str), 0, 0, 1, 1},
  {&__pyx_n_s_Each___str___locals_genexpr, __pyx_k_Each___str___locals_genexpr, sizeof(__pyx_k_Each___str___locals_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_Each_checkRecursion, __pyx_k_Each_checkRecursion, sizeof(__pyx_k_Each_checkRecursion), 0, 0, 1, 1},
  {&__pyx_n_s_Each_parseImpl, __pyx_k_Each_parseImpl, sizeof(__pyx_k_Each_parseImpl), 0, 0, 1, 1},
  {&__pyx_n_s_Each_parseImpl_locals_genexpr, __pyx_k_Each_parseImpl_locals_genexpr, sizeof(__pyx_k_Each_parseImpl_locals_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_Each_streamline, __pyx_k_Each_streamline, sizeof(__pyx_k_Each_streamline), 0, 0, 1, 1},
  {&__pyx_n_s_Each_streamline_locals_genexpr, __pyx_k_Each_streamline_locals_genexpr, sizeof(__pyx_k_Each_streamline_locals_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_Ellipsis, __pyx_k_Ellipsis, sizeof(__pyx_k_Ellipsis), 0, 0, 1, 1},
  {&__pyx_n_s_Empty, __pyx_k_Empty, sizeof(__pyx_k_Empty), 0, 0, 1, 1},
  {&__pyx_n_s_Empty___init, __pyx_k_Empty___init, sizeof(__pyx_k_Empty___init), 0, 0, 1, 1},
  {&__pyx_n_s_ErrorStop, __pyx_k_ErrorStop, sizeof(__pyx_k_ErrorStop), 0, 0, 1, 1},
  {&__pyx_kp_s_Exception_raised, __pyx_k_Exception_raised, sizeof(__pyx_k_Exception_raised), 0, 0, 1, 0},
  {&__pyx_kp_s_Exception_thrown_when_parse_exp, __pyx_k_Exception_thrown_when_parse_exp, sizeof(__pyx_k_Exception_thrown_when_parse_exp), 0, 0, 1, 0},
  {&__pyx_kp_s_Expected, __pyx_k_Expected, sizeof(__pyx_k_Expected), 0, 0, 1, 0},
  {&__pyx_kp_s_Expected_end_of_line, __pyx_k_Expected_end_of_line, sizeof(__pyx_k_Expected_end_of_line), 0, 0, 1, 0},
  {&__pyx_kp_s_Expected_end_of_text, __pyx_k_Expected_end_of_text, sizeof(__pyx_k_Expected_end_of_text), 0, 0, 1, 0},
  {&__pyx_kp_s_Expected_r_with_up_to_d_mismatch, __pyx_k_Expected_r_with_up_to_d_mismatch, sizeof(__pyx_k_Expected_r_with_up_to_d_mismatch), 0, 0, 1, 0},
  {&__pyx_kp_s_Expected_start_of_line, __pyx_k_Expected_start_of_line, sizeof(__pyx_k_Expected_start_of_line), 0, 0, 1, 0},
  {&__pyx_kp_s_Expected_start_of_text, __pyx_k_Expected_start_of_text, sizeof(__pyx_k_Expected_start_of_text), 0, 0, 1, 0},
  {&__pyx_kp_s_FAIL, __pyx_k_FAIL, sizeof(__pyx_k_FAIL), 0, 0, 1, 0},
  {&__pyx_n_s_FAIL_2, __pyx_k_FAIL_2, sizeof(__pyx_k_FAIL_2), 0, 0, 1, 1},
  {&__pyx_kp_s_FAIL_EXCEPTION, __pyx_k_FAIL_EXCEPTION, sizeof(__pyx_k_FAIL_EXCEPTION), 0, 0, 1, 0},
  {&__pyx_kp_s_FATAL, __pyx_k_FATAL, sizeof(__pyx_k_FATAL), 0, 0, 1, 0},
  {&__pyx_kp_s_FF, __pyx_k_FF, sizeof(__pyx_k_FF), 0, 0, 1, 0},
  {&__pyx_kp_s_FIGURE_SPACE, __pyx_k_FIGURE_SPACE, sizeof(__pyx_k_FIGURE_SPACE), 0, 0, 1, 0},
  {&__pyx_kp_s_FOUR_PER_EM_SPACE, __pyx_k_FOUR_PER_EM_SPACE, sizeof(__pyx_k_FOUR_PER_EM_SPACE), 0, 0, 1, 0},
  {&__pyx_n_s_FifoCache, __pyx_k_FifoCache, sizeof(__pyx_k_FifoCache), 0, 0, 1, 1},
  {&__pyx_n_s_FollowedBy, __pyx_k_FollowedBy, sizeof(__pyx_k_FollowedBy), 0, 0, 1, 1},
  {&__pyx_n_s_FollowedBy___init, __pyx_k_FollowedBy___init, sizeof(__pyx_k_FollowedBy___init), 0, 0, 1, 1},
  {&__pyx_n_s_FollowedBy_parseImpl, __pyx_k_FollowedBy_parseImpl, sizeof(__pyx_k_FollowedBy_parseImpl), 0, 0, 1, 1},
  {&__pyx_n_s_Forward, __pyx_k_Forward, sizeof(__pyx_k_Forward), 0, 0, 1, 1},
  {&__pyx_n_s_Forward___ilshift, __pyx_k_Forward___ilshift, sizeof(__pyx_k_Forward___ilshift), 0, 0, 1, 1},
  {&__pyx_n_s_Forward___init, __pyx_k_Forward___init, sizeof(__pyx_k_Forward___init), 0, 0, 1, 1},
  {&__pyx_n_s_Forward___lshift, __pyx_k_Forward___lshift, sizeof(__pyx_k_Forward___lshift), 0, 0, 1, 1},
  {&__pyx_n_s_Forward___str, __pyx_k_Forward___str, sizeof(__pyx_k_Forward___str), 0, 0, 1, 1},
  {&__pyx_n_s_Forward__setResultsName, __pyx_k_Forward__setResultsName, sizeof(__pyx_k_Forward__setResultsName), 0, 0, 1, 1},
  {&__pyx_n_s_Forward_copy, __pyx_k_Forward_copy, sizeof(__pyx_k_Forward_copy), 0, 0, 1, 1},
  {&__pyx_kp_s_Forward_declaration_of_an_expres, __pyx_k_Forward_declaration_of_an_expres, sizeof(__pyx_k_Forward_declaration_of_an_expres), 0, 0, 1, 0},
  {&__pyx_n_s_Forward_leaveWhitespace, __pyx_k_Forward_leaveWhitespace, sizeof(__pyx_k_Forward_leaveWhitespace), 0, 0, 1, 1},
  {&__pyx_n_s_Forward_streamline, __pyx_k_Forward_streamline, sizeof(__pyx_k_Forward_streamline), 0, 0, 1, 1},
  {&__pyx_n_s_Forward_validate, __pyx_k_Forward_validate, sizeof(__pyx_k_Forward_validate), 0, 0, 1, 1},
  {&__pyx_kp_s_Found_unwanted_token, __pyx_k_Found_unwanted_token, sizeof(__pyx_k_Found_unwanted_token), 0, 0, 1, 0},
  {&__pyx_n_s_GoToColumn, __pyx_k_GoToColumn, sizeof(__pyx_k_GoToColumn), 0, 0, 1, 1},
  {&__pyx_n_s_GoToColumn___init, __pyx_k_GoToColumn___init, sizeof(__pyx_k_GoToColumn___init), 0, 0, 1, 1},
  {&__pyx_n_s_GoToColumn_parseImpl, __pyx_k_GoToColumn_parseImpl, sizeof(__pyx_k_GoToColumn_parseImpl), 0, 0, 1, 1},
  {&__pyx_n_s_GoToColumn_preParse, __pyx_k_GoToColumn_preParse, sizeof(__pyx_k_GoToColumn_preParse), 0, 0, 1, 1},
  {&__pyx_n_s_Group, __pyx_k_Group, sizeof(__pyx_k_Group), 0, 0, 1, 1},
  {&__pyx_n_s_Group___init, __pyx_k_Group___init, sizeof(__pyx_k_Group___init), 0, 0, 1, 1},
  {&__pyx_n_s_Group_postParse, __pyx_k_Group_postParse, sizeof(__pyx_k_Group_postParse), 0, 0, 1, 1},
  {&__pyx_kp_s_HAIR_SPACE, __pyx_k_HAIR_SPACE, sizeof(__pyx_k_HAIR_SPACE), 0, 0, 1, 0},
  {&__pyx_n_s_HIT, __pyx_k_HIT, sizeof(__pyx_k_HIT), 0, 0, 1, 1},
  {&__pyx_kp_s_IDEOGRAPHIC_SPACE, __pyx_k_IDEOGRAPHIC_SPACE, sizeof(__pyx_k_IDEOGRAPHIC_SPACE), 0, 0, 1, 0},
  {&__pyx_n_s_ITEM, __pyx_k_ITEM, sizeof(__pyx_k_ITEM), 0, 0, 1, 1},
  {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
  {&__pyx_n_s_IndexError, __pyx_k_IndexError, sizeof(__pyx_k_IndexError), 0, 0, 1, 1},
  {&__pyx_n_s_Iterable, __pyx_k_Iterable, sizeof(__pyx_k_Iterable), 0, 0, 1, 1},
  {&__pyx_n_s_KeyError, __pyx_k_KeyError, sizeof(__pyx_k_KeyError), 0, 0, 1, 1},
  {&__pyx_n_s_Keyword, __pyx_k_Keyword, sizeof(__pyx_k_Keyword), 0, 0, 1, 1},
  {&__pyx_n_s_Keyword___init, __pyx_k_Keyword___init, sizeof(__pyx_k_Keyword___init), 0, 0, 1, 1},
  {&__pyx_n_s_Keyword_copy, __pyx_k_Keyword_copy, sizeof(__pyx_k_Keyword_copy), 0, 0, 1, 1},
  {&__pyx_n_s_Keyword_parseImpl, __pyx_k_Keyword_parseImpl, sizeof(__pyx_k_Keyword_parseImpl), 0, 0, 1, 1},
  {&__pyx_n_s_Keyword_setDefaultKeywordChars, __pyx_k_Keyword_setDefaultKeywordChars, sizeof(__pyx_k_Keyword_setDefaultKeywordChars), 0, 0, 1, 1},
  {&__pyx_n_s_L, __pyx_k_L, sizeof(__pyx_k_L), 0, 0, 1, 1},
  {&__pyx_kp_s_LF, __pyx_k_LF, sizeof(__pyx_k_LF), 0, 0, 1, 0},
  {&__pyx_n_s_L_2, __pyx_k_L_2, sizeof(__pyx_k_L_2), 0, 0, 1, 1},
  {&__pyx_n_s_LineEnd, __pyx_k_LineEnd, sizeof(__pyx_k_LineEnd), 0, 0, 1, 1},
  {&__pyx_n_s_LineEnd___init, __pyx_k_LineEnd___init, sizeof(__pyx_k_LineEnd___init), 0, 0, 1, 1},
  {&__pyx_n_s_LineEnd_parseImpl, __pyx_k_LineEnd_parseImpl, sizeof(__pyx_k_LineEnd_parseImpl), 0, 0, 1, 1},
  {&__pyx_n_s_LineStart, __pyx_k_LineStart, sizeof(__pyx_k_LineStart), 0, 0, 1, 1},
  {&__pyx_n_s_LineStart___init, __pyx_k_LineStart___init, sizeof(__pyx_k_LineStart___init), 0, 0, 1, 1},
  {&__pyx_n_s_LineStart_parseImpl, __pyx_k_LineStart_parseImpl, sizeof(__pyx_k_LineStart_parseImpl), 0, 0, 1, 1},
  {&__pyx_n_s_Literal, __pyx_k_Literal, sizeof(__pyx_k_Literal), 0, 0, 1, 1},
  {&__pyx_n_s_Literal___init, __pyx_k_Literal___init, sizeof(__pyx_k_Literal___init), 0, 0, 1, 1},
  {&__pyx_n_s_Literal_parseImpl, __pyx_k_Literal_parseImpl, sizeof(__pyx_k_Literal_parseImpl), 0, 0, 1, 1},
  {&__pyx_kp_s_Lookahead_matching_of_the_given, __pyx_k_Lookahead_matching_of_the_given, sizeof(__pyx_k_Lookahead_matching_of_the_given), 0, 0, 1, 0},
  {&__pyx_kp_s_Lookahead_to_disallow_matching_w, __pyx_k_Lookahead_to_disallow_matching_w, sizeof(__pyx_k_Lookahead_to_disallow_matching_w), 0, 0, 1, 0},
  {&__pyx_kp_s_Lookbehind_matching_of_the_given, __pyx_k_Lookbehind_matching_of_the_given, sizeof(__pyx_k_Lookbehind_matching_of_the_given), 0, 0, 1, 0},
  {&__pyx_n_s_MATCH, __pyx_k_MATCH, sizeof(__pyx_k_MATCH), 0, 0, 1, 1},
  {&__pyx_n_s_MISS, __pyx_k_MISS, sizeof(__pyx_k_MISS), 0, 0, 1, 1},
  {&__pyx_kp_s_MMSP, __pyx_k_MMSP, sizeof(__pyx_k_MMSP), 0, 0, 1, 0},
  {&__pyx_kp_s_MONGOLIAN_VOWEL_SEPARATOR, __pyx_k_MONGOLIAN_VOWEL_SEPARATOR, sizeof(__pyx_k_MONGOLIAN_VOWEL_SEPARATOR), 0, 0, 1, 0},
  {&__pyx_n_s_MULTILINE, __pyx_k_MULTILINE, sizeof(__pyx_k_MULTILINE), 0, 0, 1, 1},
  {&__pyx_n_s_Mapping, __pyx_k_Mapping, sizeof(__pyx_k_Mapping), 0, 0, 1, 1},
  {&__pyx_kp_s_Match, __pyx_k_Match, sizeof(__pyx_k_Match), 0, 0, 1, 0},
  {&__pyx_n_s_MatchFirst, __pyx_k_MatchFirst, sizeof(__pyx_k_MatchFirst), 0, 0, 1, 1},
  {&__pyx_n_s_MatchFirst___init, __pyx_k_MatchFirst___init, sizeof(__pyx_k_MatchFirst___init), 0, 0, 1, 1},
  {&__pyx_n_s_MatchFirst___init___locals_genex, __pyx_k_MatchFirst___init___locals_genex, sizeof(__pyx_k_MatchFirst___init___locals_genex), 0, 0, 1, 1},
  {&__pyx_n_s_MatchFirst___ior, __pyx_k_MatchFirst___ior, sizeof(__pyx_k_MatchFirst___ior), 0, 0, 1, 1},
  {&__pyx_n_s_MatchFirst___str, __pyx_k_MatchFirst___str, sizeof(__pyx_k_MatchFirst___str), 0, 0, 1, 1},
  {&__pyx_n_s_MatchFirst___str___locals_genexp, __pyx_k_MatchFirst___str___locals_genexp, sizeof(__pyx_k_MatchFirst___str___locals_genexp), 0, 0, 1, 1},
  {&__pyx_n_s_MatchFirst__setResultsName, __pyx_k_MatchFirst__setResultsName, sizeof(__pyx_k_MatchFirst__setResultsName), 0, 0, 1, 1},
  {&__pyx_n_s_MatchFirst__setResultsName_local, __pyx_k_MatchFirst__setResultsName_local, sizeof(__pyx_k_MatchFirst__setResultsName_local), 0, 0, 1, 1},
  {&__pyx_n_s_MatchFirst_checkRecursion, __pyx_k_MatchFirst_checkRecursion, sizeof(__pyx_k_MatchFirst_checkRecursion), 0, 0, 1, 1},
  {&__pyx_n_s_MatchFirst_parseImpl, __pyx_k_MatchFirst_parseImpl, sizeof(__pyx_k_MatchFirst_parseImpl), 0, 0, 1, 1},
  {&__pyx_n_s_MatchFirst_streamline, __pyx_k_MatchFirst_streamline, sizeof(__pyx_k_MatchFirst_streamline), 0, 0, 1, 1},
  {&__pyx_n_s_MatchFirst_streamline_locals_gen, __pyx_k_MatchFirst_streamline_locals_gen, sizeof(__pyx_k_MatchFirst_streamline_locals_gen), 0, 0, 1, 1},
  {&__pyx_kp_s_Matched, __pyx_k_Matched, sizeof(__pyx_k_Matched), 0, 0, 1, 0},
  {&__pyx_kp_s_Matches_if_current_position_is_a, __pyx_k_Matches_if_current_position_is_a, sizeof(__pyx_k_Matches_if_current_position_is_a), 0, 0, 1, 0},
  {&__pyx_kp_s_Matches_if_current_position_is_a_2, __pyx_k_Matches_if_current_position_is_a_2, sizeof(__pyx_k_Matches_if_current_position_is_a_2), 0, 0, 1, 0},
  {&__pyx_kp_s_Matches_if_current_position_is_a_3, __pyx_k_Matches_if_current_position_is_a_3, sizeof(__pyx_k_Matches_if_current_position_is_a_3), 0, 0, 1, 0},
  {&__pyx_kp_s_Matches_if_current_position_is_a_4, __pyx_k_Matches_if_current_position_is_a_4, sizeof(__pyx_k_Matches_if_current_position_is_a_4), 0, 0, 1, 0},
  {&__pyx_kp_s_Matches_if_the_current_position, __pyx_k_Matches_if_the_current_position, sizeof(__pyx_k_Matches_if_the_current_position), 0, 0, 1, 0},
  {&__pyx_kp_s_Matches_if_the_current_position_2, __pyx_k_Matches_if_the_current_position_2, sizeof(__pyx_k_Matches_if_the_current_position_2), 0, 0, 1, 0},
  {&__pyx_n_s_MethodType, __pyx_k_MethodType, sizeof(__pyx_k_MethodType), 0, 0, 1, 1},
  {&__pyx_kp_s_Missing_one_or_more_required_ele, __pyx_k_Missing_one_or_more_required_ele, sizeof(__pyx_k_Missing_one_or_more_required_ele), 0, 0, 1, 0},
  {&__pyx_n_s_MultipleMatch, __pyx_k_MultipleMatch, sizeof(__pyx_k_MultipleMatch), 0, 0, 1, 1},
  {&__pyx_n_s_MultipleMatch___init, __pyx_k_MultipleMatch___init, sizeof(__pyx_k_MultipleMatch___init), 0, 0, 1, 1},
  {&__pyx_n_s_MultipleMatch__setResultsName, __pyx_k_MultipleMatch__setResultsName, sizeof(__pyx_k_MultipleMatch__setResultsName), 0, 0, 1, 1},
  {&__pyx_n_s_MultipleMatch_parseImpl, __pyx_k_MultipleMatch_parseImpl, sizeof(__pyx_k_MultipleMatch_parseImpl), 0, 0, 1, 1},
  {&__pyx_n_s_MultipleMatch_stopOn, __pyx_k_MultipleMatch_stopOn, sizeof(__pyx_k_MultipleMatch_stopOn), 0, 0, 1, 1},
  {&__pyx_n_s_MutableMapping, __pyx_k_MutableMapping, sizeof(__pyx_k_MutableMapping), 0, 0, 1, 1},
  {&__pyx_kp_s_NBSP, __pyx_k_NBSP, sizeof(__pyx_k_NBSP), 0, 0, 1, 0},
  {&__pyx_n_s_NL, __pyx_k_NL, sizeof(__pyx_k_NL), 0, 0, 1, 1},
  {&__pyx_kp_s_NNBSP, __pyx_k_NNBSP, sizeof(__pyx_k_NNBSP), 0, 0, 1, 0},
  {&__pyx_n_s_NoMatch, __pyx_k_NoMatch, sizeof(__pyx_k_NoMatch), 0, 0, 1, 1},
  {&__pyx_n_s_NoMatch___init, __pyx_k_NoMatch___init, sizeof(__pyx_k_NoMatch___init), 0, 0, 1, 1},
  {&__pyx_n_s_NoMatch_parseImpl, __pyx_k_NoMatch_parseImpl, sizeof(__pyx_k_NoMatch_parseImpl), 0, 0, 1, 1},
  {&__pyx_kp_s_No_match_found_for, __pyx_k_No_match_found_for, sizeof(__pyx_k_No_match_found_for), 0, 0, 1, 0},
  {&__pyx_n_s_None, __pyx_k_None, sizeof(__pyx_k_None), 0, 0, 1, 1},
  {&__pyx_n_s_NotAny, __pyx_k_NotAny, sizeof(__pyx_k_NotAny), 0, 0, 1, 1},
  {&__pyx_n_s_NotAny___init, __pyx_k_NotAny___init, sizeof(__pyx_k_NotAny___init), 0, 0, 1, 1},
  {&__pyx_n_s_NotAny___str, __pyx_k_NotAny___str, sizeof(__pyx_k_NotAny___str), 0, 0, 1, 1},
  {&__pyx_n_s_NotAny_parseImpl, __pyx_k_NotAny_parseImpl, sizeof(__pyx_k_NotAny_parseImpl), 0, 0, 1, 1},
  {&__pyx_kp_s_Not_at_the_end_of_a_word, __pyx_k_Not_at_the_end_of_a_word, sizeof(__pyx_k_Not_at_the_end_of_a_word), 0, 0, 1, 0},
  {&__pyx_kp_s_Not_at_the_start_of_a_word, __pyx_k_Not_at_the_start_of_a_word, sizeof(__pyx_k_Not_at_the_start_of_a_word), 0, 0, 1, 0},
  {&__pyx_n_s_NullToken, __pyx_k_NullToken, sizeof(__pyx_k_NullToken), 0, 0, 1, 1},
  {&__pyx_n_s_NullToken___bool, __pyx_k_NullToken___bool, sizeof(__pyx_k_NullToken___bool), 0, 0, 1, 1},
  {&__pyx_n_s_NullToken___str, __pyx_k_NullToken___str, sizeof(__pyx_k_NullToken___str), 0, 0, 1, 1},
  {&__pyx_kp_s_OGHAM_SPACE_MARK, __pyx_k_OGHAM_SPACE_MARK, sizeof(__pyx_k_OGHAM_SPACE_MARK), 0, 0, 1, 0},
  {&__pyx_n_s_OneOrMore, __pyx_k_OneOrMore, sizeof(__pyx_k_OneOrMore), 0, 0, 1, 1},
  {&__pyx_n_s_OneOrMore___str, __pyx_k_OneOrMore___str, sizeof(__pyx_k_OneOrMore___str), 0, 0, 1, 1},
  {&__pyx_n_s_OnlyOnce, __pyx_k_OnlyOnce, sizeof(__pyx_k_OnlyOnce), 0, 0, 1, 1},
  {&__pyx_n_s_OnlyOnce___call, __pyx_k_OnlyOnce___call, sizeof(__pyx_k_OnlyOnce___call), 0, 0, 1, 1},
  {&__pyx_n_s_OnlyOnce___init, __pyx_k_OnlyOnce___init, sizeof(__pyx_k_OnlyOnce___init), 0, 0, 1, 1},
  {&__pyx_n_s_OnlyOnce_reset, __pyx_k_OnlyOnce_reset, sizeof(__pyx_k_OnlyOnce_reset), 0, 0, 1, 1},
  {&__pyx_n_s_Optional, __pyx_k_Optional, sizeof(__pyx_k_Optional), 0, 0, 1, 1},
  {&__pyx_n_s_Optional___init, __pyx_k_Optional___init, sizeof(__pyx_k_Optional___init), 0, 0, 1, 1},
  {&__pyx_n_s_Optional___str, __pyx_k_Optional___str, sizeof(__pyx_k_Optional___str), 0, 0, 1, 1},
  {&__pyx_n_s_Optional__optionalNotMatched, __pyx_k_Optional__optionalNotMatched, sizeof(__pyx_k_Optional__optionalNotMatched), 0, 0, 1, 1},
  {&__pyx_kp_s_Optional_matching_of_the_given_e, __pyx_k_Optional_matching_of_the_given_e, sizeof(__pyx_k_Optional_matching_of_the_given_e), 0, 0, 1, 0},
  {&__pyx_n_s_Optional_parseImpl, __pyx_k_Optional_parseImpl, sizeof(__pyx_k_Optional_parseImpl), 0, 0, 1, 1},
  {&__pyx_kp_s_Optional_repetition_of_zero_or_m, __pyx_k_Optional_repetition_of_zero_or_m, sizeof(__pyx_k_Optional_repetition_of_zero_or_m), 0, 0, 1, 0},
  {&__pyx_n_s_Or, __pyx_k_Or, sizeof(__pyx_k_Or), 0, 0, 1, 1},
  {&__pyx_n_s_Or___init, __pyx_k_Or___init, sizeof(__pyx_k_Or___init), 0, 0, 1, 1},
  {&__pyx_n_s_Or___init___locals_genexpr, __pyx_k_Or___init___locals_genexpr, sizeof(__pyx_k_Or___init___locals_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_Or___ixor, __pyx_k_Or___ixor, sizeof(__pyx_k_Or___ixor), 0, 0, 1, 1},
  {&__pyx_n_s_Or___str, __pyx_k_Or___str, sizeof(__pyx_k_Or___str), 0, 0, 1, 1},
  {&__pyx_n_s_Or___str___locals_genexpr, __pyx_k_Or___str___locals_genexpr, sizeof(__pyx_k_Or___str___locals_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_Or__setResultsName, __pyx_k_Or__setResultsName, sizeof(__pyx_k_Or__setResultsName), 0, 0, 1, 1},
  {&__pyx_n_s_Or__setResultsName_locals_genexp, __pyx_k_Or__setResultsName_locals_genexp, sizeof(__pyx_k_Or__setResultsName_locals_genexp), 0, 0, 1, 1},
  {&__pyx_n_s_Or_checkRecursion, __pyx_k_Or_checkRecursion, sizeof(__pyx_k_Or_checkRecursion), 0, 0, 1, 1},
  {&__pyx_n_s_Or_parseImpl, __pyx_k_Or_parseImpl, sizeof(__pyx_k_Or_parseImpl), 0, 0, 1, 1},
  {&__pyx_n_s_Or_streamline, __pyx_k_Or_streamline, sizeof(__pyx_k_Or_streamline), 0, 0, 1, 1},
  {&__pyx_n_s_Or_streamline_locals_genexpr, __pyx_k_Or_streamline_locals_genexpr, sizeof(__pyx_k_Or_streamline_locals_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_OrderedDict, __pyx_k_OrderedDict, sizeof(__pyx_k_OrderedDict), 0, 0, 1, 1},
  {&__pyx_n_s_OrderedDict_2, __pyx_k_OrderedDict_2, sizeof(__pyx_k_OrderedDict_2), 0, 0, 1, 1},
  {&__pyx_kp_s_PUNCTUATION_SPACE, __pyx_k_PUNCTUATION_SPACE, sizeof(__pyx_k_PUNCTUATION_SPACE), 0, 0, 1, 0},
  {&__pyx_n_s_PY_3, __pyx_k_PY_3, sizeof(__pyx_k_PY_3), 0, 0, 1, 1},
  {&__pyx_n_s_ParseBaseException, __pyx_k_ParseBaseException, sizeof(__pyx_k_ParseBaseException), 0, 0, 1, 1},
  {&__pyx_n_s_ParseBaseException___dir, __pyx_k_ParseBaseException___dir, sizeof(__pyx_k_ParseBaseException___dir), 0, 0, 1, 1},
  {&__pyx_n_s_ParseBaseException___getattr, __pyx_k_ParseBaseException___getattr, sizeof(__pyx_k_ParseBaseException___getattr), 0, 0, 1, 1},
  {&__pyx_n_s_ParseBaseException___init, __pyx_k_ParseBaseException___init, sizeof(__pyx_k_ParseBaseException___init), 0, 0, 1, 1},
  {&__pyx_n_s_ParseBaseException___repr, __pyx_k_ParseBaseException___repr, sizeof(__pyx_k_ParseBaseException___repr), 0, 0, 1, 1},
  {&__pyx_n_s_ParseBaseException___str, __pyx_k_ParseBaseException___str, sizeof(__pyx_k_ParseBaseException___str), 0, 0, 1, 1},
  {&__pyx_n_s_ParseBaseException__from_excepti, __pyx_k_ParseBaseException__from_excepti, sizeof(__pyx_k_ParseBaseException__from_excepti), 0, 0, 1, 1},
  {&__pyx_n_s_ParseBaseException_markInputline, __pyx_k_ParseBaseException_markInputline, sizeof(__pyx_k_ParseBaseException_markInputline), 0, 0, 1, 1},
  {&__pyx_n_s_ParseElementEnhance, __pyx_k_ParseElementEnhance, sizeof(__pyx_k_ParseElementEnhance), 0, 0, 1, 1},
  {&__pyx_n_s_ParseElementEnhance___init, __pyx_k_ParseElementEnhance___init, sizeof(__pyx_k_ParseElementEnhance___init), 0, 0, 1, 1},
  {&__pyx_n_s_ParseElementEnhance___str, __pyx_k_ParseElementEnhance___str, sizeof(__pyx_k_ParseElementEnhance___str), 0, 0, 1, 1},
  {&__pyx_n_s_ParseElementEnhance_checkRecursi, __pyx_k_ParseElementEnhance_checkRecursi, sizeof(__pyx_k_ParseElementEnhance_checkRecursi), 0, 0, 1, 1},
  {&__pyx_n_s_ParseElementEnhance_ignore, __pyx_k_ParseElementEnhance_ignore, sizeof(__pyx_k_ParseElementEnhance_ignore), 0, 0, 1, 1},
  {&__pyx_n_s_ParseElementEnhance_leaveWhitesp, __pyx_k_ParseElementEnhance_leaveWhitesp, sizeof(__pyx_k_ParseElementEnhance_leaveWhitesp), 0, 0, 1, 1},
  {&__pyx_n_s_ParseElementEnhance_parseImpl, __pyx_k_ParseElementEnhance_parseImpl, sizeof(__pyx_k_ParseElementEnhance_parseImpl), 0, 0, 1, 1},
  {&__pyx_n_s_ParseElementEnhance_streamline, __pyx_k_ParseElementEnhance_streamline, sizeof(__pyx_k_ParseElementEnhance_streamline), 0, 0, 1, 1},
  {&__pyx_n_s_ParseElementEnhance_validate, __pyx_k_ParseElementEnhance_validate, sizeof(__pyx_k_ParseElementEnhance_validate), 0, 0, 1, 1},
  {&__pyx_n_s_ParseException, __pyx_k_ParseException, sizeof(__pyx_k_ParseException), 0, 0, 1, 1},
  {&__pyx_n_s_ParseException_explain, __pyx_k_ParseException_explain, sizeof(__pyx_k_ParseException_explain), 0, 0, 1, 1},
  {&__pyx_n_s_ParseExpression, __pyx_k_ParseExpression, sizeof(__pyx_k_ParseExpression), 0, 0, 1, 1},
  {&__pyx_n_s_ParseExpression___init, __pyx_k_ParseExpression___init, sizeof(__pyx_k_ParseExpression___init), 0, 0, 1, 1},
  {&__pyx_n_s_ParseExpression___init___locals, __pyx_k_ParseExpression___init___locals, sizeof(__pyx_k_ParseExpression___init___locals), 0, 0, 1, 1},
  {&__pyx_n_s_ParseExpression___str, __pyx_k_ParseExpression___str, sizeof(__pyx_k_ParseExpression___str), 0, 0, 1, 1},
  {&__pyx_n_s_ParseExpression__setResultsName, __pyx_k_ParseExpression__setResultsName, sizeof(__pyx_k_ParseExpression__setResultsName), 0, 0, 1, 1},
  {&__pyx_n_s_ParseExpression_append, __pyx_k_ParseExpression_append, sizeof(__pyx_k_ParseExpression_append), 0, 0, 1, 1},
  {&__pyx_n_s_ParseExpression_copy, __pyx_k_ParseExpression_copy, sizeof(__pyx_k_ParseExpression_copy), 0, 0, 1, 1},
  {&__pyx_n_s_ParseExpression_ignore, __pyx_k_ParseExpression_ignore, sizeof(__pyx_k_ParseExpression_ignore), 0, 0, 1, 1},
  {&__pyx_n_s_ParseExpression_leaveWhitespace, __pyx_k_ParseExpression_leaveWhitespace, sizeof(__pyx_k_ParseExpression_leaveWhitespace), 0, 0, 1, 1},
  {&__pyx_n_s_ParseExpression_streamline, __pyx_k_ParseExpression_streamline, sizeof(__pyx_k_ParseExpression_streamline), 0, 0, 1, 1},
  {&__pyx_n_s_ParseExpression_validate, __pyx_k_ParseExpression_validate, sizeof(__pyx_k_ParseExpression_validate), 0, 0, 1, 1},
  {&__pyx_n_s_ParseFatalException, __pyx_k_ParseFatalException, sizeof(__pyx_k_ParseFatalException), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults, __pyx_k_ParseResults, sizeof(__pyx_k_ParseResults), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResultsWithOffset, __pyx_k_ParseResultsWithOffset, sizeof(__pyx_k_ParseResultsWithOffset), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResultsWithOffset___getite, __pyx_k_ParseResultsWithOffset___getite, sizeof(__pyx_k_ParseResultsWithOffset___getite), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResultsWithOffset___init, __pyx_k_ParseResultsWithOffset___init, sizeof(__pyx_k_ParseResultsWithOffset___init), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResultsWithOffset___repr, __pyx_k_ParseResultsWithOffset___repr, sizeof(__pyx_k_ParseResultsWithOffset___repr), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResultsWithOffset_setOffse, __pyx_k_ParseResultsWithOffset_setOffse, sizeof(__pyx_k_ParseResultsWithOffset_setOffse), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults___add, __pyx_k_ParseResults___add, sizeof(__pyx_k_ParseResults___add), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults___bool, __pyx_k_ParseResults___bool, sizeof(__pyx_k_ParseResults___bool), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults___contains, __pyx_k_ParseResults___contains, sizeof(__pyx_k_ParseResults___contains), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults___delitem, __pyx_k_ParseResults___delitem, sizeof(__pyx_k_ParseResults___delitem), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults___dir, __pyx_k_ParseResults___dir, sizeof(__pyx_k_ParseResults___dir), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults___getattr, __pyx_k_ParseResults___getattr, sizeof(__pyx_k_ParseResults___getattr), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults___getitem, __pyx_k_ParseResults___getitem, sizeof(__pyx_k_ParseResults___getitem), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults___getnewargs, __pyx_k_ParseResults___getnewargs, sizeof(__pyx_k_ParseResults___getnewargs), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults___getstate, __pyx_k_ParseResults___getstate, sizeof(__pyx_k_ParseResults___getstate), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults___iadd, __pyx_k_ParseResults___iadd, sizeof(__pyx_k_ParseResults___iadd), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults___iadd___locals_lam, __pyx_k_ParseResults___iadd___locals_lam, sizeof(__pyx_k_ParseResults___iadd___locals_lam), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults___init, __pyx_k_ParseResults___init, sizeof(__pyx_k_ParseResults___init), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults___iter, __pyx_k_ParseResults___iter, sizeof(__pyx_k_ParseResults___iter), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults___len, __pyx_k_ParseResults___len, sizeof(__pyx_k_ParseResults___len), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults___lookup, __pyx_k_ParseResults___lookup, sizeof(__pyx_k_ParseResults___lookup), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults___new, __pyx_k_ParseResults___new, sizeof(__pyx_k_ParseResults___new), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults___radd, __pyx_k_ParseResults___radd, sizeof(__pyx_k_ParseResults___radd), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults___repr, __pyx_k_ParseResults___repr, sizeof(__pyx_k_ParseResults___repr), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults___reversed, __pyx_k_ParseResults___reversed, sizeof(__pyx_k_ParseResults___reversed), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults___setitem, __pyx_k_ParseResults___setitem, sizeof(__pyx_k_ParseResults___setitem), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults___setstate, __pyx_k_ParseResults___setstate, sizeof(__pyx_k_ParseResults___setstate), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults___str, __pyx_k_ParseResults___str, sizeof(__pyx_k_ParseResults___str), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults___str___locals_gene, __pyx_k_ParseResults___str___locals_gene, sizeof(__pyx_k_ParseResults___str___locals_gene), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults__accumNames, __pyx_k_ParseResults__accumNames, sizeof(__pyx_k_ParseResults__accumNames), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults__asList, __pyx_k_ParseResults__asList, sizeof(__pyx_k_ParseResults__asList), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults__asStringList, __pyx_k_ParseResults__asStringList, sizeof(__pyx_k_ParseResults__asStringList), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults__doinit, __pyx_k_ParseResults__doinit, sizeof(__pyx_k_ParseResults__doinit), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults__iteritems, __pyx_k_ParseResults__iteritems, sizeof(__pyx_k_ParseResults__iteritems), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults__iteritems_locals_g, __pyx_k_ParseResults__iteritems_locals_g, sizeof(__pyx_k_ParseResults__iteritems_locals_g), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults__iterkeys, __pyx_k_ParseResults__iterkeys, sizeof(__pyx_k_ParseResults__iterkeys), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults__itervalues, __pyx_k_ParseResults__itervalues, sizeof(__pyx_k_ParseResults__itervalues), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults__itervalues_locals, __pyx_k_ParseResults__itervalues_locals, sizeof(__pyx_k_ParseResults__itervalues_locals), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults__lookup, __pyx_k_ParseResults__lookup, sizeof(__pyx_k_ParseResults__lookup), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults__modal, __pyx_k_ParseResults__modal, sizeof(__pyx_k_ParseResults__modal), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults__name, __pyx_k_ParseResults__name, sizeof(__pyx_k_ParseResults__name), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults__parent, __pyx_k_ParseResults__parent, sizeof(__pyx_k_ParseResults__parent), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults__tokdict, __pyx_k_ParseResults__tokdict, sizeof(__pyx_k_ParseResults__tokdict), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults__toklist, __pyx_k_ParseResults__toklist, sizeof(__pyx_k_ParseResults__toklist), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults_append, __pyx_k_ParseResults_append, sizeof(__pyx_k_ParseResults_append), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults_asDict, __pyx_k_ParseResults_asDict, sizeof(__pyx_k_ParseResults_asDict), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults_asDict_locals_genex, __pyx_k_ParseResults_asDict_locals_genex, sizeof(__pyx_k_ParseResults_asDict_locals_genex), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults_asDict_locals_toIte, __pyx_k_ParseResults_asDict_locals_toIte, sizeof(__pyx_k_ParseResults_asDict_locals_toIte), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults_asList, __pyx_k_ParseResults_asList, sizeof(__pyx_k_ParseResults_asList), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults_asXML, __pyx_k_ParseResults_asXML, sizeof(__pyx_k_ParseResults_asXML), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults_asXML_locals_genexp, __pyx_k_ParseResults_asXML_locals_genexp, sizeof(__pyx_k_ParseResults_asXML_locals_genexp), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults_clear, __pyx_k_ParseResults_clear, sizeof(__pyx_k_ParseResults_clear), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults_copy, __pyx_k_ParseResults_copy, sizeof(__pyx_k_ParseResults_copy), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults_dump, __pyx_k_ParseResults_dump, sizeof(__pyx_k_ParseResults_dump), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults_dump_locals_genexpr, __pyx_k_ParseResults_dump_locals_genexpr, sizeof(__pyx_k_ParseResults_dump_locals_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults_extend, __pyx_k_ParseResults_extend, sizeof(__pyx_k_ParseResults_extend), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults_from_dict, __pyx_k_ParseResults_from_dict, sizeof(__pyx_k_ParseResults_from_dict), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults_from_dict_locals_is, __pyx_k_ParseResults_from_dict_locals_is, sizeof(__pyx_k_ParseResults_from_dict_locals_is), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults_get, __pyx_k_ParseResults_get, sizeof(__pyx_k_ParseResults_get), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults_getName, __pyx_k_ParseResults_getName, sizeof(__pyx_k_ParseResults_getName), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults_haskeys, __pyx_k_ParseResults_haskeys, sizeof(__pyx_k_ParseResults_haskeys), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults_insert, __pyx_k_ParseResults_insert, sizeof(__pyx_k_ParseResults_insert), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults_items, __pyx_k_ParseResults_items, sizeof(__pyx_k_ParseResults_items), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults_keys, __pyx_k_ParseResults_keys, sizeof(__pyx_k_ParseResults_keys), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults_pop, __pyx_k_ParseResults_pop, sizeof(__pyx_k_ParseResults_pop), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults_pprint, __pyx_k_ParseResults_pprint, sizeof(__pyx_k_ParseResults_pprint), 0, 0, 1, 1},
  {&__pyx_n_s_ParseResults_values, __pyx_k_ParseResults_values, sizeof(__pyx_k_ParseResults_values), 0, 0, 1, 1},
  {&__pyx_n_s_ParseSyntaxException, __pyx_k_ParseSyntaxException, sizeof(__pyx_k_ParseSyntaxException), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement, __pyx_k_ParserElement, sizeof(__pyx_k_ParserElement), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement__FifoCache, __pyx_k_ParserElement__FifoCache, sizeof(__pyx_k_ParserElement__FifoCache), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement__FifoCache___init, __pyx_k_ParserElement__FifoCache___init, sizeof(__pyx_k_ParserElement__FifoCache___init), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement__FifoCache___init_2, __pyx_k_ParserElement__FifoCache___init_2, sizeof(__pyx_k_ParserElement__FifoCache___init_2), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement__FifoCache___init_3, __pyx_k_ParserElement__FifoCache___init_3, sizeof(__pyx_k_ParserElement__FifoCache___init_3), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement__FifoCache___init_4, __pyx_k_ParserElement__FifoCache___init_4, sizeof(__pyx_k_ParserElement__FifoCache___init_4), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement__FifoCache___init_5, __pyx_k_ParserElement__FifoCache___init_5, sizeof(__pyx_k_ParserElement__FifoCache___init_5), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement__UnboundedCache, __pyx_k_ParserElement__UnboundedCache, sizeof(__pyx_k_ParserElement__UnboundedCache), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement__UnboundedCache_2, __pyx_k_ParserElement__UnboundedCache_2, sizeof(__pyx_k_ParserElement__UnboundedCache_2), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement__UnboundedCache_3, __pyx_k_ParserElement__UnboundedCache_3, sizeof(__pyx_k_ParserElement__UnboundedCache_3), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement__UnboundedCache_4, __pyx_k_ParserElement__UnboundedCache_4, sizeof(__pyx_k_ParserElement__UnboundedCache_4), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement__UnboundedCache_5, __pyx_k_ParserElement__UnboundedCache_5, sizeof(__pyx_k_ParserElement__UnboundedCache_5), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement__UnboundedCache_6, __pyx_k_ParserElement__UnboundedCache_6, sizeof(__pyx_k_ParserElement__UnboundedCache_6), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement___add, __pyx_k_ParserElement___add, sizeof(__pyx_k_ParserElement___add), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement___and, __pyx_k_ParserElement___and, sizeof(__pyx_k_ParserElement___and), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement___call, __pyx_k_ParserElement___call, sizeof(__pyx_k_ParserElement___call), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement___eq, __pyx_k_ParserElement___eq, sizeof(__pyx_k_ParserElement___eq), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement___getitem, __pyx_k_ParserElement___getitem, sizeof(__pyx_k_ParserElement___getitem), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement___hash, __pyx_k_ParserElement___hash, sizeof(__pyx_k_ParserElement___hash), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement___init, __pyx_k_ParserElement___init, sizeof(__pyx_k_ParserElement___init), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement___invert, __pyx_k_ParserElement___invert, sizeof(__pyx_k_ParserElement___invert), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement___iter, __pyx_k_ParserElement___iter, sizeof(__pyx_k_ParserElement___iter), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement___mul, __pyx_k_ParserElement___mul, sizeof(__pyx_k_ParserElement___mul), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement___mul___locals_gen, __pyx_k_ParserElement___mul___locals_gen, sizeof(__pyx_k_ParserElement___mul___locals_gen), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement___mul___locals_mak, __pyx_k_ParserElement___mul___locals_mak, sizeof(__pyx_k_ParserElement___mul___locals_mak), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement___ne, __pyx_k_ParserElement___ne, sizeof(__pyx_k_ParserElement___ne), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement___or, __pyx_k_ParserElement___or, sizeof(__pyx_k_ParserElement___or), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement___radd, __pyx_k_ParserElement___radd, sizeof(__pyx_k_ParserElement___radd), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement___rand, __pyx_k_ParserElement___rand, sizeof(__pyx_k_ParserElement___rand), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement___repr, __pyx_k_ParserElement___repr, sizeof(__pyx_k_ParserElement___repr), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement___req, __pyx_k_ParserElement___req, sizeof(__pyx_k_ParserElement___req), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement___rmul, __pyx_k_ParserElement___rmul, sizeof(__pyx_k_ParserElement___rmul), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement___rne, __pyx_k_ParserElement___rne, sizeof(__pyx_k_ParserElement___rne), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement___ror, __pyx_k_ParserElement___ror, sizeof(__pyx_k_ParserElement___ror), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement___rsub, __pyx_k_ParserElement___rsub, sizeof(__pyx_k_ParserElement___rsub), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement___rxor, __pyx_k_ParserElement___rxor, sizeof(__pyx_k_ParserElement___rxor), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement___str, __pyx_k_ParserElement___str, sizeof(__pyx_k_ParserElement___str), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement___sub, __pyx_k_ParserElement___sub, sizeof(__pyx_k_ParserElement___sub), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement___xor, __pyx_k_ParserElement___xor, sizeof(__pyx_k_ParserElement___xor), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement__parseCache, __pyx_k_ParserElement__parseCache, sizeof(__pyx_k_ParserElement__parseCache), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement__parseNoCache, __pyx_k_ParserElement__parseNoCache, sizeof(__pyx_k_ParserElement__parseNoCache), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement__setResultsName, __pyx_k_ParserElement__setResultsName, sizeof(__pyx_k_ParserElement__setResultsName), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement__skipIgnorables, __pyx_k_ParserElement__skipIgnorables, sizeof(__pyx_k_ParserElement__skipIgnorables), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement__trim_traceback, __pyx_k_ParserElement__trim_traceback, sizeof(__pyx_k_ParserElement__trim_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement_addCondition, __pyx_k_ParserElement_addCondition, sizeof(__pyx_k_ParserElement_addCondition), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement_addParseAction, __pyx_k_ParserElement_addParseAction, sizeof(__pyx_k_ParserElement_addParseAction), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement_canParseNext, __pyx_k_ParserElement_canParseNext, sizeof(__pyx_k_ParserElement_canParseNext), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement_checkRecursion, __pyx_k_ParserElement_checkRecursion, sizeof(__pyx_k_ParserElement_checkRecursion), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement_copy, __pyx_k_ParserElement_copy, sizeof(__pyx_k_ParserElement_copy), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement_enablePackrat, __pyx_k_ParserElement_enablePackrat, sizeof(__pyx_k_ParserElement_enablePackrat), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement_ignore, __pyx_k_ParserElement_ignore, sizeof(__pyx_k_ParserElement_ignore), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement_inlineLiteralsUsin, __pyx_k_ParserElement_inlineLiteralsUsin, sizeof(__pyx_k_ParserElement_inlineLiteralsUsin), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement_leaveWhitespace, __pyx_k_ParserElement_leaveWhitespace, sizeof(__pyx_k_ParserElement_leaveWhitespace), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement_matches, __pyx_k_ParserElement_matches, sizeof(__pyx_k_ParserElement_matches), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement_parseFile, __pyx_k_ParserElement_parseFile, sizeof(__pyx_k_ParserElement_parseFile), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement_parseImpl, __pyx_k_ParserElement_parseImpl, sizeof(__pyx_k_ParserElement_parseImpl), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement_parseString, __pyx_k_ParserElement_parseString, sizeof(__pyx_k_ParserElement_parseString), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement_parseWithTabs, __pyx_k_ParserElement_parseWithTabs, sizeof(__pyx_k_ParserElement_parseWithTabs), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement_postParse, __pyx_k_ParserElement_postParse, sizeof(__pyx_k_ParserElement_postParse), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement_preParse, __pyx_k_ParserElement_preParse, sizeof(__pyx_k_ParserElement_preParse), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement_resetCache, __pyx_k_ParserElement_resetCache, sizeof(__pyx_k_ParserElement_resetCache), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement_runTests, __pyx_k_ParserElement_runTests, sizeof(__pyx_k_ParserElement_runTests), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement_scanString, __pyx_k_ParserElement_scanString, sizeof(__pyx_k_ParserElement_scanString), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement_searchString, __pyx_k_ParserElement_searchString, sizeof(__pyx_k_ParserElement_searchString), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement_setBreak, __pyx_k_ParserElement_setBreak, sizeof(__pyx_k_ParserElement_setBreak), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement_setBreak_locals_br, __pyx_k_ParserElement_setBreak_locals_br, sizeof(__pyx_k_ParserElement_setBreak_locals_br), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement_setDebug, __pyx_k_ParserElement_setDebug, sizeof(__pyx_k_ParserElement_setDebug), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement_setDebugActions, __pyx_k_ParserElement_setDebugActions, sizeof(__pyx_k_ParserElement_setDebugActions), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement_setDefaultWhitespa, __pyx_k_ParserElement_setDefaultWhitespa, sizeof(__pyx_k_ParserElement_setDefaultWhitespa), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement_setFailAction, __pyx_k_ParserElement_setFailAction, sizeof(__pyx_k_ParserElement_setFailAction), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement_setName, __pyx_k_ParserElement_setName, sizeof(__pyx_k_ParserElement_setName), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement_setParseAction, __pyx_k_ParserElement_setParseAction, sizeof(__pyx_k_ParserElement_setParseAction), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement_setParseAction_loc, __pyx_k_ParserElement_setParseAction_loc, sizeof(__pyx_k_ParserElement_setParseAction_loc), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement_setResultsName, __pyx_k_ParserElement_setResultsName, sizeof(__pyx_k_ParserElement_setResultsName), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement_setWhitespaceChars, __pyx_k_ParserElement_setWhitespaceChars, sizeof(__pyx_k_ParserElement_setWhitespaceChars), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement_split, __pyx_k_ParserElement_split, sizeof(__pyx_k_ParserElement_split), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement_streamline, __pyx_k_ParserElement_streamline, sizeof(__pyx_k_ParserElement_streamline), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement_suppress, __pyx_k_ParserElement_suppress, sizeof(__pyx_k_ParserElement_suppress), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement_transformString, __pyx_k_ParserElement_transformString, sizeof(__pyx_k_ParserElement_transformString), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement_tryParse, __pyx_k_ParserElement_tryParse, sizeof(__pyx_k_ParserElement_tryParse), 0, 0, 1, 1},
  {&__pyx_n_s_ParserElement_validate, __pyx_k_ParserElement_validate, sizeof(__pyx_k_ParserElement_validate), 0, 0, 1, 1},
  {&__pyx_n_s_PendingSkip, __pyx_k_PendingSkip, sizeof(__pyx_k_PendingSkip), 0, 0, 1, 1},
  {&__pyx_n_s_PositionToken, __pyx_k_PositionToken, sizeof(__pyx_k_PositionToken), 0, 0, 1, 1},
  {&__pyx_n_s_PositionToken___init, __pyx_k_PositionToken___init, sizeof(__pyx_k_PositionToken___init), 0, 0, 1, 1},
  {&__pyx_n_s_PrecededBy, __pyx_k_PrecededBy, sizeof(__pyx_k_PrecededBy), 0, 0, 1, 1},
  {&__pyx_n_s_PrecededBy___init, __pyx_k_PrecededBy___init, sizeof(__pyx_k_PrecededBy___init), 0, 0, 1, 1},
  {&__pyx_n_s_PrecededBy___init___locals_lambd, __pyx_k_PrecededBy___init___locals_lambd, sizeof(__pyx_k_PrecededBy___init___locals_lambd), 0, 0, 1, 1},
  {&__pyx_n_s_PrecededBy_parseImpl, __pyx_k_PrecededBy_parseImpl, sizeof(__pyx_k_PrecededBy_parseImpl), 0, 0, 1, 1},
  {&__pyx_n_s_QuotedString, __pyx_k_QuotedString, sizeof(__pyx_k_QuotedString), 0, 0, 1, 1},
  {&__pyx_n_s_QuotedString___init, __pyx_k_QuotedString___init, sizeof(__pyx_k_QuotedString___init), 0, 0, 1, 1},
  {&__pyx_n_s_QuotedString___init___locals_gen, __pyx_k_QuotedString___init___locals_gen, sizeof(__pyx_k_QuotedString___init___locals_gen), 0, 0, 1, 1},
  {&__pyx_n_s_QuotedString___str, __pyx_k_QuotedString___str, sizeof(__pyx_k_QuotedString___str), 0, 0, 1, 1},
  {&__pyx_n_s_QuotedString_parseImpl, __pyx_k_QuotedString_parseImpl, sizeof(__pyx_k_QuotedString_parseImpl), 0, 0, 1, 1},
  {&__pyx_n_s_RLock, __pyx_k_RLock, sizeof(__pyx_k_RLock), 0, 0, 1, 1},
  {&__pyx_kp_s_Re_s, __pyx_k_Re_s, sizeof(__pyx_k_Re_s), 0, 0, 1, 0},
  {&__pyx_n_s_RecursiveGrammarException, __pyx_k_RecursiveGrammarException, sizeof(__pyx_k_RecursiveGrammarException), 0, 0, 1, 1},
  {&__pyx_n_s_RecursiveGrammarException___init, __pyx_k_RecursiveGrammarException___init, sizeof(__pyx_k_RecursiveGrammarException___init), 0, 0, 1, 1},
  {&__pyx_n_s_RecursiveGrammarException___str, __pyx_k_RecursiveGrammarException___str, sizeof(__pyx_k_RecursiveGrammarException___str), 0, 0, 1, 1},
  {&__pyx_kp_s_RecursiveGrammarException_s, __pyx_k_RecursiveGrammarException_s, sizeof(__pyx_k_RecursiveGrammarException_s), 0, 0, 1, 0},
  {&__pyx_n_s_Regex, __pyx_k_Regex, sizeof(__pyx_k_Regex), 0, 0, 1, 1},
  {&__pyx_n_s_Regex___init, __pyx_k_Regex___init, sizeof(__pyx_k_Regex___init), 0, 0, 1, 1},
  {&__pyx_n_s_Regex___str, __pyx_k_Regex___str, sizeof(__pyx_k_Regex___str), 0, 0, 1, 1},
  {&__pyx_kp_s_Regex_may_only_be_constructed_wi, __pyx_k_Regex_may_only_be_constructed_wi, sizeof(__pyx_k_Regex_may_only_be_constructed_wi), 0, 0, 1, 0},
  {&__pyx_n_s_Regex_parseImpl, __pyx_k_Regex_parseImpl, sizeof(__pyx_k_Regex_parseImpl), 0, 0, 1, 1},
  {&__pyx_n_s_Regex_parseImplAsGroupList, __pyx_k_Regex_parseImplAsGroupList, sizeof(__pyx_k_Regex_parseImplAsGroupList), 0, 0, 1, 1},
  {&__pyx_n_s_Regex_parseImplAsMatch, __pyx_k_Regex_parseImplAsMatch, sizeof(__pyx_k_Regex_parseImplAsMatch), 0, 0, 1, 1},
  {&__pyx_n_s_Regex_sub, __pyx_k_Regex_sub, sizeof(__pyx_k_Regex_sub), 0, 0, 1, 1},
  {&__pyx_n_s_Regex_sub_locals_pa, __pyx_k_Regex_sub_locals_pa, sizeof(__pyx_k_Regex_sub_locals_pa), 0, 0, 1, 1},
  {&__pyx_kp_s_Repetition_of_one_or_more_of_the, __pyx_k_Repetition_of_one_or_more_of_the, sizeof(__pyx_k_Repetition_of_one_or_more_of_the), 0, 0, 1, 0},
  {&__pyx_kp_s_Requires_all_given_class_ParseE, __pyx_k_Requires_all_given_class_ParseE, sizeof(__pyx_k_Requires_all_given_class_ParseE), 0, 0, 1, 0},
  {&__pyx_kp_s_Requires_all_given_class_ParseEx, __pyx_k_Requires_all_given_class_ParseEx, sizeof(__pyx_k_Requires_all_given_class_ParseEx), 0, 0, 1, 0},
  {&__pyx_kp_s_Requires_that_at_least_one_class, __pyx_k_Requires_that_at_least_one_class, sizeof(__pyx_k_Requires_that_at_least_one_class), 0, 0, 1, 0},
  {&__pyx_kp_s_Requires_that_at_least_one_class_2, __pyx_k_Requires_that_at_least_one_class_2, sizeof(__pyx_k_Requires_that_at_least_one_class_2), 0, 0, 1, 0},
  {&__pyx_kp_s_SIX_PER_EM_SPACE, __pyx_k_SIX_PER_EM_SPACE, sizeof(__pyx_k_SIX_PER_EM_SPACE), 0, 0, 1, 0},
  {&__pyx_kp_s_SP, __pyx_k_SP, sizeof(__pyx_k_SP), 0, 0, 1, 0},
  {&__pyx_n_s_SimpleNamespace, __pyx_k_SimpleNamespace, sizeof(__pyx_k_SimpleNamespace), 0, 0, 1, 1},
  {&__pyx_n_s_SingleCharLiteral, __pyx_k_SingleCharLiteral, sizeof(__pyx_k_SingleCharLiteral), 0, 0, 1, 1},
  {&__pyx_n_s_SingleCharLiteral_parseImpl, __pyx_k_SingleCharLiteral_parseImpl, sizeof(__pyx_k_SingleCharLiteral_parseImpl), 0, 0, 1, 1},
  {&__pyx_n_s_SkipTo, __pyx_k_SkipTo, sizeof(__pyx_k_SkipTo), 0, 0, 1, 1},
  {&__pyx_n_s_SkipTo___init, __pyx_k_SkipTo___init, sizeof(__pyx_k_SkipTo___init), 0, 0, 1, 1},
  {&__pyx_n_s_SkipTo_parseImpl, __pyx_k_SkipTo_parseImpl, sizeof(__pyx_k_SkipTo_parseImpl), 0, 0, 1, 1},
  {&__pyx_kp_s_Special_matching_class_for_match, __pyx_k_Special_matching_class_for_match, sizeof(__pyx_k_Special_matching_class_for_match), 0, 0, 1, 0},
  {&__pyx_n_s_StringEnd, __pyx_k_StringEnd, sizeof(__pyx_k_StringEnd), 0, 0, 1, 1},
  {&__pyx_n_s_StringEnd___init, __pyx_k_StringEnd___init, sizeof(__pyx_k_StringEnd___init), 0, 0, 1, 1},
  {&__pyx_n_s_StringEnd_parseImpl, __pyx_k_StringEnd_parseImpl, sizeof(__pyx_k_StringEnd_parseImpl), 0, 0, 1, 1},
  {&__pyx_n_s_StringStart, __pyx_k_StringStart, sizeof(__pyx_k_StringStart), 0, 0, 1, 1},
  {&__pyx_n_s_StringStart___init, __pyx_k_StringStart___init, sizeof(__pyx_k_StringStart___init), 0, 0, 1, 1},
  {&__pyx_n_s_StringStart_parseImpl, __pyx_k_StringStart_parseImpl, sizeof(__pyx_k_StringStart_parseImpl), 0, 0, 1, 1},
  {&__pyx_kp_s_Structured_parse_results_to_prov, __pyx_k_Structured_parse_results_to_prov, sizeof(__pyx_k_Structured_parse_results_to_prov), 0, 0, 1, 0},
  {&__pyx_n_s_Suppress, __pyx_k_Suppress, sizeof(__pyx_k_Suppress), 0, 0, 1, 1},
  {&__pyx_n_s_Suppress_postParse, __pyx_k_Suppress_postParse, sizeof(__pyx_k_Suppress_postParse), 0, 0, 1, 1},
  {&__pyx_n_s_Suppress_suppress, __pyx_k_Suppress_suppress, sizeof(__pyx_k_Suppress_suppress), 0, 0, 1, 1},
  {&__pyx_n_s_SyntaxError, __pyx_k_SyntaxError, sizeof(__pyx_k_SyntaxError), 0, 0, 1, 1},
  {&__pyx_n_s_SyntaxWarning, __pyx_k_SyntaxWarning, sizeof(__pyx_k_SyntaxWarning), 0, 0, 1, 1},
  {&__pyx_kp_s_TAB, __pyx_k_TAB, sizeof(__pyx_k_TAB), 0, 0, 1, 0},
  {&__pyx_kp_s_THIN_SPACE, __pyx_k_THIN_SPACE, sizeof(__pyx_k_THIN_SPACE), 0, 0, 1, 0},
  {&__pyx_kp_s_THREE_PER_EM_SPACE, __pyx_k_THREE_PER_EM_SPACE, sizeof(__pyx_k_THREE_PER_EM_SPACE), 0, 0, 1, 0},
  {&__pyx_n_s_TRY, __pyx_k_TRY, sizeof(__pyx_k_TRY), 0, 0, 1, 1},
  {&__pyx_kp_s_Text_not_in_expected_column, __pyx_k_Text_not_in_expected_column, sizeof(__pyx_k_Text_not_in_expected_column), 0, 0, 1, 0},
  {&__pyx_n_s_Token, __pyx_k_Token, sizeof(__pyx_k_Token), 0, 0, 1, 1},
  {&__pyx_n_s_TokenConverter, __pyx_k_TokenConverter, sizeof(__pyx_k_TokenConverter), 0, 0, 1, 1},
  {&__pyx_n_s_TokenConverter___init, __pyx_k_TokenConverter___init, sizeof(__pyx_k_TokenConverter___init), 0, 0, 1, 1},
  {&__pyx_n_s_Token___init, __pyx_k_Token___init, sizeof(__pyx_k_Token___init), 0, 0, 1, 1},
  {&__pyx_kp_s_Token_for_matching_strings_that, __pyx_k_Token_for_matching_strings_that, sizeof(__pyx_k_Token_for_matching_strings_that), 0, 0, 1, 0},
  {&__pyx_kp_s_Token_for_matching_strings_that_2, __pyx_k_Token_for_matching_strings_that_2, sizeof(__pyx_k_Token_for_matching_strings_that_2), 0, 0, 1, 0},
  {&__pyx_kp_s_Token_for_matching_words_compose, __pyx_k_Token_for_matching_words_compose, sizeof(__pyx_k_Token_for_matching_words_compose), 0, 0, 1, 0},
  {&__pyx_kp_s_Token_for_matching_words_compose_2, __pyx_k_Token_for_matching_words_compose_2, sizeof(__pyx_k_Token_for_matching_words_compose_2), 0, 0, 1, 0},
  {&__pyx_kp_s_Token_for_skipping_over_all_unde, __pyx_k_Token_for_skipping_over_all_unde, sizeof(__pyx_k_Token_for_skipping_over_all_unde), 0, 0, 1, 0},
  {&__pyx_kp_s_Token_to_advance_to_a_specific_c, __pyx_k_Token_to_advance_to_a_specific_c, sizeof(__pyx_k_Token_to_advance_to_a_specific_c), 0, 0, 1, 0},
  {&__pyx_kp_s_Token_to_exactly_match_a_specifi, __pyx_k_Token_to_exactly_match_a_specifi, sizeof(__pyx_k_Token_to_exactly_match_a_specifi), 0, 0, 1, 0},
  {&__pyx_kp_s_Token_to_exactly_match_a_specifi_2, __pyx_k_Token_to_exactly_match_a_specifi_2, sizeof(__pyx_k_Token_to_exactly_match_a_specifi_2), 0, 0, 1, 0},
  {&__pyx_kp_s_Token_to_match_a_specified_strin, __pyx_k_Token_to_match_a_specified_strin, sizeof(__pyx_k_Token_to_match_a_specified_strin), 0, 0, 1, 0},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_UnboundedCache, __pyx_k_UnboundedCache, sizeof(__pyx_k_UnboundedCache), 0, 0, 1, 1},
  {&__pyx_kp_s_Unmatchable_token, __pyx_k_Unmatchable_token, sizeof(__pyx_k_Unmatchable_token), 0, 0, 1, 0},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_kp_s_W_s, __pyx_k_W_s, sizeof(__pyx_k_W_s), 0, 0, 1, 0},
  {&__pyx_kp_s_W_s_2, __pyx_k_W_s_2, sizeof(__pyx_k_W_s_2), 0, 0, 1, 0},
  {&__pyx_kp_s_W_s_3, __pyx_k_W_s_3, sizeof(__pyx_k_W_s_3), 0, 0, 1, 0},
  {&__pyx_kp_s_W_s_s, __pyx_k_W_s_s, sizeof(__pyx_k_W_s_s), 0, 0, 1, 0},
  {&__pyx_n_s_White, __pyx_k_White, sizeof(__pyx_k_White), 0, 0, 1, 1},
  {&__pyx_n_s_White___init, __pyx_k_White___init, sizeof(__pyx_k_White___init), 0, 0, 1, 1},
  {&__pyx_n_s_White___init___locals_genexpr, __pyx_k_White___init___locals_genexpr, sizeof(__pyx_k_White___init___locals_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_White_parseImpl, __pyx_k_White_parseImpl, sizeof(__pyx_k_White_parseImpl), 0, 0, 1, 1},
  {&__pyx_n_s_Word, __pyx_k_Word, sizeof(__pyx_k_Word), 0, 0, 1, 1},
  {&__pyx_n_s_WordEnd, __pyx_k_WordEnd, sizeof(__pyx_k_WordEnd), 0, 0, 1, 1},
  {&__pyx_n_s_WordEnd___init, __pyx_k_WordEnd___init, sizeof(__pyx_k_WordEnd___init), 0, 0, 1, 1},
  {&__pyx_n_s_WordEnd_parseImpl, __pyx_k_WordEnd_parseImpl, sizeof(__pyx_k_WordEnd_parseImpl), 0, 0, 1, 1},
  {&__pyx_n_s_WordRegex, __pyx_k_WordRegex, sizeof(__pyx_k_WordRegex), 0, 0, 1, 1},
  {&__pyx_n_s_WordRegex_parseImpl, __pyx_k_WordRegex_parseImpl, sizeof(__pyx_k_WordRegex_parseImpl), 0, 0, 1, 1},
  {&__pyx_n_s_WordStart, __pyx_k_WordStart, sizeof(__pyx_k_WordStart), 0, 0, 1, 1},
  {&__pyx_n_s_WordStart___init, __pyx_k_WordStart___init, sizeof(__pyx_k_WordStart___init), 0, 0, 1, 1},
  {&__pyx_n_s_WordStart_parseImpl, __pyx_k_WordStart_parseImpl, sizeof(__pyx_k_WordStart_parseImpl), 0, 0, 1, 1},
  {&__pyx_n_s_Word___init, __pyx_k_Word___init, sizeof(__pyx_k_Word___init), 0, 0, 1, 1},
  {&__pyx_n_s_Word___init___locals_genexpr, __pyx_k_Word___init___locals_genexpr, sizeof(__pyx_k_Word___init___locals_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_Word___str, __pyx_k_Word___str, sizeof(__pyx_k_Word___str), 0, 0, 1, 1},
  {&__pyx_n_s_Word___str___locals_charsAsStr, __pyx_k_Word___str___locals_charsAsStr, sizeof(__pyx_k_Word___str___locals_charsAsStr), 0, 0, 1, 1},
  {&__pyx_n_s_Word_parseImpl, __pyx_k_Word_parseImpl, sizeof(__pyx_k_Word_parseImpl), 0, 0, 1, 1},
  {&__pyx_kp_s_Wrapper_for_parse_actions_to_ens, __pyx_k_Wrapper_for_parse_actions_to_ens, sizeof(__pyx_k_Wrapper_for_parse_actions_to_ens), 0, 0, 1, 0},
  {&__pyx_kp_s_ZERO_WIDTH_SPACE, __pyx_k_ZERO_WIDTH_SPACE, sizeof(__pyx_k_ZERO_WIDTH_SPACE), 0, 0, 1, 0},
  {&__pyx_n_s_ZeroOrMore, __pyx_k_ZeroOrMore, sizeof(__pyx_k_ZeroOrMore), 0, 0, 1, 1},
  {&__pyx_n_s_ZeroOrMore___init, __pyx_k_ZeroOrMore___init, sizeof(__pyx_k_ZeroOrMore___init), 0, 0, 1, 1},
  {&__pyx_n_s_ZeroOrMore___str, __pyx_k_ZeroOrMore___str, sizeof(__pyx_k_ZeroOrMore___str), 0, 0, 1, 1},
  {&__pyx_n_s_ZeroOrMore_parseImpl, __pyx_k_ZeroOrMore_parseImpl, sizeof(__pyx_k_ZeroOrMore_parseImpl), 0, 0, 1, 1},
  {&__pyx_kp_s__10, __pyx_k__10, sizeof(__pyx_k__10), 0, 0, 1, 0},
  {&__pyx_kp_s__100, __pyx_k__100, sizeof(__pyx_k__100), 0, 0, 1, 0},
  {&__pyx_kp_s__101, __pyx_k__101, sizeof(__pyx_k__101), 0, 0, 1, 0},
  {&__pyx_kp_s__102, __pyx_k__102, sizeof(__pyx_k__102), 0, 0, 1, 0},
  {&__pyx_kp_s__103, __pyx_k__103, sizeof(__pyx_k__103), 0, 0, 1, 0},
  {&__pyx_kp_s__105, __pyx_k__105, sizeof(__pyx_k__105), 0, 0, 1, 0},
  {&__pyx_kp_s__107, __pyx_k__107, sizeof(__pyx_k__107), 0, 0, 1, 0},
  {&__pyx_kp_s__109, __pyx_k__109, sizeof(__pyx_k__109), 0, 0, 1, 0},
  {&__pyx_kp_s__110, __pyx_k__110, sizeof(__pyx_k__110), 0, 0, 1, 0},
  {&__pyx_kp_s__112, __pyx_k__112, sizeof(__pyx_k__112), 0, 0, 1, 0},
  {&__pyx_kp_s__113, __pyx_k__113, sizeof(__pyx_k__113), 0, 0, 1, 0},
  {&__pyx_kp_s__114, __pyx_k__114, sizeof(__pyx_k__114), 0, 0, 1, 0},
  {&__pyx_kp_s__115, __pyx_k__115, sizeof(__pyx_k__115), 0, 0, 1, 0},
  {&__pyx_kp_s__116, __pyx_k__116, sizeof(__pyx_k__116), 0, 0, 1, 0},
  {&__pyx_kp_s__117, __pyx_k__117, sizeof(__pyx_k__117), 0, 0, 1, 0},
  {&__pyx_kp_s__118, __pyx_k__118, sizeof(__pyx_k__118), 0, 0, 1, 0},
  {&__pyx_kp_s__12, __pyx_k__12, sizeof(__pyx_k__12), 0, 0, 1, 0},
  {&__pyx_kp_s__120, __pyx_k__120, sizeof(__pyx_k__120), 0, 0, 1, 0},
  {&__pyx_kp_s__14, __pyx_k__14, sizeof(__pyx_k__14), 0, 0, 1, 0},
  {&__pyx_kp_s__15, __pyx_k__15, sizeof(__pyx_k__15), 0, 0, 1, 0},
  {&__pyx_kp_s__16, __pyx_k__16, sizeof(__pyx_k__16), 0, 0, 1, 0},
  {&__pyx_kp_s__19, __pyx_k__19, sizeof(__pyx_k__19), 0, 0, 1, 0},
  {&__pyx_kp_s__2, __pyx_k__2, sizeof(__pyx_k__2), 0, 0, 1, 0},
  {&__pyx_kp_s__20, __pyx_k__20, sizeof(__pyx_k__20), 0, 0, 1, 0},
  {&__pyx_kp_s__21, __pyx_k__21, sizeof(__pyx_k__21), 0, 0, 1, 0},
  {&__pyx_kp_s__22, __pyx_k__22, sizeof(__pyx_k__22), 0, 0, 1, 0},
  {&__pyx_kp_s__25, __pyx_k__25, sizeof(__pyx_k__25), 0, 0, 1, 0},
  {&__pyx_kp_s__26, __pyx_k__26, sizeof(__pyx_k__26), 0, 0, 1, 0},
  {&__pyx_kp_s__268, __pyx_k__268, sizeof(__pyx_k__268), 0, 0, 1, 0},
  {&__pyx_kp_s__28, __pyx_k__28, sizeof(__pyx_k__28), 0, 0, 1, 0},
  {&__pyx_kp_s__29, __pyx_k__29, sizeof(__pyx_k__29), 0, 0, 1, 0},
  {&__pyx_kp_s__30, __pyx_k__30, sizeof(__pyx_k__30), 0, 0, 1, 0},
  {&__pyx_kp_s__33, __pyx_k__33, sizeof(__pyx_k__33), 0, 0, 1, 0},
  {&__pyx_kp_s__4, __pyx_k__4, sizeof(__pyx_k__4), 0, 0, 1, 0},
  {&__pyx_kp_s__456, __pyx_k__456, sizeof(__pyx_k__456), 0, 0, 1, 0},
  {&__pyx_kp_u__527, __pyx_k__527, sizeof(__pyx_k__527), 0, 1, 0, 0},
  {&__pyx_kp_u__528, __pyx_k__528, sizeof(__pyx_k__528), 0, 1, 0, 0},
  {&__pyx_kp_u__529, __pyx_k__529, sizeof(__pyx_k__529), 0, 1, 0, 0},
  {&__pyx_kp_u__530, __pyx_k__530, sizeof(__pyx_k__530), 0, 1, 0, 0},
  {&__pyx_kp_u__531, __pyx_k__531, sizeof(__pyx_k__531), 0, 1, 0, 0},
  {&__pyx_kp_u__532, __pyx_k__532, sizeof(__pyx_k__532), 0, 1, 0, 0},
  {&__pyx_kp_u__533, __pyx_k__533, sizeof(__pyx_k__533), 0, 1, 0, 0},
  {&__pyx_kp_u__534, __pyx_k__534, sizeof(__pyx_k__534), 0, 1, 0, 0},
  {&__pyx_kp_u__535, __pyx_k__535, sizeof(__pyx_k__535), 0, 1, 0, 0},
  {&__pyx_kp_u__536, __pyx_k__536, sizeof(__pyx_k__536), 0, 1, 0, 0},
  {&__pyx_kp_u__537, __pyx_k__537, sizeof(__pyx_k__537), 0, 1, 0, 0},
  {&__pyx_kp_u__538, __pyx_k__538, sizeof(__pyx_k__538), 0, 1, 0, 0},
  {&__pyx_kp_u__539, __pyx_k__539, sizeof(__pyx_k__539), 0, 1, 0, 0},
  {&__pyx_kp_u__540, __pyx_k__540, sizeof(__pyx_k__540), 0, 1, 0, 0},
  {&__pyx_kp_u__541, __pyx_k__541, sizeof(__pyx_k__541), 0, 1, 0, 0},
  {&__pyx_kp_u__542, __pyx_k__542, sizeof(__pyx_k__542), 0, 1, 0, 0},
  {&__pyx_kp_u__543, __pyx_k__543, sizeof(__pyx_k__543), 0, 1, 0, 0},
  {&__pyx_kp_u__544, __pyx_k__544, sizeof(__pyx_k__544), 0, 1, 0, 0},
  {&__pyx_kp_s__6, __pyx_k__6, sizeof(__pyx_k__6), 0, 0, 1, 0},
  {&__pyx_n_s__673, __pyx_k__673, sizeof(__pyx_k__673), 0, 0, 1, 1},
  {&__pyx_kp_s__79, __pyx_k__79, sizeof(__pyx_k__79), 0, 0, 1, 0},
  {&__pyx_kp_s__8, __pyx_k__8, sizeof(__pyx_k__8), 0, 0, 1, 0},
  {&__pyx_kp_u__80, __pyx_k__80, sizeof(__pyx_k__80), 0, 1, 0, 0},
  {&__pyx_kp_s__85, __pyx_k__85, sizeof(__pyx_k__85), 0, 0, 1, 0},
  {&__pyx_kp_s__9, __pyx_k__9, sizeof(__pyx_k__9), 0, 0, 1, 0},
  {&__pyx_kp_s__98, __pyx_k__98, sizeof(__pyx_k__98), 0, 0, 1, 0},
  {&__pyx_kp_s__99, __pyx_k__99, sizeof(__pyx_k__99), 0, 0, 1, 0},
  {&__pyx_n_s_add, __pyx_k_add, sizeof(__pyx_k_add), 0, 0, 1, 1},
  {&__pyx_n_s_addCondition, __pyx_k_addCondition, sizeof(__pyx_k_addCondition), 0, 0, 1, 1},
  {&__pyx_n_s_addParseAction, __pyx_k_addParseAction, sizeof(__pyx_k_addParseAction), 0, 0, 1, 1},
  {&__pyx_n_s_addoffset, __pyx_k_addoffset, sizeof(__pyx_k_addoffset), 0, 0, 1, 1},
  {&__pyx_n_s_adjacent, __pyx_k_adjacent, sizeof(__pyx_k_adjacent), 0, 0, 1, 1},
  {&__pyx_n_s_allResults, __pyx_k_allResults, sizeof(__pyx_k_allResults), 0, 0, 1, 1},
  {&__pyx_n_s_all_names, __pyx_k_all_names, sizeof(__pyx_k_all_names), 0, 0, 1, 1},
  {&__pyx_n_s_alphanums, __pyx_k_alphanums, sizeof(__pyx_k_alphanums), 0, 0, 1, 1},
  {&__pyx_n_s_alphas, __pyx_k_alphas, sizeof(__pyx_k_alphas), 0, 0, 1, 1},
  {&__pyx_kp_s_amp_gt_lt_quot_apos, __pyx_k_amp_gt_lt_quot_apos, sizeof(__pyx_k_amp_gt_lt_quot_apos), 0, 0, 1, 0},
  {&__pyx_n_s_aname, __pyx_k_aname, sizeof(__pyx_k_aname), 0, 0, 1, 1},
  {&__pyx_n_s_and, __pyx_k_and, sizeof(__pyx_k_and), 0, 0, 1, 1},
  {&__pyx_n_s_append, __pyx_k_append, sizeof(__pyx_k_append), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_asDict, __pyx_k_asDict, sizeof(__pyx_k_asDict), 0, 0, 1, 1},
  {&__pyx_n_s_asGroupList, __pyx_k_asGroupList, sizeof(__pyx_k_asGroupList), 0, 0, 1, 1},
  {&__pyx_n_s_asKeyword, __pyx_k_asKeyword, sizeof(__pyx_k_asKeyword), 0, 0, 1, 1},
  {&__pyx_n_s_asList, __pyx_k_asList, sizeof(__pyx_k_asList), 0, 0, 1, 1},
  {&__pyx_n_s_asMatch, __pyx_k_asMatch, sizeof(__pyx_k_asMatch), 0, 0, 1, 1},
  {&__pyx_n_s_asStringList, __pyx_k_asStringList, sizeof(__pyx_k_asStringList), 0, 0, 1, 1},
  {&__pyx_n_s_asXML, __pyx_k_asXML, sizeof(__pyx_k_asXML), 0, 0, 1, 1},
  {&__pyx_n_s_ascii_lowercase, __pyx_k_ascii_lowercase, sizeof(__pyx_k_ascii_lowercase), 0, 0, 1, 1},
  {&__pyx_n_s_ascii_uppercase, __pyx_k_ascii_uppercase, sizeof(__pyx_k_ascii_uppercase), 0, 0, 1, 1},
  {&__pyx_kp_s_at_loc, __pyx_k_at_loc, sizeof(__pyx_k_at_loc), 0, 0, 1, 0},
  {&__pyx_kp_s_b, __pyx_k_b, sizeof(__pyx_k_b), 0, 0, 1, 0},
  {&__pyx_kp_s_b_s_b, __pyx_k_b_s_b, sizeof(__pyx_k_b_s_b), 0, 0, 1, 0},
  {&__pyx_kp_s_base_exception_class_for_all_par, __pyx_k_base_exception_class_for_all_par, sizeof(__pyx_k_base_exception_class_for_all_par), 0, 0, 1, 0},
  {&__pyx_n_s_best_expr, __pyx_k_best_expr, sizeof(__pyx_k_best_expr), 0, 0, 1, 1},
  {&__pyx_n_s_bodyChars, __pyx_k_bodyChars, sizeof(__pyx_k_bodyChars), 0, 0, 1, 1},
  {&__pyx_n_s_bodyCharsOrig, __pyx_k_bodyCharsOrig, sizeof(__pyx_k_bodyCharsOrig), 0, 0, 1, 1},
  {&__pyx_n_s_bodychars, __pyx_k_bodychars, sizeof(__pyx_k_bodychars), 0, 0, 1, 1},
  {&__pyx_n_s_bool, __pyx_k_bool, sizeof(__pyx_k_bool), 0, 0, 1, 1},
  {&__pyx_n_s_breakFlag, __pyx_k_breakFlag, sizeof(__pyx_k_breakFlag), 0, 0, 1, 1},
  {&__pyx_n_s_breaker, __pyx_k_breaker, sizeof(__pyx_k_breaker), 0, 0, 1, 1},
  {&__pyx_n_s_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 0, 1, 1},
  {&__pyx_n_s_cache, __pyx_k_cache, sizeof(__pyx_k_cache), 0, 0, 1, 1},
  {&__pyx_n_s_cache_len, __pyx_k_cache_len, sizeof(__pyx_k_cache_len), 0, 0, 1, 1},
  {&__pyx_n_s_cache_size_limit, __pyx_k_cache_size_limit, sizeof(__pyx_k_cache_size_limit), 0, 0, 1, 1},
  {&__pyx_n_s_call, __pyx_k_call, sizeof(__pyx_k_call), 0, 0, 1, 1},
  {&__pyx_n_s_callDuringTry, __pyx_k_callDuringTry, sizeof(__pyx_k_callDuringTry), 0, 0, 1, 1},
  {&__pyx_n_s_callPreParse, __pyx_k_callPreParse, sizeof(__pyx_k_callPreParse), 0, 0, 1, 1},
  {&__pyx_n_s_callPreparse, __pyx_k_callPreparse, sizeof(__pyx_k_callPreparse), 0, 0, 1, 1},
  {&__pyx_n_s_callable, __pyx_k_callable, sizeof(__pyx_k_callable), 0, 0, 1, 1},
  {&__pyx_n_s_called, __pyx_k_called, sizeof(__pyx_k_called), 0, 0, 1, 1},
  {&__pyx_n_s_callers, __pyx_k_callers, sizeof(__pyx_k_callers), 0, 0, 1, 1},
  {&__pyx_n_s_canParseNext, __pyx_k_canParseNext, sizeof(__pyx_k_canParseNext), 0, 0, 1, 1},
  {&__pyx_kp_s_cannot_construct_And_with_sequen, __pyx_k_cannot_construct_And_with_sequen, sizeof(__pyx_k_cannot_construct_And_with_sequen), 0, 0, 1, 0},
  {&__pyx_kp_s_cannot_multiply_ParserElement_an, __pyx_k_cannot_multiply_ParserElement_an, sizeof(__pyx_k_cannot_multiply_ParserElement_an), 0, 0, 1, 0},
  {&__pyx_kp_s_cannot_multiply_ParserElement_an_2, __pyx_k_cannot_multiply_ParserElement_an_2, sizeof(__pyx_k_cannot_multiply_ParserElement_an_2), 0, 0, 1, 0},
  {&__pyx_kp_s_cannot_multiply_ParserElement_by, __pyx_k_cannot_multiply_ParserElement_by, sizeof(__pyx_k_cannot_multiply_ParserElement_by), 0, 0, 1, 0},
  {&__pyx_kp_s_cannot_multiply_ParserElement_by_2, __pyx_k_cannot_multiply_ParserElement_by_2, sizeof(__pyx_k_cannot_multiply_ParserElement_by_2), 0, 0, 1, 0},
  {&__pyx_kp_s_cannot_specify_a_minimum_length, __pyx_k_cannot_specify_a_minimum_length, sizeof(__pyx_k_cannot_specify_a_minimum_length), 0, 0, 1, 0},
  {&__pyx_kp_s_cannot_specify_a_minimum_length_2, __pyx_k_cannot_specify_a_minimum_length_2, sizeof(__pyx_k_cannot_specify_a_minimum_length_2), 0, 0, 1, 0},
  {&__pyx_kp_s_cannot_use_sub_with_Regex_asGrou, __pyx_k_cannot_use_sub_with_Regex_asGrou, sizeof(__pyx_k_cannot_use_sub_with_Regex_asGrou), 0, 0, 1, 0},
  {&__pyx_kp_s_cannot_use_sub_with_a_callable_w, __pyx_k_cannot_use_sub_with_a_callable_w, sizeof(__pyx_k_cannot_use_sub_with_a_callable_w), 0, 0, 1, 0},
  {&__pyx_n_s_caseless, __pyx_k_caseless, sizeof(__pyx_k_caseless), 0, 0, 1, 1},
  {&__pyx_n_s_caselessmatch, __pyx_k_caselessmatch, sizeof(__pyx_k_caselessmatch), 0, 0, 1, 1},
  {&__pyx_n_s_cause, __pyx_k_cause, sizeof(__pyx_k_cause), 0, 0, 1, 1},
  {&__pyx_n_s_chars, __pyx_k_chars, sizeof(__pyx_k_chars), 0, 0, 1, 1},
  {&__pyx_n_s_charsAsStr, __pyx_k_charsAsStr, sizeof(__pyx_k_charsAsStr), 0, 0, 1, 1},
  {&__pyx_n_s_charset, __pyx_k_charset, sizeof(__pyx_k_charset), 0, 0, 1, 1},
  {&__pyx_n_s_checkRecursion, __pyx_k_checkRecursion, sizeof(__pyx_k_checkRecursion), 0, 0, 1, 1},
  {&__pyx_n_s_check_ender, __pyx_k_check_ender, sizeof(__pyx_k_check_ender), 0, 0, 1, 1},
  {&__pyx_n_s_chr, __pyx_k_chr, sizeof(__pyx_k_chr), 0, 0, 1, 1},
  {&__pyx_n_s_class, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
  {&__pyx_n_s_clear, __pyx_k_clear, sizeof(__pyx_k_clear), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_cls, __pyx_k_cls, sizeof(__pyx_k_cls), 0, 0, 1, 1},
  {&__pyx_n_s_co_name, __pyx_k_co_name, sizeof(__pyx_k_co_name), 0, 0, 1, 1},
  {&__pyx_n_s_code, __pyx_k_code, sizeof(__pyx_k_code), 0, 0, 1, 1},
  {&__pyx_n_s_col, __pyx_k_col, sizeof(__pyx_k_col), 0, 0, 1, 1},
  {&__pyx_n_s_collect_all_And_tokens, __pyx_k_collect_all_And_tokens, sizeof(__pyx_k_collect_all_And_tokens), 0, 0, 1, 1},
  {&__pyx_n_s_collections, __pyx_k_collections, sizeof(__pyx_k_collections), 0, 0, 1, 1},
  {&__pyx_n_s_collections_abc, __pyx_k_collections_abc, sizeof(__pyx_k_collections_abc), 0, 0, 1, 1},
  {&__pyx_n_s_colno, __pyx_k_colno, sizeof(__pyx_k_colno), 0, 0, 1, 1},
  {&__pyx_n_s_column, __pyx_k_column, sizeof(__pyx_k_column), 0, 0, 1, 1},
  {&__pyx_n_s_comment, __pyx_k_comment, sizeof(__pyx_k_comment), 0, 0, 1, 1},
  {&__pyx_n_s_comments, __pyx_k_comments, sizeof(__pyx_k_comments), 0, 0, 1, 1},
  {&__pyx_n_s_compile, __pyx_k_compile, sizeof(__pyx_k_compile), 0, 0, 1, 1},
  {&__pyx_n_s_conditionAsParseAction, __pyx_k_conditionAsParseAction, sizeof(__pyx_k_conditionAsParseAction), 0, 0, 1, 1},
  {&__pyx_n_s_contains, __pyx_k_contains, sizeof(__pyx_k_contains), 0, 0, 1, 1},
  {&__pyx_n_s_context, __pyx_k_context, sizeof(__pyx_k_context), 0, 0, 1, 1},
  {&__pyx_n_s_convertWhitespaceEscapes, __pyx_k_convertWhitespaceEscapes, sizeof(__pyx_k_convertWhitespaceEscapes), 0, 0, 1, 1},
  {&__pyx_n_s_copy, __pyx_k_copy, sizeof(__pyx_k_copy), 0, 0, 1, 1},
  {&__pyx_n_s_copyDefaultWhiteChars, __pyx_k_copyDefaultWhiteChars, sizeof(__pyx_k_copyDefaultWhiteChars), 0, 0, 1, 1},
  {&__pyx_n_s_count, __pyx_k_count, sizeof(__pyx_k_count), 0, 0, 1, 1},
  {&__pyx_n_s_cpy, __pyx_k_cpy, sizeof(__pyx_k_cpy), 0, 0, 1, 1},
  {&__pyx_n_s_d, __pyx_k_d, sizeof(__pyx_k_d), 0, 0, 1, 1},
  {&__pyx_kp_s_d_d, __pyx_k_d_d, sizeof(__pyx_k_d_d), 0, 0, 1, 0},
  {&__pyx_n_s_data, __pyx_k_data, sizeof(__pyx_k_data), 0, 0, 1, 1},
  {&__pyx_n_s_debug, __pyx_k_debug, sizeof(__pyx_k_debug), 0, 0, 1, 1},
  {&__pyx_n_s_debugActions, __pyx_k_debugActions, sizeof(__pyx_k_debugActions), 0, 0, 1, 1},
  {&__pyx_n_s_debugging, __pyx_k_debugging, sizeof(__pyx_k_debugging), 0, 0, 1, 1},
  {&__pyx_n_s_default, __pyx_k_default, sizeof(__pyx_k_default), 0, 0, 1, 1},
  {&__pyx_n_s_defaultExceptionDebugAction, __pyx_k_defaultExceptionDebugAction, sizeof(__pyx_k_defaultExceptionDebugAction), 0, 0, 1, 1},
  {&__pyx_n_s_defaultStartDebugAction, __pyx_k_defaultStartDebugAction, sizeof(__pyx_k_defaultStartDebugAction), 0, 0, 1, 1},
  {&__pyx_n_s_defaultSuccessDebugAction, __pyx_k_defaultSuccessDebugAction, sizeof(__pyx_k_defaultSuccessDebugAction), 0, 0, 1, 1},
  {&__pyx_n_s_defaultValue, __pyx_k_defaultValue, sizeof(__pyx_k_defaultValue), 0, 0, 1, 1},
  {&__pyx_n_s_defaultvalue, __pyx_k_defaultvalue, sizeof(__pyx_k_defaultvalue), 0, 0, 1, 1},
  {&__pyx_n_s_delitem, __pyx_k_delitem, sizeof(__pyx_k_delitem), 0, 0, 1, 1},
  {&__pyx_n_s_depth, __pyx_k_depth, sizeof(__pyx_k_depth), 0, 0, 1, 1},
  {&__pyx_n_s_depth_2, __pyx_k_depth_2, sizeof(__pyx_k_depth_2), 0, 0, 1, 1},
  {&__pyx_n_s_deque, __pyx_k_deque, sizeof(__pyx_k_deque), 0, 0, 1, 1},
  {&__pyx_n_s_diag, __pyx_k_diag, sizeof(__pyx_k_diag), 0, 0, 1, 1},
  {&__pyx_n_s_dictvalue, __pyx_k_dictvalue, sizeof(__pyx_k_dictvalue), 0, 0, 1, 1},
  {&__pyx_n_s_dir, __pyx_k_dir, sizeof(__pyx_k_dir), 0, 0, 1, 1},
  {&__pyx_n_s_doActions, __pyx_k_doActions, sizeof(__pyx_k_doActions), 0, 0, 1, 1},
  {&__pyx_n_s_doc, __pyx_k_doc, sizeof(__pyx_k_doc), 0, 0, 1, 1},
  {&__pyx_n_s_doctag, __pyx_k_doctag, sizeof(__pyx_k_doctag), 0, 0, 1, 1},
  {&__pyx_n_s_dummy, __pyx_k_dummy, sizeof(__pyx_k_dummy), 0, 0, 1, 1},
  {&__pyx_n_s_dump, __pyx_k_dump, sizeof(__pyx_k_dump), 0, 0, 1, 1},
  {&__pyx_n_s_e, __pyx_k_e, sizeof(__pyx_k_e), 0, 0, 1, 1},
  {&__pyx_n_s_elem, __pyx_k_elem, sizeof(__pyx_k_elem), 0, 0, 1, 1},
  {&__pyx_n_s_enable, __pyx_k_enable, sizeof(__pyx_k_enable), 0, 0, 1, 1},
  {&__pyx_n_s_enablePackrat, __pyx_k_enablePackrat, sizeof(__pyx_k_enablePackrat), 0, 0, 1, 1},
  {&__pyx_n_s_enable_all_warnings, __pyx_k_enable_all_warnings, sizeof(__pyx_k_enable_all_warnings), 0, 0, 1, 1},
  {&__pyx_n_s_enable_all_warnings_2, __pyx_k_enable_all_warnings_2, sizeof(__pyx_k_enable_all_warnings_2), 0, 0, 1, 1},
  {&__pyx_n_s_enable_debug_on_named_expression, __pyx_k_enable_debug_on_named_expression, sizeof(__pyx_k_enable_debug_on_named_expression), 0, 0, 1, 1},
  {&__pyx_n_s_end, __pyx_k_end, sizeof(__pyx_k_end), 0, 0, 1, 1},
  {&__pyx_n_s_endQuoteChar, __pyx_k_endQuoteChar, sizeof(__pyx_k_endQuoteChar), 0, 0, 1, 1},
  {&__pyx_n_s_endQuoteCharLen, __pyx_k_endQuoteCharLen, sizeof(__pyx_k_endQuoteCharLen), 0, 0, 1, 1},
  {&__pyx_kp_s_endQuoteChar_cannot_be_the_empty, __pyx_k_endQuoteChar_cannot_be_the_empty, sizeof(__pyx_k_endQuoteChar_cannot_be_the_empty), 0, 0, 1, 0},
  {&__pyx_n_s_ender, __pyx_k_ender, sizeof(__pyx_k_ender), 0, 0, 1, 1},
  {&__pyx_n_s_endloc, __pyx_k_endloc, sizeof(__pyx_k_endloc), 0, 0, 1, 1},
  {&__pyx_n_s_endswith, __pyx_k_endswith, sizeof(__pyx_k_endswith), 0, 0, 1, 1},
  {&__pyx_n_s_enter, __pyx_k_enter, sizeof(__pyx_k_enter), 0, 0, 1, 1},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_n_s_eq, __pyx_k_eq, sizeof(__pyx_k_eq), 0, 0, 1, 1},
  {&__pyx_n_s_err, __pyx_k_err, sizeof(__pyx_k_err), 0, 0, 1, 1},
  {&__pyx_n_s_errmsg, __pyx_k_errmsg, sizeof(__pyx_k_errmsg), 0, 0, 1, 1},
  {&__pyx_n_s_error, __pyx_k_error, sizeof(__pyx_k_error), 0, 0, 1, 1},
  {&__pyx_n_s_errorStop, __pyx_k_errorStop, sizeof(__pyx_k_errorStop), 0, 0, 1, 1},
  {&__pyx_n_s_escChar, __pyx_k_escChar, sizeof(__pyx_k_escChar), 0, 0, 1, 1},
  {&__pyx_n_s_escCharReplacePattern, __pyx_k_escCharReplacePattern, sizeof(__pyx_k_escCharReplacePattern), 0, 0, 1, 1},
  {&__pyx_n_s_escQuote, __pyx_k_escQuote, sizeof(__pyx_k_escQuote), 0, 0, 1, 1},
  {&__pyx_n_s_escape, __pyx_k_escape, sizeof(__pyx_k_escape), 0, 0, 1, 1},
  {&__pyx_n_s_escapeRegexRangeChars, __pyx_k_escapeRegexRangeChars, sizeof(__pyx_k_escapeRegexRangeChars), 0, 0, 1, 1},
  {&__pyx_n_s_exact, __pyx_k_exact, sizeof(__pyx_k_exact), 0, 0, 1, 1},
  {&__pyx_n_s_exc, __pyx_k_exc, sizeof(__pyx_k_exc), 0, 0, 1, 1},
  {&__pyx_n_s_exceptionAction, __pyx_k_exceptionAction, sizeof(__pyx_k_exceptionAction), 0, 0, 1, 1},
  {&__pyx_kp_s_exception_raised_in_parse_action, __pyx_k_exception_raised_in_parse_action, sizeof(__pyx_k_exception_raised_in_parse_action), 0, 0, 1, 0},
  {&__pyx_kp_s_exception_thrown_by_class_Parser, __pyx_k_exception_thrown_by_class_Parser, sizeof(__pyx_k_exception_thrown_by_class_Parser), 0, 0, 1, 0},
  {&__pyx_n_s_excludeChars, __pyx_k_excludeChars, sizeof(__pyx_k_excludeChars), 0, 0, 1, 1},
  {&__pyx_n_s_exit, __pyx_k_exit, sizeof(__pyx_k_exit), 0, 0, 1, 1},
  {&__pyx_n_s_expand, __pyx_k_expand, sizeof(__pyx_k_expand), 0, 0, 1, 1},
  {&__pyx_n_s_expandtabs, __pyx_k_expandtabs, sizeof(__pyx_k_expandtabs), 0, 0, 1, 1},
  {&__pyx_n_s_explain, __pyx_k_explain, sizeof(__pyx_k_explain), 0, 0, 1, 1},
  {&__pyx_n_s_expr, __pyx_k_expr, sizeof(__pyx_k_expr), 0, 0, 1, 1},
  {&__pyx_n_s_expr1, __pyx_k_expr1, sizeof(__pyx_k_expr1), 0, 0, 1, 1},
  {&__pyx_n_s_expr_parse, __pyx_k_expr_parse, sizeof(__pyx_k_expr_parse), 0, 0, 1, 1},
  {&__pyx_n_s_exprs, __pyx_k_exprs, sizeof(__pyx_k_exprs), 0, 0, 1, 1},
  {&__pyx_n_s_exprsFound, __pyx_k_exprsFound, sizeof(__pyx_k_exprsFound), 0, 0, 1, 1},
  {&__pyx_n_s_exprtokens, __pyx_k_exprtokens, sizeof(__pyx_k_exprtokens), 0, 0, 1, 1},
  {&__pyx_n_s_extend, __pyx_k_extend, sizeof(__pyx_k_extend), 0, 0, 1, 1},
  {&__pyx_kp_s_f, __pyx_k_f, sizeof(__pyx_k_f), 0, 0, 1, 0},
  {&__pyx_n_s_f_2, __pyx_k_f_2, sizeof(__pyx_k_f_2), 0, 0, 1, 1},
  {&__pyx_n_s_f_code, __pyx_k_f_code, sizeof(__pyx_k_f_code), 0, 0, 1, 1},
  {&__pyx_n_s_f_locals, __pyx_k_f_locals, sizeof(__pyx_k_f_locals), 0, 0, 1, 1},
  {&__pyx_n_s_f_self, __pyx_k_f_self, sizeof(__pyx_k_f_self), 0, 0, 1, 1},
  {&__pyx_n_s_failAction, __pyx_k_failAction, sizeof(__pyx_k_failAction), 0, 0, 1, 1},
  {&__pyx_n_s_failOn, __pyx_k_failOn, sizeof(__pyx_k_failOn), 0, 0, 1, 1},
  {&__pyx_n_s_failed, __pyx_k_failed, sizeof(__pyx_k_failed), 0, 0, 1, 1},
  {&__pyx_n_s_failureTests, __pyx_k_failureTests, sizeof(__pyx_k_failureTests), 0, 0, 1, 1},
  {&__pyx_n_s_fatal, __pyx_k_fatal, sizeof(__pyx_k_fatal), 0, 0, 1, 1},
  {&__pyx_n_s_ff, __pyx_k_ff, sizeof(__pyx_k_ff), 0, 0, 1, 1},
  {&__pyx_n_s_file, __pyx_k_file, sizeof(__pyx_k_file), 0, 0, 1, 1},
  {&__pyx_n_s_file_contents, __pyx_k_file_contents, sizeof(__pyx_k_file_contents), 0, 0, 1, 1},
  {&__pyx_n_s_file_or_filename, __pyx_k_file_or_filename, sizeof(__pyx_k_file_or_filename), 0, 0, 1, 1},
  {&__pyx_n_s_finalResults, __pyx_k_finalResults, sizeof(__pyx_k_finalResults), 0, 0, 1, 1},
  {&__pyx_n_s_find, __pyx_k_find, sizeof(__pyx_k_find), 0, 0, 1, 1},
  {&__pyx_n_s_firstMatchChar, __pyx_k_firstMatchChar, sizeof(__pyx_k_firstMatchChar), 0, 0, 1, 1},
  {&__pyx_n_s_firstQuoteChar, __pyx_k_firstQuoteChar, sizeof(__pyx_k_firstQuoteChar), 0, 0, 1, 1},
  {&__pyx_n_s_flag, __pyx_k_flag, sizeof(__pyx_k_flag), 0, 0, 1, 1},
  {&__pyx_n_s_flags, __pyx_k_flags, sizeof(__pyx_k_flags), 0, 0, 1, 1},
  {&__pyx_n_s_flatten, __pyx_k_flatten, sizeof(__pyx_k_flatten), 0, 0, 1, 1},
  {&__pyx_n_s_fn, __pyx_k_fn, sizeof(__pyx_k_fn), 0, 0, 1, 1},
  {&__pyx_n_s_fns, __pyx_k_fns, sizeof(__pyx_k_fns), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_formatted, __pyx_k_formatted, sizeof(__pyx_k_formatted), 0, 0, 1, 1},
  {&__pyx_kp_s_found_end_of_text, __pyx_k_found_end_of_text, sizeof(__pyx_k_found_end_of_text), 0, 0, 1, 0},
  {&__pyx_kp_s_found_r, __pyx_k_found_r, sizeof(__pyx_k_found_r), 0, 0, 1, 0},
  {&__pyx_n_s_foundstr, __pyx_k_foundstr, sizeof(__pyx_k_foundstr), 0, 0, 1, 1},
  {&__pyx_n_s_frm, __pyx_k_frm, sizeof(__pyx_k_frm), 0, 0, 1, 1},
  {&__pyx_n_s_from, __pyx_k_from, sizeof(__pyx_k_from), 0, 0, 1, 1},
  {&__pyx_n_s_from_dict, __pyx_k_from_dict, sizeof(__pyx_k_from_dict), 0, 0, 1, 1},
  {&__pyx_n_s_from_exception, __pyx_k_from_exception, sizeof(__pyx_k_from_exception), 0, 0, 1, 1},
  {&__pyx_n_s_from_symbols, __pyx_k_from_symbols, sizeof(__pyx_k_from_symbols), 0, 0, 1, 1},
  {&__pyx_n_s_full, __pyx_k_full, sizeof(__pyx_k_full), 0, 0, 1, 1},
  {&__pyx_n_s_fullDump, __pyx_k_fullDump, sizeof(__pyx_k_fullDump), 0, 0, 1, 1},
  {&__pyx_kp_s_g_1, __pyx_k_g_1, sizeof(__pyx_k_g_1), 0, 0, 1, 0},
  {&__pyx_n_s_generatorType, __pyx_k_generatorType, sizeof(__pyx_k_generatorType), 0, 0, 1, 1},
  {&__pyx_n_s_genexpr, __pyx_k_genexpr, sizeof(__pyx_k_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_get, __pyx_k_get, sizeof(__pyx_k_get), 0, 0, 1, 1},
  {&__pyx_n_s_getName, __pyx_k_getName, sizeof(__pyx_k_getName), 0, 0, 1, 1},
  {&__pyx_n_s_getattr, __pyx_k_getattr, sizeof(__pyx_k_getattr), 0, 0, 1, 1},
  {&__pyx_n_s_getinnerframes, __pyx_k_getinnerframes, sizeof(__pyx_k_getinnerframes), 0, 0, 1, 1},
  {&__pyx_n_s_getitem, __pyx_k_getitem, sizeof(__pyx_k_getitem), 0, 0, 1, 1},
  {&__pyx_n_s_getnewargs, __pyx_k_getnewargs, sizeof(__pyx_k_getnewargs), 0, 0, 1, 1},
  {&__pyx_n_s_getrecursionlimit, __pyx_k_getrecursionlimit, sizeof(__pyx_k_getrecursionlimit), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_group, __pyx_k_group, sizeof(__pyx_k_group), 0, 0, 1, 1},
  {&__pyx_n_s_groupdict, __pyx_k_groupdict, sizeof(__pyx_k_groupdict), 0, 0, 1, 1},
  {&__pyx_n_s_groups, __pyx_k_groups, sizeof(__pyx_k_groups), 0, 0, 1, 1},
  {&__pyx_n_s_hasIgnoreExprs, __pyx_k_hasIgnoreExprs, sizeof(__pyx_k_hasIgnoreExprs), 0, 0, 1, 1},
  {&__pyx_n_s_hash, __pyx_k_hash, sizeof(__pyx_k_hash), 0, 0, 1, 1},
  {&__pyx_n_s_haskeys, __pyx_k_haskeys, sizeof(__pyx_k_haskeys), 0, 0, 1, 1},
  {&__pyx_n_s_hexnums, __pyx_k_hexnums, sizeof(__pyx_k_hexnums), 0, 0, 1, 1},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_n_s_iadd, __pyx_k_iadd, sizeof(__pyx_k_iadd), 0, 0, 1, 1},
  {&__pyx_n_s_id, __pyx_k_id, sizeof(__pyx_k_id), 0, 0, 1, 1},
  {&__pyx_n_s_identChars, __pyx_k_identChars, sizeof(__pyx_k_identChars), 0, 0, 1, 1},
  {&__pyx_n_s_ignore, __pyx_k_ignore, sizeof(__pyx_k_ignore), 0, 0, 1, 1},
  {&__pyx_n_s_ignoreExpr, __pyx_k_ignoreExpr, sizeof(__pyx_k_ignoreExpr), 0, 0, 1, 1},
  {&__pyx_n_s_ignoreExprs, __pyx_k_ignoreExprs, sizeof(__pyx_k_ignoreExprs), 0, 0, 1, 1},
  {&__pyx_n_s_ikey, __pyx_k_ikey, sizeof(__pyx_k_ikey), 0, 0, 1, 1},
  {&__pyx_n_s_ilshift, __pyx_k_ilshift, sizeof(__pyx_k_ilshift), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_inAccumNames, __pyx_k_inAccumNames, sizeof(__pyx_k_inAccumNames), 0, 0, 1, 1},
  {&__pyx_n_s_include, __pyx_k_include, sizeof(__pyx_k_include), 0, 0, 1, 1},
  {&__pyx_n_s_includeMatch, __pyx_k_includeMatch, sizeof(__pyx_k_includeMatch), 0, 0, 1, 1},
  {&__pyx_n_s_includeSeparators, __pyx_k_includeSeparators, sizeof(__pyx_k_includeSeparators), 0, 0, 1, 1},
  {&__pyx_n_s_include_list, __pyx_k_include_list, sizeof(__pyx_k_include_list), 0, 0, 1, 1},
  {&__pyx_n_s_indent, __pyx_k_indent, sizeof(__pyx_k_indent), 0, 0, 1, 1},
  {&__pyx_n_s_index, __pyx_k_index, sizeof(__pyx_k_index), 0, 0, 1, 1},
  {&__pyx_n_s_indices, __pyx_k_indices, sizeof(__pyx_k_indices), 0, 0, 1, 1},
  {&__pyx_n_s_init, __pyx_k_init, sizeof(__pyx_k_init), 0, 0, 1, 1},
  {&__pyx_n_s_initChars, __pyx_k_initChars, sizeof(__pyx_k_initChars), 0, 0, 1, 1},
  {&__pyx_n_s_initCharsOrig, __pyx_k_initCharsOrig, sizeof(__pyx_k_initCharsOrig), 0, 0, 1, 1},
  {&__pyx_n_s_initExprGroups, __pyx_k_initExprGroups, sizeof(__pyx_k_initExprGroups), 0, 0, 1, 1},
  {&__pyx_n_s_inlineLiteralsUsing, __pyx_k_inlineLiteralsUsing, sizeof(__pyx_k_inlineLiteralsUsing), 0, 0, 1, 1},
  {&__pyx_n_s_insStr, __pyx_k_insStr, sizeof(__pyx_k_insStr), 0, 0, 1, 1},
  {&__pyx_n_s_insert, __pyx_k_insert, sizeof(__pyx_k_insert), 0, 0, 1, 1},
  {&__pyx_n_s_inspect, __pyx_k_inspect, sizeof(__pyx_k_inspect), 0, 0, 1, 1},
  {&__pyx_n_s_instring, __pyx_k_instring, sizeof(__pyx_k_instring), 0, 0, 1, 1},
  {&__pyx_n_s_instring_slice, __pyx_k_instring_slice, sizeof(__pyx_k_instring_slice), 0, 0, 1, 1},
  {&__pyx_n_s_instrlen, __pyx_k_instrlen, sizeof(__pyx_k_instrlen), 0, 0, 1, 1},
  {&__pyx_kp_s_invalid_pattern_s_passed_to_Rege, __pyx_k_invalid_pattern_s_passed_to_Rege, sizeof(__pyx_k_invalid_pattern_s_passed_to_Rege), 0, 0, 1, 0},
  {&__pyx_n_s_invert, __pyx_k_invert, sizeof(__pyx_k_invert), 0, 0, 1, 1},
  {&__pyx_n_s_ior, __pyx_k_ior, sizeof(__pyx_k_ior), 0, 0, 1, 1},
  {&__pyx_n_s_is_iterable, __pyx_k_is_iterable, sizeof(__pyx_k_is_iterable), 0, 0, 1, 1},
  {&__pyx_n_s_isinstance, __pyx_k_isinstance, sizeof(__pyx_k_isinstance), 0, 0, 1, 1},
  {&__pyx_n_s_isspace, __pyx_k_isspace, sizeof(__pyx_k_isspace), 0, 0, 1, 1},
  {&__pyx_n_s_item, __pyx_k_item, sizeof(__pyx_k_item), 0, 0, 1, 1},
  {&__pyx_n_s_item_fn, __pyx_k_item_fn, sizeof(__pyx_k_item_fn), 0, 0, 1, 1},
  {&__pyx_n_s_itemgetter, __pyx_k_itemgetter, sizeof(__pyx_k_itemgetter), 0, 0, 1, 1},
  {&__pyx_n_s_items, __pyx_k_items, sizeof(__pyx_k_items), 0, 0, 1, 1},
  {&__pyx_n_s_itemseq, __pyx_k_itemseq, sizeof(__pyx_k_itemseq), 0, 0, 1, 1},
  {&__pyx_n_s_iter, __pyx_k_iter, sizeof(__pyx_k_iter), 0, 0, 1, 1},
  {&__pyx_n_s_iteritems, __pyx_k_iteritems, sizeof(__pyx_k_iteritems), 0, 0, 1, 1},
  {&__pyx_n_s_iteritems_2, __pyx_k_iteritems_2, sizeof(__pyx_k_iteritems_2), 0, 0, 1, 1},
  {&__pyx_n_s_iterkeys, __pyx_k_iterkeys, sizeof(__pyx_k_iterkeys), 0, 0, 1, 1},
  {&__pyx_n_s_iterkeys_2, __pyx_k_iterkeys_2, sizeof(__pyx_k_iterkeys_2), 0, 0, 1, 1},
  {&__pyx_n_s_itervalues, __pyx_k_itervalues, sizeof(__pyx_k_itervalues), 0, 0, 1, 1},
  {&__pyx_n_s_itervalues_2, __pyx_k_itervalues_2, sizeof(__pyx_k_itervalues_2), 0, 0, 1, 1},
  {&__pyx_n_s_ixor, __pyx_k_ixor, sizeof(__pyx_k_ixor), 0, 0, 1, 1},
  {&__pyx_n_s_j, __pyx_k_j, sizeof(__pyx_k_j), 0, 0, 1, 1},
  {&__pyx_n_s_join, __pyx_k_join, sizeof(__pyx_k_join), 0, 0, 1, 1},
  {&__pyx_n_s_joinString, __pyx_k_joinString, sizeof(__pyx_k_joinString), 0, 0, 1, 1},
  {&__pyx_kp_s_just_like_class_ParseFatalExcept, __pyx_k_just_like_class_ParseFatalExcept, sizeof(__pyx_k_just_like_class_ParseFatalExcept), 0, 0, 1, 0},
  {&__pyx_n_s_k, __pyx_k_k, sizeof(__pyx_k_k), 0, 0, 1, 1},
  {&__pyx_n_s_keepMatching, __pyx_k_keepMatching, sizeof(__pyx_k_keepMatching), 0, 0, 1, 1},
  {&__pyx_n_s_keepTabs, __pyx_k_keepTabs, sizeof(__pyx_k_keepTabs), 0, 0, 1, 1},
  {&__pyx_n_s_key, __pyx_k_key, sizeof(__pyx_k_key), 0, 0, 1, 1},
  {&__pyx_n_s_key_fifo, __pyx_k_key_fifo, sizeof(__pyx_k_key_fifo), 0, 0, 1, 1},
  {&__pyx_n_s_keys, __pyx_k_keys, sizeof(__pyx_k_keys), 0, 0, 1, 1},
  {&__pyx_n_s_kwargs, __pyx_k_kwargs, sizeof(__pyx_k_kwargs), 0, 0, 1, 1},
  {&__pyx_n_s_l, __pyx_k_l, sizeof(__pyx_k_l), 0, 0, 1, 1},
  {&__pyx_n_s_last, __pyx_k_last, sizeof(__pyx_k_last), 0, 0, 1, 1},
  {&__pyx_n_s_lastCR, __pyx_k_lastCR, sizeof(__pyx_k_lastCR), 0, 0, 1, 1},
  {&__pyx_n_s_lastE, __pyx_k_lastE, sizeof(__pyx_k_lastE), 0, 0, 1, 1},
  {&__pyx_n_s_last_expr, __pyx_k_last_expr, sizeof(__pyx_k_last_expr), 0, 0, 1, 1},
  {&__pyx_n_s_leaveWhitespace, __pyx_k_leaveWhitespace, sizeof(__pyx_k_leaveWhitespace), 0, 0, 1, 1},
  {&__pyx_n_s_len, __pyx_k_len, sizeof(__pyx_k_len), 0, 0, 1, 1},
  {&__pyx_n_s_line, __pyx_k_line, sizeof(__pyx_k_line), 0, 0, 1, 1},
  {&__pyx_n_s_line_column, __pyx_k_line_column, sizeof(__pyx_k_line_column), 0, 0, 1, 1},
  {&__pyx_n_s_line_str, __pyx_k_line_str, sizeof(__pyx_k_line_str), 0, 0, 1, 1},
  {&__pyx_n_s_lineno, __pyx_k_lineno, sizeof(__pyx_k_lineno), 0, 0, 1, 1},
  {&__pyx_kp_s_lineno_col_line, __pyx_k_lineno_col_line, sizeof(__pyx_k_lineno_col_line), 0, 0, 1, 0},
  {&__pyx_n_s_listAllMatches, __pyx_k_listAllMatches, sizeof(__pyx_k_listAllMatches), 0, 0, 1, 1},
  {&__pyx_n_s_literalStringClass, __pyx_k_literalStringClass, sizeof(__pyx_k_literalStringClass), 0, 0, 1, 1},
  {&__pyx_n_s_loc, __pyx_k_loc, sizeof(__pyx_k_loc), 0, 0, 1, 1},
  {&__pyx_n_s_loc1, __pyx_k_loc1, sizeof(__pyx_k_loc1), 0, 0, 1, 1},
  {&__pyx_n_s_loc2, __pyx_k_loc2, sizeof(__pyx_k_loc2), 0, 0, 1, 1},
  {&__pyx_n_s_longest, __pyx_k_longest, sizeof(__pyx_k_longest), 0, 0, 1, 1},
  {&__pyx_n_s_lookup, __pyx_k_lookup, sizeof(__pyx_k_lookup), 0, 0, 1, 1},
  {&__pyx_n_s_lookup_2, __pyx_k_lookup_2, sizeof(__pyx_k_lookup_2), 0, 0, 1, 1},
  {&__pyx_n_s_lshift, __pyx_k_lshift, sizeof(__pyx_k_lshift), 0, 0, 1, 1},
  {&__pyx_n_s_lstrip, __pyx_k_lstrip, sizeof(__pyx_k_lstrip), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_makeOptionalList, __pyx_k_makeOptionalList, sizeof(__pyx_k_makeOptionalList), 0, 0, 1, 1},
  {&__pyx_n_s_map, __pyx_k_map, sizeof(__pyx_k_map), 0, 0, 1, 1},
  {&__pyx_n_s_markInputline, __pyx_k_markInputline, sizeof(__pyx_k_markInputline), 0, 0, 1, 1},
  {&__pyx_n_s_markerString, __pyx_k_markerString, sizeof(__pyx_k_markerString), 0, 0, 1, 1},
  {&__pyx_n_s_mat, __pyx_k_mat, sizeof(__pyx_k_mat), 0, 0, 1, 1},
  {&__pyx_n_s_match, __pyx_k_match, sizeof(__pyx_k_match), 0, 0, 1, 1},
  {&__pyx_n_s_matchLen, __pyx_k_matchLen, sizeof(__pyx_k_matchLen), 0, 0, 1, 1},
  {&__pyx_n_s_matchOrder, __pyx_k_matchOrder, sizeof(__pyx_k_matchOrder), 0, 0, 1, 1},
  {&__pyx_n_s_matchString, __pyx_k_matchString, sizeof(__pyx_k_matchString), 0, 0, 1, 1},
  {&__pyx_n_s_matchWhite, __pyx_k_matchWhite, sizeof(__pyx_k_matchWhite), 0, 0, 1, 1},
  {&__pyx_n_s_match_string, __pyx_k_match_string, sizeof(__pyx_k_match_string), 0, 0, 1, 1},
  {&__pyx_n_s_match_stringloc, __pyx_k_match_stringloc, sizeof(__pyx_k_match_stringloc), 0, 0, 1, 1},
  {&__pyx_n_s_matches, __pyx_k_matches, sizeof(__pyx_k_matches), 0, 0, 1, 1},
  {&__pyx_n_s_max, __pyx_k_max, sizeof(__pyx_k_max), 0, 0, 1, 1},
  {&__pyx_n_s_maxExcLoc, __pyx_k_maxExcLoc, sizeof(__pyx_k_maxExcLoc), 0, 0, 1, 1},
  {&__pyx_n_s_maxException, __pyx_k_maxException, sizeof(__pyx_k_maxException), 0, 0, 1, 1},
  {&__pyx_n_s_maxLen, __pyx_k_maxLen, sizeof(__pyx_k_maxLen), 0, 0, 1, 1},
  {&__pyx_n_s_maxMatches, __pyx_k_maxMatches, sizeof(__pyx_k_maxMatches), 0, 0, 1, 1},
  {&__pyx_n_s_maxMismatches, __pyx_k_maxMismatches, sizeof(__pyx_k_maxMismatches), 0, 0, 1, 1},
  {&__pyx_n_s_maxSpecified, __pyx_k_maxSpecified, sizeof(__pyx_k_maxSpecified), 0, 0, 1, 1},
  {&__pyx_n_s_maxlen, __pyx_k_maxlen, sizeof(__pyx_k_maxlen), 0, 0, 1, 1},
  {&__pyx_n_s_maxloc, __pyx_k_maxloc, sizeof(__pyx_k_maxloc), 0, 0, 1, 1},
  {&__pyx_n_s_maxsize, __pyx_k_maxsize, sizeof(__pyx_k_maxsize), 0, 0, 1, 1},
  {&__pyx_n_s_maxsplit, __pyx_k_maxsplit, sizeof(__pyx_k_maxsplit), 0, 0, 1, 1},
  {&__pyx_n_s_mayIndexError, __pyx_k_mayIndexError, sizeof(__pyx_k_mayIndexError), 0, 0, 1, 1},
  {&__pyx_n_s_mayReturnEmpty, __pyx_k_mayReturnEmpty, sizeof(__pyx_k_mayReturnEmpty), 0, 0, 1, 1},
  {&__pyx_n_s_message, __pyx_k_message, sizeof(__pyx_k_message), 0, 0, 1, 1},
  {&__pyx_n_s_metaclass, __pyx_k_metaclass, sizeof(__pyx_k_metaclass), 0, 0, 1, 1},
  {&__pyx_n_s_methodCall, __pyx_k_methodCall, sizeof(__pyx_k_methodCall), 0, 0, 1, 1},
  {&__pyx_n_s_min, __pyx_k_min, sizeof(__pyx_k_min), 0, 0, 1, 1},
  {&__pyx_n_s_minElements, __pyx_k_minElements, sizeof(__pyx_k_minElements), 0, 0, 1, 1},
  {&__pyx_n_s_minLen, __pyx_k_minLen, sizeof(__pyx_k_minLen), 0, 0, 1, 1},
  {&__pyx_n_s_mismatches, __pyx_k_mismatches, sizeof(__pyx_k_mismatches), 0, 0, 1, 1},
  {&__pyx_n_s_missing, __pyx_k_missing, sizeof(__pyx_k_missing), 0, 0, 1, 1},
  {&__pyx_n_s_modal, __pyx_k_modal, sizeof(__pyx_k_modal), 0, 0, 1, 1},
  {&__pyx_n_s_modalResults, __pyx_k_modalResults, sizeof(__pyx_k_modalResults), 0, 0, 1, 1},
  {&__pyx_n_s_module, __pyx_k_module, sizeof(__pyx_k_module), 0, 0, 1, 1},
  {&__pyx_kp_s_module_2, __pyx_k_module_2, sizeof(__pyx_k_module_2), 0, 0, 1, 0},
  {&__pyx_n_s_msg, __pyx_k_msg, sizeof(__pyx_k_msg), 0, 0, 1, 1},
  {&__pyx_n_s_mul, __pyx_k_mul, sizeof(__pyx_k_mul), 0, 0, 1, 1},
  {&__pyx_n_s_multiline, __pyx_k_multiline, sizeof(__pyx_k_multiline), 0, 0, 1, 1},
  {&__pyx_n_s_multioptionals, __pyx_k_multioptionals, sizeof(__pyx_k_multioptionals), 0, 0, 1, 1},
  {&__pyx_n_s_multirequired, __pyx_k_multirequired, sizeof(__pyx_k_multirequired), 0, 0, 1, 1},
  {&__pyx_n_s_must_skip, __pyx_k_must_skip, sizeof(__pyx_k_must_skip), 0, 0, 1, 1},
  {&__pyx_n_s_mylen, __pyx_k_mylen, sizeof(__pyx_k_mylen), 0, 0, 1, 1},
  {&__pyx_kp_s_n, __pyx_k_n, sizeof(__pyx_k_n), 0, 0, 1, 0},
  {&__pyx_n_s_n_2, __pyx_k_n_2, sizeof(__pyx_k_n_2), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {&__pyx_n_s_namedItems, __pyx_k_namedItems, sizeof(__pyx_k_namedItems), 0, 0, 1, 1},
  {&__pyx_n_s_namedItemsOnly, __pyx_k_namedItemsOnly, sizeof(__pyx_k_namedItemsOnly), 0, 0, 1, 1},
  {&__pyx_n_s_ne, __pyx_k_ne, sizeof(__pyx_k_ne), 0, 0, 1, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_n_s_newloc, __pyx_k_newloc, sizeof(__pyx_k_newloc), 0, 0, 1, 1},
  {&__pyx_n_s_newself, __pyx_k_newself, sizeof(__pyx_k_newself), 0, 0, 1, 1},
  {&__pyx_n_s_nextCR, __pyx_k_nextCR, sizeof(__pyx_k_nextCR), 0, 0, 1, 1},
  {&__pyx_n_s_nextLevelIndent, __pyx_k_nextLevelIndent, sizeof(__pyx_k_nextLevelIndent), 0, 0, 1, 1},
  {&__pyx_n_s_nextLoc, __pyx_k_nextLoc, sizeof(__pyx_k_nextLoc), 0, 0, 1, 1},
  {&__pyx_n_s_nextloc, __pyx_k_nextloc, sizeof(__pyx_k_nextloc), 0, 0, 1, 1},
  {&__pyx_n_s_nl, __pyx_k_nl, sizeof(__pyx_k_nl), 0, 0, 1, 1},
  {&__pyx_n_s_nm, __pyx_k_nm, sizeof(__pyx_k_nm), 0, 0, 1, 1},
  {&__pyx_kp_s_no_defined_alternatives_to_match, __pyx_k_no_defined_alternatives_to_match, sizeof(__pyx_k_no_defined_alternatives_to_match), 0, 0, 1, 0},
  {&__pyx_n_s_nonzero, __pyx_k_nonzero, sizeof(__pyx_k_nonzero), 0, 0, 1, 1},
  {&__pyx_n_s_notChars, __pyx_k_notChars, sizeof(__pyx_k_notChars), 0, 0, 1, 1},
  {&__pyx_n_s_not_ender, __pyx_k_not_ender, sizeof(__pyx_k_not_ender), 0, 0, 1, 1},
  {&__pyx_n_s_not_in_cache, __pyx_k_not_in_cache, sizeof(__pyx_k_not_in_cache), 0, 0, 1, 1},
  {&__pyx_kp_s_not_preceded_by, __pyx_k_not_preceded_by, sizeof(__pyx_k_not_preceded_by), 0, 0, 1, 0},
  {&__pyx_n_s_notchars, __pyx_k_notchars, sizeof(__pyx_k_notchars), 0, 0, 1, 1},
  {&__pyx_n_s_nullDebugAction, __pyx_k_nullDebugAction, sizeof(__pyx_k_nullDebugAction), 0, 0, 1, 1},
  {&__pyx_kp_s_null_string_passed_to_Keyword_us, __pyx_k_null_string_passed_to_Keyword_us, sizeof(__pyx_k_null_string_passed_to_Keyword_us), 0, 0, 1, 0},
  {&__pyx_kp_s_null_string_passed_to_Literal_us, __pyx_k_null_string_passed_to_Literal_us, sizeof(__pyx_k_null_string_passed_to_Literal_us), 0, 0, 1, 0},
  {&__pyx_kp_s_null_string_passed_to_Regex_use, __pyx_k_null_string_passed_to_Regex_use, sizeof(__pyx_k_null_string_passed_to_Regex_use), 0, 0, 1, 0},
  {&__pyx_n_s_nums, __pyx_k_nums, sizeof(__pyx_k_nums), 0, 0, 1, 1},
  {&__pyx_n_s_o, __pyx_k_o, sizeof(__pyx_k_o), 0, 0, 1, 1},
  {&__pyx_n_s_obj, __pyx_k_obj, sizeof(__pyx_k_obj), 0, 0, 1, 1},
  {&__pyx_n_s_object, __pyx_k_object, sizeof(__pyx_k_object), 0, 0, 1, 1},
  {&__pyx_n_s_occurrences, __pyx_k_occurrences, sizeof(__pyx_k_occurrences), 0, 0, 1, 1},
  {&__pyx_n_s_offset, __pyx_k_offset, sizeof(__pyx_k_offset), 0, 0, 1, 1},
  {&__pyx_kp_s_only_1_or_2_index_arguments_supp, __pyx_k_only_1_or_2_index_arguments_supp, sizeof(__pyx_k_only_1_or_2_index_arguments_supp), 0, 0, 1, 0},
  {&__pyx_n_s_open, __pyx_k_open, sizeof(__pyx_k_open), 0, 0, 1, 1},
  {&__pyx_n_s_operator, __pyx_k_operator, sizeof(__pyx_k_operator), 0, 0, 1, 1},
  {&__pyx_n_s_opt1, __pyx_k_opt1, sizeof(__pyx_k_opt1), 0, 0, 1, 1},
  {&__pyx_n_s_opt1map, __pyx_k_opt1map, sizeof(__pyx_k_opt1map), 0, 0, 1, 1},
  {&__pyx_n_s_opt2, __pyx_k_opt2, sizeof(__pyx_k_opt2), 0, 0, 1, 1},
  {&__pyx_n_s_optElements, __pyx_k_optElements, sizeof(__pyx_k_optElements), 0, 0, 1, 1},
  {&__pyx_n_s_optionals, __pyx_k_optionals, sizeof(__pyx_k_optionals), 0, 0, 1, 1},
  {&__pyx_n_s_or, __pyx_k_or, sizeof(__pyx_k_or), 0, 0, 1, 1},
  {&__pyx_n_s_ordereddict, __pyx_k_ordereddict, sizeof(__pyx_k_ordereddict), 0, 0, 1, 1},
  {&__pyx_n_s_original, __pyx_k_original, sizeof(__pyx_k_original), 0, 0, 1, 1},
  {&__pyx_n_s_originalParseMethod, __pyx_k_originalParseMethod, sizeof(__pyx_k_originalParseMethod), 0, 0, 1, 1},
  {&__pyx_n_s_other, __pyx_k_other, sizeof(__pyx_k_other), 0, 0, 1, 1},
  {&__pyx_n_s_otherdictitems, __pyx_k_otherdictitems, sizeof(__pyx_k_otherdictitems), 0, 0, 1, 1},
  {&__pyx_n_s_otheritems, __pyx_k_otheritems, sizeof(__pyx_k_otheritems), 0, 0, 1, 1},
  {&__pyx_n_s_out, __pyx_k_out, sizeof(__pyx_k_out), 0, 0, 1, 1},
  {&__pyx_n_s_overlap, __pyx_k_overlap, sizeof(__pyx_k_overlap), 0, 0, 1, 1},
  {&__pyx_n_s_p1, __pyx_k_p1, sizeof(__pyx_k_p1), 0, 0, 1, 1},
  {&__pyx_n_s_p2, __pyx_k_p2, sizeof(__pyx_k_p2), 0, 0, 1, 1},
  {&__pyx_n_s_pa, __pyx_k_pa, sizeof(__pyx_k_pa), 0, 0, 1, 1},
  {&__pyx_n_s_packratEnabled, __pyx_k_packratEnabled, sizeof(__pyx_k_packratEnabled), 0, 0, 1, 1},
  {&__pyx_n_s_packrat_cache, __pyx_k_packrat_cache, sizeof(__pyx_k_packrat_cache), 0, 0, 1, 1},
  {&__pyx_n_s_packrat_cache_lock, __pyx_k_packrat_cache_lock, sizeof(__pyx_k_packrat_cache_lock), 0, 0, 1, 1},
  {&__pyx_n_s_packrat_cache_stats, __pyx_k_packrat_cache_stats, sizeof(__pyx_k_packrat_cache_stats), 0, 0, 1, 1},
  {&__pyx_n_s_par, __pyx_k_par, sizeof(__pyx_k_par), 0, 0, 1, 1},
  {&__pyx_n_s_parse, __pyx_k_parse, sizeof(__pyx_k_parse), 0, 0, 1, 1},
  {&__pyx_n_s_parseAction, __pyx_k_parseAction, sizeof(__pyx_k_parseAction), 0, 0, 1, 1},
  {&__pyx_n_s_parseAll, __pyx_k_parseAll, sizeof(__pyx_k_parseAll), 0, 0, 1, 1},
  {&__pyx_n_s_parseCache, __pyx_k_parseCache, sizeof(__pyx_k_parseCache), 0, 0, 1, 1},
  {&__pyx_n_s_parseElementList, __pyx_k_parseElementList, sizeof(__pyx_k_parseElementList), 0, 0, 1, 1},
  {&__pyx_n_s_parseElementTrace, __pyx_k_parseElementTrace, sizeof(__pyx_k_parseElementTrace), 0, 0, 1, 1},
  {&__pyx_n_s_parseFile, __pyx_k_parseFile, sizeof(__pyx_k_parseFile), 0, 0, 1, 1},
  {&__pyx_n_s_parseFn, __pyx_k_parseFn, sizeof(__pyx_k_parseFn), 0, 0, 1, 1},
  {&__pyx_n_s_parseImpl, __pyx_k_parseImpl, sizeof(__pyx_k_parseImpl), 0, 0, 1, 1},
  {&__pyx_n_s_parseImplAsGroupList, __pyx_k_parseImplAsGroupList, sizeof(__pyx_k_parseImplAsGroupList), 0, 0, 1, 1},
  {&__pyx_n_s_parseImplAsMatch, __pyx_k_parseImplAsMatch, sizeof(__pyx_k_parseImplAsMatch), 0, 0, 1, 1},
  {&__pyx_n_s_parseMethod, __pyx_k_parseMethod, sizeof(__pyx_k_parseMethod), 0, 0, 1, 1},
  {&__pyx_n_s_parseNoCache, __pyx_k_parseNoCache, sizeof(__pyx_k_parseNoCache), 0, 0, 1, 1},
  {&__pyx_n_s_parseString, __pyx_k_parseString, sizeof(__pyx_k_parseString), 0, 0, 1, 1},
  {&__pyx_n_s_parseWithTabs, __pyx_k_parseWithTabs, sizeof(__pyx_k_parseWithTabs), 0, 0, 1, 1},
  {&__pyx_n_s_parse_action_exc, __pyx_k_parse_action_exc, sizeof(__pyx_k_parse_action_exc), 0, 0, 1, 1},
  {&__pyx_kp_s_parse_actions_must_be_callable, __pyx_k_parse_actions_must_be_callable, sizeof(__pyx_k_parse_actions_must_be_callable), 0, 0, 1, 0},
  {&__pyx_n_s_parserElement, __pyx_k_parserElement, sizeof(__pyx_k_parserElement), 0, 0, 1, 1},
  {&__pyx_n_s_pattern, __pyx_k_pattern, sizeof(__pyx_k_pattern), 0, 0, 1, 1},
  {&__pyx_n_s_pbe, __pyx_k_pbe, sizeof(__pyx_k_pbe), 0, 0, 1, 1},
  {&__pyx_n_s_pdb, __pyx_k_pdb, sizeof(__pyx_k_pdb), 0, 0, 1, 1},
  {&__pyx_n_s_pe, __pyx_k_pe, sizeof(__pyx_k_pe), 0, 0, 1, 1},
  {&__pyx_n_s_pop, __pyx_k_pop, sizeof(__pyx_k_pop), 0, 0, 1, 1},
  {&__pyx_kp_s_pop_got_an_unexpected_keyword_ar, __pyx_k_pop_got_an_unexpected_keyword_ar, sizeof(__pyx_k_pop_got_an_unexpected_keyword_ar), 0, 0, 1, 0},
  {&__pyx_n_s_popitem, __pyx_k_popitem, sizeof(__pyx_k_popitem), 0, 0, 1, 1},
  {&__pyx_n_s_popleft, __pyx_k_popleft, sizeof(__pyx_k_popleft), 0, 0, 1, 1},
  {&__pyx_n_s_position, __pyx_k_position, sizeof(__pyx_k_position), 0, 0, 1, 1},
  {&__pyx_n_s_postParse, __pyx_k_postParse, sizeof(__pyx_k_postParse), 0, 0, 1, 1},
  {&__pyx_n_s_pp_value, __pyx_k_pp_value, sizeof(__pyx_k_pp_value), 0, 0, 1, 1},
  {&__pyx_n_s_pprint, __pyx_k_pprint, sizeof(__pyx_k_pprint), 0, 0, 1, 1},
  {&__pyx_n_s_preParse, __pyx_k_preParse, sizeof(__pyx_k_preParse), 0, 0, 1, 1},
  {&__pyx_n_s_preloc, __pyx_k_preloc, sizeof(__pyx_k_preloc), 0, 0, 1, 1},
  {&__pyx_n_s_prepare, __pyx_k_prepare, sizeof(__pyx_k_prepare), 0, 0, 1, 1},
  {&__pyx_n_s_preparseFn, __pyx_k_preparseFn, sizeof(__pyx_k_preparseFn), 0, 0, 1, 1},
  {&__pyx_n_s_print, __pyx_k_print, sizeof(__pyx_k_print), 0, 0, 1, 1},
  {&__pyx_n_s_printResults, __pyx_k_printResults, sizeof(__pyx_k_printResults), 0, 0, 1, 1},
  {&__pyx_n_s_print_2, __pyx_k_print_2, sizeof(__pyx_k_print_2), 0, 0, 1, 1},
  {&__pyx_n_s_printable, __pyx_k_printable, sizeof(__pyx_k_printable), 0, 0, 1, 1},
  {&__pyx_n_s_printables, __pyx_k_printables, sizeof(__pyx_k_printables), 0, 0, 1, 1},
  {&__pyx_n_s_pstr, __pyx_k_pstr, sizeof(__pyx_k_pstr), 0, 0, 1, 1},
  {&__pyx_n_s_pyparsing, __pyx_k_pyparsing, sizeof(__pyx_k_pyparsing), 0, 0, 1, 1},
  {&__pyx_n_s_qualname, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 0, 1, 1},
  {&__pyx_n_s_quoteChar, __pyx_k_quoteChar, sizeof(__pyx_k_quoteChar), 0, 0, 1, 1},
  {&__pyx_n_s_quoteCharLen, __pyx_k_quoteCharLen, sizeof(__pyx_k_quoteCharLen), 0, 0, 1, 1},
  {&__pyx_kp_s_quoteChar_cannot_be_the_empty_st, __pyx_k_quoteChar_cannot_be_the_empty_st, sizeof(__pyx_k_quoteChar_cannot_be_the_empty_st), 0, 0, 1, 0},
  {&__pyx_n_s_quotedString, __pyx_k_quotedString, sizeof(__pyx_k_quotedString), 0, 0, 1, 1},
  {&__pyx_kp_s_quoted_string_starting_with_s_en, __pyx_k_quoted_string_starting_with_s_en, sizeof(__pyx_k_quoted_string_starting_with_s_en), 0, 0, 1, 0},
  {&__pyx_n_s_r, __pyx_k_r, sizeof(__pyx_k_r), 0, 0, 1, 1},
  {&__pyx_kp_s_r_2, __pyx_k_r_2, sizeof(__pyx_k_r_2), 0, 0, 1, 0},
  {&__pyx_kp_s_r_object_is_not_iterable, __pyx_k_r_object_is_not_iterable, sizeof(__pyx_k_r_object_is_not_iterable), 0, 0, 1, 0},
  {&__pyx_n_s_radd, __pyx_k_radd, sizeof(__pyx_k_radd), 0, 0, 1, 1},
  {&__pyx_n_s_rand, __pyx_k_rand, sizeof(__pyx_k_rand), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_re, __pyx_k_re, sizeof(__pyx_k_re), 0, 0, 1, 1},
  {&__pyx_n_s_reString, __pyx_k_reString, sizeof(__pyx_k_reString), 0, 0, 1, 1},
  {&__pyx_n_s_re_match, __pyx_k_re_match, sizeof(__pyx_k_re_match), 0, 0, 1, 1},
  {&__pyx_n_s_read, __pyx_k_read, sizeof(__pyx_k_read), 0, 0, 1, 1},
  {&__pyx_n_s_ref, __pyx_k_ref, sizeof(__pyx_k_ref), 0, 0, 1, 1},
  {&__pyx_n_s_remove, __pyx_k_remove, sizeof(__pyx_k_remove), 0, 0, 1, 1},
  {&__pyx_n_s_removed, __pyx_k_removed, sizeof(__pyx_k_removed), 0, 0, 1, 1},
  {&__pyx_n_s_repl, __pyx_k_repl, sizeof(__pyx_k_repl), 0, 0, 1, 1},
  {&__pyx_n_s_replace, __pyx_k_replace, sizeof(__pyx_k_replace), 0, 0, 1, 1},
  {&__pyx_n_s_replaceWith, __pyx_k_replaceWith, sizeof(__pyx_k_replaceWith), 0, 0, 1, 1},
  {&__pyx_n_s_repr, __pyx_k_repr, sizeof(__pyx_k_repr), 0, 0, 1, 1},
  {&__pyx_n_s_req, __pyx_k_req, sizeof(__pyx_k_req), 0, 0, 1, 1},
  {&__pyx_n_s_required, __pyx_k_required, sizeof(__pyx_k_required), 0, 0, 1, 1},
  {&__pyx_n_s_res, __pyx_k_res, sizeof(__pyx_k_res), 0, 0, 1, 1},
  {&__pyx_n_s_resTag, __pyx_k_resTag, sizeof(__pyx_k_resTag), 0, 0, 1, 1},
  {&__pyx_n_s_reset, __pyx_k_reset, sizeof(__pyx_k_reset), 0, 0, 1, 1},
  {&__pyx_n_s_resetCache, __pyx_k_resetCache, sizeof(__pyx_k_resetCache), 0, 0, 1, 1},
  {&__pyx_n_s_result, __pyx_k_result, sizeof(__pyx_k_result), 0, 0, 1, 1},
  {&__pyx_n_s_resultlist, __pyx_k_resultlist, sizeof(__pyx_k_resultlist), 0, 0, 1, 1},
  {&__pyx_n_s_results, __pyx_k_results, sizeof(__pyx_k_results), 0, 0, 1, 1},
  {&__pyx_n_s_resultsName, __pyx_k_resultsName, sizeof(__pyx_k_resultsName), 0, 0, 1, 1},
  {&__pyx_n_s_ret, __pyx_k_ret, sizeof(__pyx_k_ret), 0, 0, 1, 1},
  {&__pyx_n_s_retString, __pyx_k_retString, sizeof(__pyx_k_retString), 0, 0, 1, 1},
  {&__pyx_n_s_retTokens, __pyx_k_retTokens, sizeof(__pyx_k_retTokens), 0, 0, 1, 1},
  {&__pyx_n_s_retToks, __pyx_k_retToks, sizeof(__pyx_k_retToks), 0, 0, 1, 1},
  {&__pyx_n_s_retobj, __pyx_k_retobj, sizeof(__pyx_k_retobj), 0, 0, 1, 1},
  {&__pyx_n_s_retreat, __pyx_k_retreat, sizeof(__pyx_k_retreat), 0, 0, 1, 1},
  {&__pyx_n_s_returnString, __pyx_k_returnString, sizeof(__pyx_k_returnString), 0, 0, 1, 1},
  {&__pyx_n_s_reverse, __pyx_k_reverse, sizeof(__pyx_k_reverse), 0, 0, 1, 1},
  {&__pyx_n_s_reversed, __pyx_k_reversed, sizeof(__pyx_k_reversed), 0, 0, 1, 1},
  {&__pyx_n_s_rfind, __pyx_k_rfind, sizeof(__pyx_k_rfind), 0, 0, 1, 1},
  {&__pyx_n_s_rmul, __pyx_k_rmul, sizeof(__pyx_k_rmul), 0, 0, 1, 1},
  {&__pyx_n_s_rne, __pyx_k_rne, sizeof(__pyx_k_rne), 0, 0, 1, 1},
  {&__pyx_n_s_ror, __pyx_k_ror, sizeof(__pyx_k_ror), 0, 0, 1, 1},
  {&__pyx_n_s_rstrip, __pyx_k_rstrip, sizeof(__pyx_k_rstrip), 0, 0, 1, 1},
  {&__pyx_n_s_rsub, __pyx_k_rsub, sizeof(__pyx_k_rsub), 0, 0, 1, 1},
  {&__pyx_n_s_runTests, __pyx_k_runTests, sizeof(__pyx_k_runTests), 0, 0, 1, 1},
  {&__pyx_n_s_rxor, __pyx_k_rxor, sizeof(__pyx_k_rxor), 0, 0, 1, 1},
  {&__pyx_kp_s_s, __pyx_k_s, sizeof(__pyx_k_s), 0, 0, 1, 0},
  {&__pyx_kp_s_s_2, __pyx_k_s_2, sizeof(__pyx_k_s_2), 0, 0, 1, 0},
  {&__pyx_kp_s_s_3, __pyx_k_s_3, sizeof(__pyx_k_s_3), 0, 0, 1, 0},
  {&__pyx_n_s_s_4, __pyx_k_s_4, sizeof(__pyx_k_s_4), 0, 0, 1, 1},
  {&__pyx_kp_s_s_5, __pyx_k_s_5, sizeof(__pyx_k_s_5), 0, 0, 1, 0},
  {&__pyx_kp_s_s_6, __pyx_k_s_6, sizeof(__pyx_k_s_6), 0, 0, 1, 0},
  {&__pyx_kp_s_s_7, __pyx_k_s_7, sizeof(__pyx_k_s_7), 0, 0, 1, 0},
  {&__pyx_kp_s_s_8, __pyx_k_s_8, sizeof(__pyx_k_s_8), 0, 0, 1, 0},
  {&__pyx_n_s_s_m, __pyx_k_s_m, sizeof(__pyx_k_s_m), 0, 0, 1, 1},
  {&__pyx_kp_s_s_s, __pyx_k_s_s, sizeof(__pyx_k_s_s), 0, 0, 1, 0},
  {&__pyx_kp_s_s_s_2, __pyx_k_s_s_2, sizeof(__pyx_k_s_s_2), 0, 0, 1, 0},
  {&__pyx_kp_s_s_s_3, __pyx_k_s_s_3, sizeof(__pyx_k_s_s_3), 0, 0, 1, 0},
  {&__pyx_kp_s_s_s_4, __pyx_k_s_s_4, sizeof(__pyx_k_s_s_4), 0, 0, 1, 0},
  {&__pyx_kp_s_s_s_5, __pyx_k_s_s_5, sizeof(__pyx_k_s_s_5), 0, 0, 1, 0},
  {&__pyx_kp_s_s_s_at_char_d_line_d_col_d, __pyx_k_s_s_at_char_d_line_d_col_d, sizeof(__pyx_k_s_s_at_char_d_line_d_col_d), 0, 0, 1, 0},
  {&__pyx_kp_s_s_s_d_s_s_s, __pyx_k_s_s_d_s_s_s, sizeof(__pyx_k_s_s_d_s_s_s), 0, 0, 1, 0},
  {&__pyx_kp_s_s_s_n_r_s, __pyx_k_s_s_n_r_s, sizeof(__pyx_k_s_s_n_r_s), 0, 0, 1, 0},
  {&__pyx_kp_s_s_s_s, __pyx_k_s_s_s, sizeof(__pyx_k_s_s_s), 0, 0, 1, 0},
  {&__pyx_kp_s_s_s_s_2, __pyx_k_s_s_s_2, sizeof(__pyx_k_s_s_s_2), 0, 0, 1, 0},
  {&__pyx_n_s_saveAsList, __pyx_k_saveAsList, sizeof(__pyx_k_saveAsList), 0, 0, 1, 1},
  {&__pyx_n_s_savelist, __pyx_k_savelist, sizeof(__pyx_k_savelist), 0, 0, 1, 1},
  {&__pyx_n_s_scanString, __pyx_k_scanString, sizeof(__pyx_k_scanString), 0, 0, 1, 1},
  {&__pyx_n_s_se, __pyx_k_se, sizeof(__pyx_k_se), 0, 0, 1, 1},
  {&__pyx_n_s_searchString, __pyx_k_searchString, sizeof(__pyx_k_searchString), 0, 0, 1, 1},
  {&__pyx_kp_s_second_tuple_value_must_be_great, __pyx_k_second_tuple_value_must_be_great, sizeof(__pyx_k_second_tuple_value_must_be_great), 0, 0, 1, 0},
  {&__pyx_n_s_seen, __pyx_k_seen, sizeof(__pyx_k_seen), 0, 0, 1, 1},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_n_s_selfTag, __pyx_k_selfTag, sizeof(__pyx_k_selfTag), 0, 0, 1, 1},
  {&__pyx_n_s_self_expr_parse, __pyx_k_self_expr_parse, sizeof(__pyx_k_self_expr_parse), 0, 0, 1, 1},
  {&__pyx_n_s_self_failOn_canParseNext, __pyx_k_self_failOn_canParseNext, sizeof(__pyx_k_self_failOn_canParseNext), 0, 0, 1, 1},
  {&__pyx_n_s_self_ignoreExpr_tryParse, __pyx_k_self_ignoreExpr_tryParse, sizeof(__pyx_k_self_ignoreExpr_tryParse), 0, 0, 1, 1},
  {&__pyx_n_s_self_skip_ignorables, __pyx_k_self_skip_ignorables, sizeof(__pyx_k_self_skip_ignorables), 0, 0, 1, 1},
  {&__pyx_n_s_self_type, __pyx_k_self_type, sizeof(__pyx_k_self_type), 0, 0, 1, 1},
  {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
  {&__pyx_n_s_sep, __pyx_k_sep, sizeof(__pyx_k_sep), 0, 0, 1, 1},
  {&__pyx_n_s_set, __pyx_k_set, sizeof(__pyx_k_set), 0, 0, 1, 1},
  {&__pyx_n_s_setBreak, __pyx_k_setBreak, sizeof(__pyx_k_setBreak), 0, 0, 1, 1},
  {&__pyx_n_s_setDebug, __pyx_k_setDebug, sizeof(__pyx_k_setDebug), 0, 0, 1, 1},
  {&__pyx_n_s_setDebugActions, __pyx_k_setDebugActions, sizeof(__pyx_k_setDebugActions), 0, 0, 1, 1},
  {&__pyx_n_s_setDefaultKeywordChars, __pyx_k_setDefaultKeywordChars, sizeof(__pyx_k_setDefaultKeywordChars), 0, 0, 1, 1},
  {&__pyx_n_s_setDefaultWhitespaceChars, __pyx_k_setDefaultWhitespaceChars, sizeof(__pyx_k_setDefaultWhitespaceChars), 0, 0, 1, 1},
  {&__pyx_n_s_setFailAction, __pyx_k_setFailAction, sizeof(__pyx_k_setFailAction), 0, 0, 1, 1},
  {&__pyx_n_s_setName, __pyx_k_setName, sizeof(__pyx_k_setName), 0, 0, 1, 1},
  {&__pyx_n_s_setOffset, __pyx_k_setOffset, sizeof(__pyx_k_setOffset), 0, 0, 1, 1},
  {&__pyx_n_s_setParseAction, __pyx_k_setParseAction, sizeof(__pyx_k_setParseAction), 0, 0, 1, 1},
  {&__pyx_n_s_setResultsName, __pyx_k_setResultsName, sizeof(__pyx_k_setResultsName), 0, 0, 1, 1},
  {&__pyx_n_s_setResultsName_2, __pyx_k_setResultsName_2, sizeof(__pyx_k_setResultsName_2), 0, 0, 1, 1},
  {&__pyx_n_s_setWhitespaceChars, __pyx_k_setWhitespaceChars, sizeof(__pyx_k_setWhitespaceChars), 0, 0, 1, 1},
  {&__pyx_n_s_set_trace, __pyx_k_set_trace, sizeof(__pyx_k_set_trace), 0, 0, 1, 1},
  {&__pyx_n_s_setitem, __pyx_k_setitem, sizeof(__pyx_k_setitem), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_size, __pyx_k_size, sizeof(__pyx_k_size), 0, 0, 1, 1},
  {&__pyx_n_s_skipIgnorables, __pyx_k_skipIgnorables, sizeof(__pyx_k_skipIgnorables), 0, 0, 1, 1},
  {&__pyx_n_s_skipWhitespace, __pyx_k_skipWhitespace, sizeof(__pyx_k_skipWhitespace), 0, 0, 1, 1},
  {&__pyx_kp_s_skipped, __pyx_k_skipped, sizeof(__pyx_k_skipped), 0, 0, 1, 0},
  {&__pyx_n_s_skipresult, __pyx_k_skipresult, sizeof(__pyx_k_skipresult), 0, 0, 1, 1},
  {&__pyx_n_s_skiptext, __pyx_k_skiptext, sizeof(__pyx_k_skiptext), 0, 0, 1, 1},
  {&__pyx_n_s_skipto_arg, __pyx_k_skipto_arg, sizeof(__pyx_k_skipto_arg), 0, 0, 1, 1},
  {&__pyx_n_s_sort, __pyx_k_sort, sizeof(__pyx_k_sort), 0, 0, 1, 1},
  {&__pyx_n_s_split, __pyx_k_split, sizeof(__pyx_k_split), 0, 0, 1, 1},
  {&__pyx_n_s_splitlines, __pyx_k_splitlines, sizeof(__pyx_k_splitlines), 0, 0, 1, 1},
  {&__pyx_n_s_splits, __pyx_k_splits, sizeof(__pyx_k_splits), 0, 0, 1, 1},
  {&__pyx_n_s_src, __pyx_k_src, sizeof(__pyx_k_src), 0, 0, 1, 1},
  {&__pyx_n_s_sre_constants, __pyx_k_sre_constants, sizeof(__pyx_k_sre_constants), 0, 0, 1, 1},
  {&__pyx_n_s_stacklevel, __pyx_k_stacklevel, sizeof(__pyx_k_stacklevel), 0, 0, 1, 1},
  {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
  {&__pyx_n_s_startAction, __pyx_k_startAction, sizeof(__pyx_k_startAction), 0, 0, 1, 1},
  {&__pyx_n_s_startloc, __pyx_k_startloc, sizeof(__pyx_k_startloc), 0, 0, 1, 1},
  {&__pyx_n_s_startswith, __pyx_k_startswith, sizeof(__pyx_k_startswith), 0, 0, 1, 1},
  {&__pyx_n_s_state, __pyx_k_state, sizeof(__pyx_k_state), 0, 0, 1, 1},
  {&__pyx_n_s_staticmethod, __pyx_k_staticmethod, sizeof(__pyx_k_staticmethod), 0, 0, 1, 1},
  {&__pyx_n_s_stdout, __pyx_k_stdout, sizeof(__pyx_k_stdout), 0, 0, 1, 1},
  {&__pyx_n_s_stopOn, __pyx_k_stopOn, sizeof(__pyx_k_stopOn), 0, 0, 1, 1},
  {&__pyx_n_s_str, __pyx_k_str, sizeof(__pyx_k_str), 0, 0, 1, 1},
  {&__pyx_n_s_strRepr, __pyx_k_strRepr, sizeof(__pyx_k_strRepr), 0, 0, 1, 1},
  {&__pyx_n_s_streamline, __pyx_k_streamline, sizeof(__pyx_k_streamline), 0, 0, 1, 1},
  {&__pyx_n_s_streamlined, __pyx_k_streamlined, sizeof(__pyx_k_streamlined), 0, 0, 1, 1},
  {&__pyx_n_s_strg, __pyx_k_strg, sizeof(__pyx_k_strg), 0, 0, 1, 1},
  {&__pyx_n_s_string, __pyx_k_string, sizeof(__pyx_k_string), 0, 0, 1, 1},
  {&__pyx_n_s_strip, __pyx_k_strip, sizeof(__pyx_k_strip), 0, 0, 1, 1},
  {&__pyx_n_s_sub, __pyx_k_sub, sizeof(__pyx_k_sub), 0, 0, 1, 1},
  {&__pyx_n_s_subRecCheckList, __pyx_k_subRecCheckList, sizeof(__pyx_k_subRecCheckList), 0, 0, 1, 1},
  {&__pyx_n_s_sub_2, __pyx_k_sub_2, sizeof(__pyx_k_sub_2), 0, 0, 1, 1},
  {&__pyx_n_s_success, __pyx_k_success, sizeof(__pyx_k_success), 0, 0, 1, 1},
  {&__pyx_n_s_successAction, __pyx_k_successAction, sizeof(__pyx_k_successAction), 0, 0, 1, 1},
  {&__pyx_n_s_sum, __pyx_k_sum, sizeof(__pyx_k_sum), 0, 0, 1, 1},
  {&__pyx_n_s_super, __pyx_k_super, sizeof(__pyx_k_super), 0, 0, 1, 1},
  {&__pyx_n_s_suppress, __pyx_k_suppress, sizeof(__pyx_k_suppress), 0, 0, 1, 1},
  {&__pyx_n_s_sys, __pyx_k_sys, sizeof(__pyx_k_sys), 0, 0, 1, 1},
  {&__pyx_n_s_system_version, __pyx_k_system_version, sizeof(__pyx_k_system_version), 0, 0, 1, 1},
  {&__pyx_kp_s_t, __pyx_k_t, sizeof(__pyx_k_t), 0, 0, 1, 0},
  {&__pyx_n_s_t_2, __pyx_k_t_2, sizeof(__pyx_k_t_2), 0, 0, 1, 1},
  {&__pyx_n_s_tb, __pyx_k_tb, sizeof(__pyx_k_tb), 0, 0, 1, 1},
  {&__pyx_n_s_tb_next, __pyx_k_tb_next, sizeof(__pyx_k_tb_next), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_testString, __pyx_k_testString, sizeof(__pyx_k_testString), 0, 0, 1, 1},
  {&__pyx_n_s_test_expr, __pyx_k_test_expr, sizeof(__pyx_k_test_expr), 0, 0, 1, 1},
  {&__pyx_n_s_tests, __pyx_k_tests, sizeof(__pyx_k_tests), 0, 0, 1, 1},
  {&__pyx_n_s_thiscol, __pyx_k_thiscol, sizeof(__pyx_k_thiscol), 0, 0, 1, 1},
  {&__pyx_n_s_thread, __pyx_k_thread, sizeof(__pyx_k_thread), 0, 0, 1, 1},
  {&__pyx_n_s_threading, __pyx_k_threading, sizeof(__pyx_k_threading), 0, 0, 1, 1},
  {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
  {&__pyx_n_s_throwException, __pyx_k_throwException, sizeof(__pyx_k_throwException), 0, 0, 1, 1},
  {&__pyx_n_s_tmp, __pyx_k_tmp, sizeof(__pyx_k_tmp), 0, 0, 1, 1},
  {&__pyx_n_s_tmpExprs, __pyx_k_tmpExprs, sizeof(__pyx_k_tmpExprs), 0, 0, 1, 1},
  {&__pyx_n_s_tmpLoc, __pyx_k_tmpLoc, sizeof(__pyx_k_tmpLoc), 0, 0, 1, 1},
  {&__pyx_n_s_tmpOpt, __pyx_k_tmpOpt, sizeof(__pyx_k_tmpOpt), 0, 0, 1, 1},
  {&__pyx_n_s_tmpReqd, __pyx_k_tmpReqd, sizeof(__pyx_k_tmpReqd), 0, 0, 1, 1},
  {&__pyx_n_s_tmploc, __pyx_k_tmploc, sizeof(__pyx_k_tmploc), 0, 0, 1, 1},
  {&__pyx_n_s_tmptokens, __pyx_k_tmptokens, sizeof(__pyx_k_tmptokens), 0, 0, 1, 1},
  {&__pyx_n_s_to, __pyx_k_to, sizeof(__pyx_k_to), 0, 0, 1, 1},
  {&__pyx_n_s_toItem, __pyx_k_toItem, sizeof(__pyx_k_toItem), 0, 0, 1, 1},
  {&__pyx_n_s_to_symbols, __pyx_k_to_symbols, sizeof(__pyx_k_to_symbols), 0, 0, 1, 1},
  {&__pyx_n_s_tok, __pyx_k_tok, sizeof(__pyx_k_tok), 0, 0, 1, 1},
  {&__pyx_n_s_tokenlist, __pyx_k_tokenlist, sizeof(__pyx_k_tokenlist), 0, 0, 1, 1},
  {&__pyx_n_s_tokens, __pyx_k_tokens, sizeof(__pyx_k_tokens), 0, 0, 1, 1},
  {&__pyx_n_s_tokensStart, __pyx_k_tokensStart, sizeof(__pyx_k_tokensStart), 0, 0, 1, 1},
  {&__pyx_n_s_toklist, __pyx_k_toklist, sizeof(__pyx_k_toklist), 0, 0, 1, 1},
  {&__pyx_n_s_toks, __pyx_k_toks, sizeof(__pyx_k_toks), 0, 0, 1, 1},
  {&__pyx_n_s_traceback, __pyx_k_traceback, sizeof(__pyx_k_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_transformString, __pyx_k_transformString, sizeof(__pyx_k_transformString), 0, 0, 1, 1},
  {&__pyx_n_s_trim_arity, __pyx_k_trim_arity, sizeof(__pyx_k_trim_arity), 0, 0, 1, 1},
  {&__pyx_n_s_trim_traceback, __pyx_k_trim_traceback, sizeof(__pyx_k_trim_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_tryParse, __pyx_k_tryParse, sizeof(__pyx_k_tryParse), 0, 0, 1, 1},
  {&__pyx_n_s_try_not_ender, __pyx_k_try_not_ender, sizeof(__pyx_k_try_not_ender), 0, 0, 1, 1},
  {&__pyx_n_s_tup, __pyx_k_tup, sizeof(__pyx_k_tup), 0, 0, 1, 1},
  {&__pyx_n_s_types, __pyx_k_types, sizeof(__pyx_k_types), 0, 0, 1, 1},
  {&__pyx_n_s_unquoteResults, __pyx_k_unquoteResults, sizeof(__pyx_k_unquoteResults), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_n_s_upper, __pyx_k_upper, sizeof(__pyx_k_upper), 0, 0, 1, 1},
  {&__pyx_kp_s_user_throwable_exception_thrown, __pyx_k_user_throwable_exception_thrown, sizeof(__pyx_k_user_throwable_exception_thrown), 0, 0, 1, 0},
  {&__pyx_n_s_v, __pyx_k_v, sizeof(__pyx_k_v), 0, 0, 1, 1},
  {&__pyx_n_s_validate, __pyx_k_validate, sizeof(__pyx_k_validate), 0, 0, 1, 1},
  {&__pyx_n_s_validateTrace, __pyx_k_validateTrace, sizeof(__pyx_k_validateTrace), 0, 0, 1, 1},
  {&__pyx_n_s_value, __pyx_k_value, sizeof(__pyx_k_value), 0, 0, 1, 1},
  {&__pyx_n_s_values, __pyx_k_values, sizeof(__pyx_k_values), 0, 0, 1, 1},
  {&__pyx_n_s_vars, __pyx_k_vars, sizeof(__pyx_k_vars), 0, 0, 1, 1},
  {&__pyx_n_s_verbose_stacktrace, __pyx_k_verbose_stacktrace, sizeof(__pyx_k_verbose_stacktrace), 0, 0, 1, 1},
  {&__pyx_n_s_version_info, __pyx_k_version_info, sizeof(__pyx_k_version_info), 0, 0, 1, 1},
  {&__pyx_n_s_vlist, __pyx_k_vlist, sizeof(__pyx_k_vlist), 0, 0, 1, 1},
  {&__pyx_n_s_vv, __pyx_k_vv, sizeof(__pyx_k_vv), 0, 0, 1, 1},
  {&__pyx_n_s_warn, __pyx_k_warn, sizeof(__pyx_k_warn), 0, 0, 1, 1},
  {&__pyx_n_s_warn_2, __pyx_k_warn_2, sizeof(__pyx_k_warn_2), 0, 0, 1, 1},
  {&__pyx_n_s_warn_multiple_tokens_in_named_al, __pyx_k_warn_multiple_tokens_in_named_al, sizeof(__pyx_k_warn_multiple_tokens_in_named_al), 0, 0, 1, 1},
  {&__pyx_n_s_warn_name_set_on_empty_Forward, __pyx_k_warn_name_set_on_empty_Forward, sizeof(__pyx_k_warn_name_set_on_empty_Forward), 0, 0, 1, 1},
  {&__pyx_n_s_warn_on_multiple_string_args_to, __pyx_k_warn_on_multiple_string_args_to, sizeof(__pyx_k_warn_on_multiple_string_args_to), 0, 0, 1, 1},
  {&__pyx_n_s_warn_ungrouped_named_tokens_in_c, __pyx_k_warn_ungrouped_named_tokens_in_c, sizeof(__pyx_k_warn_ungrouped_named_tokens_in_c), 0, 0, 1, 1},
  {&__pyx_n_s_warnings, __pyx_k_warnings, sizeof(__pyx_k_warnings), 0, 0, 1, 1},
  {&__pyx_n_s_weakref, __pyx_k_weakref, sizeof(__pyx_k_weakref), 0, 0, 1, 1},
  {&__pyx_n_s_whiteChars, __pyx_k_whiteChars, sizeof(__pyx_k_whiteChars), 0, 0, 1, 1},
  {&__pyx_n_s_whiteStrs, __pyx_k_whiteStrs, sizeof(__pyx_k_whiteStrs), 0, 0, 1, 1},
  {&__pyx_n_s_whitespace, __pyx_k_whitespace, sizeof(__pyx_k_whitespace), 0, 0, 1, 1},
  {&__pyx_n_s_wkref, __pyx_k_wkref, sizeof(__pyx_k_wkref), 0, 0, 1, 1},
  {&__pyx_n_s_wordChars, __pyx_k_wordChars, sizeof(__pyx_k_wordChars), 0, 0, 1, 1},
  {&__pyx_n_s_wrapper, __pyx_k_wrapper, sizeof(__pyx_k_wrapper), 0, 0, 1, 1},
  {&__pyx_n_s_write, __pyx_k_write, sizeof(__pyx_k_write), 0, 0, 1, 1},
  {&__pyx_n_s_ws, __pyx_k_ws, sizeof(__pyx_k_ws), 0, 0, 1, 1},
  {&__pyx_n_s_ws_map, __pyx_k_ws_map, sizeof(__pyx_k_ws_map), 0, 0, 1, 1},
  {&__pyx_n_s_wschar, __pyx_k_wschar, sizeof(__pyx_k_wschar), 0, 0, 1, 1},
  {&__pyx_n_s_wslit, __pyx_k_wslit, sizeof(__pyx_k_wslit), 0, 0, 1, 1},
  {&__pyx_n_s_wt, __pyx_k_wt, sizeof(__pyx_k_wt), 0, 0, 1, 1},
  {&__pyx_n_s_xmlBodyText, __pyx_k_xmlBodyText, sizeof(__pyx_k_xmlBodyText), 0, 0, 1, 1},
  {&__pyx_n_s_xml_escape, __pyx_k_xml_escape, sizeof(__pyx_k_xml_escape), 0, 0, 1, 1},
  {&__pyx_n_s_xml_escape_locals_genexpr, __pyx_k_xml_escape_locals_genexpr, sizeof(__pyx_k_xml_escape_locals_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_xor, __pyx_k_xor, sizeof(__pyx_k_xor), 0, 0, 1, 1},
  {&__pyx_n_s_zip, __pyx_k_zip, sizeof(__pyx_k_zip), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(0, 22, __pyx_L1_error)
  __pyx_builtin_vars = __Pyx_GetBuiltinName(__pyx_n_s_vars); if (!__pyx_builtin_vars) __PYX_ERR(0, 67, __pyx_L1_error)
  __pyx_builtin_object = __Pyx_GetBuiltinName(__pyx_n_s_object); if (!__pyx_builtin_object) __PYX_ERR(0, 164, __pyx_L1_error)
  __pyx_builtin_staticmethod = __Pyx_GetBuiltinName(__pyx_n_s_staticmethod); if (!__pyx_builtin_staticmethod) __PYX_ERR(0, 980, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_builtin_chr = __Pyx_GetBuiltinName(__pyx_n_s_chr); if (!__pyx_builtin_chr) __PYX_ERR(0, 98, __pyx_L1_error)
  __pyx_builtin_zip = __Pyx_GetBuiltinName(__pyx_n_s_zip); if (!__pyx_builtin_zip) __PYX_ERR(0, 120, __pyx_L1_error)
  __pyx_builtin_KeyError = __Pyx_GetBuiltinName(__pyx_n_s_KeyError); if (!__pyx_builtin_KeyError) __PYX_ERR(0, 267, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(0, 267, __pyx_L1_error)
  __pyx_builtin_IndexError = __Pyx_GetBuiltinName(__pyx_n_s_IndexError); if (!__pyx_builtin_IndexError) __PYX_ERR(0, 267, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(0, 308, __pyx_L1_error)
  __pyx_builtin_AttributeError = __Pyx_GetBuiltinName(__pyx_n_s_AttributeError); if (!__pyx_builtin_AttributeError) __PYX_ERR(0, 926, __pyx_L1_error)
  __pyx_builtin_map = __Pyx_GetBuiltinName(__pyx_n_s_map); if (!__pyx_builtin_map) __PYX_ERR(0, 1305, __pyx_L1_error)
  __pyx_builtin_Ellipsis = __Pyx_GetBuiltinName(__pyx_n_s_Ellipsis); if (!__pyx_builtin_Ellipsis) __PYX_ERR(0, 1913, __pyx_L1_error)
  __pyx_builtin_SyntaxWarning = __Pyx_GetBuiltinName(__pyx_n_s_SyntaxWarning); if (!__pyx_builtin_SyntaxWarning) __PYX_ERR(0, 1920, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 2015, __pyx_L1_error)
  __pyx_builtin_open = __Pyx_GetBuiltinName(__pyx_n_s_open); if (!__pyx_builtin_open) __PYX_ERR(0, 2334, __pyx_L1_error)
  __pyx_builtin_id = __Pyx_GetBuiltinName(__pyx_n_s_id); if (!__pyx_builtin_id) __PYX_ERR(0, 2360, __pyx_L1_error)
  __pyx_builtin_super = __Pyx_GetBuiltinName(__pyx_n_s_super); if (!__pyx_builtin_super) __PYX_ERR(0, 2558, __pyx_L1_error)
  __pyx_builtin_SyntaxError = __Pyx_GetBuiltinName(__pyx_n_s_SyntaxError); if (!__pyx_builtin_SyntaxError) __PYX_ERR(0, 3265, __pyx_L1_error)
  __pyx_builtin_sum = __Pyx_GetBuiltinName(__pyx_n_s_sum); if (!__pyx_builtin_sum) __PYX_ERR(0, 4190, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "Aspidites/_vendor/pyparsing_extension.py":98
 *     # ~  escape these chars: ^-[]
 *     for c in r"\^-[]":
 *         s = s.replace(c, chr(92) + c)             # <<<<<<<<<<<<<<
 *     s = s.replace("\n", r"\n")
 *     s = s.replace("\t", r"\t")
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_int_92); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "Aspidites/_vendor/pyparsing_extension.py":99
 *     for c in r"\^-[]":
 *         s = s.replace(c, chr(92) + c)
 *     s = s.replace("\n", r"\n")             # <<<<<<<<<<<<<<
 *     s = s.replace("\t", r"\t")
 *     return str(s)
 */
  __pyx_tuple__5 = PyTuple_Pack(2, __pyx_kp_s__4, __pyx_kp_s_n); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "Aspidites/_vendor/pyparsing_extension.py":100
 *         s = s.replace(c, chr(92) + c)
 *     s = s.replace("\n", r"\n")
 *     s = s.replace("\t", r"\t")             # <<<<<<<<<<<<<<
 *     return str(s)
 * 
 */
  __pyx_tuple__7 = PyTuple_Pack(2, __pyx_kp_s__6, __pyx_kp_s_t); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "Aspidites/_vendor/pyparsing_extension.py":242
 *                 toklist = []
 *             if isinstance(toklist, list):
 *                 self.__toklist = toklist[:]             # <<<<<<<<<<<<<<
 *             elif isinstance(toklist, _generatorType):
 *                 self.__toklist = list(toklist)
 */
  __pyx_slice__11 = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice__11)) __PYX_ERR(0, 242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__11);
  __Pyx_GIVEREF(__pyx_slice__11);

  /* "Aspidites/_vendor/pyparsing_extension.py":328
 * 
 *     def __reversed__(self):
 *         return iter(self.__toklist[::-1])             # <<<<<<<<<<<<<<
 * 
 *     def _iterkeys(self):
 */
  __pyx_slice__13 = PySlice_New(Py_None, Py_None, __pyx_int_neg_1); if (unlikely(!__pyx_slice__13)) __PYX_ERR(0, 328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__13);
  __Pyx_GIVEREF(__pyx_slice__13);

  /* "Aspidites/_vendor/pyparsing_extension.py":613
 *             item_fn = self.iteritems
 * 
 *         def toItem(obj):             # <<<<<<<<<<<<<<
 *             if isinstance(obj, ParseResults):
 *                 if obj.haskeys():
 */
  __pyx_tuple__17 = PyTuple_Pack(2, __pyx_n_s_obj, __pyx_n_s_v); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(0, 613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);
  __pyx_codeobj__18 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__17, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_toItem, 613, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__18)) __PYX_ERR(0, 613, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":867
 *         """
 * 
 *         def is_iterable(obj):             # <<<<<<<<<<<<<<
 *             try:
 *                 iter(obj)
 */
  __pyx_tuple__23 = PyTuple_Pack(1, __pyx_n_s_obj); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(0, 867, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);
  __pyx_codeobj__24 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__23, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_is_iterable, 867, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__24)) __PYX_ERR(0, 867, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":933
 *                 foundstr = ', found end of text'
 *             else:
 *                 foundstr = (', found %r' % self.pstr[self.loc:self.loc + 1]).replace(r'\\', '\\')             # <<<<<<<<<<<<<<
 *         else:
 *             foundstr = ''
 */
  __pyx_tuple__27 = PyTuple_Pack(2, __pyx_kp_s__25, __pyx_kp_s__26); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(0, 933, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);

  /* "Aspidites/_vendor/pyparsing_extension.py":1020
 *                 frm = ff[0]
 * 
 *                 f_self = frm.f_locals.get('self', None)             # <<<<<<<<<<<<<<
 *                 if isinstance(f_self, ParserElement):
 *                     if frm.f_code.co_name not in ('parseImpl', '_parseNoCache'):
 */
  __pyx_tuple__31 = PyTuple_Pack(2, __pyx_n_s_self, Py_None); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(0, 1020, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);

  /* "Aspidites/_vendor/pyparsing_extension.py":1157
 *         self.errmsg = ""
 *         self.modalResults = True  # used to mark results names as modal (report only last) or cumulative (list all)
 *         self.debugActions = (None, None, None)  # custom debug actions             # <<<<<<<<<<<<<<
 *         self.re = None
 *         self.callPreparse = True  # used to avoid redundant calls to preParse
 */
  __pyx_tuple__32 = PyTuple_Pack(3, Py_None, Py_None, Py_None); if (unlikely(!__pyx_tuple__32)) __PYX_ERR(0, 1157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__32);
  __Pyx_GIVEREF(__pyx_tuple__32);

  /* "Aspidites/_vendor/pyparsing_extension.py":1232
 *         newself = self.copy()
 *         if name.endswith("*"):
 *             name = name[:-1]             # <<<<<<<<<<<<<<
 *             listAllMatches = True
 *         newself.resultsName = name
 */
  __pyx_slice__34 = PySlice_New(Py_None, __pyx_int_neg_1, Py_None); if (unlikely(!__pyx_slice__34)) __PYX_ERR(0, 1232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__34);
  __Pyx_GIVEREF(__pyx_slice__34);

  /* "Aspidites/_vendor/pyparsing_extension.py":1246
 *             _parseMethod = self._parse
 * 
 *             def breaker(instring, loc, doActions=True, callPreParse=True):             # <<<<<<<<<<<<<<
 *                 import pdb
 *                 # this call to pdb.set_trace() is intentional, not a checkin error
 */
  __pyx_tuple__35 = PyTuple_Pack(5, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_doActions, __pyx_n_s_callPreParse, __pyx_n_s_pdb); if (unlikely(!__pyx_tuple__35)) __PYX_ERR(0, 1246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__35);
  __Pyx_GIVEREF(__pyx_tuple__35);
  __pyx_codeobj__36 = (PyObject*)__Pyx_PyCode_New(4, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__35, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_breaker, 1246, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__36)) __PYX_ERR(0, 1246, __pyx_L1_error)
  __pyx_tuple__37 = PyTuple_Pack(2, ((PyObject *)Py_True), ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__37)) __PYX_ERR(0, 1246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__37);
  __Pyx_GIVEREF(__pyx_tuple__37);

  /* "Aspidites/_vendor/pyparsing_extension.py":1304
 *         else:
 *             if not all(callable(fn) for fn in fns):
 *                 raise TypeError("parse actions must be callable")             # <<<<<<<<<<<<<<
 *             self.parseAction = list(map(_trim_arity, list(fns)))
 *             self.callDuringTry = kwargs.get("callDuringTry", False)
 */
  __pyx_tuple__38 = PyTuple_Pack(1, __pyx_kp_s_parse_actions_must_be_callable); if (unlikely(!__pyx_tuple__38)) __PYX_ERR(0, 1304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__38);
  __Pyx_GIVEREF(__pyx_tuple__38);

  /* "Aspidites/_vendor/pyparsing_extension.py":1498
 *             self.not_in_cache = not_in_cache = object()
 * 
 *             def get(self, key):             # <<<<<<<<<<<<<<
 *                 return cache.get(key, not_in_cache)
 * 
 */
  __pyx_tuple__39 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_key); if (unlikely(!__pyx_tuple__39)) __PYX_ERR(0, 1498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__39);
  __Pyx_GIVEREF(__pyx_tuple__39);
  __pyx_codeobj__40 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_get, 1498, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__40)) __PYX_ERR(0, 1498, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1501
 *                 return cache.get(key, not_in_cache)
 * 
 *             def set(self, key, value):             # <<<<<<<<<<<<<<
 *                 cache[key] = value
 * 
 */
  __pyx_tuple__41 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_key, __pyx_n_s_value); if (unlikely(!__pyx_tuple__41)) __PYX_ERR(0, 1501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__41);
  __Pyx_GIVEREF(__pyx_tuple__41);
  __pyx_codeobj__42 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_set, 1501, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__42)) __PYX_ERR(0, 1501, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1504
 *                 cache[key] = value
 * 
 *             def clear(self):             # <<<<<<<<<<<<<<
 *                 cache.clear()
 * 
 */
  __pyx_tuple__43 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__43)) __PYX_ERR(0, 1504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__43);
  __Pyx_GIVEREF(__pyx_tuple__43);
  __pyx_codeobj__44 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__43, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_clear, 1504, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__44)) __PYX_ERR(0, 1504, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1507
 *                 cache.clear()
 * 
 *             def cache_len(self):             # <<<<<<<<<<<<<<
 *                 return len(cache)
 * 
 */
  __pyx_tuple__45 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__45)) __PYX_ERR(0, 1507, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__45);
  __Pyx_GIVEREF(__pyx_tuple__45);
  __pyx_codeobj__46 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__45, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_cache_len, 1507, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__46)) __PYX_ERR(0, 1507, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1522
 *                 cache = _OrderedDict()
 * 
 *                 def get(self, key):             # <<<<<<<<<<<<<<
 *                     return cache.get(key, not_in_cache)
 * 
 */
  __pyx_tuple__47 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_key); if (unlikely(!__pyx_tuple__47)) __PYX_ERR(0, 1522, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__47);
  __Pyx_GIVEREF(__pyx_tuple__47);
  __pyx_codeobj__48 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__47, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_get, 1522, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__48)) __PYX_ERR(0, 1522, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1525
 *                     return cache.get(key, not_in_cache)
 * 
 *                 def set(self, key, value):             # <<<<<<<<<<<<<<
 *                     cache[key] = value
 *                     while len(cache) > size:
 */
  __pyx_tuple__49 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_key, __pyx_n_s_value); if (unlikely(!__pyx_tuple__49)) __PYX_ERR(0, 1525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__49);
  __Pyx_GIVEREF(__pyx_tuple__49);
  __pyx_codeobj__50 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__49, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_set, 1525, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__50)) __PYX_ERR(0, 1525, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1533
 *                             pass
 * 
 *                 def clear(self):             # <<<<<<<<<<<<<<
 *                     cache.clear()
 * 
 */
  __pyx_tuple__51 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__51)) __PYX_ERR(0, 1533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__51);
  __Pyx_GIVEREF(__pyx_tuple__51);
  __pyx_codeobj__52 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__51, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_clear, 1533, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__52)) __PYX_ERR(0, 1533, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1536
 *                     cache.clear()
 * 
 *                 def cache_len(self):             # <<<<<<<<<<<<<<
 *                     return len(cache)
 * 
 */
  __pyx_tuple__53 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__53)) __PYX_ERR(0, 1536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__53);
  __Pyx_GIVEREF(__pyx_tuple__53);
  __pyx_codeobj__54 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__53, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_cache_len, 1536, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__54)) __PYX_ERR(0, 1536, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1552
 *                 key_fifo = collections.deque([], size)
 * 
 *                 def get(self, key):             # <<<<<<<<<<<<<<
 *                     return cache.get(key, not_in_cache)
 * 
 */
  __pyx_tuple__55 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_key); if (unlikely(!__pyx_tuple__55)) __PYX_ERR(0, 1552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__55);
  __Pyx_GIVEREF(__pyx_tuple__55);
  __pyx_codeobj__56 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__55, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_get, 1552, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__56)) __PYX_ERR(0, 1552, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1555
 *                     return cache.get(key, not_in_cache)
 * 
 *                 def set(self, key, value):             # <<<<<<<<<<<<<<
 *                     cache[key] = value
 *                     while len(key_fifo) > size:
 */
  __pyx_tuple__57 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_key, __pyx_n_s_value); if (unlikely(!__pyx_tuple__57)) __PYX_ERR(0, 1555, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__57);
  __Pyx_GIVEREF(__pyx_tuple__57);
  __pyx_codeobj__58 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__57, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_set, 1555, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__58)) __PYX_ERR(0, 1555, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1561
 *                     key_fifo.append(key)
 * 
 *                 def clear(self):             # <<<<<<<<<<<<<<
 *                     cache.clear()
 *                     key_fifo.clear()
 */
  __pyx_tuple__59 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__59)) __PYX_ERR(0, 1561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__59);
  __Pyx_GIVEREF(__pyx_tuple__59);
  __pyx_codeobj__60 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__59, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_clear, 1561, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__60)) __PYX_ERR(0, 1561, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1565
 *                     key_fifo.clear()
 * 
 *                 def cache_len(self):             # <<<<<<<<<<<<<<
 *                     return len(cache)
 * 
 */
  __pyx_tuple__61 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__61)) __PYX_ERR(0, 1565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__61);
  __Pyx_GIVEREF(__pyx_tuple__61);
  __pyx_codeobj__62 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__61, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_cache_len, 1565, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__62)) __PYX_ERR(0, 1565, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1987
 *         from .pyparsing import ZeroOrMore, OneOrMore, And
 *         if other is Ellipsis:
 *             other = (0, None)             # <<<<<<<<<<<<<<
 *         elif isinstance(other, tuple) and other[:1] == (Ellipsis,):
 *             other = ((0,) + other[1:] + (None,))[:2]
 */
  __pyx_tuple__65 = PyTuple_Pack(2, __pyx_int_0, Py_None); if (unlikely(!__pyx_tuple__65)) __PYX_ERR(0, 1987, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__65);
  __Pyx_GIVEREF(__pyx_tuple__65);

  /* "Aspidites/_vendor/pyparsing_extension.py":1988
 *         if other is Ellipsis:
 *             other = (0, None)
 *         elif isinstance(other, tuple) and other[:1] == (Ellipsis,):             # <<<<<<<<<<<<<<
 *             other = ((0,) + other[1:] + (None,))[:2]
 * 
 */
  __pyx_slice__66 = PySlice_New(Py_None, __pyx_int_1, Py_None); if (unlikely(!__pyx_slice__66)) __PYX_ERR(0, 1988, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__66);
  __Pyx_GIVEREF(__pyx_slice__66);
  __pyx_tuple__67 = PyTuple_Pack(1, __pyx_builtin_Ellipsis); if (unlikely(!__pyx_tuple__67)) __PYX_ERR(0, 1988, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__67);
  __Pyx_GIVEREF(__pyx_tuple__67);

  /* "Aspidites/_vendor/pyparsing_extension.py":1989
 *             other = (0, None)
 *         elif isinstance(other, tuple) and other[:1] == (Ellipsis,):
 *             other = ((0,) + other[1:] + (None,))[:2]             # <<<<<<<<<<<<<<
 * 
 *         if isinstance(other, int):
 */
  __pyx_tuple__68 = PyTuple_Pack(1, __pyx_int_0); if (unlikely(!__pyx_tuple__68)) __PYX_ERR(0, 1989, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__68);
  __Pyx_GIVEREF(__pyx_tuple__68);
  __pyx_slice__69 = PySlice_New(__pyx_int_1, Py_None, Py_None); if (unlikely(!__pyx_slice__69)) __PYX_ERR(0, 1989, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__69);
  __Pyx_GIVEREF(__pyx_slice__69);
  __pyx_tuple__70 = PyTuple_Pack(1, Py_None); if (unlikely(!__pyx_tuple__70)) __PYX_ERR(0, 1989, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__70);
  __Pyx_GIVEREF(__pyx_tuple__70);
  __pyx_slice__71 = PySlice_New(Py_None, __pyx_int_2, Py_None); if (unlikely(!__pyx_slice__71)) __PYX_ERR(0, 1989, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__71);
  __Pyx_GIVEREF(__pyx_slice__71);

  /* "Aspidites/_vendor/pyparsing_extension.py":1995
 *         elif isinstance(other, tuple):
 *             other = tuple(o if o is not Ellipsis else None for o in other)
 *             other = (other + (None, None))[:2]             # <<<<<<<<<<<<<<
 *             if other[0] is None:
 *                 other = (0, other[1])
 */
  __pyx_tuple__72 = PyTuple_Pack(2, Py_None, Py_None); if (unlikely(!__pyx_tuple__72)) __PYX_ERR(0, 1995, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__72);
  __Pyx_GIVEREF(__pyx_tuple__72);

  /* "Aspidites/_vendor/pyparsing_extension.py":2015
 * 
 *         if minElements < 0:
 *             raise ValueError("cannot multiply ParserElement by negative value")             # <<<<<<<<<<<<<<
 *         if optElements < 0:
 *             raise ValueError("second tuple value must be greater or equal to first tuple value")
 */
  __pyx_tuple__73 = PyTuple_Pack(1, __pyx_kp_s_cannot_multiply_ParserElement_by); if (unlikely(!__pyx_tuple__73)) __PYX_ERR(0, 2015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__73);
  __Pyx_GIVEREF(__pyx_tuple__73);

  /* "Aspidites/_vendor/pyparsing_extension.py":2017
 *             raise ValueError("cannot multiply ParserElement by negative value")
 *         if optElements < 0:
 *             raise ValueError("second tuple value must be greater or equal to first tuple value")             # <<<<<<<<<<<<<<
 *         if minElements == optElements == 0:
 *             raise ValueError("cannot multiply ParserElement by 0 or (0, 0)")
 */
  __pyx_tuple__74 = PyTuple_Pack(1, __pyx_kp_s_second_tuple_value_must_be_great); if (unlikely(!__pyx_tuple__74)) __PYX_ERR(0, 2017, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__74);
  __Pyx_GIVEREF(__pyx_tuple__74);

  /* "Aspidites/_vendor/pyparsing_extension.py":2019
 *             raise ValueError("second tuple value must be greater or equal to first tuple value")
 *         if minElements == optElements == 0:
 *             raise ValueError("cannot multiply ParserElement by 0 or (0, 0)")             # <<<<<<<<<<<<<<
 * 
 *         if optElements:
 */
  __pyx_tuple__75 = PyTuple_Pack(1, __pyx_kp_s_cannot_multiply_ParserElement_by_2); if (unlikely(!__pyx_tuple__75)) __PYX_ERR(0, 2019, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__75);
  __Pyx_GIVEREF(__pyx_tuple__75);

  /* "Aspidites/_vendor/pyparsing_extension.py":2022
 * 
 *         if optElements:
 *             def makeOptionalList(n):             # <<<<<<<<<<<<<<
 *                 from .pyparsing import Optional
 *                 if n > 1:
 */
  __pyx_tuple__76 = PyTuple_Pack(2, __pyx_n_s_n_2, __pyx_n_s_Optional); if (unlikely(!__pyx_tuple__76)) __PYX_ERR(0, 2022, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__76);
  __Pyx_GIVEREF(__pyx_tuple__76);
  __pyx_codeobj__77 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__76, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_makeOptionalList, 2022, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__77)) __PYX_ERR(0, 2022, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2164
 * 
 *         if len(key) > 2:
 *             warnings.warn("only 1 or 2 index arguments supported ({0}{1})".format(key[:5],             # <<<<<<<<<<<<<<
 *                                                                                   '... [{0}]'.format(len(key))
 *                                                                                   if len(key) > 5 else ''))
 */
  __pyx_slice__78 = PySlice_New(Py_None, __pyx_int_5, Py_None); if (unlikely(!__pyx_slice__78)) __PYX_ERR(0, 2164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__78);
  __Pyx_GIVEREF(__pyx_slice__78);

  /* "Aspidites/_vendor/pyparsing_extension.py":2733
 *             self.firstMatchChar = matchString[0]
 *         except IndexError:
 *             warnings.warn("null string passed to Literal; use Empty() instead",             # <<<<<<<<<<<<<<
 *                           SyntaxWarning, stacklevel=2)
 *             self.__class__ = Empty
 */
  __pyx_tuple__81 = PyTuple_Pack(2, __pyx_kp_s_null_string_passed_to_Literal_us, __pyx_builtin_SyntaxWarning); if (unlikely(!__pyx_tuple__81)) __PYX_ERR(0, 2733, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__81);
  __Pyx_GIVEREF(__pyx_tuple__81);

  /* "Aspidites/_vendor/pyparsing_extension.py":2799
 *             self.firstMatchChar = matchString[0]
 *         except IndexError:
 *             warnings.warn("null string passed to Keyword; use Empty() instead",             # <<<<<<<<<<<<<<
 *                           SyntaxWarning, stacklevel=2)
 *         self.name = '"%s"' % self.match
 */
  __pyx_tuple__82 = PyTuple_Pack(2, __pyx_kp_s_null_string_passed_to_Keyword_us, __pyx_builtin_SyntaxWarning); if (unlikely(!__pyx_tuple__82)) __PYX_ERR(0, 2799, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__82);
  __Pyx_GIVEREF(__pyx_tuple__82);

  /* "Aspidites/_vendor/pyparsing_extension.py":3025
 * 
 *         if min < 1:
 *             raise ValueError(             # <<<<<<<<<<<<<<
 *                 "cannot specify a minimum length < 1; use Optional(Word()) if zero-length word is permitted")
 * 
 */
  __pyx_tuple__83 = PyTuple_Pack(1, __pyx_kp_s_cannot_specify_a_minimum_length); if (unlikely(!__pyx_tuple__83)) __PYX_ERR(0, 3025, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__83);
  __Pyx_GIVEREF(__pyx_tuple__83);

  /* "Aspidites/_vendor/pyparsing_extension.py":3102
 *             def charsAsStr(s):
 *                 if len(s) > 4:
 *                     return s[:4] + "..."             # <<<<<<<<<<<<<<
 *                 else:
 *                     return s
 */
  __pyx_slice__84 = PySlice_New(Py_None, __pyx_int_4, Py_None); if (unlikely(!__pyx_slice__84)) __PYX_ERR(0, 3102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__84);
  __Pyx_GIVEREF(__pyx_slice__84);

  /* "Aspidites/_vendor/pyparsing_extension.py":3100
 *         if self.strRepr is None:
 * 
 *             def charsAsStr(s):             # <<<<<<<<<<<<<<
 *                 if len(s) > 4:
 *                     return s[:4] + "..."
 */
  __pyx_tuple__86 = PyTuple_Pack(1, __pyx_n_s_s_4); if (unlikely(!__pyx_tuple__86)) __PYX_ERR(0, 3100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__86);
  __Pyx_GIVEREF(__pyx_tuple__86);
  __pyx_codeobj__87 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__86, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_charsAsStr, 3100, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__87)) __PYX_ERR(0, 3100, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3174
 *         if isinstance(pattern, str):
 *             if not pattern:
 *                 warnings.warn("null string passed to Regex; use Empty() instead",             # <<<<<<<<<<<<<<
 *                               SyntaxWarning, stacklevel=2)
 * 
 */
  __pyx_tuple__88 = PyTuple_Pack(2, __pyx_kp_s_null_string_passed_to_Regex_use, __pyx_builtin_SyntaxWarning); if (unlikely(!__pyx_tuple__88)) __PYX_ERR(0, 3174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__88);
  __Pyx_GIVEREF(__pyx_tuple__88);

  /* "Aspidites/_vendor/pyparsing_extension.py":3194
 * 
 *         else:
 *             raise TypeError("Regex may only be constructed with a string or a compiled RE object")             # <<<<<<<<<<<<<<
 * 
 *         self.re_match = self.re.match
 */
  __pyx_tuple__89 = PyTuple_Pack(1, __pyx_kp_s_Regex_may_only_be_constructed_wi); if (unlikely(!__pyx_tuple__89)) __PYX_ERR(0, 3194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__89);
  __Pyx_GIVEREF(__pyx_tuple__89);

  /* "Aspidites/_vendor/pyparsing_extension.py":3263
 *         """
 *         if self.asGroupList:
 *             warnings.warn("cannot use sub() with Regex(asGroupList=True)",             # <<<<<<<<<<<<<<
 *                           SyntaxWarning, stacklevel=2)
 *             raise SyntaxError()
 */
  __pyx_tuple__90 = PyTuple_Pack(2, __pyx_kp_s_cannot_use_sub_with_Regex_asGrou, __pyx_builtin_SyntaxWarning); if (unlikely(!__pyx_tuple__90)) __PYX_ERR(0, 3263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__90);
  __Pyx_GIVEREF(__pyx_tuple__90);

  /* "Aspidites/_vendor/pyparsing_extension.py":3268
 * 
 *         if self.asMatch and callable(repl):
 *             warnings.warn("cannot use sub() with a callable with Regex(asMatch=True)",             # <<<<<<<<<<<<<<
 *                           SyntaxWarning, stacklevel=2)
 *             raise SyntaxError()
 */
  __pyx_tuple__91 = PyTuple_Pack(2, __pyx_kp_s_cannot_use_sub_with_a_callable_w, __pyx_builtin_SyntaxWarning); if (unlikely(!__pyx_tuple__91)) __PYX_ERR(0, 3268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__91);
  __Pyx_GIVEREF(__pyx_tuple__91);

  /* "Aspidites/_vendor/pyparsing_extension.py":3273
 * 
 *         if self.asMatch:
 *             def pa(tokens):             # <<<<<<<<<<<<<<
 *                 return tokens[0].expand(repl)
 *         else:
 */
  __pyx_tuple__92 = PyTuple_Pack(1, __pyx_n_s_tokens); if (unlikely(!__pyx_tuple__92)) __PYX_ERR(0, 3273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__92);
  __Pyx_GIVEREF(__pyx_tuple__92);
  __pyx_codeobj__93 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__92, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_pa, 3273, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__93)) __PYX_ERR(0, 3273, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3276
 *                 return tokens[0].expand(repl)
 *         else:
 *             def pa(tokens):             # <<<<<<<<<<<<<<
 *                 return self.re.sub(repl, tokens[0])
 *         return self.addParseAction(pa)
 */
  __pyx_tuple__94 = PyTuple_Pack(1, __pyx_n_s_tokens); if (unlikely(!__pyx_tuple__94)) __PYX_ERR(0, 3276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__94);
  __Pyx_GIVEREF(__pyx_tuple__94);
  __pyx_codeobj__95 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__94, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_pa, 3276, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__95)) __PYX_ERR(0, 3276, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3328
 *         quoteChar = quoteChar.strip()
 *         if not quoteChar:
 *             warnings.warn("quoteChar cannot be the empty string", SyntaxWarning, stacklevel=2)             # <<<<<<<<<<<<<<
 *             raise SyntaxError()
 * 
 */
  __pyx_tuple__96 = PyTuple_Pack(2, __pyx_kp_s_quoteChar_cannot_be_the_empty_st, __pyx_builtin_SyntaxWarning); if (unlikely(!__pyx_tuple__96)) __PYX_ERR(0, 3328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__96);
  __Pyx_GIVEREF(__pyx_tuple__96);

  /* "Aspidites/_vendor/pyparsing_extension.py":3336
 *             endQuoteChar = endQuoteChar.strip()
 *             if not endQuoteChar:
 *                 warnings.warn("endQuoteChar cannot be the empty string", SyntaxWarning, stacklevel=2)             # <<<<<<<<<<<<<<
 *                 raise SyntaxError()
 * 
 */
  __pyx_tuple__97 = PyTuple_Pack(2, __pyx_kp_s_endQuoteChar_cannot_be_the_empty, __pyx_builtin_SyntaxWarning); if (unlikely(!__pyx_tuple__97)) __PYX_ERR(0, 3336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__97);
  __Pyx_GIVEREF(__pyx_tuple__97);

  /* "Aspidites/_vendor/pyparsing_extension.py":3460
 * 
 *         if min < 1:
 *             raise ValueError("cannot specify a minimum length < 1; use "             # <<<<<<<<<<<<<<
 *                              "Optional(CharsNotIn()) if zero-length char group is permitted")
 * 
 */
  __pyx_tuple__104 = PyTuple_Pack(1, __pyx_kp_s_cannot_specify_a_minimum_length_2); if (unlikely(!__pyx_tuple__104)) __PYX_ERR(0, 3460, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__104);
  __Pyx_GIVEREF(__pyx_tuple__104);

  /* "Aspidites/_vendor/pyparsing_extension.py":3656
 *     def __init__(self):
 *         super(LineEnd, self).__init__()
 *         self.setWhitespaceChars(ParserElement.DEFAULT_WHITE_CHARS.replace("\n", ""))             # <<<<<<<<<<<<<<
 *         self.errmsg = "Expected end of line"
 * 
 */
  __pyx_tuple__106 = PyTuple_Pack(2, __pyx_kp_s__4, __pyx_kp_s__12); if (unlikely(!__pyx_tuple__106)) __PYX_ERR(0, 3656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__106);
  __Pyx_GIVEREF(__pyx_tuple__106);

  /* "Aspidites/_vendor/pyparsing_extension.py":3788
 *                         tmp.append(SkipTo(skipto_arg)("_skipped*"))
 *                     else:
 *                         raise Exception("cannot construct And with sequence ending in ...")             # <<<<<<<<<<<<<<
 *                 else:
 *                     tmp.append(expr)
 */
  __pyx_tuple__108 = PyTuple_Pack(1, __pyx_kp_s_cannot_construct_And_with_sequen); if (unlikely(!__pyx_tuple__108)) __PYX_ERR(0, 3788, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__108);
  __Pyx_GIVEREF(__pyx_tuple__108);

  /* "Aspidites/_vendor/pyparsing_extension.py":3926
 *                 return best_expr._parse(instring, loc, doActions)
 * 
 *             longest = -1, None             # <<<<<<<<<<<<<<
 *             for loc1, expr1 in matches:
 *                 if loc1 <= longest[0]:
 */
  __pyx_tuple__111 = PyTuple_Pack(2, __pyx_int_neg_1, Py_None); if (unlikely(!__pyx_tuple__111)) __PYX_ERR(0, 3926, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__111);
  __Pyx_GIVEREF(__pyx_tuple__111);

  /* "Aspidites/_vendor/pyparsing_extension.py":4853
 *         try:
 *             if self.expr is not None:
 *                 retString = str(self.expr)[:1000]             # <<<<<<<<<<<<<<
 *             else:
 *                 retString = "None"
 */
  __pyx_slice__119 = PySlice_New(Py_None, __pyx_int_1000, Py_None); if (unlikely(!__pyx_slice__119)) __PYX_ERR(0, 4853, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__119);
  __Pyx_GIVEREF(__pyx_slice__119);

  /* "Aspidites/_vendor/pyparsing_extension.py":70
 * 
 * 
 * def _enable_all_warnings():             # <<<<<<<<<<<<<<
 *     __diag__.warn_multiple_tokens_in_named_alternation = True
 *     __diag__.warn_ungrouped_named_tokens_in_collection = True
 */
  __pyx_codeobj__121 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_enable_all_warnings, 70, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__121)) __PYX_ERR(0, 70, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":78
 * 
 * 
 * def _defaultStartDebugAction(instring, loc, expr):             # <<<<<<<<<<<<<<
 *     print(("Match " + str(expr) + " at loc " + str(loc) + "(%d,%d)" % (lineno(loc, instring), col(loc, instring))))
 * 
 */
  __pyx_tuple__122 = PyTuple_Pack(3, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_expr); if (unlikely(!__pyx_tuple__122)) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__122);
  __Pyx_GIVEREF(__pyx_tuple__122);
  __pyx_codeobj__123 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__122, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_defaultStartDebugAction, 78, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__123)) __PYX_ERR(0, 78, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":82
 * 
 * 
 * def _defaultSuccessDebugAction(instring, startloc, endloc, expr, toks):             # <<<<<<<<<<<<<<
 *     print("Matched " + str(expr) + " -> " + str(toks.asList()))
 * 
 */
  __pyx_tuple__124 = PyTuple_Pack(5, __pyx_n_s_instring, __pyx_n_s_startloc, __pyx_n_s_endloc, __pyx_n_s_expr, __pyx_n_s_toks); if (unlikely(!__pyx_tuple__124)) __PYX_ERR(0, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__124);
  __Pyx_GIVEREF(__pyx_tuple__124);
  __pyx_codeobj__125 = (PyObject*)__Pyx_PyCode_New(5, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__124, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_defaultSuccessDebugAction, 82, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__125)) __PYX_ERR(0, 82, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":86
 * 
 * 
 * def _defaultExceptionDebugAction(instring, loc, expr, exc):             # <<<<<<<<<<<<<<
 *     print("Exception raised:" + str(exc))
 * 
 */
  __pyx_tuple__126 = PyTuple_Pack(4, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_expr, __pyx_n_s_exc); if (unlikely(!__pyx_tuple__126)) __PYX_ERR(0, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__126);
  __Pyx_GIVEREF(__pyx_tuple__126);
  __pyx_codeobj__127 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__126, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_defaultExceptionDebugAction, 86, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__127)) __PYX_ERR(0, 86, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":90
 * 
 * 
 * def nullDebugAction(*args):             # <<<<<<<<<<<<<<
 *     """'Do-nothing' debug action, to suppress debugging output during parsing."""
 *     pass
 */
  __pyx_tuple__128 = PyTuple_Pack(1, __pyx_n_s_args); if (unlikely(!__pyx_tuple__128)) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__128);
  __Pyx_GIVEREF(__pyx_tuple__128);
  __pyx_codeobj__129 = (PyObject*)__Pyx_PyCode_New(0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__128, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_nullDebugAction, 90, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__129)) __PYX_ERR(0, 90, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":95
 * 
 * 
 * def _escapeRegexRangeChars(s):             # <<<<<<<<<<<<<<
 *     # ~  escape these chars: ^-[]
 *     for c in r"\^-[]":
 */
  __pyx_tuple__130 = PyTuple_Pack(2, __pyx_n_s_s_4, __pyx_n_s_c); if (unlikely(!__pyx_tuple__130)) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__130);
  __Pyx_GIVEREF(__pyx_tuple__130);
  __pyx_codeobj__131 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__130, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_escapeRegexRangeChars, 95, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__131)) __PYX_ERR(0, 95, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":104
 * 
 * 
 * def _flatten(L):             # <<<<<<<<<<<<<<
 *     ret = []
 *     for i in L:
 */
  __pyx_tuple__132 = PyTuple_Pack(3, __pyx_n_s_L, __pyx_n_s_ret, __pyx_n_s_i); if (unlikely(!__pyx_tuple__132)) __PYX_ERR(0, 104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__132);
  __Pyx_GIVEREF(__pyx_tuple__132);
  __pyx_codeobj__133 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_flatten, 104, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__133)) __PYX_ERR(0, 104, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":114
 * 
 * 
 * def _xml_escape(data):             # <<<<<<<<<<<<<<
 *     """Escape &, <, >, ", ', etc. in a string of data."""
 * 
 */
  __pyx_tuple__134 = PyTuple_Pack(7, __pyx_n_s_data, __pyx_n_s_from_symbols, __pyx_n_s_to_symbols, __pyx_n_s_from, __pyx_n_s_to, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__134)) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__134);
  __Pyx_GIVEREF(__pyx_tuple__134);
  __pyx_codeobj__135 = (PyObject*)__Pyx_PyCode_New(1, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__134, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_xml_escape, 114, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__135)) __PYX_ERR(0, 114, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":125
 * 
 * 
 * def col(loc, strg):             # <<<<<<<<<<<<<<
 *     """Returns current column within a string, counting newlines as line separators.
 *    The first column is number 1.
 */
  __pyx_tuple__136 = PyTuple_Pack(3, __pyx_n_s_loc, __pyx_n_s_strg, __pyx_n_s_s_4); if (unlikely(!__pyx_tuple__136)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__136);
  __Pyx_GIVEREF(__pyx_tuple__136);
  __pyx_codeobj__137 = (PyObject*)__Pyx_PyCode_New(2, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__136, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_col, 125, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__137)) __PYX_ERR(0, 125, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":140
 * 
 * 
 * def lineno(loc, strg):             # <<<<<<<<<<<<<<
 *     """Returns current line number within a string, counting newlines as line separators.
 *     The first line is number 1.
 */
  __pyx_tuple__138 = PyTuple_Pack(2, __pyx_n_s_loc, __pyx_n_s_strg); if (unlikely(!__pyx_tuple__138)) __PYX_ERR(0, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__138);
  __Pyx_GIVEREF(__pyx_tuple__138);
  __pyx_codeobj__139 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__138, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_lineno, 140, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__139)) __PYX_ERR(0, 140, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":153
 * 
 * 
 * def line(loc, strg):             # <<<<<<<<<<<<<<
 *     """Returns the line of text containing loc within a string, counting newlines as line separators.
 *        """
 */
  __pyx_tuple__140 = PyTuple_Pack(4, __pyx_n_s_loc, __pyx_n_s_strg, __pyx_n_s_lastCR, __pyx_n_s_nextCR); if (unlikely(!__pyx_tuple__140)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__140);
  __Pyx_GIVEREF(__pyx_tuple__140);
  __pyx_codeobj__141 = (PyObject*)__Pyx_PyCode_New(2, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__140, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_line, 153, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__141)) __PYX_ERR(0, 153, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":164
 * 
 * 
 * class _ParseResultsWithOffset(object):             # <<<<<<<<<<<<<<
 *     def __init__(self, p1, p2):
 *         self.tup = (p1, p2)
 */
  __pyx_tuple__142 = PyTuple_Pack(1, __pyx_builtin_object); if (unlikely(!__pyx_tuple__142)) __PYX_ERR(0, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__142);
  __Pyx_GIVEREF(__pyx_tuple__142);

  /* "Aspidites/_vendor/pyparsing_extension.py":165
 * 
 * class _ParseResultsWithOffset(object):
 *     def __init__(self, p1, p2):             # <<<<<<<<<<<<<<
 *         self.tup = (p1, p2)
 * 
 */
  __pyx_tuple__143 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_p1, __pyx_n_s_p2); if (unlikely(!__pyx_tuple__143)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__143);
  __Pyx_GIVEREF(__pyx_tuple__143);
  __pyx_codeobj__144 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__143, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 165, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__144)) __PYX_ERR(0, 165, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":168
 *         self.tup = (p1, p2)
 * 
 *     def __getitem__(self, i):             # <<<<<<<<<<<<<<
 *         return self.tup[i]
 * 
 */
  __pyx_tuple__145 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__145)) __PYX_ERR(0, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__145);
  __Pyx_GIVEREF(__pyx_tuple__145);
  __pyx_codeobj__146 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__145, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_getitem, 168, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__146)) __PYX_ERR(0, 168, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":171
 *         return self.tup[i]
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.tup[0])
 * 
 */
  __pyx_tuple__147 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__147)) __PYX_ERR(0, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__147);
  __Pyx_GIVEREF(__pyx_tuple__147);
  __pyx_codeobj__148 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__147, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_repr, 171, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__148)) __PYX_ERR(0, 171, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":174
 *         return repr(self.tup[0])
 * 
 *     def setOffset(self, i):             # <<<<<<<<<<<<<<
 *         self.tup = (self.tup[0], i)
 * 
 */
  __pyx_tuple__149 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_i); if (unlikely(!__pyx_tuple__149)) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__149);
  __Pyx_GIVEREF(__pyx_tuple__149);
  __pyx_codeobj__150 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__149, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_setOffset, 174, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__150)) __PYX_ERR(0, 174, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":178
 * 
 * 
 * class ParseResults(object):             # <<<<<<<<<<<<<<
 *     """Structured parse results, to provide multiple means of access to
 *     the parsed data:
 */
  __pyx_tuple__151 = PyTuple_Pack(1, __pyx_builtin_object); if (unlikely(!__pyx_tuple__151)) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__151);
  __Pyx_GIVEREF(__pyx_tuple__151);

  /* "Aspidites/_vendor/pyparsing_extension.py":222
 *     """
 * 
 *     def __new__(cls, toklist=None, name=None, asList=True, modal=True):             # <<<<<<<<<<<<<<
 *         if isinstance(toklist, cls):
 *             return toklist
 */
  __pyx_tuple__152 = PyTuple_Pack(6, __pyx_n_s_cls, __pyx_n_s_toklist, __pyx_n_s_name, __pyx_n_s_asList, __pyx_n_s_modal, __pyx_n_s_retobj); if (unlikely(!__pyx_tuple__152)) __PYX_ERR(0, 222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__152);
  __Pyx_GIVEREF(__pyx_tuple__152);
  __pyx_codeobj__153 = (PyObject*)__Pyx_PyCode_New(5, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__152, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_new, 222, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__153)) __PYX_ERR(0, 222, __pyx_L1_error)
  __pyx_tuple__154 = PyTuple_Pack(4, ((PyObject *)Py_None), ((PyObject *)Py_None), ((PyObject *)Py_True), ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__154)) __PYX_ERR(0, 222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__154);
  __Pyx_GIVEREF(__pyx_tuple__154);

  /* "Aspidites/_vendor/pyparsing_extension.py":231
 *     # Performance tuning: we construct a *lot* of these, so keep this
 *     # constructor as small and fast as possible
 *     def __init__(self, toklist=None, name=None, asList=True, modal=True, isinstance=isinstance):             # <<<<<<<<<<<<<<
 *         if self.__doinit:
 *             self.__doinit = False
 */
  __pyx_tuple__155 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_toklist, __pyx_n_s_name, __pyx_n_s_asList, __pyx_n_s_modal, __pyx_n_s_isinstance); if (unlikely(!__pyx_tuple__155)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__155);
  __Pyx_GIVEREF(__pyx_tuple__155);
  __pyx_codeobj__156 = (PyObject*)__Pyx_PyCode_New(6, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__155, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 231, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__156)) __PYX_ERR(0, 231, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":270
 *                         self[name] = toklist
 * 
 *     def __getitem__(self, i):             # <<<<<<<<<<<<<<
 *         if isinstance(i, (int, slice)):
 *             return self.__toklist[i]
 */
  __pyx_tuple__157 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_i, __pyx_n_s_v); if (unlikely(!__pyx_tuple__157)) __PYX_ERR(0, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__157);
  __Pyx_GIVEREF(__pyx_tuple__157);
  __pyx_codeobj__158 = (PyObject*)__Pyx_PyCode_New(2, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__157, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_getitem, 270, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__158)) __PYX_ERR(0, 270, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":279
 *                 return ParseResults([v[0] for v in self.__tokdict[i]])
 * 
 *     def __setitem__(self, k, v, isinstance=isinstance):             # <<<<<<<<<<<<<<
 *         if isinstance(v, _ParseResultsWithOffset):
 *             self.__tokdict[k] = self.__tokdict.get(k, list()) + [v]
 */
  __pyx_tuple__159 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_k, __pyx_n_s_v, __pyx_n_s_isinstance, __pyx_n_s_sub); if (unlikely(!__pyx_tuple__159)) __PYX_ERR(0, 279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__159);
  __Pyx_GIVEREF(__pyx_tuple__159);
  __pyx_codeobj__160 = (PyObject*)__Pyx_PyCode_New(4, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__159, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_setitem, 279, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__160)) __PYX_ERR(0, 279, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":292
 *             sub.__parent = wkref(self)
 * 
 *     def __delitem__(self, i):             # <<<<<<<<<<<<<<
 *         if isinstance(i, (int, slice)):
 *             mylen = len(self.__toklist)
 */
  __pyx_tuple__161 = PyTuple_Pack(10, __pyx_n_s_self, __pyx_n_s_i, __pyx_n_s_mylen, __pyx_n_s_removed, __pyx_n_s_name, __pyx_n_s_occurrences, __pyx_n_s_j, __pyx_n_s_k, __pyx_n_s_value, __pyx_n_s_position); if (unlikely(!__pyx_tuple__161)) __PYX_ERR(0, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__161);
  __Pyx_GIVEREF(__pyx_tuple__161);
  __pyx_codeobj__162 = (PyObject*)__Pyx_PyCode_New(2, 0, 10, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__161, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_delitem, 292, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__162)) __PYX_ERR(0, 292, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":313
 *             del self.__tokdict[i]
 * 
 *     def __contains__(self, k):             # <<<<<<<<<<<<<<
 *         return k in self.__tokdict
 * 
 */
  __pyx_tuple__163 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_k); if (unlikely(!__pyx_tuple__163)) __PYX_ERR(0, 313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__163);
  __Pyx_GIVEREF(__pyx_tuple__163);
  __pyx_codeobj__164 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__163, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_contains, 313, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__164)) __PYX_ERR(0, 313, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":316
 *         return k in self.__tokdict
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return len(self.__toklist)
 * 
 */
  __pyx_tuple__165 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__165)) __PYX_ERR(0, 316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__165);
  __Pyx_GIVEREF(__pyx_tuple__165);
  __pyx_codeobj__166 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__165, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_len, 316, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__166)) __PYX_ERR(0, 316, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":319
 *         return len(self.__toklist)
 * 
 *     def __bool__(self):             # <<<<<<<<<<<<<<
 *         return (not not self.__toklist)
 * 
 */
  __pyx_tuple__167 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__167)) __PYX_ERR(0, 319, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__167);
  __Pyx_GIVEREF(__pyx_tuple__167);
  __pyx_codeobj__168 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_bool, 319, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__168)) __PYX_ERR(0, 319, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":324
 *     __nonzero__ = __bool__
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return iter(self.__toklist)
 * 
 */
  __pyx_tuple__169 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__169)) __PYX_ERR(0, 324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__169);
  __Pyx_GIVEREF(__pyx_tuple__169);
  __pyx_codeobj__170 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_iter, 324, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__170)) __PYX_ERR(0, 324, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":327
 *         return iter(self.__toklist)
 * 
 *     def __reversed__(self):             # <<<<<<<<<<<<<<
 *         return iter(self.__toklist[::-1])
 * 
 */
  __pyx_tuple__171 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__171)) __PYX_ERR(0, 327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__171);
  __Pyx_GIVEREF(__pyx_tuple__171);
  __pyx_codeobj__172 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__171, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_reversed, 327, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__172)) __PYX_ERR(0, 327, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":330
 *         return iter(self.__toklist[::-1])
 * 
 *     def _iterkeys(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self.__tokdict, "iterkeys"):
 *             return self.__tokdict.iterkeys()
 */
  __pyx_tuple__173 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__173)) __PYX_ERR(0, 330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__173);
  __Pyx_GIVEREF(__pyx_tuple__173);
  __pyx_codeobj__174 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__173, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_iterkeys_2, 330, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__174)) __PYX_ERR(0, 330, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":336
 *             return iter(self.__tokdict)
 * 
 *     def _itervalues(self):             # <<<<<<<<<<<<<<
 *         return (self[k] for k in self._iterkeys())
 * 
 */
  __pyx_tuple__175 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__175)) __PYX_ERR(0, 336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__175);
  __Pyx_GIVEREF(__pyx_tuple__175);
  __pyx_codeobj__176 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__175, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_itervalues_2, 336, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__176)) __PYX_ERR(0, 336, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":339
 *         return (self[k] for k in self._iterkeys())
 * 
 *     def _iteritems(self):             # <<<<<<<<<<<<<<
 *         return ((k, self[k]) for k in self._iterkeys())
 * 
 */
  __pyx_tuple__177 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__177)) __PYX_ERR(0, 339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__177);
  __Pyx_GIVEREF(__pyx_tuple__177);
  __pyx_codeobj__178 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__177, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_iteritems_2, 339, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__178)) __PYX_ERR(0, 339, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":362
 *         """Returns an iterator of all named result key-value tuples (Python 2.x only)."""
 * 
 *         def keys(self):             # <<<<<<<<<<<<<<
 *             """Returns all named result keys (as a list in Python 2.x, as an iterator in Python 3.x)."""
 *             return list(self.iterkeys())
 */
  __pyx_tuple__179 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__179)) __PYX_ERR(0, 362, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__179);
  __Pyx_GIVEREF(__pyx_tuple__179);
  __pyx_codeobj__180 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__179, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_keys, 362, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__180)) __PYX_ERR(0, 362, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":366
 *             return list(self.iterkeys())
 * 
 *         def values(self):             # <<<<<<<<<<<<<<
 *             """Returns all named result values (as a list in Python 2.x, as an iterator in Python 3.x)."""
 *             return list(self.itervalues())
 */
  __pyx_tuple__181 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__181)) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__181);
  __Pyx_GIVEREF(__pyx_tuple__181);
  __pyx_codeobj__182 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__181, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_values, 366, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__182)) __PYX_ERR(0, 366, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":370
 *             return list(self.itervalues())
 * 
 *         def items(self):             # <<<<<<<<<<<<<<
 *             """Returns all named result key-values (as a list of tuples in Python 2.x, as an iterator in Python 3.x)."""
 *             return list(self.iteritems())
 */
  __pyx_tuple__183 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__183)) __PYX_ERR(0, 370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__183);
  __Pyx_GIVEREF(__pyx_tuple__183);
  __pyx_codeobj__184 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__183, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_items, 370, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__184)) __PYX_ERR(0, 370, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":374
 *             return list(self.iteritems())
 * 
 *     def haskeys(self):             # <<<<<<<<<<<<<<
 *         """Since keys() returns an iterator, this method is helpful in bypassing
 *            code that looks for the existence of any defined results names."""
 */
  __pyx_tuple__185 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__185)) __PYX_ERR(0, 374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__185);
  __Pyx_GIVEREF(__pyx_tuple__185);
  __pyx_codeobj__186 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__185, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_haskeys, 374, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__186)) __PYX_ERR(0, 374, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":379
 *         return bool(self.__tokdict)
 * 
 *     def pop(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         """
 *         Removes and returns item at specified index (default= ``last``).
 */
  __pyx_tuple__187 = PyTuple_Pack(8, __pyx_n_s_self, __pyx_n_s_args, __pyx_n_s_kwargs, __pyx_n_s_k, __pyx_n_s_v, __pyx_n_s_index, __pyx_n_s_ret, __pyx_n_s_defaultvalue); if (unlikely(!__pyx_tuple__187)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__187);
  __Pyx_GIVEREF(__pyx_tuple__187);
  __pyx_codeobj__188 = (PyObject*)__Pyx_PyCode_New(1, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__187, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_pop, 379, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__188)) __PYX_ERR(0, 379, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":434
 *             return defaultvalue
 * 
 *     def get(self, key, defaultValue=None):             # <<<<<<<<<<<<<<
 *         """
 *         Returns named result matching the given key, or if there is no
 */
  __pyx_tuple__189 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_key, __pyx_n_s_defaultValue); if (unlikely(!__pyx_tuple__189)) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__189);
  __Pyx_GIVEREF(__pyx_tuple__189);
  __pyx_codeobj__190 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__189, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_get, 434, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__190)) __PYX_ERR(0, 434, __pyx_L1_error)
  __pyx_tuple__191 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__191)) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__191);
  __Pyx_GIVEREF(__pyx_tuple__191);

  /* "Aspidites/_vendor/pyparsing_extension.py":457
 *             return defaultValue
 * 
 *     def insert(self, index, insStr):             # <<<<<<<<<<<<<<
 *         """
 *         Inserts new element at location index in the list of parsed tokens.
 */
  __pyx_tuple__192 = PyTuple_Pack(8, __pyx_n_s_self, __pyx_n_s_index, __pyx_n_s_insStr, __pyx_n_s_name, __pyx_n_s_occurrences, __pyx_n_s_k, __pyx_n_s_value, __pyx_n_s_position); if (unlikely(!__pyx_tuple__192)) __PYX_ERR(0, 457, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__192);
  __Pyx_GIVEREF(__pyx_tuple__192);
  __pyx_codeobj__193 = (PyObject*)__Pyx_PyCode_New(3, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__192, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_insert, 457, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__193)) __PYX_ERR(0, 457, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":478
 *                 occurrences[k] = _ParseResultsWithOffset(value, position + (position > index))
 * 
 *     def append(self, item):             # <<<<<<<<<<<<<<
 *         """
 *         Add single element to end of ParseResults list of elements.
 */
  __pyx_tuple__194 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_item); if (unlikely(!__pyx_tuple__194)) __PYX_ERR(0, 478, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__194);
  __Pyx_GIVEREF(__pyx_tuple__194);
  __pyx_codeobj__195 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__194, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_append, 478, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__195)) __PYX_ERR(0, 478, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":493
 *         self.__toklist.append(item)
 * 
 *     def extend(self, itemseq):             # <<<<<<<<<<<<<<
 *         """
 *         Add sequence of elements to end of ParseResults list of elements.
 */
  __pyx_tuple__196 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_itemseq); if (unlikely(!__pyx_tuple__196)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__196);
  __Pyx_GIVEREF(__pyx_tuple__196);
  __pyx_codeobj__197 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__196, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_extend, 493, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__197)) __PYX_ERR(0, 493, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":512
 *             self.__toklist.extend(itemseq)
 * 
 *     def clear(self):             # <<<<<<<<<<<<<<
 *         """
 *         Clear all elements and results names.
 */
  __pyx_tuple__198 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__198)) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__198);
  __Pyx_GIVEREF(__pyx_tuple__198);
  __pyx_codeobj__199 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__198, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_clear, 512, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__199)) __PYX_ERR(0, 512, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":519
 *         self.__tokdict.clear()
 * 
 *     def __getattr__(self, name):             # <<<<<<<<<<<<<<
 *         try:
 *             return self[name]
 */
  __pyx_tuple__200 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_name); if (unlikely(!__pyx_tuple__200)) __PYX_ERR(0, 519, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__200);
  __Pyx_GIVEREF(__pyx_tuple__200);
  __pyx_codeobj__201 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__200, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_getattr, 519, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__201)) __PYX_ERR(0, 519, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":525
 *             return ""
 * 
 *     def __add__(self, other):             # <<<<<<<<<<<<<<
 *         ret = self.copy()
 *         ret += other
 */
  __pyx_tuple__202 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_other, __pyx_n_s_ret); if (unlikely(!__pyx_tuple__202)) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__202);
  __Pyx_GIVEREF(__pyx_tuple__202);
  __pyx_codeobj__203 = (PyObject*)__Pyx_PyCode_New(2, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__202, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_add, 525, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__203)) __PYX_ERR(0, 525, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":530
 *         return ret
 * 
 *     def __iadd__(self, other):             # <<<<<<<<<<<<<<
 *         if other.__tokdict:
 *             offset = len(self.__toklist)
 */
  __pyx_tuple__204 = PyTuple_Pack(9, __pyx_n_s_self, __pyx_n_s_other, __pyx_n_s_offset, __pyx_n_s_addoffset, __pyx_n_s_otheritems, __pyx_n_s_otherdictitems, __pyx_n_s_k, __pyx_n_s_v, __pyx_n_s_vlist); if (unlikely(!__pyx_tuple__204)) __PYX_ERR(0, 530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__204);
  __Pyx_GIVEREF(__pyx_tuple__204);
  __pyx_codeobj__205 = (PyObject*)__Pyx_PyCode_New(2, 0, 9, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__204, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_iadd, 530, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__205)) __PYX_ERR(0, 530, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":546
 *         return self
 * 
 *     def __radd__(self, other):             # <<<<<<<<<<<<<<
 *         if isinstance(other, int) and other == 0:
 *             # useful for merging many ParseResults using sum() builtin
 */
  __pyx_tuple__206 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_other); if (unlikely(!__pyx_tuple__206)) __PYX_ERR(0, 546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__206);
  __Pyx_GIVEREF(__pyx_tuple__206);
  __pyx_codeobj__207 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__206, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_radd, 546, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__207)) __PYX_ERR(0, 546, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":554
 *             return other + self
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "(%s, %s)" % (repr(self.__toklist), repr(self.__tokdict))
 * 
 */
  __pyx_tuple__208 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__208)) __PYX_ERR(0, 554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__208);
  __Pyx_GIVEREF(__pyx_tuple__208);
  __pyx_codeobj__209 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__208, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_repr, 554, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__209)) __PYX_ERR(0, 554, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":557
 *         return "(%s, %s)" % (repr(self.__toklist), repr(self.__tokdict))
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return '[' + ', '.join(str(i) if isinstance(i, ParseResults) else repr(i) for i in self.__toklist) + ']'
 * 
 */
  __pyx_tuple__210 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__210)) __PYX_ERR(0, 557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__210);
  __Pyx_GIVEREF(__pyx_tuple__210);
  __pyx_codeobj__211 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__210, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_str, 557, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__211)) __PYX_ERR(0, 557, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":560
 *         return '[' + ', '.join(str(i) if isinstance(i, ParseResults) else repr(i) for i in self.__toklist) + ']'
 * 
 *     def _asStringList(self, sep=''):             # <<<<<<<<<<<<<<
 *         out = []
 *         for item in self.__toklist:
 */
  __pyx_tuple__212 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_sep, __pyx_n_s_out, __pyx_n_s_item); if (unlikely(!__pyx_tuple__212)) __PYX_ERR(0, 560, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__212);
  __Pyx_GIVEREF(__pyx_tuple__212);
  __pyx_codeobj__213 = (PyObject*)__Pyx_PyCode_New(2, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__212, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_asStringList, 560, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__213)) __PYX_ERR(0, 560, __pyx_L1_error)
  __pyx_tuple__214 = PyTuple_Pack(1, ((PyObject*)__pyx_kp_s__12)); if (unlikely(!__pyx_tuple__214)) __PYX_ERR(0, 560, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__214);
  __Pyx_GIVEREF(__pyx_tuple__214);

  /* "Aspidites/_vendor/pyparsing_extension.py":571
 *         return out
 * 
 *     def asList(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns the parse results as a nested list of matching tokens, all converted to strings.
 */
  __pyx_tuple__215 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_res); if (unlikely(!__pyx_tuple__215)) __PYX_ERR(0, 571, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__215);
  __Pyx_GIVEREF(__pyx_tuple__215);
  __pyx_codeobj__216 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__215, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_asList, 571, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__216)) __PYX_ERR(0, 571, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":588
 *         return [res.asList() if isinstance(res, ParseResults) else res for res in self.__toklist]
 * 
 *     def asDict(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns the named parse results as a nested dictionary.
 */
  __pyx_tuple__217 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_item_fn, __pyx_n_s_toItem, __pyx_n_s_toItem, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__217)) __PYX_ERR(0, 588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__217);
  __Pyx_GIVEREF(__pyx_tuple__217);
  __pyx_codeobj__218 = (PyObject*)__Pyx_PyCode_New(1, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__217, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_asDict, 588, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__218)) __PYX_ERR(0, 588, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":624
 *         return dict((k, toItem(v)) for k, v in item_fn())
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns a new copy of a :class:`ParseResults` object.
 */
  __pyx_tuple__219 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ret); if (unlikely(!__pyx_tuple__219)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__219);
  __Pyx_GIVEREF(__pyx_tuple__219);
  __pyx_codeobj__220 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__219, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_copy, 624, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__220)) __PYX_ERR(0, 624, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":635
 *         return ret
 * 
 *     def asXML(self, doctag=None, namedItemsOnly=False, indent="", formatted=True):             # <<<<<<<<<<<<<<
 *         """
 *         (Deprecated) Returns the parse results as XML. Tags are created for tokens and lists that have defined results names.
 */
  __pyx_tuple__221 = PyTuple_Pack(16, __pyx_n_s_self, __pyx_n_s_doctag, __pyx_n_s_namedItemsOnly, __pyx_n_s_indent, __pyx_n_s_formatted, __pyx_n_s_nl, __pyx_n_s_out, __pyx_n_s_namedItems, __pyx_n_s_nextLevelIndent, __pyx_n_s_selfTag, __pyx_n_s_i, __pyx_n_s_res, __pyx_n_s_resTag, __pyx_n_s_xmlBodyText, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__221)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__221);
  __Pyx_GIVEREF(__pyx_tuple__221);
  __pyx_codeobj__222 = (PyObject*)__Pyx_PyCode_New(5, 0, 16, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__221, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_asXML, 635, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__222)) __PYX_ERR(0, 635, __pyx_L1_error)
  __pyx_tuple__223 = PyTuple_Pack(4, ((PyObject *)Py_None), ((PyObject *)Py_False), ((PyObject*)__pyx_kp_s__12), ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__223)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__223);
  __Pyx_GIVEREF(__pyx_tuple__223);

  /* "Aspidites/_vendor/pyparsing_extension.py":696
 *         return "".join(out)
 * 
 *     def __lookup(self, sub):             # <<<<<<<<<<<<<<
 *         for k, vlist in self.__tokdict.items():
 *             for v, loc in vlist:
 */
  __pyx_tuple__224 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_sub, __pyx_n_s_k, __pyx_n_s_vlist, __pyx_n_s_v, __pyx_n_s_loc); if (unlikely(!__pyx_tuple__224)) __PYX_ERR(0, 696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__224);
  __Pyx_GIVEREF(__pyx_tuple__224);
  __pyx_codeobj__225 = (PyObject*)__Pyx_PyCode_New(2, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__224, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_lookup, 696, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__225)) __PYX_ERR(0, 696, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":703
 *         return None
 * 
 *     def getName(self):             # <<<<<<<<<<<<<<
 *         r"""
 *         Returns the results name for this token expression. Useful when several
 */
  __pyx_tuple__226 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_par); if (unlikely(!__pyx_tuple__226)) __PYX_ERR(0, 703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__226);
  __Pyx_GIVEREF(__pyx_tuple__226);
  __pyx_codeobj__227 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__226, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_getName, 703, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__227)) __PYX_ERR(0, 703, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":743
 *             return None
 * 
 *     def dump(self, indent='', full=True, include_list=True, _depth=0):             # <<<<<<<<<<<<<<
 *         """
 *         Diagnostic method for listing out the contents of
 */
  __pyx_tuple__228 = PyTuple_Pack(15, __pyx_n_s_self, __pyx_n_s_indent, __pyx_n_s_full, __pyx_n_s_include_list, __pyx_n_s_depth, __pyx_n_s_out, __pyx_n_s_NL, __pyx_n_s_items, __pyx_n_s_k, __pyx_n_s_v, __pyx_n_s_i, __pyx_n_s_vv, __pyx_n_s_genexpr, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__228)) __PYX_ERR(0, 743, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__228);
  __Pyx_GIVEREF(__pyx_tuple__228);
  __pyx_codeobj__229 = (PyObject*)__Pyx_PyCode_New(5, 0, 15, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__228, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_dump, 743, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__229)) __PYX_ERR(0, 743, __pyx_L1_error)
  __pyx_tuple__230 = PyTuple_Pack(4, ((PyObject*)__pyx_kp_s__12), ((PyObject *)Py_True), ((PyObject *)Py_True), ((PyObject *)__pyx_int_0)); if (unlikely(!__pyx_tuple__230)) __PYX_ERR(0, 743, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__230);
  __Pyx_GIVEREF(__pyx_tuple__230);

  /* "Aspidites/_vendor/pyparsing_extension.py":808
 *         return "".join(out)
 * 
 *     def pprint(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         """
 *         Pretty-printer for parsed results as a list, using the
 */
  __pyx_tuple__231 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_args, __pyx_n_s_kwargs); if (unlikely(!__pyx_tuple__231)) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__231);
  __Pyx_GIVEREF(__pyx_tuple__231);
  __pyx_codeobj__232 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__231, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_pprint, 808, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__232)) __PYX_ERR(0, 808, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":836
 * 
 *     # add support for pickle protocol
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return (self.__toklist,
 *                 (self.__tokdict.copy(),
 */
  __pyx_tuple__233 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__233)) __PYX_ERR(0, 836, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__233);
  __Pyx_GIVEREF(__pyx_tuple__233);
  __pyx_codeobj__234 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__233, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_getstate, 836, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__234)) __PYX_ERR(0, 836, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":843
 *                  self.__name))
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self.__toklist = state[0]
 *         self.__tokdict, par, inAccumNames, self.__name = state[1]
 */
  __pyx_tuple__235 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_state, __pyx_n_s_par, __pyx_n_s_inAccumNames); if (unlikely(!__pyx_tuple__235)) __PYX_ERR(0, 843, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__235);
  __Pyx_GIVEREF(__pyx_tuple__235);
  __pyx_codeobj__236 = (PyObject*)__Pyx_PyCode_New(2, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__235, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_setstate, 843, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__236)) __PYX_ERR(0, 843, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":853
 *             self.__parent = None
 * 
 *     def __getnewargs__(self):             # <<<<<<<<<<<<<<
 *         return self.__toklist, self.__name, self.__asList, self.__modal
 * 
 */
  __pyx_tuple__237 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__237)) __PYX_ERR(0, 853, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__237);
  __Pyx_GIVEREF(__pyx_tuple__237);
  __pyx_codeobj__238 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__237, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_getnewargs, 853, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__238)) __PYX_ERR(0, 853, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":856
 *         return self.__toklist, self.__name, self.__asList, self.__modal
 * 
 *     def __dir__(self):             # <<<<<<<<<<<<<<
 *         return dir(type(self)) + list(self.keys())
 * 
 */
  __pyx_tuple__239 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__239)) __PYX_ERR(0, 856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__239);
  __Pyx_GIVEREF(__pyx_tuple__239);
  __pyx_codeobj__240 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__239, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_dir, 856, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__240)) __PYX_ERR(0, 856, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":860
 * 
 *     @classmethod
 *     def from_dict(cls, other, name=None):             # <<<<<<<<<<<<<<
 *         """
 *         Helper classmethod to construct a ParseResults from a dict, preserving the
 */
  __pyx_tuple__241 = PyTuple_Pack(8, __pyx_n_s_cls, __pyx_n_s_other, __pyx_n_s_name, __pyx_n_s_is_iterable, __pyx_n_s_is_iterable, __pyx_n_s_ret, __pyx_n_s_k, __pyx_n_s_v); if (unlikely(!__pyx_tuple__241)) __PYX_ERR(0, 860, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__241);
  __Pyx_GIVEREF(__pyx_tuple__241);
  __pyx_codeobj__242 = (PyObject*)__Pyx_PyCode_New(3, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__241, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_from_dict, 860, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__242)) __PYX_ERR(0, 860, __pyx_L1_error)
  __pyx_tuple__243 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__243)) __PYX_ERR(0, 860, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__243);
  __Pyx_GIVEREF(__pyx_tuple__243);

  /* "Aspidites/_vendor/pyparsing_extension.py":894
 *     # Performance tuning: we construct a *lot* of these, so keep this
 *     # constructor as small and fast as possible
 *     def __init__(self, pstr, loc=0, msg=None, elem=None):             # <<<<<<<<<<<<<<
 *         self.loc = loc
 *         if msg is None:
 */
  __pyx_tuple__244 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_pstr, __pyx_n_s_loc, __pyx_n_s_msg, __pyx_n_s_elem); if (unlikely(!__pyx_tuple__244)) __PYX_ERR(0, 894, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__244);
  __Pyx_GIVEREF(__pyx_tuple__244);
  __pyx_codeobj__245 = (PyObject*)__Pyx_PyCode_New(5, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__244, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 894, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__245)) __PYX_ERR(0, 894, __pyx_L1_error)
  __pyx_tuple__246 = PyTuple_Pack(3, ((PyObject *)__pyx_int_0), ((PyObject *)Py_None), ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__246)) __PYX_ERR(0, 894, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__246);
  __Pyx_GIVEREF(__pyx_tuple__246);

  /* "Aspidites/_vendor/pyparsing_extension.py":906
 * 
 *     @classmethod
 *     def _from_exception(cls, pe):             # <<<<<<<<<<<<<<
 *         """
 *         internal factory method to simplify creating one type of ParseException
 */
  __pyx_tuple__247 = PyTuple_Pack(2, __pyx_n_s_cls, __pyx_n_s_pe); if (unlikely(!__pyx_tuple__247)) __PYX_ERR(0, 906, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__247);
  __Pyx_GIVEREF(__pyx_tuple__247);
  __pyx_codeobj__248 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__247, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_from_exception, 906, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__248)) __PYX_ERR(0, 906, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":913
 *         return cls(pe.pstr, pe.loc, pe.msg, pe.parserElement)
 * 
 *     def __getattr__(self, aname):             # <<<<<<<<<<<<<<
 *         """supported attributes by name are:
 *            - lineno - returns the line number of the exception text
 */
  __pyx_tuple__249 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_aname); if (unlikely(!__pyx_tuple__249)) __PYX_ERR(0, 913, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__249);
  __Pyx_GIVEREF(__pyx_tuple__249);
  __pyx_codeobj__250 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__249, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_getattr, 913, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__250)) __PYX_ERR(0, 913, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":928
 *             raise AttributeError(aname)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if self.pstr:
 *             if self.loc >= len(self.pstr):
 */
  __pyx_tuple__251 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_foundstr); if (unlikely(!__pyx_tuple__251)) __PYX_ERR(0, 928, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__251);
  __Pyx_GIVEREF(__pyx_tuple__251);
  __pyx_codeobj__252 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__251, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_str, 928, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__252)) __PYX_ERR(0, 928, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":939
 *                 (self.msg, foundstr, self.loc, self.lineno, self.column))
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return str(self)
 * 
 */
  __pyx_tuple__253 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__253)) __PYX_ERR(0, 939, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__253);
  __Pyx_GIVEREF(__pyx_tuple__253);
  __pyx_codeobj__254 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__253, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_repr, 939, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__254)) __PYX_ERR(0, 939, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":942
 *         return str(self)
 * 
 *     def markInputline(self, markerString=">!<"):             # <<<<<<<<<<<<<<
 *         """Extracts the exception line from the input string, and marks
 *            the location of the exception with a special symbol.
 */
  __pyx_tuple__255 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_markerString, __pyx_n_s_line_str, __pyx_n_s_line_column); if (unlikely(!__pyx_tuple__255)) __PYX_ERR(0, 942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__255);
  __Pyx_GIVEREF(__pyx_tuple__255);
  __pyx_codeobj__256 = (PyObject*)__Pyx_PyCode_New(2, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__255, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_markInputline, 942, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__256)) __PYX_ERR(0, 942, __pyx_L1_error)
  __pyx_tuple__257 = PyTuple_Pack(1, ((PyObject*)__pyx_kp_s__28)); if (unlikely(!__pyx_tuple__257)) __PYX_ERR(0, 942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__257);
  __Pyx_GIVEREF(__pyx_tuple__257);

  /* "Aspidites/_vendor/pyparsing_extension.py":953
 *         return line_str.strip()
 * 
 *     def __dir__(self):             # <<<<<<<<<<<<<<
 *         return "lineno col line".split() + dir(type(self))
 * 
 */
  __pyx_tuple__258 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__258)) __PYX_ERR(0, 953, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__258);
  __Pyx_GIVEREF(__pyx_tuple__258);
  __pyx_codeobj__259 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__258, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_dir, 953, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__259)) __PYX_ERR(0, 953, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":981
 * 
 *     @staticmethod
 *     def explain(exc, depth=16):             # <<<<<<<<<<<<<<
 *         """
 *         Method to take an exception and translate the Python internal traceback into a list
 */
  __pyx_tuple__260 = PyTuple_Pack(12, __pyx_n_s_exc, __pyx_n_s_depth_2, __pyx_n_s_inspect, __pyx_n_s_ret, __pyx_n_s_callers, __pyx_n_s_seen, __pyx_n_s_i, __pyx_n_s_ff, __pyx_n_s_frm, __pyx_n_s_f_self, __pyx_n_s_self_type, __pyx_n_s_code); if (unlikely(!__pyx_tuple__260)) __PYX_ERR(0, 981, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__260);
  __Pyx_GIVEREF(__pyx_tuple__260);
  __pyx_codeobj__261 = (PyObject*)__Pyx_PyCode_New(2, 0, 12, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__260, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_explain, 981, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__261)) __PYX_ERR(0, 981, __pyx_L1_error)
  __pyx_tuple__262 = PyTuple_Pack(1, ((PyObject *)__pyx_int_16)); if (unlikely(!__pyx_tuple__262)) __PYX_ERR(0, 981, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__262);
  __Pyx_GIVEREF(__pyx_tuple__262);

  /* "Aspidites/_vendor/pyparsing_extension.py":1083
 *     """
 * 
 *     def __init__(self, parseElementList):             # <<<<<<<<<<<<<<
 *         self.parseElementTrace = parseElementList
 * 
 */
  __pyx_tuple__263 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_parseElementList); if (unlikely(!__pyx_tuple__263)) __PYX_ERR(0, 1083, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__263);
  __Pyx_GIVEREF(__pyx_tuple__263);
  __pyx_codeobj__264 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__263, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 1083, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__264)) __PYX_ERR(0, 1083, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1086
 *         self.parseElementTrace = parseElementList
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "RecursiveGrammarException: %s" % self.parseElementTrace
 * 
 */
  __pyx_tuple__265 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__265)) __PYX_ERR(0, 1086, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__265);
  __Pyx_GIVEREF(__pyx_tuple__265);
  __pyx_codeobj__266 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__265, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_str, 1086, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__266)) __PYX_ERR(0, 1086, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1090
 * 
 * 
 * class ParserElement(object):             # <<<<<<<<<<<<<<
 *     """Abstract base level parser element class."""
 *     DEFAULT_WHITE_CHARS = " \n\t\r"
 */
  __pyx_tuple__267 = PyTuple_Pack(1, __pyx_builtin_object); if (unlikely(!__pyx_tuple__267)) __PYX_ERR(0, 1090, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__267);
  __Pyx_GIVEREF(__pyx_tuple__267);

  /* "Aspidites/_vendor/pyparsing_extension.py":1096
 * 
 *     @staticmethod
 *     def setDefaultWhitespaceChars(chars):             # <<<<<<<<<<<<<<
 *         r"""
 *         Overrides the default whitespace chars
 */
  __pyx_tuple__269 = PyTuple_Pack(1, __pyx_n_s_chars); if (unlikely(!__pyx_tuple__269)) __PYX_ERR(0, 1096, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__269);
  __Pyx_GIVEREF(__pyx_tuple__269);
  __pyx_codeobj__270 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__269, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_setDefaultWhitespaceChars, 1096, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__270)) __PYX_ERR(0, 1096, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1112
 * 
 *     @staticmethod
 *     def inlineLiteralsUsing(cls):             # <<<<<<<<<<<<<<
 *         """
 *         Set class to be used for inclusion of string literals into a parser.
 */
  __pyx_tuple__271 = PyTuple_Pack(1, __pyx_n_s_cls); if (unlikely(!__pyx_tuple__271)) __PYX_ERR(0, 1112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__271);
  __Pyx_GIVEREF(__pyx_tuple__271);
  __pyx_codeobj__272 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__271, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_inlineLiteralsUsing, 1112, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__272)) __PYX_ERR(0, 1112, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1134
 * 
 *     @classmethod
 *     def _trim_traceback(cls, tb):             # <<<<<<<<<<<<<<
 *         while tb.tb_next:
 *             tb = tb.tb_next
 */
  __pyx_tuple__273 = PyTuple_Pack(2, __pyx_n_s_cls, __pyx_n_s_tb); if (unlikely(!__pyx_tuple__273)) __PYX_ERR(0, 1134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__273);
  __Pyx_GIVEREF(__pyx_tuple__273);
  __pyx_codeobj__274 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__273, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_trim_traceback, 1134, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__274)) __PYX_ERR(0, 1134, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1139
 *         return tb
 * 
 *     def __init__(self, savelist=False):             # <<<<<<<<<<<<<<
 *         self.parseAction = list()
 *         self.failAction = None
 */
  __pyx_tuple__275 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_savelist); if (unlikely(!__pyx_tuple__275)) __PYX_ERR(0, 1139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__275);
  __Pyx_GIVEREF(__pyx_tuple__275);
  __pyx_codeobj__276 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__275, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 1139, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__276)) __PYX_ERR(0, 1139, __pyx_L1_error)
  __pyx_tuple__277 = PyTuple_Pack(1, ((PyObject *)Py_False)); if (unlikely(!__pyx_tuple__277)) __PYX_ERR(0, 1139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__277);
  __Pyx_GIVEREF(__pyx_tuple__277);

  /* "Aspidites/_vendor/pyparsing_extension.py":1162
 *         self.callDuringTry = False
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         """
 *         Make a copy of this :class:`ParserElement`.  Useful for defining
 */
  __pyx_tuple__278 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_cpy); if (unlikely(!__pyx_tuple__278)) __PYX_ERR(0, 1162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__278);
  __Pyx_GIVEREF(__pyx_tuple__278);
  __pyx_codeobj__279 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__278, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_copy, 1162, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__279)) __PYX_ERR(0, 1162, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1191
 *         return cpy
 * 
 *     def setName(self, name):             # <<<<<<<<<<<<<<
 *         """
 *         Define name for this expression, makes debugging and exception messages clearer.
 */
  __pyx_tuple__280 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_name); if (unlikely(!__pyx_tuple__280)) __PYX_ERR(0, 1191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__280);
  __Pyx_GIVEREF(__pyx_tuple__280);
  __pyx_codeobj__281 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__280, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_setName, 1191, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__281)) __PYX_ERR(0, 1191, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1206
 *         return self
 * 
 *     def setResultsName(self, name, listAllMatches=False):             # <<<<<<<<<<<<<<
 *         """
 *         Define name for referencing matching tokens as a nested attribute
 */
  __pyx_tuple__282 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_name, __pyx_n_s_listAllMatches); if (unlikely(!__pyx_tuple__282)) __PYX_ERR(0, 1206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__282);
  __Pyx_GIVEREF(__pyx_tuple__282);
  __pyx_codeobj__283 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__282, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_setResultsName_2, 1206, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__283)) __PYX_ERR(0, 1206, __pyx_L1_error)
  __pyx_tuple__284 = PyTuple_Pack(1, ((PyObject *)Py_False)); if (unlikely(!__pyx_tuple__284)) __PYX_ERR(0, 1206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__284);
  __Pyx_GIVEREF(__pyx_tuple__284);

  /* "Aspidites/_vendor/pyparsing_extension.py":1229
 *         return self._setResultsName(name, listAllMatches)
 * 
 *     def _setResultsName(self, name, listAllMatches=False):             # <<<<<<<<<<<<<<
 *         newself = self.copy()
 *         if name.endswith("*"):
 */
  __pyx_tuple__285 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_name, __pyx_n_s_listAllMatches, __pyx_n_s_newself); if (unlikely(!__pyx_tuple__285)) __PYX_ERR(0, 1229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__285);
  __Pyx_GIVEREF(__pyx_tuple__285);
  __pyx_codeobj__286 = (PyObject*)__Pyx_PyCode_New(3, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__285, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_setResultsName, 1229, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__286)) __PYX_ERR(0, 1229, __pyx_L1_error)
  __pyx_tuple__287 = PyTuple_Pack(1, ((PyObject *)Py_False)); if (unlikely(!__pyx_tuple__287)) __PYX_ERR(0, 1229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__287);
  __Pyx_GIVEREF(__pyx_tuple__287);

  /* "Aspidites/_vendor/pyparsing_extension.py":1238
 *         return newself
 * 
 *     def setBreak(self, breakFlag=True):             # <<<<<<<<<<<<<<
 *         """Method to invoke the Python pdb debugger when this element is
 *            about to be parsed. Set ``breakFlag`` to True to enable, False to
 */
  __pyx_tuple__288 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_breakFlag, __pyx_n_s_parseMethod, __pyx_n_s_breaker, __pyx_n_s_breaker); if (unlikely(!__pyx_tuple__288)) __PYX_ERR(0, 1238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__288);
  __Pyx_GIVEREF(__pyx_tuple__288);
  __pyx_codeobj__289 = (PyObject*)__Pyx_PyCode_New(2, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__288, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_setBreak, 1238, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__289)) __PYX_ERR(0, 1238, __pyx_L1_error)
  __pyx_tuple__290 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__290)) __PYX_ERR(0, 1238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__290);
  __Pyx_GIVEREF(__pyx_tuple__290);

  /* "Aspidites/_vendor/pyparsing_extension.py":1259
 *         return self
 * 
 *     def setParseAction(self, *fns, **kwargs):             # <<<<<<<<<<<<<<
 *         from .pyparsing import _trim_arity
 *         """
 */
  __pyx_tuple__291 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_fns, __pyx_n_s_kwargs, __pyx_n_s_trim_arity, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__291)) __PYX_ERR(0, 1259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__291);
  __Pyx_GIVEREF(__pyx_tuple__291);
  __pyx_codeobj__292 = (PyObject*)__Pyx_PyCode_New(1, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__291, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_setParseAction, 1259, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__292)) __PYX_ERR(0, 1259, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1309
 *         return self
 * 
 *     def addParseAction(self, *fns, **kwargs):             # <<<<<<<<<<<<<<
 *         from .pyparsing import _trim_arity
 *         """
 */
  __pyx_tuple__293 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_fns, __pyx_n_s_kwargs, __pyx_n_s_trim_arity); if (unlikely(!__pyx_tuple__293)) __PYX_ERR(0, 1309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__293);
  __Pyx_GIVEREF(__pyx_tuple__293);
  __pyx_codeobj__294 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_addParseAction, 1309, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__294)) __PYX_ERR(0, 1309, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1320
 *         return self
 * 
 *     def addCondition(self, *fns, **kwargs):             # <<<<<<<<<<<<<<
 *         from .pyparsing import conditionAsParseAction
 *         """Add a boolean predicate function to expression's list of parse actions. See
 */
  __pyx_tuple__295 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_fns, __pyx_n_s_kwargs, __pyx_n_s_conditionAsParseAction, __pyx_n_s_fn); if (unlikely(!__pyx_tuple__295)) __PYX_ERR(0, 1320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__295);
  __Pyx_GIVEREF(__pyx_tuple__295);
  __pyx_codeobj__296 = (PyObject*)__Pyx_PyCode_New(1, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__295, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_addCondition, 1320, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__296)) __PYX_ERR(0, 1320, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1346
 *         return self
 * 
 *     def setFailAction(self, fn):             # <<<<<<<<<<<<<<
 *         """Define action to perform if parsing fails at this expression.
 *            Fail acton fn is a callable function that takes the arguments
 */
  __pyx_tuple__297 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_fn); if (unlikely(!__pyx_tuple__297)) __PYX_ERR(0, 1346, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__297);
  __Pyx_GIVEREF(__pyx_tuple__297);
  __pyx_codeobj__298 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__297, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_setFailAction, 1346, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__298)) __PYX_ERR(0, 1346, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1359
 *         return self
 * 
 *     def _skipIgnorables(self, instring, loc):             # <<<<<<<<<<<<<<
 *         exprsFound = True
 *         while exprsFound:
 */
  __pyx_tuple__299 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_exprsFound, __pyx_n_s_e, __pyx_n_s_dummy); if (unlikely(!__pyx_tuple__299)) __PYX_ERR(0, 1359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__299);
  __Pyx_GIVEREF(__pyx_tuple__299);
  __pyx_codeobj__300 = (PyObject*)__Pyx_PyCode_New(3, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__299, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_skipIgnorables, 1359, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__300)) __PYX_ERR(0, 1359, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1372
 *         return loc
 * 
 *     def preParse(self, instring, loc):             # <<<<<<<<<<<<<<
 *         if self.ignoreExprs:
 *             loc = self._skipIgnorables(instring, loc)
 */
  __pyx_tuple__301 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_wt, __pyx_n_s_instrlen); if (unlikely(!__pyx_tuple__301)) __PYX_ERR(0, 1372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__301);
  __Pyx_GIVEREF(__pyx_tuple__301);
  __pyx_codeobj__302 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__301, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_preParse, 1372, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__302)) __PYX_ERR(0, 1372, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1384
 *         return loc
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         return loc, []
 * 
 */
  __pyx_tuple__303 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_doActions); if (unlikely(!__pyx_tuple__303)) __PYX_ERR(0, 1384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__303);
  __Pyx_GIVEREF(__pyx_tuple__303);
  __pyx_codeobj__304 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__303, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_parseImpl, 1384, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__304)) __PYX_ERR(0, 1384, __pyx_L1_error)
  __pyx_tuple__305 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__305)) __PYX_ERR(0, 1384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__305);
  __Pyx_GIVEREF(__pyx_tuple__305);

  /* "Aspidites/_vendor/pyparsing_extension.py":1387
 *         return loc, []
 * 
 *     def postParse(self, instring, loc, tokenlist):             # <<<<<<<<<<<<<<
 *         return tokenlist
 * 
 */
  __pyx_tuple__306 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_tokenlist); if (unlikely(!__pyx_tuple__306)) __PYX_ERR(0, 1387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__306);
  __Pyx_GIVEREF(__pyx_tuple__306);
  __pyx_codeobj__307 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__306, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_postParse, 1387, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__307)) __PYX_ERR(0, 1387, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1391
 * 
 *     # ~ @profile
 *     def _parseNoCache(self, instring, loc, doActions=True, callPreParse=True):             # <<<<<<<<<<<<<<
 *         TRY, MATCH, FAIL = 0, 1, 2
 *         debugging = (self.debug)  # and doActions)
 */
  __pyx_tuple__308 = PyTuple_Pack(17, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_doActions, __pyx_n_s_callPreParse, __pyx_n_s_TRY, __pyx_n_s_MATCH, __pyx_n_s_FAIL_2, __pyx_n_s_debugging, __pyx_n_s_preloc, __pyx_n_s_tokensStart, __pyx_n_s_tokens, __pyx_n_s_err, __pyx_n_s_retTokens, __pyx_n_s_fn, __pyx_n_s_parse_action_exc, __pyx_n_s_exc); if (unlikely(!__pyx_tuple__308)) __PYX_ERR(0, 1391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__308);
  __Pyx_GIVEREF(__pyx_tuple__308);
  __pyx_codeobj__309 = (PyObject*)__Pyx_PyCode_New(5, 0, 17, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__308, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_parseNoCache, 1391, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__309)) __PYX_ERR(0, 1391, __pyx_L1_error)
  __pyx_tuple__310 = PyTuple_Pack(2, ((PyObject *)Py_True), ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__310)) __PYX_ERR(0, 1391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__310);
  __Pyx_GIVEREF(__pyx_tuple__310);

  /* "Aspidites/_vendor/pyparsing_extension.py":1479
 *         return loc, retTokens
 * 
 *     def tryParse(self, instring, loc):             # <<<<<<<<<<<<<<
 *         try:
 *             return self._parse(instring, loc, doActions=False)[0]
 */
  __pyx_tuple__311 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc); if (unlikely(!__pyx_tuple__311)) __PYX_ERR(0, 1479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__311);
  __Pyx_GIVEREF(__pyx_tuple__311);
  __pyx_codeobj__312 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__311, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_tryParse, 1479, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__312)) __PYX_ERR(0, 1479, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1485
 *             raise ParseException(instring, loc, self.errmsg, self)
 * 
 *     def canParseNext(self, instring, loc):             # <<<<<<<<<<<<<<
 *         try:
 *             self.tryParse(instring, loc)
 */
  __pyx_tuple__313 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc); if (unlikely(!__pyx_tuple__313)) __PYX_ERR(0, 1485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__313);
  __Pyx_GIVEREF(__pyx_tuple__313);
  __pyx_codeobj__314 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__313, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_canParseNext, 1485, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__314)) __PYX_ERR(0, 1485, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1493
 *             return True
 * 
 *     class _UnboundedCache(object):             # <<<<<<<<<<<<<<
 *         def __init__(self):
 *             cache = {}
 */
  __pyx_tuple__315 = PyTuple_Pack(1, __pyx_builtin_object); if (unlikely(!__pyx_tuple__315)) __PYX_ERR(0, 1493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__315);
  __Pyx_GIVEREF(__pyx_tuple__315);

  /* "Aspidites/_vendor/pyparsing_extension.py":1494
 * 
 *     class _UnboundedCache(object):
 *         def __init__(self):             # <<<<<<<<<<<<<<
 *             cache = {}
 *             self.not_in_cache = not_in_cache = object()
 */
  __pyx_tuple__316 = PyTuple_Pack(11, __pyx_n_s_self, __pyx_n_s_cache, __pyx_n_s_not_in_cache, __pyx_n_s_get, __pyx_n_s_get, __pyx_n_s_set, __pyx_n_s_set, __pyx_n_s_clear, __pyx_n_s_clear, __pyx_n_s_cache_len, __pyx_n_s_cache_len); if (unlikely(!__pyx_tuple__316)) __PYX_ERR(0, 1494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__316);
  __Pyx_GIVEREF(__pyx_tuple__316);
  __pyx_codeobj__317 = (PyObject*)__Pyx_PyCode_New(1, 0, 11, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__316, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 1494, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__317)) __PYX_ERR(0, 1494, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1516
 * 
 *     if _OrderedDict is not None:
 *         class _FifoCache(object):             # <<<<<<<<<<<<<<
 *             def __init__(self, size):
 *                 self.not_in_cache = not_in_cache = object()
 */
  __pyx_tuple__318 = PyTuple_Pack(1, __pyx_builtin_object); if (unlikely(!__pyx_tuple__318)) __PYX_ERR(0, 1516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__318);
  __Pyx_GIVEREF(__pyx_tuple__318);

  /* "Aspidites/_vendor/pyparsing_extension.py":1517
 *     if _OrderedDict is not None:
 *         class _FifoCache(object):
 *             def __init__(self, size):             # <<<<<<<<<<<<<<
 *                 self.not_in_cache = not_in_cache = object()
 * 
 */
  __pyx_tuple__319 = PyTuple_Pack(12, __pyx_n_s_self, __pyx_n_s_size, __pyx_n_s_not_in_cache, __pyx_n_s_cache, __pyx_n_s_get, __pyx_n_s_get, __pyx_n_s_set, __pyx_n_s_set, __pyx_n_s_clear, __pyx_n_s_clear, __pyx_n_s_cache_len, __pyx_n_s_cache_len); if (unlikely(!__pyx_tuple__319)) __PYX_ERR(0, 1517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__319);
  __Pyx_GIVEREF(__pyx_tuple__319);
  __pyx_codeobj__320 = (PyObject*)__Pyx_PyCode_New(2, 0, 12, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__319, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 1517, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__320)) __PYX_ERR(0, 1517, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1545
 * 
 *     else:
 *         class _FifoCache(object):             # <<<<<<<<<<<<<<
 *             def __init__(self, size):
 *                 self.not_in_cache = not_in_cache = object()
 */
  __pyx_tuple__321 = PyTuple_Pack(1, __pyx_builtin_object); if (unlikely(!__pyx_tuple__321)) __PYX_ERR(0, 1545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__321);
  __Pyx_GIVEREF(__pyx_tuple__321);

  /* "Aspidites/_vendor/pyparsing_extension.py":1546
 *     else:
 *         class _FifoCache(object):
 *             def __init__(self, size):             # <<<<<<<<<<<<<<
 *                 self.not_in_cache = not_in_cache = object()
 * 
 */
  __pyx_tuple__322 = PyTuple_Pack(13, __pyx_n_s_self, __pyx_n_s_size, __pyx_n_s_not_in_cache, __pyx_n_s_cache, __pyx_n_s_key_fifo, __pyx_n_s_get, __pyx_n_s_get, __pyx_n_s_set, __pyx_n_s_set, __pyx_n_s_clear, __pyx_n_s_clear, __pyx_n_s_cache_len, __pyx_n_s_cache_len); if (unlikely(!__pyx_tuple__322)) __PYX_ERR(0, 1546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__322);
  __Pyx_GIVEREF(__pyx_tuple__322);
  __pyx_codeobj__323 = (PyObject*)__Pyx_PyCode_New(2, 0, 13, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__322, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 1546, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__323)) __PYX_ERR(0, 1546, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1580
 *     # this method gets repeatedly called during backtracking with the same arguments -
 *     # we can cache these arguments and save ourselves the trouble of re-parsing the contained expression
 *     def _parseCache(self, instring, loc, doActions=True, callPreParse=True):             # <<<<<<<<<<<<<<
 *         HIT, MISS = 0, 1
 *         lookup = (self, instring, loc, callPreParse, doActions)
 */
  __pyx_tuple__324 = PyTuple_Pack(11, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_doActions, __pyx_n_s_callPreParse, __pyx_n_s_HIT, __pyx_n_s_MISS, __pyx_n_s_lookup_2, __pyx_n_s_cache, __pyx_n_s_value, __pyx_n_s_pe); if (unlikely(!__pyx_tuple__324)) __PYX_ERR(0, 1580, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__324);
  __Pyx_GIVEREF(__pyx_tuple__324);
  __pyx_codeobj__325 = (PyObject*)__Pyx_PyCode_New(5, 0, 11, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__324, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_parseCache, 1580, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__325)) __PYX_ERR(0, 1580, __pyx_L1_error)
  __pyx_tuple__326 = PyTuple_Pack(2, ((PyObject *)Py_True), ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__326)) __PYX_ERR(0, 1580, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__326);
  __Pyx_GIVEREF(__pyx_tuple__326);

  /* "Aspidites/_vendor/pyparsing_extension.py":1606
 * 
 *     @staticmethod
 *     def resetCache():             # <<<<<<<<<<<<<<
 *         ParserElement.packrat_cache.clear()
 *         ParserElement.packrat_cache_stats[:] = [0] * len(ParserElement.packrat_cache_stats)
 */
  __pyx_codeobj__327 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_resetCache, 1606, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__327)) __PYX_ERR(0, 1606, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1613
 * 
 *     @staticmethod
 *     def enablePackrat(cache_size_limit=128):             # <<<<<<<<<<<<<<
 *         """Enables "packrat" parsing, which adds memoizing to the parsing logic.
 *            Repeated parse attempts at the same string location (which happens
 */
  __pyx_tuple__328 = PyTuple_Pack(1, __pyx_n_s_cache_size_limit); if (unlikely(!__pyx_tuple__328)) __PYX_ERR(0, 1613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__328);
  __Pyx_GIVEREF(__pyx_tuple__328);
  __pyx_codeobj__329 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__328, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_enablePackrat, 1613, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__329)) __PYX_ERR(0, 1613, __pyx_L1_error)
  __pyx_tuple__330 = PyTuple_Pack(1, ((PyObject *)__pyx_int_128)); if (unlikely(!__pyx_tuple__330)) __PYX_ERR(0, 1613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__330);
  __Pyx_GIVEREF(__pyx_tuple__330);

  /* "Aspidites/_vendor/pyparsing_extension.py":1647
 *             ParserElement._parse = ParserElement._parseCache
 * 
 *     def parseString(self, instring, parseAll=False):             # <<<<<<<<<<<<<<
 *         """
 *         Execute the parse expression with the given string.
 */
  __pyx_tuple__331 = PyTuple_Pack(10, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_parseAll, __pyx_n_s_Empty, __pyx_n_s_StringEnd, __pyx_n_s_e, __pyx_n_s_loc, __pyx_n_s_tokens, __pyx_n_s_se, __pyx_n_s_exc); if (unlikely(!__pyx_tuple__331)) __PYX_ERR(0, 1647, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__331);
  __Pyx_GIVEREF(__pyx_tuple__331);
  __pyx_codeobj__332 = (PyObject*)__Pyx_PyCode_New(3, 0, 10, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__331, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_parseString, 1647, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__332)) __PYX_ERR(0, 1647, __pyx_L1_error)
  __pyx_tuple__333 = PyTuple_Pack(1, ((PyObject *)Py_False)); if (unlikely(!__pyx_tuple__333)) __PYX_ERR(0, 1647, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__333);
  __Pyx_GIVEREF(__pyx_tuple__333);

  /* "Aspidites/_vendor/pyparsing_extension.py":1706
 *             return tokens
 * 
 *     def scanString(self, instring, maxMatches=sys.maxsize, overlap=False):             # <<<<<<<<<<<<<<
 *         """
 *         Scan the input string for expression matches.  Each match will return the
 */
  __pyx_tuple__334 = PyTuple_Pack(15, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_maxMatches, __pyx_n_s_overlap, __pyx_n_s_e, __pyx_n_s_instrlen, __pyx_n_s_loc, __pyx_n_s_preparseFn, __pyx_n_s_parseFn, __pyx_n_s_matches, __pyx_n_s_preloc, __pyx_n_s_nextLoc, __pyx_n_s_tokens, __pyx_n_s_nextloc, __pyx_n_s_exc); if (unlikely(!__pyx_tuple__334)) __PYX_ERR(0, 1706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__334);
  __Pyx_GIVEREF(__pyx_tuple__334);
  __pyx_codeobj__63 = (PyObject*)__Pyx_PyCode_New(4, 0, 15, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__334, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_scanString, 1706, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__63)) __PYX_ERR(0, 1706, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1780
 *                 raise exc
 * 
 *     def transformString(self, instring):             # <<<<<<<<<<<<<<
 *         """
 *         Extension to :class:`scanString`, to modify matching text with modified tokens that may
 */
  __pyx_tuple__335 = PyTuple_Pack(9, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_out, __pyx_n_s_lastE, __pyx_n_s_t_2, __pyx_n_s_s_4, __pyx_n_s_e, __pyx_n_s_exc, __pyx_n_s_o); if (unlikely(!__pyx_tuple__335)) __PYX_ERR(0, 1780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__335);
  __Pyx_GIVEREF(__pyx_tuple__335);
  __pyx_codeobj__336 = (PyObject*)__Pyx_PyCode_New(2, 0, 9, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__335, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_transformString, 1780, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__336)) __PYX_ERR(0, 1780, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1828
 *                 raise exc
 * 
 *     def searchString(self, instring, maxMatches=sys.maxsize):             # <<<<<<<<<<<<<<
 *         """
 *         Another extension to :class:`scanString`, simplifying the access to the tokens found
 */
  __pyx_tuple__337 = PyTuple_Pack(7, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_maxMatches, __pyx_n_s_exc, __pyx_n_s_t_2, __pyx_n_s_s_4, __pyx_n_s_e); if (unlikely(!__pyx_tuple__337)) __PYX_ERR(0, 1828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__337);
  __Pyx_GIVEREF(__pyx_tuple__337);
  __pyx_codeobj__338 = (PyObject*)__Pyx_PyCode_New(3, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__337, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_searchString, 1828, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__338)) __PYX_ERR(0, 1828, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1860
 *                 raise exc
 * 
 *     def split(self, instring, maxsplit=sys.maxsize, includeSeparators=False):             # <<<<<<<<<<<<<<
 *         """
 *         Generator method to split a string using the given expression as a separator.
 */
  __pyx_tuple__339 = PyTuple_Pack(9, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_maxsplit, __pyx_n_s_includeSeparators, __pyx_n_s_splits, __pyx_n_s_last, __pyx_n_s_t_2, __pyx_n_s_s_4, __pyx_n_s_e); if (unlikely(!__pyx_tuple__339)) __PYX_ERR(0, 1860, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__339);
  __Pyx_GIVEREF(__pyx_tuple__339);
  __pyx_codeobj__64 = (PyObject*)__Pyx_PyCode_New(4, 0, 9, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__339, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_split, 1860, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__64)) __PYX_ERR(0, 1860, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1885
 *         yield instring[last:]
 * 
 *     def __add__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of + operator - returns :class:`And`. Adding strings to a ParserElement
 */
  __pyx_tuple__340 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_other, __pyx_n_s_PendingSkip, __pyx_n_s_And); if (unlikely(!__pyx_tuple__340)) __PYX_ERR(0, 1885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__340);
  __Pyx_GIVEREF(__pyx_tuple__340);
  __pyx_codeobj__341 = (PyObject*)__Pyx_PyCode_New(2, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__340, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_add, 1885, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__341)) __PYX_ERR(0, 1885, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1924
 *         return And([self, other])
 * 
 *     def __radd__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of + operator when left operand is not a :class:`ParserElement`
 */
  __pyx_tuple__342 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_other, __pyx_n_s_SkipTo); if (unlikely(!__pyx_tuple__342)) __PYX_ERR(0, 1924, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__342);
  __Pyx_GIVEREF(__pyx_tuple__342);
  __pyx_codeobj__343 = (PyObject*)__Pyx_PyCode_New(2, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__342, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_radd, 1924, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__343)) __PYX_ERR(0, 1924, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1940
 *         return other + self
 * 
 *     def __sub__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of - operator, returns :class:`And` with error stop
 */
  __pyx_tuple__344 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_other, __pyx_n_s_And); if (unlikely(!__pyx_tuple__344)) __PYX_ERR(0, 1940, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__344);
  __Pyx_GIVEREF(__pyx_tuple__344);
  __pyx_codeobj__345 = (PyObject*)__Pyx_PyCode_New(2, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__344, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_sub_2, 1940, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__345)) __PYX_ERR(0, 1940, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1953
 *         return self + And._ErrorStop() + other
 * 
 *     def __rsub__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of - operator when left operand is not a :class:`ParserElement`
 */
  __pyx_tuple__346 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_other); if (unlikely(!__pyx_tuple__346)) __PYX_ERR(0, 1953, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__346);
  __Pyx_GIVEREF(__pyx_tuple__346);
  __pyx_codeobj__347 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__346, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_rsub, 1953, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__347)) __PYX_ERR(0, 1953, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1965
 *         return other - self
 * 
 *     def __mul__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of * operator, allows use of ``expr * 3`` in place of
 */
  __pyx_tuple__348 = PyTuple_Pack(12, __pyx_n_s_self, __pyx_n_s_other, __pyx_n_s_ZeroOrMore, __pyx_n_s_OneOrMore, __pyx_n_s_And, __pyx_n_s_minElements, __pyx_n_s_optElements, __pyx_n_s_makeOptionalList, __pyx_n_s_makeOptionalList, __pyx_n_s_ret, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__348)) __PYX_ERR(0, 1965, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__348);
  __Pyx_GIVEREF(__pyx_tuple__348);
  __pyx_codeobj__349 = (PyObject*)__Pyx_PyCode_New(2, 0, 12, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__348, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_mul, 1965, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__349)) __PYX_ERR(0, 1965, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2043
 *         return ret
 * 
 *     def __rmul__(self, other):             # <<<<<<<<<<<<<<
 *         return self.__mul__(other)
 * 
 */
  __pyx_tuple__350 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_other); if (unlikely(!__pyx_tuple__350)) __PYX_ERR(0, 2043, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__350);
  __Pyx_GIVEREF(__pyx_tuple__350);
  __pyx_codeobj__351 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__350, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_rmul, 2043, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__351)) __PYX_ERR(0, 2043, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2046
 *         return self.__mul__(other)
 * 
 *     def __or__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of | operator - returns :class:`MatchFirst`
 */
  __pyx_tuple__352 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_other, __pyx_n_s_PendingSkip, __pyx_n_s_MatchFirst); if (unlikely(!__pyx_tuple__352)) __PYX_ERR(0, 2046, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__352);
  __Pyx_GIVEREF(__pyx_tuple__352);
  __pyx_codeobj__353 = (PyObject*)__Pyx_PyCode_New(2, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__352, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_or, 2046, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__353)) __PYX_ERR(0, 2046, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2062
 *         return MatchFirst([self, other])
 * 
 *     def __ror__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of | operator when left operand is not a :class:`ParserElement`
 */
  __pyx_tuple__354 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_other); if (unlikely(!__pyx_tuple__354)) __PYX_ERR(0, 2062, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__354);
  __Pyx_GIVEREF(__pyx_tuple__354);
  __pyx_codeobj__355 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__354, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_ror, 2062, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__355)) __PYX_ERR(0, 2062, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2074
 *         return other | self
 * 
 *     def __xor__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of ^ operator - returns :class:`Or`
 */
  __pyx_tuple__356 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_other, __pyx_n_s_Or); if (unlikely(!__pyx_tuple__356)) __PYX_ERR(0, 2074, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__356);
  __Pyx_GIVEREF(__pyx_tuple__356);
  __pyx_codeobj__357 = (PyObject*)__Pyx_PyCode_New(2, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__356, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_xor, 2074, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__357)) __PYX_ERR(0, 2074, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2087
 *         return Or([self, other])
 * 
 *     def __rxor__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of ^ operator when left operand is not a :class:`ParserElement`
 */
  __pyx_tuple__358 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_other); if (unlikely(!__pyx_tuple__358)) __PYX_ERR(0, 2087, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__358);
  __Pyx_GIVEREF(__pyx_tuple__358);
  __pyx_codeobj__359 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__358, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_rxor, 2087, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__359)) __PYX_ERR(0, 2087, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2099
 *         return other ^ self
 * 
 *     def __and__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of & operator - returns :class:`Each`
 */
  __pyx_tuple__360 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_other, __pyx_n_s_Each); if (unlikely(!__pyx_tuple__360)) __PYX_ERR(0, 2099, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__360);
  __Pyx_GIVEREF(__pyx_tuple__360);
  __pyx_codeobj__361 = (PyObject*)__Pyx_PyCode_New(2, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__360, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_and, 2099, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__361)) __PYX_ERR(0, 2099, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2112
 *         return Each([self, other])
 * 
 *     def __rand__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of & operator when left operand is not a :class:`ParserElement`
 */
  __pyx_tuple__362 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_other); if (unlikely(!__pyx_tuple__362)) __PYX_ERR(0, 2112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__362);
  __Pyx_GIVEREF(__pyx_tuple__362);
  __pyx_codeobj__363 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__362, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_rand, 2112, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__363)) __PYX_ERR(0, 2112, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2124
 *         return other & self
 * 
 *     def __invert__(self):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of ~ operator - returns :class:`NotAny`
 */
  __pyx_tuple__364 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_NotAny); if (unlikely(!__pyx_tuple__364)) __PYX_ERR(0, 2124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__364);
  __Pyx_GIVEREF(__pyx_tuple__364);
  __pyx_codeobj__365 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__364, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_invert, 2124, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__365)) __PYX_ERR(0, 2124, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2131
 *         return NotAny(self)
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         # must implement __iter__ to override legacy use of sequential access to __getitem__ to
 *         # iterate over a sequence
 */
  __pyx_tuple__366 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__366)) __PYX_ERR(0, 2131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__366);
  __Pyx_GIVEREF(__pyx_tuple__366);
  __pyx_codeobj__367 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__366, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_iter, 2131, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__367)) __PYX_ERR(0, 2131, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2136
 *         raise TypeError('%r object is not iterable' % self.__class__.__name__)
 * 
 *     def __getitem__(self, key):             # <<<<<<<<<<<<<<
 *         """
 *         use ``[]`` indexing notation as a short form for expression repetition:
 */
  __pyx_tuple__368 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_key, __pyx_n_s_ret); if (unlikely(!__pyx_tuple__368)) __PYX_ERR(0, 2136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__368);
  __Pyx_GIVEREF(__pyx_tuple__368);
  __pyx_codeobj__369 = (PyObject*)__Pyx_PyCode_New(2, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__368, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_getitem, 2136, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__369)) __PYX_ERR(0, 2136, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2172
 *         return ret
 * 
 *     def __call__(self, name=None):             # <<<<<<<<<<<<<<
 *         """
 *         Shortcut for :class:`setResultsName`, with ``listAllMatches=False``.
 */
  __pyx_tuple__370 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_name); if (unlikely(!__pyx_tuple__370)) __PYX_ERR(0, 2172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__370);
  __Pyx_GIVEREF(__pyx_tuple__370);
  __pyx_codeobj__371 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__370, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_call, 2172, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__371)) __PYX_ERR(0, 2172, __pyx_L1_error)
  __pyx_tuple__372 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__372)) __PYX_ERR(0, 2172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__372);
  __Pyx_GIVEREF(__pyx_tuple__372);

  /* "Aspidites/_vendor/pyparsing_extension.py":2192
 *             return self.copy()
 * 
 *     def suppress(self):             # <<<<<<<<<<<<<<
 *         """
 *         Suppresses the output of this :class:`ParserElement`; useful to keep punctuation from
 */
  __pyx_tuple__373 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_Suppress); if (unlikely(!__pyx_tuple__373)) __PYX_ERR(0, 2192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__373);
  __Pyx_GIVEREF(__pyx_tuple__373);
  __pyx_codeobj__374 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__373, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_suppress, 2192, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__374)) __PYX_ERR(0, 2192, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2200
 *         return Suppress(self)
 * 
 *     def leaveWhitespace(self):             # <<<<<<<<<<<<<<
 *         """
 *         Disables the skipping of whitespace before matching the characters in the
 */
  __pyx_tuple__375 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__375)) __PYX_ERR(0, 2200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__375);
  __Pyx_GIVEREF(__pyx_tuple__375);
  __pyx_codeobj__376 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__375, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_leaveWhitespace, 2200, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__376)) __PYX_ERR(0, 2200, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2209
 *         return self
 * 
 *     def setWhitespaceChars(self, chars):             # <<<<<<<<<<<<<<
 *         """
 *         Overrides the default whitespace chars
 */
  __pyx_tuple__377 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_chars); if (unlikely(!__pyx_tuple__377)) __PYX_ERR(0, 2209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__377);
  __Pyx_GIVEREF(__pyx_tuple__377);
  __pyx_codeobj__378 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__377, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_setWhitespaceChars, 2209, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__378)) __PYX_ERR(0, 2209, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2218
 *         return self
 * 
 *     def parseWithTabs(self):             # <<<<<<<<<<<<<<
 *         """
 *         Overrides default behavior to expand ``<TAB>``s to spaces before parsing the input string.
 */
  __pyx_tuple__379 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__379)) __PYX_ERR(0, 2218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__379);
  __Pyx_GIVEREF(__pyx_tuple__379);
  __pyx_codeobj__380 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__379, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_parseWithTabs, 2218, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__380)) __PYX_ERR(0, 2218, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2227
 *         return self
 * 
 *     def ignore(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Define expression to be ignored (e.g., comments) while doing pattern
 */
  __pyx_tuple__381 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_other, __pyx_n_s_Suppress); if (unlikely(!__pyx_tuple__381)) __PYX_ERR(0, 2227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__381);
  __Pyx_GIVEREF(__pyx_tuple__381);
  __pyx_codeobj__382 = (PyObject*)__Pyx_PyCode_New(2, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__381, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_ignore, 2227, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__382)) __PYX_ERR(0, 2227, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2252
 *         return self
 * 
 *     def setDebugActions(self, startAction, successAction, exceptionAction):             # <<<<<<<<<<<<<<
 *         """
 *         Enable display of debugging messages while doing pattern matching.
 */
  __pyx_tuple__383 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_startAction, __pyx_n_s_successAction, __pyx_n_s_exceptionAction); if (unlikely(!__pyx_tuple__383)) __PYX_ERR(0, 2252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__383);
  __Pyx_GIVEREF(__pyx_tuple__383);
  __pyx_codeobj__384 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__383, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_setDebugActions, 2252, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__384)) __PYX_ERR(0, 2252, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2262
 *         return self
 * 
 *     def setDebug(self, flag=True):             # <<<<<<<<<<<<<<
 *         """
 *         Enable display of debugging messages while doing pattern matching.
 */
  __pyx_tuple__385 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_flag); if (unlikely(!__pyx_tuple__385)) __PYX_ERR(0, 2262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__385);
  __Pyx_GIVEREF(__pyx_tuple__385);
  __pyx_codeobj__386 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__385, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_setDebug, 2262, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__386)) __PYX_ERR(0, 2262, __pyx_L1_error)
  __pyx_tuple__387 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__387)) __PYX_ERR(0, 2262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__387);
  __Pyx_GIVEREF(__pyx_tuple__387);

  /* "Aspidites/_vendor/pyparsing_extension.py":2305
 *         return self
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return self.name
 * 
 */
  __pyx_tuple__388 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__388)) __PYX_ERR(0, 2305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__388);
  __Pyx_GIVEREF(__pyx_tuple__388);
  __pyx_codeobj__389 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__388, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_str, 2305, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__389)) __PYX_ERR(0, 2305, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2308
 *         return self.name
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return str(self)
 * 
 */
  __pyx_tuple__390 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__390)) __PYX_ERR(0, 2308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__390);
  __Pyx_GIVEREF(__pyx_tuple__390);
  __pyx_codeobj__391 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__390, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_repr, 2308, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__391)) __PYX_ERR(0, 2308, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2311
 *         return str(self)
 * 
 *     def streamline(self):             # <<<<<<<<<<<<<<
 *         self.streamlined = True
 *         self.strRepr = None
 */
  __pyx_tuple__392 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__392)) __PYX_ERR(0, 2311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__392);
  __Pyx_GIVEREF(__pyx_tuple__392);
  __pyx_codeobj__393 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__392, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_streamline, 2311, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__393)) __PYX_ERR(0, 2311, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2316
 *         return self
 * 
 *     def checkRecursion(self, parseElementList):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__394 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_parseElementList); if (unlikely(!__pyx_tuple__394)) __PYX_ERR(0, 2316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__394);
  __Pyx_GIVEREF(__pyx_tuple__394);
  __pyx_codeobj__395 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__394, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_checkRecursion, 2316, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__395)) __PYX_ERR(0, 2316, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2319
 *         pass
 * 
 *     def validate(self, validateTrace=None):             # <<<<<<<<<<<<<<
 *         """
 *         Check defined expressions for valid structure, check for infinite recursive definitions.
 */
  __pyx_tuple__396 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_validateTrace); if (unlikely(!__pyx_tuple__396)) __PYX_ERR(0, 2319, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__396);
  __Pyx_GIVEREF(__pyx_tuple__396);
  __pyx_codeobj__397 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__396, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_validate, 2319, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__397)) __PYX_ERR(0, 2319, __pyx_L1_error)
  __pyx_tuple__398 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__398)) __PYX_ERR(0, 2319, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__398);
  __Pyx_GIVEREF(__pyx_tuple__398);

  /* "Aspidites/_vendor/pyparsing_extension.py":2325
 *         self.checkRecursion([])
 * 
 *     def parseFile(self, file_or_filename, parseAll=False):             # <<<<<<<<<<<<<<
 *         """
 *         Execute the parse expression on the given file or filename.
 */
  __pyx_tuple__399 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_file_or_filename, __pyx_n_s_parseAll, __pyx_n_s_file_contents, __pyx_n_s_f_2, __pyx_n_s_exc); if (unlikely(!__pyx_tuple__399)) __PYX_ERR(0, 2325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__399);
  __Pyx_GIVEREF(__pyx_tuple__399);
  __pyx_codeobj__400 = (PyObject*)__Pyx_PyCode_New(3, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__399, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_parseFile, 2325, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__400)) __PYX_ERR(0, 2325, __pyx_L1_error)
  __pyx_tuple__401 = PyTuple_Pack(1, ((PyObject *)Py_False)); if (unlikely(!__pyx_tuple__401)) __PYX_ERR(0, 2325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__401);
  __Pyx_GIVEREF(__pyx_tuple__401);

  /* "Aspidites/_vendor/pyparsing_extension.py":2347
 *                 raise exc
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         if self is other:
 *             return True
 */
  __pyx_tuple__402 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_other); if (unlikely(!__pyx_tuple__402)) __PYX_ERR(0, 2347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__402);
  __Pyx_GIVEREF(__pyx_tuple__402);
  __pyx_codeobj__403 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__402, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_eq, 2347, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__403)) __PYX_ERR(0, 2347, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2356
 *         return False
 * 
 *     def __ne__(self, other):             # <<<<<<<<<<<<<<
 *         return not (self == other)
 * 
 */
  __pyx_tuple__404 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_other); if (unlikely(!__pyx_tuple__404)) __PYX_ERR(0, 2356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__404);
  __Pyx_GIVEREF(__pyx_tuple__404);
  __pyx_codeobj__405 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__404, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_ne, 2356, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__405)) __PYX_ERR(0, 2356, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2359
 *         return not (self == other)
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return id(self)
 * 
 */
  __pyx_tuple__406 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__406)) __PYX_ERR(0, 2359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__406);
  __Pyx_GIVEREF(__pyx_tuple__406);
  __pyx_codeobj__407 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__406, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_hash, 2359, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__407)) __PYX_ERR(0, 2359, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2362
 *         return id(self)
 * 
 *     def __req__(self, other):             # <<<<<<<<<<<<<<
 *         return self == other
 * 
 */
  __pyx_tuple__408 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_other); if (unlikely(!__pyx_tuple__408)) __PYX_ERR(0, 2362, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__408);
  __Pyx_GIVEREF(__pyx_tuple__408);
  __pyx_codeobj__409 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__408, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_req, 2362, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__409)) __PYX_ERR(0, 2362, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2365
 *         return self == other
 * 
 *     def __rne__(self, other):             # <<<<<<<<<<<<<<
 *         return not (self == other)
 * 
 */
  __pyx_tuple__410 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_other); if (unlikely(!__pyx_tuple__410)) __PYX_ERR(0, 2365, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__410);
  __Pyx_GIVEREF(__pyx_tuple__410);
  __pyx_codeobj__411 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__410, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_rne, 2365, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__411)) __PYX_ERR(0, 2365, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2368
 *         return not (self == other)
 * 
 *     def matches(self, testString, parseAll=True):             # <<<<<<<<<<<<<<
 *         """
 *         Method for quick testing of a parser against a test string. Good for simple
 */
  __pyx_tuple__412 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_testString, __pyx_n_s_parseAll); if (unlikely(!__pyx_tuple__412)) __PYX_ERR(0, 2368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__412);
  __Pyx_GIVEREF(__pyx_tuple__412);
  __pyx_codeobj__413 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__412, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_matches, 2368, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__413)) __PYX_ERR(0, 2368, __pyx_L1_error)
  __pyx_tuple__414 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__414)) __PYX_ERR(0, 2368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__414);
  __Pyx_GIVEREF(__pyx_tuple__414);

  /* "Aspidites/_vendor/pyparsing_extension.py":2388
 *             return False
 * 
 *     def runTests(self, tests, parseAll=True, comment='#',             # <<<<<<<<<<<<<<
 *                  fullDump=True, printResults=True, failureTests=False, postParse=None,
 *                  file=None):
 */
  __pyx_tuple__415 = PyTuple_Pack(26, __pyx_n_s_self, __pyx_n_s_tests, __pyx_n_s_parseAll, __pyx_n_s_comment, __pyx_n_s_fullDump, __pyx_n_s_printResults, __pyx_n_s_failureTests, __pyx_n_s_postParse, __pyx_n_s_file, __pyx_n_s_Literal, __pyx_n_s_replaceWith, __pyx_n_s_quotedString, __pyx_n_s_print, __pyx_n_s_allResults, __pyx_n_s_comments, __pyx_n_s_success, __pyx_n_s_NL, __pyx_n_s_BOM, __pyx_n_s_t_2, __pyx_n_s_out, __pyx_n_s_result, __pyx_n_s_pe, __pyx_n_s_fatal, __pyx_n_s_exc, __pyx_n_s_pp_value, __pyx_n_s_e); if (unlikely(!__pyx_tuple__415)) __PYX_ERR(0, 2388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__415);
  __Pyx_GIVEREF(__pyx_tuple__415);
  __pyx_codeobj__416 = (PyObject*)__Pyx_PyCode_New(9, 0, 26, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__415, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_runTests, 2388, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__416)) __PYX_ERR(0, 2388, __pyx_L1_error)
  __pyx_tuple__417 = PyTuple_Pack(7, ((PyObject *)Py_True), ((PyObject*)__pyx_kp_s__79), ((PyObject *)Py_True), ((PyObject *)Py_True), ((PyObject *)Py_False), ((PyObject *)Py_None), ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__417)) __PYX_ERR(0, 2388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__417);
  __Pyx_GIVEREF(__pyx_tuple__417);

  /* "Aspidites/_vendor/pyparsing_extension.py":2557
 *     """
 * 
 *     def __init__(self, exprs, savelist=False):             # <<<<<<<<<<<<<<
 *         super(ParseExpression, self).__init__(savelist)
 *         if isinstance(exprs, _generatorType):
 */
  __pyx_tuple__418 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_exprs, __pyx_n_s_savelist, __pyx_n_s_genexpr, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__418)) __PYX_ERR(0, 2557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__418);
  __Pyx_GIVEREF(__pyx_tuple__418);
  __pyx_codeobj__419 = (PyObject*)__Pyx_PyCode_New(3, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__418, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 2557, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__419)) __PYX_ERR(0, 2557, __pyx_L1_error)
  __pyx_tuple__420 = PyTuple_Pack(1, ((PyObject *)Py_False)); if (unlikely(!__pyx_tuple__420)) __PYX_ERR(0, 2557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__420);
  __Pyx_GIVEREF(__pyx_tuple__420);

  /* "Aspidites/_vendor/pyparsing_extension.py":2579
 *         self.callPreparse = False
 * 
 *     def append(self, other):             # <<<<<<<<<<<<<<
 *         self.exprs.append(other)
 *         self.strRepr = None
 */
  __pyx_tuple__421 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_other); if (unlikely(!__pyx_tuple__421)) __PYX_ERR(0, 2579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__421);
  __Pyx_GIVEREF(__pyx_tuple__421);
  __pyx_codeobj__422 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__421, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_append, 2579, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__422)) __PYX_ERR(0, 2579, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2584
 *         return self
 * 
 *     def leaveWhitespace(self):             # <<<<<<<<<<<<<<
 *         """Extends ``leaveWhitespace`` defined in base class, and also invokes ``leaveWhitespace`` on
 *            all contained expressions."""
 */
  __pyx_tuple__423 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_e); if (unlikely(!__pyx_tuple__423)) __PYX_ERR(0, 2584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__423);
  __Pyx_GIVEREF(__pyx_tuple__423);
  __pyx_codeobj__424 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__423, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_leaveWhitespace, 2584, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__424)) __PYX_ERR(0, 2584, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2593
 *         return self
 * 
 *     def ignore(self, other):             # <<<<<<<<<<<<<<
 *         from .pyparsing import Suppress
 *         if isinstance(other, Suppress):
 */
  __pyx_tuple__425 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_other, __pyx_n_s_Suppress, __pyx_n_s_e); if (unlikely(!__pyx_tuple__425)) __PYX_ERR(0, 2593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__425);
  __Pyx_GIVEREF(__pyx_tuple__425);
  __pyx_codeobj__426 = (PyObject*)__Pyx_PyCode_New(2, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__425, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_ignore, 2593, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__426)) __PYX_ERR(0, 2593, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2606
 *         return self
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         try:
 *             return super(ParseExpression, self).__str__()
 */
  __pyx_tuple__427 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__427)) __PYX_ERR(0, 2606, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__427);
  __Pyx_GIVEREF(__pyx_tuple__427);
  __pyx_codeobj__428 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__427, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_str, 2606, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__428)) __PYX_ERR(0, 2606, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2616
 *         return self.strRepr
 * 
 *     def streamline(self):             # <<<<<<<<<<<<<<
 *         super(ParseExpression, self).streamline()
 * 
 */
  __pyx_tuple__429 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_e, __pyx_n_s_other); if (unlikely(!__pyx_tuple__429)) __PYX_ERR(0, 2616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__429);
  __Pyx_GIVEREF(__pyx_tuple__429);
  __pyx_codeobj__430 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__429, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_streamline, 2616, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__430)) __PYX_ERR(0, 2616, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2650
 *         return self
 * 
 *     def validate(self, validateTrace=None):             # <<<<<<<<<<<<<<
 *         tmp = (validateTrace if validateTrace is not None else [])[:] + [self]
 *         for e in self.exprs:
 */
  __pyx_tuple__431 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_validateTrace, __pyx_n_s_tmp, __pyx_n_s_e); if (unlikely(!__pyx_tuple__431)) __PYX_ERR(0, 2650, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__431);
  __Pyx_GIVEREF(__pyx_tuple__431);
  __pyx_codeobj__432 = (PyObject*)__Pyx_PyCode_New(2, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__431, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_validate, 2650, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__432)) __PYX_ERR(0, 2650, __pyx_L1_error)
  __pyx_tuple__433 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__433)) __PYX_ERR(0, 2650, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__433);
  __Pyx_GIVEREF(__pyx_tuple__433);

  /* "Aspidites/_vendor/pyparsing_extension.py":2656
 *         self.checkRecursion([])
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         ret = super(ParseExpression, self).copy()
 *         ret.exprs = [e.copy() for e in self.exprs]
 */
  __pyx_tuple__434 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_ret, __pyx_n_s_e); if (unlikely(!__pyx_tuple__434)) __PYX_ERR(0, 2656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__434);
  __Pyx_GIVEREF(__pyx_tuple__434);
  __pyx_codeobj__435 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__434, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_copy, 2656, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__435)) __PYX_ERR(0, 2656, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2661
 *         return ret
 * 
 *     def _setResultsName(self, name, listAllMatches=False):             # <<<<<<<<<<<<<<
 *         if __diag__.warn_ungrouped_named_tokens_in_collection:
 *             for e in self.exprs:
 */
  __pyx_tuple__436 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_name, __pyx_n_s_listAllMatches, __pyx_n_s_e); if (unlikely(!__pyx_tuple__436)) __PYX_ERR(0, 2661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__436);
  __Pyx_GIVEREF(__pyx_tuple__436);
  __pyx_codeobj__437 = (PyObject*)__Pyx_PyCode_New(3, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__436, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_setResultsName, 2661, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__437)) __PYX_ERR(0, 2661, __pyx_L1_error)
  __pyx_tuple__438 = PyTuple_Pack(1, ((PyObject *)Py_False)); if (unlikely(!__pyx_tuple__438)) __PYX_ERR(0, 2661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__438);
  __Pyx_GIVEREF(__pyx_tuple__438);

  /* "Aspidites/_vendor/pyparsing_extension.py":2681
 *     """
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         super(Token, self).__init__(savelist=False)
 * 
 */
  __pyx_tuple__439 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__439)) __PYX_ERR(0, 2681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__439);
  __Pyx_GIVEREF(__pyx_tuple__439);
  __pyx_codeobj__440 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__439, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 2681, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__440)) __PYX_ERR(0, 2681, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2689
 *     """
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         super(Empty, self).__init__()
 *         self.name = "Empty"
 */
  __pyx_tuple__441 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__441)) __PYX_ERR(0, 2689, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__441);
  __Pyx_GIVEREF(__pyx_tuple__441);
  __pyx_codeobj__442 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__441, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 2689, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__442)) __PYX_ERR(0, 2689, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2700
 *     """
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         super(NoMatch, self).__init__()
 *         self.name = "NoMatch"
 */
  __pyx_tuple__443 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__443)) __PYX_ERR(0, 2700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__443);
  __Pyx_GIVEREF(__pyx_tuple__443);
  __pyx_codeobj__444 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__443, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 2700, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__444)) __PYX_ERR(0, 2700, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2707
 *         self.errmsg = "Unmatchable token"
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         raise ParseException(instring, loc, self.errmsg, self)
 * 
 */
  __pyx_tuple__445 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_doActions); if (unlikely(!__pyx_tuple__445)) __PYX_ERR(0, 2707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__445);
  __Pyx_GIVEREF(__pyx_tuple__445);
  __pyx_codeobj__446 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__445, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_parseImpl, 2707, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__446)) __PYX_ERR(0, 2707, __pyx_L1_error)
  __pyx_tuple__447 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__447)) __PYX_ERR(0, 2707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__447);
  __Pyx_GIVEREF(__pyx_tuple__447);

  /* "Aspidites/_vendor/pyparsing_extension.py":2726
 *     """
 * 
 *     def __init__(self, matchString):             # <<<<<<<<<<<<<<
 *         super(Literal, self).__init__()
 *         self.match = matchString
 */
  __pyx_tuple__448 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_matchString); if (unlikely(!__pyx_tuple__448)) __PYX_ERR(0, 2726, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__448);
  __Pyx_GIVEREF(__pyx_tuple__448);
  __pyx_codeobj__449 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__448, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 2726, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__449)) __PYX_ERR(0, 2726, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2746
 *             self.__class__ = _SingleCharLiteral
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if instring[loc] == self.firstMatchChar and instring.startswith(self.match, loc):
 *             return loc + self.matchLen, self.match
 */
  __pyx_tuple__450 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_doActions); if (unlikely(!__pyx_tuple__450)) __PYX_ERR(0, 2746, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__450);
  __Pyx_GIVEREF(__pyx_tuple__450);
  __pyx_codeobj__451 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__450, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_parseImpl, 2746, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__451)) __PYX_ERR(0, 2746, __pyx_L1_error)
  __pyx_tuple__452 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__452)) __PYX_ERR(0, 2746, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__452);
  __Pyx_GIVEREF(__pyx_tuple__452);

  /* "Aspidites/_vendor/pyparsing_extension.py":2753
 * 
 * class _SingleCharLiteral(Literal):
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if instring[loc] == self.firstMatchChar:
 *             return loc + 1, self.match
 */
  __pyx_tuple__453 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_doActions); if (unlikely(!__pyx_tuple__453)) __PYX_ERR(0, 2753, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__453);
  __Pyx_GIVEREF(__pyx_tuple__453);
  __pyx_codeobj__454 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__453, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_parseImpl, 2753, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__454)) __PYX_ERR(0, 2753, __pyx_L1_error)
  __pyx_tuple__455 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__455)) __PYX_ERR(0, 2753, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__455);
  __Pyx_GIVEREF(__pyx_tuple__455);

  /* "Aspidites/_vendor/pyparsing_extension.py":2790
 *     DEFAULT_KEYWORD_CHARS = alphanums + "_$"
 * 
 *     def __init__(self, matchString, identChars=None, caseless=False):             # <<<<<<<<<<<<<<
 *         super(Keyword, self).__init__()
 *         if identChars is None:
 */
  __pyx_tuple__457 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_matchString, __pyx_n_s_identChars, __pyx_n_s_caseless); if (unlikely(!__pyx_tuple__457)) __PYX_ERR(0, 2790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__457);
  __Pyx_GIVEREF(__pyx_tuple__457);
  __pyx_codeobj__458 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__457, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 2790, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__458)) __PYX_ERR(0, 2790, __pyx_L1_error)
  __pyx_tuple__459 = PyTuple_Pack(2, ((PyObject *)Py_None), ((PyObject *)Py_False)); if (unlikely(!__pyx_tuple__459)) __PYX_ERR(0, 2790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__459);
  __Pyx_GIVEREF(__pyx_tuple__459);

  /* "Aspidites/_vendor/pyparsing_extension.py":2811
 *         self.identChars = set(identChars)
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if self.caseless:
 *             if ((instring[loc:loc + self.matchLen].upper() == self.caselessmatch)
 */
  __pyx_tuple__460 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_doActions); if (unlikely(!__pyx_tuple__460)) __PYX_ERR(0, 2811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__460);
  __Pyx_GIVEREF(__pyx_tuple__460);
  __pyx_codeobj__461 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__460, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_parseImpl, 2811, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__461)) __PYX_ERR(0, 2811, __pyx_L1_error)
  __pyx_tuple__462 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__462)) __PYX_ERR(0, 2811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__462);
  __Pyx_GIVEREF(__pyx_tuple__462);

  /* "Aspidites/_vendor/pyparsing_extension.py":2830
 *         raise ParseException(instring, loc, self.errmsg, self)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         c = super(Keyword, self).copy()
 *         c.identChars = Keyword.DEFAULT_KEYWORD_CHARS
 */
  __pyx_tuple__463 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_c); if (unlikely(!__pyx_tuple__463)) __PYX_ERR(0, 2830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__463);
  __Pyx_GIVEREF(__pyx_tuple__463);
  __pyx_codeobj__464 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__463, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_copy, 2830, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__464)) __PYX_ERR(0, 2830, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2836
 * 
 *     @staticmethod
 *     def setDefaultKeywordChars(chars):             # <<<<<<<<<<<<<<
 *         """Overrides the default Keyword chars
 *         """
 */
  __pyx_tuple__465 = PyTuple_Pack(1, __pyx_n_s_chars); if (unlikely(!__pyx_tuple__465)) __PYX_ERR(0, 2836, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__465);
  __Pyx_GIVEREF(__pyx_tuple__465);
  __pyx_codeobj__466 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__465, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_setDefaultKeywordChars, 2836, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__466)) __PYX_ERR(0, 2836, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2854
 *     """
 * 
 *     def __init__(self, matchString):             # <<<<<<<<<<<<<<
 *         super(CaselessLiteral, self).__init__(matchString.upper())
 *         # Preserve the defining literal.
 */
  __pyx_tuple__467 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_matchString); if (unlikely(!__pyx_tuple__467)) __PYX_ERR(0, 2854, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__467);
  __Pyx_GIVEREF(__pyx_tuple__467);
  __pyx_codeobj__468 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__467, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 2854, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__468)) __PYX_ERR(0, 2854, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":2861
 *         self.errmsg = "Expected " + self.name
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if instring[loc:loc + self.matchLen].upper() == self.match:
 *             return loc + self.matchLen, self.returnString
 */
  __pyx_tuple__469 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_doActions); if (unlikely(!__pyx_tuple__469)) __PYX_ERR(0, 2861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__469);
  __Pyx_GIVEREF(__pyx_tuple__469);
  __pyx_codeobj__470 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__469, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_parseImpl, 2861, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__470)) __PYX_ERR(0, 2861, __pyx_L1_error)
  __pyx_tuple__471 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__471)) __PYX_ERR(0, 2861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__471);
  __Pyx_GIVEREF(__pyx_tuple__471);

  /* "Aspidites/_vendor/pyparsing_extension.py":2878
 *     """
 * 
 *     def __init__(self, matchString, identChars=None):             # <<<<<<<<<<<<<<
 *         super(CaselessKeyword, self).__init__(matchString, identChars, caseless=True)
 * 
 */
  __pyx_tuple__472 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_matchString, __pyx_n_s_identChars); if (unlikely(!__pyx_tuple__472)) __PYX_ERR(0, 2878, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__472);
  __Pyx_GIVEREF(__pyx_tuple__472);
  __pyx_codeobj__473 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__472, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 2878, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__473)) __PYX_ERR(0, 2878, __pyx_L1_error)
  __pyx_tuple__474 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__474)) __PYX_ERR(0, 2878, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__474);
  __Pyx_GIVEREF(__pyx_tuple__474);

  /* "Aspidites/_vendor/pyparsing_extension.py":2916
 *     """
 * 
 *     def __init__(self, match_string, maxMismatches=1):             # <<<<<<<<<<<<<<
 *         super(CloseMatch, self).__init__()
 *         self.name = match_string
 */
  __pyx_tuple__475 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_match_string, __pyx_n_s_maxMismatches); if (unlikely(!__pyx_tuple__475)) __PYX_ERR(0, 2916, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__475);
  __Pyx_GIVEREF(__pyx_tuple__475);
  __pyx_codeobj__476 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__475, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 2916, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__476)) __PYX_ERR(0, 2916, __pyx_L1_error)
  __pyx_tuple__477 = PyTuple_Pack(1, ((PyObject *)__pyx_int_1)); if (unlikely(!__pyx_tuple__477)) __PYX_ERR(0, 2916, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__477);
  __Pyx_GIVEREF(__pyx_tuple__477);

  /* "Aspidites/_vendor/pyparsing_extension.py":2925
 *         self.mayReturnEmpty = False
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         start = loc
 *         instrlen = len(instring)
 */
  __pyx_tuple__478 = PyTuple_Pack(15, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_doActions, __pyx_n_s_start, __pyx_n_s_instrlen, __pyx_n_s_maxloc, __pyx_n_s_match_string, __pyx_n_s_match_stringloc, __pyx_n_s_mismatches, __pyx_n_s_maxMismatches, __pyx_n_s_s_m, __pyx_n_s_src, __pyx_n_s_mat, __pyx_n_s_results); if (unlikely(!__pyx_tuple__478)) __PYX_ERR(0, 2925, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__478);
  __Pyx_GIVEREF(__pyx_tuple__478);
  __pyx_codeobj__479 = (PyObject*)__Pyx_PyCode_New(4, 0, 15, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__478, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_parseImpl, 2925, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__479)) __PYX_ERR(0, 2925, __pyx_L1_error)
  __pyx_tuple__480 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__480)) __PYX_ERR(0, 2925, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__480);
  __Pyx_GIVEREF(__pyx_tuple__480);

  /* "Aspidites/_vendor/pyparsing_extension.py":3006
 *     """
 * 
 *     def __init__(self, initChars, bodyChars=None, min=1, max=0, exact=0, asKeyword=False, excludeChars=None):             # <<<<<<<<<<<<<<
 *         super(Word, self).__init__()
 *         if excludeChars:
 */
  __pyx_tuple__481 = PyTuple_Pack(11, __pyx_n_s_self, __pyx_n_s_initChars, __pyx_n_s_bodyChars, __pyx_n_s_min, __pyx_n_s_max, __pyx_n_s_exact, __pyx_n_s_asKeyword, __pyx_n_s_excludeChars, __pyx_n_s_genexpr, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__481)) __PYX_ERR(0, 3006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__481);
  __Pyx_GIVEREF(__pyx_tuple__481);
  __pyx_codeobj__482 = (PyObject*)__Pyx_PyCode_New(8, 0, 11, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__481, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 3006, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__482)) __PYX_ERR(0, 3006, __pyx_L1_error)
  __pyx_tuple__483 = PyTuple_Pack(6, ((PyObject *)Py_None), ((PyObject *)__pyx_int_1), ((PyObject *)__pyx_int_0), ((PyObject *)__pyx_int_0), ((PyObject *)Py_False), ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__483)) __PYX_ERR(0, 3006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__483);
  __Pyx_GIVEREF(__pyx_tuple__483);

  /* "Aspidites/_vendor/pyparsing_extension.py":3064
 *                 self.__class__ = _WordRegex
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if instring[loc] not in self.initChars:
 *             raise ParseException(instring, loc, self.errmsg, self)
 */
  __pyx_tuple__484 = PyTuple_Pack(9, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_doActions, __pyx_n_s_start, __pyx_n_s_instrlen, __pyx_n_s_bodychars, __pyx_n_s_maxloc, __pyx_n_s_throwException); if (unlikely(!__pyx_tuple__484)) __PYX_ERR(0, 3064, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__484);
  __Pyx_GIVEREF(__pyx_tuple__484);
  __pyx_codeobj__485 = (PyObject*)__Pyx_PyCode_New(4, 0, 9, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__484, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_parseImpl, 3064, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__485)) __PYX_ERR(0, 3064, __pyx_L1_error)
  __pyx_tuple__486 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__486)) __PYX_ERR(0, 3064, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__486);
  __Pyx_GIVEREF(__pyx_tuple__486);

  /* "Aspidites/_vendor/pyparsing_extension.py":3092
 *         return loc, instring[start:loc]
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         try:
 *             return super(Word, self).__str__()
 */
  __pyx_tuple__487 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_charsAsStr, __pyx_n_s_charsAsStr); if (unlikely(!__pyx_tuple__487)) __PYX_ERR(0, 3092, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__487);
  __Pyx_GIVEREF(__pyx_tuple__487);
  __pyx_codeobj__488 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__487, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_str, 3092, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__488)) __PYX_ERR(0, 3092, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3115
 * 
 * class _WordRegex(Word):
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         result = self.re_match(instring, loc)
 *         if not result:
 */
  __pyx_tuple__489 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_doActions, __pyx_n_s_result); if (unlikely(!__pyx_tuple__489)) __PYX_ERR(0, 3115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__489);
  __Pyx_GIVEREF(__pyx_tuple__489);
  __pyx_codeobj__490 = (PyObject*)__Pyx_PyCode_New(4, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__489, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_parseImpl, 3115, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__490)) __PYX_ERR(0, 3115, __pyx_L1_error)
  __pyx_tuple__491 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__491)) __PYX_ERR(0, 3115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__491);
  __Pyx_GIVEREF(__pyx_tuple__491);

  /* "Aspidites/_vendor/pyparsing_extension.py":3130
 *     """
 * 
 *     def __init__(self, charset, asKeyword=False, excludeChars=None):             # <<<<<<<<<<<<<<
 *         super(Char, self).__init__(charset, exact=1, asKeyword=asKeyword, excludeChars=excludeChars)
 *         self.reString = "[%s]" % _escapeRegexRangeChars(''.join(self.initChars))
 */
  __pyx_tuple__492 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_charset, __pyx_n_s_asKeyword, __pyx_n_s_excludeChars); if (unlikely(!__pyx_tuple__492)) __PYX_ERR(0, 3130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__492);
  __Pyx_GIVEREF(__pyx_tuple__492);
  __pyx_codeobj__493 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__492, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 3130, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__493)) __PYX_ERR(0, 3130, __pyx_L1_error)
  __pyx_tuple__494 = PyTuple_Pack(2, ((PyObject *)Py_False), ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__494)) __PYX_ERR(0, 3130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__494);
  __Pyx_GIVEREF(__pyx_tuple__494);

  /* "Aspidites/_vendor/pyparsing_extension.py":3164
 *     """
 * 
 *     def __init__(self, pattern, flags=0, asGroupList=False, asMatch=False):             # <<<<<<<<<<<<<<
 *         """The parameters ``pattern`` and ``flags`` are passed
 *         to the ``re.compile()`` function as-is. See the Python
 */
  __pyx_tuple__495 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_pattern, __pyx_n_s_flags, __pyx_n_s_asGroupList, __pyx_n_s_asMatch); if (unlikely(!__pyx_tuple__495)) __PYX_ERR(0, 3164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__495);
  __Pyx_GIVEREF(__pyx_tuple__495);
  __pyx_codeobj__496 = (PyObject*)__Pyx_PyCode_New(5, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__495, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 3164, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__496)) __PYX_ERR(0, 3164, __pyx_L1_error)
  __pyx_tuple__497 = PyTuple_Pack(3, ((PyObject *)__pyx_int_0), ((PyObject *)Py_False), ((PyObject *)Py_False)); if (unlikely(!__pyx_tuple__497)) __PYX_ERR(0, 3164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__497);
  __Pyx_GIVEREF(__pyx_tuple__497);

  /* "Aspidites/_vendor/pyparsing_extension.py":3209
 *             self.parseImpl = self.parseImplAsMatch
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         result = self.re_match(instring, loc)
 *         if not result:
 */
  __pyx_tuple__498 = PyTuple_Pack(9, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_doActions, __pyx_n_s_result, __pyx_n_s_ret, __pyx_n_s_d, __pyx_n_s_k, __pyx_n_s_v); if (unlikely(!__pyx_tuple__498)) __PYX_ERR(0, 3209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__498);
  __Pyx_GIVEREF(__pyx_tuple__498);
  __pyx_codeobj__499 = (PyObject*)__Pyx_PyCode_New(4, 0, 9, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__498, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_parseImpl, 3209, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__499)) __PYX_ERR(0, 3209, __pyx_L1_error)
  __pyx_tuple__500 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__500)) __PYX_ERR(0, 3209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__500);
  __Pyx_GIVEREF(__pyx_tuple__500);

  /* "Aspidites/_vendor/pyparsing_extension.py":3222
 *         return loc, ret
 * 
 *     def parseImplAsGroupList(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         result = self.re_match(instring, loc)
 *         if not result:
 */
  __pyx_tuple__501 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_doActions, __pyx_n_s_result, __pyx_n_s_ret); if (unlikely(!__pyx_tuple__501)) __PYX_ERR(0, 3222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__501);
  __Pyx_GIVEREF(__pyx_tuple__501);
  __pyx_codeobj__502 = (PyObject*)__Pyx_PyCode_New(4, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__501, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_parseImplAsGroupList, 3222, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__502)) __PYX_ERR(0, 3222, __pyx_L1_error)
  __pyx_tuple__503 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__503)) __PYX_ERR(0, 3222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__503);
  __Pyx_GIVEREF(__pyx_tuple__503);

  /* "Aspidites/_vendor/pyparsing_extension.py":3231
 *         return loc, ret
 * 
 *     def parseImplAsMatch(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         result = self.re_match(instring, loc)
 *         if not result:
 */
  __pyx_tuple__504 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_doActions, __pyx_n_s_result, __pyx_n_s_ret); if (unlikely(!__pyx_tuple__504)) __PYX_ERR(0, 3231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__504);
  __Pyx_GIVEREF(__pyx_tuple__504);
  __pyx_codeobj__505 = (PyObject*)__Pyx_PyCode_New(4, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__504, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_parseImplAsMatch, 3231, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__505)) __PYX_ERR(0, 3231, __pyx_L1_error)
  __pyx_tuple__506 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__506)) __PYX_ERR(0, 3231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__506);
  __Pyx_GIVEREF(__pyx_tuple__506);

  /* "Aspidites/_vendor/pyparsing_extension.py":3240
 *         return loc, ret
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         try:
 *             return super(Regex, self).__str__()
 */
  __pyx_tuple__507 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__507)) __PYX_ERR(0, 3240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__507);
  __Pyx_GIVEREF(__pyx_tuple__507);
  __pyx_codeobj__508 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__507, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_str, 3240, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__508)) __PYX_ERR(0, 3240, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3251
 *         return self.strRepr
 * 
 *     def sub(self, repl):             # <<<<<<<<<<<<<<
 *         r"""
 *         Return Regex with an attached parse action to transform the parsed
 */
  __pyx_tuple__509 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_repl, __pyx_n_s_pa, __pyx_n_s_pa, __pyx_n_s_pa); if (unlikely(!__pyx_tuple__509)) __PYX_ERR(0, 3251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__509);
  __Pyx_GIVEREF(__pyx_tuple__509);
  __pyx_codeobj__510 = (PyObject*)__Pyx_PyCode_New(2, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__509, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_sub, 3251, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__510)) __PYX_ERR(0, 3251, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3321
 *     """
 * 
 *     def __init__(self, quoteChar, escChar=None, escQuote=None, multiline=False,             # <<<<<<<<<<<<<<
 *                  unquoteResults=True, endQuoteChar=None, convertWhitespaceEscapes=True):
 *         super(QuotedString, self).__init__()
 */
  __pyx_tuple__511 = PyTuple_Pack(10, __pyx_n_s_self, __pyx_n_s_quoteChar, __pyx_n_s_escChar, __pyx_n_s_escQuote, __pyx_n_s_multiline, __pyx_n_s_unquoteResults, __pyx_n_s_endQuoteChar, __pyx_n_s_convertWhitespaceEscapes, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__511)) __PYX_ERR(0, 3321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__511);
  __Pyx_GIVEREF(__pyx_tuple__511);
  __pyx_codeobj__512 = (PyObject*)__Pyx_PyCode_New(8, 0, 10, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__511, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 3321, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__512)) __PYX_ERR(0, 3321, __pyx_L1_error)
  __pyx_tuple__513 = PyTuple_Pack(6, ((PyObject *)Py_None), ((PyObject *)Py_None), ((PyObject *)Py_False), ((PyObject *)Py_True), ((PyObject *)Py_None), ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__513)) __PYX_ERR(0, 3321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__513);
  __Pyx_GIVEREF(__pyx_tuple__513);

  /* "Aspidites/_vendor/pyparsing_extension.py":3386
 *         self.mayReturnEmpty = True
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         result = instring[loc] == self.firstQuoteChar and self.re_match(instring, loc) or None
 *         if not result:
 */
  __pyx_tuple__514 = PyTuple_Pack(9, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_doActions, __pyx_n_s_result, __pyx_n_s_ret, __pyx_n_s_ws_map, __pyx_n_s_wslit, __pyx_n_s_wschar); if (unlikely(!__pyx_tuple__514)) __PYX_ERR(0, 3386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__514);
  __Pyx_GIVEREF(__pyx_tuple__514);
  __pyx_codeobj__515 = (PyObject*)__Pyx_PyCode_New(4, 0, 9, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__514, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_parseImpl, 3386, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__515)) __PYX_ERR(0, 3386, __pyx_L1_error)
  __pyx_tuple__516 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__516)) __PYX_ERR(0, 3386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__516);
  __Pyx_GIVEREF(__pyx_tuple__516);

  /* "Aspidites/_vendor/pyparsing_extension.py":3421
 *         return loc, ret
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         try:
 *             return super(QuotedString, self).__str__()
 */
  __pyx_tuple__517 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__517)) __PYX_ERR(0, 3421, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__517);
  __Pyx_GIVEREF(__pyx_tuple__517);
  __pyx_codeobj__518 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__517, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_str, 3421, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__518)) __PYX_ERR(0, 3421, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3454
 *     """
 * 
 *     def __init__(self, notChars, min=1, max=0, exact=0):             # <<<<<<<<<<<<<<
 *         super(CharsNotIn, self).__init__()
 *         self.skipWhitespace = False
 */
  __pyx_tuple__519 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_notChars, __pyx_n_s_min, __pyx_n_s_max, __pyx_n_s_exact); if (unlikely(!__pyx_tuple__519)) __PYX_ERR(0, 3454, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__519);
  __Pyx_GIVEREF(__pyx_tuple__519);
  __pyx_codeobj__520 = (PyObject*)__Pyx_PyCode_New(5, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__519, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 3454, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__520)) __PYX_ERR(0, 3454, __pyx_L1_error)
  __pyx_tuple__521 = PyTuple_Pack(3, ((PyObject *)__pyx_int_1), ((PyObject *)__pyx_int_0), ((PyObject *)__pyx_int_0)); if (unlikely(!__pyx_tuple__521)) __PYX_ERR(0, 3454, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__521);
  __Pyx_GIVEREF(__pyx_tuple__521);

  /* "Aspidites/_vendor/pyparsing_extension.py":3479
 *         self.mayIndexError = False
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if instring[loc] in self.notChars:
 *             raise ParseException(instring, loc, self.errmsg, self)
 */
  __pyx_tuple__522 = PyTuple_Pack(7, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_doActions, __pyx_n_s_start, __pyx_n_s_notchars, __pyx_n_s_maxlen); if (unlikely(!__pyx_tuple__522)) __PYX_ERR(0, 3479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__522);
  __Pyx_GIVEREF(__pyx_tuple__522);
  __pyx_codeobj__523 = (PyObject*)__Pyx_PyCode_New(4, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__522, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_parseImpl, 3479, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__523)) __PYX_ERR(0, 3479, __pyx_L1_error)
  __pyx_tuple__524 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__524)) __PYX_ERR(0, 3479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__524);
  __Pyx_GIVEREF(__pyx_tuple__524);

  /* "Aspidites/_vendor/pyparsing_extension.py":3495
 *         return loc, instring[start:loc]
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         try:
 *             return super(CharsNotIn, self).__str__()
 */
  __pyx_tuple__525 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__525)) __PYX_ERR(0, 3495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__525);
  __Pyx_GIVEREF(__pyx_tuple__525);
  __pyx_codeobj__526 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__525, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_str, 3495, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__526)) __PYX_ERR(0, 3495, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3545
 *     }
 * 
 *     def __init__(self, ws=" \t\r\n", min=1, max=0, exact=0):             # <<<<<<<<<<<<<<
 *         super(White, self).__init__()
 *         self.matchWhite = ws
 */
  __pyx_tuple__545 = PyTuple_Pack(8, __pyx_n_s_self, __pyx_n_s_ws, __pyx_n_s_min, __pyx_n_s_max, __pyx_n_s_exact, __pyx_n_s_genexpr, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__545)) __PYX_ERR(0, 3545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__545);
  __Pyx_GIVEREF(__pyx_tuple__545);
  __pyx_codeobj__546 = (PyObject*)__Pyx_PyCode_New(5, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__545, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 3545, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__546)) __PYX_ERR(0, 3545, __pyx_L1_error)
  __pyx_tuple__547 = PyTuple_Pack(4, ((PyObject*)__pyx_kp_s__105), ((PyObject *)__pyx_int_1), ((PyObject *)__pyx_int_0), ((PyObject *)__pyx_int_0)); if (unlikely(!__pyx_tuple__547)) __PYX_ERR(0, 3545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__547);
  __Pyx_GIVEREF(__pyx_tuple__547);

  /* "Aspidites/_vendor/pyparsing_extension.py":3565
 *             self.minLen = exact
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if instring[loc] not in self.matchWhite:
 *             raise ParseException(instring, loc, self.errmsg, self)
 */
  __pyx_tuple__548 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_doActions, __pyx_n_s_start, __pyx_n_s_maxloc); if (unlikely(!__pyx_tuple__548)) __PYX_ERR(0, 3565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__548);
  __Pyx_GIVEREF(__pyx_tuple__548);
  __pyx_codeobj__549 = (PyObject*)__Pyx_PyCode_New(4, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__548, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_parseImpl, 3565, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__549)) __PYX_ERR(0, 3565, __pyx_L1_error)
  __pyx_tuple__550 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__550)) __PYX_ERR(0, 3565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__550);
  __Pyx_GIVEREF(__pyx_tuple__550);

  /* "Aspidites/_vendor/pyparsing_extension.py":3582
 * 
 * class _PositionToken(Token):
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         super(_PositionToken, self).__init__()
 *         self.name = self.__class__.__name__
 */
  __pyx_tuple__551 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__551)) __PYX_ERR(0, 3582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__551);
  __Pyx_GIVEREF(__pyx_tuple__551);
  __pyx_codeobj__552 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__551, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 3582, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__552)) __PYX_ERR(0, 3582, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3594
 *     """
 * 
 *     def __init__(self, colno):             # <<<<<<<<<<<<<<
 *         super(GoToColumn, self).__init__()
 *         self.col = colno
 */
  __pyx_tuple__553 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_colno); if (unlikely(!__pyx_tuple__553)) __PYX_ERR(0, 3594, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__553);
  __Pyx_GIVEREF(__pyx_tuple__553);
  __pyx_codeobj__554 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__553, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 3594, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__554)) __PYX_ERR(0, 3594, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3598
 *         self.col = colno
 * 
 *     def preParse(self, instring, loc):             # <<<<<<<<<<<<<<
 *         if col(loc, instring) != self.col:
 *             instrlen = len(instring)
 */
  __pyx_tuple__555 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_instrlen); if (unlikely(!__pyx_tuple__555)) __PYX_ERR(0, 3598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__555);
  __Pyx_GIVEREF(__pyx_tuple__555);
  __pyx_codeobj__556 = (PyObject*)__Pyx_PyCode_New(3, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__555, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_preParse, 3598, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__556)) __PYX_ERR(0, 3598, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3607
 *         return loc
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         thiscol = col(loc, instring)
 *         if thiscol > self.col:
 */
  __pyx_tuple__557 = PyTuple_Pack(7, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_doActions, __pyx_n_s_thiscol, __pyx_n_s_newloc, __pyx_n_s_ret); if (unlikely(!__pyx_tuple__557)) __PYX_ERR(0, 3607, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__557);
  __Pyx_GIVEREF(__pyx_tuple__557);
  __pyx_codeobj__558 = (PyObject*)__Pyx_PyCode_New(4, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__557, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_parseImpl, 3607, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__558)) __PYX_ERR(0, 3607, __pyx_L1_error)
  __pyx_tuple__559 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__559)) __PYX_ERR(0, 3607, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__559);
  __Pyx_GIVEREF(__pyx_tuple__559);

  /* "Aspidites/_vendor/pyparsing_extension.py":3639
 *     """
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         super(LineStart, self).__init__()
 *         self.errmsg = "Expected start of line"
 */
  __pyx_tuple__560 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__560)) __PYX_ERR(0, 3639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__560);
  __Pyx_GIVEREF(__pyx_tuple__560);
  __pyx_codeobj__561 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__560, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 3639, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__561)) __PYX_ERR(0, 3639, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3643
 *         self.errmsg = "Expected start of line"
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if col(loc, instring) == 1:
 *             return loc, []
 */
  __pyx_tuple__562 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_doActions); if (unlikely(!__pyx_tuple__562)) __PYX_ERR(0, 3643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__562);
  __Pyx_GIVEREF(__pyx_tuple__562);
  __pyx_codeobj__563 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__562, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_parseImpl, 3643, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__563)) __PYX_ERR(0, 3643, __pyx_L1_error)
  __pyx_tuple__564 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__564)) __PYX_ERR(0, 3643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__564);
  __Pyx_GIVEREF(__pyx_tuple__564);

  /* "Aspidites/_vendor/pyparsing_extension.py":3654
 *     """
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         super(LineEnd, self).__init__()
 *         self.setWhitespaceChars(ParserElement.DEFAULT_WHITE_CHARS.replace("\n", ""))
 */
  __pyx_tuple__565 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__565)) __PYX_ERR(0, 3654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__565);
  __Pyx_GIVEREF(__pyx_tuple__565);
  __pyx_codeobj__566 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__565, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 3654, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__566)) __PYX_ERR(0, 3654, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3659
 *         self.errmsg = "Expected end of line"
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if loc < len(instring):
 *             if instring[loc] == "\n":
 */
  __pyx_tuple__567 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_doActions); if (unlikely(!__pyx_tuple__567)) __PYX_ERR(0, 3659, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__567);
  __Pyx_GIVEREF(__pyx_tuple__567);
  __pyx_codeobj__568 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__567, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_parseImpl, 3659, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__568)) __PYX_ERR(0, 3659, __pyx_L1_error)
  __pyx_tuple__569 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__569)) __PYX_ERR(0, 3659, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__569);
  __Pyx_GIVEREF(__pyx_tuple__569);

  /* "Aspidites/_vendor/pyparsing_extension.py":3676
 *     """
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         super(StringStart, self).__init__()
 *         self.errmsg = "Expected start of text"
 */
  __pyx_tuple__570 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__570)) __PYX_ERR(0, 3676, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__570);
  __Pyx_GIVEREF(__pyx_tuple__570);
  __pyx_codeobj__571 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__570, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 3676, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__571)) __PYX_ERR(0, 3676, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3680
 *         self.errmsg = "Expected start of text"
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if loc != 0:
 *             # see if entire string up to here is just whitespace and ignoreables
 */
  __pyx_tuple__572 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_doActions); if (unlikely(!__pyx_tuple__572)) __PYX_ERR(0, 3680, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__572);
  __Pyx_GIVEREF(__pyx_tuple__572);
  __pyx_codeobj__573 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__572, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_parseImpl, 3680, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__573)) __PYX_ERR(0, 3680, __pyx_L1_error)
  __pyx_tuple__574 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__574)) __PYX_ERR(0, 3680, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__574);
  __Pyx_GIVEREF(__pyx_tuple__574);

  /* "Aspidites/_vendor/pyparsing_extension.py":3692
 *     """
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         super(StringEnd, self).__init__()
 *         self.errmsg = "Expected end of text"
 */
  __pyx_tuple__575 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__575)) __PYX_ERR(0, 3692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__575);
  __Pyx_GIVEREF(__pyx_tuple__575);
  __pyx_codeobj__576 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__575, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 3692, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__576)) __PYX_ERR(0, 3692, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3696
 *         self.errmsg = "Expected end of text"
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if loc < len(instring):
 *             raise ParseException(instring, loc, self.errmsg, self)
 */
  __pyx_tuple__577 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_doActions); if (unlikely(!__pyx_tuple__577)) __PYX_ERR(0, 3696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__577);
  __Pyx_GIVEREF(__pyx_tuple__577);
  __pyx_codeobj__578 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__577, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_parseImpl, 3696, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__578)) __PYX_ERR(0, 3696, __pyx_L1_error)
  __pyx_tuple__579 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__579)) __PYX_ERR(0, 3696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__579);
  __Pyx_GIVEREF(__pyx_tuple__579);

  /* "Aspidites/_vendor/pyparsing_extension.py":3717
 *     """
 * 
 *     def __init__(self, wordChars=printables):             # <<<<<<<<<<<<<<
 *         super(WordStart, self).__init__()
 *         self.wordChars = set(wordChars)
 */
  __pyx_tuple__580 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_wordChars); if (unlikely(!__pyx_tuple__580)) __PYX_ERR(0, 3717, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__580);
  __Pyx_GIVEREF(__pyx_tuple__580);
  __pyx_codeobj__581 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__580, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 3717, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__581)) __PYX_ERR(0, 3717, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3722
 *         self.errmsg = "Not at the start of a word"
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if loc != 0:
 *             if (instring[loc - 1] in self.wordChars
 */
  __pyx_tuple__582 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_doActions); if (unlikely(!__pyx_tuple__582)) __PYX_ERR(0, 3722, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__582);
  __Pyx_GIVEREF(__pyx_tuple__582);
  __pyx_codeobj__583 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__582, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_parseImpl, 3722, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__583)) __PYX_ERR(0, 3722, __pyx_L1_error)
  __pyx_tuple__584 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__584)) __PYX_ERR(0, 3722, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__584);
  __Pyx_GIVEREF(__pyx_tuple__584);

  /* "Aspidites/_vendor/pyparsing_extension.py":3739
 *     """
 * 
 *     def __init__(self, wordChars=printables):             # <<<<<<<<<<<<<<
 *         super(WordEnd, self).__init__()
 *         self.wordChars = set(wordChars)
 */
  __pyx_tuple__585 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_wordChars); if (unlikely(!__pyx_tuple__585)) __PYX_ERR(0, 3739, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__585);
  __Pyx_GIVEREF(__pyx_tuple__585);
  __pyx_codeobj__586 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__585, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 3739, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__586)) __PYX_ERR(0, 3739, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3745
 *         self.errmsg = "Not at the end of a word"
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         instrlen = len(instring)
 *         if instrlen > 0 and loc < instrlen:
 */
  __pyx_tuple__587 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_doActions, __pyx_n_s_instrlen); if (unlikely(!__pyx_tuple__587)) __PYX_ERR(0, 3745, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__587);
  __Pyx_GIVEREF(__pyx_tuple__587);
  __pyx_codeobj__588 = (PyObject*)__Pyx_PyCode_New(4, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__587, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_parseImpl, 3745, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__588)) __PYX_ERR(0, 3745, __pyx_L1_error)
  __pyx_tuple__589 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__589)) __PYX_ERR(0, 3745, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__589);
  __Pyx_GIVEREF(__pyx_tuple__589);

  /* "Aspidites/_vendor/pyparsing_extension.py":3773
 * 
 *     class _ErrorStop(Empty):
 *         def __init__(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *             super(And._ErrorStop, self).__init__(*args, **kwargs)
 *             self.name = '-'
 */
  __pyx_tuple__590 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_args, __pyx_n_s_kwargs); if (unlikely(!__pyx_tuple__590)) __PYX_ERR(0, 3773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__590);
  __Pyx_GIVEREF(__pyx_tuple__590);
  __pyx_codeobj__591 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__590, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 3773, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__591)) __PYX_ERR(0, 3773, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3778
 *             self.leaveWhitespace()
 * 
 *     def __init__(self, exprs, savelist=True):             # <<<<<<<<<<<<<<
 *         exprs = list(exprs)
 *         if exprs and Ellipsis in exprs:
 */
  __pyx_tuple__592 = PyTuple_Pack(9, __pyx_n_s_self, __pyx_n_s_exprs, __pyx_n_s_savelist, __pyx_n_s_tmp, __pyx_n_s_i, __pyx_n_s_expr, __pyx_n_s_skipto_arg, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__592)) __PYX_ERR(0, 3778, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__592);
  __Pyx_GIVEREF(__pyx_tuple__592);
  __pyx_codeobj__593 = (PyObject*)__Pyx_PyCode_New(3, 0, 9, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__592, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 3778, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__593)) __PYX_ERR(0, 3778, __pyx_L1_error)
  __pyx_tuple__594 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__594)) __PYX_ERR(0, 3778, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__594);
  __Pyx_GIVEREF(__pyx_tuple__594);

  /* "Aspidites/_vendor/pyparsing_extension.py":3798
 *         self.callPreparse = True
 * 
 *     def streamline(self):             # <<<<<<<<<<<<<<
 *         # collapse any _PendingSkip's
 *         from .pyparsing import _PendingSkip
 */
  __pyx_tuple__595 = PyTuple_Pack(7, __pyx_n_s_self, __pyx_n_s_PendingSkip, __pyx_n_s_i, __pyx_n_s_e, __pyx_n_s_genexpr, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__595)) __PYX_ERR(0, 3798, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__595);
  __Pyx_GIVEREF(__pyx_tuple__595);
  __pyx_codeobj__596 = (PyObject*)__Pyx_PyCode_New(1, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__595, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_streamline, 3798, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__596)) __PYX_ERR(0, 3798, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3817
 *         return self
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         # pass False as last arg to _parse for first element, since we already
 *         # pre-parsed the string as part of our And pre-parsing
 */
  __pyx_tuple__597 = PyTuple_Pack(9, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_doActions, __pyx_n_s_resultlist, __pyx_n_s_errorStop, __pyx_n_s_e, __pyx_n_s_exprtokens, __pyx_n_s_pe); if (unlikely(!__pyx_tuple__597)) __PYX_ERR(0, 3817, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__597);
  __Pyx_GIVEREF(__pyx_tuple__597);
  __pyx_codeobj__598 = (PyObject*)__Pyx_PyCode_New(4, 0, 9, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__597, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_parseImpl, 3817, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__598)) __PYX_ERR(0, 3817, __pyx_L1_error)
  __pyx_tuple__599 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__599)) __PYX_ERR(0, 3817, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__599);
  __Pyx_GIVEREF(__pyx_tuple__599);

  /* "Aspidites/_vendor/pyparsing_extension.py":3842
 *         return loc, resultlist
 * 
 *     def __iadd__(self, other):             # <<<<<<<<<<<<<<
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 */
  __pyx_tuple__600 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_other); if (unlikely(!__pyx_tuple__600)) __PYX_ERR(0, 3842, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__600);
  __Pyx_GIVEREF(__pyx_tuple__600);
  __pyx_codeobj__601 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__600, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_iadd, 3842, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__601)) __PYX_ERR(0, 3842, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3847
 *         return self.append(other)  # And([self, other])
 * 
 *     def checkRecursion(self, parseElementList):             # <<<<<<<<<<<<<<
 *         subRecCheckList = parseElementList[:] + [self]
 *         for e in self.exprs:
 */
  __pyx_tuple__602 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parseElementList, __pyx_n_s_subRecCheckList, __pyx_n_s_e); if (unlikely(!__pyx_tuple__602)) __PYX_ERR(0, 3847, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__602);
  __Pyx_GIVEREF(__pyx_tuple__602);
  __pyx_codeobj__603 = (PyObject*)__Pyx_PyCode_New(2, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__602, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_checkRecursion, 3847, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__603)) __PYX_ERR(0, 3847, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3854
 *                 break
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */
  __pyx_tuple__604 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__604)) __PYX_ERR(0, 3854, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__604);
  __Pyx_GIVEREF(__pyx_tuple__604);
  __pyx_codeobj__605 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__604, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_str, 3854, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__605)) __PYX_ERR(0, 3854, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3882
 *     """
 * 
 *     def __init__(self, exprs, savelist=False):             # <<<<<<<<<<<<<<
 *         super(Or, self).__init__(exprs, savelist)
 *         if self.exprs:
 */
  __pyx_tuple__606 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_exprs, __pyx_n_s_savelist, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__606)) __PYX_ERR(0, 3882, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__606);
  __Pyx_GIVEREF(__pyx_tuple__606);
  __pyx_codeobj__607 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__606, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 3882, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__607)) __PYX_ERR(0, 3882, __pyx_L1_error)
  __pyx_tuple__608 = PyTuple_Pack(1, ((PyObject *)Py_False)); if (unlikely(!__pyx_tuple__608)) __PYX_ERR(0, 3882, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__608);
  __Pyx_GIVEREF(__pyx_tuple__608);

  /* "Aspidites/_vendor/pyparsing_extension.py":3889
 *             self.mayReturnEmpty = True
 * 
 *     def streamline(self):             # <<<<<<<<<<<<<<
 *         super(Or, self).streamline()
 *         if __diag__.collect_all_And_tokens:
 */
  __pyx_tuple__609 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__609)) __PYX_ERR(0, 3889, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__609);
  __Pyx_GIVEREF(__pyx_tuple__609);
  __pyx_codeobj__610 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__609, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_streamline, 3889, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__610)) __PYX_ERR(0, 3889, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3895
 *         return self
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         maxExcLoc = -1
 *         maxException = None
 */
  __pyx_tuple__611 = PyTuple_Pack(15, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_doActions, __pyx_n_s_maxExcLoc, __pyx_n_s_maxException, __pyx_n_s_matches, __pyx_n_s_e, __pyx_n_s_loc2, __pyx_n_s_err, __pyx_n_s_best_expr, __pyx_n_s_longest, __pyx_n_s_loc1, __pyx_n_s_expr1, __pyx_n_s_toks); if (unlikely(!__pyx_tuple__611)) __PYX_ERR(0, 3895, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__611);
  __Pyx_GIVEREF(__pyx_tuple__611);
  __pyx_codeobj__612 = (PyObject*)__Pyx_PyCode_New(4, 0, 15, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__611, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_parseImpl, 3895, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__612)) __PYX_ERR(0, 3895, __pyx_L1_error)
  __pyx_tuple__613 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__613)) __PYX_ERR(0, 3895, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__613);
  __Pyx_GIVEREF(__pyx_tuple__613);

  /* "Aspidites/_vendor/pyparsing_extension.py":3955
 *             raise ParseException(instring, loc, "no defined alternatives to match", self)
 * 
 *     def __ixor__(self, other):             # <<<<<<<<<<<<<<
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 */
  __pyx_tuple__614 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_other); if (unlikely(!__pyx_tuple__614)) __PYX_ERR(0, 3955, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__614);
  __Pyx_GIVEREF(__pyx_tuple__614);
  __pyx_codeobj__615 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__614, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_ixor, 3955, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__615)) __PYX_ERR(0, 3955, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3960
 *         return self.append(other)  # Or([self, other])
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */
  __pyx_tuple__616 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__616)) __PYX_ERR(0, 3960, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__616);
  __Pyx_GIVEREF(__pyx_tuple__616);
  __pyx_codeobj__617 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__616, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_str, 3960, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__617)) __PYX_ERR(0, 3960, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3969
 *         return self.strRepr
 * 
 *     def checkRecursion(self, parseElementList):             # <<<<<<<<<<<<<<
 *         subRecCheckList = parseElementList[:] + [self]
 *         for e in self.exprs:
 */
  __pyx_tuple__618 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parseElementList, __pyx_n_s_subRecCheckList, __pyx_n_s_e); if (unlikely(!__pyx_tuple__618)) __PYX_ERR(0, 3969, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__618);
  __Pyx_GIVEREF(__pyx_tuple__618);
  __pyx_codeobj__619 = (PyObject*)__Pyx_PyCode_New(2, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__618, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_checkRecursion, 3969, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__619)) __PYX_ERR(0, 3969, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":3974
 *             e.checkRecursion(subRecCheckList)
 * 
 *     def _setResultsName(self, name, listAllMatches=False):             # <<<<<<<<<<<<<<
 *         if (not __diag__.collect_all_And_tokens
 *                 and __diag__.warn_multiple_tokens_in_named_alternation):
 */
  __pyx_tuple__620 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_name, __pyx_n_s_listAllMatches, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__620)) __PYX_ERR(0, 3974, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__620);
  __Pyx_GIVEREF(__pyx_tuple__620);
  __pyx_codeobj__621 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__620, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_setResultsName, 3974, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__621)) __PYX_ERR(0, 3974, __pyx_L1_error)
  __pyx_tuple__622 = PyTuple_Pack(1, ((PyObject *)Py_False)); if (unlikely(!__pyx_tuple__622)) __PYX_ERR(0, 3974, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__622);
  __Pyx_GIVEREF(__pyx_tuple__622);

  /* "Aspidites/_vendor/pyparsing_extension.py":4005
 *     """
 * 
 *     def __init__(self, exprs, savelist=False):             # <<<<<<<<<<<<<<
 *         super(MatchFirst, self).__init__(exprs, savelist)
 *         if self.exprs:
 */
  __pyx_tuple__623 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_exprs, __pyx_n_s_savelist, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__623)) __PYX_ERR(0, 4005, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__623);
  __Pyx_GIVEREF(__pyx_tuple__623);
  __pyx_codeobj__624 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__623, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 4005, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__624)) __PYX_ERR(0, 4005, __pyx_L1_error)
  __pyx_tuple__625 = PyTuple_Pack(1, ((PyObject *)Py_False)); if (unlikely(!__pyx_tuple__625)) __PYX_ERR(0, 4005, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__625);
  __Pyx_GIVEREF(__pyx_tuple__625);

  /* "Aspidites/_vendor/pyparsing_extension.py":4012
 *             self.mayReturnEmpty = True
 * 
 *     def streamline(self):             # <<<<<<<<<<<<<<
 *         super(MatchFirst, self).streamline()
 *         if __diag__.collect_all_And_tokens:
 */
  __pyx_tuple__626 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__626)) __PYX_ERR(0, 4012, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__626);
  __Pyx_GIVEREF(__pyx_tuple__626);
  __pyx_codeobj__627 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__626, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_streamline, 4012, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__627)) __PYX_ERR(0, 4012, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4018
 *         return self
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         maxExcLoc = -1
 *         maxException = None
 */
  __pyx_tuple__628 = PyTuple_Pack(9, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_doActions, __pyx_n_s_maxExcLoc, __pyx_n_s_maxException, __pyx_n_s_e, __pyx_n_s_ret, __pyx_n_s_err); if (unlikely(!__pyx_tuple__628)) __PYX_ERR(0, 4018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__628);
  __Pyx_GIVEREF(__pyx_tuple__628);
  __pyx_codeobj__629 = (PyObject*)__Pyx_PyCode_New(4, 0, 9, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__628, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_parseImpl, 4018, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__629)) __PYX_ERR(0, 4018, __pyx_L1_error)
  __pyx_tuple__630 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__630)) __PYX_ERR(0, 4018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__630);
  __Pyx_GIVEREF(__pyx_tuple__630);

  /* "Aspidites/_vendor/pyparsing_extension.py":4042
 *                 raise ParseException(instring, loc, "no defined alternatives to match", self)
 * 
 *     def __ior__(self, other):             # <<<<<<<<<<<<<<
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 */
  __pyx_tuple__631 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_other); if (unlikely(!__pyx_tuple__631)) __PYX_ERR(0, 4042, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__631);
  __Pyx_GIVEREF(__pyx_tuple__631);
  __pyx_codeobj__632 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__631, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_ior, 4042, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__632)) __PYX_ERR(0, 4042, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4047
 *         return self.append(other)  # MatchFirst([self, other])
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */
  __pyx_tuple__633 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__633)) __PYX_ERR(0, 4047, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__633);
  __Pyx_GIVEREF(__pyx_tuple__633);
  __pyx_codeobj__634 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__633, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_str, 4047, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__634)) __PYX_ERR(0, 4047, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4056
 *         return self.strRepr
 * 
 *     def checkRecursion(self, parseElementList):             # <<<<<<<<<<<<<<
 *         subRecCheckList = parseElementList[:] + [self]
 *         for e in self.exprs:
 */
  __pyx_tuple__635 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parseElementList, __pyx_n_s_subRecCheckList, __pyx_n_s_e); if (unlikely(!__pyx_tuple__635)) __PYX_ERR(0, 4056, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__635);
  __Pyx_GIVEREF(__pyx_tuple__635);
  __pyx_codeobj__636 = (PyObject*)__Pyx_PyCode_New(2, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__635, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_checkRecursion, 4056, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__636)) __PYX_ERR(0, 4056, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4061
 *             e.checkRecursion(subRecCheckList)
 * 
 *     def _setResultsName(self, name, listAllMatches=False):             # <<<<<<<<<<<<<<
 *         if (not __diag__.collect_all_And_tokens
 *                 and __diag__.warn_multiple_tokens_in_named_alternation):
 */
  __pyx_tuple__637 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_name, __pyx_n_s_listAllMatches, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__637)) __PYX_ERR(0, 4061, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__637);
  __Pyx_GIVEREF(__pyx_tuple__637);
  __pyx_codeobj__638 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__637, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_setResultsName, 4061, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__638)) __PYX_ERR(0, 4061, __pyx_L1_error)
  __pyx_tuple__639 = PyTuple_Pack(1, ((PyObject *)Py_False)); if (unlikely(!__pyx_tuple__639)) __PYX_ERR(0, 4061, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__639);
  __Pyx_GIVEREF(__pyx_tuple__639);

  /* "Aspidites/_vendor/pyparsing_extension.py":4132
 *     """
 * 
 *     def __init__(self, exprs, savelist=True):             # <<<<<<<<<<<<<<
 *         super(Each, self).__init__(exprs, savelist)
 *         self.mayReturnEmpty = all(e.mayReturnEmpty for e in self.exprs)
 */
  __pyx_tuple__640 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_exprs, __pyx_n_s_savelist, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__640)) __PYX_ERR(0, 4132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__640);
  __Pyx_GIVEREF(__pyx_tuple__640);
  __pyx_codeobj__641 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__640, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 4132, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__641)) __PYX_ERR(0, 4132, __pyx_L1_error)
  __pyx_tuple__642 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__642)) __PYX_ERR(0, 4132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__642);
  __Pyx_GIVEREF(__pyx_tuple__642);

  /* "Aspidites/_vendor/pyparsing_extension.py":4139
 *         self.saveAsList = True
 * 
 *     def streamline(self):             # <<<<<<<<<<<<<<
 *         super(Each, self).streamline()
 *         self.mayReturnEmpty = all(e.mayReturnEmpty for e in self.exprs)
 */
  __pyx_tuple__643 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__643)) __PYX_ERR(0, 4139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__643);
  __Pyx_GIVEREF(__pyx_tuple__643);
  __pyx_codeobj__644 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__643, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_streamline, 4139, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__644)) __PYX_ERR(0, 4139, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4144
 *         return self
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if self.initExprGroups:
 *             self.opt1map = dict((id(e.expr), e) for e in self.exprs if isinstance(e, Optional))
 */
  __pyx_tuple__645 = PyTuple_Pack(21, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_doActions, __pyx_n_s_opt1, __pyx_n_s_opt2, __pyx_n_s_tmpLoc, __pyx_n_s_tmpReqd, __pyx_n_s_tmpOpt, __pyx_n_s_matchOrder, __pyx_n_s_keepMatching, __pyx_n_s_tmpExprs, __pyx_n_s_failed, __pyx_n_s_e, __pyx_n_s_missing, __pyx_n_s_resultlist, __pyx_n_s_results, __pyx_n_s_finalResults, __pyx_n_s_genexpr, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__645)) __PYX_ERR(0, 4144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__645);
  __Pyx_GIVEREF(__pyx_tuple__645);
  __pyx_codeobj__646 = (PyObject*)__Pyx_PyCode_New(4, 0, 21, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__645, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_parseImpl, 4144, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__646)) __PYX_ERR(0, 4144, __pyx_L1_error)
  __pyx_tuple__647 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__647)) __PYX_ERR(0, 4144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__647);
  __Pyx_GIVEREF(__pyx_tuple__647);

  /* "Aspidites/_vendor/pyparsing_extension.py":4193
 *         return loc, finalResults
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */
  __pyx_tuple__648 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__648)) __PYX_ERR(0, 4193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__648);
  __Pyx_GIVEREF(__pyx_tuple__648);
  __pyx_codeobj__649 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__648, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_str, 4193, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__649)) __PYX_ERR(0, 4193, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4202
 *         return self.strRepr
 * 
 *     def checkRecursion(self, parseElementList):             # <<<<<<<<<<<<<<
 *         subRecCheckList = parseElementList[:] + [self]
 *         for e in self.exprs:
 */
  __pyx_tuple__650 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_parseElementList, __pyx_n_s_subRecCheckList, __pyx_n_s_e); if (unlikely(!__pyx_tuple__650)) __PYX_ERR(0, 4202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__650);
  __Pyx_GIVEREF(__pyx_tuple__650);
  __pyx_codeobj__651 = (PyObject*)__Pyx_PyCode_New(2, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__650, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_checkRecursion, 4202, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__651)) __PYX_ERR(0, 4202, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4213
 *     """
 * 
 *     def __init__(self, expr, savelist=False):             # <<<<<<<<<<<<<<
 *         super(ParseElementEnhance, self).__init__(savelist)
 *         if isinstance(expr, str):
 */
  __pyx_tuple__652 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_expr, __pyx_n_s_savelist); if (unlikely(!__pyx_tuple__652)) __PYX_ERR(0, 4213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__652);
  __Pyx_GIVEREF(__pyx_tuple__652);
  __pyx_codeobj__653 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__652, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 4213, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__653)) __PYX_ERR(0, 4213, __pyx_L1_error)
  __pyx_tuple__654 = PyTuple_Pack(1, ((PyObject *)Py_False)); if (unlikely(!__pyx_tuple__654)) __PYX_ERR(0, 4213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__654);
  __Pyx_GIVEREF(__pyx_tuple__654);

  /* "Aspidites/_vendor/pyparsing_extension.py":4231
 *             self.ignoreExprs.extend(expr.ignoreExprs)
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if self.expr is not None:
 *             return self.expr._parse(instring, loc, doActions, callPreParse=False)
 */
  __pyx_tuple__655 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_doActions); if (unlikely(!__pyx_tuple__655)) __PYX_ERR(0, 4231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__655);
  __Pyx_GIVEREF(__pyx_tuple__655);
  __pyx_codeobj__656 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__655, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_parseImpl, 4231, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__656)) __PYX_ERR(0, 4231, __pyx_L1_error)
  __pyx_tuple__657 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__657)) __PYX_ERR(0, 4231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__657);
  __Pyx_GIVEREF(__pyx_tuple__657);

  /* "Aspidites/_vendor/pyparsing_extension.py":4237
 *             raise ParseException("", loc, self.errmsg, self)
 * 
 *     def leaveWhitespace(self):             # <<<<<<<<<<<<<<
 *         self.skipWhitespace = False
 *         self.expr = self.expr.copy()
 */
  __pyx_tuple__658 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__658)) __PYX_ERR(0, 4237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__658);
  __Pyx_GIVEREF(__pyx_tuple__658);
  __pyx_codeobj__659 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__658, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_leaveWhitespace, 4237, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__659)) __PYX_ERR(0, 4237, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4244
 *         return self
 * 
 *     def ignore(self, other):             # <<<<<<<<<<<<<<
 *         if isinstance(other, Suppress):
 *             if other not in self.ignoreExprs:
 */
  __pyx_tuple__660 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_other); if (unlikely(!__pyx_tuple__660)) __PYX_ERR(0, 4244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__660);
  __Pyx_GIVEREF(__pyx_tuple__660);
  __pyx_codeobj__661 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__660, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_ignore, 4244, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__661)) __PYX_ERR(0, 4244, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4256
 *         return self
 * 
 *     def streamline(self):             # <<<<<<<<<<<<<<
 *         super(ParseElementEnhance, self).streamline()
 *         if self.expr is not None:
 */
  __pyx_tuple__662 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__662)) __PYX_ERR(0, 4256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__662);
  __Pyx_GIVEREF(__pyx_tuple__662);
  __pyx_codeobj__663 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__662, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_streamline, 4256, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__663)) __PYX_ERR(0, 4256, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4262
 *         return self
 * 
 *     def checkRecursion(self, parseElementList):             # <<<<<<<<<<<<<<
 *         if self in parseElementList:
 *             raise RecursiveGrammarException(parseElementList + [self])
 */
  __pyx_tuple__664 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_parseElementList, __pyx_n_s_subRecCheckList); if (unlikely(!__pyx_tuple__664)) __PYX_ERR(0, 4262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__664);
  __Pyx_GIVEREF(__pyx_tuple__664);
  __pyx_codeobj__665 = (PyObject*)__Pyx_PyCode_New(2, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__664, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_checkRecursion, 4262, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__665)) __PYX_ERR(0, 4262, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4269
 *             self.expr.checkRecursion(subRecCheckList)
 * 
 *     def validate(self, validateTrace=None):             # <<<<<<<<<<<<<<
 *         if validateTrace is None:
 *             validateTrace = []
 */
  __pyx_tuple__666 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_validateTrace, __pyx_n_s_tmp); if (unlikely(!__pyx_tuple__666)) __PYX_ERR(0, 4269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__666);
  __Pyx_GIVEREF(__pyx_tuple__666);
  __pyx_codeobj__667 = (PyObject*)__Pyx_PyCode_New(2, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__666, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_validate, 4269, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__667)) __PYX_ERR(0, 4269, __pyx_L1_error)
  __pyx_tuple__668 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__668)) __PYX_ERR(0, 4269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__668);
  __Pyx_GIVEREF(__pyx_tuple__668);

  /* "Aspidites/_vendor/pyparsing_extension.py":4277
 *         self.checkRecursion([])
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         try:
 *             return super(ParseElementEnhance, self).__str__()
 */
  __pyx_tuple__669 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__669)) __PYX_ERR(0, 4277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__669);
  __Pyx_GIVEREF(__pyx_tuple__669);
  __pyx_codeobj__670 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__669, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_str, 4277, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__670)) __PYX_ERR(0, 4277, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4311
 *     """
 * 
 *     def __init__(self, expr):             # <<<<<<<<<<<<<<
 *         super(FollowedBy, self).__init__(expr)
 *         self.mayReturnEmpty = True
 */
  __pyx_tuple__671 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_expr); if (unlikely(!__pyx_tuple__671)) __PYX_ERR(0, 4311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__671);
  __Pyx_GIVEREF(__pyx_tuple__671);
  __pyx_codeobj__672 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__671, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 4311, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__672)) __PYX_ERR(0, 4311, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4315
 *         self.mayReturnEmpty = True
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         # by using self._expr.parse and deleting the contents of the returned ParseResults list
 *         # we keep any named results that were defined in the FollowedBy expression
 */
  __pyx_tuple__674 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_doActions, __pyx_n_s__673, __pyx_n_s_ret); if (unlikely(!__pyx_tuple__674)) __PYX_ERR(0, 4315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__674);
  __Pyx_GIVEREF(__pyx_tuple__674);
  __pyx_codeobj__675 = (PyObject*)__Pyx_PyCode_New(4, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__674, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_parseImpl, 4315, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__675)) __PYX_ERR(0, 4315, __pyx_L1_error)
  __pyx_tuple__676 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__676)) __PYX_ERR(0, 4315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__676);
  __Pyx_GIVEREF(__pyx_tuple__676);

  /* "Aspidites/_vendor/pyparsing_extension.py":4353
 *     """
 * 
 *     def __init__(self, expr, retreat=None):             # <<<<<<<<<<<<<<
 *         super(PrecededBy, self).__init__(expr)
 *         self.expr = self.expr().leaveWhitespace()
 */
  __pyx_tuple__677 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_expr, __pyx_n_s_retreat); if (unlikely(!__pyx_tuple__677)) __PYX_ERR(0, 4353, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__677);
  __Pyx_GIVEREF(__pyx_tuple__677);
  __pyx_codeobj__678 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__677, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 4353, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__678)) __PYX_ERR(0, 4353, __pyx_L1_error)
  __pyx_tuple__679 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__679)) __PYX_ERR(0, 4353, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__679);
  __Pyx_GIVEREF(__pyx_tuple__679);

  /* "Aspidites/_vendor/pyparsing_extension.py":4376
 *         self.parseAction.append(lambda s, l, t: t.__delitem__(slice(None, None)))
 * 
 *     def parseImpl(self, instring, loc=0, doActions=True):             # <<<<<<<<<<<<<<
 *         if self.exact:
 *             if loc < self.retreat:
 */
  __pyx_tuple__680 = PyTuple_Pack(12, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_doActions, __pyx_n_s_start, __pyx_n_s__673, __pyx_n_s_ret, __pyx_n_s_test_expr, __pyx_n_s_instring_slice, __pyx_n_s_last_expr, __pyx_n_s_offset, __pyx_n_s_pbe); if (unlikely(!__pyx_tuple__680)) __PYX_ERR(0, 4376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__680);
  __Pyx_GIVEREF(__pyx_tuple__680);
  __pyx_codeobj__681 = (PyObject*)__Pyx_PyCode_New(4, 0, 12, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__680, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_parseImpl, 4376, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__681)) __PYX_ERR(0, 4376, __pyx_L1_error)
  __pyx_tuple__682 = PyTuple_Pack(2, ((PyObject *)__pyx_int_0), ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__682)) __PYX_ERR(0, 4376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__682);
  __Pyx_GIVEREF(__pyx_tuple__682);

  /* "Aspidites/_vendor/pyparsing_extension.py":4424
 *     """
 * 
 *     def __init__(self, expr):             # <<<<<<<<<<<<<<
 *         super(NotAny, self).__init__(expr)
 *         # ~ self.leaveWhitespace()
 */
  __pyx_tuple__683 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_expr); if (unlikely(!__pyx_tuple__683)) __PYX_ERR(0, 4424, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__683);
  __Pyx_GIVEREF(__pyx_tuple__683);
  __pyx_codeobj__684 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__683, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 4424, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__684)) __PYX_ERR(0, 4424, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4431
 *         self.errmsg = "Found unwanted token, " + str(self.expr)
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if self.expr.canParseNext(instring, loc):
 *             raise ParseException(instring, loc, self.errmsg, self)
 */
  __pyx_tuple__685 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_doActions); if (unlikely(!__pyx_tuple__685)) __PYX_ERR(0, 4431, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__685);
  __Pyx_GIVEREF(__pyx_tuple__685);
  __pyx_codeobj__686 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__685, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_parseImpl, 4431, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__686)) __PYX_ERR(0, 4431, __pyx_L1_error)
  __pyx_tuple__687 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__687)) __PYX_ERR(0, 4431, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__687);
  __Pyx_GIVEREF(__pyx_tuple__687);

  /* "Aspidites/_vendor/pyparsing_extension.py":4436
 *         return loc, []
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */
  __pyx_tuple__688 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__688)) __PYX_ERR(0, 4436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__688);
  __Pyx_GIVEREF(__pyx_tuple__688);
  __pyx_codeobj__689 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__688, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_str, 4436, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__689)) __PYX_ERR(0, 4436, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4447
 * 
 * class _MultipleMatch(ParseElementEnhance):
 *     def __init__(self, expr, stopOn=None):             # <<<<<<<<<<<<<<
 *         super(_MultipleMatch, self).__init__(expr)
 *         self.saveAsList = True
 */
  __pyx_tuple__690 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_expr, __pyx_n_s_stopOn, __pyx_n_s_ender); if (unlikely(!__pyx_tuple__690)) __PYX_ERR(0, 4447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__690);
  __Pyx_GIVEREF(__pyx_tuple__690);
  __pyx_codeobj__691 = (PyObject*)__Pyx_PyCode_New(3, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__690, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 4447, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__691)) __PYX_ERR(0, 4447, __pyx_L1_error)
  __pyx_tuple__692 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__692)) __PYX_ERR(0, 4447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__692);
  __Pyx_GIVEREF(__pyx_tuple__692);

  /* "Aspidites/_vendor/pyparsing_extension.py":4455
 *         self.stopOn(ender)
 * 
 *     def stopOn(self, ender):             # <<<<<<<<<<<<<<
 *         if isinstance(ender, str):
 *             ender = self._literalStringClass(ender)
 */
  __pyx_tuple__693 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ender); if (unlikely(!__pyx_tuple__693)) __PYX_ERR(0, 4455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__693);
  __Pyx_GIVEREF(__pyx_tuple__693);
  __pyx_codeobj__694 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__693, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_stopOn, 4455, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__694)) __PYX_ERR(0, 4455, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4461
 *         return self
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         self_expr_parse = self.expr._parse
 *         self_skip_ignorables = self._skipIgnorables
 */
  __pyx_tuple__695 = PyTuple_Pack(12, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_doActions, __pyx_n_s_self_expr_parse, __pyx_n_s_self_skip_ignorables, __pyx_n_s_check_ender, __pyx_n_s_try_not_ender, __pyx_n_s_tokens, __pyx_n_s_hasIgnoreExprs, __pyx_n_s_preloc, __pyx_n_s_tmptokens); if (unlikely(!__pyx_tuple__695)) __PYX_ERR(0, 4461, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__695);
  __Pyx_GIVEREF(__pyx_tuple__695);
  __pyx_codeobj__696 = (PyObject*)__Pyx_PyCode_New(4, 0, 12, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__695, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_parseImpl, 4461, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__696)) __PYX_ERR(0, 4461, __pyx_L1_error)
  __pyx_tuple__697 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__697)) __PYX_ERR(0, 4461, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__697);
  __Pyx_GIVEREF(__pyx_tuple__697);

  /* "Aspidites/_vendor/pyparsing_extension.py":4490
 *         return loc, tokens
 * 
 *     def _setResultsName(self, name, listAllMatches=False):             # <<<<<<<<<<<<<<
 *         if __diag__.warn_ungrouped_named_tokens_in_collection:
 *             for e in [self.expr] + getattr(self.expr, 'exprs', []):
 */
  __pyx_tuple__698 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_name, __pyx_n_s_listAllMatches, __pyx_n_s_e); if (unlikely(!__pyx_tuple__698)) __PYX_ERR(0, 4490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__698);
  __Pyx_GIVEREF(__pyx_tuple__698);
  __pyx_codeobj__699 = (PyObject*)__Pyx_PyCode_New(3, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__698, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_setResultsName, 4490, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__699)) __PYX_ERR(0, 4490, __pyx_L1_error)
  __pyx_tuple__700 = PyTuple_Pack(1, ((PyObject *)Py_False)); if (unlikely(!__pyx_tuple__700)) __PYX_ERR(0, 4490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__700);
  __Pyx_GIVEREF(__pyx_tuple__700);

  /* "Aspidites/_vendor/pyparsing_extension.py":4531
 *     """
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */
  __pyx_tuple__701 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__701)) __PYX_ERR(0, 4531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__701);
  __Pyx_GIVEREF(__pyx_tuple__701);
  __pyx_codeobj__702 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__701, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_str, 4531, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__702)) __PYX_ERR(0, 4531, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4553
 *     """
 * 
 *     def __init__(self, expr, stopOn=None):             # <<<<<<<<<<<<<<
 *         super(ZeroOrMore, self).__init__(expr, stopOn=stopOn)
 *         self.mayReturnEmpty = True
 */
  __pyx_tuple__703 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_expr, __pyx_n_s_stopOn); if (unlikely(!__pyx_tuple__703)) __PYX_ERR(0, 4553, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__703);
  __Pyx_GIVEREF(__pyx_tuple__703);
  __pyx_codeobj__704 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__703, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 4553, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__704)) __PYX_ERR(0, 4553, __pyx_L1_error)
  __pyx_tuple__705 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__705)) __PYX_ERR(0, 4553, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__705);
  __Pyx_GIVEREF(__pyx_tuple__705);

  /* "Aspidites/_vendor/pyparsing_extension.py":4557
 *         self.mayReturnEmpty = True
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         try:
 *             return super(ZeroOrMore, self).parseImpl(instring, loc, doActions)
 */
  __pyx_tuple__706 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_doActions); if (unlikely(!__pyx_tuple__706)) __PYX_ERR(0, 4557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__706);
  __Pyx_GIVEREF(__pyx_tuple__706);
  __pyx_codeobj__707 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__706, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_parseImpl, 4557, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__707)) __PYX_ERR(0, 4557, __pyx_L1_error)
  __pyx_tuple__708 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__708)) __PYX_ERR(0, 4557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__708);
  __Pyx_GIVEREF(__pyx_tuple__708);

  /* "Aspidites/_vendor/pyparsing_extension.py":4563
 *             return loc, []
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */
  __pyx_tuple__709 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__709)) __PYX_ERR(0, 4563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__709);
  __Pyx_GIVEREF(__pyx_tuple__709);
  __pyx_codeobj__710 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__709, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_str, 4563, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__710)) __PYX_ERR(0, 4563, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4573
 * 
 * 
 * class _NullToken(object):             # <<<<<<<<<<<<<<
 *     def __bool__(self):
 *         return False
 */
  __pyx_tuple__711 = PyTuple_Pack(1, __pyx_builtin_object); if (unlikely(!__pyx_tuple__711)) __PYX_ERR(0, 4573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__711);
  __Pyx_GIVEREF(__pyx_tuple__711);

  /* "Aspidites/_vendor/pyparsing_extension.py":4574
 * 
 * class _NullToken(object):
 *     def __bool__(self):             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
  __pyx_tuple__712 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__712)) __PYX_ERR(0, 4574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__712);
  __Pyx_GIVEREF(__pyx_tuple__712);
  __pyx_codeobj__713 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__712, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_bool, 4574, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__713)) __PYX_ERR(0, 4574, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4579
 *     __nonzero__ = __bool__
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return ""
 * 
 */
  __pyx_tuple__714 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__714)) __PYX_ERR(0, 4579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__714);
  __Pyx_GIVEREF(__pyx_tuple__714);
  __pyx_codeobj__715 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__714, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_str, 4579, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__715)) __PYX_ERR(0, 4579, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4622
 *     __optionalNotMatched = _NullToken()
 * 
 *     def __init__(self, expr, default=__optionalNotMatched):             # <<<<<<<<<<<<<<
 *         super(Optional, self).__init__(expr, savelist=False)
 *         self.saveAsList = self.expr.saveAsList
 */
  __pyx_tuple__716 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_expr, __pyx_n_s_default); if (unlikely(!__pyx_tuple__716)) __PYX_ERR(0, 4622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__716);
  __Pyx_GIVEREF(__pyx_tuple__716);
  __pyx_codeobj__717 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__716, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 4622, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__717)) __PYX_ERR(0, 4622, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4628
 *         self.mayReturnEmpty = True
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         try:
 *             loc, tokens = self.expr._parse(instring, loc, doActions, callPreParse=False)
 */
  __pyx_tuple__718 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_doActions, __pyx_n_s_tokens); if (unlikely(!__pyx_tuple__718)) __PYX_ERR(0, 4628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__718);
  __Pyx_GIVEREF(__pyx_tuple__718);
  __pyx_codeobj__719 = (PyObject*)__Pyx_PyCode_New(4, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__718, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_parseImpl, 4628, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__719)) __PYX_ERR(0, 4628, __pyx_L1_error)
  __pyx_tuple__720 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__720)) __PYX_ERR(0, 4628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__720);
  __Pyx_GIVEREF(__pyx_tuple__720);

  /* "Aspidites/_vendor/pyparsing_extension.py":4642
 *         return loc, tokens
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */
  __pyx_tuple__721 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__721)) __PYX_ERR(0, 4642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__721);
  __Pyx_GIVEREF(__pyx_tuple__721);
  __pyx_codeobj__722 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__721, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_str, 4642, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__722)) __PYX_ERR(0, 4642, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4711
 *     """
 * 
 *     def __init__(self, other, include=False, ignore=None, failOn=None):             # <<<<<<<<<<<<<<
 *         super(SkipTo, self).__init__(other)
 *         self.ignoreExpr = ignore
 */
  __pyx_tuple__723 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_other, __pyx_n_s_include, __pyx_n_s_ignore, __pyx_n_s_failOn); if (unlikely(!__pyx_tuple__723)) __PYX_ERR(0, 4711, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__723);
  __Pyx_GIVEREF(__pyx_tuple__723);
  __pyx_codeobj__724 = (PyObject*)__Pyx_PyCode_New(5, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__723, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 4711, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__724)) __PYX_ERR(0, 4711, __pyx_L1_error)
  __pyx_tuple__725 = PyTuple_Pack(3, ((PyObject *)Py_False), ((PyObject *)Py_None), ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__725)) __PYX_ERR(0, 4711, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__725);
  __Pyx_GIVEREF(__pyx_tuple__725);

  /* "Aspidites/_vendor/pyparsing_extension.py":4724
 *         self.errmsg = "No match found for " + str(self.expr)
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         startloc = loc
 *         instrlen = len(instring)
 */
  __pyx_tuple__726 = PyTuple_Pack(14, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_doActions, __pyx_n_s_startloc, __pyx_n_s_instrlen, __pyx_n_s_expr, __pyx_n_s_expr_parse, __pyx_n_s_self_failOn_canParseNext, __pyx_n_s_self_ignoreExpr_tryParse, __pyx_n_s_tmploc, __pyx_n_s_skiptext, __pyx_n_s_skipresult, __pyx_n_s_mat); if (unlikely(!__pyx_tuple__726)) __PYX_ERR(0, 4724, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__726);
  __Pyx_GIVEREF(__pyx_tuple__726);
  __pyx_codeobj__727 = (PyObject*)__Pyx_PyCode_New(4, 0, 14, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__726, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_parseImpl, 4724, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__727)) __PYX_ERR(0, 4724, __pyx_L1_error)
  __pyx_tuple__728 = PyTuple_Pack(1, ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__728)) __PYX_ERR(0, 4724, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__728);
  __Pyx_GIVEREF(__pyx_tuple__728);

  /* "Aspidites/_vendor/pyparsing_extension.py":4800
 *     """
 * 
 *     def __init__(self, other=None):             # <<<<<<<<<<<<<<
 *         super(Forward, self).__init__(other, savelist=False)
 * 
 */
  __pyx_tuple__729 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_other); if (unlikely(!__pyx_tuple__729)) __PYX_ERR(0, 4800, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__729);
  __Pyx_GIVEREF(__pyx_tuple__729);
  __pyx_codeobj__730 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__729, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 4800, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__730)) __PYX_ERR(0, 4800, __pyx_L1_error)
  __pyx_tuple__731 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__731)) __PYX_ERR(0, 4800, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__731);
  __Pyx_GIVEREF(__pyx_tuple__731);

  /* "Aspidites/_vendor/pyparsing_extension.py":4803
 *         super(Forward, self).__init__(other, savelist=False)
 * 
 *     def __lshift__(self, other):             # <<<<<<<<<<<<<<
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 */
  __pyx_tuple__732 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_other); if (unlikely(!__pyx_tuple__732)) __PYX_ERR(0, 4803, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__732);
  __Pyx_GIVEREF(__pyx_tuple__732);
  __pyx_codeobj__733 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__732, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_lshift, 4803, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__733)) __PYX_ERR(0, 4803, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4816
 *         return self
 * 
 *     def __ilshift__(self, other):             # <<<<<<<<<<<<<<
 *         return self << other
 * 
 */
  __pyx_tuple__734 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_other); if (unlikely(!__pyx_tuple__734)) __PYX_ERR(0, 4816, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__734);
  __Pyx_GIVEREF(__pyx_tuple__734);
  __pyx_codeobj__735 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__734, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_ilshift, 4816, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__735)) __PYX_ERR(0, 4816, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4819
 *         return self << other
 * 
 *     def leaveWhitespace(self):             # <<<<<<<<<<<<<<
 *         self.skipWhitespace = False
 *         return self
 */
  __pyx_tuple__736 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__736)) __PYX_ERR(0, 4819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__736);
  __Pyx_GIVEREF(__pyx_tuple__736);
  __pyx_codeobj__737 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__736, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_leaveWhitespace, 4819, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__737)) __PYX_ERR(0, 4819, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4823
 *         return self
 * 
 *     def streamline(self):             # <<<<<<<<<<<<<<
 *         if not self.streamlined:
 *             self.streamlined = True
 */
  __pyx_tuple__738 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__738)) __PYX_ERR(0, 4823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__738);
  __Pyx_GIVEREF(__pyx_tuple__738);
  __pyx_codeobj__739 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__738, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_streamline, 4823, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__739)) __PYX_ERR(0, 4823, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4830
 *         return self
 * 
 *     def validate(self, validateTrace=None):             # <<<<<<<<<<<<<<
 *         if validateTrace is None:
 *             validateTrace = []
 */
  __pyx_tuple__740 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_validateTrace, __pyx_n_s_tmp); if (unlikely(!__pyx_tuple__740)) __PYX_ERR(0, 4830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__740);
  __Pyx_GIVEREF(__pyx_tuple__740);
  __pyx_codeobj__741 = (PyObject*)__Pyx_PyCode_New(2, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__740, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_validate, 4830, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__741)) __PYX_ERR(0, 4830, __pyx_L1_error)
  __pyx_tuple__742 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__742)) __PYX_ERR(0, 4830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__742);
  __Pyx_GIVEREF(__pyx_tuple__742);

  /* "Aspidites/_vendor/pyparsing_extension.py":4840
 *         self.checkRecursion([])
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */
  __pyx_tuple__743 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_retString); if (unlikely(!__pyx_tuple__743)) __PYX_ERR(0, 4840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__743);
  __Pyx_GIVEREF(__pyx_tuple__743);
  __pyx_codeobj__744 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__743, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_str, 4840, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__744)) __PYX_ERR(0, 4840, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4860
 *         return self.strRepr
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         if self.expr is not None:
 *             return super(Forward, self).copy()
 */
  __pyx_tuple__745 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ret); if (unlikely(!__pyx_tuple__745)) __PYX_ERR(0, 4860, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__745);
  __Pyx_GIVEREF(__pyx_tuple__745);
  __pyx_codeobj__746 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__745, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_copy, 4860, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__746)) __PYX_ERR(0, 4860, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4868
 *             return ret
 * 
 *     def _setResultsName(self, name, listAllMatches=False):             # <<<<<<<<<<<<<<
 * 
 *         if __diag__.warn_name_set_on_empty_Forward:
 */
  __pyx_tuple__747 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_name, __pyx_n_s_listAllMatches); if (unlikely(!__pyx_tuple__747)) __PYX_ERR(0, 4868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__747);
  __Pyx_GIVEREF(__pyx_tuple__747);
  __pyx_codeobj__748 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__747, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_setResultsName, 4868, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__748)) __PYX_ERR(0, 4868, __pyx_L1_error)
  __pyx_tuple__749 = PyTuple_Pack(1, ((PyObject *)Py_False)); if (unlikely(!__pyx_tuple__749)) __PYX_ERR(0, 4868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__749);
  __Pyx_GIVEREF(__pyx_tuple__749);

  /* "Aspidites/_vendor/pyparsing_extension.py":4886
 *     """
 * 
 *     def __init__(self, expr, savelist=False):             # <<<<<<<<<<<<<<
 *         super(TokenConverter, self).__init__(expr)  # , savelist)
 *         self.saveAsList = False
 */
  __pyx_tuple__750 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_expr, __pyx_n_s_savelist); if (unlikely(!__pyx_tuple__750)) __PYX_ERR(0, 4886, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__750);
  __Pyx_GIVEREF(__pyx_tuple__750);
  __pyx_codeobj__751 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__750, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 4886, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__751)) __PYX_ERR(0, 4886, __pyx_L1_error)
  __pyx_tuple__752 = PyTuple_Pack(1, ((PyObject *)Py_False)); if (unlikely(!__pyx_tuple__752)) __PYX_ERR(0, 4886, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__752);
  __Pyx_GIVEREF(__pyx_tuple__752);

  /* "Aspidites/_vendor/pyparsing_extension.py":4910
 *     """
 * 
 *     def __init__(self, expr, joinString="", adjacent=True):             # <<<<<<<<<<<<<<
 *         super(Combine, self).__init__(expr)
 *         # suppress whitespace-stripping in contained parse expressions, but re-enable it on the Combine itself
 */
  __pyx_tuple__753 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_expr, __pyx_n_s_joinString, __pyx_n_s_adjacent); if (unlikely(!__pyx_tuple__753)) __PYX_ERR(0, 4910, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__753);
  __Pyx_GIVEREF(__pyx_tuple__753);
  __pyx_codeobj__754 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__753, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 4910, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__754)) __PYX_ERR(0, 4910, __pyx_L1_error)
  __pyx_tuple__755 = PyTuple_Pack(2, ((PyObject*)__pyx_kp_s__12), ((PyObject *)Py_True)); if (unlikely(!__pyx_tuple__755)) __PYX_ERR(0, 4910, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__755);
  __Pyx_GIVEREF(__pyx_tuple__755);

  /* "Aspidites/_vendor/pyparsing_extension.py":4920
 *         self.callPreparse = True
 * 
 *     def ignore(self, other):             # <<<<<<<<<<<<<<
 *         if self.adjacent:
 *             ParserElement.ignore(self, other)
 */
  __pyx_tuple__756 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_other); if (unlikely(!__pyx_tuple__756)) __PYX_ERR(0, 4920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__756);
  __Pyx_GIVEREF(__pyx_tuple__756);
  __pyx_codeobj__757 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__756, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_ignore, 4920, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__757)) __PYX_ERR(0, 4920, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4927
 *         return self
 * 
 *     def postParse(self, instring, loc, tokenlist):             # <<<<<<<<<<<<<<
 *         retToks = tokenlist.copy()
 *         del retToks[:]
 */
  __pyx_tuple__758 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_tokenlist, __pyx_n_s_retToks); if (unlikely(!__pyx_tuple__758)) __PYX_ERR(0, 4927, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__758);
  __Pyx_GIVEREF(__pyx_tuple__758);
  __pyx_codeobj__759 = (PyObject*)__Pyx_PyCode_New(4, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__758, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_postParse, 4927, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__759)) __PYX_ERR(0, 4927, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4954
 *     """
 * 
 *     def __init__(self, expr):             # <<<<<<<<<<<<<<
 *         super(Group, self).__init__(expr)
 *         self.saveAsList = True
 */
  __pyx_tuple__760 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_expr); if (unlikely(!__pyx_tuple__760)) __PYX_ERR(0, 4954, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__760);
  __Pyx_GIVEREF(__pyx_tuple__760);
  __pyx_codeobj__761 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__760, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 4954, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__761)) __PYX_ERR(0, 4954, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":4958
 *         self.saveAsList = True
 * 
 *     def postParse(self, instring, loc, tokenlist):             # <<<<<<<<<<<<<<
 *         return [tokenlist]
 * 
 */
  __pyx_tuple__762 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_tokenlist); if (unlikely(!__pyx_tuple__762)) __PYX_ERR(0, 4958, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__762);
  __Pyx_GIVEREF(__pyx_tuple__762);
  __pyx_codeobj__763 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__762, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_postParse, 4958, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__763)) __PYX_ERR(0, 4958, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":5002
 *     """
 * 
 *     def __init__(self, expr):             # <<<<<<<<<<<<<<
 *         super(Dict, self).__init__(expr)
 *         self.saveAsList = True
 */
  __pyx_tuple__764 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_expr); if (unlikely(!__pyx_tuple__764)) __PYX_ERR(0, 5002, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__764);
  __Pyx_GIVEREF(__pyx_tuple__764);
  __pyx_codeobj__765 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__764, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 5002, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__765)) __PYX_ERR(0, 5002, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":5006
 *         self.saveAsList = True
 * 
 *     def postParse(self, instring, loc, tokenlist):             # <<<<<<<<<<<<<<
 *         for i, tok in enumerate(tokenlist):
 *             if len(tok) == 0:
 */
  __pyx_tuple__766 = PyTuple_Pack(8, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_tokenlist, __pyx_n_s_i, __pyx_n_s_tok, __pyx_n_s_ikey, __pyx_n_s_dictvalue); if (unlikely(!__pyx_tuple__766)) __PYX_ERR(0, 5006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__766);
  __Pyx_GIVEREF(__pyx_tuple__766);
  __pyx_codeobj__767 = (PyObject*)__Pyx_PyCode_New(4, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__766, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_postParse, 5006, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__767)) __PYX_ERR(0, 5006, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":5054
 *     """
 * 
 *     def postParse(self, instring, loc, tokenlist):             # <<<<<<<<<<<<<<
 *         return []
 * 
 */
  __pyx_tuple__768 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_instring, __pyx_n_s_loc, __pyx_n_s_tokenlist); if (unlikely(!__pyx_tuple__768)) __PYX_ERR(0, 5054, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__768);
  __Pyx_GIVEREF(__pyx_tuple__768);
  __pyx_codeobj__769 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__768, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_postParse, 5054, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__769)) __PYX_ERR(0, 5054, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":5057
 *         return []
 * 
 *     def suppress(self):             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __pyx_tuple__770 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__770)) __PYX_ERR(0, 5057, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__770);
  __Pyx_GIVEREF(__pyx_tuple__770);
  __pyx_codeobj__771 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__770, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_suppress, 5057, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__771)) __PYX_ERR(0, 5057, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":5061
 * 
 * 
 * class OnlyOnce(object):             # <<<<<<<<<<<<<<
 *     """Wrapper for parse actions, to ensure they are only called once.
 *     """
 */
  __pyx_tuple__772 = PyTuple_Pack(1, __pyx_builtin_object); if (unlikely(!__pyx_tuple__772)) __PYX_ERR(0, 5061, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__772);
  __Pyx_GIVEREF(__pyx_tuple__772);

  /* "Aspidites/_vendor/pyparsing_extension.py":5064
 *     """Wrapper for parse actions, to ensure they are only called once.
 *     """
 *     def __init__(self, methodCall):             # <<<<<<<<<<<<<<
 *         from .pyparsing import _trim_arity
 *         self.callable = _trim_arity(methodCall)
 */
  __pyx_tuple__773 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_methodCall, __pyx_n_s_trim_arity); if (unlikely(!__pyx_tuple__773)) __PYX_ERR(0, 5064, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__773);
  __Pyx_GIVEREF(__pyx_tuple__773);
  __pyx_codeobj__774 = (PyObject*)__Pyx_PyCode_New(2, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__773, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_init, 5064, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__774)) __PYX_ERR(0, 5064, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":5069
 *         self.called = False
 * 
 *     def __call__(self, s, l, t):             # <<<<<<<<<<<<<<
 *         if not self.called:
 *             results = self.callable(s, l, t)
 */
  __pyx_tuple__775 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_s_4, __pyx_n_s_l, __pyx_n_s_t_2, __pyx_n_s_results); if (unlikely(!__pyx_tuple__775)) __PYX_ERR(0, 5069, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__775);
  __Pyx_GIVEREF(__pyx_tuple__775);
  __pyx_codeobj__776 = (PyObject*)__Pyx_PyCode_New(4, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__775, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_call, 5069, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__776)) __PYX_ERR(0, 5069, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":5076
 *         raise ParseException(s, l, "")
 * 
 *     def reset(self):             # <<<<<<<<<<<<<<
 *         self.called = False
 */
  __pyx_tuple__777 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__777)) __PYX_ERR(0, 5076, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__777);
  __Pyx_GIVEREF(__pyx_tuple__777);
  __pyx_codeobj__778 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__777, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Aspidites__vendor_pyparsing_exte_2, __pyx_n_s_reset, 5076, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__778)) __PYX_ERR(0, 5076, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  __pyx_umethod_PyDict_Type_get.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyDict_Type_items.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyDict_Type_pop.type = (PyObject*)&PyDict_Type;
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4 = PyInt_FromLong(4); if (unlikely(!__pyx_int_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_5 = PyInt_FromLong(5); if (unlikely(!__pyx_int_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_16 = PyInt_FromLong(16); if (unlikely(!__pyx_int_16)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_92 = PyInt_FromLong(92); if (unlikely(!__pyx_int_92)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_128 = PyInt_FromLong(128); if (unlikely(!__pyx_int_128)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1000 = PyInt_FromLong(1000); if (unlikely(!__pyx_int_1000)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct__genexpr) < 0) __PYX_ERR(0, 13, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct__genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct__genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct__genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct__genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct__genexpr = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct__genexpr;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr) < 0) __PYX_ERR(0, 43, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_1_genexpr;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr) < 0) __PYX_ERR(0, 119, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_2_genexpr;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues) < 0) __PYX_ERR(0, 336, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_3__itervalues;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr) < 0) __PYX_ERR(0, 337, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_4_genexpr;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems) < 0) __PYX_ERR(0, 339, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_5__iteritems;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr) < 0) __PYX_ERR(0, 340, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_6_genexpr;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_7___iadd__) < 0) __PYX_ERR(0, 530, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_7___iadd__.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_7___iadd__.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_7___iadd__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_7___iadd__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_7___iadd__ = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_7___iadd__;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__) < 0) __PYX_ERR(0, 557, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__ = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_8___str__;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr) < 0) __PYX_ERR(0, 558, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_9_genexpr;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict) < 0) __PYX_ERR(0, 588, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_10_asDict;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr) < 0) __PYX_ERR(0, 622, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_11_genexpr;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML) < 0) __PYX_ERR(0, 635, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_12_asXML;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr) < 0) __PYX_ERR(0, 641, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_13_genexpr;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump) < 0) __PYX_ERR(0, 743, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_14_dump;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr) < 0) __PYX_ERR(0, 773, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_15_genexpr;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr) < 0) __PYX_ERR(0, 785, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_16_genexpr;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak) < 0) __PYX_ERR(0, 1238, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_17_setBreak;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction) < 0) __PYX_ERR(0, 1259, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_18_setParseAction;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr) < 0) __PYX_ERR(0, 1303, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_19_genexpr;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__) < 0) __PYX_ERR(0, 1494, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__ = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_20___init__;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__) < 0) __PYX_ERR(0, 1517, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__ = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_21___init__;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__) < 0) __PYX_ERR(0, 1546, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__ = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_22___init__;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString) < 0) __PYX_ERR(0, 1706, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_23_scanString;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split) < 0) __PYX_ERR(0, 1860, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_24_split;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__) < 0) __PYX_ERR(0, 1965, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__ = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_25___mul__;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr) < 0) __PYX_ERR(0, 1994, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_26_genexpr;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__) < 0) __PYX_ERR(0, 2557, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__ = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_27___init__;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr) < 0) __PYX_ERR(0, 2569, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_28_genexpr;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr) < 0) __PYX_ERR(0, 2570, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_29_genexpr;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__) < 0) __PYX_ERR(0, 3006, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__ = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_30___init__;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr) < 0) __PYX_ERR(0, 3010, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_31_genexpr;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr) < 0) __PYX_ERR(0, 3012, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_32_genexpr;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub) < 0) __PYX_ERR(0, 3251, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_33_sub;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__) < 0) __PYX_ERR(0, 3321, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__ = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_34___init__;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr) < 0) __PYX_ERR(0, 3361, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_35_genexpr;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__) < 0) __PYX_ERR(0, 3545, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__ = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_36___init__;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr) < 0) __PYX_ERR(0, 3548, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_37_genexpr;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr) < 0) __PYX_ERR(0, 3550, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_38_genexpr;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__) < 0) __PYX_ERR(0, 3778, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__ = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_39___init__;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr) < 0) __PYX_ERR(0, 3793, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_40_genexpr;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline) < 0) __PYX_ERR(0, 3798, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_41_streamline;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr) < 0) __PYX_ERR(0, 3802, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_42_genexpr;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr) < 0) __PYX_ERR(0, 3814, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_43_genexpr;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__) < 0) __PYX_ERR(0, 3854, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__ = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_44___str__;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr) < 0) __PYX_ERR(0, 3859, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_45_genexpr;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__) < 0) __PYX_ERR(0, 3882, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__ = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_46___init__;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr) < 0) __PYX_ERR(0, 3885, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_47_genexpr;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline) < 0) __PYX_ERR(0, 3889, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_48_streamline;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr) < 0) __PYX_ERR(0, 3892, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_49_genexpr;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__) < 0) __PYX_ERR(0, 3960, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__ = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_50___str__;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr) < 0) __PYX_ERR(0, 3965, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_51_genexpr;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName) < 0) __PYX_ERR(0, 3974, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_52__setResultsName;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr) < 0) __PYX_ERR(0, 3977, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_53_genexpr;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__) < 0) __PYX_ERR(0, 4005, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__ = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_54___init__;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr) < 0) __PYX_ERR(0, 4008, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_55_genexpr;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline) < 0) __PYX_ERR(0, 4012, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_56_streamline;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr) < 0) __PYX_ERR(0, 4015, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_57_genexpr;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__) < 0) __PYX_ERR(0, 4047, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__ = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_58___str__;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr) < 0) __PYX_ERR(0, 4052, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_59_genexpr;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName) < 0) __PYX_ERR(0, 4061, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_60__setResultsName;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr) < 0) __PYX_ERR(0, 4064, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_61_genexpr;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__) < 0) __PYX_ERR(0, 4132, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__ = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_62___init__;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr) < 0) __PYX_ERR(0, 4134, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_63_genexpr;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline) < 0) __PYX_ERR(0, 4139, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_64_streamline;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr) < 0) __PYX_ERR(0, 4141, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_65_genexpr;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl) < 0) __PYX_ERR(0, 4144, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_66_parseImpl;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr) < 0) __PYX_ERR(0, 4146, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_67_genexpr;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr) < 0) __PYX_ERR(0, 4179, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_68_genexpr;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__) < 0) __PYX_ERR(0, 4193, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__ = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_69___str__;
  if (PyType_Ready(&__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr) < 0) __PYX_ERR(0, 4198, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr.tp_dictoffset && __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr = &__pyx_type_9Aspidites_7_vendor_19pyparsing_extension___pyx_scope_struct_70_genexpr;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initpyparsing_extension(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initpyparsing_extension(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_pyparsing_extension(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_pyparsing_extension(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_pyparsing_extension(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  Py_ssize_t __pyx_t_14;
  PyObject *(*__pyx_t_15)(PyObject *);
  int __pyx_t_16;
  int __pyx_t_17;
  PyObject *__pyx_t_18 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'pyparsing_extension' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_pyparsing_extension(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("pyparsing_extension", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_Aspidites___vendor__pyparsing_extension) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name_2, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "Aspidites._vendor.pyparsing_extension")) {
      if (unlikely(PyDict_SetItemString(modules, "Aspidites._vendor.pyparsing_extension", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_type_import_code();
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "Aspidites/_vendor/pyparsing_extension.py":1
 * import collections             # <<<<<<<<<<<<<<
 * import copy
 * import pprint
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_collections, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_collections, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2
 * import collections
 * import copy             # <<<<<<<<<<<<<<
 * import pprint
 * import re
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_copy, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_copy, __pyx_t_1) < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3
 * import collections
 * import copy
 * import pprint             # <<<<<<<<<<<<<<
 * import re
 * import string
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_pprint, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pprint, __pyx_t_1) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4
 * import copy
 * import pprint
 * import re             # <<<<<<<<<<<<<<
 * import string
 * import sys
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_re, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_re, __pyx_t_1) < 0) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":5
 * import pprint
 * import re
 * import string             # <<<<<<<<<<<<<<
 * import sys
 * import types
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_string, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_string, __pyx_t_1) < 0) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":6
 * import re
 * import string
 * import sys             # <<<<<<<<<<<<<<
 * import types
 * import warnings
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_sys, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_sys, __pyx_t_1) < 0) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":7
 * import string
 * import sys
 * import types             # <<<<<<<<<<<<<<
 * import warnings
 * from operator import itemgetter
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_types, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_types, __pyx_t_1) < 0) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":8
 * import sys
 * import types
 * import warnings             # <<<<<<<<<<<<<<
 * from operator import itemgetter
 * from weakref import ref as wkref
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_warnings, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_warnings, __pyx_t_1) < 0) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":9
 * import types
 * import warnings
 * from operator import itemgetter             # <<<<<<<<<<<<<<
 * from weakref import ref as wkref
 * import sre_constants
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_itemgetter);
  __Pyx_GIVEREF(__pyx_n_s_itemgetter);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_itemgetter);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_operator, __pyx_t_1, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_itemgetter); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_itemgetter, __pyx_t_1) < 0) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":10
 * import warnings
 * from operator import itemgetter
 * from weakref import ref as wkref             # <<<<<<<<<<<<<<
 * import sre_constants
 * 
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_ref);
  __Pyx_GIVEREF(__pyx_n_s_ref);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_ref);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_weakref, __pyx_t_2, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_ref); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_wkref, __pyx_t_2) < 0) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":11
 * from operator import itemgetter
 * from weakref import ref as wkref
 * import sre_constants             # <<<<<<<<<<<<<<
 * 
 * _generatorType = type((y for y in range(1)))
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_sre_constants, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_sre_constants, __pyx_t_1) < 0) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":13
 * import sre_constants
 * 
 * _generatorType = type((y for y in range(1)))             # <<<<<<<<<<<<<<
 * 
 * from collections.abc import Iterable
 */
  __pyx_t_1 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_22genexpr(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_generatorType, ((PyObject *)Py_TYPE(__pyx_t_1))) < 0) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":15
 * _generatorType = type((y for y in range(1)))
 * 
 * from collections.abc import Iterable             # <<<<<<<<<<<<<<
 * from collections.abc import MutableMapping, Mapping
 * 
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_Iterable);
  __Pyx_GIVEREF(__pyx_n_s_Iterable);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_Iterable);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_collections_abc, __pyx_t_1, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_Iterable); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Iterable, __pyx_t_1) < 0) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":16
 * 
 * from collections.abc import Iterable
 * from collections.abc import MutableMapping, Mapping             # <<<<<<<<<<<<<<
 * 
 * system_version = tuple(sys.version_info)[:3]
 */
  __pyx_t_2 = PyList_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_MutableMapping);
  __Pyx_GIVEREF(__pyx_n_s_MutableMapping);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_MutableMapping);
  __Pyx_INCREF(__pyx_n_s_Mapping);
  __Pyx_GIVEREF(__pyx_n_s_Mapping);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_Mapping);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_collections_abc, __pyx_t_2, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_MutableMapping); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_MutableMapping, __pyx_t_2) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_Mapping); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Mapping, __pyx_t_2) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":18
 * from collections.abc import MutableMapping, Mapping
 * 
 * system_version = tuple(sys.version_info)[:3]             # <<<<<<<<<<<<<<
 * PY_3 = system_version[0] == 3
 * try:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_sys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_version_info); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PySequence_Tuple(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyTuple_GetSlice(__pyx_t_1, 0, 3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_system_version, __pyx_t_2) < 0) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":19
 * 
 * system_version = tuple(sys.version_info)[:3]
 * PY_3 = system_version[0] == 3             # <<<<<<<<<<<<<<
 * try:
 *     from _thread import RLock
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_system_version); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_2, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_t_1, __pyx_int_3, 3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PY_3, __pyx_t_2) < 0) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":20
 * system_version = tuple(sys.version_info)[:3]
 * PY_3 = system_version[0] == 3
 * try:             # <<<<<<<<<<<<<<
 *     from _thread import RLock
 * except ImportError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    /*try:*/ {

      /* "Aspidites/_vendor/pyparsing_extension.py":21
 * PY_3 = system_version[0] == 3
 * try:
 *     from _thread import RLock             # <<<<<<<<<<<<<<
 * except ImportError:
 *     from threading import RLock
 */
      __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 21, __pyx_L2_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_n_s_RLock);
      __Pyx_GIVEREF(__pyx_n_s_RLock);
      PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_RLock);
      __pyx_t_1 = __Pyx_Import(__pyx_n_s_thread, __pyx_t_2, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 21, __pyx_L2_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_RLock); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 21, __pyx_L2_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (PyDict_SetItem(__pyx_d, __pyx_n_s_RLock, __pyx_t_2) < 0) __PYX_ERR(0, 21, __pyx_L2_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":20
 * system_version = tuple(sys.version_info)[:3]
 * PY_3 = system_version[0] == 3
 * try:             # <<<<<<<<<<<<<<
 *     from _thread import RLock
 * except ImportError:
 */
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L7_try_end;
    __pyx_L2_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":22
 * try:
 *     from _thread import RLock
 * except ImportError:             # <<<<<<<<<<<<<<
 *     from threading import RLock
 * 
 */
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ImportError);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_2, &__pyx_t_7) < 0) __PYX_ERR(0, 22, __pyx_L4_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GOTREF(__pyx_t_7);

      /* "Aspidites/_vendor/pyparsing_extension.py":23
 *     from _thread import RLock
 * except ImportError:
 *     from threading import RLock             # <<<<<<<<<<<<<<
 * 
 * try:
 */
      __pyx_t_8 = PyList_New(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 23, __pyx_L4_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_INCREF(__pyx_n_s_RLock);
      __Pyx_GIVEREF(__pyx_n_s_RLock);
      PyList_SET_ITEM(__pyx_t_8, 0, __pyx_n_s_RLock);
      __pyx_t_9 = __Pyx_Import(__pyx_n_s_threading, __pyx_t_8, -1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 23, __pyx_L4_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = __Pyx_ImportFrom(__pyx_t_9, __pyx_n_s_RLock); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 23, __pyx_L4_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (PyDict_SetItem(__pyx_d, __pyx_n_s_RLock, __pyx_t_8) < 0) __PYX_ERR(0, 23, __pyx_L4_except_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L3_exception_handled;
    }
    goto __pyx_L4_except_error;
    __pyx_L4_except_error:;

    /* "Aspidites/_vendor/pyparsing_extension.py":20
 * system_version = tuple(sys.version_info)[:3]
 * PY_3 = system_version[0] == 3
 * try:             # <<<<<<<<<<<<<<
 *     from _thread import RLock
 * except ImportError:
 */
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    goto __pyx_L1_error;
    __pyx_L3_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    __pyx_L7_try_end:;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":25
 *     from threading import RLock
 * 
 * try:             # <<<<<<<<<<<<<<
 *     from collections import OrderedDict as _OrderedDict
 * except ImportError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_4, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "Aspidites/_vendor/pyparsing_extension.py":26
 * 
 * try:
 *     from collections import OrderedDict as _OrderedDict             # <<<<<<<<<<<<<<
 * except ImportError:
 *     try:
 */
      __pyx_t_7 = PyList_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 26, __pyx_L10_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_n_s_OrderedDict_2);
      __Pyx_GIVEREF(__pyx_n_s_OrderedDict_2);
      PyList_SET_ITEM(__pyx_t_7, 0, __pyx_n_s_OrderedDict_2);
      __pyx_t_2 = __Pyx_Import(__pyx_n_s_collections, __pyx_t_7, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 26, __pyx_L10_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_OrderedDict_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 26, __pyx_L10_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (PyDict_SetItem(__pyx_d, __pyx_n_s_OrderedDict, __pyx_t_7) < 0) __PYX_ERR(0, 26, __pyx_L10_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":25
 *     from threading import RLock
 * 
 * try:             # <<<<<<<<<<<<<<
 *     from collections import OrderedDict as _OrderedDict
 * except ImportError:
 */
    }
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L15_try_end;
    __pyx_L10_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":27
 * try:
 *     from collections import OrderedDict as _OrderedDict
 * except ImportError:             # <<<<<<<<<<<<<<
 *     try:
 *         from ordereddict import OrderedDict as _OrderedDict
 */
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ImportError);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_7, &__pyx_t_1) < 0) __PYX_ERR(0, 27, __pyx_L12_except_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_t_1);

      /* "Aspidites/_vendor/pyparsing_extension.py":28
 *     from collections import OrderedDict as _OrderedDict
 * except ImportError:
 *     try:             # <<<<<<<<<<<<<<
 *         from ordereddict import OrderedDict as _OrderedDict
 *     except ImportError:
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
        __Pyx_XGOTREF(__pyx_t_10);
        __Pyx_XGOTREF(__pyx_t_11);
        __Pyx_XGOTREF(__pyx_t_12);
        /*try:*/ {

          /* "Aspidites/_vendor/pyparsing_extension.py":29
 * except ImportError:
 *     try:
 *         from ordereddict import OrderedDict as _OrderedDict             # <<<<<<<<<<<<<<
 *     except ImportError:
 *         _OrderedDict = None
 */
          __pyx_t_9 = PyList_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 29, __pyx_L18_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_INCREF(__pyx_n_s_OrderedDict_2);
          __Pyx_GIVEREF(__pyx_n_s_OrderedDict_2);
          PyList_SET_ITEM(__pyx_t_9, 0, __pyx_n_s_OrderedDict_2);
          __pyx_t_8 = __Pyx_Import(__pyx_n_s_ordereddict, __pyx_t_9, -1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 29, __pyx_L18_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __pyx_t_9 = __Pyx_ImportFrom(__pyx_t_8, __pyx_n_s_OrderedDict_2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 29, __pyx_L18_error)
          __Pyx_GOTREF(__pyx_t_9);
          if (PyDict_SetItem(__pyx_d, __pyx_n_s_OrderedDict, __pyx_t_9) < 0) __PYX_ERR(0, 29, __pyx_L18_error)
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

          /* "Aspidites/_vendor/pyparsing_extension.py":28
 *     from collections import OrderedDict as _OrderedDict
 * except ImportError:
 *     try:             # <<<<<<<<<<<<<<
 *         from ordereddict import OrderedDict as _OrderedDict
 *     except ImportError:
 */
        }
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        goto __pyx_L25_try_end;
        __pyx_L18_error:;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "Aspidites/_vendor/pyparsing_extension.py":30
 *     try:
 *         from ordereddict import OrderedDict as _OrderedDict
 *     except ImportError:             # <<<<<<<<<<<<<<
 *         _OrderedDict = None
 * 
 */
        __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ImportError);
        if (__pyx_t_6) {
          __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_9, &__pyx_t_13) < 0) __PYX_ERR(0, 30, __pyx_L20_except_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_GOTREF(__pyx_t_13);

          /* "Aspidites/_vendor/pyparsing_extension.py":31
 *         from ordereddict import OrderedDict as _OrderedDict
 *     except ImportError:
 *         _OrderedDict = None             # <<<<<<<<<<<<<<
 * 
 * try:
 */
          if (PyDict_SetItem(__pyx_d, __pyx_n_s_OrderedDict, Py_None) < 0) __PYX_ERR(0, 31, __pyx_L20_except_error)
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          goto __pyx_L19_exception_handled;
        }
        goto __pyx_L20_except_error;
        __pyx_L20_except_error:;

        /* "Aspidites/_vendor/pyparsing_extension.py":28
 *     from collections import OrderedDict as _OrderedDict
 * except ImportError:
 *     try:             # <<<<<<<<<<<<<<
 *         from ordereddict import OrderedDict as _OrderedDict
 *     except ImportError:
 */
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
        goto __pyx_L12_except_error;
        __pyx_L19_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
        __pyx_L25_try_end:;
      }
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L11_exception_handled;
    }
    goto __pyx_L12_except_error;
    __pyx_L12_except_error:;

    /* "Aspidites/_vendor/pyparsing_extension.py":25
 *     from threading import RLock
 * 
 * try:             # <<<<<<<<<<<<<<
 *     from collections import OrderedDict as _OrderedDict
 * except ImportError:
 */
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_4, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L11_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_4, __pyx_t_3);
    __pyx_L15_try_end:;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":33
 *         _OrderedDict = None
 * 
 * try:             # <<<<<<<<<<<<<<
 *     from types import SimpleNamespace
 * except ImportError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    /*try:*/ {

      /* "Aspidites/_vendor/pyparsing_extension.py":34
 * 
 * try:
 *     from types import SimpleNamespace             # <<<<<<<<<<<<<<
 * except ImportError:
 *     class SimpleNamespace:
 */
      __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 34, __pyx_L28_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_n_s_SimpleNamespace);
      __Pyx_GIVEREF(__pyx_n_s_SimpleNamespace);
      PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_SimpleNamespace);
      __pyx_t_7 = __Pyx_Import(__pyx_n_s_types, __pyx_t_1, -1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 34, __pyx_L28_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_7, __pyx_n_s_SimpleNamespace); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 34, __pyx_L28_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (PyDict_SetItem(__pyx_d, __pyx_n_s_SimpleNamespace, __pyx_t_1) < 0) __PYX_ERR(0, 34, __pyx_L28_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "Aspidites/_vendor/pyparsing_extension.py":33
 *         _OrderedDict = None
 * 
 * try:             # <<<<<<<<<<<<<<
 *     from types import SimpleNamespace
 * except ImportError:
 */
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L33_try_end;
    __pyx_L28_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":35
 * try:
 *     from types import SimpleNamespace
 * except ImportError:             # <<<<<<<<<<<<<<
 *     class SimpleNamespace:
 *         pass
 */
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ImportError);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("Aspidites._vendor.pyparsing_extension", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_1, &__pyx_t_2) < 0) __PYX_ERR(0, 35, __pyx_L30_except_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_t_2);

      /* "Aspidites/_vendor/pyparsing_extension.py":36
 *     from types import SimpleNamespace
 * except ImportError:
 *     class SimpleNamespace:             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
      __pyx_t_13 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_empty_tuple, __pyx_n_s_SimpleNamespace, __pyx_n_s_SimpleNamespace, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, (PyObject *) NULL); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 36, __pyx_L30_except_error)
      __Pyx_GOTREF(__pyx_t_13);
      __pyx_t_9 = __Pyx_Py3ClassCreate(((PyObject*)&__Pyx_DefaultClassType), __pyx_n_s_SimpleNamespace, __pyx_empty_tuple, __pyx_t_13, NULL, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 36, __pyx_L30_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      if (PyDict_SetItem(__pyx_d, __pyx_n_s_SimpleNamespace, __pyx_t_9) < 0) __PYX_ERR(0, 36, __pyx_L30_except_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L29_exception_handled;
    }
    goto __pyx_L30_except_error;
    __pyx_L30_except_error:;

    /* "Aspidites/_vendor/pyparsing_extension.py":33
 *         _OrderedDict = None
 * 
 * try:             # <<<<<<<<<<<<<<
 *     from types import SimpleNamespace
 * except ImportError:
 */
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    goto __pyx_L1_error;
    __pyx_L29_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    __pyx_L33_try_end:;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":39
 *         pass
 * 
 * alphas = string.ascii_uppercase + string.ascii_lowercase             # <<<<<<<<<<<<<<
 * nums = "0123456789"
 * hexnums = nums + "ABCDEFabcdef"
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_string); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ascii_uppercase); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_string); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ascii_lowercase); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_Add(__pyx_t_1, __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_alphas, __pyx_t_2) < 0) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":40
 * 
 * alphas = string.ascii_uppercase + string.ascii_lowercase
 * nums = "0123456789"             # <<<<<<<<<<<<<<
 * hexnums = nums + "ABCDEFabcdef"
 * alphanums = alphas + nums
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_nums, __pyx_kp_s_0123456789) < 0) __PYX_ERR(0, 40, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":41
 * alphas = string.ascii_uppercase + string.ascii_lowercase
 * nums = "0123456789"
 * hexnums = nums + "ABCDEFabcdef"             # <<<<<<<<<<<<<<
 * alphanums = alphas + nums
 * printables = "".join(c for c in string.printable if c not in string.whitespace)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_nums); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = PyNumber_Add(__pyx_t_2, __pyx_n_s_ABCDEFabcdef); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_hexnums, __pyx_t_7) < 0) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":42
 * nums = "0123456789"
 * hexnums = nums + "ABCDEFabcdef"
 * alphanums = alphas + nums             # <<<<<<<<<<<<<<
 * printables = "".join(c for c in string.printable if c not in string.whitespace)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_alphas); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_nums); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyNumber_Add(__pyx_t_7, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_alphanums, __pyx_t_1) < 0) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":43
 * hexnums = nums + "ABCDEFabcdef"
 * alphanums = alphas + nums
 * printables = "".join(c for c in string.printable if c not in string.whitespace)             # <<<<<<<<<<<<<<
 * 
 * __diag__ = SimpleNamespace()
 */
  __pyx_t_1 = __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_25genexpr(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyString_Join(__pyx_kp_s__12, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_printables, __pyx_t_2) < 0) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":45
 * printables = "".join(c for c in string.printable if c not in string.whitespace)
 * 
 * __diag__ = SimpleNamespace()             # <<<<<<<<<<<<<<
 * __diag__.__doc__ = """
 * Diagnostic configuration (all default to False)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SimpleNamespace); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_diag, __pyx_t_1) < 0) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":46
 * 
 * __diag__ = SimpleNamespace()
 * __diag__.__doc__ = """             # <<<<<<<<<<<<<<
 * Diagnostic configuration (all default to False)
 *      - warn_multiple_tokens_in_named_alternation - flag to enable warnings when a results
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_diag); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_1, __pyx_n_s_doc, __pyx_kp_s_Diagnostic_configuration_all_de) < 0) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":61
 *        calls to ParserElement.setName()
 * """
 * __diag__.collect_all_And_tokens = True             # <<<<<<<<<<<<<<
 * __diag__.warn_multiple_tokens_in_named_alternation = False
 * __diag__.warn_ungrouped_named_tokens_in_collection = False
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_diag); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_1, __pyx_n_s_collect_all_And_tokens, Py_True) < 0) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":62
 * """
 * __diag__.collect_all_And_tokens = True
 * __diag__.warn_multiple_tokens_in_named_alternation = False             # <<<<<<<<<<<<<<
 * __diag__.warn_ungrouped_named_tokens_in_collection = False
 * __diag__.warn_name_set_on_empty_Forward = False
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_diag); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_1, __pyx_n_s_warn_multiple_tokens_in_named_al, Py_False) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":63
 * __diag__.collect_all_And_tokens = True
 * __diag__.warn_multiple_tokens_in_named_alternation = False
 * __diag__.warn_ungrouped_named_tokens_in_collection = False             # <<<<<<<<<<<<<<
 * __diag__.warn_name_set_on_empty_Forward = False
 * __diag__.warn_on_multiple_string_args_to_oneof = False
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_diag); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_1, __pyx_n_s_warn_ungrouped_named_tokens_in_c, Py_False) < 0) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":64
 * __diag__.warn_multiple_tokens_in_named_alternation = False
 * __diag__.warn_ungrouped_named_tokens_in_collection = False
 * __diag__.warn_name_set_on_empty_Forward = False             # <<<<<<<<<<<<<<
 * __diag__.warn_on_multiple_string_args_to_oneof = False
 * __diag__.enable_debug_on_named_expressions = False
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_diag); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_1, __pyx_n_s_warn_name_set_on_empty_Forward, Py_False) < 0) __PYX_ERR(0, 64, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":65
 * __diag__.warn_ungrouped_named_tokens_in_collection = False
 * __diag__.warn_name_set_on_empty_Forward = False
 * __diag__.warn_on_multiple_string_args_to_oneof = False             # <<<<<<<<<<<<<<
 * __diag__.enable_debug_on_named_expressions = False
 * __diag__._all_names = [nm for nm in vars(__diag__) if nm.startswith("enable_") or nm.startswith("warn_")]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_diag); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_1, __pyx_n_s_warn_on_multiple_string_args_to, Py_False) < 0) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":66
 * __diag__.warn_name_set_on_empty_Forward = False
 * __diag__.warn_on_multiple_string_args_to_oneof = False
 * __diag__.enable_debug_on_named_expressions = False             # <<<<<<<<<<<<<<
 * __diag__._all_names = [nm for nm in vars(__diag__) if nm.startswith("enable_") or nm.startswith("warn_")]
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_diag); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_1, __pyx_n_s_enable_debug_on_named_expression, Py_False) < 0) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":67
 * __diag__.warn_on_multiple_string_args_to_oneof = False
 * __diag__.enable_debug_on_named_expressions = False
 * __diag__._all_names = [nm for nm in vars(__diag__) if nm.startswith("enable_") or nm.startswith("warn_")]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_diag); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_builtin_vars, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_7)) || PyTuple_CheckExact(__pyx_t_7)) {
    __pyx_t_2 = __pyx_t_7; __Pyx_INCREF(__pyx_t_2); __pyx_t_14 = 0;
    __pyx_t_15 = NULL;
  } else {
    __pyx_t_14 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_15 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 67, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  for (;;) {
    if (likely(!__pyx_t_15)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_14 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_14); __Pyx_INCREF(__pyx_t_7); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 67, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_2, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 67, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      } else {
        if (__pyx_t_14 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_14); __Pyx_INCREF(__pyx_t_7); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 67, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_2, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 67, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      }
    } else {
      __pyx_t_7 = __pyx_t_15(__pyx_t_2);
      if (unlikely(!__pyx_t_7)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 67, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_7);
    }
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_nm, __pyx_t_7) < 0) __PYX_ERR(0, 67, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_nm); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_n_s_startswith); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_t_13 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_13)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_13);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_7 = (__pyx_t_13) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_13, __pyx_n_s_enable) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_n_s_enable);
    __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_17 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_17 < 0)) __PYX_ERR(0, 67, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (!__pyx_t_17) {
    } else {
      __pyx_t_16 = __pyx_t_17;
      goto __pyx_L39_bool_binop_done;
    }
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_nm); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_startswith); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_13))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_13);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_13, function);
      }
    }
    __pyx_t_7 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_13, __pyx_t_9, __pyx_n_s_warn_2) : __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_n_s_warn_2);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_t_17 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_17 < 0)) __PYX_ERR(0, 67, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_16 = __pyx_t_17;
    __pyx_L39_bool_binop_done:;
    if (__pyx_t_16) {
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_nm); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 67, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_7))) __PYX_ERR(0, 67, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_diag); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_2, __pyx_n_s_all_names, __pyx_t_1) < 0) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":70
 * 
 * 
 * def _enable_all_warnings():             # <<<<<<<<<<<<<<
 *     __diag__.warn_multiple_tokens_in_named_alternation = True
 *     __diag__.warn_ungrouped_named_tokens_in_collection = True
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_1_enable_all_warnings, 0, __pyx_n_s_enable_all_warnings, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__121)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_enable_all_warnings, __pyx_t_2) < 0) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":75
 *     __diag__.warn_name_set_on_empty_Forward = True
 *     __diag__.warn_on_multiple_string_args_to_oneof = True
 * __diag__.enable_all_warnings = _enable_all_warnings             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_enable_all_warnings); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_diag); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_1, __pyx_n_s_enable_all_warnings_2, __pyx_t_2) < 0) __PYX_ERR(0, 75, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":78
 * 
 * 
 * def _defaultStartDebugAction(instring, loc, expr):             # <<<<<<<<<<<<<<
 *     print(("Match " + str(expr) + " at loc " + str(loc) + "(%d,%d)" % (lineno(loc, instring), col(loc, instring))))
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_3_defaultStartDebugAction, 0, __pyx_n_s_defaultStartDebugAction, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__123)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_defaultStartDebugAction, __pyx_t_1) < 0) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":82
 * 
 * 
 * def _defaultSuccessDebugAction(instring, startloc, endloc, expr, toks):             # <<<<<<<<<<<<<<
 *     print("Matched " + str(expr) + " -> " + str(toks.asList()))
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_5_defaultSuccessDebugAction, 0, __pyx_n_s_defaultSuccessDebugAction, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__125)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_defaultSuccessDebugAction, __pyx_t_1) < 0) __PYX_ERR(0, 82, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":86
 * 
 * 
 * def _defaultExceptionDebugAction(instring, loc, expr, exc):             # <<<<<<<<<<<<<<
 *     print("Exception raised:" + str(exc))
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7_defaultExceptionDebugAction, 0, __pyx_n_s_defaultExceptionDebugAction, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__127)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_defaultExceptionDebugAction, __pyx_t_1) < 0) __PYX_ERR(0, 86, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":90
 * 
 * 
 * def nullDebugAction(*args):             # <<<<<<<<<<<<<<
 *     """'Do-nothing' debug action, to suppress debugging output during parsing."""
 *     pass
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_9nullDebugAction, 0, __pyx_n_s_nullDebugAction, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__129)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_nullDebugAction, __pyx_t_1) < 0) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":95
 * 
 * 
 * def _escapeRegexRangeChars(s):             # <<<<<<<<<<<<<<
 *     # ~  escape these chars: ^-[]
 *     for c in r"\^-[]":
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_11_escapeRegexRangeChars, 0, __pyx_n_s_escapeRegexRangeChars, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__131)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_escapeRegexRangeChars, __pyx_t_1) < 0) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":104
 * 
 * 
 * def _flatten(L):             # <<<<<<<<<<<<<<
 *     ret = []
 *     for i in L:
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13_flatten, 0, __pyx_n_s_flatten, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__133)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_flatten, __pyx_t_1) < 0) __PYX_ERR(0, 104, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":114
 * 
 * 
 * def _xml_escape(data):             # <<<<<<<<<<<<<<
 *     """Escape &, <, >, ", ', etc. in a string of data."""
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_15_xml_escape, 0, __pyx_n_s_xml_escape, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__135)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_xml_escape, __pyx_t_1) < 0) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":125
 * 
 * 
 * def col(loc, strg):             # <<<<<<<<<<<<<<
 *     """Returns current column within a string, counting newlines as line separators.
 *    The first column is number 1.
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_17col, 0, __pyx_n_s_col, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__137)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_col, __pyx_t_1) < 0) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":140
 * 
 * 
 * def lineno(loc, strg):             # <<<<<<<<<<<<<<
 *     """Returns current line number within a string, counting newlines as line separators.
 *     The first line is number 1.
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_19lineno, 0, __pyx_n_s_lineno, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__139)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_lineno, __pyx_t_1) < 0) __PYX_ERR(0, 140, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":153
 * 
 * 
 * def line(loc, strg):             # <<<<<<<<<<<<<<
 *     """Returns the line of text containing loc within a string, counting newlines as line separators.
 *        """
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_21line, 0, __pyx_n_s_line, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__141)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_line, __pyx_t_1) < 0) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":164
 * 
 * 
 * class _ParseResultsWithOffset(object):             # <<<<<<<<<<<<<<
 *     def __init__(self, p1, p2):
 *         self.tup = (p1, p2)
 */
  __pyx_t_1 = __Pyx_CalculateMetaclass(NULL, __pyx_tuple__142); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_Py3MetaclassPrepare(__pyx_t_1, __pyx_tuple__142, __pyx_n_s_ParseResultsWithOffset, __pyx_n_s_ParseResultsWithOffset, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, (PyObject *) NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "Aspidites/_vendor/pyparsing_extension.py":165
 * 
 * class _ParseResultsWithOffset(object):
 *     def __init__(self, p1, p2):             # <<<<<<<<<<<<<<
 *         self.tup = (p1, p2)
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_23_ParseResultsWithOffset_1__init__, 0, __pyx_n_s_ParseResultsWithOffset___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__144)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_init, __pyx_t_7) < 0) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":168
 *         self.tup = (p1, p2)
 * 
 *     def __getitem__(self, i):             # <<<<<<<<<<<<<<
 *         return self.tup[i]
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_23_ParseResultsWithOffset_3__getitem__, 0, __pyx_n_s_ParseResultsWithOffset___getite, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__146)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_getitem, __pyx_t_7) < 0) __PYX_ERR(0, 168, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":171
 *         return self.tup[i]
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.tup[0])
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_23_ParseResultsWithOffset_5__repr__, 0, __pyx_n_s_ParseResultsWithOffset___repr, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__148)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_repr, __pyx_t_7) < 0) __PYX_ERR(0, 171, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":174
 *         return repr(self.tup[0])
 * 
 *     def setOffset(self, i):             # <<<<<<<<<<<<<<
 *         self.tup = (self.tup[0], i)
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_23_ParseResultsWithOffset_7setOffset, 0, __pyx_n_s_ParseResultsWithOffset_setOffse, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__150)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_setOffset, __pyx_t_7) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":164
 * 
 * 
 * class _ParseResultsWithOffset(object):             # <<<<<<<<<<<<<<
 *     def __init__(self, p1, p2):
 *         self.tup = (p1, p2)
 */
  __pyx_t_7 = __Pyx_Py3ClassCreate(__pyx_t_1, __pyx_n_s_ParseResultsWithOffset, __pyx_tuple__142, __pyx_t_2, NULL, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ParseResultsWithOffset, __pyx_t_7) < 0) __PYX_ERR(0, 164, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":178
 * 
 * 
 * class ParseResults(object):             # <<<<<<<<<<<<<<
 *     """Structured parse results, to provide multiple means of access to
 *     the parsed data:
 */
  __pyx_t_1 = __Pyx_CalculateMetaclass(NULL, __pyx_tuple__151); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_Py3MetaclassPrepare(__pyx_t_1, __pyx_tuple__151, __pyx_n_s_ParseResults, __pyx_n_s_ParseResults, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_Structured_parse_results_to_prov); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "Aspidites/_vendor/pyparsing_extension.py":222
 *     """
 * 
 *     def __new__(cls, toklist=None, name=None, asList=True, modal=True):             # <<<<<<<<<<<<<<
 *         if isinstance(toklist, cls):
 *             return toklist
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_1__new__, __Pyx_CYFUNCTION_STATICMETHOD, __pyx_n_s_ParseResults___new, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__153)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_tuple__154);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_new, __pyx_t_7) < 0) __PYX_ERR(0, 222, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":231
 *     # Performance tuning: we construct a *lot* of these, so keep this
 *     # constructor as small and fast as possible
 *     def __init__(self, toklist=None, name=None, asList=True, modal=True, isinstance=isinstance):             # <<<<<<<<<<<<<<
 *         if self.__doinit:
 *             self.__doinit = False
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_3__init__, 0, __pyx_n_s_ParseResults___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__156)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_7, sizeof(__pyx_defaults), 1)) __PYX_ERR(0, 231, __pyx_L1_error)
  __pyx_t_13 = __Pyx_GetBuiltinName(__pyx_n_s_isinstance); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_t_7)->__pyx_arg_isinstance = __pyx_t_13;
  __Pyx_GIVEREF(__pyx_t_13);
  __pyx_t_13 = 0;
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_7, __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_28__defaults__);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_init, __pyx_t_7) < 0) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":270
 *                         self[name] = toklist
 * 
 *     def __getitem__(self, i):             # <<<<<<<<<<<<<<
 *         if isinstance(i, (int, slice)):
 *             return self.__toklist[i]
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_5__getitem__, 0, __pyx_n_s_ParseResults___getitem, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__158)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_getitem, __pyx_t_7) < 0) __PYX_ERR(0, 270, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":279
 *                 return ParseResults([v[0] for v in self.__tokdict[i]])
 * 
 *     def __setitem__(self, k, v, isinstance=isinstance):             # <<<<<<<<<<<<<<
 *         if isinstance(v, _ParseResultsWithOffset):
 *             self.__tokdict[k] = self.__tokdict.get(k, list()) + [v]
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_7__setitem__, 0, __pyx_n_s_ParseResults___setitem, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__160)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_7, sizeof(__pyx_defaults1), 1)) __PYX_ERR(0, 279, __pyx_L1_error)
  __pyx_t_13 = __Pyx_GetBuiltinName(__pyx_n_s_isinstance); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_t_7)->__pyx_arg_isinstance = __pyx_t_13;
  __Pyx_GIVEREF(__pyx_t_13);
  __pyx_t_13 = 0;
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_7, __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_30__defaults__);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_setitem, __pyx_t_7) < 0) __PYX_ERR(0, 279, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":292
 *             sub.__parent = wkref(self)
 * 
 *     def __delitem__(self, i):             # <<<<<<<<<<<<<<
 *         if isinstance(i, (int, slice)):
 *             mylen = len(self.__toklist)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_9__delitem__, 0, __pyx_n_s_ParseResults___delitem, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__162)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_delitem, __pyx_t_7) < 0) __PYX_ERR(0, 292, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":313
 *             del self.__tokdict[i]
 * 
 *     def __contains__(self, k):             # <<<<<<<<<<<<<<
 *         return k in self.__tokdict
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_11__contains__, 0, __pyx_n_s_ParseResults___contains, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__164)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_contains, __pyx_t_7) < 0) __PYX_ERR(0, 313, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":316
 *         return k in self.__tokdict
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return len(self.__toklist)
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_13__len__, 0, __pyx_n_s_ParseResults___len, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__166)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_len, __pyx_t_7) < 0) __PYX_ERR(0, 316, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":319
 *         return len(self.__toklist)
 * 
 *     def __bool__(self):             # <<<<<<<<<<<<<<
 *         return (not not self.__toklist)
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_15__bool__, 0, __pyx_n_s_ParseResults___bool, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__168)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 319, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_bool, __pyx_t_7) < 0) __PYX_ERR(0, 319, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":322
 *         return (not not self.__toklist)
 * 
 *     __nonzero__ = __bool__             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
  __pyx_t_7 = PyObject_GetItem(__pyx_t_2, __pyx_n_s_bool);
  if (unlikely(!__pyx_t_7)) {
    PyErr_Clear();
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_bool);
  }
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 322, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_nonzero, __pyx_t_7) < 0) __PYX_ERR(0, 322, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":324
 *     __nonzero__ = __bool__
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return iter(self.__toklist)
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_17__iter__, 0, __pyx_n_s_ParseResults___iter, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__170)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_iter, __pyx_t_7) < 0) __PYX_ERR(0, 324, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":327
 *         return iter(self.__toklist)
 * 
 *     def __reversed__(self):             # <<<<<<<<<<<<<<
 *         return iter(self.__toklist[::-1])
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_19__reversed__, 0, __pyx_n_s_ParseResults___reversed, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__172)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_reversed, __pyx_t_7) < 0) __PYX_ERR(0, 327, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":330
 *         return iter(self.__toklist[::-1])
 * 
 *     def _iterkeys(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self.__tokdict, "iterkeys"):
 *             return self.__tokdict.iterkeys()
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_21_iterkeys, 0, __pyx_n_s_ParseResults__iterkeys, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__174)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_iterkeys_2, __pyx_t_7) < 0) __PYX_ERR(0, 330, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":336
 *             return iter(self.__tokdict)
 * 
 *     def _itervalues(self):             # <<<<<<<<<<<<<<
 *         return (self[k] for k in self._iterkeys())
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_23_itervalues, 0, __pyx_n_s_ParseResults__itervalues, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__176)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_itervalues_2, __pyx_t_7) < 0) __PYX_ERR(0, 336, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":339
 *         return (self[k] for k in self._iterkeys())
 * 
 *     def _iteritems(self):             # <<<<<<<<<<<<<<
 *         return ((k, self[k]) for k in self._iterkeys())
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_25_iteritems, 0, __pyx_n_s_ParseResults__iteritems, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__178)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_iteritems_2, __pyx_t_7) < 0) __PYX_ERR(0, 339, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":342
 *         return ((k, self[k]) for k in self._iterkeys())
 * 
 *     if PY_3:             # <<<<<<<<<<<<<<
 *         keys = _iterkeys
 *         """Returns an iterator of all named result keys."""
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_PY_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_16 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_16 < 0)) __PYX_ERR(0, 342, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__pyx_t_16) {

    /* "Aspidites/_vendor/pyparsing_extension.py":343
 * 
 *     if PY_3:
 *         keys = _iterkeys             # <<<<<<<<<<<<<<
 *         """Returns an iterator of all named result keys."""
 * 
 */
    __pyx_t_7 = PyObject_GetItem(__pyx_t_2, __pyx_n_s_iterkeys_2);
    if (unlikely(!__pyx_t_7)) {
      PyErr_Clear();
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_iterkeys_2);
    }
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 343, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_keys, __pyx_t_7) < 0) __PYX_ERR(0, 343, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":346
 *         """Returns an iterator of all named result keys."""
 * 
 *         values = _itervalues             # <<<<<<<<<<<<<<
 *         """Returns an iterator of all named result values."""
 * 
 */
    __pyx_t_7 = PyObject_GetItem(__pyx_t_2, __pyx_n_s_itervalues_2);
    if (unlikely(!__pyx_t_7)) {
      PyErr_Clear();
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_itervalues_2);
    }
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 346, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_values, __pyx_t_7) < 0) __PYX_ERR(0, 346, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":349
 *         """Returns an iterator of all named result values."""
 * 
 *         items = _iteritems             # <<<<<<<<<<<<<<
 *         """Returns an iterator of all named result key-value tuples."""
 * 
 */
    __pyx_t_7 = PyObject_GetItem(__pyx_t_2, __pyx_n_s_iteritems_2);
    if (unlikely(!__pyx_t_7)) {
      PyErr_Clear();
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_iteritems_2);
    }
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 349, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_items, __pyx_t_7) < 0) __PYX_ERR(0, 349, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":342
 *         return ((k, self[k]) for k in self._iterkeys())
 * 
 *     if PY_3:             # <<<<<<<<<<<<<<
 *         keys = _iterkeys
 *         """Returns an iterator of all named result keys."""
 */
    goto __pyx_L41;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":353
 * 
 *     else:
 *         iterkeys = _iterkeys             # <<<<<<<<<<<<<<
 *         """Returns an iterator of all named result keys (Python 2.x only)."""
 * 
 */
  /*else*/ {
    __pyx_t_7 = PyObject_GetItem(__pyx_t_2, __pyx_n_s_iterkeys_2);
    if (unlikely(!__pyx_t_7)) {
      PyErr_Clear();
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_iterkeys_2);
    }
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 353, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_iterkeys, __pyx_t_7) < 0) __PYX_ERR(0, 353, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":356
 *         """Returns an iterator of all named result keys (Python 2.x only)."""
 * 
 *         itervalues = _itervalues             # <<<<<<<<<<<<<<
 *         """Returns an iterator of all named result values (Python 2.x only)."""
 * 
 */
    __pyx_t_7 = PyObject_GetItem(__pyx_t_2, __pyx_n_s_itervalues_2);
    if (unlikely(!__pyx_t_7)) {
      PyErr_Clear();
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_itervalues_2);
    }
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 356, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_itervalues, __pyx_t_7) < 0) __PYX_ERR(0, 356, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":359
 *         """Returns an iterator of all named result values (Python 2.x only)."""
 * 
 *         iteritems = _iteritems             # <<<<<<<<<<<<<<
 *         """Returns an iterator of all named result key-value tuples (Python 2.x only)."""
 * 
 */
    __pyx_t_7 = PyObject_GetItem(__pyx_t_2, __pyx_n_s_iteritems_2);
    if (unlikely(!__pyx_t_7)) {
      PyErr_Clear();
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_iteritems_2);
    }
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 359, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_iteritems, __pyx_t_7) < 0) __PYX_ERR(0, 359, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":362
 *         """Returns an iterator of all named result key-value tuples (Python 2.x only)."""
 * 
 *         def keys(self):             # <<<<<<<<<<<<<<
 *             """Returns all named result keys (as a list in Python 2.x, as an iterator in Python 3.x)."""
 *             return list(self.iterkeys())
 */
    __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_27keys, 0, __pyx_n_s_ParseResults_keys, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__180)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 362, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_keys, __pyx_t_7) < 0) __PYX_ERR(0, 362, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":366
 *             return list(self.iterkeys())
 * 
 *         def values(self):             # <<<<<<<<<<<<<<
 *             """Returns all named result values (as a list in Python 2.x, as an iterator in Python 3.x)."""
 *             return list(self.itervalues())
 */
    __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_29values, 0, __pyx_n_s_ParseResults_values, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__182)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 366, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_values, __pyx_t_7) < 0) __PYX_ERR(0, 366, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":370
 *             return list(self.itervalues())
 * 
 *         def items(self):             # <<<<<<<<<<<<<<
 *             """Returns all named result key-values (as a list of tuples in Python 2.x, as an iterator in Python 3.x)."""
 *             return list(self.iteritems())
 */
    __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_31items, 0, __pyx_n_s_ParseResults_items, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__184)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 370, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_items, __pyx_t_7) < 0) __PYX_ERR(0, 370, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __pyx_L41:;

  /* "Aspidites/_vendor/pyparsing_extension.py":374
 *             return list(self.iteritems())
 * 
 *     def haskeys(self):             # <<<<<<<<<<<<<<
 *         """Since keys() returns an iterator, this method is helpful in bypassing
 *            code that looks for the existence of any defined results names."""
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_33haskeys, 0, __pyx_n_s_ParseResults_haskeys, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__186)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_haskeys, __pyx_t_7) < 0) __PYX_ERR(0, 374, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":379
 *         return bool(self.__tokdict)
 * 
 *     def pop(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         """
 *         Removes and returns item at specified index (default= ``last``).
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_35pop, 0, __pyx_n_s_ParseResults_pop, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__188)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_pop, __pyx_t_7) < 0) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":434
 *             return defaultvalue
 * 
 *     def get(self, key, defaultValue=None):             # <<<<<<<<<<<<<<
 *         """
 *         Returns named result matching the given key, or if there is no
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_37get, 0, __pyx_n_s_ParseResults_get, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__190)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_tuple__191);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_get, __pyx_t_7) < 0) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":457
 *             return defaultValue
 * 
 *     def insert(self, index, insStr):             # <<<<<<<<<<<<<<
 *         """
 *         Inserts new element at location index in the list of parsed tokens.
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_39insert, 0, __pyx_n_s_ParseResults_insert, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__193)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 457, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_insert, __pyx_t_7) < 0) __PYX_ERR(0, 457, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":478
 *                 occurrences[k] = _ParseResultsWithOffset(value, position + (position > index))
 * 
 *     def append(self, item):             # <<<<<<<<<<<<<<
 *         """
 *         Add single element to end of ParseResults list of elements.
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_41append, 0, __pyx_n_s_ParseResults_append, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__195)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 478, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_append, __pyx_t_7) < 0) __PYX_ERR(0, 478, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":493
 *         self.__toklist.append(item)
 * 
 *     def extend(self, itemseq):             # <<<<<<<<<<<<<<
 *         """
 *         Add sequence of elements to end of ParseResults list of elements.
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_43extend, 0, __pyx_n_s_ParseResults_extend, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__197)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_extend, __pyx_t_7) < 0) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":512
 *             self.__toklist.extend(itemseq)
 * 
 *     def clear(self):             # <<<<<<<<<<<<<<
 *         """
 *         Clear all elements and results names.
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_45clear, 0, __pyx_n_s_ParseResults_clear, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__199)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_clear, __pyx_t_7) < 0) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":519
 *         self.__tokdict.clear()
 * 
 *     def __getattr__(self, name):             # <<<<<<<<<<<<<<
 *         try:
 *             return self[name]
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_47__getattr__, 0, __pyx_n_s_ParseResults___getattr, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__201)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 519, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_getattr, __pyx_t_7) < 0) __PYX_ERR(0, 519, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":525
 *             return ""
 * 
 *     def __add__(self, other):             # <<<<<<<<<<<<<<
 *         ret = self.copy()
 *         ret += other
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_49__add__, 0, __pyx_n_s_ParseResults___add, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__203)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_add, __pyx_t_7) < 0) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":530
 *         return ret
 * 
 *     def __iadd__(self, other):             # <<<<<<<<<<<<<<
 *         if other.__tokdict:
 *             offset = len(self.__toklist)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_51__iadd__, 0, __pyx_n_s_ParseResults___iadd, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__205)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_iadd, __pyx_t_7) < 0) __PYX_ERR(0, 530, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":546
 *         return self
 * 
 *     def __radd__(self, other):             # <<<<<<<<<<<<<<
 *         if isinstance(other, int) and other == 0:
 *             # useful for merging many ParseResults using sum() builtin
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_53__radd__, 0, __pyx_n_s_ParseResults___radd, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__207)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_radd, __pyx_t_7) < 0) __PYX_ERR(0, 546, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":554
 *             return other + self
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "(%s, %s)" % (repr(self.__toklist), repr(self.__tokdict))
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_55__repr__, 0, __pyx_n_s_ParseResults___repr, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__209)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_repr, __pyx_t_7) < 0) __PYX_ERR(0, 554, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":557
 *         return "(%s, %s)" % (repr(self.__toklist), repr(self.__tokdict))
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return '[' + ', '.join(str(i) if isinstance(i, ParseResults) else repr(i) for i in self.__toklist) + ']'
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_57__str__, 0, __pyx_n_s_ParseResults___str, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__211)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_str, __pyx_t_7) < 0) __PYX_ERR(0, 557, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":560
 *         return '[' + ', '.join(str(i) if isinstance(i, ParseResults) else repr(i) for i in self.__toklist) + ']'
 * 
 *     def _asStringList(self, sep=''):             # <<<<<<<<<<<<<<
 *         out = []
 *         for item in self.__toklist:
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_59_asStringList, 0, __pyx_n_s_ParseResults__asStringList, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__213)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 560, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_tuple__214);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_asStringList, __pyx_t_7) < 0) __PYX_ERR(0, 560, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":571
 *         return out
 * 
 *     def asList(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns the parse results as a nested list of matching tokens, all converted to strings.
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_61asList, 0, __pyx_n_s_ParseResults_asList, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__216)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 571, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_asList, __pyx_t_7) < 0) __PYX_ERR(0, 571, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":588
 *         return [res.asList() if isinstance(res, ParseResults) else res for res in self.__toklist]
 * 
 *     def asDict(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns the named parse results as a nested dictionary.
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_63asDict, 0, __pyx_n_s_ParseResults_asDict, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__218)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_asDict, __pyx_t_7) < 0) __PYX_ERR(0, 588, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":624
 *         return dict((k, toItem(v)) for k, v in item_fn())
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns a new copy of a :class:`ParseResults` object.
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_65copy, 0, __pyx_n_s_ParseResults_copy, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__220)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_copy, __pyx_t_7) < 0) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":635
 *         return ret
 * 
 *     def asXML(self, doctag=None, namedItemsOnly=False, indent="", formatted=True):             # <<<<<<<<<<<<<<
 *         """
 *         (Deprecated) Returns the parse results as XML. Tags are created for tokens and lists that have defined results names.
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_67asXML, 0, __pyx_n_s_ParseResults_asXML, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__222)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_tuple__223);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_asXML, __pyx_t_7) < 0) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":696
 *         return "".join(out)
 * 
 *     def __lookup(self, sub):             # <<<<<<<<<<<<<<
 *         for k, vlist in self.__tokdict.items():
 *             for v, loc in vlist:
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_69__lookup, 0, __pyx_n_s_ParseResults___lookup, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__225)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_ParseResults__lookup, __pyx_t_7) < 0) __PYX_ERR(0, 696, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":703
 *         return None
 * 
 *     def getName(self):             # <<<<<<<<<<<<<<
 *         r"""
 *         Returns the results name for this token expression. Useful when several
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_71getName, 0, __pyx_n_s_ParseResults_getName, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__227)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_getName, __pyx_t_7) < 0) __PYX_ERR(0, 703, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":743
 *             return None
 * 
 *     def dump(self, indent='', full=True, include_list=True, _depth=0):             # <<<<<<<<<<<<<<
 *         """
 *         Diagnostic method for listing out the contents of
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_73dump, 0, __pyx_n_s_ParseResults_dump, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__229)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 743, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_tuple__230);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_dump, __pyx_t_7) < 0) __PYX_ERR(0, 743, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":808
 *         return "".join(out)
 * 
 *     def pprint(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         """
 *         Pretty-printer for parsed results as a list, using the
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_75pprint, 0, __pyx_n_s_ParseResults_pprint, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__232)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_pprint, __pyx_t_7) < 0) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":836
 * 
 *     # add support for pickle protocol
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return (self.__toklist,
 *                 (self.__tokdict.copy(),
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_77__getstate__, 0, __pyx_n_s_ParseResults___getstate, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__234)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 836, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_getstate, __pyx_t_7) < 0) __PYX_ERR(0, 836, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":843
 *                  self.__name))
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self.__toklist = state[0]
 *         self.__tokdict, par, inAccumNames, self.__name = state[1]
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_79__setstate__, 0, __pyx_n_s_ParseResults___setstate, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__236)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 843, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_setstate, __pyx_t_7) < 0) __PYX_ERR(0, 843, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":853
 *             self.__parent = None
 * 
 *     def __getnewargs__(self):             # <<<<<<<<<<<<<<
 *         return self.__toklist, self.__name, self.__asList, self.__modal
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_81__getnewargs__, 0, __pyx_n_s_ParseResults___getnewargs, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__238)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 853, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_getnewargs, __pyx_t_7) < 0) __PYX_ERR(0, 853, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":856
 *         return self.__toklist, self.__name, self.__asList, self.__modal
 * 
 *     def __dir__(self):             # <<<<<<<<<<<<<<
 *         return dir(type(self)) + list(self.keys())
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_83__dir__, 0, __pyx_n_s_ParseResults___dir, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__240)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_dir, __pyx_t_7) < 0) __PYX_ERR(0, 856, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":860
 * 
 *     @classmethod
 *     def from_dict(cls, other, name=None):             # <<<<<<<<<<<<<<
 *         """
 *         Helper classmethod to construct a ParseResults from a dict, preserving the
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12ParseResults_85from_dict, __Pyx_CYFUNCTION_CLASSMETHOD, __pyx_n_s_ParseResults_from_dict, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__242)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 860, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_tuple__243);

  /* "Aspidites/_vendor/pyparsing_extension.py":859
 *         return dir(type(self)) + list(self.keys())
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def from_dict(cls, other, name=None):
 *         """
 */
  __pyx_t_13 = __Pyx_Method_ClassMethod(__pyx_t_7); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_from_dict, __pyx_t_13) < 0) __PYX_ERR(0, 860, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":178
 * 
 * 
 * class ParseResults(object):             # <<<<<<<<<<<<<<
 *     """Structured parse results, to provide multiple means of access to
 *     the parsed data:
 */
  __pyx_t_13 = __Pyx_Py3ClassCreate(__pyx_t_1, __pyx_n_s_ParseResults, __pyx_tuple__151, __pyx_t_2, NULL, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ParseResults, __pyx_t_13) < 0) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":889
 * 
 * 
 * class ParseBaseException(Exception):             # <<<<<<<<<<<<<<
 *     """base exception class for all parsing runtime exceptions"""
 * 
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 889, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
  __Pyx_GIVEREF(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 889, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_13 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_1, __pyx_n_s_ParseBaseException, __pyx_n_s_ParseBaseException, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_base_exception_class_for_all_par); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 889, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);

  /* "Aspidites/_vendor/pyparsing_extension.py":894
 *     # Performance tuning: we construct a *lot* of these, so keep this
 *     # constructor as small and fast as possible
 *     def __init__(self, pstr, loc=0, msg=None, elem=None):             # <<<<<<<<<<<<<<
 *         self.loc = loc
 *         if msg is None:
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_1__init__, 0, __pyx_n_s_ParseBaseException___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__245)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 894, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_tuple__246);
  if (__Pyx_SetNameInClass(__pyx_t_13, __pyx_n_s_init, __pyx_t_7) < 0) __PYX_ERR(0, 894, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":906
 * 
 *     @classmethod
 *     def _from_exception(cls, pe):             # <<<<<<<<<<<<<<
 *         """
 *         internal factory method to simplify creating one type of ParseException
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_3_from_exception, __Pyx_CYFUNCTION_CLASSMETHOD, __pyx_n_s_ParseBaseException__from_excepti, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__248)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 906, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":905
 *         self.args = (pstr, loc, msg)
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def _from_exception(cls, pe):
 *         """
 */
  __pyx_t_9 = __Pyx_Method_ClassMethod(__pyx_t_7); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_SetNameInClass(__pyx_t_13, __pyx_n_s_from_exception, __pyx_t_9) < 0) __PYX_ERR(0, 906, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":913
 *         return cls(pe.pstr, pe.loc, pe.msg, pe.parserElement)
 * 
 *     def __getattr__(self, aname):             # <<<<<<<<<<<<<<
 *         """supported attributes by name are:
 *            - lineno - returns the line number of the exception text
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_5__getattr__, 0, __pyx_n_s_ParseBaseException___getattr, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__250)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 913, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_13, __pyx_n_s_getattr, __pyx_t_9) < 0) __PYX_ERR(0, 913, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":928
 *             raise AttributeError(aname)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if self.pstr:
 *             if self.loc >= len(self.pstr):
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_7__str__, 0, __pyx_n_s_ParseBaseException___str, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__252)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 928, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_13, __pyx_n_s_str, __pyx_t_9) < 0) __PYX_ERR(0, 928, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":939
 *                 (self.msg, foundstr, self.loc, self.lineno, self.column))
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return str(self)
 * 
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_9__repr__, 0, __pyx_n_s_ParseBaseException___repr, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__254)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 939, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_13, __pyx_n_s_repr, __pyx_t_9) < 0) __PYX_ERR(0, 939, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":942
 *         return str(self)
 * 
 *     def markInputline(self, markerString=">!<"):             # <<<<<<<<<<<<<<
 *         """Extracts the exception line from the input string, and marks
 *            the location of the exception with a special symbol.
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_11markInputline, 0, __pyx_n_s_ParseBaseException_markInputline, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__256)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_tuple__257);
  if (__Pyx_SetNameInClass(__pyx_t_13, __pyx_n_s_markInputline, __pyx_t_9) < 0) __PYX_ERR(0, 942, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":953
 *         return line_str.strip()
 * 
 *     def __dir__(self):             # <<<<<<<<<<<<<<
 *         return "lineno col line".split() + dir(type(self))
 * 
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_18ParseBaseException_13__dir__, 0, __pyx_n_s_ParseBaseException___dir, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__259)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 953, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_13, __pyx_n_s_dir, __pyx_t_9) < 0) __PYX_ERR(0, 953, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":889
 * 
 * 
 * class ParseBaseException(Exception):             # <<<<<<<<<<<<<<
 *     """base exception class for all parsing runtime exceptions"""
 * 
 */
  __pyx_t_9 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_ParseBaseException, __pyx_t_1, __pyx_t_13, NULL, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 889, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ParseBaseException, __pyx_t_9) < 0) __PYX_ERR(0, 889, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":957
 * 
 * 
 * class ParseException(ParseBaseException):             # <<<<<<<<<<<<<<
 *     """
 *     Exception thrown when parse expressions don't match class;
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParseBaseException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_13 = __Pyx_Py3MetaclassPrepare(__pyx_t_1, __pyx_t_2, __pyx_n_s_ParseException, __pyx_n_s_ParseException, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_Exception_thrown_when_parse_exp); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);

  /* "Aspidites/_vendor/pyparsing_extension.py":981
 * 
 *     @staticmethod
 *     def explain(exc, depth=16):             # <<<<<<<<<<<<<<
 *         """
 *         Method to take an exception and translate the Python internal traceback into a list
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_14ParseException_1explain, __Pyx_CYFUNCTION_STATICMETHOD, __pyx_n_s_ParseException_explain, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__261)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 981, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_tuple__262);

  /* "Aspidites/_vendor/pyparsing_extension.py":980
 *     """
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def explain(exc, depth=16):
 *         """
 */
  __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_builtin_staticmethod, __pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 980, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_SetNameInClass(__pyx_t_13, __pyx_n_s_explain, __pyx_t_7) < 0) __PYX_ERR(0, 981, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":957
 * 
 * 
 * class ParseException(ParseBaseException):             # <<<<<<<<<<<<<<
 *     """
 *     Exception thrown when parse expressions don't match class;
 */
  __pyx_t_7 = __Pyx_Py3ClassCreate(__pyx_t_1, __pyx_n_s_ParseException, __pyx_t_2, __pyx_t_13, NULL, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ParseException, __pyx_t_7) < 0) __PYX_ERR(0, 957, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1050
 * 
 * 
 * class ParseFatalException(ParseBaseException):             # <<<<<<<<<<<<<<
 *     """user-throwable exception thrown when inconsistent parse content
 *        is found; stops all parsing immediately"""
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParseBaseException); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1050, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1050, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1050, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_13 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_1, __pyx_n_s_ParseFatalException, __pyx_n_s_ParseFatalException, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_user_throwable_exception_thrown); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1050, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_7 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_ParseFatalException, __pyx_t_1, __pyx_t_13, NULL, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1050, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ParseFatalException, __pyx_t_7) < 0) __PYX_ERR(0, 1050, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1056
 * 
 * 
 * class ParseSyntaxException(ParseFatalException):             # <<<<<<<<<<<<<<
 *     """just like :class:`ParseFatalException`, but thrown internally
 *     when an :class:`ErrorStop<And._ErrorStop>` ('-' operator) indicates
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParseFatalException); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1056, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1056, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1056, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_13 = __Pyx_Py3MetaclassPrepare(__pyx_t_1, __pyx_t_2, __pyx_n_s_ParseSyntaxException, __pyx_n_s_ParseSyntaxException, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_just_like_class_ParseFatalExcept); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1056, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_7 = __Pyx_Py3ClassCreate(__pyx_t_1, __pyx_n_s_ParseSyntaxException, __pyx_t_2, __pyx_t_13, NULL, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1056, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ParseSyntaxException, __pyx_t_7) < 0) __PYX_ERR(0, 1056, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1078
 * # ~ self.reparseLoc = restartLoc
 * 
 * class RecursiveGrammarException(Exception):             # <<<<<<<<<<<<<<
 *     """exception thrown by :class:`ParserElement.validate` if the
 *     grammar could be improperly recursive
 */
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
  __Pyx_GIVEREF(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
  __pyx_t_1 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_13 = __Pyx_Py3MetaclassPrepare(__pyx_t_1, __pyx_t_2, __pyx_n_s_RecursiveGrammarException, __pyx_n_s_RecursiveGrammarException, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_exception_thrown_by_class_Parser); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);

  /* "Aspidites/_vendor/pyparsing_extension.py":1083
 *     """
 * 
 *     def __init__(self, parseElementList):             # <<<<<<<<<<<<<<
 *         self.parseElementTrace = parseElementList
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_25RecursiveGrammarException_1__init__, 0, __pyx_n_s_RecursiveGrammarException___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__264)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1083, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_13, __pyx_n_s_init, __pyx_t_7) < 0) __PYX_ERR(0, 1083, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1086
 *         self.parseElementTrace = parseElementList
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "RecursiveGrammarException: %s" % self.parseElementTrace
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_25RecursiveGrammarException_3__str__, 0, __pyx_n_s_RecursiveGrammarException___str, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__266)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1086, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_13, __pyx_n_s_str, __pyx_t_7) < 0) __PYX_ERR(0, 1086, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1078
 * # ~ self.reparseLoc = restartLoc
 * 
 * class RecursiveGrammarException(Exception):             # <<<<<<<<<<<<<<
 *     """exception thrown by :class:`ParserElement.validate` if the
 *     grammar could be improperly recursive
 */
  __pyx_t_7 = __Pyx_Py3ClassCreate(__pyx_t_1, __pyx_n_s_RecursiveGrammarException, __pyx_t_2, __pyx_t_13, NULL, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_RecursiveGrammarException, __pyx_t_7) < 0) __PYX_ERR(0, 1078, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1090
 * 
 * 
 * class ParserElement(object):             # <<<<<<<<<<<<<<
 *     """Abstract base level parser element class."""
 *     DEFAULT_WHITE_CHARS = " \n\t\r"
 */
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_tuple__267); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1090, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_tuple__267, __pyx_n_s_ParserElement, __pyx_n_s_ParserElement, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_Abstract_base_level_parser_eleme); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1090, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "Aspidites/_vendor/pyparsing_extension.py":1092
 * class ParserElement(object):
 *     """Abstract base level parser element class."""
 *     DEFAULT_WHITE_CHARS = " \n\t\r"             # <<<<<<<<<<<<<<
 *     verbose_stacktrace = False
 * 
 */
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_DEFAULT_WHITE_CHARS, __pyx_kp_s__268) < 0) __PYX_ERR(0, 1092, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1093
 *     """Abstract base level parser element class."""
 *     DEFAULT_WHITE_CHARS = " \n\t\r"
 *     verbose_stacktrace = False             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_verbose_stacktrace, Py_False) < 0) __PYX_ERR(0, 1093, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1096
 * 
 *     @staticmethod
 *     def setDefaultWhitespaceChars(chars):             # <<<<<<<<<<<<<<
 *         r"""
 *         Overrides the default whitespace chars
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_1setDefaultWhitespaceChars, __Pyx_CYFUNCTION_STATICMETHOD, __pyx_n_s_ParserElement_setDefaultWhitespa, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__270)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1096, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);

  /* "Aspidites/_vendor/pyparsing_extension.py":1095
 *     verbose_stacktrace = False
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def setDefaultWhitespaceChars(chars):
 *         r"""
 */
  __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_builtin_staticmethod, __pyx_t_13); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1095, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_setDefaultWhitespaceChars, __pyx_t_7) < 0) __PYX_ERR(0, 1096, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1112
 * 
 *     @staticmethod
 *     def inlineLiteralsUsing(cls):             # <<<<<<<<<<<<<<
 *         """
 *         Set class to be used for inclusion of string literals into a parser.
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_3inlineLiteralsUsing, __Pyx_CYFUNCTION_STATICMETHOD, __pyx_n_s_ParserElement_inlineLiteralsUsin, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__272)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":1111
 *         ParserElement.DEFAULT_WHITE_CHARS = chars
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def inlineLiteralsUsing(cls):
 *         """
 */
  __pyx_t_13 = __Pyx_PyObject_CallOneArg(__pyx_builtin_staticmethod, __pyx_t_7); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_inlineLiteralsUsing, __pyx_t_13) < 0) __PYX_ERR(0, 1112, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1134
 * 
 *     @classmethod
 *     def _trim_traceback(cls, tb):             # <<<<<<<<<<<<<<
 *         while tb.tb_next:
 *             tb = tb.tb_next
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_5_trim_traceback, __Pyx_CYFUNCTION_CLASSMETHOD, __pyx_n_s_ParserElement__trim_traceback, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__274)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);

  /* "Aspidites/_vendor/pyparsing_extension.py":1133
 *         ParserElement._literalStringClass = cls
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def _trim_traceback(cls, tb):
 *         while tb.tb_next:
 */
  __pyx_t_7 = __Pyx_Method_ClassMethod(__pyx_t_13); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_trim_traceback, __pyx_t_7) < 0) __PYX_ERR(0, 1134, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1139
 *         return tb
 * 
 *     def __init__(self, savelist=False):             # <<<<<<<<<<<<<<
 *         self.parseAction = list()
 *         self.failAction = None
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_7__init__, 0, __pyx_n_s_ParserElement___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__276)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_tuple__277);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_init, __pyx_t_7) < 0) __PYX_ERR(0, 1139, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1162
 *         self.callDuringTry = False
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         """
 *         Make a copy of this :class:`ParserElement`.  Useful for defining
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_9copy, 0, __pyx_n_s_ParserElement_copy, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__279)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_copy, __pyx_t_7) < 0) __PYX_ERR(0, 1162, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1191
 *         return cpy
 * 
 *     def setName(self, name):             # <<<<<<<<<<<<<<
 *         """
 *         Define name for this expression, makes debugging and exception messages clearer.
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_11setName, 0, __pyx_n_s_ParserElement_setName, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__281)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_setName, __pyx_t_7) < 0) __PYX_ERR(0, 1191, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1206
 *         return self
 * 
 *     def setResultsName(self, name, listAllMatches=False):             # <<<<<<<<<<<<<<
 *         """
 *         Define name for referencing matching tokens as a nested attribute
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_13setResultsName, 0, __pyx_n_s_ParserElement_setResultsName, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__283)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_tuple__284);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_setResultsName_2, __pyx_t_7) < 0) __PYX_ERR(0, 1206, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1229
 *         return self._setResultsName(name, listAllMatches)
 * 
 *     def _setResultsName(self, name, listAllMatches=False):             # <<<<<<<<<<<<<<
 *         newself = self.copy()
 *         if name.endswith("*"):
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_15_setResultsName, 0, __pyx_n_s_ParserElement__setResultsName, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__286)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_tuple__287);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_setResultsName, __pyx_t_7) < 0) __PYX_ERR(0, 1229, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1238
 *         return newself
 * 
 *     def setBreak(self, breakFlag=True):             # <<<<<<<<<<<<<<
 *         """Method to invoke the Python pdb debugger when this element is
 *            about to be parsed. Set ``breakFlag`` to True to enable, False to
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_17setBreak, 0, __pyx_n_s_ParserElement_setBreak, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__289)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_tuple__290);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_setBreak, __pyx_t_7) < 0) __PYX_ERR(0, 1238, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1259
 *         return self
 * 
 *     def setParseAction(self, *fns, **kwargs):             # <<<<<<<<<<<<<<
 *         from .pyparsing import _trim_arity
 *         """
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_19setParseAction, 0, __pyx_n_s_ParserElement_setParseAction, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__292)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_setParseAction, __pyx_t_7) < 0) __PYX_ERR(0, 1259, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1309
 *         return self
 * 
 *     def addParseAction(self, *fns, **kwargs):             # <<<<<<<<<<<<<<
 *         from .pyparsing import _trim_arity
 *         """
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_21addParseAction, 0, __pyx_n_s_ParserElement_addParseAction, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__294)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_addParseAction, __pyx_t_7) < 0) __PYX_ERR(0, 1309, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1320
 *         return self
 * 
 *     def addCondition(self, *fns, **kwargs):             # <<<<<<<<<<<<<<
 *         from .pyparsing import conditionAsParseAction
 *         """Add a boolean predicate function to expression's list of parse actions. See
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_23addCondition, 0, __pyx_n_s_ParserElement_addCondition, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__296)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_addCondition, __pyx_t_7) < 0) __PYX_ERR(0, 1320, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1346
 *         return self
 * 
 *     def setFailAction(self, fn):             # <<<<<<<<<<<<<<
 *         """Define action to perform if parsing fails at this expression.
 *            Fail acton fn is a callable function that takes the arguments
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_25setFailAction, 0, __pyx_n_s_ParserElement_setFailAction, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__298)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1346, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_setFailAction, __pyx_t_7) < 0) __PYX_ERR(0, 1346, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1359
 *         return self
 * 
 *     def _skipIgnorables(self, instring, loc):             # <<<<<<<<<<<<<<
 *         exprsFound = True
 *         while exprsFound:
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_27_skipIgnorables, 0, __pyx_n_s_ParserElement__skipIgnorables, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__300)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_skipIgnorables, __pyx_t_7) < 0) __PYX_ERR(0, 1359, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1372
 *         return loc
 * 
 *     def preParse(self, instring, loc):             # <<<<<<<<<<<<<<
 *         if self.ignoreExprs:
 *             loc = self._skipIgnorables(instring, loc)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_29preParse, 0, __pyx_n_s_ParserElement_preParse, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__302)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_preParse, __pyx_t_7) < 0) __PYX_ERR(0, 1372, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1384
 *         return loc
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         return loc, []
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_31parseImpl, 0, __pyx_n_s_ParserElement_parseImpl, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__304)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_tuple__305);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_parseImpl, __pyx_t_7) < 0) __PYX_ERR(0, 1384, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1387
 *         return loc, []
 * 
 *     def postParse(self, instring, loc, tokenlist):             # <<<<<<<<<<<<<<
 *         return tokenlist
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_33postParse, 0, __pyx_n_s_ParserElement_postParse, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__307)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_postParse, __pyx_t_7) < 0) __PYX_ERR(0, 1387, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1391
 * 
 *     # ~ @profile
 *     def _parseNoCache(self, instring, loc, doActions=True, callPreParse=True):             # <<<<<<<<<<<<<<
 *         TRY, MATCH, FAIL = 0, 1, 2
 *         debugging = (self.debug)  # and doActions)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_35_parseNoCache, 0, __pyx_n_s_ParserElement__parseNoCache, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__309)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_tuple__310);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_parseNoCache, __pyx_t_7) < 0) __PYX_ERR(0, 1391, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1479
 *         return loc, retTokens
 * 
 *     def tryParse(self, instring, loc):             # <<<<<<<<<<<<<<
 *         try:
 *             return self._parse(instring, loc, doActions=False)[0]
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_37tryParse, 0, __pyx_n_s_ParserElement_tryParse, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__312)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_tryParse, __pyx_t_7) < 0) __PYX_ERR(0, 1479, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1485
 *             raise ParseException(instring, loc, self.errmsg, self)
 * 
 *     def canParseNext(self, instring, loc):             # <<<<<<<<<<<<<<
 *         try:
 *             self.tryParse(instring, loc)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_39canParseNext, 0, __pyx_n_s_ParserElement_canParseNext, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__314)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_canParseNext, __pyx_t_7) < 0) __PYX_ERR(0, 1485, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1493
 *             return True
 * 
 *     class _UnboundedCache(object):             # <<<<<<<<<<<<<<
 *         def __init__(self):
 *             cache = {}
 */
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_tuple__315); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_13 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_tuple__315, __pyx_n_s_UnboundedCache, __pyx_n_s_ParserElement__UnboundedCache_5, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, (PyObject *) NULL); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);

  /* "Aspidites/_vendor/pyparsing_extension.py":1494
 * 
 *     class _UnboundedCache(object):
 *         def __init__(self):             # <<<<<<<<<<<<<<
 *             cache = {}
 *             self.not_in_cache = not_in_cache = object()
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_15_UnboundedCache_1__init__, 0, __pyx_n_s_ParserElement__UnboundedCache_6, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__317)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_13, __pyx_n_s_init, __pyx_t_9) < 0) __PYX_ERR(0, 1494, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1493
 *             return True
 * 
 *     class _UnboundedCache(object):             # <<<<<<<<<<<<<<
 *         def __init__(self):
 *             cache = {}
 */
  __pyx_t_9 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_UnboundedCache, __pyx_tuple__315, __pyx_t_13, NULL, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_UnboundedCache, __pyx_t_9) < 0) __PYX_ERR(0, 1493, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1515
 *             self.__len__ = types.MethodType(cache_len, self)
 * 
 *     if _OrderedDict is not None:             # <<<<<<<<<<<<<<
 *         class _FifoCache(object):
 *             def __init__(self, size):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_OrderedDict); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1515, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_16 = (__pyx_t_7 != Py_None);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_17 = (__pyx_t_16 != 0);
  if (__pyx_t_17) {

    /* "Aspidites/_vendor/pyparsing_extension.py":1516
 * 
 *     if _OrderedDict is not None:
 *         class _FifoCache(object):             # <<<<<<<<<<<<<<
 *             def __init__(self, size):
 *                 self.not_in_cache = not_in_cache = object()
 */
    __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_tuple__318); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1516, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_13 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_tuple__318, __pyx_n_s_FifoCache, __pyx_n_s_ParserElement__FifoCache, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, (PyObject *) NULL); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1516, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);

    /* "Aspidites/_vendor/pyparsing_extension.py":1517
 *     if _OrderedDict is not None:
 *         class _FifoCache(object):
 *             def __init__(self, size):             # <<<<<<<<<<<<<<
 *                 self.not_in_cache = not_in_cache = object()
 * 
 */
    __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_1__init__, 0, __pyx_n_s_ParserElement__FifoCache___init_5, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__320)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1517, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (__Pyx_SetNameInClass(__pyx_t_13, __pyx_n_s_init, __pyx_t_9) < 0) __PYX_ERR(0, 1517, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1516
 * 
 *     if _OrderedDict is not None:
 *         class _FifoCache(object):             # <<<<<<<<<<<<<<
 *             def __init__(self, size):
 *                 self.not_in_cache = not_in_cache = object()
 */
    __pyx_t_9 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_FifoCache, __pyx_tuple__318, __pyx_t_13, NULL, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1516, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_FifoCache, __pyx_t_9) < 0) __PYX_ERR(0, 1516, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1515
 *             self.__len__ = types.MethodType(cache_len, self)
 * 
 *     if _OrderedDict is not None:             # <<<<<<<<<<<<<<
 *         class _FifoCache(object):
 *             def __init__(self, size):
 */
    goto __pyx_L42;
  }

  /* "Aspidites/_vendor/pyparsing_extension.py":1545
 * 
 *     else:
 *         class _FifoCache(object):             # <<<<<<<<<<<<<<
 *             def __init__(self, size):
 *                 self.not_in_cache = not_in_cache = object()
 */
  /*else*/ {
    __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_tuple__321); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1545, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_13 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_tuple__321, __pyx_n_s_FifoCache, __pyx_n_s_ParserElement__FifoCache, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, (PyObject *) NULL); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1545, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);

    /* "Aspidites/_vendor/pyparsing_extension.py":1546
 *     else:
 *         class _FifoCache(object):
 *             def __init__(self, size):             # <<<<<<<<<<<<<<
 *                 self.not_in_cache = not_in_cache = object()
 * 
 */
    __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10_FifoCache_3__init__, 0, __pyx_n_s_ParserElement__FifoCache___init_5, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__323)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1546, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (__Pyx_SetNameInClass(__pyx_t_13, __pyx_n_s_init, __pyx_t_9) < 0) __PYX_ERR(0, 1546, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "Aspidites/_vendor/pyparsing_extension.py":1545
 * 
 *     else:
 *         class _FifoCache(object):             # <<<<<<<<<<<<<<
 *             def __init__(self, size):
 *                 self.not_in_cache = not_in_cache = object()
 */
    __pyx_t_9 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_FifoCache, __pyx_tuple__321, __pyx_t_13, NULL, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1545, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_FifoCache, __pyx_t_9) < 0) __PYX_ERR(0, 1545, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __pyx_L42:;

  /* "Aspidites/_vendor/pyparsing_extension.py":1574
 * 
 *     # argument cache for optimizing repeated calls when backtracking through recursive expressions
 *     packrat_cache = {}  # this is set later by enabledPackrat(); this is here so that resetCache() doesn't fail             # <<<<<<<<<<<<<<
 *     packrat_cache_lock = RLock()
 *     packrat_cache_stats = [0, 0]
 */
  __pyx_t_7 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_packrat_cache, __pyx_t_7) < 0) __PYX_ERR(0, 1574, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1575
 *     # argument cache for optimizing repeated calls when backtracking through recursive expressions
 *     packrat_cache = {}  # this is set later by enabledPackrat(); this is here so that resetCache() doesn't fail
 *     packrat_cache_lock = RLock()             # <<<<<<<<<<<<<<
 *     packrat_cache_stats = [0, 0]
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_RLock); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1575, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_9 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_13))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_13);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_13, function);
    }
  }
  __pyx_t_7 = (__pyx_t_9) ? __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_t_9) : __Pyx_PyObject_CallNoArg(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1575, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_packrat_cache_lock, __pyx_t_7) < 0) __PYX_ERR(0, 1575, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1576
 *     packrat_cache = {}  # this is set later by enabledPackrat(); this is here so that resetCache() doesn't fail
 *     packrat_cache_lock = RLock()
 *     packrat_cache_stats = [0, 0]             # <<<<<<<<<<<<<<
 * 
 *     # this method gets repeatedly called during backtracking with the same arguments -
 */
  __pyx_t_7 = PyList_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  PyList_SET_ITEM(__pyx_t_7, 0, __pyx_int_0);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  PyList_SET_ITEM(__pyx_t_7, 1, __pyx_int_0);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_packrat_cache_stats, __pyx_t_7) < 0) __PYX_ERR(0, 1576, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1580
 *     # this method gets repeatedly called during backtracking with the same arguments -
 *     # we can cache these arguments and save ourselves the trouble of re-parsing the contained expression
 *     def _parseCache(self, instring, loc, doActions=True, callPreParse=True):             # <<<<<<<<<<<<<<
 *         HIT, MISS = 0, 1
 *         lookup = (self, instring, loc, callPreParse, doActions)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_41_parseCache, 0, __pyx_n_s_ParserElement__parseCache, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__325)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1580, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_tuple__326);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_parseCache, __pyx_t_7) < 0) __PYX_ERR(0, 1580, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1603
 *                 return value[0], value[1].copy()
 * 
 *     _parse = _parseNoCache             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __pyx_t_7 = PyObject_GetItem(__pyx_t_1, __pyx_n_s_parseNoCache);
  if (unlikely(!__pyx_t_7)) {
    PyErr_Clear();
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_parseNoCache);
  }
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1603, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_parse, __pyx_t_7) < 0) __PYX_ERR(0, 1603, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1606
 * 
 *     @staticmethod
 *     def resetCache():             # <<<<<<<<<<<<<<
 *         ParserElement.packrat_cache.clear()
 *         ParserElement.packrat_cache_stats[:] = [0] * len(ParserElement.packrat_cache_stats)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_43resetCache, __Pyx_CYFUNCTION_STATICMETHOD, __pyx_n_s_ParserElement_resetCache, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__327)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1606, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":1605
 *     _parse = _parseNoCache
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def resetCache():
 *         ParserElement.packrat_cache.clear()
 */
  __pyx_t_13 = __Pyx_PyObject_CallOneArg(__pyx_builtin_staticmethod, __pyx_t_7); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1605, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_resetCache, __pyx_t_13) < 0) __PYX_ERR(0, 1606, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1610
 *         ParserElement.packrat_cache_stats[:] = [0] * len(ParserElement.packrat_cache_stats)
 * 
 *     _packratEnabled = False             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_packratEnabled, Py_False) < 0) __PYX_ERR(0, 1610, __pyx_L1_error)

  /* "Aspidites/_vendor/pyparsing_extension.py":1613
 * 
 *     @staticmethod
 *     def enablePackrat(cache_size_limit=128):             # <<<<<<<<<<<<<<
 *         """Enables "packrat" parsing, which adds memoizing to the parsing logic.
 *            Repeated parse attempts at the same string location (which happens
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_45enablePackrat, __Pyx_CYFUNCTION_STATICMETHOD, __pyx_n_s_ParserElement_enablePackrat, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__329)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_13, __pyx_tuple__330);

  /* "Aspidites/_vendor/pyparsing_extension.py":1612
 *     _packratEnabled = False
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def enablePackrat(cache_size_limit=128):
 *         """Enables "packrat" parsing, which adds memoizing to the parsing logic.
 */
  __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_builtin_staticmethod, __pyx_t_13); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_enablePackrat, __pyx_t_7) < 0) __PYX_ERR(0, 1613, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1647
 *             ParserElement._parse = ParserElement._parseCache
 * 
 *     def parseString(self, instring, parseAll=False):             # <<<<<<<<<<<<<<
 *         """
 *         Execute the parse expression with the given string.
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_47parseString, 0, __pyx_n_s_ParserElement_parseString, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__332)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1647, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_tuple__333);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_parseString, __pyx_t_7) < 0) __PYX_ERR(0, 1647, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1706
 *             return tokens
 * 
 *     def scanString(self, instring, maxMatches=sys.maxsize, overlap=False):             # <<<<<<<<<<<<<<
 *         """
 *         Scan the input string for expression matches.  Each match will return the
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_49scanString, 0, __pyx_n_s_ParserElement_scanString, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__63)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_7, sizeof(__pyx_defaults2), 1)) __PYX_ERR(0, 1706, __pyx_L1_error)
  __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_sys); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_n_s_maxsize); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_CyFunction_Defaults(__pyx_defaults2, __pyx_t_7)->__pyx_arg_maxMatches = __pyx_t_9;
  __Pyx_GIVEREF(__pyx_t_9);
  __pyx_t_9 = 0;
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_7, __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_32__defaults__);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_scanString, __pyx_t_7) < 0) __PYX_ERR(0, 1706, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1780
 *                 raise exc
 * 
 *     def transformString(self, instring):             # <<<<<<<<<<<<<<
 *         """
 *         Extension to :class:`scanString`, to modify matching text with modified tokens that may
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_52transformString, 0, __pyx_n_s_ParserElement_transformString, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__336)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_transformString, __pyx_t_7) < 0) __PYX_ERR(0, 1780, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1828
 *                 raise exc
 * 
 *     def searchString(self, instring, maxMatches=sys.maxsize):             # <<<<<<<<<<<<<<
 *         """
 *         Another extension to :class:`scanString`, simplifying the access to the tokens found
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_54searchString, 0, __pyx_n_s_ParserElement_searchString, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__338)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_7, sizeof(__pyx_defaults3), 1)) __PYX_ERR(0, 1828, __pyx_L1_error)
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_sys); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_maxsize); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_CyFunction_Defaults(__pyx_defaults3, __pyx_t_7)->__pyx_arg_maxMatches = __pyx_t_13;
  __Pyx_GIVEREF(__pyx_t_13);
  __pyx_t_13 = 0;
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_7, __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_34__defaults__);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_searchString, __pyx_t_7) < 0) __PYX_ERR(0, 1828, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1860
 *                 raise exc
 * 
 *     def split(self, instring, maxsplit=sys.maxsize, includeSeparators=False):             # <<<<<<<<<<<<<<
 *         """
 *         Generator method to split a string using the given expression as a separator.
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_56split, 0, __pyx_n_s_ParserElement_split, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__64)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1860, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_7, sizeof(__pyx_defaults4), 1)) __PYX_ERR(0, 1860, __pyx_L1_error)
  __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_sys); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1860, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_n_s_maxsize); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1860, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_CyFunction_Defaults(__pyx_defaults4, __pyx_t_7)->__pyx_arg_maxsplit = __pyx_t_9;
  __Pyx_GIVEREF(__pyx_t_9);
  __pyx_t_9 = 0;
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_7, __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_36__defaults__);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_split, __pyx_t_7) < 0) __PYX_ERR(0, 1860, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1885
 *         yield instring[last:]
 * 
 *     def __add__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of + operator - returns :class:`And`. Adding strings to a ParserElement
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_59__add__, 0, __pyx_n_s_ParserElement___add, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__341)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_add, __pyx_t_7) < 0) __PYX_ERR(0, 1885, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1924
 *         return And([self, other])
 * 
 *     def __radd__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of + operator when left operand is not a :class:`ParserElement`
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_61__radd__, 0, __pyx_n_s_ParserElement___radd, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__343)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1924, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_radd, __pyx_t_7) < 0) __PYX_ERR(0, 1924, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1940
 *         return other + self
 * 
 *     def __sub__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of - operator, returns :class:`And` with error stop
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_63__sub__, 0, __pyx_n_s_ParserElement___sub, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__345)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1940, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_sub_2, __pyx_t_7) < 0) __PYX_ERR(0, 1940, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1953
 *         return self + And._ErrorStop() + other
 * 
 *     def __rsub__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of - operator when left operand is not a :class:`ParserElement`
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_65__rsub__, 0, __pyx_n_s_ParserElement___rsub, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__347)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1953, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_rsub, __pyx_t_7) < 0) __PYX_ERR(0, 1953, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1965
 *         return other - self
 * 
 *     def __mul__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of * operator, allows use of ``expr * 3`` in place of
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_67__mul__, 0, __pyx_n_s_ParserElement___mul, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__349)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1965, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_mul, __pyx_t_7) < 0) __PYX_ERR(0, 1965, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2043
 *         return ret
 * 
 *     def __rmul__(self, other):             # <<<<<<<<<<<<<<
 *         return self.__mul__(other)
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_69__rmul__, 0, __pyx_n_s_ParserElement___rmul, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__351)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2043, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_rmul, __pyx_t_7) < 0) __PYX_ERR(0, 2043, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2046
 *         return self.__mul__(other)
 * 
 *     def __or__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of | operator - returns :class:`MatchFirst`
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_71__or__, 0, __pyx_n_s_ParserElement___or, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__353)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2046, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_or, __pyx_t_7) < 0) __PYX_ERR(0, 2046, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2062
 *         return MatchFirst([self, other])
 * 
 *     def __ror__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of | operator when left operand is not a :class:`ParserElement`
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_73__ror__, 0, __pyx_n_s_ParserElement___ror, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__355)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2062, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_ror, __pyx_t_7) < 0) __PYX_ERR(0, 2062, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2074
 *         return other | self
 * 
 *     def __xor__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of ^ operator - returns :class:`Or`
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_75__xor__, 0, __pyx_n_s_ParserElement___xor, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__357)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2074, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_xor, __pyx_t_7) < 0) __PYX_ERR(0, 2074, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2087
 *         return Or([self, other])
 * 
 *     def __rxor__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of ^ operator when left operand is not a :class:`ParserElement`
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_77__rxor__, 0, __pyx_n_s_ParserElement___rxor, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__359)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2087, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_rxor, __pyx_t_7) < 0) __PYX_ERR(0, 2087, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2099
 *         return other ^ self
 * 
 *     def __and__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of & operator - returns :class:`Each`
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_79__and__, 0, __pyx_n_s_ParserElement___and, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__361)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2099, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_and, __pyx_t_7) < 0) __PYX_ERR(0, 2099, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2112
 *         return Each([self, other])
 * 
 *     def __rand__(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of & operator when left operand is not a :class:`ParserElement`
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_81__rand__, 0, __pyx_n_s_ParserElement___rand, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__363)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_rand, __pyx_t_7) < 0) __PYX_ERR(0, 2112, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2124
 *         return other & self
 * 
 *     def __invert__(self):             # <<<<<<<<<<<<<<
 *         """
 *         Implementation of ~ operator - returns :class:`NotAny`
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_83__invert__, 0, __pyx_n_s_ParserElement___invert, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__365)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_invert, __pyx_t_7) < 0) __PYX_ERR(0, 2124, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2131
 *         return NotAny(self)
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         # must implement __iter__ to override legacy use of sequential access to __getitem__ to
 *         # iterate over a sequence
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_85__iter__, 0, __pyx_n_s_ParserElement___iter, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__367)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_iter, __pyx_t_7) < 0) __PYX_ERR(0, 2131, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2136
 *         raise TypeError('%r object is not iterable' % self.__class__.__name__)
 * 
 *     def __getitem__(self, key):             # <<<<<<<<<<<<<<
 *         """
 *         use ``[]`` indexing notation as a short form for expression repetition:
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_87__getitem__, 0, __pyx_n_s_ParserElement___getitem, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__369)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_getitem, __pyx_t_7) < 0) __PYX_ERR(0, 2136, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2172
 *         return ret
 * 
 *     def __call__(self, name=None):             # <<<<<<<<<<<<<<
 *         """
 *         Shortcut for :class:`setResultsName`, with ``listAllMatches=False``.
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_89__call__, 0, __pyx_n_s_ParserElement___call, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__371)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_tuple__372);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_call, __pyx_t_7) < 0) __PYX_ERR(0, 2172, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2192
 *             return self.copy()
 * 
 *     def suppress(self):             # <<<<<<<<<<<<<<
 *         """
 *         Suppresses the output of this :class:`ParserElement`; useful to keep punctuation from
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_91suppress, 0, __pyx_n_s_ParserElement_suppress, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__374)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_suppress, __pyx_t_7) < 0) __PYX_ERR(0, 2192, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2200
 *         return Suppress(self)
 * 
 *     def leaveWhitespace(self):             # <<<<<<<<<<<<<<
 *         """
 *         Disables the skipping of whitespace before matching the characters in the
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_93leaveWhitespace, 0, __pyx_n_s_ParserElement_leaveWhitespace, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__376)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_leaveWhitespace, __pyx_t_7) < 0) __PYX_ERR(0, 2200, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2209
 *         return self
 * 
 *     def setWhitespaceChars(self, chars):             # <<<<<<<<<<<<<<
 *         """
 *         Overrides the default whitespace chars
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_95setWhitespaceChars, 0, __pyx_n_s_ParserElement_setWhitespaceChars, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__378)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_setWhitespaceChars, __pyx_t_7) < 0) __PYX_ERR(0, 2209, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2218
 *         return self
 * 
 *     def parseWithTabs(self):             # <<<<<<<<<<<<<<
 *         """
 *         Overrides default behavior to expand ``<TAB>``s to spaces before parsing the input string.
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_97parseWithTabs, 0, __pyx_n_s_ParserElement_parseWithTabs, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__380)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_parseWithTabs, __pyx_t_7) < 0) __PYX_ERR(0, 2218, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2227
 *         return self
 * 
 *     def ignore(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Define expression to be ignored (e.g., comments) while doing pattern
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_99ignore, 0, __pyx_n_s_ParserElement_ignore, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__382)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_ignore, __pyx_t_7) < 0) __PYX_ERR(0, 2227, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2252
 *         return self
 * 
 *     def setDebugActions(self, startAction, successAction, exceptionAction):             # <<<<<<<<<<<<<<
 *         """
 *         Enable display of debugging messages while doing pattern matching.
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_101setDebugActions, 0, __pyx_n_s_ParserElement_setDebugActions, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__384)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_setDebugActions, __pyx_t_7) < 0) __PYX_ERR(0, 2252, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2262
 *         return self
 * 
 *     def setDebug(self, flag=True):             # <<<<<<<<<<<<<<
 *         """
 *         Enable display of debugging messages while doing pattern matching.
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_103setDebug, 0, __pyx_n_s_ParserElement_setDebug, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__386)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_tuple__387);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_setDebug, __pyx_t_7) < 0) __PYX_ERR(0, 2262, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2305
 *         return self
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return self.name
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_105__str__, 0, __pyx_n_s_ParserElement___str, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__389)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_str, __pyx_t_7) < 0) __PYX_ERR(0, 2305, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2308
 *         return self.name
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return str(self)
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_107__repr__, 0, __pyx_n_s_ParserElement___repr, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__391)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_repr, __pyx_t_7) < 0) __PYX_ERR(0, 2308, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2311
 *         return str(self)
 * 
 *     def streamline(self):             # <<<<<<<<<<<<<<
 *         self.streamlined = True
 *         self.strRepr = None
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_109streamline, 0, __pyx_n_s_ParserElement_streamline, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__393)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_streamline, __pyx_t_7) < 0) __PYX_ERR(0, 2311, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2316
 *         return self
 * 
 *     def checkRecursion(self, parseElementList):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_111checkRecursion, 0, __pyx_n_s_ParserElement_checkRecursion, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__395)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_checkRecursion, __pyx_t_7) < 0) __PYX_ERR(0, 2316, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2319
 *         pass
 * 
 *     def validate(self, validateTrace=None):             # <<<<<<<<<<<<<<
 *         """
 *         Check defined expressions for valid structure, check for infinite recursive definitions.
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_113validate, 0, __pyx_n_s_ParserElement_validate, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__397)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2319, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_tuple__398);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_validate, __pyx_t_7) < 0) __PYX_ERR(0, 2319, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2325
 *         self.checkRecursion([])
 * 
 *     def parseFile(self, file_or_filename, parseAll=False):             # <<<<<<<<<<<<<<
 *         """
 *         Execute the parse expression on the given file or filename.
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_115parseFile, 0, __pyx_n_s_ParserElement_parseFile, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__400)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_tuple__401);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_parseFile, __pyx_t_7) < 0) __PYX_ERR(0, 2325, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2347
 *                 raise exc
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         if self is other:
 *             return True
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_117__eq__, 0, __pyx_n_s_ParserElement___eq, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__403)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_eq, __pyx_t_7) < 0) __PYX_ERR(0, 2347, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2356
 *         return False
 * 
 *     def __ne__(self, other):             # <<<<<<<<<<<<<<
 *         return not (self == other)
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_119__ne__, 0, __pyx_n_s_ParserElement___ne, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__405)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_ne, __pyx_t_7) < 0) __PYX_ERR(0, 2356, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2359
 *         return not (self == other)
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return id(self)
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_121__hash__, 0, __pyx_n_s_ParserElement___hash, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__407)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_hash, __pyx_t_7) < 0) __PYX_ERR(0, 2359, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2362
 *         return id(self)
 * 
 *     def __req__(self, other):             # <<<<<<<<<<<<<<
 *         return self == other
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_123__req__, 0, __pyx_n_s_ParserElement___req, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__409)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2362, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_req, __pyx_t_7) < 0) __PYX_ERR(0, 2362, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2365
 *         return self == other
 * 
 *     def __rne__(self, other):             # <<<<<<<<<<<<<<
 *         return not (self == other)
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_125__rne__, 0, __pyx_n_s_ParserElement___rne, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__411)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2365, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_rne, __pyx_t_7) < 0) __PYX_ERR(0, 2365, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2368
 *         return not (self == other)
 * 
 *     def matches(self, testString, parseAll=True):             # <<<<<<<<<<<<<<
 *         """
 *         Method for quick testing of a parser against a test string. Good for simple
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_127matches, 0, __pyx_n_s_ParserElement_matches, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__413)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_tuple__414);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_matches, __pyx_t_7) < 0) __PYX_ERR(0, 2368, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2388
 *             return False
 * 
 *     def runTests(self, tests, parseAll=True, comment='#',             # <<<<<<<<<<<<<<
 *                  fullDump=True, printResults=True, failureTests=False, postParse=None,
 *                  file=None):
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_13ParserElement_129runTests, 0, __pyx_n_s_ParserElement_runTests, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__416)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_tuple__417);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_runTests, __pyx_t_7) < 0) __PYX_ERR(0, 2388, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1090
 * 
 * 
 * class ParserElement(object):             # <<<<<<<<<<<<<<
 *     """Abstract base level parser element class."""
 *     DEFAULT_WHITE_CHARS = " \n\t\r"
 */
  __pyx_t_7 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_ParserElement, __pyx_tuple__267, __pyx_t_1, NULL, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1090, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ParserElement, __pyx_t_7) < 0) __PYX_ERR(0, 1090, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2552
 * 
 * 
 * class ParseExpression(ParserElement):             # <<<<<<<<<<<<<<
 *     """Abstract subclass of ParserElement, for combining and
 *     post-processing parsed tokens.
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_1, __pyx_n_s_ParseExpression, __pyx_n_s_ParseExpression, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_Abstract_subclass_of_ParserEleme); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":2557
 *     """
 * 
 *     def __init__(self, exprs, savelist=False):             # <<<<<<<<<<<<<<
 *         super(ParseExpression, self).__init__(savelist)
 *         if isinstance(exprs, _generatorType):
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_1__init__, 0, __pyx_n_s_ParseExpression___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__419)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_tuple__420);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_9) < 0) __PYX_ERR(0, 2557, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2579
 *         self.callPreparse = False
 * 
 *     def append(self, other):             # <<<<<<<<<<<<<<
 *         self.exprs.append(other)
 *         self.strRepr = None
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_3append, 0, __pyx_n_s_ParseExpression_append, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__422)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_append, __pyx_t_9) < 0) __PYX_ERR(0, 2579, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2584
 *         return self
 * 
 *     def leaveWhitespace(self):             # <<<<<<<<<<<<<<
 *         """Extends ``leaveWhitespace`` defined in base class, and also invokes ``leaveWhitespace`` on
 *            all contained expressions."""
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_5leaveWhitespace, 0, __pyx_n_s_ParseExpression_leaveWhitespace, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__424)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_leaveWhitespace, __pyx_t_9) < 0) __PYX_ERR(0, 2584, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2593
 *         return self
 * 
 *     def ignore(self, other):             # <<<<<<<<<<<<<<
 *         from .pyparsing import Suppress
 *         if isinstance(other, Suppress):
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_7ignore, 0, __pyx_n_s_ParseExpression_ignore, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__426)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_ignore, __pyx_t_9) < 0) __PYX_ERR(0, 2593, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2606
 *         return self
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         try:
 *             return super(ParseExpression, self).__str__()
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_9__str__, 0, __pyx_n_s_ParseExpression___str, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__428)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2606, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_str, __pyx_t_9) < 0) __PYX_ERR(0, 2606, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2616
 *         return self.strRepr
 * 
 *     def streamline(self):             # <<<<<<<<<<<<<<
 *         super(ParseExpression, self).streamline()
 * 
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_11streamline, 0, __pyx_n_s_ParseExpression_streamline, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__430)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_streamline, __pyx_t_9) < 0) __PYX_ERR(0, 2616, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2650
 *         return self
 * 
 *     def validate(self, validateTrace=None):             # <<<<<<<<<<<<<<
 *         tmp = (validateTrace if validateTrace is not None else [])[:] + [self]
 *         for e in self.exprs:
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_13validate, 0, __pyx_n_s_ParseExpression_validate, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__432)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2650, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_tuple__433);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_validate, __pyx_t_9) < 0) __PYX_ERR(0, 2650, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2656
 *         self.checkRecursion([])
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         ret = super(ParseExpression, self).copy()
 *         ret.exprs = [e.copy() for e in self.exprs]
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_15copy, 0, __pyx_n_s_ParseExpression_copy, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__435)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_copy, __pyx_t_9) < 0) __PYX_ERR(0, 2656, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2661
 *         return ret
 * 
 *     def _setResultsName(self, name, listAllMatches=False):             # <<<<<<<<<<<<<<
 *         if __diag__.warn_ungrouped_named_tokens_in_collection:
 *             for e in self.exprs:
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_15ParseExpression_17_setResultsName, 0, __pyx_n_s_ParseExpression__setResultsName, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__437)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_tuple__438);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_setResultsName, __pyx_t_9) < 0) __PYX_ERR(0, 2661, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2552
 * 
 * 
 * class ParseExpression(ParserElement):             # <<<<<<<<<<<<<<
 *     """Abstract subclass of ParserElement, for combining and
 *     post-processing parsed tokens.
 */
  __pyx_t_9 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_ParseExpression, __pyx_t_1, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ParseExpression, __pyx_t_9) < 0) __PYX_ERR(0, 2552, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2676
 * 
 * 
 * class Token(ParserElement):             # <<<<<<<<<<<<<<
 *     """Abstract :class:`ParserElement` subclass, for defining atomic
 *     matching patterns.
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2676, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2676, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2676, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_1, __pyx_t_2, __pyx_n_s_Token, __pyx_n_s_Token, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_Abstract_class_ParserElement_sub); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2676, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":2681
 *     """
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         super(Token, self).__init__(savelist=False)
 * 
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_5Token_1__init__, 0, __pyx_n_s_Token___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__440)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_9) < 0) __PYX_ERR(0, 2681, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2676
 * 
 * 
 * class Token(ParserElement):             # <<<<<<<<<<<<<<
 *     """Abstract :class:`ParserElement` subclass, for defining atomic
 *     matching patterns.
 */
  __pyx_t_9 = __Pyx_Py3ClassCreate(__pyx_t_1, __pyx_n_s_Token, __pyx_t_2, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2676, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Token, __pyx_t_9) < 0) __PYX_ERR(0, 2676, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2685
 * 
 * 
 * class Empty(Token):             # <<<<<<<<<<<<<<
 *     """An empty token, will always match.
 *     """
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Token); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2685, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2685, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2685, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_1, __pyx_n_s_Empty, __pyx_n_s_Empty, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_An_empty_token_will_always_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2685, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":2689
 *     """
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         super(Empty, self).__init__()
 *         self.name = "Empty"
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_5Empty_1__init__, 0, __pyx_n_s_Empty___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__442)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2689, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_9) < 0) __PYX_ERR(0, 2689, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2685
 * 
 * 
 * class Empty(Token):             # <<<<<<<<<<<<<<
 *     """An empty token, will always match.
 *     """
 */
  __pyx_t_9 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_Empty, __pyx_t_1, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2685, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Empty, __pyx_t_9) < 0) __PYX_ERR(0, 2685, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2696
 * 
 * 
 * class NoMatch(Token):             # <<<<<<<<<<<<<<
 *     """A token that will never match.
 *     """
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Token); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_1, __pyx_t_2, __pyx_n_s_NoMatch, __pyx_n_s_NoMatch, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_A_token_that_will_never_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":2700
 *     """
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         super(NoMatch, self).__init__()
 *         self.name = "NoMatch"
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7NoMatch_1__init__, 0, __pyx_n_s_NoMatch___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__444)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_9) < 0) __PYX_ERR(0, 2700, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2707
 *         self.errmsg = "Unmatchable token"
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         raise ParseException(instring, loc, self.errmsg, self)
 * 
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7NoMatch_3parseImpl, 0, __pyx_n_s_NoMatch_parseImpl, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__446)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_tuple__447);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_parseImpl, __pyx_t_9) < 0) __PYX_ERR(0, 2707, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2696
 * 
 * 
 * class NoMatch(Token):             # <<<<<<<<<<<<<<
 *     """A token that will never match.
 *     """
 */
  __pyx_t_9 = __Pyx_Py3ClassCreate(__pyx_t_1, __pyx_n_s_NoMatch, __pyx_t_2, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_NoMatch, __pyx_t_9) < 0) __PYX_ERR(0, 2696, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2711
 * 
 * 
 * class Literal(Token):             # <<<<<<<<<<<<<<
 *     """Token to exactly match a specified string.
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Token); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2711, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2711, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2711, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_1, __pyx_n_s_Literal, __pyx_n_s_Literal, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_Token_to_exactly_match_a_specifi); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2711, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":2726
 *     """
 * 
 *     def __init__(self, matchString):             # <<<<<<<<<<<<<<
 *         super(Literal, self).__init__()
 *         self.match = matchString
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7Literal_1__init__, 0, __pyx_n_s_Literal___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__449)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2726, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_9) < 0) __PYX_ERR(0, 2726, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2746
 *             self.__class__ = _SingleCharLiteral
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if instring[loc] == self.firstMatchChar and instring.startswith(self.match, loc):
 *             return loc + self.matchLen, self.match
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7Literal_3parseImpl, 0, __pyx_n_s_Literal_parseImpl, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__451)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2746, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_tuple__452);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_parseImpl, __pyx_t_9) < 0) __PYX_ERR(0, 2746, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2711
 * 
 * 
 * class Literal(Token):             # <<<<<<<<<<<<<<
 *     """Token to exactly match a specified string.
 * 
 */
  __pyx_t_9 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_Literal, __pyx_t_1, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2711, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Literal, __pyx_t_9) < 0) __PYX_ERR(0, 2711, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2752
 * 
 * 
 * class _SingleCharLiteral(Literal):             # <<<<<<<<<<<<<<
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if instring[loc] == self.firstMatchChar:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Literal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2752, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2752, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2752, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_1, __pyx_t_2, __pyx_n_s_SingleCharLiteral, __pyx_n_s_SingleCharLiteral, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, (PyObject *) NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2752, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":2753
 * 
 * class _SingleCharLiteral(Literal):
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if instring[loc] == self.firstMatchChar:
 *             return loc + 1, self.match
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_18_SingleCharLiteral_1parseImpl, 0, __pyx_n_s_SingleCharLiteral_parseImpl, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__454)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2753, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_tuple__455);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_parseImpl, __pyx_t_9) < 0) __PYX_ERR(0, 2753, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2752
 * 
 * 
 * class _SingleCharLiteral(Literal):             # <<<<<<<<<<<<<<
 *     def parseImpl(self, instring, loc, doActions=True):
 *         if instring[loc] == self.firstMatchChar:
 */
  __pyx_t_9 = __Pyx_Py3ClassCreate(__pyx_t_1, __pyx_n_s_SingleCharLiteral, __pyx_t_2, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2752, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SingleCharLiteral, __pyx_t_9) < 0) __PYX_ERR(0, 2752, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2759
 * 
 * 
 * _L = Literal             # <<<<<<<<<<<<<<
 * ParserElement._literalStringClass = Literal
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Literal); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2759, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_L_2, __pyx_t_2) < 0) __PYX_ERR(0, 2759, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2760
 * 
 * _L = Literal
 * ParserElement._literalStringClass = Literal             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Literal); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2760, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2760, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_1, __pyx_n_s_literalStringClass, __pyx_t_2) < 0) __PYX_ERR(0, 2760, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2763
 * 
 * 
 * class Keyword(Token):             # <<<<<<<<<<<<<<
 *     """Token to exactly match a specified string as a keyword, that is,
 *     it must be immediately followed by a non-keyword character.  Compare
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Token); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2763, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2763, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2763, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_1, __pyx_t_2, __pyx_n_s_Keyword, __pyx_n_s_Keyword, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_Token_to_exactly_match_a_specifi_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2763, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":2788
 *     For case-insensitive matching, use :class:`CaselessKeyword`.
 *     """
 *     DEFAULT_KEYWORD_CHARS = alphanums + "_$"             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, matchString, identChars=None, caseless=False):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_alphanums); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2788, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_13 = PyNumber_Add(__pyx_t_9, __pyx_kp_s__456); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2788, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_DEFAULT_KEYWORD_CHARS, __pyx_t_13) < 0) __PYX_ERR(0, 2788, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2790
 *     DEFAULT_KEYWORD_CHARS = alphanums + "_$"
 * 
 *     def __init__(self, matchString, identChars=None, caseless=False):             # <<<<<<<<<<<<<<
 *         super(Keyword, self).__init__()
 *         if identChars is None:
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7Keyword_1__init__, 0, __pyx_n_s_Keyword___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__458)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_13, __pyx_tuple__459);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_13) < 0) __PYX_ERR(0, 2790, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2811
 *         self.identChars = set(identChars)
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if self.caseless:
 *             if ((instring[loc:loc + self.matchLen].upper() == self.caselessmatch)
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7Keyword_3parseImpl, 0, __pyx_n_s_Keyword_parseImpl, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__461)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_13, __pyx_tuple__462);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_parseImpl, __pyx_t_13) < 0) __PYX_ERR(0, 2811, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2830
 *         raise ParseException(instring, loc, self.errmsg, self)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         c = super(Keyword, self).copy()
 *         c.identChars = Keyword.DEFAULT_KEYWORD_CHARS
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7Keyword_5copy, 0, __pyx_n_s_Keyword_copy, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__464)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_copy, __pyx_t_13) < 0) __PYX_ERR(0, 2830, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2836
 * 
 *     @staticmethod
 *     def setDefaultKeywordChars(chars):             # <<<<<<<<<<<<<<
 *         """Overrides the default Keyword chars
 *         """
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7Keyword_7setDefaultKeywordChars, __Pyx_CYFUNCTION_STATICMETHOD, __pyx_n_s_Keyword_setDefaultKeywordChars, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__466)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2836, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);

  /* "Aspidites/_vendor/pyparsing_extension.py":2835
 *         return c
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def setDefaultKeywordChars(chars):
 *         """Overrides the default Keyword chars
 */
  __pyx_t_9 = __Pyx_PyObject_CallOneArg(__pyx_builtin_staticmethod, __pyx_t_13); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_setDefaultKeywordChars, __pyx_t_9) < 0) __PYX_ERR(0, 2836, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2763
 * 
 * 
 * class Keyword(Token):             # <<<<<<<<<<<<<<
 *     """Token to exactly match a specified string as a keyword, that is,
 *     it must be immediately followed by a non-keyword character.  Compare
 */
  __pyx_t_9 = __Pyx_Py3ClassCreate(__pyx_t_1, __pyx_n_s_Keyword, __pyx_t_2, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2763, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Keyword, __pyx_t_9) < 0) __PYX_ERR(0, 2763, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2842
 * 
 * 
 * class CaselessLiteral(Literal):             # <<<<<<<<<<<<<<
 *     """Token to match a specified string, ignoring case of letters.
 *     Note: the matched results will always be in the case of the given
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Literal); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2842, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2842, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2842, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_1, __pyx_n_s_CaselessLiteral, __pyx_n_s_CaselessLiteral, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_Token_to_match_a_specified_strin); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2842, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":2854
 *     """
 * 
 *     def __init__(self, matchString):             # <<<<<<<<<<<<<<
 *         super(CaselessLiteral, self).__init__(matchString.upper())
 *         # Preserve the defining literal.
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_15CaselessLiteral_1__init__, 0, __pyx_n_s_CaselessLiteral___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__468)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2854, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_9) < 0) __PYX_ERR(0, 2854, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2861
 *         self.errmsg = "Expected " + self.name
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if instring[loc:loc + self.matchLen].upper() == self.match:
 *             return loc + self.matchLen, self.returnString
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_15CaselessLiteral_3parseImpl, 0, __pyx_n_s_CaselessLiteral_parseImpl, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__470)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_tuple__471);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_parseImpl, __pyx_t_9) < 0) __PYX_ERR(0, 2861, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2842
 * 
 * 
 * class CaselessLiteral(Literal):             # <<<<<<<<<<<<<<
 *     """Token to match a specified string, ignoring case of letters.
 *     Note: the matched results will always be in the case of the given
 */
  __pyx_t_9 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_CaselessLiteral, __pyx_t_1, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2842, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CaselessLiteral, __pyx_t_9) < 0) __PYX_ERR(0, 2842, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2867
 * 
 * 
 * class CaselessKeyword(Keyword):             # <<<<<<<<<<<<<<
 *     """
 *     Caseless version of :class:`Keyword`.
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Keyword); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2867, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2867, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2867, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_1, __pyx_t_2, __pyx_n_s_CaselessKeyword, __pyx_n_s_CaselessKeyword, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_Caseless_version_of_class_Keywo); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2867, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":2878
 *     """
 * 
 *     def __init__(self, matchString, identChars=None):             # <<<<<<<<<<<<<<
 *         super(CaselessKeyword, self).__init__(matchString, identChars, caseless=True)
 * 
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_15CaselessKeyword_1__init__, 0, __pyx_n_s_CaselessKeyword___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__473)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2878, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_tuple__474);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_9) < 0) __PYX_ERR(0, 2878, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2867
 * 
 * 
 * class CaselessKeyword(Keyword):             # <<<<<<<<<<<<<<
 *     """
 *     Caseless version of :class:`Keyword`.
 */
  __pyx_t_9 = __Pyx_Py3ClassCreate(__pyx_t_1, __pyx_n_s_CaselessKeyword, __pyx_t_2, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2867, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CaselessKeyword, __pyx_t_9) < 0) __PYX_ERR(0, 2867, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2882
 * 
 * 
 * class CloseMatch(Token):             # <<<<<<<<<<<<<<
 *     """A variation on :class:`Literal` which matches "close" matches,
 *     that is, strings with at most 'n' mismatching characters.
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Token); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2882, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2882, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2882, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_1, __pyx_n_s_CloseMatch, __pyx_n_s_CloseMatch, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_A_variation_on_class_Literal_whi); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2882, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":2916
 *     """
 * 
 *     def __init__(self, match_string, maxMismatches=1):             # <<<<<<<<<<<<<<
 *         super(CloseMatch, self).__init__()
 *         self.name = match_string
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10CloseMatch_1__init__, 0, __pyx_n_s_CloseMatch___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__476)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2916, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_tuple__477);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_9) < 0) __PYX_ERR(0, 2916, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2925
 *         self.mayReturnEmpty = False
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         start = loc
 *         instrlen = len(instring)
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10CloseMatch_3parseImpl, 0, __pyx_n_s_CloseMatch_parseImpl, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__479)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2925, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_tuple__480);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_parseImpl, __pyx_t_9) < 0) __PYX_ERR(0, 2925, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2882
 * 
 * 
 * class CloseMatch(Token):             # <<<<<<<<<<<<<<
 *     """A variation on :class:`Literal` which matches "close" matches,
 *     that is, strings with at most 'n' mismatching characters.
 */
  __pyx_t_9 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_CloseMatch, __pyx_t_1, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2882, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CloseMatch, __pyx_t_9) < 0) __PYX_ERR(0, 2882, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2952
 * 
 * 
 * class Word(Token):             # <<<<<<<<<<<<<<
 *     """Token for matching words composed of allowed character sets.
 *     Defined with string containing all allowed initial characters, an
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Token); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2952, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2952, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2952, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_1, __pyx_t_2, __pyx_n_s_Word, __pyx_n_s_Word, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_Token_for_matching_words_compose); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2952, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":3006
 *     """
 * 
 *     def __init__(self, initChars, bodyChars=None, min=1, max=0, exact=0, asKeyword=False, excludeChars=None):             # <<<<<<<<<<<<<<
 *         super(Word, self).__init__()
 *         if excludeChars:
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_4Word_1__init__, 0, __pyx_n_s_Word___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__482)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_tuple__483);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_9) < 0) __PYX_ERR(0, 3006, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3064
 *                 self.__class__ = _WordRegex
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if instring[loc] not in self.initChars:
 *             raise ParseException(instring, loc, self.errmsg, self)
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_4Word_3parseImpl, 0, __pyx_n_s_Word_parseImpl, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__485)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3064, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_tuple__486);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_parseImpl, __pyx_t_9) < 0) __PYX_ERR(0, 3064, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3092
 *         return loc, instring[start:loc]
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         try:
 *             return super(Word, self).__str__()
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_4Word_5__str__, 0, __pyx_n_s_Word___str, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__488)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3092, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_str, __pyx_t_9) < 0) __PYX_ERR(0, 3092, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":2952
 * 
 * 
 * class Word(Token):             # <<<<<<<<<<<<<<
 *     """Token for matching words composed of allowed character sets.
 *     Defined with string containing all allowed initial characters, an
 */
  __pyx_t_9 = __Pyx_Py3ClassCreate(__pyx_t_1, __pyx_n_s_Word, __pyx_t_2, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2952, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Word, __pyx_t_9) < 0) __PYX_ERR(0, 2952, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3114
 * 
 * 
 * class _WordRegex(Word):             # <<<<<<<<<<<<<<
 *     def parseImpl(self, instring, loc, doActions=True):
 *         result = self.re_match(instring, loc)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Word); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_1, __pyx_n_s_WordRegex, __pyx_n_s_WordRegex, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, (PyObject *) NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":3115
 * 
 * class _WordRegex(Word):
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         result = self.re_match(instring, loc)
 *         if not result:
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10_WordRegex_1parseImpl, 0, __pyx_n_s_WordRegex_parseImpl, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__490)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_tuple__491);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_parseImpl, __pyx_t_9) < 0) __PYX_ERR(0, 3115, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3114
 * 
 * 
 * class _WordRegex(Word):             # <<<<<<<<<<<<<<
 *     def parseImpl(self, instring, loc, doActions=True):
 *         result = self.re_match(instring, loc)
 */
  __pyx_t_9 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_WordRegex, __pyx_t_1, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_WordRegex, __pyx_t_9) < 0) __PYX_ERR(0, 3114, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3124
 * 
 * 
 * class Char(_WordRegex):             # <<<<<<<<<<<<<<
 *     """A short-cut class for defining ``Word(characters, exact=1)``,
 *     when defining a match of any single character in a string of
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_WordRegex); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_1, __pyx_t_2, __pyx_n_s_Char, __pyx_n_s_Char, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_A_short_cut_class_for_defining_W); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":3130
 *     """
 * 
 *     def __init__(self, charset, asKeyword=False, excludeChars=None):             # <<<<<<<<<<<<<<
 *         super(Char, self).__init__(charset, exact=1, asKeyword=asKeyword, excludeChars=excludeChars)
 *         self.reString = "[%s]" % _escapeRegexRangeChars(''.join(self.initChars))
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_4Char_1__init__, 0, __pyx_n_s_Char___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__493)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_tuple__494);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_9) < 0) __PYX_ERR(0, 3130, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3124
 * 
 * 
 * class Char(_WordRegex):             # <<<<<<<<<<<<<<
 *     """A short-cut class for defining ``Word(characters, exact=1)``,
 *     when defining a match of any single character in a string of
 */
  __pyx_t_9 = __Pyx_Py3ClassCreate(__pyx_t_1, __pyx_n_s_Char, __pyx_t_2, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Char, __pyx_t_9) < 0) __PYX_ERR(0, 3124, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3139
 * 
 * 
 * class Regex(Token):             # <<<<<<<<<<<<<<
 *     r"""Token for matching strings that match a given regular
 *     expression. Defined with string specifying the regular expression in
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Token); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_1, __pyx_n_s_Regex, __pyx_n_s_Regex, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_Token_for_matching_strings_that); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":3164
 *     """
 * 
 *     def __init__(self, pattern, flags=0, asGroupList=False, asMatch=False):             # <<<<<<<<<<<<<<
 *         """The parameters ``pattern`` and ``flags`` are passed
 *         to the ``re.compile()`` function as-is. See the Python
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_5Regex_1__init__, 0, __pyx_n_s_Regex___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__496)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_tuple__497);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_9) < 0) __PYX_ERR(0, 3164, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3209
 *             self.parseImpl = self.parseImplAsMatch
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         result = self.re_match(instring, loc)
 *         if not result:
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_5Regex_3parseImpl, 0, __pyx_n_s_Regex_parseImpl, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__499)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_tuple__500);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_parseImpl, __pyx_t_9) < 0) __PYX_ERR(0, 3209, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3222
 *         return loc, ret
 * 
 *     def parseImplAsGroupList(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         result = self.re_match(instring, loc)
 *         if not result:
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_5Regex_5parseImplAsGroupList, 0, __pyx_n_s_Regex_parseImplAsGroupList, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__502)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_tuple__503);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_parseImplAsGroupList, __pyx_t_9) < 0) __PYX_ERR(0, 3222, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3231
 *         return loc, ret
 * 
 *     def parseImplAsMatch(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         result = self.re_match(instring, loc)
 *         if not result:
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_5Regex_7parseImplAsMatch, 0, __pyx_n_s_Regex_parseImplAsMatch, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__505)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_tuple__506);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_parseImplAsMatch, __pyx_t_9) < 0) __PYX_ERR(0, 3231, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3240
 *         return loc, ret
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         try:
 *             return super(Regex, self).__str__()
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_5Regex_9__str__, 0, __pyx_n_s_Regex___str, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__508)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_str, __pyx_t_9) < 0) __PYX_ERR(0, 3240, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3251
 *         return self.strRepr
 * 
 *     def sub(self, repl):             # <<<<<<<<<<<<<<
 *         r"""
 *         Return Regex with an attached parse action to transform the parsed
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_5Regex_11sub, 0, __pyx_n_s_Regex_sub, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__510)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_sub, __pyx_t_9) < 0) __PYX_ERR(0, 3251, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3139
 * 
 * 
 * class Regex(Token):             # <<<<<<<<<<<<<<
 *     r"""Token for matching strings that match a given regular
 *     expression. Defined with string specifying the regular expression in
 */
  __pyx_t_9 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_Regex, __pyx_t_1, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Regex, __pyx_t_9) < 0) __PYX_ERR(0, 3139, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3281
 * 
 * 
 * class QuotedString(Token):             # <<<<<<<<<<<<<<
 *     r"""
 *     Token for matching strings that are delimited by quoting characters.
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Token); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_1, __pyx_t_2, __pyx_n_s_QuotedString, __pyx_n_s_QuotedString, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_Token_for_matching_strings_that_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":3321
 *     """
 * 
 *     def __init__(self, quoteChar, escChar=None, escQuote=None, multiline=False,             # <<<<<<<<<<<<<<
 *                  unquoteResults=True, endQuoteChar=None, convertWhitespaceEscapes=True):
 *         super(QuotedString, self).__init__()
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12QuotedString_1__init__, 0, __pyx_n_s_QuotedString___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__512)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_tuple__513);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_9) < 0) __PYX_ERR(0, 3321, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3386
 *         self.mayReturnEmpty = True
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         result = instring[loc] == self.firstQuoteChar and self.re_match(instring, loc) or None
 *         if not result:
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12QuotedString_3parseImpl, 0, __pyx_n_s_QuotedString_parseImpl, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__515)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_tuple__516);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_parseImpl, __pyx_t_9) < 0) __PYX_ERR(0, 3386, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3421
 *         return loc, ret
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         try:
 *             return super(QuotedString, self).__str__()
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_12QuotedString_5__str__, 0, __pyx_n_s_QuotedString___str, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__518)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3421, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_str, __pyx_t_9) < 0) __PYX_ERR(0, 3421, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3281
 * 
 * 
 * class QuotedString(Token):             # <<<<<<<<<<<<<<
 *     r"""
 *     Token for matching strings that are delimited by quoting characters.
 */
  __pyx_t_9 = __Pyx_Py3ClassCreate(__pyx_t_1, __pyx_n_s_QuotedString, __pyx_t_2, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_QuotedString, __pyx_t_9) < 0) __PYX_ERR(0, 3281, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3433
 * 
 * 
 * class CharsNotIn(Token):             # <<<<<<<<<<<<<<
 *     """Token for matching words composed of characters *not* in a given
 *     set (will include whitespace in matched characters if not listed in
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Token); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_1, __pyx_n_s_CharsNotIn, __pyx_n_s_CharsNotIn, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_Token_for_matching_words_compose_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":3454
 *     """
 * 
 *     def __init__(self, notChars, min=1, max=0, exact=0):             # <<<<<<<<<<<<<<
 *         super(CharsNotIn, self).__init__()
 *         self.skipWhitespace = False
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10CharsNotIn_1__init__, 0, __pyx_n_s_CharsNotIn___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__520)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3454, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_tuple__521);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_9) < 0) __PYX_ERR(0, 3454, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3479
 *         self.mayIndexError = False
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if instring[loc] in self.notChars:
 *             raise ParseException(instring, loc, self.errmsg, self)
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10CharsNotIn_3parseImpl, 0, __pyx_n_s_CharsNotIn_parseImpl, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__523)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_tuple__524);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_parseImpl, __pyx_t_9) < 0) __PYX_ERR(0, 3479, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3495
 *         return loc, instring[start:loc]
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         try:
 *             return super(CharsNotIn, self).__str__()
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10CharsNotIn_5__str__, 0, __pyx_n_s_CharsNotIn___str, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__526)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_str, __pyx_t_9) < 0) __PYX_ERR(0, 3495, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3433
 * 
 * 
 * class CharsNotIn(Token):             # <<<<<<<<<<<<<<
 *     """Token for matching words composed of characters *not* in a given
 *     set (will include whitespace in matched characters if not listed in
 */
  __pyx_t_9 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_CharsNotIn, __pyx_t_1, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CharsNotIn, __pyx_t_9) < 0) __PYX_ERR(0, 3433, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3510
 * 
 * 
 * class White(Token):             # <<<<<<<<<<<<<<
 *     """Special matching class for matching whitespace.  Normally,
 *     whitespace is ignored by pyparsing grammars.  This class is included
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Token); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_1, __pyx_t_2, __pyx_n_s_White, __pyx_n_s_White, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_Special_matching_class_for_match); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":3520
 *     """
 *     whiteStrs = {
 *         ' ': '<SP>',             # <<<<<<<<<<<<<<
 *         '\t': '<TAB>',
 *         '\n': '<LF>',
 */
  __pyx_t_9 = __Pyx_PyDict_NewPresized(23); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_9, __pyx_kp_s__29, __pyx_kp_s_SP) < 0) __PYX_ERR(0, 3520, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_kp_s__6, __pyx_kp_s_TAB) < 0) __PYX_ERR(0, 3520, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_kp_s__4, __pyx_kp_s_LF) < 0) __PYX_ERR(0, 3520, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_kp_s__103, __pyx_kp_s_CR) < 0) __PYX_ERR(0, 3520, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_kp_s__102, __pyx_kp_s_FF) < 0) __PYX_ERR(0, 3520, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_kp_u__527, __pyx_kp_s_NBSP) < 0) __PYX_ERR(0, 3520, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_kp_u__528, __pyx_kp_s_OGHAM_SPACE_MARK) < 0) __PYX_ERR(0, 3520, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_kp_u__529, __pyx_kp_s_MONGOLIAN_VOWEL_SEPARATOR) < 0) __PYX_ERR(0, 3520, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_kp_u__530, __pyx_kp_s_EN_QUAD) < 0) __PYX_ERR(0, 3520, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_kp_u__531, __pyx_kp_s_EM_QUAD) < 0) __PYX_ERR(0, 3520, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_kp_u__532, __pyx_kp_s_EN_SPACE) < 0) __PYX_ERR(0, 3520, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_kp_u__533, __pyx_kp_s_EM_SPACE) < 0) __PYX_ERR(0, 3520, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_kp_u__534, __pyx_kp_s_THREE_PER_EM_SPACE) < 0) __PYX_ERR(0, 3520, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_kp_u__535, __pyx_kp_s_FOUR_PER_EM_SPACE) < 0) __PYX_ERR(0, 3520, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_kp_u__536, __pyx_kp_s_SIX_PER_EM_SPACE) < 0) __PYX_ERR(0, 3520, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_kp_u__537, __pyx_kp_s_FIGURE_SPACE) < 0) __PYX_ERR(0, 3520, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_kp_u__538, __pyx_kp_s_PUNCTUATION_SPACE) < 0) __PYX_ERR(0, 3520, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_kp_u__539, __pyx_kp_s_THIN_SPACE) < 0) __PYX_ERR(0, 3520, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_kp_u__540, __pyx_kp_s_HAIR_SPACE) < 0) __PYX_ERR(0, 3520, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_kp_u__541, __pyx_kp_s_ZERO_WIDTH_SPACE) < 0) __PYX_ERR(0, 3520, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_kp_u__542, __pyx_kp_s_NNBSP) < 0) __PYX_ERR(0, 3520, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_kp_u__543, __pyx_kp_s_MMSP) < 0) __PYX_ERR(0, 3520, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_9, __pyx_kp_u__544, __pyx_kp_s_IDEOGRAPHIC_SPACE) < 0) __PYX_ERR(0, 3520, __pyx_L1_error)
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_whiteStrs, __pyx_t_9) < 0) __PYX_ERR(0, 3519, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3545
 *     }
 * 
 *     def __init__(self, ws=" \t\r\n", min=1, max=0, exact=0):             # <<<<<<<<<<<<<<
 *         super(White, self).__init__()
 *         self.matchWhite = ws
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_5White_1__init__, 0, __pyx_n_s_White___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__546)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_tuple__547);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_9) < 0) __PYX_ERR(0, 3545, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3565
 *             self.minLen = exact
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if instring[loc] not in self.matchWhite:
 *             raise ParseException(instring, loc, self.errmsg, self)
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_5White_3parseImpl, 0, __pyx_n_s_White_parseImpl, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__549)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_tuple__550);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_parseImpl, __pyx_t_9) < 0) __PYX_ERR(0, 3565, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3510
 * 
 * 
 * class White(Token):             # <<<<<<<<<<<<<<
 *     """Special matching class for matching whitespace.  Normally,
 *     whitespace is ignored by pyparsing grammars.  This class is included
 */
  __pyx_t_9 = __Pyx_Py3ClassCreate(__pyx_t_1, __pyx_n_s_White, __pyx_t_2, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_White, __pyx_t_9) < 0) __PYX_ERR(0, 3510, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3581
 * 
 * 
 * class _PositionToken(Token):             # <<<<<<<<<<<<<<
 *     def __init__(self):
 *         super(_PositionToken, self).__init__()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Token); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_1, __pyx_n_s_PositionToken, __pyx_n_s_PositionToken, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, (PyObject *) NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":3582
 * 
 * class _PositionToken(Token):
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         super(_PositionToken, self).__init__()
 *         self.name = self.__class__.__name__
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_14_PositionToken_1__init__, 0, __pyx_n_s_PositionToken___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__552)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_9) < 0) __PYX_ERR(0, 3582, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3581
 * 
 * 
 * class _PositionToken(Token):             # <<<<<<<<<<<<<<
 *     def __init__(self):
 *         super(_PositionToken, self).__init__()
 */
  __pyx_t_9 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_PositionToken, __pyx_t_1, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PositionToken, __pyx_t_9) < 0) __PYX_ERR(0, 3581, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3589
 * 
 * 
 * class GoToColumn(_PositionToken):             # <<<<<<<<<<<<<<
 *     """Token to advance to a specific column of input text; useful for
 *     tabular report scraping.
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_PositionToken); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3589, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3589, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3589, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_1, __pyx_t_2, __pyx_n_s_GoToColumn, __pyx_n_s_GoToColumn, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_Token_to_advance_to_a_specific_c); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3589, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":3594
 *     """
 * 
 *     def __init__(self, colno):             # <<<<<<<<<<<<<<
 *         super(GoToColumn, self).__init__()
 *         self.col = colno
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10GoToColumn_1__init__, 0, __pyx_n_s_GoToColumn___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__554)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3594, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_9) < 0) __PYX_ERR(0, 3594, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3598
 *         self.col = colno
 * 
 *     def preParse(self, instring, loc):             # <<<<<<<<<<<<<<
 *         if col(loc, instring) != self.col:
 *             instrlen = len(instring)
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10GoToColumn_3preParse, 0, __pyx_n_s_GoToColumn_preParse, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__556)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_preParse, __pyx_t_9) < 0) __PYX_ERR(0, 3598, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3607
 *         return loc
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         thiscol = col(loc, instring)
 *         if thiscol > self.col:
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10GoToColumn_5parseImpl, 0, __pyx_n_s_GoToColumn_parseImpl, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__558)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3607, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_tuple__559);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_parseImpl, __pyx_t_9) < 0) __PYX_ERR(0, 3607, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3589
 * 
 * 
 * class GoToColumn(_PositionToken):             # <<<<<<<<<<<<<<
 *     """Token to advance to a specific column of input text; useful for
 *     tabular report scraping.
 */
  __pyx_t_9 = __Pyx_Py3ClassCreate(__pyx_t_1, __pyx_n_s_GoToColumn, __pyx_t_2, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3589, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_GoToColumn, __pyx_t_9) < 0) __PYX_ERR(0, 3589, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3616
 * 
 * 
 * class LineStart(_PositionToken):             # <<<<<<<<<<<<<<
 *     r"""Matches if current position is at the beginning of a line within
 *     the parse string
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_PositionToken); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_1, __pyx_n_s_LineStart, __pyx_n_s_LineStart, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_Matches_if_current_position_is_a); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":3639
 *     """
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         super(LineStart, self).__init__()
 *         self.errmsg = "Expected start of line"
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_9LineStart_1__init__, 0, __pyx_n_s_LineStart___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__561)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_9) < 0) __PYX_ERR(0, 3639, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3643
 *         self.errmsg = "Expected start of line"
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if col(loc, instring) == 1:
 *             return loc, []
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_9LineStart_3parseImpl, 0, __pyx_n_s_LineStart_parseImpl, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__563)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_tuple__564);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_parseImpl, __pyx_t_9) < 0) __PYX_ERR(0, 3643, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3616
 * 
 * 
 * class LineStart(_PositionToken):             # <<<<<<<<<<<<<<
 *     r"""Matches if current position is at the beginning of a line within
 *     the parse string
 */
  __pyx_t_9 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_LineStart, __pyx_t_1, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_LineStart, __pyx_t_9) < 0) __PYX_ERR(0, 3616, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3649
 * 
 * 
 * class LineEnd(_PositionToken):             # <<<<<<<<<<<<<<
 *     """Matches if current position is at the end of a line within the
 *     parse string
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_PositionToken); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_1, __pyx_t_2, __pyx_n_s_LineEnd, __pyx_n_s_LineEnd, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_Matches_if_current_position_is_a_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":3654
 *     """
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         super(LineEnd, self).__init__()
 *         self.setWhitespaceChars(ParserElement.DEFAULT_WHITE_CHARS.replace("\n", ""))
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7LineEnd_1__init__, 0, __pyx_n_s_LineEnd___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__566)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_9) < 0) __PYX_ERR(0, 3654, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3659
 *         self.errmsg = "Expected end of line"
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if loc < len(instring):
 *             if instring[loc] == "\n":
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7LineEnd_3parseImpl, 0, __pyx_n_s_LineEnd_parseImpl, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__568)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3659, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_tuple__569);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_parseImpl, __pyx_t_9) < 0) __PYX_ERR(0, 3659, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3649
 * 
 * 
 * class LineEnd(_PositionToken):             # <<<<<<<<<<<<<<
 *     """Matches if current position is at the end of a line within the
 *     parse string
 */
  __pyx_t_9 = __Pyx_Py3ClassCreate(__pyx_t_1, __pyx_n_s_LineEnd, __pyx_t_2, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_LineEnd, __pyx_t_9) < 0) __PYX_ERR(0, 3649, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3671
 * 
 * 
 * class StringStart(_PositionToken):             # <<<<<<<<<<<<<<
 *     """Matches if current position is at the beginning of the parse
 *     string
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_PositionToken); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3671, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3671, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3671, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_1, __pyx_n_s_StringStart, __pyx_n_s_StringStart, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_Matches_if_current_position_is_a_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3671, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":3676
 *     """
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         super(StringStart, self).__init__()
 *         self.errmsg = "Expected start of text"
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_11StringStart_1__init__, 0, __pyx_n_s_StringStart___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__571)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3676, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_9) < 0) __PYX_ERR(0, 3676, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3680
 *         self.errmsg = "Expected start of text"
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if loc != 0:
 *             # see if entire string up to here is just whitespace and ignoreables
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_11StringStart_3parseImpl, 0, __pyx_n_s_StringStart_parseImpl, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__573)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3680, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_tuple__574);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_parseImpl, __pyx_t_9) < 0) __PYX_ERR(0, 3680, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3671
 * 
 * 
 * class StringStart(_PositionToken):             # <<<<<<<<<<<<<<
 *     """Matches if current position is at the beginning of the parse
 *     string
 */
  __pyx_t_9 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_StringStart, __pyx_t_1, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3671, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StringStart, __pyx_t_9) < 0) __PYX_ERR(0, 3671, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3688
 * 
 * 
 * class StringEnd(_PositionToken):             # <<<<<<<<<<<<<<
 *     """Matches if current position is at the end of the parse string
 *     """
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_PositionToken); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3688, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3688, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3688, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_1, __pyx_t_2, __pyx_n_s_StringEnd, __pyx_n_s_StringEnd, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_Matches_if_current_position_is_a_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3688, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":3692
 *     """
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         super(StringEnd, self).__init__()
 *         self.errmsg = "Expected end of text"
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_9StringEnd_1__init__, 0, __pyx_n_s_StringEnd___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__576)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_9) < 0) __PYX_ERR(0, 3692, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3696
 *         self.errmsg = "Expected end of text"
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if loc < len(instring):
 *             raise ParseException(instring, loc, self.errmsg, self)
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_9StringEnd_3parseImpl, 0, __pyx_n_s_StringEnd_parseImpl, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__578)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_tuple__579);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_parseImpl, __pyx_t_9) < 0) __PYX_ERR(0, 3696, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3688
 * 
 * 
 * class StringEnd(_PositionToken):             # <<<<<<<<<<<<<<
 *     """Matches if current position is at the end of the parse string
 *     """
 */
  __pyx_t_9 = __Pyx_Py3ClassCreate(__pyx_t_1, __pyx_n_s_StringEnd, __pyx_t_2, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3688, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StringEnd, __pyx_t_9) < 0) __PYX_ERR(0, 3688, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3707
 * 
 * 
 * class WordStart(_PositionToken):             # <<<<<<<<<<<<<<
 *     """Matches if the current position is at the beginning of a Word,
 *     and is not preceded by any character in a given set of
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_PositionToken); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_1, __pyx_n_s_WordStart, __pyx_n_s_WordStart, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_Matches_if_the_current_position); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":3717
 *     """
 * 
 *     def __init__(self, wordChars=printables):             # <<<<<<<<<<<<<<
 *         super(WordStart, self).__init__()
 *         self.wordChars = set(wordChars)
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_9WordStart_1__init__, 0, __pyx_n_s_WordStart___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__581)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3717, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_9, sizeof(__pyx_defaults5), 1)) __PYX_ERR(0, 3717, __pyx_L1_error)
  __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_printables); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3717, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_Defaults(__pyx_defaults5, __pyx_t_9)->__pyx_arg_wordChars = __pyx_t_13;
  __Pyx_GIVEREF(__pyx_t_13);
  __pyx_t_13 = 0;
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_9, __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_38__defaults__);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_9) < 0) __PYX_ERR(0, 3717, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3722
 *         self.errmsg = "Not at the start of a word"
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if loc != 0:
 *             if (instring[loc - 1] in self.wordChars
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_9WordStart_3parseImpl, 0, __pyx_n_s_WordStart_parseImpl, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__583)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3722, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_tuple__584);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_parseImpl, __pyx_t_9) < 0) __PYX_ERR(0, 3722, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3707
 * 
 * 
 * class WordStart(_PositionToken):             # <<<<<<<<<<<<<<
 *     """Matches if the current position is at the beginning of a Word,
 *     and is not preceded by any character in a given set of
 */
  __pyx_t_9 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_WordStart, __pyx_t_1, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_WordStart, __pyx_t_9) < 0) __PYX_ERR(0, 3707, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3730
 * 
 * 
 * class WordEnd(_PositionToken):             # <<<<<<<<<<<<<<
 *     """Matches if the current position is at the end of a Word, and is
 *     not followed by any character in a given set of ``wordChars``
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_PositionToken); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_1, __pyx_t_2, __pyx_n_s_WordEnd, __pyx_n_s_WordEnd, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_Matches_if_the_current_position_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":3739
 *     """
 * 
 *     def __init__(self, wordChars=printables):             # <<<<<<<<<<<<<<
 *         super(WordEnd, self).__init__()
 *         self.wordChars = set(wordChars)
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7WordEnd_1__init__, 0, __pyx_n_s_WordEnd___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__586)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3739, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_9, sizeof(__pyx_defaults6), 1)) __PYX_ERR(0, 3739, __pyx_L1_error)
  __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_printables); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3739, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_Defaults(__pyx_defaults6, __pyx_t_9)->__pyx_arg_wordChars = __pyx_t_13;
  __Pyx_GIVEREF(__pyx_t_13);
  __pyx_t_13 = 0;
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_9, __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_40__defaults__);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_9) < 0) __PYX_ERR(0, 3739, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3745
 *         self.errmsg = "Not at the end of a word"
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         instrlen = len(instring)
 *         if instrlen > 0 and loc < instrlen:
 */
  __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7WordEnd_3parseImpl, 0, __pyx_n_s_WordEnd_parseImpl, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__588)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3745, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_tuple__589);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_parseImpl, __pyx_t_9) < 0) __PYX_ERR(0, 3745, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3730
 * 
 * 
 * class WordEnd(_PositionToken):             # <<<<<<<<<<<<<<
 *     """Matches if the current position is at the end of a Word, and is
 *     not followed by any character in a given set of ``wordChars``
 */
  __pyx_t_9 = __Pyx_Py3ClassCreate(__pyx_t_1, __pyx_n_s_WordEnd, __pyx_t_2, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_WordEnd, __pyx_t_9) < 0) __PYX_ERR(0, 3730, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3754
 * 
 * 
 * class And(ParseExpression):             # <<<<<<<<<<<<<<
 *     """
 *     Requires all given :class:`ParseExpression` s to be found in the given order.
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParseExpression); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3754, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3754, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3754, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_1, __pyx_n_s_And, __pyx_n_s_And, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_Requires_all_given_class_ParseE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3754, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":3772
 *     """
 * 
 *     class _ErrorStop(Empty):             # <<<<<<<<<<<<<<
 *         def __init__(self, *args, **kwargs):
 *             super(And._ErrorStop, self).__init__(*args, **kwargs)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_Empty); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3772, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_13 = PyTuple_New(1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3772, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_9);
  __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CalculateMetaclass(NULL, __pyx_t_13); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3772, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_8 = __Pyx_Py3MetaclassPrepare(__pyx_t_9, __pyx_t_13, __pyx_n_s_ErrorStop, __pyx_n_s_And__ErrorStop, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, (PyObject *) NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3772, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);

  /* "Aspidites/_vendor/pyparsing_extension.py":3773
 * 
 *     class _ErrorStop(Empty):
 *         def __init__(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *             super(And._ErrorStop, self).__init__(*args, **kwargs)
 *             self.name = '-'
 */
  __pyx_t_18 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10_ErrorStop_1__init__, 0, __pyx_n_s_And__ErrorStop___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__591)); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 3773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  if (__Pyx_SetNameInClass(__pyx_t_8, __pyx_n_s_init, __pyx_t_18) < 0) __PYX_ERR(0, 3773, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3772
 *     """
 * 
 *     class _ErrorStop(Empty):             # <<<<<<<<<<<<<<
 *         def __init__(self, *args, **kwargs):
 *             super(And._ErrorStop, self).__init__(*args, **kwargs)
 */
  __pyx_t_18 = __Pyx_Py3ClassCreate(__pyx_t_9, __pyx_n_s_ErrorStop, __pyx_t_13, __pyx_t_8, NULL, 0, 1); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 3772, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_ErrorStop, __pyx_t_18) < 0) __PYX_ERR(0, 3772, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3778
 *             self.leaveWhitespace()
 * 
 *     def __init__(self, exprs, savelist=True):             # <<<<<<<<<<<<<<
 *         exprs = list(exprs)
 *         if exprs and Ellipsis in exprs:
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_3And_1__init__, 0, __pyx_n_s_And___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__593)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3778, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_13, __pyx_tuple__594);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_13) < 0) __PYX_ERR(0, 3778, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3798
 *         self.callPreparse = True
 * 
 *     def streamline(self):             # <<<<<<<<<<<<<<
 *         # collapse any _PendingSkip's
 *         from .pyparsing import _PendingSkip
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_3And_3streamline, 0, __pyx_n_s_And_streamline, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__596)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3798, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_streamline, __pyx_t_13) < 0) __PYX_ERR(0, 3798, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3817
 *         return self
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         # pass False as last arg to _parse for first element, since we already
 *         # pre-parsed the string as part of our And pre-parsing
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_3And_5parseImpl, 0, __pyx_n_s_And_parseImpl, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__598)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3817, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_13, __pyx_tuple__599);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_parseImpl, __pyx_t_13) < 0) __PYX_ERR(0, 3817, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3842
 *         return loc, resultlist
 * 
 *     def __iadd__(self, other):             # <<<<<<<<<<<<<<
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_3And_7__iadd__, 0, __pyx_n_s_And___iadd, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__601)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3842, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_iadd, __pyx_t_13) < 0) __PYX_ERR(0, 3842, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3847
 *         return self.append(other)  # And([self, other])
 * 
 *     def checkRecursion(self, parseElementList):             # <<<<<<<<<<<<<<
 *         subRecCheckList = parseElementList[:] + [self]
 *         for e in self.exprs:
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_3And_9checkRecursion, 0, __pyx_n_s_And_checkRecursion, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__603)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3847, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_checkRecursion, __pyx_t_13) < 0) __PYX_ERR(0, 3847, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3854
 *                 break
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_3And_11__str__, 0, __pyx_n_s_And___str, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__605)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3854, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_str, __pyx_t_13) < 0) __PYX_ERR(0, 3854, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3754
 * 
 * 
 * class And(ParseExpression):             # <<<<<<<<<<<<<<
 *     """
 *     Requires all given :class:`ParseExpression` s to be found in the given order.
 */
  __pyx_t_13 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_And, __pyx_t_1, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3754, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_And, __pyx_t_13) < 0) __PYX_ERR(0, 3754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3864
 * 
 * 
 * class Or(ParseExpression):             # <<<<<<<<<<<<<<
 *     """Requires that at least one :class:`ParseExpression` is found. If
 *     two expressions match, the expression that matches the longest
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParseExpression); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_1, __pyx_t_2, __pyx_n_s_Or, __pyx_n_s_Or, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_Requires_that_at_least_one_class); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":3882
 *     """
 * 
 *     def __init__(self, exprs, savelist=False):             # <<<<<<<<<<<<<<
 *         super(Or, self).__init__(exprs, savelist)
 *         if self.exprs:
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_2Or_1__init__, 0, __pyx_n_s_Or___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__607)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3882, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_13, __pyx_tuple__608);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_13) < 0) __PYX_ERR(0, 3882, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3889
 *             self.mayReturnEmpty = True
 * 
 *     def streamline(self):             # <<<<<<<<<<<<<<
 *         super(Or, self).streamline()
 *         if __diag__.collect_all_And_tokens:
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_2Or_3streamline, 0, __pyx_n_s_Or_streamline, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__610)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3889, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_streamline, __pyx_t_13) < 0) __PYX_ERR(0, 3889, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3895
 *         return self
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         maxExcLoc = -1
 *         maxException = None
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_2Or_5parseImpl, 0, __pyx_n_s_Or_parseImpl, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__612)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3895, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_13, __pyx_tuple__613);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_parseImpl, __pyx_t_13) < 0) __PYX_ERR(0, 3895, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3955
 *             raise ParseException(instring, loc, "no defined alternatives to match", self)
 * 
 *     def __ixor__(self, other):             # <<<<<<<<<<<<<<
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_2Or_7__ixor__, 0, __pyx_n_s_Or___ixor, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__615)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3955, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_ixor, __pyx_t_13) < 0) __PYX_ERR(0, 3955, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3960
 *         return self.append(other)  # Or([self, other])
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_2Or_9__str__, 0, __pyx_n_s_Or___str, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__617)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3960, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_str, __pyx_t_13) < 0) __PYX_ERR(0, 3960, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3969
 *         return self.strRepr
 * 
 *     def checkRecursion(self, parseElementList):             # <<<<<<<<<<<<<<
 *         subRecCheckList = parseElementList[:] + [self]
 *         for e in self.exprs:
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_2Or_11checkRecursion, 0, __pyx_n_s_Or_checkRecursion, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__619)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3969, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_checkRecursion, __pyx_t_13) < 0) __PYX_ERR(0, 3969, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3974
 *             e.checkRecursion(subRecCheckList)
 * 
 *     def _setResultsName(self, name, listAllMatches=False):             # <<<<<<<<<<<<<<
 *         if (not __diag__.collect_all_And_tokens
 *                 and __diag__.warn_multiple_tokens_in_named_alternation):
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_2Or_13_setResultsName, 0, __pyx_n_s_Or__setResultsName, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__621)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3974, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_13, __pyx_tuple__622);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_setResultsName, __pyx_t_13) < 0) __PYX_ERR(0, 3974, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3864
 * 
 * 
 * class Or(ParseExpression):             # <<<<<<<<<<<<<<
 *     """Requires that at least one :class:`ParseExpression` is found. If
 *     two expressions match, the expression that matches the longest
 */
  __pyx_t_13 = __Pyx_Py3ClassCreate(__pyx_t_1, __pyx_n_s_Or, __pyx_t_2, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Or, __pyx_t_13) < 0) __PYX_ERR(0, 3864, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3987
 * 
 * 
 * class MatchFirst(ParseExpression):             # <<<<<<<<<<<<<<
 *     """Requires that at least one :class:`ParseExpression` is found. If
 *     two expressions match, the first one listed is the one that will
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParseExpression); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3987, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3987, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3987, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_1, __pyx_n_s_MatchFirst, __pyx_n_s_MatchFirst, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_Requires_that_at_least_one_class_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3987, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":4005
 *     """
 * 
 *     def __init__(self, exprs, savelist=False):             # <<<<<<<<<<<<<<
 *         super(MatchFirst, self).__init__(exprs, savelist)
 *         if self.exprs:
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_1__init__, 0, __pyx_n_s_MatchFirst___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__624)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4005, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_13, __pyx_tuple__625);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_13) < 0) __PYX_ERR(0, 4005, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4012
 *             self.mayReturnEmpty = True
 * 
 *     def streamline(self):             # <<<<<<<<<<<<<<
 *         super(MatchFirst, self).streamline()
 *         if __diag__.collect_all_And_tokens:
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_3streamline, 0, __pyx_n_s_MatchFirst_streamline, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__627)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4012, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_streamline, __pyx_t_13) < 0) __PYX_ERR(0, 4012, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4018
 *         return self
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         maxExcLoc = -1
 *         maxException = None
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_5parseImpl, 0, __pyx_n_s_MatchFirst_parseImpl, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__629)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_13, __pyx_tuple__630);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_parseImpl, __pyx_t_13) < 0) __PYX_ERR(0, 4018, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4042
 *                 raise ParseException(instring, loc, "no defined alternatives to match", self)
 * 
 *     def __ior__(self, other):             # <<<<<<<<<<<<<<
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_7__ior__, 0, __pyx_n_s_MatchFirst___ior, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__632)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4042, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_ior, __pyx_t_13) < 0) __PYX_ERR(0, 4042, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4047
 *         return self.append(other)  # MatchFirst([self, other])
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_9__str__, 0, __pyx_n_s_MatchFirst___str, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__634)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4047, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_str, __pyx_t_13) < 0) __PYX_ERR(0, 4047, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4056
 *         return self.strRepr
 * 
 *     def checkRecursion(self, parseElementList):             # <<<<<<<<<<<<<<
 *         subRecCheckList = parseElementList[:] + [self]
 *         for e in self.exprs:
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_11checkRecursion, 0, __pyx_n_s_MatchFirst_checkRecursion, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__636)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4056, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_checkRecursion, __pyx_t_13) < 0) __PYX_ERR(0, 4056, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4061
 *             e.checkRecursion(subRecCheckList)
 * 
 *     def _setResultsName(self, name, listAllMatches=False):             # <<<<<<<<<<<<<<
 *         if (not __diag__.collect_all_And_tokens
 *                 and __diag__.warn_multiple_tokens_in_named_alternation):
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10MatchFirst_13_setResultsName, 0, __pyx_n_s_MatchFirst__setResultsName, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__638)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4061, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_13, __pyx_tuple__639);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_setResultsName, __pyx_t_13) < 0) __PYX_ERR(0, 4061, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":3987
 * 
 * 
 * class MatchFirst(ParseExpression):             # <<<<<<<<<<<<<<
 *     """Requires that at least one :class:`ParseExpression` is found. If
 *     two expressions match, the first one listed is the one that will
 */
  __pyx_t_13 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_MatchFirst, __pyx_t_1, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 3987, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_MatchFirst, __pyx_t_13) < 0) __PYX_ERR(0, 3987, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4074
 * 
 * 
 * class Each(ParseExpression):             # <<<<<<<<<<<<<<
 *     """Requires all given :class:`ParseExpression` s to be found, but in
 *     any order. Expressions may be separated by whitespace.
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParseExpression); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4074, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4074, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4074, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_1, __pyx_t_2, __pyx_n_s_Each, __pyx_n_s_Each, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_Requires_all_given_class_ParseEx); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4074, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":4132
 *     """
 * 
 *     def __init__(self, exprs, savelist=True):             # <<<<<<<<<<<<<<
 *         super(Each, self).__init__(exprs, savelist)
 *         self.mayReturnEmpty = all(e.mayReturnEmpty for e in self.exprs)
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_4Each_1__init__, 0, __pyx_n_s_Each___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__641)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_13, __pyx_tuple__642);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_13) < 0) __PYX_ERR(0, 4132, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4139
 *         self.saveAsList = True
 * 
 *     def streamline(self):             # <<<<<<<<<<<<<<
 *         super(Each, self).streamline()
 *         self.mayReturnEmpty = all(e.mayReturnEmpty for e in self.exprs)
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_4Each_3streamline, 0, __pyx_n_s_Each_streamline, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__644)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_streamline, __pyx_t_13) < 0) __PYX_ERR(0, 4139, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4144
 *         return self
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if self.initExprGroups:
 *             self.opt1map = dict((id(e.expr), e) for e in self.exprs if isinstance(e, Optional))
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_4Each_5parseImpl, 0, __pyx_n_s_Each_parseImpl, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__646)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_13, __pyx_tuple__647);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_parseImpl, __pyx_t_13) < 0) __PYX_ERR(0, 4144, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4193
 *         return loc, finalResults
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_4Each_7__str__, 0, __pyx_n_s_Each___str, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__649)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_str, __pyx_t_13) < 0) __PYX_ERR(0, 4193, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4202
 *         return self.strRepr
 * 
 *     def checkRecursion(self, parseElementList):             # <<<<<<<<<<<<<<
 *         subRecCheckList = parseElementList[:] + [self]
 *         for e in self.exprs:
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_4Each_9checkRecursion, 0, __pyx_n_s_Each_checkRecursion, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__651)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_checkRecursion, __pyx_t_13) < 0) __PYX_ERR(0, 4202, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4074
 * 
 * 
 * class Each(ParseExpression):             # <<<<<<<<<<<<<<
 *     """Requires all given :class:`ParseExpression` s to be found, but in
 *     any order. Expressions may be separated by whitespace.
 */
  __pyx_t_13 = __Pyx_Py3ClassCreate(__pyx_t_1, __pyx_n_s_Each, __pyx_t_2, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4074, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Each, __pyx_t_13) < 0) __PYX_ERR(0, 4074, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4208
 * 
 * 
 * class ParseElementEnhance(ParserElement):             # <<<<<<<<<<<<<<
 *     """Abstract subclass of :class:`ParserElement`, for combining and
 *     post-processing parsed tokens.
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParserElement); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_1, __pyx_n_s_ParseElementEnhance, __pyx_n_s_ParseElementEnhance, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_Abstract_subclass_of_class_Parse); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":4213
 *     """
 * 
 *     def __init__(self, expr, savelist=False):             # <<<<<<<<<<<<<<
 *         super(ParseElementEnhance, self).__init__(savelist)
 *         if isinstance(expr, str):
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_1__init__, 0, __pyx_n_s_ParseElementEnhance___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__653)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_13, __pyx_tuple__654);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_13) < 0) __PYX_ERR(0, 4213, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4231
 *             self.ignoreExprs.extend(expr.ignoreExprs)
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if self.expr is not None:
 *             return self.expr._parse(instring, loc, doActions, callPreParse=False)
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_3parseImpl, 0, __pyx_n_s_ParseElementEnhance_parseImpl, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__656)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_13, __pyx_tuple__657);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_parseImpl, __pyx_t_13) < 0) __PYX_ERR(0, 4231, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4237
 *             raise ParseException("", loc, self.errmsg, self)
 * 
 *     def leaveWhitespace(self):             # <<<<<<<<<<<<<<
 *         self.skipWhitespace = False
 *         self.expr = self.expr.copy()
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_5leaveWhitespace, 0, __pyx_n_s_ParseElementEnhance_leaveWhitesp, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__659)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_leaveWhitespace, __pyx_t_13) < 0) __PYX_ERR(0, 4237, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4244
 *         return self
 * 
 *     def ignore(self, other):             # <<<<<<<<<<<<<<
 *         if isinstance(other, Suppress):
 *             if other not in self.ignoreExprs:
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_7ignore, 0, __pyx_n_s_ParseElementEnhance_ignore, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__661)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_ignore, __pyx_t_13) < 0) __PYX_ERR(0, 4244, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4256
 *         return self
 * 
 *     def streamline(self):             # <<<<<<<<<<<<<<
 *         super(ParseElementEnhance, self).streamline()
 *         if self.expr is not None:
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_9streamline, 0, __pyx_n_s_ParseElementEnhance_streamline, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__663)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_streamline, __pyx_t_13) < 0) __PYX_ERR(0, 4256, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4262
 *         return self
 * 
 *     def checkRecursion(self, parseElementList):             # <<<<<<<<<<<<<<
 *         if self in parseElementList:
 *             raise RecursiveGrammarException(parseElementList + [self])
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_11checkRecursion, 0, __pyx_n_s_ParseElementEnhance_checkRecursi, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__665)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_checkRecursion, __pyx_t_13) < 0) __PYX_ERR(0, 4262, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4269
 *             self.expr.checkRecursion(subRecCheckList)
 * 
 *     def validate(self, validateTrace=None):             # <<<<<<<<<<<<<<
 *         if validateTrace is None:
 *             validateTrace = []
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_13validate, 0, __pyx_n_s_ParseElementEnhance_validate, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__667)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_13, __pyx_tuple__668);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_validate, __pyx_t_13) < 0) __PYX_ERR(0, 4269, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4277
 *         self.checkRecursion([])
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         try:
 *             return super(ParseElementEnhance, self).__str__()
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_19ParseElementEnhance_15__str__, 0, __pyx_n_s_ParseElementEnhance___str, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__670)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_str, __pyx_t_13) < 0) __PYX_ERR(0, 4277, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4208
 * 
 * 
 * class ParseElementEnhance(ParserElement):             # <<<<<<<<<<<<<<
 *     """Abstract subclass of :class:`ParserElement`, for combining and
 *     post-processing parsed tokens.
 */
  __pyx_t_13 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_ParseElementEnhance, __pyx_t_1, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ParseElementEnhance, __pyx_t_13) < 0) __PYX_ERR(0, 4208, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4288
 * 
 * 
 * class FollowedBy(ParseElementEnhance):             # <<<<<<<<<<<<<<
 *     """Lookahead matching of the given parse expression.
 *     ``FollowedBy`` does *not* advance the parsing position within
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParseElementEnhance); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_1, __pyx_t_2, __pyx_n_s_FollowedBy, __pyx_n_s_FollowedBy, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_Lookahead_matching_of_the_given); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":4311
 *     """
 * 
 *     def __init__(self, expr):             # <<<<<<<<<<<<<<
 *         super(FollowedBy, self).__init__(expr)
 *         self.mayReturnEmpty = True
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10FollowedBy_1__init__, 0, __pyx_n_s_FollowedBy___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__672)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_13) < 0) __PYX_ERR(0, 4311, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4315
 *         self.mayReturnEmpty = True
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         # by using self._expr.parse and deleting the contents of the returned ParseResults list
 *         # we keep any named results that were defined in the FollowedBy expression
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10FollowedBy_3parseImpl, 0, __pyx_n_s_FollowedBy_parseImpl, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__675)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_13, __pyx_tuple__676);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_parseImpl, __pyx_t_13) < 0) __PYX_ERR(0, 4315, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4288
 * 
 * 
 * class FollowedBy(ParseElementEnhance):             # <<<<<<<<<<<<<<
 *     """Lookahead matching of the given parse expression.
 *     ``FollowedBy`` does *not* advance the parsing position within
 */
  __pyx_t_13 = __Pyx_Py3ClassCreate(__pyx_t_1, __pyx_n_s_FollowedBy, __pyx_t_2, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_FollowedBy, __pyx_t_13) < 0) __PYX_ERR(0, 4288, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4324
 * 
 * 
 * class PrecededBy(ParseElementEnhance):             # <<<<<<<<<<<<<<
 *     """Lookbehind matching of the given parse expression.
 *     ``PrecededBy`` does not advance the parsing position within the
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParseElementEnhance); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_1, __pyx_n_s_PrecededBy, __pyx_n_s_PrecededBy, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_Lookbehind_matching_of_the_given); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":4353
 *     """
 * 
 *     def __init__(self, expr, retreat=None):             # <<<<<<<<<<<<<<
 *         super(PrecededBy, self).__init__(expr)
 *         self.expr = self.expr().leaveWhitespace()
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10PrecededBy_1__init__, 0, __pyx_n_s_PrecededBy___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__678)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4353, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_13, __pyx_tuple__679);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_13) < 0) __PYX_ERR(0, 4353, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4376
 *         self.parseAction.append(lambda s, l, t: t.__delitem__(slice(None, None)))
 * 
 *     def parseImpl(self, instring, loc=0, doActions=True):             # <<<<<<<<<<<<<<
 *         if self.exact:
 *             if loc < self.retreat:
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10PrecededBy_3parseImpl, 0, __pyx_n_s_PrecededBy_parseImpl, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__681)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_13, __pyx_tuple__682);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_parseImpl, __pyx_t_13) < 0) __PYX_ERR(0, 4376, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4324
 * 
 * 
 * class PrecededBy(ParseElementEnhance):             # <<<<<<<<<<<<<<
 *     """Lookbehind matching of the given parse expression.
 *     ``PrecededBy`` does not advance the parsing position within the
 */
  __pyx_t_13 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_PrecededBy, __pyx_t_1, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PrecededBy, __pyx_t_13) < 0) __PYX_ERR(0, 4324, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4400
 * 
 * 
 * class NotAny(ParseElementEnhance):             # <<<<<<<<<<<<<<
 *     """Lookahead to disallow matching with the given parse expression.
 *     ``NotAny`` does *not* advance the parsing position within the
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParseElementEnhance); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4400, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4400, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4400, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_1, __pyx_t_2, __pyx_n_s_NotAny, __pyx_n_s_NotAny, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_Lookahead_to_disallow_matching_w); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4400, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":4424
 *     """
 * 
 *     def __init__(self, expr):             # <<<<<<<<<<<<<<
 *         super(NotAny, self).__init__(expr)
 *         # ~ self.leaveWhitespace()
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_6NotAny_1__init__, 0, __pyx_n_s_NotAny___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__684)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4424, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_13) < 0) __PYX_ERR(0, 4424, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4431
 *         self.errmsg = "Found unwanted token, " + str(self.expr)
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         if self.expr.canParseNext(instring, loc):
 *             raise ParseException(instring, loc, self.errmsg, self)
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_6NotAny_3parseImpl, 0, __pyx_n_s_NotAny_parseImpl, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__686)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4431, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_13, __pyx_tuple__687);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_parseImpl, __pyx_t_13) < 0) __PYX_ERR(0, 4431, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4436
 *         return loc, []
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_6NotAny_5__str__, 0, __pyx_n_s_NotAny___str, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__689)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_str, __pyx_t_13) < 0) __PYX_ERR(0, 4436, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4400
 * 
 * 
 * class NotAny(ParseElementEnhance):             # <<<<<<<<<<<<<<
 *     """Lookahead to disallow matching with the given parse expression.
 *     ``NotAny`` does *not* advance the parsing position within the
 */
  __pyx_t_13 = __Pyx_Py3ClassCreate(__pyx_t_1, __pyx_n_s_NotAny, __pyx_t_2, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4400, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_NotAny, __pyx_t_13) < 0) __PYX_ERR(0, 4400, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4446
 * 
 * 
 * class _MultipleMatch(ParseElementEnhance):             # <<<<<<<<<<<<<<
 *     def __init__(self, expr, stopOn=None):
 *         super(_MultipleMatch, self).__init__(expr)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParseElementEnhance); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4446, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4446, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4446, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_1, __pyx_n_s_MultipleMatch, __pyx_n_s_MultipleMatch, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, (PyObject *) NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4446, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":4447
 * 
 * class _MultipleMatch(ParseElementEnhance):
 *     def __init__(self, expr, stopOn=None):             # <<<<<<<<<<<<<<
 *         super(_MultipleMatch, self).__init__(expr)
 *         self.saveAsList = True
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_14_MultipleMatch_1__init__, 0, __pyx_n_s_MultipleMatch___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__691)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_13, __pyx_tuple__692);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_13) < 0) __PYX_ERR(0, 4447, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4455
 *         self.stopOn(ender)
 * 
 *     def stopOn(self, ender):             # <<<<<<<<<<<<<<
 *         if isinstance(ender, str):
 *             ender = self._literalStringClass(ender)
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_14_MultipleMatch_3stopOn, 0, __pyx_n_s_MultipleMatch_stopOn, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__694)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_stopOn, __pyx_t_13) < 0) __PYX_ERR(0, 4455, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4461
 *         return self
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         self_expr_parse = self.expr._parse
 *         self_skip_ignorables = self._skipIgnorables
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_14_MultipleMatch_5parseImpl, 0, __pyx_n_s_MultipleMatch_parseImpl, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__696)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4461, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_13, __pyx_tuple__697);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_parseImpl, __pyx_t_13) < 0) __PYX_ERR(0, 4461, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4490
 *         return loc, tokens
 * 
 *     def _setResultsName(self, name, listAllMatches=False):             # <<<<<<<<<<<<<<
 *         if __diag__.warn_ungrouped_named_tokens_in_collection:
 *             for e in [self.expr] + getattr(self.expr, 'exprs', []):
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_14_MultipleMatch_7_setResultsName, 0, __pyx_n_s_MultipleMatch__setResultsName, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__699)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_13, __pyx_tuple__700);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_setResultsName, __pyx_t_13) < 0) __PYX_ERR(0, 4490, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4446
 * 
 * 
 * class _MultipleMatch(ParseElementEnhance):             # <<<<<<<<<<<<<<
 *     def __init__(self, expr, stopOn=None):
 *         super(_MultipleMatch, self).__init__(expr)
 */
  __pyx_t_13 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_MultipleMatch, __pyx_t_1, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4446, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_MultipleMatch, __pyx_t_13) < 0) __PYX_ERR(0, 4446, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4505
 * 
 * 
 * class OneOrMore(_MultipleMatch):             # <<<<<<<<<<<<<<
 *     """Repetition of one or more of the given expression.
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_MultipleMatch); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4505, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4505, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4505, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_1, __pyx_t_2, __pyx_n_s_OneOrMore, __pyx_n_s_OneOrMore, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_Repetition_of_one_or_more_of_the); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4505, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":4531
 *     """
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_9OneOrMore_1__str__, 0, __pyx_n_s_OneOrMore___str, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__702)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_str, __pyx_t_13) < 0) __PYX_ERR(0, 4531, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4505
 * 
 * 
 * class OneOrMore(_MultipleMatch):             # <<<<<<<<<<<<<<
 *     """Repetition of one or more of the given expression.
 * 
 */
  __pyx_t_13 = __Pyx_Py3ClassCreate(__pyx_t_1, __pyx_n_s_OneOrMore, __pyx_t_2, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4505, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_OneOrMore, __pyx_t_13) < 0) __PYX_ERR(0, 4505, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4541
 * 
 * 
 * class ZeroOrMore(_MultipleMatch):             # <<<<<<<<<<<<<<
 *     """Optional repetition of zero or more of the given expression.
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_MultipleMatch); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_1, __pyx_n_s_ZeroOrMore, __pyx_n_s_ZeroOrMore, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_Optional_repetition_of_zero_or_m); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":4553
 *     """
 * 
 *     def __init__(self, expr, stopOn=None):             # <<<<<<<<<<<<<<
 *         super(ZeroOrMore, self).__init__(expr, stopOn=stopOn)
 *         self.mayReturnEmpty = True
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10ZeroOrMore_1__init__, 0, __pyx_n_s_ZeroOrMore___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__704)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4553, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_13, __pyx_tuple__705);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_13) < 0) __PYX_ERR(0, 4553, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4557
 *         self.mayReturnEmpty = True
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         try:
 *             return super(ZeroOrMore, self).parseImpl(instring, loc, doActions)
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10ZeroOrMore_3parseImpl, 0, __pyx_n_s_ZeroOrMore_parseImpl, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__707)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_13, __pyx_tuple__708);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_parseImpl, __pyx_t_13) < 0) __PYX_ERR(0, 4557, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4563
 *             return loc, []
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10ZeroOrMore_5__str__, 0, __pyx_n_s_ZeroOrMore___str, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__710)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_str, __pyx_t_13) < 0) __PYX_ERR(0, 4563, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4541
 * 
 * 
 * class ZeroOrMore(_MultipleMatch):             # <<<<<<<<<<<<<<
 *     """Optional repetition of zero or more of the given expression.
 * 
 */
  __pyx_t_13 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_ZeroOrMore, __pyx_t_1, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ZeroOrMore, __pyx_t_13) < 0) __PYX_ERR(0, 4541, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4573
 * 
 * 
 * class _NullToken(object):             # <<<<<<<<<<<<<<
 *     def __bool__(self):
 *         return False
 */
  __pyx_t_1 = __Pyx_CalculateMetaclass(NULL, __pyx_tuple__711); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_Py3MetaclassPrepare(__pyx_t_1, __pyx_tuple__711, __pyx_n_s_NullToken, __pyx_n_s_NullToken, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, (PyObject *) NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "Aspidites/_vendor/pyparsing_extension.py":4574
 * 
 * class _NullToken(object):
 *     def __bool__(self):             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10_NullToken_1__bool__, 0, __pyx_n_s_NullToken___bool, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__713)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_bool, __pyx_t_7) < 0) __PYX_ERR(0, 4574, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4577
 *         return False
 * 
 *     __nonzero__ = __bool__             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __pyx_t_7 = PyObject_GetItem(__pyx_t_2, __pyx_n_s_bool);
  if (unlikely(!__pyx_t_7)) {
    PyErr_Clear();
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_bool);
  }
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_nonzero, __pyx_t_7) < 0) __PYX_ERR(0, 4577, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4579
 *     __nonzero__ = __bool__
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return ""
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_10_NullToken_3__str__, 0, __pyx_n_s_NullToken___str, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__715)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_str, __pyx_t_7) < 0) __PYX_ERR(0, 4579, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4573
 * 
 * 
 * class _NullToken(object):             # <<<<<<<<<<<<<<
 *     def __bool__(self):
 *         return False
 */
  __pyx_t_7 = __Pyx_Py3ClassCreate(__pyx_t_1, __pyx_n_s_NullToken, __pyx_tuple__711, __pyx_t_2, NULL, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_NullToken, __pyx_t_7) < 0) __PYX_ERR(0, 4573, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4583
 * 
 * 
 * class Optional(ParseElementEnhance):             # <<<<<<<<<<<<<<
 *     """Optional matching of the given expression.
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParseElementEnhance); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_1, __pyx_t_2, __pyx_n_s_Optional, __pyx_n_s_Optional, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_Optional_matching_of_the_given_e); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":4620
 *         FAIL: Expected end of text (at char 5), (line:1, col:6)
 *     """
 *     __optionalNotMatched = _NullToken()             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, expr, default=__optionalNotMatched):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_NullToken); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 4620, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_9);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_9, function);
    }
  }
  __pyx_t_13 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4620, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_Optional__optionalNotMatched, __pyx_t_13) < 0) __PYX_ERR(0, 4620, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4622
 *     __optionalNotMatched = _NullToken()
 * 
 *     def __init__(self, expr, default=__optionalNotMatched):             # <<<<<<<<<<<<<<
 *         super(Optional, self).__init__(expr, savelist=False)
 *         self.saveAsList = self.expr.saveAsList
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_8Optional_1__init__, 0, __pyx_n_s_Optional___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__717)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_13, sizeof(__pyx_defaults7), 1)) __PYX_ERR(0, 4622, __pyx_L1_error)
  __pyx_t_9 = PyObject_GetItem(__pyx_t_7, __pyx_n_s_Optional__optionalNotMatched);
  if (unlikely(!__pyx_t_9)) {
    PyErr_Clear();
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_Optional__optionalNotMatched);
  }
  if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 4622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_Defaults(__pyx_defaults7, __pyx_t_13)->__pyx_arg_default = __pyx_t_9;
  __Pyx_GIVEREF(__pyx_t_9);
  __pyx_t_9 = 0;
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_13, __pyx_pf_9Aspidites_7_vendor_19pyparsing_extension_42__defaults__);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_13) < 0) __PYX_ERR(0, 4622, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4628
 *         self.mayReturnEmpty = True
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         try:
 *             loc, tokens = self.expr._parse(instring, loc, doActions, callPreParse=False)
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_8Optional_3parseImpl, 0, __pyx_n_s_Optional_parseImpl, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__719)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_13, __pyx_tuple__720);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_parseImpl, __pyx_t_13) < 0) __PYX_ERR(0, 4628, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4642
 *         return loc, tokens
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_8Optional_5__str__, 0, __pyx_n_s_Optional___str, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__722)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_str, __pyx_t_13) < 0) __PYX_ERR(0, 4642, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4583
 * 
 * 
 * class Optional(ParseElementEnhance):             # <<<<<<<<<<<<<<
 *     """Optional matching of the given expression.
 * 
 */
  __pyx_t_13 = __Pyx_Py3ClassCreate(__pyx_t_1, __pyx_n_s_Optional, __pyx_t_2, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Optional, __pyx_t_13) < 0) __PYX_ERR(0, 4583, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4652
 * 
 * 
 * class SkipTo(ParseElementEnhance):             # <<<<<<<<<<<<<<
 *     """Token for skipping over all undefined text until the matched
 *     expression is found.
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParseElementEnhance); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_1, __pyx_n_s_SkipTo, __pyx_n_s_SkipTo, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_Token_for_skipping_over_all_unde); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":4711
 *     """
 * 
 *     def __init__(self, other, include=False, ignore=None, failOn=None):             # <<<<<<<<<<<<<<
 *         super(SkipTo, self).__init__(other)
 *         self.ignoreExpr = ignore
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_6SkipTo_1__init__, 0, __pyx_n_s_SkipTo___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__724)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4711, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_13, __pyx_tuple__725);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_13) < 0) __PYX_ERR(0, 4711, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4724
 *         self.errmsg = "No match found for " + str(self.expr)
 * 
 *     def parseImpl(self, instring, loc, doActions=True):             # <<<<<<<<<<<<<<
 *         startloc = loc
 *         instrlen = len(instring)
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_6SkipTo_3parseImpl, 0, __pyx_n_s_SkipTo_parseImpl, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__727)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4724, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_13, __pyx_tuple__728);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_parseImpl, __pyx_t_13) < 0) __PYX_ERR(0, 4724, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4652
 * 
 * 
 * class SkipTo(ParseElementEnhance):             # <<<<<<<<<<<<<<
 *     """Token for skipping over all undefined text until the matched
 *     expression is found.
 */
  __pyx_t_13 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_SkipTo, __pyx_t_1, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SkipTo, __pyx_t_13) < 0) __PYX_ERR(0, 4652, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4772
 * 
 * 
 * class Forward(ParseElementEnhance):             # <<<<<<<<<<<<<<
 *     """Forward declaration of an expression to be defined later -
 *     used for recursive grammars, such as algebraic infix notation.
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ParseElementEnhance); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4772, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4772, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4772, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_1, __pyx_t_2, __pyx_n_s_Forward, __pyx_n_s_Forward, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_Forward_declaration_of_an_expres); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4772, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":4800
 *     """
 * 
 *     def __init__(self, other=None):             # <<<<<<<<<<<<<<
 *         super(Forward, self).__init__(other, savelist=False)
 * 
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7Forward_1__init__, 0, __pyx_n_s_Forward___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__730)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4800, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_13, __pyx_tuple__731);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_13) < 0) __PYX_ERR(0, 4800, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4803
 *         super(Forward, self).__init__(other, savelist=False)
 * 
 *     def __lshift__(self, other):             # <<<<<<<<<<<<<<
 *         if isinstance(other, str):
 *             other = self._literalStringClass(other)
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7Forward_3__lshift__, 0, __pyx_n_s_Forward___lshift, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__733)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4803, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_lshift, __pyx_t_13) < 0) __PYX_ERR(0, 4803, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4816
 *         return self
 * 
 *     def __ilshift__(self, other):             # <<<<<<<<<<<<<<
 *         return self << other
 * 
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7Forward_5__ilshift__, 0, __pyx_n_s_Forward___ilshift, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__735)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4816, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_ilshift, __pyx_t_13) < 0) __PYX_ERR(0, 4816, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4819
 *         return self << other
 * 
 *     def leaveWhitespace(self):             # <<<<<<<<<<<<<<
 *         self.skipWhitespace = False
 *         return self
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7Forward_7leaveWhitespace, 0, __pyx_n_s_Forward_leaveWhitespace, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__737)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_leaveWhitespace, __pyx_t_13) < 0) __PYX_ERR(0, 4819, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4823
 *         return self
 * 
 *     def streamline(self):             # <<<<<<<<<<<<<<
 *         if not self.streamlined:
 *             self.streamlined = True
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7Forward_9streamline, 0, __pyx_n_s_Forward_streamline, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__739)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_streamline, __pyx_t_13) < 0) __PYX_ERR(0, 4823, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4830
 *         return self
 * 
 *     def validate(self, validateTrace=None):             # <<<<<<<<<<<<<<
 *         if validateTrace is None:
 *             validateTrace = []
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7Forward_11validate, 0, __pyx_n_s_Forward_validate, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__741)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_13, __pyx_tuple__742);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_validate, __pyx_t_13) < 0) __PYX_ERR(0, 4830, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4840
 *         self.checkRecursion([])
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if hasattr(self, "name"):
 *             return self.name
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7Forward_13__str__, 0, __pyx_n_s_Forward___str, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__744)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_str, __pyx_t_13) < 0) __PYX_ERR(0, 4840, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4860
 *         return self.strRepr
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         if self.expr is not None:
 *             return super(Forward, self).copy()
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7Forward_15copy, 0, __pyx_n_s_Forward_copy, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__746)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4860, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_copy, __pyx_t_13) < 0) __PYX_ERR(0, 4860, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4868
 *             return ret
 * 
 *     def _setResultsName(self, name, listAllMatches=False):             # <<<<<<<<<<<<<<
 * 
 *         if __diag__.warn_name_set_on_empty_Forward:
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7Forward_17_setResultsName, 0, __pyx_n_s_Forward__setResultsName, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__748)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_13, __pyx_tuple__749);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_setResultsName, __pyx_t_13) < 0) __PYX_ERR(0, 4868, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4772
 * 
 * 
 * class Forward(ParseElementEnhance):             # <<<<<<<<<<<<<<
 *     """Forward declaration of an expression to be defined later -
 *     used for recursive grammars, such as algebraic infix notation.
 */
  __pyx_t_13 = __Pyx_Py3ClassCreate(__pyx_t_1, __pyx_n_s_Forward, __pyx_t_2, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4772, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Forward, __pyx_t_13) < 0) __PYX_ERR(0, 4772, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4881
 * 
 * 
 * class TokenConverter(ParseElementEnhance):             # <<<<<<<<<<<<<<
 *     """
 *     Abstract subclass of :class:`ParseExpression`, for converting parsed results.
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ParseElementEnhance); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4881, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4881, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4881, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_1, __pyx_n_s_TokenConverter, __pyx_n_s_TokenConverter, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_Abstract_subclass_of_class_Pars); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4881, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":4886
 *     """
 * 
 *     def __init__(self, expr, savelist=False):             # <<<<<<<<<<<<<<
 *         super(TokenConverter, self).__init__(expr)  # , savelist)
 *         self.saveAsList = False
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_14TokenConverter_1__init__, 0, __pyx_n_s_TokenConverter___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__751)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4886, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_13, __pyx_tuple__752);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_13) < 0) __PYX_ERR(0, 4886, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4881
 * 
 * 
 * class TokenConverter(ParseElementEnhance):             # <<<<<<<<<<<<<<
 *     """
 *     Abstract subclass of :class:`ParseExpression`, for converting parsed results.
 */
  __pyx_t_13 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_TokenConverter, __pyx_t_1, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4881, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TokenConverter, __pyx_t_13) < 0) __PYX_ERR(0, 4881, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4891
 * 
 * 
 * class Combine(TokenConverter):             # <<<<<<<<<<<<<<
 *     """Converter to concatenate all matching tokens to a single string.
 *     By default, the matching patterns must also be contiguous in the
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_TokenConverter); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_1, __pyx_t_2, __pyx_n_s_Combine, __pyx_n_s_Combine, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_Converter_to_concatenate_all_mat); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":4910
 *     """
 * 
 *     def __init__(self, expr, joinString="", adjacent=True):             # <<<<<<<<<<<<<<
 *         super(Combine, self).__init__(expr)
 *         # suppress whitespace-stripping in contained parse expressions, but re-enable it on the Combine itself
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7Combine_1__init__, 0, __pyx_n_s_Combine___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__754)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4910, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_13, __pyx_tuple__755);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_13) < 0) __PYX_ERR(0, 4910, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4920
 *         self.callPreparse = True
 * 
 *     def ignore(self, other):             # <<<<<<<<<<<<<<
 *         if self.adjacent:
 *             ParserElement.ignore(self, other)
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7Combine_3ignore, 0, __pyx_n_s_Combine_ignore, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__757)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_ignore, __pyx_t_13) < 0) __PYX_ERR(0, 4920, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4927
 *         return self
 * 
 *     def postParse(self, instring, loc, tokenlist):             # <<<<<<<<<<<<<<
 *         retToks = tokenlist.copy()
 *         del retToks[:]
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_7Combine_5postParse, 0, __pyx_n_s_Combine_postParse, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__759)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4927, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_postParse, __pyx_t_13) < 0) __PYX_ERR(0, 4927, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4891
 * 
 * 
 * class Combine(TokenConverter):             # <<<<<<<<<<<<<<
 *     """Converter to concatenate all matching tokens to a single string.
 *     By default, the matching patterns must also be contiguous in the
 */
  __pyx_t_13 = __Pyx_Py3ClassCreate(__pyx_t_1, __pyx_n_s_Combine, __pyx_t_2, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Combine, __pyx_t_13) < 0) __PYX_ERR(0, 4891, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4938
 * 
 * 
 * class Group(TokenConverter):             # <<<<<<<<<<<<<<
 *     """Converter to return the matched tokens as a list - useful for
 *     returning tokens of :class:`ZeroOrMore` and :class:`OneOrMore` expressions.
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_TokenConverter); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4938, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4938, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4938, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_1, __pyx_n_s_Group, __pyx_n_s_Group, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_Converter_to_return_the_matched); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4938, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":4954
 *     """
 * 
 *     def __init__(self, expr):             # <<<<<<<<<<<<<<
 *         super(Group, self).__init__(expr)
 *         self.saveAsList = True
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_5Group_1__init__, 0, __pyx_n_s_Group___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__761)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4954, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_13) < 0) __PYX_ERR(0, 4954, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4958
 *         self.saveAsList = True
 * 
 *     def postParse(self, instring, loc, tokenlist):             # <<<<<<<<<<<<<<
 *         return [tokenlist]
 * 
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_5Group_3postParse, 0, __pyx_n_s_Group_postParse, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__763)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4958, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_postParse, __pyx_t_13) < 0) __PYX_ERR(0, 4958, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4938
 * 
 * 
 * class Group(TokenConverter):             # <<<<<<<<<<<<<<
 *     """Converter to return the matched tokens as a list - useful for
 *     returning tokens of :class:`ZeroOrMore` and :class:`OneOrMore` expressions.
 */
  __pyx_t_13 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_Group, __pyx_t_1, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4938, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Group, __pyx_t_13) < 0) __PYX_ERR(0, 4938, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4962
 * 
 * 
 * class Dict(TokenConverter):             # <<<<<<<<<<<<<<
 *     """Converter to return a repetitive expression as a list, but also
 *     as a dictionary. Each element can also be referenced using the first
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_TokenConverter); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4962, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4962, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4962, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_1, __pyx_t_2, __pyx_n_s_Dict, __pyx_n_s_Dict, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_Converter_to_return_a_repetitive); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 4962, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":5002
 *     """
 * 
 *     def __init__(self, expr):             # <<<<<<<<<<<<<<
 *         super(Dict, self).__init__(expr)
 *         self.saveAsList = True
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_4Dict_1__init__, 0, __pyx_n_s_Dict___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__765)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 5002, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_init, __pyx_t_13) < 0) __PYX_ERR(0, 5002, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":5006
 *         self.saveAsList = True
 * 
 *     def postParse(self, instring, loc, tokenlist):             # <<<<<<<<<<<<<<
 *         for i, tok in enumerate(tokenlist):
 *             if len(tok) == 0:
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_4Dict_3postParse, 0, __pyx_n_s_Dict_postParse, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__767)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 5006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_postParse, __pyx_t_13) < 0) __PYX_ERR(0, 5006, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":4962
 * 
 * 
 * class Dict(TokenConverter):             # <<<<<<<<<<<<<<
 *     """Converter to return a repetitive expression as a list, but also
 *     as a dictionary. Each element can also be referenced using the first
 */
  __pyx_t_13 = __Pyx_Py3ClassCreate(__pyx_t_1, __pyx_n_s_Dict, __pyx_t_2, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 4962, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Dict, __pyx_t_13) < 0) __PYX_ERR(0, 4962, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":5031
 * 
 * 
 * class Suppress(TokenConverter):             # <<<<<<<<<<<<<<
 *     """Converter for ignoring the results of a parsed expression.
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_TokenConverter); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5031, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5031, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5031, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_1, __pyx_n_s_Suppress, __pyx_n_s_Suppress, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_Converter_for_ignoring_the_resul); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5031, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "Aspidites/_vendor/pyparsing_extension.py":5054
 *     """
 * 
 *     def postParse(self, instring, loc, tokenlist):             # <<<<<<<<<<<<<<
 *         return []
 * 
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_8Suppress_1postParse, 0, __pyx_n_s_Suppress_postParse, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__769)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 5054, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_postParse, __pyx_t_13) < 0) __PYX_ERR(0, 5054, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":5057
 *         return []
 * 
 *     def suppress(self):             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __pyx_t_13 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_8Suppress_3suppress, 0, __pyx_n_s_Suppress_suppress, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__771)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 5057, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_suppress, __pyx_t_13) < 0) __PYX_ERR(0, 5057, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":5031
 * 
 * 
 * class Suppress(TokenConverter):             # <<<<<<<<<<<<<<
 *     """Converter for ignoring the results of a parsed expression.
 * 
 */
  __pyx_t_13 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_Suppress, __pyx_t_1, __pyx_t_7, NULL, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 5031, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Suppress, __pyx_t_13) < 0) __PYX_ERR(0, 5031, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":5061
 * 
 * 
 * class OnlyOnce(object):             # <<<<<<<<<<<<<<
 *     """Wrapper for parse actions, to ensure they are only called once.
 *     """
 */
  __pyx_t_1 = __Pyx_CalculateMetaclass(NULL, __pyx_tuple__772); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5061, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_Py3MetaclassPrepare(__pyx_t_1, __pyx_tuple__772, __pyx_n_s_OnlyOnce, __pyx_n_s_OnlyOnce, (PyObject *) NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_kp_s_Wrapper_for_parse_actions_to_ens); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5061, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "Aspidites/_vendor/pyparsing_extension.py":5064
 *     """Wrapper for parse actions, to ensure they are only called once.
 *     """
 *     def __init__(self, methodCall):             # <<<<<<<<<<<<<<
 *         from .pyparsing import _trim_arity
 *         self.callable = _trim_arity(methodCall)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_8OnlyOnce_1__init__, 0, __pyx_n_s_OnlyOnce___init, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__774)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5064, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_init, __pyx_t_7) < 0) __PYX_ERR(0, 5064, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":5069
 *         self.called = False
 * 
 *     def __call__(self, s, l, t):             # <<<<<<<<<<<<<<
 *         if not self.called:
 *             results = self.callable(s, l, t)
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_8OnlyOnce_3__call__, 0, __pyx_n_s_OnlyOnce___call, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__776)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5069, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_call, __pyx_t_7) < 0) __PYX_ERR(0, 5069, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":5076
 *         raise ParseException(s, l, "")
 * 
 *     def reset(self):             # <<<<<<<<<<<<<<
 *         self.called = False
 */
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9Aspidites_7_vendor_19pyparsing_extension_8OnlyOnce_5reset, 0, __pyx_n_s_OnlyOnce_reset, NULL, __pyx_n_s_Aspidites__vendor_pyparsing_exte, __pyx_d, ((PyObject *)__pyx_codeobj__778)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5076, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_reset, __pyx_t_7) < 0) __PYX_ERR(0, 5076, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":5061
 * 
 * 
 * class OnlyOnce(object):             # <<<<<<<<<<<<<<
 *     """Wrapper for parse actions, to ensure they are only called once.
 *     """
 */
  __pyx_t_7 = __Pyx_Py3ClassCreate(__pyx_t_1, __pyx_n_s_OnlyOnce, __pyx_tuple__772, __pyx_t_2, NULL, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 5061, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_OnlyOnce, __pyx_t_7) < 0) __PYX_ERR(0, 5061, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "Aspidites/_vendor/pyparsing_extension.py":1
 * import collections             # <<<<<<<<<<<<<<
 * import copy
 * import pprint
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_18);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init Aspidites._vendor.pyparsing_extension", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init Aspidites._vendor.pyparsing_extension");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (__Pyx_PyFastCFunction_Check(func)) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* KeywordStringCheck */
static int __Pyx_CheckKeywordStrings(
    PyObject *kwdict,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kwdict, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    while (PyDict_Next(kwdict, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if ((!kw_allowed) && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a - b);
            if (likely((x^a) >= 0 || (x^~b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
            }
        }
                x = a - b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla - llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("subtract", return NULL)
            result = ((double)a) - (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceSubtract : PyNumber_Subtract)(op1, op2);
}
#endif

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* SliceObject */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object is unsliceable", Py_TYPE(obj)->tp_name);
bad:
    return NULL;
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* HasAttr */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* None */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname) {
    PyErr_Format(PyExc_NameError, "free variable '%s' referenced before assignment in enclosing scope", varname);
}

/* UnpackUnboundCMethod */
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    }
#endif
    return 0;
}

/* CallUnboundCMethod0 */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_ASSUME_SAFE_MACROS
    args = PyTuple_New(1);
    if (unlikely(!args)) goto bad;
    Py_INCREF(self);
    PyTuple_SET_ITEM(args, 0, self);
#else
    args = PyTuple_Pack(1, self);
    if (unlikely(!args)) goto bad;
#endif
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    Py_DECREF(args);
bad:
    return result;
}

/* py_dict_items */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d) {
    if (PY_MAJOR_VERSION >= 3)
        return __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyDict_Type_items, d);
    else
        return PyDict_Items(d);
}

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod1 */
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* append */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x) {
    if (likely(PyList_CheckExact(L))) {
        if (unlikely(__Pyx_PyList_Append(L, x) < 0)) return -1;
    } else {
        PyObject* retval = __Pyx_PyObject_CallMethod1(L, __pyx_n_s_append, x);
        if (unlikely(!retval))
            return -1;
        Py_DECREF(retval);
    }
    return 0;
}

/* SliceObject */
static CYTHON_INLINE int __Pyx_PyObject_SetSlice(PyObject* obj, PyObject* value,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_ass_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_ass_slice(obj, cstart, cstop, value);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_ass_subscript))
#endif
    {
        int result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_ass_subscript(obj, py_slice, value);
#else
        result = value ? PyObject_SetItem(obj, py_slice, value) : PyObject_DelItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object does not support slice %.10s",
        Py_TYPE(obj)->tp_name, value ? "assignment" : "deletion");
bad:
    return -1;
}

/* FetchCommonType */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CythonFunctionShared */
#include <structmember.h>
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    Py_INCREF(m->func_qualname);
    return m->func_qualname;
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    if (unlikely(op == NULL))
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults_size = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
#if PY_MAJOR_VERSION < 3
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
#endif
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    m->defaults_size = size;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* PyIntCompare */
static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED long inplace) {
    if (op1 == op2) {
        Py_RETURN_TRUE;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        if (a == b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = Py_SIZE(op1);
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        if (intval == 0) {
            if (size == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
        } else if (intval < 0) {
            if (size >= 0)
                Py_RETURN_FALSE;
            intval = -intval;
            size = -size;
        } else {
            if (size <= 0)
                Py_RETURN_FALSE;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        if (unequal == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
        if ((double)a == (double)b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    return (
        PyObject_RichCompare(op1, op2, Py_EQ));
}

/* StringJoin */
#if !CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values) {
    return PyObject_CallMethodObjArgs(sep, __pyx_n_s_join, values, NULL);
}
#endif

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* IterNext */
static PyObject *__Pyx_PyIter_Next2Default(PyObject* defval) {
    PyObject* exc_type;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    exc_type = __Pyx_PyErr_Occurred();
    if (unlikely(exc_type)) {
        if (!defval || unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))
            return NULL;
        __Pyx_PyErr_Clear();
        Py_INCREF(defval);
        return defval;
    }
    if (defval) {
        Py_INCREF(defval);
        return defval;
    }
    __Pyx_PyErr_SetNone(PyExc_StopIteration);
    return NULL;
}
static void __Pyx_PyIter_Next_ErrorNoIterator(PyObject *iterator) {
    PyErr_Format(PyExc_TypeError,
        "%.200s object is not an iterator", Py_TYPE(iterator)->tp_name);
}
static CYTHON_INLINE PyObject *__Pyx_PyIter_Next2(PyObject* iterator, PyObject* defval) {
    PyObject* next;
    iternextfunc iternext = Py_TYPE(iterator)->tp_iternext;
    if (likely(iternext)) {
#if CYTHON_USE_TYPE_SLOTS
        next = iternext(iterator);
        if (likely(next))
            return next;
        #if PY_VERSION_HEX >= 0x02070000
        if (unlikely(iternext == &_PyObject_NextNotImplemented))
            return NULL;
        #endif
#else
        next = PyIter_Next(iterator);
        if (likely(next))
            return next;
#endif
    } else if (CYTHON_USE_TYPE_SLOTS || unlikely(!PyIter_Check(iterator))) {
        __Pyx_PyIter_Next_ErrorNoIterator(iterator);
        return NULL;
    }
#if !CYTHON_USE_TYPE_SLOTS
    else {
        next = PyIter_Next(iterator);
        if (likely(next))
            return next;
    }
#endif
    return __Pyx_PyIter_Next2Default(defval);
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* PatchModuleWithCoroutine */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchInspect */
static PyObject* __Pyx_patch_inspect(PyObject* module) {
#if defined(__Pyx_Generator_USED) && (!defined(CYTHON_PATCH_INSPECT) || CYTHON_PATCH_INSPECT)
    static int inspect_patched = 0;
    if (unlikely((!inspect_patched) && module)) {
        module = __Pyx_Coroutine_patch_module(
            module, ""
"old_types = getattr(_module.isgenerator, '_cython_generator_types', None)\n"
"if old_types is None or not isinstance(old_types, set):\n"
"    old_types = set()\n"
"    def cy_wrap(orig_func, type=type, cython_generator_types=old_types):\n"
"        def cy_isgenerator(obj): return type(obj) in cython_generator_types or orig_func(obj)\n"
"        cy_isgenerator._cython_generator_types = cython_generator_types\n"
"        return cy_isgenerator\n"
"    _module.isgenerator = cy_wrap(_module.isgenerator)\n"
"old_types.add(_cython_generator_type)\n"
        );
        inspect_patched = 1;
    }
#else
    if ((0)) return __Pyx_Coroutine_patch_module(module, NULL);
#endif
    return module;
}

/* pyfrozenset_new */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it) {
    if (it) {
        PyObject* result;
#if CYTHON_COMPILING_IN_PYPY
        PyObject* args;
        args = PyTuple_Pack(1, it);
        if (unlikely(!args))
            return NULL;
        result = PyObject_Call((PyObject*)&PyFrozenSet_Type, args, NULL);
        Py_DECREF(args);
        return result;
#else
        if (PyFrozenSet_CheckExact(it)) {
            Py_INCREF(it);
            return it;
        }
        result = PyFrozenSet_New(it);
        if (unlikely(!result))
            return NULL;
        if ((PY_VERSION_HEX >= 0x031000A1) || likely(PySet_GET_SIZE(result)))
            return result;
        Py_DECREF(result);
#endif
    }
#if CYTHON_USE_TYPE_SLOTS
    return PyFrozenSet_Type.tp_new(&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#else
    return PyObject_Call((PyObject*)&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#endif
}

/* PySetContains */
static int __Pyx_PySet_ContainsUnhashable(PyObject *set, PyObject *key) {
    int result = -1;
    if (PySet_Check(key) && PyErr_ExceptionMatches(PyExc_TypeError)) {
        PyObject *tmpkey;
        PyErr_Clear();
        tmpkey = __Pyx_PyFrozenSet_New(key);
        if (tmpkey != NULL) {
            result = PySet_Contains(set, tmpkey);
            Py_DECREF(tmpkey);
        }
    }
    return result;
}
static CYTHON_INLINE int __Pyx_PySet_ContainsTF(PyObject* key, PyObject* set, int eq) {
    int result = PySet_Contains(set, key);
    if (unlikely(result < 0)) {
        result = __Pyx_PySet_ContainsUnhashable(set, key);
    }
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* CallUnboundCMethod1 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg) {
    if (likely(cfunc->func)) {
        int flag = cfunc->flag;
        if (flag == METH_O) {
            return (*(cfunc->func))(self, arg);
        } else if (PY_VERSION_HEX >= 0x030600B1 && flag == METH_FASTCALL) {
            if (PY_VERSION_HEX >= 0x030700A0) {
                return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, &arg, 1);
            } else {
                return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
            }
        } else if (PY_VERSION_HEX >= 0x030700A0 && flag == (METH_FASTCALL | METH_KEYWORDS)) {
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
        }
    }
    return __Pyx__CallUnboundCMethod1(cfunc, self, arg);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(1);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 0, arg);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 1, arg);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(2, self, arg);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* CallUnboundCMethod2 */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2) {
    if (likely(cfunc->func)) {
        PyObject *args[2] = {arg1, arg2};
        if (cfunc->flag == METH_FASTCALL) {
            #if PY_VERSION_HEX >= 0x030700A0
            return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, args, 2);
            #else
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, 2, NULL);
            #endif
        }
        #if PY_VERSION_HEX >= 0x030700A0
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, 2, NULL);
        #endif
    }
    return __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 0, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 1, arg2);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(3);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 1, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 2, arg2);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(3, self, arg1, arg2);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* dict_getitem_default */
static PyObject* __Pyx_PyDict_GetItemDefault(PyObject* d, PyObject* key, PyObject* default_value) {
    PyObject* value;
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (unlikely(PyErr_Occurred()))
            return NULL;
        value = default_value;
    }
    Py_INCREF(value);
    if ((1));
#else
    if (PyString_CheckExact(key) || PyUnicode_CheckExact(key) || PyInt_CheckExact(key)) {
        value = PyDict_GetItem(d, key);
        if (unlikely(!value)) {
            value = default_value;
        }
        Py_INCREF(value);
    }
#endif
    else {
        if (default_value == Py_None)
            value = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyDict_Type_get, d, key);
        else
            value = __Pyx_CallUnboundCMethod2(&__pyx_umethod_PyDict_Type_get, d, key, default_value);
    }
    return value;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* None */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* py_dict_pop */
static CYTHON_INLINE PyObject *__Pyx_PyDict_Pop(PyObject *d, PyObject *key, PyObject *default_value) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX > 0x030600B3
    if ((1)) {
        return _PyDict_Pop(d, key, default_value);
    } else
#endif
    if (default_value) {
        return __Pyx_CallUnboundCMethod2(&__pyx_umethod_PyDict_Type_pop, d, key, default_value);
    } else {
        return __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyDict_Type_pop, d, key);
    }
}

/* SetItemInt */
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (!j) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               CYTHON_NCP_UNUSED int wraparound, CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o)))) {
            PyObject* old = PyList_GET_ITEM(o, n);
            Py_INCREF(v);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
            return 1;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return m->sq_ass_item(o, i, v);
        }
    }
#else
#if CYTHON_COMPILING_IN_PYPY
    if (is_list || (PySequence_Check(o) && !PyDict_Check(o)))
#else
    if (is_list || PySequence_Check(o))
#endif
    {
        return PySequence_SetItem(o, i, v);
    }
#endif
    return __Pyx_SetItemInt_Generic(o, PyInt_FromSsize_t(i), v);
}

/* GetAttr3 */
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* PyIntCompare */
static CYTHON_INLINE PyObject* __Pyx_PyInt_NeObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED long inplace) {
    if (op1 == op2) {
        Py_RETURN_FALSE;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        if (a != b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = Py_SIZE(op1);
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        if (intval == 0) {
            if (size != 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
        } else if (intval < 0) {
            if (size >= 0)
                Py_RETURN_TRUE;
            intval = -intval;
            size = -size;
        } else {
            if (size <= 0)
                Py_RETURN_TRUE;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        if (unequal != 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
        if ((double)a != (double)b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    return (
        PyObject_RichCompare(op1, op2, Py_NE));
}

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* DelItemInt */
static int __Pyx_DelItem_Generic(PyObject *o, PyObject *j) {
    int r;
    if (!j) return -1;
    r = PyObject_DelItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_DelItemInt_Fast(PyObject *o, Py_ssize_t i,
                                               CYTHON_UNUSED int is_list, CYTHON_NCP_UNUSED int wraparound) {
#if !CYTHON_USE_TYPE_SLOTS
    if (is_list || PySequence_Check(o)) {
        return PySequence_DelItem(o, i);
    }
#else
    PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
    if (likely(m && m->sq_ass_item)) {
        if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
            Py_ssize_t l = m->sq_length(o);
            if (likely(l >= 0)) {
                i += l;
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    return -1;
                PyErr_Clear();
            }
        }
        return m->sq_ass_item(o, i, (PyObject *)NULL);
    }
#endif
    return __Pyx_DelItem_Generic(o, PyInt_FromSsize_t(i));
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* SliceTupleAndList */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_crop_slice(Py_ssize_t* _start, Py_ssize_t* _stop, Py_ssize_t* _length) {
    Py_ssize_t start = *_start, stop = *_stop, length = *_length;
    if (start < 0) {
        start += length;
        if (start < 0)
            start = 0;
    }
    if (stop < 0)
        stop += length;
    else if (stop > length)
        stop = length;
    *_length = stop - start;
    *_start = start;
    *_stop = stop;
}
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject** CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject* __Pyx_PyList_GetSlice(
            PyObject* src, Py_ssize_t start, Py_ssize_t stop) {
    PyObject* dest;
    Py_ssize_t length = PyList_GET_SIZE(src);
    __Pyx_crop_slice(&start, &stop, &length);
    if (unlikely(length <= 0))
        return PyList_New(0);
    dest = PyList_New(length);
    if (unlikely(!dest))
        return NULL;
    __Pyx_copy_object_array(
        ((PyListObject*)src)->ob_item + start,
        ((PyListObject*)dest)->ob_item,
        length);
    return dest;
}
static CYTHON_INLINE PyObject* __Pyx_PyTuple_GetSlice(
            PyObject* src, Py_ssize_t start, Py_ssize_t stop) {
    PyObject* dest;
    Py_ssize_t length = PyTuple_GET_SIZE(src);
    __Pyx_crop_slice(&start, &stop, &length);
    if (unlikely(length <= 0))
        return PyTuple_New(0);
    dest = PyTuple_New(length);
    if (unlikely(!dest))
        return NULL;
    __Pyx_copy_object_array(
        ((PyTupleObject*)src)->ob_item + start,
        ((PyTupleObject*)dest)->ob_item,
        length);
    return dest;
}
#endif

/* CalculateMetaclass */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases = PyTuple_GET_SIZE(bases);
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
        tmptype = Py_TYPE(tmp);
#if PY_MAJOR_VERSION < 3
        if (tmptype == &PyClass_Type)
            continue;
#endif
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
#if PY_MAJOR_VERSION < 3
        metaclass = &PyClass_Type;
#else
        metaclass = &PyType_Type;
#endif
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

/* Py3ClassCreate */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStr(metaclass, __pyx_n_s_prepare);
        if (prep) {
            PyObject *pargs = PyTuple_Pack(2, name, bases);
            if (unlikely(!pargs)) {
                Py_DECREF(prep);
                return NULL;
            }
            ns = PyObject_Call(prep, pargs, mkw);
            Py_DECREF(prep);
            Py_DECREF(pargs);
        } else {
            if (unlikely(!PyErr_ExceptionMatches(PyExc_AttributeError)))
                return NULL;
            PyErr_Clear();
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_module, modname) < 0)) goto bad;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_qualname, qualname) < 0)) goto bad;
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_n_s_doc, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result, *margs;
    PyObject *owned_metaclass = NULL;
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_n_s_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    margs = PyTuple_Pack(3, name, bases, dict);
    if (unlikely(!margs)) {
        result = NULL;
    } else {
        result = PyObject_Call(metaclass, margs, mkw);
        Py_DECREF(margs);
    }
    Py_XDECREF(owned_metaclass);
    return result;
}

/* ClassMethod */
static PyObject* __Pyx_Method_ClassMethod(PyObject *method) {
#if CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM <= 0x05080000
    if (PyObject_TypeCheck(method, &PyWrapperDescr_Type)) {
        return PyClassMethod_New(method);
    }
#else
#if CYTHON_COMPILING_IN_PYSTON || CYTHON_COMPILING_IN_PYPY
    if (PyMethodDescr_Check(method))
#else
    #if PY_MAJOR_VERSION == 2
    static PyTypeObject *methoddescr_type = NULL;
    if (methoddescr_type == NULL) {
       PyObject *meth = PyObject_GetAttrString((PyObject*)&PyList_Type, "append");
       if (!meth) return NULL;
       methoddescr_type = Py_TYPE(meth);
       Py_DECREF(meth);
    }
    #else
    PyTypeObject *methoddescr_type = &PyMethodDescr_Type;
    #endif
    if (__Pyx_TypeCheck(method, methoddescr_type))
#endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject *)method;
        #if PY_VERSION_HEX < 0x03020000
        PyTypeObject *d_type = descr->d_type;
        #else
        PyTypeObject *d_type = descr->d_common.d_type;
        #endif
        return PyDescr_NewClassMethod(d_type, descr->d_method);
    }
#endif
    else if (PyMethod_Check(method)) {
        return PyClassMethod_New(PyMethod_GET_FUNCTION(method));
    }
    else {
        return PyClassMethod_New(method);
    }
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* Print */
#if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION < 3
static PyObject *__Pyx_GetStdout(void) {
    PyObject *f = PySys_GetObject((char *)"stdout");
    if (!f) {
        PyErr_SetString(PyExc_RuntimeError, "lost sys.stdout");
    }
    return f;
}
static int __Pyx_Print(PyObject* f, PyObject *arg_tuple, int newline) {
    int i;
    if (!f) {
        if (!(f = __Pyx_GetStdout()))
            return -1;
    }
    Py_INCREF(f);
    for (i=0; i < PyTuple_GET_SIZE(arg_tuple); i++) {
        PyObject* v;
        if (PyFile_SoftSpace(f, 1)) {
            if (PyFile_WriteString(" ", f) < 0)
                goto error;
        }
        v = PyTuple_GET_ITEM(arg_tuple, i);
        if (PyFile_WriteObject(v, f, Py_PRINT_RAW) < 0)
            goto error;
        if (PyString_Check(v)) {
            char *s = PyString_AsString(v);
            Py_ssize_t len = PyString_Size(v);
            if (len > 0) {
                switch (s[len-1]) {
                    case ' ': break;
                    case '\f': case '\r': case '\n': case '\t': case '\v':
                        PyFile_SoftSpace(f, 0);
                        break;
                    default:  break;
                }
            }
        }
    }
    if (newline) {
        if (PyFile_WriteString("\n", f) < 0)
            goto error;
        PyFile_SoftSpace(f, 0);
    }
    Py_DECREF(f);
    return 0;
error:
    Py_DECREF(f);
    return -1;
}
#else
static int __Pyx_Print(PyObject* stream, PyObject *arg_tuple, int newline) {
    PyObject* kwargs = 0;
    PyObject* result = 0;
    PyObject* end_string;
    if (unlikely(!__pyx_print)) {
        __pyx_print = PyObject_GetAttr(__pyx_b, __pyx_n_s_print_2);
        if (!__pyx_print)
            return -1;
    }
    if (stream) {
        kwargs = PyDict_New();
        if (unlikely(!kwargs))
            return -1;
        if (unlikely(PyDict_SetItem(kwargs, __pyx_n_s_file, stream) < 0))
            goto bad;
        if (!newline) {
            end_string = PyUnicode_FromStringAndSize(" ", 1);
            if (unlikely(!end_string))
                goto bad;
            if (PyDict_SetItem(kwargs, __pyx_n_s_end, end_string) < 0) {
                Py_DECREF(end_string);
                goto bad;
            }
            Py_DECREF(end_string);
        }
    } else if (!newline) {
        if (unlikely(!__pyx_print_kwargs)) {
            __pyx_print_kwargs = PyDict_New();
            if (unlikely(!__pyx_print_kwargs))
                return -1;
            end_string = PyUnicode_FromStringAndSize(" ", 1);
            if (unlikely(!end_string))
                return -1;
            if (PyDict_SetItem(__pyx_print_kwargs, __pyx_n_s_end, end_string) < 0) {
                Py_DECREF(end_string);
                return -1;
            }
            Py_DECREF(end_string);
        }
        kwargs = __pyx_print_kwargs;
    }
    result = PyObject_Call(__pyx_print, arg_tuple, kwargs);
    if (unlikely(kwargs) && (kwargs != __pyx_print_kwargs))
        Py_DECREF(kwargs);
    if (!result)
        return -1;
    Py_DECREF(result);
    return 0;
bad:
    if (kwargs != __pyx_print_kwargs)
        Py_XDECREF(kwargs);
    return -1;
}
#endif

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* PrintOne */
#if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION < 3
static int __Pyx_PrintOne(PyObject* f, PyObject *o) {
    if (!f) {
        if (!(f = __Pyx_GetStdout()))
            return -1;
    }
    Py_INCREF(f);
    if (PyFile_SoftSpace(f, 0)) {
        if (PyFile_WriteString(" ", f) < 0)
            goto error;
    }
    if (PyFile_WriteObject(o, f, Py_PRINT_RAW) < 0)
        goto error;
    if (PyFile_WriteString("\n", f) < 0)
        goto error;
    Py_DECREF(f);
    return 0;
error:
    Py_DECREF(f);
    return -1;
    /* the line below is just to avoid C compiler
     * warnings about unused functions */
    return __Pyx_Print(f, NULL, 0);
}
#else
static int __Pyx_PrintOne(PyObject* stream, PyObject *o) {
    int res;
    PyObject* arg_tuple = PyTuple_Pack(1, o);
    if (unlikely(!arg_tuple))
        return -1;
    res = __Pyx_Print(stream, arg_tuple, 1);
    Py_DECREF(arg_tuple);
    return res;
}
#endif

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CoroutineBase */
#include <structmember.h>
#include <frameobject.h>
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (Py_TYPE(ev) == (PyTypeObject*)PyExc_StopIteration) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(CYTHON_UNUSED __pyx_CoroutineObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(CYTHON_UNUSED PyObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(CYTHON_UNUSED PyObject *gen, PyObject *value, CYTHON_UNUSED int closing) {
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_type) {
        #if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
        #else
        if (exc_state->exc_traceback) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_state->exc_traceback;
            PyFrameObject *f = tb->tb_frame;
            Py_XINCREF(tstate->frame);
            assert(f->f_back == NULL);
            f->f_back = tstate->frame;
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body((PyObject *) self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *exc_tb = exc_state->exc_traceback;
    if (likely(exc_tb)) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
#else
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
#endif
    }
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(CYTHON_UNUSED PyObject* gen, PyObject *retval) {
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
static CYTHON_INLINE
PyObject *__Pyx_PyGen_Send(PyGenObject *gen, PyObject *arg) {
#if PY_VERSION_HEX <= 0x030A00A1
    return _PyGen_Send(gen, arg);
#else
    PyObject *result;
    if (PyIter_Send((PyObject*)gen, arg ? arg : Py_None, &result) == PYGEN_RETURN) {
        if (PyAsyncGen_CheckExact(gen)) {
            assert(result == Py_None);
            PyErr_SetNone(PyExc_StopAsyncIteration);
        }
        else if (result == Py_None) {
            PyErr_SetNone(PyExc_StopIteration);
        }
        else {
            _PyGen_SetStopIterationValue(result);
        }
        Py_CLEAR(result);
    }
    return result;
#endif
}
#endif
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = Py_TYPE(yf)->tp_iternext(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                PyErr_WriteUnraisable(yf);
            }
            PyErr_Clear();
        } else {
            retval = PyObject_CallFunction(meth, NULL);
            Py_DECREF(meth);
            if (!retval)
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = Py_TYPE(yf)->tp_iternext(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, CYTHON_UNUSED PyObject *arg) {
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                    gen->is_running = 0;
                    return NULL;
                }
                PyErr_Clear();
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = PyObject_CallObject(meth, args);
            } else {
                ret = PyObject_CallFunctionObjArgs(meth, typ, val, tb, NULL);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_frame);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (PyObject_CallFinalizerFromDealloc(self))
#else
        Py_TYPE(gen)->tp_del(self);
        if (Py_REFCNT(self) > 0)
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    PyObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    __Pyx_SET_REFCNT(self, 1);
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(Py_REFCNT(self) > 0);
    if (--self->ob_refcnt == 0) {
        return;
    }
    {
        Py_ssize_t refcnt = Py_REFCNT(self);
        _Py_NewReference(self);
        __Pyx_SET_REFCNT(self, refcnt);
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(Py_TYPE(self)) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_name;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_name;
    Py_INCREF(value);
    self->gi_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_qualname;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_qualname;
    Py_INCREF(value);
    self->gi_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_frame(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *frame = self->gi_frame;
    if (!frame) {
        if (unlikely(!self->gi_code)) {
            Py_RETURN_NONE;
        }
        frame = (PyObject *) PyFrame_New(
            PyThreadState_Get(),            /*PyThreadState *tstate,*/
            (PyCodeObject*) self->gi_code,  /*PyCodeObject *code,*/
            __pyx_d,                 /*PyObject *globals,*/
            0                               /*PyObject *locals*/
        );
        if (unlikely(!frame))
            return NULL;
        self->gi_frame = frame;
    }
    Py_INCREF(frame);
    return frame;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    gen->gi_frame = NULL;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchGeneratorABC */
#ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (!module) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {(char *) "gi_frame", (getter)__Pyx_Coroutine_get_frame, NULL,
     (char*) PyDoc_STR("Frame of the generator"), 0},
    {0, 0, 0, 0, 0}
};
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
};
static int __pyx_Generator_init(void) {
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
